! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General ! Public License as1 published by the Free Software Foundation; either

! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without evien the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module apetool_io

  use param
  use numkind  
  use erolib_bintable_mod, only : TCOL_READ_BIT
  use erolib_eventfile_mod
  use slib_constants
  use detect_io, only : ermldet_param_type, max_ima, & 
       i4null, r4null, r8null, cnull


  implicit none

  logical, private, parameter :: casesens = .false.
  integer, private, parameter :: ftread = 0, ftwrite = 1  ! ??
!attitude file related stuff
  type :: PSFSIZE     
     ! this are the EEFs at which radii will be estimated
     real(kind=real32), dimension (:), allocatable :: EEF
     ! this is the table with dimensions X, Y, RADIUS_4_EEF
     real(kind=real32), dimension (:,:,:), allocatable :: PSFMAP    
     ! the number of X and Y size is actuall NX+1, NY+1
     integer :: NX, NY
     ! this is the step in pixels 
     real(kind=real32) :: DX, DY
  end type PSFSIZE

  type :: APEPAR     
     type(ermldet_param_type) EPAR
     integer(int32), dimension(max_ima) :: energy_index
     character(len=256) :: apelistout, apelist
     character(len=256), dimension(max_ima) :: psfmapsets
     character(len=256), dimension(max_ima) :: sensemapsets
     integer(int32) :: n_energy_index
     integer(int32) :: n_psfmapsets
     integer(int32) :: n_sensemapsets
     real(kind=real32) ::  EEF_EXTRACT=0.7
     real(kind=real32) ::  POISSON_THRESH=4e-6
     integer(kind=int32) :: BINNING=11
     logical :: APEX_FLAG=.False.
     logical :: APESTACK_FLAG=.False.
     logical :: PSFMAP_FLAG=.False.
     logical :: SHAPEPSF_FLAG=.False.
     logical :: APESENSE_FLAG=.False.
     logical :: REMSRC_FLAG=.False.     
  end type APEPAR


!-------------------------------------------------------------------------------



contains



!----------------------------------------------------------------------------
  subroutine ape_get_param(n_imagesets, n_expimagesets, n_detmasksets, &
       n_bkgimagesets, n_srcimagesets,  n_psfmapsets, n_apesenseimagesets, &
       n_emin, n_emax, n_energy_index,& 
       mlsrclist, apesrclist, apelistout, &
       imagef, expimagf, detimagf, bkgimagef, srcimagef,  &
       mappsff, apesenseimagef, &
       emin, emax, energy_index, eef_extract, POISSON_THRESH, &
       CUTRAD, BINNING, APESENSE_FLAG, &
       APEX_FLAG, APESTACK_FLAG, PSFMAP_FLAG, shapepsf_flag, ierr)
!----------------------------------------------------------------------------
    use detect_io

    implicit none

    integer(kind=int32), intent(out) :: n_imagesets, n_psfmapsets, n_expimagesets, &
         n_detmasksets, n_bkgimagesets, n_srcimagesets, n_apesenseimagesets, &
         n_emin, n_emax, n_energy_index
    character(len=*), intent(out) :: mlsrclist
    character(len=*), intent(out) :: apesrclist
    character(len=*), intent(out) :: apelistout
    character(len=*), dimension(:), intent(out) :: imagef, mappsff, expimagf, &
         bkgimagef, srcimagef, detimagf, apesenseimagef
    real(kind=real32), dimension(:), intent(out) :: emin, emax
    integer(int32), dimension(:), intent(out) :: energy_index
    integer(int32), intent(out) :: BINNING
    real(kind=real32), intent(out) ::  eef_extract, POISSON_THRESH, CUTRAD
    logical, intent(out) :: APEX_FLAG, APESTACK_FLAG, PSFMAP_FLAG, &
         shapepsf_flag, APESENSE_FLAG
    integer(kind=int32), intent(out) :: ierr

    integer(kind=int32)   status
    character(len=*),parameter         :: rtname='ape_get_param'
    character(len=80)                  :: errstr

    !max_ima: is defined in globals of detect_io

    ierr = 0

    
    call rgetc('MLLIST', mlsrclist, status)
    if (status /= 0) ierr = max(ierr,status)   

    !write(*,*) mlsrc1list
    
    call rgetc('APELIST', apesrclist, status)
    if (status /= 0) ierr = max(ierr,status)   
    
    call rgetc('APELISTOUT', apelistout, status)
    if (status /= 0) ierr = max(ierr,status)   

    !write(*,*) mlsrclist
    !write(*,*) apesrclist
    !write(*,*) apelistout
    
    call rgetca('IMAGES', max_ima, imagef, n_imagesets, status)
    if (status /= 0) ierr = max(ierr,status)   
    
    !write(*,*) imagef


    call rgetca('PSFMAPS',max_ima , mappsff, n_psfmapsets, status)
    !write(*,*) mappsff
    if (status /= 0) ierr = max(ierr,status)   
    if (n_psfmapsets /= n_imagesets)then 
         ierr=3
       errstr = 'Number of input image and psf maps differ'
       call WRERR(rtname,ierr,errstr)   
    endif

    call rgetca('EXPIMAGES',max_ima , expimagf, n_expimagesets, status)
    if (status /= 0) ierr = max(ierr,status)   
    if (n_expimagesets /= n_imagesets)then 
         ierr=3
       errstr = 'Number of input image and exposure maps differ'
       call WRERR(rtname,ierr,errstr)   
    endif

    call rgetca('DETMASKS',max_ima, detimagf, n_detmasksets, status)
    if (status /= 0) ierr = max(ierr,status)           
!    if (n_expimagesets /= n_imagesets)then  
!         ierr=3
!       errstr = 'Number of input image and detection images differ'
!       call WRERR(rtname,ierr,errstr)   
!    endif

    call rgetca('BKGIMAGES',max_ima , bkgimagef, n_bkgimagesets, status)
    if (status /= 0) ierr = max(ierr,status)   
    if (n_bkgimagesets /= n_imagesets)then  
         ierr=3
       errstr = 'Number of input image and bakcground images differ'
       call WRERR(rtname,ierr,errstr)   
    endif

    call rgetca('SRCIMAGES',max_ima , srcimagef, n_srcimagesets, status)
    if (status /= 0) ierr = max(ierr,status)   
    if (n_srcimagesets /= n_imagesets)then  
         ierr=3
       errstr = 'Number of input image and source images differ'
       call WRERR(rtname,ierr,errstr)   
    endif

    call rgetca('APESENSEIMAGES',max_ima , apesenseimagef, n_apesenseimagesets, status)
    if (status /= 0) ierr = max(ierr,status)   
    if (n_apesenseimagesets /= n_imagesets)then  
         ierr=3
       errstr = 'Number of input image and source images differ'
       call WRERR(rtname,ierr,errstr)   
    endif
    !write(*,*) apesenseimagef, n_apesenseimagesets
    
    call rgetfa("EMIN", max_ima, emin, n_emin, ierr)
    if (status /= 0) ierr = max(ierr,status)       
    if (n_emin /= n_imagesets)then  
         ierr=3
       errstr = 'Number of input image and EMIN differ'
       call WRERR(rtname,ierr,errstr)   
    endif
   
    call rgetfa("EMAX", max_ima, emax, n_emax, ierr)
    if (status /= 0) ierr = max(ierr,status)       
    if (n_emax /= n_imagesets)then  
         ierr=3
       errstr = 'Number of input image and EMAX differ'
       call WRERR(rtname,ierr,errstr)   
    endif   
    

    ! this provides a map between the index of the input images
    ! and the band index of ermldet src lists
    call rgetja("EINDEX", max_ima, energy_index, n_energy_index, ierr)
    !call rgetfa("EINDEX", max_ima, emax, n_energy_index, ierr)
    if (status /= 0) ierr = max(ierr,status)       
    if (n_energy_index.gt.n_imagesets)then  
         ierr=3
       errstr = 'Number of input images smaller than number of EINDEX indeces'
       call WRERR(rtname,ierr,errstr)   
    endif   



    call rgetf('EEFEXTRACT', eef_extract, status)
    if (status /= 0) ierr = max(ierr,status)   

    call rgetf('PTHRESH', POISSON_THRESH, status)
    if (status /= 0) ierr = max(ierr,status)   

    call rgetf('CUTRAD',cutrad, status)
    if (status /= 0) ierr = max(ierr,status)   

    call rgetj('PSFMAPSAMPLING', BINNING, status)
    if (status /= 0) ierr = max(ierr,status)   
    if( MOD (int(BINNING),2) .eq. 0)then
       ierr=3
       errstr = 'PSMAP binning paramter should be an odd number'
       call WRERR(rtname,ierr,errstr)   
    endif
    
    !write(*,*)  eef_extract

    call rgetl('APEXFLAG',APEX_FLAG,status)
    if (status /= 0) ierr = max(ierr,status)   

    call rgetl('STACKFLAG',APESTACK_FLAG,status)
    if (status /= 0) ierr = max(ierr,status)   
    
    call rgetl('PSFMAPFLAG',PSFMAP_FLAG,status)
    if (status /= 0) ierr = max(ierr,status)   

    call rgetl('APESENSEFLAG', APESENSE_FLAG,status)
    if (status /= 0) ierr = max(ierr,status)   
    
    call rgetl('SHAPEPSF', shapepsf_flag,status)
    if (status /= 0) ierr = max(ierr,status)   

    ! write(*,*) apex_flag,  psfmap_flag, apeback_flag

  end subroutine ape_get_param
  





!----------------------------------------------------------------------------
  subroutine ape_pass_param(par, n_imagesets, n_expimagesets, &
       n_detmasksets, n_bkgimagesets, n_srcimagesets, n_psfmapsets, &
       n_sensemapsets,n_emin, n_emax, n_energy_index,& 
       mlsrclist, apesrclist, apelistout, &
       imagef, expimagf, detimagf, bkgimagef, srcimagef, &
       psfmapf, sensemapf, &
       emin, emax, energy_index,eef_extract, POISSON_THRESH, &
       CUTRAD, BINNING, APESENSE_FLAG, &
       APEX_FLAG, APESTACK_FLAG, PSFMAP_FLAG,  &
       shapepsf_flag,REMSRC_FLAG, ierr)
! this routines constructs an ermldetect param type and updates values
!----------------------------------------------------------------------------
    !use detect_io
    
    implicit none

    type(APEPAR), intent(out) :: par   
    !type(ermldet_param_type) , intent(out)  :: par
    integer(kind=int32)      , intent(out)  :: ierr
 
    integer(kind=int32), intent(in) :: n_imagesets, n_expimagesets, &
       n_detmasksets, n_bkgimagesets, n_srcimagesets, n_sensemapsets,&
       n_psfmapsets, n_emin, n_emax, n_energy_index
    character(len=*), intent(in) :: mlsrclist, apesrclist, apelistout
    character(len=*), dimension(:), intent(in) :: imagef, expimagf, &
         detimagf, bkgimagef, srcimagef, psfmapf, &
         sensemapf
    real(kind=real32), dimension(:), intent(in) :: emin, emax
    integer(int32), dimension(:), intent(in) :: energy_index
    real(kind=real32),intent(in) ::  EEF_EXTRACT, POISSON_THRESH, CUTRAD
     integer(int32), intent(in) :: BINNING
    logical, intent(in) :: shapepsf_flag, APESENSE_FLAG, APEX_FLAG, &
         APESTACK_FLAG, PSFMAP_FLAG,  REMSRC_FLAG
    character(len=*),parameter         :: rtname='APE_PASS_PARAM'
    
    par%psfmapsets=psfmapf
    par%n_psfmapsets=n_psfmapsets
    par%sensemapsets=sensemapf
    par%n_sensemapsets=n_sensemapsets
     
    par%apelistout = apelistout
    par%apelist = apesrclist

    par%energy_index = energy_index 
    par%n_energy_index = n_energy_index 
    par%EEF_EXTRACT = EEF_EXTRACT
    par%POISSON_THRESH = POISSON_THRESH
    par%BINNING = BINNING
    par%APEX_FLAG = APEX_FLAG
    par%PSFMAP_FLAG = PSFMAP_FLAG
    par%APESENSE_FLAG = APESENSE_FLAG
    par%APESTACK_FLAG = APESTACK_FLAG
    par%REMSRC_FLAG=REMSRC_FLAG
    
    par%epar%boxlistset = mlsrclist
    par%epar%mllistset = mlsrclist
    par%epar%imagesets = imagef
    par%epar%n_imagesets = n_imagesets

    par%epar%expimagesets = expimagf
    par%epar%n_expimagesets = n_expimagesets

    par%epar%bkgimagesets = bkgimagef
    par%epar%n_bkgimagesets = n_bkgimagesets

    par%epar%srcimagesets = srcimagef
    par%epar%n_srcimagesets = n_srcimagesets

    par%epar%detmasksets = detimagf
    par%epar%n_detmasksets = n_detmasksets

    par%epar%emin = emin
    par%epar%n_emin = n_emin

    par%epar%emax = emax
    par%epar%n_emax =n_emax
    
    par%epar%cutrad = CUTRAD
    par%epar%multrad = 20.0
    par%epar%nmaxfit = 3
    par%epar%nmulsou = 2


    !par%ecf = 0.0
    !par%n_ecf = n_imagesets
    !par%likemi = 8.0    
    !par%extlikemin =6.0
    !par%extmin = 6.0
    !par%extmax = 6.0
    !par%extentmodel = "beta"
    !par%nmaxfit = 3.0
    !par%nmulsou =1.0

    par%epar%detmask_flag  = .true.
    par%epar%shapelet_flag = shapepsf_flag
    par%epar%expima_flag   = .true.
    par%epar%compress_flag = .false.

    if (ierr /= 0)  call wrerr(rtname,ierr,'Error getting parameters')
 
  end subroutine ape_pass_param
  !----------------------------------------------------------------------------




  !----------------------------------------------------------------------------
  !PSFSIZE Constructor
  subroutine ape_psfsize_init(MyPSFSIZE, NX, NY, DX, DY, EEF)
  !----------------------------------------------------------------------------

    TYPE(PSFSIZE), intent(out) :: MyPSFSIZE
    integer, intent(in) :: NX, NY
    real(kind=real32), DIMENSION(:), intent(in) :: EEF
    real(kind=real32), intent(in) :: DX, DY

    if (allocated(MyPSFSIZE%PSFMAP)) deallocate(MyPSFSIZE%PSFMAP)
    allocate(MyPSFSIZE%PSFMAP(NX+1, NY+1, SIZE(EEF)))  
    if (allocated(MyPSFSIZE%EEF)) deallocate(MyPSFSIZE%EEF)
    allocate(MyPSFSIZE%EEF(SIZE(EEF)))  
    MyPSFSIZE%EEF = EEF

    MyPSFSIZE%NX = NX    
    MyPSFSIZE%NY = NY
    MyPSFSIZE%DX = DX
    MyPSFSIZE%DY = DY  
    
  end subroutine ape_psfsize_init
 !----------------------------------------------------------------------------
  

  !----------------------------------------------------------------------------
  !PSFSIZE Destructor
  subroutine ape_psfsize_kill(MyPSFSIZE)
  !----------------------------------------------------------------------------

    TYPE(PSFSIZE), intent(out) :: MyPSFSIZE
    
    if (allocated(MyPSFSIZE%PSFMAP)) deallocate(MyPSFSIZE%PSFMAP)
    if (allocated(MyPSFSIZE%EEF)) deallocate(MyPSFSIZE%EEF)
    
  end subroutine ape_psfsize_kill
 !----------------------------------------------------------------------------
  
  !----------------------------------------------------------------------------
  !PSFSIZE File Reader
  subroutine ape_psfsize_read(filename, MyPSFSIZE)
  !----------------------------------------------------------------------------

    TYPE(PSFSIZE), intent(out) :: MyPSFSIZE   
    character(*), intent(in) :: filename
    
    
    integer status,unit,blocksize,readwrite,naxes(3),maxdim,anyf,group,ierr,hdutype
    integer irow, felems, nelems, colnum
    real(kind=real32) nullval, DX, DY
    integer naxis1, naxis2, naxis3, dim1, dim2
    character comment*10
   
    ! this are the EEFs at which radii will be estimated
    real(kind=real32), dimension (:), allocatable :: EEF
    ! this is the table with dimensions X, Y, RADIUS_4_EEF
    real(kind=real32), dimension (:,:,:), allocatable :: PSFMAP    
    
    status=0
    !  Get an unused Logical Unit Number to use to open the FITS file.
    call ftgiou(unit,status)
    !write(*,*) status
    !  open the FITS file, with read-only access.  The returned BLOCKSIZE
    !  parameter is obsolete and should be ignored. 
    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)
    !write(*,*) status
    ! getx naxes from image
    maxdim=3
    call FTGISZ(unit, maxdim, naxes,status)

    naxis1=naxes(1)
    naxis2=naxes(2)
    naxis3=naxes(3)
    dim1 = naxes(1)
    dim2 = naxes(2)
    !write(*,*) "1:", naxis1, naxis2, naxis3
    allocate(PSFMAP(naxis1, naxis2, naxis3))  
    allocate(EEF(naxis3))  
    !write(*,*) "2:", naxis1, naxis2, naxis3
    call FTGKYE(unit,"APEDX",DX,comment, status)
    call FTGKYE(unit,"APEDY",DY,comment, status)
    !FTGKY[EDJKLS](unit,keyword, > keyval,comment,status)

    group=0
    nullval=0.0
    call FTG3DE(unit,group,nullval,dim1,dim2,naxis1,naxis2,naxis3, &
         psfmap,anyf,status)

    !move to next extension, should be EEF
    call ftmrhd(unit,1,hdutype,status)
    call fitserr("ftmrhd", status, ierr)    
    nelems=1
    felems=1
    colnum=1    
    do irow=1,naxis3       
       call ftgcve(unit,colnum,irow,felems,nelems,nullval,EEF(IROW), &
            anyf,status)
    end do
    
    !write(*,*) EEF

   call ape_psfsize_init(MyPSFSIZE, naxis1 -1, naxis2 -1, DX, DY, EEF)
   MyPSFSIZE%psfmap = psfmap
   
   
   deallocate(EEF, PSFMAP)
   call ftclos(unit, status)
   call fitserr("ftclos", status, ierr)   
   
   end subroutine ape_psfsize_read
 !----------------------------------------------------------------------------
  



  !----------------------------------------------------------------------------
  subroutine ape_prep_psfmap(imafile, shapelet_flag, filename, MyPSFSIZE) !, NX, NY, flagsel, shapehdr, stdhdr, refhdr, wcshdr)      
  !----------------------------------------------------------------------------
    use slib_constants
    use slib_data_src
    use slib_data_psf
    use slib_calc_space
    use expmap_legacy, only: get_vignetting, gti_t, merge_gti, opn_events, &
         get_gti, get_attitude_old, get_evt_keywords
    !use srctool_generic_avign
    !use slib_cimage ! ! this is for testing only / remove after use
    !use slib_image  ! ! this is for testing only / remove after use
    use radec2pix_mod
    use fitsimage !, only: fread_shapehdr, fread_stdhdr, fread_refhdr, fread_wcshdr, &
!         shapehdr_type, stdhdr_type, &
!         wcs_imahdr_type, wcs_refhdr_type 

    implicit none
    logical, intent(in)  :: shapelet_flag
    TYPE(PSFSIZE), intent(out) :: MyPSFSIZE
    character(*), intent(in) :: filename
    

    character(len=*) :: imafile
    integer :: NX, NY
    type(shapehdr_type) :: shapehdr
    type(stdhdr_type) :: stdhdr
    type(wcs_refhdr_type) :: refhdr
    type(wcs_imahdr_type) :: wcshdr
    integer :: flagsel  

    ! these related to number of cameras
    integer :: ncamarr
    integer, dimension(7) :: camarr, camlun
    character(len=30), dimension(7) :: instarr

    !real(kind=real32), dimension(:), allocatable :: XIMA, YIMA, RAIMA, DECIMA
    real(kind=real32) DX, DY, X, Y    
    integer NSRC, XSIZE, YSIZE, IX, IY
    integer :: unit, ierr = 0, status = 0, ftread = 0
    integer colnum,frow,felem, varidata

    character(len=30) :: comment
    integer, dimension(3) :: naxes
    character(len=10) :: ttype, tform, tunit, extname

    character(len=*),parameter         :: rtname='APE_PREP_PSFMAP'

    real(kind=real64) xrval,yrval,xrpix,yrpix,xinc,yinc,rot
    character(len=10) :: coordtype

    integer ic, clun
    character(len=30) inst, filter
    character(len=19) date_obs, date_end


    real(real64) xraw_ax, yraw_ax  
    real(real64) xraw_scal, yraw_scal 

    integer :: ngti !, egti, ogti, igti        ! number of gti pairs
    character(len=7) :: gtitype="GTI"
    type(gti_t), dimension(2) :: mgti
    type(gti_t) :: sgti
    real(real64), dimension(:), pointer :: tstart, tstop  ! gti arrays


    ! GENERAL ITERATORS
    integer I, J, K
    ! GENERAL ITERATORS

    ! THESE PARAMS HAVE TO DO WITH COORD TRANSFORMATIONS
    real(real64) :: sinr, cosr, DR2, DR2CUT
    real, parameter :: dxcen = 192.5, &      ! x/yraw center of detector map
         dycen = 192.5
    integer ixraw, iyraw
    real(real64), parameter :: pi=dacos(-1d0), deg2rad = pi/180d0
    real(real64) :: xrscal, yrscal                   ! ratio of sky to detector pixel
                                         !  plate scale
    ! THESE PARAMS HAVE TO DO WITH COORD TRANSFORMATIONS

    ! THESE PARAMS ARE RELATED  TO ATTTITUDE
    real(real64), dimension(:), pointer :: ra, dec, roll ! camera attitude
    real(real64), dimension(:), pointer :: atime ! times from attitude extension
    integer :: natt, iatt                    ! number of attitudes, attitude index
    real(real64), dimension(:), pointer :: xatt, yatt    ! attitude in image pixels
    logical, dimension(:), pointer :: flag_att    ! attitude flag that signifies if att within GTI
    real(real64) :: ra0, dec0, x0, y0, &     ! WCS for sky images
         xscal, yscal             !
    integer :: ix_size, iy_size              ! size of sky images
    ! THESE PARAMS ARE RELATED  TO ATTTITUDE

    ! these are vignetting related params
    ! i use the vignting calling function from expmap. This assumes that the vignting function
    ! is 2d array with the first array energy-band (input in expmap) and the value the 
    ! different off-axis angle. At this stage the apetool takes a single energy as input 
    ! and therefore the expmap routines have to be called with nb=1
    integer nb ! number of bands 
    real, dimension(:,:), pointer :: vig     ! array of vignetting values for different
    !   energies and off-axis angles
    real, dimension(:), pointer :: theta     ! array of off-axis angle grid points where
    real offax, out
    !   vignetting is defined
    ! these are vignetting related params


    ! THESE PARAMS ARE FOR SHAPELIB
    TYPE(STACKPSF) :: myStackPSF 
    TYPE(MODELPSF), dimension (MAX_NUM_TELESCOPES) :: myModelPSF
    TYPE(SRC) :: mySRC
    DOUBLE PRECISION IMG_PIXEL_MM
    DOUBLE PRECISION ENERGY
    INTEGER MAX_PSF_IMG_SIZE 
    double precision, ALLOCATABLE ::  COEFF0_TEMP(:),COEFF1_TEMP(:),COEFF2_TEMP(:) 
    !complex, ALLOCATABLE :: CCOEFF1_TEMP(:),  CCOEFF2_TEMP(:) ! these are for testing onlt  
    !integer NP, ISRC, IE ! these are for testing only
    ! THESE PARAMS ARE FOR SHAPELIB

    ! these are for testing, i.e. produce image and then manually measure EEF
    !character(len=256) :: calfile
    TYPE(CPSF) myCPSF
    ! these are for testing
    
    ! THESE PARAMS ARE HARD-CORED EEF
    INTEGER, parameter :: MEEF=12
    real(kind=real32), DIMENSION(MEEF) :: EEF 
    real(kind=real64), DIMENSION(MEEF) :: RADIUS !, FLUXR ! only for testing
    ! THESE PARAMS ARE HARD-CORED EEF
    EEF = (/0.4,0.45,0.5,0.55,0.6,0.65,0.7,0.75,0.8,0.85,0.9,0.95/)



    ! these are not used by the task
    ! not output image is actually made
    NX=20
    NY=20
   
    ! READ THE SHAPELET PSFS
    if( shapelet_flag ) then 
       IMG_PIXEL_MM = 0.075 !* xscal/xraw_scal 
       MAX_PSF_IMG_SIZE  = 100
       ! I initiate myModelPSF and read from CALDB
       call slib_read_alltel_model_psf(myModelPSF, myCPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)       
    end if
    


    ! get units and camera numbers from event file
    ! this is taken from the exposure map routines in expmap_io_mod
    call opn_events(1, imafile, ncamarr, camarr, camlun, instarr, ierr) 
    if (ierr > 2) call wrstop(rtname, &
         "Coulnd't obtain camera numbers")
    !write(*,*) imafile
    !write(*,*)  "ncamarr",ncamarr
    !write(*,*)  "camarr",camarr
    !write(*,*) "camlun",camlun
    !write(*,*) "instarr",instarr
    !stop


    ! get keywords from template image ===============================
    call ftgiou (unit, status)
    call fitserr("ftgiou", status, ierr)
    call ftiopn(unit,imafile, ftread, status) 
    call fitserr("ftiopn", status, ierr)
    if (status > 0) call wrstop(rtname, &
         "Couldn't open template image")
    ! the calls (eroio - fitsimage) below essentially read the
    ! image extension of the event file 
    status = 0
    ! get image shape headers
    call fread_shapehdr(unit, shapehdr, ierr) 
    ! get eSASS standard headers
    call fread_stdhdr(unit, stdhdr, ierr)
    ! get wcs keywords
    call fread_refhdr(unit, refhdr, ierr)
    call fread_wcshdr(unit, wcshdr, ierr)
    ! get flagsel keyword
    call ftgkyj(unit, 'FLAGSEL', flagsel, comment, status)
    if (status /= 0) then
       flagsel=0
       status=0
    end if

    !call FTGICS(unit, xrval,yrval,xrpix,yrpix,xinc,yinc,rot,coordtype,status)    
    !write(*,*) xrval,yrval,xrpix,yrpix,xinc,yinc,rot,coordtype,status
    !write(*,*) wcshdr%crval1, wcshdr%crval2, wcshdr%crpix1,wcshdr%crpix2, wcshdr%cdelt1, wcshdr%cdelt2, wcshdr%ctype1, wcshdr%ctype2
    !write(*,*) wcshdr%crota2

    I = SCAN(wcshdr%ctype1, '-', .TRUE.) 
    coordtype = wcshdr%ctype1(I:LEN(wcshdr%ctype1))
    xrval =  wcshdr%crval1
    yrval =  wcshdr%crval2
    xrpix =  wcshdr%crpix1
    yrpix = wcshdr%crpix2
    xinc = wcshdr%cdelt1
    yinc =  wcshdr%cdelt2
    rot = 0.D0
     
    ra0  = wcshdr%crval1
    dec0 = wcshdr%crval2
    x0   = wcshdr%crpix1
    y0   = wcshdr%crpix2
    xscal= wcshdr%cdelt1
    yscal= wcshdr%cdelt2
    ix_size = shapehdr%naxis1
    iy_size = shapehdr%naxis2

    DR2CUT = 0.51 * 0.51 / xscal / xscal

    !naxes(1) = ix_size
    !naxes(2) = iy_size

    !write(*,*) xrval,yrval,xrpix,yrpix,xinc,yinc,rot,coordtype,status
    !stop

    call ftclos(unit, status)
    call fitserr("ftclos", status, ierr)
    call FTFIOU(unit,  status)
    call fitserr("ftfiou", status, ierr)
    if (status > 0) call wrstop(rtname, "Couldn't read template header")

    NSRC = ( NX + 1 ) * ( NY + 1 )

    call slib_init_source_with_num(mySRC, NSRC)
    
    call slib_init_SPSF(myStackPSF, myModelPSF(1), NSRC)
    !initialise arrays of stacked PSF
    myStackPSF%C0=0.0
    myStackPSF%C1=0.0
    myStackPSF%C2=0.0
    allocate ( COEFF0_TEMP( SIZE(myModelPSF(1)%NX0) ) )
    allocate ( COEFF1_TEMP( SIZE(myModelPSF(1)%NX1) ) )
    allocate ( COEFF2_TEMP( SIZE(myModelPSF(1)%NX2) ) )
    !allocate ( CCOEFF1_TEMP( SIZE( myModelPSF(1)%NX1)  ), CCOEFF2_TEMP( SIZE( myModelPSF(1)%NX2)  ) )  ! this is for testing only
    COEFF0_TEMP=0.0
    COEFF1_TEMP=0.0
    COEFF2_TEMP=0.0
    ENERGY = DBLE(stdhdr%eminx(1) +  0.5 * ( stdhdr%emaxx(1) -stdhdr%eminx(1) ) ) * 1.D3

    !write(*,*) "ENERGY:",ENERGY, stdhdr%eminx(1),  stdhdr%emaxx(1)
    !write(*,*) SIZE(COEFF1_TEMP),SIZE(COEFF2_TEMP), SHAPE(myStackPSF%C1),SHAPE(myStackPSF%C2)     
    !stop
    
    XSIZE = shapehdr%naxis1
    YSIZE = shapehdr%naxis2
    
    DX = FLOAT( XSIZE / NX )
    DY = FLOAT( YSIZE / NY )

    I = 1
    ! the iteration limits are  chosen to ensure that the first 
    ! position is 0 and the last is  shapehdr%naxis1. This is 
    ! to fascilitate any interpolation later on. 
    do IX = 1, NX+1

       X = (IX - 1) * DX !+ DX/2  
        
       do IY = 1, NY+1
          
          Y = (IY - 1) * DY !+ DY/2  

          !call FTWLDP(DBLE(X), DBLE(Y), xrval,yrval,xrpix,yrpix,xinc,yinc,rot, coordtype, xpos,ypos,status)
          
          ! I use image X, Y because later on I convert attitude ra, dec to image X, Y
          mySRC%RA(I)  = X !XPOS Normally this RA, DEC but here I simply assigme X, Y
          mySRC%DEC(I) = Y !YPOS

          !write(*,*) IX, IY, X, Y,  mySRC%RA(I), mySRC%DEC(I), wcshdr%crval1, wcshdr%crpix1, wcshdr%cdelt1
          !write(*,*) IX, IY, X, Y,  mySRC%RA(I), mySRC%DEC(I), xpos, ypos

          ! Actually XD and YD should not be X and Y 
          ! but use the allocated space here to store X and Y.
          !mySRC%XD(I) = X
          !mySRC%YD(I) = Y

          I=I+1
       enddo

    enddo
    
    ! INITIALISE THE PSFMAP ARRAY
    call ape_psfsize_init(MyPSFSIZE, NX, NY, DX, DY, EEF)

    ! loop over cameras ====================================
    cam: do I = 1, ncamarr

       ic = camarr(i)
       clun = camlun(i)
       inst = instarr(i)


       !write(*,*) I, IC, CLUN, INST, SIZE(camarr), SIZE(camlun), SIZE(instarr)
       ! AGE 11.05.2018 this has dissappreard from the expmap
       ! and replaced by the following lines
       !call get_eventsdate(clun, date_obs, date_end)
       call wrchatty ("main",  "Calling get_evt_keywords...")
       call get_evt_keywords(clun, date_obs, date_end, flagsel, filter)
       call wrchatty ("main",  "Calling get_detmaps...")

 

      ! all I need realy is  xraw_ax, yraw_ax, xraw_scal, yraw_scal
      ! possibly modify get_detmaps or just call for now
      ! it is simply. Just GEOM caldb
      call  ape_get_geom(ic, inst, date_obs, date_end,  &
           xraw_ax, yraw_ax, xraw_scal, yraw_scal, ierr)
      if (ierr == 3) call wrstop (rtname, "Obtaining detector maps failed")

      

      ! this is reading event file - maybe I should directly from exposure file
      ! and maybe get away from the need to define gtitype? Ask Hermann
      call get_gti(ic, clun, gtitype, ngti, tstart, tstop, ierr)
      if (ierr == 3) call wrstop(rtname, "Obtaining gti from input event file failed")
      if (gtitype=="FLAREGTI") then
        mgti(1)%ngti=ngti
        mgti(1)%start=tstart
        mgti(1)%stop=tstop
        call get_gti(ic, clun, "QUALGTI", ngti, tstart, tstop, ierr)
        if (ierr == 3) call wrstop(rtname, "Obtaining gti from input event file failed")
        mgti(2)%ngti=ngti
        mgti(2)%start=tstart
        mgti(2)%stop=tstop
        call merge_gti(mgti, 2, sgti, -1, ierr)
        ngti=sgti%ngti
        tstart=sgti%start
        tstop=sgti%stop
      end if    

      !write(*,*) ngti, tstart, tstop

      ! reading attitude file and convering 
      call get_attitude_old(ic, clun, natt, atime, ra, dec, roll, ierr)
      if (ierr == 3) call wrstop(rtname, "Obtaining attitude failed")
      allocate (xatt(natt))
      allocate (yatt(natt))
      allocate (flag_att(natt))

      ! read vignting function
      nb=1 ! this is becuase in expmap the number of bands is an input parameter
           ! At this stage this is not the case for apetool
      call get_vignetting(ic, inst, nb, &
           REAL(stdhdr%eminx), REAL(stdhdr%emaxx), &
           date_obs, date_end, &
           theta, vig, ierr)
      if (ierr == 3) call wrstop (rtname, "Obtaining vignetting failed")

      !write(*,*) theta
      !write(*,*) vig
      !stop
 
      ! loop to indentify att times that lie outside GTIs
      ! the flag_att(K) is false for attitude times outside GTIs
      do K=1, SIZE(xatt)
         flag_att(K) = .false.
         J=1
          do while(J <= ngti .and.  .not. flag_att(K))
             if( atime(K).le.tstop(J) .and. atime(K).ge.tstart(J))&
                  flag_att(K) = .true.
             !write(*,*) J, tstart(J), ngti
             J=J+1
          enddo
          !write(*,*) atime(K), flag_att(K),  atime(K)- tstart(J-1), tstop(J-1)- atime(K)
       enddo
 
       xrscal = xscal/xraw_scal  ! xscal is the pixel scale from the WCS xscal= wcshdr%cdelt1
       yrscal = yscal/yraw_scal  ! xraw_scal is the detector pixel scale. It is produdecd
       
       ! setting the right pixel size for the EEF calculation
       !write(*,*) IMG_PIXEL_MM
       IMG_PIXEL_MM = RAW_PIX_MICRON * ABS(xscal/xraw_scal)
       myModelPSF(:)%IMG_PIXEL_MM = IMG_PIXEL_MM

       ! this is for testing
       ! the call writes arrays with the right size, which are then used
       ! to construct PSF. 
       !calfile="/local/age/MyESASS/caldb/data/erosita/fm1/bcf/fm1_slet_psf_181217v03.fits"
       !call slib_init_CPSF(calfile, myCPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
       ! this is for testing
       
       !write(*,*) "HERE",IMG_PIXEL_MM, RAW_PIX_MICRON , ABS(xscal/xraw_scal), xscal,xraw_scal
      ! stop
      ! ESTIMATE XATT,YATT FROM RA, DEC AND 
      ! XATT, YATT extended well beyond image FOV. Why?
      call radec2pix (ra, dec, natt, ra0, dec0, x0, y0, xscal, yscal, xatt, yatt) 

      ipsf: do J=1, SIZE(mySRC%RA)

         
         !write(*,*) "SOURCE:", J, mySRC%RA(J), mySRC%DEC(J), ENERGY

         !L=1 
          
          iatt: do K=1, SIZE(xatt)
             
             DR2 = (mySRC%RA(J) - xatt(K))**2 + (mySRC%DEC(J) - yatt(K))**2 
             if(DR2<=DR2CUT .and. flag_att(K))then

                !write(*,*)  mySRC%RA(J), xatt(K), mySRC%DEC(J), yatt(K)
                
                ! roll detector map and bin in image pixels  
                sinr = dsin(roll(K)*deg2rad)
                cosr = dcos(roll(K)*deg2rad)
                x = REAL(mySRC%RA(J) - xatt(K)) 
                y = REAL(mySRC%DEC(J)-  yatt(K))
                !write(*,*) mySRC%RA(J),mySRC%DEC(J),xatt(K),yatt(K), x, y
                ! these are not needed to be define here. Only
                ! once outside the loops
                !xrscal = xscal/xraw_scal  ! xscal is the pixel scale from the WCS xscal= wcshdr%cdelt1
                !yrscal = yscal/yraw_scal  ! xraw_scal is the detector pixel scale. It is produdecd
                
                !ixraw = nint((cosr*x-sinr*y)*xrscal + dxcen) ! AGE: dxcen/dycen are fixed to 192.5
                !iyraw = nint((sinr*x+cosr*y)*yrscal + dycen)

                ! this DX_RAW, DY_RAW, IE RELATIVE TO CENTER
                ixraw = nint((cosr*x-sinr*y)*xrscal)
                iyraw = nint((sinr*x+cosr*y)*yrscal)

                ! getting the vignetting at the off axis angle
                ! off-axis angle in arcmin

                offax = REAL(sqrt(((mySRC%RA(J) - xatt(K))*xscal)**2 + ((mySRC%DEC(J) - yatt(K))*yscal)**2) * 60.0)
                ! this is for testing
                !IXRAW=10.0
                !IYRAW=0.0
                call ape_calc_vign(out, offax, nb, vig, theta)

                !write(*,*) DBLE(IXRAW), DBLE(IYRAW), ENERGY, ROLL(K)

                
                call slib_get_xy(DBLE(IXRAW), DBLE(IYRAW), ENERGY, ROLL(K), myModelPSF(IC), COEFF0_TEMP, COEFF1_TEMP, COEFF2_TEMP)

                !write(*,*) "1,",theta
                !write(*,*) "2,",vig
                !write(*,*) "3,",out, offax

                !if(ENERGY>7000)then
                !   write(25,*) mySRC%RA(J),mySRC%DEC(J), offax, out, IXRAW, IYRAW
                !endif
                myStackPSF%C0(:, J) = myStackPSF%C0(:, J) + out * COEFF0_TEMP
                myStackPSF%C1(:, J) = myStackPSF%C1(:, J) + out * COEFF1_TEMP
                myStackPSF%C2(:, J) = myStackPSF%C2(:, J) + out * COEFF2_TEMP

                !write(*,*)ixraw, iyraw, x, y, xrscal, dxcen, dycen, roll(K)
                !write(*,*) K ,ra(K), dec(K), xatt(K), yatt(K), ROLL(K), mySRC%RA(J), mySRC%DEC(J), sqrt(DR2CUT)
                !write(10,*) "fk5;circle(", ra(K), ",", dec(K), ",5i)"
                
                !L=L+1

                

             endif

      !   write(*,*) J ,natt, SIZE(xatt), ra(J), dec(J), xatt(J), yatt(J), ROLL(J)
      !   write(10,*) "fk5;circle(", ra(J), ",", dec(J), ",5i)"
              
          enddo iatt

          
          !CCOEFF1_TEMP = CMPLX( myStackPSF%C1(:, J), 0)
          !CCOEFF2_TEMP = CMPLX( myStackPSF%C2(:, J), 0)    
          !NP=1
          !ISRC=J
          !IE=1
          !call slib_write2file(SIZE(myModelPSF(IC)%NX1), CCOEFF1_TEMP, &
          !     myModelPSF(IC)%NX1, myModelPSF(IC)%NY1, &
          !     real(myModelPSF(IC)%BETA1), &
          !     SIZE(myModelPSF(IC)%NX2), CCOEFF2_TEMP, myModelPSF(IC)%NX2, &
          !     myModelPSF(IC)%NY2, &
          !     real(myModelPSF(IC)%BETA2), &
          !     100, ISRC, IE, NP)
          !stop
          !WRITE(*,*) "BEFORE:",myStackPSF%C1(:, J)
          !WRITE(*,*) "BEFORE:",myStackPSF%C2(:, J)
 
          !call slib_fluxEEF(myStackPSF, myModelPSF(1), J, MEEF, EEF, RADIUS)
          
          !write(*,*) J, MEEF
          !write(*,*) "EEF:", EEF
          !write(*,*) "RADIUS:", RADIUS
          !stop

       enddo ipsf
       

      deallocate (xatt, yatt, flag_att)
      deallocate (atime, ra, dec, roll, theta, vig)
      nullify (tstart, tstop)
      !stop
   enddo cam
   
   !! this is for testing
   !! the call writes arrays with the right size, which are then used
   !! to construct PSF. 
   !calfile="/local/age/MyESASS/caldb/data/erosita/fm1/bcf/fm1_slet_psf_181217v03.fits"
   !call slib_init_CPSF(calfile, myCPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
   !! this is for testing

   
   ! write to psfmap table 
   I = 1
   do IX = 1, NX+1

      X = (IX - 1) * DX !+ DX/2  
      
      do IY = 1, NY+1
         
         Y = (IY - 1) * DY !+ DY/2  
         
         call slib_fluxEEF(myStackPSF, myModelPSF(1), I, MEEF, DBLE(EEF), RADIUS)

         ! testting only, produces images for manual inspection
         !if(RADIUS(2)>0)then 
         !   call  slib_run_write2arrays_fast(myStackPSF, myCPSF, I)
         !   write (calfile, "(I4.4, A4, I4.4, A4, I4.4 )") &
         !        int(ENERGY), "_IX_", IX, "_IY_",IY
         !   calfile="TEST_"//trim(calfile)//'.img'
         !   !write(*,*) calfile
         !   call slib_run_write2image(myCPSF%IMG, calfile)            
         !
         !   call slib_fluxR_image(myCPSF, SIZE(RADIUS), RADIUS, FLUXR)
         !
         !   write(*,*) "ENERGY, IX, IY, EEF, RADIUS", ENERGY, IX, IY, X, Y, calfile
         !   do J =1,MEEF               
         !      write(55,*) RADIUS(J), EEF(J), FLUXR(J), ENERGY, IX, IY,X, Y,calfile(1:index(calfile,' ')-1)
         !   enddo
         !   
         !endif
         

         MyPSFSIZE%PSFMAP (IX, IY, :) = REAL(RADIUS)

         I=I+1
      enddo
      
   enddo
   call slib_dealloc_CPSF(myCPSF)
   !stop
  ! write(*,*) filename

   ! write out PSF maps
   ! these are taken from expmap routines
   naxes(1) = NY+1
   naxes(2) = NY+1
   naxes(3) = SIZE(MyPSFSIZE%EEF)

   
   status = 0
   call ftgiou(unit, status)
   !write(*,*) "here1", unit, status
   call fitserr("ftgiou", status, ierr)
   call ftinit(unit, "!"//trim(filename), 0, status)
   call fitserr("ftinit", status, ierr)
   call ftphps(unit,-32,3,naxes,status)
   call fitserr("ftphps", status, ierr)
   call fwrite_stdhdr(unit, stdhdr, status)
   call ftukys(unit, "INSTRUME", "merged", " ", status)
   call fitserr("ftmkys", status, ierr)
   call ftukyj(unit, "NINST", ncamarr, " ", status)
   call fitserr("ftpkyj", status, ierr)
   do i = 1, ncamarr
      !write(inst,'(''INSTRUM'', i1)') camarr(i)
      call ftukys(unit, trim(inst), instarr(i), " ", status)
      call fitserr("ftpkys", status, ierr)
   end do
   call fwrite_refhdr(unit, refhdr, status)
   call fwrite_wcshdr(unit, wcshdr, status)
   call FTPKYE(unit,"APEDX",DX,3,"APETOOL PSFMAP DX", status)
   call FTPKYE(unit,"APEDY",DY,3,"APETOOL PSFMAP DY", status)

   !write(*,*) "here2", unit, status
   call ftp3de(unit, 0,  IX+1, IY+1, IX+1, IY+1, SIZE(MyPSFSIZE%EEF), MyPSFSIZE%PSFMAP, status)      
   call fitserr("ftp3de", status, ierr)
   !write(*,*) "here3", unit, status
   !Append/create a new empty HDU onto the end of the file and move to it.
   call ftcrhd(unit,status)

   !ttype, tform, and tunit parameters give the name, datatype, and physical units of each column, and extname giv
   ttype="EEF"
   tform="E"
   tunit="frac"
   EXTNAME="EEF"
   varidata =0
   !write(*,*) EEF, MyPSFSIZE%EEF
   call FTIBIN(unit,SIZE(MyPSFSIZE%EEF),1,ttype,tform,tunit,extname,varidata, status)
   !FTIBIN(unit,nrows,tfields,ttype,tform,tunit,extname, varidata, status)
   frow=1
   felem=1
   colnum=1
   call ftpcle(unit,colnum,frow,felem,SIZE(MyPSFSIZE%EEF),MyPSFSIZE%EEF,status)
   !write(*,*) "here", unit, status
   call ftclos(unit, status)
   call fitserr("ftclos", status, ierr)
   call ftfiou(unit, status)
   !call ftfiou(-1, status)
   
   ! write out PSF maps
   !write(*,*) "here", unit, status
   
   !stop
   do I = 1, ncamarr    
      call ftclos(camlun(i), status)
      call fitserr("ftclos", status, ierr)
      call FTFIOU(camlun(i),  status)
      call fitserr("ftfiou", status, ierr)
    enddo

   ! REMEMBER TO DEALLOCATE PSF AND SRC
   call slib_deallocate_src(mySRC)
   call slib_deallocate_myStackPSF(myStackPSF)
   deallocate(COEFF0_TEMP, COEFF1_TEMP, COEFF2_TEMP)
   !deallocate( CCOEFF1_TEMP, CCOEFF2_TEMP) ! this is for testing only
   !stop

  end subroutine ape_prep_psfmap




  subroutine ape_calc_vign(out, offax, nb, vig, theta)

    !use data_psf
    real, intent(in), dimension(:,:) :: vig     ! array of vignetting values for different
    real, intent(in), dimension(:) :: theta     ! array of off-axis angle grid points where
    real, intent(in) :: offax
    integer, intent(in) :: nb
    real out
    INTEGER I, ierr
    character(len=*),parameter         :: rtname='APE_CALC_VIGN'


   ! interpolate vignetting values (this assumes that theta 
    ! is sorted in ascending order); then multiply with detmaps  
    i=1
    do while (i<size(theta) .and. offax >= theta(i))         
       i=i+1
    end do

    !write(*,*) offax, i


    if (offax >= theta(i-1) .and. offax <= theta(i) .or. &
         i == size(theta) .and. offax > theta(size(theta))) then               

          out= vig(nb,i-1) + (vig(nb,i)-vig(nb,i-1))* &
               ((offax - theta(i-1))/(theta(i)-theta(i-1)))

       else
          ierr = 3
          call wrerr(rtname, ierr, "Interpolation of off-axis angle failed")
          return
       end if

     end subroutine ape_calc_vign
  

 

  !----------------------------------------------------------------------------
  ! given image convert ra, dec to x and y
  subroutine local_radec2xy(image, ra, dec, x, y)      
  !----------------------------------------------------------------------------
   use radec2pix_mod
   use fitsimage

    implicit none

    type(i4image_type), intent(in) :: image
    real(real64), intent(in) :: ra(:), dec(:)
    real(real64), intent(out) :: x(:), y(:)

    integer n
    real(real64) ra0, dec0, x0, y0, xd, yd
    n=SIZE(ra)

    ra0 = image%wcshdr%crval1
    dec0 = image%wcshdr%crval2
    x0 = image%wcshdr%crpix1
    y0 = image%wcshdr%crpix2
    xd = image%wcshdr%cdelt1
    yd = image%wcshdr%cdelt2

    call radec2pix (ra, dec, n, ra0, dec0, x0, y0, xd, yd, x, y)       

  end subroutine local_radec2xy

  !----------------------------------------------------------------------------
  subroutine extract_single_i4(image, ii, xsrc, ysrc, rcut2, asum)
  !----------------------------------------------------------------------------
    use fitsimage

    implicit none

    type(i4cube_type) :: image
    !type(i4image_type), intent(in) :: image
    real(real32), intent(in) :: xsrc, ysrc, rcut2
    integer, intent(in) :: ii
    integer, intent(out) :: asum
    integer i, j, imin, imax, jmin, jmax
    real(real32) dx, dy, dr2, rcut

    asum=0
    rcut = sqrt(rcut2)
    
    imin = int(xsrc - 1.5 * rcut)
    jmin = int(ysrc - 1.5 * rcut)
    if(imin.lt.1)imin=1
    if(jmin.lt.1)jmin=1
    imax = int(xsrc + 1.5 * rcut)
    jmax = int(ysrc + 1.5 * rcut)
    if(imax.gt.image%shapehdr(ii)%naxis1)imax=image%shapehdr(ii)%naxis1
    if(jmax.gt.image%shapehdr(ii)%naxis2)jmax=image%shapehdr(ii)%naxis2

    !  18/06/2018 AGE: no need to loop through whole image...
    ! x==i ; y==j
    !do i=1, image%shapehdr(ii)%naxis1
    !   do j = 1, image%shapehdr(ii)%naxis2
    do i=imin, imax
       do j = jmin, jmax
          
          dx = i - xsrc
          dy = j - ysrc
          dr2 = dx * dx + dy * dy

          !write(30,*) i,j,image%arr(i, j, ii), ii, "i4"
          
          if(dr2<rcut2)then
             asum = asum + image%arr(i, j, ii)
          endif
       enddo
    enddo
  end subroutine extract_single_i4


  !----------------------------------------------------------------------------
  subroutine extract_single_r4(image, ii, xsrc, ysrc, rcut2, asum)
  !----------------------------------------------------------------------------
    
    use fitsimage

    implicit none

    type(r4cube_type) :: image
    real(real32), intent(in) :: xsrc, ysrc, rcut2
    integer, intent(in) :: ii
    real(real32), intent(out) :: asum
    real(real32) dx, dy, dr2, rcut
    integer i, j, imin, imax, jmin, jmax

    asum=0.0
    rcut = sqrt(rcut2)
    
    imin = int(xsrc - 1.5 * rcut)
    jmin = int(ysrc - 1.5 * rcut)
    if(imin.lt.1)imin=1
    if(jmin.lt.1)jmin=1
    imax = int(xsrc + 1.5 * rcut)
    jmax = int(ysrc + 1.5 * rcut)
    if(imax.gt.image%shapehdr(ii)%naxis1)imax=image%shapehdr(ii)%naxis1
    if(jmax.gt.image%shapehdr(ii)%naxis2)jmax=image%shapehdr(ii)%naxis2

    !write(*,*) xsrc, ysrc, rcut2, rcut
    !write(*,*) imin, jmin, imax, jmax 

    
!  18/06/2018 AGE: no need to loop through whole image...
!    do i=1, image%shapehdr(ii)%naxis1
!       do j = 1, image%shapehdr(ii)%naxis2
    do i=imin, imax
       do j = jmin, jmax
          
          dx = i - xsrc
          dy = j - ysrc
          dr2 = dx * dx + dy * dy

          !write(*,*) i, j, imin, imax, jmin, jmax
          
          if(dr2<rcut2)then
             asum = asum + image%arr(i,j, ii)
          endif
       enddo
    enddo
  end subroutine extract_single_r4

  !----------------------------------------------------------------------------
  subroutine extract_single_exp(image, ii, xsrc, ysrc, rcut2, asum)
  !----------------------------------------------------------------------------
    
    use fitsimage

    implicit none

    type(r4cube_type) :: image
    real(real32), intent(in) :: xsrc, ysrc, rcut2
    integer, intent(in) :: ii
    real(real32), intent(out) :: asum
    real(real32) dx, dy, dr2, rcut
    integer i, j, imin, imax, jmin, jmax, area

    rcut = sqrt(rcut2)
    
    imin = int(xsrc - 1.5 * rcut)
    jmin = int(ysrc - 1.5 * rcut)
    if(imin.lt.1)imin=1
    if(jmin.lt.1)jmin=1
    imax = int(xsrc + 1.5 * rcut)
    jmax = int(ysrc + 1.5 * rcut)
    if(imax.gt.image%shapehdr(ii)%naxis1)imax=image%shapehdr(ii)%naxis1
    if(jmax.gt.image%shapehdr(ii)%naxis2)jmax=image%shapehdr(ii)%naxis2


    !write(*,*) xsrc, ysrc, rcut2, rcut
    !write(*,*) imin, jmin, imax, jmax 
    
    asum=0.0
    area = 0
!  18/06/2018 AGE: no need to loop through whole image...
!    do i=1, image%shapehdr(ii)%naxis1
!       do j = 1, image%shapehdr(ii)%naxis2
    do i=imin, imax
       do j = jmin, jmax
          !write(*,*) i,j,ii, imin, imax, jmin, jmax
          dx = i - xsrc
          dy = j - ysrc
          dr2 = dx * dx + dy * dy
          !write(10,*) i,j, image%arr(i,j, ii), ii, "exp"
          if(dr2<rcut2 .and. image%arr(i,j, ii).gt.0)then
             asum = asum + image%arr(i,j, ii)
             area = area + 1 
          endif
       enddo
    enddo
    if(area.gt.0)then
       asum = asum / float(area)
    else
       asum=0.0
    endif
  end subroutine extract_single_exp





  !----------------------------------------------------------------------------
  subroutine ape_read_apelist(par, filename, apelist, image, ierr)
  !----------------------------------------------------------------------------
  ! passes it to 2d boxlist type. The 2nd d to be consistent 
  ! with various ermldetect subroutine calls
  ! It passes RA, DEC and also estimates corresponding X, Y given input images.
  ! Assumes that fits extension has name "SRCLIST" and RA, DECs are stored in 
  ! columns called "RA", "DEC"
  !----------------------------------------------------------------------------
    !use detect_io
    USE detect_io, only : i4cube_type, ermldet_param_type, boxlist2d_type, init_boxlist2d, nim
    !use cfitsio_ifaces
    use radec2pix_mod

    implicit none
    type(ermldet_param_type), intent(in) :: par
    type(i4cube_type)  :: image
    character(len=*), intent(in) :: filename
    integer(kind=int32), intent(out) :: ierr
    type(boxlist2d_type), intent(out) ::apelist

    integer(kind=int32)  :: I, IROW, NROWS
    integer(kind=int32)  :: colnum_ra, colnum_dec
    integer(kind=int32)  :: unit, readwrite, blocksize, nelems, felems, anyf, status
    integer(kind=int32)  :: hdutype, extver
    real(kind=real32)    :: nullval
    character(len=*), parameter  :: rtname='APE_READ_APELIST'
    logical casesen
    real(kind=real32)    :: ra, dec
    character :: coltemplate*10, extname*10
    real(real64), dimension(:), pointer :: xatt, yatt    ! attitude in image pixels
    real(real64) :: ra0, dec0, x0, y0, &     ! WCS for sky images
         xscal, yscal             !
    integer :: ix_size, iy_size              ! size of sky images

    ! check input images
    nim=par%n_imagesets
    

    !  Get an unused Logical Unit Number to use to open the FITS file.
    call ftgiou(unit,status)

    !  open the FITS file, with read-only access.  The returned BLOCKSIZE
    !  parameter is obsolete and should be ignored. 
    readwrite=0
    call ftopen(unit,filename,readwrite,blocksize,status)

    !Move to the (first) HDU which has the specified extension type and EXTNAME (or HDUNAME) and EXTVER keyword values. The hdutype parameter may have a value of IMAGE_HDU (0), ASCII_TBL (1), BINARY_TBL (2), or ANY_HDU (-1) where ANY_HDU means that only the extname and extver values will be used to locate the correct extension. If the input value of extver is 0 then the EXTVER keyword is ignored and the first HDU with a matching EXTNAME (or HDUNAME) keyword will be found. If no matching HDU is found in the file then the current HDU will remain unchanged and a status = BAD_HDU_NUM (301) will be returned. 
    hdutype=2
    extver=0
    extname="SRCLIST"
    call FTMNHD(unit, hdutype, extname, extver, status)
    call fitserr("FTMNHD", status, ierr)   
    
    !Get the number of rows or columns in the current FITS table.
    call FTGNRW(unit, nrows, status) 

    !Get the table column number (and name) of the column whose name matches an input template name.
    casesen = .false.
    coltemplate = "RA"
    call FTGCNO(unit,casesen,coltemplate, colnum_ra,status)
    coltemplate = "DEC"
    call FTGCNO(unit,casesen,coltemplate, colnum_dec,status)

    ! write 2-dimensional source list:
    allocate(apelist%data(0:nim,nrows))
    call init_boxlist2d(apelist%data)

    apelist%shapehdr%naxis1 = nim
    apelist%shapehdr%naxis2 = nrows
    
    !write(*,*) "APELIST NROWS, NIM",colnum_ra, colnum_dec, nrows, nim
    nelems=1
    felems=1
    nullval=0.0
    do irow=1,nrows      
       call ftgcve(unit,colnum_ra,irow,felems,nelems,nullval,RA, &
            anyf,status)
       call ftgcve(unit,colnum_dec,irow,felems,nelems,nullval,DEC, &
            anyf,status)
       
       do I=1,nim
          apelist%data(I,irow)%ra = RA
          apelist%data(I,irow)%dec = DEC
          apelist%data(I,irow)%id_src = irow
       enddo

    end do



    allocate (xatt(nrows))
    allocate (yatt(nrows))
    do I=1,nim
       ! transform ra, dec to x and y
       ra0  = image%wcshdr(I)%crval1
       dec0 = image%wcshdr(I)%crval2
       x0   = image%wcshdr(I)%crpix1
       y0   = image%wcshdr(I)%crpix2
       xscal= image%wcshdr(I)%cdelt1
       yscal= image%wcshdr(I)%cdelt2
       ix_size = image%shapehdr(I)%naxis1
       iy_size = image%shapehdr(I)%naxis2
       allocate (xatt(nrows))
       allocate (yatt(nrows))
       call radec2pix (apelist%data(I,:)%ra,  apelist%data(I,:)%dec, nrows, ra0, dec0, x0, y0, xscal, yscal, xatt, yatt)               
       do irow=1,nrows      
          apelist%data(I,irow)%x_ima = REAL(xatt(irow))
          apelist%data(I,irow)%y_ima = REAL(yatt(irow))
       enddo
    end do
    
    !write(*,*) apelist%data(1,5:10)%ra
    !write(*,*) apelist%data(1,5:10)%dec
    !write(*,*) apelist%data(1,5:10)%x_ima
    !write(*,*) apelist%data(1,5:10)%y_ima

    deallocate (xatt, yatt)
    call ftclos(unit, status)
    call fitserr("ftclos", status, ierr)   
    
  end subroutine ape_read_apelist





!----------------------------------------------------------------------------
  !subroutine ape_ml2box(par, mllist, boxsrclist, ierr) 
  subroutine ape_ml2box(mllist, boxsrclist, ierr) 
!----------------------------------------------------------------------------
! convert an input mllist into a boxlist to be consistent with ermldet  
          

    use detect_io

    implicit none
    !type(ermldet_param_type), intent(in) :: par
    type(mllist_type), intent(in) :: mllist
    integer(kind=int32), intent(out) :: ierr
    type(boxlist2d_type), intent(out) :: boxsrclist

    logical(kind=4) :: has_instsum,has_bandsum   
    integer(kind=int32)  :: i,j,k,nrows,nbands
    character(len=*), parameter  :: rtname='APE_ML2BOX'
    type(boxlist_type) :: tmplist
    !integer(kind=int32),dimension(:),allocatable    :: isort
    character(len=80)                               :: errstr
    !integer(kind=int32)                             :: sla_err

    nrows=mllist%shapehdr%naxis2
    !write(*,*)"nrows",nrows
    
    ! check if input source list is empty
    if(nrows.gt.0)then

       ! check input images
       ! this is the old value. It reads the number of bands from the number of images
       !nim=par%n_imagesets
       ! this is the new, it simply finds the max value of the id_bands entry
       ! from the mllist.
       nim = maxval(mllist%data(:)%id_band)

       !write(*,*)"nim",nim, minval(mllist%data%id_band), &
       !     maxval(mllist%data%id_band),&
       !     minval(mllist%data%id_inst),&
       !     maxval(mllist%data%id_inst)
       
       ! check source list for summary rows:
       has_bandsum = .false.
       has_instsum = .false.
       nbands = maxval(mllist%data%id_band) 
       if  (minval(mllist%data%id_band) /=  maxval(mllist%data%id_band)) then
          if (minval(mllist%data%id_band) == 0) then 
             has_bandsum = .true.
          else
             ierr=3
             errstr = 'Unsupported format of input source list'
             call WRERR(rtname,ierr,errstr)
          endif
       endif
       if  (minval(mllist%data%id_inst) /=  maxval(mllist%data%id_inst)) then
          if (minval(mllist%data%id_inst) == 0) then 
             has_instsum = .true.
          else
             ierr=3
             errstr = 'Unsupported format of input source list'
             call WRERR(rtname,ierr,errstr)
          endif
       endif
       
       nrows=mllist%shapehdr%naxis2
       
       !write(*,*) has_instsum, has_bandsum
       
       if (has_instsum) then
          allocate(tmplist%data(nrows))
          call init_boxlist1d(tmplist%data)
          tmplist%shapehdr=mllist%shapehdr
          tmplist%stdhdr=mllist%stdhdr
          j=0
          do i=1,nrows
             if (mllist%data(i)%id_inst == 0) then
                j=j+1
                tmplist%data(j)=mllist%data(i)
             endif
          enddo
          nrows=j
          !      deallocate(boxlist%data)
          !      allocate(boxlist%data(nrows))
          !      boxlist%data=tmplist%data(1:nrows)
          !      deallocate(tmplist%data)
       else if (has_bandsum) then 
          allocate(tmplist%data(nrows))
          call init_boxlist1d(tmplist%data)
          tmplist%shapehdr=mllist%shapehdr
          tmplist%stdhdr=mllist%stdhdr
          j=0
          do i=1,nrows
             if (mllist%data(i)%id_band == 0) then
                j=j+1
                tmplist%data(j)=mllist%data(i)
             endif
          enddo
          nrows=j
          !      deallocate(boxlist%data)
          !      allocate(boxlist%data(nrows))
          !      boxlist%data=tmplist%data(1:nrows)
          !      deallocate(tmplist%data)
       endif
       
       ! AGE 08/2018: this is screwing up the order of sources. Commented out. 
       ! sort input list according to 'like' column:
       !if (minval(tmplist%data%like) /= maxval(tmplist%data%like)) then
       !   allocate(isort(nrows))
       !   ! sort in descending order, don't permutate array  (kflag=-1)
       !   call spsort(tmplist%data%like,nrows,isort,-1,sla_err)
       !   tmplist%data=tmplist%data(isort)
       !   deallocate(isort)
       !endif


       ! write 2-dimensional source list:
       allocate(boxsrclist%data(0:nim,nrows))
       !write(*,*) "SHAPE BOXLIST:", SHAPE(boxsrclist%data), "NIM:", nim, nrows
       call init_boxlist2d(boxsrclist%data)
       ! summary row gets first index 0:
       boxsrclist%data(0,:) = tmplist%data(1:nrows)
       deallocate(tmplist%data)
       boxsrclist%shapehdr=mllist%shapehdr
       boxsrclist%stdhdr=mllist%stdhdr
       boxsrclist%shapehdr%naxis1=nim  
       boxsrclist%shapehdr%naxis2=nrows  
       

       ! now fill the other columns 
       do j=1,nrows
          k=0
          do i=1,mllist%shapehdr%naxis2 
             !        write(*,*) 'id_src= ',boxsrclist%data(0,j)%id_src, boxlist%data(i)%id_src
             if  (boxsrclist%data(0,j)%id_src  == mllist%data(i)%id_src) then
                if (mllist%data(i)%id_inst /= 0 .and. mllist%data(i)%id_band /= 0) then 
                   k=k+1
                   
                   
                   boxsrclist%data(k,j)=mllist%data(i) 
!                   WRITE(*,*) "ERM_2BOX-1",J, I, NROWS, mllist%shapehdr%naxis2, &
!                       mllist%data(i)%x_ima, mllist%data(i)%y_ima
!                   
!                    
!                   WRITE(*,*) "ERM_2BOX-2",J, I, NROWS, mllist%shapehdr%naxis2, &
!                       boxsrclist%data(k,j)%id_band, boxsrclist%data(k,j)%ra,  boxsrclist%data(k,j)%dec
                   
                   
                endif
             endif
          enddo
       enddo
       
    else ! check if input source list is empty
       ierr=3 ! severe error; no point in continuing program execution 
       errstr = 'Input ermldet source list empty'
       call WRERR(rtname,ierr,errstr)   
    endif  ! check if input source list is empty
 
 !write(*,*) SHAPE(boxsrclist%data)

  end subroutine ape_ml2box


  !----------------------------------------------------------------------------
  subroutine ape_read_mllist(par, filename, mllist, boxsrclist, ierr)
  !----------------------------------------------------------------------------
  ! reads mllist and passes it to 2d boxlist type to be consistent 
  ! with various ermldetect subroutine calls
  !----------------------------------------------------------------------------
    use detect_io
    !use ermldet_mod

    implicit none
    type(ermldet_param_type), intent(in) :: par
    type(mllist_type), intent(in) :: mllist
    character(len=*), intent(in) :: filename
    integer(kind=int32), intent(out) :: ierr
    type(boxlist2d_type), intent(out) :: boxsrclist

    !logical(kind=4) :: has_instsum, has_bandsum   
    !integer(kind=int32)  :: i,j,k,ipoint,nrows,ord=1,nbands,nmask
    character(len=*), parameter  :: rtname='READ_MLLIST'
    !type(boxlist_type) :: tmplist
    !integer(kind=int32),dimension(:),allocatable    :: isort
    !character(len=80)                               :: errstr
    !integer(kind=int32)                             :: sla_err


    
    ! check input images
    nim=par%n_imagesets


    call fget_mllist(filename,mllist,ierr)

    
    !call ape_ml2box(par, mllist, boxsrclist, ierr) 
    call ape_ml2box(mllist, boxsrclist, ierr) 


  end subroutine ape_read_mllist



 
  !---------------------------------------------------------------------
  subroutine ape_srcmap_isrc(apar,image,eventlist,pixlist,detmask,expimage,bkgimage, &
       allStackPSF,cluPSF,isrc,srcimage, &
       allStackPSF2,cluPSF2,boxlist,srcmodrad,ierr)

  !--------------------------------------------------------------------
    use detect_io
    use ermldet_mod
    use slib_image
    implicit none

    type(APEPAR), intent(in) :: apar
    type(i4cube_type)                        :: image
    type(tEventFile),dimension(:)            :: eventlist
    type(pixlist_type),dimension(:)          :: pixlist
    type(l1cube_type)                        :: detmask
    type(r4cube_type)                        :: expimage
    type(r4cube_type)                        :: bkgimage
    type(r4cube_type)                        :: srcimage
    !type(r4cube_type)                        :: run_srcimage
    type(boxlist2d_type)                     :: boxlist
    integer(kind=int32)                      :: srcmodrad  
    integer(kind=int32)                      :: ierr
    integer(kind=int32)                      :: isrc
    integer(kind=int32)                      :: ienergy
   
    type(events_type),pointer,dimension(:)   :: events => NULL()
    !type(srcrow_type),dimension(:,:),pointer :: clulist 
    type(srcrow_type),dimension(:,:),allocatable :: clulist
    type(FPSF)                               :: allStackPSF,allStackPSF2   
    type(FPSF)                               :: cluPSF,cluPSF2   
    type(mask_type),save                     :: mask
    
    ! this is used only for testing
    !type(r4image_type)                       :: junk
    !integer(kind=int32)                      :: xsize,ysize
 
!    real(kind=real32), dimension(:,:,:), pointer   :: modelima   
    real(kind=real64), dimension(:,:,:), pointer   :: modelima  => NULL()  

    integer(kind=int32)                      :: nclu, nclu_single
    character(len=*), parameter              :: rtname='APE_SRCMAP_ISRC'

    ! these are need by model subroutine call further down
!    real(kind=real32)                        :: yfit0
    real(kind=real64)                        :: yfit0
!    real(kind=real32),dimension(:),allocatable  :: yfit1
!    real(kind=real32),dimension(max_ima)        :: yfit2 
    real(kind=real64),dimension(:),allocatable  :: yfit1
    real(kind=real64),dimension(max_ima)        :: yfit2 
    real(kind=real64), dimension(max_fit)         :: x0,y0,sig,ra,dec         
    real(kind=real64), dimension(max_ima,max_fit) :: rate
    integer(kind=int32)                           :: ipoint,ix,iy,mx,my
    logical(kind=log32)                         :: last = .FALSE.
    logical(kind=log32)                      :: newcal = .TRUE.
    integer(kind=int32)                      :: status
    logical :: testing=.false.
    ! name is used for testing
    character ( len = 256) filename, name
    real(kind=real32)                          :: pix2arcsec

    !only used for testing assignment is to suppress warning
    name=""

    !IENERGY = ierr
    ierr = 0

    !AGE ADDED 2012-05-07
    status=0
    
    ! read new cal PSF
    newcal=.true.
    last=.false.

    ! %ext is in arcsec, but in model tasks is assumed to be pixels.
    ! I therefore estimate the pixel size here
    ! look into detect_io_mod.f90: format_mllist
    pix2arcsec=(abs(REAL(image%wcshdr(1)%cdelt1))+abs(REAL(image%wcshdr(1)%cdelt2)))*1800.0

    ! allocate clulist with summary element (0:n_imagesets,maxclu)
    ! AGE:     max_fit= 10 is global 
    ! so this array holds current group of sources to fit
    !if (associated(clulist))nullify(clulist)
    if (allocated(clulist))deallocate(clulist)
    allocate(clulist(0:apar%epar%n_imagesets,max_fit))    

    call init_mllist2d(clulist)

    call ape_sou_clus(apar,boxlist,allStackPSF,allStackPSF2,isrc,clulist,cluPSF,cluPSF2,nclu,ierr)  

    
    ! set cut-out radii for source model - srcmodrad is fixed in main
    clulist(:,1:nclu)%ecutrad = real(srcmodrad)

    ! in ermldet all source that belong to the same cluster are
    ! modeled together. If I want to substract a single source model
    ! and no all sources of the cluster I have to explicitely set the
    ! nlcu = 1. I define new parameter  nclu_single=1 and is passed to
    ! model subroutine
    nclu_single=1
    !old call with nclu:
    !call fill_arrs(apar%epar,.false.,clulist,image,eventlist,pixlist,cluPSF,detmask,expimage,modelima,nclu,events,mask,status)
    ! new call with nclu_single


    call fill_arrs(apar%epar,.false.,clulist,image,eventlist,pixlist,cluPSF,detmask,expimage,modelima,nclu_single,events,mask,status)


    if (allocated(yfit1)) deallocate(yfit1)
    allocate(yfit1(mask%nev))  
    x0  = clulist(0,:)%x_ima
    y0  = clulist(0,:)%y_ima
    ra  = clulist(0,:)%ra
    dec  = clulist(0,:)%dec
    ! in model  the source extent is in pixels, but in the
    ! ermllist is in arcsec. Here I convert. See also format_mllist
    ! detect_io
    sig = clulist(0,:)%ext/pix2arcsec
    rate(1:apar%epar%n_imagesets,:) = clulist(1:apar%epar%n_imagesets,:)%rate 
    !write(*,*) "X0, Y0=",x0, y0
    !write(*,*) 'RATE=', rate(1:apar%epar%n_imagesets,:)
    !write(*,*) 'SIG:',sig
    ! see above, this is old call with nclu
    !call model(apar%epar,image,expimage,bkgimage,mask,events,clulist,cluPSF2,x0,y0,sig,rate,nclu,last,newcal,yfit0,yfit1,yfit2,modelima,status)
    ! see above, this is new call with nclu_single

    !if(nclu.ge.1)then
    !      write(*,*)"RA:",ra(1:nclu)
    !      write(*,*)"DEC:",dec(1:nclu)
    !      write(*,*)"X0:",x0(1:nclu)
    !      write(*,*)"Y0:",y0(1:nclu)
    !      write(*,*)"SIG:",sig(1:nclu)
    !      write(*,*)"RATE:",rate(1,1:nclu)
    !      write(*,*)"NEV:",mask%nev
    !      write(*,*)"LAST:",last
    !      write(*,*)"NEWCAL:",newcal
    !      write(*,*)"NCLU:",nclu
    !      do i=1,nclu
    !         write(*,*) "cluPSF",i, cluPSF%mySPSF(1)%c0(:,i)
    !         write(*,*) "cluPSF",i, cluPSF%mySPSF(1)%c1(:,i)
    !         write(*,*) "cluPSF",i, cluPSF%mySPSF(1)%c2(:,i)
    !      enddo
    !endif
    
    

    call model(apar%epar,image,expimage,bkgimage,mask,events,clulist,cluPSF2,x0,y0,sig,rate,nclu_single,last,newcal,yfit0,yfit1,yfit2,modelima,status)

    

!    if(testing)then
!    ! AGE TESTING
!       IENERGY=1
!       write (filename, "(A5, I6.6, A1, I1.1)") &
!            "SRCMODEL_", ISRC, "_",IENERGY
!       filename=trim(filename)//'_1.img'
!       call slib_run_write2image(dble(modelima(:,:,IENERGY)), filename)
!       write (filename, "(A5, I4.4)") &
!            "SRC_", ISRC
!       filename=trim(filename)//'_2.img'
!       !filename='test2.img'
!       call slib_run_write2arrays_fast(cluPSF%mySPSF(IENERGY),cluPSF%myCPSF,1)
!       call slib_run_write2image(cluPSF%myCPSF%img, filename)
!    endif

    do ipoint=1,apar%epar%n_imagesets
       my=0
       do iy = mask%iy_min, mask%iy_max
          mx=0
          my=my+1
          do ix = mask%ix_min, mask%ix_max
             mx=mx+1
             if (mask%mask(mx,my,ipoint) > 0) then
                !write(*,*) ix,iy,srcimage%arr(ix,iy,ipoint),modelima(mx,my,ipoint),  expimage%arr(ix,iy,ipoint)
              srcimage%arr(ix,iy,ipoint) = srcimage%arr(ix,iy,ipoint)  &
                     - REAL(modelima(mx,my,ipoint)) * expimage%arr(ix,iy,ipoint)

                   
             endif
          enddo
       enddo
    end do
    
    if(testing)then
       ! AGE TESTING
       !if(ISRC.eq.1.or.ISRC.eq.2.or.ISRC.eq.38.or.ISRC.eq.39)then
       if(ISRC.eq.1.or.ISRC.eq.39.or.ISRC.eq.2.or.ISRC.eq.117)then
          !if(sig(1).gt.0)then
          ienergy=1
          write (filename, "(A7, I6.6, A1, I1.1)") &
               "SRCMAP_", ISRC, "_", ienergy
          filename=trim(filename)//'_1.img'
          call slib_run_write2image(dble(srcimage%arr(:,:,IENERGY)), filename)
          write(*,*) filename, sig(1)
       endif
    endif

!    if(testing)then
!       xsize=size(srcimage%arr,1)
!       ysize=size(srcimage%arr,2)
!       allocate(junk%arr(xsize,ysize))
!       i=1
!       !do i=1,apar%epar%n_srcimagesets
!          
!          junk%arr=srcimage%arr(:,:,i)
!          junk%stdhdr=srcimage%stdhdr(i)
!          junk%refhdr=srcimage%refhdr(i)
!          junk%wcshdr=srcimage%wcshdr(i)
!          
!
!          junk%shapehdr%bitpix=-32
!          junk%shapehdr%naxis=2
!          junk%shapehdr%naxis1=xsize
!          junk%shapehdr%naxis2=ysize  
!
!          junk%stdhdr%hduclas1='IMAGE'
!          junk%stdhdr%hduclas2='GENERIC'
!          junk%stdhdr%creator='ERMLDET'
!          
!          
!          !enddo
!          ierr=0
!       call fput_r4image("!test1.fits",junk,status)
!       if (status >  0) then
!          ierr = 3
!          call wrerr(rtname,ierr,'Error writing source-map file')
!       endif
!       
!   
!
!       deallocate(junk%arr)
!    endif

    
    deallocate(yfit1)
    
  end  subroutine ape_srcmap_isrc
  !--------------------------------------------------------------------------------



!  !---------------------------------------------------------------------
!  !subroutine ape_srcmap(par,boxlist,allStackPSF,cluPSF,allStackPSF2,cluPSF2,ierr)  
!  subroutine ape_srcmap(apar,image,eventlist,pixlist,detmask,expimage,bkgimage,allStackPSF,cluPSF,allStackPSF2,cluPSF2,srcimage,boxlist,srcmodrad,nsrc,ierr)
!
!  !--------------------------------------------------------------------
!     use detect_io
!    use ermldet_mod
!    implicit none
!
!    type(APEPAR)                 :: apar
!    type(i4cube_type)                        :: image
!    type(tEventFile),dimension(:)            :: eventlist
!    type(pixlist_type),dimension(:)          :: pixlist
!    type(l1cube_type)                        :: detmask
!    type(r4cube_type)                        :: expimage
!    type(r4cube_type)                        :: bkgimage
!    type(r4cube_type)                        :: srcimage
!    type(boxlist2d_type)                     :: boxlist
!    integer(kind=int32)                      :: srcmodrad  
!    integer(kind=int32)                      :: ierr
!    integer(kind=int32)                      :: nsrc
!   
!    type(events_type),pointer,dimension(:)   :: events  => NULL()  
!    !type(srcrow_type),dimension(:,:),pointer :: clulist  => NULL()
!    type(srcrow_type),dimension(:,:),allocatable :: clulist
!    type(FPSF)                               :: allStackPSF,allStackPSF2   
!    type(FPSF)                               :: cluPSF,cluPSF2   
!    type(mask_type),save                     :: mask
!
!!    real(kind=real32), dimension(:,:,:), pointer   :: modelima   
!    real(kind=real64), dimension(:,:,:), pointer   :: modelima  => NULL()  
!
!    integer(kind=int32)                      :: nclu
!    integer(kind=int32)                      :: isrc, nsrc_in
!    character(len=*), parameter              :: rtname='APE_SRCMAP'
!
!    ! these are need by model subroutine call further down
!!    real(kind=real32)                        :: yfit0
!    real(kind=real64)                        :: yfit0
!!    real(kind=real32),dimension(:),allocatable  :: yfit1
!!    real(kind=real32),dimension(max_ima)        :: yfit2 
!    real(kind=real64),dimension(:),allocatable  :: yfit1
!    real(kind=real64),dimension(max_ima)        :: yfit2 
!    real(kind=real64), dimension(max_fit)         :: x0,y0,sig         
!    real(kind=real64), dimension(max_ima,max_fit) :: rate
!    integer(kind=int32)                           :: ipoint,ix,iy,mx,my
!    logical(kind=log32)                         :: last = .FALSE.
!    logical(kind=log32)                      :: newcal = .TRUE.
!    integer(kind=int32)                      :: status
!    ierr = 0
!    status=0
!   
!
!    ! allocate clulist with summary element (0:n_imagesets,maxclu)
!    ! AGE:     max_fit= 10 is global 
!    ! so this array holds current group of sources to fit
!    !if (.not. associated(clulist)) allocate(clulist(0:apar%epar%n_imagesets,max_fit))
!    if (.not. allocated(clulist)) allocate(clulist(0:apar%epar%n_imagesets,max_fit))  
!    call init_mllist2d(clulist)
!
!    nsrc_in = boxlist%shapehdr%naxis2
!   
!    
!    do isrc=1,nsrc_in
!       
!       !write(*,*) rtname, isrc
!
!       call ape_sou_clus(apar,boxlist,allStackPSF,allStackPSF2,isrc,clulist,cluPSF,cluPSF2,nclu,ierr)  
!  
!       ! set cut-out radii for source model - srcmodrad is fixed in main
!       clulist(:,1:nclu)%ecutrad = real(srcmodrad)
!
!       call fill_arrs(apar%epar,.false.,clulist,image,eventlist,pixlist,cluPSF,detmask,expimage,modelima,nclu,events,mask,status)
!       ! allocate yfit1 , only needed for compatibility with subroutine 'model':
!       if (allocated(yfit1)) deallocate(yfit1)
!       allocate(yfit1(mask%nev))  
!       x0  = clulist(0,:)%x_ima
!       y0  = clulist(0,:)%y_ima
!       sig = clulist(0,:)%ext
!       rate(1:apar%epar%n_imagesets,:) = clulist(1:apar%epar%n_imagesets,:)%rate
!
!       call model(apar%epar,image,expimage,bkgimage,mask,events,clulist,cluPSF,x0,y0,sig,rate,nclu,last,newcal,yfit0,yfit1,yfit2,modelima,status)
!
!       do ipoint=1,apar%epar%n_imagesets
!          my=0
!          do iy = mask%iy_min, mask%iy_max
!             mx=0
!             my=my+1
!             do ix = mask%ix_min, mask%ix_max
!                mx=mx+1
!                if (mask%mask(mx,my,ipoint) > 0) then
!                   bkgimage%arr(ix,iy,ipoint) = bkgimage%arr(ix,iy,ipoint)  &
!                        + REAL(modelima(mx,my,ipoint)) * expimage%arr(ix,iy,ipoint)
!                   !              write(*,*) "add_srcmap:", ipoint,mask%mask(mx,my,ipoint),modelima(mx,my,ipoint), expimage%arr(ix,iy,ipoint)
!                   
!                endif
!             enddo
!          enddo
!       end do
! 
!  
!       deallocate(yfit1)
!    enddo
!    
!  end  subroutine ape_srcmap
!  !--------------------------------------------------------------------------------



 !---------------------------------------------------------------------
  subroutine ape_sou_clus(apar,boxlist,allStackPSF,allStackPSF2,isrc,clulist,cluPSF,cluPSF2,nclu,ierr)  
  !--------------------------------------------------------------------
  ! this is adapted from the ermldet_mod.f90 sou_clus routine. 
  ! It defines the clulist,cluPSF,cluPSF2,nclu
  ! to be used later. 
  ! it uses id_cluster flag in mllist to identify sources that 
  ! belong to the same cluster.
  !--------------------------------------------------------------------

    use detect_io
    !use ermldet_mod
    implicit none


    type(APEPAR),intent(in)                  :: apar
    type(boxlist2d_type)                     :: boxlist
    type(FPSF)                               :: allStackPSF
    type(FPSF)                               :: allStackPSF2
    integer(kind=int32), intent(in)         :: isrc
!    type(srcrow_type),dimension(:,:),pointer :: clulist
    type(srcrow_type),dimension(:,:),allocatable :: clulist
    type(FPSF)                               :: cluPSF
    type(FPSF)                               :: cluPSF2
    integer(kind=int32)                      :: nclu,ierr
    !integer(int32), dimension(:), intent(in) :: energy_index
    !integer(int32), intent(in) :: n_energy_index

    
    integer(kind=int32)                      :: i, nsrc, iima, K
    integer(kind=int32)                      :: current_cluster_id
    character(len=*), parameter              :: rtname='APE_SOU_CLUS'
    !real(kind=real32)                        :: r_off,scut2,rad2


    
    ierr=0
    nsrc=boxlist%shapehdr%naxis2 
    nclu = 0
    !scut2=boxlist%data(0,isrc)%scutrad**2 
    
    ! at least one element in nlcu
    nclu=nclu+1
    ! clulist had 1 dimension equals to the number of input images
    ! this may be different from the number of bands in the
    ! input ermldet list (boxlist below). For that reason there is
    ! an index list that maps the input images (in the order they appear)
    ! to the index of the bands as they appear in the ermldet
    ! If no EINDEX has been provided, then it assumes that
    ! bands and images appear in the same order and are equal

    !write(*,*) "IN", n_energy_index
    ! this is to account for the possibility of
    ! less input images than number of bands in the
    ! ermldet task, eg in the case of stacking
    ! In this case a mapping should be provided
    ! between input images and ermldet bands
    !write(*,*) SHAPE(clulist), SHAPE(boxlist%data)
    if(apar%n_energy_index==0)then 
       clulist(:,nclu)=boxlist%data(:,isrc)
    else
       clulist(0,nclu)=boxlist%data(0,isrc)
       do K=1,apar%n_energy_index
          !write(*,*) K, "S:",SHAPE(clulist), "BOX:", SHAPE(boxlist%data),&
          !     "energy_index(K):",apar%energy_index(K), apar%n_energy_index
          clulist(K,nclu)=boxlist%data(apar%energy_index(K),isrc)
       enddo
    endif
    current_cluster_id = boxlist%data(0,isrc)%id_cluster

    !write(*,*) rtname, " :START: ",  current_cluster_id, nclu

    ! why is that here????
    if (apar%epar%shapelet_flag) then
       if (.not. allocated(cluPSF%myModelPSF)) then
          allocate(cluPSF%myModelPSF(size(allStackPSF%myModelPSF)))
          allocate(cluPSF2%myModelPSF(size(allStackPSF2%myModelPSF)))
          !write(*,*) rtname,"NOT ALLOCATED?"
       endif

       do iima=1,apar%epar%n_imagesets
                 !write(*,*) 'cluPSF: ',iima, size(cluPSF)
                 !write(*,*) 'cluPSF: ',iima, size(cluPSF), size(cluPSF(iima)%c1)
                 !write(*,*) 'size(cluPSF%myModelPSF),size(allStackPSF%myModelPSF:'
                 !write(*,*) size(cluPSF%myModelPSF),size(allStackPSF%myModelPSF) 
          cluPSF%myModelPSF(:) = allStackPSF%myModelPSF(:)
          !       write(*,*) 'size(NX1)=',size(cluPSF%myModelPSF(1)%NX1)
          !write(*,*) iima, nim, SHAPE(cluPSF%mySPSF), SHAPE(allStackPSF%mySPSF)
          cluPSF%mySPSF(iima)%c0(:,nclu) = allStackPSF%mySPSF(iima)%c0(:,isrc)
          cluPSF%mySPSF(iima)%c1(:,nclu) = allStackPSF%mySPSF(iima)%c1(:,isrc)
          cluPSF%mySPSF(iima)%c2(:,nclu) = allStackPSF%mySPSF(iima)%c2(:,isrc)
          cluPSF%mySPSF(iima)%nx0 = allStackPSF%mySPSF(iima)%nx0
          cluPSF%mySPSF(iima)%nx1 = allStackPSF%mySPSF(iima)%nx1
          cluPSF%mySPSF(iima)%nx2 = allStackPSF%mySPSF(iima)%nx2
          cluPSF%mySPSF(iima)%ny0 = allStackPSF%mySPSF(iima)%ny0
          cluPSF%mySPSF(iima)%ny1 = allStackPSF%mySPSF(iima)%ny1
          cluPSF%mySPSF(iima)%ny2 = allStackPSF%mySPSF(iima)%ny2
          
          cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,isrc)
          cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,isrc)
          cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,isrc)
          cluPSF2%mySPSF(iima)%nx0 = allStackPSF2%mySPSF(iima)%nx0
          cluPSF2%mySPSF(iima)%nx1 = allStackPSF2%mySPSF(iima)%nx1
          cluPSF2%mySPSF(iima)%nx2 = allStackPSF2%mySPSF(iima)%nx2
          cluPSF2%mySPSF(iima)%ny0 = allStackPSF2%mySPSF(iima)%ny0
          cluPSF2%mySPSF(iima)%ny1 = allStackPSF2%mySPSF(iima)%ny1
          cluPSF2%mySPSF(iima)%ny2 = allStackPSF2%mySPSF(iima)%ny2
       enddo
    endif
    ! look for other sources in cluster 
    ! it uses the id_cluster flag
    do i=isrc+1,nsrc

       !write(*,*) rtname, " :LOOP: ",  current_cluster_id,  boxlist%data(0,i)%id_cluster,  boxlist%data(0,i)%id_src, boxlist%data(0,i)%x_ima, boxlist%data(0,i)%y_ima
       if (boxlist%data(0,i)%id_cluster ==  current_cluster_id  ) then
       !write(*,*) rtname, " :FOUND: "
         
          nclu=nclu+1
          ! GHANGE THIS LATER ON     

          ! clulist had 1 dimension equals to the number of input images
          ! this may be different from the number of bands in the
          ! input ermldet list (boxlist below). For that reason there is
          ! an index list that maps the input images (in the order they appear)
          ! to the index of the bands as they appear in the ermldet
          ! If no EINDEX has been provided, then it assumes that
          ! bands and images appear in the same order and are equal
          if(apar%n_energy_index==0)then 
             clulist(:,nclu)=boxlist%data(:,i)
          else
             clulist(0,nclu)=boxlist%data(0,i)
             do K=1,apar%n_energy_index
                clulist(K,nclu)=boxlist%data(apar%energy_index(K),i)
             enddo
          endif
          !clulist(:,nclu)=boxlist%data(0:2,i)
          ! GHANGE THIS LATER ON     
          
          if (apar%epar%shapelet_flag) then
             do iima=1,nim
                ! change 19/11/2017:
                !luPSF%mySPSF(iima)%c0(:,nclu)  = allStackPSF%mySPSF(iima)%c0(:,isrc)
                !cluPSF%mySPSF(iima)%c1(:,nclu)  = allStackPSF%mySPSF(iima)%c1(:,isrc)
                !cluPSF%mySPSF(iima)%c2(:,nclu)  = allStackPSF%mySPSF(iima)%c2(:,isrc)
                cluPSF%mySPSF(iima)%c0(:,nclu)  = allStackPSF%mySPSF(iima)%c0(:,i)
                cluPSF%mySPSF(iima)%c1(:,nclu)  = allStackPSF%mySPSF(iima)%c1(:,i)
                cluPSF%mySPSF(iima)%c2(:,nclu)  = allStackPSF%mySPSF(iima)%c2(:,i)
                cluPSF%mySPSF(iima)%nx0 = allStackPSF%mySPSF(iima)%nx0
                cluPSF%mySPSF(iima)%nx1 = allStackPSF%mySPSF(iima)%nx1
                cluPSF%mySPSF(iima)%nx2 = allStackPSF%mySPSF(iima)%nx2
                cluPSF%mySPSF(iima)%ny0 = allStackPSF%mySPSF(iima)%ny0
                cluPSF%mySPSF(iima)%ny1 = allStackPSF%mySPSF(iima)%ny1
                cluPSF%mySPSF(iima)%ny2 = allStackPSF%mySPSF(iima)%ny2

                !change 19/11/2017
                !cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,isrc)
                !cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,isrc)
                !cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,isrc)
                cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,i)
                cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,i)
                cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,i)
                cluPSF2%mySPSF(iima)%nx0 = allStackPSF2%mySPSF(iima)%nx0
                cluPSF2%mySPSF(iima)%nx1 = allStackPSF2%mySPSF(iima)%nx1
                cluPSF2%mySPSF(iima)%nx2 = allStackPSF2%mySPSF(iima)%nx2
                cluPSF2%mySPSF(iima)%ny0 = allStackPSF2%mySPSF(iima)%ny0
                cluPSF2%mySPSF(iima)%ny1 = allStackPSF2%mySPSF(iima)%ny1
                cluPSF2%mySPSF(iima)%ny2 = allStackPSF2%mySPSF(iima)%ny2
             enddo
          endif
       endif
    enddo
 

    if (apar%epar%shapelet_flag) then
       do iima=1,apar%epar%n_imagesets
          cluPSF%mySPSF(iima)%num_src=nclu
          cluPSF2%mySPSF(iima)%num_src=nclu
       enddo
    endif


  end subroutine ape_sou_clus
  !--------------------------------------------------------------------------------





  
  !---------------------------------------------------------------------
  subroutine ape_sou_clus_old(par,boxlist,allStackPSF,allStackPSF2,isrc,clulist,cluPSF,cluPSF2,nclu,ierr)  
  !--------------------------------------------------------------------
  ! this is adapted from the ermldet_mod.f90 sou_clus routine. 
  ! It defines the clulist,cluPSF,cluPSF2,nclu
  ! to be used later. 
  ! it uses id_cluster flag in mllist to identify sources that 
  ! belong to the same cluster.
  !--------------------------------------------------------------------

    use detect_io
    !use ermldet_mod
    implicit none


    type(ermldet_param_type),intent(in)      :: par
    type(boxlist2d_type)                     :: boxlist
    type(FPSF)                               :: allStackPSF
    type(FPSF)                               :: allStackPSF2
    integer(kind=int32)                      :: isrc
    type(srcrow_type),dimension(:,:),pointer :: clulist   
    type(FPSF)                               :: cluPSF
    type(FPSF)                               :: cluPSF2
    integer(kind=int32)                      :: nclu,ierr

    integer(kind=int32)                      :: i, nsrc, iima
    integer(kind=int32)                      :: current_cluster_id
    character(len=*), parameter              :: rtname='APE_SOU_CLUS'


    ierr=0
    nsrc=boxlist%shapehdr%naxis2 
    nclu = 0
    !scut2=boxlist%data(0,isrc)%scutrad**2 

    !write(*,*) SHAPE(clulist)
    !write(*,*) SHAPE(boxlist%data)
    
    ! at least one element in nlcu
    nclu=nclu+1
    clulist(:,nclu)=boxlist%data(:,isrc)    
    current_cluster_id = boxlist%data(0,isrc)%id_cluster

    !write(*,*) rtname, " :START: ",  current_cluster_id, nclu

    ! why is that here????
    if (par%shapelet_flag) then
       if (.not. allocated(cluPSF%myModelPSF)) then
          allocate(cluPSF%myModelPSF(size(allStackPSF%myModelPSF)))
          allocate(cluPSF2%myModelPSF(size(allStackPSF2%myModelPSF)))
          !write(*,*) rtname,"NOT ALLOCATED?"
       endif

       do iima=1,nim
          !       write(*,*) 'cluPSF: ',iima, size(cluPSF)
          !       write(*,*) 'cluPSF: ',iima, size(cluPSF), size(cluPSF(iima)%c1)
          !       write(*,*) 'size(cluPSF%myModelPSF),size(allStackPSF%myModelPSF:'
          !       write(*,*) size(cluPSF%myModelPSF),size(allStackPSF%myModelPSF) 
          cluPSF%myModelPSF(:) = allStackPSF%myModelPSF(:)
          !       write(*,*) 'size(NX1)=',size(cluPSF%myModelPSF(1)%NX1)
          cluPSF%mySPSF(iima)%c0(:,nclu) = allStackPSF%mySPSF(iima)%c0(:,isrc)
          cluPSF%mySPSF(iima)%c1(:,nclu) = allStackPSF%mySPSF(iima)%c1(:,isrc)
          cluPSF%mySPSF(iima)%c2(:,nclu) = allStackPSF%mySPSF(iima)%c2(:,isrc)
          cluPSF%mySPSF(iima)%nx0 = allStackPSF%mySPSF(iima)%nx0
          cluPSF%mySPSF(iima)%nx1 = allStackPSF%mySPSF(iima)%nx1
          cluPSF%mySPSF(iima)%nx2 = allStackPSF%mySPSF(iima)%nx2
          cluPSF%mySPSF(iima)%ny0 = allStackPSF%mySPSF(iima)%ny0
          cluPSF%mySPSF(iima)%ny1 = allStackPSF%mySPSF(iima)%ny1
          cluPSF%mySPSF(iima)%ny2 = allStackPSF%mySPSF(iima)%ny2
          
          cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,isrc)
          cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,isrc)
          cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,isrc)
          cluPSF2%mySPSF(iima)%nx0 = allStackPSF2%mySPSF(iima)%nx0
          cluPSF2%mySPSF(iima)%nx1 = allStackPSF2%mySPSF(iima)%nx1
          cluPSF2%mySPSF(iima)%nx2 = allStackPSF2%mySPSF(iima)%nx2
          cluPSF2%mySPSF(iima)%ny0 = allStackPSF2%mySPSF(iima)%ny0
          cluPSF2%mySPSF(iima)%ny1 = allStackPSF2%mySPSF(iima)%ny1
          cluPSF2%mySPSF(iima)%ny2 = allStackPSF2%mySPSF(iima)%ny2
       enddo
    endif
    ! look for other sources in cluster 
    ! it uses the id_cluster flag
    do i=isrc+1,nsrc

       !write(*,*) rtname, " :LOOP: ",  current_cluster_id,  boxlist%data(0,i)%id_cluster,  boxlist%data(0,i)%id_src, boxlist%data(0,i)%x_ima, boxlist%data(0,i)%y_ima
       if (boxlist%data(0,i)%id_cluster ==  current_cluster_id  ) then
       !write(*,*) rtname, " :FOUND: "
         
          nclu=nclu+1
          clulist(:,nclu)=boxlist%data(:,i)
          if (par%shapelet_flag) then
             do iima=1,nim
                ! change 19/11/2017:
                !luPSF%mySPSF(iima)%c0(:,nclu)  = allStackPSF%mySPSF(iima)%c0(:,isrc)
                !cluPSF%mySPSF(iima)%c1(:,nclu)  = allStackPSF%mySPSF(iima)%c1(:,isrc)
                !cluPSF%mySPSF(iima)%c2(:,nclu)  = allStackPSF%mySPSF(iima)%c2(:,isrc)
                cluPSF%mySPSF(iima)%c0(:,nclu)  = allStackPSF%mySPSF(iima)%c0(:,i)
                cluPSF%mySPSF(iima)%c1(:,nclu)  = allStackPSF%mySPSF(iima)%c1(:,i)
                cluPSF%mySPSF(iima)%c2(:,nclu)  = allStackPSF%mySPSF(iima)%c2(:,i)
                cluPSF%mySPSF(iima)%nx0 = allStackPSF%mySPSF(iima)%nx0
                cluPSF%mySPSF(iima)%nx1 = allStackPSF%mySPSF(iima)%nx1
                cluPSF%mySPSF(iima)%nx2 = allStackPSF%mySPSF(iima)%nx2
                cluPSF%mySPSF(iima)%ny0 = allStackPSF%mySPSF(iima)%ny0
                cluPSF%mySPSF(iima)%ny1 = allStackPSF%mySPSF(iima)%ny1
                cluPSF%mySPSF(iima)%ny2 = allStackPSF%mySPSF(iima)%ny2

                !change 19/11/2017
                !cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,isrc)
                !cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,isrc)
                !cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,isrc)
                cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,i)
                cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,i)
                cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,i)
                cluPSF2%mySPSF(iima)%nx0 = allStackPSF2%mySPSF(iima)%nx0
                cluPSF2%mySPSF(iima)%nx1 = allStackPSF2%mySPSF(iima)%nx1
                cluPSF2%mySPSF(iima)%nx2 = allStackPSF2%mySPSF(iima)%nx2
                cluPSF2%mySPSF(iima)%ny0 = allStackPSF2%mySPSF(iima)%ny0
                cluPSF2%mySPSF(iima)%ny1 = allStackPSF2%mySPSF(iima)%ny1
                cluPSF2%mySPSF(iima)%ny2 = allStackPSF2%mySPSF(iima)%ny2
             enddo
          endif
       endif
    enddo
 

    if (par%shapelet_flag) then
       do iima=1,nim
          cluPSF%mySPSF(iima)%num_src=nclu
          cluPSF2%mySPSF(iima)%num_src=nclu
       enddo
    endif


  end subroutine ape_sou_clus_old
  !--------------------------------------------------------------------------------




  !--------------------------------------------------------------------------------
  subroutine setup_evlist(imageset, eventlist, ierr)
  !--------------------------------------------------------------------------------
  ! This sets up the tEventFile object, in two steps. The first just sets up an
  ! empty object, with the correct extension names and correct column names and
  ! types for each extension. The second step is to flag the small subset of
  ! extensions/columns required for the present task.
  !--------------------------------------------------------------------------------

    character(*),     intent(in)    :: imageset
    type(tEventFile), intent(inout) :: eventlist
    integer,          intent(out)   :: ierr

    integer, dimension(:), allocatable :: ccdnrs
    integer :: corratt_i,kk,ci

    ierr = 0

    ! IMS: set up the extensions expected in the eventlist file, plus the columns expected for each extension.
    !
    call eventlist%read_ccd_nrs(imageset, ccdnrs, ierr)
    call eventlist%init(ccdnrs, ierr)
    deallocate(ccdnrs)

    ! IMS: we only want to read the following subset of the EVENT-extension columns. Note that only X and Y columns seem directly to be used in ermldet, the others are needed by shapelib:slib_evttool_2_myevent().
    !
    call eventlist%events%ibset_col_useage("TIME",  TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("RA",    TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("DEC",   TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("PI",    TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("RAWX",  TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("RAWY",  TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("TM_NR", TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("X",     TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("Y",     TCOL_READ_BIT)
    eventlist%events%useageIsSet = .true.

    ! IMS: set up to read all CORRATT columns.
    !
    corratt_i = eventlist%get_ext_index("ext", extCORRATT)
    do kk=1,size(eventlist%ccdnr_infile)
      ci = eventlist%ccdnr_infile(kk) ! TM number
      call eventlist%extList(corratt_i)%extPerTM(ci)%ibset_all_useage(TCOL_READ_BIT) ! read all cols.
      eventlist%extList(corratt_i)%extPerTM(ci)%useageIsSet = .true.
    end do
    eventlist%extList(corratt_i)%useageFlags&
      = ibset(eventlist%extList(corratt_i)%useageFlags, TCOL_READ_BIT)

  end subroutine setup_evlist
  !--------------------------------------------------------------------------------




  !--------------------------------------------------------------------------------
  subroutine  ape_read_images(par, image,eventlist,pixlist,expimage,bkgimage,srcimage,detmask,ierr)
  !--------------------------------------------------------------------------------
  ! ! this essentially a copy of the ermldet_in in ermldet_mod
  !  - Sort accordning to column 'like' in descending order
  ! Read fits images:
  ! If par%compress_flag == true, only relevant section of fits images 
  ! as specified in detection mask header, is applied:
  !   - use fget_<type>compress for reading
  !   - set global variables ximbuf,yimbuf, ixb0, iyb0
  ! 
  !--------------------------------------------------------------------------------

    use detect_io
    
    implicit none

    type(ermldet_param_type), intent(in)            :: par
    type(i4cube_type)                               :: image
    type(tEventFile), dimension(:)                  :: eventlist
    type(pixlist_type), dimension(:)                :: pixlist
    type(l1cube_type)                               :: detmask
    type(r4cube_type)                               :: expimage
    type(r4cube_type)                               :: bkgimage
    type(r4cube_type)                               :: srcimage
    integer(kind=int32), intent(out)                :: ierr
    integer(kind=int32)                             :: status,ext
    integer(kind=int32)                             :: i,j,ipoint,ord=1,nmask
    !integer(kind=int32),dimension(:),allocatable    :: ix, iy
    type(dethdr_type),dimension(max_exp)            :: mask_dethdr
    type(stdhdr_type),dimension(max_exp)            :: mask_stdhdr
    type(stdhdr_type),dimension(max_ima)            :: image_stdhdr
    character(len=*), parameter                     :: rtname='READ_IMAGES'
    character(len=80)                               :: errstr
    !character(len=120) :: filename

    real(REAL64), dimension(:), pointer :: ptr_x=>null(),ptr_y=>null()

    status=0
    ierr=0
    ext=0


    ! check input images
    nim=par%n_imagesets

    compress_flag=par%compress_flag

    if (par%expima_flag .and. par%n_expimagesets /= nim) then
      errstr = 'Number of exposure maps and images differ'
      call WRERR(rtname,ierr,errstr)
    endif


    ! get image std headers:
    if (ierr < 3) then
      do i=1,nim
         !write(*,*) i, par%imagesets(i)
         call fget_stdhdr(par%imagesets(i),image_stdhdr(i),status)   
      enddo
    endif

    nmask=0
    if (par%detmask_flag) then
!     ximbuf=0
!     yimbuf=0
      nmask=par%n_detmasksets
      do i=1,par%n_detmasksets
         call fget_dethdr(par%detmasksets(i),mask_dethdr(i),status)
         call fget_stdhdr(par%detmasksets(i),mask_stdhdr(i),status)
!         ximbuf=max(ximbuf,mask_dethdr(i)%xdatamax-mask_dethdr(i)%xdatamin+1)
!         yimbuf=max(yimbuf,mask_dethdr(i)%ydatamax-mask_dethdr(i)%ydatamin+1)
      enddo

      if (compress_flag) then
        ximbuf=0
        yimbuf=0
        do i=1,par%n_detmasksets
          ximbuf=max(ximbuf,mask_dethdr(i)%xdatamax-mask_dethdr(i)%xdatamin+1)
          yimbuf=max(yimbuf,mask_dethdr(i)%ydatamax-mask_dethdr(i)%ydatamin+1)
        enddo
      endif
    else
      if (compress_flag) then
        compress_flag=.false.
        ierr = 1
        errstr = 'Need detection masks for compressed reading'
        call WRERR(rtname,ierr,errstr) 
      endif
    endif
!    endif


 !  create index arrays IP and EXPIP   
    call sort_ima(image_stdhdr,mask_stdhdr,nim,nmask,status)

! read images:
    if (ierr < 3) then
      do i=1,num_exp

        ixb0(i)=0
        iyb0(i)=0

        !write(*,*) i, num_exp, par%detmask_flag
        !write(*,*) num_image(i)
        
        if (par%detmask_flag) then
           !write(*,*)"fget_l1compress(par%detmasksets(i)", i
           !write(*,*) par%detmasksets(i)
           call fget_l1compress(par%detmasksets(i),num_exp,ximbuf,yimbuf,mask_dethdr(i),compress_flag,detmask,ord,i,status) 

          if (compress_flag) then
             ixb0(i)=mask_dethdr(i)%xdatamin-1
             iyb0(i)=mask_dethdr(i)%ydatamin-1
             ! set borders to .false.  to avoid problems with ixb/iyb mechanism:
             detmask%arr(i,1,:)=.false.
             detmask%arr(i,ximbuf,:)=.false.
             detmask%arr(i,:,1)=.false.
             detmask%arr(i,:,yimbuf)=.false.
          else
          ! Make sure data boundaries are present in any case:
            detmask%dethdr(i)%xdatamin = 1
            detmask%dethdr(i)%xdatamax =  detmask%shapehdr(i)%naxis1
            detmask%dethdr(i)%ydatamin = 1
            detmask%dethdr(i)%ydatamax =  detmask%shapehdr(i)%naxis2
          endif
       endif


        do j=1,num_image(i)

           !write(*,*) j

          ipoint=ip(i,j)
            ixb0(i)=mask_dethdr(i)%xdatamin-1
            iyb0(i)=mask_dethdr(i)%ydatamin-1
            !write(*,*)"fget_i4compress", ipoint,i,j, par%imagesets(ipoint)
            call fget_i4compress(par%imagesets(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i), &
                                 compress_flag,image,ord,ipoint,status)
            if (.not. compress_flag) then
              ximbuf =  image%shapehdr(1)%naxis1
              yimbuf =  image%shapehdr(1)%naxis2 
            endif
            if (status > 0) then
              ierr=3
              errstr = 'Cannot read science image'
              call WRERR(rtname,ierr,errstr) 
            endif

            ! this is exposure map
            if (par%expima_flag) then
              !write(*,*)"fget_r4compress", ipoint,i,j, par%expimagesets(ipoint)
              call fget_r4compress(par%expimagesets(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i), &
                                   compress_flag,expimage,ord,ipoint,status)

              if (status > 0) then
                ierr=3
                errstr = 'Cannot read exposure map'
                call WRERR(rtname,ierr,errstr) 
 
             else if (.not. (match_wcs(expimage%wcshdr(ipoint),image%wcshdr(ipoint))  &
                       .and. match_size(expimage%shapehdr(ipoint),image%shapehdr(ipoint)))) then
                ierr=3
                errstr = 'Mismatch of image  and exposure map '
                call WRERR(rtname,ierr,errstr)   
              endif 
           endif

           ! this is background map
           !write(*,*)"fget_r4compress", ipoint,i,j, par%bkgimagesets(ipoint)
            call fget_r4compress(par%bkgimagesets(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i), &
                                 compress_flag,bkgimage,ord,ipoint,status)
            if (status > 0) then
              ierr=3
              errstr = 'Cannot read background map'
              call WRERR(rtname,ierr,errstr)  
            else if (.not. (match_wcs(bkgimage%wcshdr(ipoint),image%wcshdr(ipoint))  &
                     .and. match_size(bkgimage%shapehdr(ipoint),image%shapehdr(ipoint)))) then
              ierr=3
              errstr = 'Mismatch of image and background map'
              call WRERR(rtname,ierr,errstr)   
           endif

           ! this is source map
           !write(*,*)"fget_r4compress", ipoint,i,j, par%srcimagesets(ipoint)
           call fget_r4compress(par%srcimagesets(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i), &
                                 compress_flag,srcimage,ord,ipoint,status)
            if (status > 0) then
              ierr=3
              errstr = 'Cannot read background map'
              call WRERR(rtname,ierr,errstr)  
            else if (.not. (match_wcs(srcimage%wcshdr(ipoint),image%wcshdr(ipoint))  &
                     .and. match_size(srcimage%shapehdr(ipoint),image%shapehdr(ipoint)))) then
              ierr=3
              errstr = 'Mismatch of image and source map'
              call WRERR(rtname,ierr,errstr)   
           endif


           

!          else
!            ierr=3
!            errstr = 'Mismatch of image and detector mask'
!            call WRERR(rtname,ierr,errstr)
!          endif
        enddo
     enddo

    ! write(*,*) "END",size(image%wcshdr), nim, par%shapelet_flag

  endif

!---------------------------------------------------------------------------------------------------- 

! read events and attitude extensions, store in erolib structure:
  
    if (par%shapelet_flag) then
       !write(*,*)image%wcshdr(nim)
       
! 20210426 IMS: this block looks identical to one in ermldet_mod:ermldet_in().

       do i=1,nim
        call setup_evlist(par%imagesets(i), eventlist(i), status)

        ! IMS: now we can open the eventlist file and read the required data.
        !
        call eventlist(i)%open(READONLY, status, filename=par%imagesets(i))
        call eventlist(i)%read(status)
        call eventlist(i)%close(status)

        call eventlist(i)%events%get_col_ptr("X", ptr_x)
        call eventlist(i)%events%get_col_ptr("Y", ptr_y)

        !write(*,*) "HERE"
        ! calcluate image pixels for events:
        ! where is that????: eroio/fitsiage_mod.f90
        ! according to that this is obsolete and ermldetr_mod:init_pixlist should be used...
        ! but I know init_pixlist does not work properly

        call skypix2ima(image%wcshdr(nim),int(ptr_x(:),kind=int32),int(ptr_y(:),kind=int32),pixlist(i),status)
        !call close_eventfile(eventlist(i), status)
        !write(*,*) 'skypix2ima: ',ierr, SIZE(pixlist(i)%pix), SIZE(eventlist(i)%events%x), SIZE(eventlist(i)%events%Y)
        !write(*,*) 'skypix2ima: ',image%wcshdr(nim)%crpix1p, image%wcshdr(nim)%crval1p,  image%wcshdr(nim)%cdelt1p
        !write(*,*) 
        !write(*,*) 'SRCIMAGE: ',srcimage%wcshdr(i)
        !write(*,*) 
        !write(*,*) 'EXPIMAGE: ',expimage%wcshdr(i)
        !write(*,*)

        nullify(ptr_x,ptr_y)
     enddo
     !write(*,*) "DONE"
    endif
    !write(*,*) "DONE"

  end subroutine ape_read_images











  !----------------------------------------------------------------------------
  subroutine apebkg(image, expimage, eefback, binning, ierr)
  !----------------------------------------------------------------------------
  ! TO DO: READ EXP MAP -> GENERATE EEFBACK OUTSIDE
  !        
    
    use fitsimage

    implicit none
    ! WHY DO I DEFINE IMAGE AS r4image_type and not i4image_type?
    type(r4image_type), intent(in) :: image
    type(r4image_type), intent(in) :: expimage
    type(r4image_type), intent(out) :: eefback
    integer(kind=int32), intent(in) :: binning
    integer(kind=int32), intent(out) :: ierr

    character(len=*),parameter        :: rtname='apebkg'

    real(kind=int32) dx, dy, dr2, rcut2
    integer(kind=int32) i, j, ix, iy, nx1, ny1, RESULT, ASUM, OFFSET
    integer(kind=int32) ixmin, ixmax, iymin, iymax

    ! this is the total area in pixels within the extraction region
    ! area-small is the area with positive exp time, i.e. to 
    ! account for edge effects, bad pixels etc. 
    integer(kind=int32) AREA, AREA_SMALL
   
    ierr=0
    eefback%arr=0.0
    rcut2 = 3.0 ! this will later on defined by psfmap
    OFFSET = 10

    RESULT = mod(binning,2)
    if(RESULT.eq.0)then 
       ierr=3
       call wrerr(rtname,ierr,'Binning should 1,3,5,7,9,11...., stopping.')
    endif

    nx1=int(binning/2)+RESULT     
    ny1=int(binning/2)+RESULT 


    do iy=ny1,image%shapehdr%naxis1,binning

         do ix=nx1,image%shapehdr%naxis2,binning         

            ASUM = 0
            AREA = 0
            AREA_SMALL = 0

            ixmin=ix-OFFSET
            if(ixmin.le.1)ixmin=1
            ixmax=ix+OFFSET
            if(ixmax.ge.image%shapehdr%naxis1)ixmax=image%shapehdr%naxis1
 
            iymin=iy-OFFSET
            if(iymin.le.1)iymin=1
            iymax=iy+OFFSET
            if(iymax.ge.image%shapehdr%naxis2)iymax=image%shapehdr%naxis2
 
            do i=ixmin, ixmax, 1               
               do j = iymin, iymax, 1
                  
                  dx = float(i - ix)
                  dy = float(j - iy)
                  dr2 = dx * dx + dy * dy
          
                  if(dr2<=rcut2)then                     
                     AREA = AREA + 1
                     
                     if(ixmin.ge.1 .and. ixmax.le.image%shapehdr%naxis1 &
                          .and. &
                          iymin.ge.1 .and. iymax.le.image%shapehdr%naxis2)then
                        if(expimage%arr(i,j) .gt. 0)then
                           AREA_SMALL = AREA_SMALL + 1
                           ASUM = ASUM + INT(image%arr(i,j))
                        endif 
                     endif 
                  endif

               enddo               
            enddo

            eefback%arr(ix,iy) = ASUM * float(AREA_SMALL)/float(AREA)             

            !fill up remaining pixels in case of binning
            if(binning.gt.1)then
               RESULT=mod(binning,2)
               ixmin=ix-(int(binning/2.)+RESULT)+1     
               iymax=ix+(int(binning/2.)+RESULT)-1     
               
               iymin=iy-(int(binning/2.)+RESULT)+1     
               iymax=iy+(int(binning/2.)+RESULT)-1     
               
               do i=iymin,iymax                     
                  do j=ixmin,ixmax                 
                     if(expimage%arr(i,j) .gt. 0.0)then
                        eefback%arr(i,j) = eefback%arr(ix,iy)
                     endif
                  enddo
               enddo
            endif
            !fill up remaining pixels in case of binning



  
         enddo

      enddo
            

    end subroutine apebkg





!----------------------------------------------------------------------------
  subroutine ape_get_geom(ic, instrume, date_obs, date_end,  &
                         xraw_ax, yraw_ax, xraw_scal, yraw_scal, ierr)
!----------------------------------------------------------------------------
! this is adapted from routine  get_detmaps of expmap_io_mod
! I only need the optical axis position  information not the full
! detmap
! may consider to place that routine in a separate directory

     implicit none


    integer, intent(in) :: ic                 ! telescope number
    character(len=30), intent(in) :: instrume 
    character(len=19), intent(in) :: date_obs, date_end
    real(real64), intent(out) :: xraw_ax, yraw_ax  
    real(real64), intent(out) :: xraw_scal, yraw_scal 
    integer, intent(out) :: ierr

    character(len=30), parameter :: telescop = "eROSITA"
    character(len=30), parameter :: filter = "-", detnam = "-"
    character(len=30) :: codenam

    integer, parameter :: maxret = 100 
    integer, dimension(maxret) :: extno

    character(len=256), dimension(maxret) :: filename , online
    character(len=256) :: expression
    character(len=10) :: strtdate, strtime, stpdate, stptime

    character(len=80) :: line
    !character(len=64) :: comment
    !character(len=5) :: cemin, cemax, cenerg

    integer :: status, unit, xscalcol, yscalcol, xaxcol, yaxcol
    integer :: nret, nfound
    real(real64)  :: nullval = 0.0
    logical :: undef

    integer :: chatter                      ! caldb verbosity

    character(len=*),parameter         :: rtname='ape_get_param'
   ! anyf = .TRUE.
    !UNDEF=.TRUE.
    ! get time for caldb access    
    call prep_gtcalf(date_obs, date_end, strtdate, strtime, &
                     stpdate, stptime, chatter)

    
! get optical axis position    
    codenam = "GEOM" 
    expression="-"
    call gtcalf (chatter, telescop, instrume, detnam, filter, &
                 codenam, strtdate, strtime, stpdate, stptime, &
                 expression, maxret, filename, extno, online, &
                 nret, nfound, status)       

    if ( status > 0 .or. nret < 1) then
      write (line, '(''Could not read GEOM data for telscope '', i1)') ic
      ierr = 3
      call wrerr(rtname, ierr, line)
      return
    end if

    if (nret > 1) then 
      call wrerr(rtname, 1, "More than one caldb entry found; using first")
    end if

    call ftgiou(unit, status)
    call fitserr("ftgiou", status, ierr)
    call fttopn(unit, filename(1), ftread, status)
    call fitserr("ftiopn", status, ierr)
    call ftgcno(unit, casesens, "X_PLATESCALE", xscalcol, status)
    call fitserr("ftgcno", status, ierr)
    call ftgcno(unit, casesens, "Y_PLATESCALE", yscalcol, status)
    call fitserr("ftgcno", status, ierr)
    call ftgcno(unit, casesens, "X_OPTAX", xaxcol, status)
    call fitserr("ftgcno", status, ierr)
    call ftgcno(unit, casesens, "Y_OPTAX", yaxcol, status)
    call fitserr("ftgcno", status, ierr)
    call ftgcvd(unit, xscalcol, 1, 1, 1, nullval, xraw_scal, undef, status)
    call fitserr("ftgcvd", status, ierr)
    xraw_scal=xraw_scal/3600d0
    call FTGCVD(unit, yscalcol, 1, 1, 1, nullval, yraw_scal, undef, status)
    call fitserr("ftgcvd", status, ierr)
    yraw_scal=yraw_scal/3600d0
    call ftgcvd(unit, xaxcol, 1, 1, 1, nullval, xraw_ax, undef, status)
    call fitserr("ftgcvd", status, ierr)
    call ftgcvd(unit, yaxcol, 1, 1, 1, nullval, yraw_ax, undef, status)
    call fitserr("ftgcvd", status, ierr)
    call ftclos(unit, status)       
    call fitserr("ftclos", status, ierr)

    !write(*,*) " xraw_ax=",xraw_ax, " yraw_ax", yraw_ax, " xraw_scal", xraw_scal, " yraw_scal", yraw_scal

  end subroutine ape_get_geom


!----------------------------------------------------------------------------
  subroutine ape_inter2d(X, Y, EEF, MyPSFSIZE, FXY)
!----------------------------------------------------------------------------
! https://en.wikipedia.org/wiki/Bilinear_interpolation
! It acctually does 3-d linear interpollatio X, Y, EEF 
! by doing two 2-d interpolations and then combining the 
! results
! IF EEF OUT OF RANGE IT PRODUCES JUNK. HAVE TO FIX THAT
    implicit none
    
    real(real32), intent(in) :: X, Y, EEF
    TYPE(PSFSIZE), intent(in) :: MyPSFSIZE
    real(kind=real32), intent(out) :: FXY

    INTEGER IX, IY
    real(kind=real32) Y11, Y12, Y21, Y22, FXY1, FXY2, FXY_Z1, FXY_Z2
    real(kind=real32) X1, X2, Y1, Y2, Z1, Z2

    INTEGER N, FLAG, NAXES(3)

    !write(*,*) X, Y, EEF
    
    NAXES = SHAPE(MyPSFSIZE%PSFMAP)
       
    
    IX = INT(X / MyPSFSIZE%DX + 1)
    if(IX .gt. (NAXES(1) - 1) )IX=NAXES(1)-1
    if(IX .le. 1) IX = 1
    IY = INT(Y / MyPSFSIZE%DY + 1)
    if(IY .gt. (NAXES(2) - 1) )IY=NAXES(2)-1
    if(IY .le. 1) IY = 1
    !write(*,*) "INTER COORDS", IX, IY
    

    N=0
    FLAG=0
    DO WHILE (FLAG==0 .AND. N<SIZE(MyPSFSIZE%EEF) )
       N=N+1
       if(EEF<MyPSFSIZE%EEF(N))FLAG=1
       !write(*,*) "INTER H",N, EEF, MyPSFSIZE%EEF(N), FLAG
    ENDDO
    !write(*,*) "INTER", N
    N=N-1
    IF(N<=0)N=1
    !EEF = MyPSFSIZE%EEF(N)
    !write(*,*) "H1", N, EEF, MyPSFSIZE%EEF(N)

    
    X1 = (IX-1) * MyPSFSIZE%DX
    X2 = IX * MyPSFSIZE%DX 
    Y1 = (IY-1) * MyPSFSIZE%DY
    Y2 = IY * MyPSFSIZE%DY 
    Z1 = MyPSFSIZE%EEF(N)
    Z2 = MyPSFSIZE%EEF(N+1)

   
    !write(*,*) X1, X, X2
    !write(*,*) Y1, Y, Y2
    !write(*,*) MyPSFSIZE%EEF(N), EEF, MyPSFSIZE%EEF(N+1)


    Y11 = MyPSFSIZE%PSFMAP(IX, IY, N)
    Y21 = MyPSFSIZE%PSFMAP(IX+1, IY, N)
    Y12 = MyPSFSIZE%PSFMAP(IX, IY+1, N)
    Y22 = MyPSFSIZE%PSFMAP(IX+1, IY+1, N)

    FXY1 = ( X2 - X ) / ( X2 - X1 ) * Y11 + ( X - X1 ) / ( X2 - X1) * Y21

    FXY2 = ( X2 - X ) / ( X2 - X1 ) * Y12 + ( X - X1 ) / ( X2 - X1) * Y22

    FXY_Z1 = ( Y2 - Y ) / ( Y2 - Y1 ) * FXY1 + ( Y - Y1 ) / ( Y2 - Y1 ) * FXY2

    !write(*,*) "INTER",FXY1, Y11, Y21
    !write(*,*) "INTER",FXY2, Y12, Y22
    
    Y11 = MyPSFSIZE%PSFMAP(IX, IY, N+1)
    Y21 = MyPSFSIZE%PSFMAP(IX+1, IY, N+1)
    Y12 = MyPSFSIZE%PSFMAP(IX, IY+1, N+1)
    Y22 = MyPSFSIZE%PSFMAP(IX+1, IY+1, N+1)



    FXY1 = ( X2 - X ) / ( X2 - X1 ) * Y11 + ( X - X1 ) / ( X2 - X1) * Y21

    FXY2 = ( X2 - X ) / ( X2 - X1 ) * Y12 + ( X - X1 ) / ( X2 - X1) * Y22

    !write(*,*) FXY1, Y11, Y21, ( X2 - X ), X - X1
    !write(*,*) FXY2, Y12, Y22

    FXY_Z2 = ( Y2 - Y ) / ( Y2 - Y1 ) * FXY1 + ( Y - Y1 ) / ( Y2 - Y1 ) * FXY2
    

    FXY = ( Z2 - EEF ) / ( Z2 - Z1 ) * FXY_Z1 + ( EEF - Z1 ) / ( Z2 - Z1 ) * FXY_Z2

    !write(*,*) Z1, EEF, Z2, FXY_Z1, FXY_Z2, FXY
    

  end subroutine ape_inter2d
!----------------------------------------------------------------------------







!----------------------------------------------------------------------------
  subroutine ape_nearest_srcs(IN_SRC, JIMA, APELIST, MyPSFSIZE, BOXLIST, OUT_SRC, NUM_OUT_SRC)
!----------------------------------------------------------------------------
! It looks for the nearest sources in the vicinity of the INSRC
! in the APELIST. Nearest means XX% EEF radius away. 
! At this stage the EEF distance is hardcored to 50%. In the future 
! a decision is needed how to imrove. 
! The IDs of the output sources are written in allocatable OUTSRC
! Remember to delallocate....
    use detect_io

    implicit none
    
    INTEGER, intent(in) :: IN_SRC
    INTEGER, intent(in) :: JIMA
    type(boxlist2d_type), intent(in)  :: APELIST
    TYPE(PSFSIZE), dimension(:), intent(in) :: MyPSFSIZE
     type(boxlist2d_type), intent(in)  :: BOXLIST
    INTEGER,  dimension (:), intent(out) :: OUT_SRC 
    INTEGER,  intent(out) ::  NUM_OUT_SRC

    real(kind=real32) xsrc, ysrc, FXY, EEF, rcut2, dr2, dx, dy, x, y
    INTEGER NSRC, ISRC, NSIZE

    OUT_SRC  = -9
    NSIZE = SIZE(OUT_SRC)
    NSRC=boxlist%shapehdr%naxis2 
    EEF = 0.5 ! look for nerarby sources within EEF RADIUS
    xsrc = apelist%data(JIMA,IN_SRC)%x_ima
    ysrc = apelist%data(JIMA,IN_SRC)%y_ima
    

    call ape_inter2d(xsrc, ysrc, EEF, MyPSFSIZE(JIMA), FXY)

    rcut2 = FXY ** 2
    
    !write(*,*) "START:", xsrc, ysrc, EEF, FXY, NSRC, NSIZE

    NUM_OUT_SRC = 0
    ISRC = 1
    do while (ISRC<=NSRC .and. NUM_OUT_SRC < NSIZE)
          x = boxlist%data(JIMA,ISRC)%x_ima
          y = boxlist%data(JIMA,ISRC)%y_ima
          
          dx = x - xsrc
          dy = y - ysrc
          dr2 = dx * dx + dy * dy
          
          if(dr2<rcut2)then
             NUM_OUT_SRC = NUM_OUT_SRC + 1 
             OUT_SRC(NUM_OUT_SRC) =ISRC

             !write(*,*) ISRC, x, y, sqrt(dr2), sqrt(rcut2), NUM_OUT_SRC
             
          endif

          ISRC = ISRC + 1
       end do
       
       
     end subroutine ape_nearest_srcs
!----------------------------------------------------------------------------





 !----------------------------------------------------------------------------
 subroutine ape_fput_apelist(filename,apelist,ierr)
 !----------------------------------------------------------------------------
 ! most of the following is adapted from the fput_* routines of detect_io
     USE detect_io, only : boxlist2d_type

    type(boxlist2d_type), intent(in)   ::  apelist
    character(len=*)                   :: filename
    integer(kind=int32), intent(out)   :: ierr

    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,irow
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    integer(kind=int32), parameter     :: ncol=13
    character(len=16), dimension(ncol) :: ttype,tunit
    character(len=4), dimension(ncol)  :: tform
    character(len=16)                  :: extname
    character(len=*),parameter         :: rtname = 'APE_FPUT_APELIST'
    character(len=80)                  :: errstr

    integer(kind=int32)                :: I, J


    !filename = par%APELISTOUT

    ttype=[ character(len=16) :: 'id_src', 'id_band',&
         'x_ima','y_ima', & 
         'ra','dec', &
         'APE_CTS', 'APE_BKG','APE_EXP', & 
         'APE_POIS', 'APE_EEF', 'APE_RADIUS', 'APE_OVER']
    tform=[ character(len=4) :: '1J  ','1J  ', &
         '1E  ','1E  ',&
         '1E  ','1E  '&
         ,'1J  ','1E  ', '1E  '&
         ,'1E  ','1E  ','1E  ','1J ']

    tunit=[ character(len=16) :: '','',&
         'pixel','pixel',&
         '','',&
         'counts','counts','',&
         '','','pixel','']    
    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=apelist%shapehdr%naxis1 * apelist%shapehdr%naxis2

    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    call FTINIT(unit,"!"//trim(filename),blocksize,status)
    fstat=max(fstat,status)
    if (fstat /= 0) then    
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else    
      status=0
      simple   = .true.
      bitpix   = 16
      naxis      = 0
      naxes(1) = 0
      !write FITS standard header:   
      call FTPHPS(unit,bitpix,naxis,naxes,status) 
      fstat=max(fstat,status)  
      ! write eROSITA standard header to primary extension :
      ! now obsolete?? (29/04/2015)
      call fwrite_stdhdr(unit,apelist%stdhdr,status)

      ! create new empty HDU:
      call FTCRHD(unit,status)
      fstat=max(fstat,status)  

      ! write header for binary table:
      call FTPHBN(unit,nrows,ncol,ttype,tform,tunit,extname,varidat,status)
      fstat=max(fstat,status)  
      ! put standard header into table extension (29/04/2015): 
      call fwrite_stdhdr(unit,apelist%stdhdr,status)
      fstat=max(fstat,status)  

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
         
         irow=0
         do I = 1, apelist%shapehdr%naxis2
            do J = 1, apelist%shapehdr%naxis1 
               irow=irow+1
    
               !write(*,*) apelist%data(J,I)%ra

               !call ftpcls(unit,colnum,frow,felem,nrows,name,status)
               call  FTPCLJ(unit,1,irow,1,1,apelist%data(J,I)%id_src,status)
               fstat = max(fstat,status)
               call  FTPCLJ(unit,2,irow,1,1,apelist%data(J,I)%id_band,status)
               fstat = max(fstat,status)

               call  FTPCLE(unit,3,irow,1,1,apelist%data(J,I)%x_ima,status)
               fstat = max(fstat,status)
               call  FTPCLE(unit,4,irow,1,1,apelist%data(J,I)%y_ima,status)
               fstat = max(fstat,status)

               call  FTPCLD(unit,5,irow,1,1,apelist%data(J,I)%ra,status)
               fstat = max(fstat,status)
               call  FTPCLD(unit,6,irow,1,1,apelist%data(J,I)%dec,status)
               fstat = max(fstat,status)

               call  FTPCLE(unit,7,irow,1,1,apelist%data(J,I)%scts,status)
               fstat = max(fstat,status)
!               call  FTPCLJ(unit,7,irow,1,1,apelist%data(J,I)%ape_cts,status)
!               fstat = max(fstat,status)

               call  FTPCLE(unit,8,irow,1,1,apelist%data(J,I)%bg_map,status)
               fstat = max(fstat,status)
!              call  FTPCLE(unit,8,irow,1,1,apelist%data(J,I)%ape_bkg,status)
!              fstat = max(fstat,status)

               call  FTPCLE(unit,9,irow,1,1,apelist%data(J,I)%exp_map,status)
               fstat = max(fstat,status)
!               call  FTPCLE(unit,9,irow,1,1,apelist%data(J,I)%ape_exp,status)
!               fstat = max(fstat,status)
              
               call  FTPCLE(unit,10,irow,1,1,apelist%data(J,I)%like,status)
               fstat = max(fstat,status)
!               call  FTPCLE(unit,10,irow,1,1,apelist%data(J,I)%ape_pois,status)
!               fstat = max(fstat,status)


               call  FTPCLE(unit,11,irow,1,1,apelist%data(J,I)%eef,status)
               fstat = max(fstat,status)              
!               call  FTPCLE(unit,11,irow,1,1,apelist%data(J,I)%ape_eef,status)
!               fstat = max(fstat,status)              

               call  FTPCLE(unit,12,irow,1,1,apelist%data(J,I)%cutrad,status)
               fstat = max(fstat,status)
!               call  FTPCLE(unit,12,irow,1,1,apelist%data(J,I)%ape_radius,status)
!               fstat = max(fstat,status)
              
!               call  FTPCLJ(unit,13,irow,1,1,apelist%data(J,I)%flag,status)
!               fstat = max(fstat,status)
            enddo
         enddo

      endif

      if (fstat /= 0) then
         ierr = 3
         errstr = 'Error writing values to table '//trim(filename)
         call WRERR(rtname,ierr,errstr)
      endif
    
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
         ierr = 3
         errstr = 'Error closing file '//filename
         call WRERR(rtname,ierr,errstr)
      endif
   endif ! 
   
 end subroutine ape_fput_apelist



subroutine read_psfmaps(par, MyPSFSIZE, mappsff)

  use detect_io

  implicit none

  type(ermldet_param_type), intent(in)  :: par

  TYPE(PSFSIZE), dimension(:), allocatable, intent(out) :: MyPSFSIZE

  character(len=*), dimension(:) :: mappsff

  integer(kind=int32) I

  ! I assume that PSFFILE are 
   if (allocated(MyPSFSIZE)) deallocate(MyPSFSIZE)
   allocate (MyPSFSIZE( par%n_imagesets ))
   do I=1,par%n_imagesets 
      !write(*,*) I, mappsff(I),SIZE(imagef)
      call ape_psfsize_read(mappsff(I), MyPSFSIZE(I))
   enddo

end subroutine read_psfmaps

!--------------------------------------------------------------------------------

  subroutine copy_r4cube(old_cube,new_cube,ierr)

  !---------------------------------------------------------------------------
  ! Duplicates the image <old_image> of type 'i4image_type' 
  ! to image <new_image>.  
  ! NOTE: The data array (new_image%arr) is allocated here.
  ! See also: copy_i1image, copy_r4image
  !---------------------------------------------------------------------------
    use detect_io

    implicit none
    
    type(r4cube_type)            :: old_cube
    type(r4cube_type)            :: new_cube
    integer(kind=int32)             :: ierr
    integer(kind=int32)             :: istat
    character(len=*),parameter      :: rtname = 'copy_r4cube'
    INTEGER, DIMENSION(3) :: A

    ierr = 0
    
    if (associated(new_cube%arr)) then
      deallocate(new_cube%arr)
      nullify(new_cube%arr)
   endif
   
    A = SHAPE(old_cube%arr)
    
    allocate(new_cube%arr(A(1), A(2), A(3)), STAT=istat)

    !write(*,*) SHAPE(old_cube%shapehdr)
    if( associated(new_cube%shapehdr) )deallocate(new_cube%shapehdr)
    allocate(new_cube%shapehdr(A(3)), STAT=istat)
    if( associated(new_cube%stdhdr) )deallocate(new_cube%stdhdr)
    allocate(new_cube%stdhdr(A(3)), STAT=istat)
    if( associated(new_cube%refhdr) )deallocate(new_cube%refhdr)
    allocate(new_cube%refhdr(A(3)), STAT=istat)
    if( associated(new_cube%wcshdr) )deallocate(new_cube%wcshdr)
    allocate(new_cube%wcshdr(A(3)), STAT=istat)
        

    if (istat == 0) then
      new_cube%arr       = old_cube%arr
      new_cube%shapehdr  = old_cube%shapehdr
      new_cube%stdhdr    = old_cube%stdhdr
      new_cube%refhdr    = old_cube%refhdr
      new_cube%wcshdr    = old_cube%wcshdr
    else 
      ierr=3
      call wrerr(rtname,ierr,'Cannot allocate memory')
    endif


  end subroutine copy_r4cube



  subroutine kill_r4cube(cube,ierr)

  !---------------------------------------------------------------------------
  ! Duplicates the image <old_image> of type 'i4image_type' 
  ! to image <new_image>.  
  ! NOTE: The data array (new_image%arr) is allocated here.
  ! See also: copy_i1image, copy_r4image
  !---------------------------------------------------------------------------
    use detect_io

    implicit none
    
    type(r4cube_type)            :: cube
    integer(kind=int32)             :: ierr
    character(len=*),parameter      :: rtname = 'kill_r4cube'

    ierr=0
    
    if (associated(cube%arr))deallocate(cube%arr)
    if( associated(cube%shapehdr) )deallocate(cube%shapehdr)
    if( associated(cube%stdhdr) )deallocate(cube%stdhdr)
    if( associated(cube%refhdr) )deallocate(cube%refhdr)
    if( associated(cube%wcshdr) )deallocate(cube%wcshdr)

  end subroutine kill_r4cube



  

! !----------------------------------------------------------------------------
  subroutine ape_sense(apar, expimage, srcimage, detmask, MyPSFSIZE)
! !----------------------------------------------------------------------------
   ! this is the soubroutine that estimates sensitivty maps as in
   ! Georgakakis+08. First it calculates background map within
   ! the EEF and then uses the inverse gammaq function to convert
   ! the background counts (within EEF) to minimum detection counts
   ! given a fixed Poisson falsape_sensee detection probability.
   use detect_io
   use ermldet_mod
   use apetool_inv

   implicit none
   type(APEPAR), intent(in) :: apar
   
   !character(len=256), dimension(:), intent(in) :: apesenseimagef
   type(r4cube_type), intent(in) :: expimage
   type(r4cube_type), intent(in) :: srcimage
   type(l1cube_type), intent(in) :: detmask

   !type(dethdr_type),dimension(max_exp)            :: mask_dethdr
 
   real(kind=real32) EEF_EXTRACT 
   integer(kind=int32) BINNING
   real(kind=real32) POISSON_THRESH   

   ! I assume that array structure MyPSFSIZE has been define outside 
   ! the subroutine and supplied to it 
   TYPE(PSFSIZE), dimension(:), intent(in) :: MyPSFSIZE

   type(r4cube_type)     :: eefexpimage
   type(r4cube_type)     :: eefbkgimage
   type(r4cube_type)     :: senseimage

   !type(r4image_type) :: tmpim
   
   integer(kind=int32) ierr, status, J, nx1, ny1, I

   real(kind=real32) xsrc, ysrc, FXY, lgp_thresh, sig_cell
   
   real(kind=real32) asum, exposure, rcut2
   integer(kind=int32) ix, iy, ix1, ix2, iy1, iy2, ixp, iyp, imodres

   !real(kind=real32), dimension (1000) :: area, lgcrate
   real(kind=real32) lgcrate0, lgcrate1, lgcrateBIN, expected
   real(kind=real32), dimension(:,:), allocatable :: area
   real(kind=real32), dimension(:), allocatable :: lgcrate
   real(kind=real64) :: mincounts

   ! write fits file keywords
   !character(len=80)  :: comment
   integer(kind=int32)    :: naxis,naxis1,naxis2,bitpix,pcount=0,gcount=1
   integer(kind=int32),dimension(2)    :: naxes
   integer(kind=int32)  :: nhdu,group,unit,blocksize
   !logical            :: anyf 
   logical :: simple=.true., extend=.true.  
   character(len=10) :: extname
   integer(kind=int32)    ::  colnum,frow,felem, varidata
   character(len=*),parameter      :: rtname = 'ape_sense'
   character*16 ttype(2),tform(2),tunit(2)

   data ttype/'LGCR','AREA'/
   data tform/'1E','1E'/
   data tunit/'rt','deg2'/
   extname = 'ACURVE'


   status = 0

   EEF_EXTRACT = apar%EEF_EXTRACT
   BINNING = apar%BINNING
   POISSON_THRESH = apar%POISSON_THRESH

   allocate (area(1000, apar%epar%n_imagesets))
   allocate (lgcrate(1000))
   area = 0.0
   lgcrate0 = -9.0
   lgcrate1 =  1.0
   lgcrateBIN = (lgcrate1 - lgcrate0) / 1000.0
   do I = 1, 1000
      lgcrate(I) = lgcrate0 + (I-1) * lgcrateBIN + lgcrateBIN/2.0
   end do

   call copy_r4cube(expimage,eefexpimage,status)
   call copy_r4cube(expimage,eefbkgimage,status)
   call copy_r4cube(expimage,senseimage,status)
   eefexpimage%arr = 0.0
   eefbkgimage%arr = 0.0
   senseimage%arr  = 0.0

   !write(*,*) detmask%arr(1000,1000,1), num_exp, apar%epar%n_detmasksets
   !stop
!   if (par%detmask_flag) then
!      write(*,*) num_exp, apar%epar%detmasksets(1), apar%epar%compress_flag
!      compress_flag=apar%epar%compress_flag
!   call fget_l1compress(apar%epar%detmasksets(i),num_exp,ximbuf,yimbuf,mask_dethdr(i),compress_flag,detmask,ord,i,status) 
!   if (compress_flag) then
!      ixb0(i)=mask_dethdr(i)%xdatamin-1
!      iyb0(i)=mask_dethdr(i)%ydatamin-1
!      ! set borders to .false.  to avoid problems with ixb/iyb mechanism:
!      detmask%arr(i,1,:)=.false.
!      detmask%arr(i,ximbuf,:)=.false.
!      detmask%arr(i,:,1)=.false.
!      detmask%arr(i,:,yimbuf)=.false.
!   else
!      ! Make sure data boundaries are present in any case:
!      detmask%dethdr(i)%xdatamin = 1
!      detmask%dethdr(i)%xdatamax =  detmask%shapehdr(i)%naxis1
!      detmask%dethdr(i)%ydatamin = 1
!      detmask%dethdr(i)%ydatamax =  detmask%shapehdr(i)%naxis2
!   endif
!endif  

   lgp_thresh = log(POISSON_THRESH)
   
   ! changed here to imodred because was generatingh warnings
   imodres= MOD(int(BINNING),2)
   nx1=int(BINNING/2.)+ imodres
   ny1=int(BINNING/2.)+ imodres

   senseimage%arr = 0.0
   
   do J = 1, apar%epar%n_imagesets ! start loop in energy
      
      do ix=nx1,expimage%shapehdr(J)%naxis1,int(BINNING)         
         do iy=ny1, expimage%shapehdr(J)%naxis2,int(BINNING)

            if(detmask%arr(ix, iy,1))then
            
               xsrc=real(ix)
               ysrc=real(iy)
         
               call ape_inter2d(xsrc, ysrc, EEF_EXTRACT, MyPSFSIZE(J), FXY)
               
               !write(*,*) "0",ix, iy, J, FXY, lgp_thresh, EEF_EXTRACT, POISSON_THRESH
               rcut2 = FXY ** 2
               
               call extract_single_exp(expimage, J, xsrc, ysrc, rcut2, exposure) 
               call extract_single_r4(srcimage, J, xsrc, ysrc, rcut2, asum)

               !write(*,*) "1",asum, exposure, expimage%wcshdr(J)%cdelt1, expimage%wcshdr(J)%cdelt2
               
               eefbkgimage%arr( ix, iy, J) = asum
               eefexpimage%arr( ix, iy, J) = exposure
               
               ! this is the old inversion of the poisson
               ! estimate min mumber of couts that
               ! exceed the POISSON_THRESH
               !sig_cell=1.0
               !nc=0
               !do while (sig_cell.ge.lgp_thresh)
               !   nc=nc+1
               !   sig_cell = -gammp(real(nc),asum)
               !   !write(*,*) "2",sig_cell, nc, asum, lgp_thresh
               !   !write(*,*) "3",-gammp(real(156271924.),asum)
               !end do
               ! this is the new inversion of the poisson
               call  poisscinv(DBLE(POISSON_THRESH), DBLE(asum), MINCOUNTS)
               !write(*,*) MINCOUNTS, NC
               senseimage%arr( ix, iy, J) = real(MINCOUNTS) + 1.E0

               !write(*,*) ix, iy, J, senseimage%arr( ix, iy, J), eefbkgimage%arr( ix, iy, J), eefexpimage%arr( ix, iy, J)
               !fill up remaining pixels in case of BINNING
               if(BINNING.gt.1)then
                  ix1=ix-(int(BINNING/2.)+imodres)+1     
                  ix2=ix+(int(BINNING/2.)+imodres)-1     
                  if(ix1.lt.1)ix1=1
                  if(ix2.gt.expimage%shapehdr(J)%naxis1)ix2=expimage%shapehdr(J)%naxis1
                  
                  iy1=iy-(int(BINNING/2.)+imodres)+1     
                  iy2=iy+(int(BINNING/2.)+imodres)-1
                  if(iy1.lt.1)iy1=1
                  if(iy2.gt.expimage%shapehdr(J)%naxis2)iy2=expimage%shapehdr(J)%naxis2
                  
                  
                  !write(*,*) ix, iy, ix1, ix2, iy1, iy2
                  do ixp=ix1,ix2                 
                     do iyp=iy1,iy2                     
                        if(expimage%arr(ixp,iyp,J).gt.0.0)then
                           eefbkgimage%arr( ixp, iyp, J) = asum
                           eefexpimage%arr( ixp, iyp, J) = exposure
                           senseimage%arr( ixp, iyp, J) =  senseimage%arr( ix, iy, J)
                           
                           !write(*,*) "E:",ix, iy, ixp, iyp, senseimage%arr( ixp, iyp, J),  senseimage%arr( ix, iy, J)!
                           
                        endif
                     enddo
                  enddo
               endif
               !fill up remaining pixels in case of BINNING

               do I=1,1000
                  expected=10**(lgcrate(I)) * EEF_EXTRACT * exposure + asum
                  
                  ! gammq return the negative ln
                  sig_cell = -gammq(senseimage%arr( ix, iy, J), expected)
                  
                  area(I, J)  =  area(I, J) + 1.0 - exp(sig_cell)
                  !if(J==2)then
                  !   write(*,*) "B=",J,"X,Y=",xsrc, ysrc, lgcrate(I), EEF_EXTRACT, "R=",senseimage%arr( ix, iy, J),"E=",exposure,"B=",asum,expected,1.0 - exp(sig_cell)
                  !endif
               enddo
               
               
            endif
         enddo
      enddo

      area(:,J) = area(:,J) * BINNING * BINNING  * REAL(abs(expimage%wcshdr(J)%cdelt1)) * REAL(abs(expimage%wcshdr(J)%cdelt2))

      !do I=1,1000
      !   write(*,*) I, lgcrate(I), area(I, J), J
      !enddo
   enddo


   ! writing results 
   status=0
   !   use default blocksize:
    blocksize=0
    ierr=0   
    nhdu=1 
    group=0
    bitpix=-32  
    naxis=2
 
   do J = 1, apar%epar%n_imagesets ! start loop in energy

      call ftgiou(unit, status)
      call ftinit(unit,"!"//apar%sensemapsets(J),blocksize,status)
      if (status > 0) then
         ierr=3
         call wrerr(rtname,ierr,'Error creating FITS file '//trim(apar%sensemapsets(J)))
      else
         naxes(1)=senseimage%shapehdr(J)%naxis1
         naxes(2)=senseimage%shapehdr(J)%naxis2
         naxis1=naxes(1)
         naxis2=naxes(2)         
         call FTPHPR(unit,simple,bitpix,naxis,naxes,pcount,gcount,extend, status)
         if (status > 0) then
            ierr = 3
            call wrerr(rtname,ierr,'Error creating FITS  image extension')
         else
            status = 0

            call fwrite_stdhdr(unit,senseimage%stdhdr(J),status)
            if (status > 0) ierr = 3 
            !call fwrite_refhdr(unit,senseimage%wcshdr(J),status)
            !if (status > 0) ierr = 3 
            call fwrite_wcshdr(unit,senseimage%wcshdr(J),status)
            if (status > 0) then 
               ierr = 3 
               call wrerr(rtname,ierr,'Error writing wcshdr') 
            endif
            !call fwrite_dethdr(unit,image%dethdr,status)
            !if (status > 0) then 
            !  ierr = 3 
            !  call wrerr(rtname,ierr,'Error writing dethdr')
            !endif 

            
            status=0
            call ftp2de(unit,group,naxis1,naxis1,naxis2,senseimage%arr(:,:,J),status)
            if (status > 0) then
               ierr=3
               call wrerr(rtname,ierr,'Error writing FITS image extension')
            endif
         endif

         status=0
         naxis=2
         bitpix=-32  
         call FTIIMG(unit,bitpix,naxis,naxes, status)
         if (status > 0) then
            ierr = 3
            call wrerr(rtname,ierr,'Error creating FITS  image extension')
         else
            status = 0
            
            call fwrite_stdhdr(unit,senseimage%stdhdr(J),status)
            if (status > 0) ierr = 3 
            !call fwrite_refhdr(unit,senseimage%wcshdr(J),status)
            !if (status > 0) ierr = 3 
            call fwrite_wcshdr(unit,senseimage%wcshdr(J),status)
            if (status > 0) then 
               ierr = 3 
               call wrerr(rtname,ierr,'Error writing wcshdr') 
            endif
            !call fwrite_dethdr(unit,image%dethdr,status)
            !if (status > 0) then 
            !  ierr = 3 
            !  call wrerr(rtname,ierr,'Error writing dethdr')
            !endif 

            status=0
            call ftp2de(unit,group,naxis1,naxis1,naxis2,eefbkgimage%arr(:,:,J),status)
            if (status > 0) then
               ierr=3
               call wrerr(rtname,ierr,'Error writing FITS image extension')
            endif
         endif
            
      
         status=0
         naxis=2
         bitpix=-32  
         call FTIIMG(unit,bitpix,naxis,naxes, status)
         if (status > 0) then
            ierr = 3
            call wrerr(rtname,ierr,'Error creating FITS  image extension')
         else
            status = 0
         
            call fwrite_stdhdr(unit,senseimage%stdhdr(J),status)
            if (status > 0) ierr = 3 
            !call fwrite_refhdr(unit,senseimage%wcshdr(J),status)
            !if (status > 0) ierr = 3 
            call fwrite_wcshdr(unit,senseimage%wcshdr(J),status)
            if (status > 0) then 
               ierr = 3 
               call wrerr(rtname,ierr,'Error writing wcshdr') 
            endif
            !call fwrite_dethdr(unit,image%dethdr,status)
            !if (status > 0) then 
            !  ierr = 3 
            !  call wrerr(rtname,ierr,'Error writing dethdr')
            !endif 

            status=0
            call ftp2de(unit,group,naxis1,naxis1,naxis2,eefexpimage%arr(:,:,J),status)
            if (status > 0) then
               ierr=3
               call wrerr(rtname,ierr,'Error writing FITS image extension')
            endif

         endif


         !Append/create a new empty HDU onto the end of the file and move to it.
         call ftcrhd(unit,status)
         if (status > 0) then
            ierr = 3
            call wrerr(rtname,ierr,'Error creating FITS  table extension')
         else
            status = 0
 
            !ttype, tform, and tunit parameters give the name, datatype, and physical units of each column, and extname giv
            !ttype="LGCR"
            !tform="E"
            !tunit="RATE"
            !EXTNAME="AREA"
            varidata =0
            !write(*,*) EEF, MyPSFSIZE%EEF
            call FTIBIN(unit,SIZE(lgcrate),2,ttype,tform,tunit,extname,varidata, status)         
            frow=1
            felem=1
            colnum=1
            call ftpcle(unit,colnum,frow,felem,SIZE(lgcrate),lgcrate,status)
            colnum=2
            call ftpcle(unit,colnum,frow,felem,SIZE(area(:,J)),area(:,J),status)

         endif
            
         call ftclos(unit,status) 
         if (status > 0) then 
            ierr = 3
         endif
        
      endif

    enddo


    call kill_r4cube(senseimage,status)
    call kill_r4cube(eefbkgimage,status)
    call kill_r4cube(eefbkgimage,status)
    deallocate (area)
    deallocate (lgcrate)


 end subroutine ape_sense




 
 
! !----------------------------------------------------------------------------
 subroutine ape_erml(apar, mllist, eventlist, &
      image, pixlist, expimage, bkgimage, srcimage, detmask, &
      MyPSFSIZE)
! !----------------------------------------------------------------------------
!! this is the soubroutine that can be called directly erml to estimate src counts. !
   use detect_io
   use slib_constants
   use slib_data_psf
   use ermldet_mod

   implicit none

   type(APEPAR) , intent(in)  :: apar
   type(mllist_type), intent(in) :: mllist
   ! I assume that eventlist is passed to the subroutine
   type(tEventFile), dimension(:), intent(in)   :: eventlist
   ! I assume that array structure MyPSFSIZE has been define outside 
   ! the subroutine and supplied to it 
   TYPE(PSFSIZE), dimension(:), intent(in) :: MyPSFSIZE
   
   

   type(i4cube_type)     :: image
   type(pixlist_type), dimension(:) :: pixlist
   type(r4cube_type)     :: expimage
   type(r4cube_type)     :: bkgimage
   type(r4cube_type)     :: srcimage
   type(l1cube_type)     :: detmask

   real(kind=real32) EEF_EXTRACT 
   logical REMSRC_FLAG
   

   ! I assume allstackPSFs are already loaded
   !type(FPSF), intent(in) :: allStackPSF,allStackPSF2
   !type(FPSF), intent(in) :: cluPSF, cluPSF2
   type(FPSF) :: allStackPSF,allStackPSF2
   type(FPSF) :: cluPSF, cluPSF2


   integer(kind=int32),parameter              :: srcmodrad=120 
   type(boxlist2d_type)  :: boxlist
   real(kind=real32), dimension(:,:,:), allocatable :: srcimage_save

   character(len=*), parameter :: rtname='APE_ERML'
   
   ! this is not used at the moment, code that uses it is commented out below
   ! to be looked into in the future
   ! they are meant to be used if multiple sources are to be removed from srcmap, 
   ! within the extraction radius not just the one for which counts are to be estimated
   ! I think this was meant to tbe "stacking" functionality now in apetool_stack_mod.f90
   !INTEGER,  dimension (1000) :: OUT_SRC 
   !INTEGER ISRC_REMOVE, NUM_OUT_SRC
   INTEGER naxis3(3), MAX_ENERGY, IE
   real(kind=real32) xsrc, ysrc, rcut2, asum, FXY, exposure
   integer(kind=int32) isum, I, J, iim, ierr, status


   !allocate(pixlist(par%n_imagesets))
   !call ape_read_images(par, image,eventlist,pixlist,expimage,bkgimage,srcimage,detmask,ierr)        
   
   EEF_EXTRACT = apar%EEF_EXTRACT
   REMSRC_FLAG = apar%REMSRC_FLAG
   ierr=0
   status=0
   ! if the option of removing nearby sources from the source image
   ! then I need to generate a running srcimage which will 
   ! store the temporary source map with the source(s) in question 
   ! removed. This map will be used to estimate the background
   ! counts at given position
   !if(REMSRC_FLAG)then
   naxis3 = SHAPE(srcimage%arr)
   nim=apar%epar%n_imagesets
   allocate(    srcimage_save( naxis3(1), naxis3(2), naxis3(3) ) )
   srcimage_save = srcimage%arr

   ! pass mllist to boxlist
   ! this modifies the nim global parameter to nim = maxval(mllist%data(:)%id_band)
    ! ie to the maximum number of bands in the mllist.
    ! this is not any more the number of input images
   !call ape_ml2box(apar%epar, mllist, boxlist, ierr)
   call ape_ml2box(mllist, boxlist, ierr)
   !write(*,*) 'SHAPE OUT:', shape(boxlist)
   !stop
    nim=apar%epar%n_imagesets 
    !if (status > 0) then
    !   ierr=3
    !   call  wrerr(rtname,status,'Cannot read mllist')
    !endif
    
    !write(*,*) ierr
    !stop
    ! Continue if ierr from ape_ml2box is < 3 
    if (ierr < 3) then

       if  (apar%epar%shapelet_flag) then
          !write(*,*) 'CALL ermldet_shapelets'
          !boxlist%data(0,1)%ra=142.28587341308594
          !boxlist%data(0,1)%dec=-2.7640185356140137
          !boxlist%data(0,2)%ra=142.28816223144531 
          !boxlist%data(0,2)%dec=-2.7518634796142578
          call ermldet_shapelets(apar%epar,boxlist,image,eventlist,allStackPSF,0,cluPSF,status)
          call ermldet_shapelets(apar%epar,boxlist,image,eventlist,allStackPSF2,srcmodrad,cluPSF2,status)
          !write(*,*) 'DONE ermldet_shapelets'
          ! this does not work because some wcshdr keywrds are not set

          
          do iim=1,apar%epar%n_imagesets
              call init_pixlist(eventlist(iim)%events,image%wcshdr(iim),pixlist(iim),status)
           enddo
       endif
       if (status > 0) then
          ierr=3
          call  wrerr(rtname,status,'Error preparing shapelet PSF')
       endif


! AGE TESTING PRINTING OUT PSF COEFFS
!   do iim=1,SIZE(allStackPSF%mySPSF)
!      
!     do i=1,allStackPSF%mySPSF(iim)%NUM_SRC
!      do j=1,SIZE(allStackPSF%mySPSF(iim)%NX0)
!         write(*,*) "ISRC:",i, "IBAND:",iim, "(NX0,NY0):",allStackPSF%mySPSF(iim)%NX0(j),allStackPSF%mySPSF(iim)%NY0(j),"C0:",allStackPSF%mySPSF(iim)%C0(j,i), allStackPSF2%mySPSF(iim)%C0(j,i)
!      enddo
       !      do j=1,SIZE(allStackPSF%mySPSF(iim)%NX1)
!         write(*,*) "ISRC:",i, "IBAND:",iim, "(NX1,NY1):",allStackPSF%mySPSF(iim)%NX1(j),allStackPSF%mySPSF(iim)%NY1(j),"C1:",allStackPSF%mySPSF(iim)%C1(j,i),allStackPSF2%mySPSF(iim)%C1(j,i)
!      enddo
!      do j=1,SIZE(allStackPSF%mySPSF(iim)%NX2)
!         write(*,*) "ISRC:",i, "IBAND:",iim, "(NX2,NY2):",allStackPSF%mySPSF(iim)%NX2(j),allStackPSF%mySPSF(iim)%NY2(j),"C2:",allStackPSF%mySPSF(iim)%C2(j,i), allStackPSF2%mySPSF(iim)%C2(j,i)
!      enddo
!!!      do j=1,SIZE(allStackPSF(iim)%mySPSF%C1(:,i))
!!         write(*,*) i, j,allStackPSF(iim)%mySPSF%C1(j,i)
!!      enddo
!!       do j=1,SIZE(allStackPSF(iim)%mySPSF%C2(:,i))
!!         write(*,*) i, j,allStackPSF(iim)%mySPSF%C2(j,i)
!!      enddo
!   enddo
!enddo
!stop
       boxlist%data(:,:)%cluflag= .false.

       MAX_ENERGY = apar%n_energy_index
       if(apar%n_energy_index==0) MAX_ENERGY  = boxlist%shapehdr%naxis1 



       ! this outer loop was on energy. however arrays include energy     
       do I = 1, boxlist%shapehdr%naxis2 ! start loop in sources
          !do I = 412,413 ! start loop in sources
          
          !      if(boxlist%data(0,I)%id_src.eq.235)then
          !write(*,*) I, boxlist%data(0,I)%ra, boxlist%data(0,I)%dec,  boxlist%data(0,I)%id_src, apar%epar%multrad,apar%epar%cutrad 
          !always remove the source in question from the background map           
          
          srcimage%arr = srcimage_save

          call ape_srcmap_isrc(apar,image,eventlist,pixlist, &
               detmask,expimage,bkgimage,&
               allStackPSF,cluPSF, I, srcimage, &
               allStackPSF2,cluPSF2,boxlist,srcmodrad,ierr)


          !!           if(REMSRC_FLAG)then
          !!              call ape_nearest_srcs(I, J, BOXLIST, MyPSFSIZE, BOXLIST, OUT_SRC, NUM_OUT_SRC)
          !!              
          !!              srcimage%arr = srcimage_save
          !!              
          !!              do ISRC_REMOVE = 1, NUM_OUT_SRC
          !!                 
          !!                 write(*,*) OUT_SRC(ISRC_REMOVE), boxlist%data(J,OUT_SRC(ISRC_REMOVE))%x_ima, &
          !!                      boxlist%data(J,OUT_SRC(ISRC_REMOVE))%x_ima
          !!                 
          !!                 call ape_srcmap_isrc(par,image,eventlist,pixlist,detmask,expimage,bkgimage,&
          !!                      allStackPSF,cluPSF, OUT_SRC(ISRC_nREMOVE), srcimage, &
          !!                      allStackPSF2,cluPSF2,boxlist,srcmodrad,ierr)
          !!              enddo
          !!              
          !!              if(NUM_OUT_SRC .gt. 0)boxlist%data(J,I)%cluflag = .true.
          !!                            
          !!           endif
          
           
          do J = 1, MAX_ENERGY ! start loop in energy 
             
         
         
             xsrc = boxlist%data(J,I)%x_ima
             ysrc = boxlist%data(J,I)%y_ima
             
             call ape_inter2d(xsrc, ysrc, EEF_EXTRACT, MyPSFSIZE(J), FXY)
         
             !FXY = 45.0
             rcut2 = FXY ** 2
         
             call extract_single_i4(image, J, xsrc, ysrc, rcut2, isum)           
             call extract_single_exp(expimage, J, xsrc, ysrc, rcut2, exposure) 
             call extract_single_r4(srcimage, J, xsrc, ysrc, rcut2, asum)
             !call extract_single_r4(bkgimage, J, xsrc, ysrc, rcut2, asum)
             
             if(apar%n_energy_index==0)then
                IE = J
             else
                IE = apar%energy_index(J)
             endif
             
             boxlist%data(IE,I)%ape_cts  = isum
             boxlist%data(IE,I)%ape_bkg = asum
             boxlist%data(IE,I)%ape_exp = exposure 
             boxlist%data(IE,I)%ape_eef = EEF_EXTRACT
             boxlist%data(IE,I)%ape_radius = FXY
             if(isum.gt.0.and.asum.gt.0)then
                boxlist%data(IE,I)%ape_pois =  exp(-gammp(real(isum),asum))
             else
                boxlist%data(IE,I)%ape_pois = -9.99
             endif
         
!             write(*,*) "HERE...",IE, I, boxlist%data(IE,I)%ape_cts, isum, xsrc, ysrc
!             write(*,*) IE, I, boxlist%data(IE,I)%ape_bkg 
!             write(*,*) IE, I, boxlist%data(IE,I)%ape_exp
!             write(*,*) IE, I, boxlist%data(IE,I)%ape_eef
!             write(*,*) IE, I, boxlist%data(IE,I)%ape_radius,  boxlist%data(IE,I)%ape_pois
             
          enddo ! end loop in energy
          
          !   endif
      
       enddo ! end loop in number of sources
   
       !write(*,*) SHAPE(boxlist%data)

       call wr_ape2mllist(apar,boxlist%shapehdr%naxis2,boxlist,mllist,ierr)
   
       !do I=1,SIZE(imagef)
       !   call ape_psfsize_kill(MyPSFSIZE(I)) 
       !enddo2
       !deallocate(MyPSFSIZE)
       
       if (allocated(srcimage_save))deallocate(srcimage_save)
       if  (apar%epar%shapelet_flag) then
          call slib_dealloc_FPSF(allStackPSF) 
          call slib_dealloc_FPSF(allStackPSF2) 
       endif
   
    endif  ! this loop is execute only if ierr from ape_ml2box is <3


 end subroutine ape_erml



  subroutine wr_ape2mllist(apar,nsrc,boxlist,mllist,ierr)
    use detect_io
    implicit none

    type(APEPAR) , intent(in)  :: apar
    !type(ermldet_param_type) , intent(in)  :: par
    integer(kind=int32) :: nsrc
    type(boxlist2d_type) :: boxlist
    type(mllist_type), intent(in) :: mllist    
    integer(kind=int32),intent(out)            :: ierr
    logical(kind=log32),dimension(max_cam)     :: instflag
    integer(kind=int32),dimension(max_cam)     :: nima_inst 
    
    logical :: file_exists

    integer(kind=int32) :: iima,isrc,iexp,iinst,ipoint,irow,status, nimlocal,firstrow
    character(len=*),parameter :: rtname = 'WR_APE2MLLIST'

    !type(srcrow_type) :: plist
    type(srcrow_type),dimension(:,:,:), allocatable :: srcbuff 

    real(kind=real32)  :: minus1,r32null
    integer(kind=int32) :: minus2,i32null


    ! this is the new, it simply finds the max value of the id_bands entry
    ! from the mllist.
    nimlocal = maxval(mllist%data(:)%id_band)
    allocate(srcbuff(max_cam,nim,nsrc))
    call init_mllist3d(srcbuff)
    
    minus1=-1
    r32null=sqrt(minus1)
    i32null=-huge(minus2)-1
    ierr=0
    instflag(1:max_cam)=.false.
    nima_inst(:)=0
    irow=0
    !write(*,*) max_cam,nim,nsrc
    !stop
    
    do isrc=1,nsrc

       do iexp=1,num_exp
          iinst=instip(iexp)
          instflag(iinst)=.true.
          do iima=1,num_image(iexp)
             ipoint = ip(iexp,iima)

             if(apar%n_energy_index.gt.0)ipoint=apar%energy_index(ipoint)             
             srcbuff(iinst,iima,isrc)  = boxlist%data(ipoint,isrc)
             !write(*,*) isrc,iima, ipoint
             !srcbuff(iinst,iima,isrc) = plist
             ! count maximum number of bands per instrument:
             nima_inst(iinst)=max(iima,nima_inst(iinst))
          enddo ! num_image
       enddo !num_exp

      ! copy to ermldet output list:
       

       irow = irow + 1
       !write(*,*) "HERE",irow
       mllist%data(irow)%ape_cts = i4null 
       mllist%data(irow)%ape_bkg = r32null !r4null
       mllist%data(irow)%ape_exp = r32null !r4null
       mllist%data(irow)%ape_eef = r32null !r4null
       mllist%data(irow)%ape_radius = r32null !r4null
       mllist%data(irow)%ape_pois = r32null !r4null

       do iinst=1,max_cam
          !write(*,*) "start",isrc, iinst, max_cam, instflag(iinst),nima_inst(iinst), nimlocal
          if (instflag(iinst)) then
             irow=irow+1
             mllist%data(irow)%ape_cts = i4null
             mllist%data(irow)%ape_bkg = r32null !r4null
             mllist%data(irow)%ape_exp = r32null !r4null
             mllist%data(irow)%ape_eef = r32null !r4null
             mllist%data(irow)%ape_radius = r32null !r4null
             mllist%data(irow)%ape_pois = r32null !r4null

            firstrow=irow
            do iima=1,nima_inst(iinst)
               !irow=irow+1
               if(apar%n_energy_index==0)then
                  irow=irow+1
               else
                  irow = firstrow + apar%energy_index(iima)                  
               endif               
               !write(*,*) "middle",isrc, irow, iima, iinst,srcbuff(iinst,iima,isrc)%ape_eef 

              mllist%data(irow)%ape_cts = srcbuff(iinst,iima,isrc)%ape_cts 
              mllist%data(irow)%ape_bkg = srcbuff(iinst,iima,isrc)%ape_bkg
              mllist%data(irow)%ape_exp = srcbuff(iinst,iima,isrc)%ape_exp
              mllist%data(irow)%ape_eef = srcbuff(iinst,iima,isrc)%ape_eef
              mllist%data(irow)%ape_radius = srcbuff(iinst,iima,isrc)%ape_radius
              mllist%data(irow)%ape_pois = srcbuff(iinst,iima,isrc)%ape_pois

              !write(*,*) iima, isrc, mllist%data(irow)%id_src, mllist%data(irow)%ape_cts, mllist%data(irow)%ra, mllist%data(irow)%dec, boxlist%data(iima,isrc)%ape_cts, boxlist%data(iima,isrc)%id_src 

              !mllist%data(irow)%ape_cts = boxlist%data(iima,isrc)%ape_cts 
              !mllist%data(irow)%ape_bkg = boxlist%data(iima,isrc)%ape_bkg
              !mllist%data(irow)%ape_exp = boxlist%data(iima,isrc)%ape_exp
              !mllist%data(irow)%ape_eef = boxlist%data(iima,isrc)%ape_eef
              !mllist%data(irow)%ape_radius = boxlist%data(iima,isrc)%ape_radius
              !mllist%data(irow)%ape_pois = boxlist%data(iima,isrc)%ape_pois
           enddo
           !write(*,*)maxval(apar%energy_index(1:apar%n_energy_index))
           !write(*,*) nima_inst(iinst)
           !write(*,*) nima_inst
           if(apar%n_energy_index>0)then
              do iima = maxval(apar%energy_index(1:apar%n_energy_index))+1, nimlocal
              !do iima=apar%energy_index(nima_inst(iinst))+1,nimlocal
                 irow = irow + 1
                 !write(*,*) "end",isrc, irow, nimlocal
              enddo
           endif
        endif
     enddo
     
  enddo 


  INQUIRE(FILE=apar%apelistout, EXIST=file_exists)

  !write(*,*)"HERE file exists", file_exists
  !write(*,*) apar%apelistout
  if(file_exists)then
     !write(*,*) "HERE NOT"
     !write(*,*) '!'//apar%apelistout
     call fput_mllist('!'//apar%apelistout,mllist,status)
  else
     !write(*,*) "HERE"
     call fput_mllist(apar%apelistout,mllist,status)
  endif

  deallocate(srcbuff)
        
 end subroutine wr_ape2mllist




end module apetool_io
