! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

program apetool

!------------------------------------------------------------------------------
! eROSITA APERTURE EXTRACTION TOOL.
! Author : AGeorgakakis  
! 
!
!
!------------------------------------------------------------------------------
  

  use param
  use ermldet_mod
  use numkind
  use fitsimage
  use detect_io
  use cal
  use erolib_eventfile_mod
  use slib_constants
  use apetool_io
  use apetool_stack
  
  implicit none
  

  integer(kind=int32)                           :: ierr,status
  type(APEPAR)                                  :: apar
  type(i4cube_type)                             :: image
  type(tEventFile), dimension(:), allocatable   :: eventlist
  type(pixlist_type), dimension(:), allocatable :: pixlist  

  type(l1cube_type)     :: detmask
  type(r4cube_type)     :: expimage
  type(r4cube_type)     :: bkgimage
  type(r4cube_type)     :: srcimage
  type(mllist_type)     :: mllist

  
  integer(kind=int32),parameter       :: srcmodrad=120 

  character(len=*), parameter :: rtname='APETOOL'

  TYPE(PSFSIZE), dimension(:), allocatable :: MyPSFSIZE

  integer(kind=int32) n_imagesets, n_expimagesets, n_psfmapsets, & 
       n_sensemapsets, n_detmasksets, n_bkgimagesets, &
       n_srcimagesets, n_emin, n_emax, n_energy_index
  character(len=256) mlsrclist
  character(len=256) apesrclist
  character(len=256) apelistout

  character(len=256), dimension(max_ima) :: imagef, psfmapf, &
       expimagf, detimagf, bkgimagef, srcimagef, apesenseimagef
  logical shapepsf_flag, APESENSE_FLAG, APESTACK_FLAG, &
       APEX_FLAG, PSFMAP_FLAG,  REMSRC_FLAG
  real(kind=real32), dimension(max_ima) :: emin, emax      ! energy boundaries 
  integer(kind=int32), dimension(max_ima) :: energy_index   ! mapping between input images and the band index of the ermldet lists 

  integer(kind=int32) I
  integer(kind=int32) BINNING

  real(kind=real32) EEF_EXTRACT, CUTRAD, POISSON_THRESH
  logical flagmllist

  ! these are for testing
  character ( len = 256) filename


  filename="" ! this is to suppress warning filename is  used for testing
  status=0
  ierr=0

  ! these should become input values
  REMSRC_FLAG = .false.


  ! get command line parameters =======================================
  call opast(ierr)
  !what is ap
  call  ape_get_param(n_imagesets, n_expimagesets, n_detmasksets, &
       n_bkgimagesets, n_srcimagesets, n_psfmapsets, n_sensemapsets, &
       n_emin, n_emax, n_energy_index,& 
       mlsrclist, apesrclist, apelistout, &
       imagef, expimagf, detimagf, bkgimagef, srcimagef, &
       psfmapf, apesenseimagef, & 
       emin, emax, energy_index, eef_extract,  POISSON_THRESH, &
       CUTRAD, BINNING, APESENSE_FLAG, &
       APEX_FLAG, APESTACK_FLAG,PSFMAP_FLAG, shapepsf_flag, ierr)
  

  call ape_pass_param(apar, n_imagesets, n_expimagesets, n_detmasksets, &
       n_bkgimagesets, n_srcimagesets,n_psfmapsets, n_sensemapsets,&
       n_emin, n_emax,  n_energy_index,& 
       mlsrclist, apesrclist, apelistout, &
       imagef, expimagf, detimagf, bkgimagef, srcimagef, &
       psfmapf, apesenseimagef,  &
       emin, emax,  energy_index, eef_extract,POISSON_THRESH, &
       CUTRAD, BINNING, APESENSE_FLAG, &
       APEX_FLAG, APESTACK_FLAG, PSFMAP_FLAG, &
       shapepsf_flag, REMSRC_FLAG, ierr)
  !filename="test1.fits"
  !write(*,*) energy_index(1:n_energy_index)
  !write(*,*) apar%PSFMAP_FLAG 
  !call stacktesting(par,energy_index,n_energy_index,filename)
  !stop

  ! this is called if PSFMAP_FLAG = True  runs the PSFMAP creation branch
  if( apar%PSFMAP_FLAG ) then 
     
     if( SIZE(apar%psfmapsets)<=0 ) then
        ierr=3
        call  wrerr(rtname,ierr,'psfmap_flag=yes but no psfmap filenames are provided')
     endif     

     if (allocated(MyPSFSIZE)) deallocate(MyPSFSIZE)
     allocate (MyPSFSIZE(SIZE(imagef)))

     do I=1, apar%epar%n_imagesets
        !write(*,*) "HERE: ",I, imagef(I), SIZE(imagef), n_imagesets
        call ape_prep_psfmap(apar%epar%imagesets(I), apar%epar%shapelet_flag, apar%psfmapsets(I), MyPSFSIZE(I))
     enddo   
  endif
  
  ! this is called if APEX_FLAG = True or APESENSEFLAG = True
  if( apar%APEX_FLAG.or.apar%APESENSE_FLAG.or.apar%APESTACK_FLAG) then 


     ! this is called if PSFMAP_FLAG = False
     !  If True then MyPSFSIZE structure has already been 
     !  created
     if( .not. (apar%PSFMAP_FLAG) ) then 
        if (allocated(MyPSFSIZE)) deallocate(MyPSFSIZE)
        allocate (MyPSFSIZE(apar%epar%n_imagesets ))
        do I=1,apar%epar%n_imagesets 
           call ape_psfsize_read(apar%psfmapsets(I), MyPSFSIZE(I))
        enddo
     endif

     if (ierr < 3) then
        if (apar%epar%shapelet_flag) then
           call alloc_eventfiles(eventlist, apar%epar%n_imagesets, status)
           if (status > 0) then
              ierr=3
              call  wrerr(rtname,status,'Cannot allocate memory for event lists')
           endif
           allocate(pixlist(apar%epar%n_imagesets))
        endif
        call ape_read_images(apar%epar,image,eventlist,pixlist,expimage,bkgimage,srcimage,detmask,ierr)        


        !that should become
        !call ape_read_images(apar,image,eventlist,pixlist,expimage,bkgimage,srcimage,detmask,ierr)        

        ! this is called if APESENSE_FLAG = True
        ! this runs the sensitivity map creation branch
        if (apar%APESENSE_FLAG) then
           !call ape_sense(apar%epar, expimage, srcimage, apesenseimagef, MyPSFSIZE, &
           !     apar%EEF_EXTRACT, apar%BINNING, apar%POISSON_THRESH)           
           ! that should become
           !call ape_sense(apar, expimage, bkgimage, MyPSFSIZE)
           call ape_sense(apar, expimage, srcimage, detmask, MyPSFSIZE)
           
        endif

        ! this is called if APEX_FLAG = True
        ! this runs the source and background count extraction branch
        ! it assumes an ermldet structure as input.
        if (apar%APEX_FLAG) then
           nim=apar%epar%n_imagesets
           call fget_mllist(apar%epar%mllistset,mllist,ierr)

           !write(*,*) apar%BINNING, apar%POISSON_THRESH, apar%epar%CUTRAD, apar%EEF_EXTRACT
           !write(*,*) SIZE(pixlist)
           call ape_erml(apar,mllist, eventlist, &
                image, pixlist, expimage, bkgimage, srcimage, detmask, &
                MyPSFSIZE)
        endif

        
        if (apar%APESTACK_FLAG) then
           nim=apar%epar%n_imagesets
           flagmllist = .FALSE.
           if(apar%epar%mllistset.ne."")then
              flagmllist = .TRUE.
              call fget_mllist(apar%epar%mllistset,mllist,ierr)
              !write(*,*)mllist%data(1)  
              if(mllist%shapehdr%naxis2.eq.0)flagmllist = .FALSE.
           endif

           call extract(apar, mllist, eventlist, &
             image, pixlist, expimage, bkgimage, srcimage, &
             detmask,flagmllist,  MyPSFSIZE)

        endif

     endif ! closing if ierr < 3

  endif ! closing if apex_flag 


  ! deallocate MyPSFSIZE structure
  ! closing down .... 
  if (allocated(MyPSFSIZE)) then
     do I=1,apar%epar%n_imagesets
        call ape_psfsize_kill(MyPSFSIZE(I)) 
     enddo
     deallocate(MyPSFSIZE)
  endif

   !deallocate events via erolib
   call dealloc_eventfiles(eventlist, status) ! IMS: this does a complete free of each eventfile.

   if (allocated(pixlist)) deallocate(pixlist)

  stop

  
  
end program apetool
