! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Fountdation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module apetool_stack

  use param
  !use ermldet_mod
  use numkind
  use erolib_eventfile_mod
  use slib_constants
  use apetool_io

  implicit none

  !logical, private, parameter :: casesens = .false.
  !integer, private, parameter :: ftread = 0, ftwrite = 1  ! ??
  !atttiude file related stuff
  type :: STACKSRC     
     ! this is the structure with the arbitrary positions 
     ! at which counts will be extracted
     ! RA, DEC are the sky coordinates
     ! RExtract is the radius within which counts will be extracted
     ! RRemove is the radius within which sources will be searched
     ! for to be removed from the SRCMAP before estimating the background
     ! at the arbitrary positions.
     ! Both radii are in arcseconds
     ! If they are <1 they are interpreted as the radius
     ! at the corresponding EEF.
     ! logical  PSFMAP determined if PSF maps need to be
     ! provided or not. 
     double precision, dimension (:), allocatable :: RA, DEC
     double precision, dimension (:), allocatable :: X, Y
     double precision, dimension (:), allocatable :: RE, RR
     integer  TOTAL ! total number of sources
     logical  PSFMAP
  end type STACKSRC


  
contains

    ! this is the destructor
    subroutine ape_kill_stacksrc(mySRC)

    TYPE(STACKSRC) :: mySRC
 
    IF (ALLOCATED (mySRC%RA)) DEALLOCATE (mySRC%RA)
    IF (ALLOCATED (mySRC%DEC)) DEALLOCATE (mySRC%DEC)
    IF (ALLOCATED (mySRC%X)) DEALLOCATE (mySRC%X)
    IF (ALLOCATED (mySRC%Y)) DEALLOCATE (mySRC%Y)
    IF (ALLOCATED (mySRC%RE)) DEALLOCATE (mySRC%RE)
    IF (ALLOCATED (mySRC%RR)) DEALLOCATE (mySRC%RR)
    mySRC%TOTAL=0
    mySRC%PSFMAP = .False.
  end subroutine ape_kill_stacksrc


  !this is the constructor
  subroutine ape_init_stacksrc(mySRC, NUMSRC)

    implicit none

    TYPE(STACKSRC) :: mySRC
    INTEGER NUMSRC

    call ape_kill_stacksrc(mySRC)
    
    mySRC%TOTAL = NUMSRC
    allocate  ( mySRC%RA(NUMSRC), mySRC%DEC(NUMSRC) )
    allocate  ( mySRC%RE(NUMSRC), mySRC%RR(NUMSRC) )
    allocate  ( mySRC%X(NUMSRC), mySRC%Y(NUMSRC) )
    mySRC%PSFMAP = .False.
    
  end subroutine ape_init_stacksrc



  
  subroutine ape_read_stacksrc(file, mySRC)

    implicit none
    CHARACTER(len=*) :: file
    TYPE(STACKSRC) :: mySRC
    integer :: unit, status, rwmode, blocksize 

    
    status=0
    rwmode=0 ! open as readonly 
    call ftgiou ( unit, status )
    write(*,*)  "apetool: reading file:", file
    call ftopen( unit, file, rwmode, blocksize, status)  
    !write(*,*) status, unit


    !call ftmnhd(unit, 2 , 'EVENTS', 0, status)
    !call ftgnrw(unit, nrows_ev, status)
    !write(*,*) unit
    call ape_read_stacksrc_extension(unit,  mySRC)

    call FTCLOS(unit,status)
    call FTFIOU(unit,status)
    
  end subroutine ape_read_stacksrc

  
  subroutine ape_read_stacksrc_extension( unit,  mySRC)
    !this reads the event extensionm of an event file. 

    implicit none

    integer unit

    TYPE(STACKSRC) :: mySRC

    integer nrows, irow, status
    
    logical :: casesn=.false., anyf

    integer :: colra, coldec, colre, colrr
    integer ::  dcra, dcdec, dcrr, dcre, repeat, width

    double precision :: dnullval, dvalues
    
    real  rnullval, rvalues

    integer inullval, nhdu, hdutype

    status=0
    dnullval=0.0 
    inullval=0 

    !call ftmnhd(unit, 2 , 'Joined', 0, status)
    !call ftmnhd(unit, 2 , 'SRCLIST', 0, status)
    ! move to first table extension nhdu=2
    nhdu=2
    call FTMAHD(unit,nhdu, hdutype,status)
    call ftgnrw(unit, nrows, status)  
    !write(*,*) nrows, status, hdutype, nhdu, unit
    mySRC%TOTAL = nrows

    write(*,*) "apetool: total positions for count extraction", nrows
    

    call ftgcno(unit,casesn,'RA',colra,status)   
    call ftgcno(unit,casesn,'DEC',coldec,status)  
    call ftgcno(unit,casesn,'RE',colre,status)  
    call ftgcno(unit,casesn,'RR',colrr,status)  

    call FTGTCL(unit,colra, dcra, repeat, width, status)
    call FTGTCL(unit,coldec, dcdec, repeat, width, status)
    call FTGTCL(unit,colre, dcrr, repeat, width, status)
    call FTGTCL(unit,colrr, dcre, repeat, width, status)
    
    call ape_init_stacksrc(mySRC, nrows)

    status=0
    !anyf=.True.
    do irow=1, nrows

       if(dcra.eq.82)then
          call ftgcvd(unit,colra,irow,1,1,dnullval,dvalues,anyf,status)
          mySRC%RA(irow)= dvalues
       else if (dcra.eq.42)then
          call ftgcve(unit,colra,irow,1,1,rnullval,rvalues,anyf,status)
          mySRC%RA(irow)= DBLE(rvalues)
       endif
       
       if(dcdec.eq.82)then
          call FTGCVD(unit,coldec,irow,1,1,dnullval,dvalues,anyf,status)
          mySRC%DEC(irow)= dvalues
       else if (dcdec.eq.42)then
          call ftgcve(unit,coldec,irow,1,1,rnullval,rvalues,anyf,status)
          mySRC%DEC(irow)= DBLE(rvalues)
       endif

       if(dcre.eq.82)then
           call FTGCVD(unit,colre,irow,1,1,dnullval,dvalues,anyf,status)
           mySRC%RE(irow)= real(dvalues)
       else if (dcre.eq.42)then
          call ftgcve(unit,colre,irow,1,1,rnullval,rvalues,anyf,status)
          mySRC%RE(irow)= rvalues
       endif

       !if at least one extraction radius is <1 then
       !it needs to read the PSF maps as well to determine
       !extraction radii. 
       if(mySRC%RE(irow).lt.1)mySRC%PSFMAP = .True.


       if(dcrr.eq.82)then
           call FTGCVD(unit,colrr,irow,1,1,dnullval,dvalues,anyf,status)
           mySRC%RR(irow)= real(dvalues)
       else if (dcrr.eq.42)then
          call ftgcve(unit,colrr,irow,1,1,rnullval,rvalues,anyf,status)
          mySRC%RR(irow)= rvalues
       endif
       !if at least one remove radius is <1 then
       !it needs to read the PSF maps as well to determine
       !extraction radii. 
       if(mySRC%RR(irow).lt.1)mySRC%PSFMAP = .True.
       !write(*,*) irow, mySRC%RA(irow),mySRC%DEC(irow),mySRC%RE(irow),mySRC%RR(irow)
     end do

  end subroutine ape_read_stacksrc_extension



!----------------------------------------------------------------------------
  subroutine ape_stack_nearest_srcs(XSRC, YSRC, RREMOVE, BOXLIST, OUT_SRC, NUM_OUT_SRC)
!----------------------------------------------------------------------------
! It looks for the nearest sources in the vicinity of the INSRC
! in the APELIST. Nearest means XX% EEF radius away. 
! At this stage the EEF distance is hardcored to 50%. In the future 
! a decision is needed how to imrove. 
! The IDs of the output sources are written in allocatable OUTSRC
! Remember to delallocate....
    use detect_io

    implicit none
    
    real(kind=real32), intent(in) :: XSRC, YSRC, RREMOVE
    type(boxlist2d_type), intent(in)  :: BOXLIST
    INTEGER,  dimension (:), intent(out) :: OUT_SRC 
    INTEGER,  intent(out) ::  NUM_OUT_SRC

    real(kind=real32) rcut2, dr2, dx, dy, x, y
    INTEGER NSRC, ISRC, NSIZE

    OUT_SRC  = -9
    NSIZE = SIZE(OUT_SRC)
    NSRC=boxlist%shapehdr%naxis2 

    rcut2 = RREMOVE**2


    write(*,*) "START:", xsrc, ysrc, rcut2, NSRC, NSIZE

    NUM_OUT_SRC = 0
    ISRC = 1
    do while (ISRC<=NSRC .and. NUM_OUT_SRC < NSIZE)
          x = boxlist%data(0,ISRC)%x_ima
          y = boxlist%data(0,ISRC)%y_ima
          
          dx = x - xsrc
          dy = y - ysrc
          dr2 = dx * dx + dy * dy
          
          if(dr2<rcut2)then
             NUM_OUT_SRC = NUM_OUT_SRC + 1 
             OUT_SRC(NUM_OUT_SRC) =ISRC

             write(*,*) ISRC, x, y, sqrt(dr2), sqrt(rcut2), NUM_OUT_SRC
             
          endif

          ISRC = ISRC + 1
       end do
       
       
     end subroutine ape_stack_nearest_srcs
!----------------------------------------------------------------------------


!  !---------------------------------------------------------------------
!  subroutine ape_stack_srcmap_isrc(par,image,eventlist,pixlist,detmask,expimage,bkgimage, &
!       allStackPSF,cluPSF,isrc,srcimage, &
!       allStackPSF2,cluPSF2,boxlist,srcmodrad,energy_index, n_energy_index,ierr)
!
!  !--------------------------------------------------------------------
!    use detect_io
!    use ermldet_mod
!    use slib_image
!    implicit none
!
!    type(ermldet_param_type)                 :: par
!    type(i4cube_type)                        :: image
!    type(tEventFile),dimension(:)            :: eventlist
!    type(pixlist_type),dimension(:)          :: pixlist
!    type(l1cube_type)                        :: detmask
!    type(r4cube_type)                        :: expimage
!    type(r4cube_type)                        :: bkgimage
!    type(r4cube_type)                        :: srcimage
!    type(r4cube_type)                        :: run_srcimage
!    type(boxlist2d_type)                     :: boxlist
!    integer(kind=int32)                      :: srcmodrad  
!    integer(kind=int32)                      :: ierr
!    integer(kind=int32)                      :: isrc
!    integer(int32), dimension(:), intent(in) :: energy_index
!    integer(int32), intent(in) :: n_energy_index
!
!    type(events_type),pointer,dimension(:)   :: events
!    type(srcrow_type),dimension(:,:),pointer :: clulist  => NULL()
!    type(FPSF)                               :: allStackPSF,allStackPSF2   
!    type(FPSF)                               :: cluPSF,cluPSF2   
!    type(mask_type),save                     :: mask
!
!    type(r4image_type)                       :: junk
!    integer(kind=int32)                      :: xsize,ysize,i
! 
!    real(kind=real32), dimension(:,:,:), pointer   :: modelima   
!
!    integer(kind=int32)                      :: nclu
!    character(len=*), parameter              :: rtname='APE_SRCMAP_ISRC'
!
!    ! these are need by model subroutine call further down
!    real(kind=real32)                        :: yfit0
!    real(kind=real32),dimension(:),allocatable  :: yfit1
!    real(kind=real32),dimension(max_ima)        :: yfit2 
!    real(kind=real64), dimension(max_fit)         :: x0,y0,sig         
!    real(kind=real64), dimension(max_ima,max_fit) :: rate
!    integer(kind=int32)                           :: ipoint,ix,iy,mx,my
!    logical(kind=log32)                         :: last = .FALSE.
!    logical(kind=log32)                      :: newcal = .TRUE.
!    integer(kind=int32)                      :: status
!    logical :: testing=.false.
!    character ( len = 256) filename, name
!
!    ierr = 0
!    status=0
!    
!    ! read new cal PSF
!    newcal=.true.
!    last=.false.
!    
!
!    ! allocate clulist with summary element (0:n_imagesets,maxclu)
!    ! AGE:     max_fit= 10 is global 
!    ! so this array holds current group of sources to fit
!    if (.not. associated(clulist)) allocate(clulist(0:par%n_imagesets,max_fit))    
!    call init_mllist2d(clulist)
!
!    call ape_stack_sou_clus(par,boxlist,allStackPSF,allStackPSF2,isrc,clulist,cluPSF,cluPSF2,nclu,energy_index, n_energy_index,ierr)  
!  
!    ! set cut-out radii for source model - srcmodrad is fixed in main
!    clulist(:,1:nclu)%ecutrad = real(srcmodrad)
!
!
!    call fill_arrs(par,.false.,clulist,image,eventlist,pixlist,cluPSF,detmask,expimage,modelima,nclu,events,mask,status)
!
!
!    if (allocated(yfit1)) deallocate(yfit1)
!    allocate(yfit1(mask%nev))  
!    x0  = clulist(0,:)%x_ima
!    y0  = clulist(0,:)%y_ima
!    sig = clulist(0,:)%ext
!    rate(1:par%n_imagesets,:) = clulist(1:par%n_imagesets,:)%rate
!
!    !write(*,*) clulist(1:par%n_imagesets,:)%rate
!    !write(*,*) x0, y0, sig
!
!    call model(par,image,expimage,bkgimage,mask,events,clulist,cluPSF2,x0,y0,sig,rate,nclu,last,newcal,yfit0,yfit1,yfit2,modelima,status)
!
!    
!    ! AGE TESTING
!    write (filename, "(A5, I4.4)") &
!         "SRC_", ISRC
!    filename=trim(filename)//'_1.img'
!    call slib_run_write2image(dble(modelima(:,:,2)), filename)
!    write (filename, "(A5, I4.4)") &
!         "SRC_", ISRC
!    filename=trim(filename)//'_2.img'
!    !filename='test2.img'
!    call slib_run_write2arrays_fast(cluPSF%mySPSF(2),cluPSF%myCPSF,1)
!    call slib_run_write2image(cluPSF%myCPSF%img, filename)
!
!
!    do ipoint=1,par%n_imagesets
!       my=0
!       do iy = mask%iy_min, mask%iy_max
!          mx=0
!          my=my+1
!          do ix = mask%ix_min, mask%ix_max
!             mx=mx+1
!             if (mask%mask(mx,my,ipoint) > 0) then
!
!
!              srcimage%arr(ix,iy,ipoint) = srcimage%arr(ix,iy,ipoint)  &
!                     - modelima(mx,my,ipoint) * expimage%arr(ix,iy,ipoint)
!
!                   
!             endif
!          enddo
!       enddo
!    end do
!    
!    ! AGE TESTING
!    write (filename, "(A7, I4.4)") &
!         "SRCMAP_", ISRC
!    filename=trim(filename)//'_1.img'
!    call slib_run_write2image(dble(srcimage%arr(:,:,2)), filename)
!    !write(*,*) filename
!    
!    if(testing)then
!       xsize=size(srcimage%arr,1)
!       ysize=size(srcimage%arr,2)
!       allocate(junk%arr(xsize,ysize))
!       i=1
!       !do i=1,par%n_srcimagesets
!          
!          junk%arr=srcimage%arr(:,:,i)
!          junk%stdhdr=srcimage%stdhdr(i)
!          junk%refhdr=srcimage%refhdr(i)
!          junk%wcshdr=srcimage%wcshdr(i)
!          
!
!          junk%shapehdr%bitpix=-32
!          junk%shapehdr%naxis=2
!          junk%shapehdr%naxis1=xsize
!          junk%shapehdr%naxis2=ysize  
!
!          junk%stdhdr%hduclas1='IMAGE'
!          junk%stdhdr%hduclas2='GENERIC'
!          junk%stdhdr%creator='ERMLDET'
!          
!          
!       !enddo
!       call fput_r4image("!test1.fits",junk,status)
!       if (status >  0) then
!          ierr = 3
!          call wrerr(rtname,ierr,'Error writing source-map file')
!       endif
!       
!    
!
!       deallocate(junk%arr)
!    endif
!
!    
!    deallocate(yfit1)
!    
!  end  subroutine ape_stack_srcmap_isrc
!  !--------------------------------------------------------------------------------


! !---------------------------------------------------------------------
!  subroutine ape_stack_sou_clus(par,boxlist,allStackPSF,allStackPSF2,isrc,clulist,cluPSF,cluPSF2,nclu,energy_index, n_energy_index,ierr)  
!  !--------------------------------------------------------------------
!  ! this is adapted from the ermldet_mod.f90 sou_clus routine. 
!  ! It defines the clulist,cluPSF,cluPSF2,nclu
!  ! to be used later. 
!  ! it uses id_cluster flag in mllist to identify sources that 
!  ! belong to the same cluster.
!  !--------------------------------------------------------------------
!
!    use detect_io
!    !use ermldet_mod
!    implicit none
!
!
!    type(ermldet_param_type),intent(in)      :: par
!    type(boxlist2d_type)                     :: boxlist
!    type(FPSF)                               :: allStackPSF
!    type(FPSF)                               :: allStackPSF2
!    integer(kind=int32)                      :: isrc
!    type(srcrow_type),dimension(:,:),pointer :: clulist   
!    type(FPSF)                               :: cluPSF
!    type(FPSF)                               :: cluPSF2
!    integer(kind=int32)                      :: nclu,ierr
!    integer(int32), dimension(:), intent(in) :: energy_index
!    integer(int32), intent(in) :: n_energy_index
!
!    
!    integer(kind=int32)                      :: i, nsrc, iima, K
!    integer(kind=int32)                      :: current_cluster_id, running_cluster_id
!    character(len=*), parameter              :: rtname='APE_SOU_CLUS'
!    real(kind=real32)                        :: r_off,scut2,rad2
! 
!    ierr=0
!    nsrc=boxlist%shapehdr%naxis2 
!    nclu = 0
!    !scut2=boxlist%data(0,isrc)%scutrad**2 
!    
!    ! at least one element in nlcu
!    nclu=nclu+1
!    ! clulist had 1 dimension equals to the number of input images
!    ! this may be different from the number of bands in the
!    ! input ermldet list (boxlist below). For that reason there is
!    ! an index list that maps the input images (in the order they appear)
!    ! to the index of the bands as they appear in the ermldet
!    ! If no EINDEX has been provided, then it assumes that
!    ! bands and images appear in the same order and are equal
!
!    !write(*,*) "IN", n_energy_index
!    if(n_energy_index==0)then 
!       clulist(:,nclu)=boxlist%data(:,isrc)
!    else
!       clulist(0,nclu)=boxlist%data(0,isrc)
!       do K=1,n_energy_index
!          !write(*,*) K, energy_index(K), "S:",SHAPE(clulist), SHAPE(boxlist%data)
!          clulist(K,nclu)=boxlist%data(energy_index(K),isrc)
!       enddo
!    endif
!    current_cluster_id = boxlist%data(0,isrc)%id_cluster
!
!    !write(*,*) rtname, " :START: ",  current_cluster_id, nclu
!
!    ! why is that here????
!    if (par%shapelet_flag) then
!       if (.not. allocated(cluPSF%myModelPSF)) then
!          allocate(cluPSF%myModelPSF(size(allStackPSF%myModelPSF)))
!          allocate(cluPSF2%myModelPSF(size(allStackPSF2%myModelPSF)))
!          !write(*,*) rtname,"NOT ALLOCATED?"
!       endif
!
!       do iima=1,nim
!                 !write(*,*) 'cluPSF: ',iima, size(cluPSF)
!                 !write(*,*) 'cluPSF: ',iima, size(cluPSF), size(cluPSF(iima)%c1)
!                 !write(*,*) 'size(cluPSF%myModelPSF),size(allStackPSF%myModelPSF:'
!                 !write(*,*) size(cluPSF%myModelPSF),size(allStackPSF%myModelPSF) 
!          cluPSF%myModelPSF(:) = allStackPSF%myModelPSF(:)
!          !       write(*,*) 'size(NX1)=',size(cluPSF%myModelPSF(1)%NX1)
!          write(*,*) iima, nim, SHAPE(cluPSF%mySPSF), SHAPE(allStackPSF%mySPSF)
!          cluPSF%mySPSF(iima)%c0(:,nclu) = allStackPSF%mySPSF(iima)%c0(:,isrc)
!          cluPSF%mySPSF(iima)%c1(:,nclu) = allStackPSF%mySPSF(iima)%c1(:,isrc)
!          cluPSF%mySPSF(iima)%c2(:,nclu) = allStackPSF%mySPSF(iima)%c2(:,isrc)
!          cluPSF%mySPSF(iima)%nx0 = allStackPSF%mySPSF(iima)%nx0
!          cluPSF%mySPSF(iima)%nx1 = allStackPSF%mySPSF(iima)%nx1
!          cluPSF%mySPSF(iima)%nx2 = allStackPSF%mySPSF(iima)%nx2
!          cluPSF%mySPSF(iima)%ny0 = allStackPSF%mySPSF(iima)%ny0
!          cluPSF%mySPSF(iima)%ny1 = allStackPSF%mySPSF(iima)%ny1
!          cluPSF%mySPSF(iima)%ny2 = allStackPSF%mySPSF(iima)%ny2
!          
!          cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,isrc)
!          cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,isrc)
!          cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,isrc)
!          cluPSF2%mySPSF(iima)%nx0 = allStackPSF2%mySPSF(iima)%nx0
!          cluPSF2%mySPSF(iima)%nx1 = allStackPSF2%mySPSF(iima)%nx1
!          cluPSF2%mySPSF(iima)%nx2 = allStackPSF2%mySPSF(iima)%nx2
!          cluPSF2%mySPSF(iima)%ny0 = allStackPSF2%mySPSF(iima)%ny0
!          cluPSF2%mySPSF(iima)%ny1 = allStackPSF2%mySPSF(iima)%ny1
!          cluPSF2%mySPSF(iima)%ny2 = allStackPSF2%mySPSF(iima)%ny2
!       enddo
!    endif
!    ! look for other sources in cluster 
!    ! it uses the id_cluster flag
!    do i=isrc+1,nsrc
!
!       !write(*,*) rtname, " :LOOP: ",  current_cluster_id,  boxlist%data(0,i)%id_cluster,  boxlist%data(0,i)%id_src, boxlist%data(0,i)%x_ima, boxlist%data(0,i)%y_ima
!       if (boxlist%data(0,i)%id_cluster ==  current_cluster_id  ) then
!       !write(*,*) rtname, " :FOUND: "
!         
!          nclu=nclu+1
!          ! GHANGE THIS LATER ON     
!
!          ! clulist had 1 dimension equals to the number of input images
!          ! this may be different from the number of bands in the
!          ! input ermldet list (boxlist below). For that reason there is
!          ! an index list that maps the input images (in the order they appear)
!          ! to the index of the bands as they appear in the ermldet
!          ! If no EINDEX has been provided, then it assumes that
!          ! bands and images appear in the same order and are equal
!          if(n_energy_index==0)then 
!             clulist(:,nclu)=boxlist%data(:,i)
!          else
!             clulist(0,nclu)=boxlist%data(0,i)
!             do K=1,n_energy_index
!                clulist(K,nclu)=boxlist%data(energy_index(K),i)
!             enddo
!          endif
!          !clulist(:,nclu)=boxlist%data(0:2,i)
!          ! GHANGE THIS LATER ON     
!          
!          if (par%shapelet_flag) then
!             do iima=1,nim
!                ! change 19/11/2017:
!                !luPSF%mySPSF(iima)%c0(:,nclu)  = allStackPSF%mySPSF(iima)%c0(:,isrc)
!                !cluPSF%mySPSF(iima)%c1(:,nclu)  = allStackPSF%mySPSF(iima)%c1(:,isrc)
!                !cluPSF%mySPSF(iima)%c2(:,nclu)  = allStackPSF%mySPSF(iima)%c2(:,isrc)
!                cluPSF%mySPSF(iima)%c0(:,nclu)  = allStackPSF%mySPSF(iima)%c0(:,i)
!                cluPSF%mySPSF(iima)%c1(:,nclu)  = allStackPSF%mySPSF(iima)%c1(:,i)
!                cluPSF%mySPSF(iima)%c2(:,nclu)  = allStackPSF%mySPSF(iima)%c2(:,i)
!                cluPSF%mySPSF(iima)%nx0 = allStackPSF%mySPSF(iima)%nx0
!                cluPSF%mySPSF(iima)%nx1 = allStackPSF%mySPSF(iima)%nx1
!                cluPSF%mySPSF(iima)%nx2 = allStackPSF%mySPSF(iima)%nx2
!                cluPSF%mySPSF(iima)%ny0 = allStackPSF%mySPSF(iima)%ny0
!                cluPSF%mySPSF(iima)%ny1 = allStackPSF%mySPSF(iima)%ny1
!                cluPSF%mySPSF(iima)%ny2 = allStackPSF%mySPSF(iima)%ny2
!
!                !change 19/11/2017
!                !cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,isrc)
!                !cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,isrc)
!                !cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,isrc)
!                cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,i)
!                cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,i)
!                cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,i)
!                cluPSF2%mySPSF(iima)%nx0 = allStackPSF2%mySPSF(iima)%nx0
!                cluPSF2%mySPSF(iima)%nx1 = allStackPSF2%mySPSF(iima)%nx1
!                cluPSF2%mySPSF(iima)%nx2 = allStackPSF2%mySPSF(iima)%nx2
!                cluPSF2%mySPSF(iima)%ny0 = allStackPSF2%mySPSF(iima)%ny0
!                cluPSF2%mySPSF(iima)%ny1 = allStackPSF2%mySPSF(iima)%ny1
!                cluPSF2%mySPSF(iima)%ny2 = allStackPSF2%mySPSF(iima)%ny2
!             enddo
!          endif
!       endif
!    enddo
! 
!
!    if (par%shapelet_flag) then
!       do iima=1,nim
!          cluPSF%mySPSF(iima)%num_src=nclu
!          cluPSF2%mySPSF(iima)%num_src=nclu
!       enddo
!    endif
!
!
!  end subroutine ape_stack_sou_clus
!  !--------------------------------------------------------------------------------






  
!  subroutine stacktesting(apar, stackfile)
!    use detect_io
!    !use slib_data_psf
!    !use slib_calc_space
!    !use radec2pix_mod
!
!    implicit none
!    
!    type(APEPAR) , intent(in)  :: apar
!    character(len=*) :: stackfile
!    ! stack sources structures
!    TYPE(STACKSRC) :: mySRC
!    !integer(int32), dimension(:), intent(in) :: energy_index
!    !integer(int32), intent(in) :: n_energy_index
!        
!    ! read the source list positions to be extracted
!    call  ape_read_stacksrc(stackfile, mySRC)
!    
!  end subroutine stacktesting

  
  subroutine extract(apar, mllist, eventlist, &
       image, pixlist, expimage, bkgimage, srcimage, &
       detmask, flagmllist, MyPSFSIZE)

    use detect_io
    use slib_data_psf
    !use slib_calc_space
    use radec2pix_mod
    use ermldet_mod

    
    implicit none
    
    character(len=*), parameter :: rtname='APETOOL_STACK_EXTRACT'
    
    type(APEPAR), intent(in) :: apar
    type(mllist_type), intent(in) :: mllist
    type(tEventFile), dimension(:), intent(in)   :: eventlist
    logical, intent(in) :: flagmllist
    type(i4cube_type) :: image
    type(pixlist_type), dimension(:) :: pixlist
    type(r4cube_type) :: expimage
    type(r4cube_type) :: bkgimage
    type(r4cube_type) :: srcimage
    type(l1cube_type) :: detmask    
    ! I assume that array structure MyPSFSIZE has been define outside 
    ! the subroutine and supplied to it 
    TYPE(PSFSIZE), dimension(:), intent(in) :: MyPSFSIZE
    !integer(int32), dimension(:), intent(in) :: energy_index
    !integer(int32), intent(in) :: n_energy_index
    
    ! PSF coefficients to be used to remove sources
    ! from SRCMAP
    ! boxlist is also needed to remove sources
    ! 
    type(FPSF) :: allStackPSF,allStackPSF2
    type(FPSF) :: cluPSF, cluPSF2
    integer(kind=int32),parameter              :: srcmodrad=120 
    type(boxlist2d_type)  :: boxlist
    real(kind=real32), dimension(:,:,:), allocatable :: srcimage_save

    
    INTEGER,  dimension (1000) :: OUT_SRC 
    INTEGER NUM_OUT_SRC, naxis3(3), MAX_ENERGY
    !INTEGER axis2(2)
    real(kind=real32) rcut2, asum, FXY, aexposure
    integer(kind=int32) isum, I, J, K, iim, ierr, status
    real(real64) :: ra0, dec0, x0, y0, &     ! WCS for sky images
         xscal, yscal             !
    integer :: ix_size, iy_size              ! size of sky images    


    ! stack sources structures
    TYPE(STACKSRC) :: mySRC
    ! boxsrclist for fascilitating storing of results and writing to file
    type(boxlist2d_type)  :: boxsrclist
    integer nrows


    ierr=0
    ! read the source list positions to be extracted
    call  ape_read_stacksrc(apar%apelist, mySRC)

    !initiate boxsrclist. currently only a single band at a time is supported.
    ! testing here with more than one input images
    !nrows=mllist%shapehdr%naxis2
    nrows=SIZE(mySRC%RA)
    !nim = maxval(mllist%data(:)%id_band)
    !nim=1
    nim=apar%epar%n_imagesets
    
    allocate(boxsrclist%data(0:nim,nrows))
    call init_boxlist2d(boxsrclist%data)

    ! use first image as template
    iim=1
    ra0  = image%wcshdr(iim)%crval1
    dec0 = image%wcshdr(iim)%crval2
    x0   = image%wcshdr(iim)%crpix1
    y0   = image%wcshdr(iim)%crpix2
    xscal= image%wcshdr(iim)%cdelt1
    yscal= image%wcshdr(iim)%cdelt2
    ix_size = image%shapehdr(iim)%naxis1
    iy_size = image%shapehdr(iim)%naxis2

    !write(*,*) abs(xscal), abs(yscal), ix_size, iy_size, ra0,dec0

    
    call radec2pix (mySRC%ra, mySRC%dec,  mySRC%TOTAL, ra0, dec0, x0, y0, xscal, yscal, mySRC%X, mySRC%Y)

    
    ! I need to generate a running srcimage which will 
    ! store the temporary source map with the source(s) in question 
    ! removed. This map will be used to estimate the background
    ! counts at given position
    naxis3 = SHAPE(srcimage%arr)
    nim=apar%epar%n_imagesets           
    allocate(    srcimage_save( naxis3(1), naxis3(2), naxis3(3) ) )
    srcimage_save = srcimage%arr

    ! this is checkimng there is an input MLLIST
    if(flagmllist)then

       ! pass mllist to boxlist
       ! this modifies the nim global parameter to nim = maxval(mllist%data(:)%id_band)
       ! ie to the maximum number of bands in the mllist.
       ! this is not any more the number of input images
       !write(*,*) "IERR",ierr
       

       !call ape_ml2box(apar%epar, mllist, boxlist, ierr)
       call ape_ml2box(mllist, boxlist, ierr)
       nim=apar%epar%n_imagesets 
       if (status > 0) then
          ierr=3
          call  wrerr(rtname,status,'Cannot read mllist')
       endif

       if (ierr < 3) then
          ! these tasks define the size of allStackPSF and cluPSF
          ! using the nim, which has been set by the ape_ml2box above
          write(*,*) 'CALL ermldet_shapelets', nim
          call ermldet_shapelets(apar%epar,boxlist,image,eventlist,allStackPSF,0,cluPSF,status)
          call ermldet_shapelets(apar%epar,boxlist,image,eventlist,allStackPSF2,srcmodrad,cluPSF2,status)
          ! in ermldet_shapelets
          ! I set nim back to the number of input images
          nim=apar%epar%n_imagesets
          write(*,*) 'DONE ermldet_shapelets', SHAPE(allStackPSF%mySPSF)
          ! this does not work because some wcshdr keywrds are not set
          do iim=1,apar%epar%n_imagesets
             call init_pixlist(eventlist(iim)%events,image%wcshdr(iim),pixlist(iim),status)
          enddo
       endif
       if (status > 0) then
          ierr=3
          call  wrerr(rtname,status,'Error preparing shapelet PSF')
       endif
       
       boxlist%data(:,:)%cluflag= .false.
    endif

    MAX_ENERGY = apar%epar%n_imagesets
    !write(*,*) "MAX_ENERGY:", MAX_ENERGY
    
    
    do I = 1, mySRC%TOTAL ! start loop in sources

       do J = 1, MAX_ENERGY !apar%epar%n_imagesets

          ! this is the aperture within sources are removed from PSFMAP before extracting background counts
          if(mySRC%RR(I)<1)then
             ! watch out this is in pixels!
             call ape_inter2d(SNGL(mySRC%X(I)), SNGL(mySRC%Y(I)), SNGL(mySRC%RR(I)), MyPSFSIZE(J), FXY)
          else
             ! watch out this is in arcsec!
             FXY = REAL(mySRC%RR(I)) / REAL((abs(xscal) * 3600.0))
          endif

          ! this is checking if there is a source list
          if(flagmllist) then
             !write(*,*) J, I,  mySRC%X(I),  mySRC%Y(I),  mySRC%RA(I), mySRC%DEC(I), FXY
             call ape_stack_nearest_srcs( SNGL(mySRC%X(I)), SNGL(mySRC%Y(I)), FXY, BOXLIST, OUT_SRC, NUM_OUT_SRC)
             !if(NUM_OUT_SRC>0)write(*,*) "Remvin dources:", NUM_OUT_SRC,OUT_SRC(1:NUM_OUT_SRC)
             srcimage%arr = srcimage_save
             !write(*,*)  NUM_OUT_SRC
             do K = 1, NUM_OUT_SRC
                !write(*,*) K, NUM_OUT_SRC
                !ierr=J
                call ape_srcmap_isrc(apar,image,eventlist,pixlist, &
                     detmask,expimage,bkgimage,&
                     allStackPSF,cluPSF, OUT_SRC(K), srcimage, &
                     allStackPSF2,cluPSF2,boxlist,srcmodrad,ierr)
             enddo
          endif

          if(mySRC%RE(I)<1)then
             ! watch out this is in pixels!
             call ape_inter2d(SNGL(mySRC%X(I)), SNGL(mySRC%Y(I)), SNGL(mySRC%RE(I)), MyPSFSIZE(J), FXY)
          else
             ! watch out this is in arcsec!
             FXY = REAL(mySRC%RE(I)) /  REAL(( abs(xscal) * 3600))
          endif
          
          rcut2 = FXY * FXY
          call extract_single_i4(image, J, SNGL(mySRC%X(I)), SNGL(mySRC%Y(I)), rcut2, isum)           
          call extract_single_exp(expimage, J, SNGL(mySRC%X(I)), SNGL(mySRC%Y(I)), rcut2, aexposure) 
          call extract_single_r4(srcimage, J, SNGL(mySRC%X(I)), SNGL(mySRC%Y(I)), rcut2, asum)
          !write(*,*)  I, mySRC%RA(I), mySRC%DEC(I), isum, asum, aexposure, FXY          


          boxsrclist%data(J,I)%id_src=I
          boxsrclist%data(J,I)%id_band = J         
          boxsrclist%data(J,I)%ra  = mySRC%RA(I)

          boxsrclist%data(J,I)%ape_cts  = i4null
          boxsrclist%data(J,I)%ape_bkg = r4null
          boxsrclist%data(J,I)%ape_exp =  r4null
          boxsrclist%data(J,I)%ape_eef = REAL(mySRC%RE(I))
          boxsrclist%data(J,I)%ape_radius = FXY

          if(aexposure.gt.0)then

             boxsrclist%data(J,I)%ape_cts  = isum
             boxsrclist%data(J,I)%ape_bkg = asum
             boxsrclist%data(J,I)%ape_exp = aexposure 
             if(isum.gt.0)then
                boxsrclist%data(J,I)%ape_pois =  exp(-gammp(real(isum),asum))
             else
                boxsrclist%data(J,I)%ape_pois = r4null
             endif
         
             !write(*,*) "HERE...",J, I, boxsrclist%data(J,I)%ape_cts, isum
             !write(*,*) J, I, boxsrclist%data(J,I)%ape_bkg 
             !write(*,*) J, I, boxsrclist%data(J,I)%ape_exp
             !write(*,*) J, I, boxsrclist%data(J,I)%ape_eef
             !write(*,*) J, I, boxsrclist%data(J,I)%ape_radius,  boxsrclist%data(J,I)%ape_pois
          endif
 
       enddo
    enddo
    
    call save_exrtact(apar, boxsrclist, ierr)

    if (allocated(srcimage_save))deallocate(srcimage_save)
    if(flagmllist)then
       call slib_dealloc_FPSF(allStackPSF)
       call slib_dealloc_FPSF(allStackPSF2) 
    endif
    call  ape_kill_stacksrc(mySRC)!
  end subroutine extract






  subroutine save_exrtact(apar, boxsrclist, ierr)

    use detect_io, only: boxlist2d_type, i4null, r4null

    implicit none

    type(APEPAR), intent(in) :: apar
    ! stack sources structures
    !TYPE(STACKSRC) :: mySRC
    type(boxlist2d_type)  :: boxsrclist
    integer(kind=int32), intent(out)   :: ierr

    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,irow,felem,nelements,icol,NIE,IE,IR
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    character(len=16)                  :: extname
    character(len=*),parameter         :: rtname = 'FPUT_MLLIST'
    character(len=80)                  :: errstr, kwd_name
    
    integer tfields
    character*16 ttype(10),tform(10),tunit(10)

    extname="STACK"
    tfields=10
    data ttype/'ID_SRC', 'ID_BAND', 'RA', 'DEC','APE_CTS'  ,'APE_BKG',  &
         'APE_EXP'  ,'APE_EEF'   ,'APE_RADIUS',  &
         'APE_POIS'/
    data tform/ '1J', '1J', '1D','1D','1J','1E','1E','1E','1E','1E'/
    data tunit/'','','deg', 'deg', 'cnts', 'cnts', 'sec', '', 'arcsec',''/
    
    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    !nrows=boxsrclist%shapehdr%naxis2
    nrows = SIZE(boxsrclist%data(1,:))
    NIE = apar%epar%n_imagesets  !SIZE(boxsrclist%data(:,1))-1


    
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    call FTINIT(unit,"!"//trim(apar%apelistout),blocksize,status)
    fstat=max(fstat,status)
    if (fstat /= 0) then    
      ierr = 3
      errstr = 'error opening file '//trim(apar%apelistout)
      call WRERR(rtname,ierr,errstr)
    else    
      status=0
      simple   = .true.
      bitpix   = 16
      naxis      = 0
      naxes(1) = 0

      !write FITS standard header:   
      call FTPHPS(unit,bitpix,naxis,naxes,status) 
      fstat=max(fstat,status)  
      ! write eROSITA standard header to primary extension :
      ! now obsolete?? (29/04/2015)
      call fwrite_stdhdr(unit,boxsrclist%stdhdr,status)
      ! create new empty HDU:
      call FTCRHD(unit,status)
      fstat=max(fstat,status)  
      ! write header for binary table:
      call FTPHBN(unit,nrows,tfields,ttype,tform,tunit,extname,varidat,status)
      !fstat=max(fstat,status)  
      ! put standard header into table extension (29/04/2015): 
      call fwrite_stdhdr(unit,boxsrclist%stdhdr,status)
      fstat=max(fstat,status)  

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(apar%apelistout)
        call WRERR(rtname,ierr,errstr) 
      else 
     
          ! Set the TNULLn keyword for column number 'n':
         icol=5
         write(kwd_name, '(a,i1)') "TNULL", icol ! or whatever, if col_num>9
         call FTUKYJ(unit, kwd_name, int(i4null,kind=int32), "NULL VALUE", ierr)
 
    
      irow=0
      IR = 0  
      IE=1
      nelements=1
      felem=1
      !write(*,*) nrows
      do while (irow < nrows .and. fstat == 0)
          IROW=IROW+1
          IE=0
          do while (IE < NIE .and. fstat == 0)          
             IR=IR+1
             IE=IE+1

          
          icol=1
          call  FTPCLJ(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%id_src,status)

          icol=icol+1
          fstat = max(fstat,status)       
          call  FTPCLJ(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%id_band,status)
          
          icol=icol+1
          fstat = max(fstat,status)       
          call  FTPCLD(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ra,status)
          icol=icol+1
          fstat = max(fstat,status)                         
          call  FTPCLD(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%dec,status) 

          icol=icol+1
          fstat = max(fstat,status)
          call  FTPCLJ(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_cts      ,status)
          icol=icol+1
          fstat = max(fstat,status)  
          !call  FTPCLE(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_bkg      ,status)
          call  FTPCNE(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_bkg,r4null,status)

          icol=icol+1
          fstat = max(fstat,status)  
          !call  FTPCLE(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_exp      ,status)
          call  FTPCNE(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_exp,r4null,status)

          icol=icol+1
          fstat = max(fstat,status)  
          !call  FTPCLE(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_eef      ,status)
          call  FTPCNE(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_eef,r4null,status)

          icol=icol+1
          fstat = max(fstat,status)  
          !call  FTPCLE(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_radius      ,status)
          call  FTPCNE(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_radius,r4null,status)

          icol=icol+1
          fstat = max(fstat,status)  
          !call  FTPCLE(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_pois      ,status)
          call  FTPCNE(unit,icol,ir,felem,nelements,boxsrclist%data(IE,irow)%ape_pois, r4null      ,status)

          fstat = max(fstat,status)
          enddo
        enddo
      endif  

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(apar%apelistout)
        call WRERR(rtname,ierr,errstr)
      endif
    
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//trim(apar%apelistout)
        call WRERR(rtname,ierr,errstr)
      endif
    endif ! 

  end subroutine save_exrtact

  
end module apetool_stack
