! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module apetool_inv

  implicit none

  private

  public :: poissinv, poisscinv

contains

  subroutine poissinv(U, LAM, X)
     implicit none
     double precision, intent(in) :: U, LAM
     double precision, intent(out) :: X
     call poissinv_core(U,1.0-U,LAM, X)  
  end subroutine poissinv
  
   
  subroutine poisscinv(V, LAM, X) 
     implicit none
     double precision, intent(in) :: V, LAM
     double precision, intent(out) :: X
     call poissinv_core(1.0-V,V,LAM, X)
   end subroutine poisscinv
  
  subroutine poissinv_core(U, V, LAM, X)

    implicit none
    
    double precision, intent(in) :: U, V, LAM
    double precision, intent(out) :: X
    integer I
    double precision W, XI, T, R, R2, DEL, LAM_ROOT, LAMI
    double precision ETA, S, S2, B0, B1

    
    X = 0.0
    LAMI= 1.0 / LAM

    if (U.eq.0.0) return

    if (V==0.0) then
       X=1D32
       return
    endif

    
    
    if (LAM.gt.4) then       
       call normcdfinv_as241(MIN(U,V), W)


       if (U.gt.V) W = -W
    
       !use polynomial approximations in central region
 
       if ( ABS(W).lt.3.0 )then
          LAM_ROOT = sqrt(LAM)

          S = LAM_ROOT * W + &
               (1.0 / 3.0 + 1.0 / 6.0 * W * W ) * &
               (1.0 - W / (12.0*LAM_ROOT))

          DEL = (1.0 /160.0)
          DEL = (1.0 / 80.0) + DEL* W * W
          DEL = (1.0 / 40.0) + DEL* W * W
          DEL = DEL * LAMI

          S = LAM + (S + DEL)
          
       ! otherwise use Newton iteration
       else
          S = W / sqrt(LAM)
          R = 1.0 + S
          if (R.lt.0.1) R = 0.1

          R2 = 1e20

          !write(*,*) R, R2,ABS(R-R2)
          do while(ABS(R-R2)>1e-8)
             T  = log(R)
             R2 = R
             S2 = sqrt( 2.0 * ( ( 1.0 - R )  + R * T ) )
             if (R.lt.1.0) S2 = -S2
             R = R2 - ( S2 - S ) * S2 / T
             if (R.lt.(0.1*R2)) R = 0.1*R2
          end do

          T  = log(R)
          S  = LAM * R + log( &
               sqrt( 2.0 * R *( ( 1.0 - R ) + R * T ) ) / &
               ABS( R - 1.0 ) ) / T
          S = S - 0.0218 / ( S + 0.065 * LAM )
          DEL = 0.01 / S
          S = S + DEL
       endif

       !if x>10, round down to nearest integer, and check accuracy
       !WRITE(*,*) "HERE1", S
       X = floor(S)
       !WRITE(*,*) "HERE2", X
       if (S.gt.10.0.AND.S.lt.(X+2.0*Del)) then

          ! correction procedure based on Temme approximation
          
          if (X.gt.(0.5*LAM).AND.X.lt.(2.0*LAM)) then
             
             XI = 1.0 / X
             ETA = X * LAMI
             ETA = sqrt( 2.0 * (1.0 - ETA + ETA * log(ETA) ) / ETA )
             if (X.gt.LAM) ETA = -ETA

             B1 =  8.0995211567045583e-16
             S = B1      
             B0 = -1.9752288294349411e-15
             S = B0 + S*ETA
             B1 = -5.1391118342426808e-16 + 25.0*B1*Xi
             S = B1 + S*ETA
             B0 =  2.8534893807047458e-14 + 24.0*B0*Xi
             S = B0 + S*ETA
             B1 = -1.3923887224181616e-13 + 23.0*B1*Xi
             S = B1 + S*ETA
             B0 =  3.3717632624009806e-13 + 22.0*B0*Xi
             S = B0 + S*ETA
             B1 =  1.1004392031956284e-13 + 21.0*B1*Xi
             S = B1 + S*ETA
             B0 = -5.0276692801141763e-12 + 20.0*B0*Xi
             S = B0 + S*ETA
             B1 =  2.4361948020667402e-11 + 19.0*B1*Xi
             S = B1 + S*ETA
             B0 = -5.8307721325504166e-11 + 18.0*B0*Xi
             S = B0 + S*ETA
             B1 = -2.5514193994946487e-11 + 17.0*B1*Xi
             S = B1 + S*ETA
             B0 =  9.1476995822367933e-10 + 16.0*B0*Xi
             S = B0 + S*ETA
             B1 = -4.3820360184533521e-09 + 15.0*B1*Xi
             S = B1 + S*ETA
             B0 =  1.0261809784240299e-08 + 14.0*B0*Xi
             S = B0 + S*ETA
             B1 =  6.7078535434015332e-09 + 13.0*B1*Xi
             S = B1 + S*ETA
             B0 = -1.7665952736826086e-07 + 12.0*B0*Xi
             S = B0 + S*ETA
             B1 =  8.2967113409530833e-07 + 11.0*B1*Xi
             S = B1 + S*ETA
             B0 = -1.8540622107151585e-06 + 10.0*B0*Xi
             S = B0 + S*ETA
             B1 = -2.1854485106799979e-06 +  9.0*B1*Xi
             S = B1 + S*ETA
             B0 =  3.9192631785224383e-05 +  8.0*B0*Xi
             S = B0 + S*ETA
             B1 = -0.00017875514403292177 +  7.0*B1*Xi
             S = B1 + S*ETA
             B0 =  0.00035273368606701921 +  6.0*B0*Xi
             S = B0 + S*ETA
             B1 =   0.0011574074074074078 +  5.0*B1*Xi
             S = B1 + S*ETA
             B0 =   -0.014814814814814815 +  4.0*B0*Xi
             S = B0 + S*ETA
             B1 =    0.083333333333333329 +  3.0*B1*Xi
             S = B1 + S*ETA
             B0 =    -0.33333333333333331 +  2.0*B0*Xi
             S = B0 + S*ETA
             S  = S / (1.0 + B1*XI)

             S = S*exp(-0.5*X*ETA*ETA)/sqrt(2.0*3.141592653589793*X)
             if (X.lt.LAM) then
                S = S + 0.5*erfc(ETA*sqrt(0.5*X))
                if (S.gt.U) X = X-1.0
             else
                S = S - 0.5*erfc(-ETA*sqrt(0.5*X))
                if (S.gt.-V) X = X-1.0
             endif

           ! sum downwards or upwards
          else
             XI = 1.0 / X
             S = - (691.0/360360.0)
             S =   (1.0/1188.0) + S*XI*XI
             S = - (1.0/1680.0) + S*XI*XI
             S =   (1.0/1260.0) + S*XI*XI
             S = - (1.0/360.0)  + S*XI*XI
             S =   (1.0/12.0)   + S*XI*XI
             S =                  S*XI
             S = (X - LAM) - X*log(X*LAMI) - S

             if (X.lt.LAM) then
                T  = exp(-0.5*S)
                S  = 1.0 - T*(U*T) * &
                     sqrt(2.0*3.141592653589793*XI) * LAM
                T  = 1.0
                XI = X
                do i=1,50
                   XI = XI - 1.0
                   T  =  T * XI *LAMI
                   S  = S + T
                enddo
                if (S .gt. 0.0) X = X - 1.0
             else
                T  = exp(-0.5*S)
                S  = 1.0 - T*(V*T) * sqrt(2.0*3.141592653589793*X)
                XI = X
                do i=1,50
                   XI = XI + 1.0
                   S   = S*XI*LAMi + 1.0
                enddo
                if (S .lt. 0.0) X = X-1.0
             endif
             
          endif
          
       endif

    endif
    !write(*,*) "HERE X", X
  ! bottom-up summation
    if (X.lt.10.0)then
       X   = 0.0
       T   = exp(0.5*LAM)
       DEL = 0.0
       if (U.gt.0.5) DEL = T*(1e-13*T)
       S   = 1.0 - T*(U*T) + DEL

       do while (S.lt.0.0) 
          X  = X+1.0
          T   = X * LAMi
          DEL = T*DEL
          S   = T*S + 1.0
       enddo
       
       ! top-down summation if needed
       
       if (S .lt. 2.0*DEL) then
          DEL = 1e13*DEL
          T   = 1e17*DEL
          DEL = V*DEL
          
          do while (DEL.lt.T) 
             X   = X + 1.0
             DEL = DEL * X * LAMI
          end do
          
          S = DEL
          T = 1.0
          do while (S.gt.0.0) 
             T =T * X*LAMI
             S =S- T
             X = X-1.0
          end do
       end if
    end if
    

  end subroutine poissinv_core


!//////////////////////////////////////////////////////////////////////
!//                                                                  //
!// The routine below is a DP F90 version of the code in                  //
!//                                                                  //
!// ALGORITHM AS241: APPLIED STATS (1988) VOL. 37, NO. 3, 477-44.    //
!// http://lib.stat.cmu.edu/apstat/241                               //
!//                                                                  //
!// The relative error is less than 1e-15, and the accuracy is       //
!// verified in the accompanying MATLAB code as241.m                 //
!//                                                                  //
!//////////////////////////////////////////////////////////////////////

  subroutine normcdfinv_as241(P, RES)

    double precision, intent(in) :: P
    double precision, intent(out) :: RES

    double precision q, r, num, den

    
    q = p - 0.5

    !write(*,*) q, p, abs(q)
    
    if (ABS(q).le.0.425) then
       r = 0.180625 - q*q

       num =         2.5090809287301226727e+3
       num = r*num + 3.3430575583588128105e+4
       num = r*num + 6.7265770927008700853e+4
       num = r*num + 4.5921953931549871457e+4
       num = r*num + 1.3731693765509461125e+4
       num = r*num + 1.9715909503065514427e+3
       num = r*num + 1.3314166789178437745e+2
       num = r*num + 3.3871328727963666080e0

       den =         5.2264952788528545610e+3
       den = r*den + 2.8729085735721942674e+4
       den = r*den + 3.9307895800092710610e+4
       den = r*den + 2.1213794301586595867e+4
       den = r*den + 5.3941960214247511077e+3
       den = r*den + 6.8718700749205790830e+2
       den = r*den + 4.2313330701600911252e+1
       den = r*den + 1.0000000000e+00

       RES = q * num / den 

    else

       if (q.lt.0.0)then
          r = p
       else
          r = 1.0 - p
       end if

       r = sqrt(-log(r))
    
       if (r.le.5.0)then
          r = r - 1.6

          num =         7.74545014278341407640e-4
          num = r*num + 2.27238449892691845833e-2
          num = r*num + 2.41780725177450611770e-1
          num = r*num + 1.27045825245236838258e0
          num = r*num + 3.64784832476320460504e0
          num = r*num + 5.76949722146069140550e0
          num = r*num + 4.63033784615654529590e0
          num = r*num + 1.42343711074968357734e0

          den =         1.05075007164441684324e-9
          den = r*den + 5.47593808499534494600e-4
          den = r*den + 1.51986665636164571966e-2
          den = r*den + 1.48103976427480074590e-1
          den = r*den + 6.89767334985100004550e-1
          den = r*den + 1.67638483018380384940e0
          den = r*den + 2.05319162663775882187e0
          den = r*den + 1.0000000000e+00

          RES = num / den
       else
          r = r - 5.0

          num =         2.01033439929228813265e-7
          num = r*num + 2.71155556874348757815e-5
          num = r*num + 1.24266094738807843860e-3
          num = r*num + 2.65321895265761230930e-2
          num = r*num + 2.96560571828504891230e-1
          num = r*num + 1.78482653991729133580e0
          num = r*num + 5.46378491116411436990e0
          num = r*num + 6.65790464350110377720e0

          den =         2.04426310338993978564e-15
          den = r*den + 1.42151175831644588870e-7
          den = r*den + 1.84631831751005468180e-5
          den = r*den + 7.86869131145613259100e-4
          den = r*den + 1.48753612908506148525e-2
          den = r*den + 1.36929880922735805310e-1
          den = r*den + 5.99832206555887937690e-1
          den = r*den + 1.0000000000e+00
          
          RES = num / den
       endif

       if (q.lt.0.0)RES = -RES
       
    endif

  end subroutine normcdfinv_as241

end module apetool_inv
