! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! 20210220 IMS: I modelled the new arrangements on those of evtool.
program catmerge_main
    use param
    use catprep_mod
    use catprep_srclist_mod
    use erolib_general_mod
    use erolib_str_mod

    implicit none

    character(*), parameter :: functionName = "catmerge_main"
    integer :: status, ierr
    integer :: ii,temp_n_bands,n_bands
    type(tCatprep_merge) :: task
    type(t_catprep_file), allocatable, dimension(:) :: catfiles
    type(t_catprep_file) :: catfile_out
    character(buffsize) :: i_str0,i_str1,i_str2,msg

    status = ok

      ! read/open the .par file
    call opast(status)
    if(status /= ok) call wrstop(functionName, "error read par file")

    ! read the input paramters
    call wrlog(functionName, "reading parameters")
    call task%read_params(status)
    ! validate the input parameters
    call wrlog(functionName, "validate parameters")
    call task%validate(status)
    ! sanity check
    if(status /= ok) then
        call wrstop(functionName, "input parameters are invalid")
    else
        call wrlog(functionName, "parameters - valid")
    end if

    call alloc_catfiles(catfiles, size(task%infile), status)

    call wrlog(functionName, "Reading infiles.")
    do ii = 1, size(task%infile)
      write(msg,'(a,i2,a)') "File ", ii, ":"
      call wrchatty(functionName, msg)

      call catfiles(ii)%read_n_bands(task%infile(ii), temp_n_bands, status)
      if(ii==1) then
        n_bands = temp_n_bands
      else
        if(temp_n_bands /= n_bands) then
          i_str0 = int2str(ii)
          i_str1 = int2str(temp_n_bands)
          i_str2 = int2str(n_bands)
call wrstop(functionName, "File "//trim(i_str0)//" has "//trim(i_str1)&
  //" energy bands, but previous have "//trim(i_str2)//".")
        end if
      end if

      call catfiles(ii)%src_list%init(n_bands, status, catfiles(ii)%info)
      call catfiles(ii)%src_list%set_all_useage(0, 1, 0, 0, srclist_group_i_normal)

      call wrchatty(functionName, "  Opening the file.")
      call catfiles(ii)%open(readonly, status, filename=task%infile(ii))

      call wrchatty(functionName, "  Reading the file.")
      call catfiles(ii)%src_list%read(status)
      call wrchatty(functionName, "  Closing the file.")
      call catfiles(ii)%close(status)
      if(status /= ok) then
        call wrerr(functionName, s_err, "ERROR with file: " // trim(task%infile(ii)))
      end if
    end do

    call catfile_out%src_list%init(n_bands, status, catfile_out%info)
    call catfile_out%src_list%set_all_useage(0, 0, 1, 0, srclist_group_i_normal)

    if(size(catfiles) > 1) call wrlog(functionName, "Merging infiles.")
    call merge_catfiles(catfiles, catfile_out, status) ! also frees each of the input extensions after use.

    call dealloc_catfiles(catfiles, status)

    ! create outfile
    call wrchatty(functionName, "Create outfile.")
    call catfile_out%open(create, status, filename=task%outfile, clobber=task%clobber)

    call wrlog(functionName, "Writing outfile.")
    call catfile_out%src_list%write(status)

    call wrchatty(functionName, "Close outfile")
    call catfile_out%close(status)

    call cpast (status, ierr)

end program catmerge_main
