! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module catprep_compcat_mod
  use param
  use erolib_bintable_mod
!  use erolib_region_mod
  use erolib_str_mod
  use catprep_srclist_mod

  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr

  implicit none

logical, private, parameter :: doTest=.false.

    ! extnames used in the files
!    character(*), parameter, private :: CompCat_ext = "CATSRCL"

  ! basic type of the catalog used for the cumulative list
  type, extends(tBinTable) :: tCompCat
    type(stdhdr_type) :: std_header
    character(buffsize) :: default_table_name = "CATSRCL"
        integer :: nsouuid ! header keyword
!        integer :: nelements_used ! number of elements filled
!        integer :: nelements ! number of elements (rows)

  contains
    procedure :: init => init_tCompCat
    procedure :: flag_region_ra_dec => flag_region_ra_dec_tCompCat
    procedure :: addnew => addnew_tCompCat
    procedure :: update => update_tCompCat
  end type tCompCat

  type, extends(tFitsFile) :: t_comp_cat_file
    type(tCompCat) :: comp_cat
  contains
  end type t_comp_cat_file

contains

  !.....................................................................
  subroutine init_tCompCat(this, status, fileInfo)
    implicit none

    class(tCompCat)                                 :: this
    integer,                           intent(inout) :: status
    type(tFitsInfo), optional, target, intent(in)    :: fileInfo

    character(*), parameter :: functionName = "init_tCompCat"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    this%info%tableName = this%default_table_name!"CATSRCL"

    call this%initColumn("A", "souUID", num_chars=14)
    call this%initColumn("A", "IAUNAM", num_chars=32)
    call this%initColumn("A", "detUID", num_chars=32)
    call this%initColumn("D", "RA", "[deg]")
    call this%initColumn("D", "DEC", "[deg]")
    call this%initColumn("E", "RADEC_ERR", "[sec]") ! 1 sigma error
    call this%initColumn("E", "ML_FLUX_0") !   [cgs]              Source flux 2)
    call this%initColumn("E", "ML_FLUX_ERR_0") !   [cgs]              1 sigma error 2)
    call this%initColumn("A", "COMMENT", num_chars=32) ! [] comment
    call this%initColumn("B", "NEW") ! indicating if a source was added in the previous run
!        type(i4_tColumn) :: active         ! currently not used !DISCUSS indicating if a source was updated in the previous run
    call this%initColumn("J", "ID_CLUSTER") ! indicating sources which have been detected in the same ermldet run and had the same id_cluster
    call this%initColumn("E", "EXT", "[sec]") ! if /=0.0_r4 source is extended
    call this%initColumn("E", "EXT_ERR", "[sec]") ! error of the size

    ! auxilliary column which is not written / saved
    call this%initColumn("J", "REGION")
    call this%initColumn("J", "ID") ! representing the row number / original order

    call this%initColumns(status)

  end subroutine init_tCompCat

  !.....................................................................
!  subroutine flag_region_ra_dec_tCompCat(this, ra_min_in, ra_max_in&
!    , dec_min_in, dec_max_in, status, overlap, value)
  subroutine flag_region_ra_dec_tCompCat(this, ra_min_in, ra_max_in&
    , dec_min_in, dec_max_in, overlap, value)
!*** NOTE! This is identical to the routine of same name in catprep_srlist. Merge them together somehow?

    implicit none

    class(tCompCat) :: this
    real(r8),           intent(in)    :: ra_min_in, ra_max_in, dec_min_in, dec_max_in
!    integer,            intent(inout) :: status
    real(r8), optional, intent(in)    :: overlap
!    integer, optional,  intent(in)    :: value
    integer(i4), optional,  intent(in)    :: value

    character(*), parameter :: functionName = "flag_region_ra_dec_tCompCat"
    real(r8) :: dec_min, dec_max
    real(r8), dimension(:), allocatable :: ras, decs, ras_min, ras_max, overlap_ras
    integer :: n_rows,ii
    real(r8), dimension(:), pointer :: ptr_ra=>null(),ptr_dec=>null()
!    real(r4), dimension(:), pointer :: ptr_region=>null()
    integer(i4), dimension(:), pointer :: ptr_region=>null()
    integer(i4) :: old_value

    n_rows = this%getNumRows()
    allocate(ras(n_rows))
    allocate(decs(n_rows))
    allocate(ras_min(n_rows))
    allocate(ras_max(n_rows))
    allocate(overlap_ras(n_rows))

    call this%get_col_ptr("RA",  ptr_ra)
    call this%get_col_ptr("DEC", ptr_dec)

    ras(:) = mod(ptr_ra, 360.0_r8)
    decs(:) = ptr_dec

    nullify(ptr_ra, ptr_dec)

    ras_min(:) = mod(ra_min_in, 360.0_r8)
    ras_max(:) = mod(ra_max_in, 360.0_r8)

    ! potentially correct for negative ra
    where (ras < tiny(0.0_r8))
      ras = ras + 360.0_r8
    end where

    where (ras_min < tiny(0.0_r8))
      ras_min = ras_min + 360.0_r8
    end where

    where (ras_max < tiny(0.0_r8))
      ras_max = ras_max + 360.0_r8
    end where

    if(present(overlap)) then
      overlap_ras = overlap/cos(decs*PI/180.0_r8)
      ! correction center for the ra overlap
      ! adjust the overlap for declination 1 / cos (angle)
      ras_min = ras_min - overlap_ras
      ras_max = ras_max + overlap_ras
      dec_min = dec_min_in - overlap
      dec_max = dec_max_in + overlap
    else
      ! ra_min / ra_max remain unchanged besides the corrections for negative values and mod 360
      dec_min = dec_min_in
      dec_max = dec_max_in
    end if

!    call this%get_col_ptr("region", ptr_region)
    call this%get_col_ptr("REGION", ptr_region)

    do ii = 1, n_rows
      ! differ two cases to replace the old region value with a custom value or 1 if it is within a region
      if(present(value)) then
        ! since we are on a sphere and 0/360 is the same we have
        ! two cases ra_min > ra_ max and ra_min < ra_max
        if(ras_min(ii) < ras_max(ii)) then
          old_value = ptr_region(ii)
          ptr_region(ii) = value
          if(ras(ii) < ras_min(ii)) ptr_region(ii) = old_value
          if(ras(ii) > ras_max(ii)) ptr_region(ii) = old_value
          if(decs(ii) < dec_min)     ptr_region(ii) = old_value
          if(decs(ii) > dec_max)     ptr_region(ii) = old_value
        else
          if(ras(ii) >= ras_min(ii)) ptr_region(ii) = value
          if(ras(ii) <= ras_max(ii)) ptr_region(ii) = value
          if(decs(ii) >= dec_min)     ptr_region(ii) = value
          if(decs(ii) <= dec_max)     ptr_region(ii) = value
        end if

      else
        if(ras_min(ii) < ras_max(ii)) then
          old_value = ptr_region(ii)
          ptr_region(ii) = 1_i4
          if(ras(ii) < ras_min(ii)) ptr_region(ii) = old_value
          if(ras(ii) > ras_max(ii)) ptr_region(ii) = old_value
          if(decs(ii) < dec_min)     ptr_region(ii) = old_value
          if(decs(ii) > dec_max)     ptr_region(ii) = old_value
        else
          if(ras(ii) >= ras_min(ii)) ptr_region(ii) = 1_i4
          if(ras(ii) <= ras_max(ii)) ptr_region(ii) = 1_i4
          if(decs(ii) >= dec_min)     ptr_region(ii) = 1_i4
          if(decs(ii) <= dec_max)     ptr_region(ii) = 1_i4
        end if
      end if
    end do

    nullify(ptr_region)

  end subroutine flag_region_ra_dec_tCompCat


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> add the new sources in a catalog to the existing comparison catalog
    !
    !
    !> @param[in] this
    !> @param[in] CatSrcList
    !> @param[inout] status
    !---------------------------------------------------------------------------

  subroutine addnew_tCompCat(this, catSrcList, status)
    implicit none

    class(tCompCat) :: this
    type(tCatSrcList), intent(inout) :: catSrcList
    integer,            intent(inout) :: status

    character(*), parameter :: functionName = "addnew_tCompCat"
    integer :: n_catlist_rows,ii,n_to_add,counter,nsouuid,n_compcat_rows
    integer :: skyfield
    integer(i4), dimension(:), pointer :: ptr_cat_new=>null()
    type(tCompCat) :: temp_comp_cat
    character(buffsize), dimension(:), pointer :: ptr_comp_souuid=>null(),ptr_cat_souuid=>null()

    character(buffsize), dimension(:), pointer :: ptr_comp_iaunam=>null(),ptr_comp_detuid=>null(),ptr_comp_comment=>null()
    real(r8), dimension(:), pointer :: ptr_comp_ra=>null(),ptr_comp_dec=>null()
    real(r4), dimension(:), pointer :: ptr_comp_rderr=>null(),ptr_comp_flux=>null(),ptr_comp_ferr=>null()
    real(r4), dimension(:), pointer :: ptr_comp_ext=>null(),ptr_comp_xerr=>null()
    integer(i1), dimension(:), pointer :: ptr_comp_new=>null()
    integer(i4), dimension(:), pointer :: ptr_comp_cluster=>null()
!,ptr_comp_region=>null(),ptr_comp_id=>null()

    character(buffsize), dimension(:), pointer :: ptr_cat_detuid=>null()
    real(r8), dimension(:), pointer :: ptr_cat_ra=>null(),ptr_cat_dec=>null()
    real(r4), dimension(:), pointer :: ptr_cat_rderr=>null()
    real(r4), dimension(:), pointer :: ptr_cat_ext=>null(),ptr_cat_xerr=>null()
    integer(i4), dimension(:), pointer :: ptr_cat_cluster=>null()
    real(r4), dimension(:), pointer :: ptr_cat_flux0=>null(),ptr_cat_ferr0=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    n_catlist_rows = catSrcList%getNumRows()
    if(n_catlist_rows < 1)&
  return

    call catSrcList%get_col_ptr("NEW", ptr_cat_new)
    n_to_add = count(ptr_cat_new==1)
    if(n_to_add < 1) then
      nullify(ptr_cat_new)
  return
    end if

    n_compcat_rows = this%getNumRows()
    call temp_comp_cat%init(status)
    call temp_comp_cat%alloc(status, n_compcat_rows)
    call temp_comp_cat%overwrite_section(this, 1, status)

    call this%realloc(status, n_compcat_rows + n_to_add)
    call this%overwrite_section(temp_comp_cat, 1, status)
    call temp_comp_cat%destroy(status)

!***        counter = this%nelements_used + 1
    skyfield = str2int(catSrcList%std_header%skyfield, status)

    call this%get_col_ptr("souUID", ptr_comp_souuid)
    call catSrcList%get_col_ptr("souUID", ptr_cat_souuid)

    call this%get_col_ptr("detUID",        ptr_comp_detuid)
    call this%get_col_ptr("IAUNAM",        ptr_comp_iaunam)
    call this%get_col_ptr("RA",            ptr_comp_ra)
    call this%get_col_ptr("DEC",           ptr_comp_dec)
    call this%get_col_ptr("RADEC_ERR",     ptr_comp_rderr)
    call this%get_col_ptr("ML_FLUX_0",     ptr_comp_flux)
    call this%get_col_ptr("ML_FLUX_ERR_0", ptr_comp_ferr)
    call this%get_col_ptr("COMMENT",       ptr_comp_comment)
    call this%get_col_ptr("NEW",           ptr_comp_new)
!                this%active%value(counter) = false
    call this%get_col_ptr("ID_CLUSTER",    ptr_comp_cluster)
    call this%get_col_ptr("EXT",           ptr_comp_ext)
    call this%get_col_ptr("EXT_ERR",       ptr_comp_xerr)

    call catSrcList%get_col_ptr("detUID", ptr_cat_detuid)
    call catSrcList%get_col_ptr("RA", ptr_cat_ra)
    call catSrcList%get_col_ptr("DEC", ptr_cat_dec)
    call catSrcList%get_col_ptr("RADEC_ERR", ptr_cat_rderr)
    call catSrcList%get_col_ptr("ML_FLUX_0", ptr_cat_flux0)
    call catSrcList%get_col_ptr("ML_FLUX_ERR_0", ptr_cat_ferr0)
    call catSrcList%get_col_ptr("ID_CLUSTER", ptr_cat_cluster)
    call catSrcList%get_col_ptr("EXT", ptr_cat_ext)
    call catSrcList%get_col_ptr("EXT_ERR", ptr_cat_xerr)

    nsouuid = this%nsouuid
    counter = n_compcat_rows
    do ii = 1, n_catlist_rows
      if(ptr_cat_new(ii) == 1) then
        counter = counter + 1
!                this%nsouuid = this%nsouuid + 1
        nsouuid = nsouuid + 1
        select case (catSrcList%std_header%obs_mode)
        case ("SURVEY")
          write(ptr_comp_souuid(counter), '(A1, I6.6, A1, I5.5)') "s",  skyfield, "_", nsouuid
        case ("POINTING")
          write(ptr_comp_souuid(counter), '(A1, I6.6, A1, I5.5)') "p",  skyfield, "_", nsouuid
        case default
          call FError(functionName, s_err,  "obs_mode "//trim(catSrcList%std_header%obs_mode)//" unknown", status)
  return
        end select

        ptr_cat_souuid(ii) = ptr_comp_souuid(counter)

        ptr_comp_detuid( counter) = ptr_cat_detuid( ii)
        ptr_comp_iaunam( counter) = ""

        ptr_comp_ra(     counter) = ptr_cat_ra(     ii)
        ptr_comp_dec(    counter) = ptr_cat_dec(    ii)
        ptr_comp_rderr(  counter) = ptr_cat_rderr(  ii)

        ptr_comp_flux(   counter) = ptr_cat_flux0(  ii)
        ptr_comp_ferr(   counter) = ptr_cat_ferr0(  ii)
        ptr_comp_comment(counter) = ""

        ptr_comp_new(    counter) = 1
!                this%active%value(counter) = false
        ptr_comp_cluster(counter) = ptr_cat_cluster(ii)

        ptr_comp_ext(    counter) = ptr_cat_ext(    ii)
        ptr_comp_xerr(   counter) = ptr_cat_xerr(   ii)
      end if
    end do

!***        this%nelements_used = counter
    this%nsouuid = nsouuid!***** what is this used for?

    nullify(ptr_cat_detuid)
    nullify(ptr_cat_ra)
    nullify(ptr_cat_dec)
    nullify(ptr_cat_rderr)
    nullify(ptr_cat_flux0)
    nullify(ptr_cat_ferr0)
    nullify(ptr_cat_cluster)
    nullify(ptr_cat_ext)
    nullify(ptr_cat_xerr)

    nullify(ptr_comp_detuid)
    nullify(ptr_comp_iaunam)
    nullify(ptr_comp_ra)
    nullify(ptr_comp_dec)
    nullify(ptr_comp_rderr)
    nullify(ptr_comp_flux)
    nullify(ptr_comp_ferr)
    nullify(ptr_comp_comment)
    nullify(ptr_comp_new)
    nullify(ptr_comp_cluster)
    nullify(ptr_comp_ext)
    nullify(ptr_comp_xerr)

    nullify(ptr_cat_souuid)
    nullify(ptr_comp_souuid)
    nullify(ptr_cat_new)

  end subroutine addnew_tCompCat

  !.....................................................................
  subroutine update_tCompCat(this, catSrcList, status)
    implicit none

    class(tCompCat) :: this
    type(tCatSrcList), intent(in)    :: catSrcList
    integer,            intent(inout) :: status

    character(*), parameter :: functionName = "update_tCompCat"
    integer :: n_rows,ii,counter

    character(buffsize), dimension(:), pointer :: ptr_comp_detuid=>null(),ptr_comp_comment=>null()
    real(r8), dimension(:), pointer :: ptr_comp_ra=>null(),ptr_comp_dec=>null()
    real(r4), dimension(:), pointer :: ptr_comp_rderr=>null(),ptr_comp_flux=>null(),ptr_comp_ferr=>null()
    real(r4), dimension(:), pointer :: ptr_comp_ext=>null(),ptr_comp_xerr=>null()
    integer(i1), dimension(:), pointer :: ptr_comp_new=>null()
    integer(i4), dimension(:), pointer :: ptr_comp_cluster=>null()

    character(buffsize), dimension(:), pointer :: ptr_cat_detuid=>null()
    real(r8), dimension(:), pointer :: ptr_cat_ra=>null(),ptr_cat_dec=>null()
    real(r4), dimension(:), pointer :: ptr_cat_rderr=>null()
    real(r4), dimension(:), pointer :: ptr_cat_ext=>null(),ptr_cat_xerr=>null()
    integer(i4), dimension(:), pointer :: ptr_cat_cluster=>null(),ptr_cat_match=>null()
    real(r4), dimension(:), pointer :: ptr_cat_flux0=>null(),ptr_cat_ferr0=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    n_rows = catSrcList%getNumRows()
    if(n_rows < 1)&
  return

    call this%get_col_ptr("detUID",        ptr_comp_detuid)
    call this%get_col_ptr("RA",            ptr_comp_ra)
    call this%get_col_ptr("DEC",           ptr_comp_dec)
    call this%get_col_ptr("RADEC_ERR",     ptr_comp_rderr)
    call this%get_col_ptr("ML_FLUX_0",     ptr_comp_flux)
    call this%get_col_ptr("ML_FLUX_ERR_0", ptr_comp_ferr)
    call this%get_col_ptr("COMMENT",       ptr_comp_comment)
    call this%get_col_ptr("NEW",           ptr_comp_new)
!                this%active%value(counter) = false
    call this%get_col_ptr("ID_CLUSTER",    ptr_comp_cluster)
    call this%get_col_ptr("EXT",           ptr_comp_ext)
    call this%get_col_ptr("EXT_ERR",       ptr_comp_xerr)

    call catSrcList%get_col_ptr("detUID",        ptr_cat_detuid)
    call catSrcList%get_col_ptr("RA",            ptr_cat_ra)
    call catSrcList%get_col_ptr("DEC",           ptr_cat_dec)
    call catSrcList%get_col_ptr("RADEC_ERR",     ptr_cat_rderr)
    call catSrcList%get_col_ptr("ML_FLUX_0",     ptr_cat_flux0)
    call catSrcList%get_col_ptr("ML_FLUX_ERR_0", ptr_cat_ferr0)
    call catSrcList%get_col_ptr("ID_CLUSTER",    ptr_cat_cluster)
    call catSrcList%get_col_ptr("EXT",           ptr_cat_ext)
    call catSrcList%get_col_ptr("EXT_ERR",       ptr_cat_xerr)
    call catSrcList%get_col_ptr("id_match",      ptr_cat_match)

    do ii = 1, n_rows
      if(ptr_cat_match(ii) > 0) then
        counter = ptr_cat_match(ii)

        ptr_comp_detuid( counter) = ptr_cat_detuid(ii)
        ptr_comp_ra(     counter) = ptr_cat_ra(    ii)
        ptr_comp_dec(    counter) = ptr_cat_dec(   ii)
        ptr_comp_rderr(  counter) = ptr_cat_rderr( ii)

        ptr_comp_flux(   counter) = ptr_cat_flux0( ii)
        ptr_comp_ferr(   counter) = ptr_cat_ferr0( ii)
        ptr_comp_comment(counter) = ""

        ptr_comp_new(    counter) = 0
!                this%active%value(counter) = true
        ptr_comp_cluster(counter) = ptr_cat_cluster(ii)

        ptr_comp_ext(    counter) = ptr_cat_ext(    ii)
        ptr_comp_xerr(   counter) = ptr_cat_xerr(   ii)
      end if
    end do

    nullify(ptr_cat_match)
    nullify(ptr_cat_detuid)
    nullify(ptr_cat_ra)
    nullify(ptr_cat_dec)
    nullify(ptr_cat_rderr)
    nullify(ptr_cat_flux0)
    nullify(ptr_cat_ferr0)
    nullify(ptr_cat_cluster)
    nullify(ptr_cat_ext)
    nullify(ptr_cat_xerr)

    nullify(ptr_comp_detuid)
    nullify(ptr_comp_ra)
    nullify(ptr_comp_dec)
    nullify(ptr_comp_rderr)
    nullify(ptr_comp_flux)
    nullify(ptr_comp_ferr)
    nullify(ptr_comp_comment)
    nullify(ptr_comp_new)
    nullify(ptr_comp_cluster)
    nullify(ptr_comp_ext)
    nullify(ptr_comp_xerr)

  end subroutine update_tCompCat

  !.....................................................................
  subroutine merge_comp_cats(filenames, comp_cat_out, status)
    implicit none

    type(t_comp_cat_file), dimension(:)&
      ,                        intent(in)    :: filenames
    type(tCompCat),           intent(out)   :: comp_cat_out
    integer,                   intent(inout) :: status

    character(*), parameter :: functionName = "merge_comp_cats"
    integer :: nfiles,fi,numRows,totalNumRows
!,nColsUsed,ii

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not. comp_cat_out%isInitialized) then
      call wrerr(functionName, s_err, "Output binary table is not initialized.")
      status = s_errRep
  return
    end if

    nfiles = size(filenames)

    totalNumRows = 0
    do fi=1,nfiles
      numRows = filenames(fi)%comp_cat%getNumRows()
      totalNumRows = totalNumRows + numRows
    end do

!    nColsUsed = filenames(1)%comp_cat%getNumCols()
!    do ii = 1, nColsUsed
!      comp_cat_out%cols(ii)%scal = filenames(1)%comp_cat%cols(ii)%scal
!    end do

    call comp_cat_out%alloc(status, totalNumRows)
    if(totalNumRows<1)&
  return

    totalNumRows = 0
    do fi=1,nfiles
      call comp_cat_out%overwrite_section(filenames(fi)%comp_cat, totalNumRows+1, status)
      numRows = filenames(fi)%comp_cat%getNumRows()
      totalNumRows = totalNumRows + numRows
    end do

  end subroutine merge_comp_cats

end module catprep_compcat_mod
