! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module catprep_match_cats_mod
    use param
    use erolib_general_mod
    use erolib_str_mod
    use erolib_sort_mod
    use erolib_math_mod

    use catprep_srclist_mod
    use catprep_compcat_mod

    implicit none

    type tQMatch_r4
        ! n_ / m_ are the prefixes for information to the n X m Matrix
        integer :: n_nelements, m_nelements
        integer, dimension(:), allocatable :: n_order, m_order
        real(r4), dimension(:, :), allocatable :: value
        type(tCatSrcList) :: cat_N
        type(tCompCat) :: cat_M

    contains
        procedure :: init => init_tQMatch_r4
        procedure :: free => free_tQMatch_r4
        procedure :: match_distance => match_distance_tQMatch_r4 ! calculate values of how likely a match between two sources is based on the distance and apply a cut to sources which are too far apart
        procedure :: determine_sc => determine_sc_tQMatch_r4 ! determine the source confusion for all n source with the m comparison sources based on the predetermined values
        procedure :: sort_sc => sort_sc_tQMatch_r4
        procedure :: sort_best => sort_best_tQMatch_r4
         ! return the id(s) of comparison sources which have been uniquely matched and are none zero (own skyfield but comparison catalog)
         ! this routine is used find the source which need to be updated in the comparison catalog file
        procedure :: get_unique_match_id => get_unique_match_id_tQMatch_r4
    end type tQMatch_r4


contains


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> routine to init a tQMatch_r4 type
    !
    !
    !
    !> @param[inout] this
    !> @param[in] nelements, melements
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tQMatch_r4(this, cat_N, cat_M, status)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this
        type(tCatSrcList) :: cat_N !***** intent(in)???
        type(tCompCat) :: cat_M

        integer :: n_n_rows,n_m_rows,ii

        character(*), parameter :: function = "init_tQMatch_r4"

        if(status > 100 .or. status == s_errRep) return

        this%cat_N = cat_N
        this%cat_M = cat_M

        n_n_rows = cat_N%getNumRows()
        n_m_rows = cat_M%getNumRows()
        this%n_nelements = n_n_rows
        this%m_nelements = n_m_rows

        call alloc(this%n_order, this%n_nelements, status)
        call alloc(this%m_order, this%m_nelements, status)
        call alloc(this%value, this%n_nelements, this%m_nelements, status)

        do ii = 1, this%n_nelements
            this%n_order(ii) = ii
        end do

        do ii = 1, this%m_nelements
            this%m_order(ii) = ii
        end do

    end subroutine init_tQMatch_r4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> routine to free a tQMatch_r4 type
    !
    !
    !
    !> @param[inout] this
    !> @param[in] nelements, melements
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tQMatch_r4(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this

        character(*), parameter :: function = "free_tQMatch_r4"

        if(status > 100 .or. status == s_errRep) return

        this%n_nelements = 0
        this%m_nelements = 0
        call dealloc(this%n_order, status)
        call dealloc(this%m_order, status)
        call dealloc(this%value, status)

        call this%cat_N%destroy(status)
        call this%cat_M%destroy(status)

    end subroutine free_tQMatch_r4

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> calculate the match matrix based on distances tQMatch_r4 type
    !
    !
    !
    !> @param[inout] this
    !> @param[in, optional] nsigma
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine match_distance_tQMatch_r4(this, status, nsigma)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this

        integer :: n_n_rows,n_m_rows,ii

        ! potential cut on the distance in units of sigma
        real(r8), optional, intent(in) :: nsigma

        character(*), parameter :: function = "match_distance_tQMatch_r4"
        real(r8), dimension(:), pointer :: ptr_cat_ra=>null(),ptr_cat_dec=>null()
        real(r4), dimension(:), pointer :: ptr_cat_rderr=>null()
        real(r8), dimension(:), pointer :: ptr_comp_ra=>null(),ptr_comp_dec=>null()
        real(r4), dimension(:), pointer :: ptr_comp_rderr=>null()
        real(r8), dimension(:), allocatable :: dist_r8

        if(status > 100 .or. status == s_errRep) return

        call alloc(dist_r8, this%n_nelements, status)

        n_n_rows = this%cat_N%getNumRows()
        if(n_n_rows < 1) then
            call wrerr(function, warn, "the input catalog has zero elements no match possible")
    return
        end if

        n_m_rows = this%cat_M%getNumRows()
        if(n_m_rows < 1) then
            call wrerr(function, warn, "the comparison catalog has zero elements no matches can be determined")
    return
        end if

        call this%cat_N%get_col_ptr("RA",        ptr_cat_ra)
        call this%cat_N%get_col_ptr("DEC",       ptr_cat_dec)
        call this%cat_N%get_col_ptr("RADEC_ERR", ptr_cat_rderr)
        call this%cat_M%get_col_ptr("RA",        ptr_comp_ra)
        call this%cat_M%get_col_ptr("DEC",       ptr_comp_dec)
        call this%cat_M%get_col_ptr("RADEC_ERR", ptr_comp_rderr)

        ! 1. calculate distances between sources
        do ii = 1, this%n_nelements
            dist_r8 = distance_gc_multi(ptr_cat_ra(ii), ptr_cat_dec(ii), ptr_comp_ra, ptr_comp_dec, true)
            this%value(ii, :) = real(dist_r8, kind=r4)
            ! convert distance into arcseconds
            this%value(ii, :) = this%value(ii, :)*3600.0_r4
            this%value(ii, :) = this%value(ii, :)&
              /sqrt(ptr_cat_rderr(ii)*ptr_cat_rderr(ii) + ptr_comp_rderr*ptr_comp_rderr)
        end do

        nullify(ptr_comp_ra,ptr_comp_dec,ptr_comp_rderr)
        nullify(ptr_cat_ra,ptr_cat_dec,ptr_cat_rderr)

        ! apply sigma cut
        if(present(nsigma)) then
            where (this%value > nsigma)
                this%value = 0
            elsewhere
                this%value = exp(-this%value/2.0)
            end where
        end if


    end subroutine match_distance_tQMatch_r4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> calculate the source confusion number for each input source N in cat_N of the tQMatch_r4 type
    !> flag new sources and set ids for uniquely matched sources
    !
    !
    !> @param[inout] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine determine_sc_tQMatch_r4(this, status)
        implicit none

        class(tQMatch_r4) :: this
        integer, intent(inout) :: status

        character(*), parameter :: function = "determine_sc_tQMatch_r4"
        integer :: ii, jj,n_n_rows,n_m_rows,max_n_matches
        integer :: sc_m
!, m_nelements
        integer, dimension(:,:), allocatable :: sc_nm
        real(r8), dimension(:), allocatable :: value
        integer, allocatable, dimension(:) :: order
        integer, allocatable, dimension(:,:) :: order2d
        character(buffsize) :: i_str

        integer(i4), dimension(:), pointer :: ptr_cat_region=>null(),ptr_cat_new=>null()
        integer(i4), dimension(:), pointer :: ptr_cat_unique=>null(),ptr_cat_souconf=>null()
        integer(i4), dimension(:), pointer :: ptr_cat_match=>null()
        character(buffsize), dimension(:), pointer :: ptr_cat_souuid=>null(),ptr_cat_iaunam=>null()
        real(r8), dimension(:), pointer :: ptr_cat_squal=>null()

        integer(i4), dimension(:), pointer :: ptr_comp_match=>null()
        character(buffsize), dimension(:), pointer :: ptr_comp_souuid=>null(),ptr_comp_iaunam=>null()

        if(status > 100 .or. status == s_errRep) return

        n_n_rows = this%cat_N%getNumRows()
        if(n_n_rows < 1) then
            call wrerr(function, warn, "the input catalog has zero elements no match possible")
    return
        end if

        call this%cat_N%get_col_ptr("region", ptr_cat_region)
        call this%cat_N%get_col_ptr("NEW",    ptr_cat_new)
        call this%cat_N%get_col_ptr("UNIQUE", ptr_cat_unique)

        n_m_rows = this%cat_M%getNumRows()
        if(n_m_rows < 1) then
            call wrerr(function, warn, "the comparison catalog has zero elements treat all found sources as new")
            do ii = 1, n_n_rows
                if(ptr_cat_region(ii) == true) then
                    ptr_cat_new(ii) = true
                    ptr_cat_unique(ii) = true
                end if
            end do
            return
        end if

        ! create a table with found matches set to 1

        allocate(sc_nm(n_n_rows, n_m_rows))

        sc_nm(:, :) = 0
        where (this%value > 0)
            sc_nm = 1
        end where

        ! determine for each source in the comparison catalog the number of potential source identifications
        ! set the elements in the SC matrix to this number
        ! all matches of any input source with the source M will be set to the determined SC
        !
        !   0101    0201
        !   0100  > 0200
        !   1000    1000
        do jj = 1, this%m_nelements
            sc_m = sum(sc_nm(:, jj))
            sc_nm(:,jj) = sc_nm(:,jj) * sc_m
        end do

        call this%cat_N%get_col_ptr("SOUCONF",  ptr_cat_souconf)
        call this%cat_N%get_col_ptr("id_match", ptr_cat_match)
        call this%cat_N%get_col_ptr("souUID",   ptr_cat_souuid)
        call this%cat_N%get_col_ptr("IAUNAM",   ptr_cat_iaunam)

        call this%cat_M%get_col_ptr("id_match", ptr_comp_match)
        call this%cat_M%get_col_ptr("souUID",   ptr_comp_souuid)
        call this%cat_M%get_col_ptr("IAUNAM",   ptr_comp_iaunam)

        ! determine the total SC number for any given input source
        ! by summing over the SCm
        ! 0201      3
        ! 0200 > SC 2
        ! 1000      1
        max_n_matches = this%cat_N%get_n_extend()
        allocate(order2d(this%n_nelements,max_n_matches))
        allocate(value(this%m_nelements))
        do ii = 1, this%n_nelements
            ptr_cat_souconf(ii) = sum(sc_nm(ii, :))

            ! if a source is new the sum equals zero and the source must be inside the region
            if(ptr_cat_souconf(ii) == 0 .and. ptr_cat_region(ii) == true) then
                ptr_cat_new(ii) = true
                ptr_cat_unique(ii) = true
                cycle
            end if

            ! if a source is uniquely identified sc must be 1
            ! make a reference of the id value which is only non zero if the source was identified within the same skyfield
            if(ptr_cat_souconf(ii) == 1) then
                order = where_i4(sc_nm(ii, :), [1], status)
                ptr_cat_match( ii) = ptr_comp_match(order(1))
                ptr_cat_souuid(ii) = ptr_comp_souuid(order(1))
                ptr_cat_iaunam(ii) = ptr_comp_iaunam(order(1))
                ! if the identified source is from the "same" skyfield the id match is > 0 and thus source is identified as unique
                if(ptr_cat_match(ii) > 0) then
                    ptr_cat_unique(ii) = true
                end if
            end if

            ! for each source sort the likelyhood for a source match in descending order
            value = this%value(ii, :)
            call sort(value, status, order)
            ! sort is ascending therefore reverse it to descending
            call reverse_array(order)

            order2d(ii,:) = order(:max_n_matches)!******** check that size(order)>max_n_matches!!
        end do

        nullify(ptr_cat_souconf,ptr_cat_match,ptr_cat_souuid,ptr_cat_iaunam)

        ! note the best matches in SOUUID and likelihood
        do jj=1,max_n_matches
          i_str = int2str(jj)
          call this%cat_N%get_col_ptr("souUID_"//trim(i_str),      ptr_cat_souuid)
          call this%cat_N%get_col_ptr("souUID_QUAL_"//trim(i_str), ptr_cat_squal)

          do ii = 1, this%n_nelements
            ptr_cat_souuid(ii) = ptr_comp_souuid(order2d(ii,jj))
            ptr_cat_squal( ii) = this%value(ii, order2d(ii,jj))
            ! go to next source if the "input" source has only one partner
!???? keep/modify this? Help!            if(count(sc_nm(ii,:) >= 1) == jj) cycle
          end do

          nullify(ptr_cat_souuid,ptr_cat_squal)
        end do

        deallocate(order2d)
        deallocate(value)
        deallocate(sc_nm)
        nullify(ptr_cat_region,ptr_cat_new,ptr_cat_unique)
        nullify(ptr_comp_match,ptr_comp_souuid,ptr_comp_iaunam)

    end subroutine determine_sc_tQMatch_r4

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> sort matrix to have the SC number in ascending order
    !
    !
    !> @param[inout] this
    !> @param[in, optional] nsigma
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine sort_sc_tQMatch_r4(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this

        integer(i4), dimension(this%n_nelements) :: sc
        integer(i4), allocatable, dimension(:) :: order

        character(*), parameter :: function = "sort_sc_tQMatch_r4"
        integer :: n_rows
        integer(i4), dimension(:), pointer :: ptr_cat_souconf=>null()

        if(status > 100 .or. status == s_errRep)&
     return

        n_rows = this%cat_N%getNumRows()
        if(n_rows < 1)&
     return

        ! first sort in ascending order of the source confusion number

        call this%cat_N%get_col_ptr("SOUCONF",  ptr_cat_souconf)
        sc = ptr_cat_souconf(:)
        nullify(ptr_cat_souconf)

        call sort(sc, status, order)
        this%n_order = this%n_order(order)
        this%value = this%value(this%n_order, :)
        call this%cat_N%reorder(order, status)

        call dealloc(order, status)

    end subroutine sort_sc_tQMatch_r4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> important: the structure must be first sorted in SC to locate the "starting" sort position
    !> important: this only applies for the case that the start row (and column) are unkown
    !> sort matrix that each source in Cat_N and Cat_M are only "used" once and the best value (max) is written on the diagonal elements
    !> the sorting only uses input sources which have more than one source "partner"
    !
    !
    !> @param[inout] this
    !> @param[in, optional] nsigma
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine sort_best_tQMatch_r4(this, status, startrow, startcol)
        implicit none

        integer, intent(inout) :: status
        class(tQMatch_r4) :: this

        integer :: ii, jj, row, col
        integer, intent(in), optional :: startrow, startcol
        integer, dimension(1) :: start
        integer, dimension(2) :: pos_best
        integer(i4), dimension(this%n_nelements) :: sc

        character(*), parameter :: function = "sort_best_tQMatch_r4"
        integer(i4), dimension(:), pointer :: ptr_cat_souconf=>null()

        if(status > 100 .or. status == s_errRep) return

        ! return the first row for which more than one source was found in the comparison catalog
        call this%cat_N%get_col_ptr("SOUCONF",  ptr_cat_souconf)
        sc = ptr_cat_souconf(:)
        nullify(ptr_cat_souconf)

        start = minloc(sc, mask = (sc > 1))

        if(start(1) < 1) then
            call wrlog(function, "sort best not possible: since no SC > 1 was found")
    return
        end if

        jj = 1

        if(present(startrow)) then
            start(1) = startrow
        end if

        if(present(startcol)) then
            jj = startcol
        end if

        ! -----------------
        ! sort the (sub) matrix
        ! -----------------
        do ii = start(1), this%n_nelements
            pos_best = maxloc(this%value(ii: , jj:))
            row = ii + pos_best(1) - 1
            col = jj + pos_best(2) - 1
            ! swap row
            call swap(this%value(ii , :), this%value(row ,:))
            call swap(this%n_order(ii), this%n_order(row))
            ! swap col
            call swap(this%value(:, jj), this%value(:, col))
            call swap(this%m_order(jj), this%m_order(col))
            jj = jj + 1
            if(jj > this%m_nelements) then
                call wrlog(function, "the number of input sources exceeds the number of sources in the comparison catalog")
                exit
            end if
        end do
        ! -------------------------

        ! apply the changes in order to the catalog lists
        call this%cat_N%reorder(this%n_order, status)
        call this%cat_M%reorder(this%m_order, status)

    end subroutine sort_best_tQMatch_r4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return the ids of unique matches (comparison catsrclist of own skyfield)
    !
    !
    !> @param[inout] this
    !> @param[in, optional] nsigma
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine get_unique_match_id_tQMatch_r4(this, id, status)
        implicit none

        integer, intent(inout) :: status
        integer :: n
        integer, allocatable, dimension(:), intent(out) :: id
        class(tQMatch_r4) :: this

        character(*), parameter :: function = "get_unique_match_id_tQMatch_r4"
        integer(i4), dimension(:), pointer :: ptr_cat_souconf=>null()

        if(status > 100 .or. status == s_errRep) return

id=0 ! 20210826 IMS: there is something screwy here - this is not set to anything sensible. I just set it to zero here to avoid -Wall complaints. I'll also add a stop and see who screams.
call wrstop(function, "I don't think this routine functions as designed.")

        call this%cat_N%get_col_ptr("SOUCONF",  ptr_cat_souconf)
        n = count(ptr_cat_souconf == 1)
        nullify(ptr_cat_souconf)

    end subroutine get_unique_match_id_tQMatch_r4

end module catprep_match_cats_mod
