! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module catprep_mod
  use param
  use erolib_ffits_mod
  use erolib_general_mod
  use erolib_file_mod
  implicit none

logical, private, parameter :: doTest=.false.
logical, private, parameter :: use_test_dirs=.false.

  type tCatprep
    character(buffsize) :: infile
    character(buffsize) :: outfile
    character(2) :: det_alg_str ! two caracter code of detection algorithm
    logical :: clobber
    integer :: p_srcdensity ! X th closest source will be used to determine the srcdensity
    logical :: do_owner_filter
    integer :: owner
    logical :: do_flux_filter
    real(r4) :: flux_threshold ! optional flux threshold for ML_flux_0 and will only be applied to outfile2
    character(buffsize) :: outfile2
    logical :: do_outfile2=.false.
    character(buffsize) :: detuid_prefix
  contains
    procedure, non_overridable :: read_params => read_params_tCatprep
    procedure, non_overridable :: validate => validate_tCatprep
  end type tCatprep

  type tCatprep_merge
    integer :: n_infile
    character(buffsize), dimension(:), allocatable :: infile
    character(buffsize) :: outfile
    logical :: clobber

  contains
    procedure, non_overridable :: read_params => read_params_tCatprep_merge
    procedure, non_overridable :: validate => validate_tCatprep_merge
  end type tCatprep_merge

  type tCatprep_addnew
    character(buffsize) :: srclist
    character(buffsize) :: srclist_out
    character(buffsize) :: catalog
    character(buffsize) :: skymap
    integer :: update
    real(r8) :: cutval
    character(buffsize) :: compcat
      ! clobbering is not supported for this task
  contains
    procedure, non_overridable :: read_params => read_params_tCatprep_addnew
    procedure, non_overridable :: validate => validate_tCatprep_addnew
  end type tCatprep_addnew

contains



  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to read in the parameters for catprep
  !
  !
  !
  !> @param[in] inParam
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_params_tCatprep(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tCatprep) :: this
    character(buffsize) :: pipeline_mode

    character(*), parameter :: function = "read_params_tCatprep"

    if(status > 100 .or. status == s_errRep) return

    call rget ("infile", this%infile, status)
    if(status /= ok) call wrstop(function, "reading infile parameter")

    call rget ("outfile", this%outfile, status)
    if(status /= ok) call wrstop(function, "reading outfile parameter")

    call rget ("det_algo", this%det_alg_str, status)
    if(status /= ok) call wrstop(function, "reading outfile parameter")

    call rget("clobber", this%clobber, status)
    if(status /= ok) call wrstop(function, "reading clobber parameter")

    call rget("outfile2", this%outfile2, status)
    if(status /= ok) call wrstop(function, "reading outfile2 parameter")

    call rget("p_srcdensity", this%p_srcdensity, status)
    if(status /= ok) call wrstop(function, "reading p_srcdensity parameter")

    call rget("detuid_prefix", this%detuid_prefix, status)
    if(status /= ok) call wrstop(function, "reading detuid_prefix parameter")


    ! adjust all read parameters
    this%infile = adjustl(this%infile)
    this%outfile = adjustl(this%outfile)
    this%outfile2 = adjustl(this%outfile2)
    this%detuid_prefix = adjustl(this%detuid_prefix)
    if(trim(this%outfile2) /= "") this%do_outfile2 = .true.

    call rget ("owner", this%owner, status)
    if(status /= ok) call wrstop(function, "reading owner parameter")
    if(this%owner <0 .or. this%owner > 2) call wrstop(function, "owner value is only valid for 0-2")
    if(this%owner == 0) then
      this%do_owner_filter = .false.
    else
      this%do_owner_filter = .true.
    end if

    call rget("flux_threshold", this%flux_threshold, status)
    if(status /= ok) call wrstop(function, "reading outfile owner parameter")

    if(this%flux_threshold > 0.0_r4) then
      this%do_flux_filter = .true.
    else
      this%do_flux_filter = .false.
    end if

    ! check for pipeline mode
!    call getenv("SASS_PARINT", pipeline_mode)
!    if(pipeline_mode == "PIPELINE") then
    call get_environment_variable("SASS_PARINT", pipeline_mode)
    if (trim(pipeline_mode) == "PIPELINE" .or. trim(pipeline_mode) == "pipeline") then
      call wrlog(function, "pipeline mode detected")

      if(this%do_owner_filter .eqv. .true.) then
        if(this%outfile2(2:2) == "b") then
          call wrlog(function, "automatic renaming of outfile2 to match selected sources")
          if(this%owner == 1) this%outfile2(2:2) = "i"
          if(this%owner == 2) this%outfile2(2:2) = "m"
        end if

        ! only generate outfile2 if either inputfilename had both source owners
        ! or if a flux threshold is applied
        if((this%infile(2:2) == "b") .or. (this%do_flux_filter .eqv. .true.)) then
          this%do_outfile2 = .true.
        else
          this%do_outfile2 = .false.
        end if

      end if
    end if

    ! check that outfile2 /= outfile
    if(trim(this%outfile2) == trim(this%outfile)) then
      call wrerr(function, warn, "outfile and outfile2 are the same")
    end if

  end subroutine read_params_tCatprep

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> validate the parameters of catprep
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine validate_tCatprep(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: exists
    class(tCatprep) :: this

    character(*), parameter :: function = "validate_tCatprep"

    if(status > 100 .or. status == s_errRep) return

    if(this%infile == "") then
      call FError(function, s_err, "infile name is invalid", status)
    end if

    call ffits_file_exists(this%infile, exists, status)
    if (exists == 0) then
      call FError(function, s_err, "infile " // trim(this%infile) // " does not exist", status)
    end if

    if(this%outfile /= "") then
              ! verify that neither of the infile / outputfilenames are duplicates
      if(this%infile == this%outfile) then
        call FError(function, s_err, "invalid filenames - infile == outfile", status)
      end if

      call ffits_file_exists(this%outfile, exists, status)
      if (exists /= 0) then
        if(this%clobber .eqv. .true.) then
          call delete_file_fortran(this%outfile, status)
        else
          call FError(function, s_err, "outfile already exists and clobber parameter set to false", status)
        end if
      end if

    else
      call FError(function, s_err, "outfile name " // trim(this%outfile) // " is invalid", status)
    end if
    
  end subroutine validate_tCatprep

     !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to read in the parameters for catprep
  !
  !
  !
  !> @param[in] inParam
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_params_tCatprep_addnew(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tCatprep_addnew) :: this

    character(*), parameter :: function = "read_params_tCatprep_addnew"

    if(status > 100 .or. status == s_errRep) return

    call rget ("srclist", this%srclist, status)
    if(status /= ok) call wrstop(function, "reading srclist parameter")
    this%srclist = adjustl(this%srclist)

    call rget ("srclist_out", this%srclist_out, status)
    if(status /= ok) call wrstop(function, "reading srclist_out parameter")
    this%srclist = adjustl(this%srclist)

    call rget ("cutval", this%cutval, status)
    if(status /= ok) call wrstop(function, "reading cut value parameter")

    call rget ("compcat", this%compcat, status)
    if(status == warn) then
      ! to handle empty strings
      status = ok
    end if
    if(status /= ok) call wrstop(function, "reading compcat parameter")

    if(this%compcat == "") then
      call rget ("skymap", this%skymap, status)
      if(status /= ok) call wrstop(function, "reading skymap parameter")
      call rget ("catalog", this%catalog, status)
      if(status /= ok) call wrstop(function, "reading catalog parameter")

      this%catalog = adjustl(this%catalog)
      this%skymap = adjustl(this%skymap)
    else
      this%compcat = adjustl(this%compcat)
    end if

    ! this is only for test purpose typically it should only be "active" if no compcat is provided manually
    call rget ("update", this%update, status)
    if(status /= ok) call wrstop(function, "reading update parameter")

  end subroutine read_params_tCatprep_addnew

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> validate the parameters of catprep
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine validate_tCatprep_addnew(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: exists
    class(tCatprep_addnew) :: this

    character(*), parameter :: function = "validate_tCatprep_addnew"

    if(status > 100 .or. status == s_errRep) return

    if(this%srclist == "") then
      call FError(function, s_err, "srclist name is invalid", status)
    end if

    call ffits_file_exists(this%srclist, exists, status)
    if (exists == 0) then
      call FError(function, s_err, "srclist " // trim(this%srclist) // " does not exist", status)
    end if

    if(this%srclist_out == "") then
      call FError(function, s_err, "srclist_out name is invalid", status)
    end if

    call ffits_file_exists(this%srclist_out, exists, status)
    if (exists == 1 .or. exists == 2) then
      call FError(function, s_err, "srclist_out " // trim(this%srclist_out) // " already exists", status)
    end if

    if(this%compcat == "") then
      call ffits_file_exists(this%skymap, exists, status)
      if (exists == 0) then
        call FError(function, s_err, "skymap " // trim(this%skymap) // " does not exist", status)
      end if
      if(this%catalog == "") then
        call FError(function, s_err, "catalog name is invalid", status)
      end if
      call ffits_file_exists(this%catalog, exists, status)
      if (exists /= 0) then
        call FError(function, s_err, "catalog " // trim(this%catalog) // " already exists", status)
      end if
    else
      call ffits_file_exists(this%compcat, exists, status)
      if (exists == 0) then
        call FError(function, s_err, "compcat " // trim(this%compcat) // " does not exist", status)
      end if
    end if

  end subroutine validate_tCatprep_addnew

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to read in the parameters for catprep_merge
  !
  !
  !
  !> @param[in] inParam
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine read_params_tCatprep_merge(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tCatprep_merge) :: this

    integer :: icount
    integer :: ninfiles = 100
    character(buffsize) :: filelist_file

    character(*), parameter :: function = "read_params_tCatprep_merge"

    if(status > 100 .or. status == s_errRep) return

    call alloc(this%infile, ninfiles, status)

    call rgeta ("infile", ninfiles, this%infile, icount, status)
    if(status /= ok) call wrstop(function, "reading infile parameter")
    if(icount == 0) then
      call wrstop(function, "no input filename(s)")
    end if

    this%infile = this%infile(:icount)
    this%n_infile = icount

    ! check whether infile is actually a file containing the infiles
    if(icount == 1) then
      if(this%infile(1)(1:1) == "@") then
        filelist_file = this%infile(1)(2:)
        call read_filelist(filelist_file, this%infile, status)
        this%n_infile = size(this%infile)
      end if
    end if

    call rget ("outfile", this%outfile, status)
    if(status /= ok) call wrstop(function, "reading outfile parameter")

    call rget("clobber", this%clobber, status)
    if(status /= ok) call wrstop(function, "reading clobber parameter")

    ! adjust all read parameters
    this%infile = adjustl(this%infile)
    this%outfile = adjustl(this%outfile)

  end subroutine read_params_tCatprep_merge

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> validate the parameters of catprep_merge
  !
  !
  !
  !> @param[in] this
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine validate_tCatprep_merge(this, status)
    implicit none

    integer, intent(inout) :: status
    integer :: exists, ii
    class(tCatprep_merge) :: this

    character(*), parameter :: function = "validate_tCatprep_merge"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(this%infile)

      if(this%infile(ii) == "") then
        call FError(function, s_err, "invalid infile name(s)", status)
      end if

      call ffits_file_exists(this%infile(ii), exists, status)
      if (exists == 0) then
        call FError(function, s_err, "infile " // trim(this%infile(ii)) // " does not exist", status)
      end if
    end do

    if(this%outfile /= "") then
      ! verify that neither of the infile / outputfilenames are duplicates
      do ii = 1, size(this%infile)
        if(this%infile(ii) == this%outfile) then
          call FError(function, s_err, "invalid filenames - one of the infiles == outfile", status)
        end if
      end do

      call ffits_file_exists(this%outfile, exists, status)
      if (exists /= 0) then
        if(this%clobber .eqv. .true.) then
          call delete_file_fortran(this%outfile, status)
        else
          call FError(function, s_err, "outfile already exists and clobber parameter set to false", status)
        end if
      end if

    else
      call FError(function, s_err, "outfile name " // trim(this%outfile) // " is invalid", status)
    end if

  end subroutine validate_tCatprep_merge

end module catprep_mod
