! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! Non-method subroutines:
!	interpret_ASCII_skyfield()
module catprep_skymap_mod

  use erolib_bintable_mod
  use erolib_str_mod
  use erolib_sort_mod
  use erolib_ffits_mod

  implicit none

!logical, private, parameter :: doTest=.false.

  type, extends(tBinTable) :: tSkymap
    character(buffsize) :: default_table_name = "SMAPS"
    integer, private :: max_fields = 9
  contains
    procedure, non_overridable :: init => init_tSkymap
    procedure, non_overridable :: get_num_fields => get_num_fields_tSkymap
    procedure, private         :: get_row_of_skyfield => get_row_of_skyfield_tSkymap
    procedure, non_overridable :: get_skyfield_boundary => get_skyfield_boundary_tSkymap
    procedure, non_overridable :: get_catalog_name => get_catalog_name_tSkymap
    procedure, non_overridable :: get_adjacent_catalog_name => get_adjacent_catalog_name_tSkymap
    procedure, non_overridable :: set_new_catalog_name => set_new_catalog_name_tSkymap
    procedure, non_overridable :: set_sscan => set_sscan_tSkymap
    procedure, non_overridable :: set_lastsscan => set_lastsscan_tSkymap

    procedure, non_overridable :: determine_skyfield => determine_skyfield_tSkymap
    procedure, non_overridable :: check_skyfield => check_skyfield_tSkymap
    procedure, non_overridable :: determine_skyfield_radec_box => determine_skyfield_radec_box_tSkymap
  end type tSkymap

  type, extends(tFitsFile) :: t_skymap_file
    type(tSkymap) :: skymap
  contains
  end type t_skymap_file

contains

  !.....................................................................
  subroutine init_tSkymap(this, status, fileInfo)
!!!extname)
    implicit none

    class(tSkymap) :: this
    integer,                           intent(inout) :: status
    type(tFitsInfo), optional, target, intent(in)    :: fileInfo
!!!    character(*), optional, intent(in) :: extname

    character(*), parameter :: functionName = "init_tSkymap"
    integer :: ii
    character(buffsize) :: i_str

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    this%info%tableName = this%default_table_name

!    if(present(extname)) then
!      this%extname = extname
!    end if

    call this%initColumn("J", "SRVMAP")
    !        call this%initColumn("J", "SMAPNR")
    call this%initColumn("D", "RA_MIN")
    call this%initColumn("D", "RA_MAX")
    call this%initColumn("D", "DE_MIN")
    call this%initColumn("D", "DE_MAX")
    call this%initColumn("I", "N_NBRS")
    call this%initColumn("I", "OWNER")

    do ii=1,this%max_fields
      i_str = int2str(ii)
      call this%initColumn("J", "FIELD"//trim(i_str))
      call this%initColumn("J", "O_FIELD"//trim(i_str))
    end do

    call this%initColumn("A", "DIR", num_chars=512)
    call this%initColumn("A", "FILENAME", num_chars=64)
    call this%initColumn("A", "LS_DIR", num_chars=512)
    call this%initColumn("A", "LS_NAME", num_chars=64)
    call this%initColumn("A", "S_DIR", num_chars=512)
    call this%initColumn("A", "S_NAME", num_chars=64)
    call this%initColumn("D", "S_TSTART")

    call this%initColumns(status)

  end subroutine init_tSkymap

  !.....................................................................
  function get_num_fields_tSkymap(this) result(num_fields)
    ! This is to allow this%max_fields to be private and thus immune from user fiddling.
    implicit none

    class(tSkymap) :: this
    integer         :: num_fields

    num_fields = this%max_fields

  end function get_num_fields_tSkymap

  !.....................................................................
  subroutine get_row_of_skyfield_tSkymap(this, skyfield, row_i, status)
    implicit none

    class(tSkymap) :: this
    integer, intent(in)    :: skyfield
    integer, intent(out)   :: row_i
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "get_row_of_skyfield_tSkymap"
    integer :: n_rows,ii
    logical :: skyfield_was_found
    integer(i4), dimension(:), pointer :: ptr_smapnr=>null()

    row_i = 0 ! just to give it some value.

    if(status > 100 .or. status == s_errRep)&
  return

    n_rows = this%getNumRows()

    if(n_rows < 1) then
      call wrerr(functionName, warn, "No rows in the skymaps table.")
      status = c_err
  return
    end if

    call this%get_col_ptr("SRVMAP", ptr_smapnr)

    skyfield_was_found = .false. ! default
    do ii = 1, n_rows
      if(ptr_smapnr(ii) == int(skyfield, kind=i4)) then
        row_i = ii
        skyfield_was_found = .true.
    exit
      end if
    end do

    nullify(ptr_smapnr)

    if(.not.skyfield_was_found) then
      call wrerr(functionName, warn, "skyfield could not be found in skymaps file")
      status = c_err
    end if

  end subroutine get_row_of_skyfield_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> return the boundary for a specific skyfield
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] ra_min, ra_max, dec_min, dec_max
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_skyfield_boundary_tSkymap(this, skyfield, ra_min, ra_max, dec_min, dec_max, status)
    implicit none

    class(tSkymap) :: this
    integer,  intent(in)    :: skyfield
    real(r8), intent(out)   :: ra_min, ra_max, dec_min, dec_max
    integer,  intent(inout) :: status

    character(*), parameter :: functionName = "get_skyfield_boundary_tSkymap"
    integer :: row_i
    real(r8), dimension(:), pointer :: ptr_ra_min=>null(),ptr_ra_max=>null()
    real(r8), dimension(:), pointer :: ptr_dec_min=>null(),ptr_dec_max=>null()

    ra_min  = 0.0d0
    ra_max  = 0.0d0
    dec_min = 0.0d0
    dec_max = 0.0d0

    if(status > 100 .or. status == s_errRep)&
  return

    call this%get_row_of_skyfield(skyfield, row_i, status)
    if(status/=0)&
  return

    call this%get_col_ptr("RA_MIN", ptr_ra_min)
    call this%get_col_ptr("RA_MAX", ptr_ra_max)
    call this%get_col_ptr("DE_MIN", ptr_dec_min)
    call this%get_col_ptr("DE_MAX", ptr_dec_max)

    ra_min  = ptr_ra_min( row_i)
    ra_max  = ptr_ra_max( row_i)
    dec_min = ptr_dec_min(row_i)
    dec_max = ptr_dec_max(row_i)

    nullify(ptr_ra_min,ptr_ra_max,ptr_dec_min,ptr_dec_max)

  end subroutine get_skyfield_boundary_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the catalog filenames of the adjacent skyfields to the target skyfield
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_adjacent_catalog_name_tSkymap(this, skyfield, filenames, status)
    implicit none

    class(tSkymap) :: this
    integer,                              intent(in)    :: skyfield
    character(buffsize), allocatable, dimension(:)&
      ,                                   intent(out)   :: filenames
    integer,                              intent(inout) :: status

    character(*), parameter :: functionName = "get_adjacent_catalog_name_tSkymap"
    integer :: row_i,n_nbrs,jj
    integer(i4) :: kk
    integer(i4), dimension(:), pointer :: ptr_field_order=>null()
    integer(i2), dimension(:), pointer :: ptr_n_nbrs=>null()
    character(buffsize), dimension(:), pointer :: ptr_dir=>null(),ptr_name=>null()
    character(buffsize) :: i_str

    call alloc(filenames, 0, status)

    if(status > 100 .or. status == s_errRep)&
  return

    call this%get_row_of_skyfield(skyfield, row_i, status)
    if(status/=0)&
  return

    call this%get_col_ptr("N_NBRS",   ptr_n_nbrs)
    call this%get_col_ptr("DIR",      ptr_dir)
    call this%get_col_ptr("FILENAME", ptr_name)

    n_nbrs = ptr_n_nbrs(row_i)
    call alloc(filenames, n_nbrs, status)
    do jj=1,n_nbrs
      i_str = int2str(jj)
      call this%get_col_ptr("O_FIELD"//trim(i_str), ptr_field_order)
      kk = ptr_field_order(row_i)
      nullify(ptr_field_order)

      filenames(jj) = trim(ptr_dir(kk))//trim(ptr_name(kk))
    end do

    nullify(ptr_n_nbrs,ptr_dir,ptr_name)

  end subroutine get_adjacent_catalog_name_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the catalog filename for the target skyfield
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_catalog_name_tSkymap(this, skyfield, filename, status)
    implicit none

    class(tSkymap) :: this
    integer,             intent(in)    :: skyfield
    character(buffsize), intent(out)   :: filename
    integer,             intent(inout) :: status

    character(*), parameter :: functionName = "get_catalog_name_tSkymap"
    integer :: row_i
    character(buffsize), dimension(:), pointer :: ptr_dir=>null(),ptr_name=>null()

    filename = ""

    if(status > 100 .or. status == s_errRep)&
  return

    call this%get_row_of_skyfield(skyfield, row_i, status)
    if(status/=0)&
  return

    call this%get_col_ptr("DIR",      ptr_dir)
    call this%get_col_ptr("FILENAME", ptr_name)
    filename = trim(ptr_dir(row_i))//trim(ptr_name(row_i))
    nullify(ptr_dir,ptr_name)

  end subroutine get_catalog_name_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write name and dir of new catalog to skymap
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine set_new_catalog_name_tSkymap(this, skyfield, filename, status)
    implicit none

    class(tSkymap) :: this
    integer,             intent(in)    :: skyfield
    character(buffsize), intent(in)    :: filename
    integer,             intent(inout) :: status

    character(*), parameter :: functionName = "set_new_catalog_name_tSkymap"
    integer :: row_i,l_index
    character(buffsize), dimension(:), pointer :: ptr_dir=>null(),ptr_name=>null()
    character(buffsize) :: cwd

    if(status > 100 .or. status == s_errRep)&
  return

    call this%get_row_of_skyfield(skyfield, row_i, status)
    if(status/=0)&
  return

    l_index = scan(filename, "/", .true.)

    call this%get_col_ptr("DIR",      ptr_dir)
    call this%get_col_ptr("FILENAME", ptr_name)

    if(l_index > 0) then
      ptr_dir(row_i) = trim(filename(:l_index))
      ptr_name(row_i) = trim(filename(l_index+1:))
    else
      call getcwd(cwd, status)
      if(status /= 0) then
        call wrerr(functionName, s_err, "Cannot get current working directory.")
      end if
      ptr_dir(row_i) = trim(cwd) // "/"
      ptr_name(row_i) = trim(filename)
    end if

    nullify(ptr_dir,ptr_name)

  end subroutine set_new_catalog_name_tSkymap

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write name and dir of new catalog to skymap
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine set_lastsscan_tSkymap(this, skyfield, filename, status)
    implicit none

    class(tSkymap) :: this
    integer,             intent(in)    :: skyfield
    character(buffsize), intent(in)    :: filename
    integer,             intent(inout) :: status

    character(*), parameter :: functionName = "set_lastsscan_tSkymap"
    integer :: row_i,l_index
    character(buffsize), dimension(:), pointer :: ptr_dir=>null(),ptr_name=>null()
    character(buffsize) :: cwd

    if(status > 100 .or. status == s_errRep)&
  return

    call this%get_row_of_skyfield(skyfield, row_i, status)
    if(status/=0)&
  return

    l_index = scan(filename, "/", .true.)

    call this%get_col_ptr("LS_DIR",  ptr_dir)
    call this%get_col_ptr("LS_NAME", ptr_name)

    if(l_index > 0) then
      ptr_dir(row_i) = trim(filename(:l_index))
      ptr_name(row_i) = trim(filename(l_index+1:))
    else
      call getcwd(cwd, status)
      if(status /= 0) then
        call wrerr(functionName, s_err, "Cannot get current working directory.")
      end if
      ptr_dir(row_i) = trim(cwd) // "/"
      ptr_name(row_i) = trim(filename)
    end if

    nullify(ptr_dir,ptr_name)

  end subroutine set_lastsscan_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write name and dir of new catalog to skymap
  !
  !
  !
  !> @param[in] this, skyfield
  !> @param[out] skyfield_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

!  subroutine set_sscan_tSkymap(this, skyfield, filename, tstart, status)
  subroutine set_sscan_tSkymap(this, skyfield, filename, status)
    implicit none

    class(tSkymap) :: this
    integer,             intent(in)    :: skyfield
    character(buffsize), intent(in)    :: filename
!    real(r8),            intent(in)    :: tstart!??????????????????? not used.
    integer,             intent(inout) :: status

    character(*), parameter :: functionName = "set_sscan_tSkymap"
    integer :: row_i,l_index
    character(buffsize), dimension(:), pointer :: ptr_dir=>null(),ptr_name=>null()
!!!    real(r8), dimension(:), pointer :: ptr_tstart=>null()
    character(buffsize) :: cwd

    if(status > 100 .or. status == s_errRep)&
  return

    call this%get_row_of_skyfield(skyfield, row_i, status)
    if(status/=0)&
  return

    l_index = scan(filename, "/", .true.)

    call this%get_col_ptr("S_DIR",  ptr_dir)
    call this%get_col_ptr("S_NAME", ptr_name)
!!!    call this%get_col_ptr("S_TSTART", ptr_tstart)

    if(l_index > 0) then
      ptr_dir(row_i) = trim(filename(:l_index))
      ptr_name(row_i) = trim(filename(l_index+1:))
    else
      call getcwd(cwd, status)
      if(status /= 0) then
        call wrerr(functionName, s_err, "Cannot get current working directory.")
      end if
      ptr_dir(row_i) = trim(cwd) // "/"
      ptr_name(row_i) = trim(filename)
    end if
!!!    ptr_tstart(row_i) = tstart

!!!    nullify(ptr_tstart)
    nullify(ptr_dir,ptr_name)

  end subroutine set_sscan_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> determine the skyfield for a specific RA / DEC coordinate
  !
  !
  !
  !> @param[in] this, ra, dec
  !> @param[out] skyfield
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine determine_skyfield_tSkymap(this, ra, dec, skyfield, status, owner)
    ! 20210419 IMS: This is used by datool.
    implicit none

    class(tSkymap)                   :: this
    real(r8),          intent(in)    :: ra, dec
    integer(i4),       intent(out)   :: skyfield
    integer,           intent(inout) :: status
    integer, optional, intent(out)   :: owner

    character(*), parameter :: functionName = "determine_skyfield_tSkymap"
    real(r8) :: ra_intern
    integer :: ii, counter,n_rows
    integer(i4), dimension(:), pointer :: ptr_smapnr=>null()
    integer(i2), dimension(:), pointer :: ptr_owner=>null()
    real(r8), dimension(:), pointer :: ptr_ra_min=>null(),ptr_ra_max=>null()
    real(r8), dimension(:), pointer :: ptr_dec_min=>null(),ptr_dec_max=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    n_rows = this%getNumRows()
    counter = 0

    ra_intern = mod(ra, 360.0_r8)

    if(ra < 0) ra_intern =  ra_intern + 360.0_r8

    if(dec < -90.0_r8 .or. dec > +90.0_r8)&
call wrstop(functionName, "Central region position is outside coordinate system of FK5.")

    call this%get_col_ptr("RA_MIN", ptr_ra_min)
    call this%get_col_ptr("RA_MAX", ptr_ra_max)
    call this%get_col_ptr("DE_MIN", ptr_dec_min)
    call this%get_col_ptr("DE_MAX", ptr_dec_max)
    call this%get_col_ptr("SRVMAP", ptr_smapnr)
    call this%get_col_ptr("OWNER",  ptr_owner)

    do ii = 1, n_rows
      if(ptr_ra_min(ii) <= ra_intern) then
        if(ptr_ra_max(ii) > ra_intern) then
          if(ptr_dec_min(ii) <= dec) then
            if(ptr_dec_max(ii) > dec) then
              skyfield = ptr_smapnr(ii)
              if(present(owner)) owner = int(ptr_owner(ii), 4)
              counter = counter + 1
            end if
          end if
        end if
      end if
    end do

    if(counter /= 1)&
call wrstop(functionName, "Could not determine skyfield - contact developers")

    nullify(ptr_smapnr,ptr_owner)
    nullify(ptr_ra_min,ptr_ra_max,ptr_dec_min,ptr_dec_max)

  end subroutine determine_skyfield_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine check_skyfield_tSkymap.
  !> @brief
  !> check whether a skyfield exists and is mpe
  !> produce comments if skyfield nr does not exists or is iki
  !
  !
  !> @param this, skyfield, exist, comment
  !> @param[in]
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine check_skyfield_tSkymap(this, skyfield, comment, warn_me, status, reject)
    ! 20210419 IMS: This is used indirectly by datool, via interpret_ASCII_skyfield().
    implicit none

    class(tSkymap)                     :: this
    integer,             intent(in)    :: skyfield
    character(buffsize), intent(out)   :: comment
    logical,             intent(out)   :: warn_me
    integer,             intent(inout) :: status
    logical, optional,   intent(out)   :: reject ! optional parameter to determine whether skyfield is "legal" to request

    character(*), parameter :: functionName = "check_skyfield_tSkymap"
    integer, dimension(:), allocatable :: order
    character(buffsize) :: skyfield_str
    integer :: owner
    integer(i4), dimension(:), pointer :: ptr_smapnr=>null()
    integer(i2), dimension(:), pointer :: ptr_owner=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    comment = ""

    call this%get_col_ptr("SRVMAP", ptr_smapnr)
    call this%get_col_ptr("OWNER",  ptr_owner)

    ! find the smapnr matching the skyfield nr
    skyfield_str = int2str(skyfield, status)
    order = where_i4(ptr_smapnr, [skyfield], status)

    if(size(order) /= 1) then
      owner = 3
    else
      owner = ptr_owner(order(1))
    end if

    warn_me = .false.
    select case (owner)
      case (0)
        comment = "selected skyfield belongs to IKI and MPE - " // trim(skyfield_str)
        if(present(reject)) reject = .false.
      case (1)
        comment = "selected skyfield belongs to IKI - " // trim(skyfield_str)
        if(present(reject)) reject = .true.
        warn_me = .true.
      case (2)
        comment = "selected skyfield belongs to MPE - " // trim(skyfield_str)
        if(present(reject)) reject = .false.
      case default
        comment = "could not find skyfield - " // trim(skyfield_str)
        if(present(reject)) reject = .true.
    end select

    call dealloc(order, status)
    nullify(ptr_smapnr,ptr_owner)

  end subroutine check_skyfield_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine get_skyfield_radec_box_tSkymap.
  !> @brief
  !> return the skyfields included in a radec box
  !
  !
  !> @param this, ra_min, dec_min, ra_max, dec_max, skyfield
  !> @param[in]
  !> @param[out, optional] n_iki, n_mpe, n_mixed
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine determine_skyfield_radec_box_tSkymap(this, ra_min, ra_max&
    , dec_min, dec_max, skyfield, mpe_skyfield, warn_me, status, n_iki, n_mpe&
    , n_mixed, comment)
    ! 20210419 IMS: This is used indirectly by datool, via interpret_ASCII_skyfield().
    implicit none

    class(tSkymap)                                    :: this
    real(r8),                           intent(in)    :: ra_min, ra_max, dec_min, dec_max
    integer, allocatable, dimension(:), intent(out)   :: skyfield
    integer, allocatable, dimension(:), intent(out)   :: mpe_skyfield
    logical,                            intent(out)   :: warn_me
    integer,                            intent(inout) :: status
    integer,             optional,      intent(out)   :: n_iki, n_mpe, n_mixed
    character(buffsize), optional,      intent(out)   :: comment

    character(*), parameter :: functionName = "determine_skyfield_radec_box_tSkymap"
    integer(i4), dimension(:), allocatable :: index_ra, index_dec, index_multi
    logical :: add_360 ! modified ra_max
    integer(i4) :: n_skyfields, n_ii, ii ! number of skyfields
    integer :: n_iki_opt,n_mpe_opt,n_mixed_opt,n_rows
    character(10) :: myString
    character(10) :: myString2

    integer(i4), dimension(:), pointer :: ptr_smapnr=>null()
    integer(i2), dimension(:), pointer :: ptr_owner=>null()
    real(r8), dimension(:), pointer :: ptr_ra_min=>null(),ptr_ra_max=>null()
    real(r8), dimension(:), pointer :: ptr_dec_min=>null(),ptr_dec_max=>null()

    n_rows = this%getNumRows()
    if(n_rows < 1)&
call wrstop(functionName, "skymap file has zero elements")

    call alloc(index_ra, n_rows, status)
    call alloc(index_dec, n_rows, status)
    call alloc(index_multi, n_rows, status)

    ! check for ra_min
    if(ra_max < ra_min) then
      add_360 = .true.
    else
      add_360 = .false.
    end if

    index_ra = 0
    index_dec = 0

    call this%get_col_ptr("RA_MIN", ptr_ra_min)
    call this%get_col_ptr("RA_MAX", ptr_ra_max)
    call this%get_col_ptr("DE_MIN", ptr_dec_min)
    call this%get_col_ptr("DE_MAX", ptr_dec_max)
    call this%get_col_ptr("SRVMAP", ptr_smapnr)
    call this%get_col_ptr("OWNER",  ptr_owner)

    do ii = 1, n_rows
      ! check whether the dec coordinates for the skyfield are within the borders
      !original code 
!      if((dec_min <= ptr_dec_min(ii)) .and. (ptr_dec_min(ii) <= dec_max)) index_dec(ii) = 1
!      if((dec_min <= ptr_dec_max(ii)) .and. (ptr_dec_max(ii) <= dec_max)) index_dec(ii) = 1

      ! AG's 8th March fix:
      if((dec_min > ptr_dec_min(ii)) .and. (ptr_dec_max(ii) > dec_min)) index_dec(ii) = 1
      if((dec_max > ptr_dec_min(ii)) .and. (ptr_dec_max(ii) > dec_max)) index_dec(ii) = 1
      if((dec_min < ptr_dec_min(ii)) .and. (ptr_dec_max(ii) < dec_max)) index_dec(ii) = 1

      if(.not.add_360) then
        if((ra_min <= ptr_ra_min(ii)) .and. (ptr_ra_min(ii) <= ra_max)) index_ra(ii) = 1   ! tile is on border of region 
        if((ra_min <= ptr_ra_max(ii)) .and. (ptr_ra_min(ii)<= ra_max)) index_ra(ii) = 1    !tile inside box 
        !bug fix add this line march 8
        if((ra_min <= ptr_ra_max(ii)) .and. (ptr_ra_max(ii) <= ra_max)) index_ra(ii) = 1   ! tile is on border of region    NEW MARCH 

      else
        if((ra_min <= ptr_ra_max(ii)) .and. (ptr_ra_max(ii) <= ra_max + 360.0_r8)) index_ra(ii) = 1   ! tile is on border of region    NEW MARCH 8 from above EXPERIMENTAL
        if((ra_min <= ptr_ra_min(ii)) .and. (ptr_ra_min(ii)<= ra_max + 360.0_r8)) index_ra(ii) = 1
        if((ra_min <= ptr_ra_max(ii)) .and. (ptr_ra_min(ii)<= ra_max + 360.0_r8)) index_ra(ii) = 1
        if((ra_min <= ptr_ra_min(ii) + 360.0_r8) .and. (ptr_ra_min(ii) + 360.0_r8 <= ra_max + 360.0_r8)) index_ra(ii) = 1
          ! this does not need to be checked since all skyfields are defined between 0 and 360 and no skyfield has a border crossing 360
          ! if ra_max would be included so does ra_min therefore we do not need to check it
          ! if(ra_min =< ptr_ra_max(ii) + 360.0_r8 =< ra_max + 360.0_r8) index_ra(ii) = 1
      end if
    end do

    index_multi = index_ra * index_dec
    n_skyfields = sum(index_multi)

    n_iki_opt = 0
    n_mpe_opt = 0
    n_mixed_opt = 0
    n_ii = 1

    call alloc(skyfield, n_skyfields, status)
    call alloc(mpe_skyfield, n_skyfields, status)

    do ii = 1, n_rows
      if(index_ra(ii) * index_dec(ii) == 1) then
        skyfield(n_ii) = ptr_smapnr(ii)
        select case (ptr_owner(ii))
          case (0)
            n_mixed_opt = n_mixed_opt + 1
            mpe_skyfield(n_mixed_opt + n_mpe_opt) =  ptr_smapnr(ii)
          case (1)
            n_iki_opt = n_iki_opt + 1
          case (2)
            n_mpe_opt = n_mpe_opt + 1
            mpe_skyfield(n_mixed_opt + n_mpe_opt) =  ptr_smapnr(ii)
          case default
            call wrerr(functionName, warn, "contact the developer - irregular skyfield found")
        end select

        n_ii = n_ii + 1
      end if
    end do

    if(present(n_mixed)) n_mixed = n_mixed_opt
    if(present(n_iki)) n_iki = n_iki_opt
    if(present(n_mpe)) n_mpe = n_mpe_opt

    warn_me = .false.
    if((n_mpe_opt + n_mixed_opt) < 1) then
      warn_me = .true.
      if(present(comment)) then
        comment = "the radec_box :" // trim(double2str(ra_min, status)) // " " // trim(double2str(ra_max, status)) // " "
        comment = trim(comment) // trim(double2str(dec_min, status)) // " " // trim(double2str(dec_max, status)) // " "
        comment = trim(comment) // " does not contain MPE data"
      end if
    end if

    if(n_mixed_opt + n_mpe_opt > 0) then
      call wrchatty(functionName,"Mixed fields + mpe fields")
      mpe_skyfield = mpe_skyfield(1:n_mixed_opt + n_mpe_opt)
    else
      call wrchatty(functionName,"Only mpe fields ")
      call alloc(mpe_skyfield, 0, status)
    end if 

    nullify(ptr_smapnr,ptr_owner)
    nullify(ptr_ra_min,ptr_ra_max,ptr_dec_min,ptr_dec_max)

    !test
    !mpe_skyfield, n_skyfields,
    write(myString,'(I10)') n_mixed_opt
    write(myString2,'(I10)') n_mpe_opt
    call wrchatty(functionName,"n_mixed_opt   "//myString // " n_mpe_opt ="// myString2)
    
    !do ii = 1, (n_mixed_opt + n_mpe_opt) !  this%nelements 
    ! write(myString,'(I10)') ii
    ! write(myString2,'(I10)') mpe_skyfield(ii)
    ! call wrchatty(functionName,"field NB  "//myString // " skyfied ="// myString2)
    !end do 
    if(status > 100 .or. status == s_errRep)&
  return

  end subroutine determine_skyfield_radec_box_tSkymap

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> interprete the content from an ASCII file as skyfields and/or skyfield boxes
  !> the skymap type must already be filled
  !
  !
  !> @param[in] filename, skymap_name
  !> @param[out, optional] comment
  !> @param[out] skyfield
  !> @param[inout,optional] status
  !---------------------------------------------------------------------------

  subroutine interpret_ASCII_skyfield(filename, skyfield, skymap_name, nreject, comment, status)
    ! 20210419 IMS: This is used by datool.
    implicit none

    character(*), intent(in) :: filename
    character(*), parameter :: function = "interpret_ASCII_skyfield"
    character(buffsize) :: iomsg, line, comment_subroutine
    character(buffsize), optional, dimension(:), allocatable, intent(out) :: comment

    character(buffsize), intent(in) :: skymap_name
    character(buffsize), dimension(:), allocatable :: regTok
    type(t_skymap_file) :: skymap_file
    integer, dimension(:), allocatable, intent(out) :: skyfield
    integer :: skyfield_int
    integer, dimension(:), allocatable :: skyfield_all, skyfield_mpe
    integer, intent(out) :: nreject ! number of total rejects

    integer :: unit, exists, iostat, pos
!    integer :: n_iki, n_mpe, n_mixed
    integer, intent(inout) :: status

    real(r8) :: ra_min, ra_max, dec_min, dec_max, switch_dec

    integer :: nlines, ncomment, success
    logical :: reject, warn_me

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) return

    ! check whether the file exists
    call ffits_file_exists(filename, exists, status)
    ! if the file does not exist return
    if (exists == 0) then
      status = s_err
      call wrerr(function, status, trim(filename) // " - does not exist")
      status = s_errRep
      return
    end if

    ! reject always if something went wrong
    reject = .true.

    call skymap_file%skymap%init(status, skymap_file%info)
    call skymap_file%skymap%set_all_useage(0, 1, 0, 0) ! mandatory read only.

    ! read the skymap file
    call skymap_file%open(readonly, status, skymap_name)
    call skymap_file%skymap%read(status)
    call skymap_file%close(status)

    call ffits_get_iounit(unit, status)
    open (unit, file=filename, status="old",action="read", iostat=iostat, iomsg=iomsg)
    call FError(function // " - open", iostat, iomsg, status)

    ! get the number of lines
    nlines = 0
    do
      read(unit, '(A)', iostat=iostat, iomsg=iomsg) line
      ! have to escape the sequence instead to stop here
      if(iostat /= 0) exit
      nlines = nlines + 1
    end do

    call alloc(comment, nlines, status)

    rewind(unit, iostat=iostat)
    if(iostat /= 0) call wrstop(function, "rewind of file " // trim(filename) // " failed")

    ncomment = 0
    nlines = 0
    nreject = 0

    ! since nothing went wrong set reject to false
    reject = .false.
    do
      read(unit, '(A)', iostat=iostat, iomsg=iomsg) line

      ! have to escape the sequence instead to stop here
      !            call FError(function // " - read " // trim(filename), iostat, iomsg, status)
      if(iostat /= 0) exit
      if(status == s_errRep) return

      ! delete the comments marked by #
      line = adjustl(line)
      pos = index(line, "#")
      select case (pos)
        case (0) ! do nothing
        case (1) ! whole line is a comment
          cycle
        case default ! reduce the line to only contain everything before the #
          line = line(1:pos-1)
      end select

      ! ignore empty lines
      pos = verify(line, " ")
      if(pos == 0) then
        cycle
      end if

      ! split input line into multiple substrings
      call strtok(line, " ", regTok)
      success = false

      select case (size(regTok))
        case (1)
          skyfield_int = str2int(trim(line), status)
          call skymap_file%skymap%check_skyfield(skyfield_int, comment_subroutine, warn_me, status, reject=reject)
          if(reject) then
            ncomment = ncomment + 1
            comment(ncomment) = trim(comment_subroutine)
            nreject = nreject + 1
          else
            if(warn_me) then
              ncomment = ncomment + 1
              comment(ncomment) = trim(comment_subroutine)
            end if
            nlines = nlines + 1
            if(nlines == 1) then
              skyfield = [skyfield_int]
            else
              skyfield = [skyfield, skyfield_int]
            end if
          end if
        case (4)
          call str2deg(regTok(1), ra_min, status, convert=.true.)
          call str2deg(regTok(2), ra_max, status, convert=.true.)
          call str2deg(regTok(3), dec_min, status)
          call str2deg(regTok(4), dec_max, status)

          if(status /= ok) then
            ncomment = ncomment + 1
            comment(ncomment) = "cannot interpret line - " // trim(line)
            status = ok
            nreject = nreject + 1
            cycle
          end if

          ! check that dec_min < dec_max
          if(dec_max < dec_min) then
            ncomment = ncomment + 1
            comment(ncomment) = "dec_min > dec_max - switched dec values for - " // trim(line)
            switch_dec = dec_max
            dec_max = dec_min
            dec_min = switch_dec
          end if

          call skymap_file%skymap%determine_skyfield_radec_box(ra_min, ra_max, dec_min, dec_max, skyfield_all, skyfield_mpe, warn_me, status, comment=comment_subroutine)

          if(.not.warn_me) then
            nlines = nlines + 1
            if(nlines == 1) then
              skyfield = skyfield_mpe
            else
              skyfield = [skyfield, skyfield_mpe]
            end if
          else
            ncomment = ncomment + 1
            comment(ncomment) = comment_subroutine
          end if

        case default
          ncomment = ncomment + 1
          comment(ncomment) = "cannot interpret line - " // trim(line)
          nreject = nreject + 1
      end select
    end do

    if(nreject > 0) then
      call alloc(skyfield, 0, status)
    else
      skyfield = skyfield(unique(skyfield, status))
    end if

    skyfield = skyfield(unique(skyfield, status))

    call skymap_file%skymap%destroy(status)

    call dealloc(regTok, status)

    close(unit, iostat=iostat, iomsg=iomsg)
    call ffits_free_iounit(unit, status)
    call FError(function // " - close "// trim(filename), iostat, iomsg, status)

  end subroutine interpret_ASCII_skyfield

end module catprep_skymap_mod
