! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module catprep_srclist_mod

  ! *** NOTE! This module is used also by varichck and specfit.

  use erolib_general_mod
  use erolib_bintable_mod
  use erolib_str_mod
  use erolib_sort_mod
  use erolib_ffits_mod
  use erolib_math_mod, only : PI, distance_gc_multi
  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr
  use catprep_mllist_mod

  implicit none

logical, private, parameter :: doTest=.false.

  type, extends(tBinTable) :: tCatSrcList
    type(stdhdr_type) :: std_header
    integer, private :: n_bands=0
    integer, private :: detuid_size=32
    integer, private :: n_ext_fields=9
    character(buffsize) :: default_table_name = "CATSRCL"
    logical :: created_by_pipeline=.false.
  contains
    procedure :: init => init_tCatSrcList
    procedure, private :: init_extend => init_extend_tCatSrcList
    procedure :: get_n_extend => get_n_extend_tCatSrcList
    procedure, private :: init_specfit => init_specfit_tCatSrcList
    procedure, private :: init_varichck => init_varichck_tCatSrcList
    procedure :: get_n_bands => get_n_bands_tCatSrcList
    procedure :: copy_col_unit => copy_col_unit_tCatSrcList
    procedure :: copy_band_col => copy_band_col_tCatSrcList
    procedure :: copy_noband_col => copy_noband_col_tCatSrcList
    procedure :: copy_col => copy_col_tCatSrcList
    procedure :: load_mllist => load_mllist_tCatSrcList
    procedure :: calc_srcdensity => calc_srcdensity_tCatSrcList
    procedure :: apply_filter_ownerflag => apply_filter_ownerflag_tCatSrcList
    procedure :: apply_filter_flux => apply_filter_flux_tCatSrcList
    procedure :: flag_region_ra_dec => flag_region_ra_dec_tCatSrcList
  end type tCatSrcList

  type, extends(tFitsFile) :: t_catprep_file
    type(tCatSrcList) :: src_list
  contains
    procedure :: read_n_bands => read_n_bands_t_catprep_file
    procedure :: read => read_t_catprep_file
    procedure :: write => write_t_catprep_file
  end type t_catprep_file

  integer, parameter :: n_HR=3

  enum, bind( C )
    enumerator :: srclist_group_i_normal=0, srclist_group_i_internal, srclist_group_i_extend
    enumerator :: srclist_group_i_specfit, srclist_group_i_varichck
  end enum

contains

  !.....................................................................
  subroutine init_tCatSrcList(this, n_bands, status, fileInfo)
    implicit none

    class(tCatSrcList)                              :: this
    integer,                           intent(in)    :: n_bands ! old-style had n_bands==4.
    integer,                           intent(inout) :: status
    type(tFitsInfo), optional, target, intent(in)    :: fileInfo

    character(*), parameter :: functionName = "init_tCatSrcList"
    integer :: ii
    character(buffsize) :: i_str

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    this%info%tableName = this%default_table_name!"CATSRCL"

    this%n_bands = n_bands

    call this%initColumn("J", "ID_SRC")
    call this%initColumn("J", "ID_CLUSTER")
    call this%initColumn("A", "detUID", num_chars=this%detuid_size)!32)
    call this%initColumn("B", "OWNER")
    ! For the ownership code, see
    !
    !	https://erosita.mpe.mpg.de/eROdoc/products/EventFiles_doc.html
    !
    ! In short, only the 2 least-significant bits have meaning. Bit 0
    ! indicates whether MPE owns the source location; bit 1, likewise
    ! for IKI. Somewhat confusingly, a zero value means 'yes', a value
    ! of unity means 'no'. A table of the four combinations is therefore:
    !
    !	OWNER	Meaning
    !	---------------
    !	0	Both MPE and IKI own it (both bits set to 0).
    !	1	IKI owns it.
    !	2	MPE owns it.
    !	3	Nobody owns it.
    !	---------------

    call this%initColumn("D", "RA",        "deg")
    call this%initColumn("E", "RA_LOWERR", "arcsec") ! optional
    call this%initColumn("E", "RA_UPERR",  "arcsec") ! optional
    call this%initColumn("D", "DEC",       "deg")
    call this%initColumn("E", "DEC_LOWERR","arcsec") ! optional
    call this%initColumn("E", "DEC_UPERR", "arcsec") ! optional
    call this%initColumn("E", "RADEC_ERR", "arcsec")
    call this%initColumn("B", "QUALITY")
    call this%initColumn("B", "QF")
    call this%initColumn("D", "LII",       "deg")
    call this%initColumn("D", "BII",       "deg")
    call this%initColumn("E", "DIST_NN")
    call this%initColumn("E", "SRCDENS",   "deg^-2")
    call this%initColumn("D", "TSTART",    "s")
    call this%initColumn("D", "TSTOP",     "s")
    call this%initColumn("E", "EXT",       "arcsec")
    call this%initColumn("E", "EXT_ERR",   "arcsec")
    call this%initColumn("E", "EXT_LOWERR","arcsec") ! optional
    call this%initColumn("E", "EXT_UPERR", "arcsec") ! optional
    call this%initColumn("E", "EXT_LIKE")
    call this%initColumn("E", "ML_RADIUS")
    call this%initColumn("E", "MASKFRAC")
    call this%initColumn("J", "RAWX",      "pixel")
    call this%initColumn("J", "RAWY",      "pixel")

    do ii=1,n_HR
      i_str = int2str(ii)
      call this%initColumn("E", "HR_"//trim(i_str))
      call this%initColumn("E", "HR_"//trim(i_str)//"_ERR")
    end do

    do ii=0,n_bands
      i_str = int2str(ii)
      call this%initColumn("E", "ML_CTS_"        //trim(i_str))
      call this%initColumn("E", "ML_CTS_ERR_"    //trim(i_str))
      call this%initColumn("E", "ML_CTS_LOWERR_" //trim(i_str)) ! optional
      call this%initColumn("E", "ML_CTS_UPERR_"  //trim(i_str)) ! optional
      call this%initColumn("E", "ML_RATE_"       //trim(i_str))
      call this%initColumn("E", "ML_RATE_ERR_"   //trim(i_str))
      call this%initColumn("E", "ML_RATE_LOWERR_"//trim(i_str)) ! optional
      call this%initColumn("E", "ML_RATE_UPERR_" //trim(i_str)) ! optional
      call this%initColumn("E", "ML_FLUX_"       //trim(i_str))
      call this%initColumn("E", "ML_FLUX_ERR_"   //trim(i_str))
      call this%initColumn("E", "ML_FLUX_LOWERR_"//trim(i_str)) ! optional
      call this%initColumn("E", "ML_FLUX_UPERR_" //trim(i_str)) ! optional

      call this%initColumn("E",    "DET_LIKE_"//trim(i_str))
      call this%initColumn("E",      "ML_BKG_"//trim(i_str))

      if(ii>0) then
        call this%initColumn("E", "ML_EXP_"//trim(i_str))
        call this%initColumn("E", "ML_EEF_"//trim(i_str))
        call this%initColumn("E", "VIGNET_"//trim(i_str))

        call this%initColumn("J", "APE_CTS_"   //trim(i_str), "count")
        call this%initColumn("E", "APE_BKG_"   //trim(i_str), "count")
        call this%initColumn("E", "APE_EXP_"   //trim(i_str), "s")
        call this%initColumn("E", "APE_EEF_"   //trim(i_str))
        call this%initColumn("E", "APE_RADIUS_"//trim(i_str), "pixel")
        call this%initColumn("E", "APE_POIS_"  //trim(i_str))
      end if
    end do

    call this%initColumn("J", "region",   group_i=srclist_group_i_internal)
    call this%initColumn("J", "id_match", group_i=srclist_group_i_internal)

    call this%init_extend()
    call this%init_specfit()
    call this%init_varichck()

    call this%initColumns(status)

!!!    call this%initHeader(status)
    call this%header%initKeyword("J", "ERODAY")
    call this%header%initKeywords(status)
    call this%header%check_keywords(status)

  end subroutine init_tCatSrcList

  !.....................................................................
  subroutine init_extend_tCatSrcList(this)
    implicit none

    class(tCatSrcList) :: this

    character(*), parameter :: functionName = "init_extend_tCatSrcList"
    integer :: ii
    character(buffsize) :: i_str

    call this%initColumn("A", "souUID", num_chars=14&
      , group_i=srclist_group_i_extend)
    call this%initColumn("A", "IAUNAM", num_chars=32&
      , group_i=srclist_group_i_extend)
    call this%initColumn("J", "NEW"&
      , group_i=srclist_group_i_extend)
    call this%initColumn("J", "UNIQUE"&
      , group_i=srclist_group_i_extend)
    call this%initColumn("J", "SOUCONF"&
      , group_i=srclist_group_i_extend)
    do ii=1,this%n_ext_fields
      i_str = int2str(ii)
      call this%initColumn("A", "souUID_"//trim(i_str), num_chars=14&
      , group_i=srclist_group_i_extend)
      call this%initColumn("D", "souUID_QUAL_"//trim(i_str)&
      , group_i=srclist_group_i_extend)
    end do

    call this%initColumn("A", "COMMENT", num_chars=32&
      , group_i=srclist_group_i_extend)

  end subroutine init_extend_tCatSrcList

  !.....................................................................
  function get_n_extend_tCatSrcList(this) result(n_ext_fields)
    ! This is to allow this%n_ext_fields to be private and thus immune from user fiddling.
    implicit none

    class(tCatSrcList) :: this
    integer             :: n_ext_fields

    n_ext_fields = this%n_ext_fields

  end function get_n_extend_tCatSrcList

  !.....................................................................
  subroutine init_specfit_tCatSrcList(this)
    implicit none

    class(tCatSrcList) :: this

    character(*), parameter :: functionName = "init_specfit_tCatSrcList"

    call this%initColumn("E", "NH",      "[cm^-2]"&
      , group_i=srclist_group_i_specfit)
    call this%initColumn("E", "NH_U",    "[cm^-2]"&
      , group_i=srclist_group_i_specfit)
    call this%initColumn("E", "NH_L",    "[cm^-2]"&
      , group_i=srclist_group_i_specfit)
    call this%initColumn("E", "PHOTON_U"&
      , group_i=srclist_group_i_specfit)
    call this%initColumn("E", "PHOTON_L"&
      , group_i=srclist_group_i_specfit)
    call this%initColumn("E", "NORM_U",  "[erg/cm^-2]"&
      , group_i=srclist_group_i_specfit)
    call this%initColumn("E", "NORM_L",  "[erg/cm^-2]"&
      , group_i=srclist_group_i_specfit)
    call this%initColumn("E", "PHOTON"&
      , group_i=srclist_group_i_specfit)
    call this%initColumn("E", "NORM",    "[erg/cm^-2]"&
      , group_i=srclist_group_i_specfit)
    call this%initColumn("E", "CHISQ"&
      , group_i=srclist_group_i_specfit)

  end subroutine init_specfit_tCatSrcList

  !.....................................................................
  subroutine init_varichck_tCatSrcList(this)
    implicit none

    class(tCatSrcList) :: this

    character(*), parameter :: functionName = "init_varichck_tCatSrcList"

    call this%initColumn("E", "MEAN_RATE"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("E", "MEAN_ERROR"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("E", "SSIGMA_RMS"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("D", "TIME4_MIN"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("D", "TIME4_MAX"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("E", "RATE1_MIN"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("E", "RATE1_MAX"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("E", "RATE1_MIN_ERR"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("E", "RATE1_MAX_ERR"&
      , group_i=srclist_group_i_varichck)

    call this%initColumn("E", "SSIGMA_RMS_SS"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("E", "AMPL_MAX"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("E", "PFACTOR_MAV"&
      , group_i=srclist_group_i_varichck)
    call this%initColumn("E", "NLV"&
      , group_i=srclist_group_i_varichck)

  end subroutine init_varichck_tCatSrcList

  !.....................................................................
  function get_n_bands_tCatSrcList(this) result(n_bands)
    ! This is to allow this%n_bands to be private and thus immune from user fiddling.
    implicit none

    class(tCatSrcList) :: this
    integer             :: n_bands

    n_bands = this%n_bands

  end function get_n_bands_tCatSrcList

  !.....................................................................
  subroutine copy_col_unit_tCatSrcList(this, mllist, ml_col_name, cp_col_name)
    implicit none

    class(tCatSrcList) :: this
    type(t_ml_list), intent(in) :: mllist
    character(*),    intent(in) :: ml_col_name,cp_col_name

    character(buffsize) :: unit_str

    call mllist%get_col_unit(ml_col_name, unit_str)
    call this%set_col_unit(cp_col_name, unit_str)

  end subroutine copy_col_unit_tCatSrcList

  !.....................................................................
  subroutine copy_band_col_tCatSrcList(this, mllist, src_ids, bi, cp_col_name&
    , ml_col_name)

    implicit none

    class(tCatSrcList)                          :: this
    type(t_ml_list),       intent(in)           :: mllist
    integer, dimension(:), intent(in)           :: src_ids
    integer,               intent(in)           :: bi
    character(*),          intent(in)           :: cp_col_name
    character(*),          intent(in), optional :: ml_col_name

    character(*), parameter :: functionName = "copy_band_col_tCatSrcList"
    character(buffsize) :: local_cp_col_name,local_ml_col_name,bi_str

    if(present(ml_col_name)) then
      local_ml_col_name = ml_col_name
    else
      local_ml_col_name = cp_col_name
    end if

    bi_str = int2str(bi)
    local_cp_col_name = trim(cp_col_name)//"_"//trim(bi_str)

    call this%copy_col(mllist, src_ids, bi, local_cp_col_name, local_ml_col_name)

  end subroutine copy_band_col_tCatSrcList

  !.....................................................................
  subroutine copy_noband_col_tCatSrcList(this, mllist, src_ids, cp_col_name&
    , ml_col_name)

    implicit none

    class(tCatSrcList)                          :: this
    type(t_ml_list),       intent(in)           :: mllist
    integer, dimension(:), intent(in)           :: src_ids
    character(*),          intent(in)           :: cp_col_name
    character(*),          intent(in), optional :: ml_col_name

    character(*), parameter :: functionName = "copy_noband_col_tCatSrcList"

    call this%copy_col(mllist, src_ids, 0, cp_col_name, ml_col_name)

  end subroutine copy_noband_col_tCatSrcList

  !.....................................................................
  subroutine copy_col_tCatSrcList(this, mllist, src_ids, bi, cp_col_name&
    , ml_col_name)

    implicit none

    class(tCatSrcList)                          :: this
    type(t_ml_list),       intent(in)           :: mllist
    integer, dimension(:), intent(in)           :: src_ids
    integer,               intent(in)           :: bi
    character(*),          intent(in)           :: cp_col_name
    character(*),          intent(in), optional :: ml_col_name

    character(*), parameter :: functionName = "copy_col_tCatSrcList"
    character(buffsize) :: local_ml_col_name,i_str,bi_str
    integer :: ml_col_i,n_ml_rows,n_cp_rows,ii,si,ierr
    character(1) :: dataTypeFITS
    integer, dimension(:), allocatable :: all_indices,indices
    logical, dimension(:), allocatable :: mask

    logical, dimension(:), pointer :: ml_is_null=>null(),cp_is_null=>null()
    logical, dimension(:), pointer :: ml_l_ptr=>null(),cp_l_ptr=>null()
    character(buffsize), dimension(:), pointer :: ml_a_ptr=>null(),cp_a_ptr=>null()
    real(r4), dimension(:), pointer :: ml_e_ptr=>null(),cp_e_ptr=>null()
    real(r8), dimension(:), pointer :: ml_d_ptr=>null(),cp_d_ptr=>null()
    integer(i1), dimension(:), pointer :: ml_b_ptr=>null(),cp_b_ptr=>null()
    integer(i2), dimension(:), pointer :: ml_i_ptr=>null(),cp_i_ptr=>null()
    integer(i4), dimension(:), pointer :: ml_j_ptr=>null(),cp_j_ptr=>null()
    integer(i8), dimension(:), pointer :: ml_k_ptr=>null(),cp_k_ptr=>null()

    integer(i4), dimension(:), pointer ::&
       ptr_id_src  =>null()&
      ,ptr_id_inst =>null()&
      ,ptr_id_band =>null()

    if(present(ml_col_name)) then
      local_ml_col_name = ml_col_name
    else
      local_ml_col_name = cp_col_name
    end if

    if(.not.mllist%read_succeeded(local_ml_col_name)) then
      call this%set_col_blank(cp_col_name, TCOL_FILL_NULLS)
  return
    end if

    ml_col_i = mllist%getColNum(local_ml_col_name)
    dataTypeFITS = mllist%cols(ml_col_i)%getDataType()
    select case(dataTypeFITS)
    case("L")
      call mllist%get_col_ptr(local_ml_col_name, ml_l_ptr, ml_is_null)
      call this%get_col_ptr(        cp_col_name, cp_l_ptr, cp_is_null)
    case("A")
      call mllist%get_col_ptr(local_ml_col_name, ml_a_ptr)!, ml_is_null)
      call this%get_col_ptr(        cp_col_name, cp_a_ptr)!, cp_is_null)
    case("E")
      call mllist%get_col_ptr(local_ml_col_name, ml_e_ptr, ml_is_null)
      call this%get_col_ptr(        cp_col_name, cp_e_ptr, cp_is_null)
    case("D")
      call mllist%get_col_ptr(local_ml_col_name, ml_d_ptr, ml_is_null)
      call this%get_col_ptr(        cp_col_name, cp_d_ptr, cp_is_null)
    case("B")
      call mllist%get_col_ptr(local_ml_col_name, ml_b_ptr, ml_is_null)
      call this%get_col_ptr(        cp_col_name, cp_b_ptr, cp_is_null)
    case("I")
      call mllist%get_col_ptr(local_ml_col_name, ml_i_ptr, ml_is_null)
      call this%get_col_ptr(        cp_col_name, cp_i_ptr, cp_is_null)
    case("J")
      call mllist%get_col_ptr(local_ml_col_name, ml_j_ptr, ml_is_null)
      call this%get_col_ptr(        cp_col_name, cp_j_ptr, cp_is_null)
    case("K")
      call mllist%get_col_ptr(local_ml_col_name, ml_k_ptr, ml_is_null)
      call this%get_col_ptr(        cp_col_name, cp_k_ptr, cp_is_null)
    case default
call wrstop(functionName, "Unrecognized choice '"//trim(dataTypeFITS)//"'")
    end select

    call mllist%get_col_ptr("ID_SRC",  ptr_id_src)
    call mllist%get_col_ptr("ID_INST", ptr_id_inst)
    call mllist%get_col_ptr("ID_BAND", ptr_id_band)

    n_ml_rows = mllist%getNumRows()
    n_cp_rows = this%getNumRows()
    allocate(mask(n_ml_rows))

    bi_str = int2str(bi)

    all_indices = [(ii,ii=1,n_ml_rows,1)]
    do si=1,n_cp_rows
      mask = .false.
      where(ptr_id_src==src_ids(si)&
      .and. ptr_id_inst/=0&
      .and. ptr_id_band==bi)&
        mask = .true.

        if(count(mask)/=1) then
          i_str = int2str(src_ids(si))
call wrstop(functionName, "ML list doesn't have exactly 1 inst==1, band=="&
  //trim(bi_str)//" entry for SRC_ID="//trim(i_str))
        end if

      indices = pack(all_indices, mask)

      if(ml_is_null(indices(1)) .and. dataTypeFITS/="A") then
        cp_is_null(si) = .true.
      else
        select case(dataTypeFITS)
        case("L")
          cp_l_ptr(si) = ml_l_ptr(indices(1))
        case("A")
          cp_a_ptr(si) = ml_a_ptr(indices(1))
        case("E")
          cp_e_ptr(si) = ml_e_ptr(indices(1))
        case("D")
          cp_d_ptr(si) = ml_d_ptr(indices(1))
        case("B")
          cp_b_ptr(si) = ml_b_ptr(indices(1))
        case("I")
          cp_i_ptr(si) = ml_i_ptr(indices(1))
        case("J")
          cp_j_ptr(si) = ml_j_ptr(indices(1))
        case("K")
          cp_k_ptr(si) = ml_k_ptr(indices(1))
        case default
call wrstop(functionName, "Unrecognized choice '"//trim(dataTypeFITS)//"'")
        end select
      end if
    end do

    ierr = 0
    call dealloc(indices,     ierr)
    call dealloc(all_indices, ierr)
    call dealloc(mask,        ierr)

    nullify(ptr_id_src,ptr_id_inst,ptr_id_band)

    nullify(ml_is_null,cp_is_null,ml_l_ptr,cp_l_ptr)
    nullify(ml_a_ptr,cp_a_ptr)
    nullify(ml_e_ptr,cp_e_ptr)
    nullify(ml_d_ptr,cp_d_ptr)
    nullify(ml_b_ptr,cp_b_ptr)
    nullify(ml_i_ptr,cp_i_ptr)
    nullify(ml_j_ptr,cp_j_ptr)
    nullify(ml_k_ptr,cp_k_ptr)

    call this%copy_col_unit(mllist, local_ml_col_name, cp_col_name)

  end subroutine copy_col_tCatSrcList

  !.....................................................................
  subroutine load_mllist_tCatSrcList(this, ml_pathname, det_alg_str&
    , detuid_prefix, status, fileInfo)
    implicit none

    class(tCatSrcList)                               :: this
    character(*),                      intent(in)    :: ml_pathname, det_alg_str, detuid_prefix
    integer,                           intent(inout) :: status
    type(tFitsInfo), optional, target, intent(in)    :: fileInfo

    character(*), parameter :: functionName = "load_mllist_tCatSrcList"
    type(t_mllist_file) :: mllist_file
    integer :: n_ml_rows,n_sources,n_bands,bi,index_last_slash,si,rdigit,ldigit
    logical, dimension(:), allocatable :: mask
    integer, dimension(:), allocatable :: src_id_indices,src_ids
    character(buffsize) :: sass_parint,ml_filename
    integer :: max_prefix_size
    character(7) :: msg

    integer(i4), dimension(:), pointer ::&
       ml_ptr_id_src  =>null()&
      ,ml_ptr_id_inst =>null()&
      ,ml_ptr_id_band =>null()&
      ,ptr_id_src     =>null()
    character(buffsize), dimension(:), pointer :: ptr_detuid=>null()

    status = 0

    call mllist_file%info%setFileName(ml_pathname)
    call mllist_file%srclist%init(status, mllist_file%info)
    call mllist_file%srclist%catprep_setup()

    call mllist_file%read(status, ml_pathname)

    this%std_header = mllist_file%srclist%std_header

    n_ml_rows = mllist_file%srclist%getNumRows()
    if(n_ml_rows < 1) then
      call wrerr(functionName, warn, "No entries found in "&
        //trim(ml_pathname)//" - exit without changes.")
      call this%init(0, status, fileInfo)

      call this%set_all_useage(0, 0, 1, 0, srclist_group_i_normal) ! write only.

      ! The s/ware meeting of 1 June 2021 decided not to write the following for the time being.
      call this%ibclr_col_useage("RAWX",    TCOL_WRITE_BIT)
      call this%ibclr_col_useage("RAWY",    TCOL_WRITE_BIT)
      call this%ibclr_col_useage("QUALITY", TCOL_WRITE_BIT)
      call this%ibclr_col_useage("QF",      TCOL_WRITE_BIT)

      call this%alloc(status, 0)
  return
    end if

    ! The ML source list table is expected to have the following structure:
    !	For each id_src value:
    !		1 row with id_inst==0
    !		N+1 rows with id_inst==1, viz:
    !			1 row with id_band==0
    !			N rows, id_band in [1,N]
    !
    ! The number of bands N will be read from the first id_src and an
    ! error will be thrown if any subsequent id_src does not conform to
    ! the above pattern.

    allocate(mask(n_ml_rows))

    ! Get list of unique src_id values:
    !
    call mllist_file%srclist%get_col_ptr("ID_SRC", ml_ptr_id_src)
    src_id_indices = unique(ml_ptr_id_src, status)
    src_ids = ml_ptr_id_src(src_id_indices)
    call sort(src_ids, status) ! just for neatness.
    n_sources = size(src_ids)

    call mllist_file%srclist%get_col_ptr("ID_INST", ml_ptr_id_inst)
    call mllist_file%srclist%get_col_ptr("ID_BAND", ml_ptr_id_band)

    mask = .false.
    where(ml_ptr_id_src==src_ids(1) .and. ml_ptr_id_inst/=0 .and. ml_ptr_id_band/=0)&
      mask = .true.

    nullify(ml_ptr_id_src,ml_ptr_id_inst,ml_ptr_id_band)

    n_bands = count(mask)

    ! Set up the cat source object:
    !
    call this%init(n_bands, status, fileInfo)
    call this%set_all_useage(0, 0, 1, 0, srclist_group_i_normal) ! write only.
!*** any internal-only columns?

    ! The s/ware meeting of 1 June 2021 decided not to write the following for the time being.
    call this%ibclr_col_useage("RAWX",    TCOL_WRITE_BIT)
    call this%ibclr_col_useage("RAWY",    TCOL_WRITE_BIT)
    call this%ibclr_col_useage("QUALITY", TCOL_WRITE_BIT)
    call this%ibclr_col_useage("QF",      TCOL_WRITE_BIT)

    call this%alloc(status, n_sources)

    ! Copy the band-0 columns:
    !
    call this%copy_noband_col(mllist_file%srclist, src_ids, "ID_SRC")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "ID_CLUSTER")
    ! GL 13/09/2021:
    call this%copy_noband_col(mllist_file%srclist, src_ids, "OWNER")
    !
    call this%copy_noband_col(mllist_file%srclist, src_ids, "RA")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "RA_LOWERR")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "RA_UPERR")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "DEC")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "DEC_LOWERR")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "DEC_UPERR")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "RADEC_ERR")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "LII")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "BII")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "DIST_NN")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "EXT")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "EXT_ERR")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "EXT_LOWERR")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "EXT_UPERR")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "EXT_LIKE")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "ML_RADIUS")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "MASKFRAC")
!    call this%copy_noband_col(mllist_file%srclist, src_ids, "RAWX")
!    call this%copy_noband_col(mllist_file%srclist, src_ids, "RAWY")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "HR_1",       "HR1")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "HR_2",       "HR2")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "HR_3",       "HR3")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "HR_1_ERR",   "HR1_ERR")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "HR_2_ERR",   "HR2_ERR")
    call this%copy_noband_col(mllist_file%srclist, src_ids, "HR_3_ERR",   "HR3_ERR")

    do bi=0,n_bands
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_CTS")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_CTS_ERR")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_CTS_LOWERR")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_CTS_UPERR")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_RATE")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_RATE_ERR")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_RATE_LOWERR")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_RATE_UPERR")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_FLUX")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_FLUX_ERR")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_FLUX_LOWERR")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_FLUX_UPERR")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "DET_LIKE")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_BKG")
    end do

    do bi=1,n_bands
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_EXP")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "VIGNET")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "ML_EEF")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "APE_CTS")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "APE_BKG")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "APE_EXP")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "APE_EEF")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "APE_RADIUS")
      call this%copy_band_col(mllist_file%srclist, src_ids, bi, "APE_POIS")
    end do

    call this%get_col_ptr("ID_SRC", ptr_id_src)
    call this%get_col_ptr("detUID", ptr_detuid)
    call get_environment_variable("SASS_PARINT", sass_parint)

    if(sass_parint == "PIPELINE") then
      this%created_by_pipeline = .true.
      index_last_slash = index(ml_pathname, "/", .true.)
      ml_filename = ml_pathname(index_last_slash+1:)
      do si = 1, n_sources
        rdigit = len_trim(ml_filename) - 5 ! -5 because we lop off '.fits'
        ldigit = rdigit - 8 ! last nine characters, which should all be digits
        write(ptr_detuid(si),'(A16,A2,I5.5,A9)') ml_filename(1:16)&
          , det_alg_str, ptr_id_src(si), ml_filename(ldigit:rdigit)
      end do
    else
      max_prefix_size = this%detuid_size - 7
      if(len_trim(detuid_prefix)>max_prefix_size) then
        do si = 1, n_sources
          write(msg,'(A2,I5.5)') det_alg_str, ptr_id_src(si)
          ptr_detuid(si) = detuid_prefix(1:max_prefix_size)//msg
        end do
      else
        do si = 1, n_sources
          write(msg,'(A2,I5.5)') det_alg_str, ptr_id_src(si)
          ptr_detuid(si) = trim(detuid_prefix)//msg
        end do
      end if
    end if

    call dealloc(mask,           status)
    call dealloc(src_id_indices, status)
    call dealloc(src_ids,        status)
    nullify(ptr_id_src,ptr_detuid)

  end subroutine load_mllist_tCatSrcList

  !.....................................................................
  subroutine calc_srcdensity_tCatSrcList(this, nth_source, status)
    implicit none

    class(tCatSrcList) :: this
    integer, intent(in)    :: nth_source
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "calc_srcdensity_tCatSrcList"
    real(r4), parameter :: PI_r4 = real(PI,kind=r4)
    integer :: n_sources,ii
    ! matrix to store all distances
    real(r4), dimension(:,:), allocatable :: dist_NN
    real(r8), dimension(:), pointer :: ptr_ra=>null(),ptr_dec=>null()
    real(r4), dimension(:), pointer :: ptr_dist_nn=>null(),ptr_src_dens=>null()
    real(r8), dimension(:), allocatable :: dist_r8

    if(status > 100 .or. status == s_errRep)&
  return

    if(nth_source < 1) then
call wrstop(functionName, "the number for the Nth closest source is < 1")
    end if

    n_sources = this%getNumRows()
    if(nth_source > n_sources) then
      call wrlog(functionName, "no local density of source can be calculated - too few input sources")
  return
    end if

    allocate(dist_NN(n_sources,n_sources))
    allocate(dist_r8(n_sources))

    call this%get_col_ptr("RA",  ptr_ra)
    call this%get_col_ptr("DEC", ptr_dec)
    call this%get_col_ptr("DIST_NN", ptr_dist_nn)
    call this%get_col_ptr("SRCDENS", ptr_src_dens)

    do ii = 1, n_sources
      dist_NN(ii, ii) = 0.0_r4

      ! exception for nelements = 1
      if(n_sources == 1) then
        ptr_dist_nn(ii) = 0.0_r4
    exit
      end if

      if(ii<n_sources) then
        dist_r8(ii+1:) = distance_gc_multi(ptr_ra(ii), ptr_dec(ii), ptr_ra(ii+1:), ptr_dec(ii+1:), true)
        dist_NN(ii,ii+1: ) = real(dist_r8(ii+1:), kind=r4)
      end if
      dist_NN(:, ii) = dist_NN(ii, :)

      ! matrix will be disturbed at this point by rearranging each row according to the distance
      ! find the closest and Xth furthest source for the distance
      ! sort the row in ascending order
      call sort(dist_NN(ii, :), status)

      ! closest source which is not itself therefore 2 not 1
      ! since ra / dec are provided in deg calculate in [sec]
      ptr_dist_nn(ii) = dist_NN(ii, 2)*3600.0_r4
      ! get the distance to Xth source
      ! including itself

      ! spherical segment 2*PI*(1-cos(THETA)) with THETA being the angle between two rays
      ptr_src_dens(ii) = 2.0_r4*PI_r4*(1.0_r4 - cos(dist_NN(ii, nth_source)*PI_r4/180.0_r4 ))
    end do

    call dealloc(dist_NN, status)
    call dealloc(dist_r8, status)
    nullify(ptr_dist_nn,ptr_src_dens)
    nullify(ptr_ra,ptr_dec)

  end subroutine calc_srcdensity_tCatSrcList

  !.....................................................................
  subroutine apply_filter_ownerflag_tCatSrcList(this, ownerflag, status)
    ! The integer 'ownerflag' comes directly from the command-line
    ! parameter --owner, and has the meaning: filter source catalog to
    ! keep only sources which belong to:
    !	1=IKI
    !	2=mpe
    !	0=nofilter
    !
    ! The first thing I'm going to do is subtract this from 3, which
    ! inverts the bit pattern in the lowest 2 bits. The new table
    ! becomes:
    !
    !	3-ownerflag	meaning:
    !	-------------------------
    !	0b10		Return IKI sources only
    !	0b01		Return MPE sources only
    !	0b11		Return everything.
    !	-------------------------
    !
    ! I do the same to the OWNER column, which then has the meaning:
    !
    !	OWNER	3_i1-OWNER	meaning:
    !	---------------------------------
    !	0b00	   0b11		Owned by both.
    !	0b01	   0b10		Owned by IKI.
    !	0b10	   0b01		Owned by MPE.
    !	0b11	   0b00		Owned by neither.
    !	---------------------------------
    !
    ! The condition for being returned by the filter is then
    !	IAND(3_i1-OWNER, 3-ownerflag)>0.
    ! **NOTE** however that this will not return sources for which
    ! OWNER==3 (i.e., owned by nobody). I am not sure if there are such
    ! sources.
    !
    implicit none

    class(tCatSrcList) :: this
    integer, intent(in)    :: ownerflag
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "apply_filter_ownerflag_tCatSrcList"
    integer :: n_sources
    logical, dimension(:), allocatable :: mask
    integer(i1), dimension(:), pointer :: ptr_owner=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    n_sources = this%getNumRows()
    if(n_sources < 1)&
  return

    call this%get_col_ptr("OWNER", ptr_owner)

    allocate(mask(n_sources))
    mask = .false.
    where(iand((3_i1-ptr_owner(:)), int(3-ownerflag,kind=i1))>0_i1) mask = .true.
    ! Subtracting from 3 is the same as inverting the 2 lowest-significance bits. E.g. 3-2=1 maps 0b10 -> 0b01.

    call this%pack(mask, status)

    deallocate(mask)
    nullify(ptr_owner)

  end subroutine apply_filter_ownerflag_tCatSrcList

  !.....................................................................
  subroutine apply_filter_flux_tCatSrcList(this, flux_th, status)
    implicit none

    class(tCatSrcList) :: this
    real(r4), intent(in) :: flux_th
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "apply_filter_flux_tCatSrcList"
    integer :: n_sources
    logical, dimension(:), allocatable :: mask
    real(r4), dimension(:), pointer :: ptr_flux0=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    n_sources = this%getNumRows()
    if(n_sources < 1)&
  return

    call alloc(mask, n_sources, status)
    mask = .false.

    call this%get_col_ptr("ML_FLUX_0", ptr_flux0)
    where(ptr_flux0 > flux_th) mask = .true.

    call this%pack(mask, status)

    nullify(ptr_flux0)
    call dealloc(mask, status)

  end subroutine apply_filter_flux_tCatSrcList

  !.....................................................................
  subroutine flag_region_ra_dec_tCatSrcList(this, ra_min_in, ra_max_in&
    , dec_min_in, dec_max_in, overlap, value)

    implicit none

    class(tCatSrcList)                :: this
    real(r8),              intent(in) :: ra_min_in, ra_max_in, dec_min_in, dec_max_in
    real(r8),    optional, intent(in) :: overlap
    integer(i4), optional, intent(in) :: value

    character(*), parameter :: functionName = "flag_region_ra_dec_tCatSrcList"
    real(r8) :: dec_min, dec_max
    real(r8), dimension(:), allocatable :: ras,decs,ras_min,ras_max,overlap_ras
    integer :: n_rows,ii,ierr
    real(r8), dimension(:), pointer :: ptr_ra=>null(),ptr_dec=>null()
    integer(i4), dimension(:), pointer :: ptr_region=>null()
    integer(i4) :: old_value

    n_rows = this%getNumRows()
    allocate(ras(n_rows))
    allocate(decs(n_rows))
    allocate(ras_min(n_rows))
    allocate(ras_max(n_rows))
    allocate(overlap_ras(n_rows))

    call this%get_col_ptr("RA",  ptr_ra)
    call this%get_col_ptr("DEC", ptr_dec)

    ras(:) = mod(ptr_ra, 360.0_r8)
    decs(:) = ptr_dec

    nullify(ptr_ra, ptr_dec)

    ras_min(:) = mod(ra_min_in, 360.0_r8)
    ras_max(:) = mod(ra_max_in, 360.0_r8)

    ! potentially correct for negative ra
    where (ras < tiny(0.0_r8))
      ras = ras + 360.0_r8
    end where

    where (ras_min < tiny(0.0_r8))
      ras_min = ras_min + 360.0_r8
    end where

    where (ras_max < tiny(0.0_r8))
      ras_max = ras_max + 360.0_r8
    end where

    if(present(overlap)) then
      overlap_ras = overlap/cos(decs*PI/180.0_r8)
      ! correction center for the ra overlap
      ! adjust the overlap for declination 1 / cos (angle)
      ras_min = ras_min - overlap_ras
      ras_max = ras_max + overlap_ras
      dec_min = dec_min_in - overlap
      dec_max = dec_max_in + overlap
    else
      ! ra_min / ra_max remain unchanged besides the corrections for negative values and mod 360
      dec_min = dec_min_in
      dec_max = dec_max_in
    end if

    call this%get_col_ptr("region", ptr_region)

    do ii = 1, n_rows
      ! differ two cases to replace the old region value with a custom value or 1 if it is within a region
      if(present(value)) then
        ! since we are on a sphere and 0/360 is the same we have
        ! two cases ra_min > ra_ max and ra_min < ra_max
        if(ras_min(ii) < ras_max(ii)) then
          old_value = ptr_region(ii)
          ptr_region(ii) = value
          if(ras(ii) < ras_min(ii)) ptr_region(ii) = old_value
          if(ras(ii) > ras_max(ii)) ptr_region(ii) = old_value
          if(decs(ii) < dec_min)     ptr_region(ii) = old_value
          if(decs(ii) > dec_max)     ptr_region(ii) = old_value
        else
          if(ras(ii) >= ras_min(ii)) ptr_region(ii) = value
          if(ras(ii) <= ras_max(ii)) ptr_region(ii) = value
          if(decs(ii) >= dec_min)     ptr_region(ii) = value
          if(decs(ii) <= dec_max)     ptr_region(ii) = value
        end if

      else
        if(ras_min(ii) < ras_max(ii)) then
          old_value = ptr_region(ii)
          ptr_region(ii) = 1_i4
          if(ras(ii) < ras_min(ii)) ptr_region(ii) = old_value
          if(ras(ii) > ras_max(ii)) ptr_region(ii) = old_value
          if(decs(ii) < dec_min)     ptr_region(ii) = old_value
          if(decs(ii) > dec_max)     ptr_region(ii) = old_value
        else
          if(ras(ii) >= ras_min(ii)) ptr_region(ii) = 1_i4
          if(ras(ii) <= ras_max(ii)) ptr_region(ii) = 1_i4
          if(decs(ii) >= dec_min)     ptr_region(ii) = 1_i4
          if(decs(ii) <= dec_max)     ptr_region(ii) = 1_i4
        end if
      end if
    end do

    nullify(ptr_region)

    ierr = 0
    call dealloc(ras, ierr)
    call dealloc(decs, ierr)
    call dealloc(ras_min, ierr)
    call dealloc(ras_max, ierr)
    call dealloc(overlap_ras, ierr)

  end subroutine flag_region_ra_dec_tCatSrcList

  !=====================================================================
  ! Methods of t_catprep_file:

  !.....................................................................
  subroutine read_n_bands_t_catprep_file(this, filename, n_bands, status)
    implicit none
    class(t_catprep_file) :: this
    character(*), intent(in) :: filename
    integer, intent(out) :: n_bands
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "read_n_bands_t_catprep_file"
    integer :: ierr,bi,dummy_col_num
    character(buffsize) :: i_str

    call this%open(readonly, status, filename=filename)

    ierr = 0
    call FTMNHD(this%info%unitNo, BINARY_TBL, this%src_list%default_table_name, 0, ierr)

    if(ierr==BAD_HDU_NUM) then
      status = c_err
      call wrerr(functionName, c_err, "Extension "&
        //trim(this%src_list%default_table_name)//" not found.")
  return
    else if(ierr /= 0) then
      status = c_err
      call report_FITS_err(functionName, ierr, status, "- extension "&
        //trim(this%src_list%default_table_name))
  return
    end if

    bi = 0
    do
      bi = bi + 1
      i_str = int2str(bi) 
      ierr = 0
      ! Try to read the column number.
      call FTGCNO(this%info%unitNo,.true.,"ML_EXP_"//trim(i_str),dummy_col_num,ierr)
      if (ierr/=0) then
        n_bands = bi - 1
    exit
      end if
    end do

    call this%close(status)

  end subroutine read_n_bands_t_catprep_file

  !.....................................................................
  subroutine read_t_catprep_file(this, status, filename)
    implicit none
    class(t_catprep_file) :: this
    integer,                intent(inout) :: status
    character(*), optional, intent(in)    :: filename

    character(*), parameter :: functionName = "read_t_catprep_file"

    if(present(filename)) call this%open(readonly, status, filename)

!*** could check extension exists. See analogous method of tEventFile.

    call this%src_list%read(status)
    call fread_stdhdr(this%info%unitNo, this%src_list%std_header, status)

    if(present(filename)) call this%close(status)

  end subroutine read_t_catprep_file

  !.....................................................................
  subroutine write_t_catprep_file(this, status, filename, clobber)
    implicit none
    class(t_catprep_file) :: this
    integer,                intent(inout) :: status
    character(*), optional, intent(in)    :: filename
    logical,      optional, intent(in)    :: clobber

    character(*), parameter :: functionName = "write_t_catprep_file"
integer :: n_cols,ii

    if(present(filename)) call this%open(create, status, filename, clobber)

!*** 20210526 IMS: should not have to do this! >:8
n_cols = this%src_list%getNumCols()
do ii=1,n_cols
  this%src_list%cols(ii)%writeNulls = .true.
end do

    call this%src_list%write(status)
    call fwrite_stdhdr(this%info%unitNo, this%src_list%std_header, status)

    if(present(filename)) call this%close(status)

  end subroutine write_t_catprep_file

  !=====================================================================
  ! Not methods, just utility subroutines:

  !.....................................................................
  subroutine dealloc_catfiles(catfiles, status)
    implicit none

    type(t_catprep_file), dimension(:), allocatable :: catfiles
    integer, intent(inout) :: status

    integer :: stat,fi
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "dealloc_catfiles"

    if(allocated(catfiles)) then
      do fi=1,size(catfiles)
        call catfiles(fi)%src_list%dealloc(status)
      end do

      errmsg=""
      deallocate(catfiles, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
    end if

  end subroutine dealloc_catfiles

  !.....................................................................
  subroutine alloc_catfiles(catfiles, n, status)
    implicit none

    type(t_catprep_file), allocatable, dimension(:) :: catfiles
    integer, intent(in) :: n
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "alloc_catfiles"
    integer :: stat
    character(buffsize) :: errmsg

    call dealloc_catfiles(catfiles, status)
    errmsg=""
    allocate(catfiles(n), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

  end subroutine alloc_catfiles

  !.....................................................................
  subroutine merge_catfiles(catfiles, catfile_out, status)
    implicit none
    type(t_catprep_file), dimension(:), intent(in) :: catfiles
    type(t_catprep_file),               intent(inout) :: catfile_out
    integer,                            intent(inout) :: status

    character(*), parameter :: functionName = "merge_catfiles"
    integer :: nfiles,fi,numRows,totalNumRows

    catfile_out%src_list%std_header = catfiles(1)%src_list%std_header

    if(.not. catfile_out%src_list%isInitialized) then
      call wrerr(functionName, s_err, "Output binary table is not initialized.")
      status = s_errRep
  return
    end if

    nfiles = size(catfiles)

    totalNumRows = 0
    do fi=1,nfiles
      numRows = catfiles(fi)%src_list%getNumRows()
      totalNumRows = totalNumRows + numRows
    end do

    call catfile_out%src_list%alloc(status, totalNumRows)
    if(totalNumRows<1)&
  return

    totalNumRows = 0
    do fi=1,nfiles
      call catfile_out%src_list%overwrite_section(catfiles(fi)%src_list, totalNumRows+1, status)
      numRows = catfiles(fi)%src_list%getNumRows()
      totalNumRows = totalNumRows + numRows
      call catfile_out%src_list%destroy(status)
    end do

  end subroutine merge_catfiles

  !.....................................................................
  function read_src_ids(cat_file_path) result(src_IDs)
    implicit none
    character(*), intent(in) :: cat_file_path
    integer(i4), dimension(:), allocatable :: src_IDs

    character(*), parameter :: functionName = "read_src_ids"
    type(t_catprep_file) :: cat_file
    integer :: exists,ierr
    integer(i4), dimension(:), pointer :: ptr_id_src=>null()

    ierr = 0

    call ffits_file_exists(cat_file_path, exists, ierr)
    if(exists == 0)&
call wrstop(functionName, "File "//trim(cat_file_path)//" does not exist.")

    call cat_file%src_list%init(0, ierr, cat_file%info)
    call cat_file%src_list%ibset_col_useage("ID_SRC", TCOL_READ_BIT)
    cat_file%src_list%useageIsSet = .true.
    call cat_file%read(ierr, cat_file_path) ! opens the file, reads the table, closes the file.

    call cat_file%src_list%get_col_ptr("ID_SRC", ptr_id_src)
    call alloc(src_IDs, size(ptr_id_src), ierr)
    src_IDs = ptr_id_src
    nullify(ptr_id_src)

  end function read_src_ids

end module catprep_srclist_mod
