! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
program catupdate_main
  use param
  use erolib_general_mod
  use erolib_str_mod

  use catprep_mod
  use catupdate_mod
  use catprep_srclist_mod
  use catprep_skymap_mod
  use catprep_match_cats_mod

  implicit none

  character(*), parameter :: functionName = "catprep_update_main"

  integer :: status, ierr, skyfield, index_pos,n_bands

  type(tCatprep_addnew) :: task

  type(t_catprep_file) :: catfile ! product
  type(t_comp_cat_file) :: comp_cat_file ! Comparison catalog
  type(t_skymap_file) :: skymap_file
  type(tQMatch_r4) :: Match

  status = ok

  ! read/open the .par file
  call opast(status)
  if(status /= ok) call wrstop(functionName, "error read par file")

  ! read the input paramters
  call wrlog(functionName, "reading parameters")
  call task%read_params(status)
  ! validate the input parameters
  call wrlog(functionName, "validate parameters")
  call task%validate(status)
  ! sanity check
  if(status /= ok) then
call wrstop(functionName, "input parameters are invalid")
  else
    call wrlog(functionName, "parameters - valid")
  end if

  !vvv could all be put in a setup routine.
  call catfile%read_n_bands(task%srclist, n_bands, status)
  call catfile%src_list%init(n_bands, status, catfile%info)

  call catfile%src_list%set_all_useage(0, 1, 0, 0, srclist_group_i_normal)
  call catfile%src_list%set_all_useage(0, 0, 0, 1, srclist_group_i_internal)
  call catfile%src_list%set_all_useage(0, 0, 0, 1, srclist_group_i_extend)
  call catfile%src_list%set_all_useage(0, 0, 0, 1, srclist_group_i_specfit)
  call catfile%src_list%set_all_useage(0, 0, 0, 1, srclist_group_i_varichck)
  !^^^

  ! read the infile and the information
  call wrlog(functionName, "read catsrclist - input")
  call catfile%open(readonly, status, filename=task%srclist)
  call catfile%src_list%read(status)
  call catfile%close(status)

  ! add the additional columns
!    CatSrcList%catextend = .true.
!    if(CatSrcList%catextend .eqv. .true.) then
!        call wrlog(functionName, "extend catsrclist with additional columns")
!        call CatSrcList%extend%init(CatSrcList%nelements, status)
!    end if
!    CatSrcList%vs_extend = .true.
!    if(CatSrcList%vs_extend .eqv. .true.) then
!        call wrlog(functionName, "extend catsrclist with varichck / specfit columns")
!        call CatSrcList%varichck%init(CatSrcList%nelements, status)
!        call CatSrcList%specfit%init(CatSrcList%nelements, status)
!    end if

  ! skyfield was read from header at this point
  ! get the comparison catalogs, assuming each filename is the same
  ! SKYMAPS.fits will be a new "Version" with skyfields, adjacent skyfields (index), CatalogName

  call comp_cat_file%comp_cat%init(status, comp_cat_file%info)
  call comp_cat_file%comp_cat%ibset_all_useage(TCOL_READ_BIT)

  if(task%compcat == "") then
    ! lock Skymap directory
    index_pos = index(trim(task%skymap), "/", back=.true.)
    call lock(task%skymap(1:index_pos-1), trim(task%skymap(index_pos+1:)) // ".LOCK")
    call wrlog(functionName, "lock "//trim(task%skymap) // ".LOCK")

    call skymap_file%skymap%init(status, skymap_file%info)
    call skymap_file%skymap%set_all_useage(0, 1, 1, 0) ! read and write
    call skymap_file%open(readwrite, status, task%skymap, .true.)
    call skymap_file%skymap%read(status)

    ! in the case of pipeline and survey processing
    skyfield = str2int(catfile%src_list%std_header%skyfield, status)

    call wrlog(functionName, "prepare cats")
    call prepare_cats(skyfield, skymap_file%skymap, task%update, comp_cat_file%comp_cat, catfile%src_list, status)

    ! check against old skyfield + catalog of adjacent skyfield(s)
    ! search for matches to identify SOUUID and new unregistered sources

    call wrlog(functionName, "matching catalogs")
    call match_cat(Match, catfile%src_list, comp_cat_file%comp_cat, task%cutval, status)

    ! the matrix and the used comparions catalog are freed at this point since they are not used anymore
    call Match%cat_M%destroy(status)
    call dealloc(Match%value, status)

    call wrlog(functionName, "create new catalogs")
    call create_new_CompCat_own_skyfield(skyfield, skymap_file%skymap, Match%cat_N, comp_cat_file, status)
    call comp_cat_file%info%setFileName(task%Catalog)
    if(task%update == true) then
      call wrlog(functionName, "update catalog")
      call comp_cat_file%comp_cat%update(Match%cat_N, status)
    end if
  else
    call wrlog(functionName, "comparison catalog was provided externally")
    call wrlog(functionName, "no update / new sources will be added to the existing cumulative catalog")
    call comp_cat_file%open(readonly, status, task%compcat)
    call comp_cat_file%comp_cat%read(status)
    call comp_cat_file%close(status)
    call wrlog(functionName, "match_cat")
    call match_cat(Match, catfile%src_list, comp_cat_file%comp_cat, task%cutval, status)

    ! the matrix and the used comparions catalog are freed at this point since they are not used anymore
    call Match%cat_M%destroy(status)
    call dealloc(Match%value, status)
  end if

  call wrlog(functionName, "write updated input catalog")
  call wrlog(functionName, "open")
  call open_FITS_file(task%srclist_out, create, Match%cat_N%info%fileInfo%unitNo, status)!???, delete)
  if(status == s_errRep)&
call wrstop(functionName, "Failed to open FITS file "//trim(task%srclist_out))

  Match%cat_N%info%fileInfo%fileOpenedForWrite = .true.

  call wrlog(functionName, "write")
  call Match%cat_N%write(status)
  call wrlog(functionName, "close")
  call close_FITS_file(Match%cat_N%info%fileInfo%unitNo, status)

  call Match%cat_N%destroy(status)

  if(task%compcat == "") then
    call comp_cat_file%comp_cat%ibset_all_useage(TCOL_WRITE_BIT)

    call wrlog(functionName, "write updated new comparison catalog")
    call comp_cat_file%open(create, status)
    call comp_cat_file%comp_cat%write(status)
    call comp_cat_file%close(status)
    call comp_cat_file%comp_cat%destroy(status)

    call wrlog(functionName, "update skymaps")
    ! now update the SKYMAPS file
    call skymap_file%skymap%set_new_catalog_name(skyfield, task%Catalog, status)
    if(task%update == false) then
      ! assume it is a single scan and update sinle scan information
      call skymap_file%skymap%set_lastsscan(skyfield, task%srclist, status)
!      call skymap_file%skymap%set_sscan(skyfield, task%srclist, Match%cat_N%std_header%tstart, status)
      call skymap_file%skymap%set_sscan(skyfield, task%srclist, status)
    end if
    call skymap_file%skymap%write(status)
    call skymap_file%close(status)
  end if

 ! unlock the SKYMAPS file
  call unlock(".", trim(task%skymap) // ".LOCK")
  call wrlog(functionName, "unlock "//trim(task%skymap) // ".LOCK")

  call cpast (status, ierr)

end program catupdate_main
