! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module catupdate_mod
  use param
  use catprep_srclist_mod
  use catprep_compcat_mod
  use catprep_skymap_mod
  use catprep_match_cats_mod
  use erolib_str_mod
  use erolib_sort_mod

  implicit none

logical, private, parameter :: doTest=.false.
logical, private, parameter :: use_test_dirs=.false.


contains


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to determine the comparison catalog by adding all source catalog and making a mask
  !> based on the skyfield internal region boundaries with an overlap area
  !
  !
  !> @param[in] skyfield, skymap
  !> @param[out] compcat
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine prepare_cats(skyfield, skymap, update, compcat_out, catsrclist, status)
    implicit none

    integer,            intent(in)    :: skyfield
    class(tSkymap),    intent(in)    :: skymap
    integer,            intent(in)    :: update ! flag signaling whether the comparison catalog of the same skyfield
                                  ! should be updated with new values
    type(tCompCat),    intent(out)   :: compcat_out
    type(tCatSrcList), intent(inout) :: catsrclist
    integer,            intent(inout) :: status

    character(*), parameter :: functionName = "prepare_cats"
    character(buffsize), dimension(:), allocatable :: filename_adj, filenames
    character(buffsize) :: filename_own,errmsg
    integer :: ii,counter,n_in_files,stat

    type(t_comp_cat_file), dimension(:), allocatable :: compcat_files

    real(r8) :: overlap = 0.5_r8 ! offset to the region mask in degrees
    real(r8) :: ra_min, ra_max, dec_min, dec_max
    integer(i4) :: value = true
!    integer(i4), dimension(:), allocatable :: order
    integer(i4), dimension(:), pointer :: ptr_comp_region=>null()

integer :: index_pos

    if(status > 100 .or. status == s_errRep) return

    ! get the filenames of the skymaps_ext file
    call skymap%get_catalog_name(skyfield, filename_own, status)
    call skymap%get_adjacent_catalog_name(skyfield, filename_adj, status)

    ! determine the comparison catalogs which are not ""
    n_in_files = 0
    do ii = 1, size(filename_adj)
      if(filename_adj(ii) /= "") then
        n_in_files = n_in_files + 1
      end if
    end do
    if(filename_own /= "") then
      n_in_files = n_in_files + 1
    end if

    ! fill the filename array
    call alloc(filenames, n_in_files, status)

    counter = 0
    if(filename_own /= "") then
      counter = counter + 1
      filenames(counter) = filename_own
    end if

    do ii = 1, size(filename_adj)
      if(filename_adj(ii) /= "") then
        counter = counter + 1
        filenames(counter) = filename_adj(ii)
      end if
    end do

if(use_test_dirs) then
  do ii = 1, n_in_files
    index_pos = index(filenames(ii), "/p")
    filenames(ii) = "/home/ims/erosita/bugs/20200721_01_catupdate_MR"//filenames(ii)(index_pos:)
  end do
end if

    errmsg=""
    allocate(compcat_files(n_in_files), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

    do ii = 1, n_in_files
      call compcat_files(ii)%comp_cat%init(status, compcat_files(ii)%info)
      call compcat_files(ii)%comp_cat%ibset_all_useage(TCOL_READ_BIT)
      call compcat_files(ii)%open(readonly, status, filename=filenames(ii))
      call compcat_files(ii)%comp_cat%read(status)
      call compcat_files(ii)%close(status)
      if(status /= ok) then
        call wrerr(functionName, s_err, "ERROR with file: "//trim(filenames(ii)))
      end if
    end do

    call merge_comp_cats(compcat_files, compcat_out, status)

    ! reduce the number of source by applying a region mask
    call skymap%get_skyfield_boundary(skyfield, ra_min, ra_max, dec_min, dec_max, status)
!    call compcat_out%flag_region_ra_dec(ra_min, ra_max, dec_min, dec_max, status, overlap, value)
    call compcat_out%flag_region_ra_dec(ra_min, ra_max, dec_min, dec_max, overlap, value)

    call compcat_out%get_col_ptr("REGION", ptr_comp_region)
    call compcat_out%pack((ptr_comp_region==value), status)
    nullify(ptr_comp_region)

    ! flag sources within the region mask of the input catalog to be able to identify new sources
!    call catsrclist%flag_region_ra_dec(ra_min, ra_max, dec_min, dec_max, status, overlap, value)
    call catsrclist%flag_region_ra_dec(ra_min, ra_max, dec_min, dec_max, overlap, value)

if(update==0) then
  ! do nothing. This is just to avoid -Wall warnings. I'll also add a stop and see who screams.
end if
call wrstop(functionName, "'update' has no function here. Time to check that the routine is doing something sensible?")

  end subroutine prepare_cats

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to match catalogs
  !
  !
  !
  !> @param[in] compCat, cutval
  !> @param[out] Match
  !> @param[inout] status, cat
  !---------------------------------------------------------------------------

  subroutine match_cat(Match, cat, compCat, cutval, status)
    implicit none

    type(tQMatch_r4),   intent(out)   :: Match
    type(tCatSrcList), intent(inout) :: cat
    type(tCompCat),    intent(in)    :: compCat
    real(r8),           intent(in)    :: cutval
    integer,            intent(inout) :: status

    character(*), parameter :: function = "match_cat"

    if(status > 100 .or. status == s_errRep) return

    call Match%init(cat, compCat, status)!*****************************

    call cat%destroy(status)
    call compCat%destroy(status)
    ! calculate a matching value and apply a cut to invalid matches
    call Match%match_distance(status, cutval)
    ! determine the number of source confusions for the input sources
    call Match%determine_sc(status)
    ! sort the input source list in ascending order of its source confusion number
    call Match%sort_sc(status)
      ! "recursive" sort sources with a SC of > 1 based on the assumption that the highest value is correct and discard the other matches for these two sources
  !        call Match%sort_best(status)

  end subroutine match_cat

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to match catalogs
  !
  !
  !
  !> @param[in] cat, skyfield, skymap
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine create_new_CompCat_own_skyfield(skyfield, skymap, cat, comp_cat_file, status)
    implicit none

    integer,               intent(in)    :: skyfield
    class(tSkymap),       intent(in)    :: skymap
    type(tCatSrcList),    intent(inout) :: cat ! Match%Cat_N
    type(t_comp_cat_file), intent(out)   :: comp_cat_file
    integer,               intent(inout) :: status

    character(buffsize) :: filename_own

    character(*), parameter :: function = "create_new_CompCat_own_skyfield"

integer :: index_pos

    if(status > 100 .or. status == s_errRep) return

    ! get the filenames of the skymaps_ext file
    call skymap%get_catalog_name(skyfield, filename_own, status)

    call comp_cat_file%comp_cat%init(status, comp_cat_file%info)
    call comp_cat_file%comp_cat%ibset_all_useage(TCOL_READ_BIT)

    if(filename_own /= "") then
if(use_test_dirs) then
  index_pos = index(filename_own, "/p")
  filename_own = "/home/ims/erosita/bugs/20200721_01_catupdate_MR"//filename_own(index_pos:)
end if
      call comp_cat_file%open(readonly, status, filename_own)
      call comp_cat_file%comp_cat%read(status)
      call comp_cat_file%close(status)
    else
      call comp_cat_file%comp_cat%alloc(status, 0)
    end if

    call comp_cat_file%comp_cat%addnew(cat, status)

  end subroutine create_new_CompCat_own_skyfield

end module catupdate_mod
