#!/usr/bin/env Rscript
# Rscript write_orbit2ascii.R asciifile mjd_ref 

options(digits=16)

### Functions --------------------------------------------------------------------------------------

date2jd <- function(year, month, day) {

   # Convert date to Julian Day (Algorithm from 'Practical Astronomy, Duffet-Smith and Zwart 2011')
   # parameters
   #   year: int  (the year before 1 A.D is 0; 10 B.C. is year -9)
   #   month: int (jan=1, feb=2 etc)
   #   day: float (day may contain fractional part)
   # returns
   #   jd: float
   #   Julian Day

   if ( month == 1 || month == 2 ) {
      yearp <- year - 1
      monthp <- month + 12
   } else {
      yearp <- year
      monthp <- month
   }
   if ( ( yearp < 1582 ) || ( yearp == 1582 && month < 10 ) || ( year == 1582 && month == 10 && month < 15 ) ) {
      B <- 0
   } else {
      A <- floor( yearp / 100 )
      B <- 2 - A + floor(A / 4)
   }

   if ( yearp < 0 ) {
      C <- floor( ( 365.25*yearp )-0.75 )
   } else {
      C <- floor(( 365.25*yearp ))
   }

   D <- floor( 30.6001 * ( monthp + 1 ) )

   jd <- B + C + D + day + 1720994.5
   return(jd)

}

jd2mjd <- function(jd) {

   # Convert Julian Day to Modified Julian Day

   return( jd - 2400000.5 )

}

utc2tt <- function(utc) {

   # Convert MJD from UTC to TT
   # parameters
   #   utc: float
   # returns
   #   tt: float

   return( utc + 69.184/86400.0 )

}

log2orbit <- function( file, dref ) {

   # Convert eROSITA orbit.log files to ascii for ftools fcreate
   # parameters
   #   file: char  (filename)
   #   dref: float (MJDREF keyword for spacecraft clock)
   # returns
   #   df: data frame 

   a <- read.table(file, header=F, skip=2)
   b <- paste(a$V1, a$V2)

   date <- strptime(b, tz="UTC", format="%d.%m.%Y %H:%M:%S")
   n <- length(date)

   year <- date$year + 1900
   month <- date$mon + 1
   day <- date$mday + date$hour/24 + date$min/1440 + date$sec/86400

   jd <- 0
   for ( kk in 1:n ) {
      jd[kk] <- date2jd(year[kk], month[kk], day[kk])
   }

   utc <- jd2mjd(jd)
   mjd <- utc2tt(utc)

   scc <- (mjd - dref) * 86400.0
   x <- a$V3 * 1000
   y <- a$V4 * 1000
   z <- a$V5 * 1000
   vx <- a$V6 * 1000
   vy <- a$V7 * 1000
   vz <- a$V8 * 1000

   df <- data.frame(date, mjd, scc, x, y, z, vx, vy, vz)
   return(df)

}
# --------------------------------------------------------------------------------------------------

args <- commandArgs(trailingOnly=TRUE)
ascii <- args[1]
mjdref <- as.double(args[2])

orbit <- log2orbit(ascii, mjdref)

coldesc <- c(
   "DATE 12A", 
   "UTC 12A", 
   "MJD D", 
   "TIME D s", 
   "X D m", 
   "Y D m", 
   "Z D m", 
   "VX D m/s", 
   "VY D m/s", 
   "VZ D m/s")

write.table(orbit, "orbit.dat", quote=T, row.names=F, col.names=F)
write.table(coldesc, "coldesc.dat", quote=F, row.names=F, col.names=F)

#warnings()

q()
