!-----------------------------------------------------------------------
! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------

 module eboxdet

 !---------------------------------------------------------------------------
 ! Module: eboxdet
 ! Routines from XMM SAS module eboxdetect_mod,
 ! adapted to be used by eROSITA SASS erbox
 !   Subroutines:
 !   eboxdetect, cmass, compl4, match, getfrac, mdetec, msnrmp, msrbkg,
 !   rcmpi4, rcmpr4
 !----------------------------------------------------------------------------

  use param
  use numkind
  use fitsimage
  use detect_io
  use cal
  use psflib
   
  implicit none 
  
!  private 

!  public eboxdetect

 contains 



!---------------------------------------------------------------------

subroutine cmass (nx,ny,lfeld,nsize,xms,yms,xerr,yerr)
 
 !----------------------------------------------------------------------------
 !  calculates center of mass of a region of interest in the field
 !
 !  update : gih        date: 16-jan-1987 17:03
 !  update : cri        date: 19-jan-1988 15:02
 !  update : ptk        date: 16-feb-1990 16:36
 !  update : exsas      date: 03-aug-1993 17:01
 !
 !  a region of interest within the array ifeld is defined by the co-
 !  ordinates nx,ny and the square half size nsize. the center of mass
 !  of the region of interest is calculated an handed over to xms,yms
 !
 !  ifeld               r4  i   input array
 !  ndimx               i4  i   x-dimension of mfeld
 !  ndimy               i4  i   y-dimension of mfeld
 !  nsize               i4  i   1/2 square size of region of interest
 !  nx                  i4  i   x-value of center of region of interest
 !  ny                  i4  i   y-value of center of region of interest
 !  xms                 r4    o x-coordinate of cms
 !  yms                 r4    o y-coordinate of cms
 !----------------------------------------------------------------------------

    real(kind=real32), dimension(:,:)  :: lfeld
!    integer(kind=int32) ::  ndimx,ndimy
    integer(kind=int32)  ::  i,j,nsize,nx,ny
    real(kind=real32)    ::  sum,xms,yms,xerr,yerr,itmp

! at the beginning sum,xms,yms has the value 0.0

    sum = 0.
    xms = 0.
    yms = 0.

    do j = ny-nsize,ny+nsize
      do i = nx-nsize,nx+nsize

! now sum sum,xms,yms
        itmp = lfeld(i,j)
        sum = sum + itmp
        xms = xms + i*itmp
        yms = yms + j*itmp

      enddo
    enddo

    xms = xms/sum
    yms = yms/sum

    xerr = 0.
    yerr = 0.

    do j = ny-nsize,ny+nsize
      do i = nx-nsize,nx+nsize

        itmp = lfeld(i,j)
        xerr = xerr + itmp*(i-xms)**2
        yerr = yerr + itmp*(j-yms)**2

      end do
    end do
   
    xerr = sqrt(xerr/(sum*(sum-1.)))
    yerr = sqrt(yerr/(sum*(sum-1.)))

    if (xerr .eq. 0.0) xerr = 0.5
    if (yerr .eq. 0.0) yerr = 0.5 

! removed (GL 12/08/2011):
!    xms = xms - 0.5
!    yms = yms - 0.5

   end subroutine cmass


!-------------------------------------------------------------------------------

subroutine compl4(nact_x,nact_y,l4feld,ncomp,ncom,thresh)

 !------------------------------------------------------------------------
 ! compression of ncomp
 !
 !  author : grh        date:
 !  update : cri        date: 23-nov-1988 10:58
 !  update : ptk        date: 17-nov-1992 12:30 l*4 data
 !  update : exsas      date: 03-aug-1993 16:19
 !
 !  compress a logical field of a factor ncomp
 !  take the mean value of ncomp*ncomp and digitalize the result
 !  depending of the threshold (0.-1.)
 !
 !  ipa_x               i4      x-dimension of the array
 !  ipa_y               i4      y-dimension of the array
 !  l4feld              l4  i/o array to be resetted
 !  nact_x              i4  i/o actual x-dimension to be compressed (on
 !                              exit is compressed of a factor ncomp)
 !  nact_y              i4  i/o actual y-dimension to be compressed (on
 !                              exit is compressed of a factor ncomp)
 !  ncomp               i4  i   compression factor
 !  thresh              r4      digital threshold (between 0. and 1.)
 !------------------------------------------------------------------------

    integer(kind=int32)                   :: i,j,k,l,merk,nact_x,nact_y,ncomp,ncom
    real(kind=real32)                     :: thresh
    integer(kind=int32)                   :: nnew_x,nnew_y,ipoint,imask
    logical(kind=log8), dimension(:,:,:)  :: l4feld
    logical(kind=log8),dimension(max_ima) :: done
    real(kind=real32)                     :: value
    character(len=*), parameter           :: rtname='COMPL4'

    nnew_x = nact_x/ncomp
    nnew_y = nact_y/ncomp
    value = ncomp*ncomp*thresh
    done(:)=.false.
    do ipoint = 1,nim
      imask=mask_ip(ipoint)
      if (.not. done(imask)) then
        done(imask)=.true.
        do j=1,nnew_y
          do i=1,nnew_x
            merk = 0
            do l=(j-1)*ncomp+1,j*ncomp
              do k=(i-1)*ncomp+1,i*ncomp
                if(l4feld(imask,ixb(k,ipoint,ncom),iyb(l,ipoint,ncom)))   merk=merk+1
              end do
            end do
            if(merk.gt.value)   then
              l4feld(imask,ixb(i,ipoint,ncom*ncomp),iyb(j,ipoint,ncom*ncomp)) = .true.
            else
              l4feld(imask,ixb(i,ipoint,ncom*ncomp),iyb(j,ipoint,ncom*ncomp)) = .false.
            end if
          end do
        end do
      endif
    end do
    nact_x = nnew_x
    nact_y = nnew_y

  end subroutine compl4


!-------------------------------------------------------------------------------

  subroutine match (par,image,detmask,lfeld,kfeld,srclist,nfeldx,nfeldy, &
                    ncom,kmerk,kold,ierr)

 !---------------------------------------------------------------
 ! 26/05/2011 (GL) snr is a local variable, not part of srclist
 ! 05/05/2014 (GL) add variable %snr to srclist
 !--------------------------------------------------------------

 !--------------------------------------------------------------------
 ! indicates source-regions and calculates cmass
 !  author : g.hasinger date:  7-mar-1985
 !  update : gih        date: 26-jan-1987 14:47
 !  update : cri        date: 22-jan-1988 12:20
 !  update : ptk        date: 16-feb-1990 17:04
 !  update : exsas      date: 03-aug-1993 17:03
 !
 !  the routine takes array lfeld which contains the flagged regions of
 !  interest and identifies sources as a number (minimal 1) of contiguos
 !  pixels. a source is regarded as 'new' if its highest pixel is at
 !  least match pixels away from the center of mass of any other source.
 !  the maximum value of lfeld within each island is handed to isnr.
 !  the center of mass of each source region is calculated
 !
 !  isnr                r4    o array with max. snr for each sour. regi.
 !  iwind               i4    o size of detection window for each source
 !  kmerk               i4  i/o actual/new number of sources yet detec.
 !  lareax              i4  i   number of 'dead' x-pixels
 !                              at the edges of ifeld
 !  lareay              i4  i   number of 'dead' y-pixels at the ed. ...
 !  lfeld               r4  i/? field with snr-flags
 !  ncom                i4  i   compression of current step
 !  ndimx               i4  ?/? x-dimension of all square fields
 !  ndimy               i4  ?/? y-dimension of all square fields
 !  nfeldx              i4  i   x-length of field to search
 !  nfeldy              i4  i   y-length of field to search
 !  xms                 r4  i/o array with x-coord. of detected sources
 !  yms                 r4  i/o array with y-coord. of detected sources
 !  cmass               sr      calculates center of mass
 !--------------------------------------------------------------------


   type(erbox_param_type)             :: par
   type(i4cube_type)                  ::  image
   type(l1cube_type)                  :: detmask
   real(kind=real32),   dimension(:,:,:)              :: kfeld
!   type(boxrow_type),dimension(:,:)   :: srclist
   type(srcrow_type),dimension(:,:)   :: srclist

   real(kind=real32), dimension(:,:)  :: lfeld
!   integer(kind=int32)               :: ndimx,ndimy
   integer(kind=int32)                :: i,l,ierr,j,k,kmerk,kold,kold1
   integer(kind=int32)                :: ncom,nfeldx,nfeldy,nfound,nlx,nly
   integer(kind=int32)                :: nx,ny,nxold,nyold,jj,mboxsize,stepsize
   real(kind=real32)                  :: maxsnr,dmatch,dmatch2,xm,ym
   real(kind=real32)                  :: xe,ye,snr
   character(len=*),parameter         ::  rtname = 'MATCH'

!   type(boxlist_type)                 :: templist

!   real(kind=real32)                  :: t1,t2,tsum
!!!   write(*,*) 'START MATCH, ncom, kmerk,kold: ',ncom, kmerk, kold



   ierr=0

!   nlx = nfeldx-lareax+1
!   nly = nfeldy-lareay+1
   nlx = nfeldx-par%boxsize
   nly = nfeldy-par%boxsize

   j = par%boxsize+1
   i = par%boxsize+1
!---------------------------------------------
! Always use 3x3 box here (eboxdetect 02/11/04 GL):

!   mboxsize=1

! or boxsize parameter:
   ! test other sizes:
    mboxsize=par%boxsize  ! full box size => lowest rate of spurious detections
   !  mboxsize=max(1,par%boxsize-1)

!  stepsize (in pixels) to advance box while looping over image
!   stepsize=(max(1,mboxsize-1))
! with (stepsize > 1) some fainter sources seem to be 
! lost , change back to 1:  


  
!  stepsize = max(1,mboxsize-1)
!  TEST 09/05/2014:
  
  stepsize = max(1,par%boxsize-2)  ! tested with good results for 9x9 box
!  TEST 09/05/2014:
!  stepsize = 1
!---------------------------------------------



   !loop over lfeld
   do while (j.le.nly .and. ierr.le.1)
     if (i.gt.nlx) i = par%boxsize+1
     do while (i.le.nlx .and. ierr.le.1)


!      if source region : lfeld(i,j) > 0
       if (lfeld(i,j).gt.0) then
! find pixel of source region with maximum value
         nx = i
         ny = j
         maxsnr = lfeld(nx,ny)
         k = nx - mboxsize
         l = ny - mboxsize

         do while (l.le.ny+mboxsize)
           if (lfeld(k,l).gt.maxsnr) then
             nx = k
             ny = l
             k = nx - mboxsize
             l = ny - mboxsize
             maxsnr = lfeld(nx,ny)
           else
             k = k + 1
             if (k.gt.nx+mboxsize) then
               k = nx - mboxsize
               l = l + 1
             endif
           endif
         enddo

         maxsnr = maxsnr / 10.


         ! compare with all sources already found before acceptance.
         ! if the difference in one coordinate is less than dmatch
         ! pixels or source too close to edge of allowed region
         ! neglect the source
         nfound = 0
         ! exclude border pixels
         if ( ((nx-par%boxsize-1).lt.2 .or. (nlx-nx).lt.2 .or. &
               (ny-par%boxsize-1).lt.2 .or. (nly-ny).lt.2) .and. &
                ncom.lt.8 ) nfound = 1

         ! Dont repeat for maximum on same pixel:
         if (nx == nxold .and. ny == nyold) nfound = 1

         

         if (nfound.eq.0) then

           call cmass (nx,ny,lfeld,mboxsize,xm,ym,xe,ye) 
           nxold=nx
           nyold=ny

           xm=xm*ncom
           ym=ym*ncom
           xe=xe*ncom
           ye=ye*ncom

           ! TEST, remove (kmerk.gt.max_sou-1) condition here:
           !          if (kmerk.gt.max_sou-1) then
           kold1 = kold+1
           if (kmerk == 0) then
             kmerk=kmerk+1
!!! v1.4             write(*,*) rtname, 'new source ', kmerk, xm,ym
             srclist(1,kmerk)%x_ima = xm
             srclist(1,kmerk)%y_ima = ym
             srclist(1,kmerk)%x_ima_err = xe
             srclist(1,kmerk)%y_ima_err = ye
             srclist(1,kmerk)%box_size = ncom
!             srclist(1,kmerk)%snr      =  maxsnr
!             write(*,*) 'call snrcalc 1: ',xm,ym,ncom
             call snrcalc (par,image,detmask,kfeld,xm,ym,nfeldx,nfeldy,ncom,snr,ierr)
!             write(*,*) 'done snrcalc 1: ',snr
             srclist(1,kmerk)%snr  =  snr 

             

!            srclist(1,kmerk)%like  = snr(kmerk)
           else  
             nfound = 0              
             do jj = 1, kmerk
               !dmatch = (0.5+par%boxsize)*(ncom+srclist(1,jj)%box_size)-1.0
               ! test smaller dmatch (28/05/2014):
               dmatch = (0.5+par%boxsize)*(0.5*(ncom+srclist(1,jj)%box_size))-1.0
               dmatch2 = dmatch*dmatch
!               if (abs(xm - srclist(1,jj)%x_ima) < dmatch .and. &
!                   abs(ym - srclist(1,jj)%y_ima) < dmatch) then
                ! changed exclusion box to exclusion circle (07/07/2014):
                if (((xm - srclist(1,jj)%x_ima)**2 + (ym - srclist(1,jj)%y_ima)**2) < dmatch2) then

                 nfound=1

                 ! drop condition (jj > kold) 20/05/2014 
                 call snrcalc (par,image,detmask,kfeld,xm,ym,nfeldx,nfeldy,ncom,snr,ierr)
 !                write(*,*) 'done snrcalc 2: ',snr
                 if (snr > srclist(1,jj)%snr) then
                   ! replace source jj by current source

!                   write(*,*)  'MATCH replace: ',jj,srclist(1,jj)%x_ima,srclist(1,jj)%y_ima,xm,ym, srclist(1,jj)%snr ,snr
                   srclist(1,jj)%x_ima = xm
                   srclist(1,jj)%y_ima = ym
                   srclist(1,jj)%x_ima_err = xe
                   srclist(1,jj)%y_ima_err = ye
                   !test 12/05/2014:
                   srclist(1,jj)%box_size = ncom
                   srclist(1,jj)%snr  =  snr 
!                  srclist(1,kmerk)%snr      =  maxsnr     
!                   write(*,*)  'new source: ',jj,srclist(1,jj)%x_ima,srclist(1,jj)%y_ima, srclist(1,jj)%snr
                 end if
               end if
             end do
!             write(*,*) 'match nfound= ',nfound  
             if (nfound == 0) then
               !append new source
               kmerk=kmerk+1
               srclist(1,kmerk)%x_ima = xm
               srclist(1,kmerk)%y_ima = ym
               srclist(1,kmerk)%x_ima_err = xe
               srclist(1,kmerk)%y_ima_err = ye
               srclist(1,kmerk)%box_size = ncom
               call snrcalc (par,image,detmask,kfeld,xm,ym,nfeldx,nfeldy,ncom,snr,ierr)
!               write(*,*) 'MATCH append: ',jj, xm,ym,maxsnr,snr
               srclist(1,jj)%snr  =  snr 
                             
!               srclist(1,kmerk)%snr      =  maxsnr  


!              srclist(1,kmerk)%like  = snr(kmerk)
             end if
            
             if (kmerk.ge.max_sou) then
               ierr = 3
               call wrerr(rtname,ierr,'Source buffer overrun')
             endif

           endif
           ! (kmerk == 0)   
         endif  ! (nfound == 0)
       endif  
       !(lfeld > 0)
       ! test step by mboxsize (GL 11/08/2011):
       i = i + stepsize 
     enddo
     ! test step by mboxsize (GL 11/08/2011):
     j = j + stepsize 
   enddo
     ! end loop over lfeld(i,j)

!   write(*,*) rtname, ' kmerk after lfeld: ', kmerk
!   allocate(templist%data(kmerk))
!   templist%data=srclist(1,1:kmerk)    
!   templist%shapehdr%naxis2=kmerk
!   call fput_boxlist('boxlist_tmp.fits',templist,ierr)
!  deallocate(templist%data)
   

   kold1 = kold+1
   do i = kold1, kmerk     
     do j = 1, kmerk
       if (i /= j .and. srclist(1,i)%snr > 0.0 .and.  srclist(1,j)%snr > 0.0) then
!        dmatch = (4.0*boxsize + min(20.0,0.4*rat))**2
!        if ((xms(i)-xms(j))**2+(yms(i)-yms(j))**2 < dmatch) then
!         dmatch = 2*boxsize*ncom
         dmatch = (0.5+par%boxsize)*(srclist(1,i)%box_size+srclist(1,j)%box_size)-1.0
         dmatch2 = dmatch*dmatch
!         if (abs(srclist(1,i)%x_ima-srclist(1,j)%x_ima) < dmatch .and. &
!             abs(srclist(1,i)%y_ima-srclist(1,j)%y_ima) < dmatch) then
         ! changed exclusion box to exclusion circle (07/07/2014):
         if (((srclist(1,i)%x_ima-srclist(1,j)%x_ima)**2 + (srclist(1,i)%y_ima-srclist(1,j)%y_ima)**2) < dmatch2) then

!            write(*,*) 'source i: ',i,srclist(1,i)%x_ima,srclist(1,i)%y_ima,srclist(1,i)%snr,srclist(1,i)%box_size
!              write(*,*) 'source j: ',j,srclist(1,j)%x_ima,srclist(1,j)%y_ima,srclist(1,j)%snr,srclist(1,j)%box_size
            if (srclist(1,i)%snr < srclist(1,j)%snr) then
!           if (srclist(1,i)%snr < srclist(1,j)%snr .or. j < kold1) then
             srclist(1,i)%snr = -1.0
!             write(*,*) 'REMOVE i=',i,srclist(1,i)%x_ima,srclist(1,i)%y_ima,' dmatch=',dmatch
            else
             srclist(1,j)%snr = -1.0
!             write(*,*) 'REMOVE j=',i,srclist(1,j)%x_ima,srclist(1,j)%y_ima,' dmatch=',dmatch
            end if
         end if
       end if
     end do
   end do

   i = kold1
   do while (i <= kmerk)
     if (srclist(1,i)%snr < 0.0) then
       srclist(1,i) =  srclist(1,kmerk)
       kmerk = kmerk-1
     else
       i = i+1
     end if
   end do

  
 end subroutine match




!------------------------------------------------------------------------------------------

  subroutine get_boxfrac(par,image,ncom,first,offax,tanphi,iima,alpha0,beta0,ierr)

  !----------------------------------------------------------------------------------------
  ! Determine fraction of PSF in detection box
  ! in a grid of NOFF offax angles, NPHI azimuth angles (tan(phi)=0.0-1.0), 
  ! and up to NIMAGES energies.
  ! 
  ! If parameter FIRST == .true.,  calculate ALPHA, BETA data cubes
  ! (This needs to be repeated  in case the values of BOXSIZE or NCOM change).
  ! return ALPHA0 (energy fraction in inner box, BETA0: source flux fraction in background frame)
  ! Limitations:
  !  Currently calculates alpha,beta values only for one octant in PHI. Assume necessary
  !  symmetry of PSF.    
  !------------------------------------------------------------------------------
 

  type(erbox_param_type),intent(in)          ::  par
  type(i4cube_type),intent(in)               ::  image
  integer(kind=int32),intent(in)             ::  iima,ncom
  real(kind=real32)                          ::  offax,tanphi
  logical(kind=log8),intent(in)              ::  first
  real(kind=real32),intent(out)              ::  alpha0, beta0
  integer(kind=int32),intent(out)            ::  ierr  
  logical(kind=log8)                         ::  survmode
  integer(kind=int32),parameter              ::  noff=16,nphi=6  
  real(kind=real32), save, dimension(nphi,noff,max_ima)  ::  alpha,beta
!  integer(kind=int32)                        ::  ioff,iphi,nx,ny,ienergy
  integer(kind=int32)                        ::  ioff,iphi
  
  integer(kind=int32)                        :: ipoint,status,ext
  real(kind=real32)                          ::  energy,pix2arcsec,boxhw,boxhw2,eef
  real(kind=real32), dimension(nphi)         ::  phi_arr
  real(kind=real32), dimension(noff)         ::  offax_arr
  real(kind=real32)                          ::  offax_max,phi
  type(cal_par_type)                         ::  cal_par
  type(psf2d_type)                           ::  psf
  character(len=clong)                       ::  psf_file
  real(kind=real32),parameter                ::  deg2rad=0.01745329252
  character(len=*),parameter                 ::  rtname = 'GET_BOXFRAC'


  ierr=0
  offax_max=30.
  
  ! mirror tan(phi) to 1st octant:
  tanphi=abs(tanphi)
  if (tanphi > 1.0) tanphi = 1.0/tanphi  


  if (first) then
    pix2arcsec = real(0.5*(abs(image%wcshdr(1)%cdelt1)+abs(image%wcshdr(1)%cdelt1))*3600.,real32)
    ! calculate box half width in arcseconds:
    boxhw = (float(par%boxsize) + 0.5) * float(ncom) * pix2arcsec
    boxhw2 = (float(par%boxsize) + 1.5) * float(ncom) * pix2arcsec
  endif


  if  (image%stdhdr(1)%obs_mode == 'SURVEY' .or.image%stdhdr(1)%obs_mode == 'survey'  ) then
    survmode = .true.
  else if  (image%stdhdr(1)%obs_mode == 'SCAN' .or. image%stdhdr(1)%obs_mode == 'scan') then
    survmode = .true. 
  else if  (image%stdhdr(1)%obs_mode == 'SLEW' .or. image%stdhdr(1)%obs_mode == 'slew') then
    survmode = .true. 
  else if  (image%stdhdr(1)%obs_mode == 'POINTING' .or. image%stdhdr(1)%obs_mode == 'pointing') then
    survmode = .false. 
  else 
    survmode = .false. 
    ierr=3
    call wrerr(rtname,ierr,'Invalid observing mode')
  endif

  if (survmode) then
    ! calculate only only one value per image:
    if (first) then
      phi=0.0
      do ipoint=1,par%n_imagesets
        call cal_setpar(image%stdhdr(ipoint),cal_par,status)    
        energy=0.5*(par%emin(ipoint)+par%emax(ipoint)) 
        cal_par%codenam='2D_PSF'
        cal_par%expression=''
        call cal_r4toex('ENERG',energy,cal_par%expression,status)
        call cal_stringtoex('OBSMODE','SURVEY',cal_par%expression,status)
        call cal_filename(cal_par,psf_file,ext,status)
        

        call get_2dpsf(psf_file,ext,phi,psf,status)    
        call  psf_boxfrac(psf,boxhw,eef,status)
        alpha(1,1,ipoint)=eef
        if (.not. par%bkgima_flag) then
          call  psf_boxfrac(psf,boxhw2,eef,status)
          beta(1,1,ipoint)=eef-alpha(1,1,ipoint)       
        endif       
      end do     
!      write(*,*)  "alpha,beta:",alpha(1,1,ipoint),beta(1,1,ipoint)
      alpha0 = alpha(1,1,iima)
      beta0  = beta(1,1,iima)
    else
      alpha0 = alpha(1,1,iima)
      beta0  = beta(1,1,iima)
!      write(*,*)  "alpha0,beta0:",alpha0,beta0 , iima
    endif
  else
    !pointing mode 
    ! on first call calculate alpha, beta data cubes:
    if (first) then
      ! sample phi in nphi steps of tan(phi)
      do iphi=1,nphi
        phi_arr(iphi)=atan(float(iphi-1)/float(nphi-1))
      end do
!??      tanlim(nphi)=1.01  
      ! 16 offaxis angles:
      ! max. offax angle = 30 arcmin

      do ioff=1,noff 
        offax_arr(ioff)=float(ioff-1)/float(noff-1) * offax_max
      end do

      do ipoint=1,par%n_imagesets
        call cal_setpar(image%stdhdr(ipoint),cal_par,status)
        energy=0.5*(par%emin(ipoint)+par%emax(ipoint)) 
        cal_par%codenam='2D_PSF'
        do ioff=1,noff        
          cal_par%expression=''
          call cal_r4toex('ENERG',energy,cal_par%expression,status)
          call cal_r4toex('THETA',offax_arr(ioff),cal_par%expression,status)
          call cal_stringtoex('OBSMODE','SURVEY',cal_par%expression,status)
          call cal_filename(cal_par,psf_file,ext,status)               
          call wrchatty(rtname,"CAL PSF: "//trim(psf_file))
          do iphi=1,nphi
            call get_2dpsf(psf_file,ext,phi,psf,status)  
            call  psf_boxfrac(psf,boxhw,eef,status)
            alpha(iphi,ioff,ipoint)=eef 
            if (.not. par%bkgima_flag) then
              call  psf_boxfrac(psf,boxhw2,eef,status)
              beta(iphi,ioff,ipoint)=eef-alpha(iphi,ioff,ipoint)       
            endif               
          end do
        end do
      end do
      ! avoid iphi=0 or iphi>6 (26/08/2019):
      iphi=min(max(1,nint(tanphi*float(nphi-1))),6)
      ioff=max(1,min(noff,nint(offax*float(noff-1)/offax_max)))
!      write(*,*) tanphi,nphi,iphi
      alpha0 = alpha(iphi,ioff,iima)
      beta0  = beta(iphi,ioff,iima) 
    else !(first)
 
!     write(*,*) "iphi:",
!      iphi=nint(tanphi*float(nphi-1))
!      ioff=min(noff,nint(offax*float(nphi-1)/offax_max))
      iphi=min(max(1,nint(tanphi*float(nphi-1))),nphi)
      ioff=max(1,min(noff,nint(offax*float(noff-1)/offax_max)))
!      write(*,*) "iphi:", offax,noff,ioff,iphi
      alpha0 = alpha(iphi,ioff,iima)
      beta0  = beta(iphi,ioff,iima) 
    endif !(first)
  endif !(survmode)

   
     


  end  subroutine get_boxfrac





!-------------------------------------------------------------------------------

!  subroutine mdetec(par,image,detmask,expimage,bkgimage,   &
!                    ndimx,ndimy,snrmin,nrun,ipa_ar,lareax, &
!                    lareay,lmap,lloc,bkscal, &
!                    kmerk,xms,yms,xerr,yerr,isnr,iwind,src,esrc, &
!                    bkg,expos,rat,erat,raw_cts,raw_bkg,eef,maskzero,ierr)

  subroutine mdetec(par,image,expimage,bkgimage,detmask,srclist,   &
                    kmerk,ierr)


  !--------------------------------------------------------------------
  ! source detection algorithm with background map
  !  author : g.hasinger date: 22-may-1985
  !  update : g.hartner  date: 10-jul-1985
  !  update : grh        date: 24-mar-1986 11:56 correct exposure
  !  update : cri        date: 22-jan-1988 12:24
  !  update : ptk        date: 19-sep-1988 11:13
  !  update : ptk        date: 17-nov-1992 21:07 absrib -> absri1
  !  update : exsas      date: 03-aug-1993 17:12
  !
  !  sliding window source detection algorithm with background map
  !  a 3x3 source window is slided over the field ifein, corresponding
  !  background is taken from the background map kfein, signal/noise
  !  ratio is estimated. whenever snr above the value snrmin a flag is
  !  set in lfeld. the isolated regions of interest in lfeld are further
  !  examined for consistency and center of mass. the mechanism is
  !  repeated a number of times with increasing  contraction.

  !  bkg                 r4  ?/? preliminary background per pixel
  !  ifein               i2  i   input field (will not be changed)
  !  ifeld               i2    o work buffer for input field
  !  ipa_ar              i4      dimension of lareax and lareay
  !  isnr                r4  ?/? maximum snr for each source region
  !  iwind               i4  ?/? size of detection window for each source
  !  kfein               r4  i   background map with the same pixel size
  !                              as ifein
  !  kfeld               r4    o work buffer for background map
  !  kmerk               i4  i/? number of detected sources
  !  lareax              i4  ?/? number of 'dead' x-rows in each step
  !  lareay              i4  ?/? number of 'dead' y-columns in each step
  !  lfeld               r4  ?/? field with snr-flag
  !  lmap                l4  i/? flag for use of an exposure map (t=yes)
  !  mfein               l1  i/? exposure map with the same pixel size
  !                              as ifein (logical*4 !)
  !  mfeld               l1    o work buffer for exposure map
  !  ndimx               i4  ?/? x-dimension of all square fields
  !  ndimy               i4  ?/? y-dimension of all square fields
  !  nfeinx              i4  i/? x-length of field to search
  !  nfeiny              i4  i/? y-length of field to search
  !  nrun                i4  i   number of compress steps to be performed
  !  snrmin              r4      signal-to-noise ratio threshold
  !  src                 r4  ?/? preliminary source counts
  !  xms                 r4  ?/? array, x-coordinates of detected sources
  !  yms                 r4  ?/? array, y-coordinates of detected sources
  !
  !  absri1              sr      gives distance to the next rib or edge
  !  abssrc              sr      gives distance to the next source
  !  hflag               r4      output flag handling routine
  !  match               sr      isolate regions of interest within lfeld
  !  msnrmp              sr      calculates the map of snr's: lfeld
  !  msrbkg              sr      gives preliminary src and bkg cts
  !  rclnr               sr      cleans an array
  !  rcmpi4              sr      compresses array by a factor of 2
  !  compl4              sr      compresses array by a factor of 2

  
   type(erbox_param_type)                             :: par
   type(i4cube_type)                                  :: image
   type(l1cube_type)                                  :: detmask
   type(r4cube_type)                                  :: expimage
   type(r4cube_type)                                  :: bkgimage  
   !type(boxrow_type),dimension(:,:)                   :: srclist
   type(srcrow_type),dimension(:,:)                   :: srclist
!   integer(kind=int32)                                :: ndimx,ndimy
   real(kind=real32)  , dimension(:,:,:), allocatable :: kfeld, kfeld_err
   real(kind=real32)  , dimension(:,:), allocatable   :: lfeld
!   type(r4image_type)                                 :: lfeld_out
!   type(r4image_type)                                 :: kfeld_out
   integer(kind=int32)     ::  i,ierr,irun,istat,j
   integer(kind=int32)     ::  kmerk,kold,ncom
   integer(kind=int32)     ::  nfeldx,nfeldy,nfelkx,nfelky,nfelmx
   integer(kind=int32)     ::  nfelmy,nfelex,nfeley
   integer(kind=int32)     ::  ipoint,ii,jj,nsum 
!   integer(kind=int32)    ::  ixd3(16),iyd3(16),ixd5(24),iyd5(24),ixx,iyy 
!   logical(kind=log8)      :: lmap,lloc
   real(kind=real32)       ::  bsum
!   real(kind=real32), dimension(max_sou) :: xms,yms,xerr,yerr
!   real(kind=real32), dimension(:,:)     :: src,esrc,rat,erat,bkg,isnr,expos
!   real(kind=real32), dimension(:,:)     :: raw_cts,raw_bkg,eef
   real(kind=real32), dimension(:,:)     ::  alpha(15,6,max_ima),beta(15,6,max_ima)   

   character(len=clong)             :: message
   character(len=*),parameter       :: rtname = 'MDETEC'  

!   real(kind=real32)    :: cpu_start,cpu_temp,cpu_stop,t1,t2

  
   ierr = 0

!  initialize nfeldx, nfeldy,
!  they will hold the sizes of the image arrays 
!  as  modified by  compression steps :
   nfeldx = image%shapehdr(1)%naxis1
   nfeldy = image%shapehdr(1)%naxis2


!  initialize kold if new list
   if(kmerk.eq.0) kold=0

!  give the arrays ifein,mfein and kfein to the work arrays ifeld,
!  mfeld and kfeld


!   write(*,*) rtname, ' ', 'allocate images ', nim,ximbuf,yimbuf, nfeldx,nfeldy

   allocate (kfeld(nim,ximbuf,yimbuf))
   allocate (kfeld_err(nim,ximbuf,yimbuf))
   allocate (lfeld(nfeldx,nfeldy))

!  write(*,*) rtname, ' ', 'done allocate images'  



!   if (par%bkgima_flag) then 
!     allocate (bfeld(nim,ximbuf,yimbuf))
!     bfeld=bg_image
!   endif


! loop over nrun steps of compression 2
   ncom = 1
   irun = 1

 
   do while (irun.le.par%nruns .and. ierr.le.1)

     if (irun.gt.1) ncom = 2*ncom



   do j = 1,nfeldy
     do i = 1,nfeldx
       do ipoint = 1,nim
         ! local mode:
         if (.not. par%bkgima_flag) then
           bsum=0.
           nsum=0 
           do ii=i-par%boxsize-2,i+par%boxsize+2 
             do jj=j-par%boxsize-2,j-par%boxsize-1
               if (ii >= 1 .and. ii<= nfeldx &
                   .and. jj >= 1 .and. jj <= nfeldy) then  
               !   ipoint -> mask_ip(ipoint) , enables sharing of detection masks
               if (.not.par%detmask_flag .or. detmask%arr(mask_ip(ipoint),ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))) then  
                 bsum=bsum+image%arr(ipoint,ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))
                 nsum=nsum+1
               end if 
               endif
             end do
             do jj=j+par%boxsize+1,j+par%boxsize+2
               if (ii >= 1 .and. ii<= nfeldx &
                   .and. jj >= 1 .and. jj <= nfeldy) then 
                !   ipoint -> mask_ip(ipoint) 
                 if (.not.par%detmask_flag .or. detmask%arr(mask_ip(ipoint),ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))) then  
                   bsum=bsum+image%arr(ipoint,ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))
                   nsum=nsum+1
                 end if  
               end if
             end do
           end do
           do jj=j-par%boxsize,j+par%boxsize 
             do ii=i-par%boxsize-2,i-par%boxsize-1
               if (ii >= 1 .and. ii<= nfeldx &
                   .and. jj >= 1 .and. jj <= nfeldy) then 
               !   ipoint -> mask_ip(ipoint)
               if (.not. par%detmask_flag .or. detmask%arr(mask_ip(ipoint),ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))) then  
                 bsum=bsum+image%arr(ipoint,ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))
                 nsum=nsum+1
               end if
               end if
             end do
             do ii=i+par%boxsize+1,i+par%boxsize+2
               if (ii >= 1 .and. ii<= nfeldx &
                  .and. jj >= 1 .and. jj <= nfeldy) then 
               !   ipoint -> mask_ip(ipoint)
               if (.not.par%detmask_flag .or. detmask%arr(mask_ip(ipoint),ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))) then  
                 bsum=bsum+image%arr(ipoint,ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))
                 nsum=nsum+1
               end if    
               end if          
             end do
           end do

           if (nsum > 0) then
             kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom)) = bsum/float(nsum)
             kfeld_err(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom)) = sqrt(bsum)/float(nsum)
           else
             kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom)) = 0.0
           endif
         else
           ! map mode:
           !average background map over detection box:
           bsum=0
           nsum=0
           do jj=j-par%boxsize,j+par%boxsize
             do ii=i-par%boxsize,i+par%boxsize
               if (ii >= 1 .and. ii<= nfeldx &
                   .and. jj >= 1 .and. jj <= nfeldy) then 
                 if (.not.par%detmask_flag .or. detmask%arr(mask_ip(ipoint),ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))) then 
                   bsum=bsum+bkgimage%arr(ipoint,ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))
                   nsum=nsum+1
                 endif
               endif
             enddo
           end do
           if (nsum > 0) then
             kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom)) = bsum/float(nsum)
           else
             kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom)) = 0.0
           endif 
!------------------------------------------------
!          kfeld(ipoint,i,j) = bfeld(ipoint,i,j)
!------------------------------------------------
         endif 
       end do 
     enddo
   enddo
 

 !  write(*,*) rtname, 'KFELD :' , kfeld(1,100,100)

   



! clean lfeld
     lfeld = 0.

!       call cpu_time(cpu_start)

! calculation of signal-to-noise map
       call msnrmp (par,image,detmask,kfeld,lfeld,nfeldx,nfeldy, &
                    ncom,istat)



!       call cpu_time(cpu_stop)
       
       write(message,*) "done msnrmp"
       call wrchatty(rtname,trim(message))
!       cpu_temp=cpu_stop


 
!!!       write(*,*) "START match: "   
   
! now perform match and cms
       call match (par,image,detmask,lfeld,kfeld,srclist,nfeldx,nfeldy, &
                   ncom,kmerk,kold, istat)
                   

!       call cpu_time(cpu_stop)
       write(message,*) "done match"
       call wrchatty(rtname,trim(message))
!       cpu_temp=cpu_stop
      


! calculate preliminary source and background counts
       call msrbkg (par,image,detmask,expimage,kfeld,kfeld_err, &
                    srclist,nfeldx,nfeldy,ncom,kold,kmerk, &
                    istat)
     

       call filter_srclist(par,kmerk,srclist,istat)



       kold =  kmerk



! compress the arrays ifeld, mfeld, kfeld, and efeld by a factor 2
            nfelmx = nfeldx
            nfelmy = nfeldy
            nfelkx = nfeldx
            nfelky = nfeldy
            nfelex = nfeldx
            nfeley = nfeldy

            call rcmpi4 (nfeldx,nfeldy,image%arr,2,ncom)
            if (par%bkgima_flag) call rcmpr4 (nfelkx,nfelky,bkgimage%arr,2,ncom)
            call rcmpr4 (nfelex,nfeley,expimage%arr,2,ncom)

            if (par%detmask_flag) call compl4(nfelmx,nfelmy,detmask%arr,2,ncom,0.55)

      irun = irun + 1
   enddo  !loop over irun 


   

  end subroutine mdetec


  !--------------------------------------------------------------------
  subroutine msnrmp (par,image,detmask,kfeld,lfeld,nfeldx, &
                     nfeldy,ncom,ierr)
  !--------------------------------------------------------------------
  !--------------------------------------------------------------------
  ! calculates signal-to-noise map lfeld
  !--------------------------------------------------------------------



    type(erbox_param_type)                             :: par
    type(i4cube_type)                                  :: image
    type(l1cube_type)                                  :: detmask

    integer(kind=int32)                    :: ibox,jbox,ncom
    real(kind=real32),   dimension(:,:,:)  :: kfeld
    real(kind=real32),   dimension(:,:)    :: lfeld
 

    integer(kind=int32)                    :: mso,ipoint,nzero 
    integer(kind=int32)                    :: i,ierr,ii,istat,j,jj
    integer(kind=int32)                    :: mlevel,nfeldx,nfeldy,nlx,nly
    real(kind=real32)                      :: bkg,ibk,snr,lsum,smso
!    logical(kind=real32)                   :: lmap
    real(kind=real32), dimension(5,5)      :: kernel
    real(kind=real32), dimension(9,9)      :: kernel9
!    real(kind=real32), dimension(11,11)    :: kernel11,kernel11data 
    real(kind=real32)                      :: ksum
    character(len=*),parameter         ::  rtname = 'MSNRMP'



!     data                       kernel/0.2,0.3,0.4,0.3,0.2, &
!                                       0.3,0.5,0.6,0.5,0.3, &
!                                       0.4,0.7,1.0,0.7,0.4, &
!                                       0.3,0.5,0.6,0.5,0.3, &
!                                       0.2,0.3,0.4,0.3,0.2/



! XMM eboxdetect kernel:
!     data                       kernel/0.2,0.3,0.4,0.3,0.2, &
!                                       0.3,0.7,0.8,0.7,0.3, &
!                                       0.4,0.8,1.0,0.8,0.4, &
!                                       0.3,0.7,0.8,0.7,0.3, &
!                                       0.2,0.3,0.4,0.3,0.2/



     kernel  =  reshape(  (/0.0288,  0.0358,  0.0388,  0.0358,  0.0288, &
                    0.0358,  0.0461,  0.0507,  0.0461,  0.0358, &  
                    0.0388,  0.0507,  0.0561,  0.0507,  0.0388, &  
                    0.0358,  0.0461,  0.0507,  0.0461,  0.0358, &  
                    0.0288,  0.0358,  0.0388,  0.0358,  0.0288/)&
                    , (/5,5/) )



     kernel9 =   reshape(  (/0.0048,0.0061,0.0075,0.0086,0.0090,0.0086,0.0075,0.0061,0.0048, &
                    0.0061,0.0082,0.0106,0.0126,0.0134,0.0126,0.0106,0.0082,0.0061, &
                    0.0075,0.0106,0.0143,0.0178,0.0192,0.0178,0.0143,0.0106,0.0075, &
                    0.0086,0.0126,0.0178,0.0229,0.0251,0.0229,0.0178,0.0126,0.0086, &
                    0.0090,0.0134,0.0192,0.0251,0.0278,0.0251,0.0192,0.0134,0.0090, &
                    0.0086,0.0126,0.0178,0.0229,0.0251,0.0229,0.0178,0.0126,0.0086, &
                    0.0075,0.0106,0.0143,0.0178,0.0192,0.0178,0.0143,0.0106,0.0075, &
                    0.0061,0.0082,0.0106,0.0126,0.0134,0.0126,0.0106,0.0082,0.0061, &
                    0.0048,0.0061,0.0075,0.0086,0.0090,0.0086,0.0075,0.0061,0.0048/)&
                    , (/9,9/) )



    
    !normalize to number of box pixels:
    kernel=kernel * 25.0
    kernel9=kernel9 * 81.0
 

!    test with flat kernel:
     kernel=1.0


    if (par%boxsize == 1) then
      mlevel=5
    else
      mlevel=15
    end if
 
! set values nlx, nly and start values for following loops
    ierr = 0
    nlx = nfeldx - par%boxsize
    nly = nfeldy - par%boxsize
    i = par%boxsize + 1
    j = par%boxsize + 1
    lsum = 0.

    lfeld = 0.


    do while (j.le.nly .and. ierr.le.1)
      if (i.gt.nlx) i =  par%boxsize + 1
      do while (i.le.nlx .and. ierr.le.1)
  
! source pixels from image ifeld
        snr = 0.
        nzero=0 
        do ipoint = 1,nim 
!!!       iso = 0
          smso = 0.0


          ibk = 0.
          mso = 0

          ibox=0
          jbox=0
          ksum=0.0
          do jj = j-par%boxsize,j+par%boxsize
            jbox=jbox+1
            ibox=0
            do ii = i-par%boxsize,i+par%boxsize
              ibox=ibox+1

              if (ii .ge. 1 .and. ii .le. nfeldx .and. &
                jj .ge. 1 .and. jj .le. nfeldy) then
                
                if (.not.par%detmask_flag .or. detmask%arr(mask_ip(ipoint),ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom)) &
                             .and. detmask%arr(mask_ip(ipoint),ixb(i,ipoint,ncom),iyb(j,ipoint,ncom))) then
                  if (par%boxsize == 2) then
!                    iso = iso + ifeld(ipoint,ii,jj)*kernel(ibox,jbox)
                     smso = smso +  real(image%arr(ipoint,ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom)))*kernel(ibox,jbox)
                     ksum = ksum +  kernel(ibox,jbox)  
!                  else 
!!!!! >>> TEST 9x9 kernel 
                  else if (par%boxsize == 4) then
                     smso = smso +  real(image%arr(ipoint,ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom)))*kernel9(ibox,jbox)
                     ksum = ksum +  kernel9(ibox,jbox) 
!!!!!  <<< TEST 9x9 kernel 
                  else
!                   iso = iso + ifeld(ipoint,ii,jj)
                    smso = smso + real(image%arr(ipoint,ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom)))
                    ksum=ksum+1
                  endif                        
!                  ibk = ibk + kfeld(ipoint,ii,jj)
                  mso = mso + 1
                endif
              end if 
            enddo
          enddo
          if (.not.par%detmask_flag .or. detmask%arr(mask_ip(ipoint),ixb(i,ipoint,ncom),iyb(j,ipoint,ncom))) then
!          ibk = max(mso*kfeld(ipoint,i,j),0.2)
!TEST:
!           changed minimum value of ibk to 0.1 (from 0.2) 1/12/04
           ibk = max(ksum*kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom)),0.1)

!          ibk = mso*kfeld(ipoint,i,j)
          endif                   

! if iso or mso smaller than critical don't continue
!           bkg = ibk*bkscal
           bkg = ibk


             if (smso > 0.0 .and. mso > mlevel) &
!!                write(*,*) rtname,  i,j, smso, bkg
!               snr = snr + poisl2 (float(iso),bkg,istat)
                snr = snr + poisl2 (smso,bkg,istat)

            if  (mso .le. mlevel) nzero=nzero+1 
               

!          endif

        end do


        if (nim > nzero) snr = gammq(real(nim-nzero),snr) 

         ! if (snr.ge.par%likemin) then
         ! test lower likelihood level for map
        if (snr .ge. par%likemin-1.0) then
          snr = snr*10.
          lfeld(i,j) = snr
          lsum = lsum + snr
        endif

        i = i + 1
      enddo
      j = j + 1
    enddo



  end subroutine msnrmp

  !--------------------------------------------------------------------
  subroutine msrbkg (par,image,detmask,expimage,kfeld,kfeld_err, &
                      srclist, nfeldx,nfeldy, ncom,kold,kmerk,ierr)
 
  !--------------------------------------------------------------------

!----------------------------------------------------------
! calculates source and background counts
! calculate detection likelihoods 
!----------------------------------------------------------

! author : g.hasinger date: 25-may-1985
! update : gih        date: 4-aug-1987 18:39
!  update : cri        date: 22-jan-1988 12:23
!  update : ptk        date: 16-nov-1990 08:19
!  update : exsas               date: 03-aug-1993 17:12

!  takes the sources detected by mdetec and calculates src and bkg
!  using the image ifeld and the background map kfeld

!  bkg                 r4    o background counts per pixel
!  ifeld               i2  i   input field
!  isnr                r4    o signal to noise ratio
!  kfeld               r4  i   background map
!  kmerk               i4  i   number of newly detected sources
!  kold                i4  i/o number of old sources in list (updated)
!  lmap                l4  i   flag for use of an exposure map (t=yes)
!  mfeld               l1  i   exposure map with same size as ifeld
!                              (logical*4 !)
!  ncom                i4  i   compression of current step
!  ndimx               i4      x-dim of all square fields (in memory)
!  ndimy               i4      y-dim of all square fields (in memory)
!  src                 r4    o source counts
!  xms                 r4  i   center of mass x-coordinate from match
!  yms                 r4  i   center of mass nate from match
!  firstpsf            l1      flag whether new psf fractions need to be calulated
!  poisl2              real64      calculates signal to noise ratio

  
   type(erbox_param_type)                             :: par
   type(i4cube_type)                                  :: image
   type(l1cube_type)                                  :: detmask
   type(r4cube_type)                                  :: expimage
!   type(r4cube_type)                                  :: bkgimage
!   type(boxrow_type),dimension(:,:)                   :: srclist
   type(srcrow_type),dimension(:,:)                   :: srclist
   real(kind=real32), dimension(:,:,:)        :: kfeld, kfeld_err
   integer(kind=int32)                        :: kmerk,ierr
   integer(kind=int32)                        :: i,ii,istat,j,jj,kold,n,imask,jmask
   integer(kind=int32)                        :: ncom,ipoint,nfeldx,nfeldy
   integer(kind=int32), dimension(max_ima)    :: mso,iso 
   real(kind=real32)                          :: offax
   real(kind=real32)                          :: aa,bb,kk,ke,alpha0,beta0,bk,bkmin
   real(kind=real32)                          :: bkgsum,isn,deltax,deltay
   real(kind=real32), dimension(max_ima)      :: imax 
   real(kind=real64)                          :: xcen,ycen
   logical(kind=log8)                         :: firstpsf
!   real(kind=real32), dimension(15,6,max_ima) :: alpha,beta
   integer(kind=int32), dimension(9,9)         :: circmask9
   real(kind=real32),parameter                :: deg2rad=0.01745329252
   character(len=*),parameter                 :: rtname = 'MSRBKG'



   ierr=0
   
   circmask9 = reshape( (/ 0,0,1,1,1,1,1,0,0, &                        
                           0,1,1,1,1,1,1,1,0, &
                           1,1,1,1,1,1,1,1,1, &
                           1,1,1,1,1,1,1,1,1, &
                           1,1,1,1,1,1,1,1,1, &
                           1,1,1,1,1,1,1,1,1, &
                           1,1,1,1,1,1,1,1,1, &
                           0,1,1,1,1,1,1,1,0, &
                           0,0,1,1,1,1,1,0,0/)&
                           ,(/ 9,9 /) )



    if (par%boxsize == 1) then
       bkmin=1./16.
    else
       bkmin=1./24.
    end if

! calculate bkg and src
!
! to calculate fluxes the amount of source flux within the source
! and background cells (alpha and beta resp.) have to be known

    ierr = 0
    mso = 9
!    n = kold + 1
! start frpm n=1 :
   n=1
   
    bkgsum = 0.

    firstpsf = .true.

    do while (n.le.kmerk .and. ierr.le.1)

! for test purposes try various locations 
     imax = -999.
!     do jjj = -1,1
!       do iii = -1,1


! 3x3 or 5x5 source pixels from image ifeld
         i = nint(srclist(1,n)%x_ima/ncom)
         j = nint(srclist(1,n)%y_ima/ncom)
!         i=max(1,i)
!         j=max(1,j) 
        

         mso = 0
         iso = 0


         jmask=0
         do jj = j-par%boxsize,j+par%boxsize
           jmask=jmask+1
           imask=0
           do ii = i-par%boxsize,i+par%boxsize
             imask=imask+1
             if (ii .ge. 1 .and. ii .le. nfeldx .and. &
                 jj .ge. 1 .and. jj .le. nfeldy) then 
               do ipoint=1,nim
                 if (.not. par%detmask_flag .or.detmask%arr(mask_ip(ipoint),ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))) then
                   ! test with circular mask:
                   if (par%boxsize .ne. 4 .or. circmask9(imask,jmask) .eq. 1) then
                     mso(ipoint) = mso(ipoint) + 1
                     iso(ipoint) = iso(ipoint) + &
                                  image%arr(ipoint,ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))
                   endif
                   !----
                 end if
               end do
             end if 
           end do
         end do
 
         do ipoint = 1,nim
         ! get fraction of psf in detection box
         !!  if (ncom == 1) then
          ! calculate for all compression steps (21/05/2014)
           if ( ncom == srclist(1,n)%box_size ) then

             call  equ2pix(image%wcshdr(ipoint),image%stdhdr(ipoint)%ra_pnt,image%stdhdr(ipoint)%dec_pnt,xcen,ycen,istat)
             deltax=real(abs(srclist(1,n)%x_ima-xcen),real32)
             deltay=real(abs(srclist(1,n)%y_ima-ycen),real32)

             ! offax in arcmin:
             offax = real(sqrt((deltax*image%wcshdr(1)%cdelt1)**2+(deltay*image%wcshdr(1)%cdelt2)**2)*60.,real32)

             call get_boxfrac(par,image,ncom,firstpsf,offax,deltay/deltax,ipoint,alpha0,beta0,ierr)   

!             write(*,*) "alpha0,beta0:",alpha0,beta0 
  
             firstpsf=.false. 

!           else
!             beta0 = 0.0
!             alpha0 = 1.0
!           end if

             if (.not. par%bkgima_flag) then
               bk = mso(ipoint)*max(kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom)),bkmin)
               isn =  poisl2(float(iso(ipoint)),bk,istat)
             else 
               bk = mso(ipoint)*kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom))
               isn =  poisl2(float(iso(ipoint)),bk,istat)
             endif
               if (mso(ipoint) .eq. 0)  srclist(ipoint,n)%maskzero=1
               imax(ipoint) = isn
               srclist(ipoint,n)%like = isn

               srclist(ipoint,n)%eef = alpha0
               srclist(ipoint,n)%box_cts = iso(ipoint) 
               srclist(ipoint,n)%bg_raw = kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom))/(ncom*ncom) 
               if (.not. par%bkgima_flag) then
!              bb = beta0/(8.0*(1+par%boxsize))
                 bb = beta0/(16*par%boxsize+24)
                 aa = alpha0-mso(ipoint)*bb
                 kk = kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom))
                 ke = kfeld_err(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom)) 
                 srclist(ipoint,n)%scts = (iso(ipoint)-mso(ipoint)*kk)/aa
                 srclist(ipoint,n)%scts_err  = sqrt(iso(ipoint)+(mso(ipoint)*ke)**2)/aa
                 srclist(ipoint,n)%bg_map = (kk-srclist(ipoint,n)%scts*bb)/real(ncom)**2
                 if (srclist(ipoint,n)%bg_map <= 0.0) then
                   srclist(ipoint,n)%scts = iso(ipoint)/alpha0
                   srclist(ipoint,n)%scts_err = sqrt(real(iso(ipoint)))/alpha0
                   srclist(ipoint,n)%bg_map = 0.0
                 end if
               else
                 srclist(ipoint,n)%scts     = (iso(ipoint)-bk) / alpha0
                 srclist(ipoint,n)%scts_err = sqrt(real(iso(ipoint)))/alpha0 
                 srclist(ipoint,n)%bg_map   = bk / (mso(ipoint)*ncom*ncom)
               end if
! CHANGE! Need average over box here!            
               srclist(ipoint,n)%exp_map   = expimage%arr(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom))/real(ncom)**2
               srclist(ipoint,n)%rate      = srclist(ipoint,n)%scts/srclist(ipoint,n)%exp_map
               srclist(ipoint,n)%rate_err  = srclist(ipoint,n)%scts_err/srclist(ipoint,n)%exp_map
             endif ! ncom  
 
          enddo
          !ipoint

     n = n + 1
   enddo ! kmerk
   kold = kmerk

!   do i=1,kmerk
!     write(*,*) "msrbkg:",xms(i),yms(i),isnr(1:5,i)
!   end do

  end subroutine msrbkg 

!-------------------------------------------------------------------------------



  !--------------------------------------------------------------------
  subroutine snrcalc (par,image,detmask,kfeld, &
                      xm,ym, nfeldx,nfeldy, ncom,  &
                      snr,ierr)
 
  !--------------------------------------------------------------------

  !----------------------------------------------------------
  ! calculates source parameters for one source 
  ! (see subroutine msrbkg)
  !----------------------------------------------------------

 
   type(erbox_param_type)                             :: par
   type(i4cube_type)                                  :: image
   type(l1cube_type)                                  :: detmask
!   type(r4cube_type)                                  :: bkgimage
   real(kind=real32),   dimension(:,:,:)              :: kfeld
!   type(srcrow_type),dimension(:,:)                   :: srclist
   real(kind=real32)                                  :: xm,ym
   real(kind=real32)                                  :: snr
   integer(kind=int32)                                :: ierr
   integer(kind=int32)                        :: isrc,ii,istat,i,j,jj,ndof
   integer(kind=int32)                        :: ncom,ipoint,nfeldx,nfeldy
   integer(kind=int32), dimension(max_ima)    :: mso,iso 
!   real(kind=real32)                          :: offax
   real(kind=real32)                          :: bk,bkmin
   real(kind=real32)                          :: bkgsum,isn
!   real(kind=real32), dimension(max_ima)      :: imax 
!   real(kind=real64)                          :: xcen,ycen
!   logical(kind=log8)                         :: firstpsf
!   real(kind=real32), dimension(15,6,max_ima) :: alpha,beta
!   real(kind=real32),parameter                :: deg2rad=0.01745329252
   character(len=*),parameter                 ::  rtname = 'SNRCALC'

    ierr = 0

    bkgsum = 0.
    mso = 0
    iso = 0


    if (par%boxsize == 1) then
       bkmin=1./16.
    else
       bkmin=1./24.
    end if



!     i = nint(srclist(1,isrc)%x_ima/ncom)
!     j = nint(srclist(1,isrc)%y_ima/ncom)
      i = nint(xm/ncom)
      j = nint(ym/ncom)


     do jj = j-par%boxsize,j+par%boxsize
       do ii = i-par%boxsize,i+par%boxsize
         if (ii .ge. 1 .and. ii .le. nfeldx .and. &
             jj .ge. 1 .and. jj .le. nfeldx) then 
           do ipoint=1,nim
             if (.not. par%detmask_flag .or.detmask%arr(mask_ip(ipoint),ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))) then
               mso(ipoint) = mso(ipoint) + 1
               iso(ipoint) = iso(ipoint) + &
                               image%arr(ipoint,ixb(ii,ipoint,ncom),iyb(jj,ipoint,ncom))
              end if
           end do
         end if 
       end do
     end do
  
      ndof=0
      snr=0.0
      do ipoint = 1,nim
        if (mso(ipoint) > 0) ndof=ndof+1
        if (.not. par%bkgima_flag) then
          bk = mso(ipoint)*max(kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom)),bkmin)
!          write(*,*) 'POISL ( ',float(iso(ipoint)),bk,' )'
          isn =  poisl2(float(iso(ipoint)),bk,istat)
        else 
          bk = mso(ipoint)*kfeld(ipoint,ixb(i,ipoint,ncom),iyb(j,ipoint,ncom))
!          write(*,*) 'POISL ( ', float(iso(ipoint)),bk,' )'
          isn =  poisl2(float(iso(ipoint)),bk,istat)
        endif 
        snr=snr+isn
      enddo
 
 
      if (ndof > 0) snr = gammq(real(ndof),snr) 

  end subroutine snrcalc 

  !------------------------------------------------------------------

  function poisl2(measur,expect,ierr) result (result_poisl2)

  !------------------------------------------------------------------
  !  poiss. likelihood
  !
  !  author : s.doebereiner date: 11-sep-1996
  !
  !  get logarithm of the cumulative poisson probability (likelihood)  
  !  to have 'measur' or more counts if 'expect' counts are expected. 
  !
  !  expect              r4  i/? expected value (background)
  !  ierr                i4    o = 0 no error
  !                              = error occured in ppois
  !  measur              i4  i/? measured value
  !  poisl2              r4      likelihood
  !
  !  ppoiss              r4      interpolates snr in table
  !  gammp               r4      calculates snr=-ln(p(measur,expect))
  !  wrflag              sr      writes to flag output stream
  !  wrerr               sr      write error message
  !------------------------------------------------------------------

    real(kind=real32)   :: result_poisl2
    integer(kind=int32) :: ierr
    real(kind=real32)   :: expect,measur

    ierr = 0

    result_poisl2 = 0.0 

    if (measur > 0) then
      result_poisl2 = gammp(measur,expect)  
    endif

  end function poisl2

 !------------------------------------------------------------------------------

   subroutine rcmpi4(nact_x,nact_y,i4arr,ncomp,ncom)


 !----------------------------------------------------------------------
 ! compression of ncomp
 !
 !  author : grh        date:
 !  update : cri        date: 23-nov-1988 10:58
 !  update : ptk        date: 06-mar-1990 13:20
 !  update : exsas      date: 03-aug-1993 16:26
 !
 !  ipa_x               i4  i   x-dimension of the array
 !  ipa_y               i4  i   y-dimension of the array
 !  i4arr               i2  i   array to be resetted
 !  nact_x              i4    o actual x-dimension to be compressed (on
 !                              exit is compressed of a factor ncomp)
 !  nact_y              i4    o actual y-dimension to be compressed (on
 !                              exit is compressed of a factor ncomp)
 !  ncomp               i4 i    compression factor
 !
 !  hflag               r4      output flag handling routine
 !  wrflag              sr      writes to flag output stream
!----------------------------------------------------------------------

    integer(kind=int32)                   ::  i,j,k,l,merk,nact_x,nact_y,ncomp,ncom
    integer(kind=int32)                   :: nnew_x,nnew_y,ipoint
    integer(kind=int32), dimension(:,:,:) :: i4arr

    nnew_x = nact_x/ncomp
    nnew_y = nact_y/ncomp
    do j=1,nnew_y
      do i=1,nnew_x
        do ipoint=1,nim
          merk = 0
          do l=(j-1)*ncomp+1,j*ncomp
            do k=(i-1)*ncomp+1,i*ncomp
              merk = merk+i4arr(ipoint,ixb(k,ipoint,ncom),iyb(l,ipoint,ncom))
            end do
          end do
          i4arr(ipoint,ixb(i,ipoint,ncom*ncomp),iyb(j,ipoint,ncom*ncomp)) = merk
        end do
      end do
    end do
    nact_x = nnew_x
    nact_y = nnew_y

  end subroutine rcmpi4

 !------------------------------------------------------------------------------
 
   subroutine rcmpr4(nact_x,nact_y,r4feld,ncomp,ncom)

  !----------------------------------------------------------------------
  !
  ! compression of ncomp
  !
  !  author : grh        date:
  !  update : cri        date: 23-nov-1988 10:58
  !  update : ptk        date: 06-mar-1990 13:20
  !  update : ptk        date: 06-mar-1990 13:46
  !  update : exsas              date: 03-aug-1993 16:26
  !
  !  compress a r4-field of a factor ncomp
  !
  !  ipa_x               i4  i   x-dimension of the array
  !  ipa_y               i4  i   y-dimension of the array
  !  r4feld              r4  i   array to be resetted
  !  nact_x              i4    o actual x-dimension to be compressed (on
  !                              exit is compressed of a factor ncomp)
  !  nact_y              i4    o actual y-dimension to be compressed (on
  !                              exit is compressed of a factor ncomp)
  !  ncomp               i4 i    compression factor
  !
  !  hflag               r4      output flag handling routine
  !  wrflag              sr      writes to flag output stream
  !----------------------------------------------------------------------

    integer(kind=int32)                 :: i,j,k,l,nact_x,nact_y,ncomp,ncom
    integer(kind=int32)                 :: nnew_x,nnew_y,ipoint
    real(kind=real32), dimension(:,:,:) :: r4feld
    real(kind=real32)                   :: merk

    nnew_x = nact_x/ncomp
    nnew_y = nact_y/ncomp
    do j=1,nnew_y
      do i=1,nnew_x
        do ipoint = 1,nim
          merk = 0.0
          do l=(j-1)*ncomp+1,j*ncomp
            do k=(i-1)*ncomp+1,i*ncomp
              merk = merk+r4feld(ipoint,ixb(k,ipoint,ncom),iyb(l,ipoint,ncom))
            end do
          end do
          r4feld(ipoint,ixb(i,ipoint,ncom*ncomp),iyb(j,ipoint,ncom*ncomp)) = merk
        end do
      end do
    end do
    nact_x = nnew_x
    nact_y = nnew_y

  end subroutine rcmpr4

end module eboxdet
