!-----------------------------------------------------------------------
! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------

module erbox_mod

  use param
  use numkind  
  use fitsimage
  use detect_io
  use cal
  use psflib

  implicit none


contains

!-------------------------------------------------------------------------------

  subroutine erbox_ini(par,ierr)

  !---------------------------------------------------------
  ! Calls subroutine OPAST to open parameter and log files
  !---------------------------------------------------------


    type(erbox_param_type)    , intent(out) :: par
    integer(kind=int32) , intent(out) :: ierr
    integer(kind=int32)               :: status
    character(len=*),parameter        :: rtname='ERBOX_INI'
    character(len=*),parameter        :: taskname='ERBOX' 

    ierr=0
    status=0

    call opast(taskname,par%lun,status)

    if (status /= 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening parameter file')
    endif 

  end subroutine erbox_ini


!-------------------------------------------------------------------------------

  subroutine erbox_par(par,ierr)
     
    type(erbox_param_type) , intent(out)     :: par
    integer(kind=int32) , intent(out)  :: ierr
    integer(kind=int32)                :: status

    character(len=*),parameter         :: rtname='ERBOX_PAR'
    
    
    ierr = 0

    call rgetca(par%lun,'IMAGES',max_ima ,par%imagesets,par%n_imagesets,status)
    if (status /= 0) ierr = max(ierr,status)
    call rgetc(par%lun,'BOXLIST'   ,par%boxlistset,status)
    if (status /= 0) ierr = max(ierr,status)      
    call rgetfa(par%lun,'EMIN', max_ima,par%emin,par%n_emin,status)  
    if (status /= 0) ierr = max(ierr,status)       
    call rgetfa(par%lun,'EMAX', max_ima,par%emax,par%n_emax,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetfa(par%lun,'ECF', max_ima,par%ecf,par%n_ecf,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetia(par%lun,'HRDEF',2*max_hr,par%hrdef,par%n_hrdef,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgeti(par%lun,'NRUNS',par%nruns,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetf(par%lun,'LIKEMIN',par%likemin,status) 
    if (status /= 0) ierr = max(ierr,status)      
    call rgeti(par%lun,'BOXSIZE',par%boxsize,status)  
    if (status /= 0) ierr = max(ierr,status) 
    call rgetl(par%lun,'COMPRESS_FLAG',par%compress_flag ,status)  
    if (status /= 0) ierr = max(ierr,status) 
    call rgetl(par%lun,'BKGIMA_FLAG',par%bkgima_flag ,status)  
    if (status /= 0) ierr = max(ierr,status) 
    call rgetl(par%lun,'EXPIMA_FLAG',par%expima_flag ,status)  
    if (status /= 0) ierr = max(ierr,status) 
    call rgetl(par%lun,'DETMASK_FLAG',par%detmask_flag,status)  
    if (status /= 0) ierr = max(ierr,status) 

    call rgetl(par%lun,'BKGIMA_FLAG',par%bkgima_flag ,status)  
    if (status /= 0) ierr = max(ierr,status) 
    if(par%bkgima_flag .eqv. .true.) then
      call rgetca(par%lun,'BKGIMAGES',max_ima,par%bkgimagesets,par%n_bkgimagesets,status)
      if (status /= 0) ierr = max(ierr,status)
    end if
    call rgetl(par%lun,'EXPIMA_FLAG',par%expima_flag ,status)  
    if (status /= 0) ierr = max(ierr,status) 
    if(par%expima_flag .eqv. .true.) then
      call rgetca(par%lun,'EXPIMAGES',max_ima,par%expimagesets,par%n_expimagesets,status)
      if (status /= 0) ierr = max(ierr,status)
    end if
    call rgetl(par%lun,'DETMASK_FLAG',par%detmask_flag,status)  
    if (status /= 0) ierr = max(ierr,status) 
    if(par%detmask_flag .eqv. .true.) then
      call rgetca(par%lun,'DETMASKS',max_ima,par%detmasksets,par%n_detmasksets,status)
      if (status /= 0) ierr = max(ierr,status)
    end if
!    call rgetl(par%lun,'CLOBBER',par%clobber,status)  
!    if (status /= 0) ierr = max(ierr,status) 

    if (ierr /= 0)  call wrerr(rtname,ierr,'Error getting parameters')

!    if (par%clobber) par%boxlistset="!"//par%boxlistset    


    if (par%expima_flag .and. par%n_expimagesets .ne. par%n_imagesets) then
      ierr=4 
      call wrerr(rtname,ierr,'Number of exposure maps not equal to number of images')
    endif
    if (par%bkgima_flag .and. par%n_bkgimagesets .ne. par%n_imagesets) then
      ierr=4 
      call wrerr(rtname,ierr,'Number of background maps not equal to number of images')
    endif
    if (par%bkgima_flag .and. par%n_bkgimagesets .ne. par%n_imagesets) then
      ierr=4 
      call wrerr(rtname,ierr,'Number of background maps not equal to number of images')
    endif
    if (par%n_emin .ne. par%n_imagesets .or. par%n_emax  .ne. par%n_imagesets) then
      ierr=4 
      call wrerr(rtname,ierr,'Number of emin or emax values not equal to number of images')
    endif
    if (par%n_ecf .ne. par%n_imagesets) then
      ierr=4 
      call wrerr(rtname,ierr,'Number of ecf values not equal to number of images')
    endif

    call rgetl(par%lun,'CLOBBER',par%clobber,status)  
    if (status /= 0) ierr = max(ierr,status) 
     
    if (par%clobber) then
       par%boxlistset="!"//trim(par%boxlistset)
    endif
    
    if (ierr > 2)  call wrerr(rtname,ierr,'Error in task parameters')

 
                       
  end subroutine erbox_par

!-------------------------------------------------------------------------------

  subroutine erbox_in(par,image,expimage,bkgimage,detmask,ierr)

  !--------------------------------------------------------------------------------
  ! Read fits images
  ! If par%compress_flag == true, only relevant section of fits images 
  ! as specified in detection mask header, is applied:
  !   - use fget_<type>compress for reading
  !   - set global variables ximbuf,yimbuf, ixb0, iyb0
  ! 
  !--------------------------------------------------------------------------------

    type(erbox_param_type), intent(in)   :: par
    type(i4cube_type)  :: image
    type(l1cube_type)  :: detmask
    type(r4cube_type)  :: expimage
    type(r4cube_type)  :: bkgimage
    integer(kind=int32), intent(out) :: ierr
    integer(kind=int32)              :: status,ext
    integer(kind=int32)              :: i,j,ipoint,nmask
    integer(kind=int32),parameter    :: order=0
    type(dethdr_type),dimension(max_exp)                :: mask_dethdr
    type(stdhdr_type),dimension(max_exp)                :: mask_stdhdr
    type(stdhdr_type),dimension(max_ima)                :: image_stdhdr
    character(len=*), parameter      :: rtname='ERBOX_IN'
    character(len=80)                :: errstr

    status=0
    ierr=0
    ext=0

 
! check input files    
    nim=par%n_imagesets
    compress_flag=par%compress_flag

    if (par%expima_flag .and. par%n_expimagesets /= nim) then
      ierr=3
      errstr = 'Number of exposure maps and images differ'
      call WRERR(rtname,ierr,errstr)
    endif



    if (par%bkgima_flag .and. par%n_bkgimagesets /= nim) then
      ierr=3
      errstr = 'Number of bkg maps and images differ'
      call WRERR(rtname,ierr,errstr)
    endif


! get image std headers:
    if (ierr < 3) then
      do i=1,nim
        call fget_stdhdr(par%imagesets(i),image_stdhdr(i),status)   
      enddo
    endif

 !  create index arrays IP and EXPIP   
 !  move after reading of detcetion masks
 !   call sort_ima(image_stdhdr,nim,status)

! get detection mask headers and determine image buffer size:

!    if (compress_flag) then


    nmask = 0
    if (par%detmask_flag) then
      nmask=par%n_detmasksets
!        ximbuf=0
!        yimbuf=0
      do i=1,par%n_detmasksets
         call fget_dethdr(par%detmasksets(i),mask_dethdr(i),status)
         call fget_stdhdr(par%detmasksets(i),mask_stdhdr(i),status)
!         ximbuf=max(ximbuf,mask_dethdr(i)%xdatamax-mask_dethdr(i)%xdatamin+1)
!         yimbuf=max(yimbuf,mask_dethdr(i)%ydatamax-mask_dethdr(i)%ydatamin+1)
      enddo
      if (compress_flag) then
        ximbuf=0
        yimbuf=0
        do i=1,par%n_detmasksets
          ximbuf=max(ximbuf,mask_dethdr(i)%xdatamax-mask_dethdr(i)%xdatamin+1)
          yimbuf=max(yimbuf,mask_dethdr(i)%ydatamax-mask_dethdr(i)%ydatamin+1)
        enddo
      endif
    else
      if (compress_flag) then
        compress_flag=.false.
        ierr = 1
        errstr = 'Need detection masks for compressed reading'
        call WRERR(rtname,ierr,errstr) 
      endif
    endif
!    endif


 !  create index arrays IP and EXPIP   
    call sort_ima(image_stdhdr,mask_stdhdr,nim,nmask,status)




    

! read images:
    if (ierr < 3) then

      do i=1,num_exp
        ixb0(i)=0
        iyb0(i)=0
        if (par%detmask_flag) then
          
          call fget_l1compress(par%detmasksets(i),num_exp,ximbuf,yimbuf,mask_dethdr(i), &
                               compress_flag,detmask,order,i,status) 

          ixb0(i)=mask_dethdr(i)%xdatamin
          iyb0(i)=mask_dethdr(i)%ydatamin
          if (compress_flag) then
             ! set borders to .false.  to avoid problems with ixb/iyb mechanism:
             detmask%arr(i,1,:)=.false.
             detmask%arr(i,ximbuf,:)=.false.
             detmask%arr(i,:,1)=.false.
             detmask%arr(i,:,yimbuf)=.false.
          endif
        endif       
        do j=1,num_image(i)
          ipoint=ip(i,j)
            ixb0(i)=mask_dethdr(i)%xdatamin-1
            iyb0(i)=mask_dethdr(i)%ydatamin-1
            call fget_i4compress(par%imagesets(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i),   &
                                 compress_flag,image,order,ipoint,status)
            if (.not. compress_flag) then
              ximbuf =  image%shapehdr(1)%naxis1
              yimbuf =  image%shapehdr(1)%naxis2 
            endif
            if (status > 0) then
              ierr=3
              errstr = 'Cannot read science image'
              call WRERR(rtname,ierr,errstr) 
            endif
            if (par%expima_flag) then
              call fget_r4compress(par%expimagesets(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i),  &
                                   compress_flag,expimage,order,ipoint,status)
              if (status > 0) then
                ierr=3
                errstr = 'Cannot read exposure map'
                call WRERR(rtname,ierr,errstr) 
              else if (.not. (match_wcs(expimage%wcshdr(ipoint),image%wcshdr(ipoint))  &
                       .and. match_size(expimage%shapehdr(ipoint),image%shapehdr(ipoint)))) then
                ierr=3
                errstr = 'Mismatch of image  and exposure map '
                call WRERR(rtname,ierr,errstr)   
              endif 
            endif
            if (par%bkgima_flag) then
              call fget_r4compress(par%bkgimagesets(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i),  &
                                   compress_flag,bkgimage,order,ipoint,status)
              if (status > 0) then
                ierr=3
                errstr = 'Cannot read background map'
                call WRERR(rtname,ierr,errstr)  
              else if (.not. (match_wcs(bkgimage%wcshdr(ipoint),image%wcshdr(ipoint))  &
                       .and. match_size(bkgimage%shapehdr(ipoint),image%shapehdr(ipoint)))) then
                ierr=3
                errstr = 'Mismatch of image and background map'
                call WRERR(rtname,ierr,errstr)   
              endif
            endif
!          else
!            ierr=3
!            errstr = 'Mismatch of image and detector mask'
!            call WRERR(rtname,ierr,errstr)
!          endif
        enddo
      enddo
    endif



 

  end subroutine erbox_in

!--------------------------------------------------------------------------------



end module  erbox_mod

