!-----------------------------------------------------------------------
! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------

module att_io


!-------------------------------------------------------------------------------
! Module containing fits table input/output routines for eROSITA attitude
! related files.
! 
! Author: Georg Lamer , AIP         date:  19-JUN-2012
!
! subroutines:
!
!    fget_corratt
!    fget_sed_raw 
!    fget_bokz_raw
!    fget_gyro_raw
!    fput_sed_raw
!    fput_bokz_raw
!    fput_gyro_raw
!    fput_interatt 
!    fput_sed_inter 
!    fput_bokz_inter
!    fput_gyro_inter
!    fput_corratt
!    interpol_corratt
!
!--------------------------------------------------------------------------------



  use param
  use fitsimage
  use cfitsio_ifaces
  use numkind
  use quaternions


 
  public 

!    integer(kind=int8)              :: i1null = -99
!    integer(kind=int16)             :: i2null = -9999
!    integer(kind=int32)             :: i4null = -999999
!    real(kind=real32)               :: r4null = -9.999E9
!    real(kind=real64)               :: r8null = -9.999D9 
!    character(len=1)                :: cnull= ''
! => moved to module fitsimage

  ! Star tracker IDs:
  integer(kind=int32), parameter :: ST_NUM_GYRO = 0
  integer(kind=int32), parameter :: ST_NUM_SED1 = 1
  integer(kind=int32), parameter :: ST_NUM_SED2 = 2
  integer(kind=int32), parameter :: ST_NUM_BOKZ = 3

  ! Gyro status bit codes:
  integer(kind=int32), parameter :: GYRO_SED1_ON_BIT = 0
  integer(kind=int32), parameter :: GYRO_SED2_ON_BIT = 1
  integer(kind=int32), parameter :: GYRO_BOKZ_ON_BIT = 2
  integer(kind=int32), parameter :: GYRO_SED1_VAL_BIT = 3
  integer(kind=int32), parameter :: GYRO_SED2_VAL_BIT = 4
  integer(kind=int32), parameter :: GYRO_BOKZ_VAL_BIT = 5
  integer(kind=int32), parameter :: GYRO_SCI_READY_BIT = 6
  integer(kind=int32), parameter :: GYRO_PREC_STAB_BIT = 7

  ! SED status bit codes:

  integer(kind=int32), parameter ::  SED_ORI_DIFF_HIGH  = 0
  integer(kind=int32), parameter ::  SED_MEAS_STARS_LOW = 1
  integer(kind=int32), parameter ::  SED_EXP_STARS_LOW = 2
  integer(kind=int32), parameter ::  SED_MOON_AVAIL     = 3
  integer(kind=int32), parameter ::  SED_ARITH_ERR_ORI  = 4
  integer(kind=int32), parameter ::  SED_STAR_QUAL      = 5
  integer(kind=int32), parameter ::  SED_ARITH_ERR_ANGV = 6
  integer(kind=int32), parameter ::  SED_ORI_QUAL       = 7
  integer(kind=int32), parameter ::  SED_SAT_VEL        = 8
  integer(kind=int32), parameter ::  SED_NOT_UPDATED    = 9
  integer(kind=int32), parameter ::  SED_SED1_DATA      = 14
  integer(kind=int32), parameter ::  SED_SED2_DATA      = 15

!  BOKZ bit status bit codes (word1):
  ! bokz word1 bit0 not used
  integer(kind=int32), parameter :: BOKZ_COVER_CLOSED   = 1
  integer(kind=int32), parameter :: BOKZ_COVER_OPENED   = 2
  integer(kind=int32), parameter :: BOKZ_OP_READY       = 10
  integer(kind=int32), parameter :: BOKZ_ORI_MODE       = 13
  integer(kind=int32), parameter :: BOKZ_IDLE_MODE      = 14
  integer(kind=int32), parameter :: BOKZ_COMM_COMPLETE  = 15
 

! BOKZ bit patterns status (word2):

  integer(kind=int16), parameter ::  BOKZ_END_ORI_PHASE1 = 256
  integer(kind=int16), parameter ::  BOKZ_END_ORI_PHASE2 = 512
  integer(kind=int16), parameter ::  BOKZ_LOC_LT3_OBJ    = 1536  
  integer(kind=int16), parameter ::  BOKZ_LOC_LT3_STARS  = 2048   
  integer(kind=int16), parameter ::  BOKZ_LOC_LT4_STARS  = 2304   
  integer(kind=int16), parameter ::  BOKZ_APR_ORI_INC    = 2560  
  integer(kind=int16), parameter ::  BOKZ_INT_CALC_ERR   = 2816
  integer(kind=int16), parameter ::  BOKZ_ROUGH_MEASURE  = 3328 
  integer(kind=int16), parameter ::  BOKZ_UNSUPP_COMM    = -32768   
  integer(kind=int16), parameter ::  BOKZ_INPUT_ERR      = -32512   
  integer(kind=int16), parameter ::  BOKZ_INPUT_TIMEOUT  = -32256  
  integer(kind=int16), parameter ::  BOKZ_CALC_TIMEOUT1  = -32000  
  integer(kind=int16), parameter ::  BOKZ_CALC_TIMEOUT2  = -31744   



  real(kind=int32), parameter    :: RADEC_ERR_UNDEFINED = -1.0
  real(kind=int32), parameter    :: ROLL_ERR_UNDEFINED  = -1.0
  integer(kind=int16), parameter :: STATUS_UNDEFINED = 0



!------------------------------------------------------------------------------------------------
  

  ! Star tracker data according to document ICD_OCS_eROSITA_eng-2.pdf
  ! and mail exchanges (http://awob.mpg.de/group/erosita_de-erocat-esass-dev_attitude/wiki/-/wiki/Main/Igor+2014-02-13)
  ! Q-Gyro
  type gyro_attrow_type
    !sequence
    integer(kind=int32)     ::  ext_ots          !  Time OTS code  [s]      
    integer(kind=int32)     ::  int_ots          !  Time OTS code  [s]      
    integer(kind=int32)     ::  otc              !  Time OTS code  [s]  
    integer(kind=int16)     ::  fracsec          !  Dt fractional seconds [ms]      
    integer(kind=int16)     ::  status           !  StatBKU status flags 
    integer(kind=int32)     ::  quat0            !  QORT[0] 
    integer(kind=int32)     ::  quat1            !  QORT[1] 
    integer(kind=int32)     ::  quat2            !  QORT[2]  
    integer(kind=int32)     ::  quat3            !  QORT[3]  
  end type gyro_attrow_type


  type gyro_att_type
    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'GYRO_ATT'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=9
    character(len=16), dimension(9)  :: cname = [ character(len=16) :: 'EXT_OTS','INT_OTS','OTC','FRACSEC','STATUS', &
                                                                       'QUAT0','QUAT1','QUAT2','QUAT3']
    character(len=4) , dimension(9) :: cform = [ character(len=4)  :: '1J','1J','1J','1I','1I','1J','1J','1J','1J']
    character(len=16), dimension(9) :: cunit=[ character(len=16)  ::  'sec','sec','sec','ms',' ',' ',' ',' ',' ']
    type(gyro_attrow_type), pointer, dimension(:)  :: data  => NULL()
    type(quaternion_type), pointer, dimension(:)   :: quat  => NULL()
    type(quaternion_type)                          :: bs_quat
    real(kind=real64)                              :: timelag=0.D0
  end type gyro_att_type



  ! SED26_1/2:
  type sed_attrow_type
    !sequence
    integer(kind=int32)     ::  otc              !  OTS code  [s]
    integer(kind=int16)     ::  fracsec          !  fractional seconds [64 microsec ?]  
    integer(kind=int32)     ::  lag              !  lag (data reception - measured)  [20 microsec] *
    integer(kind=int8)      ::  sed_id           !  SED26 identifier according to status bits 14-15 [1|2]
    integer(kind=int16)     ::  status           !  status flags  
    integer(kind=int32)     ::  xy_err           !  Orientation quality X,Y [arcsec]  *
    integer(kind=int32)     ::  roll_err         !  Orientation quality Z   [arcsec]  *
    real(kind=real64)       ::  quat0            !  Qrve[0] 
    real(kind=real64)       ::  quat1            !  Qrve[1] 
    real(kind=real64)       ::  quat2            !  Qrve[2]  
    real(kind=real64)       ::  quat3            !  Qrve[3]
    real(kind=real32)       ::  xvel             !  angular velocity X      [rad/s]
    real(kind=real32)       ::  yvel             !  angular velocity Y      [rad/s]
    real(kind=real32)       ::  zvel             !  angular velocity Z      [rad/s]
    ! *: 'unsigned int' in telemetry data, converted to int32   
  end type sed_attrow_type


  type sed_att_type
   ! sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'SED26ATT'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=14
    character(len=16), dimension(14)  :: cname = [ character(len=16) :: 'OTC','FRACSEC','LAG','SED_ID', &
                                                 'STATUS','XY_ERR','ROLL_ERR','QUAT0','QUAT1','QUAT2','QUAT3', &
                                                  'XVEL','YVEL','ZVEL']
    character(len=4) , dimension(14) :: cform = [ character(len=4)  :: '1J','1I','1J','1B','1I','1J','1J', &
                                                                      '1D','1D','1D','1D','1E','1E','1E']
    character(len=16), dimension(14) :: cunit=[ character(len=16)  :: 'sec','0.064 ms','0.02 ms',' ',' ','arcsec', &
                                              'arcsec',' ',' ',' ',' ','rad/s','rad/s',' rad/s']

    type(sed_attrow_type), pointer, dimension(:)  :: data  => NULL()
    type(quaternion_type), pointer, dimension(:)   :: quat  => NULL()
    type(quaternion_type)                          :: bs_quat
    real(kind=real64)                              :: timelag=0.D0
  end type sed_att_type


  type bokz_attrow_type 
    !sequence
    integer(kind=int32)     ::  otc              !  CW issue time  T [s]  
    integer(kind=int16)     ::  fracsec          !  Fractions of seconds dT  [ms]
    integer(kind=int16)     ::  status1          !  BOKZ status flags 1  
    integer(kind=int16)     ::  status2          !  BOKZ status flags 2  
    real(kind=real32)       ::  time_stamp       !  ?? Meaning TBC 
    ! Matrix elements are real*4 in telemetry, store as  real*8:
    real(kind=real64)       ::  mrot_00          !  matrix element (0,0)  
    real(kind=real64)       ::  mrot_01          !  matrix element (0,1)  
    real(kind=real64)       ::  mrot_02          !  matrix element (0,2)      
    real(kind=real64)       ::  mrot_10          !  matrix element (1,0)      
    real(kind=real64)       ::  mrot_11          !  matrix element (1,1)      
    real(kind=real64)       ::  mrot_12          !  matrix element (1,2)      
    real(kind=real64)       ::  mrot_20          !  matrix element [2,0)      
    real(kind=real64)       ::  mrot_21          !  matrix element [2,1)      
    real(kind=real64)       ::  mrot_22          !  matrix element [2,2)  
    ! *:  'unsigned int' in telemetry data, converted to int32       
  end type bokz_attrow_type         


  type bokz_att_type
    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'BOKZ_ATT'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=14
    character(len=16), dimension(14)  :: cname = [ character(len=16) :: 'OTC','FRACSEC','STATUS1','STATUS2', &
                                                 'TIME_STAMP','MROT_00','MROT_01','MROT_02', &
                                                  'MROT_10','MROT_11','MROT_12','MROT_20','MROT_21','MROT_22']
    character(len=4) , dimension(14) :: cform = [ character(len=4)  :: '1J','1I','1I','1I','1E','1D','1D', &
                                                                      '1D','1D','1D','1D','1D','1D','1D']
    character(len=16), dimension(14) :: cunit=[ character(len=16)  :: 'sec','ms',' ',' ','sec',' ', &
                                              ' ',' ',' ',' ',' ',' ',' ',' ']

    type(bokz_attrow_type), pointer, dimension(:)  :: data  => NULL()
    type(quaternion_type), pointer, dimension(:)   :: quat  => NULL()
    type(quaternion_type)                          :: bs_quat
    real(kind=real64)                              :: timelag=0.D0
  end type bokz_att_type
 

  !--------------------------------------------------------------------- 
  ! type definitions for eROSITA intermediate attitude file  
  ! written by task ATTPREP 
  ! Additional columns (individual star trackers, quality information may be
  ! added. 
  !--------------------------------------------------------------------

  type intattrow_type
    !sequence
    real(kind=real64)   ::  time
    real(kind=real64)   ::  ra
    real(kind=real64)   ::  dec
    real(kind=real64)   ::  roll
    real(kind=real32)   ::  radec_err
    real(kind=real32)   ::  roll_err
    integer(kind=int16) ::  st_qual
    real(kind=real32)   ::  st_maxdiff
  end type intattrow_type


  type interatt_type
    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'INTATT'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=8
    character(len=16), dimension(8)  :: cname = [ character(len=16) :: 'TIME', 'RA', 'DEC', 'ROLL','RADEC_ERR','ROLL_ERR','ST_QUAL','ST_MAXDIFF']
    character(len=4) , dimension(8) :: cform = [ character(len=4)  :: '1D','1D','1D','1D','1E','1E','1I','1E']
    character(len=16), dimension(8) :: cunit=[ character(len=16)  :: 'sec','deg','deg','deg','arcsec','arcsec','','arcsec']
    type(intattrow_type), pointer, dimension(:)   :: data  => NULL()
    type(quaternion_type), pointer, dimension(:)   :: quat  => NULL()
  end type interatt_type

!------------------------------------------------------------------------------------------------

  type corrattrow_type
    !sequence
    real(kind=real64)   ::  time
    real(kind=real64)   ::  ra
    real(kind=real64)   ::  dec
    real(kind=real64)   ::  roll    
  end type corrattrow_type


  type corratt_type
   ! sequence
    character(len=clong)            :: filename
    character(len=16)               :: extname = 'CORRATT'
    type(shapehdr_type)             :: shapehdr
    type(stdhdr_type)               :: stdhdr
    integer(kind=int32)             :: ncol=4
    character(len=16), dimension(4) :: cname = [ character(len=16) :: 'TIME', 'RA', 'DEC', 'ROLL' ]
    character(len=4) , dimension(4) :: cform = [ character(len=4)  :: '1D','1D',  '1D','1D' ]
    character(len=16), dimension(4) :: cunit=[ character(len=16)   :: 'sec','deg','deg','deg' ]
    type(corrattrow_type), pointer, dimension(:)   :: data  => NULL()
    type(quaternion_type), pointer, dimension(:)   :: quat  => NULL()
  end type corratt_type



!------------------------------------------------------------------------------------------------
!  Intermediate attitude format:

!  type interatt_row_type

!    sequence
!    real(kind=real64)   ::  TIME
!    real(kind=real64)   ::  RA
!    real(kind=real64)   ::  DEC
!    real(kind=real64)   ::  ROLL   
!    real(kind=real32)   ::  RD_E    
!    real(kind=real32)   ::  RL_E
!    integer(kind=int16) ::  SEDACT   
!    integer(kind=int16) ::  SIB_STAT   
!  end type interatt_row_type

!--------------------------------------------------------------------------------
!
!  type interatt_type
!   ! sequence
!    character(len=clong)             :: filename
!    character(len=16)                :: extname = 'INTATT'
!    type(shapehdr_type)              :: shapehdr
!    type(stdhdr_type)                :: stdhdr
!    integer(kind=int32)              :: ncol=8
!    character(len=16), dimension(8) :: cname = [ character(len=16) ::'TIME  ','RA    ','DEC',&
!                                                 'ROLL','RD_E','RL_E','SEDACT','SIB_STAT']
!
!
!    character(len=4) , dimension(8) :: cform = [ character(len=4)  :: '1D','1D','1D','1D','1E','1E','1I','1I'] 
!                                                                      
!
!    character(len=16), dimension(8) :: cunit=[ character(len=16)  :: &
!                                        'sec','deg','deg','deg','arcsec',&
!                                        'arcsec',' ',' ']
!
! 
!    type(interatt_row_type), pointer, dimension(:)  :: data  => NULL()
!    type(quaternion_type),   pointer, dimension(:)  :: quat  => NULL()
!  end type interatt_type
!  
!--------------------------------------------------------------------------------
  


  type inter_st_row_type
   
    !sequence
    real(kind=real64)   ::  TIME
    real(kind=real64)   ::  RA
    real(kind=real64)   ::  DEC 
    real(kind=real64)   ::  ROLL
    integer(kind=int16) ::  STAT1  
    integer(kind=int16) ::  STAT2   
    real(kind=real32)   ::  RADEC_ERR
    real(kind=real32)   ::  ROLL_ERR
  ! internal use, not an output column:
    type(quaternion_type)::  qcorr    
  end  type inter_st_row_type


!--------------------------------------------------------------------------------


  type inter_st_type

    !-----------------------------------------------------------------
    ! generic format for star tracker specific extensions in 
    ! intermediate  attitude file 
    ! string EXTNAME should be set to SED1/SED2/BOKZ/GYRO by 
    ! the FITS output routine.
    !-----------------------------------------------------------------

    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = ' '
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=8
    character(len=16), dimension(8) :: cname = [ character(len=16) :: 'TIME','RA','DEC','ROLL','STAT1','STAT2','RADEC_ERR',&
                                                 'ROLL_ERR']

    character(len=4) , dimension(8) :: cform = [ character(len=4)  :: '1D','1D','1D','1D','1I','1I','1E','1E']

    character(len=16) , dimension(8) :: cunit=[ character(len=16)  :: 'sec','deg','deg', &
                                        'deg',' ',' ','arcsec','arcsec']

    type(inter_st_row_type), pointer, dimension(:)  :: data  => NULL()

  end type inter_st_type




!--------------------------------------------------------------------------------


  type inter_sed_row_type

    !sequence
    real(kind=real64)   ::  TIME
    real(kind=real64)   ::  RA
    real(kind=real64)   ::  DEC 
    real(kind=real64)   ::  ROLL
    integer(kind=int16) ::  STAT   
    real(kind=real32)   ::  RADEC_ERR
    real(kind=real32)   ::  ROLL_ERR
    real(kind=real32)   ::  ANGVX 
    real(kind=real32)   ::  ANGVY
    real(kind=real32)   ::  ANGVZ
  ! internal use, not an output column:
    type(quaternion_type)::  qcorr    
  end  type inter_sed_row_type

!--------------------------------------------------------------------------------


  type inter_sed_type

    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'SED'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=10
    character(len=16), dimension(10) :: cname = [ character(len=16) :: 'TIME','RA','DEC','ROLL','STAT','RADEC_ERR',&
                                                 'ROLL_ERR','ANGVX','ANGVY','ANGVZ']

    character(len=4) , dimension(10) :: cform = [ character(len=4)  :: '1D','1D','1D','1D','1I','1E','1E','1E','1E','1E' ]

    character(len=16) , dimension(10) :: cunit=[ character(len=16)  :: 'sec','deg','deg', &
                                        'deg',' ','arcsec','arcsec','deg/sec','deg/sec','deg/sec']

    type(inter_sed_row_type), pointer, dimension(:)  :: data  => NULL()

  end type inter_sed_type




!--------------------------------------------------------------------------------


  type sunang_row_type
   
    !sequence
    real(kind=real64)   ::  TIME
    real(kind=real64)   ::  RA_SOL
    real(kind=real64)   ::  DEC_SOL 
     real(kind=real64)  ::  ANG_SOL_SRGX
    real(kind=real64)   ::  ANG_SOL_SRGZ
    real(kind=real64)   ::  ANG_SOL_SRGXZ
    real(kind=real64)   ::  RA
    real(kind=real64)   ::  DEC 
    real(kind=real64)   ::  ROLL 

  ! internal use, not an output column:
    type(quaternion_type)::  qcorr    
  end  type sunang_row_type


!--------------------------------------------------------------------------------


  type sunang_type

    !-----------------------------------------------------------------
    ! Format for solar angle  specific extensions in 
    ! intermediate  attitude file 
    ! string EXTNAME should be set to SED1/SED2/BOKZ/GYRO by 
    ! the FITS output routine.
    !-----------------------------------------------------------------

    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'SRG_SUN'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=9
    character(len=16), dimension(9) :: cname = [ character(len=16) :: 'TIME','RA_SOL','DEC_SOL','ANG_SOL_SRGX','ANG_SOL_SRGZ','ANG_SOL_SRGXZ','RA','DEC','ROLL' ]
                                                 

    character(len=4) , dimension(9) :: cform = [ character(len=4)  :: '1D','1D','1D','1D','1D','1D','1D','1D','1D' ]

    character(len=16) , dimension(9) :: cunit=[ character(len=16)  :: 'sec','deg','deg','deg','deg','deg','deg','deg','deg' ] 
    

    type(sunang_row_type), pointer, dimension(:)  :: data  => NULL()

  end type sunang_type

  
!-------------------------------------------------------------------------------

  type inter_bokz_row_type
    !sequence
    real(kind=real64)   ::  TIME
    real(kind=real64)   ::  RA
    real(kind=real64)   ::  DEC 
    real(kind=real64)   ::  ROLL
    integer(kind=int16) ::  STAT1   
    integer(kind=int16) ::  STAT2   
    ! internal use, not an output column:
    type(quaternion_type)::  qcorr 
  end  type inter_bokz_row_type

!--------------------------------------------------------------------------------

  type inter_bokz_type
    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'BOKZ'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=6
    character(len=16), dimension(6) :: cname = [ character(len=16) :: 'TIME','RA','DEC',&
                                                                       'ROLL','STAT1','STAT2']

    character(len=4) , dimension(6) :: cform = [ character(len=4)  :: '1D','1D','1D','1D','1I','1I']

    character(len=16) , dimension(6) :: cunit=[ character(len=16)  :: 'sec','deg','deg', &
                                        'deg',' ',' ']

    type(inter_bokz_row_type), pointer, dimension(:)  :: data  => NULL()

  end  type inter_bokz_type

!--------------------------------------------------------------------------------

  type inter_gyro_row_type
    !sequence
    real(kind=real64)   ::  TIME
    real(kind=real64)   ::  RA
    real(kind=real64)   ::  DEC
    real(kind=real64)   ::  ROLL
    integer(kind=int16) ::  STAT
    ! internal use, not an output column:
    type(quaternion_type)::  qcorr 
  end  type inter_gyro_row_type

!--------------------------------------------------------------------------------


  type inter_gyro_type
    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'GYRO'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=5
    character(len=16), dimension(5) :: cname = [ character(len=16) :: 'TIME','RA','DEC',&
                                                                      'ROLL','STAT']

    character(len=4) , dimension(5) :: cform = [ character(len=4)  :: '1D','1D','1D','1D','1I']

    character(len=16) , dimension(5) :: cunit=[ character(len=16)  :: 'sec','deg','deg', &
                                                                      'deg',' ']

    type(inter_gyro_row_type), pointer, dimension(:)  :: data  => NULL()
  end  type inter_gyro_type



!--------------------------------------------------------------------------------
  
contains
 


  subroutine fget_interatt(filename,ext,table,ierr)  


    implicit none

    type(interatt_type)               :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_INTERATT'
    character(len=80)                :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
    
   

    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,errstr)
      else  
!        call fread_stdhdr(unit,table%stdhdr,status)
!        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status > 0) then 
          ! go to nhdu = 2:
          status=0
          call FTMAHD(unit,2, hdutype,status)
        endif
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_stdhdr(unit,table%stdhdr,status)
          if (status /= 0) ierr=3
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            if (colnum(1) > 0) call  FTGCVD(unit,colnum(1),i1,1,rowbuf,r8null,table%data(i1:i2)%time,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(2) > 0) call  FTGCVD(unit,colnum(2),i1,1,rowbuf,r8null,table%data(i1:i2)%ra,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(3) > 0) call  FTGCVD(unit,colnum(3),i1,1,rowbuf,r8null,table%data(i1:i2)%dec,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(4) > 0) call  FTGCVD(unit,colnum(4),i1,1,rowbuf,r8null,table%data(i1:i2)%roll,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(5) > 0) call  FTGCVE(unit,colnum(5),i1,1,rowbuf,r4null,table%data(i1:i2)%radec_err,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(6) > 0) call  FTGCVE(unit,colnum(6),i1,1,rowbuf,r4null,table%data(i1:i2)%roll_err ,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(7) > 0) call  FTGCVI(unit,colnum(7),i1,1,rowbuf,i2null,table%data(i1:i2)%st_qual ,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(8) > 0) call  FTGCVE(unit,colnum(8),i1,1,rowbuf,r4null,table%data(i1:i2)%st_maxdiff ,fnull,status) 
            fstat = max(fstat,status) 
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif


  end subroutine fget_interatt

!--------------------------------------------------------------------------------


  subroutine fget_corratt(filename,ext,table,ierr)  

    implicit none

    type(corratt_type)               :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_CORRATT'
    character(len=80)                :: errstr



    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
    

    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,'error opening event list')
      else  
!        call fread_stdhdr(unit,table%stdhdr,status)
!        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status > 0) then 
          ! go to nhdu = 2:
          status=0
          call FTMAHD(unit,2, hdutype,status)
        endif
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_stdhdr(unit,table%stdhdr,status)
          if (status /= 0) ierr=3
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            if (colnum(1) > 0) call  FTGCVD(unit,colnum(1),i1,1,rowbuf,r8null,table%data(i1:i2)%time,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(2) > 0) call  FTGCVD(unit,colnum(2),i1,1,rowbuf,r8null,table%data(i1:i2)%ra,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(3) > 0) call  FTGCVD(unit,colnum(3),i1,1,rowbuf,r8null,table%data(i1:i2)%dec,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(4) > 0) call  FTGCVD(unit,colnum(4),i1,1,rowbuf,r8null,table%data(i1:i2)%roll,fnull,status) 
            fstat = max(fstat,status) 
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    


  end subroutine fget_corratt


!--------------------------------------------------------------------------------

  subroutine fget_sed_raw(filename,ext,table,ierr)  

    implicit none

    type(sed_att_type)               :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_SED_RAW'
    character(len=80)                :: errstr

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
 
    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,errstr)
      else  
!        call fread_stdhdr(unit,table%stdhdr,status)
!        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status > 0) then 
          ! go to nhdu = 2:
          status=0
          call FTMAHD(unit,2, hdutype,status)
        endif
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_stdhdr(unit,table%stdhdr,status)
          if (status /= 0) ierr=3
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          !allocate(table%quat(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=i2
            if (colnum(1) > 0) call  FTGCVJ(unit,colnum(1),i1,1,rowbuf,i4null,table%data(i1:i2)%otc,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(2) > 0) call  FTGCVI(unit,colnum(2),i1,1,rowbuf,i2null,table%data(i1:i2)%fracsec,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(3) > 0) call  FTGCVJ(unit,colnum(3),i1,1,rowbuf,i4null,table%data(i1:i2)%lag,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(4) > 0) call  FTGCVB(unit,colnum(4),i1,1,rowbuf,i1null,table%data(i1:i2)%sed_id,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(5) > 0) call  FTGCVI(unit,colnum(5),i1,1,rowbuf,i2null,table%data(i1:i2)%status,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(6) > 0) call  FTGCVJ(unit,colnum(6),i1,1,rowbuf,i4null,table%data(i1:i2)%xy_err,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(7) > 0) call  FTGCVJ(unit,colnum(7),i1,1,rowbuf,i4null,table%data(i1:i2)%roll_err,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(8) > 0) call  FTGCVD(unit,colnum(8),i1,1,rowbuf,r8null,table%data(i1:i2)%quat0,fnull,status) 
            !if (colnum(8) > 0) call  FTGCVD(unit,colnum(8),i1,1,rowbuf,r8null,table%quat(i1:i2)%arr(1),fnull,status) 

            fstat = max(fstat,status) 
            if (colnum(9) > 0) call  FTGCVD(unit,colnum(9),i1,1,rowbuf,r8null,table%data(i1:i2)%quat1,fnull,status) 
            !if (colnum(9) > 0) call  FTGCVD(unit,colnum(9),i1,1,rowbuf,r8null,table%quat(i1:i2)%arr(2),fnull,status) 

            fstat = max(fstat,status) 
            if (colnum(10) > 0) call  FTGCVD(unit,colnum(10),i1,1,rowbuf,r8null,table%data(i1:i2)%quat2,fnull,status)
            !if (colnum(10) > 0) call  FTGCVD(unit,colnum(10),i1,1,rowbuf,r8null,table%quat(i1:i2)%arr(3),fnull,status)
            fstat = max(fstat,status) 
            if (colnum(11) > 0) call  FTGCVD(unit,colnum(11),i1,1,rowbuf,r8null,table%data(i1:i2)%quat3,fnull,status)
            !if (colnum(11) > 0) call  FTGCVD(unit,colnum(11),i1,1,rowbuf,r8null,table%quat(i1:i2)%arr(4),fnull,status)
 
            fstat = max(fstat,status) 
            if (colnum(12) > 0) call  FTGCVE(unit,colnum(12),i1,1,rowbuf,r4null,table%data(i1:i2)%xvel,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(13) > 0) call  FTGCVE(unit,colnum(13),i1,1,rowbuf,r4null,table%data(i1:i2)%yvel,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(14) > 0) call  FTGCVE(unit,colnum(14),i1,1,rowbuf,r4null,table%data(i1:i2)%yvel,fnull,status) 
            fstat = max(fstat,status) 
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif

  end subroutine fget_sed_raw


!--------------------------------------------------------------------------------

  subroutine fget_bokz_raw(filename,ext,table,ierr)  

    implicit none

    type(bokz_att_type)              :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr  
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_BOKZ_RAW'
    character(len=80)                :: errstr

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
 
    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,'error opening event list')
      else  
!        call fread_stdhdr(unit,table%stdhdr,status)
!        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status > 0) then 
          ! go to nhdu = 2:
          status=0
          call FTMAHD(unit,2, hdutype,status)
        endif
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_stdhdr(unit,table%stdhdr,status)
          if (status /= 0) ierr=3
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          !allocate(table%quat(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=i2
            if (colnum(1) > 0) call  FTGCVJ(unit,colnum(1),i1,1,rowbuf,i4null,table%data(i1:i2)%otc,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(2) > 0) call  FTGCVI(unit,colnum(2),i1,1,rowbuf,i2null,table%data(i1:i2)%fracsec,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(3) > 0) call  FTGCVI(unit,colnum(3),i1,1,rowbuf,i2null,table%data(i1:i2)%status1,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(4) > 0) call  FTGCVI(unit,colnum(4),i1,1,rowbuf,i2null,table%data(i1:i2)%status2,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(5) > 0) call  FTGCVE(unit,colnum(5),i1,1,rowbuf,r4null,table%data(i1:i2)%time_stamp,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(6) > 0) call  FTGCVD(unit,colnum(6),i1,1,rowbuf,r8null,table%data(i1:i2)%mrot_00,fnull,status)
            !if (colnum(6) > 0) call  FTGCVD(unit,colnum(6),i1,1,rowbuf,r8null,table%quat(i1:i2)%matrix(1,1),fnull,status)
 
            fstat = max(fstat,status)   
            if (colnum(7) > 0) call  FTGCVD(unit,colnum(7),i1,1,rowbuf,r8null,table%data(i1:i2)%mrot_01,fnull,status)
            !if (colnum(7) > 0) call  FTGCVD(unit,colnum(7),i1,1,rowbuf,r8null,table%quat(i1:i2)%matrix(1,2),fnull,status)
 
            fstat = max(fstat,status)   
            if (colnum(8) > 0) call  FTGCVD(unit,colnum(8),i1,1,rowbuf,r8null,table%data(i1:i2)%mrot_02,fnull,status)
            !if (colnum(8) > 0) call  FTGCVD(unit,colnum(8),i1,1,rowbuf,r8null,table%quat(i1:i2)%matrix(1,3),fnull,status)
 
            fstat = max(fstat,status)   
            if (colnum(9) > 0) call  FTGCVD(unit,colnum(9),i1,1,rowbuf,r8null,table%data(i1:i2)%mrot_10,fnull,status)
            !if (colnum(9) > 0) call  FTGCVD(unit,colnum(9),i1,1,rowbuf,r8null,table%quat(i1:i2)%matrix(2,1),fnull,status)
 
            fstat = max(fstat,status)   
            if (colnum(10) > 0) call  FTGCVD(unit,colnum(10),i1,1,rowbuf,r8null,table%data(i1:i2)%mrot_11,fnull,status)
            !if (colnum(10) > 0) call  FTGCVD(unit,colnum(10),i1,1,rowbuf,r8null,table%quat(i1:i2)%matrix(2,2),fnull,status)
 
            fstat = max(fstat,status)   
            if (colnum(11) > 0) call  FTGCVD(unit,colnum(11),i1,1,rowbuf,r8null,table%data(i1:i2)%mrot_12,fnull,status)
            !if (colnum(11) > 0) call  FTGCVD(unit,colnum(11),i1,1,rowbuf,r8null,table%quat(i1:i2)%matrix(2,3),fnull,status)
 
            fstat = max(fstat,status)   
            if (colnum(12) > 0) call  FTGCVD(unit,colnum(12),i1,1,rowbuf,r8null,table%data(i1:i2)%mrot_20,fnull,status)
            !if (colnum(12) > 0) call  FTGCVD(unit,colnum(12),i1,1,rowbuf,r8null,table%quat(i1:i2)%matrix(3,1)matrix(,),fnull,status)
 
            fstat = max(fstat,status)   
            if (colnum(13) > 0) call  FTGCVD(unit,colnum(13),i1,1,rowbuf,r8null,table%data(i1:i2)%mrot_21,fnull,status)
            !if (colnum(13) > 0) call  FTGCVD(unit,colnum(13),i1,1,rowbuf,r8null,table%quat(i1:i2)%matrix(3,2),fnull,status)
 
            fstat = max(fstat,status)   
            if (colnum(14) > 0) call  FTGCVD(unit,colnum(14),i1,1,rowbuf,r8null,table%data(i1:i2)%mrot_22,fnull,status)
            !if (colnum(14) > 0) call  FTGCVD(unit,colnum(14),i1,1,rowbuf,r8null,table%quat(i1:i2)%matrix(3,3),fnull,status)

            fstat = max(fstat,status)   
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif

  end subroutine fget_bokz_raw



!--------------------------------------------------------------------------------

  subroutine fget_gyro_raw(filename,ext,table,ierr)  

    !---------------------------------------------------------------------------
    ! Read data from SRG Q-gyro
    ! Quaternion 
    !---------------------------------------------------------------------------
    implicit none

    type(gyro_att_type)               :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_GYRO_RAW'
    character(len=80)                :: errstr

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
 
    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,errstr)
      else  
!        call fread_stdhdr(unit,table%stdhdr,status)
!        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status > 0) then 
          ! go to nhdu = 2:
          status=0
          call FTMAHD(unit,2, hdutype,status)
        endif
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_stdhdr(unit,table%stdhdr,status)
          if (status /= 0) ierr=3
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          !allocate(table%quat(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=i2

            if (colnum(1) > 0) call  FTGCVJ(unit,colnum(1),i1,1,rowbuf,i4null,table%data(i1:i2)%ext_ots,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(2) > 0) call  FTGCVJ(unit,colnum(2),i1,1,rowbuf,i4null,table%data(i1:i2)%int_ots,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(3) > 0) call  FTGCVJ(unit,colnum(3),i1,1,rowbuf,i4null,table%data(i1:i2)%otc,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(4) > 0) call  FTGCVI(unit,colnum(4),i1,1,rowbuf,i2null,table%data(i1:i2)%fracsec,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(5) > 0) call  FTGCVI(unit,colnum(5),i1,1,rowbuf,i2null,table%data(i1:i2)%status,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(6) > 0) call  FTGCVJ(unit,colnum(6),i1,1,rowbuf,i4null,table%data(i1:i2)%quat0,fnull,status)
            !if (colnum(4) > 0) call  FTGCVJ(unit,colnum(4),i1,1,rowbuf,i4null,table%quat(i1:i2)%arr(4),fnull,status)
            fstat = max(fstat,status) 
            if (colnum(7) > 0) call  FTGCVJ(unit,colnum(7),i1,1,rowbuf,i4null,table%data(i1:i2)%quat1,fnull,status)
            !if (colnum(5) > 0) call  FTGCVJ(unit,colnum(5),i1,1,rowbuf,i4null,table%quat%(i1:i2)%arr(1),fnull,status)
            fstat = max(fstat,status) 
            if (colnum(8) > 0) call  FTGCVJ(unit,colnum(8),i1,1,rowbuf,i4null,table%data(i1:i2)%quat2,fnull,status)
            !if (colnum(6) > 0) call  FTGCVJ(unit,colnum(6),i1,1,rowbuf,i4null,table%quat%(i1:i2)%,arr(2)fnull,status)
            fstat = max(fstat,status) 
            if (colnum(9) > 0) call  FTGCVJ(unit,colnum(9),i1,1,rowbuf,i4null,table%data(i1:i2)%quat3,fnull,status)
            !if (colnum(7) > 0) call  FTGCVJ(unit,colnum(7),i1,1,rowbuf,i4null,table%quat%(i1:i2)%arr(3),fnull,status)
            fstat = max(fstat,status) 
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif

  end subroutine fget_gyro_raw





!-------------------------------------------------------------------------------

  subroutine fput_sed_raw(filename,append,table,ierr)  

 
    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(sed_att_type)                 :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!, nhdu
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    character(len=*),parameter         :: rtname = 'FPUT_SED_RAW'
    character(len=80)                  :: errstr

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2
 
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
 
      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLJ(unit,1,i1,1,rowbuf,table%data(i1:i2)%otc    ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,2,i1,1,rowbuf,table%data(i1:i2)%fracsec ,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,3,i1,1,rowbuf,table%data(i1:i2)%lag    ,status)
          fstat = max(fstat,status)
          call  FTPCLB(unit,4,i1,1,rowbuf,table%data(i1:i2)%sed_id  ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,5,i1,1,rowbuf,table%data(i1:i2)%status ,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,6,i1,1,rowbuf,table%data(i1:i2)%xy_err ,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,7,i1,1,rowbuf,table%data(i1:i2)%roll_err ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,8,i1,1,rowbuf,table%data(i1:i2)%quat0,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,9,i1,1,rowbuf,table%data(i1:i2)%quat1,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,10,i1,1,rowbuf,table%data(i1:i2)%quat2,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,11,i1,1,rowbuf,table%data(i1:i2)%quat3,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,12,i1,1,rowbuf,table%data(i1:i2)%xvel ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,13,i1,1,rowbuf,table%data(i1:i2)%yvel ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,14,i1,1,rowbuf,table%data(i1:i2)%zvel ,status)
          fstat = max(fstat,status)
         enddo
      endif  
       if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  

  end subroutine fput_sed_raw

!------------------------------------------------------------------------------- 

  subroutine fput_bokz_raw(filename,append,table,ierr)  

 
    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(bokz_att_type)                 :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!, nhdu
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    character(len=*),parameter         :: rtname = 'FPUT_BOKZ_RAW'
    character(len=80)                  :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2

  call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif


    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLJ(unit,1,i1,1,rowbuf,table%data(i1:i2)%otc    ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,2,i1,1,rowbuf,table%data(i1:i2)%fracsec ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,3,i1,1,rowbuf,table%data(i1:i2)%status1 ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,4,i1,1,rowbuf,table%data(i1:i2)%status2  ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,5,i1,1,rowbuf,table%data(i1:i2)%time_stamp ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,6,i1,1,rowbuf,table%data(i1:i2)%mrot_00 ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,7,i1,1,rowbuf,table%data(i1:i2)%mrot_01 ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,8,i1,1,rowbuf,table%data(i1:i2)%mrot_02 ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,9,i1,1,rowbuf,table%data(i1:i2)%mrot_10 ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,10,i1,1,rowbuf,table%data(i1:i2)%mrot_11 ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,11,i1,1,rowbuf,table%data(i1:i2)%mrot_12 ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,12,i1,1,rowbuf,table%data(i1:i2)%mrot_20  ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,13,i1,1,rowbuf,table%data(i1:i2)%mrot_21  ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,14,i1,1,rowbuf,table%data(i1:i2)%mrot_22 ,status)
          fstat = max(fstat,status)
         enddo
      endif  
       if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  



  end  subroutine fput_bokz_raw




!------------------------------------------------------------------------------- 

  subroutine fput_gyro_raw(filename,append,table,ierr)  

 
    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(gyro_att_type)                :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!, nhdu
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    character(len=*),parameter         :: rtname = 'FPUT_GYRO_RAW'
    character(len=80)                  :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2

  call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0

      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLJ(unit,1,i1,1,rowbuf,table%data(i1:i2)%otc    ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,2,i1,1,rowbuf,table%data(i1:i2)%fracsec,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,3,i1,1,rowbuf,table%data(i1:i2)%status ,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,4,i1,1,rowbuf,table%data(i1:i2)%quat0 ,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,5,i1,1,rowbuf,table%data(i1:i2)%quat1 ,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,6,i1,1,rowbuf,table%data(i1:i2)%quat2 ,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,7,i1,1,rowbuf,table%data(i1:i2)%quat3 ,status)
          fstat = max(fstat,status)
         enddo
      endif  
       if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  



  end  subroutine fput_gyro_raw



!-------------------------------------------------------------------------------

  subroutine fput_corratt(filename,append,table,ierr)  
 
    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(corratt_type)                 :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!, nhdu
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
!    character(len=16), dimension(table%ncol) :: ttype,tunit
!    character(len=4), dimension(table%ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_CORRATT'
    character(len=80)                  :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2
  
    
 
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif


    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
      

      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)


      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLD(unit,1,i1,1,rowbuf,table%data(i1:i2)%time,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,2,i1,1,rowbuf,table%data(i1:i2)%ra,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,3,i1,1,rowbuf,table%data(i1:i2)%dec,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,4,i1,1,rowbuf,table%data(i1:i2)%roll,status)
          fstat = max(fstat,status)
         enddo
      endif  
 

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  
   
  end subroutine fput_corratt

!-------------------------------------------------------------------------------



  subroutine fput_interatt(filename,append,table,ierr)  
 
  !------------------------------------------------------------
  ! Write intermediate eROSITA attitude file
  !------------------------------------------------------------

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(interatt_type)                :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!, nhdu
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
!    character(len=16), dimension(table%ncol) :: ttype,tunit
!    character(len=4), dimension(table%ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_INTERATT'
    character(len=80)                  :: errstr



    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2
  
 
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif


    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
 
      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLD(unit,1,i1,1,rowbuf,table%data(i1:i2)%time,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,2,i1,1,rowbuf,table%data(i1:i2)%ra,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,3,i1,1,rowbuf,table%data(i1:i2)%dec,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,4,i1,1,rowbuf,table%data(i1:i2)%roll,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,5,i1,1,rowbuf,table%data(i1:i2)%radec_err ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,6,i1,1,rowbuf,table%data(i1:i2)%roll_err ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,7,i1,1,rowbuf,table%data(i1:i2)%st_qual ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,8,i1,1,rowbuf,table%data(i1:i2)%st_maxdiff ,status)
          fstat = max(fstat,status)
         enddo
      endif  
 
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  
   
  end subroutine fput_interatt


!-------------------------------------------------------------------------------


  subroutine fput_st_inter(filename,append,table,ierr)  
 
  !------------------------------------------------------------
  ! Write star tracker specofic extensions of intermediate eROSITA attitude file
  !------------------------------------------------------------

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(inter_st_type)                :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!, nhdu
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
!    character(len=16), dimension(table%ncol) :: ttype,tunit
!    character(len=4), dimension(table%ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_ST_INTER'
    character(len=80)                  :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2
  
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
!      write(*,*) "1:",status,' extname ',table%extname
!      write(*,*) "table%cform:",table%cform
!      write(*,*) "table%cunit:",table%cunit
 
      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)

 !     write(*,*) "2:",status

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLD(unit,1,i1,1,rowbuf,table%data(i1:i2)%time ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,2,i1,1,rowbuf,table%data(i1:i2)%ra ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,3,i1,1,rowbuf,table%data(i1:i2)%dec ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,4,i1,1,rowbuf,table%data(i1:i2)%roll ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,5,i1,1,rowbuf,table%data(i1:i2)%stat1 ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,5,i1,1,rowbuf,table%data(i1:i2)%stat2 ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,6,i1,1,rowbuf,table%data(i1:i2)%radec_err ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,7,i1,1,rowbuf,table%data(i1:i2)%roll_err ,status)
          fstat = max(fstat,status)
          enddo
      endif  
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  

  end subroutine fput_st_inter

!-------------------------------------------------------------------------------


  subroutine fput_sed_inter(filename,append,table,ierr)  
 
  !------------------------------------------------------------
  ! Write SED 1 / 2 extensions of intermediate eROSITA attitude file
  !------------------------------------------------------------

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(inter_sed_type)                :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!, nhdu
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
!    character(len=16), dimension(table%ncol) :: ttype,tunit
!    character(len=4), dimension(table%ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_SED_INTER'
    character(len=80)                  :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2
  
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
 
      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLD(unit,1,i1,1,rowbuf,table%data(i1:i2)%time ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,2,i1,1,rowbuf,table%data(i1:i2)%ra ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,3,i1,1,rowbuf,table%data(i1:i2)%dec ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,4,i1,1,rowbuf,table%data(i1:i2)%roll ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,5,i1,1,rowbuf,table%data(i1:i2)%stat ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,6,i1,1,rowbuf,table%data(i1:i2)%radec_err ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,7,i1,1,rowbuf,table%data(i1:i2)%roll_err ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,8,i1,1,rowbuf,table%data(i1:i2)%angvx ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,9,i1,1,rowbuf,table%data(i1:i2)%angvy ,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,10,i1,1,rowbuf,table%data(i1:i2)%angvz ,status)
          fstat = max(fstat,status)
         enddo
      endif  
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  

  end subroutine fput_sed_inter

!-------------------------------------------------------------------------------

  subroutine fput_bokz_inter(filename,append,table,ierr)  
 
  !------------------------------------------------------------
  ! Write BOKZ extensions of intermediate eROSITA attitude file
  !------------------------------------------------------------

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(inter_bokz_type)                :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!, nhdu
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
!    character(len=16), dimension(table%ncol) :: ttype,tunit
!    character(len=4), dimension(table%ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_BOKZ_INTER'
    character(len=80)                  :: errstr

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2
  
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
 
      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLD(unit,1,i1,1,rowbuf,table%data(i1:i2)%time,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,2,i1,1,rowbuf,table%data(i1:i2)%ra ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,3,i1,1,rowbuf,table%data(i1:i2)%dec ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,4,i1,1,rowbuf,table%data(i1:i2)%roll ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,5,i1,1,rowbuf,table%data(i1:i2)%stat1 ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,6,i1,1,rowbuf,table%data(i1:i2)%stat2 ,status)
          fstat = max(fstat,status)
         enddo
      endif  
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  

  end subroutine fput_bokz_inter


!-------------------------------------------------------------------------------

  subroutine fput_gyro_inter(filename,append,table,ierr)  
 
  !------------------------------------------------------------
  ! Write BOKZ extensions of intermediate eROSITA attitude file
  !------------------------------------------------------------

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(inter_gyro_type)                :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!, nhdu
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
!    character(len=16), dimension(table%ncol) :: ttype,tunit
!    character(len=4), dimension(table%ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_GYRO_INTER'
    character(len=80)                  :: errstr

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2
  
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
 
      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLD(unit,1,i1,1,rowbuf,table%data(i1:i2)%time,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,2,i1,1,rowbuf,table%data(i1:i2)%ra ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,3,i1,1,rowbuf,table%data(i1:i2)%dec ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,4,i1,1,rowbuf,table%data(i1:i2)%roll ,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,5,i1,1,rowbuf,table%data(i1:i2)%stat ,status)
          fstat = max(fstat,status)

         enddo
      endif  
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  

  end subroutine fput_gyro_inter

!----------------------------------------------------------------------------------------------------


  subroutine fget_sunang(filename,ext,table,ierr)  

    !---------------------------------------------------------------------------
    ! Read data from SRG sun angles file
    ! Quaternion 
    !---------------------------------------------------------------------------
    implicit none

    type(sunang_type)                :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_SUNANG'
    character(len=80)                :: errstr

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
 
    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,errstr)
      else  
!        call fread_stdhdr(unit,table%stdhdr,status)
!        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status > 0) then 
          ! go to nhdu = 2:
          status=0
          call FTMAHD(unit,2, hdutype,status)
        endif
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_stdhdr(unit,table%stdhdr,status)
          if (status /= 0) ierr=3
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          !allocate(table%quat(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=i2
            if (colnum(1) > 0) call  FTGCVD(unit,colnum(1),i1,1,rowbuf,r8null,table%data(i1:i2)%time,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(2) > 0) call  FTGCVD(unit,colnum(2),i1,1,rowbuf,r8null,table%data(i1:i2)%ra_sol,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(3) > 0) call  FTGCVD(unit,colnum(3),i1,1,rowbuf,r8null,table%data(i1:i2)%dec_sol,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(4) > 0) call  FTGCVD(unit,colnum(4),i1,1,rowbuf,r8null,table%data(i1:i2)%ang_sol_srgx,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(5) > 0) call  FTGCVD(unit,colnum(5),i1,1,rowbuf,r8null,table%data(i1:i2)%ang_sol_srgz,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(6) > 0) call  FTGCVD(unit,colnum(6),i1,1,rowbuf,r8null,table%data(i1:i2)%ang_sol_srgxz,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(7) > 0) call  FTGCVD(unit,colnum(7),i1,1,rowbuf,r8null,table%data(i1:i2)%ra,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(8) > 0) call  FTGCVD(unit,colnum(8),i1,1,rowbuf,r8null,table%data(i1:i2)%dec,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(9) > 0) call  FTGCVD(unit,colnum(9),i1,1,rowbuf,r8null,table%data(i1:i2)%roll,fnull,status) 
            fstat = max(fstat,status)



            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif

  end subroutine fget_sunang



!----------------------------------------------------------------------------------------------------

  subroutine fput_sunang(filename,append,table,ierr)  
 
  !------------------------------------------------------------
  ! Write SRG-sun angles to fits
  !------------------------------------------------------------

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(sunang_type)                :: table
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!, nhdu
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
!    character(len=16), dimension(table%ncol) :: ttype,tunit
!    character(len=4), dimension(table%ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_SUNANG'
    character(len=80)                  :: errstr

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=table%shapehdr%naxis2
  
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,table%stdhdr,status)
      endif
    endif

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
      call FTPHBN(unit,nrows,table%ncol,table%cname,table%cform,table%cunit,table%extname,varidat,status)
      fstat=max(fstat,status)  
 
      call fwrite_stdhdr(unit,table%stdhdr,status)
      fstat=max(fstat,status)

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLD(unit,1,i1,1,rowbuf,table%data(i1:i2)%time,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,2,i1,1,rowbuf,table%data(i1:i2)%ra_sol ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,3,i1,1,rowbuf,table%data(i1:i2)%dec_sol ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,4,i1,1,rowbuf,table%data(i1:i2)%ang_sol_srgx ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,5,i1,1,rowbuf,table%data(i1:i2)%ang_sol_srgz ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,6,i1,1,rowbuf,table%data(i1:i2)%ang_sol_srgxz ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,7,i1,1,rowbuf,table%data(i1:i2)%ra ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,8,i1,1,rowbuf,table%data(i1:i2)%dec ,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,9,i1,1,rowbuf,table%data(i1:i2)%roll ,status)
          fstat = max(fstat,status)


         enddo
      endif  
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
     
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  

  end subroutine fput_sunang





!-------------------------------------------------------------------------------


  subroutine interpol_corratt(att,ev_time,ev_ra,ev_dec,ev_roll,quality,ierr)  

  !-----------------------------------------------------------------------
  ! Interpolates attitude data (RA, DEC, ROLL) for event with arrival 
  ! time EV_TIME.
  ! Uses simple linear interpolation between the 2 data points in ATT.  
  ! Simulations show that this is sufficiently accurate even very close to poles.
  ! Assumptions: 
  !    - ATT sorted in time.
  !    - ATT data spacing of the order of 1 sec or less.
  !    - ATT data set does not change between calls (save statements!)
  !-----------------------------------------------------------------------
 
    implicit none

    type(corratt_type)                    :: att
    real(kind=real64)                     :: ev_time
    real(kind=real64)                     :: ev_ra
    real(kind=real64)                     :: ev_dec
    real(kind=real64)                     :: ev_roll
    integer(kind=int32)                   :: quality
    integer(kind=int32)                   :: ierr
    integer(kind=int32)                   :: sdir
    integer(kind=int32), save             :: ind1=1, ind2=2 ,natt
    logical(kind=log32), save             :: first= .true.
    real(kind=real64)                     :: tfrac,tdelt,radiff 
    ! maximum tolerated gap in attitude time series:
    !real(kind=real64), parameter          :: tstep_max=10.0
    ! use for current test data:
    real(kind=real64), parameter          :: tstep_max=90.0
    character(len=*),parameter            :: rtname = 'INTERPOL_CORRATT'
    character(len=80)                     :: errstr    
  
    ierr = 0
    quality=0


    if (first) then
      natt= att%shapehdr%naxis2
      
      first=.false.
      if (natt < 2) then
        ierr=3
        errstr='too few attitude data points'
      endif
    else
      if (ev_time < att%data(1)%time) then
        ierr=2
        errstr = 'event time < lower limit'
      endif
      if (ev_time > att%data(natt)%time) then
        ierr=2
        errstr = 'event time > upper limit '
      endif 
!      ???? removed (17/02/2015) :
!      if (ev_time > att%data(1)%time) then
!        ierr=3
!        errstr = 'event time out of range  '
!      endif
    endif
 
! remove output in order to avoid lond error logs:
    if (ierr > 0 ) then
!      call WRERR(rtname,ierr,errstr)  
    else
      do 

         ! search attitude  time interval:
        sdir=1 
        if (ev_time >= att%data(ind1)%time .and. ev_time <= att%data(ind2)%time) exit
        if (ev_time >  att%data(ind2)%time) sdir=1
        if (ev_time <  att%data(ind1)%time) sdir=-1
        ind1 =  ind1 + sdir
        ind2 =  ind2 + sdir
        if (ind1 < 1) then
          ind1=1
          ind2=2
          ierr=2
          errstr='event time out of range'
          exit
         else if (ind2 > natt) then
          ind1=natt-1
          ind2=natt
          ierr=2
          errstr='event time out of range'
          exit
        endif
      enddo
      ! interpolate:
      if (ierr == 0) then
        tdelt=0.D0
        do 
          ! make sure that  time2!=time1
          tdelt=att%data(ind2)%time - att%data(ind1)%time
          if (tdelt > 0.D0 .or. ind2 == natt) exit
          ind2=ind2+1
        enddo
        if (tdelt .ne. 0.D0) then
          tfrac   = (ev_time-att%data(ind1)%time)/tdelt
          ! wrap ra around 0/360 deg
          radiff = att%data(ind2)%ra-att%data(ind1)%ra
          if (radiff > 180.D0) then 
            att%data(ind1)%ra =  att%data(ind1)%ra + 360.D0
            radiff = att%data(ind2)%ra-att%data(ind1)%ra
          else if (radiff <= -180.D0) then
            att%data(ind2)%ra =  att%data(ind2)%ra + 360.D0
            radiff = att%data(ind2)%ra-att%data(ind1)%ra
          endif 
          ev_ra   = att%data(ind1)%ra   + tfrac * radiff
          ev_dec  = att%data(ind1)%dec  + tfrac * (att%data(ind2)%dec-att%data(ind1)%dec)
          ev_roll = att%data(ind1)%roll + tfrac * (att%data(ind2)%roll-att%data(ind1)%roll) 
        else
          ierr=3
        endif
        ! wrap resulting RA:
        if (ev_ra > 360.D0) then
          ev_ra = mod(ev_ra,360.D0)
        else if (ev_ra < 0.D0) then
          ! should not happen, but just in case:
          ev_ra = ev_ra + 360.D0 
        endif
        if (tdelt > tstep_max) quality = 1
      endif
    endif


  end  subroutine interpol_corratt

!-------------------------------------------------------------------------------

 subroutine find_att_time(t0,tarr,tnear_index,flag,ierr)
 
   !-------------------------------------------------------------
   ! Return array indices tnear_index(1:2) of 2 entries of tarr either 
   !  - with tnear_index(1) ~= t0 (flag=0 
   !  - bracketing t0 (flag=1)
   !  - both < t0 (flag=2)
   !  - both > t0 (flag=3)
   ! On error return err=2 and flag=-1 
   ! 
   ! Uses sorting routine spsort from SLA library.
   !
   !-------------------------------------------------------------

   real(kind=real64)                               :: t0
   real(kind=real64),dimension(:)                  :: tarr
   integer(kind=int32),dimension(2)                :: tnear_index
   integer(kind=int32)                             :: ierr
   real(kind=real64)                               :: t_nearest!,t_prev,t_next
   real(kind=real64),parameter                     :: t_epsilon=1.D-6,t_large=10.0
   integer(kind=int32)                             :: ndata,kflag,sla_err
   integer(kind=int32), dimension(:), allocatable  :: iperm
   character(len=*), parameter                     :: rtname = 'find_att_time'


   ierr=0


   ndata=size(tarr)
   allocate(iperm(ndata))

   flag=-1
   kflag=1
   sla_err=0

   if (ndata < 2) then
     ierr=2
     call wrerr(rtname,ierr,'Error: too view data points in attitude time series)')
   else
     call spsort(abs(tarr-t0),ndata,iperm,kflag,sla_err)
     if (sla_err /= 0) then
       ierr = 2
       call wrerr(rtname,ierr,'Error: SLATEC error in SPSORT)')
     else
       t_nearest=tarr(iperm(1))
       if (abs(t_nearest-t0) < t_epsilon) then
         tnear_index(1)=iperm(1)
         tnear_index(2)=iperm(2)
         flag=0 ! t_nearest ~= t0
       else if (t_nearest < t0) then
         if (iperm(1) < (ndata-1)) then
           if (tarr(iperm(1)+1) >  t0) then
             tnear_index(1)=iperm(1)
             tnear_index(2)=iperm(1)+1
             flag=1 ! tnear_index(1:2) bracket t0
           endif
         else if (iperm(1) > 1) then
           if (tarr(iperm(1)-1) <  t0) then
             tnear_index(1)=iperm(1)-1
             tnear_index(1)=iperm(2)
             flag=2 ! tnear_index(1:2) < t0
           endif
         endif 
       else if (t_nearest > t0) then
         if (iperm(1) > 1) then
           if (tarr(iperm(1)-1) <  t0) then
             tnear_index(1)=iperm(1)-1
             tnear_index(2)=iperm(1)
             flag=1 !tnear_index(1:2) bracket t0
           endif
         else if (iperm(1) < (ndata-1)) then
           if (tarr(iperm(1)+1) >  t0) then
             tnear_index(1)=iperm(1)
             tnear_index(2)=iperm(1)+1
             flag=3 ! tnear_index(1:2) > t0
           endif
         endif          
       endif
     endif
   endif
      
   if (flag == -1 .or. abs(tarr(tnear_index(1))) > t_large .or. abs(tarr(tnear_index(2))) >  t_large) then
     ierr=2
     flag=-1
     call wrerr(rtname,ierr,'Error: Problems finding attitude data points close to t0)')
   endif

 end subroutine find_att_time

!-------------------------------------------------------------------------------
!
!
! subroutine interpolate_att(t0,tarr,ierr)
!
!
!   real(kind=real64)                               :: t0
!   real(kind=real64),dimension(:)                  :: tarr
!   integer(kind=int32),dimension(2)                :: tnear_index
!   integer(kind=int32)                             :: ierr
!   
!
!
! end subroutine interpolate_att
!
!
!
!-------------------------------------------------------------------------------





end module att_io
