!-----------------------------------------------------------------------
! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------

module fitsimage

!---------------------------------------------------------------------------------- 
! Provides routines for easy access to FITS images
! and headers based on CFITSIO
!
! 
! Author: GLamer, AIP,     date: 28-APR-2010
! 
! update: GLamer           date: 25-MAY-2010 (use cfitsio_ifaces,
!                                             use numkind, bug fixes)
! update: GLamer           date: 26-MAY-2010 (add functions match_size, match_wcs,
!                                              changed %image_arr to %arr)
! update: GLamer           date: 05-AUG-2010 (added subroutines copy_i2image
!                                             copy_i4image, copy_r4image)  
!          
! update: GLamer           date: 09-AUG-2010 (remove POINTER attribute from
!                                             subroutine parameters of 
!                                             type xximage_type)  
! update: GLamer           date: 25-AUG-2010 (- removed subroutine FPUT_I2IMAGE 
!                                             - added  FPUT_I1IMAGE
!                                             - corrected setting of BITPIX in 
!                                               FPUT_* routines
! update: GLamer           date: 18-NOV-2010  new header type 'dethdr_type'
!                                             + routines fread_dethdr, fget_dethdr
! update: GLamer           date: 02-MAR-2011  + include TELESCOP keyword in 
!                                               function match_expsures 
! update: GLamer           date: 07-MAR-2011  + updated fget_<xx>compress routines 
!                                               
! update: GLamer           date: 06-APR-2011    rename variable READWRITE to RWMODE 
! Update: GLamer           date: 14-SEP-2011    fix errors (array allocation) in equ2pixarr  
! Update: GLamer           date: 30-SEP-2011    new argument "order' for fget*compress routines 
! Update: GLamer           date: 22-JUN-2012    change stdhdr%tend to stdhdr%tstop
! Update: GLamer           date: 26-JUN-2012    add optional argument 'extname' to 
!                                               subroutines fget_stdhdr, fget_dethdr, fget_wcshdr
! Update: GLamer           date: 22-MAR-2013    fixed error in  PIX2EQU 
! Update: GLamer           date: 18-MAR-2014    changes to  fwrite_dethdr (check e_xxxx values) 
! Update: GLamer           date: 13-Oct-2014    added  CAM_NAMES [ FM1 .. FM7 ]

! subroutines:
!  freadkey_err
!  fread_shapehdr
!  fread_wcshdr
!  fread_xmmhdr
!  fread_refhdr
!  fread_stdhdr
!  fread_dethdr
!  fwrite_wcshdr
!  fwrite_refhdr
!  fwrite_stdhdr
!  fget_stdhdr
!  fget_dethdr
!  fget_wcshdr
!  fget_i4image
!  fget_i1image
!  fget_r4image
!  fget_r4compress
!  fget_l1compress
!  fget_i4compress
!  fput_i4image
!  fput_i1image
!  fput_r4image
!  copy_i1image
!  copy_i4image
!  copy_r4image
!  pix2equ 
!  equ2pix
!  equ2pixarr
! 
! functions:
!  match_wcs
!  match_size
!  match_exposures
!---------------------------------------------------------------------------------- 
  
  use  param
  use  cfitsio_ifaces
  use  numkind


  implicit none


  public


  integer(kind=int32)               :: decimals = 15
  integer(kind=int32),parameter     :: image_hdu=0,ascii_tbl=1,binary_tbl=2,any_hdu=-1

    integer(kind=int8)              :: i1null = -99
    integer(kind=int16)             :: i2null = -9999
    integer(kind=int32)             :: i4null = -999999
    real(kind=real32)               :: r4null = -9.999E9
    real(kind=real64)               :: r8null = -9.999D9 
    character(len=1)                :: cnull= ''

    integer(kind=int32),parameter          :: max_cam = 25 , num_tel=7   

    character(len=80), dimension(max_cam)  :: cam_names


  type pixel_type
    !sequence
    integer(kind=int32)  ::  id 
    integer(kind=int16)  ::  x
    integer(kind=int16)  ::  y
  end type pixel_type
 
  type pixlist_type
    !sequence
    integer(kind=int32)                        :: n
    type(pixel_type),dimension(:),allocatable  :: pix
  end type pixlist_type


  type shapehdr_type
    !sequence
    integer(kind=int32)   :: bitpix
    integer(kind=int32)   :: naxis
    integer(kind=int32)   :: naxis1=0
    integer(kind=int32)   :: naxis2=0 
    character(len=47) :: c_bitpix = "number of bits per data pixel" 
    character(len=47) :: c_naxis  = "number of data axes"
    character(len=47) :: c_naxis1 = "length of data axis 1"
    character(len=47) :: c_naxis2 = "length of data axis 2"
    logical(kind=log32)   :: e_bitpix
    logical(kind=log32)   :: e_naxis
    logical(kind=log32)   :: e_naxis1
    logical(kind=log32)   :: e_naxis2
  end type shapehdr_type


  type stdhdr_type
    !sequence
    character(len=8)  :: origin      = "MPE"
    character(len=8)  :: creator     = " "
    character(len=4)  :: config      = " " 
    character(len=3)  :: version     = " "
    character(len=4)  :: prefix      = " "
    character(len=3)  :: dscode      = " "
    character(len=16) :: dstype      = " "
    character(len=8)  :: owner       = " "
    character(len=8)  :: mission     = "SRG"      
    character(len=19) :: date        = " "
    character(len=8)  :: telescop    = "eROSITA"
    character(len=8)  :: instrume    = " "
    integer(kind=int32) :: ninst     = 0
    character(len=8),dimension(7)  :: instrumx    = " " 
    ! internal keyword, contains merged strings from instrumx(1:ninst)
    character(len=80)   :: instruments = " "
!    character(len=8)  :: instrum1    = " "
!    character(len=8)  :: instrum2    = " "
!    character(len=8)  :: instrum3    = " "
!    character(len=8)  :: instrum4    = " "
!    character(len=8)  :: instrum5    = " "
!    character(len=8)  :: instrum6    = " "
!    character(len=8)  :: instrum7    = " "
    character(len=8)  :: obs_mode    = " "
    character(len=8)  :: datamode    = " "      
    real(kind=real64) :: frametim    = 50.0584D0 
    character(len=8)  :: filter      = " "
    character(len=8)  :: skyfield    = " "   
    character(len=10) :: obs_id      = " "  
    character(len=8)  :: exp_id      = " "  
    character(len=20) :: observer    = " "  
    character(len=20) :: object      = " "  
    real(kind=real64) :: ra_obj
    real(kind=real64) :: dec_obj
    real(kind=real64) :: ra_cen
    real(kind=real64) :: dec_cen
    character(len=19) :: date_obs    = " "  
    character(len=19) :: date_end    = " "  
    real(kind=real64) :: tstart
    real(kind=real64) :: tstop
    real(kind=real64) :: mjdref      = 51543.875D0 
    real(kind=real64) :: timezero    = 0.D0
    character(len=1)  :: timeunit    = "s"  
    character(len=2)  :: timesys     = "TT"  
    real(kind=real64) :: ra_pnt
    real(kind=real64) :: dec_pnt
    real(kind=real64) :: pa_pnt
    character(len=3)  :: radecsys    = "FK5"  
    real(kind=real64) :: equinox  = 2000.0D0
    integer(kind=int32) :: flagsel   = 0  
!   default value should be 15:   
    integer(kind=int32) :: pat_sel   = 15   
    character(len=8)    :: gti_sel   = "NONE"
    integer(kind=int32) :: nenergy = 0
    real(kind=real64),dimension(9) :: eminx
    real(kind=real64),dimension(9) :: emaxx  
    character(len=8)  :: transmod   = " "  
    integer(kind=int32) :: pat_thr
    character(len=1)    :: t_offcor = "F"
    character(len=65536) :: history = " "
    character(len=8)  :: hduclass    = " "
    character(len=8)  :: hduvers     = "1.0.0"
    character(len=8)  :: hduclas1    = " "
    character(len=8)  :: hduclas2    = " "
    character(len=8)  :: longstr     = "OGIP 1.0"
    ! comment fields:
    character(len=47) :: c_origin    = "Origin of FITS file"
    character(len=47) :: c_creator   = "Name and version of routine"
    character(len=47) :: c_config    = "Pipeline configuration"
    character(len=47) :: c_version   = "File version"
    character(len=47) :: c_prefix    = "Standard products name prefix"
    character(len=47) :: c_dscode    = "Dataset code, codes telescope/filter/band"
    character(len=47) :: c_dstype    = "Dataset type: EventFile, Image, etc."
    character(len=47) :: c_owner     = "Sky owner: (MPE / IKI / MPE+IKI)"
    character(len=47) :: c_mission   = "Mission or test environment"
    character(len=47) :: c_date      = "Creation date"  
    character(len=47) :: c_telescop  = "Telescope / camera" 
    character(len=47) :: c_instrume  = "Instrument"
    character(len=47) :: c_ninst     = "Number of merged instruments"

    character(len=47),dimension(7) :: c_instrumx  = "Name of instrument"
    character(len=47) :: c_instrum1  = "Name of instrument 1"
    character(len=47) :: c_instrum2  = "Name of instrument 2"
    character(len=47) :: c_instrum3  = "Name of instrument 3"
    character(len=47) :: c_instrum4  = "Name of instrument 4"
    character(len=47) :: c_instrum5  = "Name of instrument 5"
    character(len=47) :: c_instrum6  = "Name of instrument 6"
    character(len=47) :: c_instrum7  = "Name of instrument 7"
    character(len=47) :: c_obs_mode  = "Observation mode (pointing/slew)"
    character(len=47) :: c_datamode  = "Data mode" 
    character(len=47) :: c_frametim  = "[ms] Nominal frame time"
    character(len=47) :: c_filter    = "filter wheel position"
    character(len=47) :: c_skyfield  = "survey field identifier"
    character(len=47) :: c_obs_id    = "Observation ID" 
    character(len=47) :: c_exp_id    = "Exposure ID"
    character(len=47) :: c_observer  = "Name of Observer" 
    character(len=47) :: c_object    = "Name of target of observation"
    character(len=47) :: c_ra_obj    = "[deg] RA of target"
    character(len=47) :: c_dec_obj   = "[deg] Dec of target"
    character(len=47) :: c_ra_cen    = "[deg] RA centre of sky field"
    character(len=47) :: c_dec_cen   = "[deg] DEC centre of sky field"
    character(len=47) :: c_date_obs  = "Start Time (UTC) of exposure"
    character(len=47) :: c_date_end  = "End Time (UTC) of exposure"
    character(len=47) :: c_tstart    = "Start time of exp. in units of TIME col"
    character(len=47) :: c_tstop      = "End time of exp. in units of TIME col"
    character(len=47) :: c_mjdref    = "[d] 2007-01-01T00:00:00 (=54101 MJD)"
    character(len=47) :: c_timezero  = "[s] Clock correction"
    character(len=47) :: c_timeunit  = "Time unit"
    character(len=47) :: c_timesys   = "Time system (TT=Terrestial Time)"
    character(len=47) :: c_ra_pnt    = "Actual (mean/median) pointing RA[deg] J2000"
    character(len=47) :: c_dec_pnt   = "Actual (mean/median) pointing DEC[deg] J2000"
    character(len=47) :: c_pa_pnt   = "Actual (mean/median) pointing pos angle"
    character(len=47) :: c_radecsys  = "Astrometric reference frame"                     
    character(len=47) :: c_equinox   = "Equinox" 
    character(len=47) :: c_flagsel   = "Flag selection"
    character(len=47) :: c_pat_sel   = "Pattern selection (1=s,3=s+d,7=s+d+t)"
    character(len=47) :: c_gti_sel   = "GTI selection" 
    character(len=47) :: c_nenergy     = "Number of energy intervals (0 .. 9)"
    character(len=47),dimension(9) :: c_eminx = "lower boundary [keV] energy interval"
    character(len=47),dimension(9) :: c_emaxx = "upper boundary [keV] energy interval"
    character(len=47) :: c_transmod  = " " 
    character(len=47) :: c_pat_thr  = " " 
    character(len=47) :: c_t_offcor  = "Time lags corrected by evprep T/F" 


!    character(len=47) :: c_emin1     = "lower boundary [keV] energy interval 1" 
!    character(len=47) :: c_emin2     = "lower boundary [keV] energy interval 2" 
!    character(len=47) :: c_emin3     = "lower boundary [keV] energy interval 3"  
!    character(len=47) :: c_emin4     = "lower boundary [keV] energy interval 4"  
!    character(len=47) :: c_emin5     = "lower boundary [keV] energy interval 5"  
!    character(len=47) :: c_emin6     = "lower boundary [keV] energy interval 6" 
!    character(len=47) :: c_emin7     = "lower boundary [keV] energy interval 7" 
!    character(len=47) :: c_emin8     = "lower boundary [keV] energy interval 8"  
!    character(len=47) :: c_emin9     = "lower boundary [keV] energy interval 9"  
!    character(len=47) :: c_emax1     = "upper boundary [keV] energy interval 1 "  
!    character(len=47) :: c_emax2     = "upper boundary [keV] energy interval 2"  
!    character(len=47) :: c_emax3     = "upper boundary [keV] energy interval 3"   
!    character(len=47) :: c_emax4     = "upper boundary [keV] energy interval 4"    
!    character(len=47) :: c_emax5     = "upper boundary [keV] energy interval 5"    
!    character(len=47) :: c_emax6     = "upper boundary [keV] energy interval 6"    
!    character(len=47) :: c_emax7     = "upper boundary [keV] energy interval 7"    
!    character(len=47) :: c_emax8     = "upper boundary [keV] energy interval 8"    
!    character(len=47) :: c_emax9     = "upper boundary [keV] energy interval 9"    
    
    character(len=47) :: c_history   = " " 
    character(len=47) :: c_hduclass  = "Conforms to HEASARC/OGIP conventions"
    character(len=47) :: c_hduvers   = "version of format"
    character(len=47) :: c_hduclas1  = " "
    character(len=47) :: c_hduclas2  = " "                   
    character(len=47) :: c_longstr   = "Support multi-line COMMENT/HISTORY records"
    logical(kind=log32)   :: e_origin      
    logical(kind=log32)   :: e_creator  
    logical(kind=log32)   :: e_config  
    logical(kind=log32)   :: e_version  
    logical(kind=log32)   :: e_prefix  
    logical(kind=log32)   :: e_dscode  
    logical(kind=log32)   :: e_dstype  
    logical(kind=log32)   :: e_owner     
    logical(kind=log32)   :: e_mission     
    logical(kind=log32)   :: e_date        
    logical(kind=log32)   :: e_telescop    
    logical(kind=log32)   :: e_instrume
! internal keyword
    logical(kind=log32)   :: e_instruments
    logical(kind=log32)   :: e_ninst  = .FALSE.
    logical(kind=log32),dimension(7)   :: e_instrumx  = .FALSE. 
!    logical(kind=log32)   :: e_instrum1  = .FALSE.  
!    logical(kind=log32)   :: e_instrum2  = .FALSE.  
!    logical(kind=log32)   :: e_instrum3  = .FALSE.  
!    logical(kind=log32)   :: e_instrum4  = .FALSE.  
!    logical(kind=log32)   :: e_instrum5  = .FALSE.  
!    logical(kind=log32)   :: e_instrum6  = .FALSE.  
!    logical(kind=log32)   :: e_instrum7  = .FALSE.   
    logical(kind=log32)   :: e_obs_mode    
    logical(kind=log32)   :: e_datamode    
    logical(kind=log32)   :: e_frametim  
    logical(kind=log32)   :: e_filter
    logical(kind=log32)   :: e_skyfield   
    logical(kind=log32)   :: e_obs_id      
    logical(kind=log32)   :: e_exp_id      
    logical(kind=log32)   :: e_observer    
    logical(kind=log32)   :: e_object      
    logical(kind=log32)   :: e_ra_obj      
    logical(kind=log32)   :: e_dec_obj
    logical(kind=log32)   :: e_ra_cen      
    logical(kind=log32)   :: e_dec_cen
    logical(kind=log32)   :: e_date_obs    
    logical(kind=log32)   :: e_date_end    
    logical(kind=log32)   :: e_tstart      
    logical(kind=log32)   :: e_tstop        
    logical(kind=log32)   :: e_mjdref      
    logical(kind=log32)   :: e_timezero    
    logical(kind=log32)   :: e_timeunit    
    logical(kind=log32)   :: e_timesys     
    logical(kind=log32)   :: e_ra_pnt      
    logical(kind=log32)   :: e_dec_pnt     
    logical(kind=log32)   :: e_pa_pnt      
    logical(kind=log32)   :: e_radecsys    
    logical(kind=log32)   :: e_equinox 
    logical(kind=log32)   :: e_flagsel
    logical(kind=log32)   :: e_pat_sel
    logical(kind=log32)   :: e_gti_sel
    logical(kind=log32)   :: e_nenergy
    logical(kind=log32), dimension(9)  ::  e_eminx = .FALSE.
    logical(kind=log32), dimension(9)  ::  e_emaxx = .FALSE.
    logical(kind=log32)   :: e_transmod
    logical(kind=log32)   :: e_pat_thr
    logical(kind=log32)   :: e_t_offcor


!    logical(kind=log32)   :: e_emin1 = .FALSE.
!    logical(kind=log32)   :: e_emin2 = .FALSE. 
!    logical(kind=log32)   :: e_emin3 = .FALSE. 
!    logical(kind=log32)   :: e_emin4 = .FALSE. 
!    logical(kind=log32)   :: e_emin5 = .FALSE. 
!    logical(kind=log32)   :: e_emin6 = .FALSE. 
!    logical(kind=log32)   :: e_emin7 = .FALSE. 
!    logical(kind=log32)   :: e_emin8 = .FALSE.
!    logical(kind=log32)   :: e_emin9 = .FALSE. 
!    logical(kind=log32)   :: e_emax1 = .FALSE.
!    logical(kind=log32)   :: e_emax2 = .FALSE. 
!    logical(kind=log32)   :: e_emax3 = .FALSE. 
!    logical(kind=log32)   :: e_emax4 = .FALSE. 
!    logical(kind=log32)   :: e_emax5 = .FALSE. 
!    logical(kind=log32)   :: e_emax6 = .FALSE. 
!    logical(kind=log32)   :: e_emax7 = .FALSE. 
!    logical(kind=log32)   :: e_emax8 = .FALSE.
!    logical(kind=log32)   :: e_emax9 = .FALSE. 
    logical(kind=log32)   :: e_history     
    logical(kind=log32)   :: e_hduclass
    logical(kind=log32)   :: e_hduvers
    logical(kind=log32)   :: e_hduclas1
    logical(kind=log32)   :: e_hduclas2
    logical(kind=log32)   :: e_longstr     


  end type stdhdr_type   

! Miscellaneous keywords used by source detection tasks 
  type dethdr_type
    !sequence
    ! erbackmap
    real(kind=real32)    :: chisq = 0.0
    ! min, max coordinates of non-zero area in detection masks,
    ! used by erbox, mldetect:  
    integer(kind=int32)  :: xdatamin = 0
    integer(kind=int32)  :: ydatamin = 0
    integer(kind=int32)  :: xdatamax = 0
    integer(kind=int32)  :: ydatamax = 0

    character(len=47) :: c_chisq  = "Chi^2 of spline fit"
    character(len=47) :: c_xdatamin  = "[image pixel] Low x boundary of valid mask area"
    character(len=47) :: c_ydatamin  = "[image pixel] Low y boundary of valid mask area"    
    character(len=48) :: c_xdatamax  = "[image pixel] High x boundary of valid mask area"
    character(len=48) :: c_ydatamax  = "[image pixel] High y boundary of valid mask area"

    logical(kind=log32)   :: e_chisq     = .false.
    logical(kind=log32)   :: e_xdatamin  = .false. 
    logical(kind=log32)   :: e_ydatamin  = .false. 
    logical(kind=log32)   :: e_xdatamax  = .false. 
    logical(kind=log32)   :: e_ydatamax  = .false. 
             
  end type dethdr_type

! Scaling keywords for RA, DEC columns in event files (columns 2+3):


  type evscalehdr_type
    !sequence
    real(kind=real64)    :: tzero_ra = 0.D0
    real(kind=real64)    :: tscal_ra = 1.D-6
    real(kind=real64)    :: tzero_dec = 0.D0    
    real(kind=real64)    :: tscal_dec = 1.D-6
!    real(kind=real64)    :: tzero_subx = -0.498046875D0 
!    real(kind=real64)    :: tscal_subx = 0.00390625D0
!    real(kind=real64)    :: tzero_suby = -0.498046875D0   
!    real(kind=real64)    :: tscal_suby =  0.00390625D0
   ! Changed 24/05/2019 (GL):
    real(kind=real64)    :: tzero_subx = -8.43333333333333D-1
    real(kind=real64)    :: tscal_subx =  6.66666666666667D-3
    real(kind=real64)    :: tzero_suby = -8.43333333333333D-1   
    real(kind=real64)    :: tscal_suby =  6.66666666666667D-3

    character(len=47)    :: c_tzero_ra = "Offset for RA scaling"
    character(len=47)    :: c_tscal_ra = "Scaling value for RA column"
    character(len=47)    :: c_tzero_dec = "Offset for DEC scaling"
    character(len=47)    :: c_tscal_dec = "Scaling value for SUBX column" 
    character(len=50)    :: c_tzero_subx = "Offset for SUBX scaling (center of first subpixel)"
    character(len=47)    :: c_tscal_subx = "Scaling value for SUBX column (1./256.)"
    character(len=50)    :: c_tzero_suby = "Offset for SUBY scaling (center of first subpixel)"
    character(len=47)    :: c_tscal_suby = "Scaling value for SUBY column (1./256.)"


    logical(kind=log32)  :: e_tzero_ra = .false.
    logical(kind=log32)  :: e_tscal_ra = .false.     
    logical(kind=log32)  :: e_tzero_dec = .false.
    logical(kind=log32)  :: e_tscal_dec = .false.
    logical(kind=log32)  :: e_tzero_subx = .false.
    logical(kind=log32)  :: e_tscal_subx = .false.
    logical(kind=log32)  :: e_tzero_suby = .false.
    logical(kind=log32)  :: e_tscal_suby = .false.

  end type evscalehdr_type

!-------------------------- 

  type twcs_hdr_type
    !sequence
    character(len=16)      :: colname_x='X'
    character(len=8)       :: tctyp_x='RA---SIN'
    character(len=8)       :: tcuni_x='deg'
    integer(kind=int32)    :: tcrpx_x=0
    real(kind=real64)      :: tcrvl_x=0.D0
    real(kind=real64)      :: tcdlt_x=1.38888890958495D-05
    character(len=47)      :: c_tctyp_x='Coordinate / projection type of axis 1'
    character(len=47)      :: c_tcuni_x='Physical unit of axis 1'
    character(len=47)      :: c_tcrpx_x='Reference pixel coordinate 1 of projected image' 
    character(len=47)      :: c_tcrvl_x='Physical coordinate 1 for reference pixel' 
    character(len=47)      :: c_tcdlt_x='[deg/pix] Increment in axis 1 at ref. pixel'  
    character(len=16)      :: colname_y='Y'
    character(len=8)       :: tctyp_y='DEC--SIN'
    character(len=8)       :: tcuni_y='deg'
    integer(kind=int32)    :: tcrpx_y=0
    real(kind=real64)      :: tcrvl_y=0.D0
    real(kind=real64)      :: tcdlt_y=1.38888890958495D-05
    character(len=47)      :: c_tctyp_y='Coordinate / projection type of axis 2'
    character(len=47)      :: c_tcuni_y='Physical unit of axis 1'
    character(len=47)      :: c_tcrpx_y='Reference pixel coordinate 2 of projected image'
    character(len=47)      :: c_tcrvl_y='Physical coordinate 1 for reference pixel'
    character(len=47)      :: c_tcdlt_y='[deg/pix] Increment in axis 1 at ref. pixel'
  end type twcs_hdr_type

!-------------------------- 

  type wcs_refhdr_type
    !sequence
    character(len=8)       :: refxctyp
    character(len=8)       :: refxcuni
    integer(kind=int32)    :: refxcrpx 
    real(kind=real64)      :: refxcrvl
    real(kind=real64)      :: refxcdlt
    real(kind=real64)      :: refxlmin
    real(kind=real64)      :: refxlmax
    real(kind=real64)      :: refxdmin
    real(kind=real64)      :: refxdmax
    character(len=8)       :: refyctyp
    character(len=8)       :: refycuni
    integer(kind=int32)    :: refycrpx
    real(kind=real64)      :: refycrvl
    real(kind=real64)      :: refycdlt
    real(kind=real64)      :: refylmin
    real(kind=real64)      :: refylmax
    real(kind=real64)      :: refydmin
    real(kind=real64)      :: refydmax
    character(len=47) :: c_refxctyp = "WCS Coord. type: RA tangent plane projection"
    character(len=47) :: c_refxcuni = "WCS Physical units of X axis" 
    character(len=47) :: c_refxcrpx = "WCS axis reference pixel of projected image"
    character(len=47) :: c_refxcrvl = "[deg] WCS coord. at X axis ref. pixel"
    character(len=47) :: c_refxcdlt = "[deg/pix] WCS X increment at ref. pixel"
    character(len=47) :: c_refxlmin = "WCS min legal QPOE proj. image X axis val"
    character(len=47) :: c_refxlmax = "WCS max legal QPOE proj. image X axis val"
    character(len=47) :: c_refxdmin = "WCS minimum projected image X axis data value"
    character(len=47) :: c_refxdmax = "WCS maximum projected image X axis data value"
    character(len=47) :: c_refyctyp = "WCS Coord. type: DEC tangent plane projection"
    character(len=47) :: c_refycuni = "WCS Physical units of Y axis"
    character(len=47) :: c_refycrpx = "WCS axis reference pixel of projected image"
    character(len=47) :: c_refycrvl = "[deg] WCS coord. at Y axis ref. pixel"
    character(len=47) :: c_refycdlt = "[deg/pix] WCS Y increment at ref. pixel"
    character(len=47) :: c_refylmin = "WCS min legal QPOE projected image Y axis value"
    character(len=47) :: c_refylmax = "WCS max legal QPOE projected image Y axis value" 
    character(len=47) :: c_refydmin = "WCS minimum projected image Y axis data value"
    character(len=47) :: c_refydmax = "WCS maximum projected image Y axis data value"
    logical(kind=log32)   :: e_refxctyp
    logical(kind=log32)   :: e_refxcuni
    logical(kind=log32)   :: e_refxcrpx
    logical(kind=log32)   :: e_refxcrvl
    logical(kind=log32)   :: e_refxcdlt
    logical(kind=log32)   :: e_refxlmin
    logical(kind=log32)   :: e_refxlmax
    logical(kind=log32)   :: e_refxdmin
    logical(kind=log32)   :: e_refxdmax
    logical(kind=log32)   :: e_refyctyp
    logical(kind=log32)   :: e_refycuni
    logical(kind=log32)   :: e_refycrpx
    logical(kind=log32)   :: e_refycrvl
    logical(kind=log32)   :: e_refycdlt
    logical(kind=log32)   :: e_refylmin
    logical(kind=log32)   :: e_refylmax
    logical(kind=log32)   :: e_refydmin
    logical(kind=log32)   :: e_refydmax

  end type wcs_refhdr_type
                         
  type wcs_xmmhdr_type
    !sequence
    character(len=8)       :: ctype1
    character(len=8)       :: cunit1
    real(kind=real64)      :: crpix1
    real(kind=real64)      :: crval1
    real(kind=real64)      :: cdelt1
    character(len=8)       :: ctype1l
    real(kind=real64)      :: crpix1l
    real(kind=real64)      :: crval1l
    real(kind=real64)      :: cdelt1l
    real(kind=real64)      :: ltv1
    real(kind=real64)      :: ltm1_1  
    character(len=8)       :: ctype2
    real(kind=real64)      :: crpix2
    real(kind=real64)      :: crval2
    character(len=8)       :: cunit2
    real(kind=real64)      :: cdelt2
    character(len=8)       :: ctype2l
    real(kind=real64)      :: crpix2l
    real(kind=real64)      :: crval2l
    real(kind=real64)      :: cdelt2l
    real(kind=real64)      :: ltv2
    real(kind=real64)      :: ltm2_2  
    character(len=47) :: c_ctype1  = "Coord. type: RA tangent plane projection"  
    character(len=47) :: c_cunit1  = "Physical units of X axis"  
    character(len=47) :: c_crpix1  = "WCS reference pixel"   
    character(len=47) :: c_crval1  = "[deg] coord. at X axis ref. pixel"        
    character(len=47) :: c_cdelt1  = "WCS pixel size"  
    character(len=47) :: c_ctype1l = "WCS coordinate name"   
    character(len=47) :: c_crpix1l = "WCS reference pixel"       
    character(len=47) :: c_crval1l = "WCS reference pixel value" 
    character(len=47) :: c_cdelt1l = "WCS pixel size" 
    character(len=47) :: c_ltv1    = "offset for logical to physical pixel (axis 1)"    
    character(len=47) :: c_ltm1_1  = "slope for logical to physical pixel (axis 1)"        
    character(len=47) :: c_ctype2  = "Coord. type: DEC tangent plane projection"  
    character(len=47) :: c_crpix2  = "WCS reference pixel"
    character(len=47) :: c_crval2  = "[deg] coord. at Y axis ref. pixel"    
    character(len=47) :: c_cunit2  = "Physical units of Y axis"       
    character(len=47) :: c_cdelt2  = "WCS pixel size"  
    character(len=47) :: c_ctype2l = "WCS coordinate name" 
    character(len=47) :: c_crpix2l = "WCS reference pixel"   
    character(len=47) :: c_crval2l = "WCS reference pixel value"       
    character(len=47) :: c_cdelt2l = "WCS pixel size"
    character(len=47) :: c_ltv2    = "offset for logical to physical pixel (axis 2)"
    character(len=47) :: c_ltm2_2  = "slope for logical to physical pixel (axis 2)" 
    logical(kind=log32)   :: e_ctype1  
    logical(kind=log32)   :: e_cunit1  
    logical(kind=log32)   :: e_crpix1  
    logical(kind=log32)   :: e_crval1  
    logical(kind=log32)   :: e_cdelt1  
    logical(kind=log32)   :: e_ctype1l 
    logical(kind=log32)   :: e_crpix1l 
    logical(kind=log32)   :: e_crval1l 
    logical(kind=log32)   :: e_cdelt1l 
    logical(kind=log32)   :: e_ltv1    
    logical(kind=log32)   :: e_ltm1_1  
    logical(kind=log32)   :: e_ctype2  
    logical(kind=log32)   :: e_crpix2  
    logical(kind=log32)   :: e_crval2  
    logical(kind=log32)   :: e_cunit2  
    logical(kind=log32)   :: e_cdelt2  
    logical(kind=log32)   :: e_ctype2l 
    logical(kind=log32)   :: e_crpix2l 
    logical(kind=log32)   :: e_crval2l 
    logical(kind=log32)   :: e_cdelt2l 
    logical(kind=log32)   :: e_ltv2    
    logical(kind=log32)   :: e_ltm2_2  
 
  end type wcs_xmmhdr_type



  !-----------------------------------------------------------------------------
  ! 
  ! Image WCS header as created by ftools (e.g. extractor v 4.36):
  !
  !-----------------------------------------------------------------------------
                        
  type wcs_imahdr_type
    !sequence
    character(len=8)       :: ctype1
    real(kind=real64)      :: crpix1 = 0.D0
    real(kind=real64)      :: drpix1 = 0.D0
    real(kind=real64)      :: crval1 = 0.D0
    real(kind=real64)      :: cdelt1 = 0.D0
    character(len=8)       :: cunit1 = 'deg     '
    real(kind=real64)      :: ddelt1 = 0.D0

    character(len=8)       :: ctype1p
    real(kind=real64)      :: crpix1p = 0.D0
    real(kind=real64)      :: crval1p = 0.D0
    real(kind=real64)      :: cdelt1p = 0.D0
    character(len=8)       :: wcsnamep = 'PHYSICAL'   
    character(len=8)       :: wcsty1p = 'PHYSICAL' 
    real(kind=real64)      :: ltv1 = 0.D0
    real(kind=real64)      :: ltm1_1 = 0.D0  
    character(len=8)       :: ctype2
    real(kind=real64)      :: crpix2 = 0.D0
    real(kind=real64)      :: drpix2 = 0.D0
    real(kind=real64)      :: crval2 = 0.D0
    real(kind=real64)      :: cdelt2 = 0.D0
    character(len=8)       :: cunit2 = 'deg    '     
    real(kind=real64)      :: ddelt2 = 0.D0
    real(kind=real64)      :: crota2 = 0.D0
    character(len=8)       :: ctype2p
    real(kind=real64)      :: crpix2p = 0.D0
    real(kind=real64)      :: crval2p = 0.D0
    real(kind=real64)      :: cdelt2p = 0.D0
    character(len=8)       :: wcsty2p  = 'PHYSICAL'    
    real(kind=real64)      :: ltv2 = 0.D0
    real(kind=real64)      :: ltm2_2 = 0.D0  
    character(len=47) :: c_ctype1  = "Coord. type: RA tangent plane projection"  
    character(len=47) :: c_crpix1  = "WCS reference pixel" 
    character(len=47) :: c_drpix1  = "original X axis reference pixel"   
    character(len=47) :: c_crval1  = "[deg] coord. at X axis ref. pixel"        
    character(len=47) :: c_cdelt1  = "WCS pixel size" 
    character(len=47) :: c_cunit1  = "WCS pixel size unit" 
    character(len=47) :: c_ddelt1  = "original X axis increment"  
    character(len=47) :: c_ctype1p = "WCS coordinate name"   
    character(len=47) :: c_crpix1p = "WCS reference pixel"       
    character(len=47) :: c_crval1p = "WCS reference pixel value" 
    character(len=47) :: c_cdelt1p = "WCS pixel size"
    character(len=47) :: c_wcsnamep = " "    
    character(len=47) :: c_wcsty1p = " "
    character(len=47) :: c_ltv1    = "offset for logical to physical pixel (axis 1)"    
    character(len=47) :: c_ltm1_1  = "slope for logical to physical pixel (axis 1)"        
    character(len=47) :: c_ctype2  = "Coord. type: DEC tangent plane projection"  
    character(len=47) :: c_crpix2  = "WCS reference pixel"
    character(len=47) :: c_drpix2  = "original Y axis reference pixel"
    character(len=47) :: c_crval2  = "[deg] coord. at Y axis ref. pixel"    
    character(len=47) :: c_cdelt2  = "WCS pixel size" 
    character(len=47) :: c_cunit2  = "WCS pixel size" 
    character(len=47) :: c_ddelt2  = "original Y axis increment"
    character(len=47) :: c_crota2  = "Sky coord rotation angle"
    character(len=47) :: c_ctype2p = "WCS coordinate name" 
    character(len=47) :: c_crpix2p = "WCS reference pixel"   
    character(len=47) :: c_crval2p = "WCS reference pixel value"       
    character(len=47) :: c_cdelt2p = "WCS pixel size"
    character(len=47) :: c_wcsty2p = ""
    character(len=47) :: c_ltv2    = "offset for logical to physical pixel (axis 2)"
    character(len=47) :: c_ltm2_2  = "slope for logical to physical pixel (axis 2)"
    logical(kind=log32)   :: e_ctype1        
    logical(kind=log32)   :: e_crpix1 
    logical(kind=log32)   :: e_drpix1 
    logical(kind=log32)   :: e_crval1  
    logical(kind=log32)   :: e_cdelt1 
    logical(kind=log32)   :: e_cunit1 
    logical(kind=log32)   :: e_ddelt1        
    logical(kind=log32)   :: e_ctype1p 
    logical(kind=log32)   :: e_crpix1p       
    logical(kind=log32)   :: e_crval1p 
    logical(kind=log32)   :: e_cdelt1p
    logical(kind=log32)   :: e_wcsnamep          
    logical(kind=log32)   :: e_wcsty1p       
    logical(kind=log32)   :: e_ltv1    
    logical(kind=log32)   :: e_ltm1_1        
    logical(kind=log32)   :: e_ctype2  
    logical(kind=log32)   :: e_crpix2 
    logical(kind=log32)   :: e_drpix2        
    logical(kind=log32)   :: e_crval2  
    logical(kind=log32)   :: e_cdelt2
    logical(kind=log32)   :: e_cunit2
    logical(kind=log32)   :: e_ddelt2
    logical(kind=log32)   :: e_crota2 
    logical(kind=log32)   :: e_ctype2p 
    logical(kind=log32)   :: e_crpix2p       
    logical(kind=log32)   :: e_crval2p 
    logical(kind=log32)   :: e_cdelt2p 
    logical(kind=log32)   :: e_wcsty2p       
    logical(kind=log32)   :: e_ltv2    
    logical(kind=log32)   :: e_ltm2_2  
  end type wcs_imahdr_type


  type l1cube_type
    !sequence
    logical(kind=log8),  pointer, dimension(:,:,:) :: arr => NULL()
    type(shapehdr_type),   pointer,dimension(:)    :: shapehdr => NULL()
    type(stdhdr_type),   pointer,dimension(:)      :: stdhdr => NULL()
    type(wcs_refhdr_type),   pointer,dimension(:)  :: refhdr => NULL()
    type(wcs_imahdr_type),   pointer,dimension(:)  :: wcshdr => NULL()
    type(dethdr_type),   pointer,dimension(:)      :: dethdr => NULL()
  end type l1cube_type

  type i1cube_type
    !sequence
    integer(kind=int8),  pointer, dimension(:,:,:) :: arr => NULL()
    type(shapehdr_type),   pointer,dimension(:)    :: shapehdr  => NULL()
    type(stdhdr_type),   pointer,dimension(:)      :: stdhdr  => NULL()
    type(wcs_refhdr_type),   pointer,dimension(:)  :: refhdr  => NULL()
    type(wcs_imahdr_type),   pointer,dimension(:)  :: wcshdr => NULL()
    type(dethdr_type),   pointer,dimension(:)      :: dethdr => NULL()
  end type i1cube_type


  type i4cube_type
    !sequence
    integer(kind=int32), pointer, dimension(:,:,:)      :: arr => NULL()
    type(shapehdr_type), pointer,dimension(:)           :: shapehdr  => NULL()
    type(stdhdr_type),   pointer,dimension(:)           :: stdhdr => NULL()
    type(wcs_refhdr_type),pointer,dimension(:)          :: refhdr => NULL()
    type(wcs_imahdr_type),pointer,dimension(:)          :: wcshdr => NULL()
    type(dethdr_type),   pointer,dimension(:)           :: dethdr => NULL()
  end type i4cube_type


  type r4cube_type
    !sequence
    real(kind=real32),  pointer, dimension(:,:,:)    :: arr => NULL()
    type(shapehdr_type),   pointer,dimension(:)      :: shapehdr => NULL()
    type(stdhdr_type),     pointer,dimension(:)      :: stdhdr => NULL()
    type(wcs_refhdr_type), pointer,dimension(:)      :: refhdr => NULL()
    type(wcs_imahdr_type), pointer,dimension(:)      :: wcshdr => NULL()
    type(dethdr_type),     pointer,dimension(:)        :: dethdr => NULL()
  end type r4cube_type



  type i1image_type
    !sequence
    integer(kind=int8),  pointer, dimension(:,:) :: arr => NULL()
    type(shapehdr_type)   :: shapehdr
    type(stdhdr_type)     :: stdhdr
    type(wcs_refhdr_type) :: refhdr
    type(wcs_imahdr_type) :: wcshdr
    type(dethdr_type)     :: dethdr
  end type i1image_type 



  type i2image_type
    !sequence
    integer(kind=int16),  pointer, dimension(:,:) :: arr => NULL()
    type(shapehdr_type)   :: shapehdr
    type(stdhdr_type)     :: stdhdr
    type(wcs_refhdr_type) :: refhdr
    type(wcs_imahdr_type) :: wcshdr
    type(dethdr_type)     :: dethdr
  end type i2image_type 


  type i4image_type
    !sequence
    integer(kind=int32),  pointer, dimension(:,:) :: arr => NULL()
    type(shapehdr_type)   :: shapehdr
    type(stdhdr_type)     :: stdhdr
    type(wcs_refhdr_type) :: refhdr
    type(wcs_imahdr_type) :: wcshdr
    type(dethdr_type)     :: dethdr
  end type i4image_type 



  type r4image_type
    !sequence
    real(kind=real32),  pointer, dimension(:,:) :: arr => NULL()
    type(shapehdr_type)   :: shapehdr
    type(stdhdr_type)     :: stdhdr
    type(wcs_refhdr_type) :: refhdr
    type(wcs_imahdr_type) :: wcshdr
    type(dethdr_type)     :: dethdr
  end type r4image_type 


  type r8image_type
    !sequence
    real(kind=real64),  pointer, dimension(:,:) :: arr => NULL()
    type(shapehdr_type)   :: shapehdr
    type(stdhdr_type)     :: stdhdr
    type(wcs_refhdr_type) :: refhdr
    type(wcs_imahdr_type) :: wcshdr
    type(dethdr_type)     :: dethdr
  end type r8image_type 




contains






!-------------------------------------------------------------------------------

  subroutine  freadkey_err(rname,keyword,cfits_status,exist,ierr)

  !-----------------------------------------------------------------------------
  ! call after each CFITSIO keyword read access
  ! in case CFITSIO (status > 0) :
  !   - increments ierr and exist=.false.
  !   - calls wrerr      
  !   - resets cfits_status to 0
  !----------------------------------------------------------------------------- 




    integer(kind=int32)   :: cfits_status,ierr
    character(len=*)      :: rname,keyword
    character(len=1024)   :: message
    logical(kind=log32)   :: exist
    
    ierr = 0
    exist = .true.
    
    if (cfits_status > 0) then 
      exist = .false.
      cfits_status = 0
      message="Warning: "//trim(rname)//  ": keyword //" // keyword//" not found"
      ! skip warning for the time being:
      !call wrchatty(trim(rname),trim(message))
    endif   
  
  end  subroutine  freadkey_err

!-------------------------------------------------------------------------------

  subroutine fread_shapehdr(unit,header,ierr)


    type(shapehdr_type)           :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)            :: status,ierr 
    character(len=17)             :: rtname = 'fread_shapeheader'

    ierr=0
    status=0

    call FTGKYJ(unit,'BITPIX',header%bitpix ,header%c_bitpix,status)
    call freadkey_err(rtname,'BITPIX',status,header%e_bitpix,ierr)  
    call FTGKYJ(unit,'NAXIS',header%naxis   ,header%c_naxis,status)
    call freadkey_err(rtname,'NAXIS',status ,header%e_naxis,ierr)  
    call FTGKYJ(unit,'NAXIS1',header%naxis1 ,header%c_naxis1,status)
    call freadkey_err(rtname,'NAXIS1',status,header%e_naxis1,ierr)
    call FTGKYJ(unit,'NAXIS2',header%naxis2 ,header%c_naxis2,status)
    call freadkey_err(rtname,'NAXIS2',status,header%e_naxis2,ierr)   
    

!   If one or more keywords are missing, ierr =1:
    if (ierr > 0) ierr=1


  end  subroutine  fread_shapehdr 

!-------------------------------------------------------------------------------

  subroutine fread_wcshdr(unit,header,ierr)

  !-----------------------------------------------------------------------------
  ! 
  ! Read  WCS header as created by ftools (e.g. extractor v 4.36)
  !
  !-----------------------------------------------------------------    
    type(wcs_imahdr_type)         :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    character(len=12)             :: rtname = 'fread_wcshdr'

    ierr=0
    status=0

    call FTGKYS(unit,'CTYPE1',header%ctype1 ,header%c_ctype1 ,status)
    call freadkey_err(rtname,'CTYPE1',status,header%e_ctype1 ,ierr)
    call FTGKYD(unit,'CRPIX1',header%crpix1 ,header%c_crpix1,status)
    call freadkey_err(rtname,'CRPIX1',status,header%e_crpix1,ierr)
    call FTGKYD(unit,'DRPIX1',header%drpix1 ,header%c_drpix1,status)
    call freadkey_err(rtname,'DRPIX1',status,header%e_drpix1,ierr)
    call FTGKYD(unit,'CRVAL1',header%crval1 ,header%c_crval1 ,status)
    call freadkey_err(rtname,'CRVAL1',status,header%e_crval1 ,ierr)
    call FTGKYD(unit,'CDELT1',header%cdelt1 ,header%c_cdelt1 ,status)
    call freadkey_err(rtname,'CDELT1',status,header%e_cdelt1, ierr)
    call FTGKYS(unit,'CUNIT1',header%cunit1 ,header%c_cunit1 ,status)
    call freadkey_err(rtname,'CUNIT1',status,header%e_cunit1 ,ierr)
    call FTGKYD(unit,'DDELT1',header%ddelt1 ,header%c_ddelt1 ,status)
    call freadkey_err(rtname,'DDELT1',status,header%e_ddelt1, ierr)
    call FTGKYS(unit,'CTYPE2',header%ctype2 ,header%c_ctype2 ,status)
    call freadkey_err(rtname,'CTYPE2',status,header%e_ctype2 ,ierr)
    call FTGKYD(unit,'CRPIX2',header%crpix2 ,header%c_crpix2,status)
    call freadkey_err(rtname,'CRPIX2',status,header%e_crpix2,ierr)
    call FTGKYD(unit,'DRPIX2',header%drpix2 ,header%c_drpix2,status)
    call freadkey_err(rtname,'DRPIX2',status,header%e_drpix2,ierr)
    call FTGKYD(unit,'CRVAL2',header%crval2 ,header%c_crval2 ,status)
    call freadkey_err(rtname,'CRVAL2',status,header%e_crval2 ,ierr)
    call FTGKYD(unit,'CDELT2',header%cdelt2 ,header%c_cdelt2 ,status)
    call freadkey_err(rtname,'CDELT2',status,header%e_cdelt2 ,ierr)
    call FTGKYS(unit,'CUNIT2',header%cunit2 ,header%c_cunit2 ,status)
    call freadkey_err(rtname,'CUNIT2',status,header%e_cunit2 ,ierr)
    call FTGKYD(unit,'DDELT2',header%ddelt2 ,header%c_ddelt2 ,status)
    call freadkey_err(rtname,'DDELT2',status,header%e_ddelt2 ,ierr)
    call FTGKYD(unit,'CROTA2',header%crota2 ,header%c_crota2 ,status)
    call freadkey_err(rtname,'CROTA2',status,header%e_crota2 ,ierr)
    call FTGKYS(unit,'WCSNAMEP',header%wcsnamep ,header%c_wcsnamep ,status)    
    call freadkey_err(rtname,'WCSNAMEP',status,header%e_wcsnamep ,ierr)
    call FTGKYS(unit,'WCSTY1P',header%wcsty1p ,header%c_wcsty1p ,status)    
    call freadkey_err(rtname,'WCSTY1P',status,header%e_wcsty1p ,ierr)
    call FTGKYS(unit,'CTYPE1P',header%ctype1p,header%c_ctype1p ,status)
    call freadkey_err(rtname,'CTYPE1P',status,header%e_ctype1p ,ierr)
    call FTGKYD(unit,'CRPIX1P',header%crpix1p,header%c_crpix1p ,status)
    call freadkey_err(rtname,'CRPIX1P',status,header%e_crpix1p ,ierr)
    call FTGKYD(unit,'CRVAL1P',header%crval1p,header%c_crval1p ,status)
    call freadkey_err(rtname,'CRVAL1P',status,header%e_crval1p ,ierr)
    call FTGKYD(unit,'CDELT1P',header%cdelt1p,header%c_cdelt1p ,status)
    call freadkey_err(rtname,'CDELT1P',status,header%e_cdelt1p ,ierr)
    call FTGKYS(unit,'WCSTY2P',header%wcsty2p ,header%c_wcsty2p ,status)
    call freadkey_err(rtname,'WCSTY2P',status,header%e_wcsty2p ,ierr)
    call FTGKYD(unit,'LTV1',header%ltv1      ,header%c_ltv1 ,status)
    call freadkey_err(rtname,'LTV1',status   ,header%e_ltv1 ,ierr)
    call FTGKYD(unit,'LTM1_1',header%ltm1_1 ,header%c_ltm1_1 ,status)
    call freadkey_err(rtname,'LTM1_1',status,header%e_ltm1_1, ierr)
    call FTGKYS(unit,'WCSTY2P',header%wcsty2p ,header%c_wcsty2p ,status)
    call freadkey_err(rtname,'WCSTY2P',status,header%e_wcsty2p ,ierr)
    call FTGKYD(unit,'LTV2',header%ltv2      ,header%c_ltv2 ,status)
    call freadkey_err(rtname,'LTV2',status   ,header%e_ltv2 ,ierr)
    call FTGKYD(unit,'LTM2_2',header%ltm2_2  ,header%c_ltm2_2 ,status)
    call freadkey_err(rtname,'LTM2_2',status ,header%e_ltm2_2 ,ierr)
    call FTGKYS(unit,'CTYPE2P',header%ctype2p,header%c_ctype2p ,status)
    call freadkey_err(rtname,'CTYPE2P',status,header%e_ctype2p ,ierr)
    call FTGKYD(unit,'CRPIX2P',header%crpix2p,header%c_crpix2p ,status)
    call freadkey_err(rtname,'CRPIX2P',status,header%e_crpix2p ,ierr)
    call FTGKYD(unit,'CRVAL2P',header%crval2p,header%c_crval2p ,status)
    call freadkey_err(rtname,'CRVAL2P',status,header%e_crval2p ,ierr)
    call FTGKYD(unit,'CDELT2P',header%cdelt2p,header%c_cdelt2p ,status)
    call freadkey_err(rtname,'CDELT2P',status,header%e_cdelt2p ,ierr)

!   If one or more keywords are missing, ierr =1:
    if (ierr > 0) ierr=1


  end subroutine fread_wcshdr


!-------------------------------------------------------------------------------


  subroutine fread_xmmhdr(unit,header,ierr)
 

    type(wcs_xmmhdr_type)         :: header 
   integer(kind=int32),intent(in) :: unit
   integer(kind=int32)            :: status,ierr 
    character(len=12)             :: rtname = 'fread_wcshdr'

    ierr=0
    status=0
    
    call FTGKYS(unit,'CTYPE1',header%ctype1 ,header%c_ctype1 ,status)
    call freadkey_err(rtname,'CTYPE1',status,header%e_ctype1 ,ierr)
    call FTGKYS(unit,'CUNIT1',header%cunit1 ,header%c_cunit1 ,status)
    call freadkey_err(rtname,'CUNIT1',status,header%e_cunit1 ,ierr)
    call FTGKYD(unit,'CRPIX1',header%crpix1 ,header%c_crpix1,status)
    call freadkey_err(rtname,'CRPIX1',status,header%e_crpix1,ierr)
    call FTGKYD(unit,'CRVAL1',header%crval1 ,header%c_crval1 ,status)
    call freadkey_err(rtname,'CRVAL1',status,header%e_crval1 ,ierr)
    call FTGKYD(unit,'CDELT1',header%cdelt1 ,header%c_cdelt1 ,status)
    call freadkey_err(rtname,'CDELT1',status,header%e_cdelt1, ierr)
    call FTGKYS(unit,'CTYPE1L',header%ctype1l,header%c_ctype1l ,status)
    call freadkey_err(rtname,'CTYPE1L',status,header%e_ctype1l ,ierr)
    call FTGKYD(unit,'CRPIX1L',header%crpix1l,header%c_crpix1l ,status)
    call freadkey_err(rtname,'CRPIX1L',status,header%e_crpix1l ,ierr)
    call FTGKYD(unit,'CRVAL1L',header%crval1l,header%c_crval1l ,status)
    call freadkey_err(rtname,'CRVAL1L',status,header%e_crval1l ,ierr)
    call FTGKYD(unit,'CDELT1L',header%cdelt1l,header%c_cdelt1l ,status)
    call freadkey_err(rtname,'CDELT1L',status,header%e_cdelt1l ,ierr)
    call FTGKYD(unit,'LTV1',header%ltv1      ,header%c_ltv1 ,status)
    call freadkey_err(rtname,'LTV1',status   ,header%e_ltv1 ,ierr)
    call FTGKYD(unit,'LTM1_1',header%ltm1_1 ,header%c_ltm1_1 ,status)
    call freadkey_err(rtname,'LTM1_1',status,header%e_ltm1_1, ierr)
    call FTGKYS(unit,'CTYPE2',header%ctype2 ,header%c_ctype2 ,status)
    call freadkey_err(rtname,'CTYPE2',status,header%e_ctype2 ,ierr)
    call FTGKYS(unit,'CUNIT2',header%cunit2 ,header%c_cunit2 ,status)
    call freadkey_err(rtname,'CUNIT2',status,header%e_cunit2 ,ierr)
    call FTGKYD(unit,'CRPIX2',header%crpix2 ,header%c_crpix2,status)
    call freadkey_err(rtname,'CRPIX2',status,header%e_crpix2,ierr)
    call FTGKYD(unit,'CRVAL2',header%crval2 ,header%c_crval2 ,status)
    call freadkey_err(rtname,'CRVAL2',status,header%e_crval2 ,ierr)
    call FTGKYD(unit,'CDELT2',header%cdelt2 ,header%c_cdelt2 ,status)
    call freadkey_err(rtname,'CDELT2',status,header%e_cdelt2 ,ierr)
    call FTGKYS(unit,'CTYPE2L',header%ctype2l,header%c_ctype2l ,status)
    call freadkey_err(rtname,'CTYPE2L',status,header%e_ctype2l ,ierr)
    call FTGKYD(unit,'CRPIX2L',header%crpix2l,header%c_crpix2l ,status)
    call freadkey_err(rtname,'CRPIX2L',status,header%e_crpix2l ,ierr)
    call FTGKYD(unit,'CRVAL2L',header%crval2l,header%c_crval2l ,status)
    call freadkey_err(rtname,'CRVAL2L',status,header%e_crval2l ,ierr)
    call FTGKYD(unit,'CDELT2L',header%cdelt2l,header%c_cdelt2l ,status)
    call freadkey_err(rtname,'CDELT2L',status,header%e_cdelt2l ,ierr)
    call FTGKYD(unit,'LTV2',header%ltv2      ,header%c_ltv2 ,status)
    call freadkey_err(rtname,'LTV2',status   ,header%e_ltv2 ,ierr)
    call FTGKYD(unit,'LTM2_2',header%ltm2_2  ,header%c_ltm2_2 ,status)
    call freadkey_err(rtname,'LTM2_2',status ,header%e_ltm2_2 ,ierr)


!   If one or more keywords are missing, ierr =1:
    if (ierr > 0) ierr=1

  end subroutine fread_xmmhdr


!-------------------------------------------------------------------------------

  subroutine fread_refhdr(unit,header,ierr)
 

    type(wcs_refhdr_type)         :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    character(len=12)             :: rtname = 'fread_refhdr'

    ierr=0
    status=0

    call FTGKYS(unit,'REFXCTYP',header%refxctyp,header%c_refxctyp,status)
    call freadkey_err(rtname,'REFXCTYP',status ,header%e_refxctyp,ierr)
    call FTGKYS(unit,'REFXCUNI',header%refxcuni,header%c_refxcuni,status)
    call freadkey_err(rtname,'REFXCUNI',status ,header%e_refxcuni,ierr)
    call FTGKYJ(unit,'REFXCRPX',header%refxcrpx,header%c_refxcrpx,status)
    call freadkey_err(rtname,'REFXCRPX',status ,header%e_refxcrpx,ierr)
    call FTGKYD(unit,'REFXCRVL',header%refxcrvl,header%c_refxcrvl,status)
    call freadkey_err(rtname,'REFXCRVL',status ,header%e_refxcrvl,ierr)
    call FTGKYD(unit,'REFXCDLT',header%refxcdlt,header%c_refxcdlt,status)
    call freadkey_err(rtname,'REFXCDLT',status, header%e_refxcdlt,ierr)
    call FTGKYD(unit,'REFXLMIN',header%refxlmin,header%c_refxlmin,status)
    call freadkey_err(rtname,'REFXLMIN',status ,header%e_refxlmin,ierr)
    call FTGKYD(unit,'REFXDMAX',header%refxdmax,header%c_refxdmax,status)
    call freadkey_err(rtname,'REFXDMAX',status ,header%e_refxdmax,ierr)
    call FTGKYS(unit,'REFYCTYP',header%refyctyp,header%c_refyctyp,status)
    call freadkey_err(rtname,'REFYCTYP',status ,header%e_refyctyp,ierr)
    call FTGKYS(unit,'REFYCUNI',header%refycuni,header%c_refycuni,status)
    call freadkey_err(rtname,'REFYCUNI',status ,header%e_refycuni,ierr)
    call FTGKYJ(unit,'REFYCRPX',header%refycrpx,header%c_refycrpx,status)
    call freadkey_err(rtname,'REFYCRPX',status ,header%e_refycrpx,ierr)
    call FTGKYD(unit,'REFYCRVL',header%refycrvl,header%c_refycrvl,status)
    call freadkey_err(rtname,'REFYCRVL',status ,header%e_refycrvl,ierr)
    call FTGKYD(unit,'REFYCDLT',header%refycdlt,header%c_refycdlt,status)
    call freadkey_err(rtname,'REFYCDLT',status ,header%e_refycdlt,ierr)
    call FTGKYD(unit,'REFYLMIN',header%refylmin,header%c_refylmin,status)
    call freadkey_err(rtname,'REFYLMIN',status ,header%e_refylmin,ierr)
    call FTGKYD(unit,'REFYDMAX',header%refydmax,header%c_refydmax,status)
    call freadkey_err(rtname,'REFYDMAX',status ,header%e_refydmax,ierr)

!   If one or more keywords are missing, ierr =1:
    if (ierr > 0) ierr=1    

  end subroutine fread_refhdr

!-------------------------------------------------------------------------------

  subroutine fread_stdhdr(unit,header,ierr)
 

    type(stdhdr_type)                 :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    integer(kind=int32)               :: i
    character(len=1)                  :: strnum
    character(len=8)                  :: keyname   
    character(len=12)                 :: rtname = 'fread_stdhdr'

    ierr=0
    status=0

    call FTGKYS(unit,'ORIGIN',header%origin ,header%c_origin,status)
    call freadkey_err(rtname,'ORIGIN',status,header%e_origin,ierr)  
    call FTGKYS(unit,'CREATOR',header%creator,header%c_creator,status)
    call freadkey_err(rtname,'CREATOR',status,header%e_creator,ierr)  
    call FTGKYS(unit,'ERCONFIG',header%config,header%c_config,status)
    call freadkey_err(rtname,'ERCONFIG',status,header%e_config,ierr)  
    call FTGKYS(unit,'ERVERS',header%version,header%c_version,status)
    call freadkey_err(rtname,'ERVERS',status,header%e_version,ierr)  
    call FTGKYS(unit,'ERPREFIX',header%prefix,header%c_prefix,status)
    call freadkey_err(rtname,'ERPREFIX',status,header%e_prefix,ierr)  
    call FTGKYS(unit,'ERDSCODE',header%dscode,header%c_dscode,status)
    call freadkey_err(rtname,'ERDSCODE',status,header%e_dscode,ierr)  
    call FTGKYS(unit,'ERDSTYPE',header%dstype,header%c_dstype,status)
    call freadkey_err(rtname,'ERDSTYPE',status,header%e_dstype,ierr)  
    call FTGKYS(unit,'OWNER',header%owner,header%c_owner,status)
    call freadkey_err(rtname,'OWNER',status,header%e_owner,ierr)  
    call FTGKYS(unit,'MISSION',header%mission,header%c_mission,status)
    call freadkey_err(rtname,'MISSION',status,header%e_mission,ierr)
    call FTGKYS(unit,'DATE',header%date      ,header%c_date,status)
    call freadkey_err(rtname,'DATE',status   ,header%e_date,ierr)   
    call FTGKYS(unit,'TELESCOP',header%telescop,header%c_telescop,status)
    call freadkey_err(rtname,'TELESCOP',status ,header%e_telescop,ierr) 
    call FTGKYS(unit,'INSTRUME',header%instrume,header%c_instrume,status)
    call freadkey_err(rtname,'INSTRUME',status ,header%e_instrume,ierr)

    call FTGKYJ(unit,'NINST',header%ninst,header%c_ninst,status) 
    call freadkey_err(rtname,'NINST',status,header%e_ninst,ierr)
    if (.not. header%e_ninst) header%ninst=0

    header%instruments=' '
    do i=1,header%ninst
      write(strnum,'(I1)') I
      keyname='INSTRUM'//strnum 
!      write(*,*) rtname, ' keyname: ',keyname
      call FTGKYS(unit,keyname,header%instrumx(i),header%c_instrumx(i),status)
!      write(*,*) rtname, 'i,  header%instrumx(i): ',i, ' ', header%instrumx(i)
      call freadkey_err(rtname,keyname,status,header%e_instrumx(i),ierr)
!      write(*,*) rtname, ' status: ',status, 'ierr: ',ierr
      if (status == 0) then
!        write(*,*) rtname, ' ' , trim(adjustl(header%instrumx(i)))
!        write(*,*) rtname, ' ' , 'x',header%instruments,'x'
        header%instruments = trim(adjustl(header%instruments))//trim(adjustl(header%instrumx(i)))
        header%e_instruments = .true.
!       write(*,*) rtname, ' ' , header%instruments
      endif
    end do
    if (header%ninst == 0 .and.  header%e_instrume) then 
      header%instruments = header%instrume
      header%e_instruments = .true.
    endif
  
    call FTGKYS(unit,'OBS_MODE',header%obs_mode,header%c_obs_mode,status)
    call freadkey_err(rtname,'OBS_MODE',status,header%e_obs_mode,ierr) 
    call FTGKYS(unit,'DATAMODE',header%datamode,header%c_datamode,status)
    call freadkey_err(rtname,'DATAMODE',status ,header%e_datamode,ierr) 
    call FTGKYD(unit,'FRAMETIM',header%frametim,header%c_frametim,status)
    call freadkey_err(rtname,'FRAMETIM',status ,header%e_frametim,ierr) 
    call FTGKYS(unit,'FILTER',header%filter    ,header%c_filter,status)
    call freadkey_err(rtname,'FILTER',status   ,header%e_filter,ierr) 
    call FTGKYS(unit,'SKYFIELD',header%skyfield,header%c_skyfield,status)
    call freadkey_err(rtname,'SKYFIELD',status   ,header%e_skyfield,ierr) 
    call FTGKYS(unit,'OBS_ID',header%obs_id    ,header%c_obs_id,status)
    call freadkey_err(rtname,'OBS_ID',status   ,header%e_obs_id,ierr) 
    call FTGKYS(unit,'EXP_ID',header%exp_id    ,header%c_exp_id,status)
    call freadkey_err(rtname,'EXP_ID',status   ,header%e_exp_id,ierr) 
    call FTGKYS(unit,'OBSERVER',header%observer,header%c_observer,status)
    call freadkey_err(rtname,'OBSERVER',status ,header%e_observer,ierr) 
    call FTGKYS(unit,'OBJECT',header%object    ,header%c_object,status)
    call freadkey_err(rtname,'OBJECT',status   ,header%e_object,ierr) 
    call FTGKYD(unit,'RA_OBJ',header%ra_obj    ,header%c_ra_obj,status)
    call freadkey_err(rtname,'RA_OBJ',status   ,header%e_ra_obj,ierr) 
    call FTGKYD(unit,'DEC_OBJ',header%dec_obj  ,header%c_dec_obj,status)
    call freadkey_err(rtname,'DEC_OBJ',status  ,header%e_dec_obj,ierr) 
    call FTGKYD(unit,'RA_CEN',header%ra_cen    ,header%c_ra_cen,status)
    call freadkey_err(rtname,'RA_CEN',status   ,header%e_ra_cen,ierr) 
    call FTGKYD(unit,'DEC_CEN',header%dec_cen,header%c_dec_cen,status)
    call freadkey_err(rtname,'DEC_CEN',status  ,header%e_dec_cen,ierr) 
    call FTGKYS(unit,'DATE-OBS',header%date_obs,header%c_date_obs,status)
    call freadkey_err(rtname,'DATE-OBS',status ,header%e_date_obs,ierr) 
    call FTGKYS(unit,'DATE-END',header%date_end,header%c_date_end,status)
    call freadkey_err(rtname,'DATE-END',status ,header%e_date_end,ierr) 
    call FTGKYD(unit,'TSTART',header%tstart    ,header%c_tstart,status)
    call freadkey_err(rtname,'TSTART',status   ,header%e_tstart,ierr) 
    call FTGKYD(unit,'TSTOP',header%tstop      ,header%c_tstop,status)
    call freadkey_err(rtname,'TSTOP',status    ,header%e_tstop,ierr) 
    call FTGKYD(unit,'MJDREF',header%mjdref    ,header%c_mjdref,status)
    call freadkey_err(rtname,'MJDREF',status   ,header%e_mjdref,ierr) 
    call FTGKYD(unit,'TIMEZERO',header%timezero,header%c_timezero,status)
    call freadkey_err(rtname,'TIMEZERO',status ,header%e_timezero,ierr) 
    call FTGKYS(unit,'TIMEUNIT',header%timeunit,header%c_timeunit,status)
    call freadkey_err(rtname,'TIMEUNIT',status ,header%e_timeunit,ierr) 
    call FTGKYS(unit,'TIMESYS',header%timesys  ,header%c_timesys,status)
    call freadkey_err(rtname,'TIMESYS',status  ,header%e_timesys,ierr) 
    call FTGKYD(unit,'RA_PNT',header%ra_pnt    ,header%c_ra_pnt,status)
    call freadkey_err(rtname,'RA_PNT',status   ,header%e_ra_pnt,ierr) 
    call FTGKYD(unit,'DEC_PNT',header%dec_pnt  ,header%c_dec_pnt,status)
    call freadkey_err(rtname,'DEC_PNT',status  ,header%e_dec_pnt,ierr) 
    call FTGKYD(unit,'PA_PNT',header%pa_pnt    ,header%c_pa_pnt,status)
    call freadkey_err(rtname,'PA_PNT',status   ,header%e_pa_pnt,ierr) 
    call FTGKYS(unit,'RADECSYS',header%radecsys,header%c_radecsys,status)
    call freadkey_err(rtname,'RADECSYS',status ,header%e_radecsys,ierr) 
    call FTGKYD(unit,'EQUINOX',header%equinox  ,header%c_equinox,status)
    call freadkey_err(rtname,'EQUINOX',status  ,header%e_equinox,ierr) 
    call FTGKYJ(unit,'FLAGSEL',header%flagsel,header%c_flagsel,status) 
    call freadkey_err(rtname,'FLAGSEL',status,header%e_flagsel,ierr)
    call FTGKYJ(unit,'PAT_SEL',header%pat_sel,header%c_pat_sel,status) 
    call freadkey_err(rtname,'PAT_SEL',status,header%e_pat_sel,ierr)
    call FTGKYS(unit,'GTI_SEL',header%gti_sel,header%c_gti_sel,status) 
    call freadkey_err(rtname,'GTI_SEL',status,header%e_gti_sel,ierr)
    call FTGKYJ(unit,'NENERGY',header%nenergy,header%c_nenergy,status) 
    call freadkey_err(rtname,'NENERGY',status,header%e_nenergy,ierr)
!    write(*,*) 'NENERGY: ',header%nenergy,' ',header%e_nenergy
    if (header%e_nenergy) then
      do i=1,header%nenergy
        write(strnum,'(I1)') I
        keyname='EMIN'//strnum 
        call FTGKYD(unit,keyname,header%eminx(i),header%c_eminx(i),status) 
        call freadkey_err(rtname,keyname,status,header%e_eminx(i),ierr)
      end do
      do i=1,header%nenergy
        write(strnum,'(I1)') I
        keyname='EMAX'//strnum 
        call FTGKYD(unit,keyname,header%emaxx(i),header%c_emaxx(i),status) 
        call freadkey_err(rtname,keyname,status,header%e_emaxx(i),ierr)
      end do
    else
      header%nenergy=0
    endif

    call FTGKYS(unit,'TRANSMOD',header%transmod,header%c_transmod,status) 
    call freadkey_err(rtname,'TRANSMOD',status,header%e_transmod,ierr)      
    call FTGKYJ(unit,'PAT_THR',header%pat_thr,header%c_pat_thr,status) 
    call freadkey_err(rtname,'PAT_THR',status,header%e_pat_thr,ierr) 
    call FTGKYS(unit,'T_OFFCOR',header%t_offcor,header%c_t_offcor,status) 
    call freadkey_err(rtname,'T_OFFCOR',status,header%e_t_offcor,ierr)   
    call FTGKYS(unit,'HDUCLASS',header%hduclass  ,header%c_hduclass,status)
    call freadkey_err(rtname,'HDUCLASS',status ,header%e_hduclass,ierr) 
    call FTGKYS(unit,'HDUVERS',header%hduvers  ,header%c_hduvers,status)
    call freadkey_err(rtname,'HDUVERS',status ,header%e_hduvers,ierr) 
    call FTGKYS(unit,'HDUCLAS1',header%hduclas1  ,header%c_hduclas1,status)
    call freadkey_err(rtname,'HDUCLAS1',status ,header%e_hduclas1,ierr) 
    call FTGKYS(unit,'HDUCLAS2',header%hduclas2  ,header%c_hduclas2,status)
    call freadkey_err(rtname,'HDUCLAS2',status ,header%e_hduclas2,ierr) 
!    call FTGKYS(unit,'HISTORY',header%history  ,header%c_history,status)
    call FTGKLS(unit,'SASSHIST',header%history  ,header%c_history,status)
!    write(*,*) rtname,' status ', status
!    write(*,*) rtname,' hist: ', trim(header%history)
    call freadkey_err(rtname,'SASSHIST',status ,header%e_history,ierr)
    call FTGKYS(unit,'LONGSTRN',header%longstr  ,header%c_longstr,status)
    call freadkey_err(rtname,'LONGSTRN',status  ,header%e_longstr,ierr) 

!   If one or more keywords are missing, ierr =1:
    if (ierr > 0) ierr=1

  end subroutine fread_stdhdr



!-------------------------------------------------------------------------------

  subroutine fread_dethdr(unit,header,ierr)
  
  !--------------------------------------------------
  !
  ! Read header keywords specific to various eROSITA 
  ! detection tasks
  !
  !--------------------------------------------------

    type(dethdr_type)                 :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    character(len=12)                 :: rtname = 'fread_dethdr'

    ierr=0
    status=0
    
    call FTGKYE(unit,'CHISQ',header%chisq ,header%c_chisq,status)
    call freadkey_err(rtname,'CHISQ',status,header%e_chisq,ierr) 
    call FTGKYJ(unit,'XDATAMIN',header%xdatamin ,header%c_xdatamin,status)
    call freadkey_err(rtname,'XDATAMIN',status,header%e_xdatamin,ierr)
    call FTGKYJ(unit,'XDATAMAX',header%xdatamax ,header%c_xdatamax,status)
    call freadkey_err(rtname,'XDATAMAX',status,header%e_xdatamax,ierr)
    call FTGKYJ(unit,'YDATAMIN',header%ydatamin ,header%c_ydatamin,status)
    call freadkey_err(rtname,'YDATAMIN',status,header%e_ydatamin,ierr)
    call FTGKYJ(unit,'YDATAMAX',header%ydatamax ,header%c_ydatamax,status)
    call freadkey_err(rtname,'YDATAMAX',status,header%e_ydatamax,ierr)
     
 
  end subroutine fread_dethdr



!-------------------------------------------------------------------------------


  subroutine fwrite_dethdr(unit,header,ierr)

  !-----------------------------------------------------------------------------
  ! 
  ! Write   header keywords  specific to  eROSITA detection tasks
  !
  !-----------------------------------------------------------------------------

    type(dethdr_type)                 :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    character(len=*),parameter        :: rtname = 'FWRITE_DETHDR'
   
    status=0
    ierr = 0
    
    if (header%e_chisq) call FTUKYE(unit,'CHISQ',header%chisq ,decimals, header%c_chisq,status)
    if (header%e_xdatamin) call FTUKYJ(unit,'XDATAMIN',header%xdatamin ,header%c_xdatamin,status)
    if (header%e_xdatamax) call FTUKYJ(unit,'XDATAMAX',header%xdatamax ,header%c_xdatamax,status)
    if (header%e_ydatamin) call FTUKYJ(unit,'YDATAMIN',header%ydatamin ,header%c_ydatamin,status)
    if (header%e_ydatamax) call FTUKYJ(unit,'YDATAMAX',header%ydatamax ,header%c_ydatamax,status)



    if (status > 0) then
      ierr = 2 
    endif


  end subroutine fwrite_dethdr

!-------------------------------------------------------------------------------

  subroutine fwrite_evsclhdr(unit,header,ierr)

  !-------------------------------------------------------
  ! write keywords TZEROn TSCALn for RA, DEC, SUBX, SUBY columns
  ! of event list.
  ! RA, DEC, SUBX, SUBY columns need to exist already  
  !------------------------------------------------------

    type(evscalehdr_type)             :: header
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    character(len=*),parameter        :: rtname = 'FWRITE_EVSCLHDR'  
    integer(kind=int32)               :: colnum
    character(len=3)                  :: c_colnum
    character(len=8)                  :: keyname    

    ierr=0

    call FTGCNO(unit,.false.,'ra',colnum,status)
    write(c_colnum,'(I3)') colnum
    keyname='TZERO'//trim(adjustl(c_colnum)) 
    call FTUKYD(unit,keyname ,header%tzero_ra ,decimals, header%c_tzero_ra,status)
    keyname='TSCAL'//trim(adjustl(c_colnum)) 
    call FTUKYD(unit,keyname ,header%tscal_ra ,decimals, header%c_tscal_ra,status)

    call FTGCNO(unit,.false.,'dec',colnum,status)
    write(c_colnum,'(I3)') colnum
    keyname='TZERO'//trim(adjustl(c_colnum))
    call FTUKYD(unit,keyname ,header%tzero_dec ,decimals, header%c_tzero_dec,status)
    keyname='TSCAL'//trim(adjustl(c_colnum))
    call FTUKYD(unit,keyname ,header%tscal_dec ,decimals, header%c_tscal_dec,status)

    call FTGCNO(unit,.false.,'subx',colnum,status)
    write(c_colnum,'(I3)') colnum
    keyname='TZERO'//trim(adjustl(c_colnum)) 
    call FTUKYD(unit,keyname ,header%tzero_subx ,decimals, header%c_tzero_subx,status)
    keyname='TSCAL'//trim(adjustl(c_colnum)) 
    call FTUKYD(unit,keyname ,header%tscal_subx ,decimals, header%c_tscal_subx,status)


    call FTGCNO(unit,.false.,'suby',colnum,status)
    write(c_colnum,'(I3)') colnum
    keyname='TZERO'//trim(adjustl(c_colnum)) 
    call FTUKYD(unit,keyname ,header%tzero_suby ,decimals, header%c_tzero_suby,status)
    keyname='TSCAL'//trim(adjustl(c_colnum)) 
    call FTUKYD(unit,keyname ,header%tscal_suby ,decimals, header%c_tscal_suby,status)



  end subroutine fwrite_evsclhdr


!-------------------------------------------------------------------------------

  subroutine fwrite_wcshdr(unit,header,ierr)
 
  !-----------------------------------------------------------------------------
  ! 
  ! Write  WCS header as created by ftools (e.g. extractor v 4.36)
  !
  !-----------------------------------------------------------------------------
 

    type(wcs_imahdr_type)           :: header 
    integer(kind=int32),intent(in)  :: unit
    integer(kind=int32)             :: status,ierr 
    character(len=*),parameter      :: rtname = 'fwrite_wcshdr'

    ierr=0
    status=0


    call FTUKYS(unit,'CTYPE1',header%ctype1   ,          header%c_ctype1  ,status)
    call FTUKYD(unit,'CRPIX1',header%crpix1   ,decimals, header%c_crpix1  ,status)
    call FTUKYD(unit,'DRPIX1',header%drpix1   ,decimals, header%c_drpix1  ,status)
    call FTUKYD(unit,'CRVAL1',header%crval1   ,decimals, header%c_crval1  ,status)
    call FTUKYD(unit,'CDELT1',header%cdelt1   ,decimals, header%c_cdelt1  ,status)
    call FTUKYS(unit,'CUNIT1',header%cunit1   ,          header%c_cunit1  ,status)
    call FTUKYD(unit,'DDELT1',header%ddelt1   ,decimals, header%c_ddelt1  ,status)
    call FTUKYS(unit,'CTYPE2',header%ctype2   ,          header%c_ctype2  ,status)
    call FTUKYD(unit,'CRPIX2',header%crpix2   ,decimals, header%c_crpix2  ,status)
    call FTUKYD(unit,'DRPIX2',header%drpix2   ,decimals, header%c_drpix2  ,status)
    call FTUKYD(unit,'CRVAL2',header%crval2   ,decimals, header%c_crval2  ,status)
    call FTUKYD(unit,'CDELT2',header%cdelt2   ,decimals, header%c_cdelt2  ,status)
    call FTUKYS(unit,'CUNIT2',header%cunit2   ,          header%c_cunit2  ,status)
    call FTUKYD(unit,'DDELT2',header%ddelt2   ,decimals, header%c_ddelt2  ,status)
    call FTUKYD(unit,'CROTA2',header%crota2   ,decimals, header%c_crota2  ,status)
    call FTUKYS(unit,'WCSNAMEP',header%wcsnamep ,          header%c_wcsnamep,status)    
    call FTUKYS(unit,'WCSTY1P',header%wcsty1p ,          header%c_wcsty1p ,status)
    call FTUKYS(unit,'CTYPE1P',header%ctype1p ,          header%c_ctype1p ,status)
    call FTUKYD(unit,'CRPIX1P',header%crpix1p ,decimals, header%c_crpix1p ,status)
    call FTUKYD(unit,'CRVAL1P',header%crval1p ,decimals, header%c_crval1p ,status)
    call FTUKYD(unit,'CDELT1P',header%cdelt1p ,decimals, header%c_cdelt1p ,status)
    call FTUKYS(unit,'WCSTY2P',header%wcsty2p ,          header%c_wcsty2p ,status)
    call FTUKYD(unit,'LTV1'   ,header%ltv1    ,decimals, header%c_ltv1    ,status)
    call FTUKYD(unit,'LTM1_1' ,header%ltm1_1  ,decimals, header%c_ltm1_1  ,status)
    call FTUKYS(unit,'WCSTY2P',header%wcsty2p ,          header%c_wcsty2p ,status)
    call FTUKYD(unit,'LTV2'   ,header%ltv2    ,decimals, header%c_ltv2    ,status)
    call FTUKYD(unit,'LTM2_2' ,header%ltm2_2  ,decimals, header%c_ltm2_2  ,status)
    call FTUKYS(unit,'CTYPE2P',header%ctype2p ,          header%c_ctype2p ,status)
    call FTUKYD(unit,'CRPIX2P',header%crpix2p ,decimals, header%c_crpix2p ,status)
    call FTUKYD(unit,'CRVAL2P',header%crval2p ,decimals, header%c_crval2p ,status)
    call FTUKYD(unit,'CDELT2P',header%cdelt2p ,decimals, header%c_cdelt2p ,status)

    if (status > 0) ierr=3


  end subroutine fwrite_wcshdr



!-------------------------------------------------------------------------------

  subroutine fwrite_refhdr(unit,header,ierr)
 

    type(wcs_refhdr_type)         :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    character(len=13)             :: rtname = 'fwrite_refhdr'

    ierr=0
    status=0

    call FTUKYS(unit,'REFXCTYP',header%refxctyp,         header%c_refxctyp,status)
    call FTUKYS(unit,'REFXCUNI',header%refxcuni,         header%c_refxcuni,status)
    call FTUKYJ(unit,'REFXCRPX',header%refxcrpx,         header%c_refxcrpx,status)
    call FTUKYD(unit,'REFXCRVL',header%refxcrvl,decimals,header%c_refxcrvl,status)
    call FTUKYD(unit,'REFXCDLT',header%refxcdlt,decimals,header%c_refxcdlt,status)
    call FTUKYD(unit,'REFXLMIN',header%refxlmin,decimals,header%c_refxlmin,status)
    call FTUKYD(unit,'REFXDMAX',header%refxdmax,decimals,header%c_refxdmax,status)
    call FTUKYS(unit,'REFYCTYP',header%refyctyp,         header%c_refyctyp,status)
    call FTUKYS(unit,'REFYCUNI',header%refycuni,         header%c_refycuni,status)
    call FTUKYJ(unit,'REFYCRPX',header%refycrpx,         header%c_refycrpx,status)
    call FTUKYD(unit,'REFYCRVL',header%refycrvl,decimals,header%c_refycrvl,status)
    call FTUKYD(unit,'REFYCDLT',header%refycdlt,decimals,header%c_refycdlt,status)
    call FTUKYD(unit,'REFYLMIN',header%refylmin,decimals,header%c_refylmin,status)
    call FTUKYD(unit,'REFYDMAX',header%refydmax,decimals,header%c_refydmax,status)


    if (status > 0) then
       ierr=3
       call wrerr(rtname,ierr,"Error writing REFHDR keywords")
    endif
    
  end subroutine fwrite_refhdr

!-------------------------------------------------------------------------------

  subroutine fread_twcshdr(unit,header,ierr)

    integer(kind=int32),intent(in)    :: unit 
    type(twcs_hdr_type)               :: header
    integer(kind=int32)               :: ierr 

    integer(kind=int32)               :: status ,colnum
    character(len=2)                  :: colnum_str
    character(len=8)                  :: keyname
    character(len=*),parameter        :: rtname = 'FREAD_TWCSHDR'
    character(len=80)                 :: errstr

    ierr=0
    status=0
    
    ! axis 1:   
    call FTGCNO(unit,.false.,trim(header%colname_x),colnum,status)
    if (status > 0) then
      ierr=3
      errstr = 'Column '//trim(header%colname_x)//' does not exist'
      call wrerr(rtname,ierr,errstr)
    else if (colnum > 99) then 
        ierr=3
        errstr = 'Column number for'//trim(header%colname_x)//' out of range [1-99]'
    else  
      write(colnum_str,'(I2)') colnum
      keyname='TCTYP'//trim(adjustl(colnum_str))
      call FTGKYS(unit,keyname,header%tctyp_x,header%c_tctyp_x,status)
      keyname='TCUNI'//trim(adjustl(colnum_str))
      call FTGKYS(unit,keyname,header%tcuni_x,header%c_tcuni_x,status)
      keyname='TCRPX'//trim(adjustl(colnum_str))
      call FTGKYJ(unit,keyname,header%tcrpx_x,header%c_tcrpx_x,status)
      keyname='TCRVL'//trim(adjustl(colnum_str))
      call FTGKYD(unit,keyname,header%tcrvl_x,header%c_tcrvl_x,status)
      keyname='TCDLT'//trim(adjustl(colnum_str))
      call FTGKYD(unit,keyname,header%tcdlt_x,header%c_tcdlt_x,status)
    endif

    status=0

    ! axis 2:       
    call FTGCNO(unit,.false.,trim(header%colname_y),colnum,status)
    if (status > 0) then
      ierr=3
      errstr = 'Column '//trim(header%colname_y)//' does not exist'
      call wrerr(rtname,ierr,errstr)
    else if (colnum > 99) then  
      ierr=3
      errstr = 'Column number for'//trim(header%colname_y)//' out of range [1-99]'
    else
           write(colnum_str,'(I2)') colnum
      keyname='TCTYP'//trim(adjustl(colnum_str))
      call FTGKYS(unit,keyname,header%tctyp_y,header%c_tctyp_y,status)
      keyname='TCUNI'//trim(adjustl(colnum_str))
      call FTGKYS(unit,keyname,header%tcuni_y,header%c_tcuni_y,status)
      keyname='TCRPX'//trim(adjustl(colnum_str))
      call FTGKYJ(unit,keyname,header%tcrpx_y,header%c_tcrpx_y,status)
      keyname='TCRVL'//trim(adjustl(colnum_str))
      call FTGKYD(unit,keyname,header%tcrvl_y,header%c_tcrvl_y,status)
      keyname='TCDLT'//trim(adjustl(colnum_str))
      call FTGKYD(unit,keyname,header%tcdlt_y,header%c_tcdlt_y,status)
    endif



  end subroutine fread_twcshdr

!-------------------------------------------------------------------------------

  subroutine fwrite_twcshdr(unit,header,ierr)

    !----------------------------------------------------------------
    ! Write WCS keywords for event tables. 
    ! The routine writes the keywords for a pair of
    ! pixel coordinates. 
    ! Keywords: TCTYPn, TCTYPm, TCUNIn, TCUNIm, TCRPXn, TCRPXm
    !           TCRVLn, TCRVLm,  TCDLTn,  TCDLTm
    ! n=column number coordinate 1, m=column number coordinate 2   
    !-----------------------------------------------------------------
 
    integer(kind=int32),intent(in)    :: unit 
    type(twcs_hdr_type)               :: header
    integer(kind=int32)               :: ierr 

    integer(kind=int32)               :: status ,colnum
    character(len=2)                  :: colnum_str
    character(len=8)                  :: keyname
    character(len=*),parameter        :: rtname = 'FWRITE_TWCSHDR'
    character(len=80)                 :: errstr

    ierr=0
    status=0
    
    ! axis 1:   
    call FTGCNO(unit,.false.,trim(header%colname_x),colnum,status)
    if (status > 0) then
      ierr=3
      errstr = 'Column '//trim(header%colname_x)//' does not exist'
      call wrerr(rtname,ierr,errstr)
    else if (colnum > 99) then 
        ierr=3
        errstr = 'Column number for'//trim(header%colname_x)//' out of range [1-99]'
    else  
      write(colnum_str,'(I2)') colnum
       keyname='TCTYP'//trim(adjustl(colnum_str))
      call FTUKYS(unit,keyname,header%tctyp_x,header%c_tctyp_x,status)
      keyname='TCUNI'//trim(adjustl(colnum_str))
      call FTUKYS(unit,keyname,header%tcuni_x,header%c_tcuni_x,status)
      keyname='TCRPX'//trim(adjustl(colnum_str))
      call FTUKYJ(unit,keyname,header%tcrpx_x,header%c_tcrpx_x,status)
      keyname='TCRVL'//trim(adjustl(colnum_str))
      call FTUKYD(unit,keyname,header%tcrvl_x,decimals,header%c_tcrvl_x,status)
      keyname='TCDLT'//trim(adjustl(colnum_str))
      call FTUKYD(unit,keyname,header%tcdlt_x,decimals,header%c_tcdlt_x,status)
    endif

    status=0

    ! axis 2:       
    call FTGCNO(unit,.false.,trim(header%colname_y),colnum,status)
    if (status > 0) then
      ierr=3
      errstr = 'Column '//trim(header%colname_y)//' does not exist'
      call wrerr(rtname,ierr,errstr)
    else if (colnum > 99) then  
      ierr=3
      errstr = 'Column number for'//trim(header%colname_y)//' out of range [1-99]'
    else
      write(colnum_str,'(I2)') colnum
      keyname='TCTYP'//trim(adjustl(colnum_str))
      call FTUKYS(unit,keyname,header%tctyp_y,header%c_tctyp_y,status)
      keyname='TCUNI'//trim(adjustl(colnum_str))
      call FTUKYS(unit,keyname,header%tcuni_y,header%c_tcuni_y,status)
      keyname='TCRPX'//trim(adjustl(colnum_str))
      call FTUKYJ(unit,keyname,header%tcrpx_y,header%c_tcrpx_y,status)
      keyname='TCRVL'//trim(adjustl(colnum_str))
      call FTUKYD(unit,keyname,header%tcrvl_y,decimals,header%c_tcrvl_y,status)
      keyname='TCDLT'//trim(adjustl(colnum_str))
      call FTUKYD(unit,keyname,header%tcdlt_y,decimals,header%c_tcdlt_y,status)
    endif



  end subroutine fwrite_twcshdr



!-------------------------------------------------------------------------------

  subroutine fwrite_stdhdr(unit,header,ierr)
 

    type(stdhdr_type)             :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    integer(kind=int32)               :: i
    character(len=1)                  :: strnum
    character(len=8)                  :: keyname  
!    character(len=64) :: version    
    character(len=80)                 :: taskname
    character(len=65536)              :: old_history = " "
    character(len=*), parameter       :: rtname = 'fwrite_stdhdr'


    ierr=0
    status=0

    
    call  get_command_argument(0, taskname)

    header%creator=to_upper(taskname)

    old_history = header%history 
 !   if (len_trim(header%history)>0) then
      header%history = trim(header%history) // ' **** ' // trim(sass_history)
!    else
!      header%history = trim(sass_history)
!    end if 



    call FTUKYS(unit,'ORIGIN',header%origin               ,header%c_origin,status)
    call FTUKYS(unit,'CREATOR',header%creator             ,header%c_creator,status)
    call FTUKYS(unit,'ERCONFIG',header%config               ,header%c_config,status)
    call FTUKYS(unit,'ERVERS',header%version             ,header%c_version,status)
    call FTUKYS(unit,'ERPREFIX',header%prefix               ,header%c_prefix,status)
    call FTUKYS(unit,'ERDSCODE',header%dscode               ,header%c_dscode,status)
    call FTUKYS(unit,'ERDSTYPE',header%dstype               ,header%c_dstype,status)
    call FTUKYS(unit,'OWNER',header%owner                 ,header%c_owner,status)
    call FTUKYS(unit,'MISSION',header%mission             ,header%c_mission,status)
    call FTUKYS(unit,'DATE',header%date                   ,header%c_date,status)
    call FTPDAT(unit,status)
    call FTUKYS(unit,'TELESCOP',header%telescop           ,header%c_telescop,status)
    call FTUKYS(unit,'INSTRUME',header%instrume           ,header%c_instrume,status)
    call FTUKYJ(unit,'NINST',header%ninst                 ,header%c_ninst,status)
    
    do i=1,header%ninst
      write(strnum,'(I1)') I
      keyname='INSTRUM'//strnum 
      call FTUKYS(unit,keyname,header%instrumx(i),header%c_instrumx(i),status)
      call freadkey_err(rtname,keyname,status,header%e_instrumx(i),ierr) 
    end do
    

    call FTUKYS(unit,'OBS_MODE',header%obs_mode            ,header%c_obs_mode,status)
    call FTUKYS(unit,'DATAMODE',header%datamode            ,header%c_datamode,status)
    call FTUKYD(unit,'FRAMETIM',header%frametim, decimals  ,header%c_frametim,status)
    call FTUKYS(unit,'FILTER',header%filter                ,header%c_filter,status)
    call FTUKYS(unit,'SKYFIELD',header%skyfield            ,header%c_skyfield,status)
    call FTUKYS(unit,'OBS_ID',header%obs_id                ,header%c_obs_id,status)
    call FTUKYS(unit,'EXP_ID',header%exp_id                ,header%c_exp_id,status)
    call FTUKYS(unit,'OBSERVER',header%observer            ,header%c_observer,status)
    call FTUKYS(unit,'OBJECT',header%object                ,header%c_object,status)
    call FTUKYD(unit,'RA_OBJ',header%ra_obj    ,decimals   ,header%c_ra_obj,status)
    call FTUKYD(unit,'DEC_OBJ',header%dec_obj  ,decimals   ,header%c_dec_obj,status)
    call FTUKYD(unit,'RA_CEN',header%ra_cen    ,decimals   ,header%c_ra_cen,status)
    call FTUKYD(unit,'DEC_CEN',header%dec_cen  ,decimals   ,header%c_dec_cen,status)
    call FTUKYS(unit,'DATE-OBS',header%date_obs            ,header%c_date_obs,status)
    call FTUKYS(unit,'DATE-END',header%date_end            ,header%c_date_end,status)
    call FTUKYD(unit,'TSTART',header%tstart    ,decimals   ,header%c_tstart,status)
    call FTUKYD(unit,'TSTOP',header%tstop      ,decimals ,header%c_tstop,status)
    call FTUKYD(unit,'MJDREF',header%mjdref    ,decimals   ,header%c_mjdref,status)
    call FTUKYD(unit,'TIMEZERO',header%timezero,decimals   ,header%c_timezero,status)
    call FTUKYS(unit,'TIMEUNIT',header%timeunit            ,header%c_timeunit,status)
    call FTUKYS(unit,'TIMESYS',header%timesys              ,header%c_timesys,status)
    call FTUKYD(unit,'RA_PNT',header%ra_pnt    ,decimals   ,header%c_ra_pnt,status)
    call FTUKYD(unit,'DEC_PNT',header%dec_pnt  ,decimals   ,header%c_dec_pnt,status)
    call FTUKYD(unit,'PA_PNT',header%pa_pnt    ,decimals   ,header%c_pa_pnt,status)
    call FTUKYS(unit,'RADECSYS',header%radecsys            ,header%c_radecsys,status)
    call FTUKYD(unit,'EQUINOX',header%equinox  ,decimals   ,header%c_equinox,status)
    call FTUKYJ(unit,'FLAGSEL',header%flagsel              ,header%c_flagsel,status)
    call FTUKYJ(unit,'PAT_SEL',header%pat_sel              ,header%c_pat_sel,status)
    call FTUKYS(unit,'GTI_SEL',header%gti_sel              ,header%c_gti_sel,status)
    call FTUKYJ(unit,'NENERGY',header%nenergy              ,header%c_nenergy,status)
    do i=1,header%nenergy
      write(strnum,'(I1)') I
      keyname='EMIN'//strnum 
      call FTUKYD(unit,keyname,header%eminx(i),decimals    ,header%c_eminx(i),status) 
    end do
    do i=1,header%nenergy
      write(strnum,'(I1)') I
      keyname='EMAX'//strnum 
      call FTUKYD(unit,keyname,header%emaxx(i),decimals   ,header%c_emaxx(i),status) 
    end do

    call FTUKYS(unit,'TRANSMOD',header%transmod            ,header%c_transmod,status)
    call FTUKYJ(unit,'PAT_THR',header%pat_thr             ,header%c_pat_thr,status)
    call FTUKYS(unit,'HDUCLASS',header%hduclass            ,header%c_hduclass,status)
    call FTUKYS(unit,'HDUVERS',header%hduvers              ,header%c_hduvers,status)
    call FTUKYS(unit,'HDUCLAS1',header%hduclas1            ,header%c_hduclas1,status)
    call FTUKYS(unit,'HDUCLAS2',header%hduclas2            ,header%c_hduclas2,status)
    call FTPKLS(unit,'SASSHIST',header%history              ,header%c_history,status)
!    call FTPHIS(unit,header%history ,status)
    call FTUKYS(unit,'LONGSTRN',header%longstr              ,header%c_longstr,status)

    header%history = old_history 

    if (status > 0) ierr=3    

  end subroutine fwrite_stdhdr

!-------------------------------------------------------------------------------

  subroutine fget_stdhdr(filename,header,ierr,extname)

    character(len=*)             :: filename
    type(stdhdr_type)            :: header
    integer(kind=int32)          :: ierr  
    character(len=*),optional    :: extname
    integer(kind=int32)          :: unit,status,rwmode,blocksize
    character(len=12)            :: rtname = 'fget_stdhdr'

    status=0
    rwmode=0
    ierr=0   

    call ftgiou(unit,status)
    call ftopen(unit,filename,rwmode,blocksize,status)

    if (status > 0) then
       ierr=3
      ! (26/05/2023): add STOP statement (temporay fix for DR1 user release): 
      call wrerr(rtname,ierr,'Error opening FITS file '//trim(filename))
      stop
    else
      if (present(extname)) then
        call FTMNHD(unit, any_hdu, extname, 0, status)
        if (status == 0) then
           call fread_stdhdr(unit,header,status)
           ierr = status
        endif 
      else
        call fread_stdhdr(unit,header,status)
        ierr = status
      endif
    endif

    call ftclos(unit,status)  

  end subroutine fget_stdhdr
       
!-------------------------------------------------------------------------------


  subroutine fget_dethdr(filename,header,ierr,extname)


    character(len=*)             :: filename
    type(dethdr_type)            :: header
    integer(kind=int32)          :: ierr  
    character(len=*),optional    :: extname
    integer(kind=int32)          :: unit,status,rwmode,blocksize
    character(len=12)   :: rtname = 'fget_dethdr'

    status=0
    rwmode=0
    ierr=0   

    call ftgiou(unit,status)
    call ftopen(unit,filename,rwmode,blocksize,status)

   if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening FITS file')
    else
      if (present(extname)) then
        call FTMNHD(unit, any_hdu, extname, 0, status)
        if (status == 0) then
           call fread_dethdr(unit,header,status)
           ierr = status
        endif 
      else
        call fread_dethdr(unit,header,status)
        ierr = status
      endif
    endif


    call ftclos(unit,status)  
    


  end subroutine fget_dethdr
       


!-------------------------------------------------------------------------------

  subroutine fget_wcshdr(filename,header,ierr,extname)

    character(len=*)          :: filename
    type(wcs_imahdr_type)     :: header
    integer(kind=int32)       :: ierr
    character(len=*),optional :: extname
    integer(kind=int32)       :: unit,status,rwmode,blocksize
    character(len=12)         :: rtname = 'fget_wcshdr'

    status=0
    rwmode=0
    ierr=0   

    call ftgiou(unit,status)
    call ftopen(unit,filename,rwmode,blocksize,status)


   if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening FITS file')
    else
      if (present(extname)) then
        call FTMNHD(unit, any_hdu, extname, 0, status)
        if (status == 0) then
           call fread_wcshdr(unit,header,status)
           ierr = status
        endif 
      else
        call fread_wcshdr(unit,header,status)
        ierr = status
      endif
    endif


    call ftclos(unit,status)  
    

    

  end subroutine fget_wcshdr
       
!-------------------------------------------------------------------------------

       
!-------------------------------------------------------------------------------



  subroutine update_stdhdr (filename, header, ierr)

  
    character(len=*), intent(in) :: filename
    type(stdhdr_type), intent(inout) :: header
    integer, intent(out) :: ierr

    integer :: i
    integer, dimension(6) :: is = (/ 5, 12, 16, 0, 0, 0 /)
    logical :: isprodname = .true.

    ierr = 0

    is(4) = index(filename(is(3)+1:), "_")
    if (is(4).lt.6) isprodname = .false. 
    is(4) = is(4)+is(3)
    is(5) = is(4)+4
    is(6) = is(5)+5

    do i=1,5
      if (filename(is(i):is(i)) .ne. "_") isprodname = .false.
    end do
    if (filename(is(6):is(6)+4) .ne. ".fits") isprodname = .false.
    if (len_trim(filename) .ne. is(6)+4) isprodname = .false.
  
    if (isprodname) then
      header%prefix  = filename(1:4)
      header%dscode  = filename(is(2)+1:is(3)-1) 
      header%dstype  = filename(is(3)+1:is(4)-1)
      header%version = filename(is(4)+1:is(5)-1)
      header%config  = filename(is(5)+1:is(6)-1)
      call wrchatty("update_stdhdr", "Standard products file name; product keywords updated")
    else
      header%prefix  = " "
      header%dscode  = " " 
      header%dstype  = " "
      header%version = " "
      header%config  = " "
      call wrchatty("update_stdhdr","Not a standard products file name; product keywords not updated")
    end if

  end subroutine update_stdhdr

!-------------------------------------------------------------------------------







  subroutine fget_i4image(filename,image,ierr)


    type(i4image_type) :: image
    character(len=*)   :: filename
    integer(kind=int32)    :: ierr
    integer(kind=int32)    :: naxis1,naxis2!naxis,
    integer(kind=int32)    :: group,unit,status,rwmode,blocksize
    integer(kind=int32)    :: null_i4 = -9999
    character(len=12)  :: rtname = 'fget_i4image'
    logical            :: anyf   

    group=0
    status=0
    rwmode=0
    ierr=0    
    

    
    if (associated(image%arr)) then
      deallocate(image%arr)
      nullify(image%arr)
    end if


    call ftgiou(unit,status)
    call ftopen(unit,filename,rwmode,blocksize,status)

    if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening FITS file')
    else
      call fread_shapehdr(unit,image%shapehdr,status)
      if (status > 0 .or. image%shapehdr%naxis /= 2) then
        ierr=3
        call wrerr(rtname,ierr,'FITS extension is not a 2D image')
      else
        naxis1 = image%shapehdr%naxis1
        naxis2 = image%shapehdr%naxis2
        call fread_stdhdr(unit,image%stdhdr,status)
        ierr=max(ierr,status)
        call fread_refhdr(unit,image%refhdr,status)
        ierr=max(ierr,status)
        call fread_wcshdr(unit,image%wcshdr,status)
        ierr=max(ierr,status)
        status=0
        allocate(image%arr(naxis1,naxis2)) 
        call ftg2dj(unit,group,null_i4,naxis1,naxis1,naxis2,image%arr,anyf,status)
        if (status > 0) then
          ierr=3
          call wrerr(rtname,ierr,'Error reading FITS image extension')
        endif 
        call ftclos(unit,status) 
      end if  
    end if 

!    write(*,*) rtname, ' obs_mode ', image%stdhdr%obs_mode

  end subroutine  fget_i4image

!-------------------------------------------------------------------------------


  subroutine fget_i1image(filename,image,ierr)


    type(i1image_type) :: image
    character(len=*)   :: filename
    integer(kind=int32)    :: naxis1,naxis2!naxis,
    integer(kind=int32)    :: group,unit,status,rwmode,blocksize,ierr
    integer(kind=int8)    :: null_i1 = -1
    character(len=12)  :: rtname = 'fget_i1image'
    logical(kind=log32)            :: anyf    

    group=0
    status=0
    rwmode=0
    ierr=0    

!    if (associated(image)) then
!      nullify(image)
!    end if
!    allocate(image)

    call ftgiou(unit,status)
    call ftopen(unit,filename,rwmode,blocksize,status)

    if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening FITS file')
    else
      call fread_shapehdr(unit,image%shapehdr,status)
      if (status > 0 .or. image%shapehdr%naxis /= 2) then
        ierr=3
        call wrerr(rtname,ierr,'FITS extension is not a 2D image')
      else
        naxis1 = image%shapehdr%naxis1
        naxis2 = image%shapehdr%naxis2
        call fread_stdhdr(unit,image%stdhdr,status)
        ierr=max(ierr,status)
        call fread_refhdr(unit,image%refhdr,status)
        ierr=max(ierr,status)
        call fread_wcshdr(unit,image%wcshdr,status)
        ierr=max(ierr,status)
        status=0
        allocate(image%arr(naxis1,naxis2)) 
        group=0   
        
        call ftg2db(unit,group,null_i1,naxis1,naxis1,naxis2,image%arr,anyf,status)
        if (status > 0) then
          ierr=3
          call wrerr(rtname,ierr,'Error reading FITS image extension')
        endif 
        call ftclos(unit,status) 
      end if  
    end if 

  end subroutine  fget_i1image

!-------------------------------------------------------------------------------

  subroutine fget_r4image(filename,image,ierr)

    implicit none

    type(r4image_type) :: image
    character(len=*)   :: filename
!    character(len=80)  :: comment
    integer(kind=int32)    :: naxis1,naxis2!naxis,
    integer(kind=int32)    :: group,unit,status,rwmode,blocksize,ierr
    real(kind=real32)       :: null_r4 = -9.9E9
    character(len=12)  :: rtname = 'fget_r4image'
    logical            :: anyf    

    group=0
    status=0
    rwmode=0
    ierr=0    

!    if (associated(image)) then
!      nullify(image)
!    end if
!    allocate(image)

    call ftgiou(unit,status)
    call ftopen(unit,filename,rwmode,blocksize,status)

    if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening FITS file')
    else
      call fread_shapehdr(unit,image%shapehdr,status)
      if (status > 0 .or. image%shapehdr%naxis /= 2) then
        ierr=3
        call wrerr(rtname,ierr,'FITS extension is not a 2D image')
      else
        naxis1 = image%shapehdr%naxis1
        naxis2 = image%shapehdr%naxis2
        call fread_stdhdr(unit,image%stdhdr,ierr)
        ierr=max(ierr,status)
        call fread_refhdr(unit,image%refhdr,status)
        ierr=max(ierr,status)
        call fread_wcshdr(unit,image%wcshdr,status)
        ierr=max(ierr,status)
        status=0
        allocate(image%arr(naxis1,naxis2)) 
        group=0   
        call ftg2de(unit,group,null_r4,naxis1,naxis1,naxis2,image%arr,anyf,status)
        if (status > 0) then
          ierr=3
          call wrerr(rtname,ierr,'Error reading FITS image extension')
        endif 
        call ftclos(unit,status)   
      end if 
     end if
  end subroutine  fget_r4image

!-------------------------------------------------------------------------------

  subroutine fget_r4compress(filename,nima,xbuf,ybuf,maskhdr,compress_flag,image,order,iima,ierr)

  !--------------------------------------------------------------------------------------
  ! Read entire or parts of fits image
  !   If compress_flag == .true. , use maskhdr keywords xdatamin,...,ydatamax
  !   to detrmine relevant part of image
  !   29/09/2011: added order parameter, 
  !   order = 0   : data cube has shape (ipoint,x,y), e.g. suited for erbox
  !   order /= 0  : data cube has shape (x,y,ipoint), e.g. suited for ermldet
  !                 In this case, allocate image%arr(maskhdr%xdatamin:maskhdr%xdatamin+xbuf,maskhdr%ydatamin:maskhdr%ydatamin+ybuf),
  !                 
  !--------------------------------------------------------------------------------------

    implicit none

    type(r4cube_type)  :: image
    type(dethdr_type)  :: maskhdr
    character(len=*)   :: filename
    logical(kind=log32):: compress_flag
    integer(kind=int32):: xbuf,ybuf,iima,nima,order
!    character(len=80)  :: comment
    integer(kind=int32)    :: naxis,naxis1,naxis2
    integer(kind=int32)    :: group,unit,status,rwmode,blocksize,ierr
    integer(kind=int32),dimension(2)             :: fpixels,lpixels,incs,naxes,imshape
    real(kind=real32), dimension(:), allocatable :: imarray 
    integer(kind=int32)     :: nxpix,nypix
    real(kind=real32)       :: null_r4 = -9.9E9
    logical                 :: anyf
    character(len=*),parameter       :: rtname = 'fget_r4compress'



    group=0
    status=0
    rwmode=0
    ierr=0   
    incs=1

    fpixels(1)=maskhdr%xdatamin
    fpixels(2)=maskhdr%ydatamin
    lpixels(1)=maskhdr%xdatamax
    lpixels(2)=maskhdr%ydatamax
    nxpix=lpixels(1)-fpixels(1)+1
    nypix=lpixels(2)-fpixels(2)+1 
    imshape(1)= nxpix
    imshape(2)= nypix 


 !   if (associated(image%arr))      deallocate(image%arr) 
 !   if (associated(image%shapehdr)) deallocate(image%shapehdr) 
 !   if (associated(image%stdhdr))   deallocate(image%stdhdr)
 !   if (associated(image%refhdr))   deallocate(image%refhdr)
 !   if (associated(image%wcshdr))   deallocate(image%wcshdr)
    !   if (associated(image%dethdr))   deallocate(image%dethdr)

    
    if (.not. associated(image%arr)) then
      allocate(image%shapehdr(nima))
      allocate(image%stdhdr(nima))
      allocate(image%refhdr(nima))
      allocate(image%wcshdr(nima))
      allocate(image%dethdr(nima))
    endif
    call ftgiou(unit,status)

    call ftopen(unit,filename,rwmode,blocksize,status)


    if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening FITS file')
    else
      call fread_shapehdr(unit,image%shapehdr(iima),status)
      if (status > 0 .or. image%shapehdr(iima)%naxis /= 2) then
        ierr=3
        call wrerr(rtname,ierr,'FITS extension is not a 2D image')
      else
        naxis  = image%shapehdr(iima)%naxis
        naxis1 = image%shapehdr(iima)%naxis1
        naxis2 = image%shapehdr(iima)%naxis2
        naxes(1) = naxis1
        naxes(2) = naxis2
        call fread_stdhdr(unit,image%stdhdr(iima),ierr)
        call fread_refhdr(unit,image%refhdr(iima),ierr)
        call fread_wcshdr(unit,image%wcshdr(iima),ierr) 
        call fread_dethdr(unit,image%dethdr(iima),ierr)        

        status=0
        group=0  
        if (compress_flag) then
          
          allocate(imarray(nxpix*nypix),STAT=status)
          call ftgsve(unit,group,naxis,naxes,fpixels,lpixels,incs,null_r4,imarray,anyf,status)

          if (order == 0) then
            if (.not. associated(image%arr)) allocate(image%arr(nima,xbuf,ybuf),STAT=status)
            image%arr(iima,:,:)=reshape(imarray,imshape)
          else
            if (.not. associated(image%arr)) then 
              allocate(image%arr(fpixels(1):fpixels(1)+xbuf,fpixels(2):fpixels(2)+ybuf,nima),STAT=status)
            endif
            image%arr(:,:,iima)=reshape(imarray,imshape)            
          endif
          deallocate(imarray)
        else
          if (order == 0) then
              if (.not. associated(image%arr)) allocate(image%arr(nima,naxis1,naxis2),STAT=status)
              call ftg2de(unit,group,null_r4,naxis1,naxis1,naxis2,image%arr(iima,:,:),anyf,status)
          else
              if (.not. associated(image%arr)) allocate(image%arr(naxis1,naxis2,nima),STAT=status)
              call ftg2de(unit,group,null_r4,naxis1,naxis1,naxis2,image%arr(:,:,iima),anyf,status)
          endif
          if (status > 0) then
            ierr=3
            call wrerr(rtname,ierr,'Error reading FITS image extension')
          endif
        endif
 
        call ftclos(unit,status)   
      end if 
     end if


  end subroutine  fget_r4compress

!-------------------------------------------------------------------------------

  subroutine fget_l1compress(filename,nima,xbuf,ybuf,maskhdr,compress_flag,image,order,iima,ierr)

  !-------------------------------------------------------------------------------------------
  ! Read images from list 'filenames' into an logical*1 image cube 'image'.
  ! The dimension of the image cube is (nima,xbuf,ybuf).
  ! If the pointers in structure 'image' are not associated, they
  ! are allocated with the dimensions nima,xbuf,ybuf (compress_flag=.true.)
  ! or dimensions nima,naxis1,naxis2  (compress_flag=.false.)

  ! Input images must be of type 'byte', values are converted to logical(kind=1) 
  ! using btest(value,0) (0 -> .false., 1 -> .true.).
  ! Appropriate for eROSITA detection masks.
  !   29/09/2011: added argument order , see fget_r4compress
  !   
  !-------------------------------------------------------------------------------------------


    implicit none

    type(l1cube_type) :: image
    type(dethdr_type)  :: maskhdr
    character(len=*)   :: filename
    logical(kind=log32):: compress_flag
    integer(kind=int32):: xbuf,ybuf
!    character(len=80)  :: comment
    integer(kind=int32)    :: naxis,naxis1,naxis2
    integer(kind=int32)    :: group,unit,status,rwmode,blocksize,ierr
    integer(kind=int32),dimension(2)   ::  fpixels,lpixels,incs,naxes,imshape
    integer(kind=int8), dimension(:),   allocatable :: imarray
    integer(kind=int8), dimension(:,:), allocatable :: byteimage   
    integer(kind=int32)     :: nxpix,nypix
    integer(kind=int32)     :: iima,nima,order
    integer(kind=int8)       :: null_i1 = -1
    character(len=15)  :: rtname = 'fget_l1compress'
    logical            :: anyf

    group=0
    status=0
    rwmode=0
    ierr=0   
    incs=1

    fpixels(1)=maskhdr%xdatamin
    fpixels(2)=maskhdr%ydatamin
    lpixels(1)=maskhdr%xdatamax
    lpixels(2)=maskhdr%ydatamax
    nxpix=lpixels(1)-fpixels(1)+1
    nypix=lpixels(2)-fpixels(2)+1 
    imshape(1)= nxpix
    imshape(2)= nypix 


 !   if (associated(image%arr))      deallocate(image%arr) 
 !   if (associated(image%shapehdr)) deallocate(image%shapehdr) 
 !   if (associated(image%stdhdr))   deallocate(image%stdhdr)
 !   if (associated(image%refhdr))   deallocate(image%refhdr)
 !   if (associated(image%wcshdr))   deallocate(image%wcshdr)
 !   if (associated(image%dethdr))   deallocate(image%dethdr)

    if (.not. associated(image%arr)) then
      allocate(image%shapehdr(nima))
      allocate(image%stdhdr(nima))
      allocate(image%refhdr(nima))
      allocate(image%wcshdr(nima))
      allocate(image%dethdr(nima))
    endif


    call ftgiou(unit,status)
    call ftopen(unit,filename,rwmode,blocksize,status)

    if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening FITS file')
    else
      call fread_shapehdr(unit,image%shapehdr(iima),status)
      if (status > 0 .or. image%shapehdr(iima)%naxis /= 2) then
        ierr=3
        call wrerr(rtname,ierr,'FITS extension is not a 2D image')
      else
        naxis  = image%shapehdr(iima)%naxis
        naxis1 = image%shapehdr(iima)%naxis1
        naxis2 = image%shapehdr(iima)%naxis2
        naxes(1) = naxis1
        naxes(2) = naxis2
        call fread_stdhdr(unit,image%stdhdr(iima),ierr)
        call fread_refhdr(unit,image%refhdr(iima),ierr)
        call fread_wcshdr(unit,image%wcshdr(iima),ierr) 
        call fread_dethdr(unit,image%dethdr(iima),ierr)        

        status=0
        group=0

        if (compress_flag) then
          
          allocate(imarray(nxpix*nypix),STAT=status)
          call ftgsvb(unit,group,naxis,naxes,fpixels,lpixels,incs,null_i1,imarray,anyf,status)

          if (order == 0) then
            if (.not. associated(image%arr)) allocate(image%arr(nima,xbuf,ybuf),STAT=status)
            image%arr(iima,:,:)=reshape(btest(imarray,0),imshape)
          else
            if (.not. associated(image%arr)) then 
              allocate(image%arr(fpixels(1):fpixels(1)+xbuf,fpixels(2):fpixels(2)+ybuf,nima),STAT=status)
            endif
            image%arr(:,:,iima)=reshape(btest(imarray,0),imshape)            
          endif
          deallocate(imarray)
        else
          allocate(byteimage(naxis1,naxis2))
          call ftg2db(unit,group,null_i1,naxis1,naxis1,naxis2,byteimage,anyf,status)
          if (order == 0) then
            if (.not. associated(image%arr)) allocate(image%arr(nima,naxis1,naxis2),STAT=status)
            image%arr(iima,:,:) = btest(byteimage,0)
          else
            if (.not. associated(image%arr)) allocate(image%arr(naxis1,naxis2,nima),STAT=status)
            image%arr(:,:,iima) = btest(byteimage,0)
          endif
          if (status > 0) then
            ierr=3
            call wrerr(rtname,ierr,'Error reading FITS image extension')
          endif
          deallocate(byteimage)
        endif
        call ftclos(unit,status)   
      end if 
     end if

     
  end subroutine  fget_l1compress


!-------------------------------------------------------------------------------

  subroutine fget_i4compress(filename,nima,xbuf,ybuf,maskhdr,compress_flag,image,order,iima,ierr)

  !---------------------------------------------------------------------------
  ! Read images from list 'filenames' into an integer*4 image cube 'image'.
  ! The dimension of the image cube is (nima,xbuf,ybuf).
  ! If the pointers in structure 'image' are not associated, they
  ! are allocated with the dimensions nima,xbuf,ybuf (compress_flag=.true.)
  ! or dimensions nima,naxis1,naxis2  (compress_flag=.false.)
  !   29/09/2011: added argument order , see fget_r4compress
  !---------------------------------------------------------------------------


    implicit none

    type(i4cube_type) :: image
    type(dethdr_type)  :: maskhdr
    character(len=*)   :: filename
    logical(kind=log32):: compress_flag
!    character(len=80)  :: comment
    integer(kind=int32)    :: xbuf,ybuf,iima,nima,order
    integer(kind=int32)    :: naxis,naxis1,naxis2
    integer(kind=int32)    :: group,unit,status,rwmode,blocksize,ierr
    integer(kind=int32), dimension(2)    ::  fpixels,lpixels,incs,naxes,imshape
    integer(kind=int32), dimension(:), allocatable :: imarray 
    integer(kind=int32)       :: nxpix,nypix
    integer(kind=int32)       :: null_i4 = -9999
    character(len=15)         :: rtname = 'fget_i4compress'
    logical                   :: anyf

    group=0
    status=0
    rwmode=0
    ierr=0   
    incs=1

    fpixels(1)=maskhdr%xdatamin
    fpixels(2)=maskhdr%ydatamin
    lpixels(1)=maskhdr%xdatamax
    lpixels(2)=maskhdr%ydatamax
    nxpix=lpixels(1)-fpixels(1)+1
    nypix=lpixels(2)-fpixels(2)+1 
    imshape(1)= nxpix
    imshape(2)= nypix

!    if (associated(image%arr))      deallocate(image%arr) 
!    if (associated(image%shapehdr)) deallocate(image%shapehdr) 
!    if (associated(image%stdhdr))   deallocate(image%stdhdr)
!    if (associated(image%refhdr))   deallocate(image%refhdr)
!    if (associated(image%wcshdr))   deallocate(image%wcshdr)
!    if (associated(image%dethdr))   deallocate(image%dethdr)


!    if (.not. associated(image%arr)) then
!      allocate(image%shapehdr(nima))
!      allocate(image%stdhdr(nima))
!      allocate(image%refhdr(nima))
!      allocate(image%wcshdr(nima))
!      allocate(image%dethdr(nima))
!    endif


    
    if (.not. associated(image%arr)) then
       allocate(image%shapehdr(nima))
      allocate(image%stdhdr(nima))
      allocate(image%refhdr(nima))
      allocate(image%wcshdr(nima))
      allocate(image%dethdr(nima))
    endif  


    
 
    call ftgiou(unit,status)
    call ftopen(unit,filename,rwmode,blocksize,status)

    if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening FITS file')
    else
      call fread_shapehdr(unit,image%shapehdr(iima),status)
      if (status > 0 .or. image%shapehdr(iima)%naxis /= 2) then
        ierr=3
        call wrerr(rtname,ierr,'FITS extension is not a 2D image')
      else
        naxis  = image%shapehdr(iima)%naxis
        naxis1 = image%shapehdr(iima)%naxis1
        naxis2 = image%shapehdr(iima)%naxis2
        naxes(1) = naxis1
        naxes(2) = naxis2
        call fread_stdhdr(unit,image%stdhdr(iima),ierr)
        call fread_refhdr(unit,image%refhdr(iima),ierr)
        call fread_wcshdr(unit,image%wcshdr(iima),ierr) 
        call fread_dethdr(unit,image%dethdr(iima),ierr)        

        status=0
        group=0  
        if (compress_flag) then
         
          allocate(imarray(nxpix*nypix),STAT=status)
          call ftgsvj(unit,group,naxis,naxes,fpixels,lpixels,incs,null_i4,imarray,anyf,status)

          if (order == 0) then
            if (.not. associated(image%arr)) allocate(image%arr(nima,xbuf,ybuf),STAT=status)
            image%arr(iima,:,:)=reshape(imarray,imshape)
          else
            if (.not. associated(image%arr)) then
              allocate(image%arr(fpixels(1):fpixels(1)+xbuf,fpixels(2):fpixels(2)+ybuf,nima),STAT=status)
            endif
            image%arr(:,:,iima)=reshape(imarray,imshape)            
          endif
          deallocate(imarray)
        else
          if (order == 0) then
              if (.not. associated(image%arr)) allocate(image%arr(nima,naxis1,naxis2),STAT=status)
              call ftg2dj(unit,group,null_i4,naxis1,naxis1,naxis2,image%arr(iima,:,:),anyf,status)
          else
              if (.not. associated(image%arr)) allocate(image%arr(naxis1,naxis2,nima),STAT=status)
              call ftg2dj(unit,group,null_i4,naxis1,naxis1,naxis2,image%arr(:,:,iima),anyf,status)
          endif
          if (status > 0) then
            ierr=3
            call wrerr(rtname,ierr,'Error reading FITS image extension')
          endif
        endif
 
        call ftclos(unit,status)   
      end if 
     end if

 
  end subroutine  fget_i4compress




!-------------------------------------------------------------------------------

  subroutine fput_i4image(filename,image,ierr)

    implicit none

    type(i4image_type) :: image
    character(len=*)   :: filename
!    character(len=80)  :: comment
    integer(kind=int32)    :: bitpix,naxis!,naxis1,naxis2
    integer(kind=int32), dimension(2) :: naxes
    integer(kind=int32)    :: nhdu,group,unit,status,blocksize,hdutype,ierr

    character(len=12)  :: rtname = 'fput_i4image'
!    logical            :: anyf    


    group=0
    status=0
!   use default blocksize:
    blocksize=0
    ierr=0   
    nhdu=1 
    bitpix=32


    call ftgiou(unit,status)
     call ftinit(unit,filename,blocksize,status)
     if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error creating FITS file '//trim(filename))
    else
      call ftmahd(unit,nhdu,hdutype,status)
       if (status > 0) then
        ierr = 3
        call wrerr(rtname,ierr,'Error creating FITS  image extension')
      else
        naxis=2
        naxes(1) = image%shapehdr%naxis1
        naxes(2) = image%shapehdr%naxis2
        status = 0
        call ftphps(unit,bitpix,naxis,naxes,status)
        if (status > 0) ierr = 3
        call fwrite_stdhdr(unit,image%stdhdr,status)
        if (status > 0) ierr = 3 
        call fwrite_refhdr(unit,image%refhdr,status)
        if (status > 0) ierr = 3 
        call fwrite_wcshdr(unit,image%wcshdr,status)
        if (status > 0) ierr = 3 
        call fwrite_dethdr(unit,image%dethdr,status)
        if (status > 0) ierr = 3 

        status=0
        call ftp2dj(unit,group,naxes(1),naxes(1),naxes(2),image%arr,status)
        if (status > 0) then
          ierr=3
          call wrerr(rtname,ierr,'Error writing FITS image extension')
        endif 
      endif
      call ftclos(unit,status) 
      if (status > 0) ierr = 3 
    end if
  end subroutine  fput_i4image

!-------------------------------------------------------------------------------

  subroutine fput_i1image(filename,image,ierr)

    implicit none

    type(i1image_type) :: image
    character(len=*)   :: filename
!    character(len=80)  :: comment
    integer(kind=int32)    :: bitpix,naxis!,naxis1,naxis2
    integer(kind=int32), dimension(2) :: naxes
    integer(kind=int32)    :: nhdu,group,unit,status,blocksize,hdutype,ierr

    character(len=12)  :: rtname = 'fput_i1image'
!    logical            :: anyf    


    group=0
    status=0
!   use default blocksize:
    blocksize=0
    ierr=0   
    nhdu=1 
!    group=0   
    bitpix=8


    call ftgiou(unit,status)
     call ftinit(unit,filename,blocksize,status)
     if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error creating FITS file')
    else
      call ftmahd(unit,nhdu,hdutype,status)
       if (status > 0) then
        ierr = 3
        call wrerr(rtname,ierr,'Error creating FITS  image extension')
      else
        naxis=2
        naxes(1) = image%shapehdr%naxis1
        naxes(2) = image%shapehdr%naxis2
        status = 0
        call ftphps(unit,bitpix,naxis,naxes,status)
        if (status > 0) ierr = 3
        call fwrite_stdhdr(unit,image%stdhdr,status)
        if (status > 0) ierr = 3 
        call fwrite_refhdr(unit,image%refhdr,status)
        if (status > 0) ierr = 3 
        call fwrite_wcshdr(unit,image%wcshdr,status)
        if (status > 0) ierr = 3 
        call fwrite_dethdr(unit,image%dethdr,status)
        if (status > 0) ierr = 3 
        status=0
        call ftp2db(unit,group,naxes(1),naxes(1),naxes(2),image%arr,status)
        if (status > 0) then
          ierr=3
        endif 
      endif
      call ftclos(unit,status) 
      if (status > 0) ierr = 3 
    end if
  end subroutine  fput_i1image

!--------------------------------------------------------------------------------

  subroutine fput_r4image(filename,image,ierr)

    implicit none

    type(r4image_type) :: image
    character(len=*)   :: filename
!    character(len=80)  :: comment
    integer(kind=int32)    :: naxis,naxis1,naxis2,bitpix,pcount=0,gcount=1
    integer(kind=int32),dimension(2)    :: naxes
    integer(kind=int32)    :: nhdu,group,unit,status,blocksize,ierr!,hdutype
    character(len=12)  :: rtname = 'fput_r4image'
    logical            :: simple=.true., extend=.true.  ! anyf, 

    status=0
!   use default blocksize:
    blocksize=0
    ierr=0   
    nhdu=1 
    group=0
    bitpix=-32  
    naxis=2
    

    call ftgiou(unit,status)
    call ftinit(unit,filename,blocksize,status)
     if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error creating FITS file '//trim(filename))
    else
      naxis1 = image%shapehdr%naxis1
      naxis2 = image%shapehdr%naxis2
      naxes(1)=naxis1
      naxes(2)=naxis2
!      call ftmahd(unit,nhdu,hdutype,status)
      call FTPHPR(unit,simple,bitpix,naxis,naxes,pcount,gcount,extend, status)
      if (status > 0) then
        ierr = 3
        call wrerr(rtname,ierr,'Error creating FITS  image extension')
      else
        status = 0

        call fwrite_stdhdr(unit,image%stdhdr,status)
        if (status > 0) ierr = 3 
!        call fwrite_refhdr(unit,image%refhdr,status)
!        if (status > 0) ierr = 3 
        call fwrite_wcshdr(unit,image%wcshdr,status)
        if (status > 0) then 
          ierr = 3 
          call wrerr(rtname,ierr,'Error writing wcshdr') 
        endif
        call fwrite_dethdr(unit,image%dethdr,status)
        if (status > 0) then 
          ierr = 3 
          call wrerr(rtname,ierr,'Error writing dethdr')
        endif 

        status=0
        call ftp2de(unit,group,naxis1,naxis1,naxis2,image%arr,status)
        if (status > 0) then
          ierr=3
          call wrerr(rtname,ierr,'Error writing FITS image extension')
        endif 
      endif
      call ftclos(unit,status) 
      if (status > 0) then 
        ierr = 3
      endif 
    end if
  end subroutine  fput_r4image

!-------------------------------------------------------------------------------

  logical function match_wcs(wcsheader1,wcsheader2)

  !------------------------------------------------------------------------
  ! Check whether keywords crpix1, crval1, cdelt1,  crpix2, crval2, cdelt2
  ! exist and are equal  in both input headers
  ! If yes, match_wcs=.true.    
  !------------------------------------------------------------------------


    type(wcs_imahdr_type) :: wcsheader1,wcsheader2
    real(kind=real64)     :: epsilon=1.D-6
    integer(kind=int32)   :: err

    match_wcs = .false.

    err = 6
    
    if (wcsheader1%e_crpix1 .and. wcsheader2%e_crpix1) then
      if  (abs(wcsheader1%crpix1 - wcsheader2%crpix1) <= epsilon*abs(wcsheader1%crpix1)) then
        err = err - 1 
      endif
    endif 
    
    if (wcsheader1%e_crpix2 .and. wcsheader2%e_crpix2) then
      if  (abs(wcsheader1%crpix2 - wcsheader2%crpix2) <= epsilon*abs(wcsheader1%crpix2)) then
        err = err - 1 
      endif
    endif 
   
    if (wcsheader1%e_crval1 .and. wcsheader2%e_crval1) then
      if  (abs(wcsheader1%crval1 - wcsheader2%crval1) <= epsilon*abs(wcsheader1%crval1)) then
        err = err - 1 
      endif
    endif 
       
    if (wcsheader1%e_crval2 .and. wcsheader2%e_crval2) then
      if  (abs(wcsheader1%crval2 - wcsheader2%crval2) <= epsilon*abs(wcsheader1%crval2)) then
        err = err - 1 
      endif
    endif 
    
  
    if (wcsheader1%e_cdelt1 .and. wcsheader2%e_cdelt1) then
      if  (abs(wcsheader1%cdelt1 - wcsheader2%cdelt1) <= epsilon*abs(wcsheader1%cdelt1)) then
        err = err - 1 
      endif
    endif 
       
    if (wcsheader1%e_cdelt2 .and. wcsheader2%e_cdelt2) then
      if  (abs(wcsheader1%cdelt2 - wcsheader2%cdelt2) <= epsilon*abs(wcsheader1%cdelt2)) then
        err = err - 1 
      endif
    endif 
    
    if (err == 0) match_wcs = .true. 


    
  end function match_wcs

!-------------------------------------------------------------------------------

  logical function match_size(header1,header2) 

  !-----------------------------------------------------------------------
  ! Check whether keywords naxis, naxis2, naxis2
  ! exist and are equal  in both input headers (shapeheader_type).
  ! If yes, match_size=.true.    
  !-----------------------------------------------------------------------


    type(shapehdr_type) ::   header1,header2
    integer(kind=int32) :: err

    match_size = .false.

    err = 3
    
    if (header1%e_naxis .and. header2%e_naxis) then
      if  (header1%naxis == header2%naxis) then
        err = err - 1 
      endif
    endif 
    
    if (header1%e_naxis1 .and. header2%e_naxis1) then
      if  (header1%naxis1 == header2%naxis1) then
        err = err - 1 
      endif
    endif 
  
     
    if (header1%e_naxis2 .and. header2%e_naxis2) then
      if  (header1%naxis2 == header2%naxis2) then
        err = err - 1 
      endif
    endif 
    
    if (err == 0) match_size = .true. 

  end function match_size


!-------------------------------------------------------------------------------

  logical function match_exposures(header1,header2) 

  !-----------------------------------------------------------------------
  ! Check whether keywords TELESCOP, INSTRUME and EXP_ID 
  ! exist and are equal  in both input headers (stdheader_type).
  ! If yes, match_exposures=.true.  
  ! 2014-12-05 GL:  Use (internal) header keyword INSTRUMENTS for comparison  
  !-----------------------------------------------------------------------


    type(stdhdr_type) ::   header1,header2
    integer(kind=int32)   :: err

    character(len=*), parameter      :: rtname='MATCH_EXPOSURES'


    match_exposures = .false.


    err = 3
   
    if (header1%e_telescop .and. header2%e_telescop) then
      if  (header1%telescop == header2%telescop) then
        err = err - 1 
      endif
    endif 
       
    if ((header1%e_instrume .and. header2%e_instrume) .or. (header1%e_instrumx(1) .and. header2%e_instrumx(1))) then
      if  (header1%instruments == header2%instruments) then
        err = err - 1 
      endif
    endif 

       
!    if (header1%e_instrume .and. header2%e_instrume) then
!      if  (header1%instrume == header2%instrume) then
!        err = err - 1 
!      endif
!    endif 


    
    if (header1%e_exp_id .and. header2%e_exp_id) then
      if  (header1%exp_id == header2%exp_id) then
        err = err - 1 
      endif
    endif 
      
!    write(*,*) rtname, ' ',header1%telescop, header2%telescop
!    write(*,*) rtname, ' ',header1%instruments, header2%instruments
!    write(*,*) rtname, ' ',header1%exp_id, header2%exp_id
!   write(*,*) rtname, ' ',err

     
    if (err == 0) match_exposures = .true. 

  end function match_exposures


!-------------------------------------------------------------------------------

  subroutine copy_i1image(old_image,new_image,ierr)

  !---------------------------------------------------------------------------
  ! Duplicates the image <old_image> of type 'i1image_type' 
  ! to image <new_image>.  
  ! NOTE: The data array (new_image%arr) is allocated here.
  !       Don't use attribute 'intent(out)' for new_image, since
  !       new_image%arr is deallocated if already associated. 
  ! See also: copy_i4image, copy_r4image
  !---------------------------------------------------------------------------

    type(i1image_type)         :: old_image
    type(i1image_type)         :: new_image 
    integer(kind=int32)        :: ierr
    integer(kind=int32)        :: istat
    character(len=*),parameter :: rtname = 'copy_i1image'
    

    ierr = 0
    istat = 0
 
    if (associated(new_image%arr)) then
      deallocate(new_image%arr)
      nullify(new_image%arr)
    endif

    allocate(new_image%arr(old_image%shapehdr%naxis1,old_image%shapehdr%naxis2), STAT=istat)
 
    if (istat == 0) then
      new_image%arr       = old_image%arr
      new_image%shapehdr  = old_image%shapehdr
      new_image%stdhdr    = old_image%stdhdr
      new_image%refhdr    = old_image%refhdr
      new_image%wcshdr    = old_image%wcshdr
    else 
      ierr=3
      call wrerr(rtname,ierr,'Cannot allocate memory')
    endif
    
  end subroutine copy_i1image 
     
!--------------------------------------------------------------------------------

  subroutine copy_i4image(old_image,new_image,ierr)

  !---------------------------------------------------------------------------
  ! Duplicates the image <old_image> of type 'i4image_type' 
  ! to image <new_image>.  
  ! NOTE: The data array (new_image%arr) is allocated here.
  ! See also: copy_i1image, copy_r4image
  !---------------------------------------------------------------------------
 
    type(i4image_type)              :: old_image
    type(i4image_type)              :: new_image 
    integer(kind=int32)             :: ierr
    integer(kind=int32)             :: istat
    character(len=*),parameter      :: rtname = 'copy_i4image'
    

    ierr = 0
    istat = 0
    
    if (associated(new_image%arr)) then
      deallocate(new_image%arr)
      nullify(new_image%arr)
    endif
    
    allocate(new_image%arr(old_image%shapehdr%naxis1,old_image%shapehdr%naxis2), STAT=istat)
 

    if (istat == 0) then
      new_image%arr       = old_image%arr
      new_image%shapehdr  = old_image%shapehdr
      new_image%stdhdr    = old_image%stdhdr
      new_image%refhdr    = old_image%refhdr
      new_image%wcshdr    = old_image%wcshdr
    else 
      ierr=3
      call wrerr(rtname,ierr,'Cannot allocate memory')
    endif


  end subroutine copy_i4image 
       
!--------------------------------------------------------------------------------


  subroutine copy_r4image(old_image,new_image,ierr)

  !---------------------------------------------------------------------------
  ! Duplicates the image <old_image> of type 'r4image_type' 
  ! to image <new_image>.  
  ! NOTE: The data array (new_image%arr) is allocated here.
  ! See also: copy_i1image, copy_i4image
  !---------------------------------------------------------------------------
 
    type(r4image_type)              :: old_image
    type(r4image_type)              :: new_image 
    integer(kind=int32)             :: ierr
    integer(kind=int32)             :: istat
    character(len=*),parameter      :: rtname = 'copy_r4image'

    ierr = 0
    istat = 0
   
    if (associated(new_image%arr)) then
      deallocate(new_image%arr)
      nullify(new_image%arr)
    endif

    allocate(new_image%arr(old_image%shapehdr%naxis1,old_image%shapehdr%naxis2), STAT=istat)
 
    if (istat == 0) then
      new_image%arr       = old_image%arr
      new_image%shapehdr  = old_image%shapehdr
      new_image%stdhdr    = old_image%stdhdr
      new_image%refhdr    = old_image%refhdr
      new_image%wcshdr    = old_image%wcshdr
    else 
      ierr=3
      call wrerr(rtname,ierr,'Cannot allocate memory')
    endif
    
  end subroutine copy_r4image 

        


!--------------------------------------------------------------------------------


  subroutine pix2equ(header,xpix,xerr,ypix,yerr,dra,ddec,radecerr,ierr)

    !------------------------------------------------------------------------------
    ! Calculates RA and DEC coordinates from image pixel coordinates for
    ! parallel (orthographic) projection.
    ! see Greisen: AIPS Memo 27 , 1994
    ! Input variables:
    !  header    : WCS header 
    !  xpix      : x image pixel coordinate  
    !  ypix      : y image pixel coordinate  
    ! Output variables:
    !  dra       : RA in decimal degrees
    !  ddec      : DEC in decimal degrees 
    !-------------------------------------------------------------------------------



    type(wcs_imahdr_type),intent(in)   :: header 
    real(kind=real32),intent(in)       :: xpix,ypix,xerr,yerr
    real(kind=real32),intent(out)      :: radecerr 
    real(kind=real64),intent(out)      :: dra,ddec
    integer(kind=int32),intent(out)    :: ierr
    real(kind=real64)                  :: ra,dec,dx,dy
    real(kind=real64)                  :: sin2,cos2,dt
    real(kind=real64)                  :: ra0,dec0,cosdec0,sindec0
    real(kind=real64)                  :: xpix2rad,ypix2rad
    real(kind=real64),parameter        :: deg2rad=0.0174532925199433 
    real(kind=real64),parameter        :: dpi=3.141592653589793,twopi=6.2831853071795862
    real(kind=real32)                  :: minus1,r32null
    
     character(len=*),parameter        :: rtname = 'PIX2EQU'  
   
    minus1=-1.0
    r32null=sqrt(minus1)
    ierr = 0
    ra=-999.D0
    dec=-999.D0
    dra  = 0.D0
    ddec = 0.D0
 
    xpix2rad=deg2rad*header%cdelt1
    ypix2rad=deg2rad*header%cdelt2

    ra0      = header%crval1 * deg2rad
    dec0     = header%crval2 * deg2rad
    cosdec0  = cos(dec0)
    sindec0  = sin(dec0)   

    dx = (xpix -  header%crpix1)*xpix2rad
    dy = (ypix -  header%crpix2)*ypix2rad



    sin2 = dx*dx + dy*dy

    if (sin2 > 1.D0) then
      ierr = 3
    else
      cos2 = sqrt (1.D0 - sin2)
      ! fix bug (22/03/2013): 
      !dt =  sindec0 * cos2
      dt =  sindec0 * cos2 + cosdec0 * dy
      if ((dt>1.D0) .or. (dt<-1.D0)) then
        ierr = 1
      else
        dec = asin(dt)
        ra = cosdec0 * cos2 - sindec0 * dy
        if ((ra == 0.D0) .and. (dx == 0.D0)) then
          ierr = 1
        else
          ra = ra0 + atan2(dx,ra)
        endif
      endif
    endif
    
    if ((ra  - ra0) > dpi)  ra = ra - twopi
    if ((ra  - ra0) < -dpi) ra = ra + twopi
    if (ra < 0.D0) ra = ra + twopi

    dra  = ra  / deg2rad
    ddec = dec / deg2rad  
    if (xerr > 0.0 .and. yerr > 0.0) then 
      radecerr =  real(SQRT((xerr*header%cdelt1)**2 + (xerr*header%cdelt2)**2) * 3600.D0 ,kind=real32  )
    else
      radecerr = r32null
    endif

  end subroutine pix2equ

        
!--------------------------------------------------------------------------------


  subroutine pix2equarr(header,xpix,ypix,dra,ddec,ierr)

    !------------------------------------------------------------------------------
    ! Calculates RA and DEC coordinates from image pixel coordinates for
    ! parallel (orthographic) projection.
    ! Input variables:
    !  header    : WCS header 
    !  xpix      : x image pixel coordinate  
    !  ypix      : y image pixel coordinate  
    ! Output variables:
    !  dra       : RA in decimal degrees
    !  ddec      : DEC in decimal degrees 
    !-------------------------------------------------------------------------------



    type(wcs_imahdr_type),intent(in)                :: header 
    real(kind=real64),dimension(:),intent(in)       :: xpix,ypix 
    real(kind=real64),dimension(:),intent(out)      :: dra,ddec 
    integer(kind=int32),intent(out)                 :: ierr
    real(kind=real64)                               :: ra,dec,dx,dy
    real(kind=real64)                  :: sin2,cos2,dt
    real(kind=real64)                  :: ra0,dec0,cosdec0,sindec0
    integer(kind=int32)                :: narr,i
    real(kind=real64)                  :: xpix2rad,ypix2rad
    real(kind=real64),parameter        :: deg2rad=0.0174532925199433 
    real(kind=real64),parameter        :: dpi=3.141592653589793,twopi=6.2831853071795862  
     character(len=*),parameter        :: rtname = 'PIX2EQUARR'  

    ierr = 0 
    ra = -999.D0
    dec = -999.D0
    
    narr = size(xpix)  
    
    dra  = 0.D0
    ddec = 0.D0

    if (size(dra) == narr .and. size(ddec) == narr .and.  size(ypix) == narr) then

  
      xpix2rad=deg2rad*header%cdelt1
      ypix2rad=deg2rad*header%cdelt2



      ra0      = header%crval1 * deg2rad
      dec0     = header%crval2 * deg2rad
      cosdec0  = cos(dec0)
      sindec0  = sin(dec0)   

      do i=lbound(dra,1),ubound(dra,1)

        dx = (xpix(i) -  header%crpix1)*xpix2rad
        dy = (ypix(i) -  header%crpix2)*xpix2rad

        sin2 = dx*dx + dy*dy

        if (sin2 > 1.D0) then
          ierr = 1
        else
          cos2 = sqrt (1.D0 - sin2)
          dt =  sindec0 * cos2
          if ((dt > 1.D0) .or. (dt < -1.D0)) then
            ierr = 1
          else
            dec = asin(dt)
            ra = cosdec0 * cos2 - sindec0 * dy
            if ((ra == 0.D0) .and. (dx == 0.D0)) then
              ierr = 1
            else
              ra = ra0 + atan2(dx,ra)
            endif
          endif
        endif
    
        if ((ra  - ra0) > dpi)  ra = ra - twopi
        if ((ra  - ra0) < -dpi) ra = ra + twopi
        if (ra < 0.D0)          ra = ra + twopi

        dra(i)  = ra  / deg2rad
        ddec(i) = dec / deg2rad   
      
      enddo 
    else 
      ierr=3
      call wrerr(rtname,ierr,'Input and output  arrays must have equal dimensions')
    endif
  
  end subroutine pix2equarr


       
!--------------------------------------------------------------------------------


  subroutine equ2pix(header,dra,ddec,xpix,ypix,ierr)

    !------------------------------------------------------------------------------
    ! Calculates image pixel coordinates from  RA and DEC coordinates
    ! for parallel (orthographic) projection.
    ! 
    ! Input variables:
    !  header    : WCS header 
    !  dra       : RA in decimal degrees
    !  ddec      : DEC in decimal degrees 
    ! Output variables:
    !  xpix      : x image pixel coordinate  
    !  ypix      : y image pixel coordinate   
    !  ierr      : error status
    ! Note: use equ2pixarr to process arrays of input/output data.
    !-------------------------------------------------------------------------------

    type(wcs_imahdr_type),intent(in)             :: header 
    real(kind=real64),intent(in)                 :: dra,ddec
    real(kind=real64),intent(out)                :: xpix,ypix 
    integer(kind=int32),intent(out)              :: ierr
    real(kind=real64)                            :: ra,dec,dx,dy
    real(kind=real64)                            :: ra0,dec0,rad2xpix,rad2ypix
    real(kind=real64),parameter                  :: deg2rad=0.0174532925199433  
    character(len=*),parameter                   :: rtname = 'EQU2PIX'

   

    ierr = 0 
 

    rad2xpix = 1.D0/deg2rad/header%cdelt1  
    rad2ypix = 1.D0/deg2rad/header%cdelt2   


    ra   = dra  * deg2rad
    dec  = ddec * deg2rad 
    ra0  = header%crval1 * deg2rad
    dec0 = header%crval2 * deg2rad
 
    
   
    dx = sin(ra-ra0) * cos(dec)
    dy = sin(dec)*cos(dec0)-cos(dec)*sin(dec0)*cos(ra-ra0)


    xpix = dx*rad2xpix + header%crpix1
    ypix = dy*rad2ypix + header%crpix2

 
  end subroutine equ2pix


       
!--------------------------------------------------------------------------------

  subroutine equ2pixarr(header,dra,ddec,xpix,ypix,ierr)

  !------------------------------------------------------------------------------
    ! Calculates image pixel coordinates from  RA and DEC coordinates
    ! for parallel (orthographic) projection.
    ! 
    ! Input variables:
    !  header    : WCS header 
    !  dra       : array of RA in decimal degrees
    !  ddec      : array of DEC in decimal degrees 
    ! Output variables:
    !  xpix      : x image pixel coordinates  
    !  ypix      : y image pixel coordinate   
    !  ierr      : error status
    ! Note: use equ2pix to process scalar input/output data.
    !-------------------------------------------------------------------------------


    type(wcs_imahdr_type),intent(in)             :: header 
    real(kind=real64),dimension(:),intent(in)    :: dra,ddec
    real(kind=real64),dimension(:),intent(out)   :: xpix,ypix 
    integer(kind=int32),intent(out)              :: ierr
    integer(kind=int32)                          :: narr
    real(kind=real64),dimension(:),allocatable   :: ra,dec,dx,dy
    real(kind=real64)                            :: ra0,dec0,rad2xpix,rad2ypix
    real(kind=real64),parameter                  :: deg2rad=0.0174532925199433  
    character(len=*),parameter                   :: rtname = 'EQU2PIXARR'

    ierr = 0 

    narr = size(dra)    
    if (size(ddec) == narr .and. size(xpix) == narr .and.  size(ypix) == narr) then

      if (allocated(ra))   deallocate(ra)
      if (allocated(dec))  deallocate(dec)
      if (allocated(dx))   deallocate(dx)
      if (allocated(dy))   deallocate(dy)     

      allocate(ra(narr))
      allocate(dec(narr))
      allocate(dx(narr))
      allocate(dy(narr))

      rad2xpix = 1.D0/deg2rad/header%cdelt1   
      rad2ypix = 1.D0/deg2rad/header%cdelt2  
 
      ra   = dra  * deg2rad
      dec  = ddec * deg2rad 
      ra0  = header%crval1 * deg2rad
      dec0 = header%crval2 * deg2rad
      dx = sin(ra-ra0) * cos(dec)
      dy = sin(dec)*cos(dec0)-cos(dec)*sin(dec0)*cos(ra-ra0)
      xpix = dx*rad2xpix + header%crpix1
      ypix = dy*rad2ypix + header%crpix2

      deallocate(ra)
      deallocate(dec)
      deallocate(dx)
      deallocate(dy)  
    else
      ierr=3
      call wrerr(rtname,ierr,'Input and output  arrays must have equal dimensions')
    endif

 
  end subroutine equ2pixarr

!--------------------------------------------------------------------------------



  subroutine skypix2ima(header,x,y,pixlist,ierr)

    !------------------------------------------------------------------------------
    !!!!! Obsolete, use ermldet_mod::init_pixlist
    ! Calculates image pixel coordinates from sky X and Y
    ! pixel coordinates using a SASS WCS header.
    ! 
    ! Input variables:
    !  header    : WCS header 
    !  x       : array of physical X positions  (32 bit integer)
    !  y      :  array of physical Y positions  (32 bit integer)
    ! Output variables:
    !  xpix      : x image pixel coordinates  (16 bit integer)
    !  ypix      : y image pixel coordinate   (16 bit integer)   
    !  ierr      : error status
    ! 
    !-------------------------------------------------------------------------------

   type(wcs_imahdr_type),intent(in)             :: header
   integer(kind=int32),dimension(:),intent(in)  :: x,y
   type(pixlist_type)                           :: pixlist
!   integer(kind=int16),dimension(:),intent(out)    :: xpix,ypix 
   integer(kind=int32),intent(out)              :: ierr

   integer(kind=int32)                          :: narr,i
   character(len=*),parameter                   :: rtname = 'SKYPIX2IMA'

   ierr=0

   narr=size(x)

   if (.not. allocated(pixlist%pix)) then
     allocate(pixlist%pix(narr))
   endif

   narr=size(x)
   if (size(pixlist%pix) == narr ) then

     do i=1, narr
       pixlist%pix(i)%id=i
       pixlist%pix(i)%x=nint(header%crpix1p+(real(x(i))-header%crval1p)/header%cdelt1p,kind=int16)
       pixlist%pix(i)%y=nint(header%crpix2p+(real(y(i))-header%crval2p)/header%cdelt2p,kind=int16) 
     enddo
   
   else
     ierr=3
     call wrerr(rtname,ierr,'Array size mismatch')
   endif

   end subroutine skypix2ima

!--------------------------------------------------------------------------------


function to_upper(str) 

!   ==============================
!   Changes a string to upper case
!   ==============================

    implicit none
    character(*), intent(In) :: str
    character(len(str))      :: to_upper

    integer :: ic, i

    character(26), parameter :: cap = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    character(26), parameter :: low = 'abcdefghijklmnopqrstuvwxyz'

!   Capitalize each letter if it is lowecase
    to_upper = str
    do i = 1, LEN_TRIM(str)
        ic = INDEX(low, str(i:i))
        if (ic > 0) to_upper(i:i) = cap(ic:ic)
    end do

end function to_upper


!--------------------------------------------------------------------------------


function to_lower(str) 

!   ==============================
!   Changes a string to lower case
!   ==============================

    implicit none
    character(*), intent(In) :: str
    character(len(str))      :: to_lower

    integer :: ic, i

    character(26), parameter :: cap = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    character(26), parameter :: low = 'abcdefghijklmnopqrstuvwxyz'

!   Capitalize each letter if it is lowercase
    to_lower = str
    do i = 1, LEN_TRIM(str)
        ic = INDEX(cap, str(i:i))
        if (ic > 0) to_lower(i:i) = low(ic:ic)
    end do

end function to_lower



!--------------------------------------------------------------------------------



end module fitsimage
