!-----------------------------------------------------------------------
! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------

module tel_io

!-------------------------------------------------------------------------------
! Module containing fits table input/output routines for eROSITA TEL pipeline 
! Author: Georg Lamer , AIP         date:  31-MAY-2012
!
!
! subroutines: 
!   fget_telgti(filename,gti,ierr) 
!   fget_hk(filename,ext,table,ierr)
!   fget_skymaps(filename,maps,ierr)
!   fget_splan(filename,plan,ierr)
!   fget_pplan(filename,plan,ierr) 
!   fget_plan(filename,plan,ierr)    
!   fget_evraw(filename,evlist,ierr)
!   fget_evlist(filename,evlist,ierr)
!   fget_evatt_list(filename,ext,table,ierr)  
!   fput_telgti(filename,append,gti,ierr)
!   fput_evraw(filename,append,evlist,ierr)
!   fput_evlist(filename,append,evlist,ierr)
!   fput_plan(filename,append,plan,ierr)
!   fput_hk(filename,append,hk,ierr)
!   eval_plan_name(filename,pdate,pstart,pstop,ierr)
!   str2int16(numstr,inum,ierr)
!   int16str6(inum,numstr,ierr)
!   ingti(time,gti,good,ierr)
!   get_instnum(instkey,instnum,ierr) 
!
!-------------------------------------------------------------------------------

  use param
  use fitsimage
  use cfitsio_ifaces
  use numkind
  use astrolib

  
  public 

  ! general tel-chain parameters:
  
  ! sky pixel size in degrees 
  real(kind=real64),parameter        ::  tel_skypixsize = 1.3888888889D-05
!  moved to cal_mod.f90:
!  real(kind=real64),parameter        ::  deg2rad=1.74532925199D-02, rad2deg=57.29577951322
 

  ! Constants for filter wheel values 
  ! Changed 2015-04-28 on request from MJF:
  integer(kind=int16),parameter      :: fw_closed=0, fw_open=1, fw_filter=2, fw_cal=3, fw_inter=4





  ! calibrated event list format:
  ! NOTE: Data scaling is used for RA, DEC, SUBX, SUBY,
  ! physical format in fits table is int32 for RA, DEC, int8 for SUBX, SUBY
  ! Columns OTS and FRACSEC belong to raw events format, can optionally be kept in calibrated format (tbd)


 

  type evrow_type
    !sequence
    real(kind=real64)      :: time
    integer(kind=int32)    :: ots
    integer(kind=int32)    :: fracsec
    real(kind=real64)      :: ra
    real(kind=real64)      :: dec
    integer(kind=int32)    :: x
    integer(kind=int32)    :: y
    real(kind=real32)      :: energy
    real(kind=real32)      :: ev_weight
    integer(kind=int16)    :: rawx
    integer(kind=int16)    :: rawy
    real(kind=real64)      :: subx
    real(kind=real64)      :: suby
!    integer(kind=int16)    :: subx
!    integer(kind=int16)    :: suby
    integer(kind=int16)    :: pha
    integer(kind=int16)    :: pat_typ
    integer(kind=int8)     :: pat_inf
    integer(kind=int8)     :: ccdnr
    integer(kind=int32)    :: flag   
  end  type evrow_type

  type evlist_type
    !sequence    
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'EVENTS'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    type(twcs_hdr_type)              :: twcs_hdr
    type(evscalehdr_type)            :: scalehdr
    integer(kind=int32)              :: ncol=18
    character(len=16), dimension(18) :: cname = [ character(len=16) :: 'TIME','OTS','FRACSEC','RA','DEC','X','Y','ENERGY','EV_WEIGHT',&
                                                'RAWX','RAWY','SUBX','SUBY','PHA','PAT_TYP','PAT_INF','CCDNR','FLAG' ]

    character(len=4) , dimension(18) :: cform = [ character(len=4)  :: &
                                                  '1D','1J','1J','1J','1J','1J','1J','1E','1E','1I','1I','1B','1B','1I','1I','1B','1B','1J']



    character(len=16), dimension(18) :: cunit = [ character(len=16)  :: &
                                                  'sec','sec','tbd','deg','deg','0.05 arcsec','0.05 arcsec','eV',' ',' ',' ',' ', &
                                                  ' ',' ',' ',' ', ' ',' ']

    type(evrow_type),  pointer, dimension(:)  :: data  => NULL()
  end type evlist_type

!----------------------------------------------------------------------
! subset of event list file used in evatt task
! for columns RA, DEC, SUBX, SUBY automatic data scaling is used
! physical FITS data formats are int32 for RA, DEC, int8 for subx, suby
! detx,dety 
!----------------------------------------------------------------------
 
  type evatt_row_type
    !sequence
    real(kind=real64)     :: time
    real(kind=real64)     :: ra
    real(kind=real64)     :: dec
    integer(kind=int16)   :: rawx
    integer(kind=int16)   :: rawy
    real(kind=real64)     :: subx
    real(kind=real64)     :: suby
!    integer(kind=int16)   :: subx    
!    integer(kind=int16)   :: suby    
    integer(kind=int32)   :: flag    
!    real(kind=real64)  pointer, dimension(:)    :: detx => NULL()
!    real(kind=real64)  pointer, dimension(:)    :: dety => NULL()
  end type evatt_row_type

  type evatt_list_type
    character(len=clong)                        :: filename   
    integer(kind=int32)                         :: lun        
    character(len=16)                           :: extname = 'EVENTS'
    type(shapehdr_type)                         :: shapehdr
    type(stdhdr_type)                           :: stdhdr
    type(evscalehdr_type)                       :: scalehdr
    integer(kind=int32)                         :: ncol=8
    type(evatt_row_type), pointer, dimension(:) :: data => NULL()
    character(len=16), dimension(8) :: cname = [ character(len=16) :: 'TIME','RA','DEC','RAWX','RAWY','SUBX','SUBY','FLAG']
  end type evatt_list_type
   


  type hkrow_type
  ! this is still incomplete, since HK format is not defined yet!
    !sequence
    real(kind=real64)   ::  time
    integer(kind=int16) ::  fwheel   
    integer(kind=int16) ::  detstat       
  end type hkrow_type


  type hk_type
    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'HK'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=3
    character(len=16), dimension(3)  :: cname = [ character(len=16) :: 'TIME','FWHEEL','DETSTAT' ]
    character(len=4) , dimension(3) :: cform = [ character(len=4)  :: '1D','1I','1I' ]
    character(len=16), dimension(3) :: cunit = [ character(len=16)  :: 'sec','','' ]
    type(hkrow_type), pointer, dimension(:)  :: data  => NULL()

  end type hk_type


  type data_block
    character(len=6)                          :: id                          
    real(kind=real64)                         :: start,stop
    integer(kind=int32)                       :: fw
    integer(kind=int32)                       :: tel
    integer(kind=int32)                       :: camstat
  end type data_block


  type obs_type
    !sequence
    type(obi_type),pointer,dimension(:)       :: obis
    character(len=1)                          :: type
    real(kind=real64)                         :: start,stop
    integer(kind=int32)                       :: obsid,n_obis,index_obs
  end type obs_type


  type obi_type
    !sequence
    integer(kind=int32)                       :: id,n,pinum
    real(kind=real64)                         :: start,stop,ra,dec,fieldlon,fieldlat
    character(len=1)                          :: type
    character(len=3)                          :: obstat
    character(len=7)                          :: fw,detstat
    character(len=20)                         :: target
    integer(kind=int32),dimension(7)          :: ifw
    integer(kind=int32),dimension(7)          :: idetstat  
  end type obi_type


  type gtirow_type
    !sequence
    real(kind=real64)   ::  start
    real(kind=real64)   ::  stop    
  end type gtirow_type


  type gti_type
    !sequence
    character(len=clong)             :: filename
    character(len=16)                :: extname = 'STDGTI'
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=2
    character(len=16), dimension(2)  :: cname = [ character(len=16) :: 'START','STOP' ]
    character(len=4) , dimension(2)  :: cform = [ character(len=4)  :: '1D','1D' ]
    character(len=16), dimension(2)  :: cunit=[ character(len=16)  :: 'sec','sec' ]
    integer(kind=int32)              :: nrows
    type(gtirow_type), pointer, dimension(:)  :: data  => NULL()
  end type gti_type


  type gtitab_type
    ! As GTI_TYPE, but omitting header. 
    ! Needed for large array allocations in TELGTI
    !sequence
    character(len=80)                :: filename
    integer(kind=int32)              :: nrows
    type(gtirow_type), pointer, dimension(:)  :: data  => NULL()
  end type gtitab_type


  type skymapsrow_type
    !sequence
    integer(kind=int32) :: smapnr
    real(kind=real64)   :: ra_min
    real(kind=real64)   :: ra_max
    real(kind=real64)   :: de_min
    real(kind=real64)   :: de_max 
    real(kind=real64)   :: ra_cen
    real(kind=real64)   :: de_cen
    real(kind=real64)   :: elon_cen
    real(kind=real64)   :: elat_cen
    real(kind=real64)   :: glon_cen
    real(kind=real64)   :: glat_cen
    real(kind=real64)   :: x_min
    real(kind=real64)   :: y_min
    integer(kind=int16) :: owner
  end type skymapsrow_type

 type skymaps_type
    !sequence
    character(len=clong)                           :: filename
    character(len=16)                              :: extname='SMAPS'   
    type(shapehdr_type)                            :: shapehdr
    type(stdhdr_type)                              :: stdhdr
    type(wcs_imahdr_type), pointer, dimension(:)   :: wcshdr => NULL()
    integer(kind=int32)                            :: ncol=14
    character(len=16), dimension(14) :: cname = [ character(len=16) :: 'SRVMAP','RA_MIN','RA_MAX','DE_MIN','DE_MAX', &
                                                                       'RA_CEN','DE_CEN','ELON_CEN','ELAT_CEN','GLON_CEN', &
                                                                        'GLAT_CEN','X_MIN','Y_MIN','OWNER' ]
    character(len=4) , dimension(14) :: cform = [ character(len=4)  :: '1J','1D','1D','1D','1D','1D','1D','1D','1D','1D',&
                                                                       '1D','1D','1D','1I' ]    
    character(len=16), dimension(14) :: cunit=[ character(len=16)   :: ' ','deg','deg','deg','deg','deg','deg','deg','deg',&
                                                                       'deg','deg','arcmin','arcmin',' ']
    type(skymapsrow_type), pointer, dimension(:)  :: data  => NULL()
 end type skymaps_type

!!---------------------------------------------------
!   obsolete:
!
!  type splanrow_type
!    !sequence
!    real(kind=real64)   :: time
!    integer(kind=int32) :: obsid
!    integer(kind=int32) :: obi
!    real(kind=real64)   :: rot_elon
!    real(kind=real64)   :: rot_elat
!    real(kind=real64)   :: pnt_elat
!    real(kind=real64)   :: drt_elon
!    real(kind=real64)   :: dpt_scan
!    real(kind=real64)   :: tstart
!    real(kind=real64)   :: tstop
!    character(len=7)    :: fwheel  
!    character(len=7)    :: detstat  
!    character(len=20)   :: remarks  
!  end type splanrow_type
!
!  type splan_type
!    !sequence
!    character(len=clong)             :: filename
!    character(len=16)                :: extname='PLAN'
!    type(shapehdr_type)              :: shapehdr
!    type(stdhdr_type)                :: stdhdr
!    integer(kind=int32)              :: ncol=13 
!    character(len=16), dimension(13) :: cname = [ character(len=16) ::  'TIME','obsid','OBI','ROT_ELON','ROT_ELAT', &
!                                        'PNT_ELAT','DRT_ELON','DPT_SCAN','TSTART','TSTOP','FWHEEL','DETSTAT','REMARKS']      
! 
!    character(len=4) , dimension(13) :: cform=[ character(len=4)  :: '1D','1J','1J','1D','1D','1D','1D','1D','1D','1D', &
!                                                                     '7A','7A','13A']
!    character(len=16), dimension(13) :: cunit=[ character(len=16) :: 'sec',' ',' ','deg','deg','deg','deg/day',        & 
!                                                                     'deg/sec','sec','sec',' ',' ',' ' ]  
! 
!    type(splanrow_type), pointer, dimension(:)  :: data  => NULL()
!  end type splan_type
!
!   /obsolete
!---------------------------------------------------




 

!-----------------------------------------------------------------------------------
! type planrow_type, adapted to new file layout as suggested to IKI on 28-Mar-2013
!-----------------------------------------------------------------------------------

  type planrow_old_type
    !sequence
    !general
    character(len=1)    :: ptype   
    character(len=3)    :: obstat 
    integer(kind=int32) :: obsid
    integer(kind=int32) :: obi
     real(kind=real64)   :: tstart
    real(kind=real64)   :: tstop
    character(len=7)    :: fwheel  
    character(len=7)    :: detstat  
    character(len=30)   :: remarks  
    ! pointing:
    character(len=30)   :: target
    integer(kind=int32) :: pinum 
    real(kind=real64)   :: ra_pnt
    real(kind=real64)   :: dec_pnt
    real(kind=real64)   :: fieldlon
    real(kind=real64)   :: fieldlat
    real(kind=real64)   :: vis_start
    real(kind=real64)   :: vis_stop
    real(kind=real64)   :: vis_dur
    real(kind=real64)   :: tact
    real(kind=real64)   :: texp
    character(len=20)   :: tcrit
    ! survey:
    real(kind=real64)   :: scp_elon
    real(kind=real64)   :: scp_elat
    real(kind=real64)   :: sca_start
    real(kind=real64)   :: rot_start
    real(kind=real64)   :: sca_omega
    real(kind=real64)   :: rot_omega
  end type planrow_old_type


  type plan_old_type
    !sequence
    character(len=64)                :: filename
    character(len=16)                :: extname='PLAN'
    integer(kind=int32)              :: pdate
    real(kind=real64)                :: pstart,pstop,vdate,vstop
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=27 
    character(len=16), dimension(27) :: cname = [ character(len=16) :: 'PTYPE','OBSTAT','OBSID','OBI','TSTART','TSTOP',&
                                                                       'FWHEEL','DETSTAT','REMARKS','TARGET','PINUM', &
                                                                       'RA_PNT','DEC_PNT','FIELDLON','FIELDLAT','VIS_START',&
                                                                       'VIS_STOP','VIS_DUR', 'TACT','TEXP','TCRIT',&
                                                                       'SCP_ELON','SCP_ELAT','SCA_START', &
                                                                       'ROT_START','SCA_OMEGA','ROT_OMEGA']
 
 
    character(len=4) , dimension(27) :: cform=[ character(len=4)  :: '1A','3A','1J','1J','1D','1D', &
                                                                     '7A','7A','30A','30A','1J', &
                                                                     '1D','1D','1D', '1D','1D', &
                                                                     '1D','1D', '1D','1D','20A', &
                                                                     '1D','1D', '1D', &
                                                                     '1D','1D','1D'] 

    character(len=16), dimension(27) :: cunit=[ character(len=16) :: ' ',' ',' ',' ','sec','sec', &
                                                                     ' ',' ',' ',' ',' ',  &
                                                                     'deg','deg','deg','deg','sec', &
                                                                     'sec','sec','sec','sec',' ', &
                                                                     'deg','deg','deg', &
                                                                     'deg','deg/day','deg/sec']

 
    type(planrow_old_type), pointer, dimension(:)  :: data  => NULL()
  end type plan_old_type

!-------------------------------------------------------------------------------------------

  !----------------------------------------------------------------------------------------
  ! New version of PLAN files
  ! as described in eRO-HS-MPL-01 v0.9, 04/2017
  !----------------------------------------------------------------------------------------



  type planrow_type
    !sequence
    !general
    character(len=1)    :: ptype   
    character(len=3)    :: obstat 
    integer(kind=int32) :: obsid
    integer(kind=int32) :: obi
    real(kind=real64)   :: tstart
    character(len=20)   :: tstart_utc
    real(kind=real64)   :: tstop
    character(len=20)   :: tstop_utc  
    character(len=7)    :: fwheel  
    character(len=7)    :: camstat  
    character(len=30)   :: erocom  
    character(len=50)   :: remarks
     ! survey:
    real(kind=real64)   :: scp_elon
    real(kind=real64)   :: scp_elat
    real(kind=real64)   :: scp_start
    real(kind=real64)   :: scp_omega
    real(kind=real64)   :: rot_start
    real(kind=real64)   :: rot_omega 
    ! IKI/NPOL parameters:
    real(kind=real64)   :: scr_alpha
    real(kind=real64)   :: scr_delta
    real(kind=real64)   :: sca_start
    real(kind=real64)   :: sca_omega 
    real(kind=real64)   :: ra_p
    real(kind=real64)   :: dec_p
    real(kind=real64)   :: ra_z0
    real(kind=real64)   :: ra_zk
    real(kind=real64)   :: dec_z0
    real(kind=real64)   :: dec_zk 

    ! pointing:
    character(len=30)   :: target
    integer(kind=int32) :: pinum 
    real(kind=real64)   :: ra_pnt
    real(kind=real64)   :: dec_pnt
    real(kind=real64)   :: fieldlon
    real(kind=real64)   :: fieldlat
    real(kind=real64)   :: vis_start
    real(kind=real64)   :: vis_stop
    real(kind=real64)   :: tact
    real(kind=real64)   :: texp
    character(len=20)   :: tcrit
    real(kind=real32)   :: prio
    character(len=11)   :: experiment 
  end type planrow_type


  type plan_type
    !sequence
    character(len=64)                :: filename
    character(len=16)                :: extname='PLAN'
    integer(kind=int32)              :: pdate
    real(kind=real64)                :: pstart,pstop,vdate,vstop
    type(shapehdr_type)              :: shapehdr
    type(stdhdr_type)                :: stdhdr
    integer(kind=int32)              :: ncol=41 
    character(len=16), dimension(41) :: cname = [ character(len=16) :: 'PTYPE','OBSTAT','OBSID','OBI','TSTART', &
                                                                       'TSTART_UTC','TSTOP','TSTOP_UTC', &
                                                                       'FWHEEL','CAMSTAT','EROCOM','REMARKS', &
                                                                       'SCP_ELON','SCP_ELAT','SCP_START', &
                                                                       'SCP_OMEGA','ROT_START','ROT_OMEGA', &
                                                                       'SCR_ALPHA','SCR_DELTA','SCA_START','SCA_OMEGA', &
                                                                       'RA_P','DEC_P','RA_Z0','RA_ZK','DEC_Z0','DEC_ZK', &
                                                                       'TARGET','PINUM', &
                                                                       'RA_PNT','DEC_PNT','FIELDLON','FIELDLAT','VIS_START', &
                                                                       'VIS_STOP', 'TACT','TEXP','TCRIT','PRIO','EXPERIMENT']
 
 
    character(len=4) , dimension(41) :: cform=[ character(len=4)  :: '1A','3A','1J','1J','1D', &
                                                                     '20A','1D','20A', &
                                                                     '7A','7A','30A','50A', &
                                                                     '1D','1D','1D', &
                                                                     '1D','1D','1D', &
                                                                     '1D','1D', '1D','1D', &
                                                                     '1D','1D', '1D','1D','1D','1D', &
                                                                     '30A','1J', &
                                                                     '1D','1D', '1D','1D', '1D', &
                                                                     '1D','1D','1D','20A','1E','11A'] 

    character(len=16), dimension(41) :: cunit=[ character(len=16) :: ' ',' ',' ',' ','sec', &
                                                                     ' ','sec',' ',  &
                                                                     ' ',' ',' ',' ', &
                                                                     'deg','deg','deg', &
                                                                     'deg/d','deg','deg/s', & 
                                                                     'deg','deg','deg','deg/s', &
                                                                     'deg','deg','deg','deg','deg','deg', &
                                                                     ' ',' ', &
                                                                     'deg','deg','deg','deg','s', &
                                                                     's','s','s',' ',' ',' ']

 
    type(planrow_type), pointer, dimension(:)  :: data  => NULL()
  end type plan_type

!---------------------------------------------------------------------------------------------------------------------------



  type evraw_type
    !sequence
    real(kind=real64)   ::  time
    integer(kind=int32) ::  frame
    integer(kind=int16) ::  rawx
    integer(kind=int16) ::  rawy
  end type evraw_type


   type evrawlist_type
    !sequence
    character(len=clong)                     :: filename
    character(len=16)                        :: extname='EVENTS'
    type(shapehdr_type)                      :: shapehdr
    type(stdhdr_type)                        :: stdhdr
    integer(kind=int32)                      :: ncol=4
    character(len=16), dimension(4)          :: cname=[ character(len=16) :: 'TIME','FRAME','RAWX','RAWY']
    character(len=4), dimension(4)           :: cform=[ character(len=4) ::  '1D','1J','1I','1I']
    character(len=16), dimension(4)          :: cunit=[ character(len=16) ::  's',' ','pixel','pixel']
    type(evraw_type), pointer, dimension(:)  :: data  => NULL()
  end type evrawlist_type


  type attprep_param_type
    !sequence
    integer(kind=int32)                     ::  lun
    character(len=clong)                    ::  sed1file
    character(len=clong)                    ::  sed2file
    character(len=clong)                    ::  bokz_file
    character(len=clong)                    ::  gyro_file
    character(len=clong)                    ::  attfile
    logical(kind=log32)                     ::  sed1flag
    logical(kind=log32)                     ::  sed2flag  
    logical(kind=log32)                     ::  bokz_flag  
    logical(kind=log32)                     ::  gyro_flag
   end type attprep_param_type  




  type telgti_param_type
    !sequence
    integer(kind=int32)                     ::  lun
    character(len=clong)                    ::  planroot
    character(len=clong)                    ::  pathname  
    character(len=clong)                    ::  eventfile
    character(len=clong)                    ::  attfile
    character(len=clong),dimension(2)       ::  hkfile 
    integer(kind=int32)                     ::  n_hkfiles 
    ! used by telselect:
    character(len=clong)                    ::  eroday    
   end type telgti_param_type  


  type telatt_param_type
    !sequence
    integer(kind=int32)                     ::  lun
    character(len=clong)                    ::  eventfile
    character(len=clong)                    ::  attfile
    character(len=clong)                    ::  gtifile
   end type telatt_param_type  




  type evatt_param_type
    !sequence
    integer(kind=int32)                     ::  lun
    character(len=clong)                    ::  eventfile
    character(len=clong)                    ::  attfile
    character(len=clong)                    ::  gtifile
   end type evatt_param_type  




!--------------------------------------------------------------------------------

  contains


  subroutine fget_evatt_list(filename,ext,table,ierr)  
  !-----------------------------------------------------------------
  ! Open event list for reading and writing, read subset of columns
  ! (defined in evatt_row_type), store file unit in table%lun
  !-----------------------------------------------------------------

    implicit none

    type(evatt_list_type)            :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,hdutype,rwmode!,blocksize
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_EVATT_LIST'
    character(len=80)                :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1     ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.

    
    call FTGIOU(unit,status)

    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      table%lun=unit
      call FTNOPN(unit,filename,rwmode,status)
       if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,errstr)
      else  
        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_stdhdr(unit,table%stdhdr,status)
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            fstat=0
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            icol=1

            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,table%data(i1:i2)%time,fnull,status) 
            fstat = max(fstat,status) 
            icol=icol+1

            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,table%data(i1:i2)%ra,fnull,status) 
            fstat = max(fstat,status) 
            icol=icol+1

            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,table%data(i1:i2)%dec,fnull,status) 
            fstat = max(fstat,status) 
            icol=icol+1

            if (colnum(icol) > 0) call  FTGCVI(unit,colnum(icol),i1,1,rowbuf,i2null,table%data(i1:i2)%rawx,fnull,status) 
            fstat = max(fstat,status) 
            icol=icol+1

            if (colnum(icol) > 0) call  FTGCVI(unit,colnum(icol),i1,1,rowbuf,i2null,table%data(i1:i2)%rawy,fnull,status) 
            fstat = max(fstat,status) 
            icol=icol+1
            
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,table%data(i1:i2)%subx,fnull,status) 
            fstat = max(fstat,status) 
            icol=icol+1

            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,table%data(i1:i2)%suby,fnull,status) 
            fstat = max(fstat,status) 
            icol=icol+1

            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,table%data(i1:i2)%flag,fnull,status) 
            fstat = max(fstat,status) 
            



            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
    endif
    
    ! leave file open
    !    call FTCLOS(unit,status)
    !    call FTFIOU(unit,status)   


  end  subroutine fget_evatt_list


!--------------------------------------------------------------------------------


  subroutine fget_hk(filename,ext,table,ierr) 

  !--------------------------------------------------------------- 
  ! read only columns TIME , DETSTAT from this this HK extension
  !---------------------------------------------------------------

    implicit none

    type(hk_type)                    :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_HK'
    character(len=80)                :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.

 
   call FTGIOU(unit,status)

     if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
       if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,errstr)
      else  
        call fread_stdhdr(unit,table%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            if (colnum(1) > 0) call  FTGCVD(unit,colnum(1),i1,1,rowbuf,r8null,table%data(i1:i2)%time,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(3) > 0) call  FTGCVI(unit,colnum(2),i1,1,rowbuf,i2null,table%data(i1:i2)%detstat,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(2) > 0) call  FTGCVI(unit,colnum(3),i1,1,rowbuf,i2null,table%data(i1:i2)%detstat,fnull,status) 
            fstat = max(fstat,status) 
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    



  end  subroutine fget_hk

!--------------------------------------------------------------------------------


  subroutine fget_hk1(filename,ext,table,ierr) 

  !--------------------------------------------------------------- 
  ! read only columns TIME , DETSTAT from this this HK extension
  !---------------------------------------------------------------

    implicit none

    type(hk_type)                    :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_HK1'
    character(len=80)                :: errstr



    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.

 
   call FTGIOU(unit,status)

     if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
       if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,errstr)
      else  
        call fread_stdhdr(unit,table%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            if (colnum(1) > 0) call  FTGCVD(unit,colnum(1),i1,1,rowbuf,r8null,table%data(i1:i2)%time,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(3) > 0) call  FTGCVI(unit,colnum(3),i1,1,rowbuf,i2null,table%data(i1:i2)%detstat,fnull,status) 
            fstat = max(fstat,status) 
!            if (colnum(2) > 0) call  FTGCVI(unit,colnum(3),i1,1,rowbuf,i2null,table%data(i1:i2)%detstat,fnull,status) 
!            fstat = max(fstat,status) 
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    



  end  subroutine fget_hk1

!--------------------------------------------------------------------------------


  subroutine fget_hk2(filename,ext,table,ierr)  

  !--------------------------------------------------------------- 
  ! read only columns TIME , FWHEEL from this this HK extension
  !---------------------------------------------------------------

    implicit none

    type(hk_type)                    :: table
    character(len=*), intent(in)     :: filename
    character(len=*), intent(in)     :: ext
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(table%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_HK2'
    character(len=80)                :: errstr



    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.

 
   call FTGIOU(unit,status)

     if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
       if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,errstr)
      else  
        call fread_stdhdr(unit,table%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,ext,0,status)
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//ext
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_shapehdr(unit,table%shapehdr,status)
          nrows=table%shapehdr%naxis2
          allocate(table%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,table%ncol
            call FTGCNO(unit,.false.,table%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//table%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            if (colnum(1) > 0) call  FTGCVD(unit,colnum(1),i1,1,rowbuf,r8null,table%data(i1:i2)%time,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(2) > 0) call  FTGCVI(unit,colnum(2),i1,1,rowbuf,i2null,table%data(i1:i2)%fwheel,fnull,status) 
            fstat = max(fstat,status) 
!            if (colnum(2) > 0) call  FTGCVI(unit,colnum(3),i1,1,rowbuf,i2null,table%data(i1:i2)%detstat,fnull,status) 
!           fstat = max(fstat,status) 
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    



  end  subroutine fget_hk2
 
  !------------------------------------------------------------------------------------------



  subroutine fget_telgti(filename,gti,ierr)  

    implicit none

    type(gti_type)                           :: gti
    character(len=*), intent(in)             :: filename
    integer(kind=int32), intent(out)         :: ierr
    logical                                  :: fnull
    integer(kind=int32)                      :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)                      :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(gti%ncol)  :: colnum 
    character(len=*),parameter               :: rtname = 'FGET_TELGTI'
    character(len=80)                        :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
 
    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,'error opening event list')
      else  
        call fread_stdhdr(unit,gti%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,gti%extname,0,status)
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//gti%extname 
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_shapehdr(unit,gti%shapehdr,status)
          nrows=gti%shapehdr%naxis2
          allocate(gti%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,gti%ncol
            call FTGCNO(unit,.false.,gti%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//gti%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          irow=0
          do while (nrows > irow)      
            
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf           
            fstat=0  
            if (colnum(1) > 0) call  FTGCVD(unit,colnum(1),i1,1,rowbuf,r8null,gti%data(i1:i2)%start,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(2) > 0) call  FTGCVD(unit,colnum(2),i1,1,rowbuf,r8null,gti%data(i1:i2)%stop,fnull,status) 
            fstat = max(fstat,status) 
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
       
  end subroutine fget_telgti
  


!--------------------------------------------------------------------------------
   

  subroutine fget_skymaps(filename,maps,ierr)  

    implicit none

    type(skymaps_type)                       :: maps
    character(len=*), intent(in)             :: filename
    integer(kind=int32), intent(out)         :: ierr
    logical                                  :: fnull
    integer(kind=int32)                      :: maxbuf,irow,rowbuf,i1,i2,icol,nrows
    integer(kind=int32)                      :: status,fstat,unit,blocksize,hdutype,rwmode
!    integer(kind=int32),dimension(maps%ncol) :: colnum 
    integer(kind=int32),dimension(14) :: colnum 
    character(len=*),parameter               :: rtname = 'FGET_SKYMAPS'
    character(len=80)                        :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
    

    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,'error opening skymaps  file')
      else  
        call fread_stdhdr(unit,maps%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,maps%extname,0,status)
        if (status > 0) then 
          ! go to nhdu = 2:
          status=0
          call FTMAHD(unit,2, hdutype,status)
        endif
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//maps%extname 
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_shapehdr(unit,maps%shapehdr,status)
          nrows=maps%shapehdr%naxis2
          allocate(maps%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,maps%ncol
            call FTGCNO(unit,.false.,maps%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//maps%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)    
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            fstat=0      
            if (colnum(1) > 0) call  FTGCVJ(unit,colnum(1),i1,1,rowbuf,i4null,maps%data(i1:i2)%smapnr,fnull,status)
            fstat = max(fstat,status)
            if (colnum(2) > 0) call  FTGCVD(unit,colnum(2),i1,1,rowbuf,r8null,maps%data(i1:i2)%ra_min,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(3) > 0) call  FTGCVD(unit,colnum(3),i1,1,rowbuf,r8null,maps%data(i1:i2)%ra_max,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(4) > 0) call  FTGCVD(unit,colnum(4),i1,1,rowbuf,r8null,maps%data(i1:i2)%de_min,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(5) > 0) call  FTGCVD(unit,colnum(5),i1,1,rowbuf,r8null,maps%data(i1:i2)%de_max,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(6) > 0) call  FTGCVD(unit,colnum(6),i1,1,rowbuf,r8null,maps%data(i1:i2)%ra_cen,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(7) > 0) call  FTGCVD(unit,colnum(7),i1,1,rowbuf,r8null,maps%data(i1:i2)%de_cen,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(8) > 0) call  FTGCVD(unit,colnum(8),i1,1,rowbuf,r8null,maps%data(i1:i2)%elon_cen,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(9) > 0) call  FTGCVD(unit,colnum(9),i1,1,rowbuf,r8null,maps%data(i1:i2)%elat_cen,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(10) > 0) call  FTGCVD(unit,colnum(10),i1,1,rowbuf,r8null,maps%data(i1:i2)%glon_cen,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(11) > 0) call  FTGCVD(unit,colnum(11),i1,1,rowbuf,r8null,maps%data(i1:i2)%glat_cen,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(12) > 0) call  FTGCVD(unit,colnum(12),i1,1,rowbuf,r8null,maps%data(i1:i2)%x_min,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(13) > 0) call  FTGCVD(unit,colnum(13),i1,1,rowbuf,r8null,maps%data(i1:i2)%y_min,fnull,status) 
            fstat = max(fstat,status)
            if (colnum(14) > 0) call  FTGCVI(unit,colnum(14),i1,1,rowbuf,i2null,maps%data(i1:i2)%owner,fnull,status)
            fstat = max(fstat,status)
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
          ! allocate array of wcs headers
          allocate(maps%wcshdr(nrows))
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    

  end subroutine fget_skymaps




  subroutine fget_plan(filename,plan,ierr)  
  
  !------------------------------------------------------------------------------
  ! Read plan file in common 'survey' and 'pointing' plan file format
  ! defined in 
  ! http://www2011.mpe.mpg.de/erosita/internal/SASS-devel/doc/planning-doc.html 
  ! from  (2012-06-18)
  !------------------------------------------------------------------------------

    implicit none

    type(plan_type)                  :: plan
    character(len=*), intent(in)     :: filename
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(plan%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_PLAN'
    character(len=80)                :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
    


    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,'error opening plan file')
      else  
        call fread_stdhdr(unit,plan%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,plan%extname,0,status)
        if (status > 0) then 
          ! go to nhdu = 2:
          status=0
          call FTMAHD(unit,2, hdutype,status)
        endif
        if (status /= 0  .or. hdutype /= 2) then
          ierr = 3
          errstr = 'error moving to extension '//plan%extname 
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_shapehdr(unit,plan%shapehdr,status)
          nrows=plan%shapehdr%naxis2
          allocate(plan%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,plan%ncol
            call FTGCNO(unit,.false.,plan%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//plan%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            fstat = 0    
            icol=1      
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%ptype,fnull,status) !PTYPE
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%obstat,fnull,status) !OBSTAT 
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,plan%data(i1:i2)%obsid,fnull,status) !OBSID
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,plan%data(i1:i2)%obi,fnull,status)   !OBI
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%tstart,fnull,status) !TSTART
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%tstart_utc,fnull,status)  !TSTART_UTC
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%tstop,fnull,status) !TSTOP
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%tstop_utc,fnull,status) !TSTOP_UTC
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%fwheel,fnull,status) !FWHEEL
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%camstat,fnull,status) !CAMSTAT
            fstat = max(fstat,status) 
            icol=icol+1  
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%erocom,fnull,status) !EROCOM
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%remarks,fnull,status) !REMARKS
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%scp_elon,fnull,status) !SCP_ELON
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%scp_elat,fnull,status) !SCP_ELAT
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%scp_start,fnull,status) !SCP_START
            fstat = max(fstat,status)  
            icol=icol+1

            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%scp_omega,fnull,status) !SCP_OMEGA
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%rot_start,fnull,status) !ROT_START
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%rot_omega,fnull,status) !ROT_OMEGA
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%scr_alpha,fnull,status) !SCR_ALPHA
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%scr_delta,fnull,status) !SCR_DELTA
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%sca_start,fnull,status) !SCA_START
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%sca_omega,fnull,status) !SCA_OMEGA
            fstat = max(fstat,status)  
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%ra_p,fnull,status) !RA_P
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%dec_p,fnull,status) !DEC_P
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%ra_z0,fnull,status) !RA_Z0
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%ra_zk,fnull,status) !RA_ZK
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%dec_z0,fnull,status) !DEC_Z0
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%dec_zk,fnull,status) !DEC_ZK
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%target,fnull,status) !TARGET
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,plan%data(i1:i2)%pinum,fnull,status) !PINUM
            fstat = max(fstat,status) 
            icol=icol+1

            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%ra_pnt,fnull,status) !RA_PNT
            fstat = max(fstat,status)  
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%dec_pnt,fnull,status) !DEC_PNT
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%fieldlon,fnull,status) !FIELDLON
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%fieldlat,fnull,status) !FIELDLAT
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%vis_start,fnull,status) !VIS_START
            fstat = max(fstat,status)  
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%vis_stop,fnull,status) !VIS_STOP
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%tact,fnull,status) !TACT
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,plan%data(i1:i2)%texp,fnull,status) !TEXP
            fstat = max(fstat,status) 
            icol=icol+1 
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%tcrit,fnull,status) !TCRIT
            fstat = max(fstat,status) 
            icol=icol+1
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,plan%data(i1:i2)%prio,fnull,status) !PRIO
            fstat = max(fstat,status)
             icol=icol+1
            if (colnum(icol) > 0) call  FTGCVS(unit,colnum(icol),i1,1,rowbuf,cnull,plan%data(i1:i2)%experiment,fnull,status) !EXPERIMENT                                  
            fstat = max(fstat,status)

            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    


  end subroutine fget_plan

!--------------------------------------------------------------------------------


  subroutine eval_plan_name(filename,pdate,pstart,pstop,ierr)

  !-------------------------------------------------------------- 
  ! Extract DATE and START, STOP values from PLAN filename.
  ! Required format of PLAN filename:
  ! e.g. F_PLAN_dddddd_ttttttttt_ttttttttt.fits
  ! dddddd : DATE, tttttttt : START / STOP in SCC seconds
  !-------------------------------------------------------------- 
 
    character(len=*), intent(in)                  :: filename
    integer(kind=int32), intent(out)              :: pdate , ierr  
    real(kind=real64),intent(out)                 :: pstart, pstop
    integer(kind=int32)                           :: status!i, 
    character(len=*),parameter                    :: numset = '0123456789'
    character(len=*),parameter                    :: rtname = 'EVAL_PLAN_NAME'

   

    ierr=0

       
    ! check filename:
       status = verify(filename(8:13),numset)
       if (status == 0)  then
         read(filename(8:13),*) pdate
        else
         ierr =3 
       endif
       status = verify(filename(15:23),numset)
       if (status == 0)  then
         read(filename(15:23),*) pstart
       else
         ierr = 3 
       endif
       status = verify(filename(25:33),numset)
       if (status == 0)  then
         read(filename(25:33),*) pstop
        else
         ierr = 3 
       endif


       if (ierr > 0) then
         call wrerr(rtname, ierr, 'Error in format of filename '//filename ) 
       endif

       


  end subroutine eval_plan_name


!--------------------------------------------------------------------------------

  subroutine fget_evraw(filename,evlist,ierr)  

    implicit none

    type(evrawlist_type)             :: evlist
    character(len=*), intent(in)     :: filename
    integer(kind=int32), intent(out) :: ierr
    logical                          :: fnull
    integer(kind=int32)              :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)              :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(evlist%ncol) :: colnum 
    character(len=*),parameter       :: rtname = 'FGET_EVRAW'
    character(len=80)                :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
        
    
    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,'error opening event list')
      else  
        call fread_stdhdr(unit,evlist%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,evlist%extname,0,status)
        if (status /= 0) then
          ierr = 3
          errstr = 'error moving to extension '//evlist%extname 
          call WRERR(rtname,ierr,errstr)
        else 
    

          call fread_shapehdr(unit,evlist%shapehdr,status)
          nrows=evlist%shapehdr%naxis2
          allocate(evlist%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,evlist%ncol
            call FTGCNO(unit,.false.,evlist%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol)=-1
              ierr=2
              errstr = 'column '//evlist%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
          enddo
          do while (nrows > irow)  
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            if (colnum(1) > 0) call  FTGCVD(unit,colnum(1),i1,1,rowbuf,r8null,evlist%data(i1:i2)%time,fnull,status) 
            fstat = max(fstat,status) 
            if (colnum(2) > 0) call  FTGCVJ(unit,colnum(2),i1,1,rowbuf,i4null,evlist%data(i1:i2)%frame,fnull,status)
            fstat = max(fstat,status)
            if (colnum(3) > 0) call  FTGCVI(unit,colnum(3),i1,1,rowbuf,i2null,evlist%data(i1:i2)%rawx,fnull,status)
            fstat = max(fstat,status)
            if (colnum(4) > 0) call  FTGCVI(unit,colnum(4),i1,1,rowbuf,i2null,evlist%data(i1:i2)%rawy,fnull,status)
            fstat = max(fstat,status)
             if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    
    

  end subroutine fget_evraw






!--------------------------------------------------------------------------------

  subroutine fget_evlist(filename,evlist,ierr)  

    implicit none

    type(evlist_type)                   :: evlist
    character(len=*), intent(in)        :: filename
    integer(kind=int32), intent(out)    :: ierr
    logical                             :: fnull
    integer(kind=int32)                 :: nrows,maxbuf,irow,rowbuf,i1,i2,icol
    integer(kind=int32)                 :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32),dimension(evlist%ncol) :: colnum    
    character(len=*),parameter          :: rtname = 'FGET_EVLIST'  
    character(len=80)                   :: errstr

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 0  ! read
    hdutype = 2    ! binary table
    fnull  = .false.
        
 
    call FTGIOU(unit,status)
    if (status /= 0) then
      ierr = 3
      call WRERR(rtname,ierr,'error getting logical unit')
    else  
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      if (status /= 0) then
        ierr = 3
        errstr = 'error opening file '//filename
        call WRERR(rtname,ierr,'error opening event list')
      else  
        call fread_stdhdr(unit,evlist%stdhdr,status)
        status=0
        call FTMNHD(unit,hdutype,evlist%extname,0,status)
        if (status /= 0) then
          ierr = 3
          errstr = 'error moving to extension '//evlist%extname 
          call WRERR(rtname,ierr,errstr)
        else 
          call fread_shapehdr(unit,evlist%shapehdr,status)
          nrows=evlist%shapehdr%naxis2
          allocate(evlist%data(nrows))
          call FTGRSZ(unit, maxbuf, status)
          irow=0
          ! get column numbers:
          status=0
          do icol=1,evlist%ncol
            call FTGCNO(unit,.false.,evlist%cname(icol),colnum(icol),status)
            if (status > 0) then 
              colnum(icol) = -1
              ierr=2
              errstr = 'column '//evlist%cname(icol)//' not found'
              call WRERR(rtname,ierr,errstr)
            endif
            status=0
          enddo
          do while (nrows > irow)  
            status=0
            rowbuf = min(maxbuf,nrows-irow)
            i1=irow+1
            i2=irow+rowbuf
            irow=irow+rowbuf
            icol=1
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,evlist%data(i1:i2)%time,fnull,status)
            fstat = max(fstat,status) 
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,evlist%data(i1:i2)%ots,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,evlist%data(i1:i2)%fracsec,fnull,status)
            fstat = max(fstat,status)          
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,evlist%data(i1:i2)%ra,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,evlist%data(i1:i2)%dec,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,evlist%data(i1:i2)%x,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,evlist%data(i1:i2)%y,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,evlist%data(i1:i2)%energy,fnull,status)
            fstat = max(fstat,status)
            if (colnum(icol) > 0) call  FTGCVE(unit,colnum(icol),i1,1,rowbuf,r4null,evlist%data(i1:i2)%ev_weight,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVI(unit,colnum(icol),i1,1,rowbuf,i2null,evlist%data(i1:i2)%rawx,fnull,status)
            fstat = max(fstat,status) 
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVI(unit,colnum(icol),i1,1,rowbuf,i2null,evlist%data(i1:i2)%rawy,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,evlist%data(i1:i2)%subx,fnull,status)
!            if (colnum(icol) > 0) call  FTGCVI(unit,colnum(icol),i1,1,rowbuf,i2null,evlist%data(i1:i2)%subx,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   

            if (colnum(icol) > 0) call  FTGCVD(unit,colnum(icol),i1,1,rowbuf,r8null,evlist%data(i1:i2)%suby,fnull,status)
!            if (colnum(icol) > 0) call  FTGCVI(unit,colnum(icol),i1,1,rowbuf,i2null,evlist%data(i1:i2)%suby,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVI(unit,colnum(icol),i1,1,rowbuf,i2null,evlist%data(i1:i2)%pha,fnull,status)
            fstat = max(fstat,status) 
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVI(unit,colnum(icol),i1,1,rowbuf,i2null,evlist%data(i1:i2)%pat_typ,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVB(unit,colnum(icol),i1,1,rowbuf,i1null,evlist%data(i1:i2)%pat_inf,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVB(unit,colnum(icol),i1,1,rowbuf,i1null,evlist%data(i1:i2)%ccdnr,fnull,status)
            fstat = max(fstat,status)
            icol=icol+1   
            if (colnum(icol) > 0) call  FTGCVJ(unit,colnum(icol),i1,1,rowbuf,i4null,evlist%data(i1:i2)%flag,fnull,status)
            fstat = max(fstat,status)
            if (fstat /= 0) then
              ierr = 3
              errstr = 'Error reading values from table '//filename
              call WRERR(rtname,ierr,errstr)
            endif
          end do
        endif 
      endif
      call FTCLOS(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif
    

  end subroutine fget_evlist



!--------------------------------------------------------------------------------

  subroutine fput_telgti(filename,append,gti,ierr)

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(gti_type)                     :: gti
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!nhdu, 
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    integer(kind=int32), parameter     :: ncol=45
!    character(len=16), dimension(ncol) :: ttype,tunit
!    character(len=4), dimension(ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_TELGTI'
    character(len=80)                  :: errstr



    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=gti%shapehdr%naxis2
    
 
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,gti%stdhdr,status)
      endif
    endif

 

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
      call FTPHBN(unit,nrows,gti%ncol,gti%cname,gti%cform,gti%cunit,gti%extname,varidat,status)
      fstat=max(fstat,status)  
      

      call fwrite_stdhdr(unit,gti%stdhdr,status)
      fstat=max(fstat,status)


      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLD(unit,1,i1,1,rowbuf,gti%data(i1:i2)%start,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,2,i1,1,rowbuf,gti%data(i1:i2)%stop,status)
          fstat = max(fstat,status)

         enddo
      endif  

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
    
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  

  end   subroutine fput_telgti


!--------------------------------------------------------------------------------

  subroutine fput_evraw(filename,append,evlist,ierr)

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(evrawlist_type)                  :: evlist
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!nhdu, 
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    integer(kind=int32), parameter     :: ncol=45
    character(len=16), dimension(ncol) :: ttype,tunit
    character(len=4), dimension(ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_MLLIST'
    character(len=80)                  :: errstr



    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=evlist%shapehdr%naxis2
    
    
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis      = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,evlist%stdhdr,status)
      endif
    endif


    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTPHBN(unit,nrows,ncol,ttype,tform,tunit,evlist%extname,varidat,status)
      fstat=max(fstat,status)  
      
      call fwrite_stdhdr(unit,evlist%stdhdr,status)
      fstat=max(fstat,status)

      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLD(unit,1,i1,1,rowbuf,evlist%data(i1:i2)%time,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,2,i1,1,rowbuf,evlist%data(i1:i2)%frame,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,3,i1,1,rowbuf,evlist%data(i1:i2)%rawx,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,4,i1,1,rowbuf,evlist%data(i1:i2)%rawy,status)
          fstat = max(fstat,status)
         enddo
      endif  

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
    
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  


  end subroutine fput_evraw       
     


!--------------------------------------------------------------------------------

  subroutine fput_evlist(filename,append,evlist,ierr)

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(evlist_type)                  :: evlist
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!nhdu, 
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    integer(kind=int32), parameter     :: ncol=45
!    character(len=16), dimension(ncol) :: ttype,tunit
!    character(len=4), dimension(ncol)  :: tform
    character(len=*), parameter        :: rtname = 'FPUT_EVLIST'
    character(len=80)                  :: errstr



    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=evlist%shapehdr%naxis2
    
    
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis      = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,evlist%stdhdr,status)
      endif
    endif


    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      fstat=max(fstat,status)  
      call FTPHBN(unit,nrows,evlist%ncol,evlist%cname,evlist%cform,evlist%cunit,evlist%extname,varidat,status)
      fstat=max(fstat,status) 
       
      if (fstat /= 0) then   
        ierr = 2
        errstr = 'error writing extension '//trim(evlist%extname)
        call WRERR(rtname,ierr,errstr)      
      endif
      status=0
      call fwrite_evsclhdr(unit,evlist%scalehdr,status)
      call fwrite_stdhdr(unit,evlist%stdhdr,status)
      call fwrite_twcshdr(unit,evlist%twcs_hdr,status)
      fstat=max(fstat,status)
      if (status > 0) then
        ierr = 2
        errstr = 'error writing standard header '
        call WRERR(rtname,ierr,errstr)             
      endif
 
      status = 0
      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2           
          call  FTPCLD(unit,1,i1,1,rowbuf,evlist%data(i1:i2)%time,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,2,i1,1,rowbuf,evlist%data(i1:i2)%ots,status)
          fstat = max(fstat,status) 
          call  FTPCLJ(unit,3,i1,1,rowbuf,evlist%data(i1:i2)%fracsec,status)
          fstat = max(fstat,status)        
          call  FTTSCL(unit,4,evlist%scalehdr%tscal_ra,evlist%scalehdr%tzero_ra,status)
          call  FTPCLD(unit,4,i1,1,rowbuf,evlist%data(i1:i2)%ra,status)
          fstat = max(fstat,status)
          call  FTTSCL(unit,5,evlist%scalehdr%tscal_dec,evlist%scalehdr%tzero_dec,status)
          call  FTPCLD(unit,5,i1,1,rowbuf,evlist%data(i1:i2)%dec,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,6,i1,1,rowbuf,evlist%data(i1:i2)%x,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,7,i1,1,rowbuf,evlist%data(i1:i2)%y,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,8,i1,1,rowbuf,evlist%data(i1:i2)%energy,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,9,i1,1,rowbuf,evlist%data(i1:i2)%ev_weight,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,10,i1,1,rowbuf,evlist%data(i1:i2)%rawx,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,11,i1,1,rowbuf,evlist%data(i1:i2)%rawy,status)
          fstat = max(fstat,status)
          call  FTTSCL(unit,12,evlist%scalehdr%tscal_subx,evlist%scalehdr%tzero_subx,status)
          call  FTPCLD(unit,12,i1,1,rowbuf,evlist%data(i1:i2)%subx,status)
          fstat = max(fstat,status)
          call  FTTSCL(unit,13,evlist%scalehdr%tscal_suby,evlist%scalehdr%tzero_suby,status)
          call  FTPCLD(unit,13,i1,1,rowbuf,evlist%data(i1:i2)%suby,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,14,i1,1,rowbuf,evlist%data(i1:i2)%pha,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,15,i1,1,rowbuf,evlist%data(i1:i2)%pat_typ,status)
          fstat = max(fstat,status)
          call  FTPCLB(unit,16,i1,1,rowbuf,evlist%data(i1:i2)%pat_inf,status)
          fstat = max(fstat,status)
          call  FTPCLB(unit,17,i1,1,rowbuf,evlist%data(i1:i2)%ccdnr,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,18,i1,1,rowbuf,evlist%data(i1:i2)%flag,status)
          fstat = max(fstat,status)
        enddo
      endif  

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
    
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  


  end subroutine fput_evlist        
   
!-------------------------------------------------------------------------------

  subroutine fput_plan(filename,append,plan,ierr)

  !-------------------------------------------------------------- 
  ! Write fits file 'PLAN'
  ! Outdated format of PLAN file, routine currenty not used by TEL chain.
  ! (GL 11-SEP-2018)
  !--------------------------------------------------------------

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(plan_old_type)                :: plan
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!nhdu, 
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    integer(kind=int32), parameter     :: ncol=27
!    character(len=16), dimension(ncol) :: ttype,tunit
!    character(len=4), dimension(ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_PLAN'
    character(len=80)                  :: errstr



    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=plan%shapehdr%naxis2
    
  
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis      = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,plan%stdhdr,status)
      endif
    endif
 

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      fstat=max(fstat,status)  
      call FTPHBN(unit,nrows,plan%ncol,plan%cname,plan%cform,plan%cunit,plan%extname,varidat,status)
      fstat=max(fstat,status) 
       
      if (fstat /= 0) then   
        ierr = 2
        errstr = 'error writing extension '//trim(plan%extname)
        call WRERR(rtname,ierr,errstr)      
      endif
      status=0
      call fwrite_stdhdr(unit,plan%stdhdr,status)
      fstat=max(fstat,status)
      if (status > 0) then
        ierr = 2
        errstr = 'error writing standard header '
        call WRERR(rtname,ierr,errstr)             
      endif

      status = 0
      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2
          call  FTPCLS(unit,1,i1,1,rowbuf,plan%data(i1:i2)%ptype,status)
          fstat = max(fstat,status)
          call  FTPCLS(unit,2,i1,1,rowbuf,plan%data(i1:i2)%obstat,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,3,i1,1,rowbuf,plan%data(i1:i2)%obsid,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,4,i1,1,rowbuf,plan%data(i1:i2)%obi,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,5,i1,1,rowbuf,plan%data(i1:i2)%tstart,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,6,i1,1,rowbuf,plan%data(i1:i2)%tstop,status)
          fstat = max(fstat,status)
          call  FTPCLS(unit,7,i1,1,rowbuf,plan%data(i1:i2)%fwheel,status)
          fstat = max(fstat,status)
          call  FTPCLS(unit,8,i1,1,rowbuf,plan%data(i1:i2)%detstat,status)
          fstat = max(fstat,status)
          call  FTPCLS(unit,9,i1,1,rowbuf,plan%data(i1:i2)%remarks,status)
          fstat = max(fstat,status)
          call  FTPCLS(unit,10,i1,1,rowbuf,plan%data(i1:i2)%target,status)
          fstat = max(fstat,status)
          call  FTPCLJ(unit,11,i1,1,rowbuf,plan%data(i1:i2)%pinum,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,12,i1,1,rowbuf,plan%data(i1:i2)%ra_pnt,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,13,i1,1,rowbuf,plan%data(i1:i2)%dec_pnt,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,14,i1,1,rowbuf,plan%data(i1:i2)%fieldlon,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,15,i1,1,rowbuf,plan%data(i1:i2)%fieldlat,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,16,i1,1,rowbuf,plan%data(i1:i2)%vis_start,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,17,i1,1,rowbuf,plan%data(i1:i2)%vis_stop,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,18,i1,1,rowbuf,plan%data(i1:i2)%vis_dur,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,19,i1,1,rowbuf,plan%data(i1:i2)%tact,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,20,i1,1,rowbuf,plan%data(i1:i2)%texp,status)
          fstat = max(fstat,status)
          call  FTPCLS(unit,21,i1,1,rowbuf,plan%data(i1:i2)%tcrit,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,22,i1,1,rowbuf,plan%data(i1:i2)%scp_elon,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,23,i1,1,rowbuf,plan%data(i1:i2)%scp_elat,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,24,i1,1,rowbuf,plan%data(i1:i2)%sca_start,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,25,i1,1,rowbuf,plan%data(i1:i2)%rot_start,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,26,i1,1,rowbuf,plan%data(i1:i2)%sca_omega,status)
          fstat = max(fstat,status)
          call  FTPCLD(unit,27,i1,1,rowbuf,plan%data(i1:i2)%rot_omega,status)
          fstat = max(fstat,status)
        enddo
      endif  

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
    
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  


  end subroutine fput_plan        
 

   
!-------------------------------------------------------------------------------

  subroutine fput_hk(filename,append,hk,ierr)


  !-------------------------------------------------------------- 
  ! Write eROSITA house keeping file
  !--------------------------------------------------------------


    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(hk_type)                      :: hk
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: varidat!nhdu, 
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    integer(kind=int32), parameter     :: ncol=21
!    character(len=16), dimension(ncol) :: ttype,tunit
!    character(len=4), dimension(ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_HK'
    character(len=80)                  :: errstr


    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=hk%shapehdr%naxis2
 
 
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis      = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,hk%stdhdr,status)
      endif
    endif


    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      fstat=max(fstat,status)  
      call FTPHBN(unit,nrows,hk%ncol,hk%cname,hk%cform,hk%cunit,hk%extname,varidat,status)
      fstat=max(fstat,status) 
       
      if (fstat /= 0) then   
        ierr = 2
         errstr = 'error writing extension '//trim(hk%extname)
        call WRERR(rtname,ierr,errstr)      
      endif
      status=0
      call fwrite_stdhdr(unit,hk%stdhdr,status)
      fstat=max(fstat,status)
      if (status > 0) then
        ierr = 2
        errstr = 'error writing standard header '
        call WRERR(rtname,ierr,errstr)             
      endif

      status = 0
      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2
          call  FTPCLD(unit,1,i1,1,rowbuf,hk%data(i1:i2)%time,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,2,i1,1,rowbuf,hk%data(i1:i2)%fwheel,status)
          fstat = max(fstat,status)
          call  FTPCLI(unit,3,i1,1,rowbuf,hk%data(i1:i2)%detstat,status)
          fstat = max(fstat,status)
        enddo
      endif  

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
    
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  




  end subroutine fput_hk   

!--------------------------------------------------------------------------------

  subroutine str2int16(numstr,inum,ierr)

  !--------------------------------------------------------
  ! Converts string (character(len=*) to integer (int16)
  !--------------------------------------------------------

    character(len=*)              :: numstr
    integer(kind=int16)           :: inum
    integer(kind=int32)           :: ierr,istat
    character(len=*),parameter    :: rtname = 'STR2INT16'


    ierr=0


    read(numstr,*,IOSTAT=istat)  inum

    if (istat > 0) then
      ierr = 3
      call wrerr(rtname,ierr,'error converting string to int16')
    endif

 
  end   subroutine str2int16

  

!--------------------------------------------------------------------------------

  subroutine obsid2owner(obsid,owner_str,ierr)
  
 
  !-----------------------------------------------------------------------
  ! Extracts data ownership information encoded in first 2 digits of OBSID
  ! Alternatively use OBSTAT2OWNER
  !-----------------------------------------------------------------------

  integer(kind=int32)           :: obsid
  character(len=8)              :: owner_str   
  integer(kind=int32)           :: ierr

  character(len=*),parameter    :: rtname='OBSID2OWNER'
   

  ierr=0
  owner_str=''
  if (obsid < 100000) then
    ierr=3
  else if (obsid < 300000) then
    ! all-sky survey
    owner_str=''
  else if (obsid < 500000) then
    owner_str='MPE'
  else if  (obsid < 700000) then
    owner_str='IKI'
  else if  (obsid < 900000) then 
    owner_str='MPE+IKI'
  else if  (obsid < 1000000) then 
    ! currently not used
    owner_str=''
  else
    ierr=3
  endif

  if (ierr == 3) call wrerr(rtname,ierr,'invalid value of OBSID')

  end subroutine obsid2owner

!--------------------------------------------------------------------------------




  subroutine obstat2owner(obstat,owner_str,ierr)
  
 
  !-----------------------------------------------------------------------
  ! Deternmine data ownership information from OBSTAT
  ! Alternatively use OBSID2OWNER
  !-----------------------------------------------------------------------

  character(len=3)              :: obstat 
  character(len=8)              :: owner_str   
  integer(kind=int32)           :: ierr

  character(len=*),parameter    :: rtname='OBSID2OWNER'
   

  ierr=0
  owner_str=''
  if (obstat == 'SRV' .or. obstat == 'srv') then
    ! all-sky survey
    owner_str=''
  else if (obstat == 'MPE' .or. obstat == 'mpe') then
    owner_str='MPE'
  else if  (obstat == 'IKI' .or. obstat == 'iki') then
    owner_str='IKI'
  else if  (obstat ==  'CAL' .or. obstat == 'cal') then 
    owner_str='MPE+IKI'
  else if  (obstat ==  'COM' .or. obstat == 'cal') then 
    owner_str='MPE+IKI'
  else if  (obstat ==  'PUB' .or. obstat == 'pub') then 
    ! Public observation , TBC:
    owner_str='MPE+IKI'
  else if  (obstat ==  'SLW' .or. obstat == 'slw') then 
    ! currently treated like survey (owner depends on sky region) tbc! 
    owner_str=''
  else
    ierr=3
  endif

  if (ierr == 3) call wrerr(rtname,ierr,'invalid value of OBSTAT')

  end subroutine obstat2owner

!--------------------------------------------------------------------------------





  subroutine int16str6(inum,numstr,ierr)
  
  !--------------------------------------------------------
  ! Converts integer (int16) to string (character(len=6) )
  !--------------------------------------------------------

    character(len=6)              :: numstr
    integer(kind=int16)           :: inum
    integer(kind=int32)           :: ierr,istat
    character(len=*),parameter    :: rtname = 'INT16STR6'


    ierr=0

    write(numstr,*,IOSTAT=istat)  inum

    if (istat > 0) then
      ierr = 3
      call wrerr(rtname,ierr,'error converting integer to string')
    endif


  end subroutine int16str6

!-------------------------------------------------------------------------------

  
!--------------------------------------------------------------------------------
! GTI related routines:
! TBD: Put into separate module (i.e gtilib) ?
! 
!-------------------------------------------------------------------------------

  subroutine ingti(time,gti,good,ierr)

  !-------------------------------------------------------
  ! Checks whether time is within good time of gti file 
  !-------------------------------------------------------

    real(kind=real64)     :: time
    type(gti_type)        :: gti
    logical(kind=log32)   :: good
    integer(kind=int32)   :: ierr

    integer(kind=int32) :: ngti,igti

    character(len=*),parameter       :: rtname='INGTI'
   
    ierr=0

    ngti = gti%shapehdr%naxis2
  
    good=.false.
    do igti=1, ngti
      if (time >= gti%data(igti)%start .and. time <= gti%data(igti)%stop) then
        good=.true.
        exit
      endif 
    end do


  end subroutine ingti


!-------------------------------------------------------------------------------
 

  subroutine get_instnum(instkey,instnum,ierr) 

    !--------------------------------------------------------------
    ! Extract telescope number as len=1 character string
    ! from string instkey (eg. INSTRUME keyword), assuming last character 
    ! is the telescope number.
    ! 
    !--------------------------------------------------------------

    implicit none

    character(len=8)               :: instkey
    character(len=1)               :: instnum
    integer(kind=int32)            :: ierr     

    integer                        :: i

    character(len=*),parameter              :: rtname='GET_INSTNUM'             

    ierr=0

    i = len(trim(instkey))

    instnum=instkey(i:i)

    if (index('1234567',instnum) == 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error: '//instnum//' not a valid telescope number')
    endif 



  end subroutine get_instnum
   

!-------------------------------------------------------------------------------
 


end module tel_io

