! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! This is a pseudo-module to illustrate how to make use of the classes in erolib_bintable_mod.f90. You can also look at the subclasses in erolib_events_mod.f90, erolib_events_pp_mod.f90, erolib_extensions_mod.f90, etc.
!
module bintable_example
  use erolib_general_mod
  use erolib_bintable_mod

  implicit none

  type, extends(tBinTable) :: tExampleTab
    integer :: numFrogs=0
  contains
    procedure :: init => init_tExampleTab
  end type tExampleTab

  type, extends(tFitsFile) :: tMyFITSfile
    type(tExampleTab) :: my_table
  contains
  end type tMyFITSfile

  character(100), parameter, private :: default_extname="MYTABNAME"

contains

  !.....................................................................
  subroutine init_tExampleTab(this, status, fileInfo, extname)
    ! The number and nature of its columns is the principal means by which an actual binary table object, subclassed from tBinTable, is distinguished. The column objects are conveniently constructed via an init routine such as the present one.
    !
    implicit none

    class(tExampleTab) :: this
    integer,         intent(inout)                :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo
    character(*),    intent(in), optional         :: extname

    character(*), parameter :: functionName = "init_tExampleTab"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    if(present(extname)) then
      this%info%tableName = extname
    else
      this%info%tableName = default_extname
    end if

    call this%initColumn("A", "NAME",   "", num_chars=20)
    call this%initColumn("D", "HEIGHT", "cm")
    call this%initColumn("E", "WEIGHT", "kg")
    call this%initColumn("I", "SCORE",  "")

    call this%initColumns(status)

  end subroutine init_tExampleTab

  !.....................................................................
  subroutine write_table(filename, scores, names, heights, weights, status)
    implicit none

    character(*), intent(in)                            :: filename
    integer(i2),  intent(in), dimension(:,:)            :: scores
    character(*), intent(in), dimension(size(scores,2)) :: names
    real(r8),     intent(in), dimension(size(scores,2)) :: heights
    real(r4),     intent(in), dimension(size(scores,2)) :: weights
    integer,      intent(inout)                         :: status

    type(tMyFITSfile) :: fileObj
    type(tExampleTab) :: tableObj
    integer :: numRows,numVecElements,ii
    integer :: i2X_SCORE
    integer(i2), dimension(:,:), pointer :: score_col_ptr=>null()
    real(r4), dimension(:), pointer :: weight_col_ptr=>null()
    real(r8), dimension(:), pointer :: height_col_ptr=>null()
    character(buffsize), dimension(:), pointer :: name_col_ptr=>null()

    numVecElements = size(scores,1)
    numRows        = size(scores,2)

    call fileObj%info%setFileName(filename)
    call tableObj%init(status, fileObj%info)

    ! Declare that all the columns are to be written to file:
    call tableObj%ibset_all_useage(writeBit)
    tableObj%useageIsSet = .true. ! without this, none of the tBinTable read or write methods will work.

    call tableObj%alloc(status, numRows)
    i2X_SCORE = tableObj%getColNum("SCORE")
    call tableObj%cols(i2X_SCORE)%allocVec(numVecElements, status)

    call tableObj%get_col_ptr("NAME",   name_col_ptr)
    call tableObj%get_col_ptr("HEIGHT", height_col_ptr)
    call tableObj%get_col_ptr("WEIGHT", weight_col_ptr)
    call tableObj%get_col_ptr("SCORE",  score_col_ptr)
    do ii=1,numRows
      name_col_ptr(ii) = names(ii)
      height_col_ptr(ii) = heights(ii)
      weight_col_ptr(ii) = weights(ii)
      score_col_ptr(:,ii) = scores(:,ii) ! note the order of indices.
    end do
    nullify(name_col_ptr)
    nullify(height_col_ptr)
    nullify(weight_col_ptr)
    nullify(score_col_ptr)

    call fileObj%open(FF_CREATE, status, clobber=.true.)
    call tableObj%write(status) ! tableObj contains the file unit number because it has a pointer to the file info attribute in which this value is stored.
    call fileObj%close(status)

    ! Clean up:
    call tableObj%destroy(status)

  end subroutine write_table

  !.....................................................................
  subroutine read_table(filename, scores, names, heights, weights, status)
    ! The calling routine needs to deallocate the arguments after use.
    implicit none

    character(*), intent(in)                               :: filename
    integer(i2),  intent(out), dimension(:,:), allocatable :: scores
    character(*), intent(out), dimension(:),   allocatable :: names
    real(r8),     intent(out), dimension(:),   allocatable :: heights
    real(r4),     intent(out), dimension(:),   allocatable :: weights
    integer,      intent(inout)                            :: status

    type(tMyFITSfile) :: fileObj
    type(tExampleTab) :: tableObj
    integer :: numRows,numVecElements,ii
    integer :: i2X_SCORE
    integer(i2), dimension(:,:), pointer :: score_col_ptr=>null()
    real(r4), dimension(:), pointer :: weight_col_ptr=>null()
    real(r8), dimension(:), pointer :: height_col_ptr=>null()
    character(buffsize), dimension(:), pointer :: name_col_ptr=>null()

    call fileObj%info%setFileName(filename)
    call tableObj%init(status, fileObj%info)

    ! Declare that all the columns are to be read from file:
    call tableObj%ibset_all_useage(readBit)
    tableObj%useageIsSet = .true. ! without this, none of the tBinTable read or write methods will work.

    call fileObj%open(FF_READONLY, status)
    call tableObj%read(status) ! tableObj contains the file unit number because it has a pointer to the file info attribute in which this value is stored.
    call fileObj%close(status)

    numRows = tableObj%getNumRows()
    i2X_SCORE = tableObj%getColNum("SCORE")
    numVecElements = tableObj%cols(i2X_SCORE)%getVectorSize()

    allocate(names(  numRows))
    allocate(heights(numRows))
    allocate(weights(numRows))
    allocate(scores(numVecElements, numRows))

    call tableObj%get_col_ptr("NAME",   name_col_ptr)
    call tableObj%get_col_ptr("HEIGHT", height_col_ptr)
    call tableObj%get_col_ptr("WEIGHT", weight_col_ptr)
    call tableObj%get_col_ptr("SCORE",  score_col_ptr)
    do ii=1,numRows
      names(ii)    = name_col_ptr(ii)
      heights(ii)  = height_col_ptr(ii)
      weights(ii)  = weight_col_ptr(ii)
      scores(:,ii) = score_col_ptr(:,ii) ! note the order of indices.
    end do
    nullify(name_col_ptr)
    nullify(height_col_ptr)
    nullify(weight_col_ptr)
    nullify(score_col_ptr)

    ! Clean up:
    call tableObj%destroy(status)

  end subroutine read_table

end module bintable_example

