! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! Non-method subroutines contained:
!  subroutine read_extensions(ccds, mjdref, extList, status)
!  subroutine read_hks(ccds, hkList, status)
!  subroutine read_gtis(ccds, mjdref, gtiList, status)
!  subroutine dealloc_eventfiles(eventfiles, status)
!  subroutine alloc_eventfiles(eventfiles, n, status)
!  subroutine get_merge_std_header_eventfile(eventFile, eventFile_out, status)
!  subroutine read_events_ext_keywords_eventfile(eventFile, status)
!  subroutine compare_stdheader_eventfile(eventfiles, status, alwayswarn)
!  subroutine compare_rawxy_masks(ccdnrs1, rawxy_mask1, ccdnrs2, rawxy_mask2, status)
!  subroutine merge_stdheader_eventfile(eventfiles, eventfile_out, status)
!  subroutine merge_ccdnrs_eventfiles()
!  subroutine merge_rawxy_masks(ccdnrs2, rawxy_mask2, rawxy_mask, ccd_is_merged, status)
!  subroutine merge_extensions(memset, eventfiles, eventfile_out, status)
!  subroutine merge_evlist_events(eventfiles, events_out, status)
!  subroutine merge_evlist_extensions(eventfiles, extI, tel_id, ext_out, status)
!  subroutine merge_evlist_hk(eventfiles, extI, tel_id, hk_out, status)
!  subroutine merge_evlist_gti(eventfiles, extI, tel_id, gti_out, status)
!  subroutine read_and_set_ccds(evlist, status)

module erolib_eventfile_mod
  use erolib_region_mod
  use erolib_wcs_mod
  use erolib_gtitab_mod
  use erolib_events_mod
  use erolib_image2_mod !******** fix up the image classes!
  use erolib_extensions_mod
  use erolib_hk_mod
  use erolib_rawxy_mod
  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr, update_stdhdr, evscalehdr_type, fwrite_evsclhdr
  use srctool_generic , only : coord_type, COORDSYS_CODE_ICRS
  use srctool_generic_regions , only : srctool_regions_test_if_coord_in_region, region_type
  use time_mod ! for sec2date.
  use radec2pix_mod

  implicit none

!logical, private, parameter :: doTest=.false.

  type :: t_filter
    logical :: do_filter_telid   = .false.
    logical :: do_filter_energy  = .false.
    logical :: do_filter_pattern = .false.
    logical :: do_filter_region  = .false.
    logical :: do_filter_gti     = .false.
    logical :: do_filter_rawxy   = .false.
    logical :: do_filter_flag    = .false.
    integer :: flag_mask=0 ! default is discard nothing - accept everything.
    logical :: flag_invert = .false.
    type(tRegions) :: regionObj
    real(r8) :: overlap = 2.0
    integer :: gti_filter_type = 0
    type(tGTI_tab) :: externalgti
    type(tRawXYSelection) :: rawxy_sel
  end type t_filter

  type :: tExtList
    type(tExtension), dimension(N_TELESCOPES) :: extPerTM
    integer :: ext_type = -1 ! valid: extDEADCOR, extBADPIX, extCORRATT or extOFFSET.
    character(buffsize) :: extNameRoot=""
    integer(i1) :: useageFlags = 0 ! LS bit = optional, 2nd is
    ! R, 3rd bit = W, 4th bit indicates internal use.
  contains
    procedure, non_overridable :: containsData => containsData_tExtList
  end type tExtList

  type :: tHKList
    type(tHK), dimension(N_TELESCOPES) :: hkPerTM
    integer :: ext_type = -1 ! valid: iHK1, iHK2, iHK3 or iHK5.
    character(buffsize) :: extNameRoot=""
    integer(i1) :: useageFlags = 0 ! LS bit = optional, 2nd is
    ! R, 3rd bit = W, 4th bit indicates internal use. Default here is read/write.
  contains
    procedure, non_overridable :: containsData => containsData_tHKList
  end type tHKList

  type :: tGTIList
    type(tGTI_tab), dimension(N_TELESCOPES) :: gtiPerTM
    integer :: ext_type = -1 ! valid: iGTI, iQUALGTI, iFLAREGTI or iREGIONGTI.
    character(buffsize) :: extNameRoot=""
    integer(i1) :: useageFlags = 0 ! LS bit = optional, 2nd is
    ! R, 3rd bit = W, 4th bit indicates internal use.
  contains
    procedure, non_overridable :: containsData => containsData_tGTIList
  end type tGTIList

  integer, parameter :: numHKtypes=4!**** define these in the respective modules!
  integer, parameter :: extTypesInEvFile(*) = [extCORRATT,extDEADCOR,extBADPIX]!*** bring in extOFFSET as soon as these extensions come into use.
  integer, parameter :: gtiTypesInEvFile(*) = [iGTI,iQUALGTI,iFLAREGTI,iREGIONGTI]

  type, extends(tFitsFile) :: tEventFile
    integer, dimension(:), allocatable :: ccdnr_infile
    logical :: isInitialized = .false.
    type(stdhdr_type) :: std_header
    type(evscalehdr_type) :: evscale_header

    type(tEvents) :: events
    logical :: write_events = .true.
    type(tImage2D_i4) :: image
    logical :: image_containsData = .false.

    ! Each element of each of these three types of list contains an array
    ! of tBinTable objects which contains N_TELESCOPES entries.
    type(tHKList),  dimension(numHKtypes)  :: hkList
    type(tExtList), dimension(size(extTypesInEvFile)) :: extList
    type(tGTIList), dimension(size(gtiTypesInEvFile)) :: gtiList
    type(tRegions) :: regionObj
    type(tRawXYMask) :: rawxy_mask
    type(tGTI_tab) :: externalgti

    logical :: filtered_pattern = .false.
    logical :: filtered_energy  = .false.
    logical :: filtered_telid   = .false.
    logical :: filtered_gti     = .false.
    logical :: filtered_flag    = .false.
    logical :: filtered_region  = .false.
    logical :: filtered_rawxy   = .false.

  contains
    procedure :: read_ccd_nrs => read_ccd_nrs_tEventFile
    procedure :: init => init_tEventFile ! to associate tExtList etc objects with extNameRoots.
    procedure :: reset_fileInfo_ptrs => reset_fileInfo_ptrs_tEventFile
    procedure :: print_me => print_me_tEventFile
    procedure :: setup_for_readwrite => setup_for_readwrite_tEventFile
    procedure :: read => read_tEventFile
    procedure :: read_rawxy_sel => read_rawxy_sel_tEventFile
    procedure :: read_events => read_events_tEventFile
    procedure :: write => write_tEventFile
    procedure :: destroy => destroy_tEventFile
    procedure :: sort_events => sort_events_tEventFile
    procedure :: read_stdheader => read_stdheader_tEventFile
    procedure :: write_stdheader => write_stdheader_tEventFile
    procedure :: apply_stdheader => apply_stdheader_tEventFile
    procedure :: apply_filter => apply_filter_tEventFile
    procedure :: delete_flagged_frames_from_GTI => delete_frames_tEventFile
    procedure :: adjust_GTI_bound_kwds => adjust_GTI_bound_kwds_tEventFile
    procedure :: apply_filter_tm_GTIs => apply_filter_tm_GTIs_tEventFile
    procedure :: apply_empty_GTI => apply_empty_GTI_tEventFile
    procedure :: apply_filter_nontm_GTI => apply_filter_nontm_GTI_tEventFile
    procedure :: apply_tm_GTI => apply_tm_GTI_tEventFile
    procedure :: apply_filter_region => apply_filter_region_tEventFile
    procedure :: derive_region_gti => derive_region_gti_tEventFile
    procedure :: record_rawxy_sel => record_rawxy_sel_tEventFile
    procedure :: get_ext_index => get_ext_index_tEventFile
  end type tEventFile

!*** set all the subroutines private except for:
!	get_merge_std_header_eventfile()
!	compare_stdheader_eventfile()
!	merge_stdheader_eventfile()
!	merge_extensions()

contains

  !=====================================================================
  ! Methods of tExtList, tHKList and tGTIList:

  !.....................................................................
  function containsData_tExtList(this) result(containsData)
    implicit none

    class(tExtList) :: this
    logical :: containsData

    integer :: i
    character(*), parameter :: functionName = "containsData_tExtList"

    containsData = .false. ! default

    do i=1,size(this%extPerTM)
      if(this%extPerTM(i)%containsData()) then
        containsData = .true.
    exit
      end if
    end do

  end function containsData_tExtList

  !.....................................................................
  function containsData_tHKList(this) result(containsData)
    implicit none

    class(tHKList) :: this
    logical :: containsData

    integer :: i
    character(*), parameter :: functionName = "containsData_tHKList"

    containsData = .false. ! default

    do i=1,size(this%hkPerTM)
      if(this%hkPerTM(i)%containsData()) then
        containsData = .true.
    exit
      end if
    end do

  end function containsData_tHKList

  !.....................................................................
  function containsData_tGTIList(this) result(containsData)
    implicit none

    class(tGTIList) :: this
    logical :: containsData

    integer :: i
    character(*), parameter :: functionName = "containsData_tGTIList"

    containsData = .false. ! default

    do i=1,size(this%gtiPerTM)
      if(this%gtiPerTM(i)%containsData()) then
        containsData = .true.
    exit
      end if
    end do

  end function containsData_tGTIList

  !=====================================================================
  ! The following routines are not methods of tExtList etc, they're just
  ! utility routines that manipulate the objects.

  !.....................................................................
  subroutine read_extensions(ccds, mjdref, extObj, status)

    implicit none

    integer, dimension(:), intent(in)    :: ccds
    real(r8),              intent(in)    :: mjdref
    class(tExtList),       intent(inout) :: extObj
    integer,               intent(inout) :: status

    character(*), parameter :: functionName = "read_extensions"
    integer :: numInst,ii,ierr,ci
    character(buffsize) :: istr

    if(status > 100 .or. status == s_errRep)&
  return

    if(minval(ccds)<1 .or. maxval(ccds)>size(extObj%extPerTM)) then
      istr = int2str(size(extObj%extPerTM))
call wrstop(functionName, "Array 'ccds' has entries outside the range [1,"//trim(istr)//"]")
    end if

    numInst = size(ccds)
    do ii = 1, numInst
      ci = ccds(ii)

      ierr = 0
      call extObj%extPerTM(ci)%read(ierr)

      if(ierr == c_err) then ! extension was not present.
        call wrerr(functionName, warn, "Extension "//trim(trim(extObj%extPerTM(ci)%info%tableName))&
          //" was expected but could not be read.")
    cycle
      else if(ierr/=0) then
call wrstop(functionName, "Error in reading "//trim(extObj%extPerTM(ci)%info%tableName))
      end if

      if(.not.extObj%extPerTM(ci)%header%keywordContainsData("MJDREF"))&
        call extObj%extPerTM(ci)%header%setKeywordValue("MJDREF",mjdref)
    end do

  end subroutine read_extensions

  !.....................................................................
  subroutine read_hks(ccds, hkList, status)

    implicit none

    integer, dimension(:), intent(in)    :: ccds
    class(tHKList),        intent(inout) :: hkList
    integer,               intent(inout) :: status

    character(*), parameter :: functionName = "read_hks"
    integer :: numInst,ii,ierr,ci
!,col_num
    character(buffsize) :: istr

    if(status > 100 .or. status == s_errRep)&
  return

    if(minval(ccds)<1 .or. maxval(ccds)>size(hkList%hkPerTM)) then
      istr = int2str(size(hkList%hkPerTM))
call wrstop(functionName, "Array 'ccds' has entries outside the range [1,"//trim(istr)//"]")
    end if

    numInst = size(ccds)
    do ii = 1, numInst
      ci = ccds(ii)

      ierr = 0
      call hkList%hkPerTM(ci)%read(ierr)

      if(ierr == c_err) then ! extension was not present.
        call wrerr(functionName, warn, "Extension "//trim(trim(hkList%hkPerTM(ci)%info%tableName))&
          //" was expected but could not be read.")
    cycle
      else if(ierr/=0) then
        call wrstop(functionName, "Error in reading "//trim(hkList%hkPerTM(ci)%info%tableName))
      end if

      if(hkList%ext_type==iHK3) then
        if(.not.hkList%hkPerTM(ci)%read_succeeded("CxTCCDPT"))&
          call hkList%hkPerTM(ci)%set_col_useage("CxTCCDPT", 0,0,0,0)
      end if
    end do

  end subroutine read_hks

  !.....................................................................
  subroutine read_gtis(ccds, mjdref, gtiList, status, nowarn, repair_gtis)

    integer, dimension(:), intent(in)    :: ccds
    real(r8),              intent(in)    :: mjdref
    class(tGTIList),       intent(inout) :: gtiList
    integer,               intent(inout) :: status
    logical, optional,     intent(in)    :: nowarn, repair_gtis

    character(*), parameter :: functionName = "read_gtis"
    character(buffsize) :: errmsg_prefix,fileName,istr
    integer :: numInst,ii,ierr,ci
    logical :: local_dowarn,local_repair_gtis,repair_was_necessary

    if(present(nowarn)) then
      local_dowarn = .not.nowarn
    else
      local_dowarn = .true.
    end if

    if(present(repair_gtis)) then
      local_repair_gtis = repair_gtis
    else
      local_repair_gtis = .false.
    end if

    fileName = ""

    if(status > 100 .or. status == s_errRep)&
  return

    if(minval(ccds)<1 .or. maxval(ccds)>size(gtiList%gtiPerTM)) then
      istr = int2str(size(gtiList%gtiPerTM))
call wrstop(functionName, "Array 'ccds' has entries outside the range [1,"//trim(istr)//"]")
    end if

    numInst = size(ccds)
    do ii = 1, numInst
      ci = ccds(ii)

      ierr = 0

      if(.not.gtiList%gtiPerTM(ci)%existsInFile()) then
        if(local_dowarn)&
          call wrerr(functionName, warn, "Extension "//trim(gtiList%gtiPerTM(ci)%info%tableName)&
            //" was not found in the file.")
    cycle
      end if

      call gtiList%gtiPerTM(ci)%read(ierr)

      if(ierr/=0)&
call wrstop(functionName, "Error in reading "//trim(gtiList%gtiPerTM(ci)%info%tableName))

      if(local_repair_gtis) then
        ierr = 0
        call gtiList%gtiPerTM(ci)%repair(ierr, repair_was_necessary)
        if(local_dowarn .and. repair_was_necessary)&
          call wrerr(functionName, warn, "Disordered GTIs in extension "&
            //trim(gtiList%gtiPerTM(ci)%info%tableName)//" have been repaired.")

      end if

      call gtiList%gtiPerTM(ci)%info%fileInfo%getFileName(fileName)
      write(errmsg_prefix, '(a)') trim(fileName)&
         //":"//trim(gtiList%gtiPerTM(ci)%info%tableName)//": "

      ! Check the ordering
      !
      ierr = 0
      call gtiList%gtiPerTM(ci)%sanityCheck(ierr, errmsg_prefix, local_dowarn)

      if(.not.gtiList%gtiPerTM(ci)%header%keywordContainsData("MJDREF"))&
        call gtiList%gtiPerTM(ci)%header%setKeywordValue("MJDREF",mjdref)
    end do

  end subroutine read_gtis

  !=====================================================================
  ! Methods of tEventFile:

  !.....................................................................
  subroutine read_ccd_nrs_tEventFile(this, fileName, ccdnrs, status)
    ! The calling routine should deallocate ccdnrs after use.
    implicit none

    class(tEventFile) :: this
    character(*),                       intent(in)    :: fileName
    integer, dimension(:), allocatable, intent(out)   :: ccdnrs
    integer,                            intent(inout) :: status

    character(*), parameter :: functionName = "read_ccd_nrs_tEventFile"
    integer :: ierr,ii,ldigit
    character(buffsize), dimension(:), allocatable :: instrum
    type(stdhdr_type) :: std_header

    call this%open(FF_READONLY, status, filename=fileName)

    ierr=0
    call FTMNHD(this%info%unitNo, BINARY_TBL, "EVENTS", 0, ierr)
    if(ierr==301) then
call wrstop(functionName, "EVENTS extension not found in file "//trim(fileName))
    else
      call report_FITS_err(functionName, ierr, s_err, "- extension EVENTS")
    end if

    call fread_stdhdr(this%info%unitNo, std_header, status)
    if(status == 1) then
      call wrlog(functionName, "Missing keywords: reading standard header")
      status = ok
    end if

    call this%close(status)

    call alloc(ccdnrs,  std_header%ninst, status)

    if(std_header%ninst == 0)&
  return

    call alloc(instrum, std_header%ninst, status)
    if(std_header%ninst > 0) then
      instrum = std_header%instrumx(:std_header%ninst)
    end if

    do ii = 1, std_header%ninst
      ldigit = len_trim(instrum(ii))
      ccdnrs(ii) = str2int(instrum(ii)(ldigit:ldigit))
    end do

    deallocate(instrum)

  end subroutine read_ccd_nrs_tEventFile

  !.....................................................................
  subroutine init_tEventFile(this, ccdnrs, status)
    implicit none

!**** 20201104 IMS: I don't think we should be setting keywords in this. They should be set (if write-only) in this%setup_for_readwrite().

    class(tEventFile) :: this
    integer, dimension(:),           intent(in)    :: ccdnrs
    integer,                         intent(inout) :: status

    character(*), parameter :: functionName = "init_tEventFile"
    integer :: ii,kk
!,jj
    character(buffsize) :: istr

    do ii=1,size(extTypesInEvFile)
      this%extList(ii)%ext_type = extTypesInEvFile(ii)
    end do

    do ii=1,numHKtypes
      this%hkList(ii)%ext_type = ii
    end do

    do ii=1,size(gtiTypesInEvFile)
      this%gtiList(ii)%ext_type = gtiTypesInEvFile(ii)
    end do

    this%hkList(iHK1)%extNameRoot = "HKx1"
    this%hkList(iHK2)%extNameRoot = "HKx2"
    this%hkList(iHK3)%extNameRoot = "HKx3"
    this%hkList(iHK5)%extNameRoot = "HKx5"

    this%ccdnr_infile = ccdnrs

    if(minval(this%ccdnr_infile)<1 .or. maxval(this%ccdnr_infile)>N_TELESCOPES) then
      istr = int2str(N_TELESCOPES)
call wrstop(functionName, "this%ccdnr_infile has entries outside the range [1,"//trim(istr)//"]")
    end if

    call this%events%init(status, this%info)

    do ii=1,size(extTypesInEvFile)
      do kk=1,size(this%extList(ii)%extPerTM)
        call this%extList(ii)%extPerTM(kk)%init(this%extList(ii)%ext_type, status, this%info)

        write(this%extList(ii)%extPerTM(kk)%info%tableName,'(a,i1)')&
          trim(this%extList(ii)%extPerTM(kk)%extNameRoot), kk

        this%extList(ii)%extPerTM(kk)%tel_id = kk
        call this%extList(ii)%extPerTM(kk)%header%setKeywordValue("MJDREF",this%std_header%mjdref)
      end do
      this%extList(ii)%extNameRoot = this%extList(ii)%extPerTM(1)%extNameRoot
    end do

    do ii=1,numHKtypes
      do kk=1,size(this%hkList(ii)%hkPerTM)
        call this%hkList(ii)%hkPerTM(kk)%init(ii, status, this%info)
        write(this%hkList(ii)%hkPerTM(kk)%info%tableName,'(a,i1,a)')&
          "HK", kk, this%hkList(ii)%hkPerTM(kk)%typeChar

        call this%hkList(ii)%hkPerTM(kk)%header%setKeywordValue("MJDREF",this%std_header%mjdref)
        call this%hkList(ii)%hkPerTM(kk)%header%setKeywordValue("TEL_ID",kk)
        call this%hkList(ii)%hkPerTM(kk)%header%setKeywordValue("APP_ID",HK_appIDs(ii))

        this%hkList(ii)%hkPerTM(kk)%tel_id = kk
      end do
    end do

    do ii=1,size(gtiTypesInEvFile)
      do kk=1,size(this%gtiList(ii)%gtiPerTM)
        call this%gtiList(ii)%gtiPerTM(kk)%init(this%gtiList(ii)%ext_type, status, this%info)
        write(this%gtiList(ii)%gtiPerTM(kk)%info%tableName,'(a,i1)')&
          trim(this%gtiList(ii)%gtiPerTM(kk)%extNameRoot), kk

        this%gtiList(ii)%gtiPerTM(kk)%tel_id = kk
      end do
      this%gtiList(ii)%extNameRoot = this%gtiList(ii)%gtiPerTM(1)%extNameRoot
    end do

    this%isInitialized = .true.

  end subroutine init_tEventFile

  !.....................................................................
  subroutine reset_fileInfo_ptrs_tEventFile(this)
    ! If you simply copy one event list to another via
    !	evlistB = evlistA
    ! then the pointers to tFitsInfo in the 'info' attribute of all the tBinTable-derived extensions of evlistB will still point to evlistA. The present subroutine sets them to point correctly to evlistB.

    implicit none

    class(tEventFile), target :: this

    integer :: ii,kk

    this%events%info%fileInfo => this%info

    do ii=1,size(this%extList)
      do kk=1,size(this%extList(ii)%extPerTM)
        this%extList(ii)%extPerTM(kk)%info%fileInfo => this%info
      end do
    end do

    do ii=1,size(this%hkList)
      do kk=1,size(this%hkList(ii)%hkPerTM)
        this%hkList(ii)%hkPerTM(kk)%info%fileInfo => this%info
      end do
    end do

    do ii=1,size(this%gtiList)
      do kk=1,size(this%gtiList(ii)%gtiPerTM)
        this%gtiList(ii)%gtiPerTM(kk)%info%fileInfo => this%info
      end do
    end do

  end subroutine reset_fileInfo_ptrs_tEventFile

  !.....................................................................
  subroutine print_me_tEventFile(this)
    ! This is for diagnostic purposes. (Note that it would be more useful if it wrote to a file object passed as an argument.)
    implicit none

    class(tEventFile) :: this

    character(*), parameter :: functionName = "print_me_tEventFile"
    integer :: ii
    character(buffsize) :: msg
    character(buffsize) :: fileName

    fileName = ""

    call this%info%getFileName(fileName)
    print *, "File name: >"//trim(fileName)//"<"
    print *, "FITS unit num:", this%info%unitNo
    print *, "File opened for read?", this%info%fileOpenedForRead
    print *, "File opened for write?", this%info%fileOpenedForWrite

    if(allocated(this%ccdnr_infile)) then
      msg = "TMs: ["
      if(size(this%ccdnr_infile)>0) then
        write(msg, '(a,i1)') trim(msg), this%ccdnr_infile(1)
        do ii=2,size(this%ccdnr_infile)
          write(msg, '(a,i1)') trim(msg)//",", this%ccdnr_infile(ii)
        end do
      end if
      msg = trim(msg)//"]."
      print *, trim(msg)

    else
      print *, "ccdnr_infile not allocated."
    end if

    print *, "File initialized?", this%isInitialized

    print *, "EVENTS initialized?", this%events%isInitialized
    print *, "write_events?", this%write_events
    print *, "image_containsData?", this%image_containsData

    print *, "regtable_containsData?", this%regionObj%containsData

    print *, "filtered_pattern?", this%filtered_pattern
    print *, "filtered_energy?",  this%filtered_energy
    print *, "filtered_telid?",   this%filtered_telid
    print *, "filtered_gti?",     this%filtered_gti
    print *, "filtered_flag?",    this%filtered_flag
    print *, "filtered_region?",  this%filtered_region
    print *, "filtered_rawxy?",   this%filtered_rawxy

    print *, "Standard extensions:"
    do ii=1,size(this%extList)!numExtTypes
      write(*, '(a,i1,a)') "  ", ii, " extNameRoot >"//trim(this%extList(ii)%extNameRoot)//"<"
      print *, "  Contains data?", this%extList(ii)%containsData()
    end do

    print *, "HK extensions:"
    do ii=1,numHKtypes
      write(*, '(a,i1,a)') "  ", ii, " extNameRoot >"//trim(this%hkList(ii)%extNameRoot)//"<"
      print *, "  Contains data?", this%hkList(ii)%containsData()
    end do

    print *, "GTI extensions:"
    do ii=1,size(this%gtiList)!numGTItypes
      write(*, '(a,i1,a)') "  ", ii, " extNameRoot >"//trim(this%gtiList(ii)%extNameRoot)//"<"
      print *, "  Contains data?", this%gtiList(ii)%containsData()
    end do

  end subroutine print_me_tEventFile

  !.....................................................................
  subroutine read_stdheader_tEventFile(this, status)
    implicit none

    class(tEventFile) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "read_stdheader_tEventFile"
    integer :: ierr
    character(buffsize) :: fileName

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not.this%info%fileOpenedForRead)&
call wrstop(functionName, "You must open the event list file for reading before calling this subroutine.")

    ierr=0
    call FTMNHD(this%info%unitNo, BINARY_TBL, "EVENTS", 0, ierr)
    if(ierr==301) then
      call this%info%getFileName(fileName)
call wrstop(functionName, "EVENTS extension not found in file "//trim(fileName))
    else
      call report_FITS_err(functionName, ierr, s_err, "- extension EVENTS")
    end if

    call fread_stdhdr(this%info%unitNo, this%std_header, status)
    if(status == 1) then
      call wrlog(functionName, "Missing keywords : reading standard header")
      status = ok
    end if

    call this%apply_stdheader(status)

  end subroutine read_stdheader_tEventFile

  !.....................................................................
  subroutine setup_for_readwrite_tEventFile(this, flagbit_list, status, exts_to_flag, gti_exts_to_flag)
    ! This covers the general-purpose case, for example for use by evtool.
    implicit none

    class(tEventFile) :: this
    integer, dimension(:),           intent(in)    :: flagbit_list
    integer,                         intent(inout) :: status
    integer, dimension(:), optional, intent(in)    :: exts_to_flag,gti_exts_to_flag

    character(*), parameter :: functionName = "setup_for_readwrite_tEventFile"
    integer :: ii,jj,kk,ci,numBits
    character(buffsize) :: iStr
    logical :: type_found

    numBits = size(flagbit_list)

    call this%events%header%set_kwd_useage("CHOPDEAD", 1, 1, 1, 0) ! read/write, but read is optional.
    call this%events%header%set_kwd_useage("T_OFFCOR", 1, 1, 1, 0) ! read/write, but read is optional.

    do jj=1,numBits
      call this%events%ibset_all_useage(flagbit_list(jj))
    end do

    ! Special useage cases:
    !
    ! These are just alternative names for the read case - we won't write them or use them.
    call this%events%set_col_useage("ENERGY",1,1,0,0)
    call this%events%set_col_useage("CCDNR", 1,1,0,0)
    !
    ! Not present in the canonical evtool event list.
    call this%events%set_col_useage("PAT_IND",    0,0,0,0)
    call this%events%set_col_useage("AMP_GAIN",   0,0,0,0)
    call this%events%set_col_useage("AMP_COR",    0,0,0,0)

    ! This column is for internal use only.
    !
    call this%events%ibclr_col_useage("frame_ID", TCOL_ISOPT_BIT)
    call this%events%ibclr_col_useage("frame_ID", TCOL_READ_BIT)
    call this%events%ibclr_col_useage("frame_ID", TCOL_WRITE_BIT)

    this%events%useageIsSet = .true.

    ! Set read/write flags for CORRATTn, DEADCORn and BADPIXn extensions:
    !
    do ii=1,size(this%extList)
      do kk=1,size(this%ccdnr_infile)
        ci = this%ccdnr_infile(kk)
        do jj=1,numBits
          call this%extList(ii)%extPerTM(ci)%ibset_all_useage(flagbit_list(jj))
        end do
        this%extList(ii)%extPerTM(ci)%useageIsSet = .true.
      end do
    end do

    do ii=1,size(this%hkList)
      do kk=1,size(this%ccdnr_infile)
        ci = this%ccdnr_infile(kk)
        do jj=1,numBits
          call this%hkList(ii)%hkPerTM(ci)%ibset_all_useage(flagbit_list(jj))
        end do

        if(ii==iHK1) then ! 20201029 IMS: These columns are present in the input H_HKn1_n_* files. I included them in the iHK1 binary table definition because I use them in non-sass software (ims_utils/extract_mod.f90). However we don't write them to the 'canonical' HKn1 extension of the event lists, so I actually ought to delete them from iHK1 and modify my private code to use its own bespoke version.
          call this%hkList(ii)%hkPerTM(ci)%set_col_useage("PnSEQMODE", 0,0,0,0)
          call this%hkList(ii)%hkPerTM(ci)%set_col_useage("PnSCTSTAT", 0,0,0,0)
        end if

        if(ii==iHK3) then ! 20210117 IMS: This column was only just introduced. We want to make it read-optional (at least for now) since some older event lists may not have it.
          call this%hkList(ii)%hkPerTM(ci)%ibset_col_useage("CnTCCDPT", isOptionalBit)
        end if

        this%hkList(ii)%hkPerTM(ci)%useageIsSet = .true.
        call this%hkList(ii)%hkPerTM(ci)%make_colnames_tel_unspecific(status)
        this%hkList(ii)%hkPerTM(ci)%header%keywords%useageFlags = 0
      end do
    end do

    do ii=1,size(this%gtiList)
      do kk=1,size(this%ccdnr_infile)
        ci = this%ccdnr_infile(kk)
        do jj=1,numBits
          call this%gtiList(ii)%gtiPerTM(ci)%ibset_all_useage(flagbit_list(jj))
        end do
        this%gtiList(ii)%gtiPerTM(ci)%useageIsSet = .true.

        if(this%gtiList(ii)%ext_type==iFLAREGTI) then ! IMS 14 May 2020: header keywords appear not to be written to FLAREGTI headers.
          this%gtiList(ii)%gtiPerTM(ci)%header%keywords%useageFlags = 0
        end if
      end do
    end do

    if(present(exts_to_flag)) then
      do kk=1,size(exts_to_flag)
        type_found = .false.
        do ii=1,size(this%extList)
          if(.not.type_found .and. this%extList(ii)%ext_type>0&
          .and. this%extList(ii)%ext_type==exts_to_flag(kk)) then
            type_found = .true.
            do jj=1,numBits
              this%extList(ii)%useageFlags = ibset(this%extList(ii)%useageFlags, flagbit_list(jj))
            end do
        exit
          end if
        end do

        if(.not.type_found) then
          iStr = int2str(exts_to_flag(kk))
call wrstop(functionName, "Extension type ID "//trim(iStr)//" not present.")
        end if
      end do

    else ! flag all ext types.
      do ii=1,size(this%extList)
        do jj=1,numBits
          this%extList(ii)%useageFlags = ibset(this%extList(ii)%useageFlags, flagbit_list(jj))
        end do
      end do
    end if

    ! Flag all HK types.
    do ii=1,size(this%hkList)
      do jj=1,numBits
        this%hkList(ii)%useageFlags = ibset(this%hkList(ii)%useageFlags, flagbit_list(jj))
      end do
    end do

    if(present(gti_exts_to_flag)) then
      do kk=1,size(gti_exts_to_flag)
        type_found = .false.
        do ii=1,size(this%gtiList)
          if(.not.type_found .and. this%gtiList(ii)%ext_type>0&
          .and. this%gtiList(ii)%ext_type==gti_exts_to_flag(kk)) then
            type_found = .true.
            do jj=1,numBits
              this%gtiList(ii)%useageFlags = ibset(this%gtiList(ii)%useageFlags, flagbit_list(jj))
            end do
        exit
          end if
        end do

        if(.not.type_found) then
          iStr = int2str(gti_exts_to_flag(kk))
call wrstop(functionName, "GTI extension type ID "//trim(iStr)//" not present.")
        end if
      end do

    else ! flag all gti ext types.
      do ii=1,size(this%gtiList)
        do jj=1,numBits
          this%gtiList(ii)%useageFlags = ibset(this%gtiList(ii)%useageFlags, flagbit_list(jj))
        end do
      end do
    end if

  end subroutine setup_for_readwrite_tEventFile

  !.....................................................................
  subroutine read_tEventFile(this, status, want_wcs, filename, nowarn, repair_gtis)
    implicit none

    class(tEventFile) :: this
    integer,                intent(inout) :: status
    logical,      optional, intent(in)    :: want_wcs
    character(*), optional, intent(in)    :: filename
    logical,      optional, intent(in)    :: nowarn, repair_gtis

    character(*), parameter :: functionName = "read_tEventFile"
    integer :: ii,ierr
    character(buffsize) :: fileName2

    fileName2 = ""

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not. this%isInitialized)&
call wrstop(functionName, "You must initialize the event-file object before attempting to read it.")

    if(present(filename)) call this%open(FF_READONLY, status, filename=filename)

    call this%info%getFileName(fileName2)
    call wrchatty(functionName, "  Moving to the EVENTS extension.")
    ierr = 0
    call FTMNHD(this%info%unitNo, BINARY_TBL, this%events%info%tableName, 0, ierr)
    if(ierr==301) then
call wrstop(functionName, "EVENTS extension not found in file "//trim(fileName2))
    else
      call report_FITS_err(functionName, ierr, s_err, "- extension "//trim(this%events%info%tableName))
    end if

    call this%read_stdheader(status) ! this sets various tEventFile attributes including the filter records and this%ccdnr_infile

    call this%read_events(status, want_wcs)

    ! Exts: set read flags then read.
    !
    do ii=1,size(this%extList)
      if(btest(this%extList(ii)%useageFlags,readBit)) then
        call read_extensions(this%ccdnr_infile, this%std_header%mjdref, this%extList(ii), status)
      end if
    end do

    ! HK: set read flags then read.
    !
    do ii=1,numHKTypes
      if(btest(this%hkList(ii)%useageFlags,readBit)) then
        call read_hks(this%ccdnr_infile, this%hkList(ii), status)
      end if
    end do

    ! GTI: set read flags then read.
    !
    do ii=1,size(this%gtiList)
      if(btest(this%gtiList(ii)%useageFlags,readBit)) then
        call read_gtis(this%ccdnr_infile, this%std_header%mjdref&
          , this%gtiList(ii), status, nowarn, repair_gtis)
      end if
    end do

!**** doesn't read AUX_EVENTS!

    call this%read_rawxy_sel(status)

    if(present(filename)) call this%close(status)

  end subroutine read_tEventFile

  !.....................................................................
  subroutine read_rawxy_sel_tEventFile(this, status)
    implicit none

    class(tEventFile) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "read_rawxy_sel_tEventFile"
    logical :: extension_exists=.false.

    if(status > 100 .or. status == s_errRep)&
  return

    this%filtered_rawxy = .false. ! default

    call this%rawxy_mask%read(this%info%unitNo, rawxy_ext, extension_exists, status)

    if(status == 0 .and. extension_exists) then
      if(    size(this%rawxy_mask%values,1) /= N_PIXELS_CCD&
        .or. size(this%rawxy_mask%values,2) /= N_PIXELS_CCD) then

          call wrerr(functionName, s_err, "RAWX/Y selection image has wrong dimensions")
          status = s_errRep
  return
      end if

      this%filtered_rawxy = .true.
    end if

  end subroutine read_rawxy_sel_tEventFile

  !.....................................................................
  subroutine read_events_tEventFile(this, status, want_wcs)

!*** too little meat on this now to justify a separate routine.

    implicit none

    class(tEventFile) :: this
    integer,           intent(inout) :: status
    logical, optional, intent(in)    :: want_wcs

    character(*), parameter :: functionName = "read_events_tEventFile"

    if(status > 100 .or. status == s_errRep)&
  return

    if(present(want_wcs)) then
      this%events%want_wcs = want_wcs
    else
      this%events%want_wcs = .false.
    end if

    call this%events%read(status)
    call this%events%read_extras(status) ! reads WCS if this%want_wcs.

  end subroutine read_events_tEventFile

  !.....................................................................
  subroutine write_stdheader_tEventFile(this, status)
    implicit none

    class(tEventFile) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "write_stdheader_tEventFile"
    character(buffsize) :: sass_parint
    character(buffsize) :: fileName

    fileName = ""

    if(status > 100 .or. status == s_errRep)&
  return

    call this%info%getFileName(fileName)

    call get_environment_variable("SASS_PARINT", sass_parint)
    if (trim(sass_parint) == "PIPELINE" .or. trim(sass_parint) == "pipeline") then
      call update_stdhdr(fileName, this%std_header, status)
    end if

    ! keywords written by write stdheader should be written with update routine instead of write routine
    call fwrite_stdhdr(this%info%unitNo, this%std_header, status)

  end subroutine write_stdheader_tEventFile

  !.....................................................................
  subroutine write_tEventFile(this, status, filename, clobber)
    implicit none

    class(tEventFile) :: this
    integer,                intent(inout) :: status
    character(*), optional, intent(in)    :: filename
    logical,      optional, intent(in)    :: clobber

    character(*), parameter :: functionName = "write_tEventFile"
    character(buffsize) :: sass_parint
    integer :: ii,jj,ci,ierr
    integer, dimension(1) :: dummy_naxes
    character(buffsize) :: fileName2

    dummy_naxes(1) = 0
    fileName2 = ""

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not. this%isInitialized) then
      call wrerr(functionName, s_err, "You must first initialize the event file.")
      status = s_errRep
  return
    end if

    if(present(filename)) call this%open(FF_CREATE, status, filename=filename, clobber=clobber)
    call this%info%getFileName(fileName2)

    ! update std_header with filename
    call get_environment_variable("SASS_PARINT", sass_parint)
    if (trim(sass_parint) == "PIPELINE" .or. trim(sass_parint) == "pipeline") then
      call update_stdhdr(fileName2, this%std_header, status)
    end if

    if(this%image_containsData) then
      call wrchatty(functionName, "Writing image.")
      call this%image%write(this%info%unitNo, status)
      call fwrite_stdhdr(this%info%unitNo,this%std_header,status)
    else
      ierr = 0
      call FTIIMG(this%info%unitNo, 32, 1, dummy_naxes, ierr) ! 20200429 IMS: FITSIO cocks up if we try to write a keyword to the PHDU before creating some sort of image.
      call report_FITS_err(functionName, ierr, status)
      call fwrite_stdhdr(this%info%unitNo,this%std_header,status)
    end if

    if(this%write_events) then
      if(this%events%isAllocated()) then
        call wrchatty(functionName, "Writing events.")
        call this%events%write(status)
        call wrchatty(functionName, "Writing events-extension extras.")
        call this%events%write_extras(status)
      else
        call wrlog(functionName, "Write of events was requested, but the events array is not allocated.")
      end if
    else
      call wrlog(functionName, "Write of events was not requested; no events written.")
    end if

    call wrchatty(functionName, "Writing std header.")
    call fwrite_stdhdr(this%info%unitNo, this%std_header, status)
    ! write OGIP HDUCLAS1
    call ffits_update_key_str(this%info%unitNo, "HDUCLAS1", "EVENTS", "", status)

    if(this%write_events) then
      do ii=1,size(this%extList)
        do jj = 1, size(this%ccdnr_infile)
          ci = this%ccdnr_infile(jj)

          if(.not.this%extList(ii)%extPerTM(ci)%isAllocated())&
        cycle !**** should label when are nested loops.

          call wrchatty(functionName, "Writing extension "&
            //trim(this%extList(ii)%extPerTM(ci)%info%tableName)//".")

          call this%extList(ii)%extPerTM(ci)%write(status)
          call this%write_stdheader(status)
        end do
      end do

      do ii=1,numHKTypes
        if(this%hkList(ii)%containsData()) then
          do jj = 1, size(this%ccdnr_infile)
            ci = this%ccdnr_infile(jj)

            if(.not.this%hkList(ii)%hkPerTM(ci)%isAllocated())&
          cycle

            call wrchatty(functionName, "Writing extension "&
              //trim(this%hkList(ii)%hkPerTM(ci)%info%tableName)//".")
            call this%hkList(ii)%hkPerTM(ci)%write(status)
            call this%write_stdheader(status)
          end do
        end if
      end do

    else
      call wrlog(functionName, "Since events were not requested, HK and other extensions are also not written.")
    end if

    TYP: do ii=1,size(this%gtiList)
      if(.not.(this%write_events .or. this%gtiList(ii)%ext_type==iGTI))&
    cycle TYP

      CCD: do jj = 1, size(this%ccdnr_infile)
        ci = this%ccdnr_infile(jj)

        if(.not.this%gtiList(ii)%gtiPerTM(ci)%isAllocated())&
      cycle CCD

        call wrchatty(functionName, "Writing extension "//trim(this%gtiList(ii)%gtiPerTM(ci)%info%tableName)//".")
        call this%gtiList(ii)%gtiPerTM(ci)%write(status)
        call this%write_stdheader(status)
      end do CCD
    end do TYP

    if(this%regionObj%containsData) then
      call wrchatty(functionName, "Writing REGION extension.")
      call this%regionObj%write("REGION", status, this%info)
      call this%write_stdheader(status)
    end if

    if(this%filtered_rawxy)&
      call this%rawxy_mask%write(this%info%unitNo, status)

    if(present(filename)) call this%close(status)

  end subroutine write_tEventFile

  !.....................................................................
  subroutine destroy_tEventFile(this, status)
    implicit none

    class(tEventFile) :: this
    integer, intent(inout) :: status

    integer :: ii,ci
    character(*), parameter :: functionName = "destroy_tEventFile"

    call this%events%destroy(status)
    call this%image%free(status)

    do ii=1,size(this%extList)
      do ci = 1, size(this%extList(ii)%extPerTM)
        call this%extList(ii)%extPerTM(ci)%destroy(status)
      end do
    end do

    do ii=1,numHKTypes
      do ci = 1, size(this%hkList(ii)%hkPerTM)
        call this%hkList(ii)%hkPerTM(ci)%destroy(status)
      end do
    end do

    do ii=1,size(this%gtiList)
      do ci = 1, size(this%gtiList(ii)%gtiPerTM)
        call this%gtiList(ii)%gtiPerTM(ci)%destroy(status)
      end do
    end do

    call this%regionObj%free(status)
    call this%rawxy_mask%free(status)
    call this%externalgti%dealloc(status)

    call dealloc(this%ccdnr_infile, status)

  end subroutine destroy_tEventFile

  !.....................................................................
  subroutine sort_events_tEventFile(this, status)
    implicit none

    class(tEventFile) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "sort_events_tEventFile"
!    integer, dimension(5) :: sort_sequence

    if(status > 100 .or. status == s_errRep)&
  return

!    sort_sequence(1) = this%events%getColNum("TM_NR")
!    sort_sequence(2) = this%events%getColNum("PHA")
!    sort_sequence(3) = this%events%getColNum("RAWX")
!    sort_sequence(4) = this%events%getColNum("RAWY")
!    sort_sequence(5) = this%events%getColNum("TIME")

!    call this%events%sort_in_sequence(status, sort_sequence)
    call this%events%sort_in_sequence(["TIME ","RAWY ","RAWX ","PHA  ","TM_NR"], status)

  end subroutine sort_events_tEventFile

  !.....................................................................
  subroutine apply_filter_tm_GTIs_tEventFile(this, extI, status, overlap, do_events)
    ! This filters the event list by each of the TM-specific GTIs of type specified by extI - i.e. those in the GTIn extensions, originally derived from HK; those in the QUALGTIn extensions, created (if they exist) by the task of that ilk; or those in FLAREGTIn extensions. After event filtering, the GTIs are padded slightly with an overlap factor before being merged (logical AND) with those of 'higher order' in the GTI type food chain.
    !
    implicit none

    class(tEventFile) :: this
    integer,            intent(in)    :: extI
    integer,            intent(inout) :: status
    real(r8), optional, intent(in)    :: overlap
    logical,  optional, intent(in)    :: do_events

    character(*), parameter :: function = "apply_filter_tm_GTIs_tEventFile"
    type(tGTI_tab) :: gti_with_overlap
    integer :: ci,jj
    logical :: local_do_events!,local_do_sort
!    real(r8), dimension(:), pointer :: times=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not.this%gtiList(extI)%containsData()) then
      call wrerr(function, warn, trim(this%gtiList(extI)%extNameRoot)&
        //" selected as filter but the file doesn't contain these extensions. "&
        //"This is now treated as empty GTIs, thus discarding everything.")
      call this%apply_empty_GTI(extI, status, do_events)
  return
    end if

    if(present(do_events)) then
      local_do_events = do_events
    else
      local_do_events = .true. ! for backward-compatibility. Just in case.
    end if

!    if(local_do_events) then
!      call this%events%get_col_ptr("TIME", times)
!      call this%events%sort(times, status)
!      nullify(times)
!    end if

    do jj = 1, size(this%ccdnr_infile)
      ci = this%ccdnr_infile(jj)
      if(local_do_events)&
        call this%events%apply_gti(this%gtiList(extI)%gtiPerTM(ci), this%std_header%mjdref, status)

      call get_gti_with_overlap(this%gtiList(extI)%gtiPerTM(ci), gti_with_overlap, overlap)
      call this%apply_tm_GTI(this%gtiList(extI)%gtiPerTM(ci), gti_with_overlap, ci, status, extI)
    end do

    call gti_with_overlap%destroy(status)

  end subroutine apply_filter_tm_GTIs_tEventFile

  !.....................................................................
  subroutine apply_empty_GTI_tEventFile(this, extI, status, do_events)
    ! Throws everything away. Whee!
    !
    implicit none

    class(tEventFile) :: this
    integer,            intent(in)    :: extI
    integer,            intent(inout) :: status
    logical,  optional, intent(in)    :: do_events

    character(*), parameter :: function = "apply_empty_GTI_tEventFile"
    type(tGTI_tab) :: tmp_GTI
    integer :: ci,jj
    logical :: local_do_events

    if(status > 100 .or. status == s_errRep)&
  return

    if(present(do_events)) then
      local_do_events = do_events
    else
      local_do_events = .true. ! for backward-compatibility. Just in case.
    end if

    call setup_tmp_GTI(0, tmp_GTI, status)
    call tmp_GTI%header%setKeywordValue("MJDREF", this%std_header%mjdref)
    call tmp_GTI%sanityCheck(status, do_warn=.false.) ! just to set the flag that it is done.

    do jj = 1, size(this%ccdnr_infile)
      ci = this%ccdnr_infile(jj)
      if(local_do_events)&
        call this%events%apply_gti(tmp_GTI, this%std_header%mjdref, status)

      call this%apply_tm_GTI(tmp_GTI, tmp_GTI, ci, status, extI)
    end do

    call tmp_GTI%destroy(status)

  end subroutine apply_empty_GTI_tEventFile

  !.....................................................................
  subroutine apply_filter_nontm_GTI_tEventFile(this, externalgti, status, overlap)
    ! This does similar to apply_filter_tm_GTIs_tEventFile() except the GTI, which should apply equally across all TMs, is supplied directly.
!*** it's presently hard-coded to print messages at least concerning only externalgti
    !
    implicit none

    class(tEventFile)                 :: this
    type(tGTI_tab),     intent(in)    :: externalgti
    integer,            intent(inout) :: status
    real(r8), optional, intent(in)    :: overlap

    character(*), parameter :: functionName = "apply_filter_nontm_GTI_tEventFile"
    integer :: ci,jj
    type(tGTI_tab) :: gti_with_overlap

    if(status > 100 .or. status == s_errRep)&
  return

    call wrchatty(functionName, "-- external gti filter of events")
    call this%events%apply_gti(externalgti, this%std_header%mjdref, status)
    call get_gti_with_overlap(externalgti, gti_with_overlap, overlap)
    call wrchatty(functionName, "-- external gti filter of other extensions")
    do jj = 1, size(this%ccdnr_infile)
      ci = this%ccdnr_infile(jj)
      call this%apply_tm_GTI(externalgti, gti_with_overlap, ci, status)
    end do

  end subroutine apply_filter_nontm_GTI_tEventFile

  !.....................................................................
  subroutine apply_tm_GTI_tEventFile(this, gti, gti_with_overlap, tel_id, status, presentExtI)
    ! This does 2 things with the supplied GTI:
    !	(1) Merges the GTI (logical AND) with those of the specified TM which also have extI type numbers < presentExtI.
    !	(2) Filters (using the padded GTI) the extensions CORRATT, BADPIX, DEADCORR plus the HK extensions (all of specified TM).
!*** change presentExtI to a list of extI integers for the GTI types to AND with the submitted one - with the action to do all of them if the list is not submitted.
!*** make gti_with_overlap optional, call get_gti_with_overlap() if it is not present.
    implicit none

    class(tEventFile)                 :: this
    type(tGTI_tab),     intent(in)    :: gti,gti_with_overlap
    integer,            intent(in)    :: tel_id
    integer,            intent(inout) :: status
    integer, optional,  intent(in)    :: presentExtI

    integer :: last_extJ,extJ,ii

    if(present(presentExtI)) then
      last_extJ = presentExtI-1
    else
      last_extJ = size(this%gtiList)
    end if

    do extJ=1,last_extJ
      if(this%gtiList(extJ)%containsData())&
        call this%gtiList(extJ)%gtiPerTM(tel_id)%and_with(gti, status)
    end do

    do ii=1,size(this%extList)
      if(this%extList(ii)%containsData())&
        call this%extList(ii)%extPerTM(tel_id)%apply_gti(gti_with_overlap, status)
    end do

    do ii=1,numHKTypes
      if(this%hkList(ii)%containsData())&
        call this%hkList(ii)%hkPerTM(tel_id)%apply_gti(gti_with_overlap, status)
    end do

  end subroutine apply_tm_GTI_tEventFile

  !.....................................................................
  subroutine apply_filter_region_tEventFile(this, regionObj, status)
    implicit none

    class(tEventFile) :: this
    type(tRegions), intent(in)    :: regionObj
    integer,        intent(inout) :: status

    character(*), parameter :: function = "apply_filter_region_tEventFile"
    integer(i4), dimension(:), allocatable :: iMask
    integer :: ii, numRows
    real(r8), dimension(:), pointer :: ras=>null(),decs=>null(),xs=>null(),ys=>null()
    integer(i2), dimension(:), pointer :: rawxs=>null(),rawys=>null()

    ! for using srctool spatial region selections
    type(region_type) :: srctool_region
    type(coord_type) :: srctool_coord

    if(status > 100 .or. status == s_errRep)&
  return

    this%regionObj = regionObj

    numRows = this%events%getNumRows()
    call alloc(iMask, numRows, status)
    iMask = 0

    call this%regionObj%get_srctool_region(srctool_region, status)

    call this%events%get_col_ptr("RA",   ras)
    call this%events%get_col_ptr("DEC",  decs)
    call this%events%get_col_ptr("X",    xs)
    call this%events%get_col_ptr("Y",    ys)
    call this%events%get_col_ptr("RAWX", rawxs)
    call this%events%get_col_ptr("RAWY", rawys)

    ! TODO update when srctool can handle the other formats
    ! THIS ASSUMES THAT ONLY ONE TYOE OF COORDSYS in all regions
    do ii = 1, numRows
      srctool_coord%long = ras(ii)
      srctool_coord%lat  = decs(ii)
      srctool_coord%x    = xs(ii)
      srctool_coord%y    = ys(ii)
      srctool_coord%rawx = rawxs(ii)
      srctool_coord%rawy = rawys(ii)
      srctool_coord%coordsys_code = COORDSYS_CODE_ICRS
      ! test
      !      if(srctool_region%pregcomp(1)%coordsys_code == COORDSYS_CODE_PHYSICAL) then
      !        srctool_coord%coordsys_code = COORDSYS_CODE_PHYSICAL
      !      end if

      call srctool_regions_test_if_coord_in_region(srctool_coord, srctool_region, iMask(ii), status)
      if(status /= ok) status = s_errRep
    end do

    call this%events%pack((iMask>0), status)

    call dealloc(iMask, status)

    nullify(ras)
    nullify(decs)
    nullify(xs)
    nullify(ys)
    nullify(rawxs)
    nullify(rawys)

  end subroutine apply_filter_region_tEventFile

  !.....................................................................
  subroutine apply_stdheader_tEventFile(this, status)
    implicit none

    class(tEventFile) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "apply_stdheader_tEventFile"
    integer :: ii, ldigit
    character(buffsize), dimension(:), allocatable :: instrum

    if(status > 100 .or. status == s_errRep)&
  return

    this%filtered_pattern = .false.
    this%filtered_flag    = .false.
    this%filtered_gti     = .false.
    this%filtered_energy  = .false.
    this%filtered_telid   = .false.
    if(this%std_header%pat_sel /= 0)      this%filtered_pattern = .true.
    if(this%std_header%flagsel /= 0)      this%filtered_flag    = .true.
    if(this%std_header%gti_sel /= "NONE") this%filtered_gti     = .true.
    if(this%std_header%nenergy /= 0)      this%filtered_energy  = .true.
    if(this%std_header%ninst   /= 0)      this%filtered_telid   = .true.

    if(allocated(this%ccdnr_infile)) then
      if(size(this%ccdnr_infile) /= this%std_header%ninst)&
call wrstop(functionName, "Existing this%ccdnr_infile has a size which is not equal to std_header%ninst.")

    else
      call alloc(instrum, this%std_header%ninst, status)
      call alloc(this%ccdnr_infile, this%std_header%ninst, status)
    
      if(this%std_header%ninst == 0)&
  return

      if(this%std_header%ninst > 0) then
        instrum = this%std_header%instrumx(:this%std_header%ninst)
      end if

      do ii = 1, this%std_header%ninst
        ldigit = len_trim(instrum(ii))
        this%ccdnr_infile(ii) = str2int(instrum(ii)(ldigit:ldigit))
      end do

      deallocate(instrum)
    end if

  end subroutine apply_stdheader_tEventFile

  !.....................................................................
  subroutine apply_filter_tEventFile(this, filter_obj, status, image_specs, do_sort)
    ! Filters the event list but also then constructs an image, if required.
    !
    implicit none

    class(tEventFile)                         :: this
    type(t_filter),      intent(in)           :: filter_obj
    integer,             intent(inout)        :: status
    type(t_image_specs), intent(in), optional :: image_specs
    logical,             intent(in), optional :: do_sort

    character(*), parameter :: functionName = "apply_filter_tEventFile"
    integer :: nenergy,ci,ext_i,ii,jj
    character(buffsize) :: msg
    logical :: local_sort

    if(status > 100 .or. status == s_errRep)&
  return

    local_sort = .false.
    if(present(do_sort)) local_sort = do_sort
    if(local_sort) call this%sort_events(status)

    if(filter_obj%do_filter_telid .and. status == ok) then
      call wrchatty(functionName, "telid filter")
      call this%events%apply_telid(this%ccdnr_infile, status)
    end if

    if(filter_obj%do_filter_energy .and. status == ok) then
      call wrchatty(functionName, "energy filter")
      nenergy = this%std_header%nenergy

      ! Only apply energy unit conversion on std_header here, since
      ! std_header is keV while energy column are in eV.
      !
      call this%events%apply_energy(&
          this%std_header%eminx(:nenergy)*1000.0_r8&
        , this%std_header%emaxx(:nenergy)*1000.0_r8, status)
    end if

    if(filter_obj%do_filter_flag .and. status == ok) then
      if(filter_obj%flag_invert) then
        call wrerr(functionName, warn&
          , "Cannot calculate proper exposure if inversion of flag-mask is requested!")
      else
        call wrchatty(functionName, "Chopping holes in GTI where whole frame is flag-filtered.")
!        call this%delete_flagged_frames_from_GTI(this%std_header%flagsel, status, .true.)
        call this%delete_flagged_frames_from_GTI(filter_obj%flag_mask, status, .true.)
        ! => delete_frames_tEventFile
      end if

      call wrchatty(functionName, "flag filter")
!      call this%events%apply_flag(this%std_header%flagsel, status, invert=filter_obj%flag_invert)
      call this%events%apply_flag(filter_obj%flag_mask, status, invert=filter_obj%flag_invert)
    end if

    if(filter_obj%do_filter_pattern .and. status == ok) then
      call wrchatty(functionName, "pattern filter")
      call this%events%apply_pattern(this%std_header%pat_sel, status)
    end if

    if(filter_obj%do_filter_region .and. status == ok) then
      call wrchatty(functionName, "region filter")
      call this%apply_filter_region(filter_obj%regionObj, status)

!!!      this%regionObj%reg_file%info = this%info

      ! if its a "physical" filter a region gti cannot be derived -> no conversion from rawx/y to ra/dec
      ! it is enough to check first region, since only one type of wcs is allowed
      if(this%regionObj%regRows(1)%wcs /= "physical") then
        call this%derive_region_gti(status)

        call this%apply_filter_tm_GTIs(iREGIONGTI, status, filter_obj%overlap, do_events=.false.)
        call this%adjust_GTI_bound_kwds(status)
      end if
    end if

    if(filter_obj%do_filter_gti .and. status == ok) then
      call wrchatty(functionName, "gti filter")

      if(  filter_obj%gti_filter_type == gti_type&
      .or. filter_obj%gti_filter_type == qualgti_type&
      .or. filter_obj%gti_filter_type == flaregti_type) then
        select case(filter_obj%gti_filter_type)
          case(gti_type)
            write(msg, '(a)') "- internal gti filter"
            ext_i = iGTI
          case(qualgti_type)
            write(msg, '(a)') "- qual gti filter"
            ext_i = iQUALGTI
          case(flaregti_type)
            write(msg, '(a)') "- flare gti filter"
            ext_i = iFLAREGTI
          case default
call wrstop(functionName, "bug")
        end select

        call wrchatty(functionName, msg)
        call this%apply_filter_tm_GTIs(ext_i, status, filter_obj%overlap, do_events=.true.)
        call this%adjust_GTI_bound_kwds(status)

        jj = this%get_ext_index("gti", iFLAREGTI)
        do ii=1,size(this%ccdnr_infile)
          ci = this%ccdnr_infile(ii)
          call this%gtiList(jj)%gtiPerTM(ci)%header%setKeywordValue("TSTART"&
            , this%std_header%tstart)
          call this%gtiList(jj)%gtiPerTM(ci)%header%setKeywordValue("TSTOP"&
            , this%std_header%tstop)
        end do

      else if(filter_obj%gti_filter_type == externalgti_type) then
        call wrchatty(functionName, "- external gti filter")
        call this%apply_filter_nontm_GTI(filter_obj%externalgti, status, filter_obj%overlap)
        ! Note that we can't call apply_filter_tm_GTIs() with GTI type iEXTERNAL because this call runs over all the TMs.
        call this%adjust_GTI_bound_kwds(status)
        call filter_obj%externalgti%header%setKeywordValue("TSTART",this%std_header%tstart)
        call filter_obj%externalgti%header%setKeywordValue("TSTOP", this%std_header%tstop)

      else
call wrstop(functionName, "Unrecognized GTI filter type.")
      end if

      call filter_obj%externalgti%destroy(status)
    end if

    if(filter_obj%do_filter_rawxy .and. status == ok) then
      call wrchatty(functionName, "RAWX/Y filter")
      call this%events%apply_rawxy(filter_obj%rawxy_sel, status)
    end if

    if(present(image_specs) .and. status == ok) then
      if(.not. this%events%wcs_containsData) then
        call wrerr(functionName, c_err, "No X/Y WCS was read from the events header, can't make image.")
  return
      end if

      call wrlog(functionName, "Creating image.")
      call this%image%create_from_events(this%events, image_specs, status)
      this%image_containsData = .true.
    end if

  end subroutine apply_filter_tEventFile

  !.....................................................................
  subroutine get_prev_chopper_value(numHKRows, hk_times, hk_chopper&
    , default_chopper_value, eventTime, jj, chopper_value)
    ! This assumes that eventTimes will be submitted in an ascending sequence.

!*** 20210907 IMS: note that this is similar to routines in evprep. Indeed there is a general class of problems which involve incrementally matching two (assumed sorted) lists.

    implicit none

    integer,                        intent(in)    :: numHKRows
    real(r8), dimension(numHKRows), intent(in)    :: hk_times
    real(r4), dimension(numHKRows), intent(in)    :: hk_chopper
    real,                           intent(in)    :: default_chopper_value
    real(r8),                       intent(in)    :: eventTime
    integer,                        intent(inout) :: jj
    real(r4),                       intent(out)   :: chopper_value

    character(*), parameter :: functionName = "get_prev_chopper_value"

    if(jj<1 .or. jj>numHKRows)&
call wrstop(functionName, "Index to HK1 out of range.")

    if(jj==1 .and. eventTime<hk_times(1)) then
      chopper_value = default_chopper_value
  return
    end if

    do
      if(jj==numHKRows)&
    exit

      if(eventTime<hk_times(jj+1))&
    exit

      jj = jj + 1
    end do

    chopper_value = hk_chopper(jj)

  end subroutine get_prev_chopper_value

  !.....................................................................
  subroutine delete_frames_tEventFile(this, flagsel, status, check_sort)
    ! This deletes from GTI any frame flagged bad, and filtered.
    implicit none

    class(tEventFile)             :: this
    integer, intent(in)           :: flagsel
    integer, intent(inout)        :: status
    logical, intent(in), optional :: check_sort

    character(*), parameter :: functionName = "delete_frames_tEventFile"
    real(r4), parameter :: margin_frac = 1.05
    logical :: local_check_sort,scale_frame_length_by_chopper,frames_abut
    integer(i4) :: flag_mask
    integer :: n_events,n_non_null,ii,jj,ci,ei,fi,n_unique_frames,n_bad_frames,numHKRows,n_btis,bfi,fi0,fi1
    real(r8), dimension(:), allocatable :: temp_times,temp_times_tm
    character(20) :: str_value
    integer(i1), dimension(:), allocatable :: temp_tms
    integer(i4), dimension(:), allocatable :: temp_flag,temp_flag_tm,frame_id
    integer, dimension(:), allocatable :: u_frametime_indices
    real(r8), dimension(:), allocatable :: times_of_frame,starts,stops
    logical, dimension(:), allocatable :: frame_is_bad
    real(r4), dimension(:), allocatable :: frame_lengths
    type(tGTI_tab) :: BTIs
    real(r4) :: delta_t,av_frame_duration

    real(r4), dimension(:), pointer :: ptr_hk1_chopper=>null()
    real(r8), dimension(:), pointer ::&
       ptr_frametime =>null()&
      ,ptr_hk1_times =>null()&
      ,ptr_bti_starts=>null()&
      ,ptr_bti_stops =>null()
    logical, dimension(:), pointer :: frametime_is_null=>null()
    integer(i1), dimension(:), pointer :: ptr_tms=>null()
    integer(i4), dimension(:), pointer :: ptr_flag=>null()

    if(present(check_sort)) then
      local_check_sort = check_sort
    else
      local_check_sort = .false.
    end if

    flag_mask = 0_i4
    if(btest(flagsel, BIT_CORRUPT_FRAME)) flag_mask = ibset(flag_mask, BIT_CORRUPT_FRAME)
    if(btest(flagsel, BIT_ARTEFACT     )) flag_mask = ibset(flag_mask, BIT_ARTEFACT)

    if(flag_mask==0_i4)&
  return ! no selection on these frame-specific bits has been made.

    n_events = this%events%getNumCols()
    if(n_events < 1)&
  return

!    call this%events%get_col_ptr("FRAMETIME", ptr_frametime, frametime_is_null)
    call this%events%get_col_ptr("TIME", ptr_frametime, frametime_is_null)

    n_non_null = size(.not.frametime_is_null)
    if(n_non_null < 1) then
      nullify(ptr_frametime, frametime_is_null)
  return
    end if

    temp_times = pack(ptr_frametime, .not.frametime_is_null)

    if(local_check_sort .and. n_non_null>1) then
      if(any(temp_times(2:n_non_null) < temp_times(1:n_non_null-1)))&
call wrstop(functionName, "TIME values of the event table appear not to be sorted.")
    end if

    scale_frame_length_by_chopper = .false.
    call this%events%header%getKeywordValue("CHOPDEAD", str_value) !**** use fitsio routine.
    if(trim(str_value)=="T") scale_frame_length_by_chopper = .true.

    call this%events%get_col_ptr("TM_NR", ptr_tms)
    call this%events%get_col_ptr("FLAG",  ptr_flag)
    temp_tms  = pack(ptr_tms,  .not.frametime_is_null)
    temp_flag = pack(ptr_flag, .not.frametime_is_null)
    nullify(ptr_flag, ptr_tms)
    nullify(ptr_frametime, frametime_is_null)

    L_TMS: do ii=1,size(this%ccdnr_infile)
      ci = int(this%ccdnr_infile(ii), kind=i1)

      ! Get a list of unique frame times for this TM, then work out
      ! which of these needs to be omitted.
      !
      temp_times_tm = pack(temp_times, (temp_tms==ci))
      u_frametime_indices = unique(temp_times_tm, status)
      n_unique_frames = size(u_frametime_indices)

      if(n_unique_frames < 1)&
    cycle L_TMS

      call alloc(frame_is_bad,   n_unique_frames, status)
      call alloc(times_of_frame, n_unique_frames, status)
      call alloc(frame_id,       n_unique_frames, status)

      temp_flag_tm = pack(temp_flag, (temp_tms==ci))

      frame_id(:) = 0 ! these values should never be accessed, but it is better not to leave them hanging.
      frame_is_bad(:) = .false. ! default
      fi = 1
      times_of_frame(fi) = temp_times_tm(u_frametime_indices(fi))

      bfi = 0
      ei = 1
      if(iand(temp_flag_tm(ei), flag_mask) > 0_i4) then
        frame_is_bad(fi) = .true.
        bfi = bfi + 1
        frame_id(bfi) = fi
      end if

      ! Note: the following only works if the non-NULL FRAMETIMEs occur
      ! in increasing order.
      !
      L_EVENTS: do ei=2,size(temp_times_tm)
        if((temp_times_tm(ei) == times_of_frame(fi)))&
      cycle L_EVENTS

        ! If we get to here, we have run through all events in this
        ! frame. Look for the next frame.
        !
        fi = fi + 1
        times_of_frame(fi) = temp_times_tm(u_frametime_indices(fi))
        if(temp_times_tm(ei) /= times_of_frame(fi)) then
          if(local_check_sort) then
call wrstop(functionName, "Well, this is pretty weird. You'd better contact the developer.")
          else
call wrstop(functionName, "Maybe your EVENTS extension is not correctly sorted?")
          end if
        end if

        ! We assume that evprep has flagged all events of the frame with
        ! any frame-specific flag, so we can just test the first event
        ! with a new time. Subsequent events of this frame will test
        ! .true. in the following, thus will.
        !
        if(iand(temp_flag_tm(ei), flag_mask) /= 0_i4) then
          frame_is_bad(fi) = .true.
          bfi = bfi + 1
          frame_id(bfi) = fi
        end if
      end do L_EVENTS

      if(.not.any(frame_is_bad))&
    cycle L_TMS

      times_of_frame = pack(times_of_frame, frame_is_bad)
      n_bad_frames = count(frame_is_bad)
      call alloc(frame_lengths, n_bad_frames, status)
      frame_lengths = 1.0 ! we'll multiply them by the canonical frame
      ! duration of ~50 ms after dealing with chopper effects.

      ! Now we have to look at the chopper settings and the CHOPDEAD
      ! keyword to calculate the length of frame to cut out.
      !
      if(scale_frame_length_by_chopper) then
        call this%hkList(iHK1)%hkPerTM(ci)%get_col_ptr("PxCHOPPER", ptr_hk1_chopper)
        call this%hkList(iHK1)%hkPerTM(ci)%get_col_ptr("TIME",      ptr_hk1_times)
        numHKRows = this%hkList(iHK1)%hkPerTM(ci)%getNumRows()
        jj = 1
        L_FRAME: do bfi = 1, n_bad_frames
          call get_prev_chopper_value(numHKRows, ptr_hk1_times, ptr_hk1_chopper, 1.0&
            , times_of_frame(bfi), jj, frame_lengths(bfi))
        end do L_FRAME
        nullify(ptr_hk1_times)
        nullify(ptr_hk1_chopper)

        where(frame_lengths==0.0) frame_lengths=1.0 ! CHOPPER==0 implies =1.
      end if

      frame_lengths = frame_lengths*FRAME_DURATION

      ! Now set up BTIs covering the bad frames. We will try not to have
      ! separate BTI for frames that appear likely to be contiguous.
      !
      call alloc(starts, n_bad_frames, status)
      call alloc(stops,  n_bad_frames, status)

      bfi = 1
      n_btis = 1
      starts(n_btis) = times_of_frame(bfi) - real(0.5*frame_lengths(bfi),kind=r8)

      L_FRAME2: do bfi = 2, n_bad_frames
        fi0 = frame_id(bfi-1)
        fi1 = frame_id(bfi)

        frames_abut = .true. ! default
        if(fi1-fi0>1) then
          frames_abut = .false.
        else
          delta_t = real(times_of_frame(bfi) - times_of_frame(bfi-1), kind=r4)
          av_frame_duration = 0.5*(frame_lengths(bfi) + frame_lengths(bfi-1))
          if(delta_t > av_frame_duration*margin_frac)&
            frames_abut = .false.

        end if

        if(.not.frames_abut) then
          stops( n_btis) = times_of_frame(bfi-1) + real(0.5*frame_lengths(bfi-1),kind=r8)
          n_btis = n_btis + 1
          starts(n_btis) = times_of_frame(bfi  ) - real(0.5*frame_lengths(bfi  ),kind=r8)
        end if
      end do L_FRAME2

      stops(n_btis) = times_of_frame(n_bad_frames) + real(0.5*frame_lengths(n_bad_frames),kind=r8)

      call BTIs%init(iTemp, status)
      call BTIs%ibset_col_useage("START",TCOL_INTRN_BIT)
      call BTIs%ibset_col_useage("STOP", TCOL_INTRN_BIT)
      call BTIs%alloc(status, n_btis)
      call BTIs%get_col_ptr("START", ptr_bti_starts)
      call BTIs%get_col_ptr("STOP",  ptr_bti_stops)
      ptr_bti_starts = starts(:n_btis)
      ptr_bti_stops  = stops( :n_btis)
      nullify(ptr_bti_starts, ptr_bti_stops)
      call dealloc(starts, status)
      call dealloc(stops, status)

      call BTIs%sanityCheck(status)

      ! Finally, do GTI<ci> .and..not.BTI
      !
      call this%gtiList(iGTI)%gtiPerTM(ci)%and_with_not(BTIs, status)

      call BTIs%destroy(status)

      call dealloc(frame_lengths, status)
      call dealloc(temp_times_tm, status)
      call dealloc(u_frametime_indices, status)
      call dealloc(frame_is_bad, status)
      call dealloc(frame_id, status)
      call dealloc(times_of_frame, status)
      call dealloc(temp_flag_tm, status)
    end do L_TMS

    call dealloc(temp_tms, status)
    call dealloc(temp_flag, status)
    call dealloc(temp_times, status)

    call this%adjust_GTI_bound_kwds(status)

  end subroutine delete_frames_tEventFile

  !.....................................................................
  subroutine adjust_GTI_bound_kwds_tEventFile(this, status)
    ! If any value in one of the GTIn extensions has been changed (e.g.
    ! by filtering with an external GTI or a GTI extension of greater
    ! specificity), the TSTART and TSTOP keywords of the standard header
    ! need to be adjusted accordingly, and following from that, also
    ! the DATE-OBS and DATE-END keywords.
    !
    implicit none

    class(tEventFile)      :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "adjust_GTI_bound_kwds_tEventFile"
    integer :: type_i,n_rows,ii,ci,n_rows_ci
    real(r8) :: min_time,max_time
    real(r8), dimension(:), pointer :: ptr_starts=>null(),ptr_stops=>null()

    call wrchatty(functionName, "Set new DATE-OBS, DATE-END keyword values.")

    type_i = this%get_ext_index("gti", iGTI)

    n_rows = 0
    min_time = -huge(1.0d0) ! These defaults are arbitrary since, with the
    max_time =  huge(1.0d0) ! code as written, they will never be accessed.
    do ii=1,size(this%ccdnr_infile)
      ci = this%ccdnr_infile(ii)

      n_rows_ci = this%gtiList(type_i)%gtiPerTM(ci)%getNumRows()
      if(n_rows_ci > 0) then
        call this%gtiList(type_i)%gtiPerTM(ci)%get_col_ptr("START", ptr_starts)
        call this%gtiList(type_i)%gtiPerTM(ci)%get_col_ptr("STOP",  ptr_stops)

        if(n_rows > 0) then ! means we have previous values of min_time and max_time.
          min_time = min(min_time, minval(ptr_starts))
          max_time = max(max_time, maxval(ptr_stops))
        else
          min_time = minval(ptr_starts)
          max_time = maxval(ptr_stops)
        end if

        nullify(ptr_starts,ptr_stops)
      end if

      n_rows = n_rows + n_rows_ci
    end do

    if(n_rows > 0) then
      this%std_header%tstart = min_time
      this%std_header%tstop  = max_time
    else
      ! Set TSTOP = TSTART. This is arguably better (and certainly
      ! easier, with the present interface) than either setting the
      ! keywords to NULL or omitting them altogether, both of which
      ! would require downstream tasks to include special behaviour to
      ! cope.
      !
      this%std_header%tstop = this%std_header%tstart
    end if

    call sec2date(this%std_header%tstart, this%std_header%mjdref&
      , this%std_header%date_obs, status)
    call sec2date(this%std_header%tstop,  this%std_header%mjdref&
      , this%std_header%date_end, status)

  end subroutine adjust_GTI_bound_kwds_tEventFile

  !.....................................................................
  subroutine derive_region_gti_tEventFile(this, status)
    implicit none

    class(tEventFile)      :: this
    integer, intent(inout) :: status

    character(*), parameter :: function = "derive_region_gti_tEventFile"
    integer :: ii,ci

    if(status > 100 .or. status == s_errRep)&
  return

    if(size(this%regionObj%regRows) <= 0)&
  return
    if(.not.this%extList(extCORRATT)%containsData()) then
      call wrlog(function, "Region gti cannot be derived / no CORRAT ext was found")
  return
    end if

    do ii = 1, size(this%ccdnr_infile)
      ci = this%ccdnr_infile(ii)
      call this%extList(extCORRATT)%extPerTM(ci)%derive_region_GTI(this%regionObj&
        , this%gtiList(iREGIONGTI)%gtiPerTM(ci), status, this%info)
    end do

  end subroutine derive_region_gti_tEventFile

  !.....................................................................
  subroutine record_rawxy_sel_tEventFile(this, rawxy_sel, status, is_new)
    implicit none

    class(tEventFile) :: this
    type(tRawXYSelection), intent(in)    :: rawxy_sel
    integer,               intent(inout) :: status
    logical, optional,     intent(in)    :: is_new

    character(*), parameter :: functionName = "record_rawxy_sel_tEventFile"

    integer :: xiLo,xiHi,yiLo,yiHi
    integer, dimension(2) :: naxis
!*** instead declare it a parameter in erolib_general?

    if(status > 100 .or. status == s_errRep)&
  return

    if(present(is_new)) then
      if(is_new) then ! Can't concatenate the two 'ifs' since with Fortran there is no guarantee that B is not evaluated if .not.A in an expression such as 'A .and. B'.
        naxis(1) = N_PIXELS_CCD
        naxis(2) = N_PIXELS_CCD
        call this%rawxy_mask%init(status)
        this%rawxy_mask%extname = rawxy_ext
        this%rawxy_mask%values = this%rawxy_mask%all_bits_set
      end if
    end if

    if(allocated(rawxy_sel%mask%values)) then
      if(rawxy_sel%invert) then
        if(rawxy_sel%ccdnr_defined) then
          where(rawxy_sel%mask%values)&
            this%rawxy_mask%values = ibclr(this%rawxy_mask%values, rawxy_sel%ccdnr-1)
        else ! .not. rawxy_sel%ccdnr_defined. Clear all CCD bits inside the selected rectangle (including the boundary).
          where(rawxy_sel%mask%values)&
            this%rawxy_mask%values = 0
        end if
      else
        if(rawxy_sel%ccdnr_defined) then
          where(.not.rawxy_sel%mask%values)&
            this%rawxy_mask%values = ibclr(this%rawxy_mask%values, rawxy_sel%ccdnr-1)
        else ! .not. rawxy_sel%ccdnr_defined. Clear all CCD bits inside the selected rectangle (including the boundary).
          where(.not.rawxy_sel%mask%values)&
            this%rawxy_mask%values = 0
        end if
      end if

  return
    end if

    if(rawxy_sel%axes(1)%lo%is_defined) then
      xiLo = rawxy_sel%axes(1)%lo%value
    else
      xiLo = 1
    end if
    if(rawxy_sel%axes(1)%hi%is_defined) then
      xiHi = rawxy_sel%axes(1)%hi%value
    else
      xiHi = N_PIXELS_CCD
    end if
    if(rawxy_sel%axes(2)%lo%is_defined) then
      yiLo = rawxy_sel%axes(2)%lo%value
    else
      yiLo = 1
    end if
    if(rawxy_sel%axes(2)%hi%is_defined) then
      yiHi = rawxy_sel%axes(2)%hi%value
    else
      yiHi = N_PIXELS_CCD
    end if

!*** Note that the output image array this%image seems to be constructed in evtool_image_mod.create_from_events_tImage2D_i4() with the indices in (x,y) order rather than (y,x). I have followed that convention in constructing this%rawxy_mask%values, but it probably warrants further digging - for example, to see what order the WCS keywords come in: which is 1 and which 2.

    if(rawxy_sel%invert) then
      if(rawxy_sel%ccdnr_defined) then
        this%rawxy_mask%values(xiLo:xiHi,yiLo:yiHi) = ibclr(this%rawxy_mask%values(xiLo:xiHi,yiLo:yiHi), rawxy_sel%ccdnr-1)
      else ! .not. rawxy_sel%ccdnr_defined. Clear all CCD bits inside the selected rectangle (including the boundary).
        this%rawxy_mask%values(xiLo:xiHi,yiLo:yiHi) = 0
      end if
    else
      if(rawxy_sel%ccdnr_defined) then
        this%rawxy_mask%values(:,1:yiLo-1)&
          = ibclr(this%rawxy_mask%values(:,1:yiLo-1),            rawxy_sel%ccdnr-1)
        this%rawxy_mask%values(1:xiLo-1,:)&
          = ibclr(this%rawxy_mask%values(1:xiLo-1,:),            rawxy_sel%ccdnr-1)
        this%rawxy_mask%values(xiHi+1:N_PIXELS_CCD,:)&
          = ibclr(this%rawxy_mask%values(xiHi+1:N_PIXELS_CCD,:), rawxy_sel%ccdnr-1)
        this%rawxy_mask%values(:,yiHi+1:N_PIXELS_CCD)&
          = ibclr(this%rawxy_mask%values(:,yiHi+1:N_PIXELS_CCD), rawxy_sel%ccdnr-1)

      else ! .not. rawxy_sel%ccdnr_defined. Clear all CCD bits outside the selected rectangle.
        this%rawxy_mask%values(:,1:yiLo-1) = 0
        this%rawxy_mask%values(1:xiLo-1,yiLo:yiHi) = 0
        this%rawxy_mask%values(xiHi+1:N_PIXELS_CCD,yiLo:yiHi) = 0
        this%rawxy_mask%values(:,yiHi+1:N_PIXELS_CCD) = 0
      end if
    end if

  end subroutine record_rawxy_sel_tEventFile

  !.....................................................................
  integer function get_ext_index_tEventFile(this, ext_class_str, ext_type_i) result(index_i)
    ! Given the class of the extension (currently either GTI-type or the type 'ext' that includes DEADCOR, CORRATT, BADPIX and eventually OFFSET), and the type integer within that class, this function returns the index of the appropriate list of tBinTable objects which corresponds to the type.
    !
    implicit none

    class(tEventFile) :: this
    character(*), intent(in) :: ext_class_str
    integer,      intent(in) :: ext_type_i

    character(*), parameter :: functionName = "get_ext_index_tEventFile"
    integer :: ii
    character(buffsize) :: iStr

    index_i = -1 ! default, indicates not found.

    select case(ext_class_str)
    case("gti")
      do ii=1,size(this%gtiList)
        if(this%gtiList(ii)%ext_type==ext_type_i) then
          index_i = ii
      exit
        end if
      end do

    case("ext")
      do ii=1,size(this%extList)
        if(this%extList(ii)%ext_type==ext_type_i) then
          index_i = ii
      exit
        end if
      end do

    case default
call wrstop(functionName, "Unrecognized extension class "//trim(ext_class_str))
    end select

    if(index_i<0) then
      iStr = int2str(ext_type_i)
call wrstop(functionName, "Could not find extension type "//trim(iStr))
    end if

  end function get_ext_index_tEventFile

  !=====================================================================
  ! The following routines are not methods of tEventFile, they're just
  ! utility routines that manipulate the objects.

  !.....................................................................
  subroutine dealloc_eventfiles(eventfiles, status)
    implicit none

    type(tEventFile), dimension(:), allocatable :: eventfiles
    integer, intent(inout) :: status

    integer :: stat,fi
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "dealloc_eventfiles"

    if(allocated(eventfiles)) then
      do fi=1,size(eventfiles)
        call eventfiles(fi)%destroy(status)
      end do

      errmsg=""
      deallocate(eventfiles, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
    end if

  end subroutine dealloc_eventfiles

  !.....................................................................
  subroutine alloc_eventfiles(eventfiles, n, status)
    implicit none

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    type(tEventFile), allocatable, dimension(:) :: eventfiles

    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "alloc_eventfiles"

    call dealloc_eventfiles(eventfiles, status)
    errmsg=""
    allocate(eventfiles(n), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

  end subroutine alloc_eventfiles

  !.....................................................................
  subroutine get_merge_std_header_eventfile(eventFile, eventFile_out, status)
    implicit none

    type(tEventFile), dimension(:), intent(inout) :: eventFile
    type(tEventFile),               intent(inout) :: eventFile_out
    integer,                        intent(inout) :: status

    character(*), parameter :: function = "get_merge_std_header_eventfile"

    if(status > 100 .or. status == s_errRep)&
  return

    call read_events_ext_keywords_eventfile(eventFile, status)
    call compare_stdheader_eventfile(eventFile, status)
    call merge_stdheader_eventfile(eventFile, eventFile_out, status)

  end subroutine get_merge_std_header_eventfile

  !.....................................................................
  subroutine read_events_ext_keywords_eventfile(eventFile, status)
!*** might need upgrading
    implicit none

    type(tEventFile), dimension(:) :: eventFile
    integer, intent(inout) :: status

    character(*), parameter :: function = "read_events_ext_keywords_eventfile"
    integer :: ii

    if(status > 100 .or. status == s_errRep)&
  return

    do ii = 1, size(eventFile)
      call ffits_movnam_hdu(eventFile(ii)%info%unitNo, BINARY_TBL, "EVENTS", status)
      call eventFile(ii)%read_stdheader(status)
      call eventFile(ii)%events%wcs%read(eventFile(ii)%info%unitNo, BINARY_TBL, "X", "Y", status)
    end do

  end subroutine read_events_ext_keywords_eventfile

  !.....................................................................
  subroutine compare_stdheader_eventfile(eventfiles, status, alwayswarn)
    implicit none

    integer :: ii, warnlevel, same
    integer, intent(inout) :: status
    logical, intent(in), optional :: alwayswarn

    integer :: compare_emin, compare_emax, nenergy, nenergy_1

    type(tEventFile), dimension(:), intent(in) :: eventfiles

    character(*), parameter :: function = "compare_stdheader_eventfile"
    character(buffsize) :: warnmsg

    if(status > 100 .or. status == s_errRep)&
  return

    if(size(eventfiles) < 1) then
      call wrerr(function, s_err, "Eventfiles array is empty.")
      status = s_errRep
  return
    end if

    if(size(eventfiles) == 1)&
  return

    warnlevel = 0
    if(present(alwayswarn)) then
      if(alwayswarn) warnlevel = ibset(warnlevel, 8)
    end if

    do ii = 2, size(eventfiles)
      if(compare(eventfiles(ii)%ccdnr_infile, eventfiles(1)%ccdnr_infile) /= true) then
        warnlevel = ibset(warnlevel, 0)
        warnmsg = "Instruments in input files are different."
        if(ibclr(warnlevel, 0) > 0) then
          call wrerr(function, status, warnmsg)
          status = s_errRep
  return
        end if
      end if

      if(eventfiles(ii)%std_header%pat_sel /= eventfiles(1)%std_header%pat_sel) then
        warnlevel = ibset(warnlevel, 1)
        warnmsg = "pat_sel in input files are different."
        if(ibclr(warnlevel, 1) > 0) then
          call wrerr(function, status, warnmsg)
          status = s_errRep
  return
        end if
      end if

      if(eventfiles(ii)%std_header%flagsel /= eventfiles(1)%std_header%flagsel) then
        warnlevel = ibset(warnlevel, 2)
        warnmsg = "flagsel in input files are different."
        if(ibclr(warnlevel, 2) > 0) then
          call wrerr(function, status, warnmsg)
          status = s_errRep
  return
        end if
      end if

      if(eventfiles(ii)%std_header%gti_sel /= eventfiles(1)%std_header%gti_sel) then
        warnmsg = "gti_sel in input files are different."
        call wrerr(function, status, warnmsg)
      end if

      call compare_rawxy_masks(eventfiles(1)%ccdnr_infile, eventfiles(1)%rawxy_mask,&
        eventfiles(ii)%ccdnr_infile, eventfiles(ii)%rawxy_mask, status)

      nenergy = eventfiles(ii)%std_header%nenergy
      nenergy_1 = eventfiles(1)%std_header%nenergy
      compare_emin = compare(eventfiles(ii)%std_header%eminx(:nenergy)&
        , eventfiles(1)%std_header%eminx(:nenergy_1))
      compare_emax = compare(eventfiles(ii)%std_header%emaxx(:nenergy)&
        , eventfiles(1)%std_header%emaxx(:nenergy_1))

      if(compare_emin == false .or. compare_emax == false) then
        warnlevel = ibset(warnlevel, 3)
        warnmsg = "emin / emax in input files are different."
        if(ibclr(warnlevel, 3) > 0) then
          call wrerr(function, status, warnmsg)
          status = s_errRep
  return
        end if
      end if

      ! compare std_header
      call compare_str(eventfiles(1)%std_header%obs_mode&
        , eventfiles(ii)%std_header%obs_mode, same, status, true)
      if(same == false) call wrerr(function, warn, "Different OBS_MODE in infiles.")
      call compare_str(eventfiles(1)%std_header%datamode&
        , eventfiles(ii)%std_header%datamode, same, status, true)
      if(same == false) call wrerr(function, warn, "Different DATAMODE in infiles.")
      call compare_str(eventfiles(1)%std_header%filter&
        , eventfiles(ii)%std_header%filter, same, status, true)
      if(same == false) call wrerr(function, warn, "Different FILTER in infiles.")

      !      if(EventFile(1)%std_header%obs_mode /= EventFile(ii)%std_header%obs_mode) call wrerr(function, warn, "Different OBS_MODE in infiles")
      !      if(EventFile(1)%std_header%datamode /= EventFile(ii)%std_header%datamode) call wrerr(function, warn, "Different DATAMODE in infiles")
      ! TODO implement filter comparison if available
      !            if(EventFile(1)%std_header%filter /= EventFile(ii)%std_header%filter) call wrerr(function, warn, "Different FILTER in infiles")


      if(abs(eventfiles(1)%std_header%frametim - eventfiles(ii)%std_header%frametim)&
        > tiny(eventfiles(1)%std_header%frametim)) then
        call wrerr(function, warn, "Different FRAMETIM in infiles.")
      end if

      if(eventfiles(1)%std_header%radecsys /= eventfiles(ii)%std_header%radecsys) then
        call wrerr(function, s_err, "Different RADECSYS in infiles.")
        status = s_errRep
  return
      end if

      if(abs(eventfiles(1)%std_header%equinox - eventfiles(ii)%std_header%equinox)&
        > tiny(eventfiles(1)%std_header%equinox)) then
        call wrerr(function, s_err, "Different EQUINOX in infiles.")
        status = s_errRep
  return
      end if

      if(abs(eventfiles(1)%std_header%timezero - eventfiles(ii)%std_header%timezero)&
        > tiny(eventfiles(1)%std_header%timezero)) then
        call wrerr(function, s_err, "Different TIMEZERO in infiles.")
        status = s_errRep
  return
      end if

      if(eventfiles(1)%std_header%timeunit /= eventfiles(ii)%std_header%timeunit) then
        call wrerr(function, s_err, "Different TIMEUNIT in infiles.")
        status = s_errRep
  return
      end if

      if(eventfiles(1)%std_header%timesys /= eventfiles(ii)%std_header%timesys) then
        call wrerr(function, s_err, "Different TIMESYS in infiles")
        status = s_errRep
  return
      end if
    end do

  end subroutine compare_stdheader_eventfile

  !.....................................................................
  subroutine compare_rawxy_masks(ccdnrs1, rawxy_mask1, ccdnrs2, rawxy_mask2, status)
    implicit none

    integer, dimension(:), intent(in) :: ccdnrs1, ccdnrs2
    type(tRawXYMask), intent(in) :: rawxy_mask1, rawxy_mask2
    integer, intent(inout) :: status

    character(*), parameter :: function = "compare_rawxy_masks"

    logical :: is_alloc1,is_alloc2
    integer :: ii,jj,ccdnr
    logical, dimension(N_PIXELS_CCD,N_PIXELS_CCD) :: mask1,mask2

    if(status > 100 .or. status == s_errRep)&
  return

    ! What we don't want to allow is if evlist 1 and evlist 2 share a CCD number,
    ! but have different rawxy masks for that CCD. (Note that if one mask is not
    ! allocated, and the other mask is all .true. for the relevant CCD, we will
    ! allow that to pass.
    !
    is_alloc1 = allocated(rawxy_mask1%values)
    is_alloc2 = allocated(rawxy_mask2%values)

    if(.not. (is_alloc1 .or. is_alloc2))&
  return

!*** I am assuming here that there are no repeated elements in either ccdnrs1 or ccdnrs2.
    do ii=1,size(ccdnrs1)
      ccdnr = ccdnrs1(ii)
      do jj=1,size(ccdnrs2)
        if(ccdnr==ccdnrs2(jj)) then
          if(is_alloc1) then
            mask1 = btest(rawxy_mask1%values, ccdnr-1)
            if(is_alloc2) then
              mask2 = btest(rawxy_mask2%values, ccdnr-1)
              if(.not. all(mask1 .eqv. mask2)) then
                call wrerr(function, s_err, "RawXY selection masks not equivalent for 2 input event lists.")
                status = s_errRep
  return
              end if
            else
              if(.not. all(mask1)) then
                call wrerr(function, s_err, "RawXY selection masks not equivalent for 2 input event lists.")
                status = s_errRep
  return
              end if
            end if

          else ! then rawxy_mask2 is allocated
            mask2 = btest(rawxy_mask2%values, ccdnr-1)
            if(is_alloc1) then
              mask1 = btest(rawxy_mask1%values, ccdnr-1)
              if(.not. all(mask1 .eqv. mask2)) then
                call wrerr(function, s_err, "RawXY selection masks not equivalent for 2 input event lists.")
                status = s_errRep
  return
              end if
            else
              if(.not. all(mask2)) then
                call wrerr(function, s_err, "RawXY selection masks not equivalent for 2 input event lists.")
                status = s_errRep
  return
              end if
            end if
          end if
        end if
      end do
    end do

  end subroutine compare_rawxy_masks

  !.....................................................................
  subroutine merge_stdheader_eventfile(eventfiles, eventfile_out, status)
    implicit none

    type(tEventFile), dimension(:), intent(in)    :: eventfiles
    type(tEventFile),               intent(inout) :: eventfile_out
    integer,                        intent(inout) :: status

    integer :: ii, same

    real(r8), dimension(:), allocatable :: emin, emax
    logical, dimension(N_TELESCOPES) :: ccd_is_merged

    character(*), parameter :: functionName = "merge_stdheader_eventfile"

    if(status > 100 .or. status == s_errRep)&
  return

    ! make sanity check
    if(size(eventfiles) == 0) then
      call wrerr(functionName, s_err, "number of input eventfiles for merging = 0")
      status = s_errRep
  return
    end if

    if(.not.eventfile_out%isInitialized)&
call wrstop(functionName, "You must initialize the output event list before calling this subroutine.")

    ! do not completely copy the EventFile but just the important parameters
    !        eventfile_out = eventfiles(1)
    eventfile_out%filtered_pattern = eventfiles(1)%filtered_pattern
    eventfile_out%filtered_energy  = eventfiles(1)%filtered_energy
    eventfile_out%filtered_telid   = eventfiles(1)%filtered_telid
    eventfile_out%filtered_gti     = eventfiles(1)%filtered_gti
    eventfile_out%filtered_flag    = eventfiles(1)%filtered_flag
    eventfile_out%filtered_region  = eventfiles(1)%filtered_region
    eventfile_out%filtered_rawxy   = eventfiles(1)%filtered_rawxy

    eventfile_out%rawxy_mask = eventfiles(1)%rawxy_mask

    eventfile_out%std_header = eventfiles(1)%std_header

    if(size(eventfiles) < 2)&
  return

    eventfile_out%std_header%ninst = size(eventfile_out%ccdnr_infile)
    if(size(eventfile_out%ccdnr_infile) > 1) then
      eventfile_out%std_header%instrume = "merged"
    end if

    do ii = 2, size(eventfiles)
      call compare_str(eventfile_out%std_header%filter&
        , eventfiles(ii)%std_header%filter, same, status, true)
      if(same == false) then
        eventfile_out%std_header%filter = "mixed"
        call wrerr(functionName, warn, "Keyword 'filter' will be set to mixed.")
      end if

      eventfile_out%std_header%pat_sel =  ior(eventfile_out%std_header%pat_sel&
        , eventfiles(ii)%std_header%pat_sel)
      eventfile_out%std_header%flagsel = iand(eventfile_out%std_header%flagsel&
        , eventfiles(ii)%std_header%flagsel)

      call merge_or_r8(eventfile_out%std_header%eminx, eventfile_out%std_header%emaxx&
        , eventfile_out%std_header%nenergy, eventfiles(ii)%std_header%eminx&
        , eventfiles(ii)%std_header%emaxx, eventfiles(ii)%std_header%nenergy&
        , emin, emax, status)

      eventfile_out%std_header%nenergy = size(emin)

      if(size(emin) > size(eventfile_out%std_header%eminx)) then
        call wrerr(functionName, warn, "Merged file would contain more energy"&
          //" intervals, number of energy intervals are truncated")
      end if

      eventfile_out%std_header%eminx(:size(emin)) = emin
      eventfile_out%std_header%emaxx(:size(emax)) = emax

      ! update the tstart and tstop
      if(eventfiles(ii)%std_header%tstart < eventfile_out%std_header%tstart)&
        eventfile_out%std_header%tstart = eventfiles(ii)%std_header%tstart
      if(eventfiles(ii)%std_header%tstop > eventfile_out%std_header%tstop)&
        eventfile_out%std_header%tstop = eventfiles(ii)%std_header%tstop
      call sec2date(eventfile_out%std_header%tstart, eventfile_out%std_header%mjdref&
        , eventfile_out%std_header%date_obs, status)
      call sec2date(eventfile_out%std_header%tstop,  eventfile_out%std_header%mjdref&
        , eventfile_out%std_header%date_end, status)

      ! GTI
      if(eventfiles(ii)%std_header%gti_sel /= eventfile_out%std_header%gti_sel) then
        eventfile_out%std_header%gti_sel = "undef"
      end if

      ! compare skyfields
      if(trim(eventfile_out%std_header%skyfield) .ne. trim(eventfiles(ii)%std_header%skyfield)) then
        eventfile_out%std_header%skyfield = "merged"
      end if
    end do

    do ii = 1, size(eventfile_out%ccdnr_infile)
      write(eventfile_out%std_header%instrumx(ii), '(a,i1)') "TM", eventfile_out%ccdnr_infile(ii)
    end do

    do ii=1,N_TELESCOPES
      ccd_is_merged(ii) = .false. ! the default value
    end do

    ! Merge the RAWXY masks last, because we need the full list of CCD numbers. Note that we do this here rather than in get_merged_extension_eventfile() because we need the full RAWXY extension before we apply the current RAWXY filter to it in update_filter_settings_eventfile().
    !
    do ii = 2, size(eventfiles)
      call merge_rawxy_masks(eventfiles(ii)%ccdnr_infile, eventfiles(ii)%rawxy_mask, &
        eventfile_out%rawxy_mask, ccd_is_merged, status)
    end do
    if(allocated(eventfile_out%rawxy_mask%values))&
      eventfile_out%filtered_rawxy = .true.

    call dealloc(emin, status)
    call dealloc(emax, status)

  end subroutine merge_stdheader_eventfile

  !.....................................................................
  subroutine merge_ccdnrs_eventfiles(eventfiles, ccdnrs, status)
    ! The calling routine should deallocate ccdnrs after us.

    implicit none

    type(tEventFile), dimension(:),     intent(in)    :: eventfiles
    integer, dimension(:), allocatable, intent(out)   :: ccdnrs
    integer,                            intent(inout) :: status

    character(*), parameter :: function = "merge_ccdnrs_eventfile"
    integer, dimension(:), allocatable :: temp_ccdnrs
    integer :: ii

    if(size(eventfiles)<1) then
      call alloc(ccdnrs, 0, status)
  return
    end if

    ccdnrs = eventfiles(1)%ccdnr_infile

    do ii = 2, size(eventfiles)
      call merge_arrays(ccdnrs, eventfiles(ii)%ccdnr_infile&
        , temp_ccdnrs, status)

      ccdnrs = temp_ccdnrs
    end do
    call dealloc(temp_ccdnrs, status)

  end subroutine merge_ccdnrs_eventfiles

  !.....................................................................
  subroutine merge_rawxy_masks(ccdnrs2, rawxy_mask2, rawxy_mask, ccd_is_merged, status)
    implicit none

    integer, dimension(:), intent(in)    :: ccdnrs2
    type(tRawXYMask),      intent(in)    :: rawxy_mask2
    type(tRawXYMask),      intent(inout) :: rawxy_mask
    logical, dimension(:), intent(inout) :: ccd_is_merged
    integer,               intent(inout) :: status

    character(*), parameter :: function = "merge_rawxy_masks"

    integer :: ii,ccdnr,xi,yi
    logical, dimension(N_PIXELS_CCD,N_PIXELS_CCD) :: mask

    if(status > 100 .or. status == s_errRep)&
  return

    do ii=1,size(ccdnrs2)
      ccdnr = ccdnrs2(ii)
      if(.not.ccd_is_merged(ccdnr)) then
        if(allocated(rawxy_mask2%values)) then
          if(.not.allocated(rawxy_mask%values)) then
            call rawxy_mask%init(status)
            rawxy_mask%extname = rawxy_mask2%extname
          end if

          mask = btest(rawxy_mask2%values, ccdnr-1)
!*** 2019-09-11 IMS: note that overload to int8 for all bit manipulation functions appears to be gnu-specific (i.e., non-standard f90). Will it work on all O/S?

          ! IMS: the following will surely be slow, but so far I can't think of a funky array method.
          !
          do xi=1,N_PIXELS_CCD
            do yi=1,N_PIXELS_CCD
              if(mask(xi,yi)) then
                rawxy_mask%values(xi,yi) = ibset(rawxy_mask%values(xi,yi), ccdnr-1)
              else
                rawxy_mask%values(xi,yi) = ibclr(rawxy_mask%values(xi,yi), ccdnr-1)
              end if
            end do
          end do
        end if

        ccd_is_merged(ccdnr) = .true.
      end if
    end do

  end subroutine merge_rawxy_masks

  !.....................................................................
  subroutine merge_extensions(plenty_memory, eventfiles, eventfile_out&
    , status, repair_gtis)
!*** stupid name - rename it to 'merge_eventlists'. (Note that expmerge calls this.)
    implicit none

    logical,                        intent(in)    :: plenty_memory ! true for small files, reopen not needed.
    type(tEventFile), dimension(:), intent(inout) :: eventfiles
    type(tEventFile),               intent(inout) :: eventfile_out
    integer,                        intent(inout) :: status
    logical, optional,              intent(in)    :: repair_gtis

    character(*), parameter :: functionName = "merge_extensions"
    real(r8), parameter :: x0=0d0, y0=0d0, xd=-0.05/3600d0, yd=0.05/3600d0
    integer :: ci,ci1
    integer :: numFiles,fi,ii,jj,fi_first_events,numRows
    logical :: first_events_found,can_merge,wcs_same_as_1
    real(r8) :: correction,mjd_diff
    type(tWcs) :: wcs_1
    character(buffsize) :: errmsg_prefix,numStr
    character(buffsize) :: fileName
    logical, dimension(:), allocatable :: ext_found
    real(r8), dimension(:), pointer :: ras=>null(),decs=>null(),xs=>null(),ys=>null(),times=>null()

    fileName = ""

    if(status > 100 .or. status == s_errRep)&
  return

    numFiles = size(eventfiles)
    if(numFiles<1) then
      call wrerr(functionName, s_err, "There are no event lists to merge.")
      status = s_errRep
  return
    end if

    numStr = int2str(numFiles)
    call wrlog(functionName, "Merging "//trim(numStr)//" files.")

    if(.not.plenty_memory)&
      call wrlog(functionName, "Infiles will be read 1 extension at a time.")

    ! Check if we can merge events.
    !
    can_merge = .true. ! default
    first_events_found = .false. ! default
    fi_first_events = -99 ! don't like to leave it floating.
    do fi=1,numFiles
      if(.not.plenty_memory) then
        call eventfiles(fi)%open(FF_READONLY, status)
        call eventfiles(fi)%read_events(status, want_wcs=.true.)
        call eventfiles(fi)%close(status)
      end if

      call eventfiles(fi)%info%getFileName(fileName)
      if(eventfiles(fi)%events%containsData()) then
        if(first_events_found) then
          if(.not.checkMergable_bintables(eventfiles(fi_first_events)%events, eventfiles(fi)%events)) then
            call wrerr(functionName, warn, "Event table of file "//trim(fileName)//" not mergeable.")
            can_merge = .false.
    exit
          end if

        else
          first_events_found = .true.
          fi_first_events = fi
        end if

      else ! file fi contains no events.
        call wrerr(functionName, warn, "Event table of file "//trim(fileName)//" contains no data.")
      end if

!*** if we are really concerned about memory, we should free each event list at this point and re-read them within merge_evlist_events() below.
    end do ! file

    if(.not.can_merge) then
      call wrerr(functionName, s_err, "Event tables are not mergeable.")
      status = s_errRep
  return
    end if

    if(.not.first_events_found) then
      call wrerr(functionName, warn, "No events were found in any input file.")
      fi_first_events = 1 ! because we need to copy over WCS from something.
    end if

    call wrchatty(functionName, "Merging events.")

    ! Process events extensions.
    !
    eventfile_out%events%wcs = eventfiles(fi_first_events)%events%wcs
    eventfile_out%events%wcs_containsData = eventfiles(fi_first_events)%events%wcs_containsData

    wcs_1 = eventfiles(fi_first_events)%events%wcs
    do fi=fi_first_events+1,numFiles
      if(.not.eventfiles(fi)%events%containsData())&
    cycle

      call eventfiles(fi)%events%get_col_ptr("RA",   ras)
      call eventfiles(fi)%events%get_col_ptr("DEC",  decs)
      call eventfiles(fi)%events%get_col_ptr("X",    xs)
      call eventfiles(fi)%events%get_col_ptr("Y",    ys)
      call eventfiles(fi)%events%get_col_ptr("TIME", times)

      ! Correct the X and Y values for a file if its WCS is different to that for file 1.
      !
      numRows = eventfiles(fi)%events%getNumRows()
      wcs_same_as_1 = .true. ! default
      if(compare(eventfiles(fi)%events%wcs%refpix, wcs_1%refpix) == false) wcs_same_as_1 = .false.
      if(compare(eventfiles(fi)%events%wcs%refval, wcs_1%refval) == false) wcs_same_as_1 = .false.
      if(compare(eventfiles(fi)%events%wcs%delta,  wcs_1%delta ) == false) wcs_same_as_1 = .false.
      if(.not. wcs_same_as_1) then
        call wrchatty(functionName, "Calling radec2pix.")

        call radec2pix(ras, decs, numRows, wcs_1%refval(1), wcs_1%refval(2)&
          , x0, y0, xd, yd, xs, ys)
      end if

      ! Apply a time correction if necessary.
      !
      correction = 0_r8
      mjd_diff = eventfiles(fi)%std_header%mjdref - eventfiles(fi_first_events)%std_header%mjdref

      if(abs(mjd_diff) > tiny(eventfiles(fi_first_events)%std_header%mjdref)) then
        call wrchatty(functionName, "mjdref correction applied to events extension")
        correction = mjd_diff*3600.0_r8*24.0_r8
        times = times + correction
      end if

      nullify(ras)
      nullify(decs)
      nullify(xs)
      nullify(ys)
      nullify(times)
    end do

    eventfile_out%events%header = eventfiles(1)%events%header

    call merge_evlist_events(eventfiles, eventfile_out%events, status)

    ! The following also sorts the rows in the (ascending) order given.
    !
    call eventfile_out%events%delete_duplicates(["TIME ","RAWX "&
      ,"RAWY ","PHA  ","TM_NR"], status, do_sort=.true.)

    do fi=1,numFiles
      call eventfiles(fi)%events%destroy(status)
    end do

    ! Merge BADPIX, DEADCOR etc
    !
    do ii=1,size(eventfile_out%extList)
      if(.not.btest(eventfile_out%extList(ii)%useageFlags,writeBit))&
    cycle

      can_merge = .true. ! default
      do fi = 1, numFiles
        if(.not.plenty_memory) then
          call eventfiles(fi)%open(FF_READONLY, status)
          call read_extensions(eventfiles(fi)%ccdnr_infile, eventfiles(fi)%std_header%mjdref&
            , eventfiles(fi)%extList(ii), status)
          call eventfiles(fi)%close(status)
        end if

        ci1 = eventfiles(fi)%ccdnr_infile(1)
        if(.not.eventfiles(fi)%extList(ii)%extPerTM(ci1)%foundInFile) then
          call wrerr(functionName, warn, "Extension type '"&
            //trim(eventfiles(fi)%extList(ii)%extNameRoot)&
            //"' is not present in all files: can't merge these extensions.")

          can_merge = .false.
      exit
        end if
      end do ! file

      if(can_merge) then
        ! Merge extensions of this type, one TM at a time.
        !
        do jj=1,size(eventfile_out%ccdnr_infile)
          ci = eventfile_out%ccdnr_infile(jj)
          call wrchatty(functionName, "Merging "&
            //trim(eventfile_out%extList(ii)%extPerTM(ci)%info%tableName)//" extensions.")

          call merge_evlist_extensions(eventfiles, ii, ci, eventfile_out%extList(ii)%extPerTM(ci), status)
          call eventfile_out%extList(ii)%extPerTM(ci)%delete_duplicates(status, do_sort=.true.)
        end do

        do fi=1,numFiles
          do jj=1,size(eventfiles(fi)%extList(ii)%extPerTM)
            call eventfiles(fi)%extList(ii)%extPerTM(jj)%destroy(status)
          end do
        end do
      end if ! can merge.
    end do ! loop over extension types (badpix, deadcor etc)

    ! Merge HK extensions.
    !
    do ii=1,numHKtypes
      can_merge = .true. ! default
      do fi = 1, numFiles
        if(.not.plenty_memory) then
          call eventfiles(fi)%open(FF_READONLY, status)
          call read_hks(eventfiles(fi)%ccdnr_infile, eventfiles(fi)%hkList(ii), status)
          call eventfiles(fi)%close(status)
        end if

        ci1 = eventfiles(fi)%ccdnr_infile(1)
        if(.not.eventfiles(fi)%hkList(ii)%hkPerTM(ci1)%foundInFile) then

          call wrerr(functionName, warn, "HK extension type '"&
            //trim(eventfiles(fi)%hkList(ii)%extNameRoot)&
            //"' is not present in all files: can't merge these extensions.")
          can_merge = .false.
      exit
        end if
      end do ! file

      if(can_merge) then
        ! Merge extensions of this type, one TM at a time.
        !
        do jj=1,size(eventfile_out%ccdnr_infile)
          ci = eventfile_out%ccdnr_infile(jj)

          call wrchatty(functionName, "Merging "&
            //trim(eventfile_out%hkList(ii)%hkPerTM(ci)%info%tableName)//" extensions.")

          call merge_evlist_hk(eventfiles, ii, ci, eventfile_out%hkList(ii)%hkPerTM(ci), status)

!*** original code had adding this as a TODO:
!          call eventfile_out%hkList(ii)%hkPerTM(ci)%delete_duplicate_times(status)
          !**** rather call eventfile_out%hkList(ii)%hkPerTM(ci)%delete_duplicates(status, [this%ind%r8HK_RecordTime])? 20210719 IMS: now we can do as follows:
          call eventfile_out%hkList(ii)%hkPerTM(ci)%delete_duplicates(["RecordTime"], status, do_sort=.true.)!, dotest=.true.)
        end do

        do fi=1,numFiles
          do jj=1,size(eventfiles(fi)%hkList(ii)%hkPerTM)
            call eventfiles(fi)%hkList(ii)%hkPerTM(jj)%destroy(status)
          end do
        end do
      end if ! can merge.
    end do ! loop over HK types.

    ! Merge GTI extensions.
    !
    call alloc(ext_found, numFiles, status)
    do ii=1,size(eventfile_out%gtiList)
      if(.not.btest(eventfile_out%gtiList(ii)%useageFlags,writeBit))&
    cycle

      can_merge = .true. ! default
      ext_found(:) = .false. ! default
      do fi = 1, numFiles
        if(.not.plenty_memory) then
          call eventfiles(fi)%open(FF_READONLY, status)
          call read_gtis(eventfiles(fi)%ccdnr_infile, eventfiles(fi)%std_header%mjdref&
            , eventfiles(fi)%gtiList(ii), status, nowarn=.true., repair_gtis=repair_gtis)
          call eventfiles(fi)%close(status)
        end if

        ci1 = eventfiles(fi)%ccdnr_infile(1) !********************************* shouldn't we really check them all??
        if(eventfiles(fi)%gtiList(ii)%gtiPerTM(ci1)%foundInFile)&
          ext_found(fi) = .true.

      end do ! file

      if(.not.all(ext_found)) then ! we can't merge them, whatever.
        can_merge = .false.

        ! Usually none of the files has the required extension. There is no need to squawk in this case, only when some of them lack it, but others have it, viz:
        !
        if(any(ext_found)) then
          call wrerr(functionName, warn, "GTI extension type '"&
            //trim(eventfiles(1)%gtiList(ii)%extNameRoot)&
            //"' is not present in all files: can't merge these extensions.")
        end if
      end if

      if(can_merge) then
        ! Merge extensions of this type, one TM at a time.
        !
        do jj=1,size(eventfile_out%ccdnr_infile)
          ci = eventfile_out%ccdnr_infile(jj)

          call wrchatty(functionName, "Merging "&
            //trim(eventfile_out%gtiList(ii)%gtiPerTM(ci)%info%tableName)//" extensions.")

          call merge_evlist_gti(eventfiles, ii, ci, eventfile_out%gtiList(ii)%gtiPerTM(ci), status)
        end do

        do fi=1,numFiles
          do jj=1,size(eventfiles(fi)%gtiList(ii)%gtiPerTM)
            call eventfiles(fi)%gtiList(ii)%gtiPerTM(jj)%destroy(status)
          end do
        end do

        do jj=1,size(eventfile_out%ccdnr_infile)
          ci = eventfile_out%ccdnr_infile(jj)
          call eventfile_out%gtiList(ii)%gtiPerTM(ci)%info%fileInfo%getFileName(fileName)
          write(errmsg_prefix, '(a)') trim(fileName)&
            //":"//trim(eventfile_out%gtiList(ii)%gtiPerTM(ci)%info%tableName)//": "

          call eventfile_out%gtiList(ii)%gtiPerTM(ci)%sanityCheck(status, errmsg_prefix)
        end do
      end if ! can merge.
    end do ! loop over GTI types.

  end subroutine merge_extensions

  !.....................................................................
  subroutine merge_evlist_events(eventfiles, events_out, status)
    implicit none

    type(tEventFile), dimension(:), intent(in)    :: eventfiles
    type(tEvents),                  intent(inout) :: events_out
    integer,                        intent(inout) :: status

    integer :: nfiles,fi,numRows,ii,nColsUsed,totalNumRows
    character(*), parameter :: functionName = "merge_evlist_events"

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not. events_out%isInitialized) then
      call wrerr(functionName, s_err, "Output binary table is not initialized.")
      status = s_errRep
  return
    end if

    nfiles = size(eventfiles)

    totalNumRows = 0
    do fi=1,nfiles
      numRows = eventfiles(fi)%events%getNumRows()
      totalNumRows = totalNumRows + numRows
    end do

    nColsUsed = eventfiles(1)%events%getNumCols()
    do ii = 1, nColsUsed
      events_out%cols(ii)%scal = eventfiles(1)%events%cols(ii)%scal
    end do

    call events_out%alloc(status, totalNumRows)
    if(totalNumRows<1)&
  return

    totalNumRows = 0
    do fi=1,nfiles
      call events_out%overwrite_section(eventfiles(fi)%events, totalNumRows+1, status)
      numRows = eventfiles(fi)%events%getNumRows()
      totalNumRows = totalNumRows + numRows
    end do

  end subroutine merge_evlist_events

  !.....................................................................
  subroutine merge_evlist_extensions(eventfiles, extI, tel_id, ext_out, status)
    implicit none

    type(tEventFile), dimension(:), intent(in)    :: eventfiles
    integer,                        intent(in)    :: extI
    integer,                        intent(in)    :: tel_id
    type(tExtension),               intent(inout) :: ext_out
    integer,                        intent(inout) :: status

    character(*), parameter :: functionName = "merge_evlist_extensions"
    integer :: nfiles,fi,numRows,first_fi,totalNumRows
    type(tExtension) :: firstExt
    logical, dimension(size(eventfiles)) :: has_this_tm

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not. ext_out%isInitialized) then
      call wrerr(functionName, s_err, "Output binary table is not initialized.")
      status = s_errRep
  return
    end if

    nfiles = size(eventfiles)

    first_fi = 0 ! default, indicates has not been found yet.
    totalNumRows = 0
    do fi=1,nfiles
      has_this_tm(fi) = .false. ! default
      if(eventfiles(fi)%extList(extI)%extPerTM(tel_id)%containsData()) then
        has_this_tm(fi) = .true.
        if(first_fi<1) first_fi = fi
        numRows = eventfiles(fi)%extList(extI)%extPerTM(tel_id)%getNumRows()
        totalNumRows = totalNumRows + numRows
      end if
    end do

    call ext_out%alloc(status, totalNumRows)

    if(.not.any(has_this_tm))&
  return

    firstExt = eventfiles(first_fi)%extList(extI)%extPerTM(tel_id)
    do fi=first_fi+1,nfiles
      if(has_this_tm(fi)) then
        if(.not.checkMergable_bintables(firstExt, eventfiles(fi)%extList(extI)%extPerTM(tel_id))) then
          call wrerr(functionName, s_err, "Extensions are not mergable.")
          status = s_errRep
  return
        end if
      end if
    end do

    totalNumRows = 0
    do fi=1,nfiles
      if(has_this_tm(fi)) then
        call ext_out%overwrite_section(eventfiles(fi)%extList(extI)%extPerTM(tel_id), totalNumRows+1, status)
        numRows = eventfiles(fi)%extList(extI)%extPerTM(tel_id)%getNumRows()
        totalNumRows = totalNumRows + numRows
      end if
    end do

  end subroutine merge_evlist_extensions

  !.....................................................................
  subroutine merge_evlist_hk(eventfiles, extI, tel_id, hk_out, status)
    implicit none

    type(tEventFile), dimension(:), intent(in)    :: eventfiles
    integer,                        intent(in)    :: extI
    integer,                        intent(in)    :: tel_id
    type(tHK),                      intent(inout) :: hk_out
    integer,                        intent(inout) :: status

    character(*), parameter :: functionName = "merge_evlist_hk"
    integer :: nfiles,fi,numRows,first_fi,totalNumRows
    type(tHK) :: firstHK
    logical, dimension(size(eventfiles)) :: has_this_tm

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not. hk_out%isInitialized) then
      call wrerr(functionName, s_err, "Output binary table is not initialized.")
      status = s_errRep
  return
    end if

    nfiles = size(eventfiles)

    first_fi = 0 ! default, indicates has not been found yet.
    totalNumRows = 0
    do fi=1,nfiles
      has_this_tm(fi) = .false. ! default
      if(eventfiles(fi)%hkList(extI)%hkPerTM(tel_id)%containsData()) then
        has_this_tm(fi) = .true.
        if(first_fi<1) first_fi = fi
        numRows = eventfiles(fi)%hkList(extI)%hkPerTM(tel_id)%getNumRows()
        totalNumRows = totalNumRows + numRows
      end if
    end do

    call hk_out%alloc(status, totalNumRows)

    if(.not.any(has_this_tm))&
  return

    firstHK = eventfiles(first_fi)%hkList(extI)%hkPerTM(tel_id)
    do fi=first_fi+1,nfiles
      if(has_this_tm(fi)) then
        if(.not.checkMergable_bintables(firstHK, eventfiles(fi)%hkList(extI)%hkPerTM(tel_id))) then
          call wrerr(functionName, s_err, "HK extensions are not mergable.")
          status = s_errRep
  return
        end if
      end if
    end do

    totalNumRows = 0
    do fi=1,nfiles
      if(has_this_tm(fi)) then
        call hk_out%overwrite_section(eventfiles(fi)%hkList(extI)%hkPerTM(tel_id), totalNumRows+1, status)
        numRows = eventfiles(fi)%hkList(extI)%hkPerTM(tel_id)%getNumRows()
        totalNumRows = totalNumRows + numRows
      end if
    end do

  end subroutine merge_evlist_hk

  !.....................................................................
  subroutine merge_evlist_gti(eventfiles, extI, tel_id, gti_out, status)
    implicit none

    type(tEventFile), dimension(:), intent(in)    :: eventfiles
    integer,                        intent(in)    :: extI
    integer,                        intent(in)    :: tel_id
    type(tGTI_tab),                     intent(inout) :: gti_out
    integer,                        intent(inout) :: status

    character(*), parameter :: functionName = "merge_evlist_gti"
    integer :: nfiles,fi,first_fi,numRows
    logical, dimension(size(eventfiles)) :: has_this_tm
    real(r8), dimension(:), pointer :: in_starts=>null(),in_stops=>null(),out_starts=>null(),out_stops=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not. gti_out%isInitialized) then
      call wrerr(functionName, s_err, "Output binary table is not initialized.")
      status = s_errRep
  return
    end if

    nfiles = size(eventfiles)

    first_fi = 0 ! default, indicates has not been found yet.
    do fi=1,nfiles
      has_this_tm(fi) = .false. ! default
      if(eventfiles(fi)%gtiList(extI)%gtiPerTM(tel_id)%containsData()) then
        has_this_tm(fi) = .true.
        if(first_fi<1) first_fi = fi
      end if
    end do

    if(.not.any(has_this_tm)) then
      call gti_out%realloc(status, 0)
  return
    end if

    gti_out%header = eventfiles(first_fi)%gtiList(extI)%gtiPerTM(tel_id)%header

    call eventfiles(first_fi)%gtiList(extI)%gtiPerTM(tel_id)%get_col_ptr("START", in_starts)
    call eventfiles(first_fi)%gtiList(extI)%gtiPerTM(tel_id)%get_col_ptr("STOP",  in_stops)

    numRows = eventfiles(first_fi)%gtiList(extI)%gtiPerTM(tel_id)%getNumRows()
    call gti_out%realloc(status, numRows)
    call gti_out%get_col_ptr("START", out_starts)
    call gti_out%get_col_ptr("STOP",  out_stops)
    out_starts = in_starts
    out_stops  = in_stops

    do fi=first_fi+1,nfiles
      if(has_this_tm(fi)) then
        if(eventfiles(fi)%gtiList(extI)%gtiPerTM(tel_id)%isAllocated()) then
          call gti_out%or_with(eventfiles(fi)%gtiList(extI)%gtiPerTM(tel_id), status)
        end if
      end if
    end do

    nullify(in_starts)
    nullify(in_stops)
    nullify(out_starts)
    nullify(out_stops)

  end subroutine merge_evlist_gti
end module erolib_eventfile_mod
