! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_events_mod

  use ero_general ! need for flag masks.
  use erolib_general_mod
  use erolib_bintable_mod
  use erolib_wcs_mod
  use erolib_gtitab_mod
  use erolib_rawxy_mod

  implicit none

logical, private, parameter :: doTest=.false.

  type tRawXYSelection
    type(tRangeI), dimension(2) :: axes
    integer :: ccdnr = 0
    logical :: ccdnr_defined = .false.
    logical :: invert = .false.
    type(tRawXYMask_logical) :: mask
  contains
    procedure, non_overridable :: print_me => print_tRawXYSelection
    procedure, non_overridable :: get_entry_strs => get_entry_strs_tRawXYSelection
  end type tRawXYSelection

  type, extends(tBinTable) :: tEvents
    type(tWcs) :: wcs
    logical :: want_wcs = .false.
    logical :: wcs_containsData = .false.
  contains
    procedure :: init => init_tEvents
    procedure :: initHeader => initHeader_tEvents
    procedure :: read_extras => read_extras_tEvents
    procedure :: write_extras => write_extras_tEvents
    procedure :: delete_flagged => delete_flagged_tEvents
    procedure :: apply_gti => apply_gti_tEvents
    procedure, non_overridable :: apply_telid => apply_telid_tEvents
    procedure, non_overridable :: apply_energy => apply_energy_tEvents
    procedure, non_overridable :: apply_flag => apply_flag_tEvents
    procedure, non_overridable :: apply_pattern => apply_pattern_tEvents
    procedure, non_overridable :: apply_rawxy => apply_rawxy_tEvents
  end type tEvents

!  integer, parameter :: strEvH_T_OFFCOR = 1
contains

  !.....................................................................
  subroutine print_tRawXYSelection(this)
    class(tRawXYSelection) :: this

    character(buffsize) :: entryStr,ccdnrStr,invertStr

    call this%get_entry_strs(entryStr, ccdnrStr, invertStr)

    print *, "rawxy = "//trim(entryStr)
    print *, "rawxy_telid = "//trim(ccdnrStr)
    print *, "rawxy_invert = "//trim(invertStr)

  end subroutine print_tRawXYSelection

  !.....................................................................
  subroutine get_entry_strs_tRawXYSelection(this, entryStr, ccdnrStr, invertStr)
    character(*), intent(out) :: entryStr,ccdnrStr,invertStr
    class(tRawXYSelection) :: this

    character(buffsize) :: xLoStr,xHiStr,yLoStr,yHiStr

    if(this%axes(1)%lo%is_defined) then
      write (xLoStr, '(i3)') this%axes(1)%lo%value
    else
      xLoStr = 'inf'
    end if
    if(this%axes(1)%hi%is_defined) then
      write (xHiStr, '(i3)') this%axes(1)%hi%value
    else
      xHiStr = 'inf'
    end if
    if(this%axes(2)%lo%is_defined) then
      write (yLoStr, '(i3)') this%axes(2)%lo%value
    else
      yLoStr = 'inf'
    end if
    if(this%axes(2)%hi%is_defined) then
      write (yHiStr, '(i3)') this%axes(2)%hi%value
    else
      yHiStr = 'inf'
    end if

    entryStr = trim(xLoStr)//' '//trim(xHiStr)//' '//trim(yLoStr)//' '//trim(yHiStr)

    if(this%ccdnr_defined) then
      write (ccdnrStr, '(i1)') this%ccdnr
    else
      ccdnrStr = 'undefined'
    end if

    if(this%invert) then
      invertStr = "yes"
    else
      invertStr = "no"
    end if

  end subroutine get_entry_strs_tRawXYSelection

  !.....................................................................
  subroutine init_tEvents(this, status, fileInfo)
    implicit none

    class(tEvents) :: this
    integer, intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "init_tEvents"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    this%info%tableName = "EVENTS"

    call this%initColumn("D", "TIME", "s")
    call this%initColumn("D", "RA", "deg")
    call this%initColumn("D", "DEC", "deg")
    call this%initColumn("D", "X")
    call this%initColumn("D", "Y")
    call this%initColumn("E", "PI")
    call this%initColumn("E", "EV_WEIGHT")
    call this%initColumn("I", "RAWX")
    call this%initColumn("I", "RAWY")
    call this%initColumn("D", "SUBX")
    call this%initColumn("D", "SUBY")
    call this%initColumn("I", "PHA")
    call this%initColumn("I", "PAT_TYP")
    call this%initColumn("B", "PAT_INF")
    call this%initColumn("B", "TM_NR")
    call this%initColumn("J", "FLAG")
    call this%initColumn("D", "FRAMETIME", "s")
    call this%initColumn("D", "RECORDTIME")
    call this%initColumn("I", "PAT_IND")
    call this%initColumn("E", "AMP_GAIN")
    call this%initColumn("E", "AMP_COR")

    call this%initColumn("E", "ENERGY")
    call this%initColumn("B", "CCDNR")

    call this%initColumn("J", "frame_ID") ! just for internal use.

    call this%initColumns(status)

    call this%initHeader(status)

  end subroutine init_tEvents

  !.....................................................................
  subroutine initHeader_tEvents(this, status)
    implicit none

    class(tEvents), target :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "initHeader_tEvents"

    this%header%tabInfo => this%info

    ! Since the following keywords are not present in early versions of
    ! the EVENTS extension, I will set both their useageFlags to
    ! 'read optional' in setup_for_read_tEventFile().
    call this%header%initKeyword("A", "CHOPDEAD")
    call this%header%initKeyword("A", "T_OFFCOR")

    call this%header%initKeywords(status)
    call this%header%check_keywords(status)

  end subroutine initHeader_tEvents

  !.....................................................................
  subroutine read_extras_tEvents(this, status)
    implicit none

    class(tEvents) :: this
    integer, intent(inout) :: status

    character(*), parameter :: function = "read_extras_tEvents"

    if(this%want_wcs) then
      status = 0
      call this%wcs%read(this%info%fileInfo%unitNo, BINARY_TBL, "X", "Y", status)
      if(status==0) this%wcs_containsData = .true.
    end if

    ! Note that I'm attaching the stdhdr to the event file.

  end subroutine read_extras_tEvents

  !.....................................................................
  subroutine write_extras_tEvents(this, status)
    implicit none

    class(tEvents) :: this
    integer, intent(inout) :: status

    character(*), parameter :: function = "write_extras_tEvents"

    if(this%wcs_containsData) then
      call wrchatty(function, "Writing WCS for events.")
      call this%wcs%write(this%info%fileInfo%unitNo, BINARY_TBL, status)
    end if

    ! Note that I'm attaching the stdhdr to the event file.

  end subroutine write_extras_tEvents

  !.....................................................................
  subroutine delete_flagged_tEvents(this, status)
    implicit none

    class(tEvents) :: this
    integer, intent(inout) :: status

    logical, dimension(:), allocatable :: mask

    character(*), parameter :: function = "delete_flagged_tEvents"
    integer :: numRows
    integer(i4), dimension(:), pointer :: flags=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    numRows = this%getNumRows()
    call alloc(mask, numRows, status)
    mask = .false.
    call this%get_col_ptr("FLAG", flags)
    where (iand(flags,badFlagMask) == 0) ! The iand will return non-zero if this%flag has any bits set which are set in badFlagMask. 
      mask = .true.
    end where
    call this%pack(mask, status)
    call dealloc(mask, status)
    nullify(flags)

  end subroutine delete_flagged_tEvents

  !.....................................................................
  subroutine apply_gti_tEvents(this, GTI, mjdref, status)
!*** 20200614 IMS: previous version used this%mjdref.
    implicit none

    class(tEvents)                :: this
    type(tGTI_tab), intent(in)    :: GTI
    real(r8),       intent(in)    :: mjdref
    integer,        intent(inout) :: status

    character(*), parameter :: functionName = "apply_gti_tEvents"
    integer :: numRows,i1EvTM_NR
    type(tGTI_tab) :: tempGTI
    logical, dimension(:), allocatable :: mask, mask_ccdnr
    real(r8) :: mjdref_temp
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null(),times=>null()
    integer(i1), dimension(:), pointer :: tm_nrs=>null()
    logical, dimension(:), pointer :: time_is_null=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    numRows = this%getNumRows()
    call alloc(mask, numRows, status)
    call alloc(mask_ccdnr, numRows, status)

    tempGTI = GTI
    ! internally correct the GTI

    call tempGTI%get_col_ptr("START", starts)
    call tempGTI%get_col_ptr("STOP",  stops)
    call this%get_col_ptr(   "TIME",  times, time_is_null)
    call this%get_col_ptr(   "TM_NR", tm_nrs)

    call tempGTI%header%getKeywordValue("MJDREF",mjdref_temp)
    if(mjdref /= mjdref_temp) then
      call wrchatty(functionName, "Applying mjdref correction.")
      starts = (mjdref_temp - mjdref)*3600.0_r8*24.0_r8 + starts
      stops  = (mjdref_temp - mjdref)*3600.0_r8*24.0_r8 + stops
      call tempGTI%header%setKeywordValue("MJDREF",mjdref)
    end if

!    call this%sort(times, status)!**** why TIME and not FRAMETIME? Check TIME has values, use FRAMETIME if not.
    mask = get_GTI_mask(times, tempGTI, status, .false., time_is_null) ! no sort.

    if(tempGTI%tel_id >= 1 .and. tempGTI%tel_id <= 7) then
      mask_ccdnr = .false.
      i1EvTM_NR = this%getColNum("TM_NR")
      where(tm_nrs == tempGTI%tel_id)
        mask_ccdnr = .true.
      end where
    else
!**** should really send in a boolean argument to indicate whether GTI is ccd-specific or not.
      mask_ccdnr = .true.
    end if

    mask = .not.(.not.mask .and. mask_ccdnr)

    call this%pack(mask, status)
!***** better to update and return mask.
    call tempGTI%destroy(status)

    call dealloc(mask, status)
    call dealloc(mask_ccdnr, status)

    nullify(starts)
    nullify(stops)
    nullify(times,time_is_null)
    nullify(tm_nrs)

  end subroutine apply_gti_tEvents

  !.....................................................................
  subroutine apply_telid_tEvents(this, telid, status)
    implicit none

    class(tEvents)                       :: this
    integer, dimension(:), intent(in)    :: telid
    integer,               intent(inout) :: status

    character(*), parameter :: functionName = "apply_telid_tEvents"
    integer :: ii,numRows
    logical, dimension(:), allocatable :: mask
    integer(i1), dimension(:), pointer :: tm_nrs=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    call this%get_col_ptr("TM_NR", tm_nrs)

    numRows = this%getNumRows()
    call alloc(mask, numRows, status)
    mask = .false.
    do ii = 1, size(telid)
      where (tm_nrs == telid(ii))
        mask = .true.
      end where
    end do

    call this%pack(mask, status)
    call dealloc(mask, status)

    nullify(tm_nrs)

  end subroutine apply_telid_tEvents

  !.....................................................................
  subroutine apply_energy_tEvents(this, emin, emax, status)
    implicit none

    class(tEvents)                        :: this
    real(r8), dimension(:), intent(in)    :: emin, emax
    integer,                intent(inout) :: status

    character(*), parameter :: functionName = "apply_energy_tEvents"
    integer :: ii,numRows,n_bands
    logical, dimension(:), allocatable :: mask
    real(r4), dimension(:), pointer :: pis=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    n_bands = size(emin)
    if(n_bands /= size(emax))&
call wrstop(functionName, "emin and emax don't have the same number of elements.")

    if(n_bands<1)&
  return

    call this%get_col_ptr("PI", pis)

    numRows = this%getNumRows()
    call alloc(mask, numRows, status)
    mask = .false.
    do ii = 1, n_bands
      where (pis <= emax(ii) .and. pis >= emin(ii))
        mask = .true.
      end where
    end do

    call this%pack(mask, status)
    call dealloc(mask, status)

    nullify(pis)

  end subroutine apply_energy_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a flag filter to the events type
  !
  !
  !
  !> @param[in] this
  !> @param[in] flag
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_flag_tEvents(this, flagsel, status, invert)
    implicit none

    class(tEvents) :: this
    integer, intent(in)           :: flagsel
    integer, intent(inout)        :: status
    logical, intent(in), optional :: invert

    character(*), parameter :: functionName = "apply_flag_tEvents"
    integer :: numRows
    logical, dimension(:), allocatable :: mask
    integer(i4), dimension(:), pointer :: flags=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    numRows = this%getNumRows()
    call alloc(mask, numRows, status)
    mask = .false.
    call this%get_col_ptr("FLAG", flags)
    where (iand(flags, flagsel) == 0) ! Only events will pass for which the FLAG value has none of the bits set which are 'on' in flagsel, since only for such will the iand() yield zero.
      mask = .true.
    end where

    if(present(invert)) then
      if(invert) mask = .not.mask
    end if

    call this%pack(mask, status)
    call dealloc(mask, status)
    nullify(flags)

  end subroutine apply_flag_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a pattern filter to the events type
  !
  !
  !
  !> @param[in] this
  !> @param[in] flag
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_pattern_tEvents(this, pat_sel, status)
    implicit none

    class(tEvents) :: this
    integer, intent(in) :: pat_sel
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "apply_pattern_tEvents"
    integer :: numRows
    integer(i2), dimension(:), allocatable :: pattern, pat_typ
    logical, dimension(:), allocatable :: mask
    integer(i2), dimension(:), pointer :: pat_typs=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    ! selection == 0 invalid and valid pattern are copied
    if(pat_sel == 0)&
  return

    call this%get_col_ptr("PAT_TYP", pat_typs)

    numRows = this%getNumRows()
    call alloc(pattern, numRows, status)
    call alloc(pat_typ, numRows, status)
    call alloc(mask, numRows, status)
    mask = .false.

    if(pat_sel < 0) then
      where(pat_typs <= 0)
        mask = .true.
      end where
      call this%pack(mask, status)
      nullify(pat_typs)
  return
    end if

    pattern = pat_typs ! copies the values.

    ! set all invalid patterns to 0
    where(pattern <= 0_i2)
      pattern = 0_i2
      pat_typ = 0_i2!zero
    end where

    ! transform all patterns to bitmask starting at 0
    where(pattern > 0_i2)
      pat_typ = ibset(0_i2, pattern - 1_i2)
    end where

    ! set mask to true if pat_typ is found in bitmask pat_sel
    where (iand(pat_typ, int(pat_sel, kind(pat_typ))) /= 0)
      mask = .true.
    end where

    call this%pack(mask, status)

    call dealloc(pattern, status)
    call dealloc(pat_typ, status)
    call dealloc(mask, status)

    nullify(pat_typs)

  end subroutine apply_pattern_tEvents

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> apply a RAWX filter to the events type
  !
  !
  !
  !> @param[in] this
  !> @param[in] rawx_sel
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine apply_rawxy_tEvents(this, rawxy_sel, status)
    implicit none

    class(tEvents) :: this
    type(tRawXYSelection), intent(in) :: rawxy_sel
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "apply_rawxy_tEvents"
    integer(i4), dimension(:), allocatable :: order
    logical, dimension(:), allocatable :: mask, ccd_mask
    integer :: ii,jj,numRows
    logical :: inside_sel,outside_sel
    integer(i1), dimension(:), pointer :: tm_nrs=>null()
    integer(i2), dimension(:), pointer :: rawxs=>null(),rawys=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not. (rawxy_sel%axes(1)%lo%is_defined .or. rawxy_sel%axes(1)%hi%is_defined &
         .or. rawxy_sel%axes(2)%lo%is_defined .or. rawxy_sel%axes(2)%hi%is_defined &
         .or. allocated(rawxy_sel%mask%values)))&
  return

    if(rawxy_sel%invert) then
      inside_sel  = .false.
    else
      inside_sel  = .true.
    end if
    outside_sel = .not.inside_sel

    call this%get_col_ptr("TM_NR", tm_nrs)
    call this%get_col_ptr("RAWX",  rawxs)
    call this%get_col_ptr("RAWY",  rawys)

    numRows = this%getNumRows()
    call alloc(mask, numRows, status)
    call alloc(ccd_mask, numRows, status)

    if(rawxy_sel%ccdnr_defined) then ! no rawx/y selection done where event CCD /= rawxy_sel%ccdnr.
      mask = .true.
      where (tm_nrs == rawxy_sel%ccdnr)&
        mask = inside_sel ! default.
    else ! same rawx/y selection will be done regardless of event CCD.
      mask = inside_sel ! default.
    end if

    if(rawxy_sel%mask%containsData) then
      if(rawxy_sel%ccdnr_defined) then
        ccd_mask = .false.
        where (tm_nrs == rawxy_sel%ccdnr)&
          ccd_mask = .true.

        order = [(ii,ii=1, numRows,1)]
        order = pack(order, ccd_mask)

        do ii=1,size(order)
          jj = order(ii)
          if(.not.rawxy_sel%mask%values(rawxs(jj),rawys(jj))) then
            mask(jj) = outside_sel
          end if
        end do

      else
        do jj=1,numRows
          if(.not.rawxy_sel%mask%values(rawxs(jj),rawys(jj))) then
            mask(jj) = outside_sel
          end if
        end do
      end if
      
    else
      if(rawxy_sel%ccdnr_defined) then
        if(rawxy_sel%axes(1)%lo%is_defined) then
          where (rawxs < rawxy_sel%axes(1)%lo%value .and. tm_nrs == rawxy_sel%ccdnr)
            mask = outside_sel
          end where
        end if
        if(rawxy_sel%axes(1)%hi%is_defined) then
          where (rawxs > rawxy_sel%axes(1)%hi%value .and. tm_nrs == rawxy_sel%ccdnr)
            mask = outside_sel
          end where
        end if

        if(rawxy_sel%axes(2)%lo%is_defined) then
          where (rawys < rawxy_sel%axes(2)%lo%value .and. tm_nrs == rawxy_sel%ccdnr)
            mask = outside_sel
          end where
        end if
        if(rawxy_sel%axes(2)%hi%is_defined) then
          where (rawys > rawxy_sel%axes(2)%hi%value .and. tm_nrs == rawxy_sel%ccdnr)
            mask = outside_sel
          end where
        end if

      else ! apply the selection to all CCDs.
        if(rawxy_sel%axes(1)%lo%is_defined) then
          where (rawxs < rawxy_sel%axes(1)%lo%value)
            mask = outside_sel
          end where
        end if
        if(rawxy_sel%axes(1)%hi%is_defined) then
          where (rawxs > rawxy_sel%axes(1)%hi%value)
            mask = outside_sel
          end where
        end if

        if(rawxy_sel%axes(2)%lo%is_defined) then
          where (rawys < rawxy_sel%axes(2)%lo%value)
            mask = outside_sel
          end where
        end if
        if(rawxy_sel%axes(2)%hi%is_defined) then
          where (rawys > rawxy_sel%axes(2)%hi%value)
            mask = outside_sel
          end where
        end if
      end if ! apply to a specific CCD or to all.
    end if ! choice between mask or rectangle.

    call this%pack(mask, status)

    call dealloc(order, status)
    call dealloc(mask, status)
    call dealloc(ccd_mask, status)

    nullify(tm_nrs)
    nullify(rawxs)
    nullify(rawys)

  end subroutine apply_rawxy_tEvents

end module erolib_events_mod
