! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_extensions_mod

  use erolib_bintable_mod
  use erolib_gtitab_mod
  use erolib_region_mod
  use srctool_generic , only : coord_type, COORDSYS_CODE_ICRS
  use srctool_generic_regions , only : srctool_regions_test_if_coord_in_region, region_type

  implicit none

!logical, private, parameter :: doTest=.true.

  type, extends(tBinTable) :: tExtension
    integer :: tel_id = -1
    integer :: ext_type = -1
    character(buffsize) :: extNameRoot=""
  contains
    procedure :: initHeader => initHeader_tExtension
    procedure :: init => init_tExtension
    procedure :: apply_gti => apply_gti_tExtension
    procedure :: derive_region_GTI => derive_region_GTI_tExtension
  end type tExtension

  integer, parameter :: ext_numTypes = 4

  ! Extension indices:
  enum, bind(c)
    enumerator :: extCORRATT=1,extDEADCOR,extBADPIX,extOFFSET
  end enum
  character(7), parameter :: ext_nameRoots(*) = ["CORRATT","DEADCOR","BADPIX ","OFFSET "]

  real(r8), parameter :: d_gap = 10.0_r8
  character(*), parameter :: region_gti_ext = "REGGTI"

contains

  !.....................................................................
  subroutine initHeader_tExtension(this, status)
    implicit none

    class(tExtension), target :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "initHeader_tExtension"

    this%header%tabInfo => this%info

    call this%header%initKeyword("D", "MJDREF")
    call this%header%initKeywords(status)
    call this%header%check_keywords(status) ! sets this%isInitialized

  end subroutine initHeader_tExtension

  !.....................................................................
  subroutine init_tExtension(this, extI, status, fileInfo, extNameRoot)
    implicit none

    class(tExtension) :: this
    integer,                           intent(in)    :: extI
    integer,                           intent(inout) :: status
    type(tFitsInfo), optional, target, intent(in)    :: fileInfo
    character(*), optional,            intent(in)    :: extNameRoot

    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "init_tExtension"
    integer :: ii, nColsUsed
    integer :: i2BP_RAWX,i2BP_RAWY,i2BP_YEXTENT,i2BP_TYPE,i2BP_BADFLAG,r8BP_TIMEMIN
    integer :: r8BP_TIMEMAX,i2BP_PHAMAX,i2BP_PHAMIN,r4BP_PHAMED

    if(present(fileInfo)) this%info%fileInfo => fileInfo
!    this%info%tableName = "***"
    this%ext_type = extI

    if(present(extNameRoot)) then
      this%extNameRoot = extNameRoot
    else
      this%extNameRoot = trim(ext_nameRoots(extI))
    end if

    select case(extI)
    case(extBADPIX)
      call this%initColumn("I",  "RAWX")
      call this%initColumn("I",  "RAWY")
      call this%initColumn("I",  "YEXTENT")
      call this%initColumn("I",  "TYPE")
      call this%initColumn("I",  "BADFLAG")
      call this%initColumn("D",  "TIMEMIN")
      call this%initColumn("D",  "TIMEMAX")
      call this%initColumn("I",  "PHAMIN")
      call this%initColumn("I",  "PHAMAX")
      call this%initColumn("E",  "PHAMED")

    case(extCORRATT)
      call this%initColumn("D", "TIME",  "s")
      call this%initColumn("D", "RA",    "deg")
      call this%initColumn("D", "DEC",   "deg")
      call this%initColumn("D", "ROLL",  "deg")

    case(extDEADCOR)
      call this%initColumn("D", "TIME")
      call this%initColumn("E", "DEADC")

    case(extOFFSET)
      call this%initColumn("D",  "TIME")
      call this%initColumn("I",  "RAWX")
      call this%initColumn("I",  "OFFSETX")

    case default
      errmsg = int2str(extI)
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(errmsg)//"'")
      status = s_errRep
  return
    end select

    call this%initColumns(status)
    nColsUsed = size(this%cols)

    select case(extI)
    case(extBADPIX)
      r8BP_TIMEMAX = this%getColNum("TIMEMAX")
      r8BP_TIMEMIN = this%getColNum("TIMEMIN")
      r4BP_PHAMED  = this%getColNum("PHAMED")
      i2BP_YEXTENT = this%getColNum("YEXTENT")
      i2BP_TYPE    = this%getColNum("TYPE")
      i2BP_BADFLAG = this%getColNum("BADFLAG")
      i2BP_PHAMIN  = this%getColNum("PHAMIN")
      i2BP_PHAMAX  = this%getColNum("PHAMAX")
      i2BP_RAWY    = this%getColNum("RAWY")
      i2BP_RAWX    = this%getColNum("RAWX")

      this%sort_sequence = [r8BP_TIMEMIN,r8BP_TIMEMAX,i2BP_RAWX,i2BP_RAWY,i2BP_YEXTENT&
        ,i2BP_TYPE,i2BP_BADFLAG,i2BP_PHAMIN,r4BP_PHAMED,i2BP_PHAMAX]
!      this%sort_sequence = [r8BP_TIMEMAX,r8BP_TIMEMIN,r4BP_PHAMED,i2BP_YEXTENT&
!        ,i2BP_TYPE,i2BP_BADFLAG,i2BP_PHAMIN,i2BP_PHAMAX,i2BP_RAWY,i2BP_RAWX]
!!      this%sort_sequence = [i2BP_RAWX,i2BP_RAWY,i2BP_PHAMAX,i2BP_PHAMIN&
!!         ,i2BP_BADFLAG,i2BP_TYPE,i2BP_YEXTENT,r4BP_PHAMED,r8BP_TIMEMIN,r8BP_TIMEMAX]

    case(extCORRATT)
      call alloc(this%sort_sequence, nColsUsed, status)
      do ii=1,nColsUsed ! Note that this relies on the order of initColumn() calls being unchanged.
        this%sort_sequence(ii) = ii
      end do

    case(extDEADCOR)
      call alloc(this%sort_sequence, nColsUsed-1, status)
      do ii=1,nColsUsed-1 ! Note that this relies on the order of initColumn() calls being unchanged.
        this%sort_sequence(ii) = ii
      end do

    case(extOFFSET)
      call alloc(this%sort_sequence, nColsUsed, status)
      do ii=1,nColsUsed ! Note that this relies on the order of initColumn() calls being unchanged.
        this%sort_sequence(ii) = ii
      end do

    case default
      errmsg = int2str(extI)
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(errmsg)//"'")
      status = s_errRep
  return
    end select

    call this%initHeader(status)

  end subroutine init_tExtension

  !.....................................................................
  subroutine derive_region_GTI_tExtension(this, regionsObj, gti, status, fileInfo)
    implicit none

    class(tExtension) :: this
    type(tRegions),  intent(in)    :: regionsObj
    type(tGTI_tab),      intent(out)   :: gti
    integer,         intent(inout) :: status
    type(tFitsInfo), optional, target, intent(in) :: fileInfo

    character(*), parameter :: functionName = "derive_region_GTI_tExtension"
    integer :: nContSeq,ii,jj,dummyI,nContSeqInReg,nGTIs
    integer, dimension(:), allocatable :: csStartIs,csStopIs,csirStartIs,csirStopIs
    logical, dimension(:), allocatable :: isInside
    type(region_type) :: srctool_region
    type(coord_type) :: srctool_coord
    character(buffsize) :: comment
    logical :: presentlyInsideReg
    real(r8) :: mjdref
    integer :: numRows,numGTIRows
    real(r8), dimension(:), pointer :: ras=>null(),decs=>null(),times=>null()
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null()
    character(buffsize) :: errmsg_prefix

    if(status > 100 .or. status == s_errRep)&
  return

    if(this%ext_type /= extCORRATT)&
call wrstop(functionName, "This call is only valid for CORRATT extensions.")

    numRows = this%getNumRows()
    call alloc(csStartIs, numRows, status)
    call alloc(csStopIs, numRows, status)
    call alloc(csirStartIs, numRows, status)
    call alloc(csirStopIs, numRows, status)
    call alloc(isInside, numRows, status)

    ! Default setup of GTI.
    !
    call gti%init(iREGIONGTI, status, fileInfo)
    call gti%set_all_useage(0,0,writeFlag=1,useInternallyFlag=1)
    call gti%alloc(status, 0)
    call this%header%getKeywordValue("MJDREF",mjdref)
    call gti%header%setKeywordValue("MJDREF",mjdref)
    gti%tel_id = this%tel_id
    gti%info%tableName = region_gti_ext//trim(int2str(gti%tel_id, status))

    ! Only if a region is found is the gti valid.
    !
    if(.not.allocated(regionsObj%regRows) .or. numRows<1) then
      write(errmsg_prefix, '(a)') trim(gti%info%tableName)//": "
      call gti%sanityCheck(status, errmsg_prefix) ! just to set the flag.
  return
    end if

    ! Find which CORRATT times fall inside or outside the region:
    !
    call this%get_col_ptr("RA",  ras)
    call this%get_col_ptr("DEC", decs)

    call regionsObj%get_srctool_encap_circle_region(srctool_region, status)
    srctool_coord%coordsys_code = COORDSYS_CODE_ICRS
    do ii = 1,numRows
      srctool_coord%long = ras( ii)
      srctool_coord%lat  = decs(ii)
      call srctool_regions_test_if_coord_in_region(srctool_coord, srctool_region, dummyI, status)
      isInside(ii) = (dummyI/=0)
    end do

    nullify(ras)
    nullify(decs)

    ! Determine the start and stop indices of time sequences within which successive times are separated by less than d_gap:
    !
    call this%get_col_ptr("TIME",  times)

    csStartIs(1) = 1
    nContSeq = 1
    do ii = 2, numRows
      if( abs(times(ii) - times(ii-1)) > d_gap) then
        csStopIs(nContSeq) = ii - 1
        nContSeq = nContSeq + 1
        csStartIs(nContSeq) = ii
      end if
    end do
    csStopIs(nContSeq) = numRows

    ! Now we want to divide these sequences further. We excise time gaps > d_gap out of GTIs, but essentially then we want to AND the result with GTIS for the attitude falling inside the region. We'll do that here by setting up new start and stop indices for entering or leaving respectively a period in which the attitude falls inside the region.
    !
    nContSeqInReg = 0
    do ii=1, nContSeq
      presentlyInsideReg = .false.

      do jj=csStartIs(ii),csStopIs(ii)
        if(isInside(jj)) then
          if(.not.presentlyInsideReg) then
            nContSeqInReg = nContSeqInReg + 1
            csirStartIs(nContSeqInReg) = jj
            presentlyInsideReg = .true.
          end if
        else ! att(jj) is outside the region.
          if(presentlyInsideReg) then
            csirStopIs(nContSeqInReg) = jj - 1 ! because att(jj-1) was the last att still inside the region.
            presentlyInsideReg = .false.
          end if
        end if
      end do

      if(presentlyInsideReg) then
        csirStopIs(nContSeqInReg) = csStopIs(ii)
      end if
    end do

    ! Calculate the number of GTIs. It would be nContSeqInReg, but we discard sequences which only have 1 member.
    !
    nGTIs = 0
    do ii=1, nContSeqInReg
      if(csirStopIs(ii)>csirStartIs(ii)) nGTIs = nGTIs + 1
    end do

    if(nGTIs<1) then
      nullify(times)
      write(errmsg_prefix, '(a)') trim(gti%info%tableName)//": "
      call gti%sanityCheck(status, errmsg_prefix) ! just to set the flag.
  return
    end if

    call gti%dealloc(status)
    call gti%alloc(status, nGTIs)

    ! Now set the GTIs.
    !
    call gti%get_col_ptr("START",  starts)
    call gti%get_col_ptr("STOP",   stops)

    jj = 0
    do ii=1, nContSeqInReg
      if(csirStopIs(ii)>csirStartIs(ii)) then
        jj = jj + 1
        starts(jj) = times(csirStartIs(ii))
        stops( jj) = times(csirStopIs(ii))
        if(csirStartIs(ii)>1) then
          if( abs(times(csirStartIs(ii)) - times(csirStartIs(ii)-1)) > d_gap) then
            comment = "TM"//trim(int2str(this%tel_id,status))&
              //": not enough corrat information before derived region gti -"
            comment = trim(comment)//" "//double2str(times(csirStartIs(ii)), status)
            call wrerr(functionName, warn, comment)
          end if
        end if
      end if
    end do

    nullify(stops)
    nullify(starts)
    nullify(times)

    numGTIRows = gti%getNumRows()

    write(errmsg_prefix, '(a)') trim(gti%info%tableName)//": "
    call gti%sanityCheck(status, errmsg_prefix)

    call dealloc(csStartIs, status)
    call dealloc(csStopIs, status)
    call dealloc(csirStartIs, status)
    call dealloc(csirStopIs, status)
    call dealloc(isInside, status)

  end subroutine derive_region_GTI_tExtension

  !.....................................................................
  subroutine apply_gti_tExtension(this, GTI, status)
    implicit none

    class(tExtension) :: this
    type(tGTI_tab), intent(in)    :: GTI
    integer,    intent(inout) :: status

    character(*), parameter :: functionName = "apply_gti_tExtension"
    integer :: numRows,ei,numGTIs,ii,jj
    type(tGTI_tab) :: tempGTI
    logical, dimension(:), allocatable :: mask,mask2
    real(r8) :: mjdref_this,mjdref_temp
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null(),times=>null()
    real(r8), dimension(:), pointer :: ptr_time_min=>null(),ptr_time_max=>null()
    logical, dimension(:), pointer :: time_is_null=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    numRows = this%getNumRows()
    if(numRows<1) then
      call wrerr(functionName, warn, "Zero rows in "//trim(this%info%tableName)&
        //" - can't apply GTIs.")
  return
    end if

    tempGTI = GTI

    call this%header%getKeywordValue("MJDREF",mjdref_this)
    call tempGTI%header%getKeywordValue("MJDREF",mjdref_temp)

    call tempGTI%get_col_ptr("START", starts)
    call tempGTI%get_col_ptr("STOP",  stops)

    if(mjdref_this /= mjdref_temp) then
      starts = starts + (mjdref_temp - mjdref_this)*3600.0_r8*24.0_r8
      stops  = stops  + (mjdref_temp - mjdref_this)*3600.0_r8*24.0_r8
      call tempGTI%header%setKeywordValue("MJDREF",mjdref_this)
    end if

    if(this%ext_type==extBADPIX) then
      numGTIs = size(starts)

      call this%get_col_ptr("TIMEMIN", ptr_time_min)
      call this%get_col_ptr("TIMEMAX", ptr_time_max)

      call alloc(mask, numRows, status)
      mask = .false.

      do ii=1,numRows
        do jj=1,numGTIs
          if(ptr_time_max(ii)>=starts(jj) .and. ptr_time_min(ii)<=stops(jj)) then
            mask(ii) = .true.
        exit
          end if
        end do
      end do

      nullify(ptr_time_min,ptr_time_max)
      call this%pack(mask, status)

      call dealloc(mask, status)
      nullify(starts,stops)
      call tempGTI%destroy(status)

  return
    end if ! this%ext_type==extBADPIX

    ! Deal with other sorts of extension:

    nullify(starts,stops)

    call this%get_col_ptr("TIME", times, time_is_null)
    call alloc(mask, numRows, status)
    mask = get_GTI_mask(times, tempGTI, status, .true., time_is_null, .true.)
    nullify(times, time_is_null)

    ! The task expmap will want to interpolate both DEADCOR and CORRATT values. To make this easier, we here add back in to the mask 1 value into the out-of-GTI region for each GTI start and stop.
    !
    call alloc(mask2, numRows, status)
    mask2 = .false.
    ei = 1
    GLOOP: do
      ! We want to find the table rows (ei values) that correspond to both the first and last rows within the iith GTI.
      !
      TLOOP1: do while(.not.mask(ei))
!if(doTest) print *, ei
        ei = ei + 1
        if(ei>numRows)&
    exit GLOOP
      end do TLOOP1

      ! If we reached here, ei is the number of the first table row within the current overlapping GTI.

      if(ei==1) then
        call wrerr(functionName, warn, "Extension "//trim(this%info%tableName)&
          //": no row before start of first overlapping GTI.")
      else
        mask2(ei-1) = .true.
      end if

      TLOOP2: do while(mask(ei))
        ei = ei + 1
        if(ei>numRows) then
          call wrerr(functionName, warn, "Extension "//trim(this%info%tableName)&
            //": no row after end of last overlapping GTI.")
    exit GLOOP
        end if
      end do TLOOP2

      ! If we reached here, ei is 1 more than the number of the last table row within the current overlapping GTI. (Note that we don't have to check that ei<numRows, since that is checked in TLOOP2.)

      mask2(ei) = .true.
    end do GLOOP

    mask = mask .or. mask2

    call this%pack(mask, status)
    call dealloc(mask2, status)
    call dealloc(mask, status)

    call tempGTI%destroy(status)

  end subroutine apply_gti_tExtension

  !.....................................................................
  subroutine dealloc_extensions(tables, status)
!**** haven't been able to figure out how to work the polymorphism with (de)alloc_bintables, so I've included this temporarily.
    implicit none

    type(tExtension), allocatable, dimension(:) :: tables
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "dealloc_extensions"

    if(allocated(tables)) then
      errmsg=""
      deallocate(tables, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
    end if

  end subroutine dealloc_extensions

  !.....................................................................
  subroutine alloc_extensions(tables, n, status)
!**** haven't been able to figure out how to work the polymorphism with (de)alloc_bintables, so I've included this temporarily.
    implicit none

    type(tExtension), allocatable, dimension(:) :: tables
    integer, intent(in) :: n
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "alloc_extensions"

    call dealloc_extensions(tables, status)
    errmsg=""
    allocate(tables(n), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

  end subroutine alloc_extensions

end module erolib_extensions_mod
