! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! MODULE: erolib_ffits_mod
!
! DESCRIPTION:
!> more functions using the fits files which could be useful
!> most functions are just wrappers including automatic error reports
!
!------------------------------------------------------------------------------
!
!  subroutine ffits_write_history(unit, history_str, status)
!  subroutine ffits_get_iounit(iounit, status)
!  subroutine ffits_free_iounit(iounit, status)
!  subroutine ffits_err(function, status, ierr, comment)
!  subroutine ffits_flush(unit, status)
!  subroutine ffits_get_time_str(datestr, timeref, status)
!  subroutine ffits_return_date(datestr, year, month, day, hour, minute, second, status)
!  subroutine ffits_create_img(unit, bitpix, naxis, naxes, status)
!  subroutine ffits_write_key_sequence(unit, keyroot, startno, max_keys, keyvals, comments, status)
!  subroutine ffits_read_key_sequence(unit, keyroot, startno, max_keys, keyvals, nfound, status)
!  subroutine ffits_delete_rows(unit, firstrow, nrows, status)
!  subroutine ffits_delete_rowlist(unit, rowlist, nrows, status)
!  subroutine ffits_get_rowsize(unit, rowsize, status)
!  subroutine ffits_read_col_name(unit, colname, values, status, colnum_return)
!  subroutine ffits_read_col(unit, colnum, frow, felem, nelements, nullval, values, status, anyf)
!  subroutine ffits_read_col_str(unit, colnum, frow, felem, nelements, nullval, values, status, anyf)
!  subroutine ffits_read_col_str_ll(unit, colnum, frow, felem, nelements, nullval, values, status, anyf)
!  subroutine ffits_read_col_ll(unit, colnum, frow, felem, nelements, nullval, values, status, anyf)
!  subroutine ffits_write_col(unit, colnum, frow, felem, nelements, values, status, nullval)
!  subroutine ffits_write_col_str(unit, colnum, frow, felem, nelements, values, status)
!  subroutine ffits_write_col_ll(unit, colnum, frow, felem, nelements, values, status)
!  subroutine ffits_write_col_str_ll(unit, colnum, frow, felem, nelements, values, status)
!  subroutine ffits_write_img_2D(unit, dim1, naxis1, naxis2, image, status, group)
!  subroutine ffits_read_img_2D(unit, dim1, naxis1, naxis2, image, status, group)
!  subroutine ffits_get_num_rows(unit, nrows, status)
!  subroutine ffits_get_num_cols(unit, ncols, status)
!  subroutine ffits_get_colnum(unit, casesen, coltemplate, colnum, status, not_found)
!  subroutine ffits_create_tbl(unit, nrows, tfields, ttype, tform, tunit, extname, varidat, status)
!  subroutine ffits_file_exists(filename, exists, status)
!  subroutine ffits_open_table(unit, filename, rwmode, status)
!  subroutine ffits_read_key_str(unit, keyword, keyval, status, comment)
!  subroutine ffits_read_key(unit, keyword, keyval, status, comment)
!  subroutine ffits_get_num_hdu(unit, hdunum, status)
!  subroutine ffits_get_name_bin_table(unit, hdu_name, status)
!  subroutine ffits_create_file(unit, filename, blocksize, status)
!  subroutine ffits_copy_file(iunit, ounit, previous, current, following, status)
!  subroutine ffits_copy_header(inunit, outunit, status)
!  subroutine ffits_close_file(unit, status)
!  subroutine close_FITS_file(unit, status)
!  subroutine ffits_get_num_hdus(unit, nhdu, status)
!  subroutine ffits_get_hdu_type(unit, hdutype, status)
!  subroutine ffits_movabs_hdu(unit, hdu, hdutype, status)
!  subroutine ffits_movnam_hdu(unit, hdutype, extname, status, extver)
!  subroutine ffits_open_file(unit, filename, rwmode, status)
!  subroutine open_FITS_file(filename, rwmode, unit, status, delete)
!  subroutine ffits_file_name(unit, filename, status)
!  subroutine ffits_delete_hdu(unit, hdutype, status)
!  subroutine ffits_modify_key_longstr(unit, keyword, keyval, status, comment)
!  subroutine ffits_delete_file(unit, status)
!  subroutine ffits_update_key(unit, keyword, keyval, comment, status)
!  subroutine ffits_update_key_str(unit, keyword, keyval, comment, status)
!  subroutine ffits_get_col_info(unit, colnum, ttype, tunit, datatype, repeat_val, tscal, tzero, tnull, tdisp, status)
!  subroutine ffits_get_col_datatype(unit, colnum, datacode, repeat_val, width, status)
!  subroutine ffits_insert_col(unit, colnum, ttype, tform, status)
!  subroutine ffits_write_tdim(unit, colnum, naxis, naxes, status)

module erolib_ffits_mod

  use param, only : wrerr
  use erolib_general_mod
  use erolib_str_mod, only : int2str

  implicit none

!logical, private, parameter :: doTest=.false.

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> attach history to a file - similar to cfitsio-routine fits_write_history
  !
  !
  !
  !> @param[in] unit, history_str
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_write_history(unit, history_str, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit

    character(*), parameter :: function = "ffits_write_history"
    character(*), intent(in) :: history_str

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTPHIS(unit, history_str, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_write_history

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get an unused iounit - similar to cfitsio-routine
  !
  !
  !
  !> @param[inout] iounit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_get_iounit(iounit, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status, iounit

    character(*), parameter :: function = "ffits_get_iounit"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTGIOU(iounit, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_get_iounit

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free a used iounit - similar to cfitsio-routine
  !
  !
  !
  !> @param[inout] iounit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_free_iounit(iounit, status)
    implicit none

    integer :: ierr
!    integer, intent(inout) :: status, iounit
    integer, intent(in) :: iounit
    integer, intent(inout) :: status

    character(*), parameter :: function = "ffits_free_iounit"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTFIOU(iounit, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_free_iounit

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> ffits_err - similar to fitserr except ierr
  !
  !
  !
  !> @param[in] function
  !> @param[inout] status, ierr
  !---------------------------------------------------------------------------

  subroutine ffits_err(function, status, ierr, comment)
    implicit none

    integer, intent(inout) :: status, ierr
    character(*) :: function
    character(30) :: errtext
    character(*), intent(in), optional :: comment

    if(status /= ok) then
      select case (ierr)
        case (warn, c_err)
          call ftgerr(status, errtext)
          if(present(comment)) then
            call wrerr(function, ierr, trim(errtext) // " - " // trim(comment))
          else
            call wrerr(function, ierr, errtext)
          end if
          status = ierr
        case (0, s_err)
          !                    call fitserr(function, status, ierr)
          call ftgerr(status, errtext)
          !                    call wrerr(function, ierr, errtext)
          if(present(comment)) then
            call wrerr(function, s_err_no_exit, trim(errtext) // " - " // trim(comment))
          else
            call wrerr(function, s_err_no_exit, errtext)
          end if
          status = s_errRep
        case (ignore)
            ! ignore the report
        case (s_errRep)
            ! severe error was already reported
      end select
    else
      ! set status back to ierr
      status = ierr
    end if

  end subroutine ffits_err

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> flush all to the fits file, useful after writing keywords for tscale ..., or inserting tables / images
  !> which may not be written to the file (unit) yet
  !
  !
  !> @param[in] inParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_flush(unit, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(inout) :: status

    character(*), parameter :: function = "ffits_flush"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTFLUS(unit, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_flush

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the current date as a string - similar to cfitsio-routine
  !
  !
  !
  !> @param[in] datestr, timeref
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_get_time_str(datestr, timeref, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status

    character(buffsize), intent(out) :: datestr
    integer, intent(out) :: timeref

    character(*), parameter :: function = "ffits_get_time_str"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTGSTM(datestr, timeref, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_get_time_str

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> break a date string into numbers
  !
  !
  !
  !> @param[in] inParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_return_date(datestr, year, month, day, hour, minute, second, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status

    character(*), intent(in) :: datestr

    integer, intent(out) :: year, month, day, hour, minute, second

    character(*), parameter :: function = "ffits_return_date"


    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls

    call FTS2TM(datestr, year, month, day, hour, minute, second, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_return_date

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create an image / primary extension - similar to cfitsio-routine name
  !> One can force a new primary array to be inserted at the beginning of the
  !> FITS file by setting status = -9 prior to calling the routine
  !
  !
  !> @param[in] inParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_create_img(unit, bitpix, naxis, naxes, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit
    integer, intent(in) :: bitpix
    integer, intent(in) :: naxis
    integer, dimension(naxis), intent(in) :: naxes

    character(*), parameter :: function = "ffits_create_img"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! exception for status = -9
    if(ierr == -9) then
      ierr = 0
      status = -9
    end if

    ! the fitsio calls
    call FTIIMG(unit, bitpix, naxis, naxes, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_create_img

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write sequential numbered keywords - no similar cfitsio-routine exists
  !
  !
  !
  !> @param[in] inParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_write_key_sequence(unit, keyroot, startno, max_keys, keyvals, comments, status)
    implicit none

    integer :: ierr, lencomment
    integer, intent(inout) :: status
    integer, intent(in) :: startno, max_keys, unit

    character(*), parameter :: function = "ffits_write_key_sequence"
    character(*), intent(in) :: keyroot

    character(*), dimension(:) :: comments

    class(*), dimension(:), intent(inout) :: keyvals

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    if(size(comments) == 1) then
      lencomment = len_trim(comments(1)(:))
      if(comments(1)(lencomment:lencomment) /= "&") then
        comments(1) = trim(comments(1)) // "&"
      end if
    end if

    ! the fitsio calls
    select type(keyvals)
      type is ( integer )
      call FTPKNJ(unit, keyroot, startno, max_keys, keyvals, comments, status)
      type is ( character(*) )
      call FTPKNS(unit, keyroot, startno, max_keys, keyvals, comments, status)
      class default
      print *, "write key sequence for that type is not yet supported"
      status = s_errRep
    end select

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_write_key_sequence

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read sequential numbered keywords - no similar cfitsio-routine exists
  !
  !
  !
  !> @param[in] inParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_read_key_sequence(unit, keyroot, startno, max_keys, keyvals, nfound, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(out) :: nfound
    integer, intent(in) :: startno, max_keys, unit

    character(*), parameter :: function = "ffits_read_key_sequence"
    character(*), intent(in) :: keyroot

    class(*), dimension(:), intent(inout) :: keyvals

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    select type(keyvals)
      type is ( real )
      call FTGKNE(unit, keyroot, startno, max_keys, keyvals, nfound, status)
      type is ( real(dp) )
      call FTGKND(unit, keyroot, startno, max_keys, keyvals, nfound, status)
      type is ( integer )
      call FTGKNJ(unit, keyroot, startno, max_keys, keyvals, nfound, status)
      type is ( character(*) )
      call FTGKNS(unit, keyroot, startno, max_keys, keyvals, nfound, status)
      class default
      print *, "read key sequence for that type is not yet supported"
      status = s_errRep
    end select

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_read_key_sequence

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete nrows starting from firstrow in the CHDU - similar to cfitsio-routine fits_delete_rows
  !
  !
  !
  !> @param[in] inParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_delete_rows(unit, firstrow, nrows, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: firstrow, nrows, unit

    character(*), parameter :: function = "ffits_delete_rows"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTDROW(unit, firstrow, nrows, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_delete_rows

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete a list of rows starting from the CHDU - similar to cfitsio-routine fits_delete_rows
  !
  !
  !
  !> @param[in] inParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_delete_rowlist(unit, rowlist, nrows, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: nrows, unit
    integer, dimension(:), intent(in) :: rowlist

    character(*), parameter :: function = "ffits_delete_rowlist"


    ! return automatically
    if (status > 0) then
      return
    end if

    ierr = -status
    status = 0

    ! the fitsio calls
    call FTDRWS(unit, rowlist, nrows, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_delete_rowlist

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> determine the optimal number to read/write rows - similar to cfitsio-routine fits_get_rowsize
  !
  !
  !
  !> @param[in] unit, rowsize
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_get_rowsize(unit, rowsize, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(out) :: rowsize
    integer, intent(in) :: unit

    character(*), parameter :: function = "ffits_get_rowsize"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTGRSZ(unit, rowsize, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_get_rowsize


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read in the whole column of a fits table
  !
  !
  !
  !> @param[in] inParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_read_col_name(unit, colname, values, status, colnum_return)
    implicit none

    integer :: ierr, nelements, colnum, rowsize, nrows
    integer, intent(inout) :: status
    integer, intent(in) :: unit
    integer :: frow, lrow, felem

    integer, optional, intent(out) :: colnum_return

    character(*), intent(in) :: colname
    character(*), parameter :: function = "ffits_read_col_name"

    class(*), dimension(:), intent(inout) :: values

    logical :: anyf
    integer(i1), dimension(1) :: i1_values

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
    end if

    call ffits_get_num_rows(unit, nelements, status)
    call ffits_get_rowsize(unit, rowsize, status)
    call ffits_get_colnum(unit, .false., colname, colnum, status)

    ! in case you want to save the column number
    if(present(colnum_return)) colnum_return = colnum

    anyf = .false.
    felem = 1

    ! the fitsio calls
    do frow = 1, nelements, rowsize
      nrows = min(nelements - frow + 1, rowsize)
      lrow = frow + nrows - 1
      select type(values)
        type is ( real(r4) )
        call FTGCVE(unit, colnum, frow, felem, nrows, 0.0_r4, values(frow:lrow), anyf, status)
        type is ( real(r8) )
        call FTGCVD(unit, colnum, frow, felem, nrows, 0.0_r8, values(frow:lrow), anyf, status)
        type is ( integer(i1))
        call FTGCVB(unit, colnum, frow, felem, nrows, 0_i1, values(frow:lrow), anyf, status)
        type is ( integer(i4) )
        call FTGCVJ(unit, colnum, frow, felem, nrows, 0_i4, values(frow:lrow), anyf, status)
        type is ( character(*) )
        if(len(values) == 1)  then
!          call FTGCVB(unit, colnum, frow, felem, nrows, 0_i1, values(frow:lrow), anyf, status)
          i1_values(1) = iachar(values(1),kind=i1)
          call FTGCVB(unit, colnum, frow, felem, nrows, 0_i1, i1_values(frow:lrow), anyf, status)
        else
          call FTGCVS(unit, colnum, frow, felem, nrows, "", values(frow:lrow), anyf, status)
        end if
        type is ( integer(i2) )
        call FTGCVI(unit, colnum, frow, felem, nrows, 0_i2, values(frow:lrow), anyf, status)
        class default
        print *, "read col for that type is not yet supported"
        status = s_errRep
      end select
    end do

    ! error report
    if(status /= ok) call wrerr(function, ierr, "reading column - " // trim(colname) )

    if(anyf .eqv. .true.) then
      call wrerr(function, s_err, "values in column " // trim(colname)// " are undefined")
      status = s_errRep
    end if

  end subroutine ffits_read_col_name

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read a column from a fits binary table - similar to cfitsio-routine fits_read_col
  !> undefined values are not supported, if anyf is provided automatic error is disabled
  !
  !
  !> @param[in] unit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_read_col(unit, colnum, frow, felem, nelements, nullval, values, status, anyf)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum, frow, felem, nelements

    class(*), intent(in) :: nullval
    class(*),  dimension(:), intent(inout) :: values

    integer(i1) :: nullval_i1
    integer(i2) :: nullval_i2
    integer(i4) :: nullval_i4
    integer(i8) :: nullval_i8
    real(r4) :: nullval_r4
    real(r8) :: nullval_r8

    logical, intent(out), optional :: anyf
    logical :: anyf_tmp

    character(*), parameter :: function = "ffits_read_col"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    select type (nullval)
      type is ( real(r4) )
      nullval_r4 = nullval
      type is ( real(r8) )
      nullval_r8 = nullval
      type is ( integer(i8))
      nullval_i8 = nullval
      type is ( integer(i4) )
      nullval_i4 = nullval
      type is ( integer(i2) )
      nullval_i2 = nullval
      type is ( integer(i1) )
      nullval_i1 = nullval
    end select

    ! the fitsio calls
    select type(values)
      type is ( real(r4) )
      call FTGCVE(unit, colnum, frow, felem, nelements, nullval_r4, values, anyf_tmp, status)
      type is ( real(r8) )
      call FTGCVD(unit, colnum, frow, felem, nelements, nullval_r8, values, anyf_tmp, status)
      type is ( integer(i8))
      call FTGCVK(unit, colnum, frow, felem, nelements, nullval_i8, values, anyf_tmp, status)
      type is ( integer(i4) )
      call FTGCVJ(unit, colnum, frow, felem, nelements, nullval_i4, values, anyf_tmp, status)
      type is ( integer(i2) )
      call FTGCVI(unit, colnum, frow, felem, nelements, nullval_i2, values, anyf_tmp, status)
      type is ( integer(i1) )
      call FTGCVB(unit, colnum, frow, felem, nelements, nullval_i1, values, anyf_tmp, status)
      class default
      print *, "read col for that type is not yet supported"
      status = s_errRep
    end select

    ! error report
    call ffits_err(function, status, ierr)

    if(present(anyf)) then
      anyf = anyf_tmp
      return
    end if

  !    if(anyf_tmp .eqv. .true.) then
  !      call wrerr(function, s_err, "values in column number" // trim(int2str(colnum)) // " are undefined")
  !      status = s_errRep
  !    end if

  end subroutine ffits_read_col

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read an asci column from a fits binary table - similar to cfitsio-routine fits_read_col
  !> undefined values are not supported, if anyf is provided automatic error is disabled
  !
  !
  !> @param[in] unit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_read_col_str(unit, colnum, frow, felem, nelements, nullval, values, status, anyf)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum, frow, felem, nelements

    character(*), intent(in) :: nullval
    character(*),  dimension(:), intent(inout) :: values

    logical, intent(out), optional :: anyf
    logical :: anyf_tmp

    character(*), parameter :: function = "ffits_read_col_str"
    integer(i1), dimension(1) :: i1_values

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    if(len(values) == 1)  then
!      call FTGCVB(unit, colnum, frow, felem, nelements, 0, values, anyf_tmp, status)
      call FTGCVB(unit, colnum, frow, felem, nelements, 0_i1, i1_values, anyf_tmp, status)
      values(1) = char(i1_values(1), kind=i1)
    else
      call FTGCVS(unit, colnum, frow, felem, nelements, nullval, values, anyf_tmp, status)
    end if

    ! error report
    call ffits_err(function, status, ierr)

    if(present(anyf)) then
      anyf = anyf_tmp
      return
    end if

    if(anyf_tmp .eqv. .true.) then
      call wrerr(function, s_err, "values in column number" // trim(int2str(colnum)) // " are undefined")
      status = s_errRep
    end if

  end subroutine ffits_read_col_str

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read an asci column from a fits binary table - similar to cfitsio-routine fits_read_col
  !> undefined values are not supported, if anyf is provided automatic error is disabled
  !
  !
  !> @param[in] unit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_read_col_str_ll(unit, colnum, frow, felem, nelements, nullval, values, status, anyf)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum
    integer(i8), intent(in) :: frow, felem, nelements

    character(*), intent(in) :: nullval
    character(*),  dimension(:), intent(inout) :: values

    logical, intent(out), optional :: anyf
    logical :: anyf_tmp

    character(*), parameter :: function = "ffits_read_col_str_ll"
    integer(i1), dimension(1) :: i1_values

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    if(len(values) == 1)  then
!      call FTGCVBLL(unit, colnum, frow, felem, nelements, 0, values, anyf_tmp, status)
      call FTGCVBLL(unit, colnum, frow, felem, nelements, 0_i1, i1_values, anyf_tmp, status)
      values(1) = char(i1_values(1), kind=i1)
    else
      call FTGCVSLL(unit, colnum, frow, felem, nelements, nullval, values, anyf_tmp, status)
    end if

    ! error report
    call ffits_err(function, status, ierr)

    if(present(anyf)) then
      anyf = anyf_tmp
      return
    end if

    if(anyf_tmp .eqv. .true.) then
      call wrerr(function, s_err, "values in column number" // trim(int2str(colnum)) // " are undefined")
      status = s_errRep
    end if

  end subroutine ffits_read_col_str_ll


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read a column from a fits binary table - similar to cfitsio-routine fits_read_col
  !> undefined values are not supported, if anyf is provided automatic error is disabled
  !> can be used to read in tables with more than Integer(I4) elements
  !
  !> @param[in] unit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_read_col_ll(unit, colnum, frow, felem, nelements, nullval, values, status, anyf)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum
    integer(i8), intent(in) :: frow, felem, nelements

    class(*), intent(in) :: nullval
    class(*),  dimension(:), intent(inout) :: values

    integer(i1) :: nullval_i1
    integer(i2) :: nullval_i2
    integer(i4) :: nullval_i4
    integer(i8) :: nullval_i8
    real(r4) :: nullval_r4
    real(r8) :: nullval_r8

    logical, intent(out), optional :: anyf
    logical :: anyf_tmp

    character(*), parameter :: function = "ffits_read_col_ll"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    select type (nullval)
      type is ( real(r4) )
      nullval_r4 = nullval
      type is ( real(r8) )
      nullval_r8 = nullval
      type is ( integer(i8))
      nullval_i8 = nullval
      type is ( integer(i4) )
      nullval_i4 = nullval
      type is ( integer(i2) )
      nullval_i2 = nullval
      type is ( integer(i1) )
      nullval_i1 = nullval
    end select

    ! the fitsio calls
    select type(values)
      type is ( real(r4) )
      call FTGCVELL(unit, colnum, frow, felem, nelements, nullval_r4, values, anyf_tmp, status)
      type is ( real(r8) )
      call FTGCVDLL(unit, colnum, frow, felem, nelements, nullval_r8, values, anyf_tmp, status)
      type is ( integer(i8))
      call FTGCVKLL(unit, colnum, frow, felem, nelements, nullval_i8, values, anyf_tmp, status)
      type is ( integer(i4) )
      call FTGCVJLL(unit, colnum, frow, felem, nelements, nullval_i4, values, anyf_tmp, status)
      type is ( integer(i2) )
      call FTGCVILL(unit, colnum, frow, felem, nelements, nullval_i2, values, anyf_tmp, status)
      type is ( integer(i1) )
      call FTGCVBLL(unit, colnum, frow, felem, nelements, nullval_i1, values, anyf_tmp, status)
      class default
      print *, "read col for that type is not yet supported"
      status = s_errRep
    end select

    ! error report
    call ffits_err(function, status, ierr)

    if(present(anyf)) then
      anyf = anyf_tmp
      return
    end if

  !    if(anyf_tmp .eqv. .true.) then
  !      call wrerr(function, s_err, "values in column number" // trim(int2str(colnum)) // " are undefined")
  !      status = s_errRep
  !    end if

  end subroutine ffits_read_col_ll

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write a column to a fits binary table - similar to cfitsio-routine fits_write_col
  !> undefined values are not supported, if anyf is provided automatic error is disabled
  !
  !
  !> @param[in] unit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_write_col(unit, colnum, frow, felem, nelements, values, status, nullval)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum, frow, felem, nelements

    integer(i1) :: nullval_i1
    integer(i2) :: nullval_i2
    integer(i4) :: nullval_i4
    integer(i8) :: nullval_i8
    real(r4) :: nullval_r4
    real(r8) :: nullval_r8

    class(*), intent(in), optional :: nullval

    class(*),  dimension(:), intent(in) :: values

    character(*), parameter :: function = "ffits_write_col"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    if(present(nullval)) then
      select type (nullval)
        type is ( real(r4) )
        nullval_r4 = nullval
        type is ( real(r8) )
        nullval_r8 = nullval
        type is ( integer(i8))
        nullval_i8 = nullval
        type is ( integer(i4) )
        nullval_i4 = nullval
        type is ( integer(i2) )
        nullval_i2 = nullval
        type is ( integer(i1) )
        nullval_i1 = nullval
      end select

      ! the fitsio calls
      select type(values)
        type is ( real(r4) )
        call FTPCNE(unit,colnum,frow,felem,nelements,values, nullval_r4, status)
        type is ( real(r8) )
        call FTPCND(unit,colnum,frow,felem,nelements,values, nullval_r8, status)
        type is ( integer(i8))
        call FTPCNK(unit,colnum,frow,felem,nelements,values, nullval_i8, status)
        type is ( integer(i4))
        call FTPCNJ(unit,colnum,frow,felem,nelements,values, nullval_i4, status)
        type is ( integer(i2))
        call FTPCNI(unit,colnum,frow,felem,nelements,values, nullval_i2, status)
        type is ( integer(i1))
        call FTPCNB(unit,colnum,frow,felem,nelements,values, nullval_i1, status)
        class default
        print *, "write col for that type is not yet supported"
        status = s_errRep
      end select
    else
          ! the fitsio calls
      select type(values)
        type is ( real(r4) )
        call FTPCLE(unit,colnum,frow,felem,nelements,values, status)
        type is ( real(r8) )
        call FTPCLD(unit,colnum,frow,felem,nelements,values, status)
        type is ( integer(i8))
        call FTPCLK(unit,colnum,frow,felem,nelements,values, status)
        type is ( integer(i4))
        call FTPCLJ(unit,colnum,frow,felem,nelements,values, status)
        type is ( integer(i2))
        call FTPCLI(unit,colnum,frow,felem,nelements,values, status)
        type is ( integer(i1))
        call FTPCLB(unit,colnum,frow,felem,nelements,values, status)
        class default
        print *, "write col for that type is not yet supported"
        status = s_errRep
      end select
    end if


    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_write_col

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write an ascii column to a fits binary table - similar to cfitsio-routine fits_write_col
  !> undefined values are not supported, if anyf is provided automatic error is disabled
  !
  !
  !> @param[in] unit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_write_col_str(unit, colnum, frow, felem, nelements, values, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum, frow, felem, nelements

    character(*),  dimension(:), intent(in) :: values

    character(*), parameter :: function = "ffits_write_col_str"
    integer(i1), dimension(1) :: i1_values

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    if(len(values) == 1)  then
!      call FTPCLB(unit,colnum,frow,felem,nelements,values, status)
      i1_values(1) = iachar(values(1),kind=i1)
      call FTPCLB(unit,colnum,frow,felem,nelements,i1_values, status)
    else
      call FTPCLS(unit,colnum,frow,felem,nelements,values, status)
    end if

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_write_col_str

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write a column to a fits binary table - similar to cfitsio-routine fits_write_col
  !> undefined values are not supported, if anyf is provided automatic error is disabled
  !
  !
  !> @param[in] unit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_write_col_ll(unit, colnum, frow, felem, nelements, values, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum
    integer(i8), intent(in) :: frow, felem, nelements

    class(*),  dimension(:), intent(in) :: values

    character(*), parameter :: function = "ffits_write_col_ll"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    select type(values)
      type is ( real )
      call FTPCLELL(unit,colnum,frow,felem,nelements,values, status)
      type is ( real(dp) )
      call FTPCLDLL(unit,colnum,frow,felem,nelements,values, status)
      type is ( integer(i8))
      call FTPCLKLL(unit,colnum,frow,felem,nelements,values, status)
      type is ( integer(i4))
      call FTPCLJLL(unit,colnum,frow,felem,nelements,values, status)
      type is ( integer(i2))
      call FTPCLILL(unit,colnum,frow,felem,nelements,values, status)
      type is ( integer(i1))
      call FTPCLBLL(unit,colnum,frow,felem,nelements,values, status)
      class default
      print *, "write col for that type is not yet supported"
      status = s_errRep
    end select

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_write_col_ll

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write an ascii column to a fits binary table - similar to cfitsio-routine fits_write_col
  !> undefined values are not supported, if anyf is provided automatic error is disabled
  !
  !
  !> @param[in] unit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_write_col_str_ll(unit, colnum, frow, felem, nelements, values, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum
    integer(i8), intent(in) :: frow, felem, nelements

    character(*),  dimension(:), intent(in) :: values

    character(*), parameter :: function = "ffits_write_col_str_ll"
    integer(i1), dimension(1) :: i1_values

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    if(len(values) == 1)  then
!      call FTPCLBLL(unit,colnum,frow,felem,nelements,values, status)
      i1_values(1) = iachar(values(1),kind=i1)
      call FTPCLBLL(unit,colnum,frow,felem,nelements,i1_values, status)
    else
      call FTPCLSLL(unit,colnum,frow,felem,nelements,values, status)
    end if

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_write_col_str_ll


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> put a 2D image into a data array (primary / image extension) which has at least the size of the input image - similar to cfitsio-routine fits_write_img_2D
  !
  !
  !
  !> @param[in] unit, image
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_write_img_2D(unit, dim1, naxis1, naxis2, image, status, group)
    implicit none

    integer :: ierr, grp
    integer, intent(inout) :: status
    integer, intent(in), optional :: group
    integer, intent(in) :: naxis1, naxis2
    integer, intent(in) :: dim1, unit

    class(*), dimension(:,:), intent(in) :: image

    character(*), parameter :: functionName = "ffits_write_img_2D"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    grp = 0
    if(present(group)) grp = group

    ! the fitsio calls
    select type(image)
      type is ( integer(i1))
        call FTP2DB(unit, grp, dim1, naxis1, naxis2, image, status)
      type is ( integer(i2))
        call FTP2DI(unit, grp, dim1, naxis1, naxis2, image, status)
      type is ( integer(i4))
        call FTP2DJ(unit, grp, dim1, naxis1, naxis2, image, status)
      type is ( real(r4))
        call FTP2DE(unit, grp, dim1, naxis1, naxis2, image, status)
      type is ( real(r8))
        call FTP2DD(unit, grp, dim1, naxis1, naxis2, image, status)
    !                reshape(image, [naxis1*naxis2]), status)
      class default
call wrstop(functionName, "Unrecognized choice of image data type.")
    end select

    ! error report
    call ffits_err(functionName, status, ierr)

  end subroutine ffits_write_img_2D

  subroutine ffits_write_img_3D(unit, dim1, dim2, naxis1, naxis2, naxis3, cube, status, group)
    implicit none

    integer :: ierr, grp
    integer, intent(inout) :: status
    integer, intent(in), optional :: group
    integer, intent(in) :: naxis1, naxis2, naxis3
    integer, intent(in) :: dim1, dim2, unit

    class(*), dimension(:,:,:), intent(in) :: cube

    character(*), parameter :: functionName = "ffits_write_img_3D"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    grp = 0
    if(present(group)) grp = group

    ! the fitsio calls
    select type(cube)
      type is ( integer(i1))
        call FTP3DB(unit, grp, dim1, dim2, naxis1, naxis2, naxis3, cube, status)
      type is ( integer(i2))
        call FTP3DI(unit, grp, dim1, dim2, naxis1, naxis2, naxis3, cube, status)
      type is ( integer(i4))
        call FTP3DJ(unit, grp, dim1, dim2, naxis1, naxis2, naxis3, cube, status)
      type is ( real(r4))
        call FTP3DE(unit, grp, dim1, dim2, naxis1, naxis2, naxis3, cube, status)
      type is ( real(r8))
        call FTP3DD(unit, grp, dim1, dim2, naxis1, naxis2, naxis3, cube, status)
    !                reshape(image, [naxis1*naxis2]), status)
      class default
call wrstop(functionName, "Unrecognized choice of cube data type.")
    end select

    ! error report
    call ffits_err(functionName, status, ierr)

  end subroutine ffits_write_img_3D

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read a 2D image from a data array (primary / image extension) which has at least the size of the input image - similar to cfitsio-routine fits_read_img_2D
  !
  !
  !
  !> @param[in] unit, image
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_read_img_2D(unit, dim1, naxis1, naxis2, image, status, group)
    implicit none

    integer :: ierr, grp
    integer, intent(inout) :: status
    integer, intent(in), optional :: group
    integer, intent(in) :: naxis1, naxis2
    integer, intent(in) :: dim1, unit

    logical :: anyf

!    class(*), dimension(:,:), intent(out) :: image
    class(*), dimension(:,:) :: image

    character(*), parameter :: function = "ffits_read_img_2D"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    grp = 0
    if(present(group)) grp = group

    ! the fitsio calls
    select type(image)
      type is ( integer(i1))
      call FTG2DB(unit, grp, 0, dim1, naxis1, naxis2, image, anyf, status)
      type is ( integer(i2))
      call FTG2DI(unit, grp, 0, dim1, naxis1, naxis2, image, anyf, status)
      type is ( integer(i4))
      call FTG2DJ(unit, grp, 0, dim1, naxis1, naxis2, image, anyf, status)
    end select

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_read_img_2D

  subroutine ffits_read_img_3D(unit, dim1, dim2, naxis1, naxis2, naxis3, cube, status, group)
    implicit none

    integer :: ierr, grp
    integer, intent(inout) :: status
    integer, intent(in), optional :: group
    integer, intent(in) :: naxis1, naxis2, naxis3
    integer, intent(in) :: dim1, dim2, unit

    logical :: anyf

    class(*), dimension(:,:,:), intent(in) :: cube

    character(*), parameter :: function = "ffits_read_img_3D"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    grp = 0
    if(present(group)) grp = group

    ! the fitsio calls
    select type(cube)
      type is ( integer(i1))
      call FTG3DB(unit, grp, 0, dim1, dim2, naxis1, naxis2, naxis3, cube, anyf, status)
      type is ( integer(i2))
      call FTG3DI(unit, grp, 0, dim1, dim2, naxis1, naxis2, naxis3, cube, anyf, status)
      type is ( integer(i4))
    !      call FTP3DJ(unit, grp, dim1, naxis1, naxis2, cube, status)
!*** 2019-09-11 IMS: why is this commented out?
    end select

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_read_img_3D

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the number of rows - similar to cfitsio-routine fits_get_num_rows
  !
  !
  !
  !> @param[in] unit
  !> @param[out] nrows
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_get_num_rows(unit, nrows, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(out) :: nrows
    integer, intent(inout) :: status

    character(*), parameter :: function = "ffits_get_num_rows"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTGNRW(unit, nrows, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_get_num_rows

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the number of columns - similar to cfitsio-routine fits_get_num_cols
  !
  !
  !
  !> @param[in] unit
  !> @param[out] ncols
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_get_num_cols(unit, ncols, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(out) :: ncols
    integer, intent(inout) :: status
    character(*), parameter :: function = "ffits_get_num_cols"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTGNCL(unit, ncols, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_get_num_cols

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the column number - similar to cfitsio-routine fits_get_colnum
  !
  !
  !
  !> @param[in] unit
  !> @param[out] casesen
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_get_colnum(unit, casesen, coltemplate, colnum, status, not_found)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(out) :: colnum
    integer, intent(inout) :: status
    logical, intent(out), optional :: not_found

    character(*), parameter :: function = "ffits_get_colnum"
    character(*), intent(in) :: coltemplate

    logical, intent(in) :: casesen

    character(buffsize) :: comment

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if
        
    ! the fitsio calls
    call FTGCNO(unit, casesen, coltemplate, colnum, status)
    if (status == 219) then
      if (present(not_found)) not_found=.true.
      comment = "Column name '"//trim(coltemplate)//"'."
      if(ierr /= ignore) ierr = c_err
    end if

    ! error report
    call ffits_err(function, status, ierr, comment) ! does status=ierr if status is not bad.

  end subroutine ffits_get_colnum

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> insert a new binary table extension - similar to cfitsio-routine fits_create_tbl
  !
  !
  !
  !> @param[in] unit, nrows, tfields, ttype, tfor, tunit, extname
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_create_tbl(unit, nrows, tfields, ttype, tform, tunit, extname, varidat, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, nrows, tfields, varidat

    character(*), parameter :: function = "ffits_create_tbl"
    character(*), dimension(:), intent(in) :: ttype, tform, tunit
    character(*), intent(in) :: extname

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTIBIN(unit, nrows, tfields, ttype, tform, tunit, extname, varidat, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_create_tbl




  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> wrapper of FTEXISTS
  !
  !
  !
  !> @param[in] filename
  !> @param[out] exsists
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_file_exists(filename, exists, status)
    implicit none

    integer :: ierr
    integer, intent(out) :: exists
    integer, intent(inout) :: status

    character(*), parameter :: function = "ffits_file_exists"
    character(*), intent(in) :: filename

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    call FTEXIST(filename, exists, status)
    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_file_exists


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> wrapper to open FITS binary table - similar to cfitsio-routine fits_open_table
  !
  !
  !
  !> @param[in] unit
  !> @param[in] filename
  !> @param[in] rwmode
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_open_table(unit, filename, rwmode, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: rwmode
    integer, intent(in) :: unit
    integer, intent(inout) :: status

    character(*), intent(in) :: filename
    character(*), parameter :: function = "ffits_open_table"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTTOPN(unit ,filename, rwmode, status)
    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_open_table

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the keyword as a character string from a FITS file - similar to cfitsio-routine fits_read_str
  !
  !> @param[in] unit, keyword
  !> @param[out] keyval
  !> @param[inout] status
  !> @param[out, optional] comment
  !---------------------------------------------------------------------------

  subroutine ffits_read_key_str(unit, keyword, keyval, status, comment)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(inout) :: status

    character(*), parameter :: function = "ffits_read_key_str"
    character(*), intent(in) :: keyword
    character(fkeylength) :: fkeyword
    character(*), intent(out) :: keyval
    character(fcommentlength) :: tmp_comment
    character(*), intent(out), optional :: comment

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    fkeyword = keyword
    call FTGKYS(unit, keyword, keyval, tmp_comment, status)
    if(present(comment)) comment = tmp_comment

    ! error report
    call ffits_err(function, status, ierr, "keyword = " // trim(keyword))

  end subroutine ffits_read_key_str

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the integer number of a keyword in a FITS file - similar to cfitsio-routine fits_read_key
  !
  !
  !
  !> @param[in] unit
  !> @param[out] outParam
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_read_key(unit, keyword, keyval, status, comment)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(inout) :: status

    class(*), intent(inout) :: keyval

    character(fkeylength) :: fkeyword
    character(*), intent(in) :: keyword

    character(*), intent(out), optional :: comment
    character(fcommentlength) :: tmp_comment
    character(*), parameter :: function = "ffits_read_key"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    fkeyword = keyword
    select type(keyval)
      type is ( real )
      call FTGKYE(unit, fkeyword, keyval, tmp_comment, status)
      type is ( real(dp) )
      call FTGKYD(unit, fkeyword, keyval, tmp_comment, status)
      type is ( integer )
      call FTGKYJ(unit, fkeyword, keyval, tmp_comment, status)
      class default
      print *, "read key for that type is not yet supported"
      status = s_errRep
    end select
    if(present(comment)) comment = tmp_comment

    ! error report
    call ffits_err(function, status, ierr , "keyword = " // trim(keyword))

  end subroutine ffits_read_key

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> wrapper around FTTHDU(unit, hdunum, status)
  !
  !
  !
  !> @param[in] unit
  !> @param[inout] hdunum
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_get_num_hdu(unit, hdunum, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(out) :: hdunum
    integer, intent(inout) :: status
    character(*), parameter :: function = "ffits_get_num_hdu"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if
        
    ierr = -status
    status = 0

    call FTTHDU(unit, hdunum, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_get_num_hdu

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the names of the binary table extension in a FITS file
  !
  !
  !
  !> @param[in] unit
  !> @param[inout] hdu_name
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_get_name_bin_table(unit, hdu_name, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(inout) :: status
    integer :: hdunum, nbintab, ii, hdutype

    character(fkeylength), intent(inout), allocatable, dimension(:) :: hdu_name
    character(fkeylength) :: key
    character(*), parameter :: function = "ffits_get_name_bin_table"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if
        
    ! workfunctions
    call ffits_get_num_hdu(unit, hdunum, status)

    ! determine the number of binary table hdu extension
    nbintab = 0
    do ii = 1, hdunum
      call ffits_movabs_hdu(unit, ii, hdutype, status)
      ! 2 == binary table
      if (hdutype == 2) then
        nbintab = nbintab + 1
      end if
    end do

    ! allocate memory for the binary table hdu extension names
    call alloc(hdu_name, nbintab, status)

    nbintab = 0
    do ii = 1, hdunum
      call ffits_movabs_hdu(unit, ii, hdutype, status)
      ! 2 == binary table
      if (hdutype == 2) then
        nbintab = nbintab + 1
        key = "EXTNAME"
        call ffits_read_key_str(unit, key, hdu_name(nbintab), status)
      end if
    end do

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_get_name_bin_table

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create a fits file - similar to cfitsio-routine fits_create_file
  !
  !
  !
  !> @param[in] unit, blocksize, filename
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_create_file(unit, filename, blocksize, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(in) :: blocksize
    integer, intent(inout) :: status

    character(*), intent(in) :: filename
    character(*), parameter :: function = "ffits_create_file"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTINIT(unit, filename, blocksize, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_create_file

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> copy the parts / complete hdu from input to output - similar to cfitsio-routine fits_copy_file
  !
  !
  !
  !> @param[in] inunit, ounit, previous, current, following
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_copy_file(iunit, ounit, previous, current, following, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: iunit, ounit, previous, current, following
    integer, intent(inout) :: status

    character(*), parameter :: function = "ffits_copy_file"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTCPFL(iunit, ounit, previous, current, following, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_copy_file

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> copy the fits header without the data from one file to the other - similar to cfitsio-routine fits_copy_header
  !
  !
  !
  !> @param[in] inunit, outunit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_copy_header(inunit, outunit, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: inunit, outunit

    character(*), parameter :: function = "ffits_copy_header"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTCPHD(inunit, outunit, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_copy_header

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close a FITS file unit descriptor - similar to cfitsio-routine fits_close_file
  !
  !
  !
  !> @param[in] unit
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_close_file(unit, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit

    character(*), parameter :: function = "ffits_close_file"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTCLOS(unit, status)

    ! error report
    call ffits_err(function, status, ierr)
        
  end subroutine ffits_close_file

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> Closes a FITS file and frees the unit.
  !
  !
  !
  !> @param[inout] this
  !> @param[in] filename, rwmode
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_FITS_file(unit, status)
    implicit none

    integer,      intent(in)           :: unit
    integer,      intent(inout)        :: status

    character(*), parameter :: function = "close_FITS_file"
    character(buffsize) :: iStr

    if(status > 100 .or. status == s_errRep) return

    call ffits_close_file(unit, status)
    call ffits_free_iounit(unit, status)
    iStr = int2str(unit)
    if(status == ok) call wrchatty(function, "File unit number "//trim(iStr)&
      //" successfully closed and freed.")

  end subroutine close_FITS_file

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> return the number of hdu - similar to cfitsio-routine fits_get_num_hdus
  !
  !
  !
  !> @param[in] unit
  !> @param[inout] nhdu
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_get_num_hdus(unit, nhdu, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(out) :: nhdu
    integer, intent(in) :: unit

    character(*), parameter :: function = "ffits_get_num_hdus"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTTHDU(unit, nhdu, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the hdu type of the CHDU - similar to cfitsio-routine fits_get_hdu_type
  !
  !
  !
  !> @param[in] unit
  !> @param[out] hdutype
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_get_hdu_type(unit, hdutype, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(out) :: hdutype
    integer, intent(in) :: unit

    character(*), parameter :: function = "ffits_get_hdu_type"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTGHDT(unit, hdutype, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_get_hdu_type

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> move to a specified hdu by number - similar to cfitsio-routine fits_movabs_hdu
  !
  !
  !
  !> @param[in] unit, hdu
  !> @param[inout] hdutype
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_movabs_hdu(unit, hdu, hdutype, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(out) :: hdutype
    integer, intent(in) :: unit, hdu

    character(*), parameter :: function = "ffits_movabs_hdu"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTMAHD(unit, hdu, hdutype, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_movabs_hdu

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> move to a specified hdu extension by name - similar to cfitsio-routine fits_movnam_hdu
  !> extver is optional, move to the first as default
  !
  !
  !> @param[in] unit, hdutype, extname
  !> @param[in, optional] extver
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_movnam_hdu(unit, hdutype, extname, status, extver)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, hdutype
    integer, intent(in), optional :: extver

    character(*), intent(in) :: extname
    character(*), parameter :: function = "ffits_movnam_hdu"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    if (present(extver)) then
      call FTMNHD(unit, hdutype, extname, extver, status)
    else
      call FTMNHD(unit, hdutype, extname, 0, status)
    end if

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_movnam_hdu

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> open a fits file - similar to cfitsio-routine fits_open_file
  !
  !
  !
  !> @param[in] unit, filename, rwmode
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_open_file(unit, filename, rwmode, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, rwmode

    character(*), intent(in) :: filename
    character(*), parameter :: function = "ffits_open_file"
    character(buffsize) :: comment

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTNOPN(unit, trim(filename), rwmode, status)

    ! error report
    comment = "filename = " // trim(filename)
    call ffits_err(function, status, ierr, comment)

  end subroutine ffits_open_file

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> Opens a FITS file and returns the unit.
  !
  !
  !
  !> @param[inout] this
  !> @param[in] filename, rwmode
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_FITS_file(filename, rwmode, unit, status, delete)
    implicit none

    character(*), intent(in)           :: filename
    integer,      intent(in)           :: rwmode
    integer,      intent(out)          :: unit
    integer,      intent(inout)        :: status
    integer,      intent(in), optional :: delete

    integer :: exists

    character(*), parameter :: function = "open_FITS_file"

    if(status > 100 .or. status == s_errRep) return

    call ffits_get_iounit(unit, status)
!    if(rwmode /= create) then
    if(rwmode /= FF_CREATE) then
      call ffits_file_exists(filename, exists, status)
      if(exists==0)&
call wrstop(function, "File '"//trim(filename)//"' does not exist.")

      call ffits_open_file(unit, filename, rwmode, status)
      if(status == ok) call wrchatty(function, trim(filename) // " successfully opened")
    else
      call ffits_file_exists(filename, exists, status)
      if(present(delete) .and. exists == 1) then
        if(delete == true) call delete_file_fortran(filename, status)
      end if
      call ffits_create_file(unit, filename, fblocksize, status)
      if(status == ok) call wrchatty(function, trim(filename) // " successfully created")
    end if

  end subroutine open_FITS_file

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the filename of an input unit - similar to cfitsio-routine fits_file_name
  !
  !
  !
  !> @param[in] unit, filename
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_file_name(unit, filename, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit

    character(*), parameter :: function = "ffits_file_name"
    character(*), intent(inout) :: filename

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTFLNM(unit, filename, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_file_name

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete the CHDU from the FITS unit descriptor - similar to cfitsio-routine fits_delete_hdu
  !
  !
  !
  !> @param[in] unit
  !> @param[inout] hdutype
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_delete_hdu(unit, hdutype, status)
    implicit none

    integer :: ierr
    integer, intent(inout) ::  status
    integer, intent(out) :: hdutype
    integer, intent(in) :: unit
    character(*), parameter :: function = "ffits_delete_hdu"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTDHDU(unit, hdutype,status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_delete_hdu

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> rename current keyword with - similar to cfitsio-routine fits_modify_key_longstr
  !
  !
  !
  !> @param[in] unit, keyword, keyval
  !> @param[in, optional] comment
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_modify_key_longstr(unit, keyword, keyval, status, comment)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit

    character(*), intent(in) :: keyword, keyval
    character(*), intent(in), optional :: comment
    character(*), parameter :: function = "ffits_modify_key_longstr"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    if (present(comment)) then
      call FTMKLS(unit, keyword, keyval, comment, status)
    else
      call FTMKLS(unit, keyword, keyval, "", status)
    end if

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_modify_key_longstr

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete the FITS file unit descriptor and delete the associate file - similar to cfitsio-routine fits_delete_file
  !
  !
  !
  !> @param[in] unit
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_delete_file(unit, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(inout) :: status

    character(*), parameter :: function = "ffits_delete_file"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTDELT(unit, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_delete_file

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> update the header keyword of a string key / or create it - similar to cfitsio-routine fits_update_key
  !
  !
  !
  !> @param[in] unit, keyword, keyval, comment
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_update_key(unit, keyword, keyval, comment, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(inout) :: status

    integer :: decimals

    class(*) :: keyval

    character(*), parameter :: function = "ffits_update_key"
    character(*), intent(in) :: keyword
    character(fkeylength) :: fkeyword
    character(*), intent(in) :: comment
    !    integer(i2) :: i2_val
    !    integer(i1) :: i1_val

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    decimals = -14

    ! the fitsio calls
    fkeyword = keyword
    select type (keyval)
      type is ( real(r4) )
      call FTUKYE(unit, fkeyword, keyval, decimals, comment, status)
      type is ( real(r8) )
      call FTUKYD(unit, fkeyword, keyval, decimals, comment, status)
      type is ( integer(i8) )
      call FTUKYK(unit, fkeyword, keyval, comment, status)
      type is ( integer(i4) )
      call FTUKYJ(unit, fkeyword, keyval, comment, status)
      !      type is ( integer(i2) )
      !      i2_val = int(keyval, kind(i2_val))
      !      call FTUKYJ(unit, fkeyword, i2_val, comment, status)
      !      type is ( integer(i1) )
      !      i1_val = keyval
      !      call FTUKYJ(unit, fkeyword, i1_val, comment, status)
      class default
      print *, "update key for that type is not yet supported"
      status = s_errRep
    end select


    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_update_key

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> update the header keyword of a string key / or create it - similar to cfitsio-routine fits_update_key
  !
  !
  !
  !> @param[in] unit, keyword, keyval, comment
  !> @param[inout] status
  !> @param[inout, optional] ierr
  !---------------------------------------------------------------------------

  subroutine ffits_update_key_str(unit, keyword, keyval, comment, status)
    implicit none

    integer :: ierr
    integer, intent(in) :: unit
    integer, intent(inout) :: status

    character(*) :: keyval

    character(*), parameter :: function = "ffits_update_key"

    character(*), intent(in) :: keyword
    character(fkeylength) :: fkeyword
    character(*), intent(in) :: comment


    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    fkeyword = keyword
    call FTUKYS(unit, fkeyword, keyval, comment, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_update_key_str

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read in column information
  !
  !
  !
  !> @param[in] inParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_get_col_info(unit, colnum, ttype, tunit, datatype, repeat_val, tscal, tzero, tnull, tdisp, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum
    integer, intent(out) :: tnull
    integer, intent(out) :: repeat_val

    character(*), parameter :: function = "ffits_get_col_info"
    character(buffsize), intent(out) :: ttype
    character(buffsize), intent(out) :: tunit
    character(buffsize), intent(out) :: tdisp
    character(buffsize), intent(out) :: datatype

    real(dp), intent(out) :: tzero
    real(dp), intent(out) :: tscal

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTGBCL(unit, colnum, ttype, tunit, datatype, repeat_val, tscal, tzero, tnull, tdisp, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_get_col_info

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the datatype of a column
  !
  !
  !
  !> @param[in] unit, colnum, datacode, repeat_val, width
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_get_col_datatype(unit, colnum, datacode, repeat_val, width, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum
    integer, intent(out) :: datacode, repeat_val, width

    character(*), parameter :: function = "ffits_get_col_datatype"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTEQTY(unit, colnum, datacode, repeat_val, width, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_get_col_datatype

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> insert a column into an ASCII or binary table - similar to cfitsio-routine fits_insert_col
  !
  !
  !
  !> @param[in] unit,
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_insert_col(unit, colnum, ttype, tform, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum
    character(*), dimension(:), intent(in) :: ttype, tform

    character(*), parameter :: function = "ffits_insert_col"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTICOL(unit, colnum, ttype, tform, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_insert_col

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write the tdim keyword - similar to cfitsio-routine fits_write_tdim
  !
  !
  !
  !> @param[in] unit,
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine ffits_write_tdim(unit, colnum, naxis, naxes, status)
    implicit none

    integer :: ierr
    integer, intent(inout) :: status
    integer, intent(in) :: unit, colnum, naxis
    integer, intent(in), dimension(naxis) :: naxes

    character(*), parameter :: function = "ffits_write_tdim"

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep) then
      return
    else ! extract the ierr from the status and reset the internal function status to 0
      ierr = status ! reporting status
      status = 0 ! internal status
    end if

    ! the fitsio calls
    call FTPTDM(unit, colnum, naxis, naxes, status)

    ! error report
    call ffits_err(function, status, ierr)

  end subroutine ffits_write_tdim

end module erolib_ffits_mod


