! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! MODULE: erolib_file_mod
!
! DESCRIPTION:
!> functions concerning files for use in eSASS
!
!------------------------------------------------------------------------------

module erolib_file_mod

  use param
  use erolib_general_mod

  implicit none

  integer, parameter, private :: max_ninfiles = 100

contains
  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read an input filelist
  !
  !
  !
  !> @param[in] file
  !> @param[out]
  !> @param[inout] status, filelist
  !---------------------------------------------------------------------------

  subroutine read_filelist (file, filelist, status)
    implicit none

    integer, intent(inout) :: status
    character(*), intent(in) :: file
    character(*), dimension(:), allocatable, intent(inout) :: filelist

    integer :: in_FileUnit, in_counter, in_FileNum
    character(256) :: line

    character(*), parameter :: function = "read_filelist"

    if(status > 100 .or. status == s_errRep) return

    in_counter = 0
    in_FileNum = 0

    open(newunit=in_FileUnit, file=file, action="read")
    ! first count how many files are in the list
    do while (status == ok)
      read(in_FileUnit,'(A)',iostat=status) line
      if (status == ok) then
        if (line(1:1) .ne. "#") then
          in_FileNum = in_FileNum + 1
        endif
      else
      endif
    end do
    rewind(in_FileUnit)

    ! reset status to be okay
    status = ok

    call alloc(filelist, in_FileNum, status)

    do while (status == ok)
      read(in_FileUnit,'(A)',iostat = status) line
      if (status == ok) then
        if (line(1:1) .ne. "#") then
          if (len_trim(line) > 0) then
            in_counter = in_counter + 1
            filelist(in_counter) = line
          endif
        endif
      endif
    end do

    close(in_FileUnit)

    if (in_counter > 0) then
      status = ok
    else
      call wrerr(function, s_err, "input filelist appears to be empty")
      status = s_errRep
    end if

  end subroutine read_filelist

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine rget_filenames.
  !> @brief
  !> wrapper of rgeta to read an array of input strings which can also be stored in an ASCII file
  !
  !
  !> @param parameter_name, parameter_str
  !> @param[in] parameter_name
  !> @param[out] parameter_str
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine rget_filenames(parameter_name, parameter_str, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    character(buffsize), intent(in) :: parameter_name
    character(buffsize), dimension(:), allocatable, intent(out) :: parameter_str

    !internal param

    integer :: icount
    character(buffsize) :: filelist_file

    character(*), parameter :: function = "erolib_rget_filenames"

    if(status > 100 .or. status == s_errRep) return

    if(.not. allocated(parameter_str)) then
      call alloc(parameter_str, max_ninfiles, status)
    end if

    call rgeta (parameter_name, max_ninfiles, parameter_str, icount, status)
    if(status /= ok) then
      status = s_err
      call wrerr(function, status, "reading filename parameter: " // trim(parameter_name))
    end if
    parameter_str = parameter_str(:icount)

    ! check whether infile is actually a file containing the infiles
    if(icount == 1) then
      if(parameter_str(1)(1:1) == "@") then
        filelist_file = parameter_str(1)(2:)
        call read_filelist(filelist_file, parameter_str, status)
      end if
    end if

    if(size(parameter_str) == 0) then
      status = s_err
      call wrerr(function, status, "empty filename parameter: " // trim(parameter_name))
    end if

  end subroutine rget_filenames

end module erolib_file_mod
