! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_fit_mod
  use param
  use erolib_general_mod
  use erolib_like_mod
  use erolib_linklist2_mod

  implicit none

logical, parameter, private :: doTest=.false.

  private

  type, public :: t_report
    logical :: has_converged=.false.
    integer :: n_iters=0
    real :: best_fit_value=0.0
  end type t_report

  type :: t_step_arrow
    real :: xlo=0.0
    real :: xhi=0.0
    real :: ylo=0.0
    real :: yhi=0.0
    integer :: ci=0
  end type t_step_arrow

  type, public :: t_step_plot
    private
    integer :: default_xsize_i = 300
    integer :: default_ysize_i = 300
    character(3) :: pgdev='/xs'
    integer :: ci_good_diag=4
    integer :: ci_bad_diag=5
    integer :: ci_good=6
    integer :: ci_bad=7

    class(t_likelihood), pointer :: likelihood_obj=>null()
    integer :: xpar_i=0
    integer :: ypar_i=0
    real :: xlo=0.0
    real :: xhi=0.0
    real :: ylo=0.0
    real :: yhi=0.0
    integer :: xsize_i=0
    integer :: ysize_i=0

    real, dimension(6) :: pgTrMatrix
    real, dimension(:,:), allocatable :: image
    integer :: dev_id=0
    logical :: is_initialized=.false.
    logical :: plot_started=.false.
  contains
    procedure, non_overridable :: init => init_t_step_plot
    procedure :: draw => draw_t_step_plot
    procedure :: destroy => destroy_t_step_plot
  end type t_step_plot

  type :: t_record
    real :: f=0.0
    real :: lambda=0.0
    real, dimension(:), allocatable :: params
    logical :: trans_available=.false.
  contains
    procedure :: destroy => destroy_t_record
  end type t_record

  type, public :: t_control
    private
    real :: start_lambda=0.001
    real :: max_lambda=1.0e10
    real :: lambda_multiplier=10.0
    real :: shrink_frac=0.95
    integer :: max_n_success_before_diagonalize=3
    integer :: max_n_iters=20
    logical :: step_plot_required=.false.

    integer :: n_success_since_diagonalize=0
    logical :: diagonalize_required=.false.
    logical :: is_initialized=.false.
    type(t_report), public :: report
    type(t_step_plot), public :: plot_obj
    type(t_linked_list) :: records_LL
!    class(t_linked_list), pointer :: records_LL=>null()
  contains
    procedure, non_overridable :: init => init_t_control
    procedure, non_overridable :: add_record => add_record_t_control
    procedure, non_overridable :: get_current_record => get_current_record_t_control
    procedure, non_overridable :: do_diagonalize => do_diagonalize_t_control
    procedure, non_overridable :: report_bad_solve => report_bad_solve_t_control
    procedure, non_overridable :: has_converged => has_converged_t_control
    procedure, non_overridable :: report_not_converged => report_not_converged_t_control
    procedure, non_overridable :: destroy => destroy_t_control
  end type t_control

  public :: levenberg_marquardt

contains

  !.....................................................................
  subroutine destroy_t_record(this)
    implicit none
    class(t_record) :: this

    if(allocated(this%params)) deallocate(this%params)

  end subroutine destroy_t_record

  !.....................................................................
  subroutine init_t_step_plot(this, likelihood_obj, xpar_i, ypar_i&
    , xlo, xhi, ylo, yhi, xsize_i, ysize_i)

    implicit none
    class(t_step_plot)            :: this
    class(t_likelihood), target   :: likelihood_obj
    integer,           intent(in) :: xpar_i, ypar_i
    real,              intent(in) :: xlo, xhi, ylo, yhi
    integer, optional, intent(in) :: xsize_i, ysize_i

    real :: frac0

    this%likelihood_obj => likelihood_obj
    this%xpar_i = xpar_i
    this%ypar_i = ypar_i
    this%xlo = xlo
    this%xhi = xhi
    this%ylo = ylo
    this%yhi = yhi

    if(present(xsize_i)) then
      this%xsize_i = xsize_i
    else
      this%xsize_i = this%default_xsize_i
    end if
    if(present(ysize_i)) then
      this%ysize_i = ysize_i
    else
      this%ysize_i = this%default_ysize_i
    end if

    allocate(this%image(this%xsize_i,this%ysize_i))

    frac0 = -0.5/real(this%xsize_i)
    this%pgTrMatrix(1) = this%xlo*(1.0-frac0) + this%xhi*frac0
    this%pgTrMatrix(2) = (this%xhi - this%xlo)/real(this%xsize_i)
    this%pgTrMatrix(3) = 0.0

    frac0 = -0.5/real(this%ysize_i)
    this%pgTrMatrix(4) = this%ylo*(1.0-frac0) + this%yhi*frac0
    this%pgTrMatrix(5) = 0.0
    this%pgTrMatrix(6) = (this%yhi - this%ylo)/real(this%ysize_i)

    this%is_initialized = .true.

  end subroutine init_t_step_plot

  !.....................................................................
  subroutine draw_t_step_plot(this, stored_records)
    implicit none
    class(t_step_plot)                       :: this
    type(t_record), dimension(:), intent(in) :: stored_records

    character(*), parameter :: functionName = "draw_t_step_plot"
!    real, dimension(6) :: pgTrMatrix = (/0.0, 1.0, 0.0, 0.0, 0.0, 1.0/)
    integer :: PGOPEN,n_records,ii,xi,yi,i_last_good_params
    real, dimension(:), allocatable :: last_good_params,grad,deltas
    real :: frac,f
    type(t_step_arrow), dimension(:), allocatable :: step_arrows

    if(.not.this%is_initialized)&
call wrstop(functionName, "You must initialize the plot_obj attribute of the control object.")

    if(.not.this%plot_started) then
      this%dev_id = PGOPEN(this%pgdev)
      if(this%dev_id <= 0)&
call wrstop(functionName, "Could not open pgplot.")

      call PGSVP(0.0, 0.999, 0.0, 0.999)
      call PGSWIN(this%xlo, this%xhi, this%ylo, this%yhi)
      call PGASK(.true.) ! sets things up so that PGPAGE waits for user input before continuing.

      call PGSCR(this%ci_good_diag, 0.4,1.0,0.2) ! bright, fresh green
      call PGSCR(this%ci_bad_diag,  1.0,0.4,0.2) ! bright, hot red
      call PGSCR(this%ci_good,      0.1,0.8,0.1) ! green
      call PGSCR(this%ci_bad,       0.8,0.1,0.1) ! red

      this%plot_started = .true.
    end if

    n_records = size(stored_records)
    if(n_records<1)&
  return

    i_last_good_params = 1
    if(n_records>1) then
      ! Construct the various step arrows:
      !
      allocate(step_arrows(n_records-1))
      i_last_good_params = 1
      do ii=2,n_records
        step_arrows(ii-1)%xlo = stored_records(i_last_good_params)%params(this%xpar_i)
        step_arrows(ii-1)%xhi = stored_records(ii)%params(this%xpar_i)
        step_arrows(ii-1)%ylo = stored_records(i_last_good_params)%params(this%ypar_i)
        step_arrows(ii-1)%yhi = stored_records(ii)%params(this%ypar_i)

        if(stored_records(ii)%f > stored_records(ii-1)%f) then
          i_last_good_params = ii

          if(stored_records(ii)%trans_available) then
            step_arrows(ii-1)%ci = this%ci_good_diag
          else
            step_arrows(ii-1)%ci = this%ci_good
          end if

        else if(stored_records(ii)%trans_available) then
          step_arrows(ii-1)%ci = this%ci_bad_diag
        else
          step_arrows(ii-1)%ci = this%ci_bad
        end if
      end do
    end if

    last_good_params = stored_records(i_last_good_params)%params
    print *, "*** Iteration", n_records
    print *, "  Last good params:"
    call print_vec(last_good_params, 2)

    f = this%likelihood_obj%calc_value(last_good_params)
    grad = this%likelihood_obj%calc_gradient(last_good_params)

    print *, ''
    print *, "  Last good likelihood:", f
    print *, ''
    print *, "  Last good gradient:"
    call print_vec(grad, 2)

    if(i_last_good_params < n_records) then
      deltas = stored_records(n_records)%params - last_good_params
      print *, ''
      print *, "  Deltas:"
      call print_vec(deltas, 2)
    end if
    print *, ''

    do xi=1,this%xsize_i
      frac = (real(xi)-0.5)/real(this%xsize_i)
      last_good_params(this%xpar_i) = this%xlo*(1.0-frac) + this%xhi*frac

      do yi=1,this%ysize_i
        frac = (real(yi)-0.5)/real(this%ysize_i)
        last_good_params(this%ypar_i) = this%ylo*(1.0-frac) + this%yhi*frac

        this%image(xi,yi) = this%likelihood_obj%calc_value(last_good_params)
      end do
    end do

    call PGBBUF()
    call PGPAGE() ! prompts the user before continuing.

    call PGGRAY(this%image, this%xsize_i, this%xsize_i, 1, this%xsize_i, 1, this%xsize_i&
      , maxval(this%image), minval(this%image), this%pgTrMatrix)

    if(n_records<2)&
  return

    ! Now draw the arrows.
    !
    do ii=1,n_records-1
      call PGSCI(step_arrows(ii)%ci)
      call PGARRO (step_arrows(ii)%xlo, step_arrows(ii)%ylo, step_arrows(ii)%xhi, step_arrows(ii)%yhi)
    end do

    call PGEBUF()

  end subroutine draw_t_step_plot

  !.....................................................................
  subroutine destroy_t_step_plot(this)
    implicit none
    class(t_step_plot) :: this

    if(allocated(this%image)) deallocate(this%image)
    if(associated(this%likelihood_obj)) nullify(this%likelihood_obj)! then
!      if(allocated(this%likelihood_obj)) deallocate(this%likelihood_obj)
!      if(associated(this%likelihood_obj)) nullify(this%likelihood_obj)
!    end if

  end subroutine destroy_t_step_plot

  !.....................................................................
  subroutine init_t_control(this, max_n_iters, start_lambda, max_lambda&
    , lambda_multiplier, shrink_frac, max_n_success_before_diagonalize, step_plot_required)
    implicit none

    class(t_control)              :: this
    integer, optional, intent(in) :: max_n_iters
    real,    optional, intent(in) :: start_lambda
    real,    optional, intent(in) :: max_lambda
    real,    optional, intent(in) :: lambda_multiplier
    real,    optional, intent(in) :: shrink_frac
    integer, optional, intent(in) :: max_n_success_before_diagonalize
    logical, optional, intent(in) :: step_plot_required

    if(present(max_n_iters      )) this%max_n_iters       = max_n_iters
    if(present(start_lambda     )) this%start_lambda      = start_lambda
    if(present(max_lambda       )) this%max_lambda        = max_lambda
    if(present(lambda_multiplier)) this%lambda_multiplier = lambda_multiplier
    if(present(shrink_frac      )) this%shrink_frac       = shrink_frac

    if(present(max_n_success_before_diagonalize)) then
      this%max_n_success_before_diagonalize = max_n_success_before_diagonalize
      this%diagonalize_required = .true.
    else
      this%diagonalize_required = .false.
    end if

    if(present(step_plot_required)) then
      this%step_plot_required = step_plot_required
    else
      this%step_plot_required = .false.
    end if

!    if(.not. associated(this%records_LL)) allocate(this%records_LL)

    this%is_initialized=.true.

  end subroutine init_t_control

  !.....................................................................
  subroutine add_record_t_control(this, params, f, trans_available)
    implicit none

    class(t_control)               :: this
    real, dimension(:), intent(in) :: params
    real,               intent(in) :: f
    logical, optional,  intent(in) :: trans_available

!    real :: lambda
!    class(t_record), pointer :: last_record
!    class(t_record), pointer :: next_record
    type(t_record) :: last_record
    type(t_record) :: next_record
    logical :: is_first_step,last_step_was_successful
    integer :: n_params,n_steps,ii
    type(t_record), dimension(:), allocatable :: stored_records

    next_record%f = f
    allocate(next_record%params(size(params)))
    next_record%params = params
    if(present(trans_available)) next_record%trans_available = trans_available

    if(this%records_LL%isEmpty()) then
      is_first_step = .true.
      next_record%lambda = this%start_lambda
    else
      is_first_step = .false.

      call this%records_LL%set_curr_to_last()
      last_record = this%get_current_record()

      if(f > last_record%f) then
        last_step_was_successful = .true.
        next_record%lambda = last_record%lambda/this%lambda_multiplier
        if(next_record%trans_available)&
          this%n_success_since_diagonalize = this%n_success_since_diagonalize + 1
      else
        last_step_was_successful = .false.
        next_record%lambda = last_record%lambda*this%lambda_multiplier
      end if
    end if

    call this%records_LL%appendLink(next_record)

    if(this%step_plot_required) then! .and..not.is_first_step) then
      ! Extract the stored
      n_params = size(params)
      n_steps = this%records_LL%getNumLinks()
      allocate(stored_records(n_steps))

      call this%records_LL%set_curr_to_first()
      do ii = 1, n_steps
        stored_records(ii) = this%get_current_record()
        call this%records_LL%set_curr_to_next()
      end do

      call this%plot_obj%draw(stored_records)

      deallocate(stored_records) !*** do the individual params arrays get deallocated??
    end if

  end subroutine add_record_t_control

  !.....................................................................
  function get_current_record_t_control(this) result(record)
    ! The user should not call this before ascertaining that the list contains some elements.
    implicit none
    class(t_control) :: this
    type(t_record) :: record

    character(*), parameter :: functionName = "get_current_link_object"
    class(*), pointer :: tmp=>null()

    tmp => this%records_LL%getCurrObj()
    if(associated(tmp)) then
      select type(tmp)
      type is (t_record)
        record = tmp
      class default
call wrstop(functionName, "Unrecognized class or type.")
      end select
    else
call wrstop(functionName, "No items in list.")
    end if

  end function get_current_record_t_control

  !.....................................................................
  logical function do_diagonalize_t_control(this)
    implicit none

    class(t_control) :: this

    if(this%diagonalize_required&
    .and. (this%n_success_since_diagonalize > this%max_n_success_before_diagonalize)) then
      do_diagonalize_t_control = .true.
    else
      do_diagonalize_t_control = .false.
    end if

  end function do_diagonalize_t_control

  !.....................................................................
  subroutine report_bad_solve_t_control(this, ierr)
    implicit none
    class(t_control) :: this
    integer, intent(in) :: ierr

print *, "Solve returned error status", ierr

    call this%destroy()

  end subroutine report_bad_solve_t_control

  !.....................................................................
  logical function has_converged_t_control(this)
    ! The LM is judged to have converged if 3 successful steps in a row have occurred, if (f_n-f_{n-1})<0.001 and if ()<(f_n-f_{n-1}) < (f_{n-1} - f_{n-2}).
    implicit none
    class(t_control) :: this

    integer, parameter :: rnd=3
    real, dimension(rnd+1) :: fs
    real, dimension(rnd) :: dfs
    integer :: ii,n_steps
    type(t_record) :: record

    has_converged_t_control = .false.
    n_steps = this%records_LL%getNumLinks()
    if(n_steps > rnd) then
      call this%records_LL%set_curr_to_last()
      do ii=1,rnd+1
        record = this%get_current_record()
        fs(rnd+2-ii) = record%f
        call this%records_LL%set_curr_to_prev()
      end do

      dfs = fs(2:rnd+1) - fs(1:rnd)

      if(all(dfs>0.0)& ! i.e., last rnd steps were successful
      .and. dfs(rnd)<dfs(rnd-1) .and. dfs(rnd)<0.001)&
        has_converged_t_control = .true.
    end if

    this%report%has_converged = has_converged_t_control

  end function has_converged_t_control

  !.....................................................................
  subroutine report_not_converged_t_control(this)
    implicit none
    class(t_control) :: this

    this%report%has_converged = .false.

    call this%destroy()

  end subroutine report_not_converged_t_control

  !.....................................................................
  subroutine destroy_t_control(this)
    implicit none
    class(t_control) :: this

    integer :: ierr

    if(this%step_plot_required) then
      call PGCLOS()
    end if

!    call this%report%destroy()
    call this%plot_obj%destroy()
    ierr = this%records_LL%delAllLinks()

  end subroutine destroy_t_control

  !.....................................................................
  subroutine levenberg_marquardt(likelihood_obj, control_obj, start_params, best_params)
    implicit none
    !
    ! This seeks for a maximum in the value of the supplied likelihood object, with the aim of estimating the best-fitting model parameters. The routine won't work correctly unless the curvature matrix (matrix of second derivatives of the likelihood relative to the model parameters) is negative-definite.
    !
    ! The code is the result of a long evolutionary process that began with the routine 'mrqmin' in chapter 15.5 of Press et al, "Numerical Recipes in Fortran". Basically we have an function f(q_) implemented in likelihood_obj which, when supplied with a list of parameter values q_, can return a value scalar f, a gradient vector g_(q_) and a curvature matrix H(q_). The intent is to find the set of parameter values which maximizes f. The assumption is that the function varies smoothly, and that the maximum is quadratic.
    !
    ! With a pure quadratic function, one could go direct to the maximum, given a single sample of f, g_ and H, by
    !
    !	q_max = q_ - H^{-1}(q_) * g_(q_).				(1)
    !
    ! However, in other cases, the f surface may be far from quadratic sufficiently far from the maximum, and application of formula (1) might in fact take us further from the maximum.
    !
    ! The slow and robust way to find a maximum is to follow the gradient uphill in small, cautious steps:
    !
    !	q_{i+1} = q_i - g_(q_i)/lambda.
    !
    ! The Levenberg-Marquardt algorithm offers a cunning way to mix the bold and cautious approaches via the mixing parameter lambda. The iteration algorithm is given by
    !
    !	q_{i+1} = q_i - [H(q_i) - I*lambda]^{-1} * g_(q_i).		(2)
    !
    ! Each iteration is tested to see if f(q_{i+1}) has increased. If so, lambda is decreased, emboldening the next iteration; if not, lambda is increased, for a more safe and timid next try.
    !
    ! I have built in a further innovation to address the problem that, if the shape of the maximum is thin, an uphill gradient search becomes quite inefficient, since the algorithm spends most of its time rattling back and forth across a ridge while making only very slow progress up it. The solution is to transform the parameter space so as to diagonalize the local curvature matrix H, such that the 1-sigma contour is spheroidal and of unity radius. For a symmetric matrix H,
    !
    !	H = P * L * P^T
    !
    ! where the columns of P are eigenvectors, the diagonal entries of L are the matching eigenvalues, and T denotes transpose. For negative-definite H, all the eigenvalues will be negative, and L can be written as
    !
    !	L = -(S*S)^{-1}.
    !
    ! In this case,
    !
    !	H = -P*S^{-1}*I*S^{-1}*P^T.
    !
    ! Inserting this in (1) and premultiplying by S^{-1}*P^T gives
    !
    !	(S^{-1}*P^T)*q_min = (S^{-1}*P^T)*q_ + (S*P^T)*g_.
    !
    ! (Note that P^{-1} = P^T.) Therefore we modify the algorithm in (2) to
    !
    !	q'_{i+1} = q'_i - [H' - I*lambda]^{-1} * g'_
    !
    ! where q'_ = (S^{-1}*P^T)*q_, H' = (P*S)^T*H*(P*S) and g'_ = (P*S)^T*g_.

    class(t_likelihood)                              :: likelihood_obj
    type(t_control),                     intent(in)  :: control_obj
    real, dimension(:),                  intent(in)  :: start_params
    real, dimension(size(start_params)), intent(out) :: best_params

    character(*), parameter :: functionName = "levenberg_marquardt"
    real, dimension(size(start_params)) :: params_old,params_new,grad,delta_params,trans_delta_params,trans_grad
    real, dimension(size(start_params),size(start_params)) :: curv_mat,id_mat,trans_mat,untrans_mat,mat
    real :: f_old,f_new,lambda,length
    logical :: trans_available
    integer :: iter,n_successive_failures,ierr,ii,i_hi,i_lo
    type(tRange) :: line_range

    params_old = start_params

    id_mat = 0.0
    do ii=1,size(start_params)
      id_mat(ii,ii) = 1.0
    end do

    f_old = likelihood_obj%calc_value(params_old)
    grad = likelihood_obj%calc_gradient(params_old)
    curv_mat = likelihood_obj%calc_curvature(params_old)

    call control_obj%add_record(params_old, f_old)

    trans_available = .false. ! default
    if(control_obj%do_diagonalize()) then
      call calc_trans_mats(curv_mat, trans_mat, untrans_mat, trans_available)
    end if
    if(.not.trans_available) then
      call likelihood_obj%calc_scale(curv_mat)
    end if

    lambda = control_obj%start_lambda

if(doTest) then
  print *, "Starting values:"
  print *, "  f=", f_old
  print *, "  lambda=", lambda
  print *, "  params:"
  call print_vec(params_old, 2)
  print *, "  grad:"
  call print_vec(grad, 2)
  print *, ''
end if

    n_successive_failures = 0
    do iter=1, control_obj%max_n_iters
      ! Solve linear equation to obtain params_new.
      if(trans_available) then
        mat = matmul(transpose(untrans_mat),matmul(curv_mat,untrans_mat)) - id_mat*lambda
        trans_grad = matmul(transpose(untrans_mat), grad)
        call solve_symm(mat, -trans_grad, trans_delta_params, ierr)
        delta_params = matmul(untrans_mat, trans_delta_params)
      else ! at least scale the parameters so that the diagonal terms of curv_mat become -1.
        call likelihood_obj%apply_scale(grad, curv_mat)
        mat = curv_mat - id_mat*lambda
        call solve_symm(mat, -grad, delta_params, ierr)
        call likelihood_obj%unscale(delta_params)
      end if

      if(ierr/=0) then
        call control_obj%report_bad_solve(ierr)
  return
      end if

if(doTest) then
  print *, "Iter", iter
  print *, "  deltas:"
  call print_vec(delta_params, 2)
end if

      ! If params_new out of model limits, shrink them back a bit.
      !
      call likelihood_obj%model_obj%get_bounds_on_line(params_old, delta_params, length, line_range, i_hi, i_lo)
if(doTest) print *, "  length of deltas=", length
      if(line_range%hi%isUndefined) then
if(doTest) print *, "  upper bound in delta direction is undefined."
        params_new = params_old + delta_params
      else if(line_range%hi%value > length) then
if(doTest) print *, "  upper bound in delta direction =", line_range%hi%value
        params_new = params_old + delta_params
      else
if(doTest) then
  print *, "  upper bound in delta direction =", line_range%hi%value
  print *, "  deltas multiplied by =", control_obj%shrink_frac*line_range%hi%value
end if
        params_new = params_old + delta_params*(control_obj%shrink_frac*real(line_range%hi%value))
      end if

      f_new = likelihood_obj%calc_value(params_new)

if(doTest) then
  print *, "  trial f=", f_new
  print *, "  trial params:"
  call print_vec(params_new, 2)
end if

      call control_obj%add_record(params_new, f_new, trans_available)

      if(f_new > f_old) then ! success
if(doTest) print *, "  *** SUCCESS ***"
        if(control_obj%has_converged()) then
          best_params = params_new
  return
        end if

        params_old = params_new
        f_old = f_new
        lambda = lambda/control_obj%lambda_multiplier
        grad = likelihood_obj%calc_gradient(params_old)
        curv_mat = likelihood_obj%calc_curvature(params_old)

if(doTest) then
  print *, "  lambda=", lambda
  print *, "  grad:"
  call print_vec(grad, 2)
  print *, ''
end if

      else ! failure. Increase lambda and try again.
if(doTest) print *, "  *** failure. ***"
        lambda = lambda*control_obj%lambda_multiplier
        if(lambda > control_obj%max_lambda)&
    exit
      end if

if(doTest) then
  print *, "  lambda=", lambda
  print *, ''
end if

      if(control_obj%do_diagonalize()) then
        call calc_trans_mats(curv_mat, trans_mat, untrans_mat, trans_available)
      end if
    end do

    ! If we got to here, we ran out of either iterations or lambda.

    best_params = params_old
    call control_obj%report_not_converged()

  end subroutine levenberg_marquardt

  !.....................................................................
  subroutine calc_trans_mats(curv_mat, trans_mat, untrans_mat, trans_available)
    implicit none
!*** make this private so don't have to worry about any checks on array dimensions.
    real, dimension(:,:), intent(in)  :: curv_mat
    real, dimension(:,:), intent(out) :: trans_mat, untrans_mat
    logical,              intent(out) :: trans_available

    real, dimension(size(curv_mat,1)) :: eigvals
    real, dimension(size(curv_mat,1),size(curv_mat,1)) :: eigvecs,sigma_mat,inv_sigma_mat
    integer :: ierr,ii,n_params

    n_params = size(curv_mat,1)

    ! defaults:
    trans_mat = 0.0
    untrans_mat = 0.0
    trans_available = .false.

    call calc_eigen_symm(curv_mat, eigvals, eigvecs, ierr)

    if(ierr/=0)&
  return

    if(.not.all(eigvals<0.0))&
  return

    sigma_mat = 0.0
    inv_sigma_mat = 0.0
    do ii=1,n_params
      inv_sigma_mat(ii,ii) = sqrt(-eigvals(ii))
      sigma_mat(ii,ii) = 1.0/inv_sigma_mat(ii,ii)
    end do

    trans_mat = matmul(inv_sigma_mat, transpose(eigvecs))  ! S^{-1}*P^T
    untrans_mat = matmul(eigvecs, sigma_mat)               ! P*S
    trans_available = .true.

  end subroutine calc_trans_mats

  !.....................................................................
  subroutine print_vec(vec, n_space)
    implicit none
    real, dimension(:), intent(in) :: vec
    integer, optional,  intent(in) :: n_space

    character(buffsize) :: str,space_str
    integer :: n_elements,ii

    space_str = ''

    n_elements = size(vec)
    str = ''
    if(n_elements>0) then
      write(str, '(e10.3)') vec(1)
      do ii=2,n_elements
        write(str, '(a,e11.4)') trim(str)//"  ",vec(ii)
      end do
    end if

    if(present(n_space)) then
      str = space_str(1:n_space)//trim(str)
    end if

    print *, trim(str)

  end subroutine print_vec
end module erolib_fit_mod

