! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! MODULE: erolib_general_mod
!
! DESCRIPTION:
!> often used parameters, definitions and general functions
!
! Contents:
!  subroutine FError(function, stat, msg, status)
!  subroutine report_FITS_err(functionName, fitsErrI, generalStatus, addendum)
!  subroutine alloc(this, n, status, src)
!  subroutine dealloc(this, status)
!  subroutine intersect(array1, array2, outarray, status)
!  function complement_int_order(array1, array2, status)
!  subroutine delete_file_fortran(filename, status)
!------------------------------------------------------------------------------

module erolib_general_mod
    
  use ISO_FORTRAN_ENV ! gfortran module for real64 and other definitions
  use param
  use ero_general ! need for flag masks.

  implicit none

logical, private, parameter :: doTest = .false.

  integer, parameter :: fblocksize  = 1
  integer, parameter :: buffsize = 2048
  integer, parameter :: fkeylength = 8
  integer, parameter :: fcolnamelength = 71
  integer, parameter :: fcommentlength = 80
  integer, parameter :: sp = REAL32, r4 = REAL32
  integer, parameter :: dp = REAL64, r8 = REAL64
  integer, parameter :: qp = REAL128, r16 = REAL128
  integer, parameter :: i1 = INT8
  integer, parameter :: i2 = INT16
  integer, parameter :: i4 = INT32
  integer, parameter :: i8 = INT64
  integer, parameter :: tel_mask_len = INT8
  integer(i2), parameter :: pha_max = 12288 ! 20191011 IMS: According to KOD (see mail of 11 Oct 2019).
  real(r8), parameter :: eroday_duration_s = 4.0_r8*3600.0_r8
  real(r8), parameter :: ssCounterPeriod = real(2**19, kind=r8)/5.0e7_r8 ! subSec tick occurs at the overflow of a 19-bit counter driven by a 50MHz clock.

  integer, parameter :: N_TELESCOPES = 7
  integer, parameter :: N_PIXELS_CCD = 384 ! in both X and Y directions
  real,    parameter :: FRAME_DURATION = 0.0500584 ! email to IMS from WB, 26 Jul 2020.
!*** 20201204 IMS: I'll temporarily put these here. Later they should go in ero_general_mod.f90.

  integer(i1), parameter :: null_i1 = 0_i1
  integer(i2), parameter :: null_i2 = 0_i2
  integer(i4), parameter :: null_i4 = 0_i4
  integer(i8), parameter :: null_i8 = 0_i8
  real(r4), parameter :: null_r4 = 0.0_r4
  real(r8), parameter :: null_r8 = 0.0_r8
  character(*), parameter :: null_c = " "

  integer, parameter :: BAD_HDU_NUM=301 !*** should be obtainable from fitsio, no?
  integer, parameter :: COL_NOT_FOUND=219 !*** should be obtainable from fitsio, no?

  ! currently extend the region by 1 degree (Whitebook Merloni) - TODO better number
  !    real (dp), parameter :: eFoV = 45_dp / 3600_dp
  real(r8), parameter :: eFoV = 1.0_r8

  enum, bind( C )
    enumerator :: FF_READONLY, FF_READWRITE, FF_CREATE
  end enum
  ! ^ should replace v
  enum, bind( C )
    enumerator :: readonly, readwrite, create
  end enum

enum, bind( C )
enumerator :: ANY_HDU = -1, IMAGE_HDU = 0, ASCII_TBL = 1, BINARY_TBL = 2
end enum

enum, bind( C )
enumerator :: false = 0, true = 1
end enum

enum, bind( C )
enumerator :: and = 0, or = 1
end enum

enum, bind( C )
enumerator :: s_errRep = -3, ok = 0, warn = 1, c_err = 2, s_err_no_exit=3, s_err = 4, ignore = 10
end enum

enum, bind( C )
enumerator :: gti_type = 0, qualgti_type = 1, flaregti_type = 2, externalgti_type = 3
end enum

  type tBoundI
    logical :: is_defined = .false.
!    real(dp) :: value = 0.0
    integer :: value = 0
  end type tBoundI

  type tRangeI
    type(tBoundI) :: lo,hi
  end type tRangeI

  type :: t_timer
    integer, private :: start_time=0
    integer, private :: stop_time=0
!    real, private :: ticks_per_sec=0.0
    integer, private :: ticks_per_sec=0
    logical, private :: started=.false.
  contains
    procedure :: start => start_t_timer
    procedure :: stopp => stopp_t_timer
  end type t_timer

  type :: t_array_iterator
    private
    integer, dimension(:), allocatable :: sizes
    integer, dimension(:), allocatable :: indices
    integer :: n_dims=0
    integer :: ii=0
    logical :: is_initialized = .false.
  contains
    procedure :: create  => create_t_array_iterator
    procedure :: get => get_t_array_iterator
    procedure :: iterate => iterate_t_array_iterator
    procedure :: destroy => destroy_t_array_iterator
  end type t_array_iterator

! general routine to allocate memory
! always tries deallocating the memory first

interface alloc
  module procedure alloc_r8, alloc_r8_2d, alloc_character, alloc_r4,&
    alloc_i4_2d, alloc_logical, alloc_i1, alloc_i2, alloc_i4, alloc_i8, alloc_i8_2d, &
    alloc_i1_2d, alloc_r4_2d, alloc_r4_3d, alloc_i4_3d, alloc_i2_2d, alloc_i2_3d&
    , alloc_logical_2d
end interface alloc

! general routine do deallocate memory

interface dealloc
  module procedure dealloc_r8, dealloc_r8_2d, dealloc_r4,&
    dealloc_i4_2d, dealloc_character, dealloc_logical, &
    dealloc_i1, dealloc_i2, dealloc_i4, dealloc_i8, dealloc_i8_2d, &
    dealloc_i1_2d, dealloc_r4_2d, dealloc_r4_3d, dealloc_i4_3d, dealloc_i2_2d, dealloc_i2_3d&
    , dealloc_logical_2d
end interface dealloc

! general routine to intersect arrays
interface intersect
  module procedure intersect_int, intersect_real, intersect_double
end interface intersect

  integer(i4), parameter :: badFlagMask = ior(FLAG_CORRUPT_EVENT, FLAG_CORRUPT_FRAME)

contains

  !.....................................................................
  subroutine print_erolib_version(chatty)
    implicit none
    logical, intent(in) :: chatty

    if(chatty) then
      call wrchatty("erolib","Version 2.68.")
    else
      call wrlog("erolib","Version 2.68.")
    end if

  end subroutine print_erolib_version

  !.....................................................................
  subroutine FError(function, stat, msg, status)
    
    character(*), intent(in) :: function, msg
        
    integer, intent(in) :: stat
    integer, intent(inout), optional :: status

if(doTest) print *, ">>> FError() stat, status:", stat, status
if(doTest) print *, "+++ FError() function="//trim(function)//"<<"
if(doTest) print *, "+++ FError() msg="//trim(msg)//"<<"

    if(stat /= 0) then
      if(present(status)) then
        select case (status)
          case (0)
            status = s_err
            call wrerr (function, status, trim(msg))
            status = s_errRep
          case (ignore)
              ! ignore the error
          case default
            call wrerr(function, status, trim(msg))
        end select
      else
        call wrlog (function, trim(msg))
        stop
      end if
    end if

if(doTest) print *, "<<< FError()"

  end subroutine FError

  !.....................................................................
  subroutine report_FITS_err(functionName, fitsErrI, generalStatus, addendum)
    implicit none

    character(*), intent(in) :: functionName
    integer,      intent(in) :: fitsErrI, generalStatus
    character(*), intent(in), optional :: addendum

    character(len=30) :: errtext

    if(fitsErrI/=0) then
      call FTGERR(fitsErrI, errtext)
      if(present(addendum)) then
        call wrerr(functionName, generalStatus, trim(errtext)//trim(addendum))
      else
        call wrerr(functionName, generalStatus, errtext)
      end if
    end if

  end subroutine report_FITS_err

  !.....................................................................
  ! sub to allocate 1d double type arrays
  subroutine alloc_r8(this, n, status, src)

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    real(r8), allocatable, dimension(:) :: this
    real(r8), dimension(:), optional, intent(in) :: src

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_r8"

    call dealloc(this, status)
    errmsg=""
    if(present(src)) then
      allocate(this(n), stat=stat, errmsg=errmsg, source=src)
    else
      allocate(this(n), stat=stat, errmsg=errmsg)
      this = null_r8
    end if
    call FError(function, stat, errmsg, status)

  end subroutine alloc_r8;

      ! sub to allocate 2d double type arrays
  subroutine alloc_r8_2d(this, n, m, status)

    integer :: stat
    integer, intent(in) :: n, m
    integer, intent(inout) :: status

    real(r8), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_r8_2d"

    call dealloc(this, status)
    errmsg=""
    allocate(this(n, m), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)
    this(:,:) = null_r8

  end subroutine alloc_r8_2d;

  ! sub to deallocate double type arrays
  subroutine dealloc_r8(this, status)

    integer :: stat
    integer, intent(inout) :: status

    real(r8), allocatable, dimension(:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_r8"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_r8

      ! sub to deallocate double type arrays 2d
  subroutine dealloc_r8_2d(this, status)

    integer :: stat
    integer, intent(inout) :: status

    real(r8), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_r8_2d"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_r8_2d

  ! sub to allocate 1d real type arrays
  subroutine alloc_r4(this, n, status, src)

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    real(r4), allocatable, dimension(:) :: this
    real(r4), dimension(:), optional, intent(in) :: src

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_r4"

    call dealloc(this, status)
    errmsg=""
    if(present(src)) then
      allocate(this(n), stat=stat, errmsg=errmsg, source=src)
    else
      allocate(this(n), stat=stat, errmsg=errmsg)
      this = null_r4
    end if
    call FError(function, stat, errmsg, status)

  end subroutine alloc_r4;

          ! sub to allocate 2d real type arrays
  subroutine alloc_r4_2d(this, n, m, status, s_n, s_m)

    integer :: stat, n0, n1, m0, m1
    integer, intent(in) :: n, m
    integer, intent(inout) :: status

    integer, intent(in), optional :: s_n, s_m

    real(r4), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_r4_2d"

    n0 = 1
    m0 = 1
    n1 = n
    m1 = m
    if(present(s_n)) then
      n0 = s_n
      n1 = n1 + s_n - 1
    end if

    if(present(s_m)) then
      m0 = s_m
      m1 = n1 + s_m - 1
    end if

    call dealloc(this, status)
    errmsg=""
    allocate(this(n0:n1, m0:m1), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

    this(:,:) = null_r4

  end subroutine alloc_r4_2d;

          ! sub to allocate 3d real type arrays
  subroutine alloc_r4_3d(this, n, m, l, status, s_n, s_m, s_l)

    integer :: stat, n0, n1, m0, m1, l0, l1
    integer, intent(in) :: n, m, l
    integer, intent(inout) :: status

    integer, intent(in), optional :: s_n, s_m, s_l

    real(r4), allocatable, dimension(:,:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_r4_3d"

    n0 = 1
    m0 = 1
    l0 = 1
    n1 = n
    m1 = m
    l1 = l
    if(present(s_n)) then
      n0 = s_n
      n1 = n1 + s_n - 1
    end if

    if(present(s_m)) then
      m0 = s_m
      m1 = n1 + s_m - 1
    end if

    if(present(s_l)) then
      l0 = s_l
      l1 = l1 + s_l - 1
    end if

    call dealloc(this, status)
    errmsg=""
    allocate(this(n0:n1, m0:m1, l0:l1), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

    this(:,:,:) = null_r4

  end subroutine alloc_r4_3d;

  ! sub to deallocate real type arrays
  subroutine dealloc_r4(this, status)

    integer :: stat
    integer, intent(inout) :: status

    real(r4), allocatable, dimension(:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_r4"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_r4

          ! sub to deallocate real type arrays 2d
  subroutine dealloc_r4_2d(this, status)

    integer :: stat
    integer, intent(inout) :: status

    real(r4), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_r4_2d"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_r4_2d

          ! sub to deallocate real type arrays 3d
  subroutine dealloc_r4_3d(this, status)

    integer :: stat
    integer, intent(inout) :: status

    real(r4), allocatable, dimension(:,:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_r4_3d"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_r4_3d

      ! sub to allocate 1d integer type arrays
  subroutine alloc_i1(this, n, status, src)

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    integer(i1), allocatable, dimension(:) :: this
    integer(i1), dimension(:), optional, intent(in) :: src

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_i1"

    call dealloc(this, status)
    errmsg=""
    if(present(src)) then
      allocate(this(n), stat=stat, errmsg=errmsg, source=src)
    else
      allocate(this(n), stat=stat, errmsg=errmsg)
      this = null_i1
    end if
    call FError(function, stat, errmsg, status)

  end subroutine alloc_i1;

             ! sub to allocate 1d integer type arrays
  subroutine alloc_i1_2d(this, n, m, status)

    integer :: stat
    integer, intent(in) :: n, m
    integer, intent(inout) :: status

    integer(i1), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_i1_2d"

    call dealloc(this, status)
    errmsg=""
    allocate(this(n, m), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)
    this(:,:) = null_i1

  end subroutine alloc_i1_2d;

      ! sub to allocate 1d integer type arrays
  subroutine alloc_i2(this, n, status, src)

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    integer(i2), allocatable, dimension(:) :: this
    integer(i2), dimension(:), optional, intent(in) :: src

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_i2"

    call dealloc(this, status)
    errmsg=""
    if(present(src)) then
      allocate(this(n), stat=stat, errmsg=errmsg, source=src)
    else
      allocate(this(n), stat=stat, errmsg=errmsg)
      this = null_i2
    end if
    call FError(function, stat, errmsg, status)

  end subroutine alloc_i2;

  ! sub to allocate 2d integer type arrays
  subroutine alloc_i2_2d(this, n, m, status, s_n, s_m)

    integer :: stat, n0, n1, m0, m1
    integer, intent(in) :: n, m
    integer, intent(inout) :: status

    integer, intent(in), optional :: s_n, s_m

    integer(i2), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_i2_2d"

    n0 = 1
    m0 = 1
    n1 = n
    m1 = m
    if(present(s_n)) then
      n0 = s_n
      n1 = n1 + s_n - 1
    end if

    if(present(s_m)) then
      m0 = s_m
      m1 = n1 + s_m - 1
    end if

    call dealloc(this, status)
    errmsg=""
    allocate(this(n0:n1, m0:m1), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

    this(:,:) = null_i2

  end subroutine alloc_i2_2d;

     ! sub to allocate 3d integer type arrays
  subroutine alloc_i2_3d(this, n, m, l, status, s_n, s_m, s_l)

    integer :: stat, n0, n1, m0, m1, l0, l1
    integer, intent(in) :: n, m, l
    integer, intent(inout) :: status

    integer, intent(in), optional :: s_n, s_m, s_l

    integer(i2), allocatable, dimension(:,:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_i2_3d"

    n0 = 1
    m0 = 1
    l0 = 1
    n1 = n
    m1 = m
    l1 = l
    if(present(s_n)) then
      n0 = s_n
      n1 = n1 + s_n - 1
    end if

    if(present(s_m)) then
      m0 = s_m
      m1 = n1 + s_m - 1
    end if

    if(present(s_l)) then
      l0 = s_l
      l1 = l1 + s_l - 1
    end if

    call dealloc(this, status)
    errmsg=""
    allocate(this(n0:n1, m0:m1, l0:l1), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

    this(:,:,:) = null_i2

  end subroutine alloc_i2_3d;

          ! sub to allocate 1d integer type arrays
  subroutine alloc_i4(this, n, status, src)

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    integer(i4), allocatable, dimension(:), intent(inout) :: this
    integer(i4), dimension(:), optional, intent(in) :: src

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_i4"

    call dealloc(this, status)
    errmsg=""
    if(present(src)) then
      allocate(this(n), stat=stat, errmsg=errmsg, source=src)
    else
      allocate(this(n), stat=stat, errmsg=errmsg)
      this = null_i4
    end if
    call FError(function, stat, errmsg, status)

  end subroutine alloc_i4;

              ! sub to allocate 2d integer type arrays
  subroutine alloc_i4_2d(this, n, m, status, s_n, s_m)

    integer :: stat, n0, n1, m0, m1
    integer, intent(in) :: n, m
    integer, intent(inout) :: status

    integer, intent(in), optional :: s_n, s_m

    integer(i4), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_i4_2d"

    n0 = 1
    m0 = 1
    n1 = n
    m1 = m
    if(present(s_n)) then
      n0 = s_n
      n1 = n1 + s_n - 1
    end if

    if(present(s_m)) then
      m0 = s_m
      m1 = n1 + s_m - 1
    end if

    call dealloc(this, status)
    errmsg=""
    allocate(this(n0:n1, m0:m1), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

    this(:,:) = null_i4

  end subroutine alloc_i4_2d;

   ! sub to allocate 3d integer type arrays
  subroutine alloc_i4_3d(this, n, m, l, status, s_n, s_m, s_l)

    integer :: stat, n0, n1, m0, m1, l0, l1
    integer, intent(in) :: n, m, l
    integer, intent(inout) :: status

    integer, intent(in), optional :: s_n, s_m, s_l

    integer(i4), allocatable, dimension(:,:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_i4_3d"

    n0 = 1
    m0 = 1
    l0 = 1
    n1 = n
    m1 = m
    l1 = l
    if(present(s_n)) then
      n0 = s_n
      n1 = n1 + s_n - 1
    end if

    if(present(s_m)) then
      m0 = s_m
      m1 = n1 + s_m - 1
    end if

    if(present(s_l)) then
      l0 = s_l
      l1 = l1 + s_l - 1
    end if

    call dealloc(this, status)
    errmsg=""
    allocate(this(n0:n1, m0:m1, l0:l1), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

    this(:,:,:) = null_i4

  end subroutine alloc_i4_3d;

              ! sub to allocate 1d integer type arrays
  subroutine alloc_i8(this, n, status, src)

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    integer(i8), allocatable, dimension(:) :: this
    integer(i8), dimension(:), optional, intent(in) :: src

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_i8"

    call dealloc(this, status)
    errmsg=""
    if(present(src)) then
      allocate(this(n), stat=stat, errmsg=errmsg, source=src)
    else
      allocate(this(n), stat=stat, errmsg=errmsg)
      this = null_i8
    end if
    call FError(function, stat, errmsg, status)

  end subroutine alloc_i8;

              ! sub to allocate 2d integer type arrays
  subroutine alloc_i8_2d(this, n, m, status, s_n, s_m)

    integer :: stat, n0, n1, m0, m1
    integer, intent(in) :: n, m
    integer, intent(inout) :: status

    integer, intent(in), optional :: s_n, s_m

    integer(i8), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_i8_2d"

    n0 = 1
    m0 = 1
    n1 = n
    m1 = m
    if(present(s_n)) then
      n0 = s_n
      n1 = n1 + s_n - 1
    end if

    if(present(s_m)) then
      m0 = s_m
      m1 = n1 + s_m - 1
    end if

    call dealloc(this, status)
    errmsg=""
    allocate(this(n0:n1, m0:m1), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

    this(:,:) = null_i8

  end subroutine alloc_i8_2d;


      ! sub to deallocate integer type arrays
  subroutine dealloc_i1(this, status)

    integer :: stat
    integer, intent(inout) :: status
    integer(i1), allocatable, dimension(:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_i1"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_i1

              ! sub to deallocate integer type arrays 2d
  subroutine dealloc_i1_2d(this, status)

    integer :: stat
    integer, intent(inout) :: status

    integer(i1), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_i1_2d"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_i1_2d


          ! sub to deallocate integer type arrays
  subroutine dealloc_i2(this, status)

    integer :: stat
    integer, intent(inout) :: status
    integer(i2), allocatable, dimension(:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_i2"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_i2

         ! sub to deallocate integer type arrays 2d
  subroutine dealloc_i2_2d(this, status)

    integer :: stat
    integer, intent(inout) :: status

    integer(i2), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_i2_2d"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_i2_2d

            ! sub to deallocate real type arrays 3d
  subroutine dealloc_i2_3d(this, status)

    integer :: stat
    integer, intent(inout) :: status

    integer(i2), allocatable, dimension(:,:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_i2_3d"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_i2_3d

          ! sub to deallocate integer type arrays
  subroutine dealloc_i4(this, status)

    integer :: stat
    integer, intent(inout) :: status
    integer(i4), allocatable, dimension(:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_i4"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_i4

              ! sub to deallocate integer type arrays 2d
  subroutine dealloc_i4_2d(this, status)

    integer :: stat
    integer, intent(inout) :: status

    integer(i4), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_i4_2d"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_i4_2d

          ! sub to deallocate real type arrays 3d
  subroutine dealloc_i4_3d(this, status)

    integer :: stat
    integer, intent(inout) :: status

    integer(i4), allocatable, dimension(:,:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_i4_3d"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_i4_3d

              ! sub to deallocate integer type arrays
  subroutine dealloc_i8(this, status)

    integer :: stat
    integer, intent(inout) :: status
    integer(i8), allocatable, dimension(:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_i8"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_i8

  ! sub to deallocate integer type arrays 2d
  subroutine dealloc_i8_2d(this, status)

    integer :: stat
    integer, intent(inout) :: status

    integer(i8), allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_i8_2d"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_i8_2d

  ! sub to allocate 1d character type arrays
  subroutine alloc_character(this, n, status)

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    character(*), allocatable, dimension(:) :: this
    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_character"

    call dealloc(this, status)
    errmsg=""
    allocate(this(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

    this = null_c

  end subroutine alloc_character;

  ! sub to deallocate character type arrays
  subroutine dealloc_character(this, status)

    integer :: stat
    integer, intent(inout) :: status

    character(*), allocatable, dimension(:) :: this
    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_character"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_character

      ! sub to allocate 1d logical type arrays
  subroutine alloc_logical(this, n, status)

    integer :: stat
    integer, intent(in) :: n
    integer, intent(inout) :: status

    logical, allocatable, dimension(:) :: this
    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_logical"

    call dealloc(this, status)
    errmsg=""
    allocate(this(n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_logical;

      ! sub to allocate 2d logical type arrays
  subroutine alloc_logical_2d(this, m, n, status)

    integer :: stat
    integer, intent(in) :: m,n
    integer, intent(inout) :: status

    logical, allocatable, dimension(:,:) :: this
    character(buffsize) :: errmsg
    character(*), parameter :: function = "alloc_logical_2d"

    call dealloc(this, status)
    errmsg=""
    allocate(this(m,n), stat=stat, errmsg=errmsg)
    call FError(function, stat, errmsg, status)

  end subroutine alloc_logical_2d

  ! sub to deallocate logical type arrays
  subroutine dealloc_logical(this, status)

    integer :: stat
    integer, intent(inout) :: status

    logical, allocatable, dimension(:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_logical"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_logical

  ! sub to deallocate 2d logical type arrays
  subroutine dealloc_logical_2d(this, status)

    integer :: stat
    integer, intent(inout) :: status

    logical, allocatable, dimension(:,:) :: this

    character(buffsize) :: errmsg
    character(*), parameter :: function = "dealloc_logical_2d"

    if(allocated(this)) then
      errmsg=""
      deallocate(this, stat=stat, errmsg=errmsg)
      call FError(function, stat, errmsg, status)
    end if

  end subroutine dealloc_logical_2d

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> intersect two integer arrays
  !
  !
  !
  !> @param[in] array1, array2
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine intersect_int(array1, array2, outarray, status)
    implicit none

    integer :: ii
    integer, dimension(:), intent(in) :: array1, array2

    integer, dimension(size(array1)) :: mask
    integer, dimension(:), allocatable, intent(out) :: outarray

    integer, intent(inout) :: status

    if(status > 100 .or. status == s_errRep) return

    mask(:) = false
    do ii = 1, size(array2)
      where(array1 == array2(ii))
        mask = true
      end where
    end do

    call alloc(outarray, sum(mask), status)

    outarray = pack(array1, mask == true)

  end subroutine intersect_int

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> intersect two real arrays
  !
  !
  !
  !> @param[in] array1, array2
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine intersect_real(array1, array2, outarray, status)
    implicit none

    integer :: ii
    real, dimension(:), intent(in) :: array1, array2

    integer, dimension(size(array1)) :: mask
    real, dimension(:), allocatable, intent(out) :: outarray

    integer, intent(inout) :: status

    if(status > 100 .or. status == s_errRep) return

    mask(:) = false
    do ii = 1, size(array2)
      where(array1 == array2(ii))
        mask = true
      end where
    end do

    call alloc(outarray, sum(mask), status)
    outarray = pack(array1, mask == true)

  end subroutine intersect_real

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> intersect two double arrays
  !
  !
  !
  !> @param[in] array1, array2
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine intersect_double(array1, array2, outarray, status)
    implicit none

    integer :: ii
    real(r8), dimension(:), intent(in) :: array1, array2

    integer, dimension(size(array1)) :: mask
    real(r8), dimension(:), allocatable, intent(out) :: outarray

    integer, intent(inout) :: status

    if(status > 100 .or. status == s_errRep) return

    mask(:) = false
    do ii = 1, size(array2)
      where(array1 == array2(ii))
        mask = true
      end where
    end do

    call alloc(outarray, sum(mask), status)
    outarray = pack(array1, mask == true)

  end subroutine intersect_double

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> intersect two integer arrays
  !> out: indices of array1
  !
  !
  !> @param[in] array1, array2
  !> @param[inout] status
  !---------------------------------------------------------------------------

  function intersect_int_order(array1, array2, status)
    implicit none

    integer :: ii
    integer, dimension(:), intent(in) :: array1, array2
    integer, dimension(:), allocatable :: intersect_int_order, mask
    integer, intent(inout) :: status

    call alloc(intersect_int_order, size(array1), status)
    call alloc(mask, size(array1), status)

    intersect_int_order = [(ii, ii=1, size(array1), 1)]

    mask(:) = false
    do ii = 1, size(array2)
      where(array1 == array2(ii))
        mask = true
      end where
    end do

    intersect_int_order = pack(intersect_int_order, mask == true)
    call dealloc(mask, status)

  end function intersect_int_order

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the complement of two integer arrays
  !> the complement of array1 which is not in array2
  !> out: indices of array1
  !
  !
  !> @param[in] array1, array2
  !> @param[inout] status
  !---------------------------------------------------------------------------

  function complement_int_order(array1, array2, status)
    implicit none

    integer :: ii
    integer, dimension(:), intent(in) :: array1, array2
    integer, dimension(:), allocatable :: complement_int_order, mask
    integer, intent(inout) :: status

    call alloc(complement_int_order, size(array1), status)
    call alloc(mask, size(array1), status)

    complement_int_order = [(ii, ii=1, size(array1), 1)]

    mask(:) = true
    do ii = 1, size(array2)
      where(array1 == array2(ii))
        mask = false
      end where
    end do

    complement_int_order = pack(complement_int_order, mask == true)
    call dealloc(mask, status)

  end function complement_int_order

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> delete a file using fortran intrinsic
  !
  !
  !
  !> @param[in] filename
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine delete_file_fortran(filename, status)
    implicit none

    integer, intent(inout) :: status
    integer :: stat, unit
    logical :: file_exists

    character(*), intent(in) :: filename
    character(*), parameter :: function = "delete_file_fortran"

    if(status > 100 .or. status == s_errRep) return

    inquire(file=filename, exist=file_exists)
    if(file_exists .eqv. .false.) return

    open(newunit = unit, iostat = stat, file = filename, status = 'old')
    call FError(function, stat, "opening file " // trim(filename) // " to be deleted", status)

    if (status == s_errRep) return

    close(unit, iostat = stat, status = 'delete')
    call FError(function, stat, "deleting file " // trim(filename), status)

  end subroutine delete_file_fortran

  !.....................................................................
  subroutine start_t_timer(this)
    implicit none
    class(t_timer) :: this

    call system_clock(count_rate=this%ticks_per_sec)
    call system_clock(count=this%start_time)
    this%started = .true.

  end subroutine start_t_timer

  !.....................................................................
  real function stopp_t_timer(this) result(tdiff)
    implicit none
    class(t_timer) :: this

    tdiff = 0.0

    if(.not.this%started)&
call wrstop("stopp_t_timer", "Timer has not been started.")

    call system_clock(count=this%stop_time)
!    tdiff = real(this%stop_time - this%start_time)/this%ticks_per_sec
    tdiff = real(this%stop_time - this%start_time)/real(this%ticks_per_sec)

    this%started = .false.

  end function stopp_t_timer

  !.....................................................................
  subroutine create_t_array_iterator(this, sizes)
    implicit none

    class(t_array_iterator) :: this
    integer, dimension(:), intent(in) :: sizes

    character(*), parameter :: functionName = "create_t_array_iterator"
    integer :: ierr,ii

    if(this%is_initialized)&
call wrstop(functionName, "You must destroy the previous object before you can construct a new one.")

    if(any(sizes<1))&
call wrstop(functionName, "Non-positive array size values are not allowed.")

    ierr = 0
    this%n_dims = size(sizes)
    call alloc(this%sizes,   this%n_dims, ierr)
    call alloc(this%indices, this%n_dims, ierr)
    do ii=1,this%n_dims
      this%sizes(ii) = sizes(ii)
      this%indices(ii) = 1
    end do

    this%ii = 1
    this%is_initialized = .true.

  end subroutine create_t_array_iterator

  !.....................................................................
  subroutine get_t_array_iterator(this, indices)
    implicit none

    class(t_array_iterator) :: this
    integer, dimension(this%n_dims), intent(out) :: indices

    character(*), parameter :: functionName = "get_t_array_iterator"

    indices = this%indices

  end subroutine get_t_array_iterator

  !.....................................................................
  subroutine iterate_t_array_iterator(this, indices)
    ! Supposing we have an N-dimensional orthogonal array whose dimensions are given by the N-element sequence S={s_1,s_2,...,s_N}. Clearly the total number of array elements M is given by
    !
    !	    __N
    !	M = ||  s_i.
    !	    i=1
    !
    ! We want to be able to iterate through all M of the array elements by generating, at each step, an N-element series T={t_1,t_2,...,t_} such that 1<=t_i<=N for all i. That is the function of the present method.
    !
    implicit none

    class(t_array_iterator) :: this
    integer, dimension(this%n_dims), intent(out), optional :: indices

    character(*), parameter :: functionName = "iterate_t_array_iterator"
    integer :: ii

    do ii=1,this%n_dims
      this%indices(ii) = this%indices(ii) + 1
      if(this%indices(ii) > this%sizes(ii)) then
        this%indices(ii) = 1
      else
    exit
      end if
    end do

    if(present(indices))&
      indices = this%indices

  end subroutine iterate_t_array_iterator

  !.....................................................................
  subroutine destroy_t_array_iterator(this)
    implicit none

    class(t_array_iterator) :: this

    character(*), parameter :: functionName = "destroy_t_array_iterator"
    integer :: ierr

    ierr = 0

    call dealloc(this%sizes, ierr)
    call dealloc(this%indices, ierr)
    this%n_dims=0
    this%ii=0
    this%is_initialized = .false.

  end subroutine destroy_t_array_iterator

end module erolib_general_mod
