! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! Non-method subroutines contained:
!  subroutine read_gtis()

module erolib_gti_mod
  use param
  use erolib_general_mod
  use erolib_ranges_mod

  implicit none

logical, private, parameter :: doTest=.false.

  type :: tGTIs
    real(r8), dimension(:), allocatable :: starts
    real(r8), dimension(:), allocatable :: stops
  contains
    procedure, non_overridable :: get_num => get_num_tGTIs
    procedure, non_overridable :: alloc => alloc_tGTIs
    procedure, non_overridable :: dealloc => dealloc_tGTIs
    procedure, non_overridable :: realloc => realloc_tGTIs
    procedure, non_overridable :: is_legal => is_legal_tGTIs
    procedure, private         :: convert_to_ranges => convert_to_ranges_tGTIs
    procedure, private         :: load_from_ranges => load_from_ranges_tGTIs
    procedure, non_overridable :: and_with => and_with_tGTIs
    procedure, non_overridable :: or_with => or_with_tGTIs
    procedure, non_overridable :: and_with_not => and_with_not_tGTIs
    procedure, non_overridable :: contain => contain_tGTIs
  end type tGTIs

contains

  !.....................................................................
  function get_num_tGTIs(this) result(num_gtis)
    implicit none

    class(tGTIs) :: this
    integer :: num_gtis

    character(*), parameter :: functionName = "get_num_tGTIs"

    if(allocated(this%starts)) then
      if(allocated(this%stops)) then
        num_gtis = size(this%starts)
        if(num_gtis /= size(this%stops))&
call wrstop(functionName, "Arrays 'starts' and 'stops' have different sizes.")
      else
call wrstop(functionName, "Array 'starts' is allocated but 'stops' is not.")
      end if
    else
      if(allocated(this%stops)) then
call wrstop(functionName, "Array 'stops' is allocated but 'starts' is not.")
      else
        num_gtis = 0
      end if
    end if

  end function get_num_tGTIs

  !.....................................................................
  subroutine alloc_tGTIs(this, n)
    implicit none

    class(tGTIs) :: this
    integer, intent(in) :: n

    character(*), parameter :: functionName = "alloc_tGTIs"
    integer :: stat
    character(buffsize) :: errmsg

    if(n < 0)&
call wrstop(functionName, "You can't allocate an array to less than zero size.")

    call this%dealloc()
    errmsg = ""
    stat = 0

    allocate(this%starts(n), stat=stat, errmsg=errmsg)
    if(stat/=0)&
call wrstop(functionName, "Allocation error: "//trim(errmsg))

    allocate(this%stops(n),  stat=stat, errmsg=errmsg)
    if(stat/=0)&
call wrstop(functionName, "Allocation error: "//trim(errmsg))

    this%starts = 0.0d0
    this%stops  = 0.0d0

  end subroutine alloc_tGTIs

  !.....................................................................
  subroutine dealloc_tGTIs(this)
    implicit none

    class(tGTIs) :: this

    character(*), parameter :: functionName = "dealloc_tGTIs"
    integer :: stat
    character(buffsize) :: errmsg

    errmsg = ""
    stat = 0

    if(allocated(this%starts)) then
      deallocate(this%starts, stat=stat, errmsg=errmsg)
      if(stat/=0)&
call wrstop(functionName, "Deallocation error: "//trim(errmsg))
    end if

    if(allocated(this%stops)) then
      deallocate(this%stops, stat=stat, errmsg=errmsg)
      if(stat/=0)&
call wrstop(functionName, "Deallocation error: "//trim(errmsg))
    end if

  end subroutine dealloc_tGTIs

  !.....................................................................
  subroutine realloc_tGTIs(this, n)
    implicit none

    class(tGTIs) :: this
    integer, intent(in) :: n

    character(*), parameter :: functionName = "realloc_tGTIs"

    call this%dealloc()
    call this%alloc(n)

  end subroutine realloc_tGTIs

  !.....................................................................
  function is_legal_tGTIs(this) result(is_legal)
    implicit none

    class(tGTIs) :: this
    logical :: is_legal

    character(*), parameter :: functionName = "is_legal_tGTIs"
    integer :: num_gtis
    type(tRange), dimension(:), allocatable :: ranges

    num_gtis = this%get_num() ! checks that both starts and stops are allocated and of the same size.
    if(num_gtis<1) then
      is_legal = .true.
  return
    end if

!    if(any(this%starts>=this%stops)) then
!      is_legal = .false.
!  return
!    end if

!    if(num_gtis<2) then
!      is_legal = .true.
!  return
!    end if

!    if(any(this%starts(2:num_gtis)<=this%stops(1:num_gtis-1))) then
!      is_legal = .false.
!  return
!    end if

!    is_legal = .true.

    ranges = this%convert_to_ranges()
    is_legal = ranges_are_legal(ranges)
    deallocate(ranges)

  end function is_legal_tGTIs

  !.....................................................................
  function convert_to_ranges_tGTIs(this) result(ranges)
    implicit none

    class(tGTIs) :: this
    type(tRange), dimension(:), allocatable :: ranges

    character(*), parameter :: functionName = "convert_to_ranges_tGTIs"
    integer :: num_gtis

    num_gtis = this%get_num() ! checks that both starts and stops are allocated and of the same size.
    allocate(ranges(num_gtis))

    if(num_gtis<1)&
  return

    ranges%lo%isUndefined = .false.
    ranges%hi%isUndefined = .false.
    ranges%lo%value = this%starts
    ranges%hi%value = this%stops

  end function convert_to_ranges_tGTIs

  !.....................................................................
  subroutine load_from_ranges_tGTIs(this,ranges)
    implicit none

    class(tGTIs) :: this
    type(tRange), dimension(:), intent(in) :: ranges

    character(*), parameter :: functionName = "load_from_ranges_tGTIs"
    integer :: num_ranges

    num_ranges = size(ranges)
    call this%realloc(num_ranges)

    if(num_ranges<1)&
  return

    if(any(ranges%lo%isUndefined) .or. any(ranges%hi%isUndefined))&
call wrstop(functionName, "Input ranges may not have any undefined bounds.")

    this%starts = ranges%lo%value
    this%stops  = ranges%hi%value

  end subroutine load_from_ranges_tGTIs

  !.....................................................................
  subroutine and_with_tGTIs(this, gtis)!, check_legality)
    implicit none

    class(tGTIs) :: this
    type(tGTIs),       intent(in) :: gtis
!    logical, optional, intent(in) :: check_legality

    character(*), parameter :: functionName = "and_with_tGTIs"
    type(tRange), dimension(:), allocatable :: rangesA,rangesB,ANDed_ranges

    rangesA = this%convert_to_ranges()
    rangesB = gtis%convert_to_ranges()
    call and_ranges(rangesA, rangesB, ANDed_ranges, check_legality=.true.)
    call this%load_from_ranges(ANDed_ranges)

    deallocate(rangesA, rangesB, ANDed_ranges)

  end subroutine and_with_tGTIs

  !.....................................................................
  subroutine or_with_tGTIs(this, gtis)!, check_legality)
    implicit none

    class(tGTIs) :: this
    type(tGTIs),       intent(in) :: gtis
!    logical, optional, intent(in) :: check_legality

    character(*), parameter :: functionName = "or_with_tGTIs"
    type(tRange), dimension(:), allocatable :: rangesA,rangesB,ORed_ranges

    rangesA = this%convert_to_ranges()
    rangesB = gtis%convert_to_ranges()
    call or_ranges(rangesA, rangesB, ORed_ranges, check_legality=.true.)
    call this%load_from_ranges(ORed_ranges)

    deallocate(rangesA, rangesB, ORed_ranges)

  end subroutine or_with_tGTIs

  !.....................................................................
  subroutine and_with_not_tGTIs(this, gtis)!, check_legality)
    implicit none

    class(tGTIs) :: this
    type(tGTIs), intent(in) :: gtis
!    logical, optional, intent(in) :: check_legality

    character(*), parameter :: functionName = "and_with_not_tGTIs"
    type(tRange), dimension(:), allocatable :: rangesA,rangesB,not_rangesB,ANDed_ranges

    rangesA = this%convert_to_ranges()
    rangesB = gtis%convert_to_ranges()
    call negate_ranges(rangesB, not_rangesB) ! performs a logical 'not'. **NOTE** this performs a mandatory check of legality.
    call and_ranges(rangesA, not_rangesB, ANDed_ranges, check_legality=.true.)
    call this%load_from_ranges(ANDed_ranges)

    deallocate(rangesA, rangesB, not_rangesB, ANDed_ranges)

  end subroutine and_with_not_tGTIs

  !.....................................................................
  logical function contain_tGTIs(this, time) result(time_in_gti)
    ! Is the input time contained in any of the GTIs? (Slow would be: any(time>=this%starts .and. time<this%stops))
    implicit none

    class(tGTIs)         :: this
    real(r8), intent(in) :: time

    integer :: ilo,ihi,imid,n_GTIs

    n_GTIs = this%get_num()

    ! are there any GTIs?
    if( n_GTIs <= 0 ) then
       time_in_gti = .false.
  return
    end if

    ! is event outside total GTI range?
    if( time<this%starts(1) .or. time>=this%stops(n_GTIs) ) then
       time_in_gti = .false.
  return
    end if

    ! binary search to see whether time is within range
    ilo = 1
    ihi = n_GTIs
    do
       if(ilo == ihi) then
          ! check single GTI
          time_in_gti = (time>=this%starts(ilo) .and. time<this%stops(ilo))
  return
       end if

       imid = (ilo+ihi)/2
       if(imid == ilo) then
          ! only two elements to look at so we just check both
          time_in_gti = (time>=this%starts(ilo) .and. time<this%stops(ilo)) .or. &
               (time>=this%starts(ihi) .and. time<this%stops(ihi))
  return
       end if

       if(time < this%starts(imid)) then
          ihi = imid-1
       else
          ilo = imid
       end if

    end do

  end function contain_tGTIs

  !.....................................................................
  subroutine read_gtis(gti_tab_root, tmi, unitNo, gtis, status)
    ! This has been adapted from the routine get_gti() which was present
    ! in expmap at least up to version 2.4. It is intended for use by
    ! expmap itself and by tasks such as apetool and flaregti which
    ! until now have called the expmap routine.

    implicit none

    character(len=*), intent(in)    :: gti_tab_root
    integer,          intent(in)    :: tmi, unitNo
    type(tGTIs),      intent(out)   :: gtis
    integer,          intent(inout) :: status

    character(*), parameter :: functionName = "read_gtis"
    character(len=9) :: gtinam
    integer :: ierr,ngti
    integer :: tstartcol, tendcol
    integer :: igti, ival
    logical :: undef
    integer,parameter :: nblock = 7200

    ierr = 0
    call gtis%alloc(0)

    !if (gti_tab_root /= "FLAREGTI") then
    write (gtinam, '(a, i1)') trim(gti_tab_root), tmi
    !else
    !  gtinam=gti_tab_root
    !end if
    call ftmnhd(unitNo, BINARY_TBL, gtinam, 0, ierr)
    if (ierr > 0) then
      status = 1
      call wrerr(functionName, status, "Missing " // trim(gtinam) // " extension")
  return
    end if 
    call ftgnrw(unitNo, ngti, ierr)
    call fitserr("ftgnrw", ierr, status)
    call ftgcno(unitNo, .false., "START", tstartcol, ierr)
    call fitserr("ftgcno", ierr, status, " (colname=START)")
    call ftgcno(unitNo, .false., "STOP", tendcol, ierr)
    call fitserr("ftgcno", ierr, status, " (colname=STOP)")

    call gtis%realloc(ngti)

    do igti = 1, ngti, nblock
      ival = min(nblock,ngti-igti+1) 
      call ftgcvd(unitNo, tstartcol, igti, 1, ival, null_r8, &
                  gtis%starts(igti:igti+ival-1), undef, ierr)
      call fitserr("ftgcvd", ierr, status)
      call ftgcvd(unitNo, tendcol, igti, 1, ival, null_r8, &
                  gtis%stops(igti:igti+ival-1), undef, ierr)
      call fitserr("ftgcvd", ierr, status)
    end do

    if (ierr > 0) then
      status = 3
      call wrerr(functionName, status, "Error reading GTI extension")
  return
    end if 
    
!    do igti = 1, ngti
!      if (gtis%stops(igti)-gtis%starts(igti) > 1d7) then
!         status = 3
!         call wrerr(functionName, status, "Wrong gti")
!    exit 
!      end if
!    end do 

    if(.not.gtis%is_legal()) then
      status = 3
      call wrerr(functionName, status, "GTIs are not legal.")
    end if

  end subroutine read_gtis

end module erolib_gti_mod
