! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! Non-method subroutines contained:
!  subroutine get_gti_with_overlap(gti, gti_with_overlap, overlap)
!  function get_GTI_mask(time, GTI, status) result (mask)
!  subroutine setup_tmp_GTI(numRows, tmp_GTI, status)
!  subroutine dealloc_gtis(tables, status)
!  subroutine alloc_gtis(tables, n, status)

module erolib_gtitab_mod

  use erolib_general_mod
  use erolib_sort_mod
  use erolib_str_mod
  use erolib_bintable_mod
  use erolib_ranges_mod

  implicit none

logical, private, parameter :: doTest=.false.

  type, extends(tBinTable) :: tGTI_tab
    integer :: tel_id = -1
    integer :: ext_type = -1
    character(buffsize) :: extNameRoot=""
    logical :: hasBeenSanityChecked = .false.
!*** sort out these two! What they are meant to do, etc.
logical :: use_mjdref = .false.
logical :: use_timezero = .false.
  contains
    procedure :: init => init_tGTI_tab
    procedure :: initHeader => initHeader_tGTI_tab
    procedure, private :: convert_to_ranges => convert_to_ranges_tGTI_tab
    procedure, private :: set_limit_keywords => set_limit_keywords_tGTI_tab
    procedure, private :: load_from_ranges => load_from_ranges_tGTI_tab
    procedure :: sanityCheck => sanityCheck_tGTI_tab
    procedure :: repair => repair_tGTI_tab
    procedure :: rebase => rebase_tGTI_tab
    procedure :: rebase_mjdref => rebase_mjdref_tGTI_tab
    procedure :: or_with => or_with_tGTI_tab
    procedure :: and_with => and_with_tGTI_tab
    procedure :: and_with_not => and_with_not_tGTI_tab
  end type tGTI_tab

  type, extends(tFitsFile) :: tGTIFile
    type(tGTI_tab) :: gti
  contains
    procedure :: read => read_tGTIFile
    procedure :: free => free_tGTIFile
  end type tGTIFile

  ! Extension indices:
  enum, bind(c)
    enumerator :: iGTI=1,iQUALGTI,iFLAREGTI,iREGIONGTI,iEXTERNAL,iTemp
  end enum

contains

  !.....................................................................
  subroutine init_tGTI_tab(this, extI, status, fileInfo, extNameRoot)
    implicit none

    class(tGTI_tab) :: this
    integer,                           intent(in)    :: extI
    integer,                           intent(inout) :: status
    type(tFitsInfo), optional, target, intent(in)    :: fileInfo
    character(*), optional,            intent(in)    :: extNameRoot

    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "init_tGTI_tab"

    select case(extI)
    case(iGTI)
      if(present(extNameRoot)) then
        this%extNameRoot = extNameRoot
      else
        this%extNameRoot = "GTI"
      end if
    case(iQUALGTI)
      if(present(extNameRoot)) then
        this%extNameRoot = extNameRoot
      else
        this%extNameRoot = "QUALGTI"
      end if
    case(iFLAREGTI)
      if(present(extNameRoot)) then
        this%extNameRoot = extNameRoot
      else
        this%extNameRoot = "FLAREGTI"
      end if
    case(iREGIONGTI)
      if(present(extNameRoot)) then
        this%extNameRoot = extNameRoot
      else
        this%extNameRoot = "REGIONGTI"
      end if
    case(iEXTERNAL)
      if(present(extNameRoot)) then
        this%extNameRoot = extNameRoot
      else
        this%extNameRoot = "STDGTI"
      end if
    case(iTemp) ! Not meant to be read or written - just for internal use.
      if(present(extNameRoot)) then
        this%extNameRoot = extNameRoot
      else
        this%extNameRoot = "dummy"
      end if
    case default
      errmsg = int2str(extI)
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(errmsg)//"'")
      status = s_errRep
  return
    end select

    if(present(fileInfo)) this%info%fileInfo => fileInfo
!    this%info%tableName = "GTI" !*** it won't always be exactly this.
    call this%initColumn("D", "START", "s")
    call this%initColumn("D", "STOP",  "s")

    call this%initColumns(status)

    call this%initHeader(status)

  end subroutine init_tGTI_tab

  !.....................................................................
  subroutine initHeader_tGTI_tab(this, status)
    implicit none

    class(tGTI_tab), target :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "initHeader_tGTI_tab"

    this%header%tabInfo => this%info

    call this%header%initKeyword("D", "MJDREF")
    call this%header%initKeyword("D", "TIMEZERO")
    call this%header%initKeyword("D", "TSTART")
    call this%header%initKeyword("D", "TSTOP")
    call this%header%initKeywords(status)

    call this%header%check_keywords(status)

  end subroutine initHeader_tGTI_tab

  !.....................................................................
  function convert_to_ranges_tGTI_tab(this) result(ranges)
    implicit none

    class(tGTI_tab) :: this
    type(tRange), dimension(:), allocatable :: ranges

    character(*), parameter :: functionName = "convert_to_ranges_tGTI_tab"
    integer :: num_gtis
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null()

    num_gtis = this%getNumRows() ! checks that both starts and stops are allocated and of the same size.
    allocate(ranges(num_gtis))

    if(num_gtis<1)&
  return

    call this%get_col_ptr("START", starts)
    call this%get_col_ptr("STOP",  stops)

    ranges%lo%isUndefined = .false.
    ranges%hi%isUndefined = .false.
    ranges%lo%value = starts
    ranges%hi%value = stops

    nullify(starts, stops)

  end function convert_to_ranges_tGTI_tab

  !.....................................................................
  subroutine set_limit_keywords_tGTI_tab(this)
    implicit none

    class(tGTI_tab), target :: this

    character(*), parameter :: functionName = "set_limit_keywords_tGTI_tab"
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null()

    call this%get_col_ptr("START", starts)
    call this%get_col_ptr("STOP",  stops)

    call this%header%setKeywordValue("TSTART", minval(starts))
    call this%header%setKeywordValue("TSTOP",  maxval(stops))

    nullify(starts, stops)

  end subroutine set_limit_keywords_tGTI_tab

  !.....................................................................
  subroutine load_from_ranges_tGTI_tab(this,ranges,status)
    implicit none

    class(tGTI_tab) :: this
    type(tRange), dimension(:), intent(in) :: ranges
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "load_from_ranges_tGTI_tab"
    integer :: num_ranges
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null()

    num_ranges = size(ranges)
    call this%realloc(status, num_ranges)

    if(num_ranges<1)&
  return

    if(any(ranges%lo%isUndefined) .or. any(ranges%hi%isUndefined))&
call wrstop(functionName, "Input ranges may not have any undefined bounds.")

    call this%get_col_ptr("START", starts)
    call this%get_col_ptr("STOP",  stops)

    starts(:) = ranges(:)%lo%value
    stops(:)  = ranges(:)%hi%value

    nullify(starts, stops)

  end subroutine load_from_ranges_tGTI_tab

  !.....................................................................
  subroutine sanityCheck_tGTI_tab(this, status, errmsg_prefix, do_warn)
    ! Performs a sanity check of the GTI; also calculates this%tstart and this%tstop.

    implicit none

    class(tGTI_tab) :: this
    integer,                intent(inout) :: status
    character(*), optional, intent(in)    :: errmsg_prefix
    logical,      optional, intent(in)    :: do_warn

    character(*), parameter :: functionName = "sanityCheck_tGTI_tab"
    integer :: numGTIs
    character(buffsize) :: errmsg
    logical :: local_do_warn
    type(tRange), dimension(:), allocatable :: ranges

    if(status > 100 .or. status == s_errRep)&
  return

    if(present(do_warn)) then
      local_do_warn = do_warn
    else
      local_do_warn = .true.
    end if

    if(present(errmsg_prefix)) then
      write(errmsg, '(a)') errmsg_prefix
    else
      errmsg = ""
    end if

    if(.not. this%isInitialized) then
      call wrerr(functionName, s_err, trim(errmsg)//" You must first designate the column properties.")
      status = s_errRep
  return
    else if(.not. allocated(this%cols)) then
      call wrerr(functionName, s_err, trim(errmsg)//" Your initialization routine should allocate this%cols.")
      status = s_errRep
  return
    end if

    numGTIs = this%getNumRows()

    if(numGTIs<1) then
      if(local_do_warn) call wrerr(functionName, warn, trim(errmsg)//" Zero GTI rows found.")
      this%hasBeenSanityChecked = .true.
  return
    end if

    ranges = this%convert_to_ranges()
    if(ranges_are_legal(ranges)) then
      this%hasBeenSanityChecked = .true.
      call this%set_limit_keywords()
    else
call wrstop(functionName, trim(errmsg)//"GTIs are not legal.")
    end if

    deallocate(ranges)

  end subroutine sanityCheck_tGTI_tab

  !.....................................................................
  subroutine repair_tGTI_tab(this, status, repair_was_needed)
    ! Repairs any disordered GTIs.

    implicit none

    class(tGTI_tab)                      :: this
    integer,           intent(inout) :: status
    logical, optional, intent(out)   :: repair_was_needed

    character(*), parameter :: functionName = "repair_tGTI_tab"
    integer :: numInputGTIs,numGTIs0,numGTIs1,gi
    type(tGTI_tab) :: temp_GTIs
    logical :: local_repair_was_needed
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null(),temp_starts=>null(),temp_stops=>null()
    integer, dimension(:), allocatable :: order
    character(buffsize) :: errmsg

    if(present(repair_was_needed)) repair_was_needed=.false. ! default

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not. this%isInitialized) then
      call wrerr(functionName, s_err, trim(errmsg)//" You must first designate the column properties.")
      status = s_errRep
  return
    else if(.not. allocated(this%cols)) then
      call wrerr(functionName, s_err, trim(errmsg)//" Your initialization routine should allocate this%cols.")
      status = s_errRep
  return
    end if

    call this%get_col_ptr("START", starts)
    call this%get_col_ptr("STOP",  stops)

    numInputGTIs = this%getNumRows()

    if(numInputGTIs<1) then
      nullify(stops)
      nullify(starts)
  return
    end if

    local_repair_was_needed = .false. ! default
    if(any(starts >= stops)) local_repair_was_needed=.true.
    if(numInputGTIs>1) then
      if(any(starts(2:) <= stops(1:numInputGTIs-1))) local_repair_was_needed=.true.
    end if

    if(present(repair_was_needed)) repair_was_needed=local_repair_was_needed

    if(.not.local_repair_was_needed) then
      nullify(stops)
      nullify(starts)
  return
    end if

    ! First excise any GTIs for which start is >= stop:
    !
    if(any(starts >= stops)) then
      call setup_tmp_GTI(numInputGTIs, temp_GTIs, status)
      call temp_GTIs%get_col_ptr("START", temp_starts)
      call temp_GTIs%get_col_ptr("STOP",  temp_stops)

      numGTIs0 = 0
      do gi=1,numInputGTIs
        if(starts(gi) < stops(gi)) then
          numGTIs0 = numGTIs0 + 1
          temp_starts(numGTIs0) = starts(gi)
          temp_stops( numGTIs0) = stops( gi)
        end if
      end do

      nullify(starts)
      nullify(stops)
      call this%realloc(status, numGTIs0)
      if(numGTIs0>0) then
        call this%get_col_ptr("START", starts)
        call this%get_col_ptr("STOP",  stops)
        starts = temp_starts(:numGTIs0)
        stops  = temp_stops( :numGTIs0)
      end if

      numInputGTIs = numGTIs0

      nullify(temp_starts)
      nullify(temp_stops)
      call temp_GTIs%destroy(status)
    end if

    if(numInputGTIs>1) then
      if(any(starts(2:) <= stops(1:numInputGTIs-1))) then
        call setup_tmp_GTI(numInputGTIs, temp_GTIs, status)
        call temp_GTIs%get_col_ptr("START", temp_starts)
        call temp_GTIs%get_col_ptr("STOP",  temp_stops)

        call sort(starts, status, order)
        stops = stops(order)
        deallocate(order)

        numGTIs1 = 1
        temp_starts(numGTIs1) = starts(1)
        temp_stops( numGTIs1) = stops(1)
        do gi=2,numInputGTIs
          if(starts(gi)>temp_stops(numGTIs1)) then
            numGTIs1 = numGTIs1 + 1
            temp_starts(numGTIs1) = starts(gi)
            temp_stops( numGTIs1) = stops( gi)
          else if(stops(gi)>temp_stops(numGTIs1)) then
            temp_stops( numGTIs1) = stops( gi)
          end if
        end do

        nullify(starts)
        nullify(stops)
        call this%realloc(status, numGTIs1)
        call this%get_col_ptr("START", starts)
        call this%get_col_ptr("STOP",  stops)
        starts = temp_starts(:numGTIs1)
        stops  = temp_stops( :numGTIs1)

        numInputGTIs = numGTIs1

        nullify(temp_starts)
        nullify(temp_stops)
        call temp_GTIs%destroy(status)
      end if
    end if

    call this%header%setKeywordValue("TSTART", minval(starts))
    call this%header%setKeywordValue("TSTOP",  maxval(stops))
    this%hasBeenSanityChecked = .true.

    nullify(stops)
    nullify(starts)

  end subroutine repair_tGTI_tab

  !.....................................................................
  subroutine rebase_mjdref_tGTI_tab(this, mjdref_old, mjdref_new)
    implicit none
    class(tGTI_tab) :: this
    real(r8), intent(in) :: mjdref_old, mjdref_new

    character(*), parameter :: functionName = "rebase_mjdref_tGTI_tab"
    real(r8) :: offset
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null()

    offset = (mjdref_old - mjdref_new)*3600.0_r8*24.0_r8

    call this%get_col_ptr("START", starts)
    call this%get_col_ptr("STOP",  stops)
    starts = offset + starts
    stops  = offset + stops
    nullify(stops)
    nullify(starts)

  end subroutine rebase_mjdref_tGTI_tab

  !.....................................................................
  subroutine rebase_tGTI_tab(this, ref_gti, status)
    ! This changes the mjdref or timezero in 'this' to equal those in
    ! the ref_gti, then recalculates times in 'this' to conform to this
    ! new reference.

    implicit none

    class(tGTI_tab) :: this
    type(tGTI_tab), intent(in)    :: ref_gti
    integer,    intent(inout) :: status

    character(*), parameter :: functionName = "rebase_tGTI_tab"
    real(r8) :: offset,mjdref_this,mjdref_ref,tzero_this,tzero_ref
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null()

    if(.not.this%hasBeenSanityChecked) then
      call wrerr(functionName, s_err, "GTIs have not been sanity checked.")
      status = s_errRep
  return
    end if

    if(this%header%keywordContainsData("MJDREF")) then
      call this%header%getKeywordValue("MJDREF",mjdref_this)
      if(ref_gti%header%keywordContainsData("MJDREF")) then
        call ref_gti%header%getKeywordValue("MJDREF",mjdref_ref)
        if(mjdref_ref /= mjdref_this) then
          call this%rebase_mjdref(mjdref_this, mjdref_ref)

          call this%header%setKeywordValue("MJDREF",mjdref_ref)
        end if

      else if(ref_gti%header%keywordContainsData("TIMEZERO")) then
        call wrerr(functionName, s_err, "GTI and reference GTI: time references don't match.")
        status = s_errRep
  return
      else
        call wrerr(functionName, s_err, "Reference GTI has no valid time reference.")
        status = s_errRep
  return
      end if

    else if(this%header%keywordContainsData("TIMEZERO")) then
      call this%header%getKeywordValue("TIMEZERO",tzero_this)
      if(ref_gti%header%keywordContainsData("TIMEZERO")) then
        call ref_gti%header%getKeywordValue("TIMEZERO",tzero_ref)
        if(tzero_ref /= tzero_this) then
          offset = tzero_this - tzero_ref

          call this%get_col_ptr("START", starts)
          call this%get_col_ptr("STOP",  stops)
          starts = offset + starts
          stops  = offset + stops
          nullify(stops)
          nullify(starts)

          call this%header%setKeywordValue("TIMEZERO",tzero_ref)
        end if

      else if(ref_gti%header%keywordContainsData("MJDREF")) then
        call wrerr(functionName, s_err, "GTI and reference GTI: time references don't match.")
        status = s_errRep
  return
      else
        call wrerr(functionName, s_err, "Reference GTI has no valid time reference.")
        status = s_errRep
  return
      end if

    else
      call wrerr(functionName, s_err, "GTI has no valid time reference.")
      status = s_errRep
  return
    end if

  end subroutine rebase_tGTI_tab

  !.....................................................................
  subroutine or_with_tGTI_tab(this, gtis, status)
    implicit none
    class(tGTI_tab) :: this
    type(tGTI_tab), intent(in)    :: gtis
    integer,    intent(inout) :: status

    character(*), parameter :: functionName = "or_with_tGTI_tab"
    type(tRange), dimension(:), allocatable :: rangesA,rangesB,ORed_ranges

    rangesA = this%convert_to_ranges()
    rangesB = gtis%convert_to_ranges()
    call or_ranges(rangesA, rangesB, ORed_ranges, check_legality=.true.)
    call this%load_from_ranges(ORed_ranges, status)

    deallocate(rangesA, rangesB, ORed_ranges)

  end subroutine or_with_tGTI_tab

  !.....................................................................
  subroutine and_with_tGTI_tab(this, gtis, status)
    implicit none
    class(tGTI_tab) :: this
    type(tGTI_tab), intent(in)    :: gtis
    integer,    intent(inout) :: status

    character(*), parameter :: functionName = "and_with_tGTI_tab"
    type(tRange), dimension(:), allocatable :: rangesA,rangesB,ANDed_ranges

    rangesA = this%convert_to_ranges()
    rangesB = gtis%convert_to_ranges()
    call and_ranges(rangesA, rangesB, ANDed_ranges, check_legality=.true.)
    call this%load_from_ranges(ANDed_ranges, status)

    deallocate(rangesA, rangesB, ANDed_ranges)

  end subroutine and_with_tGTI_tab

  !.....................................................................
  subroutine and_with_not_tGTI_tab(this, btis, status)
    implicit none
    class(tGTI_tab) :: this
    type(tGTI_tab), intent(in)    :: btis ! bad time intervals.
    integer,    intent(inout) :: status

    character(*), parameter :: functionName = "and_with_not_tGTI_tab"
    type(tRange), dimension(:), allocatable :: rangesA,rangesB,not_rangesB,ANDed_ranges

    rangesA = this%convert_to_ranges()
    rangesB = btis%convert_to_ranges()
    call negate_ranges(rangesB, not_rangesB) ! performs a logical 'not'. **NOTE** this performs a mandatory check of legality.
    call and_ranges(rangesA, not_rangesB, ANDed_ranges, check_legality=.true.)
    call this%load_from_ranges(ANDed_ranges, status)

    deallocate(rangesA, rangesB, not_rangesB, ANDed_ranges)

  end subroutine and_with_not_tGTI_tab

  !=====================================================================
  ! The remaining routines are not methods of tGTI_tab, they're just
  ! utility routines that manipulate the objects.

  !.....................................................................
  subroutine get_gti_with_overlap(gti, gti_with_overlap, overlap)
    implicit none
    type(tGTI_tab),         intent(in)  :: gti
    type(tGTI_tab),         intent(out) :: gti_with_overlap
    real(r8), optional, intent(in)  :: overlap

    real(r8), dimension(:), pointer :: starts=>null(),stops=>null()

    gti_with_overlap = gti
    if(present(overlap)) then
      call gti_with_overlap%get_col_ptr("START", starts)
      call gti_with_overlap%get_col_ptr("STOP",  stops)
      starts = starts - overlap
      stops  = stops  + overlap
      nullify(stops)
      nullify(starts)
    end if

  end subroutine get_gti_with_overlap

  !.....................................................................
  function get_GTI_mask(times, GTI, status, do_sort, time_is_null, add_nearest_outer) result (mask)
    ! A time is taken to be within a GTI if it is >= the start time and < the stop time.
    !
    ! Presence of the optional argument 'add_nearest_outer' (it does not
    ! have to be .true.) will cause the algorithm to attempt to add
    ! 'nearest outer' time samples to the returned mask. This is
    ! designed to deal with cases in which the GTI intervals may be too
    ! few and/or narrow to contain any of the samples. For downstream
    ! tasks it may be undesirable to have no returned rows of extensions
    ! such as DEADCOR after GTI filtering.
    !
    implicit none

    real(r8), dimension(:), intent(in)           :: times
    type(tGTI_tab),         intent(in)           :: GTI
    integer,                intent(inout)        :: status
    logical,                intent(in), optional :: do_sort
    logical, dimension(size(times))&
      ,                     intent(in), optional :: time_is_null
    logical,                intent(in), optional :: add_nearest_outer
    logical, dimension(size(times)) :: mask

    character(*), parameter :: functionName = "get_GTI_mask"
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null()
    integer :: numGTIs,num_times,ti,gi,tti,ii,ti_first_non_NULL
    logical :: local_sort
    real(r8), dimension(:), allocatable :: times_copy
    logical, dimension(:), allocatable :: ltime_is_null
    integer, dimension(:), allocatable :: times_order,first_after_start,first_after_end

    if(status > 100 .or. status == s_errRep)&
  return

    if(.not.GTI%hasBeenSanityChecked) then
      call wrerr(functionName, s_err, "GTIs have not been sanity checked.")
      status = s_errRep
  return
    end if

    mask(:) = .false.
    numGTIs = GTI%getNumRows()
    num_times = size(times)
    if(numGTIs<1 .or. num_times<1)&
  return

    call alloc(first_after_start,numGTIs,status)
    call alloc(first_after_end,  numGTIs,status)
    first_after_start(:) = 0 ! default indicates
    first_after_end(:) = 0   ! not yet set.

    local_sort = .false.
    if(present(do_sort))&
      local_sort = do_sort

    if(local_sort) then
      times_copy = times
      call sort(times_copy, status, times_order)
      call dealloc(times_copy, status)
    else
      times_order = [(ii,ii=1,num_times)]
    end if

    call alloc(ltime_is_null, num_times, status)
    ltime_is_null(:) = .false.
    if(present(time_is_null)) ltime_is_null = time_is_null

    if(all(ltime_is_null)) then
      call dealloc(ltime_is_null, status)
  return
    end if

    call GTI%get_col_ptr("START", starts)
    call GTI%get_col_ptr("STOP",  stops)

    ! Find first non-null TIME value.
    !
    ti_first_non_NULL = num_times + 1 ! the test on all(time_is_null) should mean it never gets this value.
    do ti=1,num_times
      tti = times_order(ti)
      if(.not.ltime_is_null(tti)) then
        ti_first_non_NULL = ti
    exit
      end if
    end do

    ti = ti_first_non_NULL
    tti = times_order(ti)
    LGTI: do gi=1, numGTIs
      LTIMES: do
        if(times(tti) >= starts(gi) .and. times(tti) < stops(gi)) then
          ! TIME of this row is inside the GTI.

          mask(tti) = .true.

          if(first_after_start(gi)==0)&
            first_after_start(gi) = ti

        else if(times(tti) >= stops(gi)) then
          first_after_end(gi) = ti

    cycle LGTI
        end if

        if(ti >= num_times)&
    exit LGTI

        ti = ti + 1
        tti = times_order(ti)
      end do LTIMES
    end do LGTI

    nullify(stops)
    nullify(starts)
    call dealloc(ltime_is_null, status)

    if(present(add_nearest_outer)) then
      do gi=1, numGTIs
        if(first_after_start(gi)>0) then
          if(first_after_start(gi) > ti_first_non_NULL) then
            ti = first_after_start(gi) - 1
            tti = times_order(ti)
            mask(tti) = .true.
          end if
        else ! first_after_start(gi)==0. This may happen if the GTI is short enough to fall between two time samples.
          if(first_after_end(gi) > ti_first_non_NULL) then
            ti = first_after_end(gi) - 1
            tti = times_order(ti)
            mask(tti) = .true.
          end if
        end if
      end do
    end if

    call dealloc(times_order,status)
    call dealloc(first_after_start,status)
    call dealloc(first_after_end,  status)

  end function get_GTI_mask

  !.....................................................................
  subroutine setup_tmp_GTI(numRows, tmp_GTI, status)
    ! Inits, then sets useage flags for tmp_GTI and finally allocates the columns.

    implicit none

    integer,         intent(in)    :: numRows
    type(tGTI_tab),      intent(inout) :: tmp_GTI
    integer,         intent(inout) :: status

    character(*), parameter :: functionName = "setup_tmp_GTI"

    call tmp_GTI%init(iTemp, status)

    ! Internal use only.
    call tmp_GTI%set_all_useage(0,0,0,1)
    tmp_GTI%useageIsSet = .true.

    call tmp_GTI%alloc(status, numRows)

  end subroutine setup_tmp_GTI

  !.....................................................................
  subroutine dealloc_gtis(tables, status)
!**** haven't been able to figure out how to work the polymorphism with (de)alloc_bintables, so I've included this temporarily.
    implicit none

    type(tGTI_tab), allocatable, dimension(:) :: tables
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "dealloc_gtis"

    if(allocated(tables)) then
      deallocate(tables, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
    end if

  end subroutine dealloc_gtis

  !.....................................................................
  subroutine alloc_gtis(tables, n, status)
!**** haven't been able to figure out how to work the polymorphism with (de)alloc_bintables, so I've included this temporarily.
    implicit none

    type(tGTI_tab), allocatable, dimension(:) :: tables
    integer, intent(in) :: n
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "alloc_gtis"

    call dealloc_gtis(tables, status)
    allocate(tables(n), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

  end subroutine alloc_gtis

  !=====================================================================
  ! Methods of tGTIFile.

  !.....................................................................
  subroutine read_tGTIFile(this, status, fileName, tableName)
    implicit none

    class(tGTIFile) :: this
    integer,                intent(inout) :: status
    character(*), optional, intent(in)    :: fileName,tableName

    character(len=*),parameter :: functionName='read_tGTIFile'
    integer :: ierr
    character(buffsize) :: local_fileName

    call this%gti%init(iEXTERNAL, status, this%info)
    if(present(tableName)) then
      this%gti%info%tableName = tableName
    else
      this%gti%info%tableName = "STDGTI"
    end if

    call this%gti%ibset_all_useage(TCOL_READ_BIT)
    this%gti%useageIsSet = .true.

    if(present(fileName)) call this%open(FF_READONLY, status, filename=fileName)
    call this%gti%read(ierr)
    call this%info%getFileName(local_fileName)
    call this%gti%sanityCheck(status, "GTI read from file "//trim(local_fileName)//": ")

    if(ierr/=0)&
call wrstop(functionName, "Error in reading "//trim(this%gti%info%tableName))

    if(present(fileName)) call this%close(status)

  end subroutine read_tGTIFile

  !.....................................................................
  subroutine free_tGTIFile(this, status)
    class(tGTIFile) :: this
    integer, intent(inout) :: status

    character(len=*),parameter :: functionName='free_tGTIFile'

    call this%gti%destroy(status)

    call this%info%setFileName("")
    this%info%unitNo = 0 ! fitsio unit to access the file
    this%info%fileOpenedForRead = .false.
    this%info%fileOpenedForWrite = .false.
    this%info%isRenamed=.false.

  end subroutine free_tGTIFile
end module erolib_gtitab_mod
