! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_hk_mod

  use erolib_general_mod
  use erolib_str_mod
  use erolib_sort_mod
  use erolib_bintable_mod
  use erolib_gtitab_mod

  implicit none

logical, parameter, private :: doTest=.false.

  type, extends(tBinTable) :: tHK
    integer :: tel_id = -1
    integer :: ext_type = -1
    character(2) :: extNameRoot = "HK"
    character(1) :: typeChar = "x"
  contains
    procedure :: initHeader => initHeader_tHK
    procedure :: init => init_tHK
    procedure :: derive_frametime => derive_frametime_tHK
    procedure :: derive_columns => derive_columns_tHK
    procedure :: make_colnames_tel_specific => make_colnames_tel_specific_tHK
    procedure :: make_colnames_tel_unspecific => make_colnames_tel_unspecific_tHK
    procedure :: apply_gti => apply_gti_tHK
  end type tHK

  type, extends(tFitsFile) :: tHKFile
    type(tHK) :: hk
  end type tHKFile

  ! Extension indices:
  enum, bind(c)
    enumerator :: iHK1=1,iHK2,iHK3,iHK5
  end enum
  integer, parameter :: nHKTypes = 4

  character(4), parameter :: HK_appIDs(*) = ["HKx1","HKx2","HKx3","HKx5"]

  type, extends(tBinTable) :: tSlowSciHK
  contains
    procedure :: initHeader => initHeader_tSlowSciHK
  end type tSlowSciHK

  type, extends(tSlowSciHK) :: tSlowHK
  contains
    procedure :: init => init_tSlowHK
  end type tSlowHK

  type, extends(tFitsFile) :: tSlowHKFile
    type(tSlowHK) :: slowHK
  end type tSlowHKFile

  type, extends(tSlowSciHK) :: tSciHK
  contains
    procedure :: calc_frame_times => calc_frame_times_tSciHK
  end type tSciHK

  type, extends(tSciHK) :: tSciHK00
  contains
    procedure :: init => init_tSciHK00
  end type tSciHK00

  type, extends(tFitsFile) :: tSciHKFile00
    type(tSciHK00) :: sciHK
  end type tSciHKFile00

  type, extends(tSciHK) :: tSciHK05
  contains
    procedure :: init => init_tSciHK05
  end type tSciHK05

  type, extends(tFitsFile) :: tSciHKFile05
    type(tSciHK05) :: sciHK
  end type tSciHKFile05

  type, extends(tSciHK) :: tSciHK06
  contains
    procedure :: init => init_tSciHK06
  end type tSciHK06

  type, extends(tFitsFile) :: tSciHKFile06
    type(tSciHK06) :: sciHK
  end type tSciHKFile06

  type, extends(tSciHK) :: tSciHK8
  contains
    procedure :: init => init_tSciHK8
  end type tSciHK8

  type, extends(tFitsFile) :: tSciHKFile8
    type(tSciHK8) :: sciHK
  end type tSciHKFile8

  type, extends(tSciHK) :: tSciHK9
  contains
    procedure :: init => init_tSciHK9
  end type tSciHK9

  type, extends(tFitsFile) :: tSciHKFile9
    type(tSciHK9) :: sciHK
  end type tSciHKFile9

  type, extends(tSciHK) :: tSciHKA
  contains
    procedure :: init => init_tSciHKA
  end type tSciHKA

  type, extends(tFitsFile) :: tSciHKFileA
    type(tSciHKA) :: sciHK
  end type tSciHKFileA

!  integer, parameter :: i4_SHH_TEL_ID=1

  ! Column indices for tSciHK06:
!  enum, bind(c)
!    enumerator :: i4SciHKFrame=1,r8SciHKFrameTime,i4SciHKNumLost
!  end enum

  ! Column indices for tSlowHK:
!  enum, bind(c)
!    enumerator :: r8SlowHKRecordTime=1,i8SlowHKHkIDX23
!  end enum

  ! Column indices for tSciHK[8,9,A]:
!  enum, bind(c)
!    enumerator :: r8_ScA_RecordTime=1,i4_ScA_EXT_OTS,i4_ScA_INT_OTS,i4_ScA_SCTime,i2_ScA_SubSec
!  end enum

contains

  !.....................................................................
  subroutine initHeader_tHK(this, status)
    implicit none

    class(tHK), target :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "initHeader_tHK"

    this%header%tabInfo => this%info

    call this%header%initKeyword("D", "MJDREF")
    call this%header%initKeyword("J", "TEL_ID")
    call this%header%initKeyword("A", "APP_ID")
    call this%header%initKeywords(status)

    call this%header%check_keywords(status)

  end subroutine initHeader_tHK

  !.....................................................................
  subroutine init_tHK(this, extI, status, fileInfo)
    implicit none

    class(tHK) :: this
    integer,                           intent(in)    :: extI
    integer,                           intent(inout) :: status
    type(tFitsInfo), target, optional, intent(in)    :: fileInfo

    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "init_tHK"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
!    this%info%tableName = "***"
    this%ext_type = extI

    select case(extI)
    case(iHK1)
      this%typeChar = "1"
 
      call this%initColumn("D", "TIME",       "s")
      call this%initColumn("D", "RecordTime", "s")
      call this%initColumn("B", "PnSEQMODE")
      call this%initColumn("B", "PnSCTSTAT")
      call this%initColumn("B", "PnPROCMODE")
      call this%initColumn("J", "PnPROCOPTS")
      call this%initColumn("E", "PnTMFRAMES")
      call this%initColumn("E", "PnCCDFRMS")
      call this%initColumn("E", "PnEVTCTR")
      call this%initColumn("E", "PnMIPCTR")
      call this%initColumn("E", "PnFWSTEPCT")
      call this%initColumn("B", "PnFWSTAT")
      call this%initColumn("E", "PnCHOPPER")

      ! requested by KD
      call this%initColumn("E", "PnSPLT2NOI")
      call this%initColumn("E", "PnEVTQLIMT")

      ! derivative hk columns
      call this%initColumn("I", "DxPROCMODE")
      call this%initColumn("I", "DxPROCOPTS")
      call this%initColumn("I", "DxFWSTAT")

      ! frametime from record information
      call this%initColumn("D", "RFRAMETIME", "s")
      call this%initColumn("I", "QRFRAMETIME")

      ! columns required by telgti/telselect
      !    type(i2_tColumn) :: fwheel
      call this%initColumn("I", "DETSTAT")

    case(iHK2)
      this%typeChar = "2"

      call this%initColumn("D", "TIME",       "s")
      call this%initColumn("D", "RecordTime", "s")
      call this%initColumn("B", "FWHEEL")
      call this%initColumn("B", "CnFWPOS")
      call this%initColumn("B", "CnHVFWSW")
      call this%initColumn("E", "CnSTPVOLT")
      call this%initColumn("E", "PnMIPDELN")
      call this%initColumn("E", "PnMIPDELS")
      call this%initColumn("E", "PnMIPDELOW")
      call this%initColumn("B", "PnMIPIMGMX")
      call this%initColumn("I", "DxFWPOS")
      call this%initColumn("D", "RFRAMETIME", "s")
      call this%initColumn("I", "QRFRAMETIME")

    case(iHK3)
      this%typeChar = "3"

      call this%initColumn("D", "TIME",       "s")
      call this%initColumn("D", "RecordTime", "s")
      call this%initColumn("E", "PnFWDELAY")
      call this%initColumn("E", "PnENVTAB")
      call this%initColumn("E", "CnTCCDPT")
      call this%initColumn("I", "DxENVTAB")
      call this%initColumn("D", "RFRAMETIME", "s")
      call this%initColumn("I", "QRFRAMETIME")

    case(iHK5)
      this%typeChar = "5"

      call this%initColumn("D", "TIME",       "s")
      call this%initColumn("D", "RecordTime", "s")
      call this%initColumn("E", "PnMIPTHR")
      call this%initColumn("E", "PnPRIMTHR")
      call this%initColumn("E", "PnFWSTEPS")
      call this%initColumn("E", "PnSPLTCT")
      call this%initColumn("E", "PnEVTBMAX")
      call this%initColumn("E", "PnOUTBFREE")
      call this%initColumn("E", "PnMIPCTRPT")
      call this%initColumn("I", "DxMIPTHR")
      call this%initColumn("I", "DxPRIMTHR")
      call this%initColumn("I", "DxMIPCTRPT")
      call this%initColumn("D", "RFRAMETIME", "s")
      call this%initColumn("I", "QRFRAMETIME")

    case default
      errmsg = int2str(extI)
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(errmsg)//"'")
      status = s_errRep
  return
    end select

    call this%initColumns(status)

    call this%initHeader(status)

  end subroutine init_tHK

  !.....................................................................
  subroutine derive_frametime_tHK(this, nTimes, frameTimes, recordTimes, rGTI, status)
    implicit none
!**** 20200312 IMS: maybe we should rename these RFRAMETIME and QRFRAMETIME columns to RTIME because the conversion is now done using the pp_events TIME column, not FRAMETIME.

    class(tHK) :: this
    integer,                     intent(in)    :: nTimes
    real(r8), dimension(nTimes), intent(in)    :: frameTimes, recordTimes
    type(tGTI_tab),                  intent(in)    :: rGTI
    integer,                     intent(inout) :: status

    character(*), parameter :: functionName = "derive_frametime_tHK"
    integer :: ii, jj, kk, n_hkx, n_uni_rtime, n_evt, n_evt_complete, counter,numGTIs
    integer, allocatable, dimension(:) :: order, order_ftime
    real(r8) :: ftime_duration, rtime_duration, time_scaling
    real(r8), allocatable, dimension(:) :: rtime, uni_rtime, ftime_event, ftime
    integer(i2), allocatable, dimension(:) :: rec_type  ! 1 = hkx
    integer(i4), allocatable, dimension(:) :: order_hkx ! index of hk record
    logical, allocatable, dimension(:) :: within_gti, evt_within_gti
    real(r8), dimension(nTimes) :: temp_ft,temp_rt
    real(r8), dimension(:), pointer :: record_times=>null(),rframetimes=>null()
    integer(i2), dimension(:), pointer :: qual_flags=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    status = this%check(functionName)
    if(status/=0)&
  return

    if(nTimes<1) then
      call wrerr(functionName, c_err, "Can't convert HK times because there are no events.")
      status = c_err
  return
    end if

    numGTIs = rGTI%getNumRows()
    if(numGTIs<1) then
      call wrerr(functionName, c_err, "Can't convert HK times because there are no GTIs.")
      status = c_err
  return
    end if

    ! alternative to loop which is commented below
    ! sort events in frametime and afterwards in recordtime
    temp_ft = frameTimes
    temp_rt = recordTimes
    call sortTimes(nTimes, temp_ft, temp_rt)

    uni_rtime = temp_rt(unique(temp_rt, status))

    n_hkx = this%getNumRows()
    n_uni_rtime = size(uni_rtime)

    ! create combined array of recordtimes
    call alloc(rec_type, n_uni_rtime + n_hkx, status)
    call alloc(rtime, n_uni_rtime + n_hkx, status)
    call alloc(order_hkx, n_uni_rtime + n_hkx, status)
    call alloc(ftime, n_uni_rtime + n_hkx, status)

    rec_type(n_uni_rtime + 1: n_uni_rtime + n_hkx) = 1
    order_hkx(n_uni_rtime + 1 : n_uni_rtime + n_hkx) = [(ii,ii=1,n_hkx,1)]
    call this%get_col_ptr("RecordTime", record_times)
    rtime = [uni_rtime, record_times]
    nullify(record_times)

    counter = 1
    do ii = 1, nTimes
      if( abs(temp_rt(ii) - uni_rtime(counter)) < tiny(uni_rtime(counter))) then
        ftime(counter) = temp_ft(ii)
!    cycle
      else
        counter = counter + 1
        ftime(counter) = temp_ft(ii)
      end if
    end do
    ! write last frametime to last recordtime - this might not be correct
    ftime(n_uni_rtime) = temp_ft(nTimes)

    call sort(rtime, status, order)
    rec_type = rec_type(order)
    order_hkx = order_hkx(order)
    ftime = ftime(order)

    ! all entries from events_pp + hk extension is checked
    call alloc(within_gti, size(rtime), status)
    within_gti = get_GTI_mask(rtime, rGTI, status, do_sort=.false.)

    ! only check for the recordtimes in the events extension
    ! to get the maximum number of unique recordtimes as an indicator on how to calculate the frametime
    call alloc(evt_within_gti, size(uni_rtime), status)
    evt_within_gti = get_GTI_mask(uni_rtime, rGTI, status, do_sort=.true.)
    n_evt_complete = count(evt_within_gti)
    call dealloc(evt_within_gti, status)

    if(n_evt_complete<2) then
      call wrerr(functionName, c_err, "Can't interpolate HK times because there are less than 2 events within GTI.")
      status = c_err
  return
    end if

! 20200817 IMS: this is an ingenious algorithm, but it could lead to massive cantilevers. :) However only in regions far from GTI, so I suppose it is ok.

    call this%get_col_ptr("RFRAMETIME", rframetimes)
    call this%get_col_ptr("QRFRAMETIME", qual_flags)

    ! number of processed evts with complete records (all frametimes are available)
    ! complete record are always within gti
    n_evt = 0
    do ii = 1, size(rtime)

      ! check if its a record and within the recordtime gti
      if(rec_type(ii) == 0) then
        if(within_gti(ii)) n_evt = n_evt + 1
      end if

      ! check if record is hkx
      if(rec_type(ii) == 1) then
        ! no complete record yet -> get the next two complete record
        if(n_evt == 0) then
          call get_next_complete_event_record(rec_type, within_gti, ii, jj, status)
          call get_next_complete_event_record(rec_type, within_gti, jj, kk, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          rframetimes(order_hkx(ii)) = ftime(ii)
          qual_flags( order_hkx(ii)) = 3
        ! all complete records have been processed -> get the last two complete records
        else if(n_evt == n_evt_complete) then
          call get_last_complete_event_record(rec_type, within_gti, ii, kk, status)
          call get_last_complete_event_record(rec_type, within_gti, kk, jj, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          rframetimes(order_hkx(ii)) = ftime(ii)
          qual_flags( order_hkx(ii)) = 4
        ! position between two complete event records
        else
          call get_next_complete_event_record(rec_type, within_gti, ii, kk, status)
          call get_last_complete_event_record(rec_type, within_gti, ii, jj, status)
          time_scaling = (ftime(jj) - ftime(kk)) / (rtime(jj) - rtime(kk))
          rtime_duration = rtime(jj) - rtime(ii)
          ftime_duration = rtime_duration * time_scaling
          ftime(ii) = ftime(jj) - ftime_duration
          rframetimes(order_hkx(ii)) = ftime(ii)
          ! if a recordtime of HK is within the rGTI
          ! since a frametime / recordtime exists for the start / stop of a rGTI
          ! not much interpolation is needed and thus the quality should be higher
          if(within_gti(ii)) then
            qual_flags(order_hkx(ii)) = 1
          else
            qual_flags(order_hkx(ii)) = 2
          end if
        end if
      end if
    end do

    nullify(qual_flags)
    nullify(rframetimes)
    call dealloc(rtime, status)
    call dealloc(uni_rtime, status)
    call dealloc(ftime_event, status)
    call dealloc(ftime, status)
    call dealloc(order, status)
    call dealloc(order_ftime, status)
    call dealloc(rec_type, status)
    call dealloc(order_hkx, status)
    call dealloc(within_gti, status)

  end subroutine derive_frametime_tHK

  !.....................................................................
  subroutine sortTimes(nTimes, frameTimes, recordTimes)
    implicit none

    integer,                     intent(in)    :: nTimes
    real(r8), dimension(nTimes), intent(inout) :: frameTimes, recordTimes

    real(r8), dimension(nTimes) :: tmparray_r8
    integer, dimension(:), allocatable :: order
    integer :: status
    character(*), parameter :: function = "sortTimes"

    tmparray_r8 = frameTimes
    call sort_double(tmparray_r8, status, order)
    frameTimes  = frameTimes( order)
    recordTimes = recordTimes(order)

    tmparray_r8 = recordTimes
    call sort_double(tmparray_r8, status, order)
    frameTimes  = frameTimes( order)
    recordTimes = recordTimes(order)

    call dealloc(order, status)

  end subroutine sortTimes

  !.....................................................................
  subroutine get_next_complete_event_record(rec_types, within_gti, ii, jj, status)
!**** private?
    implicit none

    integer(i2), dimension(:), intent(in)    :: rec_types
    logical,     dimension(:), intent(in)    :: within_gti
    integer,                   intent(in)    :: ii
    integer,                   intent(out)   :: jj
    integer,                   intent(inout) :: status

    integer :: kk

    character(*), parameter :: functionName = "get_next_complete_event_record"

    if(status > 100 .or. status == s_errRep)&
  return

    do kk = ii+1, size(rec_types)
      ! check if type is events
      if(rec_types(kk) == 0) then
        ! check if recordtime is within gti
        if(within_gti(kk)) then
          jj = kk
  return
        end if
      end if
    end do

call wrstop(functionName, "Contact developer - algorithm should never reach this point.")

  end subroutine get_next_complete_event_record

  !.....................................................................
  subroutine get_last_complete_event_record(rec_types, within_gti, ii, jj, status)
    implicit none

    integer(i2), dimension(:), intent(in)    :: rec_types
    logical,     dimension(:), intent(in)    :: within_gti
    integer,                   intent(in)    :: ii
    integer,                   intent(out)   :: jj
    integer,                   intent(inout) :: status

    integer :: kk

    character(*), parameter :: functionName = "get_last_complete_event_record"

    if(status > 100 .or. status == s_errRep)&
  return
    do kk = ii-1, 1, -1
      ! check if type is events
      if(rec_types(kk) == 0) then
        ! check if recordtime is within gti
        if(within_gti(kk)) then
          jj = kk
  return
        end if
      end if
    end do

call wrstop(functionName, "contact developer - algorithm should never reach this point")

  end subroutine get_last_complete_event_record

  !.....................................................................
  subroutine make_colnames_tel_specific_tHK(this, tel_id, status)
    ! Most column names in the input HK files begin with Pn or Cn, where
    ! 'n' represents an integer in the range [1,7] which designates the
    ! telescope. However since the TM number may not be known when the
    ! tHK objects are initialized, the column names are specified with
    ! an actual 'n' in place of the TM number. The present routine
    ! finds all such instances and replaces the 'n' with the TM number.
    !
    ! Note that we supply tel_id rather than reading the header value
    ! because we will usually want to call this routine before we have
    ! read the extension. (Plus the extension may not contain tel_id as
    ! a header keyword.)

    implicit none

    class(tHK) :: this
    integer, intent(in)    :: tel_id
    integer, intent(inout) :: status

    integer :: ii,nColsUsed
    character(2) :: cnStr, pnStr
    character(1) :: tmStr
    character(*), parameter :: functionName = "make_colnames_tel_specific_tHK"

    status = this%check(functionName)
    if(status/=0)&
  return

    write(tmStr,'(i1)') tel_id
    cnStr = "Cn"
    pnStr = "Pn"

    nColsUsed = this%getNumCols()

    do ii=1,nColsUsed
      if(this%cols(ii)%nameStr(:2)==cnStr .or. this%cols(ii)%nameStr(:2)==pnStr) then
        this%cols(ii)%nameStr(2:2) = tmStr
      end if
    end do

  end subroutine make_colnames_tel_specific_tHK

  !.....................................................................
  subroutine make_colnames_tel_unspecific_tHK(this, status, tel_id)
    ! Most column names in the input HK files begin with Pn or Cn, where
    ! 'n' represents an integer in the range [1,7] which designates the
    ! telescope. The present routine replaces the integer with 'x'.

    implicit none

    class(tHK) :: this
    integer,           intent(inout) :: status
    integer, optional, intent(in)    :: tel_id

    integer :: ii,nColsUsed
    character(2) :: cStr,pStr,cnStr,pnStr
    character(*), parameter :: functionName = "make_colnames_tel_unspecific_tHK"

    status = this%check(functionName)
    if(status/=0)&
  return

    if(.not.this%header%containsData()) then
      call wrerr(functionName, s_err, "Header keyword values have neither been read nor set.")
      status = s_errRep
  return
    end if

    nColsUsed = this%getNumCols()

    if(present(tel_id)) then
      write(cStr,'(a,i1)') 'C', tel_id
      write(pStr,'(a,i1)') 'P', tel_id

      do ii=1,nColsUsed
        if(  this%cols(ii)%nameStr(:2)==cStr  .or. this%cols(ii)%nameStr(:2)==pStr) then
          this%cols(ii)%nameStr(2:2) = "x"
        end if
      end do

    else
      cnStr = "Cn"
      pnStr = "Pn"

      do ii=1,nColsUsed
        if(this%cols(ii)%nameStr(:2)==cnStr .or. this%cols(ii)%nameStr(:2)==pnStr) then
          this%cols(ii)%nameStr(2:2) = "x"
        end if
      end do
    end if

  end subroutine make_colnames_tel_unspecific_tHK

  !.....................................................................
  subroutine derive_columns_tHK(this, status)
    implicit none

    ! input / output param
    integer, intent(inout) :: status
    class(tHK) :: this

    character(*), parameter :: functionName = "derive_columns_tHK"
    character(buffsize) :: errmsg
    integer :: numRows
    real(r8), dimension(:), pointer :: times=>null(),record_times=>null()
    integer(i1), dimension(:), pointer :: procmodes=>null(),fwpos=>null(),fwheel=>null()
    integer(i2), dimension(:), pointer :: detstats=>null()
!*** get rid of the leading 'a' as soon as the globals become obsolete.
    integer :: i4H1_PxPROCOPTS,i2H1_DxPROCOPTS,i1H1_PxPROCMODE
    integer :: i2H1_DxPROCMODE,i1H1_PxFWSTAT,i2H1_DxFWSTAT
    integer :: i1H2_CxFWPOS,i2H2_DxFWPOS,r4H3_PxENVTAB,i2H3_DxENVTAB
    integer :: r4H5_PxMIPTHR,i2H5_DxMIPTHR,r4H5_PxPRIMTHR
    integer :: i2H5_DxPRIMTHR,r4H5_PxMIPCTRPT,i2H5_DxMIPCTRPT
    character(1) :: tmStr

    if(status > 100 .or. status == s_errRep)&
  return

    status = this%check(functionName)
    if(status/=0)&
  return

    numRows = this%getNumRows()
    if(numRows < 1)&
  return

    if(this%tel_id<1) then
      call wrerr(functionName, warn, "Columns should be telescope-specific at this point!")
      tmStr = "n"
    else
      write(tmStr, '(i1)') this%tel_id
    end if

    if(.not.this%col_is_allocated("TIME")) then
      call this%get_col_ptr("TIME", times)
      call this%get_col_ptr("RecordTime", record_times)
      times = record_times
      nullify(record_times)
      nullify(times)
    end if

    select case(this%ext_type)
    case(iHK1)
      call this%get_col_ptr("P"//tmStr//"PROCMODE", procmodes)
      call this%get_col_ptr("DETSTAT",    detstats)

      where(procmodes == 1 .or. procmodes == 4 .or. procmodes == 9 .or. procmodes == 10) detstats = 1

      nullify(procmodes)
      nullify(detstats)

      i4H1_PxPROCOPTS = this%getColNum("P"//tmStr//"PROCOPTS")
      i2H1_DxPROCOPTS = this%getColNum("DxPROCOPTS")
      i1H1_PxPROCMODE = this%getColNum("P"//tmStr//"PROCMODE")
      i2H1_DxPROCMODE = this%getColNum("DxPROCMODE")
      i1H1_PxFWSTAT   = this%getColNum("P"//tmStr//"FWSTAT")
      i2H1_DxFWSTAT   = this%getColNum("DxFWSTAT")

      call derive_diffs_column(this%cols(i4H1_PxPROCOPTS), this%cols(i2H1_DxPROCOPTS), status)
      call derive_diffs_column(this%cols(i1H1_PxPROCMODE), this%cols(i2H1_DxPROCMODE), status)
      call derive_diffs_column(this%cols(i1H1_PxFWSTAT  ), this%cols(i2H1_DxFWSTAT  ), status)

    case(iHK2)
      call this%get_col_ptr("C"//tmStr//"FWPOS", fwpos)
      call this%get_col_ptr("FWHEEL",  fwheel)

      fwheel = 127 ! default, means 'unknown'.
      where(fwpos == 14) fwheel = 0 ! closed
      where(fwpos == 12) fwheel = 1 ! open
      where(fwpos ==  8) fwheel = 2 ! filter
      where(fwpos == 13) fwheel = 3 ! calib
      where(fwpos == 15) fwheel = 4 ! moving

      nullify(fwheel)
      nullify(fwpos)

      i1H2_CxFWPOS = this%getColNum("C"//tmStr//"FWPOS")
      i2H2_DxFWPOS = this%getColNum("DxFWPOS")
      call derive_diffs_column(this%cols(i1H2_CxFWPOS), this%cols(i2H2_DxFWPOS), status)

    case(iHK3)
      r4H3_PxENVTAB = this%getColNum("P"//tmStr//"ENVTAB")
      i2H3_DxENVTAB = this%getColNum("DxENVTAB")
      call derive_diffs_column(this%cols(r4H3_PxENVTAB), this%cols(i2H3_DxENVTAB), status)

    case(iHK5)
      r4H5_PxMIPTHR   = this%getColNum("P"//tmStr//"MIPTHR")
      i2H5_DxMIPTHR   = this%getColNum("DxMIPTHR")
      r4H5_PxPRIMTHR  = this%getColNum("P"//tmStr//"PRIMTHR")
      i2H5_DxPRIMTHR  = this%getColNum("DxPRIMTHR")
      r4H5_PxMIPCTRPT = this%getColNum("P"//tmStr//"MIPCTRPT")
      i2H5_DxMIPCTRPT = this%getColNum("DxMIPCTRPT")
      call derive_diffs_column(this%cols(r4H5_PxMIPTHR),   this%cols(i2H5_DxMIPTHR), status)
      call derive_diffs_column(this%cols(r4H5_PxPRIMTHR),  this%cols(i2H5_DxPRIMTHR), status)
      call derive_diffs_column(this%cols(r4H5_PxMIPCTRPT), this%cols(i2H5_DxMIPCTRPT), status)

    case default
      errmsg = int2str(this%ext_type)
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(errmsg)//"'")
      status = s_errRep
  return
    end select

  end subroutine derive_columns_tHK

  !.....................................................................
  subroutine apply_gti_tHK(this, GTI, status)
    implicit none

    class(tHK) :: this
    type(tGTI_tab), intent(in)    :: GTI
    integer,    intent(inout) :: status

    character(*), parameter :: functionName = "apply_gti_tHK"
    type(tGTI_tab) :: tempGTI
    logical, dimension(:), allocatable :: mask
    integer :: numRows
    real(r8) :: mjdref_this,mjdref_temp
    real(r8), dimension(:), pointer :: times=>null()
    logical, dimension(:), pointer :: time_is_null=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    tempGTI = GTI
    ! internally correct the GTI

    call this%header%getKeywordValue("MJDREF",mjdref_this)
    call tempGTI%header%getKeywordValue("MJDREF",mjdref_temp)
    if(mjdref_this /= mjdref_temp) then
      call tempGTI%rebase_mjdref(mjdref_temp, mjdref_this)
      call tempGTI%header%setKeywordValue("MJDREF",mjdref_this)
    end if

    numRows = this%getNumRows()
    call alloc(mask, numRows, status)
    call this%get_col_ptr("TIME", times, time_is_null)
    mask = get_GTI_mask(times, tempGTI, status, .true., time_is_null, .true.)
    call this%pack(mask, status)
    nullify(times, time_is_null)
    call dealloc(mask, status)

    call tempGTI%destroy(status)

  end subroutine apply_gti_tHK

  !.....................................................................
  subroutine dealloc_hks(tables, status)
!**** haven't been able to figure out how to work the polymorphism with (de)alloc_bintables, so I've included this temporarily.
    implicit none

    type(tHK), allocatable, dimension(:) :: tables
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "dealloc_hks"

    if(allocated(tables)) then
      deallocate(tables, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
    end if

  end subroutine dealloc_hks

  !.....................................................................
  subroutine alloc_hks(tables, n, status)
!**** haven't been able to figure out how to work the polymorphism with (de)alloc_bintables, so I've included this temporarily.
    implicit none

    type(tHK), allocatable, dimension(:) :: tables
    integer, intent(in) :: n
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "alloc_hks"

    call dealloc_hks(tables, status)
    allocate(tables(n), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

  end subroutine alloc_hks

  !.....................................................................
  subroutine initHeader_tSlowSciHK(this, status)
    implicit none

    class(tSlowSciHK), target :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "initHeader_tSlowSciHK"

    this%header%tabInfo => this%info

    call this%header%initKeyword("J", "TEL_ID")
    call this%header%initKeywords(status)

    call this%header%check_keywords(status)

  end subroutine initHeader_tSlowSciHK

  !.....................................................................
  subroutine calc_frame_times_tSciHK(this, frame_times, flagged_good, status, use_ext)
    ! The calling routine should deallocate frame_times and flagged_good after use.
!**** 20200824 IMS: note that, although this method is also accessible for tSciHK06 objects, at present the tSciHK06 init does not recognize the required columns. Some work needed to close this trap.
    implicit none

    class(tSciHK) :: this
    real(r8), dimension(:), allocatable, intent(out) :: frame_times
    logical, dimension(:), allocatable, intent(out) :: flagged_good
    integer, intent(inout) :: status
    logical, optional, intent(in) :: use_ext

    character(*), parameter :: functionName = "calc_frame_times_tSciHK"
    integer(i4) :: bitmask
    logical :: local_use_ext
    integer(i4), dimension(:), allocatable :: ft_seconds
    integer :: numRows,ii
    logical, dimension(:), allocatable :: mask
    character(buffsize) :: msg
    integer(i4), dimension(:), pointer :: ext_ots=>null(),int_ots=>null(),sctimes=>null()
    integer(i2), dimension(:), pointer :: subsecs=>null()
    logical, dimension(:), pointer :: ext_ots_nulls=>null(),int_ots_nulls=>null(),sctime_nulls=>null(),subsec_nulls=>null()
    real(r8), dimension(:), pointer :: recordtimes=>null()

    ! Generate the mask 0xffff0000:
    !
    if(bit_size(bitmask)/=32)&
call wrstop(functionName, "Something is screwy here. Expected a 4-byte integer.")

    do ii=0,15
      bitmask = ibclr(bitmask,ii)
    end do
    do ii=16,31
      bitmask = ibset(bitmask,ii)
    end do

    if(present(use_ext)) then
      local_use_ext = use_ext
    else
      local_use_ext = .false.
    end if

    call this%get_col_ptr("EXT_OTS",    ext_ots, ext_ots_nulls)
    call this%get_col_ptr("INT_OTS",    int_ots, int_ots_nulls)
    call this%get_col_ptr("SCTime",     sctimes, sctime_nulls)
    call this%get_col_ptr("SubSec",     subsecs, subsec_nulls)
    call this%get_col_ptr("RecordTime", recordtimes)

    if(local_use_ext) then
      ft_seconds = iand(ext_ots, bitmask) + sctimes
    else
      ft_seconds = iand(int_ots, bitmask) + sctimes
    end if

    numRows = this%getNumRows()
    call alloc(flagged_good, numRows, status)

    flagged_good = .true.
    where(ext_ots_nulls) flagged_good = .false.
    where(int_ots_nulls) flagged_good = .false.
    where(subsec_nulls) flagged_good = .false.
    where(sctime_nulls ) flagged_good = .false.
    where(subsecs > 95_i2) flagged_good = .false.
    ! 20200818IMS: There may be the need to consider additional criteria.

    frame_times = real(ft_seconds, kind=r8) + real(subsecs, kind=r8)*ssCounterPeriod

    mask = (abs(frame_times - recordtimes)<0.9*real(2**16, kind=r8))
    if(any(flagged_good .and. .not. mask)) then
      msg = int2str(count(flagged_good .and. .not. mask))
      call wrlog(functionName, trim(msg)//" bad time wraps omitted.")
      flagged_good = flagged_good .and. mask
    end if

    nullify(ext_ots, ext_ots_nulls)
    nullify(int_ots, int_ots_nulls)
    nullify(sctimes, sctime_nulls)
    nullify(subsecs, subsec_nulls)
    nullify(recordtimes)

  end subroutine calc_frame_times_tSciHK

  !.....................................................................
  subroutine init_tSciHK00(this, status, fileInfo)
    implicit none

!*** 20211012 IMS: 00, 05 and 06 differ only in the extension name.

    class(tSciHK00) :: this
    integer, intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "init_tSciHK00"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    ! this%info%tableName should here be set by the calling routine, because there are 2 tables of identical form to read.
    this%info%tableName = "SciHKA00"

    call this%initColumn("J", "FRAME")
    call this%initColumn("D", "FRAMETIME", "s")
    call this%initColumn("J", "NumLost")

    call this%initColumns(status)

    call this%initHeader(status)

    ! Since we only have a single use in mind for this class, we can
    ! also set the %useageFlags here.
    !
    !                            - optional.
    !                           / - read.
    !                          / / - write.
    !                         / / / - no read, write, but used internally.
    call this%set_all_useage(0,1,0,0)

  end subroutine init_tSciHK00

  !.....................................................................
  subroutine init_tSciHK05(this, status, fileInfo)
    implicit none

    class(tSciHK05) :: this
    integer, intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "init_tSciHK05"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    ! this%info%tableName should here be set by the calling routine, because there are 2 tables of identical form to read.
    this%info%tableName = "SciHKA05"

    call this%initColumn("J", "FRAME")
    call this%initColumn("D", "FRAMETIME", "s")
    call this%initColumn("J", "NumLost")

    call this%initColumns(status)

    call this%initHeader(status)

    ! Since we only have a single use in mind for this class, we can
    ! also set the %useageFlags here.
    !
    !                            - optional.
    !                           / - read.
    !                          / / - write.
    !                         / / / - no read, write, but used internally.
    call this%set_all_useage(0,1,0,0)

  end subroutine init_tSciHK05

  !.....................................................................
  subroutine init_tSciHK06(this, status, fileInfo)
    implicit none

    class(tSciHK06) :: this
    integer, intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "init_tSciHK06"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    ! this%info%tableName should here be set by the calling routine, because there are 2 tables of identical form to read.
    this%info%tableName = "SciHKA06"

    call this%initColumn("J", "FRAME")
    call this%initColumn("D", "FRAMETIME", "s")
    call this%initColumn("J", "NumLost")

    call this%initColumns(status)

    call this%initHeader(status)

    ! Since we only have a single use in mind for this class, we can
    ! also set the %useageFlags here.
    !
    !                            - optional.
    !                           / - read.
    !                          / / - write.
    !                         / / / - no read, write, but used internally.
    call this%set_all_useage(0,1,0,0)

  end subroutine init_tSciHK06

  !.....................................................................
  subroutine init_tSlowHK(this, status, fileInfo)
    implicit none

    class(tSlowHK) :: this
    integer, intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "init_tSlowHK"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    ! this%info%tableName should be set by the calling routine, because it depends on TEL_ID.

    call this%initColumn("D", "RecordTime", "s")
    call this%initColumn("K", "HKIDX23")

    call this%initColumns(status)

    call this%initHeader(status)

    ! Since we only have a single use in mind for this class, we can
    ! also set the %useageFlags here.
    !
    !                                         - optional.
    !                                        / - read.
    !                                       / / - write.
    !                                      / / / - no read, write, but used internally.
    call this%set_col_useage("RecordTime",0,1,0,0)
    call this%set_col_useage("HKIDX23",   0,1,0,0)

    this%useageIsSet = .true.

  end subroutine init_tSlowHK

  !.....................................................................
  subroutine init_tSciHK8(this, status, fileInfo)
    implicit none

    class(tSciHK8) :: this
    integer, intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "init_tSciHK8"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    this%info%tableName = "SciHKA8"

    call this%initColumn("J", "FRAME")
    call this%initColumn("D", "RecordTime", "s")
    call this%initColumn("J", "EXT_OTS")
    call this%initColumn("J", "INT_OTS")
    call this%initColumn("J", "SCTime")
    call this%initColumn("I", "SubSec")
    call this%initColumn("J", "MIP-pixel")

    call this%initColumns(status)

    call this%initHeader(status)

    ! Since we only have a single use in mind for this class, we can
    ! also set the %useageFlags here.
    !
    call this%set_all_useage(0, 1, 0, 0) ! =read only.

  end subroutine init_tSciHK8

  !.....................................................................
  subroutine init_tSciHK9(this, status, fileInfo)
    implicit none

    class(tSciHK9) :: this
    integer, intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "init_tSciHK9"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    this%info%tableName = "SciHKA9"

    call this%initColumn("J", "FRAME")
    call this%initColumn("D", "RecordTime", "s")
    call this%initColumn("J", "EXT_OTS")
    call this%initColumn("J", "INT_OTS")
    call this%initColumn("J", "SCTime")
    call this%initColumn("I", "SubSec")
    call this%initColumn("J", "Sum")

    call this%initColumns(status)

    call this%initHeader(status)

    ! Since we only have a single use in mind for this class, we can
    ! also set the %useageFlags here.
    !
    call this%set_all_useage(0, 1, 0, 0) ! =read only.

  end subroutine init_tSciHK9

  !.....................................................................
  subroutine init_tSciHKA(this, status, fileInfo)
    implicit none

    class(tSciHKA) :: this
    integer, intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "init_tSciHKA"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    this%info%tableName = "SciHKAA"

    call this%initColumn("J", "FRAME")
    call this%initColumn("D", "RecordTime", "s")
    call this%initColumn("J", "EXT_OTS")
    call this%initColumn("J", "INT_OTS")
    call this%initColumn("J", "SCTime")
    call this%initColumn("I", "SubSec")
    call this%initColumn("I", "MIP-Pixel")
    call this%initColumn("I", "Sum")

    call this%initColumns(status)

    call this%initHeader(status)

    ! Since we only have a single use in mind for this class, we can
    ! also set the %useageFlags here.
    !
    call this%set_all_useage(0, 1, 0, 0) ! =read only.

  end subroutine init_tSciHKA

end module erolib_hk_mod
