! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_image2_mod
    use erolib_general_mod
    use erolib_ffits_mod
    use erolib_events_mod
    use erolib_wcs_mod
    implicit none

logical, private, parameter :: doTest=.false.

    private

        character(*), parameter :: image_ext = "IMAGE"

  type, public :: t_image_specs
    integer, dimension(2)  :: pixel_size_XY  ! was 'rebin'
    integer, dimension(2)  :: dimensions     ! was 'resolution'
    real(r8), dimension(2) :: centre_XY      ! was 'center_position'
    logical :: auto_centre    = .false.
    logical :: auto_dimension = .false.
  end type t_image_specs

    type tImage2D_i4
        type(tWcs) :: wcs
        integer(i4) :: nelements = 0
        integer(i4), dimension(2) :: rebin
        integer(i4), dimension(2) :: resolution ! naxis
        real(r8), dimension(2) :: center_position
        integer(i4), dimension(:,:), allocatable :: value
        character(buffsize) :: extname = ""
    contains
        !        procedure, non_overridable :: readFromFits => read_tImage2D_i4
        procedure, non_overridable :: write => write_tImage2D_i4
        procedure, non_overridable :: create_extension => create_extension_tImage2D_i4
        procedure, non_overridable :: free => free_tImage2D_i4
        procedure, non_overridable :: init => init_tImage2D_i4
        procedure, non_overridable :: create_from_events => create_from_events_tImage2D_i4
!        procedure, non_overridable :: wcs_check => wcs_check_tImage2D_i4
    end type tImage2D_i4

    public :: tImage2D_i4

    integer, parameter :: max_image_dimension = 18000 ! equivalent to 20 degrees of 4 arcsecond pixels.
    real, parameter :: auto_dim_pad = 1.01 ! 1% added to auto image dimensions.

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloc the content of tImage2D_i4 type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tImage2D_i4(this, n, status)
        implicit none

        integer, dimension(2), intent(in) :: n
        integer, intent(inout) :: status

        class(tImage2D_i4) :: this

        character(*), parameter :: function = "init_tImage2D_i4"

        call alloc(this%value, n(1), n(2), status)
        this%value = 0
        this%nelements = n(1) * n(2)
        this%resolution = n
        this%extname = ""

    end subroutine init_tImage2D_i4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free the content of tImage2D_i4 type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tImage2D_i4(this, status)
        implicit none

        integer, intent(inout) :: status

        class(tImage2D_i4) :: this

        character(*), parameter :: function = "free_tImage2D_i4"

        call dealloc(this%value, status)

    end subroutine free_tImage2D_i4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> write the image to a fits file
    !
    !
    !
    !> @param[in] this
    !> @param[out] unit
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tImage2D_i4(this, unit, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: unit

        class(tImage2D_i4) :: this

        character(*), parameter :: function = "write_tImage2D_i4"

        if(status > 100 .or. status == s_errRep) return

        ! move to the correct extension
        status = ignore
        call ffits_movnam_hdu(unit, IMAGE_HDU, this%extname, status)
        if(status /= ignore) then
            status = ok
            call this%create_extension(unit, status)
        else
            status = ok
        end if

        call ffits_write_img_2D(unit, this%resolution(1), this%resolution(1), this%resolution(2), this%value, status)

        call this%wcs%write(unit, IMAGE_HDU, status)

    end subroutine write_tImage2D_i4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create an image extension
    !
    !
    !
    !> @param[in] this
    !> @param[out] unit
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine create_extension_tImage2D_i4(this, unit, status, abshdu, extname)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: unit
        integer, intent(in), optional :: abshdu
        integer :: moved, hdutype

        class(tImage2D_i4) :: this

        character(*), parameter :: function = "create_extension_tImage2D_i4"
        character(*), optional, intent(in) :: extname

        if(status > 100 .or. status == s_errRep) return

        moved = false

        if(present(extname) .and. present(abshdu)) then
            call wrerr(function, warn, "abshdu and extname provided, abshdu has the priority")
            call ffits_movabs_hdu(unit, abshdu, hdutype, status)
            moved = true
        end if

        if(present(abshdu) .and. moved == false) then
            call ffits_movabs_hdu(unit, abshdu, hdutype, status)
            moved = true
        end if

        if(present(extname) .and. moved == false) then
            call ffits_movnam_hdu(unit, BINARY_TBL, extname, status)
!**** 2019-09-12 IMS: shouldn't this be IMAGE_HDU not BINARY_TBL?
            moved = true
        end if

        ! go to the first primary extension by default
        if(moved == false) call ffits_movabs_hdu(unit, 1, hdutype, status)

        ! create img extension
        call ffits_create_img(unit, 32, 2, this%resolution, status)
        call ffits_flush(unit, status)

    end subroutine create_extension_tImage2D_i4

  !.....................................................................
  subroutine create_from_events_tImage2D_i4(this, events, im_specs, status)
    implicit none

    class(tImage2D_i4)                 :: this
    type(tEvents),       intent(in)    :: events
    type(t_image_specs), intent(in)    :: im_specs
    integer,             intent(inout) :: status

    character(*), parameter :: functionName = "create_from_events_tImage2D_i4"
    integer :: numRows,ii,numGoodEvents
    logical, dimension(:), allocatable :: mask
    real(r8), dimension(:), allocatable :: xs, ys
    integer, dimension(:), allocatable :: xis, yis
    real(r8) :: meanX,meanY,xPixelsLo,xPixelsHi,yPixelsLo,yPixelsHi,xLo,xHi,yLo,yHi
    real(r8) :: dxPixelsMin,dxPixelsMax,dyPixelsMin,dyPixelsMax
    character(buffsize) :: meanXStr,meanYStr,ctrXStr,ctrYStr,num0Str,num1Str
    character(buffsize) :: xPixelsSpanStr,yPixelsSpanStr,xDimStr,yDimStr
    real(r8), dimension(:), pointer :: ras=>null(),decs=>null(),xs_ptr=>null(),ys_ptr=>null()
    logical, dimension(:), pointer :: ra_nulls=>null(),dec_nulls=>null()
    integer, dimension(2) :: local_im_dim
    real(r8), dimension(2) :: local_im_ctr

    if(status > 100 .or. status == s_errRep) return
    numRows = events%getNumRows()

    if(numRows< 1) then
      call wrerr(functionName, warn, "Image will be empty because zero events were found.")

      if(im_specs%auto_centre .or. im_specs%auto_dimension)&
        call wrerr(functionName, warn, "Image auto parameters cannot be used when there are no events.")

      this%rebin = im_specs%pixel_size_XY
      call this%init(im_specs%dimensions, status)
      this%extname = image_ext

      call this%wcs%adjustForImage_new(events%wcs, im_specs%pixel_size_XY, im_specs%dimensions, im_specs%centre_XY, status)

  return
    end if

    call events%get_col_ptr("RA",   ras,  ra_nulls)
    call events%get_col_ptr("DEC",  decs, dec_nulls)
    call events%get_col_ptr("X",    xs_ptr)
    call events%get_col_ptr("Y",    ys_ptr)

    call alloc(mask, numRows, status)
    mask = .true.
    where((ras==0.0 .and. decs==0.0) .or. ra_nulls .or. dec_nulls)&
      mask = .false.

!*** 20200915 IMS: complain about ra==0 .and. dec==0. GL should set them to null.

    numGoodEvents = count(mask)
    if(numRows > numGoodEvents) then
      num0Str = int2str(numRows-numGoodEvents)
      call wrerr(functionName, warn, trim(num0Str)&
        //" events found with RA==0 and DEC==0. These should be flagged or set to NULL!")
    end if

    xs = pack(xs_ptr, mask)
    ys = pack(ys_ptr, mask)
    call dealloc(mask, status)

    ! Report comparison between mean and range of X and Y values on the one
    ! hand, and what the user has specified for the image centre and bounds
    ! on the other.
    !
    meanX = sum(xs/float(numGoodEvents))
    meanY = sum(ys/float(numGoodEvents))
    xLo = minval(xs)
    xHi = maxval(xs)
    yLo = minval(ys)
    yHi = maxval(ys)
    xPixelsLo = xLo/im_specs%pixel_size_XY(1)
    xPixelsHi = xHi/im_specs%pixel_size_XY(1)
    yPixelsLo = yLo/im_specs%pixel_size_XY(2)
    yPixelsHi = yHi/im_specs%pixel_size_XY(2)

    if(im_specs%auto_centre) then
      local_im_ctr(1) = 0.5*(xLo + xHi)
      local_im_ctr(2) = 0.5*(yLo + yHi)
    else
      local_im_ctr = im_specs%centre_XY

      write(meanXStr, *) meanX
      write(meanYStr, *) meanY
      write(ctrXStr, *) im_specs%centre_XY(1)
      write(ctrYStr, *) im_specs%centre_XY(2)

      call wrchatty(functionName, "User has specified centre ("//trim(adjustl(ctrXStr))&
        //","//trim(adjustl(ctrYStr))//") whereas mean X/Y are ("&
        //trim(adjustl(meanXStr))//","//trim(adjustl(meanYStr))//").")
    end if

    dxPixelsMin = (xLo - local_im_ctr(1))/im_specs%pixel_size_XY(1)
    dxPixelsMax = (xHi - local_im_ctr(1))/im_specs%pixel_size_XY(1)
    dyPixelsMin = (yLo - local_im_ctr(2))/im_specs%pixel_size_XY(2)
    dyPixelsMax = (yHi - local_im_ctr(2))/im_specs%pixel_size_XY(2)

    if(im_specs%auto_dimension) then
      ! X:
      !
      local_im_dim(1) = 2*nint(auto_dim_pad*max(abs(dxPixelsMin), abs(dxPixelsMax)))

      ! Y:
      !
      local_im_dim(2) = 2*nint(auto_dim_pad*max(abs(dyPixelsMin), abs(dyPixelsMax)))

    else
      local_im_dim = im_specs%dimensions

      write(xPixelsSpanStr, *) xPixelsHi-xPixelsLo
      write(yPixelsSpanStr, *) yPixelsHi-yPixelsLo
      write(xDimStr, *) im_specs%dimensions(1)
      write(yDimStr, *) im_specs%dimensions(2)

      call wrchatty(functionName, "User has specified ("&
        //trim(adjustl(xDimStr))//","//trim(adjustl(yDimStr))&
        //") for image size in pixels, whereas event X/Y pixellised extents are ("&
        //trim(adjustl(xPixelsSpanStr))//","//trim(adjustl(yPixelsSpanStr))//").")
    end if

    if(local_im_dim(1) > max_image_dimension) then
      call wrerr(functionName, warn, "Image X dimension is larger than"&
        //" the maximum allowed. Truncating to maximum.")
      local_im_dim(1) = max_image_dimension
    end if
    if(local_im_dim(2) > max_image_dimension) then
      call wrerr(functionName, warn, "Image Y dimension is larger than"&
        //" the maximum allowed. Truncating to maximum.")
      local_im_dim(2) = max_image_dimension
    end if

    ! Set up image object:
    !
    this%rebin = im_specs%pixel_size_XY
    call this%init(local_im_dim, status)
    this%extname = image_ext

    call this%wcs%adjustForImage_new(events%wcs, im_specs%pixel_size_XY, local_im_dim, local_im_ctr, status)

    ! Now assign the events to image pixels.
    !
    allocate(xis(size(xs)))
    allocate(yis(size(ys)))

    ! Note that 'image pixel coordinates' (IPC) in the WCS scheme are such that the first pixel extends from 0.5 IPC to 1.5 IPC (see section 2.1.4 of Greisen & Calabretta, 'Representation of world coordinates in FITS', A&A, 395, 1061 (2002)). Fortran function nint() is therefore appropriate for deriving the pixel number from the IPC.
    !
    xis = nint(this%wcs%crpix1p + (xs - this%wcs%crval1p)/this%wcs%cdelt1p)
    yis = nint(this%wcs%crpix2p + (ys - this%wcs%crval2p)/this%wcs%cdelt2p)
    deallocate(xs)
    deallocate(ys)

    ! Check that at least some events are within the image bounds. If so, then check for spillover. If not, then warn, with a suggestion for a better image centre.
    !
    if(all(xis<1 .or. xis>local_im_dim(1)) .and. all(yis<1 .or. yis>local_im_dim(2))) then
      call wrerr(functionName, warn, "Zero events fall within the image bounds.")

    else if (any(xis<1 .or. xis>local_im_dim(1) .or. yis<1 .or. yis>local_im_dim(2))) then
      num0Str = int2str(numGoodEvents)
      num1Str = int2str(count(xis<1 .or. xis>local_im_dim(1) .or. yis<1 .or. yis>local_im_dim(2)))
      call wrerr(functionName, warn, trim(num1Str)//" out of "//trim(num0Str)&
        //" possible events fall outside the image bounds.")
    end if

    ! Add events to image:
    !
    do ii = 1, numGoodEvents
      if(   xis(ii) > 0 .and. xis(ii) <= local_im_dim(1)&
      .and. yis(ii) > 0 .and. yis(ii) <= local_im_dim(2)) then
         this%value(xis(ii), yis(ii)) = this%value(xis(ii), yis(ii)) + 1
      end if
    end do

    deallocate(xis)
    deallocate(yis)

    nullify(ras)
    nullify(decs)
    nullify(xs_ptr)
    nullify(ys_ptr)
    nullify(ra_nulls)
    nullify(dec_nulls)

  end subroutine create_from_events_tImage2D_i4

end module erolib_image2_mod
