! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_image_mod
  use erolib_general_mod
  use erolib_ffits_mod

  implicit none

  ! basic type of the catalog used for the cumulative list
  type, public :: tImage
    integer :: unit=0
    character(buffsize) :: filename=""
    character(buffsize) :: extname=""
    integer :: nelements=0
    integer, dimension(2) :: naxis=[0,0]
  contains
    procedure, non_overridable :: open => open_tImage
    procedure, non_overridable :: close => close_tImage
    procedure, non_overridable :: init => init_tImage
    procedure, non_overridable :: free => free_tImage
  end type tImage

  type, extends(tImage) :: tImage_i2
    integer :: bitpix = 16
    integer(i2), dimension(:,:), allocatable :: value
  contains
    procedure, non_overridable :: create => create_extension_tImage_i2
    procedure, non_overridable :: read => read_tImage_i2
    procedure, non_overridable :: write => write_tImage_i2
  end type tImage_i2

  type, extends(tImage) :: tImage_i4
    integer :: bitpix = 32
    integer(i4), dimension(:,:), allocatable :: value
  contains
    procedure, non_overridable :: create => create_extension_tImage_i4
    procedure, non_overridable :: read => read_tImage_i4
    procedure, non_overridable :: write => write_tImage_i4
  end type tImage_i4

  type, extends(tImage) :: tImage_r4
    integer :: bitpix = -32
    real(r4), dimension(:,:), allocatable :: value
  contains
    procedure, non_overridable :: create => create_extension_tImage_r4
    procedure, non_overridable :: read => read_tImage_r4
    procedure, non_overridable :: write => write_tImage_r4
  end type tImage_r4

contains


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> routine to open the Image File
  !
  !
  !
  !> @param[inout] this
  !> @param[in] filename, rwmode
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_tImage(this, filename, rwmode, status, delete, extname)
    implicit none

    integer, intent(inout) :: status
    class(tImage) :: this

    character(*), intent(in) :: filename
    character(*), intent(in), optional :: extname
    integer, intent(in) :: rwmode

    integer :: exists

    integer, intent(in), optional :: delete

    character(*), parameter :: function = "open_tImage"

    if(status > 100 .or. status == s_errRep) return

    this%filename = filename
    this%extname = ""
    if(present(extname)) this%extname = extname

    call ffits_get_iounit(this%unit, status)
    if(rwmode /= FF_CREATE) then
      call ffits_open_file(this%unit, filename, rwmode, status)
      if(status == ok) call wrlog(function, trim(this%filename) // " successfully opened")
    else
      call ffits_file_exists(filename, exists, status)
      if(present(delete) .and. exists == 1) then
        if(delete == true) call delete_file_fortran(filename, status)
      end if
      call ffits_create_file(this%unit, filename, fblocksize, status)
      if(status == ok) call wrlog(function, trim(this%filename) // " successfully created")
    end if

  end subroutine open_tImage

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close the tImage file
  !
  !
  !
  !> @param[inout] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_tImage(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tImage) :: this

    character(*), parameter :: function = "close_tImage"

    if(status > 100 .or. status == s_errRep) return

    call ffits_close_file(this%unit, status)
    call ffits_free_iounit(this%unit, status)

  end subroutine close_tImage

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> init the tImage class
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine init_tImage(this, naxis, status, extname)
    implicit none

    integer, intent(inout) :: status
    class(tImage) :: this
    integer, dimension(2), intent(in) :: naxis
    character(*), optional, intent(in) :: extname

    character(*), parameter :: functionName = "init_tImage"

    if(status > 100 .or. status == s_errRep) return

    this%naxis = naxis
    this%nelements = naxis(1) * naxis(2)
    if(present(extname)) this%extname = extname

    select type (this)
!**** 2019-09-11 IMS: why have the select type? Why not just alloc?
      type is (tImage_i2)
        call alloc(this%value, naxis(1), naxis(2), status)
      type is (tImage_i4)
        call alloc(this%value, naxis(1), naxis(2), status)
      type is (tImage_r4)
        call alloc(this%value, naxis(1), naxis(2), status)
      class default
call wrstop(functionName, "Unrecognized choice of image data type.")
    end select

  end subroutine init_tImage

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the tImage class
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine free_tImage(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tImage) :: this

    character(*), parameter :: functionName = "free_tImage"

    if(status > 100 .or. status == s_errRep) return

    this%nelements = 0
    this%naxis = [0,0]

    select type (this)
!**** 2019-09-11 IMS: why have the select type? Why not just alloc?
      type is (tImage_i2)
        call dealloc(this%value, status)
      type is (tImage_i4)
        call dealloc(this%value, status)
      type is (tImage_r4)
        call dealloc(this%value, status)
      class default
call wrstop(functionName, "Unrecognized choice of image data type.")
    end select

  end subroutine free_tImage

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create the tImage extension
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine create_extension_tImage_i2(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tImage_i2) :: this

    character(*), parameter :: function = "create_extension_tImage"

    if(status > 100 .or. status == s_errRep) return

    call ffits_create_img(this%unit, this%bitpix, size(this%naxis), this%naxis, status)
    if(this%extname /= "") call ffits_update_key_str(this%unit, "EXTNAME", this%extname, "", status)
    ! just to make sure the image extension was created
    call ffits_flush(this%unit, status)

  end subroutine create_extension_tImage_i2

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> create the tImage extension
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine create_extension_tImage_i4(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tImage_i4) :: this
    character(*), parameter :: function = "create_extension_tImage"

    if(status > 100 .or. status == s_errRep) return

    call ffits_create_img(this%unit, this%bitpix, size(this%naxis), this%naxis, status)
    if(this%extname /= "") call ffits_update_key_str(this%unit, "EXTNAME", this%extname, "", status)
    ! just to make sure the image extension was created
    call ffits_flush(this%unit, status)

  end subroutine create_extension_tImage_i4

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read tImage class
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine read_tImage_i2(this, status, filename, extname)
    implicit none

    integer, intent(inout) :: status
    class(tImage_i2) :: this
    character(*), optional, intent(in) :: filename
    character(*), optional, intent(in) :: extname

    character(*), parameter :: function = "read_tImage_i2"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) call this%open(filename, FF_READONLY, status)
    if(present(extname)) then
        call ffits_movnam_hdu(this%unit, IMAGE_HDU, extname, status)
    end if

    ! get naxis
    call FTGISZ(this%unit, 2, this%naxis, status)
    call this%init(this%naxis, status)
    call ffits_read_img_2D(this%unit, this%naxis(1), this%naxis(1), this%naxis(2), this%value, status)

  end subroutine read_tImage_i2

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read tImage class
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine read_tImage_i4(this, status, filename, extname)
    implicit none

    integer, intent(inout) :: status
    class(tImage_i4) :: this
    character(*), optional, intent(in) :: filename
    character(*), optional, intent(in) :: extname

    character(*), parameter :: function = "read_tImage_i4"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) call this%open(filename, FF_READONLY, status)
    if(present(extname)) then
        call ffits_movnam_hdu(this%unit, IMAGE_HDU, extname, status)
    end if

!*** 2019-09-11 IMS: no init as in i2?
! get naxis
call FTGISZ(this%unit, 2, this%naxis, status)
call this%init(this%naxis, status)
    call ffits_read_img_2D(this%unit, size(this%naxis), this%naxis(1), this%naxis(2), this%value, status)

  end subroutine read_tImage_i4

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write tImage class
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine write_tImage_i2(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tImage_i2) :: this

    character(*), parameter :: function = "write_tImage_i2"

    if(status > 100 .or. status == s_errRep) return

!    call ffits_write_img_2D(this%unit, size(this%naxis), this%naxis(1), this%naxis(2), this%value, status)
    call ffits_write_img_2D(this%unit, this%naxis(1), this%naxis(1), this%naxis(2), this%value, status)

  end subroutine write_tImage_i2

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> write tImage class
  !
  !
  !> @param[in] this
  !> @param[out] n
  !> @param[inout] status
  !> @param[in, optional]
  !---------------------------------------------------------------------------

  subroutine write_tImage_i4(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tImage_i4) :: this

    character(*), parameter :: function = "write_tImage_i4"

    if(status > 100 .or. status == s_errRep) return

!    call ffits_write_img_2D(this%unit, size(this%naxis), this%naxis(1), this%naxis(2), this%value, status)
    call ffits_write_img_2D(this%unit, this%naxis(1), this%naxis(1), this%naxis(2), this%value, status)

  end subroutine write_tImage_i4


  subroutine create_extension_tImage_r4(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tImage_r4) :: this
    character(*), parameter :: function = "create_extension_tImage"

    if(status > 100 .or. status == s_errRep) return

    call ffits_create_img(this%unit, this%bitpix, size(this%naxis), this%naxis, status)
    if(this%extname /= "") call ffits_update_key_str(this%unit, "EXTNAME", this%extname, "", status)
    ! just to make sure the image extension was created
    call ffits_flush(this%unit, status)

  end subroutine create_extension_tImage_r4

  subroutine read_tImage_r4(this, status, filename, extname)
    implicit none

    integer, intent(inout) :: status
    class(tImage_r4) :: this
    character(*), optional, intent(in) :: filename
    character(*), optional, intent(in) :: extname

    character(*), parameter :: function = "read_tImage_r4"

    if(status > 100 .or. status == s_errRep) return

    if(present(filename)) call this%open(filename, FF_READONLY, status)
    if(present(extname)) then
        call ffits_movnam_hdu(this%unit, IMAGE_HDU, extname, status)
    end if

!*** 2019-09-11 IMS: no init as in i2?
! get naxis
call FTGISZ(this%unit, 2, this%naxis, status)
call this%init(this%naxis, status)
    call ffits_read_img_2D(this%unit, size(this%naxis), this%naxis(1), this%naxis(2), this%value, status)

  end subroutine read_tImage_r4

  subroutine write_tImage_r4(this, status)
    implicit none

    integer, intent(inout) :: status
    class(tImage_r4) :: this

    character(*), parameter :: function = "write_tImage_r4"

    if(status > 100 .or. status == s_errRep) return

!    call ffits_write_img_2D(this%unit, size(this%naxis), this%naxis(1), this%naxis(2), this%value, status)
    call ffits_write_img_2D(this%unit, this%naxis(1), this%naxis(1), this%naxis(2), this%value, status)

  end subroutine write_tImage_r4

end module erolib_image_mod
