! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
!  subroutine calc_eigen_symm(array, evals, evecs)
!  subroutine solve_symm(symm_mat, b_vec, x_vec, ierr)
module erolib_lapack_mod

    use erolib_general_mod
!    use erolib_str_mod
    use erolib_math_mod
    implicit none

contains

  !.....................................................................
  subroutine calc_eigen_symm(symm_mat, evals, evecs, ierr)
    ! Calculates the eigenvalues and eigenvectors of a symmetric matrix.
    ! (Note that only the upper triangular half of the input matrix
    ! symm_mat is read.) The returned matrix 'evecs' contains the ith
    ! eigenvector in its ith column, whereas the ith element of 'evals'
    ! is the corresponding eigenvalue. The following identity is true:
    !
    !	symm_mat = matmul(matmul(evecs,diag_evals),transpose(evecs))
    !
    ! where 'diag_evals' is a square matrix containing the eigenvalues
    ! along its diagonal and zero elsewhere.
    !
    implicit none

    real, dimension(:,:),                  intent(in)  :: symm_mat
    real, dimension(size(symm_mat,1))&
      ,                                    intent(out) :: evals
    real, dimension(size(symm_mat,1),size(symm_mat,1))&
      ,                                    intent(out) :: evecs
    integer,                               intent(out) :: ierr

    character(*), parameter :: functionName = "calc_eigen_symm"
    integer, parameter :: lwmax = 1000
    integer :: size_i,size_work,size_iwork
    integer, dimension(:), allocatable :: iwork
    real, dimension(:), allocatable :: work
    real, dimension(size(symm_mat,1),size(symm_mat,1)) :: tmp_mat

    external :: SSYEVD

    size_i = size(symm_mat,1)
    if(size_i /= size(symm_mat,2))&
call wrstop(functionName,"Input matrix is not square.")

    tmp_mat = symm_mat

    ! Query the optimal workspace.
    !
    allocate(work(1))
    allocate(iwork(1))
    size_work = -1
    size_iwork = -1
    CALL SSYEVD('Vectors', 'Upper', size_i, tmp_mat, size_i, evals, work, size_work, iwork, size_iwork, ierr )
    size_work = MIN( lwmax, INT( work( 1 ) ) )
    size_iwork = MIN( lwmax, iwork( 1 ) )

    deallocate(work)
    deallocate(iwork)
    allocate(work(size_work))
    allocate(iwork(size_iwork))

    tmp_mat = symm_mat

    ! Solve eigenproblem.
    !
    call SSYEVD('Vectors', 'Upper', size_i, tmp_mat, size_i, evals, work, size_work, iwork, size_iwork, ierr )

    ! Check for convergence.
    !
    if(ierr/=0) THEN
      call wrerr(functionName, warn, "SSYEVD failed to compute eigenvalues.")
    end if

    evecs = tmp_mat

  end subroutine calc_eigen_symm

  !.....................................................................
  subroutine solve_symm(symm_mat, b_vec, x_vec, ierr)
    implicit none

    real, dimension(:,:),              intent(in)  :: symm_mat
    real, dimension(size(symm_mat,1)), intent(in)  :: b_vec
    real, dimension(size(symm_mat,1)), intent(out) :: x_vec
    integer,                           intent(out) :: ierr

    character(*), parameter :: functionName = "solve_symm"
    integer, parameter :: lwmax = 100
    integer :: size_i,size_work
    real, dimension(:), allocatable :: work
    real, dimension(size(symm_mat,1),size(symm_mat,1)) :: tmp_mat
    real, dimension(size(symm_mat,1)) :: ipiv
    real, dimension(size(symm_mat,1),1) :: tmp_b_vec

    external :: SSYSV

    size_i = size(symm_mat,1)
    if(size_i /= size(symm_mat,2))&
call wrstop(functionName,"Input matrix is not square.")

    tmp_mat = symm_mat
    tmp_b_vec(:,1) = b_vec

    ! Query the optimal workspace.
    !
    allocate(work(1))
    size_work = -1
    call SSYSV( 'Lower', size_i, 1, tmp_mat, size_i, ipiv, tmp_b_vec, size_i, work, size_work, ierr )
    size_work = min( lwmax, int( work( 1 ) ) )

    deallocate(work)
    allocate(work(size_work))

    tmp_mat = symm_mat
    tmp_b_vec(:,1) = b_vec

    ! Solve the equations A*X = B.
    !
    call SSYSV( 'Lower', size_i, 1, tmp_mat, size_i, ipiv, tmp_b_vec, size_i, work, size_work, ierr )

    ! Check for convergence.
    !
    if(ierr/=0) THEN
      call wrerr(functionName, warn, "SSYSV failed to solve the equations.")
    end if

    x_vec = tmp_b_vec(:,1)

  end subroutine solve_symm
end module erolib_lapack_mod
