! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_like_mod
  use param
  use erolib_general_mod
  use erolib_ranges_mod
  use erolib_str_mod
  use erolib_math_mod

  implicit none

logical, parameter, private :: doTest=.false.

  integer, parameter :: GAUSSIAN_PROB=0
  integer, parameter :: POISSON_PROB=1
  integer, parameter :: EVENT_PROB=2

  type :: t_grid_axes
    real, dimension(:), allocatable :: samples
  end type t_grid_axes

  type :: t_data
    logical, private :: is_initialized=.false.
    integer, private :: n_points=0
    integer, private :: n_dims=0
    real, dimension(:,:), allocatable, private :: locations
    real, dimension(:), allocatable, private :: real_values,uncerts
    integer, dimension(:), allocatable, private :: int_values
    type(tRange), dimension(:), allocatable :: limits
  contains
    procedure, non_overridable :: init => init_t_data
    procedure, non_overridable :: destroy => destroy_t_data
    procedure, non_overridable :: get_n_points => get_n_points_t_data
    procedure, non_overridable :: get_n_dims => get_n_dims_t_data
    procedure, non_overridable :: get_locations => get_locations_t_data
    procedure, non_overridable :: get_real_values => get_real_values_t_data
    procedure, non_overridable :: get_uncerts => get_uncerts_t_data
    procedure, non_overridable :: get_int_values => get_int_values_t_data
    procedure, non_overridable :: has_locations => has_locations_t_data
    procedure, non_overridable :: has_reals => has_reals_t_data
    procedure, non_overridable :: has_uncerts => has_uncerts_t_data
    procedure, non_overridable :: has_ints => has_ints_t_data
  end type t_data

  type, abstract :: t_model
    logical, private :: is_initialized=.false.
    integer, private :: n_params=0
    integer, private :: n_dims=0
    logical, private :: checks_done=.false.
    type(tRange), dimension(:), allocatable :: param_limits
  contains
    procedure, non_overridable :: init => init_t_model
    procedure, non_overridable :: destroy => destroy_t_model
    procedure, non_overridable :: check_params_size => check_params_size_t_model
    procedure, non_overridable :: params_are_within_limits => params_are_within_limits_t_model
    procedure, non_overridable :: check_array_sizes => check_array_sizes_t_model
    procedure, non_overridable :: get_n_params => get_n_params_t_model
    procedure, non_overridable :: get_n_dims => get_n_dims_t_model
    procedure, non_overridable :: get_bounds_on_line => get_bounds_on_line_t_model
    procedure, non_overridable :: fdiff_1st => fdiff_1st_t_model
    procedure, non_overridable :: fdiff_2nd => fdiff_2nd_t_model
    procedure, non_overridable :: fdiff_integral_1st => fdiff_integral_1st_t_model
    procedure, non_overridable :: fdiff_integral_2nd => fdiff_integral_2nd_t_model
    procedure, non_overridable :: simulate => simulate_t_model
    procedure, private         :: simulate_gaussian => simulate_gaussian_t_model
    procedure, private         :: simulate_poisson => simulate_poisson_t_model
    procedure, private         :: simulate_events => simulate_events_t_model
    procedure, non_overridable :: integral => integral_t_model
    procedure, non_overridable :: integral_approx => integral_approx_t_model

    procedure(calc_values_t_model),        deferred :: calc_values
    procedure(calc_1st_derivs_t_model),    deferred :: calc_1st_derivs
    procedure(calc_2nd_derivs_t_model),    deferred :: calc_2nd_derivs
    procedure(integral_indef_t_model),     deferred :: integral_indef
    procedure(inverse_integral_t_model),   deferred :: inverse_integral
    procedure(integral_1st_deriv_t_model), deferred :: integral_1st_deriv
    procedure(integral_2nd_deriv_t_model), deferred :: integral_2nd_deriv
  end type t_model

  type, abstract :: t_likelihood
    logical, private :: is_initialized=.false.
    integer, private :: n_params=0
    class(t_model), pointer :: model_obj=>null()
    type(t_data) :: data_obj
    real, dimension(:), allocatable :: param_scales
  contains
    procedure, non_overridable :: init => init_t_likelihood
    procedure, non_overridable :: params_are_within_limits => params_are_within_limits_t_likelihood
    procedure, non_overridable :: destroy => destroy_t_likelihood
    procedure, non_overridable :: fdiff_1st => fdiff_1st_t_likelihood
    procedure, non_overridable :: fdiff_2nd => fdiff_2nd_t_likelihood
    procedure, non_overridable :: calc_scale => calc_scale_t_likelihood
    procedure, non_overridable :: apply_scale => apply_scale_t_likelihood
    procedure, non_overridable :: unscale => unscale_t_likelihood

    procedure(calc_value_t_likelihood),     deferred :: calc_value
    procedure(calc_gradient_t_likelihood),  deferred :: calc_gradient
    procedure(calc_curvature_t_likelihood), deferred :: calc_curvature
  end type t_likelihood

!*** 20210402 IMS: Note that gfortran seg faults if one includes 'non_overridable' in the subclass methods below.

  type, extends(t_likelihood) :: t_gauss_like
  contains
!    procedure, non_overridable :: calc_value => calc_value_t_gauss_like
!    procedure, non_overridable :: calc_gradient => calc_gradient_t_gauss_like
!    procedure, non_overridable :: calc_curvature => calc_curvature_t_gauss_like
    procedure :: calc_value => calc_value_t_gauss_like
    procedure :: calc_gradient => calc_gradient_t_gauss_like
    procedure :: calc_curvature => calc_curvature_t_gauss_like
  end type t_gauss_like

  type, extends(t_likelihood) :: t_poisson_like
  contains
!    procedure, non_overridable :: calc_value => calc_value_t_poisson_like
!    procedure, non_overridable :: calc_gradient => calc_gradient_t_poisson_like
!    procedure, non_overridable :: calc_curvature => calc_curvature_t_poisson_like
    procedure :: calc_value => calc_value_t_poisson_like
    procedure :: calc_gradient => calc_gradient_t_poisson_like
    procedure :: calc_curvature => calc_curvature_t_poisson_like
  end type t_poisson_like

  type, extends(t_likelihood) :: t_event_like
  contains
!    procedure, non_overridable :: calc_value => calc_value_t_event_like
!    procedure, non_overridable :: calc_gradient => calc_gradient_t_event_like
!    procedure, non_overridable :: calc_curvature => calc_curvature_t_event_like
    procedure :: calc_value => calc_value_t_event_like
    procedure :: calc_gradient => calc_gradient_t_event_like
    procedure :: calc_curvature => calc_curvature_t_event_like
  end type t_event_like


  abstract interface
    function calc_values_t_model(this, locations, params) result(values)
      import t_model
      implicit none
      class(t_model)                   :: this
      real, dimension(:,:), intent(in) :: locations
      real, dimension(:),   intent(in) :: params
      real, dimension(:), allocatable  :: values
    end function
  end interface

  abstract interface
    function calc_1st_derivs_t_model(this, locations, params, ii) result(derivs_1st)
      import t_model
      implicit none
      class(t_model)                   :: this
      real, dimension(:,:), intent(in) :: locations
      real, dimension(:),   intent(in) :: params
      integer,              intent(in) :: ii
      real, dimension(:), allocatable  :: derivs_1st
    end function
  end interface

  abstract interface
    function calc_2nd_derivs_t_model(this, locations, params, ii, jj) result(derivs_2nd)
      import t_model
      implicit none
      class(t_model)                   :: this
      real, dimension(:,:), intent(in) :: locations
      real, dimension(:),   intent(in) :: params
      integer,              intent(in) :: ii,jj
      real, dimension(:), allocatable  :: derivs_2nd
    end function
  end interface

  abstract interface
    function integral_indef_t_model(this, params, xs) result(integrals)
      import t_model
      implicit none
      class(t_model)                  :: this
      real, dimension(:), intent(in)  :: params
      real, dimension(:), intent(in)  :: xs
      real, dimension(:), allocatable :: integrals
    end function
  end interface

  abstract interface
    function inverse_integral_t_model(this, params, integrals) result(xs)
      import t_model
      implicit none
      class(t_model)                  :: this
      real, dimension(:), intent(in)  :: params
      real, dimension(:), intent(in)  :: integrals
      real, dimension(:), allocatable :: xs
    end function
  end interface

  abstract interface
    real function integral_1st_deriv_t_model(this, data_limits, params, ii) result(deriv)
      use erolib_ranges_mod
      import t_model
      implicit none
      class(t_model)                         :: this
      type(tRange), dimension(:), intent(in) :: data_limits
      real, dimension(:),         intent(in) :: params
      integer,                    intent(in) :: ii
    end function
  end interface

  abstract interface
    real function integral_2nd_deriv_t_model(this, data_limits, params, ii, jj) result(deriv)
      use erolib_ranges_mod
      import t_model
      implicit none
      class(t_model)                         :: this
      type(tRange), dimension(:), intent(in) :: data_limits
      real, dimension(:),         intent(in) :: params
      integer,                    intent(in) :: ii,jj
    end function
  end interface

  abstract interface
    real function calc_value_t_likelihood(this, params) result(value)
      import t_likelihood
      implicit none
      class(t_likelihood) :: this
      real, dimension(this%n_params), intent(in)  :: params
    end function
  end interface

  abstract interface
    function calc_gradient_t_likelihood(this, params) result(gradient)
      import t_likelihood
      implicit none
      class(t_likelihood) :: this
      real, dimension(this%n_params), intent(in)  :: params
      real, dimension(this%n_params)              :: gradient
    end function
  end interface

  abstract interface
    function calc_curvature_t_likelihood(this, params) result (curv_mat)
      import t_likelihood
      implicit none
      class(t_likelihood)                         :: this
      real, dimension(this%n_params),   intent(in) :: params
      real, dimension(this%n_params,this%n_params) :: curv_mat
    end function
  end interface

contains

  !=====================================================================
  ! Methods of t_model:

  !.....................................................................
  subroutine init_t_model(this, n_params, n_dims, param_limits)
    implicit none
!*** better to call this 'set_param_limits' and get rid of the is_initialized flag.

    class(t_model) :: this
    integer, intent(in) :: n_params, n_dims
    type(tRange), dimension(n_params), optional, intent(in) :: param_limits

    character(*), parameter :: functionName = "init_t_model"
!    character(buffsize) :: iStr0,iStr1

    this%n_params = n_params
    this%n_dims = n_dims

    if(present(param_limits)) then
      this%param_limits = param_limits
    else
      allocate(this%param_limits(this%n_params)) ! All elements are tRange objects with, by default, undefined (i.e., open) bounds.
    end if

    this%is_initialized = .true.

  end subroutine init_t_model

  !.....................................................................
  subroutine destroy_t_model(this)
    implicit none

    class(t_model) :: this

    character(*), parameter :: functionName = "destroy_t_model"

    if(allocated(this%param_limits)) deallocate(this%param_limits)

  end subroutine destroy_t_model

  !.....................................................................
  subroutine check_params_size_t_model(this, params)
    implicit none

    class(t_model)                 :: this
    real, dimension(:), intent(in) :: params

    character(*), parameter :: functionName = "check_params_size_t_model"
    integer :: n_params,expected_n_params
    character(buffsize) :: iStr0,iStr1

    n_params = size(params)
    expected_n_params = this%get_n_params()

    if(n_params /= expected_n_params) then
      iStr0 = int2str(n_params)
      iStr1 = int2str(expected_n_params)
call wrstop(functionName, "Expected "//trim(iStr1)//" parameters, got "//trim(iStr0))
    end if

  end subroutine check_params_size_t_model

  !.....................................................................
  logical function params_are_within_limits_t_model(this, params) result(params_ok)
    implicit none

    class(t_model) :: this
    real, dimension(this%n_params), intent(in) :: params

    character(*), parameter :: functionName = "params_are_within_limits_t_model"
    integer :: ii
    character(buffsize) :: iStr

    params_ok = .false. ! default

    call this%check_params_size(params)

    do ii=1,this%n_params
      if(.not.this%param_limits(ii)%contains_value(real(params(ii),kind=r8))) then
        iStr = int2str(ii)
        call wrerr(functionName, warn, "Parameter "//trim(iStr)//" is outside limits.")
  return
      end if
    end do

    params_ok = .true.

  end function params_are_within_limits_t_model

  !.....................................................................
  subroutine check_array_sizes_t_model(this, n_dims, params)
    implicit none

    class(t_model)                 :: this
    integer,            intent(in) :: n_dims
    real, dimension(:), intent(in) :: params

    character(*), parameter :: functionName = "check_array_sizes_t_model"
    integer :: expected_n_dims
    character(buffsize) :: iStr0,iStr1

    expected_n_dims = this%get_n_dims()

    if(n_dims /= expected_n_dims) then
      iStr0 = int2str(n_dims)
      iStr1 = int2str(expected_n_dims)
call wrstop(functionName, "Expected "//trim(iStr1)//"-dimensional data, got "//trim(iStr0))
    end if

    call this%check_params_size(params)

  end subroutine check_array_sizes_t_model

  !.....................................................................
  integer function get_n_params_t_model(this) result(n_params)
    implicit none

    class(t_model) :: this

    character(*), parameter :: functionName = "get_n_params_t_model"

    n_params = this%n_params

  end function get_n_params_t_model

  !.....................................................................
  integer function get_n_dims_t_model(this) result(n_dims)
    implicit none

    class(t_model) :: this

    character(*), parameter :: functionName = "get_n_dims_t_model"

    n_dims = this%n_dims

  end function get_n_dims_t_model

  !.....................................................................
  subroutine get_bounds_on_line_t_model(this, p, delta_p, length, line_range, i_hi, i_lo)
    ! This subroutine draws a line through the parameter space in the direction of the input vector. It returns a tRange object with bounds corresponding to the first limit encountered (if any) in forward and reverse directions, as well as the indices of the respective parameters associated with those limits. An error is thrown if the starting position in parameter space falls outside the parameter limits.
    !
    implicit none

    class(t_model)                        :: this
    real, dimension(:),       intent(in)  :: p
    real, dimension(size(p)), intent(in)  :: delta_p
    real,                     intent(out) :: length
    type(tRange),             intent(out) :: line_range
    integer,                  intent(out) :: i_hi, i_lo

    character(*), parameter :: functionName = "get_bounds_on_line_t_model"
    integer :: n_params,ii
    logical :: first_bounds_set
    type(tRange) :: present_range
    real :: abs_frac_delta_this_axis

    if(.not.this%params_are_within_limits(p))&
call wrstop(functionName, "Parameters are outside model limits.")

    length = sqrt(sum(delta_p*delta_p))
    if(length<=0.0)&
call wrstop(functionName, "Delta has zero length.")

    n_params = size(p)
    first_bounds_set = .false.
    do ii=1, n_params
      if(delta_p(ii)==0.0)&
    cycle ! the vector has no component in the ii-axis direction, so it can't encounter any limits of the iith parameter.

      abs_frac_delta_this_axis = length/abs(delta_p(ii))

      ! Get the line bounds imposed by the present, iith parameter limits:
      !
      if(delta_p(ii)>0.0) then
        present_range%lo%isUndefined = this%param_limits(ii)%lo%isUndefined
        present_range%hi%isUndefined = this%param_limits(ii)%hi%isUndefined
        if(.not.this%param_limits(ii)%lo%isUndefined) then
          present_range%lo%value = (this%param_limits(ii)%lo%value - p(ii))*abs_frac_delta_this_axis
        end if
        if(.not.this%param_limits(ii)%hi%isUndefined) then
          present_range%hi%value = (this%param_limits(ii)%hi%value - p(ii))*abs_frac_delta_this_axis
        end if

      else ! delta_p(ii)<0.0
        present_range%hi%isUndefined = this%param_limits(ii)%lo%isUndefined
        present_range%lo%isUndefined = this%param_limits(ii)%hi%isUndefined
        if(.not.this%param_limits(ii)%lo%isUndefined) then
          present_range%hi%value = (p(ii) - this%param_limits(ii)%lo%value)*abs_frac_delta_this_axis
        end if
        if(.not.this%param_limits(ii)%hi%isUndefined) then
          present_range%lo%value = (p(ii) - this%param_limits(ii)%hi%value)*abs_frac_delta_this_axis
        end if
      end if

      if(.not.first_bounds_set) then
        i_hi = ii
        i_lo = ii
        line_range = present_range
        first_bounds_set = .true.

      else
        if(.not.present_range%lo%isUndefined) then ! that's the only way it can be more restrictive than the previous minimum.
          if(line_range%lo%isUndefined) then
            line_range%lo = present_range%lo
            i_lo = ii
          else if(line_range%lo%value < present_range%lo%value) then
            line_range%lo = present_range%lo
            i_lo = ii
          end if
        end if

        if(.not.present_range%hi%isUndefined) then ! that's the only way it can be more restrictive than the previous maximum.
          if(line_range%hi%isUndefined) then
            line_range%hi = present_range%hi
            i_hi = ii
          else if(line_range%hi%value > present_range%hi%value) then
            line_range%hi = present_range%hi
            i_hi = ii
          end if
        end if
      end if
    end do

  end subroutine get_bounds_on_line_t_model

  !.....................................................................
  function fdiff_1st_t_model(this, locations, params, deltas, ii) result(derivs_1st)
    ! Returns approximate derivative w.r.t. params(ii) via finite difference.
    !
    ! The calling routine should deallocate derivs_1st after use.
    implicit none

    class(t_model)                             :: this
    real, dimension(:,:),           intent(in) :: locations
    real, dimension(this%n_params), intent(in) :: params,deltas
    integer,                        intent(in) :: ii
    real, dimension(:), allocatable            :: derivs_1st

    character(*), parameter :: functionName = "fdiff_1st_t_model"
    real, dimension(this%n_params) :: local_params
    real, dimension(:), allocatable :: values_lo,values_hi

    call this%check_params_size(params)

    local_params = params
    local_params(ii) = params(ii) - deltas(ii)
    values_lo  = this%calc_values(locations, local_params)
    local_params(ii) = params(ii) + deltas(ii)
    values_hi  = this%calc_values(locations, local_params)
    derivs_1st = 0.5*(values_hi - values_lo)/deltas(ii)

    deallocate(values_lo,values_hi)

  end function fdiff_1st_t_model

  !.....................................................................
  function fdiff_2nd_t_model(this, locations, params, deltas, ii, jj) result(derivs_2nd)
    ! Returns approximate derivative w.r.t. params(ii) via finite difference.
    !
    ! The calling routine should deallocate derivs_2nd after use.
    implicit none

    class(t_model) :: this
    real, dimension(:,:),           intent(in) :: locations
    real, dimension(this%n_params), intent(in) :: params,deltas
    integer,                        intent(in) :: ii,jj
    real, dimension(:), allocatable            :: derivs_2nd
!*** deallocate them

    character(*), parameter :: functionName = "fdiff_2nd_t_model"
    real, dimension(this%n_params) :: local_params
    real, dimension(:), allocatable :: values_lo,values_hi,values_mid

    call this%check_params_size(params)

    if(ii==jj) then
      local_params = params
      values_mid = this%calc_values(locations, local_params)
      local_params(ii) = params(ii) - deltas(ii)
      values_lo  = this%calc_values(locations, local_params)
      local_params(ii) = params(ii) + deltas(ii)
      values_hi  = this%calc_values(locations, local_params)
      derivs_2nd = (values_hi + values_lo - 2.0*values_mid)/deltas(ii)/deltas(ii)
      deallocate(values_mid)
    else
      local_params = params
      local_params(jj) = params(jj) - deltas(jj)
      values_lo = this%fdiff_1st(locations, local_params, deltas, ii)
      local_params(jj) = params(jj) + deltas(jj)
      values_hi = this%fdiff_1st(locations, local_params, deltas, ii)
      derivs_2nd = 0.5*(values_hi - values_lo)/deltas(jj)
    end if

    deallocate(values_lo,values_hi)

  end function fdiff_2nd_t_model

  !.....................................................................
  function fdiff_integral_1st_t_model(this, data_limits, params, deltas, ii) result(deriv_1st)
    ! Returns approximate derivative w.r.t. params(ii) via finite difference.
    implicit none

    class(t_model)                             :: this
    type(tRange), dimension(:),     intent(in) :: data_limits
    real, dimension(this%n_params), intent(in) :: params,deltas
    integer,                        intent(in) :: ii
    real                                       :: deriv_1st

    character(*), parameter :: functionName = "fdiff_integral_1st_t_model"
    real, dimension(this%n_params) :: local_params
    real :: value_lo,value_hi

    call this%check_params_size(params)

    local_params = params
    local_params(ii) = params(ii) - deltas(ii)
    value_lo  = this%integral(data_limits, local_params)
    local_params(ii) = params(ii) + deltas(ii)
    value_hi  = this%integral(data_limits, local_params)
    deriv_1st = 0.5*(value_hi - value_lo)/deltas(ii)

  end function fdiff_integral_1st_t_model

  !.....................................................................
  function fdiff_integral_2nd_t_model(this, data_limits, params, deltas, ii, jj) result(deriv_2nd)
    ! Returns approximate derivative w.r.t. params(ii) via finite difference.
    implicit none

    class(t_model) :: this
    type(tRange), dimension(:),     intent(in) :: data_limits
    real, dimension(this%n_params), intent(in) :: params,deltas
    integer,                        intent(in) :: ii,jj
    real                                       :: deriv_2nd

    character(*), parameter :: functionName = "fdiff_integral_2nd_t_model"
    real, dimension(this%n_params) :: local_params
    real :: value_lo,value_hi,value_mid

    call this%check_params_size(params)

    if(ii==jj) then
      local_params = params
      value_mid = this%integral(data_limits, local_params)
      local_params(ii) = params(ii) - deltas(ii)
      value_lo  = this%integral(data_limits, local_params)
      local_params(ii) = params(ii) + deltas(ii)
      value_hi  = this%integral(data_limits, local_params)
      deriv_2nd = (value_hi + value_lo - 2.0*value_mid)/deltas(ii)/deltas(ii)
    else
      local_params = params
      local_params(jj) = params(jj) - deltas(jj)
      value_lo = this%integral_1st_deriv(data_limits, local_params, ii)
      local_params(jj) = params(jj) + deltas(jj)
      value_hi = this%integral_1st_deriv(data_limits, local_params, ii)
      deriv_2nd = 0.5*(value_hi - value_lo)/deltas(jj)
    end if

  end function fdiff_integral_2nd_t_model

  !.....................................................................
  function simulate_gaussian_t_model(this, in_data_obj, params) result(out_data_obj)
    ! This generates real random values at each of the locations in in_data_obj. The values follow the Gaussian probability distribution, with respective means calculated from the model and sigma values being taken from the uncertainties stored in in_data_obj.
    implicit none

    class(t_model)                 :: this
    type(t_data),       intent(in) :: in_data_obj
    real, dimension(:), intent(in) :: params
    type(t_data)                   :: out_data_obj

    character(*), parameter :: functionName = "simulate_gaussian_t_model"
    real, dimension(:), allocatable :: sigmas,means,gauss_randoms
    real, dimension(:,:), allocatable :: locations

    if(.not.this%params_are_within_limits(params))&
call wrstop(functionName, "Parameters are outside model limits.")

    ! Since we are in the same module, we could just access these private attributes of the t_data struct. However we will use the calls so as to fail with an orderly error if the respective vectors are not allocated.
    sigmas = in_data_obj%get_uncerts()
    locations = in_data_obj%get_locations()

    means = this%calc_values(locations, params)

    allocate(gauss_randoms(size(means)))
    call getGaussianRandoms(gauss_randoms) ! returned numbers have zero mean and unity std dev.

    out_data_obj = in_data_obj
    out_data_obj%real_values = gauss_randoms*sigmas + means

    deallocate(sigmas,means,gauss_randoms)
    deallocate(locations)

  end function simulate_gaussian_t_model

  !.....................................................................
  function simulate_poisson_t_model(this, in_data_obj, params) result(out_data_obj)
    ! This generates integer random values at each of the locations in in_data_obj. The values follow the Poisson probability distribution, with respective expectation values calculated from the model.
    implicit none

    class(t_model)                 :: this
    type(t_data),       intent(in) :: in_data_obj
    real, dimension(:), intent(in) :: params
    type(t_data)                   :: out_data_obj

    character(*), parameter :: functionName = "simulate_poisson_t_model"
    real, dimension(:), allocatable :: means
    real, dimension(:,:), allocatable :: locations
    integer :: n_points,ii

    if(.not.this%params_are_within_limits(params))&
call wrstop(functionName, "Parameters are outside model limits.")

    ! Since we are in the same module, we could just access these private attributes of the t_data struct. However we will use the calls so as to fail with an orderly error if the respective vectors are not allocated.
    locations = in_data_obj%get_locations()

    means = this%calc_values(locations, params)

    n_points = size(means)
    out_data_obj = in_data_obj
    if(.not.allocated(out_data_obj%int_values)) allocate(out_data_obj%int_values(n_points))
    do ii=1, n_points
      out_data_obj%int_values(ii) = poissonRandom(means(ii))
    end do

    deallocate(means)
    deallocate(locations)

  end function simulate_poisson_t_model

  !.....................................................................
  function simulate_events_t_model(this, in_data_obj, params) result(out_data_obj)
    ! This returns a vector of locations which follow a probability
    ! distribution prescribed by the model function f(x). The general
    ! prescription for obtaining such random numbers is first to obtain
    ! the indefinite integral I(x) of f. This is scaled according to
    ! the range [x0,x1] within which we want the random samples to give
    !
    !	     I(x) - I(x0)
    !	y = ---------------.
    !	     I(x1) - I(x0)
    !
    ! If we then generate evenly-distributed randoms y between 0 and 1,
    ! inverting the expression will provide randoms x with a
    ! probability density within [x0,x1] given by f(x):
    !
    !	x = I^{-1}[I(x0)*(1-y) + I(x1)*y].
    !
    ! Note that f(x) must be >= 0 everywhere in [x0,x1] (so that I(x) is
    ! monotonic in this interval), and (I(x1) - I(x0)) must be >0.
    !
    implicit none

!*** at present only for 1D

    class(t_model)                 :: this
    type(t_data),       intent(in) :: in_data_obj
    real, dimension(:), intent(in) :: params
    type(t_data)                   :: out_data_obj

    character(*), parameter :: functionName = "simulate_events_t_model"
    integer, parameter :: n_dims=1
    integer :: n_points
    real, dimension(1) :: x0s,x1s,integrals_lo,integrals_hi
    real, dimension(:), allocatable :: randoms,args
    real, dimension(:,:), allocatable :: locations

    if(.not.this%params_are_within_limits(params))&
call wrstop(functionName, "Parameters are outside model limits.")

    if(in_data_obj%limits(1)%lo%isUndefined .or. in_data_obj%limits(1)%hi%isUndefined)&
call wrstop(functionName, "Input data object must have definite limits.")

    x0s(1) = real(in_data_obj%limits(1)%lo%value) ! Just for
    x1s(1) = real(in_data_obj%limits(1)%hi%value) ! brevity.

    n_points = in_data_obj%get_n_points()
    allocate(randoms(n_points))
    allocate(locations(n_points,1))
    call getRandomNumbers(randoms)

    integrals_lo = this%integral_indef(params, x0s)
    integrals_hi = this%integral_indef(params, x1s)

    args = integrals_lo(1)*(1.0 - randoms) + integrals_hi(1)*randoms
    locations(:,1) = this%inverse_integral(params, args)

    call out_data_obj%init(n_points, n_dims, locations, limits=in_data_obj%limits)

    deallocate(randoms,args)
    deallocate(locations)

  end function simulate_events_t_model

  !.....................................................................
  function simulate_t_model(this, in_data_obj, type_i, params) result(out_data_obj)
    implicit none

    class(t_model)                 :: this
    type(t_data),       intent(in) :: in_data_obj
    integer,            intent(in) :: type_i
    real, dimension(:), intent(in) :: params
    type(t_data)                   :: out_data_obj

    character(*), parameter :: functionName = "simulate_t_model"

    if(.not.this%params_are_within_limits(params))&
call wrstop(functionName, "Parameters are outside model limits.")

    if(.not.in_data_obj%is_initialized)&
call wrstop(functionName, "Input data object is not initialized.")

    select case(type_i)
    case(GAUSSIAN_PROB)
      out_data_obj = this%simulate_gaussian(in_data_obj, params)
    case(POISSON_PROB)
      out_data_obj = this%simulate_poisson(in_data_obj, params)
    case(EVENT_PROB)
      out_data_obj = this%simulate_events(in_data_obj, params)
    case default
call wrstop(functionName, "Unrecognized simulation type.")
    end select

  end function simulate_t_model

  !.....................................................................
  subroutine increment(n_dims, max_indices, indices)
    implicit none

    integer,                    intent(in)    :: n_dims
    integer, dimension(n_dims), intent(in)    :: max_indices
    integer, dimension(n_dims), intent(inout) :: indices

    integer :: di

    do di=1,n_dims
      if(indices(di)<max_indices(di)) then
        indices(di) = indices(di) + 1
    exit
      else
        indices(di) = 1
      end if
    end do

  end subroutine increment

  !.....................................................................
  real function integral_t_model(this, data_limits, params) result(integral)
!*** at present only for 1D
    implicit none

    class(t_model)                             :: this
    type(tRange), dimension(1),     intent(in) :: data_limits ! range in each dimension
    real, dimension(:),             intent(in) :: params

    character(*), parameter :: functionName = "integral_t_model"
    real, dimension(1) :: x_los,x_his,dummy

    x_los(1) = real(data_limits(1)%lo%value)
    x_his(1) = real(data_limits(1)%hi%value)
    dummy = this%integral_indef(params, x_his) - this%integral_indef(params, x_los)
    integral = dummy(1)

  end function integral_t_model

  !.....................................................................
  real function integral_approx_t_model(this, data_limits, params&
    , n_samples_per_dim) result(integral)

    implicit none

    class(t_model)                             :: this
    type(tRange), dimension(:),     intent(in) :: data_limits ! range in each dimension
    real, dimension(:),             intent(in) :: params
    integer, dimension(size(data_limits))&
      ,                             intent(in) :: n_samples_per_dim ! number of samples per dimension, to be distributed evenly across the range.

    character(*), parameter :: functionName = "integral_approx_t_model"
    integer :: n_dims,n_values,ii,di
    real, dimension(:,:), allocatable :: locations
    type(t_grid_axes), dimension(:), allocatable :: grid_axes
    real, dimension(:), allocatable :: fracs,values
    integer, dimension(:), allocatable :: indices

    n_dims = size(data_limits)
    n_values = product(n_samples_per_dim)
    allocate(locations(n_values,n_dims))

    allocate(grid_axes(n_dims))
    do di=1,n_dims
      fracs = [(real(ii-0.5)/real(n_samples_per_dim(di)),ii=1,n_samples_per_dim(di))]
      allocate(grid_axes(di)%samples(n_samples_per_dim(di)))
      grid_axes(di)%samples = real(data_limits(di)%lo%value)*(1.0-fracs)&
                            + real(data_limits(di)%hi%value)*     fracs
    end do

    allocate(indices(n_dims))
    indices(:) = 1
    do ii=1,n_values
      call increment(n_dims, n_samples_per_dim, indices)
      do di=1,n_dims
        locations(ii,di) = grid_axes(di)%samples(indices(di))
      end do
    end do

    values = this%calc_values(locations, params)
    integral = sum(values)/real(n_values)
    do di=1,n_dims
      integral = integral*real(data_limits(di)%hi%value - data_limits(di)%lo%value)
    end do

  end function integral_approx_t_model

  !=====================================================================
  ! Methods of t_data:

  !.....................................................................
  subroutine init_t_data(this, n_points, n_dims, locations, real_values, uncerts, int_values, limits)
    implicit none

    class(t_data) :: this
    integer,                           intent(in) :: n_points,n_dims
    real, dimension(n_points,n_dims)&
      ,                      optional, intent(in) :: locations
    real, dimension(n_points)&
      ,                      optional, intent(in) :: real_values
    real, dimension(n_points)&
      ,                      optional, intent(in) :: uncerts
    integer, dimension(n_points)&
      ,                      optional, intent(in) :: int_values
    type(tRange), dimension(n_dims)&
      ,                      optional, intent(in) :: limits

    character(*), parameter :: functionName = "init_t_data"

    this%n_points = n_points
    this%n_dims   = n_dims

    if(present(locations)) then
      this%locations = locations
    end if

    ! For Gaussian or similar stats:
    if(present(real_values)) then
      this%real_values = real_values
    end if ! ignore uncerts otherwise.

    if(present(uncerts)) then
      this%uncerts = uncerts
    end if ! ignore uncerts otherwise.

    ! For Poisson stats:
    if(present(int_values)) then
      this%int_values = int_values
    end if

    if(present(limits)) then
      this%limits = limits
    else
      allocate(this%limits(this%n_dims)) ! All elements are tRange objects with, by default, undefined (i.e., open) bounds.
    end if

    this%is_initialized = .true.

  end subroutine init_t_data

  !.....................................................................
  subroutine destroy_t_data(this)
    implicit none

    class(t_data) :: this

    character(*), parameter :: functionName = "destroy_t_data"

    if(allocated(this%locations)) deallocate(this%locations)
    if(allocated(this%real_values)) deallocate(this%real_values)
    if(allocated(this%uncerts)) deallocate(this%uncerts)
    if(allocated(this%int_values)) deallocate(this%int_values)
    if(allocated(this%limits)) deallocate(this%limits)

  end subroutine destroy_t_data

  !.....................................................................
  integer function get_n_points_t_data(this) result(n_points)
    implicit none

    class(t_data) :: this

    character(*), parameter :: functionName = "get_n_points_t_data"

    n_points = this%n_points

  end function get_n_points_t_data

  !.....................................................................
  integer function get_n_dims_t_data(this) result(n_dims)
    implicit none

    class(t_data) :: this

    character(*), parameter :: functionName = "get_n_dims_t_data"

    n_dims = this%n_dims

  end function get_n_dims_t_data

  !.....................................................................
  function get_locations_t_data(this) result(locations)
    ! The calling routine should deallocate locations after use.
    implicit none

    class(t_data) :: this
    real, dimension(:,:), allocatable :: locations

    character(*), parameter :: functionName = "get_locations_t_data"

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    if(.not.allocated(this%locations))&
call wrstop(functionName, "There are no location values in the data object.")

    locations = this%locations

  end function get_locations_t_data

  !.....................................................................
  function get_real_values_t_data(this) result(real_values)
    ! The calling routine should deallocate real_values after use.
    implicit none

    class(t_data) :: this
    real, dimension(:), allocatable :: real_values

    character(*), parameter :: functionName = "get_real_values_t_data"

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    if(.not.allocated(this%real_values))&
call wrstop(functionName, "There are no real values in the data object.")

    real_values = this%real_values

  end function get_real_values_t_data

  !.....................................................................
  function get_uncerts_t_data(this) result(uncerts)
    ! The calling routine should deallocate uncerts after use.
    implicit none

    class(t_data) :: this
    real, dimension(:), allocatable :: uncerts

    character(*), parameter :: functionName = "get_uncerts_t_data"

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    if(.not.allocated(this%uncerts))&
call wrstop(functionName, "There are no uncerts in the data object.")

    uncerts = this%uncerts

  end function get_uncerts_t_data

  !.....................................................................
  function get_int_values_t_data(this) result(int_values)
    ! The calling routine should deallocate int_values after use.
    implicit none

    class(t_data) :: this
    integer, dimension(:), allocatable :: int_values

    character(*), parameter :: functionName = "get_int_values_t_data"

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    if(.not.allocated(this%int_values))&
call wrstop(functionName, "There are no integer values in the data object.")

    int_values = this%int_values

  end function get_int_values_t_data

  !.....................................................................
  logical function has_locations_t_data(this)
    implicit none
    class(t_data) :: this

    has_locations_t_data = allocated(this%locations)

  end function has_locations_t_data

  !.....................................................................
  logical function has_reals_t_data(this)
    implicit none
    class(t_data) :: this

    has_reals_t_data = allocated(this%real_values)

  end function has_reals_t_data

  !.....................................................................
  logical function has_uncerts_t_data(this)
    implicit none
    class(t_data) :: this

    has_uncerts_t_data = allocated(this%uncerts)

  end function has_uncerts_t_data

  !.....................................................................
  logical function has_ints_t_data(this)
    implicit none
    class(t_data) :: this

    has_ints_t_data = allocated(this%int_values)

  end function has_ints_t_data

  !=====================================================================
  ! Methods of t_likelihood:

  !.....................................................................
  subroutine init_t_likelihood(this, model_obj, data_obj)
    implicit none

    class(t_likelihood)      :: this
    class(t_model), target    :: model_obj
    type(t_data),  intent(in) :: data_obj

    character(*), parameter :: functionName = "init_t_likelihood"

    this%model_obj => model_obj ! We have to do it this way because t_model is an abstract class.
    this%data_obj = data_obj ! t_data is not an abstract class, so we can just pass it in directly.

    this%n_params = model_obj%get_n_params()

    allocate(this%param_scales(this%n_params))
    this%param_scales = 1.0

    this%is_initialized = .true.

  end subroutine init_t_likelihood

  !.....................................................................
  logical function params_are_within_limits_t_likelihood(this, params) result(params_ok)
    implicit none

    class(t_likelihood) :: this
    real, dimension(this%n_params), intent(in) :: params

    character(*), parameter :: functionName = "params_are_within_limits_t_likelihood"

    params_ok = this%model_obj%params_are_within_limits(params)

  end function params_are_within_limits_t_likelihood

  !.....................................................................
  subroutine destroy_t_likelihood(this)
    implicit none

    class(t_likelihood) :: this

    character(*), parameter :: functionName = "destroy_t_likelihood"

    call this%model_obj%destroy()
    nullify(this%model_obj)
    call this%data_obj%destroy()

  end subroutine destroy_t_likelihood

  !.....................................................................
  real function fdiff_1st_t_likelihood(this, params, deltas, ii) result(deriv_1st)
    ! Returns approximate derivative w.r.t. params(ii) via finite difference.
    implicit none

    class(t_likelihood)            :: this
    real, dimension(:), intent(in) :: params,deltas
    integer,            intent(in) :: ii

    character(*), parameter :: functionName = "fdiff_1st_t_likelihood"
    real, dimension(:), allocatable :: local_params
    real :: value_lo,value_hi

    call this%model_obj%check_params_size(params)
    allocate(local_params(size(params)))

    local_params = params
    local_params(ii) = params(ii) - deltas(ii)
    value_lo = this%calc_value(local_params)
    local_params(ii) = params(ii) + deltas(ii)
    value_hi = this%calc_value(local_params)
    deriv_1st = 0.5*(value_hi - value_lo)/deltas(ii)

  end function fdiff_1st_t_likelihood

  !.....................................................................
  real function fdiff_2nd_t_likelihood(this, params, deltas, ii, jj) result(deriv_2nd)
    ! Returns approximate derivative w.r.t. params(ii), params(jj) via finite difference.
    implicit none

    class(t_likelihood)            :: this
    real, dimension(:), intent(in) :: params,deltas
    integer,            intent(in) :: ii,jj

    character(*), parameter :: functionName = "fdiff_2nd_t_likelihood"
    real, dimension(:), allocatable :: local_params
    real :: value_lo,value_hi,value_mid

    call this%model_obj%check_params_size(params)
    allocate(local_params(size(params)))

    if(ii==jj) then
      local_params = params
      value_mid = this%calc_value(local_params)

      local_params(ii) = params(ii) - deltas(ii)
      value_lo = this%calc_value(local_params)

      local_params(ii) = params(ii) + deltas(ii)
      value_hi = this%calc_value(local_params)

      deriv_2nd = (value_hi + value_lo - 2.0*value_mid)/deltas(ii)/deltas(ii)
    else
      local_params = params
      local_params(jj) = params(jj) - deltas(jj)
      value_lo = this%fdiff_1st(local_params, deltas, ii)

      local_params(jj) = params(jj) + deltas(jj)
      value_hi = this%fdiff_1st(local_params, deltas, ii)

      deriv_2nd = 0.5*(value_hi - value_lo)/deltas(jj)
    end if

  end function fdiff_2nd_t_likelihood

  !.....................................................................
  subroutine calc_scale_t_likelihood(this, curv_mat)
    ! This calculates a scale factor for each of the model parameters such that the diagonal terms on the curvature matrix (i.e., the double derivatives of the likelihood w.r.t that parameter) are all equal to -1. **NOTE** that this requires that the curvature matrix be negative-definite.
    implicit none

    class(t_likelihood)              :: this
    real, dimension(:,:), intent(in) :: curv_mat

    character(*), parameter :: functionName = "calc_scale_t_likelihood"
    integer :: ii
    real, dimension(:), allocatable :: diagonal

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    if(size(curv_mat,1)/=this%n_params .or. size(curv_mat,2)/=this%n_params)&
call wrstop(functionName, "Bad dimensions input matrix.")

    allocate(diagonal(this%n_params))
    do ii=1,this%n_params
      diagonal(ii) = curv_mat(ii,ii)
    end do

    if(any(diagonal>=0.0)) then
      call wrerr(functionName, warn, "Some of the diagonal elements of the curvature matrix are non-negative.")
  return
    end if

    this%param_scales = sqrt(-diagonal)

  end subroutine calc_scale_t_likelihood

  !.....................................................................
  subroutine apply_scale_t_likelihood(this, grad, curv_mat)
    implicit none

    class(t_likelihood)                                   :: this
    real, dimension(:),                     intent(inout) :: grad
    real, dimension(size(grad),size(grad)), intent(inout) :: curv_mat

    character(*), parameter :: functionName = "apply_scale_t_likelihood"
    integer :: ii

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    if(size(grad)/=this%n_params)&
call wrstop(functionName, "Bad dimensions input arrays.")

    grad = grad/this%param_scales

    do ii=1,this%n_params
      curv_mat(ii,:) = curv_mat(ii,:)/this%param_scales(ii)
      curv_mat(:,ii) = curv_mat(:,ii)/this%param_scales(ii)
    end do

  end subroutine apply_scale_t_likelihood

  !.....................................................................
  subroutine unscale_t_likelihood(this, params)
    implicit none

    class(t_likelihood)               :: this
    real, dimension(:), intent(inout) :: params

    character(*), parameter :: functionName = "unscale_t_likelihood"

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    if(size(params)/=this%n_params)&
call wrstop(functionName, "Bad dimension input vector.")

    params = params/this%param_scales

  end subroutine unscale_t_likelihood

  !=====================================================================
  ! Methods of t_gauss_like:

  !.....................................................................
  real function calc_value_t_gauss_like(this, params) result (value)
    ! Calculates log-likelihood where the noise is Gaussian.

    implicit none

    class(t_gauss_like) :: this
    real, dimension(this%n_params), intent(in)  :: params

    character(*), parameter :: functionName = "calc_value_t_gauss_like"
    real, dimension(:,:), allocatable :: data_locations
    real, dimension(:), allocatable :: data_values,data_uncerts,model_values,arg

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    data_locations = this%data_obj%get_locations()
    data_values = this%data_obj%get_real_values()
    data_uncerts = this%data_obj%get_uncerts()
    model_values = this%model_obj%calc_values(data_locations, params)
    arg = (model_values - data_values)/data_uncerts
    value = -sum(0.5*arg*arg)

    deallocate(data_locations)
    deallocate(data_values,data_uncerts,model_values,arg)

  end function calc_value_t_gauss_like

  !.....................................................................
  function calc_gradient_t_gauss_like(this, params) result (gradient)
    implicit none

    class(t_gauss_like) :: this
    real, dimension(this%n_params), intent(in) :: params
    real, dimension(this%n_params)             :: gradient

    character(*), parameter :: functionName = "calc_gradient_t_gauss_like"
    real, dimension(:,:), allocatable :: data_locations
    real, dimension(:), allocatable :: data_values,data_uncerts,model_values,args
    real, dimension(:), allocatable :: model_derivs
    integer :: ii

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    data_locations = this%data_obj%get_locations()
    data_values = this%data_obj%get_real_values()
    data_uncerts = this%data_obj%get_uncerts()
    model_values = this%model_obj%calc_values(data_locations, params)
    args = -(model_values - data_values)/data_uncerts/data_uncerts

    gradient(:) = 0.0
    do ii=1, this%n_params
      model_derivs = this%model_obj%calc_1st_derivs(data_locations, params, ii)
      gradient(ii) = sum(args*model_derivs)
    end do

    deallocate(data_locations)
    deallocate(data_values,data_uncerts,model_values,args)
    deallocate(model_derivs)

  end function calc_gradient_t_gauss_like

  !.....................................................................
  function calc_curvature_t_gauss_like(this, params) result (curv_mat)
    implicit none

    class(t_gauss_like)                          :: this
    real, dimension(this%n_params),   intent(in) :: params
    real, dimension(this%n_params,this%n_params) :: curv_mat

    character(*), parameter :: functionName = "calc_curvature_t_gauss_like"
    real, dimension(:,:), allocatable :: data_locations
    real, dimension(:), allocatable :: data_values,data_uncerts,model_values
    real, dimension(:,:), allocatable :: model_derivs
    real, dimension(:), allocatable :: model_2nd_derivs,dxs,data_variance
    integer :: ii,jj,n_locations

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    data_locations = this%data_obj%get_locations()
    data_values = this%data_obj%get_real_values()
    data_uncerts = this%data_obj%get_uncerts()
    model_values = this%model_obj%calc_values(data_locations, params)
    dxs = (model_values - data_values)
    data_variance = data_uncerts*data_uncerts

    curv_mat(:,:) = 0.0

    n_locations = size(data_values)
    allocate(model_derivs(n_locations,this%n_params))

    do ii=1, this%n_params
      model_derivs(:,ii) = this%model_obj%calc_1st_derivs(data_locations, params, ii)

      model_2nd_derivs = this%model_obj%calc_2nd_derivs(data_locations, params, ii, ii)
      curv_mat(ii,ii) = -sum((model_derivs(:,ii)*model_derivs(:,ii)&
                      + dxs*model_2nd_derivs)/data_variance)

      do jj=1, ii-1
        model_2nd_derivs = this%model_obj%calc_2nd_derivs(data_locations, params, ii, jj)

        curv_mat(ii,jj) = -sum((model_derivs(:,ii)*model_derivs(:,jj)&
                        + dxs*model_2nd_derivs)/data_variance)

        curv_mat(jj,ii) = curv_mat(ii,jj)
      end do
    end do

    deallocate(data_locations)
    deallocate(data_values,data_uncerts,model_values)
    deallocate(model_derivs,model_2nd_derivs)

  end function calc_curvature_t_gauss_like

  !=====================================================================
  ! Methods of t_poisson_like:

  !.....................................................................
  real function calc_value_t_poisson_like(this, params) result(value)
    ! The expression for the Poissonian likelihood of a value f given n counts measured is
    !
    !	          f^n * exp(-f)
    !	p(r|n) = ---------------.
    !	               n!
    ! Log of this gives
    !
    !	ln(p) = n*ln(f) - f - ln(n!).
    !
    implicit none

    class(t_poisson_like) :: this
    real, dimension(this%n_params), intent(in)  :: params

    character(*), parameter :: functionName = "calc_value_t_poisson_like"
    real, dimension(:,:), allocatable :: data_locations
    integer, dimension(:), allocatable :: data_values
    real, dimension(:), allocatable :: model_values,arg,cum_sum_lns
    integer :: max_n,n_points,ii

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    data_locations = this%data_obj%get_locations()
    data_values = this%data_obj%get_int_values()
    model_values = this%model_obj%calc_values(data_locations, params)

    ! To calculate ln(n!) where we have an array of n values, we find the highest value of n, store the values ln(n) for n in {1...n_max}, calculate the cumulative sum, then choose the matching value for each input n.

    max_n = maxval(data_values)
    allocate(cum_sum_lns(max_n))
    cum_sum_lns(1) = 0.0
    do ii=2,max_n
      cum_sum_lns(ii) = cum_sum_lns(ii-1) + log(real(ii))
    end do

    n_points = size(data_values)
    allocate(arg(n_points))
    arg(:) = 0.0
    where(data_values>0)
      arg = cum_sum_lns(data_values)
    end where

    arg = data_values*log(model_values) - model_values - arg
    value = sum(arg)

    deallocate(data_locations,cum_sum_lns)
    deallocate(data_values,model_values,arg)

  end function calc_value_t_poisson_like

  !.....................................................................
  function calc_gradient_t_poisson_like(this, params) result (gradient)
    ! Given the log-Poissonian likelihood
    !
    !	ln(p) = n*ln(f) - f - ln(n!),
    !
    ! the gradient is
    !
    !	                    df
    !	g_(i) = (n/f - 1)*------.
    !	                   dq_i
    !
    implicit none

    class(t_poisson_like) :: this
    real, dimension(this%n_params), intent(in) :: params
    real, dimension(this%n_params)             :: gradient

    character(*), parameter :: functionName = "calc_gradient_t_poisson_like"
    real, dimension(:,:), allocatable :: data_locations
    integer, dimension(:), allocatable :: data_values
    real, dimension(:), allocatable :: model_values,arg,cum_sum_lns
    real, dimension(:), allocatable :: model_derivs
    integer :: ii

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    data_locations = this%data_obj%get_locations()
    data_values = this%data_obj%get_int_values()
    model_values = this%model_obj%calc_values(data_locations, params)

    allocate(arg(this%n_params))
    arg(:) = -1.0
    where(model_values>0.0)&
      arg = arg + real(data_values)/model_values

    gradient(:) = 0.0
    do ii=1, this%n_params
      model_derivs = this%model_obj%calc_1st_derivs(data_locations, params, ii)
      gradient(ii) = sum(arg*model_derivs)
    end do

    deallocate(data_locations)
    deallocate(data_values)
    deallocate(model_values,arg,cum_sum_lns)
    deallocate(model_derivs)

  end function calc_gradient_t_poisson_like

  !.....................................................................
  function calc_curvature_t_poisson_like(this, params) result (curv_mat)
    ! Given the log-Poissonian likelihood
    !
    !	ln(p) = n*ln(f) - f - ln(n!),
    !
    ! the curvature is
    !
    !	                      d^2f         n     df     df
    !	H(i,j) = (n/f - 1)*----------- - -----*------*------.
    !	                    dq_i dq_j     f^2   dq_i   dq_j
    !
    implicit none

    class(t_poisson_like)                        :: this
    real, dimension(this%n_params),   intent(in) :: params
    real, dimension(this%n_params,this%n_params) :: curv_mat

    character(*), parameter :: functionName = "calc_curvature_t_poisson_like"
    real, dimension(:,:), allocatable :: data_locations
    integer, dimension(:), allocatable :: data_values
    real, dimension(:), allocatable :: model_values,arg0,arg1
    real, dimension(:,:), allocatable :: model_derivs
    real, dimension(:), allocatable :: model_2nd_derivs
    integer :: ii,jj,n_locations

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    data_locations = this%data_obj%get_locations()
    data_values = this%data_obj%get_int_values()
    model_values = this%model_obj%calc_values(data_locations, params)

    allocate(arg0(this%n_params))
    allocate(arg1(this%n_params))
    arg0(:) = -1.0
    arg1(:) = 0.0
    where(model_values>0.0)
      arg0 = real(data_values)/model_values
      arg1 = arg0/model_values
      arg0 = arg0 - 1.0
    end where

    curv_mat(:,:) = 0.0

    n_locations = size(data_values)
    allocate(model_derivs(n_locations,this%n_params))

    do ii=1, this%n_params
      model_derivs(:,ii) = this%model_obj%calc_1st_derivs(data_locations, params, ii)

      model_2nd_derivs = this%model_obj%calc_2nd_derivs(data_locations, params, ii, ii)
      curv_mat(ii,ii) = sum(arg0*model_2nd_derivs - arg1*model_derivs(:,ii)*model_derivs(:,ii))

      do jj=1, ii-1
        model_2nd_derivs = this%model_obj%calc_2nd_derivs(data_locations, params, ii, jj)
        curv_mat(ii,jj) = sum(arg0*model_2nd_derivs - arg1*model_derivs(:,ii)*model_derivs(:,jj))

        curv_mat(jj,ii) = curv_mat(ii,jj)
      end do
    end do

    deallocate(data_locations)
    deallocate(data_values,arg0,arg1,model_values)
    deallocate(model_derivs,model_2nd_derivs)

  end function calc_curvature_t_poisson_like

  !=====================================================================
  ! Methods of t_event_like:

  ! The broad situation is where one has a set of N events, each
  ! associated with a 'location', i.e. a value of the independent
  ! variable r_ (scalar or vector), and one wants to fit a model
  ! function f(r_) inside a closed region of r_, such that f(r_) is the
  ! probability density for the occurrence of events.
  !
  ! We have a range of interest [rLo,rHi], and a set of M events at
  ! values {r_i}, where r_i in [rLo, rHi] for all i. We start by
  ! dividing the range into N bins of width (rHi-rLo)/N. The likelihood
  ! function for obtaining n_j events in bin j is the Poissonion one,
  !
  ! 	       f_j^{n_j} * exp(-f_j)
  ! 	p_j = -----------------------,
  ! 	               n_j!
  ! where
  ! 	     r_{j+1}
  ! 	      /
  ! 	f_j = | dr f(r)
  ! 	      /
  ! 	     r_j
  ! for
  ! 	             j
  ! 	r_j = rLo + ---*(rHi-rLo).
  ! 	             N
  !
  ! The log-likelihood for obtaining all the count values n_j is
  !
  ! 	        _N-1
  ! 	        \
  ! 	ln(p) =  > ln(p_j)
  ! 	        /_
  ! 	        j=0
  ! 
  ! 	        _N-1
  ! 	        \
  ! 	      =  > [n_j*ln(f_j) - ln(n_j!) - f_j]
  ! 	        /_
  ! 	        j=0
  ! 
  ! 	        _N-1                          rHi
  ! 	        \                             /
  ! 	      =  > [n_j*ln(f_j) - ln(n_j!)] - | dr f(r).
  ! 	        /_                            /
  ! 	        j=0                          rLo
  !
  ! In the limit of large N, eventually nearly all of the bins will
  ! contain zero counts, plus there will be M bins which contain 1 count
  ! each. In this limit, too, we have
  !
  ! 	 lim          (rHi-rLo)
  ! 	 ...   f_j = ----------- * f(r_{j+1/2}).
  ! 	N->inf            N
  ! Thus
  ! 	               _M              rHi
  ! 	 lim           \               /
  ! 	 ...   ln(p) =  > ln[f(r_i)] - | dr f(r) + K,
  ! 	N->inf         /_              /
  ! 	               i=1            rLo
  !
  ! where K is all the terms that don't depend on the model parameters
  ! q_, and which can therefore be ignored for fitting purposes.

  !.....................................................................
  real function calc_value_t_event_like(this, params) result (value)
    implicit none

    class(t_event_like) :: this
    real, dimension(this%n_params), intent(in)  :: params

    character(*), parameter :: functionName = "calc_value_t_event_like"
    real, dimension(:,:), allocatable :: data_locations
    real, dimension(:), allocatable :: model_values

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    data_locations = this%data_obj%get_locations()
    model_values = this%model_obj%calc_values(data_locations, params) ! should not allow it to be <=zero at any point.
    value = sum(log(model_values)) - this%model_obj%integral(this%data_obj%limits, params)

    deallocate(data_locations,model_values)

  end function calc_value_t_event_like

  !.....................................................................
  function calc_gradient_t_event_like(this, params) result(gradient)
    implicit none

    class(t_event_like)                        :: this
    real, dimension(this%n_params), intent(in) :: params
    real, dimension(this%n_params)             :: gradient

    character(*), parameter :: functionName = "calc_gradient_t_event_like"
    real, dimension(:,:), allocatable :: data_locations
    real, dimension(:), allocatable :: model_values,model_derivs
    integer :: ii
    real :: deriv

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    data_locations = this%data_obj%get_locations()
    model_values = this%model_obj%calc_values(data_locations, params) ! should not allow it to be <=zero at any point.

    gradient(:) = 0.0
    do ii=1, this%n_params
      model_derivs = this%model_obj%calc_1st_derivs(data_locations, params, ii)
      deriv = this%model_obj%integral_1st_deriv(this%data_obj%limits, params, ii)
      gradient(ii) = sum(model_derivs/model_values) - deriv
    end do

    deallocate(data_locations,model_values,model_derivs)

  end function calc_gradient_t_event_like

  !.....................................................................
  function calc_curvature_t_event_like(this, params) result (curv_mat)
    implicit none

    class(t_event_like)                          :: this
    real, dimension(this%n_params),   intent(in) :: params
    real, dimension(this%n_params,this%n_params) :: curv_mat

    character(*), parameter :: functionName = "calc_curvature_t_event_like"
    real, dimension(:,:), allocatable :: data_locations
    real, dimension(:), allocatable :: model_values,arg
    real, dimension(:,:), allocatable :: model_derivs
    real, dimension(:), allocatable :: model_2nd_derivs
    integer :: ii,jj,n_locations
    real :: deriv_2nd

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    data_locations = this%data_obj%get_locations()
    model_values = this%model_obj%calc_values(data_locations, params) ! should not allow it to be <=zero at any point.

    curv_mat(:,:) = 0.0

    n_locations = size(data_locations,1)
    allocate(model_derivs(n_locations,this%n_params))

    do ii=1, this%n_params
      model_derivs(:,ii) = this%model_obj%calc_1st_derivs(data_locations, params, ii)
      model_2nd_derivs = this%model_obj%calc_2nd_derivs(data_locations, params, ii, ii)
      arg = model_2nd_derivs - model_derivs(:,ii)*model_derivs(:,ii)/model_values
      deriv_2nd = this%model_obj%integral_2nd_deriv(this%data_obj%limits, params, ii, ii)

      curv_mat(ii,ii) = sum(arg/model_values) - deriv_2nd

      do jj=1, ii-1
        model_2nd_derivs = this%model_obj%calc_2nd_derivs(data_locations, params, jj, ii)!ii, jj)
        arg = model_2nd_derivs - model_derivs(:,ii)*model_derivs(:,jj)/model_values
        deriv_2nd = this%model_obj%integral_2nd_deriv(this%data_obj%limits, params, jj, ii)!ii, jj)

        curv_mat(ii,jj) = sum(arg/model_values) - deriv_2nd

        curv_mat(jj,ii) = curv_mat(ii,jj)
      end do
    end do

    deallocate(data_locations,model_values,model_derivs,model_2nd_derivs,arg)

  end function calc_curvature_t_event_like

end module erolib_like_mod

