! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! See:
!  function example_get_current_link_object(my_LL) result(ii)
!
module erolib_linklist2_mod
  use param
  use erolib_general_mod
  use erolib_str_mod

  implicit none

logical, private, parameter :: doTest = .false.

  private    ! information hiding
  public :: t_link, t_linked_list

  type :: t_link
        private   ! information hiding
        class(*), pointer :: obj => null()
        class(t_link), pointer :: next => null()
        class(t_link), pointer :: previous => null()
  contains
        procedure, non_overridable :: nextLink    ! get the link after this link
        procedure, non_overridable :: previousLink    ! get the previous link
!        procedure, non_overridable :: setNextLink ! set the link after this link
!        procedure, non_overridable :: setPreviousLink ! set the previous link to this link
        procedure, non_overridable :: getObj ! get the type/class associated with this link
  end type t_link

  type :: t_linked_list
        class(t_link),pointer :: firstLink => null() ! first link in list
        class(t_link),pointer :: lastLink => null()  ! last link in list
        class(t_link),pointer :: currLink => null()  ! list iterator
  contains
        procedure, non_overridable :: appendLink    ! add class(*) to the end of the linked list
        procedure, non_overridable :: insertLink    ! insert class(*) in the linked list before the current link.
        procedure, non_overridable :: deleteLink    ! delete this link without retaining connections to other links
        procedure, non_overridable :: getCurrObj    ! returns the object contained in the current link.
!        procedure, non_overridable :: getFirstObj  ! return object associated with firstLink
!        procedure, non_overridable :: getLastObj  ! return object associated with firstLink
        procedure, non_overridable :: getNumLinks ! get the number of associated Links in the List
        procedure, non_overridable :: isEmpty     ! return true if list is empty
        procedure, non_overridable :: set_curr_to_first
        procedure, non_overridable :: set_curr_to_last
        procedure, non_overridable :: set_curr_to_next
        procedure, non_overridable :: set_curr_to_prev
        procedure, non_overridable :: delAllLinks ! delete all links
  end type t_linked_list

  ! creates errors in eclipse only
  interface t_link
    module procedure constructor ! construct/initialize a link
  end interface
  ! end errors in eclipse

contains

  !=====================================================================
  ! Methods of t_linked_list:

  !.....................................................................
  subroutine appendLink(this, obj)
    ! Leaves currLink pointing to the newly appended link.
    class(t_linked_list) :: this
    class(*) :: obj

    class(t_link), pointer :: newLink=>null()

    if(this%isEmpty()) then
      this%currLink => t_link(obj, null(), null())
      this%firstLink => this%currLink
      this%lastLink => this%currLink

    else
      newLink => t_link(obj, this%lastLink, null())

      this%lastLink%next => newLink
!!!      call this%lastLink%setNextLink(newLink)
      this%lastLink => newLink
      this%currLink => newLink
    end if
  end subroutine appendLink

  !.....................................................................
  subroutine insertLink(this, obj)
    ! Inserts a link before the current link. Current link is unchanged.
    class(t_linked_list) :: this
    class(*) :: obj

    class(t_link), pointer :: newLink=>null()

    if(this%isEmpty()) then
      this%currLink => t_link(obj, null(), null())
      this%firstLink => this%currLink
      this%lastLink => this%currLink

    else
      if(associated(this%currLink%previous)) then
        newLink => t_link(obj, this%currLink%previous, this%currLink)
        this%currLink%previous%next => newLink
!!!        call this%currLink%previous%setNextLink(newLink)
      else ! currLink must be the first.
        newLink => t_link(obj, null(), this%currLink)
        this%firstLink => newLink
      end if

      this%currLink%previous => newLink
!!!      call this%currLink%setPreviousLink(newLink)
    end if
  end subroutine insertLink

  !.....................................................................
  integer function deleteLink(this) result(ierr)
    ! Deletes the currLink, leaves currLink pointing to the link before if there is one, the link after if not.
    class(t_linked_list) :: this

    character(*), parameter :: functionName = "deleteLink"
    class(t_link), pointer :: tempLink=>null()

    if(associated(this%currLink)) then
      deallocate(this%currLink%obj, stat=ierr)
      if(ierr /= 0) then
        call wrerr(functionName, warn, "Could not deallocate object of link!")
!  return
      end if

      if(associated(this%currLink%previous)) then ! currLink is not the first.
        tempLink => this%currLink%previous

        if(associated(this%currLink%next)) then ! currLink is not the last.
          this%currLink%next%previous => this%currLink%previous
          this%currLink%previous%next => this%currLink%next
        else ! currLink is the last.
          this%currLink%previous%next => null()
          this%lastLink => tempLink
        end if

        deallocate(this%currLink, stat=ierr)
        this%currLink => tempLink

      else ! currLink is the first.
        if(associated(this%currLink%next)) then ! currLink is not the last.
          tempLink => this%currLink%next
          this%currLink%next%previous => null()
          this%firstLink => tempLink
          deallocate(this%currLink, stat=ierr)
          this%currLink => tempLink

        else ! currLink is the first, last and only link.
          deallocate(this%currLink, stat=ierr)
          this%currLink  => null()
          this%firstLink => null()
          this%lastLink  => null()
        end if
      end if
    end if
  end function deleteLink

  !.....................................................................
  function getCurrObj(this) result(currObj)
    class(t_linked_list) :: this
    class(*), pointer :: currObj

    if(associated(this%currLink)) then
      currObj => this%currLink%getObj()
    else
      currObj => null()
    end if

  end function getCurrObj

  !.....................................................................
!  function getFirstObj(this) result(firstObj)
!    class(t_linked_list) :: this
!    class(*), pointer :: firstObj
!
!    if(associated(this%firstLink)) then
!      firstObj => this%firstLink%getObj()
!    else
!      firstObj => null()
!    end if
!
!  end function getFirstObj

  !.....................................................................
!  function getLastObj(this) result(lastObj)
!    class(t_linked_list) :: this
!    class(*), pointer :: lastObj
!
!    if(associated(this%lastLink)) then
!      lastObj => this%lastLink%getObj()
!    else
!      lastObj => null()
!    end if
!
!  end function getLastObj

  !.....................................................................
  integer function getNumLinks(this)
    class(t_linked_list) :: this

    getNumLinks = 0
    call this%set_curr_to_first()
    do while(associated(this%currLink))
      getNumLinks = getNumLinks + 1
      call this%set_curr_to_next()
    end do
    call this%set_curr_to_first()
  end function

  !.....................................................................
  logical function isEmpty(this)
    class(t_linked_list) :: this

    if (associated(this%firstLink)) then
      isEmpty = .false.
    else
      isEmpty = .true.
    end if
  end function isEmpty

  !.....................................................................
  subroutine set_curr_to_first(this)
    class(t_linked_list) :: this
    this%currLink => this%firstLink
  end subroutine set_curr_to_first

  !.....................................................................
  subroutine set_curr_to_last(this)
    class(t_linked_list) :: this
    this%currLink => this%lastLink
  end subroutine set_curr_to_last

  !.....................................................................
  subroutine set_curr_to_next(this)
    class(t_linked_list) :: this
    if(associated(this%currLink)) then
      this%currLink => this%currLink%next!Link()
    else
      this%currLink => this%firstLink
    end if
  end subroutine set_curr_to_next

  !.....................................................................
  subroutine set_curr_to_prev(this)
    class(t_linked_list) :: this
    if(associated(this%currLink)) then
      this%currLink => this%currLink%previous!Link()
    else
      this%currLink => this%lastLink
    end if
  end subroutine set_curr_to_prev

  !.....................................................................
  integer function delAllLinks(this) result(ierr)
    class(t_linked_list) :: this

    character(*), parameter :: functionName = "delAllLinks"
    character(buffsize) :: iStr
    integer :: ii

    ierr = 0
    ii = 0
    call this%set_curr_to_first()
    do while (associated(this%currLink))
      ii = ii + 1
      ierr = this%deleteLink()

      if(ierr /= 0) then
        iStr = int2str(ii)
        call wrerr(functionName, warn, "Could not delete "//trim(iStr)//"th link!")
    exit
      end if
    end do

  end function delAllLinks

  !=====================================================================
  ! Methods of t_link:

    function nextLink(this)
        class(t_link) :: this
        class(t_link), pointer :: nextLink
        nextLink => this%next
    end function nextLink

    ! ---------------------------------------------------------------- !

    function previousLink(this)
        class(t_link) :: this
        class(t_link), pointer :: previousLink
        previousLink => this%previous
    end function previousLink

    ! ---------------------------------------------------------------- !

!    subroutine setNextLink(this,next)
!        class(t_link) :: this
!        class(t_link), pointer :: next
!        this%next => next
!
!    end subroutine setNextLink

    ! ---------------------------------------------------------------- !

!    subroutine setPreviousLink(this,previous)
!        class(t_link) :: this
!        class(t_link), pointer :: previous
!        this%previous => previous
!    end subroutine setPreviousLink

    ! ---------------------------------------------------------------- !

    function getObj(this)
        class(t_link) :: this
        class(*), pointer :: getObj
        getObj => this%obj
    end function getObj

  !=====================================================================
  ! Utility functions:

    function constructor(obj, previous, next)
        class(t_link),pointer :: constructor
        class(*) :: obj
        class(t_link), pointer :: next, previous

        allocate(constructor)
        constructor%next => next
        constructor%previous => previous
        allocate(constructor%obj, source=obj)
    end function constructor

  !.....................................................................
!  function example_get_current_link_object(my_LL) result(ii)
!    implicit none
!    type(t_linked_list), intent(in) :: my_LL
!    integer :: ii

!    character(*), parameter :: functionName = "example_get_current_link_object"
!    class(*), pointer :: tmp=>null()

!    tmp => my_LL%getCurrObj()
!    if(associated(tmp)) then
!      select type(tmp)
!      type is (integer)
!        ii = tmp
!      class default
!call wrstop(functionName, "Unrecognized class or type.")
!      end select

!    else
!call wrstop(functionName, "No items in list.")
!    end if

!  end function example_get_current_link_object

end module erolib_linklist2_mod
