! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_linklist_mod
    implicit none

logical, private, parameter :: doTest = .false.

    private    ! information hiding
    public :: link, list

    type link
        private   ! information hiding
        class(*), pointer :: obj => null()
        class(link), pointer :: next => null()
        class(link), pointer :: previous => null()
    contains
        procedure, non_overridable :: nextLink    ! get the link after this link
        procedure, non_overridable :: previousLink    ! get the previous link
        procedure, non_overridable :: setNextLink ! set the link after this link
        procedure, non_overridable :: setPreviousLink ! set the previous link to this link
        procedure, non_overridable :: getObj ! get the type/class associated with this link
    end type link

    type list

        class(link),pointer :: firstLink => null() ! first link in list
        class(link),pointer :: lastLink => null()  ! last link in list
        class(link),pointer :: currLink => null()  ! list iterator
    contains
        procedure, non_overridable :: addObj    ! add class(*) to the end of the linked list
        procedure, non_overridable :: currObj    ! add class(*) to the end of the linked list
        procedure, non_overridable :: firstObj  ! return object associated with firstLink
        procedure, non_overridable :: isEmpty     ! return true if list is empty
        procedure, non_overridable :: reset       ! reset list iterator
        procedure, non_overridable :: last       ! set list iterator to last link
        procedure, non_overridable :: next        !iterate to next value in list
        procedure, non_overridable :: previous        !iterate to next value in list
        procedure, non_overridable :: delLink ! delete this link without retaining connections to other links
        procedure, non_overridable :: delAllLinks ! delete all links
        procedure, non_overridable :: getNumLinks ! get the number of associated Links in the List
    end type list

  ! creates errors in eclipse only
  interface link
   module procedure constructor ! construct/initialize a link
  end interface
  ! end errors in eclipse

contains

    ! BEGIN -------------- functions for Lists -------------------

    function getNumLinks (this)
      integer :: getNumLinks
      class(list) :: this
      getNumLinks = 0;
      call this%reset();
      do while(associated(this%currLink))
        getNumLinks = getNumLinks + 1
        call this%next()
      end do
      call this%reset();
    end function

    subroutine addObj(this, obj)
        class(list) :: this
        class(*) :: obj
        class(link), pointer :: newLink

        if (.not. associated(this%firstLink)) then
if(doTest) print *, "this%firstLink is not associated."
            this%currLink => link(obj, this%firstLink, this%firstLink) !**** 20211123 IMS: this causes a memory leak.
            this%firstLink => this%currLink
            this%lastLink => this%currLink
        else
if(doTest) print *, "this%firstLink is already associated."

            newLink => link(obj, this%lastLink%nextLink(), this%lastLink%previousLink())
            call newLink%setPreviousLink(this%currLink)
            call this%lastLink%setNextLink(newLink)
            this%lastLink => newLink
        end if

    end subroutine addObj

    ! ---------------------------------------------------------------- !

    function firstObj(this)
        class(list) :: this
        class(*), pointer :: firstObj

        firstObj => this%firstLink%getObj()

    end function firstObj

    ! ---------------------------------------------------------------- !
    function currObj(this)
        class(list) :: this
        class(*), pointer :: currObj

        currObj => this%currLink%getObj()

    end function currObj

    ! ---------------------------------------------------------------- !

    function isEmpty(this)
        class(list) :: this
        logical isEmpty

        if (associated(this%firstLink)) then
            isEmpty = .false.
        else
            isEmpty = .true.
        end if
    end function isEmpty

    ! ---------------------------------------------------------------- !

    subroutine next (this)
        class(list) :: this
        if(associated(this%currLink)) then
            this%currLink => this%currLink%nextLink()
        else
            this%currLink => this%firstLink
        end if
    end subroutine next

    ! ---------------------------------------------------------------- !

    subroutine reset (this)
        class(list) :: this
        this%currLink => this%firstLink
    end subroutine reset

    ! ---------------------------------------------------------------- !

    subroutine last (this)
        class(list) :: this
        this%currLink => this%lastLink
    end subroutine last

   ! ---------------------------------------------------------------- !

    function delLink (this)
        class(list) :: this
        
        integer :: delLink
        
        delLink = 0

        if(associated(this%currLink, this%firstLink)) then
            if(associated(this%currLink%nextLink())) then
                this%firstLink => this%currLink%next
                this%currLink%next%previous=>null()
                deallocate(this%currLink, stat=delLink)
                call this%reset()
                return
            else
                this%firstLink => null()
                this%lastLink => null()
                deallocate(this%currLink, stat=delLink)
                return
            end if
        end if

        if(associated(this%currLink, this%lastLink)) then
            this%lastLink => this%currLink%previous
            this%currLink%previous%next => null()
            deallocate(this%currLink, stat=delLink)
            call this%reset()
            return
        end if

        if(associated(this%currLink)) then
            ! check if it is the first Link
            this%currLink%next%previous => this%currLink%previous
            this%currLink%previous%next => this%currLink%next
            deallocate(this%currLink, stat=delLink)
            call this%reset()
        end if

    end function delLink

    ! ---------------------------------------------------------------- !

    function delAllLinks (this)
        class(list) :: this
        
        integer :: delAllLinks
        
        delAllLinks = 0
        if(associated(this%currLink)) then
            call this%reset()
            do while (associated(this%currLink))
                delAllLinks = this%delLink()
                if(delAllLinks /= 0) return
            end do
        end if

    end function delAllLinks

    ! ---------------------------------------------------------------- !

    subroutine previous (this)
        class(list) :: this
        if(associated(this%currLink)) then
            this%currLink => this%currLink%previousLink()
        else
            this%currLink => this%firstLink
        end if
    end subroutine previous

    ! END -------------- functions for Lists -------------------

    ! BEGIN -------------- functions for Links -------------------

    function nextLink(this)
        class(link) :: this
        class(link), pointer :: nextLink
        nextLink => this%next
    end function nextLink

    ! ---------------------------------------------------------------- !

    function previousLink(this)
        class(link) :: this
        class(link), pointer :: previousLink
        previousLink => this%previous
    end function previousLink

    ! ---------------------------------------------------------------- !

    subroutine setNextLink(this,next)
        class(link) :: this
        class(link), pointer :: next
        this%next => next

    end subroutine setNextLink

    ! ---------------------------------------------------------------- !

    subroutine setPreviousLink(this,previous)
        class(link) :: this
        class(link), pointer :: previous
        this%previous => previous
    end subroutine setPreviousLink

    ! ---------------------------------------------------------------- !

    function getObj(this)
        class(link) :: this
        class(*), pointer :: getObj
        getObj => this%obj
    end function getObj

    ! END -------------- functions for Links -------------------

    function constructor(obj, next, previous)
        class(link),pointer :: constructor
        class(*) :: obj
        class(link), pointer :: next, previous
        allocate(constructor)
        constructor%next => next
        constructor%previous => previous
        allocate(constructor%obj, source=obj)
    end function constructor

end module erolib_linklist_mod
