! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
!    function deg2rad (alpha)
!    function rad2deg(rad)
!    function ratio_double(x, a, b)
!    recursive function binarysearch_bin_double(valdp, array) result(bin)
!    function interpol_linear_double_sorted (values, x, y, status)
!    real(dp)  function interpol_linear_double(value, xmin, xmax, ymin, ymax)
!    subroutine array_loop (array, loopvalue, status)
!    function mod_double(array, modval)
!    subroutine interpol_gc_double_sorted(newtime, newlat, newlon, time, lat, lon, status)
!    subroutine intermediate_point_gc(ratio, lat1, lon1, lat2, lon2, lat, lon, status)
!    function distance_gc(lat1, lon1, lat2,lon2, deg2rad)
!    subroutine deg2ra(degree, HH, MM, SS, ms, status)
!    subroutine deg2dec(degree, DD, MM, SS, ms, minus, status)
!    subroutine deg2hms_str(degree, str, status)
!    subroutine deg2dms_str(degree, str, status)
!    function distance_gc_multi(lat0, lon0, lat, lon, deg2rad)
!  subroutine getGaussianRandoms(numbers)
!  function poissonRandom(average)
!  function exponentialRandom()
!  function randomNumber()
!  subroutine getExponentialRandoms(numbers)
!  subroutine getRandomNumbers(harvest)
!  real(r8) function inverf(y)
module erolib_math_mod

    use erolib_general_mod
    use erolib_str_mod
    implicit none

    real(dp), parameter :: PI = 4.0_dp*DATAN(1.0_dp)
    real(dp), parameter :: ROOTPI = sqrt(PI)
    real(dp), parameter :: ROOT2 = sqrt(2.0_dp)
    real(dp), parameter :: ROOT3 = sqrt(3.0_dp)
    real(dp), parameter :: ROOT_HALF = 1.0_dp/ROOT2

!    private :: exponentialRandom, poissonRandom
    integer, private, parameter :: bufferSize = 1000

    type :: t_NR_inverf
      real(r8) :: twoOnRootPi = 2.0_r8/ROOTPI
      real(r8) :: y=0.0_r8
      real(r8) :: oneMinusY=0.0_r8
      real(r8) :: zeta=0.0_r8
      integer :: max_n_iter=0
      logical :: is_initialized=.false.
    contains
      procedure :: init => init_inverf
      procedure :: f  =>  f_inverf
      procedure :: df => df_inverf
      procedure :: df_FD => df_FD_inverf
      procedure :: solve => solve_inverf
    end type t_NR_inverf

    interface inverf
        module procedure inverf_r4, inverf_r8
    end interface

contains

    function deg2rad (alpha)

        real(dp) :: deg2rad
        real(dp), intent(in) :: alpha
        deg2rad = alpha * PI / 180.0_dp

    end function deg2rad

    function rad2deg(rad)

        real(dp), intent(in) :: rad
        real(dp) :: rad2deg
        rad2deg = rad * 180.0_dp / PI

    end function rad2deg

    function ratio_double(x, a, b)

        real(dp), intent(in) :: x, a, b
        real(dp) :: ratio_double

        if(x < a .or. x > b) then
            call wrlog("ratio_double", "value is outside of the interval")
        end if

        ratio_double = (x-a) / (b-a)

    end function ratio_double

        ! binarysearch for the bin which fullfill the criteria : value > array(bin) and value < array(bin)
    recursive function binarysearch_bin_double(valdp, array) result(bin)
        implicit none

        real(dp), dimension(:), intent(in) :: array
        real(dp), intent(in) :: valdp
        integer :: n, midbin, bin

        bin = 0 ! just to avoid -Wmaybe-uninitialized warning.

        n = size(array)
        if(n == 0) then
!            bin = 0
    return
        end if
        ! check whether valdp is inside the the array
        if(valdp < array(1) .or. valdp > array(n)) then
!            bin = 0
    return
        end if

        midbin = (n + 1) / 2

        if(n == 2) then
            bin = midbin
            return
        end if

        if(valdp < array(midbin)) then
            bin = binarysearch_bin_double(valdp, array(1:midbin))
        else if (valdp > array(midbin)) then
            bin = binarysearch_bin_double(valdp, array(midbin:n))
            if(bin /= 0) bin = bin + midbin - 1
        end if

    end function binarysearch_bin_double


    ! function to get the interpolated y values for a custom array
    ! requires input array of x values and y values - the input x array must be sorted
    function interpol_linear_double_sorted (values, x, y, status)
        implicit none

        real(dp), intent(in), dimension(:) :: values, x, y
        real(dp), allocatable, dimension(:) :: interpol_linear_double_sorted

        integer :: firstbin, nval, nx, ny, ii
        integer, intent(inout) :: status

        character(*),parameter :: function = "interpol_linear_double_sorted"

        if(status > 100 .or. status == s_errRep) return

        firstbin = binarysearch_bin_double(values(1), x)
        nval = size(values)
        nx = size(x)
        ny = size(y)

        ! report error here, for now only error with print and program stop
        if(nx /= ny) then
            call wrerr(function, s_err, "length of input x and y array are different")
            status = s_errRep
            return
        end if

        call alloc(interpol_linear_double_sorted, nval, status)
        if(status == s_errRep) return

        do ii = 1, nval
            do while (values(ii) > x(firstbin+1) .and. firstbin < nx)
                firstbin = firstbin+1
            end do
            interpol_linear_double_sorted(ii) = &
                interpol_linear_double(values(ii), x(firstbin), x(firstbin+1), y(firstbin), y(firstbin+1))
        end do

    end function interpol_linear_double_sorted

    ! function to linear interpolate between values

    real(dp)  function interpol_linear_double(value, xmin, xmax, ymin, ymax)
        implicit none
        real(dp), intent(in) :: value, xmin, xmax, ymin, ymax
        real(dp) :: ratio
        ratio = (value - xmin) / (xmax - xmin)
        interpol_linear_double = (1.0_dp - ratio) * ymin + ratio * ymax
    end function interpol_linear_double

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> convert an input array of a parameter to account for "looping"
    !> e.g. degrees [0:360] but 360 == 0 or 359.8 == -0.2
    !> assume two consecutive values have always the smallest distance
    !> create a continues array of values accounting for that
    !
    !> @param[in] array
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine array_loop (array, loopvalue, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii, n

        real(dp), dimension(:), intent(inout) :: array
        real(dp), intent(in) :: loopvalue

        character(*), parameter :: function = "array_loop"

        if(status > 100 .or. status == s_errRep) return

        array = mod_double(array, loopvalue)

        n = size(array)

        do ii = 2, n
            if(abs(array(ii) - array(ii-1)) > loopvalue / 2.0_dp) then
                if(array(ii) > array(ii-1)) then
                    array(:ii-1) = array(:ii-1) + loopvalue
                else
                    array(:ii-1) = array(:ii-1) - loopvalue
                end if
            end if
        end do


    end subroutine array_loop

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> apply a mod(ulo) to an array
    !
    !
    !
    !> @param[in] array
    !> @param[out] modval
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function mod_double(array, modval)
        implicit none

        integer :: ii

        real(dp), dimension(:), intent(in) :: array
        real(dp), intent(in) :: modval
        real(dp), dimension(size(array)) :: mod_double

        do ii = 1, size(array)
            mod_double(ii) = mod(array(ii), modval)
        end do

    end function mod_double


    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> interpolate positions on great circle
    !
    !
    !
    !> @param[in] newtime, time, ra, dec
    !> @param[out] outParam
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine interpol_gc_double_sorted(newtime, newlat, newlon, time, lat, lon, status)
        implicit none

        integer, intent(inout) :: status
        integer :: bin, nval, n, ii

        real(dp), dimension(:), intent(in) :: newtime, time
        real(dp), dimension(size(time)), intent(in) :: lon, lat
        real(dp), dimension(size(newtime)), intent(out) :: newlon, newlat
        real(dp) :: ratio

        character(*), parameter :: function = "interpol_gc_double_sorted"

        if(status > 100 .or. status == s_errRep) return

        bin = binarysearch_bin_double(newtime(1), time)
        nval = size(newtime)
        n = size(time)
        
        do ii = 1, nval
            do while (newtime(ii) > time(bin+1) .and. bin < n)
                bin = bin+1
            end do
            ratio = ratio_double(newtime(ii), time(bin), time(bin+1))
            call intermediate_point_gc(ratio, lat(bin), lon(bin), lat(bin+1), lon(bin+1),&
                newlat(ii), newlon(ii), status)
        end do

    end subroutine interpol_gc_double_sorted

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> intermediate points on a great circle
    !
    !
    !
    !> @param[in] ratio, lat1, lon1, lat2, lon2
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine intermediate_point_gc(ratio, lat1, lon1, lat2, lon2, lat, lon, status)
        implicit none

        integer, intent(inout) :: status
        real(dp), intent(in) :: ratio, lat1, lon1, lat2, lon2
        real(dp) :: A, B, d, x, y, z
        real(dp) :: rad_lat1, rad_lon1, rad_lat2, rad_lon2
        real(dp), intent(out) :: lat, lon

        character(*), parameter :: function = "intermediate_point_gc"

        if(status > 100 .or. status == s_errRep) return

        rad_lat1 = deg2rad(lat1)
        rad_lon1 = deg2rad(lon1)
        rad_lat2 = deg2rad(lat2)
        rad_lon2 = deg2rad(lon2)

        d = distance_gc (rad_lat1, rad_lon1, rad_lat2, rad_lon2, false)
        
        A=sin((1-ratio)*d)/sin(d)
        B=sin(ratio*d)/sin(d)

        x = A*cos(rad_lat1)*cos(rad_lon1) +  B*cos(rad_lat2)*cos(rad_lon2)
        y = A*cos(rad_lat1)*sin(rad_lon1) +  B*cos(rad_lat2)*sin(rad_lon2)
        z = A*sin(rad_lat1)           +  B*sin(rad_lat2)
        lat=atan2(z,sqrt(x*x+y*y))
        lon=atan2(y,x)

        lat = rad2deg(lat)
        lon = rad2deg(lon)

       ! call wrlog(function, trim(double2str(lat)) // " " // trim(double2str(lon)))

    end subroutine intermediate_point_gc

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> calculate the distance between two points on a sphere
    !
    !
    !,
    !> @param[in] lat1, lon1, lat2,lon2, deg2rad
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function distance_gc(lat1, lon1, lat2,lon2, deg2rad)
        implicit none

        real(dp), intent(in) :: lon1, lat1, lon2, lat2
        real(dp) :: distance_gc, a, b, c, d, e, dlat, convert
        integer, intent(in) :: deg2rad

        character(*), parameter :: function = "distance_gc"

        convert = 1.0_dp
        if(deg2rad == true) convert = PI / 180.0_dp

        dlat = (lat2 - lat1)* convert
        a = cos(lon2 * convert) * sin(dlat)
        b = cos(lon1 * convert) * sin(lon2 * convert)
        c = sin(lon1 * convert) * cos(lon2 * convert) * cos(dlat)
        d = sin(lon1 * convert) * sin(lon2 * convert)
        e = cos(lon1 * convert) * cos(lon2 * convert) * cos(dlat)

        distance_gc = atan2(sqrt(a*a + (b-c)*(b-c)), d + e)

        if(deg2rad == true) distance_gc = distance_gc / convert

    end function distance_gc

!    !---------------------------------------------------------------------------
!    !> @author
!    !> Christoph Grossberger
!    !
!    ! DESCRIPTION:
!    !> Brief description of routine.
!    !> @brief
!    !> the uncertainty distance between two source positions for which both sources may be identical
!    !
!    !
!    !
!    !> @param[in] radec0_err, sys_err, radec_err
!    !> @param[out] outParam
!    !> @param[inout] status
!    !---------------------------------------------------------------------------
!
!    function distance_err(radec0_err, radec_err, nsigma, sys_err)
!        implicit none
!
!        character(*), parameter :: function = "distance_err"
!        real(r4), intent(in) :: sys_err, radec0_err
!        real(r4), intent(in) :: radec_err
!        integer, intent(in) :: nsigma
!
!        real(r8) :: distance_err
!
!        distance_err = sqrt(nsigma**2 * (radec0_err**2 + radec_err**2) + sys_err**2)
!
!    end function distance_err

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> convert an input number [deg] into HH MM SS . ms
    !
    !
    !
    !> @param[in] degree
    !> @param[out] HH, MM, SS, ms
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine deg2ra(degree, HH, MM, SS, ms, status)
        implicit none

        integer, intent(inout) :: status
        real(r8), intent(in) :: degree
        integer, intent(out) :: HH, MM, SS
        real(r8) :: hours
        real(r8), intent(out) :: ms

        character(*), parameter :: function = "deg2ra"

        if(status > 100 .or. status == s_errRep) return

        hours = degree / 15.0_r8
        HH = int(hours)
        MM = int((hours - HH) * 60.0_r8)
        SS = int((hours - HH - MM / 60.0_r8) * 3600.0_r8)
        ms = (hours - HH - MM/60.0_r8 - SS/3600.0_r8) * 3600.0_r8 * 1000.0_r8

    end subroutine deg2ra

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> convert an input number [deg] into DD MM SS . ms
    !
    !
    !
    !> @param[in] degree
    !> @param[out] DD, MM, SS, ms
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine deg2dec(degree, DD, MM, SS, ms, minus, status)
        implicit none

        integer, intent(inout) :: status
        real(r8), intent(in) :: degree
        integer, intent(out) :: DD, MM, SS, minus
        real(r8), intent(out) :: ms

        character(*), parameter :: function = "deg2dec"

        if(status > 100 .or. status == s_errRep) return

        minus = false
        if(degree < 0.0_r8) minus = true

        DD = int(abs(degree))
        MM = int((abs(degree) - DD) * 60.0_r8)
        SS = int((abs(degree) - DD - MM / 60.0_r8) * 3600.0_r8)
        ms = (abs(degree) - DD - MM/60.0_r8 - SS/3600.0_r8) * 3600.0_r8 * 1000.0_r8

    end subroutine deg2dec

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> construct a hms string from degrees with two decimal
    !
    !
    !
    !> @param[in] degrees
    !> @param[out] str
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine deg2hms_str(degree, str, status)
        implicit none

        integer, intent(inout) :: status
        real(r8), intent(in) :: degree

        integer :: HH, MM, SS
        real(r8) :: ms
        character(buffsize) :: str, HH_str, MM_str, SS_str, ms_str

        character(*), parameter :: function = "deg2hms_str"

        if(status > 100 .or. status == s_errRep) return

        call deg2ra(degree, HH, MM, SS, ms, status)

        HH_str = int2str(HH, status)
        MM_str = int2str(MM, status)
        SS_str = int2str(SS, status)
        ms_str = double2str(ms, status)

        if(len_trim(HH_str) < 2) HH_str = "0" // trim(HH_str)
        if(len_trim(MM_str) < 2) MM_str = "0" // trim(MM_str)
        if(len_trim(SS_str) < 2) SS_str = "0" // trim(SS_str)

        ! need two decimal for coordinate string
        ms = ms / 10.0_r8
        ms_str = int2str(nint(ms), status)
        if(len_trim(ms_str) < 2) ms_str = "0" // trim(ms_str)

        str = trim(HH_str) // trim(MM_str) // trim(SS_str) // "." // trim(ms_str)

    end subroutine deg2hms_str

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> construct a hms string from degrees with two decimal
    !
    !
    !
    !> @param[in] degrees
    !> @param[out] str
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine deg2dms_str(degree, str, status)
        implicit none

        integer, intent(inout) :: status
        real(r8), intent(in) :: degree

        integer :: DD, MM, SS, minus
        real(r8) :: ms
        character(buffsize) :: str, DD_str, MM_str, SS_str, ms_str

        character(*), parameter :: function = "deg2dms_str"

        if(status > 100 .or. status == s_errRep) return

        call deg2dec(degree, DD, MM, SS, ms, minus, status)

        DD_str = int2str(DD, status)
        MM_str = int2str(MM, status)
        SS_str = int2str(SS, status)
        ms_str = double2str(ms, status)

        if(len_trim(DD_str) < 2) DD_str = "0" // trim(DD_str)
        if(len_trim(MM_str) < 2) MM_str = "0" // trim(MM_str)
        if(len_trim(SS_str) < 2) SS_str = "0" // trim(SS_str)

        ! need only one decimal for coordinate string
        ms = ms / 100.0_r8
        ms_str = int2str(nint(ms), status)

        if(minus == true) then
            str = "-"//trim(DD_str) // trim(MM_str) // trim(SS_str) // "." // trim(ms_str)
        else
            str = "+"//trim(DD_str) // trim(MM_str) // trim(SS_str) // "." // trim(ms_str)
        end if

    end subroutine deg2dms_str

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> calculate the distance between two points on a sphere
    !
    !
    !,
    !> @param[in] lat1, lon1, lat2,lon2, deg2rad
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function distance_gc_multi(lat0, lon0, lat, lon, deg2rad)
        implicit none

        real(r8), intent(in) :: lon0, lat0
        real(r8), intent(in), dimension(:) :: lat, lon
        real(r8), dimension(size(lat)) :: distance_gc_multi
        real(r8), dimension(size(lat)) :: a, b, c, d, e, dlat, convert
        integer, intent(in) :: deg2rad

        character(*), parameter :: function = "distance_gc_multi"

        convert = 1.0_r8
        ! convert degrees to rad
        if(deg2rad == true) convert = PI / 180.0_r8

        dlat = (lat - lat0) * convert
        a = cos(lon * convert) * sin(dlat)
        b = cos(lon0 * convert) * sin(lon * convert)
        c = sin(lon0 * convert) * cos(lon * convert) * cos(dlat)
        d = sin(lon0 * convert) * sin(lon * convert)
        e = cos(lon0 * convert) * cos(lon * convert) * cos(dlat)

        distance_gc_multi = atan2(sqrt(a*a + (b-c)*(b-c)), d + e)

        ! convert rad to degress
        if(deg2rad == true) distance_gc_multi = distance_gc_multi / convert

    end function distance_gc_multi

!    !---------------------------------------------------------------------------
!    !> @author
!    !> Christoph Grossberger
!    !
!    ! DESCRIPTION:
!    !> Brief description of routine.
!    !> @brief
!    !> the uncertainty distance between two source positions for which both sources may be identical
!    !
!    !
!    !
!    !> @param[in] radec0_err, sys_err, radec_err
!    !> @param[out] outParam
!    !> @param[inout] status
!    !---------------------------------------------------------------------------
!
!    function distance_err_multi(radec0_err, radec_err, nsigma, sys_err)
!        implicit none
!
!        character(*), parameter :: function = "distance_err_multi"
!        real(r4), intent(in) :: sys_err, radec0_err
!        real(r4), intent(in), dimension(:) :: radec_err
!        integer, intent(in) :: nsigma
!
!        real(r8), dimension(size(radec_err)) :: distance_err_multi
!
!        distance_err_multi = sqrt(nsigma**2 * (radec0_err**2 + radec_err**2) + sys_err**2)
!
!    end function distance_err_multi

  !.....................................................................
  subroutine getGaussianRandoms(numbers)
    ! Returns a vector of values which have an gaussian probability
    ! distribution about zero with standard deviation equal to 1.0.

    implicit none

    real, dimension(:), intent(out) :: numbers

    integer :: ii, n_rand
    real fac, rsq, v1, v2

    n_rand = size(numbers)
!print *, "... getGaussianRandoms(). n_rand=", n_rand

    if (n_rand / 2.0 .eq. floor(n_rand / 2.0)) then ! even
      do ii = 1, n_rand - 1, 2
        do
          v1 = 2.0 * randomNumber() - 1.0
          v2 = 2.0 * randomNumber() - 1.0
          rsq = v1 * v1 + v2 * v2

        if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
        end do

        fac = sqrt(-2.0 * log(rsq) / rsq)
        numbers(ii)     = v1 * fac
        numbers(ii + 1) = v2 * fac
      end do

    else ! odd
      do ii = 1, n_rand - 2, 2
        do
          v1 = 2.0 * randomNumber() - 1.0
          v2 = 2.0 * randomNumber() - 1.0
          rsq = v1 * v1 + v2 * v2

        if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
        end do

        fac = sqrt(-2.0 * log(rsq) / rsq)
        numbers(ii)     = v1 * fac
        numbers(ii + 1) = v2 * fac
      end do

      do
        v1 = 2.0 * randomNumber() - 1.0
        v2 = 2.0 * randomNumber() - 1.0
        rsq = v1 * v1 + v2 * v2

      if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
      end do

      fac = sqrt(-2.0 * log(rsq) / rsq)
      numbers(n_rand) = v1 * fac
    end if

  end subroutine getGaussianRandoms

  !.....................................................................
  function poissonRandom(average)
    ! Returns a random value with a Poissonian distribution. The algorithm is
    ! not efficient for large values of 'average'.

    implicit none

    real, intent(in) :: average
    integer          :: poissonRandom

    real :: trial

    poissonRandom = 0
    trial = exponentialRandom()
    do while (trial < average)
      poissonRandom = poissonRandom + 1
      trial = trial + exponentialRandom()
    end do

  end function poissonRandom

  !.....................................................................
  function exponentialRandom()
    implicit none

    real :: exponentialRandom
    real :: harvest(bufferSize)
    integer :: ii
    save ii, harvest
    data ii /1/

    if (ii .eq. 1) then
      call getExponentialRandoms(harvest)
    end if
    exponentialRandom = harvest(ii)
    ii = ii + 1
    if (ii .gt. bufferSize) ii = 1

  end function exponentialRandom

  !.....................................................................
  function randomNumber()
    implicit none

    real :: randomNumber
    real :: harvest(bufferSize)
    integer :: ii
    save ii, harvest
    data ii /1/

    if (ii .eq. 1) then
      call getRandomNumbers(harvest)
    end if
    randomNumber = harvest(ii)
    ii = ii + 1
    if (ii .gt. bufferSize) ii = 1

  end function randomNumber

  !.....................................................................
  subroutine getExponentialRandoms(numbers)
    ! Returns a vector of values which have an exponential probability
    ! distribution with a mean value of 1.0.

    implicit none

    real, intent(out) :: numbers(:)

    real, allocatable :: harvest(:)
    integer :: n_rands,ii

    n_rands = size(numbers)
    allocate(harvest(n_rands))

    call getRandomNumbers(harvest)
    do ii = 1, n_rands
      numbers(ii) = -log(1.0 - harvest(ii)) ! the subtraction of the
      ! random number from 1.0 is to prevent the occurrence of zero-
      ! valued argument to log().
    end do

  end subroutine getExponentialRandoms

  !.....................................................................
  subroutine getRandomNumbers(harvest)
    implicit none

    real, intent(out) :: harvest(:)

    call random_seed()
    call random_number(harvest)

  end subroutine getRandomNumbers


  !.....................................................................
  subroutine init_inverf(this, y, max_n_iter)
    implicit none
    class(t_NR_inverf)   :: this
    real(r8), intent(in) :: y
    integer,  intent(in) :: max_n_iter

    this%y = y
    this%oneMinusY = 1.0_r8 - y
    this%zeta = 1.0_r8/sqrt(-log(this%oneMinusY))
    this%max_n_iter = max_n_iter

    this%is_initialized=.true.

  end subroutine init_inverf

  !.....................................................................
  real(r8) function f_inverf(this, x, ii) result(f)
    implicit none
    class(t_NR_inverf)   :: this
    real(r8), intent(in) :: x
    integer,  intent(in) :: ii

    character(*), parameter :: functionName = "f_inverf"

    if(.not.this%is_initialized)&
call wrstop(functionName, "Not initialized.")

    select case(ii)
    case(0)
      f = erf(x) - this%y
    case(1)
      f = erfc(x) - this%oneMinusY
    case(2)
      f = this%zeta*sqrt(-log(erfc(x))) - 1.0_r8
    case default
call wrstop(functionName, "Bad type.")
    end select

  end function f_inverf

  !.....................................................................
  real(r8) function df_inverf(this, x, ii) result(df)
    implicit none
    class(t_NR_inverf)   :: this
    real(r8), intent(in) :: x
    integer,  intent(in) :: ii

    character(*), parameter :: functionName = "df_inverf"
    real(r8) :: erfcx

    if(.not.this%is_initialized)&
call wrstop(functionName, "Not initialized.")

    select case(ii)
    case(0)
      df = (2.0_r8/ROOTPI)*exp(-x*x)
    case(1)
      df = -(2.0_r8/ROOTPI)*exp(-x*x)
    case(2)
      erfcx = erfc(x)
      df = this%zeta*exp(-x*x)/sqrt(-log(erfcx))/erfcx/ROOTPI
    case default
call wrstop(functionName, "Bad type.")
    end select

  end function df_inverf

  !.....................................................................
  real(r8) function df_FD_inverf(this, x, ii, frac_delta) result(df)
    implicit none
    class(t_NR_inverf)   :: this
    real(r8), intent(in) :: x
    integer,  intent(in) :: ii
    real(r8), intent(in), optional :: frac_delta

    character(*), parameter :: functionName = "df_FD_inverf"
    real(r8) :: delta_x,f_lo,f_hi

    if(present(frac_delta)) then
      delta_x = x*frac_delta
    else
      delta_x = x*0.001
    end if

    f_lo = this%f(x-delta_x, ii)
    f_hi = this%f(x+delta_x, ii)
    df = (f_hi - f_lo)/2.0_r8/delta_x

  end function df_FD_inverf

  !.....................................................................
  subroutine solve_inverf(this, type_i, x, n_iter)
    implicit none
    class(t_NR_inverf)             :: this
    integer,           intent(in)  :: type_i
    real(r8),          intent(out) :: x
    integer, optional, intent(out) :: n_iter

    character(*), parameter :: functionName = "solve_inverf"
    real(r8) :: x_old,x_new,f,df
    integer :: ii

    if(.not.this%is_initialized)&
call wrstop(functionName, "Not initialized.")

!*** could speed it up a lot with better guesses for initial x_old - and maybe change of variable for very close to 1.
    if(type_i == 2) then
      x_old = 1.0_r8/this%zeta
    else
      x_old = 0.0_r8
    end if

    do ii=1,this%max_n_iter
      f = this%f(x_old, type_i)
      df = this%df(x_old, type_i)
      x_new = x_old - f/df

      if(abs(x_new - x_old)<epsilon(1.0_r8))&
    exit

      x_old = x_new
    end do
    x = x_new
    if(present(n_iter)) n_iter = ii

  end subroutine solve_inverf

  !.....................................................................
  real(r8) recursive function inverf_scalar(y) result(x)
    ! This calculates the inverse error function following the 'master function' prescription of Blair, Edwards and Johnson, Mathematics of Computation, 30, 186, pp827-830 (1976).
    implicit none

    real(r8), intent(in) :: y

    character(*), parameter :: functionName = "inverf_scalar"
    integer, parameter :: max_n_iter=100
    type(t_NR_inverf) :: nr
    integer :: type_i

    if(y<0.0_r8) then
      x = -inverf_scalar(-y)
    else
      if(y<0.9975) then
        ! Invert erf(x) - y = 0.
        !
        type_i = 0

      else if(1.0_r8-y<1.0d-12) then
        ! Invert erfc(x) - (1 - y) = 0.
        !
        type_i = 1

      else if(y<1.0_r8) then
        ! Invert sqrt(-ln[erfc{x}])/sqrt(-ln[1-y]) - 1 = 0.
        !
        type_i = 2

      else
call wrstop(functionName, "y out of range.")
      end if

      call nr%init(y, max_n_iter)
      call nr%solve(type_i, x)
    end if

  end function inverf_scalar

  !.....................................................................
  function inverf_r4(ys) result(xs)
    implicit none

    real(r4), dimension(:), intent(in) :: ys
    real(r4), dimension(:), allocatable :: xs

    character(*), parameter :: functionName = "inverf_r4"
    integer :: n_points,ii

    n_points = size(ys)
    allocate(xs(n_points))
    do ii=1,n_points
      xs(ii) = real(inverf_scalar(real(ys(ii),kind=r8)),kind=r4)
    end do

  end function inverf_r4

  !.....................................................................
  function inverf_r8(ys) result(xs)
    implicit none

    real(r8), dimension(:), intent(in) :: ys
    real(r8), dimension(:), allocatable :: xs

    character(*), parameter :: functionName = "inverf_r8"
    integer :: n_points,ii

    n_points = size(ys)
    allocate(xs(n_points))
    do ii=1,n_points
      xs(ii) = inverf_scalar(ys(ii))
    end do

  end function inverf_r8

  ! From cephes (I have not yet got the inverse erf to work):
  !.....................................................................
  real(r8) function ndtri(y0) result(x)
    !                                                     ndtri.c
    !
    !     Inverse of Normal distribution function
    !
    !
    !
    ! SYNOPSIS:
    !
    ! double x, y, ndtri();
    !
    ! x = ndtri( y );
    !
    !
    !
    ! DESCRIPTION:
    !
    ! Returns the argument, x, for which the area under the
    ! Gaussian probability density function (integrated from
    ! minus infinity to x) is equal to y.
    !
    !
    ! For small arguments 0 < y < exp(-2), the program computes
    ! z = sqrt( -2.0    ! log(y) );  then the approximation is
    ! x = z - log(z)/z  - (1/z) P(1/z) / Q(1/z).
    ! There are two rational functions P/Q, one for 0 < y < exp(-32)
    ! and the other for y up to exp(-2).  For larger arguments,
    ! w = y - 0.5, and  x/sqrt(2pi) = w + w**3 R(w**2)/S(w**2)).
    !
    !
    ! ACCURACY:
    !
    !                      Relative error:
    ! arithmetic   domain        # trials      peak         rms
    !    IEEE     0.125, 1        20000       7.2e-16     1.3e-16
    !    IEEE     3e-308, 0.135   50000       4.6e-16     9.8e-17
    !
    !
    ! ERROR MESSAGES:
    !
    !   message         condition    value returned
    ! ndtri domain       x < 0        NPY_NAN
    ! ndtri domain       x > 1        NPY_NAN
    !
    implicit none
    real(r8), intent(in) :: y0

    character(*), parameter :: functionName = "ndtri"

    ! Approximation for 0 <= |y - 0.5| <= 3/8
    !
    real(r8), dimension(5) :: P0=[&
       -5.99633501014107895267E1&
      , 9.80010754185999661536E1&
      ,-5.66762857469070293439E1&
      , 1.39312609387279679503E1&
      ,-1.23916583867381258016E0&
    ]

    !/* 1.00000000000000000000E0, */
    real(r8), dimension(8) :: Q0=[&
        1.95448858338141759834E0&
      , 4.67627912898881538453E0&
      , 8.63602421390890590575E1&
      ,-2.25462687854119370527E2&
      , 2.00260212380060660359E2&
      ,-8.20372256168333339912E1&
      , 1.59056225126211695515E1&
      ,-1.18331621121330003142E0&
    ]

    ! Approximation for interval z = sqrt(-2 log y ) between 2 and 8
    ! i.e., y between exp(-2) = .135 and exp(-32) = 1.27e-14.
    !
    real(r8), dimension(9) :: P1=[&
        4.05544892305962419923E0&
      , 3.15251094599893866154E1&
      , 5.71628192246421288162E1&
      , 4.40805073893200834700E1&
      , 1.46849561928858024014E1&
      , 2.18663306850790267539E0&
      ,-1.40256079171354495875E-1&
      ,-3.50424626827848203418E-2&
      ,-8.57456785154685413611E-4&
    ]

    !/*  1.00000000000000000000E0, */
    real(r8), dimension(8) :: Q1=[&
        1.57799883256466749731E1&
      , 4.53907635128879210584E1&
      , 4.13172038254672030440E1&
      , 1.50425385692907503408E1&
      , 2.50464946208309415979E0&
      ,-1.42182922854787788574E-1&
      ,-3.80806407691578277194E-2&
      ,-9.33259480895457427372E-4&
    ]

    ! Approximation for interval z = sqrt(-2 log y ) between 8 and 64
    ! i.e., y between exp(-32) = 1.27e-14 and exp(-2048) = 3.67e-890.
    !
    real(r8), dimension(9) :: P2=[&
        3.23774891776946035970E0&
      , 6.91522889068984211695E0&
      , 3.93881025292474443415E0&
      , 1.33303460815807542389E0&
      , 2.01485389549179081538E-1&
      , 1.23716634817820021358E-2&
      , 3.01581553508235416007E-4&
      , 2.65806974686737550832E-6&
      , 6.23974539184983293730E-9&
    ]

!    /*  1.00000000000000000000E0, */
    real(r8), dimension(8) :: Q2=[&
        6.02427039364742014255E0&
      , 3.67983563856160859403E0&
      , 1.37702099489081330271E0&
      , 2.16236993594496635890E-1&
      , 1.34204006088543189037E-2&
      , 3.28014464682127739104E-4&
      , 2.89247864745380683936E-6&
      , 6.79019408009981274425E-9&
    ]

    real(r8) :: y, z, y2, x0, x1
    integer :: code

    if(y0<=0.0 .or. y0>=1.0)&
call wrstop(functionName, "Input out of bound.")

    code = 1
    y = y0
    if(y > (1.0 - 0.13533528323661269189)) then !/* 0.135... = exp(-2) */
      y = 1.0 - y
      code = 0
    end if

    if(y > 0.13533528323661269189) then
      y = y - 0.5
      y2 = y * y
      x = y + y * (y2 * polevl(y2, P0) / p1evl(y2, Q0))
      x = x * ROOTPI
  return
    end if

    x = sqrt(-2.0 * log(y))
    x0 = x - log(x) / x

    z = 1.0 / x
    if(x < 8.0) then ! /* y > exp(-32) = 1.2664165549e-14 */
      x1 = z * polevl(z, P1) / p1evl(z, Q1)
    else
      x1 = z * polevl(z, P2) / p1evl(z, Q2)
    end if
    x = x0 - x1
    if (code /= 0)&
      x = -x

  end function ndtri

  !.....................................................................
  real(r8) function polevl(x, coef) result(ans)
    !     Evaluate polynomial
    !
    !
    !
    ! SYNOPSIS:
    !
    ! int N;
    ! double x, y, coef[N+1], polevl[];
    !
    ! y = polevl( x, coef, N );
    !
    !
    !
    ! DESCRIPTION:
    !
    ! Evaluates polynomial of degree N:
    !
    !                     2          N
    ! y  =  C  + C x + C x  +...+ C x
    !        0    1     2          N
    !
    ! Coefficients are stored in reverse order:
    !
    ! coef[0] = C  , ..., coef[N] = C  .
    !            N                   0
    !
    !  The function p1evl() assumes that coef[N] = 1.0 and is
    ! omitted from the array.  Its calling arguments are
    ! otherwise the same as polevl().
    !
    !
    ! SPEED:
    !
    ! In the interest of speed, there are no checks for out
    ! of bounds arithmetic.  This routine is used by most of
    ! the functions in the library.  Depending on available
    ! equipment features, the user may wish to rewrite the
    ! program in microcode or assembly language.
    !
    implicit none
    real(r8),               intent(in) :: x
    real(r8), dimension(:), intent(in) :: coef

    integer :: N
    integer :: ii

    N = size(coef)

    ans = coef(1)
    do ii=2, N
      ans = x*ans + coef(ii)
    end do

  end function polevl

  !.....................................................................
  real(r8) function p1evl(x, coef) result(ans)
    ! Evaluate polynomial when coefficient of x^N is 1.0.
    ! Otherwise same as polevl.
    !
    implicit none

    real(r8),               intent(in) :: x
    real(r8), dimension(:), intent(in) :: coef

    integer :: N
    integer :: ii

    N = size(coef)

    ans = x + coef(1)
    do ii=2, N
      ans = x*ans + coef(ii)
    end do
  end function p1evl

  !.....................................................................
  real(r8) function erfinv(y) result(x)
    ! Inverse of the error function.
    !
    ! Computes the inverse of the error function on the restricted domain
    ! -1 < y < 1. This restriction ensures the existence of a unique result
    ! such that erf(erfinv(y)) = y.
    !
    implicit none
    real(r8), intent(in) :: y

    character(*), parameter :: functionName = "erfinv"

    if(y <= -1.0_r8 .or. y >= 1.0_r8)&
call wrstop(functionName, "Outside bounds.")

    x = ndtri(0.5_r8 * (y+1.0_r8)) * ROOT_HALF
  end function erfinv

  !.....................................................................
  real(r8) function erfcinv(y) result(x)
    ! Inverse of the complementary error function.
    !
    ! Computes the inverse of the complimentary error function on the restricted
    ! domain 0 < y < 2. This restriction ensures the existence of a unique result
    ! such that erfc(erfcinv(y)) = y.
    !
    implicit none
    real(r8), intent(in) :: y

    character(*), parameter :: functionName = "erfcinv"

    if(y <= -1.0_r8 .or. y >= 1.0_r8)&
call wrstop(functionName, "Outside bounds.")

    x = -ndtri(0.5_r8 * y) * ROOT_HALF
  end function erfcinv
end module erolib_math_mod
