! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_quickplot_mod
  ! The purpose of the present module is to provide routines for 'quick and dirty' plots, probably in use mostly of diagnostic data. How to use it:
  !
  !	(1) Include in the subroutine the line
  !		type(t_quickplot) :: quickplot
  !
  !	(2a) For each graph, do
  !		call quickplot%add_graph(xs, ys, format_str)
  !
  !	(2b) For each image, do
  !		call quickplot%add_image(2d_array, x_lo, x_hi, y_lo, y_hi, with_colourbar)
  !
  !	(3) Finally,
  !		call quickplot%show()
  !
  use param
  use erolib_general_mod
  use erolib_linklist_mod
  use erolib_ranges_mod

  implicit none

  private

  type, public :: t_histo_bin
    integer :: counts=0
    real :: x=0.0
    real :: y=0.0
    real :: yerr=0.0
  end type t_histo_bin

  type, public :: t_histo
    private
    real, dimension(:), allocatable :: bin_edges
    integer, dimension(:), allocatable :: counts
    real, dimension(:), allocatable :: ys
    real, dimension(:), allocatable :: yerrs
    real :: expected_integral=1.0
    integer :: n_under=0
    integer :: n_over=0
    logical :: is_initialized=.false.
  contains
    procedure :: init => init_t_histo
    procedure :: get => get_t_histo
    procedure :: destroy => destroy_t_histo
  end type t_histo

  type :: t_plot
    real, dimension(:,:), allocatable :: points ! 1=x, 2=y, 3=y_err, 4=x_err
    integer :: colour_index=0
    integer :: point_style=0
!**    integer :: line_style=0
    logical :: poly=.false.
  contains
    procedure :: init => init_t_plot
    procedure :: destroy => destroy_t_plot
  end type t_plot

  type, public :: t_quickplot
    private
    type(t_plot), dimension(:), allocatable :: plots
    character(3) :: pgdev="/xs"
    integer :: dev_id=0
    class(list), pointer :: plots_LL=>null()
    real, dimension(:,:), allocatable :: bkg_image
    real, dimension(6) :: pgTrMatrix = (/0.0, 1.0, 0.0, 0.0, 0.0, 1.0/)
    logical :: with_colourbar
    type(tRange) :: x_range,y_range
    real :: w_x_min=0.0
    real :: w_x_max=1.0
    real :: w_y_min=0.0
    real :: w_y_max=1.0
  contains
    procedure, non_overridable :: add_graph => add_graph_t_quickplot
    procedure, non_overridable :: add_image => add_image_t_quickplot
    procedure, non_overridable :: show => show_t_quickplot
    procedure, private         :: deleteLinkedList => deleteLinkedList_t_quickplot
    procedure, private         :: init_plots => init_plots_t_quickplot
    procedure, private         :: destroy => destroy_t_quickplot
  end type t_quickplot

  real, parameter :: vp_x_lo=0.1
  real, parameter :: vp_x_hi=0.999
  real, parameter :: vp_y_lo=0.1
  real, parameter :: vp_y_hi=0.999
  real, parameter :: img_vp_x_lo=vp_x_lo
  real, parameter :: img0_vp_x_hi=vp_x_hi
  real, parameter :: img1_vp_x_hi=0.81
  real, parameter :: img_vp_y_lo=vp_y_lo
  real, parameter :: img_vp_y_hi=vp_y_hi
  real, parameter :: cbar_vp_x_lo = 0.82
  real, parameter :: cbar_vp_x_hi = 0.84
  real, parameter :: cbar_vp_y_lo = img_vp_y_lo
  real, parameter :: cbar_vp_y_hi = img_vp_y_hi

  real, parameter :: w_buffer_frac=0.05

contains

  !=====================================================================
  ! Methods of t_histo:

  !.....................................................................
  subroutine init_t_histo(this, xs, n_bins, x_lo, x_hi, expected_integral)
    implicit none

    class(t_histo) :: this
    real, dimension(:), intent(in) :: xs
    integer,            intent(in) :: n_bins
    real,               intent(in) :: x_lo, x_hi
    real, optional,     intent(in) :: expected_integral

    character(*), parameter :: functionName = "init_t_histo"
    real, dimension(:), allocatable :: fracs
    integer :: n_points,ii
    integer, dimension(:), allocatable :: bin_is
    real :: bin_width,scale_factor

    if(present(expected_integral)) this%expected_integral = expected_integral

    allocate(this%bin_edges(n_bins+1))
    fracs = [(real(ii-1)/real(n_bins),ii=1,n_bins+1)]
    this%bin_edges = x_lo*(1.0 - fracs) + x_hi*fracs

    allocate(this%counts(n_bins))
!    allocate(this%ys(n_bins))
!    allocate(this%yerrs(n_bins))

    bin_is = 1 + floor(n_bins*((xs - x_lo)/(x_hi - x_lo)))
    this%n_under = count(bin_is<1)
    do ii=1,n_bins
      this%counts(ii) = count(bin_is==ii)
    end do
    this%n_over  = count(bin_is>n_bins)

    bin_width = this%bin_edges(2) - this%bin_edges(1)
    n_points = size(xs)
    scale_factor = this%expected_integral/real(n_points)/bin_width

    this%ys    =      real(this%counts) *scale_factor
    this%yerrs = sqrt(real(this%counts))*scale_factor

    deallocate(fracs, bin_is)

    this%is_initialized=.true.

  end subroutine init_t_histo

  !.....................................................................
  subroutine destroy_t_histo(this)
    implicit none

    class(t_histo) :: this

    character(*), parameter :: functionName = "destroy_t_histo"

    if(allocated(this%bin_edges))&
      deallocate(this%bin_edges)

    if(allocated(this%counts))&
      deallocate(this%counts)

    if(allocated(this%ys))&
      deallocate(this%ys)

    if(allocated(this%yerrs))&
      deallocate(this%yerrs)

    this%is_initialized=.false.

  end subroutine destroy_t_histo

  !.....................................................................
  subroutine get_t_histo(this, histo_values, bin_edges, n_under, n_over)
    ! Calling routine should deallocate histo_values and bin_edges after use (and destroy the histogram).
    implicit none

    class(t_histo) :: this
    type(t_histo_bin), dimension(:)&
      , allocatable,                   intent(out) :: histo_values
    real, dimension(:)&
      ,         allocatable, optional, intent(out) :: bin_edges
    integer,                 optional, intent(out) :: n_under,n_over

    character(*), parameter :: functionName = "get_t_histo"
    integer :: n_bins
!,ii

    if(.not.this%is_initialized)&
call wrstop(functionName, "You must initialize the histogram first.")

    n_bins = size(this%counts)
    allocate(histo_values(n_bins))
!    do ii=1,n_bins
!      histo_values(ii)%counts = this%counts(ii)
!      histo_values(ii)%x = this%counts(ii)
!      histo_values(ii)%y = this%counts(ii)
!      histo_values(ii)%yerr = this%counts(ii)
!    end do
    histo_values(:)%counts = this%counts(:)
    histo_values(:)%x = 0.5*(this%bin_edges(1:n_bins) + this%bin_edges(2:n_bins+1))
    histo_values(:)%y = this%ys(:)
    histo_values(:)%yerr = this%yerrs(:)

    if(present(bin_edges)) bin_edges = this%bin_edges
    if(present(n_under))   n_under   = this%n_under
    if(present(n_over))    n_over    = this%n_over

  end subroutine get_t_histo

  !=====================================================================
  ! Methods of t_plot:

  !.....................................................................
  subroutine init_t_plot(this, xs, ys, style_str, yerr)
    implicit none

    class(t_plot) :: this
    real, dimension(:),        intent(in) :: xs
    real, dimension(size(xs)), intent(in) :: ys
    character(2), optional,    intent(in) :: style_str
    real, dimension(size(xs))&
      , optional,              intent(in) :: yerr

    character(*), parameter :: functionName = "init_t_plot"
    integer :: n_dims,n_points

    n_points = size(xs)
    n_dims = 2 ! default
    if(present(yerr)) n_dims = 3

    allocate(this%points(n_points,n_dims))
    this%points(:,1) = xs
    this%points(:,2) = ys
    if(present(yerr))&
      this%points(:,3) = yerr

    if(present(style_str)) then
      ! Work out colour_index.
      !
      select case(style_str(1:1))
      case('r')
        this%colour_index = 2
      case('o')
        this%colour_index = 8
      case('y')
        this%colour_index = 7
      case('g')
        this%colour_index = 3
      case('b')
        this%colour_index = 4
      case('v')
        this%colour_index = 12
      case('k')
        this%colour_index = 0
      case('w')
        this%colour_index = 1
      case default
call wrstop(functionName, "Colour letter not recognized.")
      end select

      ! Work out point style
      !
      select case(style_str(2:2))
      case('.')
        this%point_style = -1
      case('+')
        this%point_style = 2
      case('*')
        this%point_style = 3
      case('o')
        this%point_style = 4
      case('x')
        this%point_style = 5
      case('s')
        this%point_style = 6 ! open square
      case('t')
        this%point_style = 7 ! open triangle
      case('d')
        this%point_style = 11 ! open diamond
      case('-')
        this%point_style = 0 ! draw line.
      case default
call wrstop(functionName, "Point symbol not recognized.")
      end select

    else ! style_str not present.
      this%colour_index = 4 ! blue
      this%point_style = 0 ! draw line.
    end if

!***    integer :: line_type=0
!***    logical :: poly=.false.

  end subroutine init_t_plot

  !.....................................................................
  subroutine destroy_t_plot(this, status)
    implicit none

    class(t_plot) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "destroy_t_plot"

    status = 0

    if(allocated(this%points))&
      deallocate(this%points) !*** should get error status etc

  end subroutine destroy_t_plot

  !=====================================================================
  ! Methods of t_quickplot:

  !.....................................................................
  subroutine add_graph_t_quickplot(this, xs, ys, style_str, yerr)
    ! This stores a t_plot object in a linked list.
    implicit none

    class(t_quickplot) :: this
    real, dimension(:),        intent(in) :: xs
    real, dimension(size(xs)), intent(in) :: ys
    character(*), optional,    intent(in) :: style_str
    real, dimension(size(xs))&
      , optional,              intent(in) :: yerr

    character(*), parameter :: functionName = "add_graph_t_quickplot"
    type(t_plot) :: plot
    integer :: status

    call plot%init(xs, ys, style_str, yerr)
    if(.not. associated(this%plots_LL)) allocate(this%plots_LL)
    call this%plots_LL%addObj(plot)
    call plot%destroy(status)

  end subroutine add_graph_t_quickplot

  !.....................................................................
  subroutine add_image_t_quickplot(this, image, x_lo, x_hi, y_lo, y_hi, with_colourbar)
    implicit none

    class(t_quickplot)               :: this
    real, dimension(:,:), intent(in) :: image
!    type(tRange), optional, intent(in) :: x_range,y_range
    real, optional,       intent(in) :: x_lo,x_hi,y_lo,y_hi
    logical, optional,    intent(in) :: with_colourbar

    character(*), parameter :: functionName = "add_image_t_quickplot"
    integer :: size_xi,size_yi

    size_xi = size(image,1)
    size_yi = size(image,2)
    allocate(this%bkg_image(size_xi,size_yi))
    this%bkg_image = image

    this%with_colourbar = .false. ! default
!    if(present(x_range)) this%x_range = x_range
!    if(present(y_range)) this%y_range = y_range
    this%x_range = construct_tRange(real(x_lo, kind=r8), real(x_hi, kind=r8))
    this%y_range = construct_tRange(real(y_lo, kind=r8), real(y_hi, kind=r8))

    if(present(with_colourbar)) this%with_colourbar = with_colourbar

  end subroutine add_image_t_quickplot

  !.....................................................................
  subroutine show_t_quickplot(this)
    implicit none

    class(t_quickplot) :: this

    character(*), parameter :: functionName = "show_t_quickplot"
    integer :: n_plots,ii,jj,PGOPEN,n_points,n_dims,status
!    class(*), pointer :: tmp
    real :: xmin,xmin_this_plot,xmax,xmax_this_plot
    real :: ymin,ymin_this_plot,ymax,ymax_this_plot
    real :: buffer_x,buffer_y,min_val,max_val
    integer :: size_xi,size_yi

    status = 0

    call this%init_plots()
    n_plots = size(this%plots)

    if(n_plots<1 .and..not.allocated(this%bkg_image))&
call wrstop(functionName, "Nothing to plot.")

    ! Open pgplot and set up viewport.
    !
    this%dev_id = PGOPEN(this%pgdev)
    if(this%dev_id <= 0)&
call wrstop(functionName, "Could not open pgplot.")

    if(n_plots>0) then
      call PGSVP(vp_x_lo, vp_x_hi, vp_y_lo, vp_y_hi)
    else
      if(this%with_colourbar) then
        call PGSVP(img_vp_x_lo, img1_vp_x_hi, img_vp_y_lo, img_vp_y_hi)
      else
        call PGSVP(img_vp_x_lo, img0_vp_x_hi, img_vp_y_lo, img_vp_y_hi)
      end if
    end if

    if(allocated(this%bkg_image)) then
      size_xi = size(this%bkg_image,1)
      size_yi = size(this%bkg_image,2)
      min_val = minval(this%bkg_image)
      max_val = maxval(this%bkg_image)

      this%w_x_min = 0.5
      this%w_x_max = real(size_xi)+0.5
      this%w_y_min = 0.5
      this%w_y_max = real(size_yi)+0.5

      call PGSWIN(this%w_x_min, this%w_x_max, this%w_y_min, this%w_y_max)

      call PGGRAY(this%bkg_image, size_xi, size_yi, 1, size_xi, 1, size_yi&
        , max_val, min_val, this%pgTrMatrix)

      if(.not.(this%x_range%lo%isUndefined .or. this%x_range%hi%isUndefined)) then
        this%w_x_min = real(this%x_range%lo%value)
        this%w_x_max = real(this%x_range%hi%value)
      end if
      if(.not.(this%y_range%lo%isUndefined .or. this%y_range%hi%isUndefined)) then
        this%w_y_min = real(this%y_range%lo%value)
        this%w_y_max = real(this%y_range%hi%value)
      end if

      call PGSWIN(this%w_x_min, this%w_x_max, this%w_y_min, this%w_y_max)

print *, "Image min, max:", min_val, max_val

      if(this%with_colourbar) then
        call draw_colour_bar(min_val, max_val, "")
      end if
    end if

    if(n_plots<1) then
      call PGSCI(1)
      call PGBOX("BCNT", 0.0, 0, "BCNT", 0.0, 0)
      call PGCLOS()
      call this%destroy(status)

  return
    end if

    ! Find extrema of plot points.
    !
    do ii=1,n_plots
      if(size(this%plots(ii)%points, 2)>2) then
        ymin_this_plot = minval(this%plots(ii)%points(:,2)-this%plots(ii)%points(:,3))
        ymax_this_plot = maxval(this%plots(ii)%points(:,2)+this%plots(ii)%points(:,3))
      else
        ymin_this_plot = minval(this%plots(ii)%points(:,2))
        ymax_this_plot = maxval(this%plots(ii)%points(:,2))
      end if
      xmin_this_plot = minval(this%plots(ii)%points(:,1))
      xmax_this_plot = maxval(this%plots(ii)%points(:,1))

      if(ii>1) then
        xmin = min(xmin, xmin_this_plot)
        xmax = max(xmax, xmax_this_plot)
        ymin = min(ymin, ymin_this_plot)
        ymax = max(ymax, ymax_this_plot)
      else
        xmin = xmin_this_plot
        xmax = xmax_this_plot
        ymin = ymin_this_plot
        ymax = ymax_this_plot
      end if
    end do

    if(xmax<=xmin .or. ymax<=ymin)&
call wrstop(functionName, "Plots are of zero size.")

    buffer_x = (xmax - xmin)*w_buffer_frac
    buffer_y = (ymax - ymin)*w_buffer_frac
    this%w_x_min = xmin - buffer_x
    this%w_x_max = xmax + buffer_x
    this%w_y_min = ymin - buffer_y
    this%w_y_max = ymax + buffer_y

    ! Draw all the plots.
    !
    call PGSWIN(this%w_x_min, this%w_x_max, this%w_y_min, this%w_y_max)

    do ii=1,n_plots
      call PGSCI(this%plots(ii)%colour_index)
      n_points = size(this%plots(ii)%points, 1)
      n_dims   = size(this%plots(ii)%points, 2)

      if(this%plots(ii)%point_style==0) then ! draw a polyline.
        call PGLINE(n_points, this%plots(ii)%points(:,1), this%plots(ii)%points(:,2))
      else ! draw scatter plot.
        call PGPT(n_points, this%plots(ii)%points(:,1), this%plots(ii)%points(:,2), this%plots(ii)%point_style)
      end if

      if(n_dims>2) then ! draw error bars.
        do jj=1,n_points
          call PGMOVE(this%plots(ii)%points(jj,1), this%plots(ii)%points(jj,2)-this%plots(ii)%points(jj,3))
          call PGDRAW(this%plots(ii)%points(jj,1), this%plots(ii)%points(jj,2)+this%plots(ii)%points(jj,3))
        end do
      end if
    end do

    call PGSCI(1)
    call PGBOX("BCNT", 0.0, 0, "BCNT", 0.0, 0)
    call PGCLOS()

    call this%destroy(status)

  end subroutine show_t_quickplot

  !.....................................................................
  subroutine init_plots_t_quickplot(this)
    implicit none

    class(t_quickplot) :: this

    character(*), parameter :: functionName = "init_plots_t_quickplot"
    integer :: n_plots,ii,status
    class(*), pointer :: tmp

    if(associated(this%plots_LL)) then
      n_plots = this%plots_LL%getNumLinks()
    else
      n_plots = 0
    end if

    allocate(this%plots(n_plots))

    if(n_plots < 1)&
  return

    call this%plots_LL%reset() ! resets the leaf index to the start of the list.

    do ii = 1, n_plots
      tmp => this%plots_LL%currObj()
      select type(tmp)
        type is (t_plot)
          this%plots(ii) = tmp
      end select

      call this%plots_LL%next()
    end do

    call this%deleteLinkedList(status)

  end subroutine init_plots_t_quickplot

  !.....................................................................
  subroutine deleteLinkedList_t_quickplot(this, status)
    implicit none

    class(t_quickplot)     :: this
    integer, intent(inout) ::  status

    character(*), parameter :: functionName = "deleteLinkedList_t_quickplot"
    integer :: stat
    class(*), pointer :: tmp

    if(status > 100 .or. status == s_errRep) return

    if(associated(this%plots_LL)) then
      call this%plots_LL%reset()
      do while (associated(this%plots_LL%currLink))
        tmp => this%plots_LL%currLink%getObj()

        select type (tmp)
          type is (t_plot)
            call tmp%destroy(status)
            if(status == s_errRep)&
  return
        end select

        call this%plots_LL%next()
      end do

      if(associated(tmp)) deallocate(tmp)

      call this%plots_LL%reset()
      stat =  this%plots_LL%delAllLinks()
      if(stat /= 0) then
        call wrerr(functionName, s_err, "deallocation of LinkedList failed")
        status = s_errRep
      end if

      deallocate(this%plots_LL)
    end if

  end subroutine deleteLinkedList_t_quickplot

  !.....................................................................
  subroutine draw_colour_bar(valueLo, valueHi, y_label, isLog)
    implicit none

    real,              intent(in) :: valueLo, valueHi
    character(*),      intent(in) :: y_label
    logical, optional, intent(in) :: isLog

    integer :: ciLo,ciHi
    real :: vp_x_lo_old,vp_x_hi_old,vp_y_lo_old,vp_y_hi_old,oldCharHeight
    real :: xMin_old,xMax_old,yMin_old,yMax_old,yLo0,yLo1
    integer :: fsi_old,ci_old,numCIs,ci,i

    call PGQCIR(ciLo,ciHi)
print *, "+++ draw_colour_bar(). ciLo, ciHi:", ciLo,ciHi

    call PGQVP(0, vp_x_lo_old,vp_x_hi_old,vp_y_lo_old,vp_y_hi_old)
    call PGQWIN(xMin_old, xMax_old, yMin_old, yMax_old)
    call PGQFS(fsi_old)
    call PGQCI(ci_old)
    call PGQCH(oldCharHeight)

    call PGSCH(0.8)
    call PGSVP(cbar_vp_x_lo,cbar_vp_x_hi,cbar_vp_y_lo,cbar_vp_y_hi)
    call PGSWIN(0.0,1.0,0.0,1.0)
    call PGSFS(1)
    numCIs = 1 + ciHi - ciLo
    yLo1 = 0.0
    do i=1,numCIs
!      ci = i - 1 + ciLo
      ci = ciHi + 1 - i
      call PGSCI(ci)
      yLo0 = yLo1
      yLo1 = i/real(numCIs)
      call PGRECT(0.0, 1.0, yLo0, yLo1)
    end do

    call PGSCI(1)
    call PGSFS(2)
    call PGSWIN(0.0,1.0,valueLo,valueHi)

    if(present(isLog)) then
      call PGBOX("BC", 0.0, 0, "BCLMT", 0.0, 0)
    else
      call PGBOX("BC", 0.0, 0, "BCMT", 0.0, 0)
    end if

    call PGLAB("",y_label,"")

    call PGSVP(vp_x_lo_old,vp_x_hi_old,vp_y_lo_old,vp_y_hi_old)
    call PGSWIN(xMin_old, xMax_old, yMin_old, yMax_old)
    call PGSCI(ci_old)
    call PGSFS(fsi_old)
    call PGSCH(oldCharHeight)

  end subroutine draw_colour_bar

  !.....................................................................
  subroutine destroy_t_quickplot(this, status)
    implicit none

    class(t_quickplot) :: this
    integer, intent(inout) :: status

    integer :: n_plots,ii

    if(allocated(this%plots)) then
      n_plots = size(this%plots)
      do ii=1,n_plots
        call this%plots(ii)%destroy(status)
      end do
      deallocate(this%plots)
    end if

    call this%deleteLinkedList(status)

    if(allocated(this%bkg_image)) deallocate(this%bkg_image)

  end subroutine destroy_t_quickplot

end module erolib_quickplot_mod

