! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! Non-method subroutines:
 !  subroutine a_or_b_primitive(range_A, range_B, ored_ranges, num_output_ranges)
 !  subroutine a_and_b_primitive(range_A, range_B, anded_range, with_out)
!  subroutine old_or_new_primitive(range_A, range_B, ored_ranges, num_output_ranges)
!  subroutine old_and_new_primitive(range_A, range_B, anded_range, with_out)
!  subroutine or_ranges_primitive(rangesA, rangesB, outRanges)
!  subroutine and_ranges_primitive(rangesA, rangesB, outRanges)
!  subroutine range_relation(range_A, range_B, bound_A_lo, bound_A_hi)
!  subroutine lo_range_relation(range_A, range_B, bound_A_lo)
!  subroutine hi_range_relation(range_A, range_B, bound_A_hi)
!  subroutine a_or_b(range_A, range_B, ored_ranges, num_output_ranges)
!  subroutine a_and_b(range_A, range_B, anded_range, with_out)
!  subroutine old_or_new(range_A, range_B, ored_ranges, num_output_ranges)
!  subroutine old_and_new(range_A, range_B, anded_range, with_out)
!  subroutine or_ranges(rangesA, rangesB, outRanges, check_legality)
!  subroutine and_ranges(rangesA, rangesB, outRanges, check_legality)
!  subroutine negate_ranges(ranges, outRanges)!, check_legality)
!  logical function a_extends_higher_than_b(range_A, range_B)
!  logical function ranges_are_legal(ranges)
!  logical function ranges_are_equivalent(range_A, range_B)
!  subroutine primitive_to_range(range_in, range_out)
module erolib_ranges_mod

  use erolib_general_mod
  use erolib_sort_mod ! for swap().
  use erolib_str_mod ! for int2str().

!**** TODO: there is still merge_and_* etc stuff in sort_mod. Can these be expressed as ranges?

  implicit none

logical, private, parameter :: doTest=.false.

  private

  type :: tBound
    real(r8) :: value=0.0
    logical :: isUndefined=.true.
  contains
    procedure, non_overridable :: str => str_tBound
  end type tBound

  type, public :: tRange_primitive
    real(r8) :: lo=0.0
    real(r8) :: hi=0.0
  end type tRange_primitive

  type :: tRanges_primitive
    type(tRange_primitive), dimension(:), allocatable :: range_list
  end type tRanges_primitive

  type, public :: tRange
    type(tBound) :: lo,hi
  contains
    procedure, non_overridable :: is_legal => is_legal_tRange
    procedure, non_overridable :: contains_value => contains_value_tRange
    procedure, non_overridable :: print_me => print_me_tRange
    procedure, non_overridable :: str => str_tRange
!    procedure, non_overridable :: construct => construct_tRange
  end type tRange

  type :: tRanges
    type(tRange), dimension(:), allocatable :: range_list
  contains
    procedure, non_overridable :: get_number => get_number_tRanges
    procedure, non_overridable :: is_legal => is_legal_tRanges
  end type tRanges

  type, public, abstract :: t_range_iterator
    logical :: is_initialized=.false.
    integer :: last_index=0
    integer :: n_elements=0
  contains
    procedure(reset_t_range_iterator),    deferred :: reset
    procedure(destroy_t_range_iterator),  deferred :: destroy
    procedure(get_next_t_range_iterator), deferred :: get_next
  end type t_range_iterator

  abstract interface
    subroutine reset_t_range_iterator(this)
      import t_range_iterator
      implicit none
      class(t_range_iterator) :: this
    end subroutine
  end interface

  abstract interface
    subroutine destroy_t_range_iterator(this)
      import t_range_iterator
      implicit none
      class(t_range_iterator) :: this
    end subroutine
  end interface

  abstract interface
    subroutine get_next_t_range_iterator(this, lo, hi, eof)
      import t_range_iterator
      implicit none
      ! It is assumed that this obeys the following conditions:
      !		- hi > lo.
      !		- lo > previous hi, if there was one.
      class(t_range_iterator) :: this
      real,    intent(out) :: lo, hi
      logical, intent(out) :: eof
    end subroutine
  end interface

  public :: a_or_b_primitive, a_and_b_primitive, or_ranges_primitive, and_ranges_primitive
  public :: a_or_b, a_and_b, or_ranges, and_ranges, negate_ranges
  public :: ranges_are_legal, ranges_are_equivalent, primitive_to_range
  public :: construct_tRange

  public :: test_ranges,test_primitive_ranges

  enum, bind(c)
    enumerator :: LT_B_LO=1,EQ_B_LO,BETWEEN_B,EQ_B_HI,GT_B_HI
  end enum
  private :: LT_B_LO,EQ_B_LO,BETWEEN_B,EQ_B_HI,GT_B_HI

contains

  !.....................................................................
  function str_tBound(this) result(str)
    implicit none
    class(tBound) :: this
    character(256) :: str

    if(this%isUndefined)then
      str = "  undefined"
    else
      write(str, '(a,e10.3)') "  ", this%value
    end if

  end function str_tBound

  !.....................................................................
  subroutine a_or_b_primitive(range_A, range_B, ored_ranges, num_output_ranges)
    implicit none
    type(tRange_primitive),               intent(in)  :: range_A, range_B
    type(tRange_primitive), dimension(2), intent(out) :: ored_ranges
    integer,                              intent(out) :: num_output_ranges

    ! Just so as not to leave them floating if num_output_ranges = 1:
    ored_ranges(2)%lo = 0.0
    ored_ranges(2)%hi = 0.0

    if(range_A%lo < range_B%lo) then
      if(range_A%hi < range_B%lo) then
        num_output_ranges = 2
        ored_ranges(1)%lo = range_A%lo
        ored_ranges(1)%hi = range_A%hi
        ored_ranges(2)%lo = range_B%lo
        ored_ranges(2)%hi = range_B%hi
      else if(range_A%hi < range_B%hi) then
        num_output_ranges = 1
        ored_ranges(1)%lo = range_A%lo
        ored_ranges(1)%hi = range_B%hi
      else ! range_A%hi >= range_B%hi
        num_output_ranges = 1
        ored_ranges(1)%lo = range_A%lo
        ored_ranges(1)%hi = range_A%hi
      end if
    else if(range_A%lo < range_B%hi) then
      if(range_A%hi < range_B%hi) then
        num_output_ranges = 1
        ored_ranges(1)%lo = range_B%lo
        ored_ranges(1)%hi = range_B%hi
      else ! range_A%hi >= range_B%hi
        num_output_ranges = 1
        ored_ranges(1)%lo = range_B%lo
        ored_ranges(1)%hi = range_A%hi
      end if
    else ! range_A%lo >= range_B%hi
      num_output_ranges = 2
      ored_ranges(1)%lo = range_B%lo
      ored_ranges(1)%hi = range_B%hi
      ored_ranges(2)%lo = range_A%lo
      ored_ranges(2)%hi = range_A%hi
    end if

  end subroutine a_or_b_primitive

  !.....................................................................
  subroutine a_and_b_primitive(range_A, range_B, anded_range, with_out)
    implicit none
    type(tRange_primitive), intent(in)  :: range_A, range_B
    type(tRange_primitive), intent(out) :: anded_range
    logical,                intent(out) :: with_out

    ! Just so as not to leave them floating if with_out=.false.:
    anded_range%lo = 0.0
    anded_range%hi = 0.0

    if(range_A%lo < range_B%lo) then
      if(range_A%hi < range_B%lo) then
        with_out = .false.
      else if(range_A%hi < range_B%hi) then
        with_out = .true.
        anded_range%lo = range_B%lo
        anded_range%hi = range_A%hi
      else ! range_A%hi >= range_B%hi
        with_out = .true.
        anded_range%lo = range_B%lo
        anded_range%hi = range_B%hi
      end if
    else if(range_A%lo < range_B%hi) then
      if(range_A%hi < range_B%hi) then
        with_out = .true.
        anded_range%lo = range_A%lo
        anded_range%hi = range_A%hi
      else ! range_A%hi >= range_B%hi
        with_out = .true.
        anded_range%lo = range_A%lo
        anded_range%hi = range_B%hi
      end if
    else ! range_A%lo >= range_B%hi
      with_out = .false.
    end if

  end subroutine a_and_b_primitive

  !.....................................................................
  subroutine old_or_new_primitive(range_A, range_B, ored_ranges, num_output_ranges)
    ! This works on the assumption that range_A%hi < range_B%hi. If this is not true, the routine will fail.
    implicit none
    type(tRange_primitive),               intent(in)  :: range_A, range_B
    type(tRange_primitive), dimension(2), intent(out) :: ored_ranges
    integer,                              intent(out) :: num_output_ranges

    ! Just so as not to leave them floating if num_output_ranges = 1:
    ored_ranges(2)%lo = 0.0
    ored_ranges(2)%hi = 0.0

    if(range_A%lo < range_B%lo) then
      ! Possibilities:
      !		A:   |---|
      !		A:   |------|
      !		A:   |--------|
      !		B:          |------|
      if(range_A%hi < range_B%lo) then
        !	A:   |---|
        !	B:          |------|
        num_output_ranges = 2
        ored_ranges(1)%lo = range_A%lo
        ored_ranges(1)%hi = range_A%hi
        ored_ranges(2)%lo = range_B%lo
        ored_ranges(2)%hi = range_B%hi
      else ! assume range_B%lo <= range_A%hi < range_B%hi
        ! Possibilities:
        !	A:   |------|
        !	A:   |--------|
        !	B:          |------|
        num_output_ranges = 1
        ored_ranges(1)%lo = range_A%lo
        ored_ranges(1)%hi = range_B%hi
      end if
    else ! assume range_B%lo <= range_A%lo < range_A%hi < range_B%hi
      ! Possibilities:
      !		A:   |------|
      !		A:      |---|
      !		B:   |--------|
      num_output_ranges = 1
      ored_ranges(1)%lo = range_B%lo
      ored_ranges(1)%hi = range_B%hi
    end if

  end subroutine old_or_new_primitive

  !.....................................................................
  subroutine old_and_new_primitive(range_A, range_B, anded_range, with_out)
    ! This works on the assumption that range_A%hi < range_B%hi. If this is not true, the routine will fail.
    implicit none
    type(tRange_primitive), intent(in)  :: range_A, range_B
    type(tRange_primitive), intent(out) :: anded_range
    logical,                intent(out) :: with_out

    ! Just so as not to leave them floating if with_out=.false.:
    anded_range%lo = 0.0
    anded_range%hi = 0.0

    if(range_A%lo < range_B%lo) then
      ! Possibilities:
      !		A:   |---|
      !		A:   |------|
      !		A:   |--------|
      !		B:          |------|
      if(range_A%hi <= range_B%lo) then
        ! Possibilities:
        !	A:   |---|
        !	A:   |------|
        !	B:          |------|
        with_out = .false.
      else ! assume range_B%lo < range_A%hi < range_B%hi
        !	A:   |--------|
        !	B:          |------|
        with_out = .true.
        anded_range%lo = range_B%lo
        anded_range%hi = range_A%hi
      end if
    else ! assume range_B%lo <= range_A%lo < range_A%hi < range_B%hi
      ! Possibilities:
      !		A:   |------|
      !		A:      |---|
      !		B:   |--------|
      with_out = .true.
      anded_range%lo = range_A%lo
      anded_range%hi = range_A%hi
    end if

  end subroutine old_and_new_primitive

  !.....................................................................
  subroutine or_ranges_primitive(rangesA, rangesB, outRanges)
    implicit none

    type(tRange_primitive), dimension(:),              intent(in)  :: rangesA, rangesB
    type(tRange_primitive), dimension(:), allocatable, intent(out) :: outRanges

    character(*), parameter :: functionName = "or_ranges_primitive"
    type(tRanges_primitive), dimension(2) :: ranges
    integer, dimension(2) :: ind,maxind
    integer :: kk,ii,olderI,newerI,num_output_ranges,start_ind_new
    type(tRange_primitive), dimension(2) :: ored_ranges

    maxind(1) = size(rangesA)
    maxind(2) = size(rangesB)

    if(maxind(1)<1) then
      allocate(outRanges(maxind(2)))
      outRanges(:) = rangesB(:)
  return
    end if

    if(maxind(2)<1) then
      allocate(outRanges(maxind(1)))
      outRanges(:) = rangesA(:)
  return
    end if

    allocate(outRanges(maxind(1)+maxind(2)))

    kk = 1
    ind(:) = 1
    olderI = 1
    newerI = 2

    ranges(olderI)%range_list = rangesA
    ranges(newerI)%range_list = rangesB
    if(ranges(newerI)%range_list(ind(newerI))%hi < ranges(olderI)%range_list(ind(olderI))%hi)&
      call swap(olderI, newerI)

    outRanges(kk) = ranges(olderI)%range_list(ind(olderI))

    do
      call old_or_new_primitive(outRanges(kk)&
        , ranges(newerI)%range_list(ind(newerI)), ored_ranges, num_output_ranges)
      ! Note that num_output_ranges can only be 1 or 2.

      outRanges(kk) = ored_ranges(1)

      ind(olderI) = ind(olderI) + 1

      if(ind(olderI) > maxind(olderI)) then
        if(num_output_ranges==2) then
          start_ind_new = ind(newerI) ! ranges(newerI)%range_list(ind(newerI)) will already equal ored_ranges(2), so we don't have to get the latter explicitly.
        else
          start_ind_new = ind(newerI) + 1
        end if

        do ii=start_ind_new,maxind(newerI)
          kk = kk + 1
          outRanges(kk) = ranges(newerI)%range_list(ii)
        end do
    exit
      end if

      if(ranges(newerI)%range_list(ind(newerI))%hi < ranges(olderI)%range_list(ind(olderI))%hi)&
        call swap(olderI, newerI)

      if(num_output_ranges==2) then
        kk = kk + 1
        outRanges(kk) = ranges(olderI)%range_list(ind(olderI))
      end if
    end do

    outRanges = outRanges(:kk)

  end subroutine or_ranges_primitive

  !.....................................................................
  subroutine and_ranges_primitive(rangesA, rangesB, outRanges)
    implicit none

    type(tRange_primitive), dimension(:),              intent(in)  :: rangesA, rangesB
    type(tRange_primitive), dimension(:), allocatable, intent(out) :: outRanges

    character(*), parameter :: functionName = "and_ranges_primitive"

    type(tRanges_primitive), dimension(2) :: ranges
    integer, dimension(2) :: ind,maxind
    integer :: kk,olderI,newerI
    type(tRange_primitive) :: anded_range
    logical :: with_out

    maxind(1) = size(rangesA)
    maxind(2) = size(rangesB)

    if(maxind(1)<1 .or. maxind(2)<1) then
      allocate(outRanges(0))
  return
    end if

    allocate(outRanges(maxind(1)+maxind(2)))

    kk = 1
    ind(:) = 1
    olderI = 1
    newerI = 2

    ranges(olderI)%range_list = rangesA
    ranges(newerI)%range_list = rangesB
    if(ranges(newerI)%range_list(ind(newerI))%hi < ranges(olderI)%range_list(ind(olderI))%hi)&
      call swap(olderI, newerI)

    outRanges(kk) = ranges(olderI)%range_list(ind(olderI))

    do
      call old_and_new_primitive(outRanges(kk)&
        , ranges(newerI)%range_list(ind(newerI)), anded_range, with_out)

      if(with_out) then
        outRanges(kk) = anded_range
      else
        kk = kk - 1
      end if

      ind(olderI) = ind(olderI) + 1
      if(ind(olderI) > maxind(olderI))&
    exit

      if(ranges(newerI)%range_list(ind(newerI))%hi < ranges(olderI)%range_list(ind(olderI))%hi)&
        call swap(olderI, newerI)

      kk = kk + 1
      outRanges(kk) = ranges(olderI)%range_list(ind(olderI))
    end do

    outRanges = outRanges(:kk)

  end subroutine and_ranges_primitive

  !.....................................................................
  subroutine range_relation(range_A, range_B, bound_A_lo, bound_A_hi)
    implicit none
    type(tRange), intent(in)  :: range_A, range_B
    integer,      intent(out) :: bound_A_lo, bound_A_hi

    call lo_range_relation(range_A, range_B, bound_A_lo)
    call hi_range_relation(range_A, range_B, bound_A_hi)

  end subroutine range_relation

  !.....................................................................
  subroutine lo_range_relation(range_A, range_B, bound_A_lo)
    implicit none
    type(tRange), intent(in)  :: range_A, range_B
    integer,      intent(out) :: bound_A_lo

    if(range_A%lo%isUndefined) then
      if(range_B%lo%isUndefined) then
        bound_A_lo = EQ_B_LO
      else
        bound_A_lo = LT_B_LO
      end if
    else
      if(range_B%lo%isUndefined) then
        if(range_B%hi%isUndefined) then
          bound_A_lo = BETWEEN_B
        else if(range_A%lo%value <  range_B%hi%value) then
          bound_A_lo = BETWEEN_B
        else if(range_A%lo%value == range_B%hi%value) then
          bound_A_lo = EQ_B_HI
        else !  range_A%lo%value >  range_B%hi%value
          bound_A_lo = GT_B_HI
        end if

      else if(range_A%lo%value <  range_B%lo%value) then
        bound_A_lo = LT_B_LO
      else if(range_A%lo%value == range_B%lo%value) then
        bound_A_lo = EQ_B_LO
      else !  range_A%lo%value >  range_B%lo%value
        if(range_B%hi%isUndefined) then
          bound_A_lo = BETWEEN_B
        else if(range_A%lo%value <  range_B%hi%value) then
          bound_A_lo = BETWEEN_B
        else if(range_A%lo%value == range_B%hi%value) then
          bound_A_lo = EQ_B_HI
        else !  range_A%lo%value >  range_B%hi%value
          bound_A_lo = GT_B_HI
        end if
      end if
    end if

  end subroutine lo_range_relation

  !.....................................................................
  subroutine hi_range_relation(range_A, range_B, bound_A_hi)
    implicit none
    type(tRange), intent(in)  :: range_A, range_B
    integer,      intent(out) :: bound_A_hi

    if(range_A%hi%isUndefined) then
      if(range_B%hi%isUndefined) then
        bound_A_hi = EQ_B_HI
      else
        bound_A_hi = GT_B_HI
      end if
    else
      if(range_B%hi%isUndefined) then
        if(range_B%lo%isUndefined) then
          bound_A_hi = BETWEEN_B
        else if(range_A%hi%value >  range_B%lo%value) then
          bound_A_hi = BETWEEN_B
        else if(range_A%hi%value == range_B%lo%value) then
          bound_A_hi = EQ_B_LO
        else !  range_A%hi%value <  range_B%lo%value
          bound_A_hi = LT_B_LO
        end if

      else if(range_A%hi%value >  range_B%hi%value) then
        bound_A_hi = GT_B_HI
      else if(range_A%hi%value == range_B%hi%value) then
        bound_A_hi = EQ_B_HI
      else !  range_A%hi%value <  range_B%hi%value
        if(range_B%lo%isUndefined) then
          bound_A_hi = BETWEEN_B
        else if(range_A%hi%value >  range_B%lo%value) then
          bound_A_hi = BETWEEN_B
        else if(range_A%hi%value == range_B%lo%value) then
          bound_A_hi = EQ_B_LO
        else !  range_A%hi%value <  range_B%lo%value
          bound_A_hi = LT_B_LO
        end if
      end if
    end if

  end subroutine hi_range_relation

  !.....................................................................
  subroutine a_or_b(range_A, range_B, ored_ranges, num_output_ranges)
    implicit none
    type(tRange),               intent(in)  :: range_A, range_B
    type(tRange), dimension(2), intent(out) :: ored_ranges
    integer,                              intent(out) :: num_output_ranges

    integer :: bound_A_lo, bound_A_hi

    call range_relation(range_A, range_B, bound_A_lo, bound_A_hi)

    if(bound_A_lo==LT_B_LO) then
      if(bound_A_hi==LT_B_LO) then
        num_output_ranges = 2
        ored_ranges(1)%lo = range_A%lo
        ored_ranges(1)%hi = range_A%hi
        ored_ranges(2)%lo = range_B%lo
        ored_ranges(2)%hi = range_B%hi
      else if(bound_A_hi==EQ_B_LO .or. bound_A_hi==BETWEEN_B) then
        num_output_ranges = 1
        ored_ranges(1)%lo = range_A%lo
        ored_ranges(1)%hi = range_B%hi
      else ! range_A%hi >= range_B%hi
        num_output_ranges = 1
        ored_ranges(1)%lo = range_A%lo
        ored_ranges(1)%hi = range_A%hi
      end if
    else if(bound_A_lo==EQ_B_LO .or. bound_A_lo==BETWEEN_B) then
      if(bound_A_hi==BETWEEN_B) then
        num_output_ranges = 1
        ored_ranges(1)%lo = range_B%lo
        ored_ranges(1)%hi = range_B%hi
      else ! range_A%hi >= range_B%hi
        num_output_ranges = 1
        ored_ranges(1)%lo = range_B%lo
        ored_ranges(1)%hi = range_A%hi
      end if
    else ! range_A%lo >= range_B%hi
      num_output_ranges = 2
      ored_ranges(1)%lo = range_B%lo
      ored_ranges(1)%hi = range_B%hi
      ored_ranges(2)%lo = range_A%lo
      ored_ranges(2)%hi = range_A%hi
    end if

  end subroutine a_or_b

  !.....................................................................
  subroutine a_and_b(range_A, range_B, anded_range, with_out)
    implicit none
    type(tRange), intent(in)  :: range_A, range_B
    type(tRange), intent(out) :: anded_range
    logical,                intent(out) :: with_out

    integer :: bound_A_lo, bound_A_hi

    call range_relation(range_A, range_B, bound_A_lo, bound_A_hi)

    if(bound_A_lo==LT_B_LO) then
      if(bound_A_hi==LT_B_LO) then
        with_out = .false.
      else if(bound_A_hi==EQ_B_LO .or. bound_A_hi==BETWEEN_B) then
        with_out = .true.
        anded_range%lo = range_B%lo
        anded_range%hi = range_A%hi
      else ! range_A%hi >= range_B%hi
        with_out = .true.
        anded_range%lo = range_B%lo
        anded_range%hi = range_B%hi
      end if
    else if(bound_A_lo==EQ_B_LO .or. bound_A_lo==BETWEEN_B) then
      if(bound_A_hi==BETWEEN_B) then
        with_out = .true.
        anded_range%lo = range_A%lo
        anded_range%hi = range_A%hi
      else ! range_A%hi >= range_B%hi
        with_out = .true.
        anded_range%lo = range_A%lo
        anded_range%hi = range_B%hi
      end if
    else ! range_A%lo >= range_B%hi
      with_out = .false.
    end if

  end subroutine a_and_b

  !.....................................................................
  subroutine old_or_new(range_A, range_B, ored_ranges, num_output_ranges)
    ! This works on the assumption that old_range%hi < new_range%hi. If this is not true, the routine will fail.
    implicit none
    type(tRange),               intent(in)  :: range_A, range_B
    type(tRange), dimension(2), intent(out) :: ored_ranges
    integer,                    intent(out) :: num_output_ranges

    integer :: bound_A_lo, bound_A_hi

    call range_relation(range_A, range_B, bound_A_lo, bound_A_hi)

    if(bound_A_lo==LT_B_LO) then
      ! Possibilities:
      !		A:   |---|
      !		A:   |------|
      !		A:   |--------|
      !		B:          |------|
      if(bound_A_hi==LT_B_LO) then
        !	A:   |---|
        !	B:          |------|
        num_output_ranges = 2
        ored_ranges(1)%lo = range_A%lo
        ored_ranges(1)%hi = range_A%hi
        ored_ranges(2)%lo = range_B%lo
        ored_ranges(2)%hi = range_B%hi
      else ! assume range_B%lo <= range_A%hi < range_B%hi
        ! Possibilities:
        !	A:   |------|
        !	A:   |--------|
        !	B:          |------|
        num_output_ranges = 1
        ored_ranges(1)%lo = range_A%lo
        ored_ranges(1)%hi = range_B%hi
      end if
    else ! assume range_B%lo <= range_A%lo < range_A%hi < range_B%hi
      ! Possibilities:
      !		A:   |------|
      !		A:      |---|
      !		B:   |--------|
      num_output_ranges = 1
      ored_ranges(1)%lo = range_B%lo
      ored_ranges(1)%hi = range_B%hi
    end if

  end subroutine old_or_new

  !.....................................................................
  subroutine old_and_new(range_A, range_B, anded_range, with_out)
    ! This works on the assumption that old_range%hi < new_range%hi. If this is not true, the routine will fail.
    implicit none
    type(tRange), intent(in)  :: range_A, range_B
    type(tRange), intent(out) :: anded_range
    logical,      intent(out) :: with_out

    integer :: bound_A_lo, bound_A_hi

    call range_relation(range_A, range_B, bound_A_lo, bound_A_hi)

    if(bound_A_lo==LT_B_LO) then
      ! Possibilities:
      !		A:   |---|
      !		A:   |------|
      !		A:   |--------|
      !		B:          |------|
      if(bound_A_hi==LT_B_LO .or. bound_A_hi==EQ_B_LO) then
        ! Possibilities:
        !	A:   |---|
        !	A:   |------|
        !	B:          |------|
        with_out = .false.
      else ! assume range_B%lo < range_A%hi < range_B%hi
        !	A:   |--------|
        !	B:          |------|
        with_out = .true.
        anded_range%lo = range_B%lo
        anded_range%hi = range_A%hi
      end if
    else ! assume range_B%lo <= range_A%lo < range_A%hi < range_B%hi
      ! Possibilities:
      !		A:   |------|
      !		A:      |---|
      !		B:   |--------|
      with_out = .true.
      anded_range%lo = range_A%lo
      anded_range%hi = range_A%hi
    end if

  end subroutine old_and_new

  !.....................................................................
  subroutine or_ranges(rangesA, rangesB, outRanges, check_legality)
    implicit none

    type(tRange), dimension(:),              intent(in)  :: rangesA, rangesB
    type(tRange), dimension(:), allocatable, intent(out) :: outRanges
    logical, optional,                       intent(in)  :: check_legality

    character(*), parameter :: functionName = "or_ranges"
    logical :: local_check_legality
    type(tRanges), dimension(2) :: ranges
    integer, dimension(2) :: ind,maxind
    integer :: kk,ii,olderI,newerI,num_output_ranges,newer_bound_hi,start_ind_new
    type(tRange), dimension(2) :: ored_ranges

    if(present(check_legality)) then
      local_check_legality = check_legality
    else
      local_check_legality = .true.
    end if

    if(local_check_legality) then
      if(.not.(ranges_are_legal(rangesA) .and. ranges_are_legal(rangesB)))&
call wrstop(functionName, "One or both of the input lists of ranges are not legal.")
    end if

    maxind(1) = size(rangesA)
    maxind(2) = size(rangesB)

    if(maxind(1)<1) then
      allocate(outRanges(maxind(2)))
      outRanges(:) = rangesB(:)
  return
    end if

    if(maxind(2)<1) then
      allocate(outRanges(maxind(1)))
      outRanges(:) = rangesA(:)
  return
    end if

    allocate(outRanges(maxind(1)+maxind(2)))

    ind(:) = 1
    olderI = 1
    newerI = 2

    ranges(olderI)%range_list = rangesA
    ranges(newerI)%range_list = rangesB

    call hi_range_relation(ranges(newerI)%range_list(ind(newerI))&
      , ranges(olderI)%range_list(ind(olderI)), newer_bound_hi)

    if(.not.(newer_bound_hi==EQ_B_HI .or. newer_bound_hi==GT_B_HI))&
      call swap(olderI, newerI)

    outRanges(1) = ranges(olderI)%range_list(ind(olderI))

    kk = 1
    do
      call old_or_new(outRanges(kk)&
        , ranges(newerI)%range_list(ind(newerI)), ored_ranges, num_output_ranges)
      ! Note that num_output_ranges can only be 1 or 2.

      outRanges(kk) = ored_ranges(1)

      ind(olderI) = ind(olderI) + 1

      if(ind(olderI) > maxind(olderI)) then
        if(num_output_ranges==2) then
          start_ind_new = ind(newerI) ! ranges(newerI)%range_list(ind(newerI)) will already equal ored_ranges(2), so we don't have to get the latter explicitly.
        else
          start_ind_new = ind(newerI) + 1
        end if

        do ii=start_ind_new,maxind(newerI)
          kk = kk + 1
          outRanges(kk) = ranges(newerI)%range_list(ii)
        end do
    exit
      end if

      call hi_range_relation(ranges(newerI)%range_list(ind(newerI))&
        , ranges(olderI)%range_list(ind(olderI)), newer_bound_hi)

      if(.not.(newer_bound_hi==EQ_B_HI .or. newer_bound_hi==GT_B_HI))&
        call swap(olderI, newerI)

      if(num_output_ranges==2) then
        kk = kk + 1
        outRanges(kk) = ranges(olderI)%range_list(ind(olderI))
      end if
    end do

    outRanges = outRanges(:kk)

  end subroutine or_ranges

  !.....................................................................
  subroutine and_ranges(rangesA, rangesB, outRanges, check_legality)
    implicit none

    type(tRange), dimension(:),              intent(in)  :: rangesA, rangesB
    type(tRange), dimension(:), allocatable, intent(out) :: outRanges
    logical, optional,                       intent(in)  :: check_legality

    character(*), parameter :: functionName = "and_ranges"
    logical :: local_check_legality
    type(tRanges), dimension(2) :: ranges
    integer, dimension(2) :: ind,maxind
    integer :: kk,olderI,newerI,newer_bound_hi
    type(tRange) :: anded_range
    logical :: with_out

    if(present(check_legality)) then
      local_check_legality = check_legality
    else
      local_check_legality = .true.
    end if

    if(local_check_legality) then
      if(.not.(ranges_are_legal(rangesA) .and. ranges_are_legal(rangesB)))&
call wrstop(functionName, "One or both of the input lists of ranges are not legal.")
    end if

    maxind(1) = size(rangesA)
    maxind(2) = size(rangesB)

    if(maxind(1)<1 .or. maxind(2)<1) then
      allocate(outRanges(0))
  return
    end if

    allocate(outRanges(maxind(1)+maxind(2)))

    kk = 1
    ind(:) = 1
    olderI = 1
    newerI = 2

    ranges(olderI)%range_list = rangesA
    ranges(newerI)%range_list = rangesB
    call hi_range_relation(ranges(newerI)%range_list(ind(newerI))&
      , ranges(olderI)%range_list(ind(olderI)), newer_bound_hi)

    if(.not.(newer_bound_hi==EQ_B_HI .or. newer_bound_hi==GT_B_HI))&
      call swap(olderI, newerI)

    outRanges(kk) = ranges(olderI)%range_list(ind(olderI))

    do
      call old_and_new(outRanges(kk), ranges(newerI)%range_list(ind(newerI))&
        , anded_range, with_out)

      if(with_out) then
        outRanges(kk) = anded_range
      else
        kk = kk - 1
      end if

      ind(olderI) = ind(olderI) + 1
      if(ind(olderI) > maxind(olderI))&
    exit

      call hi_range_relation(ranges(newerI)%range_list(ind(newerI))&
        , ranges(olderI)%range_list(ind(olderI)), newer_bound_hi)

      if(.not.(newer_bound_hi==EQ_B_HI .or. newer_bound_hi==GT_B_HI))&
        call swap(olderI, newerI)

      kk = kk + 1
      outRanges(kk) = ranges(olderI)%range_list(ind(olderI))
    end do

    outRanges = outRanges(:kk)

  end subroutine and_ranges

  !.....................................................................
  subroutine negate_ranges(ranges, not_ranges)!, check_legality)
    implicit none

    type(tRange), dimension(:),              intent(in)  :: ranges
    type(tRange), dimension(:), allocatable, intent(out) :: not_ranges
!    logical, optional,                       intent(in)  :: check_legality

    character(*), parameter :: functionName = "negate_ranges"
    integer :: num_ranges,num_not_ranges,nri,ri

    if(.not.(ranges_are_legal(ranges)))&
call wrstop(functionName, "The input list of ranges is not legal.")

    num_ranges = size(ranges)
    if(num_ranges<1) then ! includes nothing.
      allocate(not_ranges(1)) ! by default, both bounds of the new tRange element are undefined - i.e., includes everything.
  return
    end if

    num_not_ranges = num_ranges + 1 ! default
    if(ranges(         1)%lo%isUndefined) num_not_ranges = num_not_ranges - 1
    if(ranges(num_ranges)%hi%isUndefined) num_not_ranges = num_not_ranges - 1

    allocate(not_ranges(num_not_ranges))

    nri = 0
    ri = 1
    if(.not.ranges(ri)%lo%isUndefined) then
      nri = nri + 1
      not_ranges(nri)%lo%isUndefined = .true.
      not_ranges(nri)%hi%isUndefined = .false.
      not_ranges(nri)%hi%value = ranges(ri)%lo%value
    end if

    do ri=1,num_ranges-1
      nri = nri + 1
      not_ranges(nri)%lo%isUndefined = .false.
      not_ranges(nri)%lo%value = ranges(ri  )%hi%value
      not_ranges(nri)%hi%isUndefined = .false.
      not_ranges(nri)%hi%value = ranges(ri+1)%lo%value
    end do

    ri = num_ranges
    if(.not.ranges(ri)%hi%isUndefined) then
      nri = nri + 1
      not_ranges(nri)%lo%isUndefined = .false.
      not_ranges(nri)%lo%value = ranges(ri)%hi%value
      not_ranges(nri)%hi%isUndefined = .true.
    end if

  end subroutine negate_ranges

  !.....................................................................
  logical function a_extends_higher_than_b(range_A, range_B)
    implicit none
    type(tRange), intent(in) :: range_A, range_B

    if(range_A%hi%isUndefined) then
      if(range_B%hi%isUndefined) then
        a_extends_higher_than_b = .false.
      else
        a_extends_higher_than_b = .true.
      end if
    else
      if(range_B%hi%isUndefined) then
        a_extends_higher_than_b = .false.
      else if(range_A%hi%value > range_B%hi%value) then
        a_extends_higher_than_b = .true.
      else
        a_extends_higher_than_b = .false.
      end if
    end if

  end function a_extends_higher_than_b

  !.....................................................................
  logical function is_legal_tRange(this)
    ! This just checks that the high bound is greater than the low bound.
    implicit none
    class(tRange) :: this

    is_legal_tRange = .true. ! default

    if(.not.this%lo%isUndefined .and. .not.this%hi%isUndefined) then
      if(this%lo%value>=this%hi%value) is_legal_tRange = .false.
    end if

  end function is_legal_tRange

  !.....................................................................
  logical function contains_value_tRange(this, value)
    ! This just checks that the high bound is greater than the low bound.
    implicit none
    class(tRange) :: this
    real(r8), intent(in) :: value

    if(this%lo%isUndefined) then
      if(this%hi%isUndefined) then
        contains_value_tRange = .true.
      else if(value >= this%hi%value) then
        contains_value_tRange = .false.
      else
        contains_value_tRange = .true.
      end if
    else if(value <= this%lo%value) then
      contains_value_tRange = .false.
    else
      if(this%hi%isUndefined) then
        contains_value_tRange = .true.
      else if(value >= this%hi%value) then
        contains_value_tRange = .false.
      else
        contains_value_tRange = .true.
      end if
    end if

  end function contains_value_tRange

  !.....................................................................
  subroutine print_me_tRange(this, num_leading_spaces)
    implicit none
    class(tRange) :: this
    integer, optional, intent(in) :: num_leading_spaces

    integer :: local_NLS

    if(present(num_leading_spaces)) then
      local_NLS = num_leading_spaces
    else
      local_NLS = 0
    end if
!*** not yet doing anything with it.

    print *, "Lo:"
    if(this%lo%isUndefined)then
      print *, "  undefined"
    else
      write(*, '(a,e10.3)') "  ", this%lo%value
    end if

    print *, "Hi:"
    if(this%hi%isUndefined)then
      print *, "  undefined"
    else
      write(*, '(a,e10.3)') "  ", this%hi%value
    end if

  end subroutine print_me_tRange

  !.....................................................................
  function str_tRange(this) result(str)
    implicit none
    class(tRange) :: this
    character(256) :: str

    str = "Lo:"//new_line('a')//this%lo%str()//new_line('a')//"Hi:"//trim(this%hi%str())

  end function str_tRange

  !.....................................................................
  function construct_tRange(lo, hi) result(myrange)
    implicit none
    real(r8), optional, intent(in) :: lo, hi
    type(tRange) :: myrange

    if(present(lo)) then
      myrange%lo%isUndefined = .false.
      myrange%lo%value = lo
    end if
    if(present(hi)) then
      myrange%hi%isUndefined = .false.
      myrange%hi%value = hi
    end if

  end function construct_tRange

  !.....................................................................
  integer function get_number_tRanges(this) result(num_ranges)
    implicit none
    class(tRanges) :: this

    num_ranges = 0 ! default

    if(.not.allocated(this%range_list))&
  return

    num_ranges = size(this%range_list)

  end function get_number_tRanges

  !.....................................................................
  logical function is_legal_tRanges(this)
    ! This performs some sanity checks on the list of ranges.
    implicit none
    class(tRanges) :: this

    integer :: num_ranges,ii

    is_legal_tRanges = .true. ! default

    num_ranges = this%get_number()

    if(num_ranges < 1)&
  return

    is_legal_tRanges = .false. ! new default

    do ii=1,num_ranges
      if(.not.this%range_list(ii)%is_legal())& ! checks that lo<hi for each range.
  return
    end do

    if(num_ranges>1) then
      ! A list of ranges does not make sense if internal bounds are undefined. Only bound lo of the 1st range and bound hi of the last are permitted to be undefined.
      ! 
      if(any(this%range_list(1:num_ranges-1)%hi%isUndefined))&
  return
      if(any(this%range_list(2:num_ranges)%lo%isUndefined))&
  return

      ! The final check makes sure that successive ranges don't overlap.
      !
      if(any(this%range_list(1:num_ranges-1)%hi%value >= this%range_list(2:num_ranges)%lo%value))&
  return
    end if

    is_legal_tRanges = .true. ! only if we made it to the finish line.

  end function is_legal_tRanges

  !.....................................................................
  logical function ranges_are_legal(ranges)
    implicit none
    type(tRange), dimension(:), intent(in) :: ranges

    character(*), parameter :: functionName="ranges_are_legal"
    integer :: num_ranges,ii
    character(buffsize) :: iStr0,msg

    num_ranges = size(ranges)
    do ii=1,num_ranges
      if(.not.ranges(ii)%is_legal()) then ! implies that both its bounds are defined.
        iStr0 = int2str(ii)
        write(msg, '(a,e12.5,a,e12.5)') "Range "//trim(iStr0)//" not legal. Lo="&
          , ranges(ii)%lo%value, " hi=", ranges(ii)%hi%value
        call wrerr(functionName, warn, msg)
        ranges_are_legal = .false.
  return
      end if
    end do

    ranges_are_legal = .true.

    if(num_ranges<2)&
  return

    do ii=1,num_ranges-1
      if(ranges(ii)%hi%isUndefined) then
        iStr0 = int2str(ii)
        write(msg, '(a)') "The upper boundary of range "//trim(iStr0)&
          //" is undefined. Internal undefined bounds are not legal."
        call wrerr(functionName, warn, msg)
        ranges_are_legal = .false.
  return
      end if
    end do

    do ii=2,num_ranges
      if(ranges(ii)%lo%isUndefined) then
        iStr0 = int2str(ii)
        write(msg, '(a)') "The lower boundary of range "//trim(iStr0)&
          //" is undefined. Internal undefined bounds are not legal."
        call wrerr(functionName, warn, msg)
        ranges_are_legal = .false.
  return
      end if
    end do

    do ii=1,num_ranges-1
      if(ranges(ii)%hi%value >= ranges(ii+1)%lo%value) then
        iStr0 = int2str(ii)
        write(msg, '(a,e10.3,a,e10.3,a)') "High bound ", ranges(ii)%hi%value&
          , " of range "//trim(iStr0)//" overlaps low bound "&
          , ranges(ii+1)%lo%value, " of the following."
        call wrerr(functionName, warn, msg)
        ranges_are_legal = .false.
  return
      end if
    end do

  end function ranges_are_legal

  !.....................................................................
  logical function ranges_are_equivalent(range_A, range_B)
    implicit none
    type(tRange), intent(in) :: range_A, range_B

    ranges_are_equivalent = .false. ! default

    if(range_A%lo%isUndefined) then
      if(.not.range_B%lo%isUndefined)&
  return
    else
      if(range_B%lo%isUndefined)&
  return
      if(range_A%lo%value /= range_B%lo%value)&
  return
    end if

    if(range_A%hi%isUndefined) then
      if(.not.range_B%hi%isUndefined)&
  return
    else
      if(range_B%hi%isUndefined)&
  return
      if(range_A%hi%value /= range_B%hi%value)&
  return
    end if

    ranges_are_equivalent = .true.

  end function ranges_are_equivalent

  !.....................................................................
  subroutine primitive_to_range(range_in, range_out)
    ! Converts an object of type tRange_primitive into type tRange.
    implicit none

    type(tRange_primitive), intent(in)  :: range_in
    type(tRange),           intent(out) :: range_out

    range_out%lo%value = range_in%lo
    range_out%hi%value = range_in%hi

  end subroutine primitive_to_range

  !.....................................................................
  subroutine test_primitive_ranges()
    use erolib_str_mod

    implicit none

    character(*), parameter :: functionName = "tests"
    integer, parameter :: num_ranges_A=10, num_ranges_B=6
    type(tRange_primitive), dimension(num_ranges_A) :: ranges_p_A
    type(tRange_primitive), dimension(num_ranges_B) :: ranges_p_B
    type(tRange_primitive), dimension(8) :: expected_p_OR
    type(tRange_primitive), dimension(8) :: expected_p_AND
    type(tRange_primitive), dimension(:), allocatable :: actual_p_OR
    type(tRange_primitive), dimension(:), allocatable :: actual_p_AND
    integer :: num_actual_OR,num_actual_AND,ii
    character(buffsize) :: numStr
    logical :: test_passed

    ranges_p_A( 1)%lo = 0.1
    ranges_p_A( 1)%hi = 0.2
    ranges_p_A( 2)%lo = 0.3
    ranges_p_A( 2)%hi = 0.4
    ranges_p_A( 3)%lo = 0.5
      ranges_p_B( 1)%lo = 0.6
    ranges_p_A( 3)%hi = 0.7
    ranges_p_A( 4)%lo = 0.8
    ranges_p_A( 4)%hi = 0.9
    ranges_p_A( 5)%lo = 1.0
      ranges_p_B( 1)%hi = 1.1
    ranges_p_A( 5)%hi = 1.2
    ranges_p_A( 6)%lo = 1.3
      ranges_p_B( 2)%lo = 1.4
      ranges_p_B( 2)%hi = 1.5
      ranges_p_B( 3)%lo = 1.6
      ranges_p_B( 3)%hi = 1.7
    ranges_p_A( 6)%hi = 1.8
      ranges_p_B( 4)%lo = 1.9
    ranges_p_A( 7)%lo = 2.0
    ranges_p_A( 7)%hi = 2.1
      ranges_p_B( 4)%hi = 2.2
    ranges_p_A( 8)%lo = 2.3
      ranges_p_B( 5)%lo = 2.4
    ranges_p_A( 8)%hi = 2.5
    ranges_p_A( 9)%lo = 2.6
      ranges_p_B( 5)%hi = 2.7
    ranges_p_A( 9)%hi = 2.8
      ranges_p_B( 6)%lo = 2.9
      ranges_p_B( 6)%hi = 3.0
    ranges_p_A(10)%lo = 3.1
    ranges_p_A(10)%hi = 3.2

    expected_p_OR(1)%lo = ranges_p_A( 1)%lo
    expected_p_OR(1)%hi = ranges_p_A( 1)%hi
    expected_p_OR(2)%lo = ranges_p_A( 2)%lo
    expected_p_OR(2)%hi = ranges_p_A( 2)%hi
    expected_p_OR(3)%lo = ranges_p_A( 3)%lo
    expected_p_OR(3)%hi = ranges_p_A( 5)%hi
    expected_p_OR(4)%lo = ranges_p_A( 6)%lo
    expected_p_OR(4)%hi = ranges_p_A( 6)%hi
    expected_p_OR(5)%lo = ranges_p_B( 4)%lo
    expected_p_OR(5)%hi = ranges_p_B( 4)%hi
    expected_p_OR(6)%lo = ranges_p_A( 8)%lo
    expected_p_OR(6)%hi = ranges_p_A( 9)%hi
    expected_p_OR(7)%lo = ranges_p_B( 6)%lo
    expected_p_OR(7)%hi = ranges_p_B( 6)%hi
    expected_p_OR(8)%lo = ranges_p_A(10)%lo
    expected_p_OR(8)%hi = ranges_p_A(10)%hi

    expected_p_AND(1)%lo = ranges_p_B( 1)%lo
    expected_p_AND(1)%hi = ranges_p_A( 3)%hi
    expected_p_AND(2)%lo = ranges_p_A( 4)%lo
    expected_p_AND(2)%hi = ranges_p_A( 4)%hi
    expected_p_AND(3)%lo = ranges_p_A( 5)%lo
    expected_p_AND(3)%hi = ranges_p_B( 1)%hi
    expected_p_AND(4)%lo = ranges_p_B( 2)%lo
    expected_p_AND(4)%hi = ranges_p_B( 2)%hi
    expected_p_AND(5)%lo = ranges_p_B( 3)%lo
    expected_p_AND(5)%hi = ranges_p_B( 3)%hi
    expected_p_AND(6)%lo = ranges_p_A( 7)%lo
    expected_p_AND(6)%hi = ranges_p_A( 7)%hi
    expected_p_AND(7)%lo = ranges_p_B( 5)%lo
    expected_p_AND(7)%hi = ranges_p_A( 8)%hi
    expected_p_AND(8)%lo = ranges_p_A( 9)%lo
    expected_p_AND(8)%hi = ranges_p_B( 5)%hi

    call or_ranges_primitive(ranges_p_A, ranges_p_B, actual_p_OR)

    num_actual_OR = size(actual_p_OR)
    if(num_actual_OR/=8) then
      numStr = int2str(num_actual_OR)
call wrstop(functionName, "num_actual_OR should equal 8 but equals "//trim(numStr))
    end if

    test_passed = .true.
    do ii=1,num_actual_OR
      if(  expected_p_OR(ii)%lo/=actual_p_OR(ii)%lo&
      .or. expected_p_OR(ii)%hi/=actual_p_OR(ii)%hi) then
        print *, ii, "Lo:", expected_p_OR(ii)%lo, actual_p_OR(ii)%lo
        print *, ii, "Hi:", expected_p_OR(ii)%hi, actual_p_OR(ii)%hi
        print *, ""
        test_passed = .false.
      end if
    end do

    if(test_passed) print *, "Primitive OR test passed."

    deallocate(actual_p_OR)

    call and_ranges_primitive(ranges_p_A, ranges_p_B, actual_p_AND)

    num_actual_AND = size(actual_p_AND)
    if(num_actual_AND/=8) then
      numStr = int2str(num_actual_AND)
call wrstop(functionName, "num_actual_AND should equal 8 but equals "//trim(numStr))
    end if

    test_passed = .true.
    do ii=1,num_actual_AND
      if(  expected_p_AND(ii)%lo/=actual_p_AND(ii)%lo&
      .or. expected_p_AND(ii)%hi/=actual_p_AND(ii)%hi) then
        print *, ii, "Lo:", expected_p_AND(ii)%lo, actual_p_AND(ii)%lo
        print *, ii, "Hi:", expected_p_AND(ii)%hi, actual_p_AND(ii)%hi
        print *, ""
        test_passed = .false.
      end if
    end do

    if(test_passed) print *, "Primitive AND test passed."
    deallocate(actual_p_AND)

  end subroutine test_primitive_ranges

  !.....................................................................
  subroutine test_ranges()
    use erolib_str_mod

    implicit none

    character(*), parameter :: functionName = "tests"
    integer, parameter :: num_ranges_A=10,num_ranges_B=6,max_num_expected=12
    type(tRange), dimension(num_ranges_A) :: ranges_A
    type(tRange), dimension(num_ranges_B) :: ranges_B
    type(tRange), dimension(:), allocatable :: expected_OR
    type(tRange), dimension(:), allocatable :: expected_AND
    type(tRange), dimension(:), allocatable :: actual_OR
    type(tRange), dimension(:), allocatable :: actual_AND
    integer :: num_actual_OR,num_actual_AND,ii,flagI,num_exp_OR,num_exp_AND
    character(buffsize) :: numStr1,numStr2
    logical :: test_passed

    ranges_A( 1)%lo%value = 0.1
    ranges_A( 1)%hi%value = 0.2
    ranges_A( 2)%lo%value = 0.3
    ranges_A( 2)%hi%value = 0.4
    ranges_A( 3)%lo%value = 0.5
      ranges_B( 1)%lo%value = 0.6
    ranges_A( 3)%hi%value = 0.7
    ranges_A( 4)%lo%value = 0.8
    ranges_A( 4)%hi%value = 0.9
    ranges_A( 5)%lo%value = 1.0
      ranges_B( 1)%hi%value = 1.1
    ranges_A( 5)%hi%value = 1.2
    ranges_A( 6)%lo%value = 1.3
      ranges_B( 2)%lo%value = 1.4
      ranges_B( 2)%hi%value = 1.5
      ranges_B( 3)%lo%value = 1.6
      ranges_B( 3)%hi%value = 1.7
    ranges_A( 6)%hi%value = 1.8
      ranges_B( 4)%lo%value = 1.9
    ranges_A( 7)%lo%value = 2.0
    ranges_A( 7)%hi%value = 2.1
      ranges_B( 4)%hi%value = 2.2
    ranges_A( 8)%lo%value = 2.3
      ranges_B( 5)%lo%value = 2.4
    ranges_A( 8)%hi%value = 2.5
    ranges_A( 9)%lo%value = 2.6
      ranges_B( 5)%hi%value = 2.7
    ranges_A( 9)%hi%value = 2.8
      ranges_B( 6)%lo%value = 2.9
      ranges_B( 6)%hi%value = 3.0
    ranges_A(10)%lo%value = 3.1
    ranges_A(10)%hi%value = 3.2

    allocate(expected_OR( max_num_expected))
    allocate(expected_AND(max_num_expected))

    test_passed = .true.
    do flagI = 0,15
      if(btest(flagI, 0)) then
        ranges_A(1)%lo%isUndefined = .true.
      else
        ranges_A(1)%lo%isUndefined = .false.
      end if

      if(btest(flagI, 1)) then
        ranges_B(1)%lo%isUndefined = .true.
      else
        ranges_B(1)%lo%isUndefined = .false.
      end if

      if(btest(flagI, 2)) then
        ranges_A(num_ranges_A)%hi%isUndefined = .true.
      else
        ranges_A(num_ranges_A)%hi%isUndefined = .false.
      end if

      if(btest(flagI, 3)) then
        ranges_B(num_ranges_B)%hi%isUndefined = .true.
      else
        ranges_B(num_ranges_B)%hi%isUndefined = .false.
      end if

      num_exp_OR = 0
      num_exp_AND = 0
      if(ranges_B(1)%lo%isUndefined) then
        num_exp_OR = num_exp_OR + 1
        expected_OR(num_exp_OR)%lo = ranges_B(1)%lo
        expected_OR(num_exp_OR)%hi = ranges_A(5)%hi

        num_exp_AND = num_exp_AND + 1
        expected_AND(num_exp_AND) = ranges_A(1)
        num_exp_AND = num_exp_AND + 1
        expected_AND(num_exp_AND) = ranges_A(2)
        num_exp_AND = num_exp_AND + 1
        expected_AND(num_exp_AND) = ranges_A(3)

      else
        num_exp_OR = num_exp_OR + 1
        expected_OR(num_exp_OR) = ranges_A(1)
        num_exp_OR = num_exp_OR + 1
        expected_OR(num_exp_OR) = ranges_A(2)
        num_exp_OR = num_exp_OR + 1
        expected_OR(num_exp_OR)%lo = ranges_A(3)%lo
        expected_OR(num_exp_OR)%hi = ranges_A(5)%hi

        num_exp_AND = num_exp_AND + 1
        expected_AND(num_exp_AND)%lo = ranges_B(1)%lo
        expected_AND(num_exp_AND)%hi = ranges_A(3)%hi
      end if

      num_exp_OR = num_exp_OR + 1
      expected_OR(num_exp_OR) = ranges_A( 6)
      num_exp_OR = num_exp_OR + 1
      expected_OR(num_exp_OR) = ranges_B( 4)
      num_exp_OR = num_exp_OR + 1
      expected_OR(num_exp_OR)%lo = ranges_A( 8)%lo
      expected_OR(num_exp_OR)%hi = ranges_A( 9)%hi

      num_exp_AND = num_exp_AND + 1
      expected_AND(num_exp_AND) = ranges_A( 4)
      num_exp_AND = num_exp_AND + 1
      expected_AND(num_exp_AND)%lo = ranges_A( 5)%lo
      expected_AND(num_exp_AND)%hi = ranges_B( 1)%hi
      num_exp_AND = num_exp_AND + 1
      expected_AND(num_exp_AND) = ranges_B( 2)
      num_exp_AND = num_exp_AND + 1
      expected_AND(num_exp_AND) = ranges_B( 3)
      num_exp_AND = num_exp_AND + 1
      expected_AND(num_exp_AND) = ranges_A( 7)
      num_exp_AND = num_exp_AND + 1
      expected_AND(num_exp_AND)%lo = ranges_B( 5)%lo
      expected_AND(num_exp_AND)%hi = ranges_A( 8)%hi
      num_exp_AND = num_exp_AND + 1
      expected_AND(num_exp_AND)%lo = ranges_A( 9)%lo
      expected_AND(num_exp_AND)%hi = ranges_B( 5)%hi

      if(ranges_B(num_ranges_B)%hi%isUndefined) then
        num_exp_OR = num_exp_OR + 1
        expected_OR(num_exp_OR) = ranges_B(num_ranges_B)

        num_exp_AND = num_exp_AND + 1
        expected_AND(num_exp_AND) = ranges_A(num_ranges_A)

      else
        num_exp_OR = num_exp_OR + 1
        expected_OR(num_exp_OR) = ranges_B(num_ranges_B)
        num_exp_OR = num_exp_OR + 1
        expected_OR(num_exp_OR) = ranges_A(num_ranges_A)
      end if

      call or_ranges(ranges_A, ranges_B, actual_OR)

      num_actual_OR = size(actual_OR)
      if(num_actual_OR/=num_exp_OR) then
        numStr1 = int2str(num_exp_OR)
        numStr2 = int2str(num_actual_OR)
call wrstop(functionName, "num_actual_OR should equal "//trim(numStr1)//" but equals "//trim(numStr2))
      end if

      do ii=1,num_actual_OR
        if(.not.ranges_are_equivalent(expected_OR(ii), actual_OR(ii))) then
          print *, ii, "Lo:", expected_OR(ii)%lo, actual_OR(ii)%lo
          print *, ii, "Hi:", expected_OR(ii)%hi, actual_OR(ii)%hi
          print *, ""
          test_passed = .false.
        end if
      end do

      deallocate(actual_OR)

      call and_ranges(ranges_A, ranges_B, actual_AND)

      num_actual_AND = size(actual_AND)
      if(num_actual_AND/=num_exp_AND) then
        numStr1 = int2str(num_exp_AND)
        numStr2 = int2str(num_actual_AND)
call wrstop(functionName, "num_actual_AND should equal "//trim(numStr1)//" but equals "//trim(numStr2))
      end if

      do ii=1,num_actual_AND
        if(.not.ranges_are_equivalent(expected_AND(ii), actual_AND(ii))) then
          print *, ii, "Lo:", expected_AND(ii)%lo, actual_AND(ii)%lo
          print *, ii, "Hi:", expected_AND(ii)%hi, actual_AND(ii)%hi
          print *, ""
          test_passed = .false.
        end if
      end do
      deallocate(actual_AND)
    end do

    if(test_passed) print *, "OR test passed."
    if(test_passed) print *, "AND test passed."

    deallocate(expected_OR)
    deallocate(expected_AND)
  end subroutine test_ranges

end module erolib_ranges_mod
