! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_rawxy_mod
!  use ero_general
  use erolib_general_mod
  use erolib_ffits_mod

  implicit none

  private

  character(*), parameter :: rawxy_ext = "RAWXY_SEL"

  type tRawXYMask
    integer(tel_mask_len), dimension(:,:), allocatable :: values
    integer(tel_mask_len) :: all_bits_set = 0
    character(buffsize) :: extname = ""
!    integer, dimension(2) :: naxis
    integer :: bitpix = 0
    integer :: data_type = tel_mask_len
  contains
    procedure, non_overridable :: init => init_tRawXYMask
    procedure, non_overridable :: read => read_tRawXYMask
    procedure, non_overridable :: create_extension => create_extension_tRawXYMask
    procedure, non_overridable :: write => write_tRawXYMask
    procedure, non_overridable :: free => free_tRawXYMask
  end type tRawXYMask

  type tRawXYMask_logical
    character(buffsize) :: filename = ""
    integer :: unit = 0 ! fitsio unit to access the file
    character(buffsize) :: extname = ""
    logical, dimension(:,:), allocatable :: values
    logical :: containsData=.false.
  contains
    procedure, non_overridable :: open => open_tRawXYMask_logical
    procedure, non_overridable :: read => read_tRawXYMask_logical
    procedure, non_overridable :: close => close_tRawXYMask_logical
    procedure, non_overridable :: free => free_tRawXYMask_logical
  end type tRawXYMask_logical

  public :: rawxy_ext,tRawXYMask,tRawXYMask_logical

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> alloc the content of tRawXYMask type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

  subroutine init_tRawXYMask(this, status)
    implicit none

    integer, intent(inout) :: status

    class(tRawXYMask) :: this

    character(*), parameter :: function = "init_tRawXYMask"
    integer :: ii

    if(status > 100 .or. status == s_errRep) return

    call alloc(this%values, N_PIXELS_CCD, N_PIXELS_CCD, status)
    this%values = 0
    this%extname = ""

    select case (tel_mask_len)
      case (INT8)
        this%bitpix = 8
      case (INT16)
        this%bitpix = 16
      case default
        call wrerr(function, s_err, "Type of RAWXY mask elements is not (yet) supported.")
        status = s_errRep
  return
    end select

    this%all_bits_set = 0
    do ii=1,N_TELESCOPES
      this%all_bits_set = ibset(this%all_bits_set, ii-1)
    end do

  end subroutine init_tRawXYMask

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> read the rawxy selection mask from a fits file
    !
    !
    !
    !> @param[in] this
    !> @param[out] unit
    !> @param[inout] status
    !---------------------------------------------------------------------------

  subroutine read_tRawXYMask(this, unit, extname, exists, status)
    implicit none

    integer,      intent(inout) :: unit
    character(*), intent(in)    :: extname
    logical,      intent(out)   :: exists
    integer,      intent(inout) :: status

    class(tRawXYMask) :: this

    character(*), parameter :: function = "read_tRawXYMask"
    integer, dimension(2) :: naxis
    integer :: old_status

    if(status > 100 .or. status == s_errRep) return

    old_status = status
    status = ignore
    call ffits_movnam_hdu(unit, IMAGE_HDU, extname, status)
    if(status /= ignore) then
      exists = .false.
      status = old_status
  return
    end if

    exists = .true.
    status = old_status

    call FTGISZ(unit, 2, naxis, status)
    if(naxis(1) /= N_PIXELS_CCD .or. naxis(2) /= N_PIXELS_CCD) then
      call wrerr(function, s_err, "RAWXY mask just read has the wrong dimensions.")
      status = s_errRep
  return
    end if

    call this%init(status)
    this%extname = extname
    call ffits_read_img_2D(unit, naxis(1), naxis(1), naxis(2), this%values, status)

  end subroutine read_tRawXYMask

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create an image extension
    !
    !
    !
    !> @param[in] this
    !> @param[out] unit
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine create_extension_tRawXYMask(this, unit, status)
      implicit none

      integer, intent(inout) :: status
      integer, intent(in) :: unit

      class(tRawXYMask) :: this

      character(*), parameter :: function = "create_extension_tRawXYMask"
      integer, dimension(2) :: naxis

      if(status > 100 .or. status == s_errRep) return

      naxis = shape(this%values)

      call ffits_create_img(unit, this%bitpix, size(naxis), naxis, status)
      if(this%extname /= "") call ffits_update_key_str(unit, "EXTNAME", this%extname, "", status)
      ! just to make sure the image extension was created
      call ffits_flush(unit, status)

    end subroutine create_extension_tRawXYMask

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> write the rawxy selection mask to a fits file
    !
    !
    !
    !> @param[in] this
    !> @param[out] unit
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tRawXYMask(this, unit, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: unit

        class(tRawXYMask) :: this

        character(*), parameter :: function = "write_tRawXYMask"
        integer, dimension(2) :: naxis

        if(status > 100 .or. status == s_errRep) return

        if(.not. allocated(this%values)) then
          call wrerr(function, warn, "RAWXY mask is not allocated - nothing to write.")
    return
        end if

        naxis = shape(this%values)

        ! move to the correct extension
        status = ignore
        call ffits_movnam_hdu(unit, IMAGE_HDU, this%extname, status)
        if(status /= ignore) then
            status = ok
            call this%create_extension(unit, status)
        else
            status = ok
        end if

        call ffits_write_img_2D(unit, naxis(1), naxis(1), naxis(2), this%values, status)

    end subroutine write_tRawXYMask

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free the content of tRawXYMask type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tRawXYMask(this, status)
        implicit none

        integer, intent(inout) :: status

        class(tRawXYMask) :: this

        character(*), parameter :: function = "free_tRawXYMask"

        call dealloc(this%values, status)

    end subroutine free_tRawXYMask

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> open an eventfile
  !
  !
  !
  !> @param[in] EventFile, filename
  !> @param[out] iotype
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine open_tRawXYMask_logical(this, iotype, status, delete)
    implicit none

    class(tRawXYMask_logical)     :: this
    integer, intent(in)           :: iotype
    integer, intent(inout)        :: status
    logical, intent(in), optional :: delete

    character(*), parameter :: function = "open_tRawXYMask_logical"

    if(status > 100 .or. status == s_errRep) return

    call ffits_get_iounit(this%unit, status)

    if(iotype /= FF_CREATE) then
      call ffits_open_file(this%unit, this%filename, iotype, status)
    else
      if(present(delete)) then
        if(delete) call delete_file_fortran(this%filename, status)
      end if
      call ffits_create_file(this%unit, this%filename, fblocksize, status)
    end if

  end subroutine open_tRawXYMask_logical

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the rawxy selection mask from a fits file
  !
  !
  !
  !> @param[in] this
  !> @param[out] unit
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine fits_error_handler(functionName, status)
!*** put it somewhere else.

    implicit none

    character(*), intent(in) :: functionName
    integer,      intent(in) :: status

    character(buffSize) :: errtext

    if(status /= 0) then
      call FTGERR(status, errtext)
      call wrstop(functionName, errtext)
    end if

  end subroutine fits_error_handler

  subroutine read_tRawXYMask_logical(this, status)
    implicit none

    class(tRawXYMask_logical) :: this
    integer, intent(inout) :: status

    character(*), parameter :: function = "read_tRawXYMask_logical"
    integer :: hdutype,bitpix,ierr
    integer, dimension(2) :: naxes
    logical :: anyf
    integer(i1), dimension(:,:), allocatable :: i1_image
    integer(i2), dimension(:,:), allocatable :: i2_image
    integer(i4), dimension(:,:), allocatable :: i4_image
!    integer(i8), dimension(:,:), allocatable :: i8_image
    real(r4),    dimension(:,:), allocatable :: r4_image
    real(r8),    dimension(:,:), allocatable :: r8_image

    if(status > 100 .or. status == s_errRep) return

!**** do some check that the file is open.

    if(this%extname == "") then
      ! Move to the primary HDU.
      !
      ierr = 0
      call FTMAHD(this%unit,1,hdutype,ierr)
      call fits_error_handler(function, ierr)

      if(hdutype /= IMAGE_HDU) then
        call wrstop(function, "The primary HDU is coming up as not of type IMAGE_HDU! WTF?")
      end if

    else
!*** move to that extension, check it is an image.
call wrstop(function, "Not yet implemented.")
    end if

    ierr = 0
    call FTGIDT(this%unit, bitpix,ierr) ! Possible returned values of bitpix are: 8, 16, 32, 64, -32, or -64 corresponding to unsigned byte, signed 2-byte integer, signed 4-byte integer, signed 8-byte integer, real, and double. 
    call fits_error_handler(function, ierr)

    ierr = 0
    call FTGISZ(this%unit, 2, naxes, ierr)
    call fits_error_handler(function, ierr)
    if(naxes(1) /= N_PIXELS_CCD .or. naxes(2) /= N_PIXELS_CCD) then
      call wrerr(function, s_err, "RAWXY mask just read has the wrong dimensions.")
      status = s_errRep
  return
    end if

    call alloc(this%values, naxes(1), naxes(2), status)
    this%values = .true.
    ierr = 0
    select case(bitpix)
      case(8)
        call alloc(i1_image, naxes(1), naxes(2), status)
        call FTG2DB(this%unit, 0, 0, naxes(1), naxes(1), naxes(2), i1_image, anyf, ierr)
        call fits_error_handler(function, ierr)
        where(i1_image==0_i1) this%values = .false.
        call dealloc(i1_image, status)
      case(16)
        call alloc(i2_image, naxes(1), naxes(2), status)
        call FTG2DI(this%unit, 0, 0, naxes(1), naxes(1), naxes(2), i2_image, anyf, ierr)
        call fits_error_handler(function, ierr)
        where(i2_image==0_i2) this%values = .false.
        call dealloc(i2_image, status)
      case(32)
        call alloc(i4_image, naxes(1), naxes(2), status)
        call FTG2DJ(this%unit, 0, 0, naxes(1), naxes(1), naxes(2), i4_image, anyf, ierr)
        call fits_error_handler(function, ierr)
        where(i4_image==0_i4) this%values = .false.
        call dealloc(i4_image, status)
!      case(64)
!        call alloc(i8_image, naxes(1), naxes(2), status)
!        call FTG2DK(this%unit, 0, 0, naxes(1), naxes(1), naxes(2), i8_image, anyf, ierr)
!        call fits_error_handler(function, ierr)
!        where(i8_image==0_i8) this%values = .false.
!        call dealloc(i8_image, status)
      case(-32)
        call alloc(r4_image, naxes(1), naxes(2), status)
        call FTG2DE(this%unit, 0, 0_r4, naxes(1), naxes(1), naxes(2), r4_image, anyf, ierr)
        call fits_error_handler(function, ierr)
        where(r4_image==0.0_r4) this%values = .false.
        call dealloc(r4_image, status)
      case(-64)
        call alloc(r8_image, naxes(1), naxes(2), status)
        call FTG2DD(this%unit, 0, 0_r8, naxes(1), naxes(1), naxes(2), r8_image, anyf, ierr)
        call fits_error_handler(function, ierr)
        where(r8_image==0.0_r8) this%values = .false.
        call dealloc(r8_image, status)
      case default
        call wrerr(function, s_err, "Unrecognized data type of RAWXY mask.")
        status = s_errRep
  return
    end select

  end subroutine read_tRawXYMask_logical

  !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> close the tImage file
  !
  !
  !
  !> @param[inout] this
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine close_tRawXYMask_logical(this, status)
    implicit none

    class(tRawXYMask_logical) :: this
    integer, intent(inout) :: status

    character(*), parameter :: function = "close_tRawXYMask_logical"

    if(status > 100 .or. status == s_errRep) return

    call ffits_close_file(this%unit, status)
    call ffits_free_iounit(this%unit, status)

  end subroutine close_tRawXYMask_logical

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free the content of tRawXYMask_logical type
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tRawXYMask_logical(this, status)
        implicit none

        class(tRawXYMask_logical) :: this
        integer, intent(inout) :: status

        character(*), parameter :: function = "free_tRawXYMask_logical"

        call dealloc(this%values, status)

    end subroutine free_tRawXYMask_logical
end module erolib_rawxy_mod
