! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
!Non-method subroutines:
!  subroutine dealloc_regRows(regRows, status)
!  subroutine alloc_regRows(regRows, n, status)
!  function convert_regcomp_treg(regcomp, status) result(regRow)
!  subroutine convert_list2regRows(list_d, regRows, status)
!  subroutine addCircleRegionFK5(list_d, regTok, reject, warn, status, comment)
!  subroutine addRegion(list_d, regTok, status)
!  subroutine deleteLinkedListRegions(list_d, status)
!  function convert_regtype_tregtable(regtype, status)
!  subroutine alloc_regions(list_of_regions, n, status)
!  subroutine dealloc_regions(list_of_regions, status)
!  subroutine getRegionsFromStr(str_in, regionsObj, status)
!  subroutine interpret_ASCII_regions_circle(filename, regionsObj, nreject, status, comment)
!  subroutine interpret_ASCII_regions(filename, regionsObj, status)
!  subroutine interpret_expression_regions(expression, regionsObj, status)
!  subroutine str2deg(str, deg, status, convert)
!  subroutine validateShapeParams(rshape, n, status)

module erolib_region_mod

  use ISO_FORTRAN_ENV ! gfortran module for real64 and other definitions
  use param
  use erolib_general_mod
  use erolib_ffits_mod
  use erolib_bintable_mod
  use erolib_str_mod
  use erolib_sort_mod
  use erolib_linklist_mod
  use detect_io, only : calc_sourceowner
  use srctool_generic_regions, only : region_type, regcomp_type, &
        srctool_regions_init_region, srctool_regions_init_regcomp, srctool_regions_add_regcomp, &
        REGCOMP_SHAPE_CODE_CIRCLE, REGCOMP_SHAPE_CODE_ELLIPSE, REGCOMP_SHAPE_CODE_BOX, &
        REGCOMP_SHAPE_CODE_ANNULUS, REGCOMP_SHAPE_CODE_ELLANNULUS, REGCOMP_SHAPE_CODE_RECTANGLE, &
        REGCOMP_NUM_PARS_CIRCLE, REGCOMP_NUM_PARS_ELLIPSE, REGCOMP_NUM_PARS_BOX, &
        REGCOMP_NUM_PARS_ANNULUS, REGCOMP_NUM_PARS_ELLANNULUS, REGCOMP_NUM_PARS_RECTANGLE
  use srctool_generic, only : &
        COORDSYS_CODE_ICRS, COORDSYS_CODE_FK4, COORDSYS_CODE_GAL, COORDSYS_CODE_ECL, COORDSYS_CODE_IMAGE, &
        COORDSYS_CODE_IMAGE, COORDSYS_CODE_PHYSICAL, COORDSYS_CODE_UNITVEC, COORDSYS_CODE_ERROR

  implicit none

!logical, parameter, private :: doTest=.false.

  ! We need 4 objects because, although tFitsFile/tBinTable is the standard paring, many of the old routines were written in terms of an object equivalent to a single row in the region table and a second object that contained a list of such rows. In order to avoid re-writing the module completely I have merged these two schemes and added appropriate code to translate between them.
  !
  ! Note that region information is stored internally in the tRegRow objects, but translated to or from a tRegTab object when respectively writing to or reading from a FITS region file.

  type, extends(tBinTable) :: tRegTab
  contains
    procedure, private :: init => init_tRegTab
  end type tRegTab

  type, extends(tFitsFile) :: tRegFile
    type(tRegTab) :: regext
  contains
  end type tRegFile

  type tRegRow
    character(16) :: rshape
    real(r8), allocatable, dimension(:) :: x
    real(r8), allocatable, dimension(:) :: y
    real(r8), allocatable, dimension(:) :: r
    real(r8), allocatable, dimension(:) :: rotang
    character(buffsize) :: wcs
    logical :: add
    ! optional / additional parameter
    integer :: skyfield
    integer :: order
  contains
    procedure, non_overridable :: fill => fill_tRegRow
    procedure, non_overridable :: getEncapsulCircleReg => getEncapsulCircleReg_tRegRow
    procedure, non_overridable :: getRegExp => getRegExp_tRegRow
    procedure, non_overridable :: free => free_tRegRow
    procedure, non_overridable :: get_srctool_regcomp
  end type tRegRow

  type tRegions
    type(tRegFile) :: reg_file
    type(tRegRow), allocatable, dimension(:) :: regRows
    logical :: containsData=.false.
  contains
    procedure, non_overridable :: free => free_tRegions
    procedure, non_overridable :: read => read_tRegions
    procedure, non_overridable :: write => write_tRegions
    procedure, non_overridable :: get_srctool_region
    procedure, non_overridable :: get_srctool_encap_circle_region
    procedure, non_overridable :: sort_skyfield => sort_skyfield_tRegions
  end type tRegions

contains

  !=====================================================================
  ! Methods of tRegTab:

  !.....................................................................
  subroutine init_tRegTab(this, status, fileInfo, extname)
    implicit none

    class(tRegTab) :: this
    integer, intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo
    character(*),    intent(in), optional         :: extname

    character(*), parameter :: functionName = "init_tRegTab"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    if(present(extname)) then
      this%info%tableName = extname
    else
      this%info%tableName = "REGION"
    end if

    call this%initColumn("A", "SHAPE",  "", num_chars=16)
    call this%initColumn("D", "XC",     "deg")
    call this%initColumn("D", "YC",     "deg")
    call this%initColumn("D", "R",      "deg")
    call this%initColumn("D", "ROTANG", "deg")

    call this%initColumns(status)


  end subroutine init_tRegTab

  !=====================================================================
  ! Methods of tRegRow:

  !.....................................................................
  subroutine fill_tRegRow(this, rshape, xs, ys, status, rs, rotangs, add, wcs)
    implicit none

    class(tRegRow) :: this
    character(*),                     intent(in)    :: rshape
    real(r8), dimension(:),           intent(in)    :: xs, ys
    integer,                          intent(inout) :: status
    real(r8), dimension(:), optional, intent(in)    :: rs, rotangs
    logical,                optional, intent(in)    :: add
    character(*),           optional, intent(in)    :: wcs

    character(*), parameter :: functionName = "fill_tRegRow"

    !return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep)&
  return

    ! the default
    this%wcs = "fk5"
    this%add = .true.
    this%rshape = rshape

    !        if(rshape == "box") this%rshape = "rotbox"

    if(this%rshape(1:1) == "!") then
      this%add = .false.
      call strdel_substr(this%rshape, "!")
      if(present(add)) then
        if(add .eqv. .true.) then
          status = s_err
          call wrerr(functionName, status, "conflicting add sign of region and region name")
  return
        end if
      end if
    end if

    if(present(add)) this%add = add
    if(present(wcs)) this%wcs = wcs

    ! allocate x and y
    select case (this%rshape)
      case ("rectangle", "rotrectangle")
                if(size(xs)<2 .or. size(ys)<2) then
                  call wrerr(functionName, s_err, "xs and ys need 2 entries for shape "// trim(rshape))
                  status = s_errRep
  return
                end if

                call alloc(this%x, 2, status)
                call alloc(this%y, 2, status)
                this%x(1) = xs(1)
                this%y(1) = ys(1)
                this%x(2) = xs(2)
                this%y(2) = ys(2)
      case ("polygon")
                call alloc(this%x, size(xs), status)
                this%x = xs
                call alloc(this%y, size(ys), status)
                this%y = ys
      case ("circle", "point", "annulus", "elliptannulus", "pie", &
                "sector", "ellipse", "rotbox", "rotdiamond", "rotrhombus", "box")
                if(size(xs)<1 .or. size(ys)<1) then
                  call wrerr(functionName, s_err, "xs and ys need 1 entry for shape "// trim(rshape))
                  status = s_errRep
  return
                end if

                call alloc(this%x, 1, status)
                call alloc(this%y, 1, status)
                this%x(1) = xs(1)
                this%y(1) = ys(1)
      case default
                call wrerr(functionName, s_err, "region " // trim(rshape) // " is not supported")
                status = s_errRep
    end select

    ! allocate r
    select case (this%rshape)
      case ("point", "rectangle", "rotrectangle", "polygon", "pie", "sector")
                call alloc(this%r, 0, status)
      case ("circle")
                if(.not. present(rs)) then
                  call wrerr(functionName, s_err, "cannot initialize shape "// trim(rshape)//" - no rs provided")
                  status = s_errRep
  return
                end if
                if(size(rs)<1) then
                  call wrerr(functionName, s_err, "rs needs 1 entry for shape "// trim(rshape))
                  status = s_errRep
  return
                end if

                call alloc(this%r, 1, status)
                this%r(1) = rs(1)
      case ("elliptannulus")
                if(.not. present(rs)) then
                  call wrerr(functionName, s_err, "cannot initialize shape "// trim(rshape)//" - no rs provided")
                  status = s_errRep
  return
                end if
                if(size(rs)<4) then
                  call wrerr(functionName, s_err, "rs needs 4 entries for shape "// trim(rshape))
                  status = s_errRep
  return
                end if

                call alloc(this%r, 4, status)
                this%r(1) = rs(1)
                this%r(2) = rs(2)
                this%r(3) = rs(3)
                this%r(4) = rs(4)
      case default
                if(.not. present(rs)) then
                  call wrerr(functionName, s_err, "cannot initialize shape "// trim(rshape)//" - no rs provided")
                  status = s_errRep
  return
                end if
                if(size(rs)<2) then
                  call wrerr(functionName, s_err, "rs needs 2 entries for shape "// trim(rshape))
                  status = s_errRep
  return
                end if

                call alloc(this%r, 2, status)
                this%r(1) = rs(1)
                this%r(2) = rs(2)
    end select

    ! allocate rotang
    select case (this%rshape)
      case("circle", "point", "annulus", "polygon", "rectangle")
                call alloc(this%rotang, 0, status)
      case ("elliptannulus", "pie", "sector")
                if(.not. present(rotangs)) then
                  call wrerr(functionName, s_err, "cannot initialize shape "&
                    // trim(rshape)//" - no rotangs provided")
                  status = s_errRep
  return
                end if
                call alloc(this%rotang, 2, status)
                this%rotang(1) = rotangs(1)
                select case (size(rotangs))
                    case (1)
                        this%rotang(2) = rotangs(1)
                    case (2)
                        this%rotang(2) = rotangs(2)
                    case default
call wrstop(functionName, "invalid number of angles for "// trim(this%rshape))
                end select
      case ("ellipse", "rotbox", "rotrectangle", "rotdiamond", "rotrhombus", "box")
                if(.not. present(rotangs)) then
                  call wrerr(functionName, s_err, "cannot initialize shape "&
                   //trim(rshape)//" - no rotangs provided")
                  status = s_errRep
  return
                end if
                if(size(rotangs)<1) then
                  call wrerr(functionName, s_err, "rotangs needs 1 entry for shape "// trim(rshape))
                  status = s_errRep
  return
                end if

                call alloc(this%rotang, 1, status)
                this%rotang(1) = rotangs(1)
      case default
      ! no rotang needed
    end select

  end subroutine fill_tRegRow

  !.....................................................................
  subroutine free_tRegRow(this, status)
        implicit none

        class(tRegRow) :: this
        integer, intent(inout) :: status

        character(*), parameter :: function = "free_tRegRow"


        !return automatically (fitsio error start at 101, 0 = ok)
        if (status > 100 .or. status == s_errRep) return

        call dealloc(this%x, status)
        call dealloc(this%y, status)
        call dealloc(this%r, status)
        call dealloc(this%rotang, status)

  end subroutine free_tRegRow

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> convert a Region file to a srctool type of handling regions
    !
    !  type :: regcomp_type
    !    integer(kind=int32) :: shape_code
    !    integer(kind=int32) :: coordsys_code
    !    integer(kind=int32) :: num_pars
    !    integer(kind=int32) :: num_pars_read
    !    real(kind=real64)   :: pars(1:MAX_REGCOMP_PARS) ! meaning is shape-specific
    !    logical(kind=bool)  :: wildcards(1:MAX_REGCOMP_PARS) ! set to true if the param should be interpreted as wild card
    !    integer(kind=int32) :: weight ! set to -1 if component is to be subtracted
    !    real(kind=real64)   :: max_radius  ! maximum distance of any point in regcomp from src coords - deg
    !  end type regcomp_type
    !
    !> @param[in] this
    !> @param[out] regComp
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine get_srctool_regcomp(this, regcomp, status)
        ! 20200715 IMS: Note that there used to be minus signs on the values of this%rotang in the cases (ellipse, box, rotbox, elliptannulus) in which the region is rotatable, i.e. has no azimthal symmetry. According to Tom Dwelly, this was done to conform with the convention in ftools. However, the minuses have been removed so as to conform to the convention of ds9. Task srctool also matches the latter convention.
        !
        implicit none

        class(tRegRow) :: this
        type(regcomp_type), intent(out)   :: regcomp
        integer,            intent(inout) :: status

        character(*), parameter :: function = "get_srctool_regcomp"

        if(status > 100 .or. status == s_errRep) return

        call srctool_regions_init_regcomp(regcomp)
        ! convert shape
        select case (this%rshape)
            case ("circle")
                regcomp%shape_code = REGCOMP_SHAPE_CODE_CIRCLE
                regcomp%num_pars = REGCOMP_NUM_PARS_CIRCLE
                regcomp%pars = [this%x, this%y, this%r, this%rotang]
            case ("ellipse")
                regcomp%shape_code = REGCOMP_SHAPE_CODE_ELLIPSE
                regcomp%num_pars = REGCOMP_NUM_PARS_ELLIPSE
                regcomp%pars = [this%x, this%y, this%r, this%rotang]
            case ("box", "rotbox")
                regcomp%shape_code = REGCOMP_SHAPE_CODE_BOX
                regcomp%num_pars = REGCOMP_NUM_PARS_BOX
                regcomp%pars = [this%x, this%y, this%r, this%rotang]
            case ("annulus")
                regcomp%shape_code = REGCOMP_SHAPE_CODE_ANNULUS
                regcomp%num_pars = REGCOMP_NUM_PARS_ANNULUS
                regcomp%pars = [this%x, this%y, this%r, this%rotang]
            case ("elliptannulus")
                regcomp%shape_code = REGCOMP_SHAPE_CODE_ELLANNULUS
                regcomp%num_pars = REGCOMP_NUM_PARS_ELLANNULUS
                regcomp%pars = [this%x, this%y, this%r, this%rotang]
            case ("rectangle")
                regcomp%shape_code = REGCOMP_SHAPE_CODE_RECTANGLE
                regcomp%num_pars = REGCOMP_NUM_PARS_RECTANGLE
                regcomp%pars = [this%x, this%y, this%r, this%rotang]
            case default
                call wrerr(function, s_err, "converting of region " //&
                    trim(this%rshape) // " is not supported")
                status = s_errRep
        end select

        ! convert coordsys
        select case (trim(this%wcs))
            case ( "fk5","j2000","FK5","J2000" )
                regcomp%coordsys_code = COORDSYS_CODE_ICRS
            case ( "fk4","b1950", "FK4","B1950" )
                regcomp%coordsys_code = COORDSYS_CODE_FK4
            case ( "gal","galactic","Gal","GAL" )
                regcomp%coordsys_code = COORDSYS_CODE_GAL
            case ( "ecl","ecliptic","ECL" )
                regcomp%coordsys_code = COORDSYS_CODE_ECL
            case ( "image" )
                regcomp%coordsys_code = COORDSYS_CODE_IMAGE
            case ( "physical" )
                regcomp%coordsys_code = COORDSYS_CODE_PHYSICAL
            case ( "unitvec" )
                regcomp%coordsys_code = COORDSYS_CODE_UNITVEC
            case  default
                regcomp%coordsys_code = COORDSYS_CODE_ERROR
        end select

        ! weight (add or subtract)
        if(this%add .eqv. .true.) then
            regcomp%weight = 1
        else
            regcomp%weight = -1
        end if

        !        regcomp%pars = [this%x, this%y, this%r, this%rotang]
        regcomp%num_pars_read = size(regcomp%pars)

    end subroutine get_srctool_regcomp

  !.....................................................................
  function getRegExp_tRegRow(this, status) result(regExp)
    implicit none

    class(tRegRow) :: this
    integer, intent(inout) :: status
    character(buffsize) :: regExp

    character(*), parameter :: function = "getRegExp_tRegRow"
    integer :: ii, pos

        !return automatically (fitsio error start at 101, 0 = ok)
        if (status > 100 .or. status == s_errRep) return

        call wrlog(function, "WCS keyword can cause problems due to projection - cfitsio error message")
        if(this%add .eqv. .true.) then
            regExp = trim(this%wcs) // "; " // trim(this%rshape) // "("
        else
            regExp = trim(this%wcs) // "; -" // trim(this%rshape) // "("
        end if

        ! check whether x and y of the region are present and mandatory
        if(.not. allocated(this%x)) then
            status = s_err
            call wrerr(function, status, "no x field found for this region")
            return
        end if

        if(.not. allocated(this%y)) then
            status = s_err
            call wrerr(function, status, "no y field found for this region")
            return
        end if

        do ii = 1, size(this%x)
            regExp = trim(regExp) // trim(double2str(this%x(ii), status=status)) // ", "
            ! return if error
            if (status == s_errRep) return
        end do

        do ii = 1, size(this%y)
            regExp = trim(regExp) // trim(double2str(this%y(ii), status=status)) // ", "
            ! return if error
            if (status == s_errRep) return
        end do

        if(allocated(this%r)) then
            do ii = 1, size(this%r)
                regExp = trim(regExp) // trim(double2str(this%r(ii), status=status)) // ", "
                ! return if error
                if (status == s_errRep) return
            end do
        end if

        if(allocated(this%rotang)) then
            do ii = 1, size(this%rotang)
                regExp = trim(regExp) // trim(double2str(this%rotang(ii), status=status)) // ", "
                ! return if error
                if (status == s_errRep) return
            end do
        end if

        pos = index(regExp, ",", .true.)
        if(len_trim(regExp) == pos) regExp = regExp(1:pos-1)

        regExp = trim(regExp) // ");"

  end function getRegExp_tRegRow

  !.....................................................................
  function getEncapsulCircleReg_tRegRow(this, status) result(circRegRow)
    implicit none

    class(tRegRow) :: this
    integer, intent(inout) :: status
    type(tRegRow) :: circRegRow

        character(*), parameter :: function = "getEncapsulCircleReg_tRegRow"

        real(kind=8) :: width, height
        real(kind=8), allocatable, dimension(:) :: x, y, r

        !return automatically (fitsio error start at 101, 0 = ok)
        if (status > 100 .or. status == s_errRep) return

        call alloc(x, 1, status)
        call alloc(y, 1, status)
        call alloc(r, 1, status)

        select case (this%wcs)
            case ('fk5', 'fk4', 'galactic', 'ecliptic')
              ! everythings fine here
            case default
                ! for these WCS no encapsulating circle can be calculated based on eFoV
call wrstop(function, "wrong wcs - must be either fk4/fk5/galactic/ecliptic, for getting an encapsulating circle of region + field of view")
        end select

        select case (this%rshape)
            case ('circle', 'annulus')
                x = this%x(1)
                y = this%y(1)
                r = this%r(1) + eFoV
                call circRegRow%fill('circle', x, y, status, r)
            case ('ellipse', 'elliptannulus')
                x = this%x(1)
                y = this%y(1)
                r = max(this%r(1), this%r(2))+ eFoV
                call circRegRow%fill('circle', x, y, status, r)

            case ('box', 'rotbox')
                x = this%x(1)
                y = this%y(1)
                r = sqrt((this%r(1) * this%r(1)) + (this%r(2) * this%r(2))) / 2.0_dp  + eFoV
                call circRegRow%fill('circle', x, y, status, r)
            case ('rectangle')
                x = (this%x(1) + this%x(2)) / dble(2.0)
                y = (this%y(1) + this%y(2)) / dble(2.0)
                width  = abs(this%x(1) - this%x(2))
                height = abs(this%y(1) - this%y(2))
                r = sqrt(width * width + height * height) / 2.0_dp  + eFoV
                call circRegRow%fill('circle', x, y, status, r)
            case default
                status = s_err
                call wrerr (function, status, "Get encapsulated circular region is not supported for "&
                  //trim(this%rshape))
        end select

  end function getEncapsulCircleReg_tRegRow

  !=====================================================================
  ! The remaining routines are not methods of tRegRow, they're just
  ! utility routines that manipulate the objects.

  !.....................................................................
  subroutine dealloc_regRows(regRows, status)
    implicit none

    type(tRegRow), allocatable, dimension(:), intent(out)   :: regRows
    integer,                                  intent(inout) :: status

    character(*), parameter :: function = "dealloc_regRows"
    integer :: stat,ii

    if(allocated(regRows)) then
      do ii=1,size(regRows)
        call regRows(ii)%free(status)
      end do
      deallocate(regRows, stat=stat)
      call FError(function, stat, "dealloc regRow type", status)
    end if

  end subroutine dealloc_regRows

  !.....................................................................
  subroutine alloc_regRows(regRows, n, status)
    implicit none

    type(tRegRow), allocatable, dimension(:), intent(out)   :: regRows
    integer,                                  intent(in)    :: n
    integer,                                  intent(inout) :: status

        character(*), parameter :: function = "alloc_regRows"
        integer :: stat
        character(buffsize) :: errmsg

        call dealloc_regRows(regRows, status)
        allocate(regRows(n), stat=stat, errmsg=errmsg)
        call FError(function, stat, errmsg, status)

  end subroutine alloc_regRows

  !.....................................................................
  function convert_regcomp_treg(regcomp, status) result(regRow)
    implicit none

    type(regcomp_type), intent(in)    :: regcomp
    integer,            intent(inout) :: status
    type(tRegRow) :: regRow

    character(*), parameter :: function = "convert_regcomp_treg"
    character(buffsize) :: rshape, wcs
    logical :: add

    if(status > 100 .or. status == s_errRep) return

        ! convert coordsys
        select case (regcomp%coordsys_code)
            case (COORDSYS_CODE_ICRS)
                wcs = "fk5"
            case (COORDSYS_CODE_FK4)
                wcs = "fk4"
            case (COORDSYS_CODE_GAL)
                wcs = "galactic"
            case (COORDSYS_CODE_ECL)
                wcs = "ecliptic"
            case (COORDSYS_CODE_IMAGE)
                wcs = "image"
            case (COORDSYS_CODE_PHYSICAL)
                wcs = "physical"
            case (COORDSYS_CODE_UNITVEC)
                wcs = "univtec"
            case  default
                call wrerr(function, s_err, "converting of regcomp coordsys code " //&
                    trim(int2str(regcomp%coordsys_code)) // " is not supported")
                status = s_errRep
        end select

        ! weight (add or subtract)
        if(regcomp%weight == 0) then
            add = .true.
        else
            add = .false.
        end if

        ! convert shape
        select case (regcomp%shape_code)
          case (REGCOMP_SHAPE_CODE_CIRCLE)
            rshape = "circle"
            !                Region, rshape, x, y, status, r, rotang, add, wcs
            call regRow%fill(rshape, regcomp%pars(1:1), regcomp%pars(2:2)&
              , status, regcomp%pars(3:3), add=add, wcs=wcs)
          case (REGCOMP_SHAPE_CODE_ELLIPSE)
            rshape = "ellipse"
            call regRow%fill(rshape, regcomp%pars(1:1), regcomp%pars(2:2)&
              , status, regcomp%pars(3:4), regcomp%pars(5:5), add=add, wcs=wcs)
          case (REGCOMP_SHAPE_CODE_BOX)
            rshape = "box"
            call regRow%fill(rshape, regcomp%pars(1:1), regcomp%pars(2:2)&
              , status, regcomp%pars(3:4), regcomp%pars(5:5), add=add, wcs=wcs)
          case (REGCOMP_SHAPE_CODE_ANNULUS)
            rshape = "annulus"
            call regRow%fill(rshape, regcomp%pars(1:1), regcomp%pars(2:2)&
              , status, regcomp%pars(3:4), add=add, wcs=wcs)
          case (REGCOMP_SHAPE_CODE_ELLANNULUS)
            rshape = "ellannulus"
            call regRow%fill(rshape, regcomp%pars(1:1), regcomp%pars(2:2)&
              , status, regcomp%pars(3:6), regcomp%pars(7:8), add=add, wcs=wcs)
          case (REGCOMP_SHAPE_CODE_RECTANGLE)
            rshape = "rectangle"
            call regRow%fill(rshape, regcomp%pars(1:1), regcomp%pars(2:2)&
              , status, regcomp%pars(3:4), add=add, wcs=wcs)
          case default
            call wrerr(function, s_err, "converting of regcomp shape code "&
              //trim(int2str(regcomp%shape_code)) // " is not supported")
            status = s_errRep
        end select

  end function convert_regcomp_treg

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> convert a list with the associated object of type tReg type to an allocatable tReg type
    !
    !
    !
    !> @param[in] list_d
    !> @param[out] filename
    !---------------------------------------------------------------------------

  subroutine convert_list2regRows(list_d, regRows, status)
    implicit none

    type(list),                               intent(inout) :: list_d
    type(tRegRow), allocatable, dimension(:), intent(out)   :: regRows
    integer,                                  intent(inout) :: status

    character(*), parameter :: functionName = "convert_list2regRows"
    integer :: ii, n
    class(*), pointer :: tmp

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep)&
  return

    n = list_d%getNumLinks()

    call alloc_regRows(regRows, n, status)

    ! reset the list
    call list_d%reset()

    do ii = 1, n
      tmp => list_d%currObj()
      select type(tmp)
        type is (tRegRow)
          regRows(ii) = tmp
      end select
      call list_d%next()
    end do

  end subroutine convert_list2regRows

  !=====================================================================
  ! Routines concerning linked lists of tRegRow objects.

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine addCircleRegion.
    !> @brief
    !> add an array of substring containing circle expressions to a linked List
    !> the circle must be smaller than 15 arc min and be within the mpe half of the sky
    !
    !> @param list_d, regTok
    !> @param[in]
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

  subroutine addCircleRegionFK5(list_d, regTok, reject, warn, status, comment)
    implicit none

    class(list), pointer,               intent(inout) :: list_d
    character (buffsize), dimension(:), intent(in)    :: regTok
    integer,                            intent(out)   :: reject, warn
    integer,                            intent(inout) :: status
    character(buffsize), optional,      intent(out)   :: comment

    character(*), parameter :: function = "addCircleRegionFK5"
    real(dp) :: x, y, r, glat, glon
    integer(i1) :: flag
    integer :: ii, n, stat
!, add_region
    logical :: add_region
    logical :: conv_hour_to_deg = .true.
    character(buffsize) :: substr, errmsg, i_comment
    character (buffsize), dimension(:), allocatable :: substr_tok
    type(tRegRow), pointer :: regRow

        if(status > 100 .or. status == s_errRep) return

        ! number of potential regions
        n = size(regTok)
        if(n > 1) then
            i_comment = "multiple regions in input line detected - might not show all errors "
        end if

        reject = false
        warn = false
        i_comment = ""

        do ii = 1, n

            substr = adjustl(regTok(ii))
            call str_lower(substr)

            ! modify the individual region tokens to be able to extract the information
            ! commas and brackets are deleted / replaced with spaces
            call strdel_substr(substr, ")", .false.)
            call strreplace_substr(substr, ",", " ", .false.)
            call strreplace_substr(substr, "(", " ")
            call strreplace_substr(substr, ",", " ", .false.)
            ! split the new substr into tokens separated by " " (space)
            call strtok(substr, " ", substr_tok)

            ! first check for global ds9 parameter and cycle that command
            if(trim(substr_tok(1)) == "global") then
                !    print *, "global detected"
                cycle
            end if

            select case (size(substr_tok))
                case (1)
                    if((trim(substr_tok(1)) /= "fk5") .and. (trim(substr_tok(1)) /= "j2000")) then
                        reject = true
                        if(present(comment)) then
                            i_comment = trim(i_comment) // " false WCS selected - " // trim(substr)
                        else
                            call wrerr(function, s_err_no_exit, "false WCS selected - " // trim(substr))
                            status = s_errRep
                            return
                        end if
                    end if

                    cycle
                case (3)

                    call str2deg(substr_tok(1), x, status, convert=conv_hour_to_deg)
                    call str2deg(substr_tok(2), y, status)
                    call str2deg(substr_tok(3), r, status)
                    add_region = .true. ! default

                    if(status /= ok) then
                        reject = true
                        status = ok
                        if(present(comment)) then
                            i_comment = trim(i_comment) // " cannot interpret - " // trim(substr)
                            cycle
                        else
                            call wrerr(function, s_err_no_exit, "cannot interpret - " // trim(substr))
                            status = s_errRep
                            return
                        end if
                    end if

                    if(r > 0.25_r8) then
                        add_region = .false.
                        reject = true
                        i_comment = trim(i_comment) // " size of region > 15' - " // trim(substr)
                    end if

                    if(add_region) then
                        ! always allocate a new region (pointer, since it is added to the linked list)
                        allocate(regRow, stat=stat, errmsg=errmsg)
                        ! error allocating the pointer
                        if(stat /= 0) then
                            if(present(comment)) then
                                i_comment = "error allocating region pointer - contact developers - " // trim(errmsg)
                            else
                                call wrerr(function, s_err_no_exit, errmsg)
                            end if
                            status = s_errRep
                            return
                        end if

                        call regRow%fill("circle", [x], [y], status, [r], wcs="fk5", add=.true.)
                    end if
                case (4)
                    if(trim(substr_tok(1)) /= "circle") then
                        reject = true
                        if(present(comment)) then
                            i_comment = trim(i_comment) // " is not a circle - " // trim(substr)
                            cycle
                        else
                            call wrerr(function, s_err_no_exit, "is not a circle - " // trim(substr))
                            status = s_errRep
                            return
                        end if
                    end if

                    call str2deg(substr_tok(2), x, status, convert=conv_hour_to_deg)
                    call str2deg(substr_tok(3), y, status)
                    call str2deg(substr_tok(4), r, status)

                    add_region = .true. ! default

                    if(status /= ok) then
                        reject = true
                        status = ok
                        if(present(comment)) then
                            i_comment = trim(i_comment) // " cannot interpret - " // trim(substr)
                            cycle
                        else
                            call wrerr(function, s_err_no_exit, "cannot interpret - " // trim(substr))
                            status = s_errRep
                            return
                        end if
                    end if

                    if(r > 0.25_r8) then
                        reject = true
                        add_region = .false.
                        i_comment = trim(i_comment) // " size of region > 15' - " // trim(substr)
                       ! cycle
                    end if

                    if(add_region) then
                        ! always allocate a new region (pointer, since it is added to the linked list)
                        allocate(regRow, stat=stat, errmsg=errmsg)
                        ! error allocating the pointer
                        if(stat /= 0) then
                            if(present(comment)) then
                                i_comment = "error allocating region pointer - contact developers - " // trim(errmsg)
                            else
                                call wrerr(function, s_err_no_exit, errmsg)
                            end if
                            status = s_errRep
                            return
                        end if
                        call regRow%fill("circle", [x], [y], status, [r], wcs="fk5", add=.true.)
                    end if

                case default
                    reject = true
                    if(present(comment)) then
                        i_comment = trim(i_comment) // " cannot interpret - " // trim(substr)
                        cycle
                    else
                        call wrerr(function, s_err_no_exit, "cannot interpret - " // trim(substr))
                        status = s_errRep
                        return
                    end if
            end select

            if(reject == false) then
                call calc_sourceowner(regRow%x(1),regRow%y(1),regRow%r(1), glon, glat, flag, status)
                select case (flag)
                    case (0)
                        ! i_comment = trim(i_comment) // " selected region belongs to IKI and MPE - " // adjustl(regTok(ii))
                        call list_d%addObj(regRow)
                    case (1)
                        i_comment = trim(i_comment) // " selected region belongs to IKI - " // adjustl(regTok(ii))
!                        warn = true
                        reject = true
                        ! TODO remove the addObj later
                 !       print *, "TODO remove addObj after development phase - remove IKI requests"
                !        call list_d%addObj(regRow)
                    case (2)
                        ! i_comment = trim(i_comment) // " selected region belongs to MPE - " // adjustl(regTok(ii))
                        call list_d%addObj(regRow)
                end select
            end if

        end do

        if(present(comment)) comment = i_comment
        !            call list_d%addObj(regRow)

        call dealloc(substr_tok, status)

        if(associated(regRow)) then
            call regRow%free(status)
            deallocate(regRow)
        end if

  end subroutine addCircleRegionFK5

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !>  ! add an array of substrings (regTok), containing region information, to a LinkedList of regions
    !> only one wcs is allowed for all regions
    !
    !
    !> @param list_d
    !> @param region_buffer
    !---------------------------------------------------------------------------

  subroutine addRegion(list_d, regTok, status)
    implicit none

    class(list), pointer,               intent(inout) :: list_d
    character (buffsize), dimension(:), intent(in)    :: regTok
    integer,                            intent(inout) :: status

    character (*), parameter :: function = "addRegion"
    integer :: stat
!    integer :: wcs_is_set = false ! only set to false the first time routine is called
    logical :: wcs_is_set = .false. ! only set to false the first time routine is called
    logical :: conv_hour_to_deg = .false.
    character (buffsize) :: wcs
    character (buffsize) :: substr, rshape, errmsg
    character (buffsize), dimension(:), allocatable :: substr_tok
    logical :: add
    real(dp), allocatable, dimension(:) :: x
    real(dp), allocatable, dimension(:) :: y
    real(dp), allocatable, dimension(:) :: r
    real(dp), allocatable, dimension(:) :: rotang
    type(tRegRow), pointer :: regRow
    integer :: n, ii

    if(status > 100 .or. status == s_errRep)&
  return

    ! number of potential regions
    n = size(regTok)
    !        wcs_is_set = .false.
    do ii = 1, n
      substr = adjustl(regTok(ii))
      call str_lower(substr)
      ! modify the individual region tokens to be able to extract the information
      ! commas and brackets are deleted / replaced with spaces
      call strdel_substr(substr, ")", .false.)
      call strreplace_substr(substr, ",", " ", .false.)
      call strreplace_substr(substr, "(", " ")
      call strreplace_substr(substr, ",", " ", .false.)

      ! split the new substr into tokens separated by " " (space)
      call strtok(substr, " ", substr_tok)
      ! check whether the first keyword is a coordinate system

      if(size(substr_tok) == 1) then
        if(wcs_is_set) then
          if(substr_tok(1) /= wcs)&
call wrstop(function, "multiple WCS in input region is not allowed")

        else
          select case (trim(substr_tok(1)))
          case ('galactic', 'ecliptic')
                            wcs = substr_tok(1)
                            wcs_is_set = .true.
                            conv_hour_to_deg = .false.
          case ('fk5','j2000')
                            wcs = 'fk5'
                            wcs_is_set = .true.
                            conv_hour_to_deg = .true.
          case ('fk4','b1950')
                            wcs = 'fk4'
                            wcs_is_set = .true.
                            conv_hour_to_deg = .true.
          case ('image')
                            wcs = substr_tok(1)
                            wcs_is_set = .true.
                            conv_hour_to_deg = .false.
call wrstop(function, "WCS - " // trim(wcs) // " - is not yet supported")
          case ('linear', 'icrs', 'amplifier') ! 'physical' previously
                            wcs = substr_tok(1)
                            wcs_is_set = .true.
                            conv_hour_to_deg = .false.
call wrstop(function, "WCS - " // trim(wcs) // " - is not yet supported")
          case ('physical', 'detector')
                            ! TODO other name for coordinate system since its not consistend with ds9 region files
                            ! here RAWX and RAWY are filtered
                            wcs ='physical'
                            wcs_is_set = .true.
                            conv_hour_to_deg = .false.
          case ('global')
                            ! the global parameter is ignored
    cycle

          case default
call wrstop(function, "Supplied region is either the name of a file that does not exist or a region expression I can't interpret.")
          end select
        end if
    cycle ! cycle if only the wcs is set
      end if ! size(substr_tok) == 1

      ! since global settings might be larger in size check additionally if the first word is global and ignore it
      if(trim(substr_tok(1)) == "global") then
        call wrlog(function, "Global region file settings ignored")
    cycle
      end if

      ! no coordinate system found and not set before set default to fk5
      if(.not.wcs_is_set) then
        wcs = "fk5"
        conv_hour_to_deg = .true.
      end if

      ! check whether region has to be added or deleted
      if( substr_tok(1)(1:1) == "-") then
        add = .false.
        substr_tok(1) = substr_tok(1)(2:len_trim(substr_tok(1)))
      else
        add = .true.
      end if

      call validateShapeParams(substr_tok(1), size(substr_tok), status)
      if(status == s_errRep) return
      ! always allocate a new region (pointer, since it is added to the linked list)
      allocate(regRow, stat=stat, errmsg=errmsg)
      call FError(function // " allocate pointer", stat, errmsg, status)

      rshape = substr_tok(1)

      if(status == s_errRep) return

      select case (rshape)
      case('circle')
                    call alloc(x, 1, status)
                    call alloc(y, 1, status)
                    call alloc(r, 1, status)

                    call str2deg(substr_tok(2), x(1), status, convert=conv_hour_to_deg)
                    call str2deg(substr_tok(3), y(1), status)
                    call str2deg(substr_tok(4), r(1), status)

                    call regRow%fill(rshape, x, y, status, r, wcs=wcs, add=add)
      case('annulus')
                    call alloc(x, 1, status)
                    call alloc(y, 1, status)
                    call alloc(r, 2, status)

                    call str2deg(substr_tok(2), x(1), status, convert=conv_hour_to_deg)
                    call str2deg(substr_tok(3), y(1), status)
                    call str2deg(substr_tok(4), r(1), status)
                    call str2deg(substr_tok(5), r(2), status)

                    call regRow%fill(rshape, x, y, status, r, wcs=wcs, add=add)
                !                case('rectangle')
                !                    ! CHECK WITM HB and TOM
                !                    if(wcs=='fk5' .or. wcs == 'fk4') then
                !                        call wrerr(function, s_err, "rectangle regions are not supported in fk4/fk5")
                !                        status = s_errRep
                !                        return
                !                    end if
                !
                !                    call alloc(x, 2, status)
                !                    call alloc(y, 2, status)
                !
                !                    if(status == s_errRep) return
                !
                !                    call str2deg(substr_tok(2), x(1), status, convert=conv_hour_to_deg)
                !                    call str2deg(substr_tok(3), x(2), status, convert=conv_hour_to_deg)
                !                    call str2deg(substr_tok(4), y(1), status)
                !                    call str2deg(substr_tok(5), y(2), status)
                !
                !                    call region%init(rshape, x=x, y=y, wcs=wcs, add=add, status=status)

      case('ellipse','box')

                    call alloc(x, 1, status)
                    call alloc(y, 1, status)
                    call alloc(r, 2, status)
                    call alloc(rotang, 1, status)

                    call str2deg(substr_tok(2), x(1), status, convert=conv_hour_to_deg)
                    call str2deg(substr_tok(3), y(1), status)
                    call str2deg(substr_tok(4), r(1), status)
                    call str2deg(substr_tok(5), r(2), status)
                    call str2deg(substr_tok(6), rotang(1), status)

                    call regRow%fill(rshape, x, y, status, r, rotang, wcs=wcs, add=add)

      case ('elliptannulus')

                    call alloc(x, 1, status)
                    call alloc(y, 1, status)
                    call alloc(r, 4, status)

                    call str2deg(substr_tok(2), x(1), status, convert=conv_hour_to_deg)
                    call str2deg(substr_tok(3), y(1), status)
                    call str2deg(substr_tok(4), r(1), status)
                    call str2deg(substr_tok(5), r(2), status)
                    call str2deg(substr_tok(6), r(3), status)
                    call str2deg(substr_tok(7), r(4), status)

                    select case (size(substr_tok))
                        case (8)
                            call alloc(rotang, 1, status)
                            call str2deg(substr_tok(8), rotang(1), status)
                        case (9)
                            call alloc(rotang, 2, status)
                            call str2deg(substr_tok(8), rotang(1), status)
                            call str2deg(substr_tok(9), rotang(2), status)
                        case default
call wrstop(function, "too many parameters when trying to interpret substrings to region: elliptannulus")
                    end select

                    call regRow%fill(rshape, x, y, status, r, rotang, wcs=wcs, add=add)

      case default
call wrstop(function, "region is " // trim(rshape) // " - only circle, ellipse and box regions are fully supported")
      end select

      call list_d%addObj(regRow)

      call dealloc(x, status)
      call dealloc(y, status)
      call dealloc(r, status)
      call dealloc(rotang, status)
      call dealloc(substr_tok, status)

      if(associated(regRow)) then
                call regRow%free(status)
                deallocate(regRow)
      end if
    end do

  end subroutine addRegion

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> delete all regions in a linkedList - freeMemory
    !
    !
    !
    !> @param[in] list_d
    !---------------------------------------------------------------------------

    subroutine deleteLinkedListRegions(list_d, status)
        implicit none

        integer, intent(inout) ::  status
        integer :: stat

        character(*), parameter :: function = "deleteLinkedListRegions"

        class(list), pointer :: list_d
        class(*), pointer :: tmp

        if(status > 100 .or. status == s_errRep) return

        if(associated(list_d)) then

            call list_d%reset()
            do while (associated(list_d%currLink))
                tmp => list_d%currLink%getObj()
                select type (tmp)
!                    type is (tReg)
                    type is (tRegRow)
                    call tmp%free(status)
                    if(status == s_errRep) return
                end select
                call list_d%next()
            end do

            if(associated(tmp)) deallocate(tmp)

        end if

        call list_d%reset()
        stat =  list_d%delAllLinks()
        if(stat /= 0) then
            call wrerr(function, s_err, "deallocation of LinkedList failed")
            status = s_errRep
        end if

        if(associated(list_d)) deallocate(list_d)


    end subroutine deleteLinkedListRegions

  !=====================================================================
  ! Methods of tRegions:

  !.....................................................................
  subroutine free_tRegions(this, status)
    implicit none

    class(tRegions) :: this
    integer, intent(inout) :: status

    character(*), parameter :: function = "free_tRegions"

    call dealloc_regRows(this%regRows, status)

  end subroutine free_tRegions

  !.....................................................................
!  subroutine copy_vals_to_col(in_vals, num_vec_elements, row_i, col_obj)
!  subroutine copy_vals_to_col(in_vals, num_vec_elements, row_i, col_name, table)
!**** this should be in erolib_bintable_mod.f90!!!
!    implicit none
!    real(r8), dimension(:), intent(in)    :: in_vals
!    integer,                intent(in)    :: num_vec_elements,row_i
!    type(tColumn),          intent(inout) :: col_obj
!    character(*),           intent(in)    :: col_name
!    type(tBinTable),        intent(inout) :: table
!
!    integer :: iSize
!    real(r8), dimension(:), pointer :: colvals=>null()
!    real(r8), dimension(:,:), pointer :: veccolvals=>null()
!
!    iSize = size(in_vals)
!    if(iSize>0) then
!      if(num_vec_elements>0) then
!        call table%get_col_ptr(col_name, veccolvals)
!        veccolvals(1:iSize,row_i) = in_vals
!        nullify(veccolvals)
!      else
!        call table%get_col_ptr(col_name, colvals)
!        colvals(row_i) = in_vals(1)
!      end if
!    end if
!
!  end subroutine copy_vals_to_col

  !.....................................................................
  subroutine write_tRegions(this, extname, status, fileInfo)
    ! This requires that the FITS file has already been opened for write via the 'open' method of the this%reg_file object; the file should be closed by the calling routine via this%reg_file%close() after it has finished with it.

    implicit none

    class(tRegions) :: this
    character(*), intent(in)    :: extname
    integer,      intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "write_tRegions"
    integer, parameter :: numVecCols = 4
    integer, dimension(numVecCols) :: nve
    integer :: ii,numRows
    real(r8), dimension(:,:), pointer :: veccolvals=>null()
!    integer :: r8Reg_XC,r8Reg_YC,r8Reg_R,r8Reg_ROTANG
    character(buffsize), dimension(:), pointer :: shape_ptr=>null()

    ! Loop through the regions and work out the maximum number of vector elements for each of the column types.
    !
    numRows = size(this%regRows)
    nve = 0
    do ii=1,numRows
      nve(1) = max(nve(1), size(this%regRows(ii)%x))
      nve(2) = max(nve(2), size(this%regRows(ii)%y))
      nve(3) = max(nve(3), size(this%regRows(ii)%r))
      nve(4) = max(nve(4), size(this%regRows(ii)%rotang))
    end do

    ! Now set up the binary table object:
    !
    if(present(fileInfo)) then
      call this%reg_file%regext%init(status, fileInfo, extname)
    else
      call this%reg_file%regext%init(status, this%reg_file%info, extname)
    end if
    call this%reg_file%regext%ibset_all_useage(writeBit)
    this%reg_file%regext%useageIsSet = .true.

    call this%reg_file%regext%alloc(status, numRows)
!    r8Reg_XC     = this%reg_file%regext%getColNum("XC")
!    r8Reg_YC     = this%reg_file%regext%getColNum("YC")
!    r8Reg_R      = this%reg_file%regext%getColNum("R")
!    r8Reg_ROTANG = this%reg_file%regext%getColNum("ROTANG")
!    call this%reg_file%regext%cols(r8Reg_XC    )%allocVec(nve(1), status)
!    call this%reg_file%regext%cols(r8Reg_YC    )%allocVec(nve(2), status)
!    call this%reg_file%regext%cols(r8Reg_R     )%allocVec(nve(3), status)
!    call this%reg_file%regext%cols(r8Reg_ROTANG)%allocVec(nve(4), status)
    call this%reg_file%regext%allocVec("XC",     nve(1), status)
    call this%reg_file%regext%allocVec("YC",     nve(2), status)
    call this%reg_file%regext%allocVec("R",      nve(3), status)
    call this%reg_file%regext%allocVec("ROTANG", nve(4), status)

    ! Copy over all the values:
    !
    call this%reg_file%regext%get_col_ptr("SHAPE", shape_ptr)
    do ii=1,numRows
      shape_ptr(ii) = this%regRows(ii)%rshape
    end do
    nullify(shape_ptr)

    if(nve(1)>0) then
      call this%reg_file%regext%get_col_ptr("XC", veccolvals)
      do ii=1,numRows
        veccolvals(1:size(this%regRows(ii)%x),ii) = this%regRows(ii)%x
      end do
      nullify(veccolvals)
    end if

    if(nve(2)>0) then
      call this%reg_file%regext%get_col_ptr("YC", veccolvals)
      do ii=1,numRows
        veccolvals(1:size(this%regRows(ii)%y),ii) = this%regRows(ii)%y
      end do
      nullify(veccolvals)
    end if

    if(nve(3)>0) then
      call this%reg_file%regext%get_col_ptr("R", veccolvals)
      do ii=1,numRows
        veccolvals(1:size(this%regRows(ii)%r),ii) = this%regRows(ii)%r
      end do
      nullify(veccolvals)
    end if

    if(nve(4)>0) then
      call this%reg_file%regext%get_col_ptr("ROTANG", veccolvals)
      do ii=1,numRows
        veccolvals(1:size(this%regRows(ii)%rotang),ii) = this%regRows(ii)%rotang
      end do
      nullify(veccolvals)
    end if

    ! Now open, write, and close the file.
    !
    call this%reg_file%regext%write(status)

!**** include in header?    call ffits_update_key_str(unit, "HDUCLAS1", "REGION", "", status)

!*** destroy the tRegTab object? Probably a good idea.
    call this%reg_file%regext%destroy(status)

  end subroutine write_tRegions

  !.....................................................................
  subroutine read_tRegions(this, extname, status)
    ! This requires that the FITS file has already been opened for read via the 'open' method of the this%reg_file object; the file should be closed by the calling routine via this%reg_file%close() after it has finished with it.

    implicit none

    class(tRegions) :: this
    character(*), intent(in)    :: extname
    integer,      intent(inout) :: status

    character(*), parameter :: functionName = "read_tRegions"
    integer :: ii,numRows,stat
    character(buffsize) :: errmsg
    real(r8), dimension(:,:), pointer :: x_ptr=>null(),y_ptr=>null(),r_ptr=>null(),a_ptr=>null()
    character(buffsize), dimension(:), pointer :: shape_ptr=>null()

    call this%reg_file%regext%init(status, this%reg_file%info, extname)
    call this%reg_file%regext%ibset_all_useage(readBit)
    this%reg_file%regext%useageIsSet = .true.

    call this%reg_file%regext%read(status)

    ! Now we have to translate the values from the tRegTab object to the array of tRegRow objects.
    !
    numRows = this%reg_file%regext%getNumRows()

    if(allocated(this%regRows)) deallocate(this%regRows, stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

    allocate(this%regRows(numRows), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

    call this%reg_file%regext%get_col_ptr("XC",     x_ptr)
    call this%reg_file%regext%get_col_ptr("YC",     y_ptr)
    call this%reg_file%regext%get_col_ptr("R",      r_ptr)
    call this%reg_file%regext%get_col_ptr("ROTANG", a_ptr)
    call this%reg_file%regext%get_col_ptr("SHAPE",  shape_ptr)
    do ii = 1, numRows
      call str_lower(shape_ptr(ii))
      call this%regRows(ii)%fill(shape_ptr(ii)&
        , x_ptr(:,ii), y_ptr(:,ii), status, r_ptr(:,ii), a_ptr(:,ii))
    end do
    nullify(shape_ptr)
    nullify(a_ptr)
    nullify(r_ptr)
    nullify(y_ptr)
    nullify(x_ptr)

!*** destroy the tRegTab object? Probably a good idea.
    call this%reg_file%regext%destroy(status)

    this%containsData = .true.

  end subroutine read_tRegions

  !.....................................................................
  subroutine get_srctool_region(this, srctool_region, status)
    implicit none

    class(tRegions)                  :: this
    type(region_type), intent(out)   :: srctool_region
    integer,           intent(inout) :: status

    character(*), parameter :: functionName = "get_srctool_region"
    integer :: ii
    type(regcomp_type) :: regcomp

    if(status > 100 .or. status == s_errRep) return

    call srctool_regions_init_region(srctool_region)

    do ii = 1, size(this%regRows)
      call this%regRows(ii)%get_srctool_regcomp(regcomp, status)
      call srctool_regions_add_regcomp(srctool_region, regcomp)
    end do

  end subroutine get_srctool_region

  !.....................................................................
  subroutine get_srctool_encap_circle_region(this, encap_region, status)
    implicit none

    class(tRegions) :: this
    type(region_type), intent(out)   :: encap_region
    integer,           intent(inout) :: status

    character(*), parameter :: function = "get_srctool_encap_circle_region"
    integer :: ii
    type(tRegRow), dimension(size(this%regRows)) :: circleRegRows
    type(regcomp_type) :: regcomp

    if(status > 100 .or. status == s_errRep) return

    call srctool_regions_init_region(encap_region)

    do ii = 1, size(this%regRows)
      circleRegRows(ii) = this%regRows(ii)%getEncapsulCircleReg(status)
      call circleRegRows(ii)%get_srctool_regcomp(regcomp, status)
      call srctool_regions_add_regcomp(encap_region, regcomp)
    end do

  end subroutine get_srctool_encap_circle_region

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine sort_skyfield_tRegions.
    !> @brief
    !> sort the Region Type per skyfield
    !
    !
    !> @param this
    !> @param[in]
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine sort_skyfield_tRegions(this, status)
        implicit none

        class(tRegions) :: this
        integer, intent(inout) :: status

        !internal param
        integer, dimension(:), allocatable :: skyfield
        integer, dimension(:), allocatable :: order

        character(*), parameter :: function = "sort_skyfield_tRegions"

        if(status > 100 .or. status == s_errRep) return

        skyfield = this%regRows(:)%skyfield
        call sort(skyfield, status, order)

        this%regRows = this%regRows(order)

        call dealloc(skyfield, status)
        call dealloc(order, status)

    end subroutine sort_skyfield_tRegions

  !=====================================================================
  ! The remaining routines are not methods of tRegions, they're just
  ! utility routines that manipulate the objects.

  !.....................................................................
  function convert_regtype_tregtable(regtype, status)
    implicit none

    type(region_type), intent(in) :: regtype
    integer,           intent(inout) :: status
    type(tRegions) :: convert_regtype_tregtable

    character(*), parameter :: function = "convert_regtype_tregtable"
    integer :: n, ii

    if(status > 100 .or. status == s_errRep) return

    n = regtype%num_regcomps
    allocate(convert_regtype_tregtable%regRows(n))
    do ii = 1, n
      convert_regtype_tregtable%regRows(ii) = convert_regcomp_treg(regtype%pregcomp(ii), status)
    end do

  end function convert_regtype_tregtable

  !.....................................................................
  subroutine alloc_regions(list_of_regions, n, status)
    implicit none

    type(tRegions), allocatable, dimension(:) :: list_of_regions
    integer, intent(in) :: n
    integer, intent(inout) :: status

        character(*), parameter :: function = "alloc_regions"
        integer :: stat
        character(buffsize) :: errmsg

        call dealloc_regions(list_of_regions, status)
        allocate(list_of_regions(n), stat=stat, errmsg=errmsg)
        call FError(function, stat, errmsg, status)

  end subroutine alloc_regions

  !.....................................................................
  subroutine dealloc_regions(list_of_regions, status)
    implicit none

    type(tRegions), allocatable, dimension(:) :: list_of_regions
    integer, intent(inout) :: status

        character(*), parameter :: function = "dealloc_regions"
        integer :: stat

        if(allocated(list_of_regions)) then
            deallocate(list_of_regions, stat=stat)
            call FError(function, stat, "dealloc tRegions type", status)
        end if

  end subroutine dealloc_regions

  !.....................................................................
  subroutine getRegionsFromStr(str_in, regionsObj, status)
    ! The string can either be a region expression or the name of a FITS region file.
    implicit none

    character(*),   intent(in)    :: str_in
    type(tRegions), intent(out)   :: regionsObj
    integer,        intent(inout) :: status

    character(*), parameter :: functionName = "getRegionsFromStr"
    character(buffsize) :: str
    integer :: exists

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep)&
  return

    ! get rid of preceeding spaces
    str = adjustl(str_in)
    ! check whether the str is a file and exists
    call ffits_file_exists(str, exists, status)
    ! if the file does not exist
    if (exists == 0) then
      call interpret_expression_regions(str, regionsObj, status)
    else
      ! try to interpret str as a FITS file
      call regionsObj%reg_file%info%setFileName(str)
      status = ignore
      call regionsObj%reg_file%open(FF_READONLY, status)
!!!            call regionsObj%read(str, status)
      ! if not a FITS file it is treated as ascii
      if(status /= ignore) then ! it is not a FITS file.
        ! reset status to ok 
        status = ok
        call interpret_ASCII_regions(str, regionsObj, status)
      else
        ! reset status from ignore to ok again
        status = ok
        call regionsObj%read('REGION', status)
        call regionsObj%reg_file%close(status)
      end if
    end if

    regionsObj%containsData = .true.

  end subroutine getRegionsFromStr

  !.....................................................................
  subroutine interpret_ASCII_regions_circle(filename, regionsObj, nreject, status, comment)
    ! 20210424 IMS: this is used by datool.
    implicit none

    character(*),                                             intent(in)    :: filename
    type(tRegions),                                           intent(out)   :: regionsObj
    integer,                                                  intent(out)   :: nreject
    integer,                                                  intent(inout) :: status
    character(buffsize), optional, dimension(:), allocatable, intent(out)   :: comment

    character(*), parameter :: function = "interpret_ASCII_regions_circle"
    character(buffsize) :: iomsg, line
    character(buffsize), dimension(:), allocatable :: regTok
    class(list), pointer :: list_d=>null()
    integer :: unit, exists, iostat, pos
    integer :: nlines, reject, warn, ncomment

        ! return automatically (fitsio error start at 101, 0 = ok)
        if (status > 100 .or. status == s_errRep) return

        ! read the ascii file on a per line bases
!        if(.not. associated(list_d)) allocate(list_d)
        allocate(list_d)

        ! check whether the file exists
        call ffits_file_exists(filename, exists, status)
        ! if the file does not exist return
        if (exists == 0) then
            status = s_err
            call wrerr(function, status, trim(filename) // " - does not exist")
            status = s_errRep
            return
        end if

        call ffits_get_iounit(unit, status)
        open (unit, file=filename, status="old",action="read", iostat=iostat, iomsg=iomsg)
        call FError(function // " - open", iostat, iomsg, status)

        ! get the number of lines
        nlines = 0
        do
            read(unit, '(A)', iostat=iostat, iomsg=iomsg) line
            ! have to escape the sequence instead to stop here
            if(iostat /= 0) exit
            nlines = nlines + 1
        end do

        if(present(comment)) call alloc(comment, nlines, status)

        rewind(unit, iostat=iostat)
        if(iostat /= 0)&
call wrstop(function, "rewind of file " // trim(filename) // " failed")

        nreject = 0
        ncomment = 0

        do
            read(unit, '(A)', iostat=iostat, iomsg=iomsg) line
            if(IS_IOSTAT_END(iostat)) exit
            if(iostat /= 0) then
              call FError(function // " - read " // trim(filename), iostat, iomsg, status)
              exit
            end if
            if(status == s_errRep) return

            ! delete the comments marked by #
            line = adjustl(line)
            pos = index(line, "#")
            select case (pos)
                case (0) ! do nothing
                case (1) ! whole line is a comment
                    cycle
                case default ! reduce the line to only contain everything before the #
                    line = line(1:pos-1)
            end select

            ! ignore empty lines
            pos = verify(line, " ")
            if(pos == 0) then
                cycle
            end if

            ! split input line into multiple substrings, each command is characterized by a ;
            call strtok(line, ";", regTok)

            if(present(comment)) then
                call addCircleRegionFK5(list_d, regTok, reject, warn, status, comment(nreject+1))
            else
                call addCircleRegionFK5(list_d, regTok, reject, warn, status)
            end if

            if(reject == true) then
                nreject = nreject + 1
                ncomment = ncomment + 1
            end if
            if(warn == true) then
                ncomment = ncomment + 1
            end if

            call dealloc(regTok, status)
        end do

        if(present(comment)) then

            if(ncomment > 0) then
                comment = comment(1:ncomment)
            else
                call alloc(comment, 0, status)
            end if
        end if

        close(unit, iostat=iostat, iomsg=iomsg)
        call ffits_free_iounit(unit, status)
        call FError(function // " - close "// trim(filename), iostat, iomsg, status)

        call convert_list2regRows(list_d, regionsObj%regRows, status)

        call deleteLinkedListRegions(list_d, status)

  end subroutine interpret_ASCII_regions_circle

  !.....................................................................
  subroutine interpret_ASCII_regions(filename, regionsObj, status)
    implicit none

    character(*),   intent(in)    :: filename
    type(tRegions), intent(out)   :: regionsObj
    integer,        intent(inout) :: status

    character(*), parameter :: function = "interpret_ASCII_regions"
    character(buffsize) :: iomsg, line
    character(buffsize), dimension(:), allocatable :: regTok
    class(list), pointer :: list_d=>null()
    integer :: unitNo, exists, iostat, pos

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep)&
  return

    ! read the ascii file on a per line bases
!    if(.not. associated(list_d)) allocate(list_d)
    allocate(list_d)

    ! check whether the file exists
    call ffits_file_exists(filename, exists, status)
    ! if the file does not exist return
    if (exists == 0) then
        status = s_err
        call wrerr(function, status, trim(filename) // " - does not exist")
        status = s_errRep
  return
    end if

    call ffits_get_iounit(unitNo, status)
    open (unitNo, file=filename, status="old",action="read", iostat=iostat, iomsg=iomsg)
    call FError(function // " - open", iostat, iomsg, status)

        ! get the number of lines
!        nlines = 0
!        do
!            read(unit, '(A)', iostat=iostat, iomsg=iomsg) line
!            ! have to escape the sequence instead to stop here
!            if(iostat /= 0) exit
!            nlines = nlines + 1
!        end do

!        rewind(unit, iostat=iostat)
!        if(iostat /= 0) call wrstop(function, "rewind of file " // trim(filename) // " failed")

    do
      read(unitNo, '(A)', iostat=iostat, iomsg=iomsg) line
      if(IS_IOSTAT_END(iostat))&
    exit

      if(iostat /= 0) then
        call FError(function // " - read " // trim(filename), iostat, iomsg, status)
    exit
      end if
      if(status == s_errRep)&
  return

      ! delete the comments marked by #
      line = adjustl(line)
      pos = index(line, "#")
      select case (pos)
      case (0) ! do nothing
      case (1) ! whole line is a comment
    cycle
      case default ! reduce the line to only contain everything before the #
        line = line(1:pos-1)
      end select

      ! ignore empty lines
      pos = verify(line, " ")
      if(pos == 0)&
    cycle

      ! split input line into multiple substrings, each command is characterized by a ;
      call strtok(line, ";", regTok)
      call addRegion(list_d, regTok, status)
      call dealloc(regTok, status)
    end do

    close(unitNo, iostat=iostat, iomsg=iomsg)
    call ffits_free_iounit(unitNo, status)
    call FError(function // " - close "// trim(filename), iostat, iomsg, status)

    call convert_list2regRows(list_d, regionsObj%regRows, status)

    call deleteLinkedListRegions(list_d, status)

  end subroutine interpret_ASCII_regions

  !.....................................................................
  subroutine interpret_expression_regions(expression, regionsObj, status)
    ! 20210424 IMS: This is called by datool.
    implicit none

    character(*),      intent(in)    :: expression
    type(tRegions),    intent(out)   :: regionsObj
    integer, optional, intent(inout) :: status

    character(*), parameter :: function = "interpret_expression_regions"
    character(buffsize), dimension(:), allocatable :: regTok
    class(list), pointer :: list_d=>null()

    ! return automatically (fitsio error start at 101, 0 = ok)
    if (status > 100 .or. status == s_errRep)&
  return

!    if(.not. associated(list_d)) allocate(list_d)
    allocate(list_d)

    ! interprete the filename as an expression
    call strtok(expression, ";", regTok)
    call addRegion(list_d, regTok, status)

    call convert_list2regRows(list_d, regionsObj%regRows, status)

    call deleteLinkedListRegions(list_d, status)

  end subroutine interpret_expression_regions

  !=====================================================================
  ! General-purpose utility routines.

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> verify that the number of substrings (parameters)  matches the number of parameters for a specific regions
    !
    !
    !
    !> @param[in] rshape
    !> @param[out] n
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine validateShapeParams(rshape, n, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: n
        integer :: stat

        character(*), intent(inout) :: rshape
        character(*), parameter :: function = "validateShapeParams"

        if(status > 100 .or. status == s_errRep) return

        stat = 0
        select case (rshape)
            case ("circle")
                if(n == 4) stat = 1
            case ("rectangle", "annulus")
                if(n == 5) stat = 1
            case ("ellipse", "box", "rotbox")
                if(n == 6) stat = 1
                if(n == 8) then
                    stat = 1
                    rshape = "elliptannulus"
                    call wrlog(function, "DS9 elliptical annulus detected")
                end if
            case ("elliptannulus")
                if(n == 8) stat = 1 ! 1 angle for both ellipses
                if(n == 9) stat = 1 ! 2 different angles
            case default
                call  wrerr(function, s_err, "region of type " // trim(rshape) // &
                    " is not supported and/or wrong expression used")
                status = s_errRep
                return
        end select

        if(stat == 0) then
            call wrerr(function, s_err, "the number of parameters does not match the region type")
            status = s_errRep
        end if

    end subroutine validateShapeParams

end module erolib_region_mod
