! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_sort_mod

    use ISO_FORTRAN_ENV ! gfortran module for real64 and other definitions
    use erolib_general_mod
    use param
    
    implicit none

    interface swap
        module procedure swap_i1, swap_i2, swap_i4, swapreal, swapstr, swapdouble, swap_i4_array, swap_r4_array
    end interface

    interface mergesort
        module procedure mergesort_i1, mergesort_i2, mergesort_i4, mergesort_real, mergesort_double
        module procedure mergesort_i1_order, mergesort_i2_order, mergesort_i4_order, mergesort_real_order, mergesort_double_order
    end interface

    interface sort
        module procedure sort_i1, sort_i2, sort_i4, sort_real, sort_double
    end interface

    interface compare
        module procedure compare_r4_array, compare_r8_array, compare_i1_array, compare_i2_array, compare_i4_array
    end interface

    interface unique
        module procedure unique_i2, unique_i4, unique_r8
    end interface unique

contains

    subroutine reverse_array(array)
        implicit none

        class(*), dimension(:) :: array
        integer :: n

        n = size(array)

        select type (array)
            type is (integer(i1))
            array = array( n:1:-1)
            type is (integer(i2))
            array = array( n:1:-1)
            type is (integer(i4))
            array = array( n:1:-1)
            type is (real(sp))
            array = array( n:1:-1)
            type is (real(dp))
            array = array( n:1:-1)
            type is (character(*))
            array = array(n:1:-1)
        end select

    end subroutine reverse_array

    subroutine sort_i1 (array, status, order)
        implicit none

        integer(i1), dimension(:), intent(inout) :: array
        integer, dimension(:), optional, allocatable , intent(out):: order
        integer(i1), dimension(:), allocatable :: subarray
        integer, dimension(:), allocatable :: suborder

        integer :: n, ii
        integer, optional, intent(inout) :: status

        character(*), parameter :: function = "sort_i1"

        n = size(array)
!        call alloc(subarray, (n+1 / 2), status)
        call alloc(subarray, ceiling(real(n)/2.0), status)

        if(present(order)) then
            call alloc(order, n, status)
            order = [(ii, ii=1, n)]
!            call alloc(suborder, (n+1 / 2), status)
            call alloc(suborder, ceiling(real(n)/2.0), status)
            call mergesort(array, n, subarray, order, suborder)
        else
            call mergesort(array, n, subarray)
        end if

        call dealloc (subarray, status)
        call dealloc (suborder, status)

    end subroutine sort_i1

    subroutine sort_i2 (array, status, order)
        implicit none

        integer(i2), dimension(:), intent(inout) :: array
        integer, dimension(:), optional, allocatable , intent(out):: order
        integer(i2), dimension(:), allocatable :: subarray
        integer, dimension(:), allocatable :: suborder

        integer :: n, ii
        integer, optional, intent(inout) :: status

        character(*), parameter :: function = "sort_i2"

        n = size(array)
!        call alloc(subarray, (n+1 / 2), status)
        call alloc(subarray, ceiling(real(n)/2.0), status)

        if(present(order)) then
            call alloc(order, n, status)
            order = [(ii, ii=1, n)]
!            call alloc(suborder, (n+1 / 2), status)
            call alloc(suborder, ceiling(real(n)/2.0), status)
            call mergesort(array, n, subarray, order, suborder)
        else
            call mergesort (array, n, subarray)
        end if

        call dealloc (subarray, status)
        call dealloc (suborder, status)

    end subroutine sort_i2

    subroutine sort_i4 (array, status, order)
        implicit none

        integer, dimension(:), intent(inout) :: array
        integer(i4), dimension(:), optional, allocatable , intent(out):: order
        integer, dimension(:), allocatable :: subarray
        integer(i4), dimension(:), allocatable :: suborder

        integer :: n, ii
        integer, optional, intent(inout) :: status

        character(*), parameter :: function = "sort_i4"

        n = size(array)
!        call alloc(subarray, (n+1 / 2), status)
        call alloc(subarray, ceiling(real(n)/2.0), status)

        if(present(order)) then
            call alloc(order, n, status)
            order = [(ii, ii=1, n)]
!            call alloc(suborder, (n+1 / 2), status)
            call alloc(suborder, ceiling(real(n)/2.0), status)
            call mergesort(array, n, subarray, order, suborder)
        else
            call mergesort (array, n, subarray)
        end if

        call dealloc (subarray, status)
        call dealloc (suborder, status)

    end subroutine sort_i4

    subroutine sort_real (array, status, order)
        implicit none

        real, dimension(:), intent(inout) :: array
        real, dimension(:), allocatable :: subarray

        integer, dimension(:), allocatable, optional , intent(out):: order
        integer, dimension(:), allocatable :: suborder

        integer :: n, ii
        integer, optional, intent(inout) :: status

        character(*), parameter :: function = "sort_real"

        n = size(array)
!        call alloc(subarray, (n+1 / 2), status)
        call alloc(subarray, ceiling(real(n)/2.0), status)

        if(present(order)) then
            call alloc(order, n, status)
            order = [(ii, ii=1, n)]
!            call alloc(suborder, (n+1 / 2), status)
            call alloc(suborder, ceiling(real(n)/2.0), status)
            call mergesort(array, n, subarray, order, suborder)
        else
            call mergesort (array, n, subarray)
        end if

        call dealloc (subarray, status)
        call dealloc (suborder, status)

    end subroutine sort_real

    subroutine sort_double (array, status, order)
        implicit none

        real(dp), dimension(:), intent(inout) :: array
        integer, optional, intent(inout) :: status
        integer, dimension(:), allocatable, optional , intent(out):: order

        real(dp), dimension(:), allocatable :: subarray
        integer, dimension(:), allocatable :: suborder
        integer :: n, ii
        character(*), parameter :: function = "sort_double"

        n = size(array)
!        call alloc(subarray, (n+1 / 2), status)
        call alloc(subarray, ceiling(real(n)/2.0), status)

        if(present(order)) then
            call alloc(order, n, status)
            order = [(ii, ii=1, n)]
!            call alloc(suborder, (n+1 / 2), status)
            call alloc(suborder, ceiling(real(n)/2.0), status)
            call mergesort(array, n, subarray, order, suborder)
        else
            call mergesort(array, n, subarray)
        end if

        call dealloc (subarray, status)
        call dealloc (suborder, status)

    end subroutine sort_double

    subroutine swap_i1(a, b)
        implicit none
        integer(i1), intent(inout) :: a,b
        integer(i1) :: tmp
        tmp = a
        a = b
        b = tmp;
    end subroutine swap_i1

    subroutine swap_i2(a, b)
        implicit none
        integer(i2), intent(inout) :: a,b
        integer(i2) :: tmp
        tmp = a
        a = b
        b = tmp;
    end subroutine swap_i2

    subroutine swap_i4(a, b)
        implicit none
        integer(i4), intent(inout) :: a,b
        integer(i4) :: tmp
        tmp = a
        a = b
        b = tmp;
    end subroutine swap_i4

    subroutine swap_i4_array(a, b)
        implicit none
        integer(i4), intent(inout), dimension(:) :: a
        integer(i4), intent(inout), dimension(size(a)) :: b
        integer(i4), dimension(size(a)) :: tmp
        tmp = a
        a = b
        b = tmp;
    end subroutine swap_i4_array

    subroutine swap_r4_array(a, b)
        implicit none
        real(r4), intent(inout), dimension(:) :: a
        real(r4), intent(inout), dimension(size(a)) :: b
        real(r4), dimension(size(a)) :: tmp
        tmp = a
        a = b
        b = tmp;
    end subroutine swap_r4_array


    subroutine swapreal(a, b)
        implicit none
        real, intent(inout) :: a,b
        real :: tmp
        tmp = a
        a = b
        b = tmp;
    end subroutine swapreal

    subroutine swapdouble(a, b)
        implicit none
        real(dp), intent(inout) :: a,b
        real(dp) :: tmp
        tmp = a
        a = b
        b = tmp;
    end subroutine swapdouble

    subroutine swapstr(a, b)
        implicit none
        character(*), intent(inout) :: a,b
        character(len(a)) :: tmp
        tmp = a
        a = b
        b = tmp;
    end subroutine swapstr

    subroutine merge_i1 (array_l, nl, array_r, nr, array, n)

        integer, intent(in) :: nl, nr, n ! dimension of arrays
        integer(i1), intent(inout), dimension(n) :: array
        integer(i1), intent(in), dimension(nl) :: array_l
        integer(i1), intent(in), dimension(nr) :: array_r
        integer :: ll, rr, nn

        ! indices in the arrays
        ll = 1; rr = 1; nn = 1;
        do while (ll <= nl .and. rr <= nr)
            if(array_l(ll) <= array_r(rr)) then
                array(nn) = array_l(ll)
                ll = ll + 1
            else
                array(nn) = array_r(rr)
                rr = rr + 1
            end if
            nn = nn + 1
        end do

        do while ( ll <= nl)
            array(nn) = array_l(ll)
            ll = ll + 1
            nn = nn + 1
        end do

    end subroutine merge_i1

    recursive subroutine mergesort_i1(array, n, subarray)
        implicit none

        integer, intent(in) :: n
        integer(i1), dimension(n), intent(inout) :: array
        integer(i1), dimension((n+1)/2), intent(out) :: subarray
        integer :: index_l, index_r

        if(n<2)  return
        if(n == 2) then
            if(array(1) > array(2)) call swap(array(1), array(2))
            return
        end if

        index_l = (n+1)/2
        index_r = n - index_l

        call mergesort_i1(array(1:index_l), index_l, subarray)
        call mergesort_i1(array((index_l + 1) : n), index_r, subarray)
        if(array(index_l) > array(index_l+1)) then
            subarray(1:index_l) = array(1:index_l)
            call merge_i1(subarray, index_l, array(index_l+1), index_r, array, n)
        end if

    end subroutine mergesort_i1

    subroutine merge_i2 (array_l, nl, array_r, nr, array, n)

        integer, intent(in) :: nl, nr, n ! dimension of arrays
        integer(i2), intent(inout), dimension(n) :: array
        integer(i2), intent(in), dimension(nl) :: array_l
        integer(i2), intent(in), dimension(nr) :: array_r
        integer :: ll, rr, nn

        ! indices in the arrays
        ll = 1; rr = 1; nn = 1;
        do while (ll <= nl .and. rr <= nr)
            if(array_l(ll) <= array_r(rr)) then
                array(nn) = array_l(ll)
                ll = ll + 1
            else
                array(nn) = array_r(rr)
                rr = rr + 1
            end if
            nn = nn + 1
        end do

        do while ( ll <= nl)
            array(nn) = array_l(ll)
            ll = ll + 1
            nn = nn + 1
        end do

    end subroutine merge_i2

    recursive subroutine mergesort_i2(array, n, subarray)
        implicit none

        integer, intent(in) :: n
        integer(i2), dimension(n), intent(inout) :: array
        integer(i2), dimension((n+1)/2), intent(out) :: subarray
        integer :: index_l, index_r

        if(n<2)  return
        if(n == 2) then
            if(array(1) > array(2)) call swap(array(1), array(2))
            return
        end if

        index_l = (n+1)/2
        index_r = n - index_l

        call mergesort_i2(array(1:index_l), index_l, subarray)
        call mergesort_i2(array((index_l + 1) : n), index_r, subarray)
        if(array(index_l) > array(index_l+1)) then
            subarray(1:index_l) = array(1:index_l)
            call merge_i2(subarray, index_l, array(index_l+1), index_r, array, n)
        end if

    end subroutine mergesort_i2

    subroutine merge_i4 (array_l, nl, array_r, nr, array, n)

        integer, intent(in) :: nl, nr, n ! dimension of arrays
        integer(i4), intent(inout), dimension(n) :: array
        integer(i4), intent(in), dimension(nl) :: array_l
        integer(i4), intent(in), dimension(nr) :: array_r
        integer :: ll, rr, nn

        ! indices in the arrays
        ll = 1; rr = 1; nn = 1;
        do while (ll <= nl .and. rr <= nr)
            if(array_l(ll) <= array_r(rr)) then
                array(nn) = array_l(ll)
                ll = ll + 1
            else
                array(nn) = array_r(rr)
                rr = rr + 1
            end if
            nn = nn + 1
        end do

        do while ( ll <= nl)
            array(nn) = array_l(ll)
            ll = ll + 1
            nn = nn + 1
        end do

    end subroutine merge_i4

    recursive subroutine mergesort_i4(array, n, subarray)
        implicit none

        integer, intent(in) :: n
        integer(i4), dimension(n), intent(inout) :: array
        integer(i4), dimension((n+1)/2), intent(out) :: subarray
        integer :: index_l, index_r

        if(n<2)  return
        if(n == 2) then
            if(array(1) > array(2)) call swap(array(1), array(2))
            return
        end if

        index_l = (n+1)/2
        index_r = n - index_l

        call mergesort_i4(array(1:index_l), index_l, subarray)
        call mergesort_i4(array((index_l + 1) : n), index_r, subarray)
        if(array(index_l) > array(index_l+1)) then
            subarray(1:index_l) = array(1:index_l)
            call merge_i4(subarray, index_l, array(index_l+1), index_r, array, n)
        end if

    end subroutine mergesort_i4

    subroutine merge_real (array_l, nl, array_r, nr, array, n)

        integer, intent(in) :: nl, nr, n ! dimension of arrays
        real, intent(inout), dimension(n) :: array
        real, intent(in), dimension(nl) :: array_l
        real, intent(in), dimension(nr) :: array_r
        integer :: ll, rr, nn

        ! indices in the arrays
        ll = 1; rr = 1; nn = 1;
        do while (ll <= nl .and. rr <= nr)
            if(array_l(ll) <= array_r(rr)) then
                array(nn) = array_l(ll)
                ll = ll + 1
            else
                array(nn) = array_r(rr)
                rr = rr + 1
            end if
            nn = nn + 1
        end do

        do while ( ll <= nl)
            array(nn) = array_l(ll)
            ll = ll + 1
            nn = nn + 1
        end do

    end subroutine merge_real

    recursive subroutine mergesort_real(array, n, subarray)
        implicit none

        integer, intent(in) :: n
        real, dimension(n), intent(inout) :: array
        real, dimension((n+1)/2), intent(out) :: subarray
        integer :: index_l, index_r

        if(n<2)  return
        if(n == 2) then
            if(array(1) > array(2)) call swap(array(1), array(2))
            return
        end if

        index_l = (n+1)/2
        index_r = n - index_l

        call mergesort_real(array(1:index_l), index_l, subarray)
        call mergesort_real(array((index_l + 1) : n), index_r, subarray)
        if(array(index_l) > array(index_l+1)) then
            subarray(1:index_l) = array(1:index_l)
            call merge_real(subarray, index_l, array(index_l+1), index_r, array, n)
        end if

    end subroutine mergesort_real

    subroutine merge_double (array_l, nl, array_r, nr, array, n)

        integer, intent(in) :: nl, nr, n ! dimension of arrays
        real(dp), intent(inout), dimension(n) :: array
        real(dp), intent(in), dimension(nl) :: array_l
        real(dp), intent(in), dimension(nr) :: array_r
        integer :: ll, rr, nn

        ! indices in the arrays
        ll = 1; rr = 1; nn = 1;
        do while (ll <= nl .and. rr <= nr)
            if(array_l(ll) <= array_r(rr)) then
                array(nn) = array_l(ll)
                ll = ll + 1
            else
                array(nn) = array_r(rr)
                rr = rr + 1
            end if
            nn = nn + 1
        end do

        do while ( ll <= nl)
            array(nn) = array_l(ll)
            ll = ll + 1
            nn = nn + 1
        end do

    end subroutine merge_double

    recursive subroutine mergesort_double(array, n, subarray)
        implicit none

        integer, intent(in) :: n
        real(dp), dimension(n), intent(inout) :: array
        real(dp), dimension((n+1)/2), intent(out) :: subarray
        integer :: index_l, index_r

        if(n<2)  return
        if(n == 2) then
            if(array(1) > array(2)) call swap(array(1), array(2))
            return
        end if

        index_l = (n+1)/2
        index_r = n - index_l

        call mergesort_double(array(1:index_l), index_l, subarray)
        call mergesort_double(array((index_l + 1) : n), index_r, subarray)
        if(array(index_l) > array(index_l+1)) then
            subarray(1:index_l) = array(1:index_l)
            call merge_double(subarray, index_l, array(index_l+1), index_r, array, n)
        end if

    end subroutine mergesort_double

    subroutine merge_double_order(array_l, nl, array_r, nr, array, n, order_l, order_r, order)

        integer, intent(in) :: nl, nr, n ! dimension of arrays
        integer, intent(inout), dimension(n) :: order
        integer, intent(in), dimension(nl) :: order_l
        integer, intent(in), dimension(nr) :: order_r

        real(dp), intent(inout), dimension(n) :: array
        real(dp), intent(in), dimension(nl) :: array_l
        real(dp), intent(in), dimension(nr) :: array_r
        integer :: ll, rr, nn

        ! indices in the arrays
        ll = 1; rr = 1; nn = 1;
        do while (ll <= nl .and. rr <= nr)
            if(array_l(ll) <= array_r(rr)) then
                array(nn) = array_l(ll)
                order(nn) = order_l(ll)
                ll = ll + 1
            else
                array(nn) = array_r(rr)
                order(nn) = order_r(rr)
                rr = rr + 1
            end if
            nn = nn + 1
        end do

        do while ( ll <= nl)
            array(nn) = array_l(ll)
            order(nn) = order_l(ll)
            ll = ll + 1
            nn = nn + 1
        end do

    end subroutine merge_double_order

    recursive subroutine mergesort_double_order(array, n, subarray, order, suborder)
        implicit none

        integer, intent(in) :: n
        integer, dimension(n), intent(inout) :: order
        integer, dimension((n+1)/2), intent(out) :: suborder

        real(dp), dimension(n), intent(inout) :: array
        real(dp), dimension((n+1)/2), intent(out) :: subarray
        integer :: index_l, index_r

        if(n<2)  return
        if(n == 2) then
            if((array(1) - array(2)) > tiny(array(1))) then
                call swap(array(1), array(2))
                call swap(order(1), order(2))
            end if
            return
        end if

        index_l = (n+1)/2
        index_r = n - index_l

        call mergesort_double_order(array(1:index_l), index_l, subarray, order(1:index_l), suborder)
        call mergesort_double_order(array((index_l + 1) : n), index_r, subarray, order((index_l+1):n), suborder)

        if((array(index_l) - array(index_l+1)) > tiny(array(index_l))) then
            subarray(1:index_l) = array(1:index_l)
            suborder(1:index_l) = order(1:index_l)
            call merge_double_order(subarray, index_l, array(index_l+1), index_r, array, n, suborder, order(index_l+1), order)
        end if

    end subroutine mergesort_double_order

    subroutine merge_real_order(array_l, nl, array_r, nr, array, n, order_l, order_r, order)

        integer, intent(in) :: nl, nr, n ! dimension of arrays
        integer, intent(inout), dimension(n) :: order
        integer, intent(in), dimension(nl) :: order_l
        integer, intent(in), dimension(nr) :: order_r

        real(sp), intent(inout), dimension(n) :: array
        real(sp), intent(in), dimension(nl) :: array_l
        real(sp), intent(in), dimension(nr) :: array_r
        integer :: ll, rr, nn

        ! indices in the arrays
        ll = 1; rr = 1; nn = 1;
        do while (ll <= nl .and. rr <= nr)
            if(array_l(ll) <= array_r(rr)) then
                array(nn) = array_l(ll)
                order(nn) = order_l(ll)
                ll = ll + 1
            else
                array(nn) = array_r(rr)
                order(nn) = order_r(rr)
                rr = rr + 1
            end if
            nn = nn + 1
        end do

        do while ( ll <= nl)
            array(nn) = array_l(ll)
            order(nn) = order_l(ll)
            ll = ll + 1
            nn = nn + 1
        end do

    end subroutine merge_real_order

    recursive subroutine mergesort_real_order(array, n, subarray, order, suborder)
        implicit none

        integer, intent(in) :: n
        integer, intent(inout), dimension(n) :: order
        integer, intent(out), dimension((n+1)/2) :: suborder

        real(sp), dimension(n), intent(inout) :: array
        real(sp), dimension((n+1)/2), intent(out) :: subarray

        integer :: index_l, index_r

        if(n<2)  return
        if(n == 2) then
            if(array(1) > array(2)) then
                call swap(array(1), array(2))
                call swap(order(1), order(2))
            end if
            return
        end if

        index_l = (n+1)/2
        index_r = n - index_l

        call mergesort_real_order(array(1:index_l), index_l, subarray, order(1:index_l), suborder)
        call mergesort_real_order(array((index_l + 1) : n), index_r, subarray, order((index_l+1):n), suborder)

        if(array(index_l) > array(index_l+1)) then
            subarray(1:index_l) = array(1:index_l)
            suborder(1:index_l) = order(1:index_l)
            call merge_real_order(subarray, index_l, array(index_l+1), index_r, array, n, suborder, order(index_l+1), order)
        end if

    end subroutine mergesort_real_order

    subroutine merge_i4_order(array_l, nl, array_r, nr, array, n, order_l, order_r, order)

        integer, intent(in) :: nl, nr, n ! dimension of arrays
        integer(i4), intent(inout), dimension(n) :: array
        integer(i4), intent(in), dimension(nl) :: array_l
        integer(i4), intent(in), dimension(nr) :: array_r
        integer, intent(inout), dimension(n) :: order
        integer, intent(in), dimension(nl) :: order_l
        integer, intent(in), dimension(nr) :: order_r
        integer :: ll, rr, nn

        ! indices in the arrays
        ll = 1; rr = 1; nn = 1;
        do while (ll <= nl .and. rr <= nr)
            if(array_l(ll) <= array_r(rr)) then
                array(nn) = array_l(ll)
                order(nn) = order_l(ll)
                ll = ll + 1
            else
                array(nn) = array_r(rr)
                order(nn) = order_r(rr)
                rr = rr + 1
            end if
            nn = nn + 1
        end do

        do while ( ll <= nl)
            array(nn) = array_l(ll)
            order(nn) = order_l(ll)
            ll = ll + 1
            nn = nn + 1
        end do

    end subroutine merge_i4_order

    subroutine merge_i2_order(array_l, nl, array_r, nr, array, n, order_l, order_r, order)

        integer, intent(in) :: nl, nr, n ! dimension of arrays
        integer(i2), intent(inout), dimension(n) :: array
        integer(i2), intent(in), dimension(nl) :: array_l
        integer(i2), intent(in), dimension(nr) :: array_r
        integer, intent(inout), dimension(n) :: order
        integer, intent(in), dimension(nl) :: order_l
        integer, intent(in), dimension(nr) :: order_r
        integer :: ll, rr, nn

        ! indices in the arrays
        ll = 1; rr = 1; nn = 1;
        do while (ll <= nl .and. rr <= nr)
            if(array_l(ll) <= array_r(rr)) then
                array(nn) = array_l(ll)
                order(nn) = order_l(ll)
                ll = ll + 1
            else
                array(nn) = array_r(rr)
                order(nn) = order_r(rr)
                rr = rr + 1
            end if
            nn = nn + 1
        end do

        do while ( ll <= nl)
            array(nn) = array_l(ll)
            order(nn) = order_l(ll)
            ll = ll + 1
            nn = nn + 1
        end do

    end subroutine merge_i2_order

    subroutine merge_i1_order(array_l, nl, array_r, nr, array, n, order_l, order_r, order)

        integer, intent(in) :: nl, nr, n ! dimension of arrays
        integer(i1), intent(inout), dimension(n) :: array
        integer(i1), intent(in), dimension(nl) :: array_l
        integer(i1), intent(in), dimension(nr) :: array_r
        integer, intent(inout), dimension(n) :: order
        integer, intent(in), dimension(nl) :: order_l
        integer, intent(in), dimension(nr) :: order_r
        integer :: ll, rr, nn

        ! indices in the arrays
        ll = 1; rr = 1; nn = 1;
        do while (ll <= nl .and. rr <= nr)
            if(array_l(ll) <= array_r(rr)) then
                array(nn) = array_l(ll)
                order(nn) = order_l(ll)
                ll = ll + 1
            else
                array(nn) = array_r(rr)
                order(nn) = order_r(rr)
                rr = rr + 1
            end if
            nn = nn + 1
        end do

        do while ( ll <= nl)
            array(nn) = array_l(ll)
            order(nn) = order_l(ll)
            ll = ll + 1
            nn = nn + 1
        end do

    end subroutine merge_i1_order

    recursive subroutine mergesort_i1_order(array, n, subarray, order, suborder)
        implicit none

        integer, intent(in) :: n
        integer(i1), dimension(n), intent(inout) :: array
        integer(i1), dimension((n+1)/2), intent(out) :: subarray
        integer, dimension(n), intent(inout) :: order
        integer, dimension((n+1)/2), intent(out) :: suborder
        integer :: index_l, index_r

        if(n<2)  return
        if(n == 2) then
            if(array(1) > array(2)) then
                call swap(array(1), array(2))
                call swap(order(1), order(2))
            end if
            return
        end if

        index_l = (n+1)/2
        index_r = n - index_l

        call mergesort_i1_order(array(1:index_l), index_l, subarray, order(1:index_l), suborder)
        call mergesort_i1_order(array((index_l + 1) : n), index_r, subarray, order((index_l+1):n), suborder)

        if(array(index_l) > array(index_l+1)) then
            subarray(1:index_l) = array(1:index_l)
            suborder(1:index_l) = order(1:index_l)
            call merge_i1_order(subarray, index_l, array(index_l+1), index_r, array, n, suborder, order(index_l+1), order)
        end if

    end subroutine mergesort_i1_order

    recursive subroutine mergesort_i2_order(array, n, subarray, order, suborder)
        implicit none

        integer, intent(in) :: n
        integer(i2), dimension(n), intent(inout) :: array
        integer(i2), dimension((n+1)/2), intent(out) :: subarray
        integer, dimension(n), intent(inout) :: order
        integer, dimension((n+1)/2), intent(out) :: suborder
        integer :: index_l, index_r

        if(n<2)  return
        if(n == 2) then
            if(array(1) > array(2)) then
                call swap(array(1), array(2))
                call swap(order(1), order(2))
            end if
            return
        end if

        index_l = (n+1)/2
        index_r = n - index_l

        call mergesort_i2_order(array(1:index_l), index_l, subarray, order(1:index_l), suborder)
        call mergesort_i2_order(array((index_l + 1) : n), index_r, subarray, order((index_l+1):n), suborder)

        if(array(index_l) > array(index_l+1)) then
            subarray(1:index_l) = array(1:index_l)
            suborder(1:index_l) = order(1:index_l)
            call merge_i2_order(subarray, index_l, array(index_l+1), index_r, array, n, suborder, order(index_l+1), order)
        end if

    end subroutine mergesort_i2_order

    recursive subroutine mergesort_i4_order(array, n, subarray, order, suborder)
        implicit none

        integer, intent(in) :: n
        integer(i4), dimension(n), intent(inout) :: array
        integer(i4), dimension((n+1)/2), intent(out) :: subarray
        integer, dimension(n), intent(inout) :: order
        integer, dimension((n+1)/2), intent(out) :: suborder
        integer :: index_l, index_r

        if(n<2)  return
        if(n == 2) then
            if(array(1) > array(2)) then
                call swap(array(1), array(2))
                call swap(order(1), order(2))
            end if
            return
        end if

        index_l = (n+1)/2
        index_r = n - index_l

        call mergesort_i4_order(array(1:index_l), index_l, subarray, order(1:index_l), suborder)
        call mergesort_i4_order(array((index_l + 1) : n), index_r, subarray, order((index_l+1):n), suborder)

        if(array(index_l) > array(index_l+1)) then
            subarray(1:index_l) = array(1:index_l)
            suborder(1:index_l) = order(1:index_l)
            call merge_i4_order(subarray, index_l, array(index_l+1), index_r, array, n, suborder, order(index_l+1), order)
        end if

    end subroutine mergesort_i4_order

    integer function compare_r4_array(a1, a2)
        implicit none

        integer :: ii

        real(r4), dimension(:), intent(in) :: a1, a2

        if(size(a1) /= size(a2)) then
            compare_r4_array = false
            return
        end if

        if(size(a1) == 0 .or. size(a2) == 0) then
            compare_r4_array = false
            if(size(a1) == 0 .and. size(a2) == 0) compare_r4_array = true
            return
        end if

        compare_r4_array = true
        do ii = 1,size(a1)
            if(abs(a1(ii) - a2(ii)) <= tiny(a1(ii))) then
                cycle
            else
                compare_r4_array = false
                exit
            end if
        end do

    end function compare_r4_array

    integer function compare_r8_array(a1, a2)
        implicit none

        integer :: ii

        real(r8), dimension(:), intent(in) :: a1, a2

        if(size(a1) /= size(a2)) then
            compare_r8_array = false
            return
        end if

        if(size(a1) == 0 .or. size(a2) == 0) then
            compare_r8_array = false
            if(size(a1) == 0 .and. size(a2) == 0) compare_r8_array = true
            return
        end if

        compare_r8_array = true
        do ii = 1,size(a1)
            if(abs(a1(ii) - a2(ii)) <= tiny(a1(ii))) then
                cycle
            else
                compare_r8_array = false
                exit
            end if
        end do

    end function compare_r8_array

    integer function compare_i4_array(a1, a2)
        implicit none

        integer :: ii

        integer(i4), dimension(:), intent(in) :: a1, a2

        compare_i4_array = true

        if(size(a1) /= size(a2)) then
            compare_i4_array = false
            return
        end if

        if(size(a1) == 0 .or. size(a2) == 0) then
            compare_i4_array = false
            if(size(a1) == 0 .and. size(a2) == 0) compare_i4_array = true
            return
        end if

        do ii = 1, size(a1)
            if(a1(ii) == a2(ii)) then
                cycle
            else
                compare_i4_array = false
                exit
            end if
        end do

    end function compare_i4_array

    integer function compare_i2_array(a1, a2)
        implicit none

        integer :: ii

        integer(i2), dimension(:), intent(in) :: a1, a2

        if(size(a1) /= size(a2)) then
            compare_i2_array = false
            return
        end if

        if(size(a1) == 0 .or. size(a2) == 0) then
            compare_i2_array = false
            if(size(a1) == 0 .and. size(a2) == 0) compare_i2_array = true
            return
        end if

        compare_i2_array = true
        do ii = 1,size(a1)
            if(a1(ii) == a2(ii)) then
                cycle
            else
                compare_i2_array = false
                exit
            end if
        end do

    end function compare_i2_array

    integer function compare_i1_array(a1, a2)
        implicit none

        integer :: ii

        integer(i1), dimension(:), intent(in) :: a1, a2

        if(size(a1) /= size(a2)) then
            compare_i1_array = false
            return
        end if

        if(size(a1) == 0 .or. size(a2) == 0) then
            compare_i1_array = false
            if(size(a1) == 0 .and. size(a2) == 0) compare_i1_array = true
            return
        end if

        compare_i1_array = true
        do ii = 1,size(a1)
            if(a1(ii) == a2(ii)) then
                cycle
            else
                compare_i1_array = false
                exit
            end if
        end do

    end function compare_i1_array

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> create the union of two integer arrays, with unique entries
    !
    !
    !
    !> @param[in] a
    !> @param[out] b, c
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine union_i4(a, b, c, status)
        implicit none

        integer, intent(inout) :: status
        integer(i4), dimension(:) :: a, b
        integer(i4), allocatable, dimension(:) :: c
        integer(i4), dimension(size(a) + size(b)) :: tmp
        logical, allocatable, dimension(:) :: mask
        integer :: ii

        character(*), parameter :: function = "union_i4"

        if(status > 100 .or. status == s_errRep) return

        call alloc(c, size(a) + size(b), status)
        call alloc(mask, size(a) + size(b), status)

        tmp = [a, b]
        call sort_i4(tmp, status=status)

        mask = .true.
        do ii = 2, size(tmp)
            if(tmp(ii) == tmp(ii-1)) mask(ii) = .false.
        end do

        c = pack(tmp, mask)

    end subroutine union_i4

        ! merge double arrays like in gti, this routine is used to merge energy intervals

    subroutine merge_or_r8(a1, a2, na, b1, b2, nb, c1, c2, status)
        implicit none

        integer, intent(inout) :: status
        integer :: nI, ii, n

        integer, intent(in) :: na, nb

        real(r8), dimension(na), intent(in) :: a1, a2
        real(r8), dimension(nb), intent(in) :: b1, b2
        real(r8), dimension(na + nb) :: tmp1, tmp2

        ! allocatable
        real(r8), dimension(:), allocatable, intent(out) :: c1, c2
        integer, allocatable, dimension(:) :: order

        character(*), parameter :: function = "merge_or_r8"

        if(status > 100 .or. status == s_errRep) return

        n = na + nb

        call alloc(c1, n, status)
        call alloc(c2, n, status)

        tmp1 = [a1, b1]
        tmp2 = [a2, b2]

!        call sort(tmp1, order, status);
        call sort(tmp1, status, order);
        tmp2 = tmp2(order)
        call dealloc(order, status)

        if(na == 0 .or. nb == 0) then
            return
        end if

        nI = 1
        c1(nI) = tmp1(1)
        c2(nI) = tmp2(1)

        do ii = 2, n
            if(c2(nI) < tmp1(ii)) then
                nI = nI + 1
                c1(nI) = tmp1(ii)
                c2(nI) = tmp2(ii)
            else
                if(c2(nI) >= tmp2(ii)) cycle
                ! else c2(nI) > tmp1(ii) .and. c2(nI) < tmp2(ii)
                c2(nI) = tmp2(ii)
            end if
        end do

        c1 = c1(:nI)
        c2 = c2(:nI)

    end subroutine merge_or_r8

    ! merge double arrays like in gti, this routine is used to merge energy intervals

    subroutine merge_and_r8(a1, a2, na, b1, b2, nb, c1, c2, status)
        implicit none

        integer, intent(in) :: na, nb
        real(r8), dimension(na), intent(in) :: a1, a2
        real(r8), dimension(nb), intent(in) :: b1, b2
        real(r8), dimension(na + nb) :: tmp1, tmp2
        real(r8) :: last_start, last_stop


        integer, intent(inout) :: status
        integer :: nI, ii, n, last
        integer, dimension(na) :: a_index
        integer, dimension(size(b1)) :: b_index
        integer, dimension(na + size(b1)) :: c_index

        real(r8), dimension(:), allocatable, intent(out) :: c1, c2
        integer, dimension(:), allocatable :: order


        character(*), parameter :: function = "merge_and_r8"

        if(status > 100 .or. status == s_errRep) return

        n = na + nb

        a_index(:) = 0
        b_index(:) = 1
        c_index = [a_index, b_index]

        call alloc(c1, n, status)
        call alloc(c2, n, status)

        tmp1 = [a1, b1]
        tmp2 = [a2, b2]

!        call sort(tmp1, order, status)
        call sort(tmp1, status, order)
        tmp2 = tmp2(order)
        c_index = c_index(order)
        call dealloc(order, status)

        nI = 1 ! next merged Interval number
        last_start = tmp1(1)
        last_stop = tmp2(1)
        last = c_index(1)

        do ii = 2, n

            if(last == c_index(ii)) then
                last_start = tmp1(ii)
                last_stop = tmp2(ii)
                cycle ! check the next interval
            end if

            c1(nI) = max(last_start, tmp1(ii))
            c2(nI) = min(last_stop, tmp2(ii))
            if(c1(nI) < c2(nI)) then
                last_start = c2(nI)
                nI = nI + 1
                if(last_stop < tmp2(ii)) then
                    last_stop = tmp2(ii)
                    last = c_index(ii)
                end if
            else ! no intersection
                last_start = tmp1(ii)
                last_stop = tmp2(ii)
                last = c_index(ii)
            end if

        end do

        if(nI == 1) then
            call wrerr(function, s_err_no_exit, "No intersecting intervals.")
! 20210819 IMS: should change all these error codes from s_err to s_err_no_exit, because the former stops execution.
            status = s_errRep
        end if
        ! since its next Interval number subtract 1
        c1 = c1(:nI-1)
        c2 = c2(:nI-1)

    end subroutine merge_and_r8

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return an array of indices indicating unique entries in an array
    !
    !
    !
    !> @param[in] array
    !> @param[out] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function unique_r8(array, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii
        ! the input array
        real(r8), dimension(:), intent(in) :: array
        real(r8), dimension(size(array)) :: tmp_array
        integer, dimension(:), allocatable :: unique_r8, mask

        character(*), parameter :: function = "unique_r8"

        if(status > 100 .or. status == s_errRep) return

        ! check if size = 1 since nothing needs to be done here
        if(size(array) == 1) then
            call alloc(unique_r8, 1, status)
            unique_r8 = 1
            return
        end if

        tmp_array = array

!        call sort(tmp_array, unique_r8, status)
        call sort(tmp_array, status, unique_r8)
        call alloc(mask, size(tmp_array), status)

        mask = true
        do ii = 2, size(tmp_array)
            if(nearest(tmp_array(ii), 1.0_r8) == nearest(tmp_array(ii-1), 1.0_r8)) mask(ii) = false
        end do

        unique_r8 = pack(unique_r8, mask == true)
        call dealloc(mask, status)

    end function unique_r8

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return an array of indices indicating unique entries in an array
    !
    !
    !
    !> @param[in] array
    !> @param[out] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function unique_i4(array, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii
        ! the input array
        integer(i4), dimension(:), intent(in) :: array
        integer(i4), dimension(size(array)) :: tmp_array
        integer, dimension(:), allocatable :: unique_i4, mask

        character(*), parameter :: function = "unique_i4"

        if(status > 100 .or. status == s_errRep) return

        ! check if size = 1 since nothing needs to be done here
        if(size(array) == 1) then
            call alloc(unique_i4, 1, status)
            unique_i4 = 1
            return
        end if

        tmp_array = array

!        call sort(tmp_array, unique_i4, status)
        call sort(tmp_array, status, unique_i4)
        call alloc(mask, size(tmp_array), status)

        mask = true
        do ii = 2, size(tmp_array)
            if(tmp_array(ii) == tmp_array(ii-1)) mask(ii) = false
        end do

        unique_i4 = pack(unique_i4, mask == true)
        call dealloc(mask, status)

    end function unique_i4

    function unique_i2(array, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii
        ! the input array
        integer(i2), dimension(:), intent(in) :: array
        integer(i2), dimension(size(array)) :: tmp_array
        integer, dimension(:), allocatable :: unique_i2, mask

        character(*), parameter :: function = "unique_i2"

        if(status > 100 .or. status == s_errRep) return

        ! check if size = 1 since nothing needs to be done here
        if(size(array) == 1) then
            call alloc(unique_i2, 1, status)
            unique_i2 = 1
            return
        end if

        tmp_array = array

        call sort(tmp_array, status, unique_i2)
        call alloc(mask, size(tmp_array), status)

        mask = true
        do ii = 2, size(tmp_array)
            if(tmp_array(ii) == tmp_array(ii-1)) mask(ii) = false
        end do

        unique_i2 = pack(unique_i2, mask == true)
        call dealloc(mask, status)

    end function unique_i2

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return an array of indices of entries in an array which are contained in a second array
    !
    !
    !
    !> @param[in] array
    !> @param[out] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function where_i2(array, comp_array, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii
        ! the input array
        integer(i2), dimension(:), intent(in) :: array, comp_array
        integer, dimension(:), allocatable :: where_i2, mask

        character(*), parameter :: function = "where_i2"

        if(status > 100 .or. status == s_errRep) return

        call alloc(where_i2, size(array), status)
        if(size(where_i2) == 0) return

        call alloc(mask, size(array), status)
        where_i2 = [(ii,ii=1,size(array),1)]
        mask = false
        do ii = 1, size(comp_array)
            where (array == comp_array(ii))
                mask = true
            end where
        end do

        where_i2 = pack(where_i2, mask == true)
        call dealloc(mask, status)

    end function where_i2

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return an array of indices of entries in an array which are contained in a second array
    !
    !
    !
    !> @param[in] array
    !> @param[out] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function where_i4(array, comp_array, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii
        ! the input array
        integer(i4), dimension(:), intent(in) :: array, comp_array
        integer, dimension(:), allocatable :: where_i4, mask

        character(*), parameter :: function = "where_i4"

        if(status > 100 .or. status == s_errRep) return

        call alloc(where_i4, size(array), status)
        if(size(where_i4) == 0) return

        call alloc(mask, size(array), status)
        where_i4 = [(ii,ii=1,size(array),1)]
        mask = false
        do ii = 1, size(comp_array)
            where (array == comp_array(ii))
                mask = true
            end where
        end do

        where_i4 = pack(where_i4, mask == true)
        call dealloc(mask, status)

    end function where_i4

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return an array of indices of entries in an array which are contained in a second array
    !
    !
    !
    !> @param[in] array
    !> @param[out] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function where_r8(array, comp_array, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii
        ! the input array
        real(r8), dimension(:), intent(in) :: array, comp_array
        integer, dimension(:), allocatable :: where_r8, mask

        character(*), parameter :: function = "where_r8"

        if(status > 100 .or. status == s_errRep) return

        call alloc(where_r8, size(array), status)
        if(size(where_r8) == 0) return

        call alloc(mask, size(array), status)
        where_r8 = [(ii,ii=1,size(array),1)]
        mask = false
        do ii = 1, size(comp_array)
            where (abs(array - comp_array(ii)) < epsilon(comp_array(ii)))
                mask = true
            end where
        end do

        where_r8 = pack(where_r8, mask == true)
        call dealloc(mask, status)

    end function where_r8

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return an array of indices of entries in an array which are contained in a second array
    !
    !
    !
    !> @param[in] array
    !> @param[out] order
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine where_i4_sub(array, comp_array, order, status, complement)
        implicit none

        integer, intent(inout) :: status
        integer :: ii
        ! the input array
        integer(i4), dimension(:), intent(in) :: array, comp_array
        integer, dimension(:), allocatable :: order, mask
        integer, dimension(:), allocatable, intent(out), optional :: complement

        character(*), parameter :: function = "where_i4_sub"

        if(status > 100 .or. status == s_errRep) return

        call alloc(order, size(array), status)
        if(size(order) == 0) then
            ! if the size of the original array was zero so must be the complement
            if(present(complement)) call alloc(complement, 0, status)
            return
        end if

        call alloc(mask, size(array), status)
        order = [(ii,ii=1,size(array),1)]
        mask = false
        do ii = 1, size(comp_array)
            where (array == comp_array(ii))
                mask = true
            end where
        end do

        if(present(complement)) then
            call alloc(complement, size(array), status)
            complement = [(ii,ii=1,size(array),1)]
            complement = pack(complement, mask == false)
        end if

        order = pack(order, mask == true)
        call dealloc(mask, status)

    end subroutine where_i4_sub

      !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return an array of indices of entries in an array which are contained in a second array of (characters)
    !
    !
    !
    !> @param[in] array, comp_array
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    function where_str(array, comp_array, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii
        ! the input array
        character(*), dimension(:), intent(in) :: array, comp_array
        integer, dimension(:), allocatable :: where_str, mask

        character(*), parameter :: function = "where_str"

        if(status > 100 .or. status == s_errRep) return

        call alloc(where_str, size(array), status)
        if(size(where_str) == 0) return

        call alloc(mask, size(array), status)
        where_str = [(ii,ii=1,size(array),1)]
        mask = false
        do ii = 1, size(comp_array)
            where (array == comp_array(ii))
                mask = true
            end where
        end do

        where_str = pack(where_str, mask == true)
        call dealloc(mask, status)

    end function where_str

  !.....................................................................
  subroutine merge_arrays(array1, array2, outarray, status)
    ! Only implemented for integers at present.

    implicit none

    integer, intent(inout) :: status
    integer, dimension(:), intent(in) :: array1, array2
    integer, allocatable, dimension(:), intent(out) :: outarray
    integer, dimension(size(array1) + size(array2)) :: tmp
    integer, allocatable, dimension(:) :: order

    logical, allocatable, dimension(:) :: mask
    integer :: ii

    character(*), parameter :: function = "merge_arrays"

    if(status > 100 .or. status == s_errRep) return

    call alloc(mask, size(array1) + size(array2), status)

    tmp = [array1, array2]

    call sort(tmp, status, order)

    mask = .true.
    do ii = 2, size(tmp)
      if(tmp(ii) == tmp(ii-1)) mask(ii) = .false.
    end do

    call alloc(outarray, size(array1) + size(array2), status)

    outarray = pack(tmp, mask)

    call dealloc(order, status)
    call dealloc(mask, status)

  end subroutine merge_arrays

end module erolib_sort_mod
