! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! MODULE: strmod
!
! DESCRIPTION:
!> string manipulation functions and subroutines
!
!------------------------------------------------------------------------------
! Subroutines contained:
!  subroutine split_on_whitespace(in_string, out_strings)
!  subroutine split_file_path(path, dir, file_name)
!  subroutine get_next_noncomment_line(uno, line, eof)
!    subroutine strtok(str, c, outstr)
!    subroutine strdel_substr(string, substring, first)
!    subroutine strreplace_substr(string, substring, replace, first)
!    subroutine init_random_seed()
!    function str_tmpfname (dir)
!    subroutine print_array(array)
!    subroutine str_upper(str)
!    subroutine str_lower(str)
!    interface int2str
!    character(buffsize) function real2str(k, status)
!    character(buffsize) function double2str(k, status)
!    character(buffsize) function hex2str(k, status)
!    integer function str2int(k, status)
!    real function str2real(k, status)
!    real(dp) function str2double(k, status)
!    integer function str2hex(k, status)
!    logical function str2bool(k)
!    subroutine compare_str (str_1, str_2, same, status, upper)
!
module erolib_str_mod
    
    use erolib_general_mod
    implicit none

    interface int2str
        module procedure i1_2str, i2_2str, i4_2str, i8_2str
    end interface int2str

contains

  !.....................................................................
  subroutine binary_str_to_ints(bin_str, bin_int_0s, bin_int_1s)
    ! The input argument is expected to be a string of the form '0bccccc' in which each 'c' may be either '0', '1', 'x', or a space, and in which the number of non-space cs must not exceed the number of bits in an integer type. It is interpreted as a combination of two binary masks, one of which sets the 0s, the other the 1s. The last or right-most character is taken to be the least significant bit of the masks. The routine separates out the masks and returns both. 
    !
    implicit none
    character(*), intent(in)  :: bin_str
    integer,      intent(out) :: bin_int_0s, bin_int_1s

    character(*), parameter :: functionName = "binary_str_to_ints"
    integer :: n_char,n_bits,bi,ii
    character(buffsize) :: istr

    bin_int_0s = 0
    bin_int_1s = 0

    n_char = len_trim(bin_str)
    if(n_char < 1)&
call wrstop(functionName, "Input string is empty.")

    if(bin_str(1:2) /= "0b" .and. bin_str(1:2) /= "0B")&
call wrstop(functionName, "Input string does not begin with '0b'.")

    n_bits = bit_size(bin_int_0s)
    istr = int2str(n_bits)

    bi = 0
    do ii=n_char,3,-1
      if(bin_str(ii:ii)==' ')&
    cycle

      bi = bi + 1
      if(bi > n_bits)&
call wrstop(functionName, "The input string may not contain more than "//trim(istr)//" bit positions.")

      if(     bin_str(ii:ii)=='0') then
        bin_int_0s = ibset(bin_int_0s, bi-1) ! bi-1 because bits are counted from 0.
      else if(bin_str(ii:ii)=='1') then
        bin_int_1s = ibset(bin_int_1s, bi-1) ! bi-1 because bits are counted from 0.
      ! Note that we in fact allow any character other than 0, 1 or space to count as 'x'.
      end if
    end do

  end subroutine binary_str_to_ints

  !.....................................................................
  subroutine binary_ints_to_str(bin_int_0s, bin_int_1s, bin_str)
    ! This is the reverse of binary_str_to_ints(). Here the 0s and 1s masks are supplied, and the corresponding string is returned. Note that an error occurs if the same bit is set in both masks. 
    !
    implicit none
    integer,      intent(in)  :: bin_int_0s, bin_int_1s
    character(*), intent(out) :: bin_str

    character(*), parameter :: functionName = "binary_ints_to_str"
    integer :: n_bits,bi
    character(1) :: mychar

    if(iand(bin_int_0s, bin_int_1s)/=0)&
call wrstop(functionName, "The same bit is set in both input masks.")

    bin_str = "0b"
    n_bits = bit_size(bin_int_0s)
    do bi=n_bits,1,-1
      if(     btest(bin_int_0s, bi-1)) then
        mychar = '0'
      else if(btest(bin_int_1s, bi-1)) then
        mychar = '1'
      else
        mychar = 'x'
      end if

      bin_str = trim(bin_str)//mychar
    end do

  end subroutine binary_ints_to_str

  !.....................................................................
  subroutine split_on_whitespace(in_string, out_strings)
    implicit none
    character(*),                            intent(in)  :: in_string
    character(*), dimension(:), allocatable, intent(out) :: out_strings

    character(*), parameter :: tab=char(9)
    integer :: n_chars,n_non_space,ii,jj,i_start,i_finis
    logical :: in_a_space_sequence

    ! Go twice over every character in the string - the first time to calculate the number of non-space stretches, the second to fill the out array.
    !
    n_chars = len_trim(in_string)

    n_non_space = 0
    in_a_space_sequence = .true.
    do ii=1,n_chars
      if(in_string(ii:ii)==" " .or. in_string(ii:ii)==tab) then
        if(.not.in_a_space_sequence) then
          in_a_space_sequence = .true.
        end if
      else
        if(in_a_space_sequence) then
          in_a_space_sequence = .false.
          n_non_space = n_non_space + 1
        end if
      end if
    end do

    allocate(out_strings(n_non_space))

    jj = 0
    in_a_space_sequence = .true.
    i_start = 0
    do ii=1,n_chars
      if(in_string(ii:ii)==" " .or. in_string(ii:ii)==tab) then
        if(.not.in_a_space_sequence) then
          in_a_space_sequence = .true.
          i_finis = ii - 1
          if(jj>0) then
            out_strings(jj) = in_string(i_start:i_finis)
          end if
        end if
      else
        if(in_a_space_sequence) then
          in_a_space_sequence = .false.
          jj = jj + 1
          i_start = ii
        end if
      end if
    end do
    if(.not.in_a_space_sequence) then
      i_finis = n_chars
      if(jj>0) then
        out_strings(jj) = in_string(i_start:i_finis)
      end if
    end if

  end subroutine split_on_whitespace

  !.....................................................................
  subroutine split_file_path(path, dir, file_name)
    implicit none
    character(*), intent(in) :: path
    character(*), intent(out) :: dir, file_name

    integer :: pos_last_slash

    dir = ""
    file_name = ""

    if(len_trim(path)<1)&
  return

    pos_last_slash = index(path, "/", back=.true.)
    if(pos_last_slash==0) then
      file_name = trim(path)
  return
    end if

    dir       = path(:pos_last_slash-1)
    file_name = path(pos_last_slash+1:)

  end subroutine split_file_path

  !.....................................................................
  subroutine get_next_noncomment_line(uno, line, eof)
    ! This skips lines that consist either entirely of white space or
    ! that begin with white space followed by a # character.
    !
!*** should put it in erolib_str_mod.
    implicit none
    integer,      intent(in)  :: uno
    character(*), intent(out) :: line
    logical,      intent(out) :: eof

    integer :: hashloc

    eof = .false. ! default

    do
      read(uno,'(a)',end=1) line
      ! Loop if the line begins with any number of blanks optionally followed by a hash:
      hashloc = index(line, "#")
      if(hashloc>0)&
        line = line(1:hashloc-1)

      if(trim(line)/="")&
  return
    end do

1   eof = .true.

  end subroutine get_next_noncomment_line

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> split the input string into substring indicated by a supplied character
    !
    !
    !
    !> @param[in] str
    !> @param[in] c
    !> @param[out] outstr
    !---------------------------------------------------------------------------

    subroutine strtok(str, c, outstr)
        implicit none

        character(*), intent( in ):: str
        character, intent( in ) :: c
        character(buffsize), allocatable, dimension(:), intent( out ) :: outstr
        character(*), parameter :: function = "strtok"
        character(buffsize) :: errmsg
        
        integer :: status, ii
        integer :: counter, cmax, maxlength, stat
        integer :: pos ! last position in string
        integer :: newpos ! new position in string
        integer :: first, next ! first and next position of character in substring

        ! determine max length of the input string
        maxlength = len_trim(str)

        ! initialise the default values in each run of the function
        counter = 0
        cmax = 0
        pos = 0
        newpos = 1
        status = 0

        ! determine the number of the characters c in string str
        newpos = index(str(pos + 1 : maxlength), c)
        ! the character was not found in the str
        if (newpos == 0) then
            call alloc(outstr, 1, status)
            outstr = str
            !            if(allocated(outstr)) deallocate(outstr)
            ! allocate(outstr(1), source=str)
            return
        end if

        counter = 1
        do ii = 1, len_trim(str)
            if(str(ii:ii) == c) then
                if(ii > 1) then
                    if(str(ii-1:ii-1) /= str(ii:ii)) counter = counter + 1
                end if
                if(ii == 1) then
                    counter = counter + 1
                end if
                if(ii == len_trim(str)) counter = counter - 1
            end if
        end do

        !        if(allocated(outstr)) deallocate(outstr, stat=stat, errmsg=errmsg)
        !        call FError(function, stat, errmsg, status)

        ! determine the number substrings and allocate them
        cmax = counter
        errmsg=""
        allocate(outstr(cmax), stat=stat, errmsg=errmsg)
        call FError(function, stat, errmsg, status)
        if(status == s_err) return

        ! reset the positions and counter
        next = 1
        first = 1
        newpos = 1
        counter = 0

        ! fill the array of strings with the substrings
        do while(counter < cmax)
            first = next
            counter = counter + 1
            newpos = index(str( first : maxlength), c)
            do while(newpos == 1)
                first = first + 1
                newpos = index(str(first : maxlength), c)
            end do

            next = first + newpos
            if(newpos /= 0) then
                outstr(counter) = str(first : next - 2)
            else
                outstr(counter) = str(first : maxlength)
            end if

        end do

    end subroutine strtok

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> remove first substring occurence in string and shift all characters to fill the gap
    !> optional parameter to delete all occurences
    !
    !
    !> @param[in] string
    !> @param[out] substring
    !> @param[inout,optional] status
    !---------------------------------------------------------------------------

    subroutine strdel_substr(string, substring, first)
        implicit none

        character(*), parameter :: function = "strdel_substr"
        character(*), intent(inout) :: string
        character(*), intent(in) :: substring
        
        logical, intent(in), optional :: first

        integer :: single, pos1
        integer :: len_substr;
        integer :: counter
        
        counter = 0
        single = 1;
        pos1 = 1;
        len_substr = len_trim(substring)
        if(len_substr == 0) len_substr = len(substring)

        if(present(first)) then
            if(first .eqv. .false.) single = 0
        end if

        do while (pos1 /= 0)
            counter = counter + 1
            pos1 = index(trim(string), substring)
            if(pos1 == 0) return
            if(pos1 == 1) then
                string = string(pos1 + len_substr : len_trim(string))
            else
                string = string(1 : pos1 - 1) // string(pos1 + len_substr : len_trim(string))
            end if
            if(single == 1) return
        end do

    end subroutine strdel_substr

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> replace first substring occurence in string with a new string
    !> optional parameter to delete all occurences
    !
    !
    !> @param[in] string
    !> @param[out] substring
    !> @param[inout,optional] status
    !---------------------------------------------------------------------------

    subroutine strreplace_substr(string, substring, replace, first)
        implicit none

        character(*), parameter :: function = "strdel_substr"
        character(*), intent(inout) :: string
        character(*), intent(in) :: substring, replace
        logical, intent(in), optional :: first

        integer :: single, pos1
        integer :: len_substr;
        integer :: counter

        counter = 0
        single = 1;
        pos1 = 1;
        len_substr = len_trim(substring)
        if(len_substr == 0) len_substr = len(substring)

        if(present(first)) then
            if(first .eqv. .false.) single = 0
        end if

        do while (pos1 /= 0)
            counter = counter + 1
            pos1 = index(trim(string), substring)
            if(pos1 == 0) return
            if(pos1 == 1) then
                string = replace // string(pos1 + len_substr : len_trim(string))
            else
                string = string(1 : pos1 - 1) // replace // string(pos1 + len_substr : len_trim(string))
            end if
            if(single == 1) return
        end do

    end subroutine strreplace_substr

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> init a random seed for the random number generator
    !> is the example given by the gfortran manpage
    !
    !
    !---------------------------------------------------------------------------

    subroutine init_random_seed()
        implicit none
        
        integer, allocatable :: seed(:)
        integer :: i, n, dt(8), pid, t(2), s, stat, status
        integer(8) :: count, tms
		
        character(*), parameter :: function = "init_random_seed"
        character(buffsize) :: errmsg

        stat = 0
        call random_seed(size = n)
        errmsg=""
        allocate(seed(n), stat=stat, errmsg=errmsg)
        call FError(function // " allocate seed(" // trim(int2str(n)) // ")", stat, errmsg, status)
        
        ! think about a better error handling
        if(stat /= 0) then
            print *, "critical error - memory allocation failed"
            stop
        end if
        
        call system_clock(count)
        if (count /= 0) then
            t = transfer(count, t)
        else
            call date_and_time(values=dt)
            tms = (dt(1) - 1970) * 365_8 * 24 * 60 * 60 * 1000 &
                + dt(2) * 31_8 * 24 * 60 * 60 * 1000 &
                + dt(3) * 24 * 60 * 60 * 60 * 1000 &
                + dt(5) * 60 * 60 * 1000 &
                + dt(6) * 60 * 1000 + dt(7) * 1000 &
                + dt(8)
            t = transfer(tms, t)
        end if
        s = ieor(t(1), t(2))
        pid = getpid() + 1099279 ! Add a prime
        s = ieor(s, pid)
        if (n >= 3) then
            seed(1) = t(1) + 36269
            seed(2) = t(2) + 72551
            seed(3) = pid
            if (n > 3) then
                seed(4:) = s + 37 * (/ (i, i = 0, n - 4) /)
            end if
        else
            seed = s + 37 * (/ (i, i = 0, n - 1 ) /)
        end if

        call random_seed(put=seed)
        
    end subroutine init_random_seed


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return a temporary file name
    !
    !
    !
    !> @param[out] filename
    !---------------------------------------------------------------------------

    function str_tmpfname (dir)
        implicit none
        
        character(9) :: fmt
        character(buffsize) :: str_tmpfname
        character(*), optional, intent(in) :: dir
        character(buffsize) :: tmp
        
        real :: r_number
        
        integer :: i_number

        call init_random_seed()
        call random_number(r_number)
        i_number = int(r_number * 10**8)
        fmt = '(I8.8)'
        write (tmp, fmt) i_number
        str_tmpfname = "tmp_" // trim(tmp)

        if(present(dir)) str_tmpfname = trim(dir) // "/tmp_" // trim(tmp)

    end function str_tmpfname

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> print an array of strings with trim
    !
    !
    !
    !> @param[in] array
    !> @param[out] status
    !---------------------------------------------------------------------------

    subroutine print_array(array)
        implicit none
        
        integer :: ii
        
        character(*), dimension(:), intent(in) :: array

        do ii = 1, size(array)
            print *, trim(array(ii))
        end do

    end subroutine print_array

    ! routines taken from http://rosettacode.org/wiki/String_case#Fortran

    subroutine str_upper(str)
        implicit none
        character(*), intent(in out) :: str
        integer :: i

        do i = 1, len(str)
            select case(str(i:i))
                case("a":"z")
                    str(i:i) = achar(iachar(str(i:i))-32)
            end select
        end do
    end subroutine str_upper

    subroutine str_lower(str)
        implicit none
        
        character(*), intent(in out) :: str
        integer :: i

        do i = 1, len(str)
            select case(str(i:i))
                case("A":"Z")
                    str(i:i) = achar(iachar(str(i:i))+32)
            end select
        end do
    end subroutine str_lower

            !   "Convert an integer to string."
    character(buffsize) function i8_2str(k, status, format)
        implicit none
        
        integer :: iostat
        integer(i8), intent(in) :: k
        integer, intent(inout), optional :: status
        character(*), intent(in), optional :: format
        
        character(*), parameter :: function = "i8_2str"
        character(buffsize) :: iomsg
        
        iomsg=""
        if(.not. present (format)) then
            write (i8_2str, *, iostat=iostat, iomsg=iomsg) k
        else
            write(i8_2str, format, iostat=iostat, iomsg=iomsg) k
        end if

        if(present(status)) call FError(function, iostat, iomsg, status)
        i8_2str = adjustl(i8_2str)

    end function i8_2str

            !   "Convert an integer to string."
    character(buffsize) function i4_2str(k, status, format)
        implicit none
        
        integer :: iostat
        integer, intent(in) :: k
        integer, intent(inout), optional :: status
        character(*), intent(in), optional :: format
        
        character(*), parameter :: function = "i4_2str"
        character(buffsize) :: iomsg
        
        iomsg=""
        if(.not. present (format)) then
            write (i4_2str, *, iostat=iostat, iomsg=iomsg) k
        else
            write(i4_2str, format, iostat=iostat, iomsg=iomsg) k
        end if

        if(present(status)) call FError(function, iostat, iomsg, status)
        i4_2str = adjustl(i4_2str)

    end function i4_2str

    character(buffsize) function i2_2str(k, status, format)
        implicit none

        integer :: iostat
        integer(i2), intent(in) :: k
        integer, intent(inout), optional :: status
        character(*), intent(in), optional :: format

        character(*), parameter :: function = "i2_2str"
        character(buffsize) :: iomsg

        iomsg=""
        if(.not. present (format)) then
            write (i2_2str, *, iostat=iostat, iomsg=iomsg) k
        else
            write(i2_2str, format, iostat=iostat, iomsg=iomsg) k
        end if
        if(present(status)) call FError(function, iostat, iomsg, status)
        i2_2str = adjustl(i2_2str)

    end function i2_2str

    character(buffsize) function i1_2str(k, status, format)
        implicit none

        integer :: iostat
        integer(i1), intent(in) :: k
        integer, intent(inout), optional :: status
        character(*), intent(in), optional :: format

        character(*), parameter :: function = "i1_2str"
        character(buffsize) :: iomsg

        iomsg=""
        if(.not. present (format)) then
            write (i1_2str, *, iostat=iostat, iomsg=iomsg) k
        else
            write(i1_2str, format, iostat=iostat, iomsg=iomsg) k
        end if
        if(present(status)) call FError(function, iostat, iomsg, status)
        i1_2str = adjustl(i1_2str)

    end function i1_2str


        !   "Convert a real to string."
    character(buffsize) function real2str(k, status, max_n_dec)
        implicit none

        real,    intent(in)              :: k
        integer, intent(inout), optional :: status
        integer, intent(in),    optional :: max_n_dec ! maximum number of digits desired after the decimal point.

        character(*), parameter :: function = "real2str"
        character(buffsize) :: iomsg
        integer :: iostat
        integer :: posn_of_dec_point,n_dec

        iomsg=""
        write (real2str, *, iostat=iostat, iomsg=iomsg) k
        if(present(status)) call FError(function, iostat, iomsg, status)
        real2str = adjustl(real2str)

        if(present(max_n_dec)) then
          posn_of_dec_point = index(real2str, ".")
          if(posn_of_dec_point>0) then
            n_dec = len_trim(real2str) - posn_of_dec_point
            if(n_dec > max_n_dec) real2str = real2str(1:(posn_of_dec_point + max_n_dec))
          end if
        end if

    end function real2str

            !   "Convert a double to string."
    character(buffsize) function double2str(k, status, max_n_dec)
        implicit none

        real(kind=8), intent(in)              :: k
        integer,      intent(inout), optional :: status
        integer,      intent(in),    optional :: max_n_dec ! maximum number of digits desired after the decimal point.

        character(*), parameter :: function = "double2str"
        character(buffsize) :: iomsg
        integer :: iostat
        integer :: posn_of_dec_point,n_dec

        iomsg=""
        write (double2str, *, iostat=iostat, iomsg=iomsg) k
        if(present(status)) call FError(function, iostat, iomsg, status)
        double2str = adjustl(double2str)

        if(present(max_n_dec)) then
          posn_of_dec_point = index(double2str, ".")
          if(posn_of_dec_point>0) then
            n_dec = len_trim(double2str) - posn_of_dec_point
            if(n_dec > max_n_dec) double2str = double2str(1:(posn_of_dec_point + max_n_dec))
          end if
        end if

    end function double2str

                !   "Convert a hex to string."
    character(buffsize) function hex2str(k, status)
        implicit none

        integer :: iostat
        integer, intent(in) :: k
        integer, intent(inout), optional :: status

        character(*), parameter :: function = "hex2str"
        character(buffsize) :: iomsg

        iomsg=""
        write (hex2str, '(Z8)', iostat=iostat, iomsg=iomsg) k
        if(present(status)) call FError(function, iostat, iomsg, status)
        hex2str = adjustl(hex2str)

    end function hex2str

        ! convert a string to int
    integer function str2int(k, status)
        implicit none
        
        integer :: iostat
        integer, intent(inout), optional :: status
        
        character(*), intent(in) :: k
        character(*), parameter :: function = "str2int"
        character(buffsize) :: iomsg
        
        iomsg=""
        read (k, *, iostat=iostat, iomsg=iomsg) str2int
        if(present(status)) call FError(function, iostat, iomsg, status)
    
    end function

    ! convert a string to real
    real function str2real(k, status)
        implicit none
        
        integer :: iostat!, tmp_str2int
        integer, intent(inout), optional :: status
        
        character(*), intent(in) :: k
        character(*), parameter :: function = "str2real"
        character(buffsize) :: iomsg
        
        iomsg=""
        read (k, *, iostat=iostat, iomsg=iomsg) str2real
        
        if(present(status)) call FError(function, iostat, iomsg, status)
        
    end function

    ! convert a string to double
    real(dp) function str2double(k, status)
        implicit none
        
        integer :: iostat!, tmp_str2int
        integer, intent(inout), optional :: status
        
        character(*), intent(in) :: k
        character(*), parameter :: function = "str2double"
        character(buffsize) :: iomsg
        
        iomsg=""
        read (k, *, iostat=iostat, iomsg=iomsg) str2double
        
        if(present(status)) then 
            call FError(function, iostat, iomsg, status)
        end if
        
    end function

    ! convert a string into a number in hex
    integer function str2hex(k, status)
        implicit none

        integer :: iostat!, tmp_str2int
        integer, intent(inout), optional :: status

        character(*), intent(in) :: k
        character(buffsize) :: strhex
        character(*), parameter :: function = "str2hex"
        character(buffsize) :: iomsg

        strhex = k
        if(k(1:2) == "0x") strhex = strhex(3:len_trim(k))

        iomsg=""
        read (strhex, '(Z8)', iostat=iostat, iomsg=iomsg) str2hex

        if(present(status)) then
            call FError(function, iostat, iomsg, status)
        end if

    end function

    ! convert a string into a boolean
    logical function str2bool(k)
      implicit none

      character(*), intent(in) :: k
      character(*), parameter :: function = "str2bool"
      character(buffsize) :: lcase_k!,iomsg

      str2bool = .false. ! just to avoid -Wmaybe-uninitialized warning.

      lcase_k = k
      call str_lower(lcase_k)

      if(trim(lcase_k)=="t" .or. trim(lcase_k)=="true" .or. trim(lcase_k)==".true."&
      .or. trim(lcase_k)=="y" .or. trim(lcase_k)=="yes" .or. trim(lcase_k)=="1") then
        str2bool = .true.

      else if(trim(lcase_k)=="f" .or. trim(lcase_k)=="false" .or. trim(lcase_k)==".false."&
      .or. trim(lcase_k)=="n" .or. trim(lcase_k)=="no" .or. trim(lcase_k)=="0") then
        str2bool = .false.

      else
call wrstop(function, "Cannot interpret supposed boolean string "//trim(k))
      end if

    end function

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !>  compare two strings, performing adjustl before and optional compare the capitalized letters
    !
    !
    !
    !> @param[out] same
    !> @param[in] str_1, str_2
    !> @param[inout] status,
    !---------------------------------------------------------------------------

    subroutine compare_str (str_1, str_2, same, status, upper)
        implicit none

        integer, intent(inout) :: status

        character(*), parameter :: function = "compare_str"
        character(*), intent(in) :: str_1, str_2
        character(len(str_1)) :: str_1_adjustl
        character(len(str_2)) :: str_2_adjustl
        integer, intent(out) :: same
        integer, intent(in), optional :: upper

        if(status > 100 .or. status == s_errRep) return

        if(len(str_1) /= len(str_2)) then
            same = false
            return
        end if

        str_1_adjustl = adjustl(str_1)
        str_2_adjustl = adjustl(str_2)

        if(present(upper)) then
            if(upper == true) then
                call str_upper(str_1_adjustl)
                call str_upper(str_2_adjustl)
            end if
        end if

        if(str_1_adjustl /= str_2_adjustl) then
            same = false
        else
            same = true
        end if

    end subroutine compare_str

  !.....................................................................
  subroutine str2deg(str, deg, status, convert)
    ! Accepted formats for 'str' (\d denotes an integer, \s means white space, a following * means 'to any number', a following + means 'one or more', square brackets denote optional, a comma-separated list in curly brackets means 'or'):
    !
    !   {+,-,}\d+[.\d+] (e.g. 4, -90, 270, +104.23)
    !
    !	\d\d:\d\d:\d\d[.\d+] (Note that if convert is present and .true., the first field is treated as degrees; as hours of right ascension if absent or .false..) (E.g. 41:10:21.4)
    !
    !	{+,-,}\d+{h,d}[\s*\d+{m,'}\s*[\d+[.\d+]{s,"}]] (e.g. 20h 12m 10.9s, -15d5'13")
    !
    implicit none

    character(*), intent(in)           :: str
    real(dp),     intent(out)          :: deg
    integer,      intent(inout)        :: status
    logical,      intent(in), optional :: convert

    character(*), parameter :: functionName = "str2deg"
    character (buffsize), allocatable, dimension(:) :: tok
    character (buffsize) :: field_1,field_2,field_3,remainder
    real(dp) :: value_1,value_2,value_3
    logical :: treat_first_field_as_h,is_hours_RA
    integer :: pos,ii

    if(len_trim(str)>buffsize)&
call wrstop(functionName, "String '"//trim(str)//"' is too long.")

    if(present(convert)) then ! valid only for xx:xx:xx format.
      treat_first_field_as_h = convert
    else
      treat_first_field_as_h = .false.
    end if

    pos = verify (trim(str), " +-01234556789.:""'dhms")
    if (pos /= 0)&
call wrstop(functionName, "String '"//trim(str)&
  //"' contains the invalid character '"//str(pos:pos)//"'.")

    ! if pure number it is by default degrees
    pos = verify(trim(str), " +-0123456789.")
    if(pos == 0) then
      deg = str2double(str, status)
  return
    end if

    if(index(trim(str), ":")==0) then
      ! String contains no ':' characters. See if it contains 'd' or 'h':
      !
      is_hours_RA = .false. ! default
      pos = index(trim(str), "h")
      if(pos>0) then
        is_hours_RA = .true.
      else
        pos = index(trim(str), "d")
      end if

      if(pos==0)&
call wrstop(functionName, "String '"//trim(str)//"' contains neither 'd' nor 'h'.")

      field_1 = str(1:pos-1)
      remainder = str(pos+1:)
      ii = verify(trim(field_1), " +-0123456789")
      if(ii/=0)&
call wrstop(functionName, "First field '"//trim(field_1)&
  //"' contains the invalid character '"//field_1(ii:ii)//"'.")

      value_1 = str2double(field_1)

      pos = scan(trim(remainder), "m'")
      if(pos==0) then
        ! Ignore the remainder. We could check that it contains only spaces...
        value_2 = 0.0d0
        value_3 = 0.0d0

      else
        field_2 = remainder(1:pos-1)
        remainder = remainder(pos+1:)
        ii = verify(trim(field_2), " 0123456789")
        if(ii/=0)&
call wrstop(functionName, "Second field '"//trim(field_2)&
  //"' contains the invalid character '"//field_2(ii:ii)//"'.")

        value_2 = str2double(field_2)

        pos = scan(trim(remainder), "s''")!***********************
        if(pos==0) then
          ! Ignore the remainder. We could check that it contains only spaces...
          value_3 = 0.0d0

        else
          field_3 = remainder(1:pos-1)
          remainder = remainder(pos+1:)
          ii = verify(trim(field_3), " 0123456789.")
          if(ii/=0)&
call wrstop(functionName, "Third field '"//trim(field_3)&
  //"' contains the invalid character '"//field_3(ii:ii)//"'.")

          value_3 = str2double(field_3)
        end if
      end if

      deg = value_1 + (sign(value_2, value_1) + sign(value_3, value_1)/60.0_dp)/60.0_dp
      if(is_hours_RA) deg = 15.0_dp*deg

    else
      ! String contains ':' characters. It must contain 2 of them, i.e. 3 fields separated by ':'.
      !
      call strtok(trim(str), ":", tok)
      if(size(tok) /= 3)&
call wrstop(functionName, "String '"//trim(str)//"' should contain 3 fields separated by ':' characters.")

      value_1 = str2double(tok(1))
      value_2 = str2double(tok(2))
      value_3 = str2double(tok(3))

      deg = value_1 + (sign(value_2, value_1) + sign(value_3, value_1)/60.0_dp)/60.0_dp
      if(treat_first_field_as_h) deg = 15.0_dp*deg
    end if

  end subroutine str2deg

end module erolib_str_mod

