! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! This implements a sorting algorithm of the 'timsort' type (see https://svn.python.org/projects/python/trunk/Objects/listsort.txt), invented by Tim Peters. It is basically a merge sort that makes use of the fact that the input list may already be partially sorted. Note that it is a 'stable' sort, in that elements of the sortable object that have the same value of the sorting property (but which may be distinguished by values of other, hidden, non-sorted properties) maintain their initial ordering.
!
! The following example lines indicate how timsort should be used.
!
!	use erolib_timsort
!
!	...
!
!	type(t_sort_scalar) :: sortable ! or any other subclass of t_sortable
!	call sortable%init(<values to be sorted>)
!	call timsort(sortable)
!	call sortable%check() ! if thought necessary
!	! do something with sortable%indices
!	call sortable%destroy()
!
module erolib_timsort
  use param
  use erolib_general_mod

  implicit none

logical, parameter, private :: doTest=.false.

  private

  type, public, abstract :: t_sortable
    integer, dimension(:), allocatable :: indices
  contains
    procedure(get_n_elements_t_sortable), deferred :: get_n_elements
    procedure(compare_t_sortable), deferred :: compare
    procedure(destroy_t_sortable), deferred :: destroy
    procedure :: is_sorted => is_sorted_t_sortable
  end type t_sortable

  abstract interface
    integer function get_n_elements_t_sortable(this) result(n_elements)
      import t_sortable
      implicit none
      class(t_sortable) :: this
    end function
  end interface

  abstract interface
    integer function compare_t_sortable(this, i1, i2) result(comp)
      ! If sortable_obj(i1) > sortable_obj(i2), function should return COMP_GT, etc.
      import t_sortable
      implicit none
      class(t_sortable)   :: this
      integer, intent(in) :: i1, i2
    end function
  end interface

  abstract interface
    subroutine destroy_t_sortable(this)
      import t_sortable
      implicit none
      class(t_sortable) :: this
    end subroutine
  end interface

  type, extends(t_sortable) :: t_sort_scalar
!*** better to write this with a generic array? Might have to make it a pointer... to an array of scalars???
    private
    real(r4), dimension(:), allocatable :: r4_values
    real(r8), dimension(:), allocatable :: r8_values
    integer(i1), dimension(:), allocatable :: i1_values
    integer(i2), dimension(:), allocatable :: i2_values
    integer(i4), dimension(:), allocatable :: i4_values
    integer(i8), dimension(:), allocatable :: i8_values
    character(1) :: dataTypeFITS = "" ! from the set ['E','D','B','I','J','K'].
  contains
    procedure :: init => init_t_sort_scalar
    procedure :: get_n_elements => get_n_elements_t_sort_scalar
    procedure :: compare => compare_t_sort_scalar
    procedure :: destroy => destroy_t_sort_scalar
  end type t_sort_scalar

  ! The nomenclature that I will use is that *ei signifies the location
  ! in this%indices, whereas *ii signifies the value this%indices(*ei).
  ! Just *i serves only to remind that the quantity is integer. Note that
  ! *ii = *ei right after t_runstack%init().

  type :: t_run
    integer :: start_ei=0
    integer :: length_i=0
  end type t_run

  integer, parameter :: min_gallop_i = 7

  type :: t_runstack
    type(t_run), dimension(:), allocatable :: runs
    class(t_sortable), pointer :: sortable_obj=>null()
    integer :: n_elements = 0
    integer :: minrun_size = 0
    integer :: stack_size = 0
    integer, dimension(:), allocatable :: scratch
    integer :: ei = 1 ! next element after the last run on the stack.
    integer :: si = 0 ! last run on the stack.
    integer :: gallop_mult = 1
  contains
    procedure :: calc_minrun => calc_minrun_t_runstack
    procedure :: calc_stack_size => calc_stack_size_t_runstack
    procedure :: init => init_t_runstack
    procedure :: finished => finished_t_runstack
    procedure :: append_run => append_run_t_runstack
    procedure :: invert => invert_t_runstack
    procedure :: bsearch_A_pinching_B => bsearch_A_pinching_B_t_runstack
    procedure :: bsearch_B_pinching_A => bsearch_B_pinching_A_t_runstack
    procedure :: insert => insert_t_runstack
    procedure :: get_next_run => get_next_run_t_runstack
    procedure :: calc_len_run_A_in_B => calc_len_run_A_in_B_t_runstack
    procedure :: calc_len_run_B_in_A => calc_len_run_B_in_A_t_runstack
    procedure :: merge_core => merge_core_t_runstack
    procedure :: merge_2_runs => merge_2_runs_t_runstack
    procedure :: conditional_merges => conditional_merges_t_runstack
    procedure :: forced_merges => forced_merges_t_runstack
    procedure :: destroy => destroy_t_runstack
  end type t_runstack

  enum, bind( C )
    enumerator :: COMP_LT=1, COMP_EQ, COMP_GT
  end enum


  type, extends(t_sortable) :: t_sort_test
    real, dimension(:), allocatable :: values
    real, dimension(:), allocatable :: values_temp
    type(t_run), dimension(:), allocatable :: runs
  contains
    procedure :: init => init_t_sort_test
    procedure :: get_n_elements => get_n_elements_t_sort_test
    procedure :: compare => compare_t_sort_test
    procedure :: destroy => destroy_t_sort_test
  end type t_sort_test

  interface median
    module procedure median_r4, median_r8
  end interface

  public :: timsort,COMP_LT,COMP_EQ,COMP_GT,test_timsort,median

contains

  !=====================================================================
  ! Methods of t_sortable:

  !.....................................................................
  logical function is_sorted_t_sortable(this, bad_row_i) result(is_sorted)
    implicit none
    class(t_sortable)              :: this
    integer, intent(out), optional :: bad_row_i

    integer :: n_elements,ii

    if(present(bad_row_i)) bad_row_i = 0 ! just so as to not leave it dangling.

    is_sorted = .true. ! default
    n_elements = this%get_n_elements()
    if(n_elements < 2)&
  return

    do ii=2,n_elements
      if(this%compare(this%indices(ii-1),this%indices(ii)) == COMP_GT) then ! should be <=.
        is_sorted = .false.
        if(present(bad_row_i)) bad_row_i = ii
    exit
      end if
    end do

  end function is_sorted_t_sortable

  !=====================================================================
  ! Methods of t_sort_scalar:

  !.....................................................................
  subroutine init_t_sort_scalar(this, values)
    implicit none
    class(t_sort_scalar)               :: this
    class(*), dimension(:), intent(in) :: values

    character(*), parameter :: functionName = "init_t_sort_scalar"

    select type(values)
    type is ( real(r4) )
      this%dataTypeFITS = "E"
      this%r4_values = values
    type is ( real(r8) )
      this%dataTypeFITS = "D"
      this%r8_values = values
    type is ( integer(i1) )
      this%dataTypeFITS = "B"
      this%i1_values = values
    type is ( integer(i2) )
      this%dataTypeFITS = "I"
      this%i2_values = values
    type is ( integer(i4) )
      this%dataTypeFITS = "J"
      this%i4_values = values
    type is ( integer(i8) )
      this%dataTypeFITS = "K"
      this%i8_values = values
    class default
call wrstop(functionName, "Sorting scalars of the given array type is not supported.")
    end select

  end subroutine init_t_sort_scalar

  !.....................................................................
  integer function get_n_elements_t_sort_scalar(this) result(n_elements)
    implicit none
    class(t_sort_scalar) :: this

    character(*), parameter :: functionName = "get_n_elements_t_sort_scalar"

    select case(this%dataTypeFITS)
    case("E")
      n_elements = size(this%r4_values)
    case("D")
      n_elements = size(this%r8_values)
    case("B")
      n_elements = size(this%i1_values)
    case("I")
      n_elements = size(this%i2_values)
    case("J")
      n_elements = size(this%i4_values)
    case("K")
      n_elements = size(this%i8_values)
    case default
call wrstop(functionName, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
    end select

  end function get_n_elements_t_sort_scalar

  !.....................................................................
  integer function compare_t_sort_scalar(this, i1, i2) result(comp)
    ! If sortable_obj(i1) > sortable_obj(i2), function should return COMP_GT, etc.
    implicit none
    class(t_sort_scalar)   :: this
    integer, intent(in) :: i1, i2

    character(*), parameter :: functionName = "compare_t_sort_scalar"

    select case(this%dataTypeFITS)
    case("E")
      if(this%r4_values(i1) > this%r4_values(i2)) then
        comp = COMP_GT
      else if(this%r4_values(i1) < this%r4_values(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("D")
      if(this%r8_values(i1) > this%r8_values(i2)) then
        comp = COMP_GT
      else if(this%r8_values(i1) < this%r8_values(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("B")
      if(this%i1_values(i1) > this%i1_values(i2)) then
        comp = COMP_GT
      else if(this%i1_values(i1) < this%i1_values(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("I")
      if(this%i2_values(i1) > this%i2_values(i2)) then
        comp = COMP_GT
      else if(this%i2_values(i1) < this%i2_values(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("J")
      if(this%i4_values(i1) > this%i4_values(i2)) then
        comp = COMP_GT
      else if(this%i4_values(i1) < this%i4_values(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("K")
      if(this%i8_values(i1) > this%i8_values(i2)) then
        comp = COMP_GT
      else if(this%i8_values(i1) < this%i8_values(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case default
call wrstop(functionName, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
    end select

  end function compare_t_sort_scalar

  !.....................................................................
  subroutine destroy_t_sort_scalar(this)
    implicit none
    class(t_sort_scalar) :: this

    character(*), parameter :: functionName = "destroy_t_sort_scalar"
    integer :: ierr

    ierr = 0

    select case(this%dataTypeFITS)
    case("E")
      call dealloc(this%r4_values, ierr)
    case("D")
      call dealloc(this%r8_values, ierr)
    case("B")
      call dealloc(this%i1_values, ierr)
    case("I")
      call dealloc(this%i2_values, ierr)
    case("J")
      call dealloc(this%i4_values, ierr)
    case("K")
      call dealloc(this%i8_values, ierr)
    case default
call wrstop(functionName, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
    end select

    this%dataTypeFITS = ""

  end subroutine destroy_t_sort_scalar

  !=====================================================================
  ! Methods of t_runstack:

  !.....................................................................
  subroutine calc_minrun_t_runstack(this)
    ! We pick a minrun_size in range(32, 65) such that
    ! n_elements/minrun_size is exactly a power of 2; or if that isn't
    ! possible, is close to, but strictly less than, a power of 2. This
    ! is easier to do than it may sound: take the first 6 bits of
    ! n_elements, and add 1 if any of the remaining bits are set.
    !
    implicit none
    class(t_runstack) :: this

    character(*), parameter :: functionName="calc_minrun_t_runstack"
    integer, parameter :: n_bits_in_MRS = 6
    integer :: pos_1st_set_bit,frompos

    pos_1st_set_bit = bit_size(this%n_elements) - leadz(this%n_elements)
    ! This counts from the LSB, counting that bit as 1. Example:
    !
    !	n_elements = 2112 dec
    !	           = 0b 0000 1000 0100 0000 (supposing the type is i2,
    !                                        for brevity)
    !
    ! For this, bit_size() should return 16, and leadz() returns 4. Thus
    ! the position of the MS set bit, when counted from the LSB, is 12.

    if(pos_1st_set_bit < n_bits_in_MRS) then
      this%minrun_size = this%n_elements ! There will just be 1 run which
      ! will be sorted by insertion, no merging necessary.
  return
    end if

    frompos = pos_1st_set_bit - n_bits_in_MRS
    this%minrun_size = 0
    call mvbits(this%n_elements, frompos, n_bits_in_MRS, this%minrun_size, 0)

    if(modulo(this%n_elements, this%minrun_size)/=0)&
      this%minrun_size = this%minrun_size + 1

  end subroutine calc_minrun_t_runstack

  !.....................................................................
  subroutine calc_stack_size_t_runstack(this)
    ! The stack can never grow larger than about log_base_phi(N)
    ! entries, where phi = (1+sqrt(5))/2 ~= 1.618.
    !
    implicit none
    class(t_runstack) :: this

    character(*), parameter :: functionName="calc_stack_size_t_runstack"
    real :: phi

    phi = 0.5*(1.0 + sqrt(5.0))
    this%stack_size = 2*int(log(real(this%n_elements))/log(phi))
    ! The extra factor of 2 is just for some elbow room. I don't
    ! anticipate that memory will be a bottleneck in sass applications.

  end subroutine calc_stack_size_t_runstack

  !.....................................................................
  subroutine init_t_runstack(this, sortable_obj)
    implicit none
    class(t_runstack)         :: this
    class(t_sortable), target :: sortable_obj

    character(*), parameter :: functionName="init_t_runstack"
    integer :: ii

    this%sortable_obj => sortable_obj
    ! We have to do it this way because t_sortable is an abstract class.

    this%n_elements = sortable_obj%get_n_elements()
    this%sortable_obj%indices = [(ii,ii=1,this%n_elements,1)]
    call this%calc_minrun()
    call this%calc_stack_size()
    allocate(this%runs(this%stack_size))!*** error stuff
    allocate(this%scratch(this%n_elements))!*** error stuff
    this%scratch(:) = 0

    this%ei = 1 ! next element after the last run on the stack.
    this%si = 0 ! last run on the stack.

  end subroutine init_t_runstack

  !.....................................................................
  logical function finished_t_runstack(this)
    implicit none
    class(t_runstack) :: this

    finished_t_runstack = .false. ! default
    if(this%ei > this%n_elements) finished_t_runstack = .true.

  end function finished_t_runstack

  !.....................................................................
  subroutine append_run_t_runstack(this, start_ei, length_i)
    implicit none
    class(t_runstack) :: this
    integer, intent(in) :: start_ei,length_i

    character(*), parameter :: functionName="append_run_t_runstack"

    this%si = this%si + 1
    if(this%si > this%stack_size)&
call wrstop(functionName, "Not enough room on stack.")
!**** eventually check stack size and allocate it larger if necessary before appending the run.

    this%runs(this%si)%start_ei = start_ei
    this%runs(this%si)%length_i = length_i

  end subroutine append_run_t_runstack

  !.....................................................................
  subroutine invert_t_runstack(this, start_ei, finis_ei)
    ! Inverts the order of the sequence.
    implicit none
    class(t_runstack) :: this
    integer, intent(in) :: start_ei,finis_ei

    character(*), parameter :: functionName="invert_t_runstack"
    integer :: len_i,lower_len_i,lower_finis_ei,upper_start_ei

    len_i = 1 + finis_ei - start_ei
    if(mod(len_i,2)==0) then ! even number of elements in the sequence to be inverted.
      lower_len_i = len_i/2
    else ! odd number of elements
      lower_len_i = (len_i-1)/2
    end if

    lower_finis_ei = start_ei + lower_len_i - 1
    upper_start_ei = finis_ei - lower_len_i + 1
    this%scratch(1:lower_len_i) = this%sortable_obj%indices(finis_ei:upper_start_ei:-1)
    this%sortable_obj%indices(upper_start_ei:finis_ei)&
      = this%sortable_obj%indices(lower_finis_ei:start_ei:-1)
    this%sortable_obj%indices(start_ei:lower_finis_ei)&
      = this%scratch(1:lower_len_i)

    ! Examples:
    !	start_i = 10, finis_i = 15
    !	len_i = 6 (i.e. even)
    !	lower_len_i = 3
    !	lower_finis_i = 10 + 3 - 1 = 12
    !	upper_start_i = 15 - 3 + 1 = 13
    !	scratch(1:3) = this%indices(15:13:-1)
    !	this%indices(13:15) = this%indices(12:10:-1)
    !	this%indices(10:12) = scratch(1:3)
    !
    !	start_i = 10, finis_i = 16
    !	len_i = 7 (i.e. odd)
    !	lower_len_i = 3
    !	lower_finis_i = 10 + 3 - 1 = 12
    !	upper_start_i = 16 - 3 + 1 = 14
    !	scratch(1:3) = this%indices(16:14:-1)
    !	this%indices(14:16) = this%indices(12:10:-1)
    !	this%indices(10:12) = scratch(1:3)
    ! Here this%indices(13) is left unmoved.

  end subroutine invert_t_runstack

  !.....................................................................
  integer function bsearch_B_pinching_A_t_runstack(this, b_start_ei, b_finis_ei&
    , a_ii, take_lo_bound) result(b_ei)

    ! You should only call this if
    !	B(1) < A(a_ii) <= B(N).
    ! The routine applies a binary search to find j such that
    !	B(j-1) < A(a_ii) <= B(j).
    !
    implicit none
    class(t_runstack)   :: this
    integer, intent(in) :: b_start_ei, b_finis_ei, a_ii
    logical, intent(in) :: take_lo_bound

    character(*), parameter :: functionName="bsearch_B_pinching_A"
    integer :: lo_ei,hi_ei,len_i,half_len_i,mid_ei

    if((b_finis_ei - b_start_ei) < 1)&
call wrstop(functionName, "The sequence has fewer than 2 members.")

    lo_ei = b_start_ei
    hi_ei = b_finis_ei
    do
      len_i = 1 + hi_ei - lo_ei
      if(len_i < 3)&
    exit

      half_len_i = len_i/2 ! integer division, truncates.
      mid_ei = lo_ei + half_len_i

      if(this%sortable_obj%compare(this%sortable_obj%indices(mid_ei), a_ii)/=COMP_LT) then
        hi_ei = mid_ei
      else
        lo_ei = mid_ei
      end if
    end do

    if(take_lo_bound) then
      b_ei = lo_ei
    else
      b_ei = hi_ei
    end if

  end function bsearch_B_pinching_A_t_runstack

  !.....................................................................
  integer function bsearch_A_pinching_B_t_runstack(this, a_start_ei, a_finis_ei&
    , b_ii, take_lo_bound) result(a_ei)

    ! You should only call this if
    !	A(1) <= B(b_ii) < A(N).
    ! The routine applies a binary search to find j such that
    !	A(j-1) <= B(b_ii) < A(j).
    !
    implicit none
    class(t_runstack)   :: this
    integer, intent(in) :: a_start_ei, a_finis_ei, b_ii
    logical, intent(in) :: take_lo_bound

    character(*), parameter :: functionName="bsearch_A_pinching_B"
    integer :: lo_ei,hi_ei,len_i,half_len_i,mid_ei

    if((a_finis_ei - a_start_ei) < 1)&
call wrstop(functionName, "The sequence has fewer than 2 members.")

    lo_ei = a_start_ei
    hi_ei = a_finis_ei
    do
      len_i = 1 + hi_ei - lo_ei
      if(len_i < 3)&
    exit

      half_len_i = len_i/2 ! integer division, truncates.
      mid_ei = lo_ei + half_len_i

      if(this%sortable_obj%compare(this%sortable_obj%indices(mid_ei), b_ii)==COMP_GT) then
        hi_ei = mid_ei
      else
        lo_ei = mid_ei
      end if
    end do

    if(take_lo_bound) then
      a_ei = lo_ei
    else
      a_ei = hi_ei
    end if

  end function bsearch_A_pinching_B_t_runstack

  !.....................................................................
  subroutine insert_t_runstack(this, start_ei, test_ei, omit_lo_check&
    , omit_hi_check)
    ! Inserts element ii in its correct order in the sequence
    ! {start_i,ii-1}, which is assumed to be ascending
    ! (i.e., a0 <= a1 <= a2 <= ...). A binary search is used.
    implicit none
    class(t_runstack)             :: this
    integer, intent(in)           :: start_ei,test_ei
    logical, intent(in), optional :: omit_lo_check, omit_hi_check
    ! In fact these could have any data type, since their presence alone
    ! is used as the boolean flag.

    character(*), parameter :: functionName="insert_t_runstack"
    integer :: start_ii,finis_ii,test_ii!,lo_ei,len_i,mid_ei!,half_len_i,hi_ei
    integer :: ei_to_replace

    if((test_ei-start_ei) < 2)&
call wrstop(functionName, "The sequence has length < 2.")

    start_ii = this%sortable_obj%indices(start_ei)
    finis_ii = this%sortable_obj%indices(test_ei-1)
    test_ii = this%sortable_obj%indices(test_ei)

    if(.not.present(omit_lo_check)) then
      if(this%sortable_obj%compare(start_ii, test_ii)==COMP_GT) then
        ! Note that if sortable_obj(start_ii)==sortable_object(test_ii),
        ! in the interests of preserving a stable sort, we don't want
        ! to place the latter before the former.
        this%sortable_obj%indices(start_ei+1:test_ei)&
          = this%sortable_obj%indices(start_ei:test_ei-1) 
        this%sortable_obj%indices(start_ei) = test_ii
  return
      end if
    end if

    if(.not.present(omit_hi_check)) then
      if(this%sortable_obj%compare(test_ii, finis_ii)/=COMP_LT) then
        ! Leave the sequence alone, it is already entirely sorted.
  return
      end if
    end if

    ! If we get to here, we have a start index and a finish index, which
    ! define a range [lo,hi] of the sortable_obj value; if we also
    ! assume that the calling routine has made correct use of the 'omit'
    ! arguments, then the test element x=sortable_obj(indices(ii)) obeys
    ! lo <= x < hi.

    ei_to_replace = this%bsearch_A_pinching_B(start_ei, test_ei-1, test_ii, take_lo_bound=.false.)
    this%sortable_obj%indices(ei_to_replace+1:test_ei)&
      = this%sortable_obj%indices(ei_to_replace:test_ei-1)
    this%sortable_obj%indices(ei_to_replace) = test_ii

  end subroutine insert_t_runstack

  !.....................................................................
  subroutine get_next_run_t_runstack(this)
    implicit none
    class(t_runstack) :: this

    character(*), parameter :: functionName="get_next_run_t_runstack"
    logical :: initially_descending,do_insertion,next_pair_descend
    integer :: rei

    if(this%ei == this%n_elements) then
      call this%append_run(this%ei, 1)
      this%ei = this%n_elements + 1
      ! will signal finished when the eponymous function is next called.
  return
    end if

    ! If we get to here, this%ei should be < this%n_elements.
    !
    ! We proceed along the list from index ei, testing to see how far
    ! the elements are ordered, either in decreasing (a0 > a1 > a2 >...)
    ! or increasing (a0 <= a1 <= a2 <= ...) sequence. The moment
    ! elements occur out of the order set by the first two, we begin an
    ! insertion sort. The run terminates as soon as either its length
    ! equals minrun_size or we run out of list elements - whichever
    ! comes first.
    !
    ! The only tweak to this is that if the run starts in descending
    ! order, it is inverted to become ascending as soon as the order
    ! breaks, before entering the insertion sort. (Note that the strict
    ! definition of descending avoids changing the order of sort-equal
    ! entries through this inversion - because the strict definition
    ! ensures that there can be *no* successive sort-equal entries in
    ! the stretch that is inverted.)
    !
    ! Note that at this point, indices(ei) still == ei for the range of
    ! ei values in play.
    !
    if(this%sortable_obj%compare(this%ei, this%ei+1)==COMP_GT) then
      initially_descending = .true.
    else
      initially_descending = .false.
    end if

    do_insertion = .false.
    rei = this%ei + 1
    do
      rei = rei + 1
      if(rei > this%n_elements .or. (1 + rei - this%ei)>this%minrun_size) then
        call this%append_run(this%ei, rei - this%ei)
        this%ei = rei ! If rei>this%n_elements, setting this%ei = rei
        ! ensures that finished() will return .true. the next time it
        ! is called.
    exit
      end if

      if(do_insertion) then
        call this%insert(this%ei, rei)
      else
        next_pair_descend = (this%sortable_obj%compare(rei-1, rei)==COMP_GT)
        if(next_pair_descend .and. .not.initially_descending) then
          ! This pair breaks the ascending run.

          do_insertion = .true.
          call this%insert(this%ei, rei, omit_hi_check=.true.)

        else if(.not.next_pair_descend .and. initially_descending) then
          ! This pair breaks the descending run.

          call this%invert(this%ei, rei-1)
          do_insertion = .true.
          call this%insert(this%ei, rei, omit_lo_check=.true.)
        end if
      end if
    end do

  end subroutine get_next_run_t_runstack

  !.....................................................................
  subroutine calc_len_run_A_in_B_t_runstack(this, a_start_ei, a_finis_ei, b_ei&
    , delta_i, no_more_As)
    ! Returns the largest i such that A(i) <= B(b_ei) < A(i+1).

    implicit none
    class(t_runstack)    :: this
    integer, intent(in)  :: a_start_ei, a_finis_ei, b_ei
    integer, intent(out) :: delta_i
    logical, intent(out) :: no_more_As

    ! For brevity I substitute
    !
    !	a_start_ei -> 1
    !	a_finis_ei -> N
    !	b_ei -> j
    !
    ! On entry, we know that:
    !
    !	B(j-1) < A(1) <= B(j).
    !
    ! Either of two situations may pertain:
    !
    !	B(j-1) < A(1) ... A(i) <= B(j) < A(i+1)
    !
    ! where it could be that i==1, or
    !
    !	B(j-1) < A(1) ... A(N) <= B(j)
    !
    ! where equally, it could be that N==1. For the first case we return
    !
    !	delta_i = i
    !	no_more_As = .false.
    !
    ! For the second case we return
    !
    !	delta_i = N
    !	no_more_As = .true.

    character(*), parameter :: functionName="calc_len_run_A_in_B"
    integer :: a_ei,b_ii,a_ii,a_trial_ei,a_last_ei,ii

    no_more_As = .false. ! default
    a_ei = a_start_ei
    b_ii = this%sortable_obj%indices(b_ei)

    L_OUTER: do ! so long as A(a_ii)<=B(b_ii).
      ! Begin by incrementing by 1:
      !
      L_WALK: do ii = 1, min_gallop_i
        a_ei = a_ei + 1
        if(a_ei > a_finis_ei)&
      exit L_WALK

        a_ii = this%sortable_obj%indices(a_ei)
        if(this%sortable_obj%compare(a_ii, b_ii)==COMP_GT) then
          delta_i = a_ei - a_start_ei
    exit L_OUTER
        end if
      end do L_WALK

      if(a_ei >= a_finis_ei) then
        no_more_As = .true.
        delta_i = 1 + a_finis_ei - a_start_ei
    exit L_OUTER
      end if

      ! If we have got to here, then it is guaranteed that A(a_ii) <= B(b_ii).

      ! Now we commence an exponential search:
      !
      L_GALLOP: do
        this%gallop_mult = 2*this%gallop_mult
        a_trial_ei = a_ei + this%gallop_mult

        if(a_trial_ei >= a_finis_ei) then
          a_trial_ei = a_finis_ei
          no_more_As = .true.
        end if

        a_ii = this%sortable_obj%indices(a_trial_ei)
        if(this%sortable_obj%compare(a_ii, b_ii)==COMP_GT) then
          ! We have A(a_ii) <= B(b_ii) < A(trial_ii).
          a_last_ei = this%bsearch_A_pinching_B(a_ei, a_trial_ei, b_ii, take_lo_bound=.true.)
          if(a_last_ei < a_finis_ei) no_more_As = .false.
          delta_i = 1 + a_last_ei - a_start_ei
          this%gallop_mult = 1
    exit L_OUTER
        end if

        if(no_more_As) then
          delta_i = 1 + a_finis_ei - a_start_ei
          this%gallop_mult = 1
    exit L_OUTER
        end if

        a_ei = a_trial_ei
      end do L_GALLOP
    end do L_OUTER

  end subroutine calc_len_run_A_in_B_t_runstack

  !.....................................................................
  subroutine calc_len_run_B_in_A_t_runstack(this, a_ei, b_start_ei, b_finis_ei&
    , delta_i, no_more_Bs)
    ! Returns the largest i such that B(i) < A(a_ii) <= B(i+1).

    implicit none
    class(t_runstack)    :: this
    integer, intent(in)  :: a_ei,b_start_ei,b_finis_ei
    integer, intent(out) :: delta_i
    logical, intent(out) :: no_more_Bs

    ! For brevity I substitute
    !
    !	a_ei -> j
    !	b_start_ei -> 1
    !	b_finis_ei -> N
    !
    ! On entry, we know that:
    !
    !	A(j-1) <= B(1) < A(j).
    !
    ! Either of two situations may pertain:
    !
    !	A(j-1) <= B(1) ... B(i) < A(j) <= B(i+1)
    !
    ! where it could be that i==1, or
    !
    !	A(j-1) <= B(1) ... B(N) < A(j)
    !
    ! where equally, it could be that N==1. For the first case we return
    !
    !	delta_i = i
    !	no_more_Bs = .false.
    !
    ! For the second case we return
    !
    !	delta_i = N
    !	no_more_Bs = .true.

    character(*), parameter :: functionName="calc_len_run_B_in_A"
    integer :: b_ei,a_ii,b_ii,b_trial_ei,b_last_ei,ii

    no_more_Bs = .false. ! default
    b_ei = b_start_ei
    a_ii = this%sortable_obj%indices(a_ei)

    L_OUTER: do ! so long as B(b_ii)<A(a_ii).
      ! Begin by incrementing by 1:
      !
      L_WALK: do ii = 1, min_gallop_i
        b_ei = b_ei + 1
        if(b_ei > b_finis_ei)&
      exit L_WALK

        b_ii = this%sortable_obj%indices(b_ei)
        if(this%sortable_obj%compare(b_ii, a_ii)/=COMP_LT) then
          delta_i = b_ei - b_start_ei
    exit L_OUTER
        end if
      end do L_WALK

      if(b_ei >= b_finis_ei) then
        no_more_Bs = .true.
        delta_i = 1 + b_finis_ei - b_start_ei
    exit L_OUTER
      end if

      ! If we have got to here, then it is guaranteed that B(b_ii) < A(a_ii).

      ! Now we commence an exponential search:
      !
      L_GALLOP: do
        this%gallop_mult = 2*this%gallop_mult
        b_trial_ei = b_ei + this%gallop_mult

        if(b_trial_ei >= b_finis_ei) then
          b_trial_ei = b_finis_ei
          no_more_Bs = .true.
        end if

        b_ii = this%sortable_obj%indices(b_trial_ei)
        if(this%sortable_obj%compare(b_ii, a_ii)/=COMP_LT) then
          ! We have B(b_ii) < A(a_ii) <= B(trial_ii).
          b_last_ei = this%bsearch_B_pinching_A(b_ei, b_trial_ei, a_ii, take_lo_bound=.true.)
          if(b_last_ei < b_finis_ei) no_more_Bs = .false.
          delta_i = 1 + b_last_ei - b_start_ei
          this%gallop_mult = 1
    exit L_OUTER
        end if

        if(no_more_Bs) then
          delta_i = 1 + b_finis_ei - b_start_ei
          this%gallop_mult = 1
    exit L_OUTER
        end if

        b_ei = b_trial_ei
      end do L_GALLOP
    end do L_OUTER

  end subroutine calc_len_run_B_in_A_t_runstack

  !.....................................................................
  subroutine merge_core_t_runstack(this, a_start_ei, a_finis_ei, b_start_ei&
    , b_finis_ei, scr_start_i, scr_finis_i, start_B_in_A)
    ! This merges two sequences from this%indices: a late-occurring
    ! sequence B with an earlier-occurring sequence A. ('Early' and
    ! 'late' refer to their respective ranges of ei values.) The two are
    ! merged onto this%scratch, commencing from the given position.
    !
    implicit none
    class(t_runstack)   :: this
    integer, intent(in) :: a_start_ei,a_finis_ei
    integer, intent(in) :: b_start_ei,b_finis_ei
    integer, intent(in) :: scr_start_i,scr_finis_i
    logical, intent(in) :: start_B_in_A

    character(*), parameter :: functionName="merge_core_t_runstack"
    logical :: count_B_in_A,no_more_As,no_more_Bs
    integer :: scr_i,a_ei,b_ei,a_ei_1,b_ei_1,scr_i_1,delta_i

    count_B_in_A = start_B_in_A
    ! If start_B_in_A, it means we know cast-iron that
    !	A(a_start_ei-1) <= B(b_start_ei) < A(a_start_ei).
    ! Otherwise, that
    !	B(b_start_ei-1) < A(a_start_ei) <= B(b_start_ei).

    scr_i = scr_start_i
    a_ei = a_start_ei
    b_ei = b_start_ei

    do
      if(count_B_in_A) then
        ! Returns the largest i such that B(i) < A(a_ii) <= B(i+1).
        call this%calc_len_run_B_in_A(a_ei, b_ei, b_finis_ei, delta_i, no_more_Bs)

        b_ei_1  = b_ei  + delta_i - 1
        scr_i_1 = scr_i + delta_i - 1
        this%scratch(scr_i:scr_i_1) = this%sortable_obj%indices(b_ei:b_ei_1)

        if(no_more_Bs) then
          if(a_ei <= a_finis_ei)& ! Tack on the rest of the As.
            this%scratch(scr_i_1+1:scr_finis_i) = this%sortable_obj%indices(a_ei:a_finis_ei)

    exit
        end if

        b_ei = b_ei_1 + 1
        count_B_in_A = .not.count_B_in_A ! toggles it

      else ! count_A_in_B.
        !         ! Returns the largest i such that A(i) <= B(b_ei) < A(i+1).
        call this%calc_len_run_A_in_B(a_ei, a_finis_ei, b_ei, delta_i, no_more_As)

        a_ei_1  = a_ei  + delta_i - 1
        scr_i_1 = scr_i + delta_i - 1
        this%scratch(scr_i:scr_i_1) = this%sortable_obj%indices(a_ei:a_ei_1)

        if(no_more_As) then
          if(b_ei <= b_finis_ei)& ! Tack on the rest of the Bs.
            this%scratch(scr_i_1+1:scr_finis_i) = this%sortable_obj%indices(b_ei:b_finis_ei)

    exit
        end if

        a_ei = a_ei_1 + 1
        count_B_in_A = .not.count_B_in_A ! toggles it
      end if

      scr_i = scr_i_1 + 1
    end do

  end subroutine merge_core_t_runstack

  !.....................................................................
  subroutine merge_2_runs_t_runstack(this, si)
    ! This merges run si with si+1. The calling routine needs to test
    ! that si+1 exists. Note that we can be assured that both runs are
    ! in ascending order; what we don't know is how they overlap. This
    ! has to be determined first.
    !
    ! We recognize 13 cases, which are diagrammed below by the position
    ! and size of B (i.e., the run that starts at
    ! this%runs(si+1)%start_ei) with respect to a fixed A (the run that
    ! starts at this%runs(si)%start_ei).
    !
    !	    A:        |-----| 
    !	 1  B:  |---|
    !
    !	 2  B:  |-----|
    !	 3  B:  |-------|
    !	 4  B:  |-----------|
    !	 5  B:  |-----------------|
    !	 6  B:        |---|
    !	 7  B:        |-----|
    !	 8  B:        |-----------|
    !	 9  B:          |-|
    !	10  B:          |---|
    !	11  B:          |---------|
    !
    !	12  B:              |-----|
    !	13  B:                |---|
    !
    implicit none
    class(t_runstack)   :: this
    integer, intent(in) :: si

    character(*), parameter :: functionName="merge_2_runs_t_runstack"
    integer :: a_start_ei,a_finis_ei,b_start_ei,b_finis_ei
    integer :: a_start_ii,a_finis_ii,b_start_ii,b_finis_ii
    integer :: len_i
    integer :: scratch_start_i1,scratch_finis_i1,scratch_start_i2,scratch_finis_i2
    integer :: ei_first_A_in_B,ei_last_A_in_B,ei_first_B_in_A,ei_last_B_in_A
    integer :: si1, si2

    si1 = si
    si2 = si + 1

    a_start_ei = this%runs(si1)%start_ei
    a_finis_ei = this%runs(si1)%start_ei + this%runs(si1)%length_i - 1
    b_start_ei = this%runs(si2)%start_ei
    b_finis_ei = this%runs(si2)%start_ei + this%runs(si2)%length_i - 1
    a_start_ii = this%sortable_obj%indices(a_start_ei)
    a_finis_ii = this%sortable_obj%indices(a_finis_ei)
    b_start_ii = this%sortable_obj%indices(b_start_ei)
    b_finis_ii = this%sortable_obj%indices(b_finis_ei)

    len_i = this%runs(si1)%length_i + this%runs(si2)%length_i ! just for brevity.

    if(this%sortable_obj%compare(a_finis_ii, b_start_ii)==COMP_GT) then
      ! excludes cases 12 and 13, which don't require any further action.
      !
      if(this%sortable_obj%compare(b_finis_ii, a_start_ii)==COMP_LT) then
        ! Case 1 - there is zero overlap, but the order of the two runs is inverted.
        !
        this%scratch(:this%runs(si2)%length_i) = this%sortable_obj%indices(b_start_ei:b_finis_ei)
        this%scratch(this%runs(si2)%length_i+1:len_i) = this%sortable_obj%indices(a_start_ei:a_finis_ei)
        this%sortable_obj%indices(a_start_ei:b_finis_ei) = this%scratch(1:len_i)
        !*** One could do this with fewer copies, e.g. through making use of the descending size of runs.

      else
        ! This leaves cases 2-11, all of which involve some overlap.
        !
        if(this%sortable_obj%compare(b_start_ii, a_start_ii)==COMP_LT) then
          ! Cases 2-5. Look for the element of B that A(1) should replace.
          !
          ei_first_A_in_B = this%bsearch_B_pinching_A(b_start_ei, b_finis_ei, a_start_ii, take_lo_bound=.false.)
          if(this%sortable_obj%compare(b_finis_ii, a_finis_ii)==COMP_LT) then
            ! Cases 2 & 3. Look for the element of A that B(-1) should replace.
            !
            !	    A:        |=====| 
            !	 2  B:  |-----|
            !	 3  B:  |-------|
            !
            ei_last_B_in_A = this%bsearch_A_pinching_B(a_start_ei, a_finis_ei&
              , b_finis_ii, take_lo_bound=.false.)
            !
            !	                     lbia
            !	                     v
            !	         a0 .... aj, aj+1 ... aN
            !	b0 ... bi, bi+1 ... bN
            !	           ^
            !	           faib
            !
            scratch_start_i1 = 1
            scratch_finis_i1 = ei_first_A_in_B - b_start_ei
            scratch_start_i2 = len_i - (a_finis_ei - ei_last_B_in_A)
            scratch_finis_i2 = len_i
            this%scratch(scratch_start_i1:scratch_finis_i1)&
              = this%sortable_obj%indices(b_start_ei:ei_first_A_in_B-1)
            this%scratch(scratch_start_i2:scratch_finis_i2)&
              = this%sortable_obj%indices(ei_last_B_in_A:a_finis_ei)

            ! We know cast-iron that
            !	B(ei_first_A_in_B-1) < A(a_start_ei) <= B(ei_first_A_in_B)
            ! and that
            !	A(ei_last_B_in_A-1) <= B(b_finis_ei) < A(ei_last_B_in_A).

            call this%merge_core(a_start_ei, ei_last_B_in_A-1, ei_first_A_in_B, b_finis_ei&
              , scratch_finis_i1+1, scratch_start_i2-1, start_B_in_A=.false.)

            this%sortable_obj%indices(a_start_ei:b_finis_ei)&
              = this%scratch(scratch_start_i1:scratch_finis_i2)

          else
            ! Cases 4 & 5. Look for the element of B that A(-1) should replace.
            !
            !	    A:        |=====| 
            !	 4  B:  |-----------|
            !	 5  B:  |-----------------|
            !
            ei_last_A_in_B = this%bsearch_B_pinching_A(ei_first_A_in_B-1&
              , b_finis_ei, a_finis_ii, take_lo_bound=.false.)
            !
            !	         a0 ......... aN
            !	b0 ... bi, bi+1 ... bj, bj+1 ... bN
            !	           ^            ^
            !	        faib            laib
            !
            scratch_start_i1 = 1
            scratch_finis_i1 = ei_first_A_in_B - b_start_ei
            scratch_start_i2 = len_i - (b_finis_ei - ei_last_A_in_B)
            this%scratch(scratch_start_i1:scratch_finis_i1)&
              = this%sortable_obj%indices(b_start_ei:ei_first_A_in_B-1)

            ! We know cast-iron that
            !	B(ei_first_A_in_B-1) < A(a_start_ei) <= B(ei_first_A_in_B)
            ! and that
            !	B(ei_last_A_in_B-1)  < A(a_finis_ei) <= B(ei_last_A_in_B).

            call this%merge_core(a_start_ei, a_finis_ei, ei_first_A_in_B, ei_last_A_in_B-1&
              , scratch_finis_i1+1, scratch_start_i2-1, start_B_in_A=.false.)

            this%sortable_obj%indices(a_start_ei:ei_last_A_in_B-1)&
              = this%scratch(scratch_start_i1:scratch_start_i2-1)
          end if

        else
          ! Cases 6-11. Look for the element of A that B(1) should replace.
          !
          ei_first_B_in_A = this%bsearch_A_pinching_B(a_start_ei, a_finis_ei, b_start_ii, take_lo_bound=.false.)
          if(this%sortable_obj%compare(b_finis_ii, a_finis_ii)==COMP_LT) then
            ! Cases 6 & 9. Look for the element of A that B(-1) should replace.
            !
            !	    A:        |=====| 
            !	 6  B:        |---|
            !	 9  B:          |-|
            !
            ei_last_B_in_A = this%bsearch_A_pinching_B(ei_first_B_in_A-1&
              , a_finis_ei, b_finis_ii, take_lo_bound=.false.)
            !
            !	        fbia            lbia
            !	           v            v
            !	a0 ... ai, ai+1 ... aj, aj+1 ... aN
            !	          b0 ..........bN
            !
            scratch_finis_i1 = ei_first_B_in_A - a_start_ei
            scratch_start_i2 = len_i - (a_finis_ei - ei_last_B_in_A)
            scratch_finis_i2 = len_i
            this%scratch(scratch_start_i2:scratch_finis_i2)&
              = this%sortable_obj%indices(ei_last_B_in_A:a_finis_ei)

            ! We know cast-iron that
            !	A(ei_first_B_in_A-1) <= B(b_start_ei) < A(ei_first_B_in_A)
            ! and that
            !	A(ei_last_B_in_A-1)  <= B(b_finis_ei) <= A(ei_last_B_in_A).

            call this%merge_core(ei_first_B_in_A, ei_last_B_in_A-1, b_start_ei, b_finis_ei&
              , scratch_finis_i1+1, scratch_start_i2-1, start_B_in_A=.true.)

            this%sortable_obj%indices(ei_first_B_in_A:b_finis_ei)&
              = this%scratch(scratch_finis_i1+1:scratch_finis_i2)

          else
            ! Cases 7, 8, 10, 11. Look for the element of B that A(-1) should replace.
            !
            !	    A:        |=====| 
            !	 7  B:        |-----|
            !	 8  B:        |-----------|
            !	10  B:          |---|
            !	11  B:          |---------|
            !
            ei_last_A_in_B = this%bsearch_B_pinching_A(b_start_ei, b_finis_ei&
              , a_finis_ii, take_lo_bound=.false.)
            !
            !	        fbia
            !	           v
            !	a0 ... ai, ai+1 ... aN
            !	          b0 ......bj, bj+1 ... bN
            !	                       ^
            !	                       laib
            scratch_finis_i1 = ei_first_B_in_A - a_start_ei
            scratch_start_i2 = len_i - (b_finis_ei - ei_last_A_in_B)

            ! We know cast-iron that
            !	A(ei_first_B_in_A-1) <= B(b_start_ei) < A(ei_first_B_in_A)
            ! and that
            !	B(ei_last_A_in_B-1)  < A(a_finis_ei) <= B(ei_last_A_in_B).

            call this%merge_core(ei_first_B_in_A, a_finis_ei, b_start_ei, ei_last_A_in_B-1&
              , scratch_finis_i1+1, scratch_start_i2-1, start_B_in_A=.true.)

            this%sortable_obj%indices(ei_first_B_in_A:ei_last_A_in_B-1)&
              = this%scratch(scratch_finis_i1+1:scratch_start_i2-1)
          end if
        end if
      end if
    end if

    this%runs(si1)%length_i = this%runs(si1)%length_i + this%runs(si2)%length_i
    this%si = si1

  end subroutine merge_2_runs_t_runstack

  !.....................................................................
  subroutine conditional_merges_t_runstack(this)
    implicit none
    class(t_runstack) :: this

    character(*), parameter :: functionName="conditional_merges_t_runstack"
    logical :: stack_is_clean

    stack_is_clean = .false. ! default
    do while(.not.stack_is_clean)
      ! The idea is to maintain the stack such that every sequence of
      ! runs {A,B,C} obeys two 'invariants':
      !
      ! 1.  len(A) > len(B) + len(C).
      ! 2.  len(B) > len(C).
      !
      ! Wherever len(A) <= len(B) + len(C), the shorter of A and C is
      ! merged with B (ties favor C, for the freshness-in-cache reason),
      ! and the new run replaces the A,B or B,C entries.
      !
      stack_is_clean = .true.

      if(this%si > 2) then
        if(this%runs(this%si-2)%length_i&
          <= (this%runs(this%si-1)%length_i + this%runs(this%si)%length_i)) then

          stack_is_clean = .false.

          if(this%runs(this%si-2)%length_i < this%runs(this%si)%length_i) then
            call this%merge_2_runs(this%si-2)
          else
            call this%merge_2_runs(this%si-1)
          end if
          ! Note that any call to merge_2_runs() decrements si.
        end if
      end if

      if(this%si > 1) then
        if(this%runs(this%si-1)%length_i <= this%runs(this%si)%length_i) then
          stack_is_clean = .false.

          if(this%si > 2) then
            if(this%runs(this%si-2)%length_i < this%runs(this%si)%length_i) then
              call this%merge_2_runs(this%si-2)
            else
              call this%merge_2_runs(this%si-1)
            end if
          else ! si must be == 2.
           call this%merge_2_runs(this%si-1)
          end if
        end if
      end if
    end do

  end subroutine conditional_merges_t_runstack

  !.....................................................................
  subroutine forced_merges_t_runstack(this)
    implicit none
    class(t_runstack) :: this

    do while(this%si > 1)
      call this%merge_2_runs(this%si-1)
    end do

  end subroutine forced_merges_t_runstack

  !.....................................................................
  subroutine destroy_t_runstack(this)
    implicit none
    class(t_runstack) :: this

!    call this%sortable_obj%destroy()
    nullify(this%sortable_obj)
    deallocate(this%runs) !*** add error stuff
    deallocate(this%scratch) !*** add error stuff

  end subroutine destroy_t_runstack

  !=====================================================================
  ! The actual timsort() routine, followed by some test stuff:

  !.....................................................................
  subroutine timsort(sortable_obj)
    implicit none
    class(t_sortable), target :: sortable_obj

    type(t_runstack) :: runstack

    call runstack%init(sortable_obj)

    do while(.not.runstack%finished())
      call runstack%get_next_run()
      call runstack%conditional_merges()
    end do
    call runstack%forced_merges()

    call runstack%destroy()

  end subroutine timsort


  !.....................................................................
  subroutine init_t_sort_test(this)
    use erolib_math_mod

    implicit none
    class(t_sort_test) :: this

    integer, parameter :: n_elements = 70

    allocate(this%values(n_elements))
    allocate(this%runs(14))

    ! Mostly I want to test the 13 merge cases. Sequences of sorted randoms will be constructed here for that purpose.
    !
    ! A:
    !
    this%values( 1) = .25680
    this%values( 2) = .30111
    this%values( 3) = .33907
    this%values( 4) = .40629
    this%values( 5) = .46551
    this%values( 6) = .49082
    this%values( 7) = .54927
    this%values( 8) = .58195
    this%values( 9) = .63332
    this%values(10) = .68881
    this%values(11) = .74793

    this%runs(1)%start_ei = 1
    this%runs(1)%length_i = 11

    !	    A:        |-----| 
    !	 1  B:  |---|
    !
    this%values(12) = .00494
    this%values(13) = .02980
    this%values(14) = .05501
    this%values(15) = .08779
    this%values(16) = .10125

    this%runs(2)%start_ei = 12
    this%runs(2)%length_i = 5

! Final indices() should be:
!           1          12
!           2          13
!           3          14
!           4          15
!           5          16
!           6           1
!           7           2
!           8           3
!           9           4
!          10           5
!          11           6
!          12           7
!          13           8
!          14           9
!          15          10
!          16          11

    !	    A:        |-----| 
    !	 2  B:  |-----|
    !
    this%values(17) = .16039
    this%values(18) = .17691
    this%values(19) = .19025
    this%values(20) = .21596
    this%values(21) = .21949
    this%values(22) = .25680

    this%runs(3)%start_ei = 17
    this%runs(3)%length_i = 6

! Final indices() should be:
!           1          12
!           2          13
!           3          14
!           4          15
!           5          16
!           6           1
!           7          17
!           8           2
!           9           3
!          10           4
!          11           5
!          12           6
!          13           7
!          14           8
!          15           9
!          16          10
!          17          11

    !	    A:        |-----| 
    !	 3  B:  |-------|
    !
    this%values(23) = .17522
    this%values(24) = .20837
    this%values(25) = .22482
    this%values(26) = .28484
    this%values(27) = .31157

    this%runs(4)%start_ei = 23
    this%runs(4)%length_i = 5

! Final indices() should be:
!           1          12
!           2          13
!           3          14
!           4           1
!           5          15
!           6           2
!           7          16
!           8           3
!           9           4
!          10           5
!          11           6
!          12           7
!          13           8
!          14           9
!          15          10
!          16          11

    !	    A:        |-----| 
    !	 4  B:  |-----------|
    !
    this%values(28) = .19025
    this%values(29) = .21949
    this%values(30) = .26585
    this%values(31) = .43853
    this%values(32) = .57949
    this%values(33) = .74793

    this%runs(5)%start_ei = 28
    this%runs(5)%length_i = 6

! Final indices() should be:
!           1          12
!           2          13
!           3           1
!           4          14
!           5           2
!           6           3
!           7           4
!           8          15
!           9           5
!          10           6
!          11           7
!          12          16
!          13           8
!          14           9
!          15          10
!          16          11
!          17          17

    !	    A:        |-----| 
    !	 5  B:  |-----------------|
    !
    this%values(34) = .20687
    this%values(35) = .20837
    this%values(36) = .79796

    this%runs(6)%start_ei = 34
    this%runs(6)%length_i = 3

! Final indices() should be:
!           1          12
!           2          13
!           3           1
!           4           2
!           5           3
!           6           4
!           7           5
!           8           6
!           9           7
!          10           8
!          11           9
!          12          10
!          13          11
!          14          14

    !	    A:        |-----| 
    !	 6  B:        |---|
    !
    this%values(37) = .25680
    this%values(38) = .64641
    this%values(39) = .64768
    this%values(40) = .65085
    this%values(41) = .65138

    this%runs(7)%start_ei = 37
    this%runs(7)%length_i = 5

! Final indices() should be:
!           1           1
!           2          12
!           3           2
!           4           3
!           5           4
!           6           5
!           7           6
!           8           7
!           9           8
!          10           9
!          11          13
!          12          14
!          13          15
!          14          16
!          15          10
!          16          11

    !	    A:        |-----| 
    !	 7  B:        |-----|
    !
    this%values(42) = .25680
    this%values(43) = .60569
    this%values(44) = .60717
    this%values(45) = .74793

    this%runs(8)%start_ei = 42
    this%runs(8)%length_i = 4

! Final indices() should be:
!           1           1
!           2          12
!           3           2
!           4           3
!           5           4
!           6           5
!           7           6
!           8           7
!           9           8
!          10          13
!          11          14
!          12           9
!          13          10
!          14          11
!          15          15

    !	    A:        |-----| 
    !	 8  B:        |-----------|
    !
    this%values(46) = .25680
    this%values(47) = .71905
    this%values(48) = .72070
    this%values(49) = .81381
    this%values(50) = .82062

    this%runs(9)%start_ei = 46
    this%runs(9)%length_i = 5

! Final indices() should be:
!           1           1
!           2          12
!           3           2
!           4           3
!           5           4
!           6           5
!           7           6
!           8           7
!           9           8
!          10           9
!          11          10
!          12          13
!          13          14
!          14          11
!          15          15
!          16          16

    !	    A:        |-----| 
    !	 9  B:          |-|
    !
    this%values(51) = .72971
    this%values(52) = .73074

    this%runs(10)%start_ei = 51
    this%runs(10)%length_i = 2

! Final indices() should be:
!           1           1
!           2           2
!           3           3
!           4           4
!           5           5
!           6           6
!           7           7
!           8           8
!           9           9
!          10          10
!          11          12
!          12          13
!          13          11

    !	    A:        |-----| 
    !	10  B:          |---|
    !
    this%values(53) = .66548
    this%values(54) = .66568
    this%values(55) = .66966
    this%values(56) = .74793

    this%runs(11)%start_ei = 53
    this%runs(11)%length_i = 4

! Final indices() should be:
!           1           1
!           2           2
!           3           3
!           4           4
!           5           5
!           6           6
!           7           7
!           8           8
!           9           9
!          10          12
!          11          13
!          12          14
!          13          10
!          14          11
!          15          15

    !	    A:        |-----| 
    !	11  B:          |---------|
    !
    this%values(57) = .58195
    this%values(58) = .63762
    this%values(59) = .89733
    this%values(60) = .94284

    this%runs(12)%start_ei = 57
    this%runs(12)%length_i = 4

! Final indices() should be:
!           1           1
!           2           2
!           3           3
!           4           4
!           5           5
!           6           6
!           7           7
!           8           8
!           9          12
!          10           9
!          11          13
!          12          10
!          13          11
!          14          14
!          15          15

    !	    A:        |-----| 
    !	12  B:              |-----|
    !
    this%values(61) = .74793
    this%values(62) = .95358
    this%values(63) = .95366

    this%runs(13)%start_ei = 61
    this%runs(13)%length_i = 3

! Final indices() should be:
!           1           1
!           2           2
!           3           3
!           4           4
!           5           5
!           6           6
!           7           7
!           8           8
!           9           9
!          10          10
!          11          11
!          12          12
!          13          13
!          14          14

    !	    A:        |-----| 
    !	13  B:                |---|
    !
    this%values(64) = .92077
    this%values(65) = .92863
    this%values(66) = .92947
    this%values(67) = .93094
    this%values(68) = .93273
    this%values(69) = .93696
    this%values(70) = .99914

    this%runs(14)%start_ei = 64
    this%runs(14)%length_i = 7

! Final indices() should be:
!           1           1
!           2           2
!           3           3
!           4           4
!           5           5
!           6           6
!           7           7
!           8           8
!           9           9
!          10          10
!          11          11
!          12          12
!          13          13
!          14          14
!          15          15
!          16          16
!          17          17
!          18          18

  end subroutine init_t_sort_test

  !.....................................................................
  integer function get_n_elements_t_sort_test(this) result(n_elements)
    implicit none
    class(t_sort_test) :: this

!    n_elements = size(this%values)
    n_elements = size(this%values_temp)

  end function get_n_elements_t_sort_test

  !.....................................................................
  integer function compare_t_sort_test(this, i1, i2) result(comp)
    implicit none
    class(t_sort_test) :: this
    integer, intent(in)  :: i1, i2

!    if(this%values(i1) > this%values(i2)) then
    if(this%values_temp(i1) > this%values_temp(i2)) then
      comp = COMP_GT
!    else if(this%values(i1) < this%values(i2)) then
    else if(this%values_temp(i1) < this%values_temp(i2)) then
      comp = COMP_LT
    else
      comp = COMP_EQ
    end if

  end function compare_t_sort_test

  !.....................................................................
  subroutine destroy_t_sort_test(this)
    implicit none
    class(t_sort_test) :: this

    deallocate(this%indices) !*** add error stuff
    deallocate(this%values) !*** add error stuff
    deallocate(this%values_temp) !*** add error stuff
    deallocate(this%runs) !*** add error stuff

  end subroutine destroy_t_sort_test

  !.....................................................................
  subroutine test_bsearch()
    implicit none

    integer, parameter :: n_elements = 6
    type(t_sort_test) :: sortme
    type(t_runstack) :: runstack
    integer :: ii,ei

    sortme%indices = [(ii,ii=1,n_elements)]
    allocate(sortme%values_temp(n_elements))

    sortme%values_temp(1) = 0.1
    sortme%values_temp(2) = 0.2
    sortme%values_temp(3) = 0.3
    sortme%values_temp(4) = 0.4
    sortme%values_temp(5) = 0.5
    sortme%values_temp(6) = 0.3

    call runstack%init(sortme)

    ei = runstack%bsearch_A_pinching_B(1, 5, 6, take_lo_bound=.true.)
    print *, "Binary search returned lower bound", ei, "expect 3"
    ei = runstack%bsearch_A_pinching_B(1, 5, 6, take_lo_bound=.false.)
    print *, "Binary search returned upper bound", ei, "expect 4"

    ei = runstack%bsearch_B_pinching_A(1, 5, 6, take_lo_bound=.true.)
    print *, "Binary search returned lower bound", ei, "expect 2"
    ei = runstack%bsearch_B_pinching_A(1, 5, 6, take_lo_bound=.false.)
    print *, "Binary search returned upper bound", ei, "expect 3"

  end subroutine test_bsearch

  !.....................................................................
  subroutine test_CLR()
    implicit none

    integer, parameter :: n_elements = 23
    integer, parameter :: n_in_run = 20
    type(t_sort_test) :: sortme
    type(t_runstack) :: runstack
    integer :: ii,delta_i
    logical :: no_more_As

    sortme%indices = [(ii,ii=1,n_elements)]
    allocate(sortme%values_temp(n_elements))
    do ii=1,n_in_run
      sortme%values_temp(ii) = real(ii)
    end do
    sortme%values_temp(n_in_run+1) =  5.5
    sortme%values_temp(n_in_run+2) = 10.5
    sortme%values_temp(n_in_run+3) = 10.0

    call runstack%init(sortme)

    call runstack%calc_len_run_A_in_B(1, 4, 21, delta_i, no_more_As)
    print *, "delta_i, no_more_As (expect 4 T):", delta_i,no_more_As
    call runstack%calc_len_run_A_in_B(1, 5, 21, delta_i, no_more_As)
    print *, "delta_i, no_more_As (expect 5 T):", delta_i,no_more_As
    call runstack%calc_len_run_A_in_B(1, 6, 21, delta_i, no_more_As)
    print *, "delta_i, no_more_As (expect 5 F):", delta_i,no_more_As

    call runstack%calc_len_run_A_in_B(1, 9, 22, delta_i, no_more_As)
    print *, "delta_i, no_more_As (expect 9 T):", delta_i,no_more_As
    call runstack%calc_len_run_A_in_B(1, 10, 22, delta_i, no_more_As)
    print *, "delta_i, no_more_As (expect 10 T):", delta_i,no_more_As
    call runstack%calc_len_run_A_in_B(1, 20, 22, delta_i, no_more_As)
    print *, "delta_i, no_more_As (expect 10 F):", delta_i,no_more_As

    call runstack%calc_len_run_A_in_B(1, 20, 23, delta_i, no_more_As)
    print *, "delta_i, no_more_As (expect 10 F):", delta_i,no_more_As

  end subroutine test_CLR

  !.....................................................................
  subroutine test_merge_cases()
    use erolib_sort_mod
    implicit none

    integer, parameter :: case_i=8
    type(t_sort_test) :: sortme
    type(t_runstack) :: runstack
    type(t_run) :: run_A,run_B
    integer :: ii,ierr
    real, dimension(:), allocatable :: vals
    integer, dimension(:), allocatable :: indices

    call sortme%init()

    run_A = sortme%runs(1)
    run_B = sortme%runs(case_i+1)
    allocate(sortme%values_temp(run_A%length_i + run_B%length_i))
    sortme%values_temp(1:run_A%length_i) = sortme%values(1:run_A%length_i)
    sortme%values_temp(run_A%length_i+1:)&
      = sortme%values(run_B%start_ei:run_B%start_ei+run_B%length_i-1)

    call runstack%init(sortme)

    deallocate(runstack%runs)!*** error stuff
    allocate(runstack%runs(2))!*** error stuff
    runstack%runs(1) = sortme%runs(1)
    runstack%runs(2)%start_ei = sortme%runs(1)%length_i + 1
    runstack%runs(2)%length_i = sortme%runs(case_i+1)%length_i

    ierr = 0
    vals = sortme%values_temp
    call sort(vals, ierr, indices)

    do ii=1,size(sortme%values_temp)
!      print *, ii, sortme%values_temp(ii)
      print *, ii, indices(ii), sortme%values_temp(indices(ii))
    end do
    print *, ""

    call runstack%merge_2_runs(1)

    print *, ""
    do ii=1,size(sortme%values_temp)
      print *, ii, sortme%indices(ii), sortme%values_temp(sortme%indices(ii))
    end do

  end subroutine test_merge_cases

  !.....................................................................
  subroutine test_timsort()
    implicit none

    integer, parameter :: test_i=2

    select case(test_i)
    case(1) ! test the binary search routines.
      call test_bsearch()
    case(2) ! test the calculation of the length of a sequence of values that fits between two others.
      call test_CLR()
    case(3)
      call test_merge_cases()
!    case(99)
!      call test_all()
    case default
call wrstop("test","unrecognized test_i")
    end select

  end subroutine test_timsort

  !=====================================================================
  ! Median overloads:

  !.....................................................................
  real(r4) function median_r4(values)
    real(r4), dimension(:), intent(in) :: values

    character(*), parameter :: functionName="median_r4"
    type(t_sort_scalar) :: sortable
    integer :: n_values,half_i
    logical :: is_even

    n_values = size(values)
    if(n_values<1)&
call wrstop(functionName, "Cannot calculate median of a zero-length list.")

    is_even = (modulo(n_values, 2)==0)

    call sortable%init(values)
    call timsort(sortable)
!    call sortable%check()

    if(is_even) then
      half_i = n_values/2
      median_r4 = 0.5_r4*(values(sortable%indices(half_i)) + values(sortable%indices(half_i+1)))
    else
      half_i = (n_values + 1)/2
      median_r4 = values(sortable%indices(half_i))
    end if

    call sortable%destroy()
  end function median_r4

  !.....................................................................
  real(r8) function median_r8(values)
    real(r8), dimension(:), intent(in) :: values

    character(*), parameter :: functionName="median_r8"
    type(t_sort_scalar) :: sortable
    integer :: n_values,half_i
    logical :: is_even

    n_values = size(values)
    if(n_values<1)&
call wrstop(functionName, "Cannot calculate median of a zero-length list.")

    is_even = (modulo(n_values, 2)==0)

    call sortable%init(values)
    call timsort(sortable)
!    call sortable%check()

    if(is_even) then
      half_i = n_values/2
      median_r8 = 0.5_r8*(values(sortable%indices(half_i)) + values(sortable%indices(half_i+1)))
    else
      half_i = (n_values + 1)/2
      median_r8 = values(sortable%indices(half_i))
    end if

    call sortable%destroy()
  end function median_r8

end module erolib_timsort

