! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_wcs_mod

    use erolib_str_mod
    use erolib_ffits_mod
    use erolib_general_mod
    !  use evtool_general_mod
    implicit none

      ! WCS parameter characters

    character(*), parameter :: wcs_tbl_refpix = "tcrpx"
    character(*), parameter :: wcs_tbl_refval = "tcrvl"
    character(*), parameter :: wcs_tbl_delta = "tcdlt"
    character(*), parameter :: wcs_tbl_unit = "tcuni"
    character(*), parameter :: wcs_tbl_type = "tctyp"
    character(*), parameter :: wcs_tbl_col = "ttype"

    character(*), parameter :: wcs_img_refpix = "crpix"
    character(*), parameter :: wcs_img_refval = "crval"
    character(*), parameter :: wcs_img_delta = "cdelt"
    character(*), parameter :: wcs_img_unit = "cunit"
    character(*), parameter :: wcs_img_type = "ctype"

    character(*), parameter :: wcsnamep = "PHYSICAL"
    character(*), parameter :: wcsty1p = "PHYSICAL"
    character(*), parameter :: wcsty2p = "PHYSICAL"
    character(*), parameter :: wcs_img_refpix_1p = "crpix1p"
    character(*), parameter :: wcs_img_refval_1p = "crval1p"
    character(*), parameter :: wcs_img_delta_1p = "cdelt1p"
    character(*), parameter :: wcs_img_type_1p = "ctype1p"
    character(*), parameter :: wcs_img_refpix_2p = "crpix2p"
    character(*), parameter :: wcs_img_refval_2p = "crval2p"
    character(*), parameter :: wcs_img_delta_2p = "cdelt2p"
    character(*), parameter :: wcs_img_type_2p = "ctype2p"

    type tWcs
        character(buffsize), dimension(2) :: col
        character(buffsize), dimension(2) :: type
        character(buffsize), dimension(2) :: unit
        integer, dimension(2) :: colnum
        real(dp), dimension(2) :: refpix, refval, delta
        character(buffsize), dimension(2) :: col_comm, type_comm, unit_comm, refpix_comm, refval_comm, delta_comm
        real(dp) :: crpix1p, crpix2p, crval1p, crval2p, cdelt1p, cdelt2p
        character(8) :: ctype1p, ctype2p
    contains
        procedure :: read => read_tWcs
        procedure :: write => write_tWcs
        procedure :: adjustForImage_new => adjustForImage_new_tWcs
        procedure :: adjustForImage => adjustForImage_tWcs
    end type tWcs

    interface alloc_wcs
        module procedure alloc_wcs_single, alloc_wcs_array
    end interface alloc_wcs

    interface dealloc_wcs
        module procedure dealloc_wcs_single, dealloc_wcs_array
    end interface dealloc_wcs

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> allocate wcs types
    !
    !> @param[in] wcs
    !> @param[out] n
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine alloc_wcs_single(wcs, status)
        implicit none

        integer :: stat
        integer, intent(inout) :: status

        type(tWcs), allocatable :: wcs

        character(buffsize) :: errmsg
        character(*), parameter :: function = "alloc_wcs_single"

        call dealloc_wcs(wcs, status)
        allocate(wcs, stat=stat, errmsg=errmsg)
        call FError(function, stat, errmsg, status)

    end subroutine alloc_wcs_single

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> dealloacte wcs structures
    !
    !
    !
    !> @param[inout] wcs
    !> @param[inout] status
    !---------------------------------------------------------------------------
    subroutine dealloc_wcs_single(wcs, status)

        integer :: stat
        integer, intent(inout) :: status

        type(tWCS), allocatable :: wcs

        character(*), parameter :: function = "dealloc_wcs_single"

        if(allocated(wcs)) then
            deallocate(wcs, stat=stat)
            call FError(function, stat, "dealloc wcs type", status)
        end if

    end subroutine dealloc_wcs_single

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> allocate wcs types
    !
    !> @param[in] wcs
    !> @param[out] n
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine alloc_wcs_array(wcs, n, status)
        implicit none

        integer :: stat
        integer, intent(in) :: n
        integer, intent(inout) :: status

        type(tWcs), allocatable, dimension(:) :: wcs

        character(buffsize) :: errmsg
        character(*), parameter :: function = "alloc_wcs_array"

        call dealloc_wcs(wcs, status)
        allocate(wcs(n), stat=stat, errmsg=errmsg)
        call FError(function, stat, errmsg, status)

    end subroutine alloc_wcs_array

        !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> dealloacte wcs structures
    !
    !
    !
    !> @param[inout] wcs
    !> @param[inout] status
    !---------------------------------------------------------------------------
    subroutine dealloc_wcs_array(wcs, status)

        integer :: stat
        integer, intent(inout) :: status

        type(tWCS), allocatable, dimension(:) :: wcs

        character(*), parameter :: function = "dealloc_wcs_single"

        if(allocated(wcs)) then
            deallocate(wcs, stat=stat)
            call FError(function, stat, "dealloc wcs type", status)
        end if

    end subroutine dealloc_wcs_array


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> return a new wcs with adjusted values to account for an image with input of rebin, resolution, and new center position
    !
    !
    !
    !> @param[in] rebin, resolution, center_position
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

  subroutine adjustForImage_new_tWcs(this, events_wcs, pixel_size_XY, im_dimensions, im_centre_XY, status)
    ! The event list has X and Y columns, which have associated WCS keywords to translate between such values and RA/dec. In the present routine we convert these WCS keywords into ones suitable for an image of the dimensions given. 

    implicit none

    class(tWcs) :: this
    type(tWcs), intent(in) :: events_wcs
    integer(i4), intent(in), dimension(2) :: pixel_size_XY, im_dimensions
    real(r8), intent(in), dimension(2) :: im_centre_XY
    integer, intent(inout) :: status

    character(*), parameter :: function = "adjustForImage_new_tWcs"
    real(r8), dimension(2) :: im_centre_pixels
    integer :: ii

    if(status > 100 .or. status == s_errRep) return

    ! For the standard WCS (the ones that convert between pixel number and RA/dec), we're going to leave CRVAL unchanged and alter CRPIX.
    !
    do ii=1,2
      this%refval(ii) = events_wcs%refval(ii)
      this%delta( ii) = events_wcs%delta( ii)*pixel_size_XY(ii)

      ! The WCS convention is that the centre of the 1st pixel has coordinate value 1.0; thus the leading edge of that pixel (since image pixels are by definition of width 1 in pixel coordinates) has coordinate value 0.5, and the trailing edge, 1.5. We take here the centre of the image in these coordinates to be a floating-point number exactly half-way between the leading edge of the first image pixel (LEFP) and the trailing edge of the last (TELP). E.g. for an image dimension of size 2, LEFP= (as always) 0.5 and TELP=2.5. Thus the centre is at 1.5, which is (as expected) on the boundary between the 1st and 2nd pixels.
      !
      im_centre_pixels(ii) = 0.5d0*(1 + im_dimensions(ii))

      this%refpix(ii) = im_centre_pixels(ii)&
        + (events_wcs%refpix(ii) - im_centre_XY(ii))/float(pixel_size_XY(ii))

      this%colnum(ii) = ii

      this%type(       ii) = events_wcs%type(ii)
      this%type_comm(  ii) = events_wcs%type_comm(ii)
      this%unit(       ii) = events_wcs%unit(ii)
      this%unit_comm(  ii) = events_wcs%unit_comm(ii)
      this%refpix_comm(ii) = events_wcs%refpix_comm(ii)
      this%refval_comm(ii) = events_wcs%refval_comm(ii)
      this%delta_comm( ii) = events_wcs%delta_comm(ii)
    end do

    ! For the 'physical' WCS which connects the image pixel coordinates with the X/Y values, we keep CRPIX1P at the centre of the image and shift CRVAL1P instead.
    !
    this%crpix1p = im_centre_pixels(1)
    this%crpix2p = im_centre_pixels(2)
    this%crval1p = im_centre_XY(1)
    this%crval2p = im_centre_XY(2)
    this%cdelt1p = pixel_size_XY(1)
    this%cdelt2p = pixel_size_XY(2)
    this%ctype1p = "X"
    this%ctype2p = "Y"

  end subroutine adjustForImage_new_tWcs

  !.....................................................................
    subroutine adjustForImage_tWcs(this, rebin, resolution, center_position, status)
!*** 20200703 IMS: retain this so long as evtool_mod.f90 is kept. Nothing else needs to call it.
        implicit none

        integer, intent(inout) :: status
        integer(i4), intent(in), dimension(2) :: rebin, resolution
        class(tWcs) :: this
        real(r8), intent(in), dimension(2) :: center_position

        character(*), parameter :: function = "adjustForImage_tWcs"

        if(status > 100 .or. status == s_errRep) return

        this%delta(:) = this%delta(:) * rebin
        this%refpix(1) = this%refpix(1) - nint(center_position(1))
        this%refpix(2) = this%refpix(2) - nint(center_position(2))

        ! shift to account for image and refactor
        this%refpix(1) = this%refpix(1) + resolution(1) * rebin(1) / 2.0_r8
        this%refpix(2) = this%refpix(2) + resolution(2) * rebin(2) / 2.0_r8

        ! -0.5 to move from center in pix to bottom left
        ! 0.5 / rebin to account for rebin
        this%refpix(1) = this%refpix(1) / rebin(1) - 0.5_r8 + 0.5_r8 / rebin(1)
        this%refpix(2) = this%refpix(2) / rebin(2) - 0.5_r8 + 0.5_r8 / rebin(2)

        this%colnum(1) = 1
        this%colnum(2) = 2

        ! experimental code to determine values for the Physical WCS
        this%crpix1p = this%refpix(1)
        this%crpix2p = this%refpix(2)
        this%crval1p = nint(center_position(1))
        this%crval2p = nint(center_position(2))
        this%cdelt1p = rebin(1)
        this%cdelt2p = rebin(2)
        this%ctype1p = "X"
        this%ctype2p = "Y"

    end subroutine adjustForImage_tWcs

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> read the appropriate WCS header keywords for two coordinate columns
    !
    !
    !
    !> @param[in] unit
    !> @param[out] col1, col2
    !> @param[inout] status
    !---------------------------------------------------------------------------

  subroutine read_tWcs(this, unitNum, hdutype, col1, col2, status)
    implicit none

    class(tWCS) :: this
    integer,      intent(in)    :: unitNum
    integer,      intent(in)    :: hdutype
    character(*), intent(in)    :: col1, col2
    integer,      intent(inout) :: status

    character(*), parameter :: functionName = "read_tWcs"
    integer, parameter :: fitsErrKwdNotFound=202
    character(buffsize), dimension(2) :: colnum_str
    integer :: ii,ierr

    if(status > 100 .or. status == s_errRep) return
        
    call ffits_get_colnum(unitNum, .false., col1, this%colnum(1), status)
    call ffits_get_colnum(unitNum, .false., col2, this%colnum(2), status)
        
    colnum_str(1) = int2str(this%colnum(1), status)
    colnum_str(2) = int2str(this%colnum(2), status)

    ! if for some reason the column numbers could not be read / converted return
    if(status == s_errRep) return

    ierr = 0
    if(hdutype == BINARY_TBL) then
      do ii = 1, 2
        call FTGKYS(unitNum, wcs_tbl_col&
          //trim(colnum_str(ii)), this%col(ii), this%col_comm(ii), ierr)
        if(ierr==fitsErrKwdNotFound) then
          call wrchatty(functionName, wcs_tbl_col//trim(colnum_str(ii))//" keyword not found.")
      exit
        end if

        call FTGKYS(unitNum, wcs_tbl_type&
          //trim(colnum_str(ii)), this%type(ii), this%type_comm(ii), ierr)
        if(ierr==fitsErrKwdNotFound) then
          call wrchatty(functionName, wcs_tbl_type//trim(colnum_str(ii))//" keyword not found.")
      exit
        end if

        call FTGKYS(unitNum, wcs_tbl_unit&
          //trim(colnum_str(ii)), this%unit(ii), this%unit_comm(ii), ierr)
        if(ierr==fitsErrKwdNotFound) then
          call wrchatty(functionName, wcs_tbl_unit//trim(colnum_str(ii))//" keyword not found.")
      exit
        end if

        call FTGKYD(unitNum, wcs_tbl_refpix&
          //trim(colnum_str(ii)), this%refpix(ii), this%refpix_comm(ii), ierr)
        if(ierr==fitsErrKwdNotFound) then
          call wrchatty(functionName, wcs_tbl_refpix//trim(colnum_str(ii))//" keyword not found.")
      exit
        end if

        call FTGKYD(unitNum, wcs_tbl_refval&
          //trim(colnum_str(ii)), this%refval(ii), this%refval_comm(ii), ierr)
        if(ierr==fitsErrKwdNotFound) then
          call wrchatty(functionName, wcs_tbl_refval//trim(colnum_str(ii))//" keyword not found.")
      exit
        end if

        call FTGKYD(unitNum, wcs_tbl_delta&
          //trim(colnum_str(ii)), this%delta(ii), this%delta_comm(ii), ierr)
        if(ierr==fitsErrKwdNotFound) then
          call wrchatty(functionName, wcs_tbl_delta//trim(colnum_str(ii))//" keyword not found.")
      exit
        end if
      end do

      if(ierr==fitsErrKwdNotFound) then
        call wrerr(functionName, warn, "Could not read WCS (columns "//trim(col1)//", "//trim(col2)//").")
        status = c_err
  return
      else if(ierr/=0) then
        call wrerr(functionName, s_err, "Error reading WCS (columns "//trim(col1)//", "//trim(col2)//").")
        status = s_errRep
  return
      end if
    end if

        if(hdutype == IMAGE_HDU) then
          do ii = 1, 2
            call ffits_read_key_str(unitNum, wcs_img_type&
              //trim(colnum_str(ii)), this%type(ii), status, this%type_comm(ii))
            call ffits_read_key_str(unitNum, wcs_img_unit&
              //trim(colnum_str(ii)), this%unit(ii), status, this%unit_comm(ii))
            call ffits_read_key(unitNum, wcs_img_refpix&
              //trim(colnum_str(ii)), this%refpix(ii), status, this%refpix_comm(ii))
            call ffits_read_key(unitNum, wcs_img_refval&
              //trim(colnum_str(ii)), this%refval(ii), status, this%refval_comm(ii))
            call ffits_read_key(unitNum, wcs_img_delta&
              //trim(colnum_str(ii)), this%delta(ii), status, this%delta_comm(ii))
          end do

            ! experimental code writing physical WCS for image extension
!            call ffits_read_key_str(unitNum, "WCSNAMEP", wcsnamep, status)
!            call ffits_read_key_str(unitNum, "WCSTY1P", wcsty1p, status)
!            call ffits_read_key_str(unitNum, "WCSTY2P", wcsty2p, status)
            call ffits_read_key_str(unitNum, wcs_img_type_1p, this%ctype1p, status)
            call ffits_read_key_str(unitNum, wcs_img_type_2p, this%ctype2p, status)
            call ffits_read_key(    unitNum, wcs_img_refpix_1p, this%crpix1p, status)
            call ffits_read_key(    unitNum, wcs_img_refval_1p, this%crval1p, status)
            call ffits_read_key(    unitNum, wcs_img_delta_1p, this%cdelt1p, status)
            call ffits_read_key(    unitNum, wcs_img_refpix_2p, this%crpix2p, status)
            call ffits_read_key(    unitNum, wcs_img_refval_2p, this%crval2p, status)
            call ffits_read_key(    unitNum, wcs_img_delta_2p, this%cdelt2p, status)

        end if

    end subroutine read_tWcs

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> write the appropriate WCS header keywords for two coordinate columns
    !
    !
    !
    !> @param[in] unit
    !> @param[out] col1, col2
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine write_tWcs(this, unit, hdutype, status)
        implicit none

        integer, intent(inout) :: status
        integer, intent(in) :: unit
        integer, intent(in) :: hdutype
        integer :: ii

        class(tWCS) :: this

        character(buffsize), dimension(2) :: colnum_str
        character(*), parameter :: function = "write_tWcs"

        if(status > 100 .or. status == s_errRep) return
        
        if(hdutype == BINARY_TBL) then
            ! update column numbers
            call ffits_get_colnum(unit, .false., this%col(1), this%colnum(1), status)
            call ffits_get_colnum(unit, .false., this%col(2), this%colnum(2), status)
        
            colnum_str(1) = int2str(this%colnum(1), status)
            colnum_str(2) = int2str(this%colnum(2), status)
        
            do ii = 1, 2
                call ffits_update_key_str(unit, wcs_tbl_col&
                  // trim(colnum_str(ii)), this%col(ii), this%col_comm(ii), status)
                call ffits_update_key_str(unit, wcs_tbl_type&
                  // trim(colnum_str(ii)), this%type(ii), this%type_comm(ii), status)
                call ffits_update_key_str(unit, wcs_tbl_unit&
                  // trim(colnum_str(ii)), this%unit(ii), this%unit_comm(ii), status)
                call ffits_update_key(unit, wcs_tbl_refpix&
                  // trim(colnum_str(ii)), this%refpix(ii), this%refpix_comm(ii), status)
                call ffits_update_key(unit, wcs_tbl_refval&
                  // trim(colnum_str(ii)), this%refval(ii), this%refval_comm(ii), status)
                call ffits_update_key(unit, wcs_tbl_delta&
                  // trim(colnum_str(ii)), this%delta(ii), this%delta_comm(ii), status)
            end do
        end if

        if(hdutype == IMAGE_HDU) then
            ! since its an image the colnum would be 1 and 2
            colnum_str(1) = "1"
            colnum_str(2) = "2"
            do ii = 1, 2
                call ffits_update_key_str(unit, wcs_img_type&
                  // trim(colnum_str(ii)), this%type(ii), this%type_comm(ii), status)
                call ffits_update_key_str(unit, wcs_img_unit&
                  // trim(colnum_str(ii)), this%unit(ii), this%unit_comm(ii), status)
                call ffits_update_key(unit, wcs_img_refpix&
                  // trim(colnum_str(ii)), this%refpix(ii), this%refpix_comm(ii), status)
                call ffits_update_key(unit, wcs_img_refval&
                  // trim(colnum_str(ii)), this%refval(ii), this%refval_comm(ii), status)
                call ffits_update_key(unit, wcs_img_delta&
                  // trim(colnum_str(ii)), this%delta(ii), this%delta_comm(ii), status)
            end do

            ! experimental code writing physical WCS for image extension
            call ffits_update_key_str(unit, "WCSNAMEP", wcsnamep, "", status)
            call ffits_update_key_str(unit, "WCSTY1P", wcsty1p, "", status)
            call ffits_update_key_str(unit, "WCSTY2P", wcsty2p, "", status)
            call ffits_update_key_str(unit, wcs_img_type_1p, this%ctype1p, "", status)
            call ffits_update_key_str(unit, wcs_img_type_2p, this%ctype2p, "", status)
            call ffits_update_key(unit, wcs_img_refpix_1p, this%crpix1p, "", status)
            call ffits_update_key(unit, wcs_img_refval_1p, this%crval1p, "", status)
            call ffits_update_key(unit, wcs_img_delta_1p, this%cdelt1p, "", status)
            call ffits_update_key(unit, wcs_img_refpix_2p, this%crpix2p, "", status)
            call ffits_update_key(unit, wcs_img_refval_2p, this%crval2p, "", status)
            call ffits_update_key(unit, wcs_img_delta_2p, this%cdelt2p, "", status)

        end if

    end subroutine write_tWcs

end module erolib_wcs_mod
