!-----------------------------------------------------------------------
! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------

module evatt_mod

!------------------------------------------------------------------------------
! eROSITA TEL chain, apply corrected attitude to events.
! Author : GLamer, AIP         date:  20-MAR-2014 : v1.0    
! 
!
! subroutines:
!  evatt_ini(par,ierr)
!  evatt_par(par,ierr)
!  evatt_in(par,evlist,attitude,gti,instpar,ierr) 
!  evatt_calc(par,evlist,attitude,gti,instpar,ierr)
!  evatt_wrcol(evlist,ierr)
!------------------------------------------------------------------------------
  

!  use proc
  use param
  use numkind  
  use fitsimage
  use cal
  use att_io
  use tel_io
  use astrolib


  
  implicit none

  public

!  type(corratt_type)                      :: att
!  type(evlist_type)                       :: evlist
!  type(instpar_type)                      :: instpar


contains




  subroutine evatt_ini(par,ierr)

  !------------------------------------------------------------
  ! Calls subroutine OPAST to open parameter and log files
  !------------------------------------------------------------

    implicit none

    type(evatt_param_type), intent(out) :: par
    integer(kind=int32) , intent(out)   :: ierr
    integer(kind=int32)                 :: status
    character(len=*),parameter          :: rtname='EVATT_INI'
    character(len=*),parameter          :: taskname='EVATT' 

    ierr=0
    status=0

    call opast(taskname,par%lun,status)

    if (status /= 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening parameter file')
    endif 

  end subroutine evatt_ini


!-------------------------------------------------------------------------------


  subroutine evatt_par(par,ierr)
 
  !---------------------------------------------------------------------
  ! Get task parameters 
  !
  !---------------------------------------------------------------------  
  
    implicit none

    type(evatt_param_type)                  :: par
    integer(kind=int32)      , intent(out)  :: ierr
    integer(kind=int32)                     :: status

    character(len=*),parameter              :: rtname='EVATT_PAR'
    
    

    ierr = 0

    call rgetc(par%lun,'EVENTFILE'   ,par%eventfile,status)
    if (status /= 0) ierr = max(ierr,status)   

    call rgetc(par%lun,'ATTFILE'   ,par%attfile,status)
    if (status /= 0) ierr = max(ierr,status)   

    call rgetc(par%lun,'GTIFILE'   ,par%gtifile,status)
    if (status /= 0) ierr = max(ierr,status)   

   

    if (ierr /= 0)  call wrerr(rtname,ierr,'Error getting parameters')


  end  subroutine evatt_par


!-------------------------------------------------------------------------------


  subroutine evatt_in(par,evlist,attitude,gti,instpar,ierr) 

  !---------------------------------------------------------------------
  ! Open event list file par%eventfile, read event list
  ! Open attitude file par%attfile , read attitude data 
  !---------------------------------------------------------------------    


    implicit none

    type(evatt_param_type)                  :: par
    type(evatt_list_type)                   :: evlist
    type(corratt_type)                      :: attitude
    type(gti_type)                          :: gti
    type(instpar_type)                      :: instpar
    integer(kind=int32),       intent(out)  :: ierr

    type(cal_par_type)                      :: cal_par
    character(len=clong)                    :: cal_file
    integer(kind=int32)                     :: status,ext
    ! make gtiext a task parameter?
    character(len=16)                       :: attext
    character(len=16)                       :: hkext
    character(len=16)                       :: gtiext
    character(len=1)                        :: instnum
    character(len=1024)                     :: wstr
    character(len=*),parameter              :: rtname='EVATT_IN'

    ierr = 0
    status=0   


    call fget_evatt_list(par%eventfile,'EVENTS',evlist,status)

    if (status > 2) then
      ierr = 3
      call wrerr(rtname,ierr,'Error reading EVENTS file')
    endif
   
    if (ierr < 3) then 
      call get_instnum(evlist%stdhdr%instrume,instnum,status)
      attext='CORRATT'//instnum
      hkext='HK'//instnum
      gtiext='GTI'//instnum
    endif

    if (status > 2) then
      ierr = 3
      call wrerr(rtname,ierr,'Error reading EVENTS file')
    else
      

      status=0
      call fget_corratt(par%attfile,attext,attitude,status)
      if (status > 2) then
        ierr = 3
        call wrerr(rtname,ierr,'Error reading ATTITUDE file')
      endif
    endif


    if (ierr < 2) then 
      status=0
      gti%extname=gtiext
      call fget_telgti(par%gtifile,gti,status)
      if (status > 2) then
        ierr = 3
        call wrerr(rtname,ierr,'Error reading GTI file '//trim(par%gtifile)//':'//gti%extname)
      endif
    endif

 
    ! get instrument parameters from CAL file:
 
    if (ierr < 2) then 
      call cal_setpar(evlist%stdhdr,cal_par,status)
      cal_par%codenam='GEOM'
      cal_par%expression='-'
      ! empty expression causes problems, set OBSMODE='SURVEY'
      ! (not relevant for instpar file)
!!!      call cal_stringtoex('OBSMODE','SURVEY',cal_par%expression,status)
      call cal_filename(cal_par,cal_file,ext,status)
      ! ignore status=1,status=2 warnings:
      if (status > 2) then
        ierr = 3
        call wrerr(rtname,ierr,'CAL file not found')
      else
        call cal_getinstpar(cal_file,ext,instpar,status)
        if (status > 0) then
          ierr = 3
          call wrerr(rtname,ierr,'Cannot read CAL file')
        endif
      endif  
    endif
 
    ! Write warning about correct use of INSTPAR files
    wstr='General warning: With pipeline version 946 the handling of camera time jumps has changed:'
    call wrerr(rtname,1,trim(wstr))
    wstr='In v946 time jumps are now corrected by evprep and new (v15+) INSTPAR files have to be used.'
    call wrerr(rtname,1,trim(wstr))
    if (.not. evlist%stdhdr%e_t_offcor .or. (evlist%stdhdr%t_offcor == 'F')) then
      wstr='According to the header the TIME values in this file are NOT corrected.'
      call wrerr(rtname,1,trim(wstr))
      wstr='INSTPAR files with version number below 15 have to be used.'
    else if (evlist%stdhdr%e_t_offcor .and. (evlist%stdhdr%t_offcor == 'T')) then
      wstr='According to the header the TIME values in this file are corrected by evprep.'
      call wrerr(rtname,1,trim(wstr))
      wstr='INSTPAR files  v15 have to  be used.'  
    endif
    wstr='The following INSTPAR file is used:'
    call wrerr(rtname,1,trim(wstr))
    call wrerr(rtname,1,trim(cal_file))   

  end subroutine evatt_in

!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------


!  subroutine evatt_calc(par,evlist,attitude,gti,instpar,ierr) 
  subroutine evatt_calc(par,evlist,attitude,gti,instpar,ierr) 
    
  !----------------------------------------------------------------------  
  ! Interpolates attitude data at event times.
  ! Applies distortion correction to raw coordinates 
  ! and calculates RA, DEC for each event.
  !
  !----------------------------------------------------------------------  


    implicit none

    type(evatt_param_type)                   :: par
    type(evatt_list_type)                    :: evlist
    type(corratt_type)                       :: attitude
    type(gti_type)                           :: gti
    type(instpar_type)                       :: instpar
    integer(kind=int32),       intent(out)   :: ierr

    integer(kind=int32)                      :: iev,nev,status,quality
    integer(kind=int32)                      :: err_counter_noatt,err_counter_qual
    real(kind=real64)                        :: ev_rapnt,ev_decpnt,ev_roll,rotang
    real(kind=real64)                        :: ra, dec, dx, dy, xdet, ydet
    real(kind=real64)                        :: xcorr, ycorr,xrot,yrot,xpix2rad,ypix2rad
    logical(kind=log32)                      :: good
    character(len=80)                        :: errstr
    character(len=10)                        :: numstr    
    character(len=*),parameter               :: rtname='EVATT_CALC'

    ierr = 0
    err_counter_noatt=0
    err_counter_qual=0

    xpix2rad=instpar%x_platescale*cal_deg2rad/3600.D0
    ypix2rad=instpar%y_platescale*cal_deg2rad/3600.D0

    nev=evlist%shapehdr%naxis2
   !test:
    !open(1,file='att_interp.dat')

    do iev=1,nev
      
      if (ierr < 3) then
        ierr=0
      else
        exit
      endif

      call  ingti(evlist%data(iev)%time,gti,good,status)

      status=0
      quality=0
     call interpol_corratt(attitude,evlist%data(iev)%time+instpar%timelag,ev_rapnt,ev_decpnt,ev_roll,quality,status)
     ! TEST without time lags
!      call interpol_corratt(attitude,evlist%data(iev)%time,ev_rapnt,ev_decpnt,ev_roll,quality,status)
     ! write(*,*) 'done interp: ',quality, status
      if (status > 2) then
        ierr=3  
        call wrerr(rtname,ierr,'Error interpolating attitude data')        
      else if (status > 0) then
        ierr=2 
        err_counter_noatt=err_counter_noatt+1
!        call wrerr(rtname,ierr,'No attitude data for event')
      else
        if (quality > 0 .and. good) then
          ierr=2
          err_counter_qual=err_counter_qual+1  
!          call wrerr(rtname,ierr,'Warning: Low attitude quality within good GTI') 
        endif 
        status=0
      endif


      if (ierr == 0) then

         !----------------------------------------------------------
         ! Definition of roll angle ev_roll:
         ! counter-clockwise angle between the north direction
         ! and the X direction of the RAWX/RAWY coordinate system
         ! => 90 deg offset
         ! rotang = -ev_roll + 90.D0  (GL 21/02/2019)
         ! changed sign of RAWY direction (15/01/2019):
         ! Orientation of sky image in RAWX/RAWY system for roll=0: north => -RAWX, east => +RAWY
         !
         ! Rotation by -roll+90.deg + flip in Y axis :  
         !         north => +dy,  east => +dx
         !         (required by cal_tan2sphere)
         !----------------------------------------------------------

         ! 
         rotang = -ev_roll + 90.D0

         xdet = dble(evlist%data(iev)%rawx) + evlist%data(iev)%subx
         ydet = dble(evlist%data(iev)%rawy) + evlist%data(iev)%suby


         status=0
         call cal_distcorr(instpar,xdet,ydet,xcorr,ycorr,status)
         ierr = max(ierr,status)

         dx=(xcorr-instpar%x_ref)*xpix2rad
         dy=(ycorr-instpar%y_ref)*ypix2rad

         

         status=0 
         ! apply rotang counter clock-wise:
!         call cal_xyrotate(xcorr,xcorr,instpar%x_ref,instpar%y_ref,rotang,xrot,yrot,status)
         call cal_xyrotate(dx,dy,0.D0,0.D0,rotang,xrot,yrot,status)
         ! one axis needs to be flipped, : 
         yrot=-yrot  
         ierr = max(ierr,status)
         status=0 


         call cal_tan2sphere(xrot,yrot,ev_rapnt*cal_deg2rad,ev_decpnt*cal_deg2rad,ra,dec,status)
         ierr = max(ierr,status)
         status=0 

         evlist%data(iev)%ra  = ra  * cal_rad2deg
         evlist%data(iev)%dec = dec * cal_rad2deg

         ! Keep RA in [0,360]:
         if (evlist%data(iev)%ra > 360.D0) then
           evlist%data(iev)%ra = mod(evlist%data(iev)%ra,360.D0)
         else if ( evlist%data(iev)%ra < 0.D0) then
           evlist%data(iev)%ra = evlist%data(iev)%ra + 360.D0 
         endif        

         call calc_ownerflag(evlist%data(iev)%ra,evlist%data(iev)%dec,evlist%data(iev)%flag,status) 
         ierr = max(ierr,status)

      else
      
        evlist%data(iev)%ra = 0.0
        evlist%data(iev)%dec = 0.0 

 
      endif     
      

      
    enddo
    
    if (err_counter_noatt > 0) then 
      ierr=2
      write(numstr,"(I10)") err_counter_noatt
      errstr='Missing attitude data for '//numstr//' event(s)'
      call wrerr(rtname,ierr,errstr)
    endif
    
    if (err_counter_qual > 0) then 
      ierr=2
      write(numstr,"(I10)") err_counter_qual
      errstr='Low quality attitude data for '//numstr//' event(s)'
      call wrerr(rtname,ierr,errstr)
    endif


    !test:
    !close(1)

   end subroutine evatt_calc

!-------------------------------------------------------------------------------



  subroutine evatt_wrcol(evlist,ierr) 
  
    !-----------------------------------------------------------
    ! Writes RA, DEC, FLAG columns to existing event file.
    ! 
    !
    !-----------------------------------------------------------

    implicit none

    type(evatt_list_type)              :: evlist
    integer(kind=int32), intent(out)   :: ierr
!    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2,colnum
!    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: status,fstat,unit,hdutype,rwmode
    
!    integer(kind=int32)                :: nhdu, varidat
!    real(kind=real64), pointer, dimension(:)  :: ra,dec

    character(len=80)                  :: errstr
    character(len=*),parameter         :: rtname='EVATT_WRCOL'

    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    rwmode= 0    !readonly
    hdutype = 2    ! binary table
!    fnull  = .false.



    !-----------------------------------------------------
    ! Closing the file in fget_evatt_list and re-opening
    ! it here leades to segmenation fault in FTNOPN
    !
    ! call FTGIOU(unit,status)
    ! call FTNOPN(unit,evlist%filename,rwmode,status)
    !------------------------------------------------------  

    ! file is still  open, here's the lun: 
    unit=evlist%lun

 
    call FTMNHD(unit,hdutype,'EVENTS',0,status)
    fstat=max(fstat,status)


    nrows=evlist%shapehdr%naxis2

    if (fstat /= 0) then   
      ierr = 3
      errstr = 'error writing  file '//trim(evlist%filename)
      call WRERR(rtname,ierr,errstr) 
    else 
      call FTGRSZ(unit, maxbuf, status)
      irow=0 
      do while (irow < nrows .and. fstat == 0)
        rowbuf = min(maxbuf,nrows-irow)
        i1=irow+1
        i2=irow+rowbuf
        irow=i2  
        
        call  FTGCNO(unit,.false.,'RA',colnum,status)
        call  FTPCLD(unit,colnum,i1,1,rowbuf,evlist%data(i1:i2)%ra,status) 
        call  FTGCNO(unit,.false.,'DEC',colnum,status)
        call  FTPCLD(unit,colnum,i1,1,rowbuf,evlist%data(i1:i2)%dec,status) 
        call  FTGCNO(unit,.false.,'FLAG',colnum,status)
        call  FTPCLJ(unit,colnum,i1,1,rowbuf,evlist%data(i1:i2)%flag,status) 
      enddo
    endif

    fstat=max(fstat,status)

    if (fstat /= 0) then
      ierr = 3
      errstr = 'Error writing data to table '//trim(evlist%filename)
      call WRERR(rtname,ierr,errstr)
    else
      status = 0
      call fwrite_stdhdr(unit,evlist%stdhdr,status)
      
    endif
 
    call FTCLOS(unit,status) 
    call   FTFIOU(unit,status)   

  end subroutine evatt_wrcol 

!-------------------------------------------------------------------------------

  subroutine calc_ownerflag(ra,dec,flag,ierr) 

 
    !--------------------------------------------------------------------
    ! Determines proprietary sky area (IKI, MPE or MPE_IKI from 
    ! RA, DEC [degrees], and sets  bits 0 and 1 of FLAG accordingly    
    ! NOTE: owner=MPE=> bit0=0, owner=IKI=> bit1=0, MPE+IKI: both bits =0 
    !
    !--------------------------------------------------------------------


    implicit none
 
    real(kind=real64)                  :: ra    
    real(kind=real64)                  :: dec
    integer(kind=int32)                :: flag
    integer(kind=int32)                :: ierr

    integer(kind=int32)                :: status
    real(kind=real64)                  :: glon,glat,dlon_dual
    logical(kind=log32)                :: ikiflag,mpeflag
    real(kind=real64),parameter        :: sin_hw_dual = 0.0055268536447D0
    ! sin(19 arcmin), dual ownership stripe currently set to +- 19 arcmin
    
    character(len=*),parameter              :: rtname='CALC_OWNERFLAG'  


    ierr=0
    ikiflag=.false.
    mpeflag=.false.

    call equgal(2451544.5d0,.true.,ra*cal_deg2rad, dec*cal_deg2rad, glon, glat, status)

    if (status == 0) then


!      dlon_dual =  asin(sin_hw_dual/cos(glat))*rad2deg    
!      if (glon*rad2deg  >= 360.D0-dlon_dual .or.  glon*rad2deg >= 0d0    .and. glon*rad2deg <= 180d0+dlon_dual) ikiflag=.true.
!      if (glon*rad2deg  >= 180.D0-dlon_dual .and. glon*rad2deg <= 360d0  .or.  glon*rad2deg <= dlon_dual) mpeflag=.true.
      if (sin_hw_dual/cos(glat) < 1.D0) then 
        dlon_dual =  asin(sin_hw_dual/cos(glat))    
        if (glon  >= d2pi-dlon_dual .or.  glon >= 0d0    .and. glon <= dpi+dlon_dual) ikiflag=.true.
        if (glon  >= dpi -dlon_dual .and. glon <= d2pi  .or.  glon <= dlon_dual) mpeflag=.true.
      else
        ! event is close to galactic poles and within dual OS stripe:
        ikiflag=.true.
        mpeflag=.true.
      endif    
     
      flag = ior(flag,3)
      if (mpeflag) flag = ibclr(flag,0)
      if (ikiflag) flag = ibclr(flag,1)
    else
      ierr=3
      call wrerr(rtname,ierr,'Error getting galactic coordinates')
    endif

  end subroutine calc_ownerflag

!-------------------------------------------------------------------------------  


end module evatt_mod
