! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
program evtool_main

    use param
    use erolib_ffits_mod
    use erolib_str_mod
    use erolib_general_mod
    use erolib_eventfile_mod
    use evtool_mod

    implicit none

    integer :: status, ii, ierr

    type(tEvtool) :: task
    type(tEventFile), allocatable, dimension(:) :: eventfiles
    type(tEventFile) :: eventfile_out
    character(buffsize) :: msg,fileName
    integer, dimension(:), allocatable :: ccdnrs
    real*4 :: cpuTimeStart, cpuTimeFinis

    character(*), parameter :: functionName = "evtool_main"

    status = 0

    call cpu_time(cpuTimeStart)

    ! read/open the .par file and the logs
    call opast(status)
    if(status /= ok) call wrstop(functionName, "error read par file")

    call wrchatty(functionName, "Read params.")
    call task%init(status)
    call task%read_params(status)
    !call task%print_params(status)
    call task%validate(status)

    call alloc_eventfiles(eventfiles, size(task%infiles), status)
    do ii = 1, size(task%infiles)
      call eventfiles(ii)%read_ccd_nrs(task%infiles(ii), ccdnrs, status)
      call eventfiles(ii)%init(ccdnrs, status)
      deallocate(ccdnrs)
      call eventfiles(ii)%setup_for_readwrite([TCOL_READ_BIT], status) ! This initializes and sets useage for all extensions.

      if(.not.task%fsp%do_filter_gti) then
        if(.not.eventfiles(ii)%filtered_gti) then
          msg = int2str(ii)
          call wrerr(functionName, warn, "GTI filtering not yet done on file "//trim(msg))
        end if
      end if
    end do

    if(task%memset) then
      call wrlog(functionName, "memset==.true., so memory setting for small files is used.")

      call wrlog(functionName, "Reading infiles.")
      do ii = 1, size(task%infiles)
        write(msg,'(a,i2,a)') "File ", ii, ":"
        call wrchatty(functionName, msg)

        call wrchatty(functionName, "  Opening the file.")
        call eventfiles(ii)%open(FF_READONLY, status, filename=task%infiles(ii))

        call wrchatty(functionName, "  Reading the file.")
        call eventfiles(ii)%read(status, want_wcs=.true.,repair_gtis=task%repair_gtis)
        call wrchatty(functionName, "  Closing the file.")
        call eventfiles(ii)%close(status)
        if(status /= ok) then
          call wrerr(functionName, s_err, "ERROR with file: " // trim(task%infiles(ii)))
        end if
      end do

    else ! memset==.false., the current default
      call wrlog(functionName, "memset==.false., so memory setting for large files is used.")

      ! read in the informations for the stdheader and wcs
      do ii = 1, size(task%infiles)
        write(msg,'(a,i2,a)') "File ", ii, ":"
        call wrchatty(functionName, msg)

        call wrchatty(functionName, "  Opening the file.")
        call eventfiles(ii)%open(FF_READONLY, status, filename=task%infiles(ii))
        call eventfiles(ii)%read_stdheader(status)

        call wrchatty(functionName, "  Moving to the EVENTS extension.")
        ierr = 0
        call FTMNHD(eventfiles(ii)%info%unitNo, BINARY_TBL, "EVENTS", 0, ierr)
        if(ierr==301) then
          call eventfiles(ii)%info%getFileName(fileName)
call wrstop(functionName, "EVENTS extension not found in file "//trim(fileName))
        else
          call report_FITS_err(functionName, ierr, s_err, "- extension EVENTS")
        end if

        call wrchatty(functionName, "  Reading the rawx/y selection information.")
        call eventfiles(ii)%read_rawxy_sel(status)
        call wrchatty(functionName, "  Closing the file.")
        call eventfiles(ii)%close(status)
!*** destroy eventfiles(ii) extensions, to clear memory?
        if(status /= ok) then
          call wrerr(functionName, s_err, "ERROR with file: " // trim(task%infiles(ii)))
        end if
      end do
    end if ! memset

    call merge_ccdnrs_eventfiles(eventfiles, ccdnrs, status)
    call eventfile_out%init(ccdnrs, status)
    deallocate(ccdnrs)
    call eventfile_out%setup_for_readwrite([TCOL_WRITE_BIT], status) ! This initializes and sets useage for all extensions.

    if(task%fsp%do_filter_gti .and. task%fsp%gti_filter_type==externalgti_type)&
      call task%fsp%externalgti%header%setKeywordValue("MJDREF",eventfile_out%std_header%mjdref)

    if(size(eventfiles) > 1) call wrlog(functionName, "Merging infiles.")
    call compare_stdheader_eventfile(eventfiles, status)
    call wrlog(functionName, "Merging headers.")
    call merge_stdheader_eventfile(eventfiles, eventfile_out, status)
    call wrlog(functionName, "Update filter settings.")
    call update_filter_settings_eventfile(task, eventfile_out, status)

    call wrlog(functionName, "Merging extensions.")
    call merge_extensions(task%memset, eventfiles, eventfile_out, status, repair_gtis=task%repair_gtis) ! also destroys each of the input extensions after use. *** sorts events in order "TIME","RAWX","RAWY","PHA ".

    call dealloc_eventfiles(eventfiles, status)

    call wrlog(functionName, "Applying filters.")
    if(task%filter_image == true) then
      call eventfile_out%apply_filter(task%fsp, status, task%im) !**** sorts
    else
      call eventfile_out%apply_filter(task%fsp, status) !**** sorts
    end if

    ! if the events extension should not be written set it here
    ! since events might be used before to create images and such
    if(task%write_events) then
      eventfile_out%write_events = .true.
    else
      eventfile_out%write_events = .false.
    end if

    if(.not.eventfile_out%events%header%keywordContainsData("T_OFFCOR"))&
      call eventfile_out%events%header%setKeywordValue("T_OFFCOR","F")

    ! create outfile
    call wrlog(functionName, "Create outfile.")
    call eventfile_out%open(FF_CREATE, status, filename=task%outfile, clobber=(task%delete_outfile>0))

    ! use temporary external skyfield
    if(task%external_skyfield /= "") eventfile_out%std_header%skyfield = trim(task%external_skyfield)

    call wrlog(functionName, "Writing outfile.")
    call eventfile_out%write(status)
    !  call eventfile_out%free(status)

    call wrlog(functionName, "Close outfile")
    call eventfile_out%close(status)

!*** 2019-09-11 IMS: eventfile_out never seems to be freed?

    call task%free(status)

    call cpu_time(cpuTimeFinis)
    write(msg,'(f10.2,a)') cpuTimeFinis-cpuTimeStart, ' s CPU run time.'
    call wrlog(functionName, msg)

    call cpast (status, ierr)

end program evtool_main
