! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! Non-method subroutines contained:
!  subroutine extractRawXYBounds(rawxy_str, rawxy_sel, status)
!  subroutine validateEnergyInt(emin, emax, status)
!  subroutine interpret_str_gti(str, gti, status)
!  subroutine update_filter_settings_eventfile(task, eventFile_out, status)
!  subroutine update_filter_gti_settings(Evtool, eventFile, status)
!  subroutine update_filter_telid_settings(Evtool, EventFile, status)
!  subroutine update_filter_energy_settings(Evtool, EventFile, status)
!  subroutine update_filter_flag_settings(Evtool, EventFile, status)
!  subroutine update_filter_pattern_settings(Evtool, EventFile, status)
!  subroutine update_filter_rawxy_settings(Evtool, EventFile, status)

module evtool_mod

  use erolib_general_mod
  use erolib_str_mod
  use erolib_ffits_mod
  use erolib_region_mod
  use erolib_file_mod
  use erolib_extensions_mod
  use erolib_gtitab_mod
  use erolib_rawxy_mod
  use erolib_events_mod
  use erolib_eventfile_mod
  use erolib_image2_mod
  use time_mod
  use param

  implicit none

logical, private, parameter :: doTest=.true.

  private :: extractRawXYBounds,validateEnergyInt,interpret_str_gti
  private :: update_filter_gti_settings,update_filter_telid_settings
  private :: update_filter_energy_settings,update_filter_flag_settings
  private :: update_filter_pattern_settings,update_filter_rawxy_settings

  type tEvtool
    ! Command-line parameters:

    character(buffsize), dimension(:), allocatable :: infiles
    character(buffsize) :: outfile = ""
    logical :: clobber = .false.
    logical :: write_events = .false.
    logical :: write_image = .false.
    character(buffsize) :: image_dimensions_str = ""
    character(buffsize) :: image_centre_str = ""
    ! --rebin => this%im%pixel_size_XY
    real(dp), dimension(:), allocatable :: emin, emax
    integer, dimension(:), allocatable :: ccdnr
    character(buffsize) :: region = ""
    character(buffsize) :: gti_str = ""
    character(buffsize) :: flag = ""
    ! --flag_invert => this%fsp%flag_invert
    integer :: pattern = 0
    character(buffsize) :: rawxy_str = ""
    ! --rawxy_telid => this%fsp%rawxy_sel%ccdnr
    ! --rawxy_invert => this%fsp%rawxy_sel%invert
    logical :: memset = .false. ! the memory setting to read in all files at once.
    real(r8) :: overlap = 0.0
    character(buffsize) :: external_skyfield = "" ! temporary?
    logical :: repair_gtis = .false.

    ! The remainder are internal attributes rather than command-line parameters.

    type(t_filter) :: fsp ! type defined in erolib_eventfile_mod.f90.
    type(t_image_specs) :: im
    integer :: nfile = 0
    integer :: delete_outfile = 0
    integer :: nenergy = 0
    integer :: ninst = 0
    character(buffsize), dimension(:), allocatable :: instrum
!    integer :: flag_hex = 0
    integer :: filter_image = 0
    logical :: write_externalgti = .false.

  contains
    procedure :: read_params => read_params_tEvtool
    procedure :: print_params => print_params_tEvtool
    procedure :: validate => validate_tEvtool
    procedure :: free => free_tEvtool
    procedure :: init => init_tEvtool
  end type tEvtool

  character(*), parameter :: gti_ext = "GTI"
  character(*), parameter :: qualgti_ext = "QUALGTI"
  character(*), parameter :: flaregti_ext = "FLAREGTI"
  character(*), parameter :: external_gti_ext = "STDGTI"
  integer, parameter :: canonical_rebin=80

contains

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> init the Evtool type structure with preset of allocated arrays
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine init_tEvtool(this, status, nfile, ninst, nenergy)
        implicit none

        class(tEvtool) :: this
        integer, intent(inout) :: status
        integer, intent(in), optional :: nfile, ninst, nenergy

        character(*), parameter :: functionName = "init_tEvtool"
        integer :: local_ninst

        if(status > 100 .or. status == s_errRep)&
    return

        if(present(nfile)) then
            this%nfile = nfile
        else
            this%nfile = 100
        end if

        if(present(ninst)) then
            local_ninst = ninst
        else
            local_ninst = 7
        end if

        if(present(nfile)) then
            this%nenergy = nenergy
        else
            this%nenergy = 9
        end if

        call alloc(this%infiles, this%nfile, status)
        call alloc(this%emin, this%nenergy, status)
        call alloc(this%emax, this%nenergy, status)
        call alloc(this%ccdnr, local_ninst, status)
!        call alloc(this%rebin, 2, status)

    end subroutine init_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> free all allocated arrays in the Evtool type
    !
    !
    !
    !> @param[in] this
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine free_tEvtool(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tEvtool) :: this

        character(*), parameter :: function = "free_tEvtool"

        if(status > 100 .or. status == s_errRep)&
    return

!        call dealloc(this%rebin, status)
        call dealloc(this%instrum, status)
        call dealloc(this%ccdnr, status)
        call dealloc(this%emin, status)
        call dealloc(this%emax, status)
        call dealloc(this%infiles, status)

        call this%fsp%rawxy_sel%mask%free(status)

    end subroutine free_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> read the parameters of the Evtool
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine read_params_tEvtool(this, status)
        implicit none

        class(tEvtool) :: this
        integer, intent(inout) :: status

        character(*), parameter :: functionName = "read_params_tEvtool"
        integer :: icount,ierr
        integer :: ninfiles = 100
        character(buffsize) :: filelist_file
!        integer, dimension(4) :: tempRawXYSel
!        integer :: ii

        if(status > 100 .or. status == s_errRep)&
    return

        call rgeta ("eventfiles", ninfiles, this%infiles, icount, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading infile parameter")
        this%infiles = this%infiles(:icount)

        ! check whether infile is actually a file containing the infiles
        if(icount == 1) then
            if(this%infiles(1)(1:1) == "@") then
                filelist_file = this%infiles(1)(2:)
                call read_filelist(filelist_file, this%infiles, status)
            end if
        end if

        call rget ("outfile", this%outfile, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading outfile parameter")

        call rget ("clobber", this%clobber, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading clobber parameter")

        call rget("events", this%write_events, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading events parameter")

        call rget("image", this%write_image, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading image parameter")

        call rget("size", this%image_dimensions_str, ierr)

        call rget("center_position", this%image_centre_str, ierr)

        call rgeta("rebin", 2, this%im%pixel_size_XY, icount, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading image rebin parameter")
        if(icount == 1) this%im%pixel_size_XY(2) = this%im%pixel_size_XY(1)
        if(icount == 0) this%im%pixel_size_XY = this%im%pixel_size_XY(:icount)

        ! max number of nenergy will be set to 9 due to std_header settings
        ! should be updated if std_header will be updated
        !    call rgeti ("nenergy", this%nenergy, ierr)
        !    if(ierr /= 0) call wrstop(functionName, "reading nenergy parameter")

        call rgeta ("emin", this%nenergy, this%emin, icount, ierr)
        if(ierr == 1) ierr = 0
        if(ierr /= 0) call wrstop(functionName, "reading emin parameter")
        this%emin = this%emin(:icount) ! changed conversion to be done only in the events extension conversion
        !    this%emin = this%emin(:icount) * 1000.0_r8 ! keV to eV conversion

        call rgeta ("emax", this%nenergy, this%emax, icount, ierr)
        if(ierr == 1) ierr = 0
        if(ierr /= 0) call wrstop(functionName, "reading emax parameter")
        this%emax = this%emax(:icount) ! changed conversion to be done only in the events extension conversion
        !    this%emax = this%emax(:icount) * 1000.0_r8 ! keV to eV conversion

        ! update the nenergy keyword
        this%nenergy = icount

        call rgeta("telid", 7, this%ccdnr, icount, ierr)
        if(ierr == 1) ierr = 0
        if(ierr /= 0) call wrstop(functionName, "reading telid parameter")
        this%ccdnr = this%ccdnr(:icount)

        call rget ("region", this%region, ierr)
        if(ierr == 1) then
            this%region = ""
        else if (ierr /= 0) then
            call wrstop(functionName, "reading region parameter")
        end if

        call rget ("gti", this%gti_str, ierr)
        if(ierr == 1) then
            this%gti_str = ""
        else if (ierr /= 0) then
            call wrstop(functionName, "reading gti parameter")
        end if

        call rget ("flag", this%flag, ierr)
        if(ierr == 1) then
            this%flag = ""
        else if (ierr /= 0) then
            call wrstop(functionName, "reading flag parameter")
        end if

        call rget ("flag_invert", this%fsp%flag_invert, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading flag_invert parameter")

        call rget ("pattern", this%pattern, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading pattern parameter")

        call rget ("rawxy", this%rawxy_str, ierr)
        if(ierr == 1) then
            this%rawxy_str = ""
        else if(ierr /= 0) then
          call wrstop(functionName, "reading rawxy parameter")
        end if

        this%fsp%rawxy_sel%ccdnr_defined = .true. ! default
        call rget ("rawxy_telid", this%fsp%rawxy_sel%ccdnr, ierr)
        if(this%fsp%rawxy_sel%ccdnr < 1 .or. this%fsp%rawxy_sel%ccdnr > N_TELESCOPES) then
            this%fsp%rawxy_sel%ccdnr = 0
            this%fsp%rawxy_sel%ccdnr_defined = .false.
        else if (ierr /= 0) then
            call wrstop(functionName, "reading rawxy_ccdnr parameter")
        end if

        call rget ("rawxy_invert", this%fsp%rawxy_sel%invert, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading rawxy_invert parameter")

        call rget ("memset", this%memset, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading memset parameter")

        call rget ("overlap", this%overlap, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading overlap parameter")

        call rget ("skyfield", this%external_skyfield, ierr)
        if(ierr == 1) then
            this%external_skyfield = ""
        else if (ierr /= 0) then
            call wrstop(functionName, "reading skyfield parameter")
        end if

        call rget ("repair_gtis", this%repair_gtis, ierr)
        if(ierr /= 0) call wrstop(functionName, "reading repair_gtis parameter")

        ! adjust all read parameters
        this%infiles = adjustl(this%infiles)
        this%outfile = adjustl(this%outfile)
        this%region  = adjustl(this%region)
        this%gti_str = adjustl(this%gti_str)
        this%rawxy_str = adjustl(this%rawxy_str)
!*** check other str params.

    end subroutine read_params_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> print the parameters of the Evtool
    !
    !
    !
    !> @param[in] this
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine print_params_tEvtool(this, status)
        implicit none

        integer, intent(inout) :: status
        class(tEvtool) :: this
        integer :: ii

        character(*), parameter :: function = "print_params_tEvtool"
        character(buffsize) :: entryStr, ccdnrStr, invertStr

        if(status > 100 .or. status == s_errRep)&
    return

        call wrlog(function, "printing parameters")
        print *, "EVENTFILES: - @FILELISTS are already read in"
        do ii = 1, size(this%infiles)
            print *, "*" // trim(this%infiles(ii)) // "*"
        end do

        print *, "OUTFILE:"
        print *, "*" // trim(this%outfile) // "*"

        print *, "CLOBBER:"
        print * , this%clobber

        print *, "WRITE EVENTS:"
        print * , this%write_events

        print *, "WRITE IMAGE:"
        print * , this%write_image

        print *, "RESOLUTION"
        print *, this%im%dimensions

        print *, "CENTER POSITION"
        print *, this%im%centre_XY

        print *, "REBIN"
        print *, this%im%pixel_size_XY

        print *, "EMIN"
        print *, this%emin

        print *, "EMAX"
        print *, this%emax

        print *, "TELID"
        print *, this%ccdnr

        print *, "REGION"
        print *, "*" // trim(this%region) // "*"

        print *, "GTI"
        print *, "*" // trim(this%gti_str) // "*"

        print *, "FLAG"
        print *, "*" // trim(this%flag) // "*"

        print *, "PATTERN"
        print *, this%pattern

        print *, "RAWXY"
        call this%fsp%rawxy_sel%get_entry_strs(entryStr, ccdnrStr, invertStr)
        print trim(entryStr)

        print *, "RAWXY_TELID"
        print *, trim(ccdnrStr)

        print *, "RAWXY_INVERT"
        print *, trim(invertStr)

    end subroutine print_params_tEvtool

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> validate the input parameters of Evtool
    !
    !
    !
    !> @param[in] this
    !> @param[out]
    !> @param[inout] status
    !---------------------------------------------------------------------------

  subroutine validate_tEvtool(this, status)
    implicit none

    class(tEvtool) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "validate_tEvtool"
    integer :: exists, ii
    type(tFitsFile) :: external_GTI_file
    character(buffsize) :: msg

    if(status > 100 .or. status == s_errRep)&
  return

    do ii = 1, size(this%infiles)
            if(trim(this%infiles(ii)) == "") call wrerr(functionName, s_err, "one of the infiles is blank")

            call ffits_file_exists(this%infiles(ii), exists, status)
            if (exists == 0) then
                call wrerr(functionName, s_err, "infile " // trim(this%infiles(ii)) // " does not exist")
                status = s_errRep
  return
            end if
    end do

    if(trim(this%outfile) /= "") then
            this%delete_outfile = false
            call ffits_file_exists(this%outfile, exists, status)
            if (exists /= 0) then
                if(this%clobber .eqv. .false.) then
                    call wrerr(functionName, s_err, "outfile " // trim(this%outfile) // " already exists")
                    status = s_errRep
  return
                else
                    this%delete_outfile = true
                end if
            end if
    else
            call wrerr(functionName, s_err, "outfile is blank")
    end if

    this%filter_image = false
    if(this%write_image) then
          call extract_pair_dbl(this%image_centre_str, this%im%auto_centre, this%im%centre_XY)
          call extract_pair_int(this%image_dimensions_str, this%im%auto_dimension, this%im%dimensions)

          if(.not.this%im%auto_dimension .and.(this%im%dimensions(1) < 1 .or. this%im%dimensions(2) < 1)) then
            call wrerr(functionName, s_err, "zero or negative resolution selected")
            status = s_errRep
  return
          end if

          if(this%im%pixel_size_XY(1) < 1 .or. this%im%pixel_size_XY(2) < 1) then
            call wrerr(functionName, s_err, "zero or negative rebin selected")
            status = s_errRep
  return
          end if

          if(this%im%pixel_size_XY(1) /= canonical_rebin .or. this%im%pixel_size_XY(2) /= canonical_rebin) then
            msg = int2str(canonical_rebin)
            call wrerr(functionName, warn, "If pixel size (parameter --rebin) is not left at "&
              //trim(msg)//", reliable source detection is not guaranteed.")
          end if

          this%filter_image = true
    end if

    if(this%nenergy < 0 .and. this%nenergy > 9) then
            call wrerr(functionName, s_err, "max number of energy intervals"&
              //" cannot be less than zero or more than 9")
            status = s_errRep
  return
    end if

    if(size(this%emin) /= size(this%emax)) then
            call wrerr(functionName, s_err, "number of given energy boundaries"&
              //" (emin / emax) is inconsistent")
            status = s_errRep
    end if

    if(size(this%emin) > this%nenergy .or. size(this%emax) > this%nenergy) then
            call wrerr(functionName, s_err, "max number of energy intervals is"&
              //" lower than the max number of energy bands provided")
            status = s_errRep
    end if

    call validateEnergyInt(this%emin, this%emax, status)

!        this%filter_energy = false
    if(size(this%emin) > 0) this%fsp%do_filter_energy = .true.

        ! FIX for HB evtool call

        if(size(this%ccdnr) > 0) then
            this%fsp%do_filter_telid = .true.

            if(minval(this%ccdnr) < 1) then
                call wrerr(functionName, s_err, "telescope id is illegal, cannot be less than 1")
                status = s_errRep
            end if

            if(maxval(this%ccdnr) > 7) then
                call wrerr(functionName, s_err, "telescope id is illegal, cannot be greater than 7")
                status = s_errRep
            end if

        else if (size(this%ccdnr) > 7) then
            call wrerr(functionName, s_err, "more telescope ids selected than max number of instruments (7)")
            status = s_errRep
        else
            this%fsp%do_filter_telid = .false.
        end if

        if(trim(this%region) /= "") then
            call getRegionsFromStr(this%region, this%fsp%regionObj, status)
            if(status == ok) then
              if(all(this%fsp%regionObj%regRows(:)%wcs == "fk5")) then
                this%fsp%do_filter_region = .true.
              else
                call wrerr(functionName, warn, "Only FK5 regions can yet be processed, sorry!")
              end if
            end if
        else
            call alloc_regRows(this%fsp%regionObj%regRows, 0, status)
        end if

        if(trim(this%gti_str) /= "") then
            ! determine the input format of the gti
            ! type is per telescope
            if(this%gti_str == gti_ext) then
                this%fsp%gti_filter_type = gti_type
            else if(this%gti_str == qualgti_ext) then
                this%fsp%gti_filter_type = qualgti_type
            ! type is global
            else if(this%gti_str == flaregti_ext) then
                this%fsp%gti_filter_type = flaregti_type
            else
                this%fsp%gti_filter_type = externalgti_type
                ! assume type is a FITS file
                call ffits_file_exists(this%gti_str, exists, status)
                if (exists == 1) then
                  call external_GTI_file%info%setFileName(this%gti_str)
                  call this%fsp%externalgti%init(iEXTERNAL, status, external_GTI_file%info)
                  this%fsp%externalgti%info%tableName = this%fsp%externalgti%extNameRoot
                  call this%fsp%externalgti%set_all_useage(0, 1, 0, 0)
                  call external_GTI_file%open(readonly, status)
                  call this%fsp%externalgti%read(status)
                  call this%fsp%externalgti%sanityCheck(status, "External GTI:")
                  if(.not. this%fsp%externalgti%containsData()) then
                    call wrerr(functionName, s_err, external_gti_ext&
                      //" in file "//trim(this%gti_str)//" could not be read")
                      status = s_errRep
    return
                  end if
                  call external_GTI_file%close(status)
                else
                    ! interpret string as two values indicating tstart and tstop
                    call interpret_str_gti(this%gti_str, this%fsp%externalgti, status)

                end if

            end if
            if(status == ok) this%fsp%do_filter_gti = .true.
        end if

        if(trim(this%flag) /= "") then
!          this%flag_hex = str2hex(this%flag, status)
          this%fsp%flag_mask = str2hex(this%flag, status)
          if(status == ok) this%fsp%do_filter_flag = .true.
        end if

        ! single = 1, double = 2, triple = 4, quadruple = 8
        if(this%pattern > 0 .and. this%pattern < 16) then
          this%fsp%do_filter_pattern = .true.
        end if

        if(trim(this%rawxy_str) /= "") then
          ! Determine which of the two acceptable strings it is - either (mask) file name or 'xLo xHi yLo yHi'.
          !
          call ffits_file_exists(this%rawxy_str, exists, status)
          if (exists == 1) then
            this%fsp%rawxy_sel%mask%filename = this%rawxy_str
            call this%fsp%rawxy_sel%mask%open(readonly, status)
            call this%fsp%rawxy_sel%mask%read(status)
            call this%fsp%rawxy_sel%mask%close(status)
            this%fsp%rawxy_sel%mask%containsData = .true.
          else
            call extractRawXYBounds(this%rawxy_str, this%fsp%rawxy_sel, status)
          end if

          if(status == ok) then
            ! Added this just so the pipeline call to evtool doesn't automatically end up in a RAWXY extension being stored.
            !
            if(  this%fsp%rawxy_sel%axes(1)%lo%value > 1&
            .or. this%fsp%rawxy_sel%axes(2)%lo%value > 1&
            .or. this%fsp%rawxy_sel%axes(1)%hi%value < N_PIXELS_CCD&
            .or. this%fsp%rawxy_sel%axes(2)%hi%value < N_PIXELS_CCD)&
              this%fsp%do_filter_rawxy = .true.
          end if
        end if

        if(this%fsp%rawxy_sel%ccdnr_defined) then
          if(this%fsp%rawxy_sel%ccdnr < 1 .or. this%fsp%rawxy_sel%ccdnr > N_TELESCOPES) then
            call wrerr(functionName, s_err, "RAWXY ccd number is out of range.")
            status = s_errRep
    return
          end if
        end if

  end subroutine validate_tEvtool

  !=====================================================================
  ! Utility routines, not methods of tEvtool.

    !---------------------------------------------------------------------------
    !> @author
    !> Ian M Stewart
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> blah
    !
    !
    !
    !> @param[in] emin
    !> @param[out] emax
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine extractRawXYBounds(rawxy_str, rawxy_sel, status)
        implicit none

        character(*),             intent(in)    :: rawxy_str
        type(tRawXYSelection),    intent(inout) :: rawxy_sel
        integer,                  intent(inout) :: status

        integer :: ii
        character(buffsize), allocatable, dimension(:) :: listOfStrs
        integer :: icount,readstat

        character(*), parameter :: function = "extractRawXYBounds"

        if(status > 100 .or. status == s_errRep)&
    return

        ! Divide the rawxy string into its space-separated substrings:
        !
        call strtok(rawxy_str, ' ', listOfStrs)
        icount = size(listOfStrs)

        if(icount == 4) then
          ! Read the 4 values into the relevant struct elements:
          !
          ii = 1
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(1)%lo%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          ii = 2
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(1)%hi%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          ii = 3
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(2)%lo%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          ii = 4
          read(listOfStrs(ii),*,iostat=readstat) rawxy_sel%axes(2)%hi%value
          if (readstat>0) then
            call wrerr(function, s_err, "Cannot read " // trim(listOfStrs(ii)))
          status = s_errRep
    return
          end if

          do ii = 1, 2
            if(rawxy_sel%axes(ii)%lo%value < 1)&
              rawxy_sel%axes(ii)%lo%value = 1
            if(rawxy_sel%axes(ii)%hi%value > N_PIXELS_CCD)&
              rawxy_sel%axes(ii)%hi%value = N_PIXELS_CCD

            rawxy_sel%axes(ii)%lo%is_defined = .true.
            rawxy_sel%axes(ii)%hi%is_defined = .true.
          end do

          ! Fill the mask.
          !
          allocate(rawxy_sel%mask%values(N_PIXELS_CCD,N_PIXELS_CCD))
          rawxy_sel%mask%values = .false.
          rawxy_sel%mask%values(rawxy_sel%axes(1)%lo%value:rawxy_sel%axes(1)%hi%value&
                           ,rawxy_sel%axes(2)%lo%value:rawxy_sel%axes(2)%hi%value) = .true.

        else
            call wrerr(function, warn, "Invalid rawxy argument "//trim(rawxy_str)&
              //". It is neither the name of an existing file or a list of 4 substrings.")
            do ii = 1, 2
              rawxy_sel%axes(ii)%lo%is_defined = .false.
              rawxy_sel%axes(ii)%lo%value = 0
              rawxy_sel%axes(ii)%hi%is_defined = .false.
              rawxy_sel%axes(ii)%hi%value = 0
            end do
        end if

        call dealloc(listOfStrs, status)

        do ii=1, 2
          if(rawxy_sel%axes(ii)%lo%is_defined .and. rawxy_sel%axes(ii)%hi%is_defined) then
            if(rawxy_sel%axes(ii)%lo%value > rawxy_sel%axes(ii)%hi%value) then
              call wrerr(function, s_err, "RAWXY filter: lo may not be greater than hi.")
              status = s_errRep
    return
            end if
          end if
        end do

    end subroutine extractRawXYBounds

  !.....................................................................
  subroutine extract_pair_dbl(image_centre_str, auto_centre_image, center_position)
    implicit none

    character(*),           intent(in)  :: image_centre_str
    logical,                intent(out) :: auto_centre_image
    real(r8), dimension(2), intent(out) :: center_position

    character(*), parameter :: functionName = "extract_pair_dbl"
    integer :: ii,ierr
    character(buffsize), allocatable, dimension(:) :: listOfStrs
    integer :: icount,readstat


    if(trim(image_centre_str)=='auto' .or. trim(image_centre_str)=='AUTO') then
      auto_centre_image = .true.
      center_position = 0.0d0

    else
      auto_centre_image = .false.

      ! Divide the string into its space-separated substrings:
      !
      call strtok(image_centre_str, ' ', listOfStrs)
      icount = size(listOfStrs)

      if(icount == 2) then
        ! Read the 2 values into the relevant struct elements:
        !
        do ii = 1,2
          read(listOfStrs(ii),*,iostat=readstat) center_position(ii)
          if (readstat>0) then
            call wrerr(functionName, s_err, "Cannot read " // trim(listOfStrs(ii)))
    return
          end if
        end do

      else if(icount == 1) then
        ii = 1
        read(listOfStrs(ii),*,iostat=readstat) center_position(ii)
        if (readstat>0) then
          call wrerr(functionName, s_err, "Cannot read " // trim(listOfStrs(ii)))
    return
        end if

        center_position(2) = center_position(1)

      else
        call wrerr(functionName, s_err, "Invalid 'center_position' parameter "&
          //trim(image_centre_str)//". It must either equal 'auto' or be a list of 2 numbers.")
      end if

      call dealloc(listOfStrs, ierr)
    end if

  end subroutine extract_pair_dbl

  !.....................................................................
  subroutine extract_pair_int(image_dimensions_str, auto_dimension_image, resolution)
    implicit none

    character(*),          intent(in)  :: image_dimensions_str
    logical,               intent(out) :: auto_dimension_image
    integer, dimension(2), intent(out) :: resolution

    character(*), parameter :: functionName = "extract_pair_int"
    integer :: ii,ierr
    character(buffsize), allocatable, dimension(:) :: listOfStrs
    integer :: icount,readstat


    if(trim(image_dimensions_str)=='auto' .or. trim(image_dimensions_str)=='AUTO') then
      auto_dimension_image = .true.
      resolution = 0

    else
      auto_dimension_image = .false.

      ! Divide the string into its space-separated substrings:
      !
      call strtok(image_dimensions_str, ' ', listOfStrs)
      icount = size(listOfStrs)

      if(icount == 2) then
        ! Read the 2 values into the relevant struct elements:
        !
        do ii = 1,2
          read(listOfStrs(ii),*,iostat=readstat) resolution(ii)
          if (readstat>0) then
            call wrerr(functionName, s_err, "Cannot read " // trim(listOfStrs(ii)))
    return
          end if
        end do

      else if(icount == 1) then
        ii = 1
        read(listOfStrs(ii),*,iostat=readstat) resolution(ii)
        if (readstat>0) then
          call wrerr(functionName, s_err, "Cannot read " // trim(listOfStrs(ii)))
    return
        end if

        resolution(2) = resolution(1)

      else
        call wrerr(functionName, s_err, "Invalid 'size' parameter "&
          //trim(image_dimensions_str)//". It must either equal 'auto' or be a list of 2 numbers.")
      end if

      call dealloc(listOfStrs, ierr)
    end if

  end subroutine extract_pair_int

    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> validate energy intervalls
    !
    !
    !
    !> @param[in] emin
    !> @param[out] emax
    !> @param[inout] status
    !---------------------------------------------------------------------------

    subroutine validateEnergyInt(emin, emax, status)
        implicit none

        integer, intent(inout) :: status
        integer :: ii

        real(dp), dimension(:), intent(in) :: emin, emax

        character(*), parameter :: function = "validateEnergyInt"

        if(status > 100 .or. status == s_errRep)&
    return

        if(size(emin) /= size(emax)) then
            call wrerr(function, s_err, "number of min/max energy entries is different")
            status = s_errRep
    return
        end if

        do ii = 1, size(emin)
            if(emin(ii) > emax(ii)) then
                call wrerr(function, s_err, "selected min energy > max energy")
                status = s_errRep
    return
            end if
            if(emin(ii) < 0.0_dp .or. emax(ii) < 0.0_dp) then
                call wrerr(function, s_err, "selected min and/or max energy is smaller than 0 eV")
                status = s_errRep
    return
            end if
        end do

    end subroutine validateEnergyInt


    !---------------------------------------------------------------------------
    !> @author
    !> Christoph Grossberger
    !
    ! DESCRIPTION:
    !> Brief description of routine.
    !> @brief
    !> interprete string as gti
    !
    !
    !
    !> @param[in] file
    !> @param[out]
    !> @param[inout] status, filelist
    !---------------------------------------------------------------------------

  subroutine interpret_str_gti(str, gti, status)
    implicit none

    character(*), intent(in)    :: str
    type(tGTI_tab),   intent(out)   :: gti
    integer,      intent(inout) :: status

    character(*), parameter :: functionName = "interpret_str_gti"
    character(buffsize), dimension(:), allocatable :: gti_str
    real(r8), dimension(:), pointer :: starts=>null(),stops=>null()

    if(status > 100 .or. status == s_errRep)&
  return

    call strtok(str, " ", gti_str)
    if(size(gti_str) /= 2) then
      call wrerr(functionName, s_err, "could not interprete " // trim(str) // " as external gti")
      status = s_errRep
  return
    end if

    call gti%init(iTemp, status)
    call gti%set_all_useage(0, 0, 0, 1) ! set for internal useage.
    call gti%alloc(status, 1)
    call gti%get_col_ptr("START", starts)
    call gti%get_col_ptr("STOP",  stops)
    starts(1) = str2double(gti_str(1), status)
    stops( 1) = str2double(gti_str(2), status)
    call gti%header%setKeywordValue("TSTART",starts(1))
    call gti%header%setKeywordValue("TSTOP", stops( 1))
!        ! when set to negative mjdref, the mjdref value is ignored when gti is applied
!        gti%use_mjdref = .false.
!        gti%use_timezero = .false.
    call gti%sanityCheck(status, "User input GTI:")

    nullify(stops)
    nullify(starts)

  end subroutine interpret_str_gti

  !.....................................................................
  subroutine update_filter_settings_eventfile(task, eventFile_out, status)
    implicit none

    type(tEvtool),    intent(in)    :: task
    type(tEventFile), intent(inout) :: eventFile_out
    integer,          intent(inout) :: status

    character(*), parameter :: function = "update_filter_settings_eventfile"

    if(status > 100 .or. status == s_errRep)&
  return

    call update_filter_telid_settings(  task, eventFile_out, status)
    call update_filter_energy_settings( task, eventFile_out, status)
    call update_filter_flag_settings(   task, eventFile_out, status)
    call update_filter_pattern_settings(task, eventFile_out, status)
    call update_filter_rawxy_settings(  task, eventFile_out, status)

    !        call update_filter_region_settings(task, eventFile, status)
    call update_filter_gti_settings(    task, eventFile_out, status)

  end subroutine update_filter_settings_eventfile

  !.....................................................................
  subroutine update_filter_gti_settings(task, eventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: task
    type(tEventFile), intent(inout) :: eventFile
    integer,          intent(inout) :: status

    character(*), parameter :: function = "update_filter_gti_settings"

    if(status > 100 .or. status == s_errRep)&
  return

    if(task%fsp%do_filter_gti .and. .not. eventFile%filtered_gti) then
      eventFile%filtered_gti = .true.
      if(task%fsp%gti_filter_type == gti_type)      eventFile%std_header%gti_sel = gti_ext
      if(task%fsp%gti_filter_type == qualgti_type)  eventFile%std_header%gti_sel = qualgti_ext
      if(task%fsp%gti_filter_type == flaregti_type) eventFile%std_header%gti_sel = flaregti_ext
      if(task%fsp%gti_filter_type == externalgti_type) then
        eventFile%std_header%gti_sel = external_gti_ext
      end if
    end if

  end subroutine update_filter_gti_settings

  !.....................................................................
  subroutine update_filter_telid_settings(task, EventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: task
    type(tEventFile), intent(inout) :: EventFile
    integer,          intent(inout) :: status

    integer, dimension(:), allocatable :: intersect_array, complement_array
    integer :: ii
    character(*), parameter :: function = "update_filter_telid_settings"

    if(status > 100 .or. status == s_errRep)&
  return

    if(task%fsp%do_filter_telid .and. .not. EventFile%filtered_telid) then
      call wrerr(function, s_err, "EventFile out has no input ccdnr, this should never happen")
      status = s_errRep
  return
    end if

    if(task%fsp%do_filter_telid .and. EventFile%filtered_telid) then
      intersect_array = intersect_int_order(EventFile%ccdnr_infile, task%ccdnr, status)
      complement_array = complement_int_order(task%ccdnr, EventFile%ccdnr_infile, status)
      EventFile%ccdnr_infile = EventFile%ccdnr_infile(intersect_array)

      if(size(EventFile%ccdnr_infile) < size(task%ccdnr)) then
        do ii = 1, size(complement_array)
          call wrerr(function, warn, "Selected TELID "&
            //trim(int2str(task%ccdnr(complement_array(ii)), status))&
            //" is not present in the infile/s")
        end do
      end if

      !update stdheader
      EventFile%std_header%ninst = size(EventFile%ccdnr_infile)
      do ii=1, size(EventFile%ccdnr_infile)
        write(EventFile%std_header%instrumx(ii), '(a,i1)') "TM", EventFile%ccdnr_infile(ii)
      end do

      if(EventFile%std_header%ninst == 1) then
        EventFile%std_header%instrume = EventFile%std_header%instrumx(1)
      else
        EventFile%std_header%instrume = "merged"
      end if

      call dealloc(intersect_array, status)
      call dealloc(complement_array, status)

    end if
  end subroutine update_filter_telid_settings

  !.....................................................................
  subroutine update_filter_energy_settings(task, eventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: task
    type(tEventFile), intent(inout) :: eventFile
    integer,          intent(inout) :: status

    real(r8), dimension(:), allocatable :: emin, emax

    character(*), parameter :: function = "update_filter_energy_settings"
    character(buffsize) :: str0,str1

    if(status > 100 .or. status == s_errRep)&
  return

    if(task%fsp%do_filter_energy .and. .not. eventFile%filtered_energy) then
      eventFile%filtered_energy = .true.
      eventFile%std_header%nenergy = task%nenergy
      eventFile%std_header%eminx(:task%nenergy) = task%emin
      eventFile%std_header%emaxx(:task%nenergy) = task%emax
  return
    end if

    if(task%fsp%do_filter_energy .and. eventFile%filtered_energy) then
      call merge_and_r8(task%emin, task%emax, task%nenergy&
        , eventFile%std_header%eminx, eventFile%std_header%emaxx&
        , eventFile%std_header%nenergy, emin, emax, status)

      if(status/=0) then
        write(str0, '(es8.2,a,es8.2)') task%emin(1), "-", task%emax(task%nenergy)
        write(str1, '(es8.2,a,es8.2)') eventFile%std_header%eminx(1)&
          , "-", eventFile%std_header%emaxx(eventFile%std_header%nenergy)
call wrstop(function, "Could not merge new energy selection (lo-hi = "&
  //trim(str0)//") with previous (lo-hi = "//trim(str1)//").")
      end if

      if(size(emin) > 9) then
        call wrerr(function, s_err, "The number of energy intervals after user applied energy settings exceeds the max of 9 energy intervals.")
        status = s_errRep
  return
      end if
      eventFile%std_header%nenergy = size(emin)
      eventFile%std_header%eminx(:size(emin)) = emin
      eventFile%std_header%emaxx(:size(emin)) = emax
      call dealloc(emin, status)
      call dealloc(emax, status)
    end if

  end subroutine update_filter_energy_settings

  !.....................................................................
  subroutine update_filter_flag_settings(task, EventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: task
    type(tEventFile), intent(inout) :: EventFile
    integer,          intent(inout) :: status

!    integer :: flag
    character(*), parameter :: function = "update_filter_flag_settings"

    if(status > 100 .or. status == s_errRep)&
  return

    if(task%fsp%do_filter_flag .and. .not.EventFile%filtered_flag) then
      EventFile%filtered_flag = .true.
!      EventFile%std_header%flagsel = task%flag_hex
      EventFile%std_header%flagsel = task%fsp%flag_mask
      if(task%fsp%flag_invert) then
        EventFile%std_header%c_flagsel = "Flag selection: "//trim(task%flag)//" (inverted)"
      else
        EventFile%std_header%c_flagsel = "Flag selection: "//trim(task%flag)
      end if
  return
    end if

    if(task%fsp%do_filter_flag .and. EventFile%filtered_flag) then
!      flag = ior(EventFile%std_header%flagsel, task%flag_hex)
!      EventFile%std_header%flagsel = flag
      EventFile%std_header%flagsel = ior(EventFile%std_header%flagsel, task%fsp%flag_mask)
    end if

!if(doTest) print *, "xxx update_filter_flag_settings(), task%flag_hex=", task%flag_hex
  end subroutine update_filter_flag_settings

  !.....................................................................
  subroutine update_filter_pattern_settings(task, EventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: task
    type(tEventFile), intent(inout) :: EventFile
    integer,          intent(inout) :: status

    integer :: pattern
    character(*), parameter :: function = "update_filter_pattern_settings"

    if(status > 100 .or. status == s_errRep)&
  return

    if(task%fsp%do_filter_pattern .and. .not. EventFile%filtered_pattern) then
      EventFile%filtered_pattern = .true.
      EventFile%std_header%pat_sel = task%pattern
  return
    end if

    if(task%fsp%do_filter_pattern .and. EventFile%filtered_pattern) then
      pattern = iand(EventFile%std_header%pat_sel, task%pattern)
      EventFile%std_header%pat_sel = pattern
    end if

  end subroutine update_filter_pattern_settings

  !.....................................................................
  subroutine update_filter_rawxy_settings(task, EventFile, status)
    implicit none

    type(tEvtool),    intent(in)    :: task
    type(tEventFile), intent(inout) :: EventFile
    integer,          intent(inout) :: status

    character(*), parameter :: function = "update_filter_rawxy_settings"

    if(status > 100 .or. status == s_errRep)&
  return

    if(task%fsp%do_filter_rawxy .and. .not. EventFile%filtered_rawxy) then
      EventFile%filtered_rawxy = .true.
      call EventFile%record_rawxy_sel(task%fsp%rawxy_sel, status, .true.)
  return
    end if

    if(task%fsp%do_filter_rawxy .and. EventFile%filtered_rawxy) then
      call EventFile%record_rawxy_sel(task%fsp%rawxy_sel, status)
    end if

  end subroutine update_filter_rawxy_settings

  !.....................................................................
  subroutine test_print_extensions(eventFile, msg)
    implicit none
    type(tEventFile), intent(in) :: EventFile
    character(*),     intent(in) :: msg

    integer :: ii,jj,ci

    do ii=1,size(eventFile%extList)
      do jj = 1, size(eventFile%ccdnr_infile)
        ci = eventFile%ccdnr_infile(jj)
        if(eventFile%extList(ii)%extPerTM(ci)%isAllocated()) then
          print *, trim(msg)//trim(eventFile%extList(ii)%extPerTM(ci)%info%tableName)//": allocated."
        else
          print *, trim(msg)//trim(eventFile%extList(ii)%extPerTM(ci)%info%tableName)//": NOT allocated."
        end if
      end do
    end do

  end subroutine test_print_extensions

end module evtool_mod
