! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module expmap_io

  use param
  use ero_general
  use erolib_general_mod ! for report_FITS_err(), BINARY_TBL, buffsize.
  use erolib_str_mod, only: str_upper
  use srctool_generic_avign
  use iso_fortran_env
  use fitsimage, only: fread_shapehdr, fread_stdhdr, fread_wcshdr, fread_refhdr, &
                       fwrite_stdhdr, fwrite_wcshdr, fwrite_refhdr, &
                       shapehdr_type, stdhdr_type, &
                       wcs_imahdr_type, wcs_refhdr_type, update_stdhdr
  use srctool_generic, only: wrschatty, wrslog, wrserr

  use erolib_ranges_mod

  implicit none

logical, private, parameter :: doTest=.false.
logical, private, parameter :: doTest_detmap=.false.

  private
  public :: lininterp_att, linint_deadc, get_detmaps, &
            remove_badpix, get_ima_keywords, &
            get_expmaps, get_area, get_vignetting2, &
            write_cam_expmaps, write_merged_expmaps

public :: quantize_image ! to stop -Wall complaints.

public :: imaout ! 20201204 IMS: just for diagnostic purposes.

!  real(real64), private, parameter :: nullval = 0d0
  real(real64), private, parameter :: min_dE_keV=0.05d0 ! width of the energy intervals in the estimation of the vignetting function.
  integer, private, parameter :: ftread = 0, ftwrite = 1
  logical, private, parameter :: casesens = .false.
  integer, public, parameter :: max_n_bands = 50
  integer, public, parameter :: max_n_exp = max_n_bands*N_TELESCOPES ! buffer size for expmap file names

  type, public :: t_taskpars
    character(buffsize) :: templateimage = ''
    logical             :: withmerged = .false.
    logical             :: withsingle = .false.
    character(buffsize) :: gtitype = ''
    logical             :: withvig = .false.
    logical             :: withinmaps = .false.
    logical             :: withweights = .false.
    logical             :: withdetmaps = .false.
    logical             :: withfilebadpix = .false.
    character(buffsize) :: accuracy = ''
    real                :: plindex=0.0
    character(buffsize), dimension(N_TELESCOPES) :: evfiles
    character(buffsize), dimension(max_n_exp) :: exp_out_files, exp_in_files, merged_out_files
    real, dimension(max_n_bands) :: e_min, e_max
  contains
    procedure :: readme => read_t_taskpars
  end type t_taskpars

  type, public :: tAtt
    real(real64) :: time  = 0.0d0
    real(real64) :: ra    = 0.0d0 ! deg
    real(real64) :: dec   = 0.0d0 ! deg
    real(real64) :: roll  = 0.0d0 ! deg
    real(real64) :: x_pix = 0.0d0 ! Pointing location in
    real(real64) :: y_pix = 0.0d0 ! image pixel coordinates.
    logical :: is_valid=.false.
  end type tAtt

  type, public :: tDeadCorr
    real(real64) :: time
    real :: value
  end type tDeadCorr

  type, public :: gti_t
    real(kind=real64), dimension(:), allocatable :: starts
    real(kind=real64), dimension(:), allocatable :: stops
    integer, private :: ngti=0
  contains
    procedure :: alloc => alloc_gti_t
    procedure :: dealloc => dealloc_gti_t
    procedure :: get_num => get_num_gti_t
    procedure :: or_with => or_with_gti_t
    procedure :: and_with_not => and_with_not_gti_t
  end type gti_t

  type, public :: badpixt
    integer(int16) :: rawx
    integer(int16) :: rawy
    integer(int16) :: yextent
    real(real64) :: tmin
    real(real64) :: tmax
    integer(int16) :: phamin
    integer(int16) :: phamax
    integer(int16) :: type
    integer(int16) :: badflag 
  end type badpixt

  type, public :: t_tm_info
    logical           :: is_present = .false.
    integer           :: flagsel = 0
    integer           :: patsel = 0
    character(len=30) :: instrum = ''
    character(len=19) :: date_obs = ''
    character(len=19) :: date_end = ''
    character(len=30) :: filter = ''
    type(gti_t)   :: gtis ! GTIs read from event list, extension type given by t_taskpars%gtitype.
    type(badpixt), allocatable, dimension(:) :: badpix ! bad pixels, read from BADPIXn
    type(tAtt), dimension(:), allocatable :: atts  ! attitudes, read from CORRATTn
    type(tDeadCorr), dimension(:), allocatable :: deadcs ! deadcorr data, read from DEADCORn
  end type t_tm_info

  type, public :: t_taskinfo
    type(t_taskpars) :: params
    type(t_tm_info), dimension(N_TELESCOPES) :: tm_info
    integer :: n_evfiles = 0
    integer :: n_bands = 0
    integer :: n_exp_out = 0
    integer :: n_exp_in = 0
    integer :: n_TMs = 0
  contains
    procedure :: obtain => obtain_t_taskinfo
  end type t_taskinfo

  ! Bad pixel types:
  enum, bind( C )
    enumerator :: BP_hot=1, BP_flicker, BP_dead
  end enum

  ! 'BADFLAG' meanings:
  enum, bind( C )
    enumerator :: BP_uplinked=1, BP_fromCALDB, BP_fromFile
  end enum

  type, public :: timer_t
    real(real64), private :: start_time=0.0d0
    logical, private :: timer_was_started=.false.
  contains
    procedure :: start => start_timer_t
    procedure :: stop => stop_timer_t
  end type timer_t

  contains

  !.....................................................................
  subroutine read_t_taskpars(this, n_evfiles, n_bands, n_exp_out, n_exp_in)
    implicit none

    class(t_taskpars) :: this
    integer, intent(out) :: n_evfiles, n_bands, n_exp_out, n_exp_in

    character(*), parameter :: functionName = "read_t_taskpars"
    integer :: ierr,temp_n_bands,n_merged,ib

    ierr = 0
    call rgetca("inputdatasets", N_TELESCOPES, this%evfiles, n_evfiles, ierr)

    ! get WCS keywords from first eventfile if no template image is given
    call rgetc("templateimage", this%templateimage, ierr)

    call rgetfa("emin", max_n_bands, this%e_min, n_bands, ierr)
    call rgetfa("emax", max_n_bands, this%e_max, temp_n_bands, ierr)
    if (n_bands /= temp_n_bands)&
call wrstop(functionName, "Upper and lower energy bounds don't match.")

    call rgetl("withmergedmaps", this%withmerged, ierr)
    call rgetl("withsinglemaps", this%withsingle, ierr)
    if (.not. this%withmerged .and. .not. this%withsingle)&
call wrstop(functionName, "No output selected. Set either --withmergedmaps or --withsinglemaps to true.")      

    call rgetl("withvignetting", this%withvig, ierr)
    call rgetl("withinputmaps", this%withinmaps, ierr)
    call rgetl("withweights", this%withweights, ierr)
    call rgetl("withdetmaps", this%withdetmaps, ierr)
    call rgetl("withfilebadpix", this%withfilebadpix, ierr)

    call rgetc("gtitype", this%gtitype, ierr)
    call str_upper(this%gtitype)

    if (trim(this%gtitype)=="QUAL")&
      this%gtitype ="QUALGTI"
    if (trim(this%gtitype)=="FLARE")&
      this%gtitype ="FLAREGTI"

    if (.not.(     trim(this%gtitype)=="QUALGTI"&
              .or. trim(this%gtitype)=="FLAREGTI"&
              .or. trim(this%gtitype)=="GTI"))&
call wrstop(functionName, "Don't recognize --gtitype '"//trim(this%gtitype)//"'.")      

    call rgetc("accuracy", this%accuracy, ierr)
    if(.not. (     trim(this%accuracy)/="original"&
              .or. trim(this%accuracy)/="approximate"&
              .or. trim(this%accuracy)/="exact"))&
call wrstop(functionName, "--accuracy type '"//trim(this%accuracy)//"' not recognized.")

    call rgetf("plindex", this%plindex, ierr)

    if (this%withsingle) then
      ! Single-TM expmap file names must be given in ascending order
      ! of energy band and camera number, starting with the energy
      ! bands of the first camera, followed by the energy bands of
      ! the second camera, etc. 
      call rgetca("singlemaps", max_n_exp, this%exp_out_files, n_exp_out, ierr)
      if (ierr > 0)&
call wrstop(functionName, "Could not read list of --singlemaps.")
    end if

    ! obtain file names for averaged all-eRO exposure maps (one per energy band)
    if (this%withmerged) then
      call rgetca("mergedmaps", max_n_bands, this%merged_out_files, n_merged, ierr)
      if (ierr > 0)&
call wrstop(functionName,"Could not read file name.")
      if (n_merged /= n_bands)&
call wrstop(functionName, "Number of energy bands and file names don't match.")
      do ib=1,n_bands
        if (len_trim(this%merged_out_files(ib)) == 0)&
call wrstop(functionName, "No name for total exposure map given")
      end do
    end if

    if (this%withinmaps) then
      ! Single-TM expmap file names must be given in ascending order
      ! of energy band and camera number, starting with the energy
      ! bands of the first camera, followed by the energy bands of
      ! the second camera, etc. 
      call rgetca("inputmaps", max_n_exp, this%exp_in_files, n_exp_in, ierr)
      if (ierr > 0)&
call wrstop(functionName,"Could not read list of --inputmaps.")
    end if

  end subroutine read_t_taskpars

  !.....................................................................
  subroutine obtain_t_taskinfo(this, status)
    implicit none

    class(t_taskinfo) :: this
    integer, intent(out) :: status

    character(*), parameter :: functionName = "obtain_t_taskinfo"
    integer :: ii,jj,ierr,unit_num,flagsel,patsel,dummy_hdu_type,ngti,ndc
    logical, dimension(N_TELESCOPES) :: tm_is_in_this_file
    character(len=64) :: comment
    character(len=30) :: instrum,gtisel
    character(len=8) :: extnam
    character(1) :: tmStr
    character(4) :: file1Str
    character(len=19) :: date_obs, date_end
    character(len=30) :: filter
    character(len=16) :: istr

    call this%params%readme(this%n_evfiles, this%n_bands, this%n_exp_out, this%n_exp_in)

    do ii = 1, this%n_evfiles
      call stop_if_not_exists(functionName, this%params%evfiles(ii))

      call wrlog(functionName, "Reading keywords and GTI from "//trim(this%params%evfiles(ii)))

      ! Open the event list.
      !
      ierr = 0
      call FTGIOU(unit_num, ierr)
      call fitserr("ftgiou", ierr, status)
      call FTNOPN(unit_num, this%params%evfiles(ii), ftread, ierr)
      call fitserr("ftnopn", ierr, status)

      ! Find out which TMs are in this file. We do this by looking for the respective DEADCORn extension.
      !
      tm_is_in_this_file(:) = .false.
      do jj = 1, N_TELESCOPES
        write (extnam,'(''DEADCOR'', i1)') jj
        ierr = 0
        call FTMNHD(unit_num, BINARY_TBL, extnam, 0, ierr)
        if ( ierr == 0 ) then
          if (this%tm_info(jj)%is_present) then
            write (tmStr, '(i1)') jj
            write (file1Str, '(i4)') ii
call wrstop(functionName, "Extensions for TM"//trim(adjustl(tmStr))&
  //" were already found in an event list previous to "//trim(adjustl(file1Str))&
  //". Duplication of this sort is not permitted.")
          end if

          this%tm_info(jj)%is_present = .true.
          tm_is_in_this_file(jj) = .true.

          write(this%tm_info(jj)%instrum, '(a,i1)') "TM", jj
        end if
      end do

      ! Get important keywords.
      !
      ! Try to move to the PHDU.
      !
      ierr = 0
      call FTMAHD(unit_num, 1, dummy_hdu_type, ierr)
      call report_FITS_err(functionName, ierr, status, "- couldn't move to PHDU.")

      call FTGKYS(unit_num, "DATE-OBS", date_obs, comment, ierr)
      if(ierr/=0) then
        ierr=0 ! IMS: this seems to be necessary.
!        call FTMNHD(unit_num, BINARY_TBL, "EVENTS", 0, ierr)
!        call fitserr("ftmnhd", ierr, status)
        call FTMAHD(unit_num, 2, dummy_hdu_type, ierr) ! move to 2nd HDU.
        call report_FITS_err(functionName, ierr, status, "- couldn't move to HDU 2.")
        call FTGKYS(unit_num, "DATE-OBS", date_obs, comment, ierr)
        call report_FITS_err(functionName, ierr, status, "- couldn't read keyword 'DATE-OBS'.")
      end if

      call FTGKYS(unit_num, "DATE-END", date_end, comment, ierr)
      call fitserr("ftgkys", ierr, status)

      call FTGKYJ(unit_num, "FLAGSEL", flagsel, comment, ierr)
      if (ierr /= 0) then
        call wrlog(functionName,"FLAGSEL keyword could not be read; set to 0.")
        flagsel=0
        ierr=0
      else
        if(btest(flagsel, BIT_OUT_OF_FOV) .and. .not.this%params%withdetmaps) then
          call wrerr(functionName, warn&
            , "Event list excludes out-of-FOV events, but you specified --withdetmaps=no!")
        else if(.not.btest(flagsel, BIT_OUT_OF_FOV) .and. this%params%withdetmaps) then
          call wrerr(functionName, warn&
            , "Event list includes out-of-FOV events, but you specified --withdetmaps=yes!")
        end if
      end if

      call FTGKYS(unit_num, "FILTER", filter, comment, ierr)
      call fitserr("ftgkys", ierr, status)
      if (trim(adjustl(filter)) == "FILTER" .or. trim(adjustl(filter)) == "filter") then
        filter = "FILTER"
      else if(trim(adjustl(filter)) == "OPEN" .or. trim(adjustl(filter)) == "open") then
        filter = "OPEN"
      else
        call wrerr(functionName,2,"Unrecognized value '"//trim(filter)&
          //"' of FILTER keyword. The CALDB will return you files for all "&
          //"filters, which may not be what you want.")
        filter ="-"
      end if

      call FTGKYJ(unit_num, "PAT_SEL", patsel, comment, ierr)
      if (ierr /= 0) then
        call wrlog(functionName,"PAT_SEL keyword could not be read; set to 0.")
        patsel=0
        ierr=0
      end if

      call FTGKYS(unit_num, "GTI_SEL", gtisel, comment, ierr)
      if (ierr /= 0 .or. gtisel=="NONE") then
        call wrerr(functionName, warn&
          , "Event list appears not to have been selected on GTI. Exposure may differ from expmap.")
        ierr=0
      end if

!      call FTGKYS(unit_num,'INSTRUME',instrum,comment,ierr)
!      call fitserr("ftgkys", ierr, status)
!      if (instrum == "merged") then
!        call FTGKYS(unit_num,'INSTRUM1',instrum,comment,ierr)
!        call fitserr("ftgkys", ierr, status)
!        if (ierr > 0) then
!          status = 3
!          call wrerr(functionName, status, "Can't read INSTRUME keyword")
!          ierr = 0
!        end if
!      end if

!      instrum = instrum(1:len_trim(instrum)-1)

      do jj = 1, N_TELESCOPES
        if(tm_is_in_this_file(jj)) then
          this%tm_info(jj)%date_obs = date_obs
          this%tm_info(jj)%date_end = date_end
!*** 20210626 IMS - why are these keywords in the TM-specific section?
          this%tm_info(jj)%flagsel  = flagsel
          this%tm_info(jj)%filter   = filter
          this%tm_info(jj)%patsel   = patsel
!          this%tm_info(jj)%instrum  = instrum
        end if
      end do

      ! Read various extensions of the event list.
      !
      do jj = 1, N_TELESCOPES
        if(.not.tm_is_in_this_file(jj))&
      cycle

        instrum = this%tm_info(jj)%instrum

        if (this%params%withfilebadpix) then
          call wrschatty (functionName, trim(instrum)&
            //": Calling get_badpix...")
          call get_badpix(jj, unit_num, this%tm_info(jj)%badpix, ierr)
          if (ierr == 3)&
call wrstop (functionName, trim(instrum)//": Obtaining badpix list failed")
        end if

        call wrschatty (functionName, trim(instrum)&
          //": Calling get_gti...")
        call get_gti(jj, unit_num, this%params%gtitype, this%tm_info(jj)%gtis, ierr)
        if (ierr == 3)&
call wrstop(functionName, trim(instrum)//": Obtaining gti from input event file failed")

        ngti = this%tm_info(jj)%gtis%get_num()
        write(istr, *) ngti
        call wrslog(functionName, trim(instrum)&
          //": Number of GTIs this TM: "//adjustl(trim(istr)))

        if(ngti <= 0) then ! We will just make a blank map for this TM. We don't want to exit under these circumstances.
          call wrerr(functionName, warn, trim(instrum)&
          //": No GTIs found. The maps for this TM will be empty.")

          allocate(this%tm_info(jj)%atts(0)) 

        else
          call wrschatty (functionName,  trim(instrum)&
              //": Calling get_attitude...")
          call get_attitude(jj, unit_num, this%tm_info(jj)%atts, ierr)
          if (ierr /= 0)&
call wrstop(functionName, trim(instrum)//": Obtaining attitude failed")

          if (size(this%tm_info(jj)%atts) < 2)&
call wrstop(functionName, trim(instrum)&
            //": Less than 2 rows in CORRATT this CCD - can't interpolate attitudes.")
        end if

        call wrschatty (functionName, trim(instrum)//": Calling get_deadc...")
        call get_deadc(jj, unit_num, this%tm_info(jj)%deadcs, ierr)
        if (ierr == 3)&
call wrstop(functionName, trim(instrum)//": Obtaining deadcorr failed")

        ndc = size(this%tm_info(jj)%deadcs)
        if(ndc<1) then
          call wrserr(functionName, warn, trim(instrum)&
            //": No deadcorr values obtained. Will assume deadcorr==1.")
        else if(ndc==1) then
          call wrserr(functionName, warn, trim(instrum)&
            //": Only one deadcorr value obtained. This will be used for all.")
        end if
      end do

      ! Close the event list file and release the unit number.
      !
      ierr = 0
      call FTCLOS(unit_num, ierr)
      call fitserr("ftclos", ierr, status)
      call FTFIOU(unit_num, ierr)
      call fitserr("ftfiou", ierr, status)
    end do

    this%n_TMs = 0
    do jj = 1, N_TELESCOPES
      if (this%tm_info(jj)%is_present) this%n_TMs = this%n_TMs + 1
    end do

    if (this%params%withsingle .and. this%n_exp_out /= this%n_TMs*this%n_bands)&
call wrstop(functionName, "Number of --singlemaps must equal n_TMs x n_bands.")

    if (this%params%withinmaps .and. this%n_exp_in  /= this%n_TMs*this%n_bands)&
call wrstop(functionName, "Number of --inputmaps must equal n_TMs x n_bands.")

  end subroutine obtain_t_taskinfo

  !.....................................................................
  subroutine alloc_gti_t(this, n)
    implicit none

    class(gti_t) :: this
    integer, intent(in) :: n

    call this%dealloc()
    allocate(this%starts(n))
    allocate(this%stops(n))
    this%ngti = n

  end subroutine alloc_gti_t

  !.....................................................................
  subroutine dealloc_gti_t(this)
    implicit none

    class(gti_t) :: this

    if(allocated(this%starts))&
      deallocate(this%starts)
    if(allocated(this%stops))&
      deallocate(this%stops)
    this%ngti = 0

  end subroutine dealloc_gti_t

  !.....................................................................
  integer function get_num_gti_t(this) result(ngti)
    implicit none

    class(gti_t) :: this

    ngti = this%ngti

  end function get_num_gti_t

  !.....................................................................
  subroutine or_with_gti_t(this, gtis)
    ! This has been adapted from erolib_gti_mod:or_with_tGTI()

    implicit none
    class(gti_t) :: this
    type(gti_t), intent(in) :: gtis

    character(*), parameter :: functionName = "or_with_gti_t"
    integer :: numRowsA,numRowsB,ii,numOutRows
    type(tRange), dimension(:), allocatable :: rangesA,rangesB,outRanges
    integer :: ngtis

    numRowsA = this%get_num()
    numRowsB = gtis%get_num()

    if(numRowsA<1) then
      if(numRowsB>0) then
        call this%dealloc()
        ngtis = gtis%get_num()
        call this%alloc(ngtis)
        this%starts = gtis%starts
        this%stops = gtis%stops
      end if
  return
    else if(numRowsB<1) then
  return
    end if

    ! If we've got to here, it means that both 'this' and gtis have >0 rows.

    ! Convert both to tRange objects.
    !
    allocate(rangesA(numRowsA))
    allocate(rangesB(numRowsB))
    rangesA%lo%isUndefined = .false.
    rangesA%hi%isUndefined = .false.
    rangesB%lo%isUndefined = .false.
    rangesB%hi%isUndefined = .false.
    do ii=1,numRowsA
      rangesA(ii)%lo%value = this%starts(ii)
      rangesA(ii)%hi%value = this%stops( ii)
    end do
    do ii=1,numRowsB
      rangesB(ii)%lo%value = gtis%starts(ii)
      rangesB(ii)%hi%value = gtis%stops( ii)
    end do

    call or_ranges(rangesA, rangesB, outRanges)

    ! Convert back from outRanges
    !
    numOutRows = size(outRanges)
    call this%dealloc()
    call this%alloc(numOutRows)
    do ii=1,numOutRows
      this%starts(ii) = outRanges(ii)%lo%value
      this%stops( ii) = outRanges(ii)%hi%value
    end do

    deallocate(rangesA)
    deallocate(rangesB)
    deallocate(outRanges)

  end subroutine or_with_gti_t

  !.....................................................................
  subroutine and_with_not_gti_t(this, exclude_GTIs)
    ! This has been adapted from erolib_gti_mod:and_with_not_tGTI()

    implicit none

    class(gti_t) :: this
    type(gti_t), intent(in) :: exclude_GTIs

    character(*), parameter :: functionName = "and_with_not_tGTI"
    integer :: numRowsA,numRowsB,ii,numOutRows
    type(tRange), dimension(:), allocatable :: rangesA,rangesB,outRanges

    numRowsA = this%get_num()
    numRowsB = exclude_GTIs%get_num()

    if(numRowsA<1)&
  return

    if(numRowsB<1)&
  return

    ! If we've got to here, it means that both 'this' and exclude_GTIs have >0 rows.

    if(this%stops(numRowsA) <= exclude_GTIs%starts(1)) then
  return
    end if

    if(this%starts(1) >= exclude_GTIs%stops(numRowsB)) then
  return
    end if

    ! Convert 'this' to a tRanges object.
    !
    allocate(rangesA(numRowsA))
    rangesA%lo%isUndefined = .false.
    rangesA%hi%isUndefined = .false.
    do ii=1,numRowsA
      rangesA(ii)%lo%value = this%starts(ii)
      rangesA(ii)%hi%value = this%stops(ii)
    end do

    ! Convert the gaps between exclude_GTIs to a tRanges object with the first and last bound being undefined.
    !
    allocate(rangesB(numRowsB+1))
    rangesB%lo%isUndefined = .false.
    rangesB%hi%isUndefined = .false.
    ii = 1
    rangesB(ii)%lo%isUndefined = .true.
    rangesB(ii)%hi%value = exclude_GTIs%starts(ii)
    do ii=2,numRowsB
      rangesB(ii)%lo%value = exclude_GTIs%stops(ii-1)
      rangesB(ii)%hi%value = exclude_GTIs%starts(ii)
    end do
    ii = numRowsB+1
    rangesB(ii)%lo%value = exclude_GTIs%stops(ii-1)
    rangesB(ii)%hi%isUndefined = .true.

    call and_ranges(rangesA, rangesB, outRanges)

    ! Convert back from outRanges
    !
    numOutRows = size(outRanges)
!    call this%realloc(numOutRows)
    call this%dealloc()
    call this%alloc(numOutRows)
    do ii=1,numOutRows
      this%starts(ii) = outRanges(ii)%lo%value
      this%stops( ii) = outRanges(ii)%hi%value
    end do

    deallocate(rangesA)
    deallocate(rangesB)
    deallocate(outRanges)

  end subroutine and_with_not_gti_t

  !.....................................................................
  subroutine start_timer_t(this)
    implicit none

    class(timer_t) :: this

    call cpu_time(this%start_time)
    this%timer_was_started = .true.

  end subroutine start_timer_t

  !.....................................................................
  subroutine stop_timer_t(this, msg)
    implicit none

    class(timer_t) :: this
    character(*), optional, intent(in) :: msg

    real(real64) :: stop_time
    character(len=80) :: line

    if(this%timer_was_started) then
      call cpu_time(stop_time)

      write (line, '(f10.3,a)') stop_time - this%start_time, " seconds."
      if(present(msg)) then
        line = trim(msg)//" "//trim(adjustl(line))
      else
        line = trim(adjustl(line))
      end if
      call wrlog("main", line)

      this%timer_was_started = .false.

    else
      call wrerr("stop_timer_t", 1, "You need to start the timer before you stop it.")
    end if

  end subroutine stop_timer_t

!----------------------------------------------------------------------------
  subroutine stop_if_not_exists(function_name, file_name)
!----------------------------------------------------------------------------
    implicit none

    character(*), intent(in) :: function_name, file_name

    integer :: ierr,dummy_rtn,exists

    ierr = 0
    call FTEXIST(file_name, exists, ierr)
    if(ierr/=0) then
      call fitserr("FTEXIST", ierr, dummy_rtn, " - "//trim(file_name))
    else if(exists==0) then
call wrstop(function_name, "File "//trim(file_name)//" not found.")
    end if

  end subroutine stop_if_not_exists

!----------------------------------------------------------------------------
  subroutine get_gti(ic, clun, gtitype, gtis, ierr)
!----------------------------------------------------------------------------

    integer,          intent(in)  :: ic, clun
    character(len=*), intent(in)  :: gtitype
    type(gti_t),      intent(out) :: gtis
    integer,          intent(out) :: ierr

    character(*), parameter :: functionName = "get_gti"
    real(real64), parameter :: nullval_d = 0d0
    character(len=9) :: gtinam
    integer :: status
    integer :: tstartcol, tendcol
    integer :: igti, ival
    logical :: undef
    integer,parameter :: nblock = 7200
    integer :: ngti

    status = 0
    ierr = 0

    !if (gtitype /= "FLAREGTI") then
    write (gtinam, '(a, i1)') trim(gtitype), ic
    !else
    !  gtinam=gtitype
    !end if
    call ftmnhd(clun, binary_tbl, gtinam, 0, status)
    if (status > 0) then
      ierr = 1
      call wrerr(functionName, ierr, "Missing " // trim(gtinam) // " extension")
  return
    end if 
    call ftgnrw(clun, ngti, status)
    call fitserr("ftgnrw", status, ierr)
    call ftgcno(clun, casesens, "START", tstartcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=START)")
    call ftgcno(clun, casesens, "STOP", tendcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=STOP)")

    call gtis%alloc(ngti)

    do igti = 1, ngti, nblock
      ival = min(nblock,ngti-igti+1) 
      call ftgcvd(clun, tstartcol, igti, 1, ival, nullval_d, &
                  gtis%starts(igti:igti+ival-1), undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, tendcol, igti, 1, ival, nullval_d, &
                  gtis%stops(igti:igti+ival-1), undef, status)
      call fitserr("ftgcvd", status, ierr)
    end do

    if (status > 0) then
      ierr = 3
      call wrerr(functionName, ierr, "Error reading GTI extension")
  return
    end if 
    
    do igti = 1, ngti
      if (gtis%stops(igti)-gtis%starts(igti) > 1d7) then
         ierr = 3
         call wrerr(functionName, ierr, "Wrong gti")
    exit 
      end if
    end do 

  end subroutine get_gti

!----------------------------------------------------------------------------
  subroutine put_gti(ic, unit, ogtis, ierr)
!----------------------------------------------------------------------------

    integer, intent(in) :: ic, unit
    type(gti_t), intent(in) :: ogtis
    integer, intent(out) :: ierr

    character(len=64), dimension(2) :: ttype
    character(len=64), dimension(2) :: tform, tunit
    character(len=64) :: gtinam

    integer :: igti,status,ngti

    status = 0

    ngti = ogtis%get_num()

    ttype(1) = "START"
    ttype(2) = "STOP"
    tform = "1D"
    tunit = ""
    write (gtinam, '(''GTI'', i1)') ic
    call ftibin(unit, ngti, 2, ttype, tform, tunit, gtinam, 0, status) 
    call fitserr("ftibin", status, ierr)

    do igti = 1, ngti
      call ftpcld(unit, 1, igti, 1, 1, ogtis%starts(igti), status)
      call fitserr("ftpcld", status, ierr)
      call ftpcld(unit, 2, igti, 1, 1, ogtis%stops(igti), status)
      call fitserr("ftpcld", status, ierr)
    end do   

    if (status > 0) then
      ierr = 3
      call wrerr("put_gti", ierr, "Can't write GTI")
    end if

  end subroutine put_gti

!----------------------------------------------------------------------------
  subroutine get_attitude(ic, clun, atts, ierr)
!----------------------------------------------------------------------------
    ! Note that the calling routine should deallocate atts when it is finished with it.

    implicit none

    integer,                               intent(in)  :: ic, clun
    type(tAtt), dimension(:), allocatable, intent(out) :: atts
    integer,                               intent(out) :: ierr

    character(*), parameter :: functionName="get_attitude"
    real(real64), parameter :: nullval = 0d0
    character(len=8) :: attnam
    integer :: status
    integer :: timcol, racol, deccol, rollcol
    integer :: natt,iatt,ival
    logical :: undef
    integer,parameter :: nblock = 3200

    status = 0
    ierr = 0

    write (attnam, '(''CORRATT'', i1)') ic
    call ftmnhd(clun, binary_tbl, attnam, 0, status)
    if (status > 0) then
      ierr = 1
      call wrerr(functionName, ierr, "Missing "//trim(attnam)//" extension")
      allocate (atts(0))
  return
    end if 

    call ftgnrw(clun, natt, status)
    call fitserr("ftgnrw", status, ierr)
    call ftgcno(clun, casesens, "TIME", timcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=TIME)")
    call ftgcno(clun, casesens, "RA", racol, status)
    call fitserr("ftgcno", status, ierr, " (colname=RA)")
    call ftgcno(clun, casesens, "DEC", deccol, status)
    call fitserr("ftgcno", status, ierr, " (colname=DEC)")
    call ftgcno(clun, casesens, "ROLL", rollcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=ROLL)")

    allocate (atts(natt))

    do iatt = 1, natt, nblock
      ival = min(nblock,natt-iatt+1)
      call ftgcvd(clun, timcol, iatt, 1, ival, nullval, &
                  atts(iatt:iatt+ival-1)%time, undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, racol, iatt, 1, ival, nullval, &
                  atts(iatt:iatt+ival-1)%ra, undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, deccol, iatt, 1, ival, nullval, &
                  atts(iatt:iatt+ival-1)%dec, undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, rollcol, iatt, 1, ival, nullval, &
                  atts(iatt:iatt+ival-1)%roll, undef, status)
      call fitserr("ftgcvd", status, ierr)
    end do

    if(natt>1) then
      if(any(atts(2:natt)%time < atts(1:natt-1)%time)) then
call wrstop(functionName, "Extension "//trim(attnam)//": TIME column values not in increasing order.")
      end if
    end if

  end subroutine get_attitude

!----------------------------------------------------------------------------
  subroutine get_deadc(ic, clun, deadcs, ierr)
!----------------------------------------------------------------------------
    ! Note that the calling routine should deallocate deadc after use.

    integer, intent(in) :: ic, clun
    type(tDeadCorr), dimension(:), allocatable, intent(out) :: deadcs
    integer, intent(out) :: ierr

    character(*), parameter :: functionName="get_deadc"
    real, parameter :: nullval_e = 0.0
    real(real64), parameter :: nullval_d = 0d0
    character(len=8) :: dcnam
    integer :: status,ndc
    integer :: timcol, dccol
    integer :: idc, ival
    logical :: undef
    integer,parameter :: nblock = 3200

    ierr = 0
    status = 0

    write (dcnam, '(''DEADCOR'', i1)') ic
    call ftmnhd(clun, binary_tbl, dcnam, 0, status)
    call fitserr("ftmnhd", status, ierr)
    call ftgnrw(clun, ndc, status)
    call fitserr("ftgnrw", status, ierr)
    call ftgcno(clun, casesens, "TIME", timcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=TIME)")
    call ftgcno(clun, casesens, "DEADC", dccol, status)
    call fitserr("ftgcno", status, ierr, " (colname=DEADC)")

    allocate (deadcs(ndc))

    do idc = 1, ndc, nblock
      ival = min(nblock,ndc-idc+1)
      call ftgcvd(clun, timcol, idc, 1, ival, nullval_d, &
                  deadcs(idc:idc+ival-1)%time, undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcve(clun, dccol, idc, 1, ival, nullval_e, &
                  deadcs(idc:idc+ival-1)%value, undef, status)
      call fitserr("ftgcve", status, ierr)
    end do

    if (status > 0) then
      ierr = 3
      call wrerr(functionName, ierr, "Error reading DEADC extension")
  return
    end if 

  end subroutine get_deadc

!----------------------------------------------------------------------------
  recursive subroutine lininterp_att(atts, iatime, att_interp, ierr, do_angles)
!----------------------------------------------------------------------------

    type(tAtt), dimension(:)&
      ,                intent(in)    :: atts       ! times from attitude extension
    integer,           intent(inout) :: iatime     ! index to current attitude 
                                                !   table column
    type(tAtt),        intent(inout) :: att_interp
    integer,           intent(out)   :: ierr
    logical, optional, intent(in)    :: do_angles

    integer :: natts
    real(real64), parameter :: delta = 2.5d0
    real(real64) :: bfrac

    ! defaults:
    ierr = 0

    ! Probably don't need the following, since the struct is defined already with these defaults:
    att_interp%ra    = 0.0d0
    att_interp%dec   = 0.0d0
    att_interp%roll  = 0.0d0
    att_interp%x_pix = 0.0d0
    att_interp%y_pix = 0.0d0
    att_interp%is_valid = .false.

    natts = size(atts)
    if (iatime >= natts) then
      ierr = 3
  return
    end if
 
    do while (iatime < natts-1 .and. atts(iatime+1)%time < att_interp%time) 
      iatime = iatime+1
    end do

    if (att_interp%time > atts(iatime)%time-delta .and. att_interp%time < atts(iatime+1)%time+delta) then
      bfrac = (att_interp%time - atts(iatime)%time)&
        /(atts(iatime+1)%time - atts(iatime)%time)

      if(present(do_angles)) then
        att_interp%ra    = atts(iatime)%ra    + bfrac*(atts(iatime+1)%ra    - atts(iatime)%ra)
        att_interp%dec   = atts(iatime)%dec   + bfrac*(atts(iatime+1)%dec   - atts(iatime)%dec)
        att_interp%roll  = atts(iatime)%roll  + bfrac*(atts(iatime+1)%roll  - atts(iatime)%roll)
      end if
      att_interp%x_pix = atts(iatime)%x_pix + bfrac*(atts(iatime+1)%x_pix - atts(iatime)%x_pix)
      att_interp%y_pix = atts(iatime)%y_pix + bfrac*(atts(iatime+1)%y_pix - atts(iatime)%y_pix)
    else
      ierr = 2
  return
    end if

    att_interp%is_valid = .true.

    if (att_interp%time>atts(natts)%time .or. att_interp%time<atts(1)%time) then
      ierr = 1 
      att_interp%is_valid = .false.
    end if

  end subroutine lininterp_att

!----------------------------------------------------------------------------
  recursive subroutine linint_deadc(gtime, deadcs, idctime, deadc, ierr)
!----------------------------------------------------------------------------
  
    real(real64),       intent(in)    :: gtime   ! time (endpoint of bin) for which
                                                 !   deadtime is to be interpolated 
    type(tDeadCorr), dimension(:)&
      ,                 intent(in)    :: deadcs  ! deadc from deadtime extension
    integer,            intent(inout) :: idctime ! index to current deadtime 
                                                 !   table row
    real,               intent(out)   :: deadc   ! interpolated deadc at time gtime
    integer,            intent(out)   :: ierr

    integer :: ndc
    real(real64), parameter :: delta = 5d0
!    real(real64) :: bfrac
    real :: bfrac

    ierr = 0
    deadc = 1.0

    ndc = size(deadcs)

    if(ndc<1) then
      deadc = 1.0
  return
    else if(ndc==1) then
      deadc = deadcs(1)%value
  return
    end if

    if (idctime >= ndc) then
      ierr = 3  
  return
    end if 

    do while (idctime<ndc-1 .and.  deadcs(idctime+1)%time < gtime)
      idctime = idctime+1
    end do   
    
    if (gtime>deadcs(ndc)%time .or. gtime<deadcs(1)%time) then
      ierr = 1

      if(gtime>deadcs(ndc)%time) then ! extrapolate.
        deadc = deadcs(ndc)%value
      else
        deadc = deadcs(1)%value
      end if

  return
    end if

    if (gtime > deadcs(idctime)%time-delta .and. gtime < deadcs(idctime+1)%time+delta) then
      bfrac = real(gtime - deadcs(idctime)%time)&
        /real(deadcs(idctime+1)%time - deadcs(idctime)%time)
      deadc = deadcs(idctime)%value + bfrac*(deadcs(idctime+1)%value-deadcs(idctime)%value)
    else
      ierr = 2
  return
    end if

  end subroutine linint_deadc

!----------------------------------------------------------------------------
  function process_caldb_status(caldb_status, nret, tel_id, function_name&
    , cal_code, ib) result(ierr)
!----------------------------------------------------------------------------
    implicit none

    integer,           intent(in) :: caldb_status,nret,tel_id
    character(*),      intent(in) :: function_name,cal_code
    integer, optional, intent(in) :: ib
    integer :: ierr

    character(len=80) :: line

    ierr = 0 ! default

    if (caldb_status /= 0 .and. caldb_status /= 1) then ! caldb_status==1 means nret<1, which we deal with below.
      if(present(ib)) then
        write (line, '(a,i1,a,i1)') "Error finding cal component "&
          //trim(cal_code)//" for telescope ", tel_id, ", band ", ib
      else
        write (line, '(a,i1,a,i1)') "Error finding cal component "&
          //trim(cal_code)//" for telescope ", tel_id, "."
      end if
      call wrerr(function_name, 3, line)
      ierr = 3
  return
    end if

    if (nret > 1) then 
      call wrerr(function_name, 1, "More than one "//trim(cal_code)&
        //" found; using first.")
    end if

    if (nret < 1) then
      if(present(ib)) then
        write (line, '(a,i1,a,i1)') "Could find no cal component "&
          //trim(cal_code)//" for telescope ", tel_id, ", band ", ib
      else
        write (line, '(a,i1,a,i1)') "Could find no cal component "&
          //trim(cal_code)//" for telescope ", tel_id, "."
      end if
      call wrerr(function_name, 3, line)
      ierr = 3
  return
    end if

  end function process_caldb_status


!----------------------------------------------------------------------------
  subroutine get_detmaps(ic, instrume, nb, e_min, e_max, date_obs, date_end, flagsel, &
                         patsel, detmap, xraw_ax, yraw_ax, xraw_scal, yraw_scal, ierr)
!----------------------------------------------------------------------------
    integer,            intent(in)  :: ic                 ! telescope number
    character(len=30),  intent(in)  :: instrume 
    integer,            intent(in)  :: nb                 ! number of energy bands
    real, dimension(:), intent(in)  :: e_min, e_max ! bounds of energy bands
    integer,            intent(in)  :: flagsel,patsel
    character(len=19),  intent(in)  :: date_obs, date_end
    real, dimension(N_PIXELS_CCD,N_PIXELS_CCD,nb)&
      ,                 intent(out) :: detmap ! detector maps for nb bands 
    real(real64),       intent(out) :: xraw_ax, yraw_ax ! optic axis in CCD coords.  
    real(real64),       intent(out) :: xraw_scal, yraw_scal ! CCD pixel dimensions in deg.
    integer,            intent(out) :: ierr

    character(*), parameter :: functionName = "get_detmaps"
    character(len=30), parameter :: telescop = "eROSITA"
    character(len=30), parameter :: filter = "-", detnam = "-"
    character(len=30) :: codenam

    integer, parameter :: group = 1, maxret = 100 
    integer, parameter :: nullval_j = 0
    real, parameter :: nullval_e = 0.0
    real(real64), parameter :: nullval_d = 0.0d0
    integer, dimension(maxret) :: extno
    integer, dimension(:,:), allocatable :: fovmap
    character(len=256), dimension(maxret) :: filename , online
    character(len=256) :: expression
    character(len=10) :: strtdate, strtime, stpdate, stptime
    character(len=80) :: line
    character(len=64) :: comment
    character(len=5) :: cenerg
    real(real64), dimension(1) :: temp_dbl

    integer :: i, ib, status, unit, hdutype, xscalcol, yscalcol, xaxcol, yaxcol
    integer :: naxis1, naxis2
    integer :: nret, nfound
    logical :: anyf, undef

    integer :: chatter                      ! caldb verbosity

    ierr = 0 ! default

    ! Just so as to not leave these floating:
    xraw_ax = -99.9
    yraw_ax = -99.9
    xraw_scal = -99.9
    yraw_scal = -99.9

    ! get time for caldb access    
    call prep_gtcalf(date_obs, date_end, strtdate, strtime, &
                     stpdate, stptime, chatter)

    ! loop through energy bands
    do ib = 1, nb

      ! create ascii energy bounds; remove leading blanks
      write(cenerg,'(i5)') nint(0.5*(e_min(ib)+e_max(ib))*1e3)
      do while (cenerg(1:1) == " ")
        cenerg = cenerg(2:)
      end do

      ! create test expression (this is preliminary - needs to be improved       
      expression = "ENERG.eq." // trim(cenerg)

      codenam = "DETMAP"

      call wrchatty(functionName, "Searching CALDB for component codename "//trim(codenam))
      call gtcalf (chatter, telescop, instrume, detnam, filter, &
                   codenam, strtdate, strtime, stpdate, stptime, &
                   expression, maxret, filename, extno, online, &
                   nret, nfound, status)

      ierr = process_caldb_status(status, nret, ic, functionName, codenam, ib)
      if(ierr/=0) then
  return
      end if

      call stop_if_not_exists(functionName, filename(1))

      ! open cal file, move to extension, get keywords and data array
      status = 0
      call ftgiou (unit, status)
      call fitserr("ftiou", status, ierr)
      call ftiopn(unit, filename(1), ftread, status)
      call fitserr("ftiopn", status, ierr)
      ! is first extno 0 or 1??
      call ftmahd(unit, extno(1)+1, hdutype, status)
      call fitserr("ftmahd", status, ierr)

      call ftgkyj(unit, "NAXIS1", naxis1, comment, status)
      call fitserr("ftgkyj", status, ierr)
      call ftgkyj(unit, "NAXIS2", naxis2, comment, status)
      call fitserr("ftgkyj", status, ierr)
      if (naxis1 /= N_PIXELS_CCD .or. naxis2 /= N_PIXELS_CCD) then
        ierr = 3
        call wrerr(functionName, ierr, "Wrong detmap size")
  return
      end if 

      call ftg2de(unit,group,nullval_e,naxis1,naxis1,naxis2,detmap(:,:,ib),anyf,status)
      call fitserr("ftg2de", status, ierr)

      call ftclos(unit, status)       
      call fitserr("ftclos", status, ierr)

      if (status > 0) then
        ierr = 3
        write (line, '(''Could not read DETMAP for telescope '', i1, '', band '', i1)') ic, ib
        call wrerr(functionName, ierr, line)       
      end if

    end do

    if (btest(flagsel,BIT_OUT_OF_FOV)) then

      ! get fov map
      codenam="FOVMAP"
      expression="-"
      call wrchatty(functionName, "Searching CALDB for component codename "//trim(codenam))
      call gtcalf (chatter, telescop, instrume, detnam, filter, &
                   codenam, strtdate, strtime, stpdate, stptime, &
                   expression, maxret, filename, extno, online, &
                   nret, nfound, status)       

      allocate (fovmap(N_PIXELS_CCD, N_PIXELS_CCD))

      if ( status > 0 .or. nret < 1) then
        write (line, '(a,i1,a)') "Could not read FOVMAP for telescope ", ic, ". Setting it to unity."
        ierr = 2
        call wrerr(functionName, ierr, line)
        fovmap = 1
      else

        ! open cal file, move to extension, get keywords and data array
        status = 0
        call ftiopn(unit, filename(1), ftread, status)
        call fitserr("ftiopn", status, ierr)
        ! is first extno 0 or 1??
        call ftmahd(unit, extno(1)+1, hdutype, status)
        call fitserr("ftmahd", status, ierr)
        call ftgkyj(unit, "NAXIS1", naxis1, comment, status)
        call fitserr("ftgkyj", status, ierr)
        call ftgkyj(unit, "NAXIS2", naxis2, comment, status)
        call fitserr("ftgkyj", status, ierr)
        if (naxis1 /= N_PIXELS_CCD .or. naxis2 /= N_PIXELS_CCD) then
          ierr = 3
          call wrerr(functionName, ierr, "Wrong detmap size")
          return
        end if 

        call ftg2dj(unit,group,nullval_j,naxis1,naxis1,naxis2,fovmap,anyf,status)
        call fitserr("ftg2dj", status, ierr)

        call ftclos(unit, status)       
        call fitserr("ftclos", status, ierr)

        if (status > 0) then
          ierr = 3
          write (line, '(''Could not read FOVMAP for telescope '', i1, '', band '', i1)') ic, ib
          call wrerr(functionName, ierr, line)       
        end if
      end if

      do i=1,nb
        detmap(:,:,i)=detmap(:,:,i)*real(fovmap) 
      end do

      deallocate(fovmap)
   
    end if

    if (btest(flagsel,BIT_NEXT_TO_BORDER) .or. patsel>0) then
      detmap(1:N_PIXELS_CCD,           1,1:nb)=0.
      detmap(1:N_PIXELS_CCD,N_PIXELS_CCD,1:nb)=0.
      detmap(           1,1:N_PIXELS_CCD,1:nb)=0.
      detmap(N_PIXELS_CCD,1:N_PIXELS_CCD,1:nb)=0.
    end if

    ! get optical axis position    
    codenam = "GEOM" 
    expression="-"
    call wrchatty(functionName, "Searching CALDB for component codename "//trim(codenam))
    call gtcalf (chatter, telescop, instrume, detnam, filter, &
                 codenam, strtdate, strtime, stpdate, stptime, &
                 expression, maxret, filename, extno, online, &
                 nret, nfound, status)       

    ierr = process_caldb_status(status, nret, ic, functionName, codenam)!, ib)
    if(ierr/=0)&
  return

    call stop_if_not_exists(functionName, filename(1))

    status = 0
    call ftgiou(unit, status)
    call fitserr("ftgiou", status, ierr)
    call fttopn(unit, filename(1), ftread, status)
    call fitserr("ftiopn", status, ierr)
    call ftgcno(unit, casesens, "X_PLATESCALE", xscalcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=X_PLATESCALE)")
    call ftgcno(unit, casesens, "Y_PLATESCALE", yscalcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=Y_PLATESCALE)")

! 20201130 IMS: evatt uses X/Y_REF as the pointing direction of the TM, not X/Y_OPTAX. The present values seem to be the same, so there should be no functional difference; for the time being it remains only a formal correction.
! 20201208 IMS: oops, X/Y_REF is 0.5 pixels displaced from X/Y_OPTAX in TM4. They're the same for TM1.
!    call ftgcno(unit, casesens, "X_OPTAX", xaxcol, status)
!    call fitserr("ftgcno", status, ierr, " (colname=X_OPTAX)")
!    call ftgcno(unit, casesens, "Y_OPTAX", yaxcol, status)
!    call fitserr("ftgcno", status, ierr, " (colname=Y_OPTAX)")
    call ftgcno(unit, casesens, "X_REF", xaxcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=X_REF)")
    call ftgcno(unit, casesens, "Y_REF", yaxcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=Y_REF)")

!    call ftgcvd(unit, xscalcol, 1, 1, 1, nullval_d, xraw_scal, undef, status) ! CCD pixel size in arcsec
    call ftgcvd(unit, xscalcol, 1, 1, 1, nullval_d, temp_dbl, undef, status) ! CCD pixel size in arcsec
    call fitserr("ftgcvd", status, ierr)
!    xraw_scal=xraw_scal/3600d0 ! CCD pixel size in deg.
    xraw_scal=temp_dbl(1)/3600d0 ! CCD pixel size in deg.
!    call ftgcvd(unit, yscalcol, 1, 1, 1, nullval_d, yraw_scal, undef, status) ! CCD pixel size in arcsec
    call ftgcvd(unit, yscalcol, 1, 1, 1, nullval_d, temp_dbl, undef, status) ! CCD pixel size in arcsec
    call fitserr("ftgcvd", status, ierr)
!    yraw_scal=yraw_scal/3600d0 ! CCD pixel size in deg.
    yraw_scal=temp_dbl(1)/3600d0 ! CCD pixel size in deg.
!    call ftgcvd(unit, xaxcol, 1, 1, 1, nullval_d, xraw_ax, undef, status)
    call ftgcvd(unit, xaxcol, 1, 1, 1, nullval_d, temp_dbl, undef, status)
    call fitserr("ftgcvd", status, ierr)
    xraw_ax = temp_dbl(1)
!    call ftgcvd(unit, yaxcol, 1, 1, 1, nullval_d, yraw_ax, undef, status)
    call ftgcvd(unit, yaxcol, 1, 1, 1, nullval_d, temp_dbl, undef, status)
    call fitserr("ftgcvd", status, ierr)
    yraw_ax = temp_dbl(1)
    call ftclos(unit, status)       
    call fitserr("ftclos", status, ierr)

  end subroutine get_detmaps

!----------------------------------------------------------------------------
  subroutine get_badpix(ic, clun, badpix, ierr)
!----------------------------------------------------------------------------
    integer,                        intent(in)  :: ic, clun
    type(badpixt), allocatable, dimension(:)&
      ,                             intent(out) :: badpix
    integer,                        intent(out) :: ierr

    integer, parameter :: maxret = 100 
    character(len=30), parameter :: telescop = "eROSITA", codenam = "BADPIX"
    character(len=30), parameter :: filter = "-", detnam = "-"
    character(len=7) :: badnam

    integer(int16), parameter :: nullvali = -1
    real(real64), parameter :: nullvald = -1d0
    integer :: i, nrow, unitNo, status
    integer :: rxcol, rycol, yextcol, tmincol, tmaxcol, tpcol, bdcol
!    integer*2 :: badflag, type
    logical :: undef

    ierr = 0 ! default
    status = 0

    ! read badpixels from BADPIX extension
    unitNo = clun
    write (badnam, '(''BADPIX'', i1)') ic
    status = 0
    call ftmnhd(unitNo, binary_tbl, badnam, 0, status)
    call fitserr("ftmnhd", status, ierr)

    ! get table info
    status = 0
    call ftgnrw(unitNo, nrow, status)
    call fitserr("ftgnrw", status, ierr)
    call ftgcno(unitNo, casesens, "RAWX", rxcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=RAWX)")
    call ftgcno(unitNo, casesens, "RAWY", rycol, status)
    call fitserr("ftgcno", status, ierr, " (colname=RAWY)")
    call ftgcno(unitNo, casesens, "YEXTENT", yextcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=YEXTENT)")
    call ftgcno(unitNo, casesens, "TIMEMIN", tmincol, status)
    call fitserr("ftgcno", status, ierr, " (colname=TIMEMIN)")
    call ftgcno(unitNo, casesens, "TIMEMAX", tmaxcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=TIMEMAX)")
    call ftgcno(unitNo, casesens, "TYPE", tpcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=TYPE)")
    call ftgcno(unitNo, casesens, "BADFLAG", bdcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=BADFLAG)")
    
    allocate(badpix(nrow))

    do i=1,nrow
!      call ftgcvi(unitNo, tpcol, i, 1, 1, nullvali,  type, undef, status)   
      call ftgcvi(unitNo, tpcol, i, 1, 1, nullvali,  badpix(i:i)%type, undef, status)   
      call fitserr("ftgcvi", status, ierr)
!      call ftgcvi(unitNo, bdcol, i, 1, 1, nullvali,  badflag, undef, status)   
      call ftgcvi(unitNo, bdcol, i, 1, 1, nullvali,  badpix(i:i)%badflag, undef, status)   
      call fitserr("ftgcvi", status, ierr)

!      badpix(i)%type=type
!      badpix(i)%badflag=badflag
      call ftgcvi(unitNo, rxcol,   i, 1, 1, nullvali, badpix(i:i)%rawx,    undef, status)   
      call fitserr("ftgcvi", status, ierr)
      call ftgcvi(unitNo, rycol,   i, 1, 1, nullvali, badpix(i:i)%rawy,    undef, status)   
      call fitserr("ftgcvi", status, ierr)
      call ftgcvi(unitNo, yextcol, i, 1, 1, nullvali, badpix(i:i)%yextent, undef, status)
      call fitserr("ftgcvi", status, ierr)
      call ftgcvd(unitNo, tmincol, i, 1, 1, nullvald, badpix(i:i)%tmin,    undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(unitNo, tmaxcol, i, 1, 1, nullvald, badpix(i:i)%tmax,    undef, status)
      call fitserr("ftgcvd", status, ierr)
    end do

  end subroutine  get_badpix

!----------------------------------------------------------------------------
  recursive subroutine remove_badpix(badpix, flagsel, patsel, t0, t1&
    , nb, detmap, baddetmap)
!----------------------------------------------------------------------------

  type(badpixt), dimension(:), intent(in)  :: badpix
  integer,                     intent(in)  :: flagsel, patsel, nb
  real(real64),                intent(in)  :: t0, t1
  real, dimension(N_PIXELS_CCD,N_PIXELS_CCD,nb)&
    ,                          intent(in)  :: detmap
  real, dimension(N_PIXELS_CCD,N_PIXELS_CCD,nb)&
    ,                          intent(out) :: baddetmap

  integer :: ii,bxi,byi,yext,yi,nbad
  integer :: x_start_i,x_finis_i,y_start_i,y_finis_i
  integer :: px_start_i,px_finis_i,py_start_i,py_finis_i
  real, dimension(3,3,nb) :: bad_neigh

  nbad = size(badpix)

  baddetmap = detmap

  ! Set just the orthogonal neighbours bad:
  !
  bad_neigh(:,:,:) = 1.0
  bad_neigh(2,1,:) = 0.0
  bad_neigh(2,3,:) = 0.0
  bad_neigh(1,2,:) = 0.0
  bad_neigh(3,2,:) = 0.0

  ! 20200916 IMS: deletion of CCD pixels on the basis of filtering of
  ! event PAT_TYP has been added. This works because you will not get
  ! valid pattern events either on declared bad pixels themselves or on
  ! their orthogonal neighbours.
  !
  do ii=1,nbad
    if (t0<badpix(ii)%tmax.or.t1>badpix(ii)%tmin) then 
      bxi=badpix(ii)%rawx 
      byi=badpix(ii)%rawy
      yext=badpix(ii)%yextent

      ! Test whether to remove centre:
      !
      if(badpix(ii)%badflag==BP_uplinked&
      .or. (badpix(ii)%type==BP_hot     .and. btest(flagsel,BIT_ON_BADPIX))&
      .or. (badpix(ii)%type==BP_flicker .and. btest(flagsel,BIT_ON_FLICKERING))&
      .or. (badpix(ii)%type==BP_dead    .and. btest(flagsel,BIT_ON_DEADPIX))&
      .or. patsel>0) then
        baddetmap(bxi,byi:byi+yext-1,1:nb)=0.
      end if

      ! Test whether to remove neighbours:
      !
      if(  (badpix(ii)%type==BP_hot     .and. btest(flagsel,BIT_NEXT_TO_BRIGHT_PIX))&
      .or. (badpix(ii)%type==BP_flicker .and. btest(flagsel,BIT_NEXT_TO_FLICKERING))&
      .or. (badpix(ii)%type==BP_dead    .and. btest(flagsel,BIT_NEXT_TO_DEAD_PIX))&
      .or. patsel>0) then
        x_start_i = max(           1, bxi - 1)
        x_finis_i = min(N_PIXELS_CCD, bxi + 1)
        px_start_i = x_start_i - bxi + 2
        px_finis_i = x_finis_i - bxi + 2
        do yi=byi,byi+yext-1
          y_start_i = max(           1, yi - 1)
          y_finis_i = min(N_PIXELS_CCD, yi + 1)
          py_start_i = y_start_i - yi + 2
          py_finis_i = y_finis_i - yi + 2

          baddetmap(x_start_i:x_finis_i,y_start_i:y_finis_i,1:nb)&
            = baddetmap(x_start_i:x_finis_i,y_start_i:y_finis_i,1:nb)&
            *bad_neigh(px_start_i:px_finis_i,py_start_i:py_finis_i,1:nb)
        end do
      end if
    end if ! event time is within bad pixel validity.
  end do ! loop over bad pixels.

  end subroutine remove_badpix 
    
!----------------------------------------------------------------------------
  subroutine get_area(ic, nbands, instrume, date_obs, date_end, e_min, e_max, filt, weights, ierr)
!----------------------------------------------------------------------------
    ! This routine doesn't, as written, calculate a value for effective area;
    ! rather it estimates an average value for the contribution of the TM
    ! (instrume) in question, this average being weighted by the effective
    ! area. The average is performed over the energy band, i.e. what is
    ! approximated is
    !
    !	              e_hi
    !	              /
    !	              | de area(e)*contrib(e)
    !	              /
    !	             e_lo
    !	<contrib> = --------------------------.
    !	                  e_hi
    !	                  /
    !	                  | de area(e)
    !	                  /
    !	                 e_lo
    !
    integer,                         intent(in)  :: ic           ! telescope number 
    integer,                         intent(in)  :: nbands
    character(len=30),               intent(in)  :: instrume, filt
    character(len=19),               intent(in)  :: date_obs, date_end
    real(real32), dimension(nbands), intent(in)  :: e_min, e_max ! energy band boundaries in PI units
    real(real32), dimension(nbands), intent(out) :: weights      ! area weight factor array
    integer,                         intent(out) :: ierr

    character(*), parameter :: functionName="get_area"
    real, parameter :: nullval = -1.0

    character(len=30), parameter :: telescop = "eROSITA", detnam = "-"
    character(len=256), parameter :: expression = "-"
    character(len=30) :: codenam
    character(len=10) :: strtdate, strtime, stpdate, stptime
    integer, parameter :: maxret = 100 
    character(len=256), dimension(maxret) :: filename , online
    integer, dimension(maxret) :: extno
    logical :: undef
    integer :: narea
    type(tRange_primitive), dimension(:), allocatable :: arf_energies
    ! 'Primitive' range objects can't have undefined bounds.

    real(real32) :: csum, snorm, width
    real(real32), dimension(:), allocatable :: elos, ehis, areas, conts
    integer :: chatter, nret, nfound
    integer :: unitNo, status, hdutype
    integer :: elocol, ehicol, areacol, concol
    integer :: iarea, ib
    type(tRange_primitive), dimension(nbands) :: energies
    type(tRange_primitive) :: anded_range
    logical :: ranges_overlap

    codenam = "SPECRESP"

    ! Convert energies to range objects:
    !
    do ib=1,nbands
      energies(ib)%lo = real(e_min(ib), kind=real64)
      energies(ib)%hi = real(e_max(ib), kind=real64)
    end do

    call prep_gtcalf(date_obs, date_end, strtdate, strtime, &
                     stpdate, stptime, chatter)

    call wrchatty(functionName, "Searching CALDB for component codename "//trim(codenam))
    call gtcalf (chatter, telescop, instrume, detnam, filt, &
                 codenam, strtdate, strtime, stpdate, stptime, &
                 expression, maxret, filename, extno, online, &
                 nret, nfound, status)       

    ierr = process_caldb_status(status, nret, ic, functionName, codenam)!, ib)
    if(ierr/=0)&
  return

    call stop_if_not_exists(functionName, filename(1))

    ! open cal file, move to extension, get keywords and data array
    status = 0
    call ftgiou(unitNo, status)
    call fitserr("ftgiou", status, ierr)
    call ftiopn(unitNo, filename(1), ftread, status)
    call fitserr("ftiopn", status, ierr)
! is first extno 0 or 1??
    call ftmahd(unitNo, extno(1)+1, hdutype, status)
    call fitserr("ftmahd", status, ierr)

    call ftgnrw(unitNo, narea, status)
    call fitserr("ftgnrw", status, ierr)
    call ftgcno(unitNo, casesens, "ENERG_LO", elocol, status)
    call fitserr("ftgcno", status, ierr, " (colname=ENERG_LO)")
    call ftgcno(unitNo, casesens, "ENERG_HI", ehicol, status)
    call fitserr("ftgcno", status, ierr, " (colname=ENERG_HI)")
    call ftgcno(unitNo, casesens, "SPECRESP", areacol, status)
    call fitserr("ftgcno", status, ierr, " (colname=SPECRESP)")
    call ftgcno(unitNo, casesens, "CONTRIB", concol, status)
    call fitserr("ftgcno", status, ierr, " (colname=CONTRIB)")

    allocate(elos( narea))
    allocate(ehis( narea))
    allocate(areas(narea))
    allocate(conts(narea))

    call FTGCVE(unitNo, elocol, 1, 1, narea, nullval, elos, undef, status)
    call fitserr("FTGCVE", status, ierr, " (colname=ENERG_LO)")
    call FTGCVE(unitNo, ehicol, 1, 1, narea, nullval, ehis, undef, status)
    call fitserr("FTGCVE", status, ierr, " (colname=ENERG_HI)")
    call FTGCVE(unitNo, areacol, 1, 1, narea, nullval, areas, undef, status)
    call fitserr("FTGCVE", status, ierr, " (colname=SPECRESP)")
    call FTGCVE(unitNo, concol, 1, 1, narea, nullval, conts, undef, status)
    call fitserr("FTGCVE", status, ierr, " (colname=CONTRIB)")

    call ftclos(unitNo, status)       
    call fitserr("ftclos", status, ierr)

    allocate(arf_energies(narea))
    arf_energies%lo = real(elos, kind=real64)
    arf_energies%hi = real(ehis, kind=real64)

    do ib=1,nbands
      csum = 0.; snorm = 0.

      do iarea = 1, narea
        ! Calc overlap between the ARF and user energy ranges:
        call a_and_b_primitive(arf_energies(iarea), energies(ib), anded_range, ranges_overlap)
        if(.not.ranges_overlap)&
      cycle

        width = real(anded_range%hi - anded_range%lo, kind=real32)
        csum = csum + areas(iarea)*conts(iarea)*width
        snorm = snorm + areas(iarea)*width
      end do

      weights(ib) = csum/snorm
    end do

    deallocate(elos, ehis, areas, conts)
    deallocate(arf_energies)

  end subroutine get_area

!----------------------------------------------------------------------------
  subroutine get_vignetting2(instrume, nbands, e_min, e_max, date_obs, date_end, &
                              plindex, detmap, ierr)
!----------------------------------------------------------------------------
    ! In this subroutine we estimate a spectrum-weighted average value of the
    ! vignetting function for each energy band, at each pixel location. The
    ! spectrum used is a power law, with index supplied via the parameter
    ! --plindex. In formal terms we want (for each pixel) to approximate the
    ! integral
    !
    !	        e_hi
    !	        /
    !	        | de e**k v(e)
    !	        /
    !	       e_lo
    !	<v> = ----------------
    !	          e_hi
    !	          /
    !	          | de e**k
    !	          /
    !	         e_lo
    !
    ! where e_lo and e_hi are the edges of the band. The first step to this
    ! is to break the integral into a sum of integrals as follows:
    !
    !	        __N  e_{i+1}
    !	        \     /
    !	         >    | de e**k v(e)
    !	        /_    /
    !	        i=1  e_i
    !	<v> = ----------------------
    !	         __N  e_{i+1}
    !	         \     /
    !	          >    | de e**k
    !	         /_    /
    !	         i=1  e_i
    !
    ! with e_1 = e_lo and e_{N+1} = e_hi. The approximation is to replace each
    ! integral term by (using a generic example)
    !
    !	e_{i+1}
    !	 /
    !	 | de f(e) ~ f(<e>_i) * (e_{i+1} - e_i),
    !	 /
    !	e_i
    !
    ! where f(e) here is e**k v(e) in the numerator and e**k in the denominator,
    ! and <e>_i = (e_{i+1} + e_i)/2.0. In present practice, (e_{i+1} - e_i)
    ! = constant for each i, therefore cancels between numerator and denominator.
    !
    character(len=30),       intent(in)    :: instrume
    integer,                 intent(in)    :: nbands
    real, dimension(nbands), intent(in)    :: e_min, e_max
    character(len=19),       intent(in)    :: date_obs, date_end
    real,                    intent(in)    :: plindex
    real, dimension(:,:,:),  intent(inout) :: detmap
    integer,                 intent(out)   :: ierr

    character(*), parameter :: functionName="get_vignetting2"

    integer :: xiSize,yiSize,vectorShape(1),ib,n_subband,ii
    integer, dimension(size(detmap,1)) :: xiVector
    integer, dimension(size(detmap,2)) :: yiVector
    real(kind=real64), dimension(size(detmap,1)*size(detmap,2)) :: xpos,ypos,vign
    type(avign_type) :: avign  ! array of avigns for all instruments
    real :: width_band,frac,pl_spec,denom
    real, dimension(:), allocatable :: subband_bounds
    real, dimension(size(detmap,1),size(detmap,2)) :: av_vig
!    real(kind=real64) :: energy
    real :: energy

    xiSize = size(detmap,1)
    yiSize = size(detmap,2)
!*** check that size(detmap,3)==nbands
    vectorShape(1) = xiSize*yiSize

    xiVector = [(ii,ii=1,xiSize,1)]
    yiVector = [(ii,ii=1,yiSize,1)]
    xpos = reshape(spread(real(xiVector, kind=real64), 2, yiSize), vectorShape)
    ypos = reshape(spread(real(yiVector, kind=real64), 1, xiSize), vectorShape)

    call avign_empty_avign(avign)
    call avign_read_avign_from_caldb(avign, instrume, date_obs, date_end, ierr)
    if ( ierr .ne. 0 ) then
      call wrerr (functionName, ierr, "Problem reading AVIGN from CALDB")
  return
    end if
           
    do ib = 1,nbands
      width_band = e_max(ib) - e_min(ib)
      if(width_band<=0.0d0)&
call wrstop(functionName, "Width of energy band must be >0.")

      n_subband = ceiling(width_band/min_dE_keV) ! minimum value will be 1.
      allocate(subband_bounds(n_subband+1))
      subband_bounds(1          ) = e_min(ib)
      subband_bounds(n_subband+1) = e_max(ib)
      do ii=2,n_subband
        frac = (ii-1)/real(n_subband)
        subband_bounds(ii) = e_max(ib)*frac + e_min(ib)*(1.0-frac)
      end do

!      av_vig = 0.0d0
      av_vig = 0.0
      denom = 0.0d0
      do ii = 1, n_subband
!        energy = 0.5d0*(subband_bounds(ii) + subband_bounds(ii+1))
        energy = 0.5*(subband_bounds(ii) + subband_bounds(ii+1))
        call avign_calc_vign_at_xye(avign, xpos, ypos, real(energy, kind=real64), vign, ierr)
        pl_spec = energy**plindex
        av_vig = av_vig + pl_spec*reshape(real(vign), [size(detmap,1),size(detmap,2)])
        denom = denom + pl_spec
      end do
      av_vig = av_vig/denom

      detmap(:,:,ib) = detmap(:,:,ib)*av_vig(:,:)

      deallocate(subband_bounds)
    end do

    call avign_empty_avign(avign)

if(doTest_detmap) then
  detmap(:,:,:) = 0.5
  detmap(1,1,:) = 1.0
  detmap(1,N_PIXELS_CCD,:) = 1.0
  detmap(N_PIXELS_CCD,1,:) = 1.0
  detmap(N_PIXELS_CCD,N_PIXELS_CCD,:) = 1.0
  detmap(192,192,:) = 1.0
end if

  end subroutine get_vignetting2

!----------------------------------------------------------------------------
  subroutine get_ima_keywords (imafile, shapehdr, stdhdr, refhdr, wcshdr)
!----------------------------------------------------------------------------

    character(len=256),    intent(in)  :: imafile
    type(shapehdr_type),   intent(out) :: shapehdr
    type(stdhdr_type),     intent(out) :: stdhdr
    type(wcs_refhdr_type), intent(out) :: refhdr
    type(wcs_imahdr_type), intent(out) :: wcshdr

    integer :: unit, status, ierr

    call stop_if_not_exists("get_ima_keywords", imafile)

    status = 0
    call ftgiou (unit, status)
    call fitserr("ftgiou", status, ierr)
    call ftiopn(unit,imafile, ftread, status) 
    call fitserr("ftiopn", status, ierr)
    if (status > 0) call wrstop("get_ima_keywords", "Couldn't open template image")

    ! get image shape headers
    call fread_shapehdr(unit, shapehdr, ierr) 
    ! get eSASS standard headers
    call fread_stdhdr(unit, stdhdr, ierr)
    call fread_refhdr(unit, refhdr, ierr)
    ! get wcs keywords
    call fread_wcshdr(unit, wcshdr, ierr)

    call ftclos(unit, status)
    call fitserr("ftclos", status, ierr)
    if (status > 0) call wrstop("get_ima_keywords", "Couldn't read template header")
 
  end subroutine get_ima_keywords

!----------------------------------------------------------------------------
  subroutine quantize_image (level, naxis1, naxis2, image)
!----------------------------------------------------------------------------

    real, intent(in) :: level
    integer, intent(in) :: naxis1, naxis2
    real, dimension(:,:), intent(inout) :: image
  
    integer :: ix, iy, isave
    real :: save, smean, level2

    level2 = 0.5*level
    do iy=1,naxis2
      save = image(1,iy)
      isave = 1
      do ix=2,naxis1
        if (abs(image(ix,iy)-save) > level) then
          smean=sum(image(isave:ix-1,iy))/(ix-isave)
          if (smean < level2) smean = 0.
          image(isave:ix-1,iy) = smean
          save = image(ix,iy)
          isave = ix  
        end if
      end do
    end do
    do ix=1,naxis1
      save = image(ix,1)
      isave = 1
      do iy=2,naxis2
        if (abs(image(ix,iy)-save) > level) then
          smean=sum(image(ix,isave:iy-1))/(iy-isave)
          if (smean < level2) smean = 0.
          image(ix,isave:iy-1) = smean
          save = image(ix,iy)
          isave = iy  
        end if
      end do
    end do
    
  end subroutine quantize_image
      
!----------------------------------------------------------------------------
  subroutine write_cam_expmaps (ic, ilist, inst, nb, e_min, e_max&
    , stdhdr, refhdr, wcshdr, ix_size, iy_size, ogtis, expfiles, camexp)
!----------------------------------------------------------------------------

    integer,               intent(in) :: ic, ilist
    character(len=30),     intent(in) :: inst
    integer,               intent(in) :: nb
    real, dimension(:)&
      ,                    intent(in) :: e_min, e_max
    type(stdhdr_type),     intent(in) :: stdhdr
    type(wcs_refhdr_type), intent(in) :: refhdr
    type(wcs_imahdr_type), intent(in) :: wcshdr
    integer,               intent(in) :: ix_size, iy_size
    type(gti_t),           intent(in) :: ogtis
    character(*), dimension(:)&
      ,                    intent(in) :: expfiles
    real, allocatable, dimension(:,:,:)&
      ,                    intent(in) :: camexp

    type(stdhdr_type) :: local_stdhdr
    character(len=80) :: line
    integer :: ib, ierr, unit, status
    integer, dimension(2) :: naxes

    status = 0
    naxes(1) = ix_size
    naxes(2) = iy_size

    ! write out camera exposure maps
    do ib=1,nb
      status = 0
      call ftgiou(unit, status)
      call fitserr("ftgiou", status, ierr)
      call ftinit(unit, "!"//trim(expfiles(nb*(ilist-1)+ib)), 0, status)
      call fitserr("ftinit", status, ierr)
      call ftphps(unit,-32,2,naxes,status)
      call fitserr("ftphps", status, ierr)
      local_stdhdr = stdhdr
      call update_stdhdr(expfiles(nb*(ilist-1)+ib), local_stdhdr, ierr)
      ! CG update stdhdr with instrument and energy bands
      local_stdhdr%ninst = 1
      local_stdhdr%instrumx(ib) = trim(inst)
      local_stdhdr%instrume = trim(inst)
      local_stdhdr%nenergy = 1
      local_stdhdr%eminx(1) = e_min(ib)
      local_stdhdr%emaxx(1) = e_max(ib)

      call fwrite_stdhdr(unit, local_stdhdr, status)

      status = 0
      call fwrite_refhdr(unit, refhdr, status)
      call fwrite_wcshdr(unit, wcshdr, status)

!!!!!!!!!! Quantized images compress better:
!!!   call quantize_image(0.1, ix_size, iy_size, camexp(:,ib,:))
!!!!!!!!!!

      status = 0
      call ftp2de(unit, 0, ix_size, ix_size, iy_size, camexp(:,ib,:), status)            
      call fitserr("ftp2de", status, ierr)
      call wrchatty ("write_cam_expmaps",  "Calling put_gti...")
      call put_gti(ic, unit, ogtis, ierr)

      call ftclos(unit, status)
      call fitserr("ftclos", status, ierr)
      if (status > 0) then
        write(line, &
          '(''Could not write exposure map for telescope '', i1, '', band '', i1)') ic, ib
        call wrstop("write_cam_expmaps", line)
      end if

    end do

  end subroutine write_cam_expmaps

!----------------------------------------------------------------------------
  subroutine write_merged_expmaps(instarr, camarr, nb, e_min, e_max, stdhdr&
    , refhdr, wcshdr, ix_size, iy_size, ogtis, txpfiles, texp)
!----------------------------------------------------------------------------

    character(len=30), dimension(:), intent(in) :: instarr
    integer,           dimension(:), intent(in) :: camarr
    integer,                         intent(in) :: nb
    real, dimension(:),              intent(in) :: e_min, e_max
    type(stdhdr_type),               intent(in) :: stdhdr
    type(wcs_refhdr_type),           intent(in) :: refhdr
    type(wcs_imahdr_type),           intent(in) :: wcshdr
    integer,                         intent(in) :: ix_size, iy_size 
    type(gti_t), dimension(N_TELESCOPES)&
      ,                              intent(in) :: ogtis
    character(*), dimension(:),      intent(in) :: txpfiles
    real, allocatable, dimension(:,:,:)&
      ,                              intent(in) :: texp

    type(stdhdr_type) :: local_stdhdr
    character(len=80) :: line
    character(len=30) :: instn
    integer :: ic, ilist, ib, ierr, unit, status, ncamarr
    integer, dimension(2) :: naxes

    ncamarr = size(camarr)

    naxes(1) = ix_size
    naxes(2) = iy_size

    do ib=1,nb
      status = 0
      call ftgiou(unit, status)
      call fitserr("ftgiou", status, ierr)
      call ftinit(unit, "!"//trim(txpfiles(ib)), 0, status)
      call fitserr("ftinit", status, ierr)
      call ftphps(unit,-32,2,naxes,status)
      call fitserr("ftphps", status, ierr)
      local_stdhdr = stdhdr
      call update_stdhdr(txpfiles(ib), local_stdhdr, ierr)
      local_stdhdr%nenergy = 1
      local_stdhdr%eminx(1) = e_min(ib)
      local_stdhdr%emaxx(1) = e_max(ib)
      call fwrite_stdhdr(unit, local_stdhdr, status)
      status = 0
      call ftukys(unit, "INSTRUME", "merged", " ", status)
      call fitserr("ftmkys", status, ierr)
      call ftukyj(unit, "NINST", ncamarr, " ", status)
      call fitserr("ftpkyj", status, ierr)
      do ilist = 1, ncamarr
        write(instn,'(''INSTRUM'', i1)') ilist
        call ftukys(unit, trim(instn), instarr(ilist), " ", status)
        call fitserr("ftpkys", status, ierr)
      end do  
      call fwrite_refhdr(unit, refhdr, status)
      call fwrite_wcshdr(unit, wcshdr, status)

!!!!!!!! Quantized images compress better:
!!!   call quantize_image(0.1, ix_size, iy_size, texp(:,ib,:))
!!!!!!!!

      status = 0
      call ftp2de(unit, 0,  ix_size, ix_size, iy_size, texp(:,ib,:), status)      
      call fitserr("ftp2de", status, ierr)

      do ilist = 1, ncamarr
        ic = camarr(ilist)
        call wrchatty ("write_merged_expmaps",  "Calling put_gti...")
        call put_gti(ic, unit, ogtis(ic), ierr)!*****
      end do  
    
        call ftclos(unit, status)
        call fitserr("ftclos", status, ierr)
        if (status > 0) then
          write(line, &
          '(''Could not write all telescope exposure map for band '', i1)') ib
          call wrstop("expmap_main", line)
        end if
      end do


  end subroutine write_merged_expmaps

!----------------------------------------------------------------------------
  subroutine get_expmaps (cam_list_i, cam_i, nb, expinfiles, ix_size, iy_size, camexp, &
                          gtis, ierr)
!----------------------------------------------------------------------------

    integer,                 intent(in)  :: cam_list_i, cam_i, nb
    character(*), dimension(:)&
      ,                      intent(in)  :: expinfiles
    integer,                 intent(in)  :: ix_size, iy_size
    real, dimension(ix_size,nb,iy_size)&
      ,                      intent(out) :: camexp 
    type(gti_t),             intent(out) :: gtis
    integer,                 intent(out) :: ierr

    integer, parameter :: group = 1
    real, parameter :: nullval = -1.0
    integer :: ib, unit, naxis1, naxis2, status
    character(len=256) :: expnam
    character(len=64) :: comment
    logical :: anyf, fexist 
    real(real64) :: tmax, tmax_old
    integer :: noexp,ngti

    tmax = 0d0
    noexp = 0   
    do ib = 1, nb
      expnam = trim(expinfiles(nb*(cam_list_i-1)+ib))
      inquire (file=expnam, exist=fexist)
      if (fexist) then
        status = 0
        call ftgiou(unit, status)
        call fitserr("ftgiou", status, ierr)
        call ftiopn(unit, expnam, ftread, status)
        call fitserr("ftiopn", status, ierr)
        call ftgkyj(unit, "NAXIS1", naxis1, comment, status)
        call fitserr("ftgkyj", status, ierr)
        call ftgkyj(unit, "NAXIS2", naxis2, comment, status)
        call fitserr("ftgkyj", status, ierr)
        if (ix_size /= naxis1 .or. iy_size /= naxis2) then
          ierr = 3
          call wrerr("get_expmaps", ierr, "Wrong exposure map size")
  return
        end if 
        call ftg2de(unit,group,nullval,naxis1,naxis1,naxis2,camexp(:,ib,:),anyf,status)
        call fitserr("ftg2de", status, ierr)

        if (ib > 1) call gtis%dealloc() ! probably don't need this, the argument to get_gti() is intent(out).
        call get_gti(cam_i, unit, "GTI", gtis, ierr)
        if ( ierr == 3) then
          call wrerr("get_expmaps", ierr, "Can't read gti extension of input exposure map")
  return
        end if

        ! check whether end of last gti agrees in all energy bands 
        !   (energy dependent gti are not supported)  
        tmax_old = tmax 
        ngti = gtis%get_num()
        tmax = gtis%stops(ngti)
        if (tmax_old > 0. .and. tmax /= tmax_old) then
          ierr = 3
          call wrerr("get_expmaps", ierr, "Inconsistent gti of input exposure maps")
  return
        end if

        call ftclos(unit, status)       
      else ! file not found.
        tmax = 0d0
        camexp(:,ib,:) = 0.
 
        call wrerr("get_expmaps", 1, "No input exposure maps, yet")
        noexp = noexp + 1
      end if
    end do 

    ! Input exposure maps need to be present for all bands
    if (noexp /= nb .and. noexp /= 0) then
      call gtis%alloc(0)
      ierr = 3
      call wrerr("get_expmaps", ierr, "Input exposure maps only available for some energy bands")
  return
    end if 

  end subroutine get_expmaps

!----------------------------------------------------------------------------
  subroutine imaout (ima, name)
!----------------------------------------------------------------------------

    real, dimension(:,:), intent(in) :: ima        
    character(len=*) :: name

    integer :: ierr, unit, status, hdutype
    integer, dimension(2) :: naxes

    print *, "Writing image file "//trim(name)

    naxes = shape(ima)

    status = 0
    call ftgiou(unit, status)
    call fitserr("ftgiou", status, ierr)
    call ftinit(unit, "!"//name, 0, status)
    call fitserr("ftinit", status, ierr)
    call ftmahd(unit,1,hdutype,status)
    call fitserr("ftmahd", status, ierr)
    call ftphps(unit,-32,2,naxes,status)
    call fitserr("ftphps", status, ierr)
    call ftp2de(unit, 0, naxes(1),naxes(1),naxes(2),ima, status)
    call fitserr("ftp2de", status, ierr)
    call ftclos(unit, status)
    call fitserr("ftclos", status, ierr)

  end subroutine imaout

end module expmap_io
