! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! 20210430 IMS: this contains code used not by expmap but by other packages.
module expmap_legacy

  use param
  use ero_general
  use erolib_general_mod ! for report_FITS_err() and BINARY_TBL.
  use srctool_generic_avign
  use iso_fortran_env
  use fitsimage, only: fread_shapehdr, fread_stdhdr, fread_wcshdr, &
                       fwrite_stdhdr, fwrite_wcshdr, &
                       shapehdr_type, stdhdr_type, &
                       wcs_imahdr_type, update_stdhdr
  use srctool_generic, only: wrschatty, wrslog, wrserr

  use erolib_ranges_mod

  implicit none

logical, private, parameter :: doTest=.false.
logical, private, parameter :: doTest_detmap=.false.

  private

  public :: gti_t
  public :: opn_events, get_gti, merge_gti, & 
            get_attitude, get_attitude_old, linint_att2, &
            get_vignetting, get_evt_keywords

  real(real64), private, parameter :: nullval = 0d0
  integer, private, parameter :: ftread = 0, ftwrite = 1
  logical, private, parameter :: casesens = .false.

  type, public :: tAtt
    real(real64) :: time  = 0.0d0
    real(real64) :: ra    = 0.0d0 ! deg
    real(real64) :: dec   = 0.0d0 ! deg
    real(real64) :: roll  = 0.0d0 ! deg
    real(real64) :: x_pix = 0.0d0 ! Pointing location in
    real(real64) :: y_pix = 0.0d0 ! image pixel coordinates.
    logical :: is_valid=.false.
  end type tAtt

  type gti_t ! Legacy, used by apetool, transprep, flaregti
    real(kind=real64), dimension(:), pointer :: start
    real(kind=real64), dimension(:), pointer :: stop
    integer :: ngti
  contains
    procedure :: alloc => alloc_gti_t
    procedure :: dealloc => dealloc_gti_t
  end type gti_t

  contains

  !.....................................................................
  subroutine alloc_gti_t(this, n)
    implicit none

    class(gti_t) :: this
    integer, intent(in) :: n

    call this%dealloc()
    allocate(this%start(n))
    allocate(this%stop(n))
    this%ngti = n

  end subroutine alloc_gti_t

  !.....................................................................
  subroutine dealloc_gti_t(this)
    implicit none

    class(gti_t) :: this

    deallocate(this%start)
    deallocate(this%stop)
    this%ngti = 0

  end subroutine dealloc_gti_t

!----------------------------------------------------------------------------
  subroutine stop_if_not_exists(function_name, file_name)
!----------------------------------------------------------------------------
    implicit none

    character(*), intent(in) :: function_name, file_name

    integer :: ierr,dummy_rtn,exists

    ierr = 0
    call FTEXIST(file_name, exists, ierr)
    if(ierr/=0) then
      call fitserr("FTEXIST", ierr, dummy_rtn, " - "//trim(file_name))
    else if(exists==0) then
call wrstop(function_name, "File "//trim(file_name)//" not found.")
    end if

  end subroutine stop_if_not_exists

!----------------------------------------------------------------------------
  subroutine opn_events(nfiles, evfiles, ncamarr, camarr, camlun, instarr, ierr)
!----------------------------------------------------------------------------
    ! Note that it is forbidden for any TM to be represented more than once in
    ! the input list of events. The task will fail with an error if this
    ! condition is detected.
    !
    ! This is used in apetool.

    integer,                         intent(in) :: nfiles
    character*(*), dimension(*),     intent(in) :: evfiles 
    integer,                         intent(out) :: ncamarr
    integer, dimension(N_TELESCOPES)&
      ,                              intent(out) :: camarr
    integer, dimension(N_TELESCOPES)&
      ,                              intent(out) :: camlun
    character(*), dimension(N_TELESCOPES)&
      ,                              intent(out) :: instarr
    integer,                         intent(out) :: ierr

    logical :: tmAlreadyFound
    integer, dimension(N_TELESCOPES) :: evlun
    integer :: i, j, status, fileInWhichPreviousTMFound
    character(len=64) :: comment
    character(len=30), dimension(N_TELESCOPES) :: instrum
    character(len=8) :: extnam
    character(1) :: tmStr
    character(4) :: file1Str,file2Str

    ! open event files for reading
    do i = 1, nfiles
      call stop_if_not_exists("opn_events", evfiles(i))

      status = 0
      call ftgiou(evlun(i), status)
      call fitserr("ftgiou", status, ierr)
      call ftnopn(evlun(i), evfiles(i), ftread, status)
      call fitserr("ftnopn", status, ierr)
      ! get instrume
      call ftmnhd(evlun(i), binary_tbl, "EVENTS", 0, status)
      call fitserr("ftmnhd", status, ierr)
      call ftgkys(evlun(i),'INSTRUME',instrum(i),comment,status)
      call fitserr("ftgkys", status, ierr)
      if (instrum(i) == "merged") then
        call ftgkys(evlun(i),'INSTRUM1',instrum(i),comment,status)
        call fitserr("ftgkys", status, ierr)
        if (status > 0) then
          ierr = 3
          call wrerr("opn_events", ierr, "Can't read INSTRUME keyword")
          status = 0
        end if
      end if
      instrum(i) = instrum(i)(1:len_trim(instrum(i))-1)
    end do
    if (status > 0) then
      ierr = 3
      call wrerr("opn_events", ierr, "Can't open event file")
  return
    end if
 
    ! which cameras are in which input file: 
    ! search for DEADCOR extensions of each telescope 
    ncamarr = 0
    do j = 1, N_TELESCOPES
      write (extnam,'(''DEADCOR'', i1)') j
      tmAlreadyFound = .false. ! the default.
      do i = 1, nfiles
        status = 0
        call ftmnhd(evlun(i), binary_tbl, extnam, 0, status)
        if ( status == 0 ) then
          if (tmAlreadyFound) then
            write (tmStr, '(i1)') j
            write (file1Str, '(i4)') i
            write (file2Str, '(i4)') fileInWhichPreviousTMFound
            ierr = 3
            call wrerr("opn_events", ierr, "Extensions for TM "//trim(adjustl(tmStr))&
              //" were found in both event lists "//trim(adjustl(file1Str))//" and "&
              //trim(adjustl(file2Str))//". Duplication of this sort is not permitted.")
  return
          end if

          tmAlreadyFound = .true.
          fileInWhichPreviousTMFound = i

          ncamarr = ncamarr+1
          camarr(ncamarr) = j
          camlun(ncamarr) = evlun(i)
          write(instarr(ncamarr),'(a,i1)') trim(instrum(i)), j 
        end if
      end do
    end do
    if (ncamarr == 0) then
      ierr = 3
      call wrerr("opn_events", ierr, "No DEADCOR extensions found")
      return
    end if
 
  end subroutine opn_events

!----------------------------------------------------------------------------
  subroutine get_gti(ic, clun, gtitype, ngti, tstart, tend, ierr)
!----------------------------------------------------------------------------
    ! Used by apetool.

    integer, intent(in) :: ic, clun
    integer, intent(out) :: ngti
    real(real64), dimension(:), pointer :: tstart, tend  ! gti arrays
    integer, intent(out) :: ierr

    character(len=*), intent(in) :: gtitype
    character(len=9) :: gtinam
    integer :: status
    integer :: tstartcol, tendcol
    integer :: igti, ival
    logical :: undef
    integer,parameter :: nblock = 7200

    status = 0

    !if (gtitype /= "FLAREGTI") then
    write (gtinam, '(a, i1)') trim(gtitype), ic
    !else
    !  gtinam=gtitype
    !end if
    call ftmnhd(clun, binary_tbl, gtinam, 0, status)
    if (status > 0) then
      ierr = 1
      call wrerr("get_gti", ierr, "Missing " // trim(gtinam) // " extension")
  return
    end if 
    call ftgnrw(clun, ngti, status)
    call fitserr("ftgnrw", status, ierr)
    call ftgcno(clun, casesens, "START", tstartcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=START)")
    call ftgcno(clun, casesens, "STOP", tendcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=STOP)")

    allocate (tstart(ngti), tend(ngti))

    do igti = 1, ngti, nblock
      ival = min(nblock,ngti-igti+1) 
      call ftgcvd(clun, tstartcol, igti, 1, ival, nullval, &
                  tstart(igti:igti+ival-1), undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, tendcol, igti, 1, ival, nullval, &
                  tend(igti:igti+ival-1), undef, status)
      call fitserr("ftgcvd", status, ierr)
    end do

    if (status > 0) then
      ierr = 3
      call wrerr("get_gti", ierr, "Error reading GTI extension")
  return
    end if 
    
    do igti = 1, ngti
      if (tend(igti)-tstart(igti) > 1d7) then
         ierr = 3
         call wrerr("get_gti", ierr, "Wrong gti")
    exit 
      end if
    end do 

  end subroutine get_gti

!----------------------------------------------------------------------------
  subroutine merge_gti(gti, num_gti, gti_out, merge_type, ierr)
!----------------------------------------------------------------------------
    ! Not used in expmap any more but still used in apetool and flaregti.

  type(gti_t), dimension(:), intent(in) :: gti
  integer, intent(in) :: num_gti
  type(gti_t), intent(out) :: gti_out         
  integer, intent(in) :: merge_type ! +1: union; -1: intersection
  integer, intent(out) :: ierr

  real(kind=real64), dimension(:), allocatable :: time_list
  integer, dimension(:), allocatable :: plusminus, isort
  integer :: i, igti, ngti_sum, nlist, index, add, add_old, sla_err, ngti

  ierr = 0

  ngti_sum=2*sum(gti%ngti)
  allocate (time_list(ngti_sum), plusminus(ngti_sum), isort(ngti_sum))

  index=1
  do i=1,num_gti
    time_list(index:index+gti(i)%ngti-1)=gti(i)%start(1:gti(i)%ngti)
    plusminus(index:index+gti(i)%ngti-1)=1
    index=index+gti(i)%ngti
    time_list(index:index+gti(i)%ngti-1)=gti(i)%stop(1:gti(i)%ngti)
    plusminus(index:index+gti(i)%ngti-1)=-1
    index=index+gti(i)%ngti
  end do
  nlist=index-1

  call dpsort(time_list,nlist,isort,1,sla_err)
!***  ierr = sla_err?

  ! determine required size for gti_out
  ngti=0
  do i=1,num_gti
     ngti=ngti+gti(i)%ngti
  end do

  call gti_out%alloc(ngti)

  ! Need to understand this again (20140326) ===>
  add=0
  igti=0      
  do i=1,nlist
    add_old=add 
    add=add+plusminus(isort(i))    
    if (merge_type>0) then
      if (add>0.and.add_old==0) then
        igti=igti+1
        gti_out%start(igti)=time_list(isort(i)) 
      end if
      if (add==0.and.add_old>0) then
        gti_out%stop(igti)=time_list(isort(i)) 
      end if
    else
      if (add==num_gti.and.add_old<num_gti) then
        igti=igti+1
        gti_out%start(igti)=time_list(isort(i))
      end if
      if (add<num_gti.and.add_old==num_gti) then
        gti_out%stop(igti)=time_list(isort(i))
      end if
    end if
  end do
  gti_out%ngti=igti  

  deallocate(time_list, plusminus, isort)

  end subroutine merge_gti

!----------------------------------------------------------------------------
  subroutine get_attitude(ic, clun, atts, ierr)
!----------------------------------------------------------------------------
    ! Used in flaregti.
    !
    ! Note that the calling routine should deallocate atts when it is finished with it.

    implicit none

    integer,                               intent(in) :: ic, clun
    type(tAtt), dimension(:), allocatable, intent(out) :: atts
    integer,                               intent(out) :: ierr

    character(len=8) :: attnam
    integer :: status
    integer :: timcol, racol, deccol, rollcol
    integer :: natt,iatt,ival
    logical :: undef
    integer,parameter :: nblock = 3200

    status = 0
    ierr = 0

    write (attnam, '(''CORRATT'', i1)') ic
    call ftmnhd(clun, binary_tbl, attnam, 0, status)
    if (status > 0) then
      ierr = 1
      call wrerr("get_attitude", ierr, "Missing " // trim(attnam) // " extension")
      allocate (atts(0))
  return
    end if 

    call ftgnrw(clun, natt, status)
    call fitserr("ftgnrw", status, ierr)
    call ftgcno(clun, casesens, "TIME", timcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=TIME)")
    call ftgcno(clun, casesens, "RA", racol, status)
    call fitserr("ftgcno", status, ierr, " (colname=RA)")
    call ftgcno(clun, casesens, "DEC", deccol, status)
    call fitserr("ftgcno", status, ierr, " (colname=DEC)")
    call ftgcno(clun, casesens, "ROLL", rollcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=ROLL)")

    allocate (atts(natt))

    do iatt = 1, natt, nblock
      ival = min(nblock,natt-iatt+1)
      call ftgcvd(clun, timcol, iatt, 1, ival, nullval, &
                  atts(iatt:iatt+ival-1)%time, undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, racol, iatt, 1, ival, nullval, &
                  atts(iatt:iatt+ival-1)%ra, undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, deccol, iatt, 1, ival, nullval, &
                  atts(iatt:iatt+ival-1)%dec, undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, rollcol, iatt, 1, ival, nullval, &
                  atts(iatt:iatt+ival-1)%roll, undef, status)
      call fitserr("ftgcvd", status, ierr)
    end do

  end subroutine get_attitude

!----------------------------------------------------------------------------
  subroutine get_attitude_old(ic, clun, natt, atime, ra, dec, roll, ierr)
!----------------------------------------------------------------------------
    ! 20200828 IMS: used by apetool.
    integer, intent(in) :: ic, clun
    integer, intent(out) :: natt
    real(real64), dimension(:), pointer :: atime, ra, dec, roll
    integer, intent(out) :: ierr
    character(len=8) :: attnam
    integer :: status
    integer :: timcol, racol, deccol, rollcol
    integer :: iatt, ival
    logical :: undef
    integer,parameter :: nblock = 3200

    status = 0

    write (attnam, '(''CORRATT'', i1)') ic
    call ftmnhd(clun, binary_tbl, attnam, 0, status)
    if (status > 0) then
      ierr = 1
      call wrerr("get_attitude", ierr, "Missing " // trim(attnam) // " extension")
  return
    end if 

    call ftgnrw(clun, natt, status)
    call fitserr("ftgnrw", status, ierr)
    call ftgcno(clun, casesens, "TIME", timcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=TIME)")
    call ftgcno(clun, casesens, "RA", racol, status)
    call fitserr("ftgcno", status, ierr, " (colname=RA)")
    call ftgcno(clun, casesens, "DEC", deccol, status)
    call fitserr("ftgcno", status, ierr, " (colname=DEC)")
    call ftgcno(clun, casesens, "ROLL", rollcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=ROLL)")

    allocate (atime(natt), ra(natt), dec(natt), roll(natt))

    do iatt = 1, natt, nblock
      ival = min(nblock,natt-iatt+1)
      call ftgcvd(clun, timcol, iatt, 1, ival, nullval, &
                  atime(iatt:iatt+ival-1), undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, racol, iatt, 1, ival, nullval, &
                  ra(iatt:iatt+ival-1), undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, deccol, iatt, 1, ival, nullval, &
                  dec(iatt:iatt+ival-1), undef, status)
      call fitserr("ftgcvd", status, ierr)
      call ftgcvd(clun, rollcol, iatt, 1, ival, nullval, &
                  roll(iatt:iatt+ival-1), undef, status)
      call fitserr("ftgcvd", status, ierr)
    end do

  end subroutine get_attitude_old

!----------------------------------------------------------------------------
  subroutine linint_att2(gtime, atts, xatt, yatt, iatime, xatt1, yatt1, ierr)
!----------------------------------------------------------------------------
!*** kept because used by flaregti.

    real(real64), intent(in) :: gtime        ! time (midpoint of bin) for which
                                             !   attitude is to be interpolated 
    type(tAtt), dimension(:), intent(in) :: atts ! times from attitude extension
    real(real64), dimension(size(atts)), intent(in) :: xatt, yatt ! attitude in image pixels
    integer, intent(inout) :: iatime         ! index to current attitude 
                                             !   table column
    real(real64), intent(out) :: xatt1, yatt1    ! attitude at time gtime
    integer, intent(out) :: ierr

    integer :: natts
    real(real64), parameter :: delta = 2.5d0
    real(real64) :: bfrac

    ierr = 0
    xatt1 = 0.0
    yatt1 = 0.0

    natts = size(atts)
    if (iatime >= natts) then
      ierr = 3
  return
    end if

    do while (iatime < natts-1 .and. atts(iatime+1)%time < gtime) 
      iatime = iatime+1
    end do

    if (gtime > atts(iatime)%time-delta .and. gtime < atts(iatime+1)%time+delta) then
      bfrac = (gtime-atts(iatime)%time)/(atts(iatime+1)%time-atts(iatime)%time)
      xatt1 = xatt(iatime) + bfrac*(xatt(iatime+1)-xatt(iatime))
      yatt1 = yatt(iatime) + bfrac*(yatt(iatime+1)-yatt(iatime))
    else
      ierr = 2
  return
    end if

    if (gtime>atts(natts)%time .or. gtime<atts(1)%time) then
      ierr = 1 
    end if

  end subroutine linint_att2

!----------------------------------------------------------------------------
  function process_caldb_status(caldb_status, nret, tel_id, function_name&
    , cal_code, ib) result(ierr)
!----------------------------------------------------------------------------
    implicit none

    integer,           intent(in) :: caldb_status,nret,tel_id
    character(*),      intent(in) :: function_name,cal_code
    integer, optional, intent(in) :: ib
    integer :: ierr

    character(len=80) :: line

    ierr = 0 ! default

    if (caldb_status /= 0 .and. caldb_status /= 1) then ! caldb_status==1 means nret<1, which we deal with below.
      if(present(ib)) then
        write (line, '(a,i1,a,i1)') "Error finding cal component "&
          //trim(cal_code)//" for telescope ", tel_id, ", band ", ib
      else
        write (line, '(a,i1,a,i1)') "Error finding cal component "&
          //trim(cal_code)//" for telescope ", tel_id, "."
      end if
      call wrerr(function_name, 3, line)
      ierr = 3
  return
    end if

    if (nret > 1) then 
      call wrerr(function_name, 1, "More than one "//trim(cal_code)&
        //" found; using first.")
    end if

    if (nret < 1) then
      if(present(ib)) then
        write (line, '(a,i1,a,i1)') "Could find no cal component "&
          //trim(cal_code)//" for telescope ", tel_id, ", band ", ib
      else
        write (line, '(a,i1,a,i1)') "Could find no cal component "&
          //trim(cal_code)//" for telescope ", tel_id, "."
      end if
      call wrerr(function_name, 3, line)
      ierr = 3
  return
    end if

  end function process_caldb_status


!----------------------------------------------------------------------------
  subroutine get_vignetting(ic, instrume, nb, e_min, e_max, date_obs, date_end, theta, vig, ierr)
!----------------------------------------------------------------------------
    ! 20201116 IMS: Used by apetool.

    integer, intent(in) :: ic
    character(len=30), intent(in) :: instrume 
    integer, intent(in) :: nb
    real, dimension(:) :: e_min, e_max
    character(len=19), intent(in) :: date_obs, date_end
    real, dimension(:), pointer :: theta
    real, dimension(:,:), pointer :: vig
    integer, intent(out) :: ierr

    character(len=30), parameter :: telescop = "eROSITA", codenam = "TVIGNET"
    character(len=30), parameter :: filter = "-", detnam = "-"
    character(len=256) :: expression = "-"

    integer, parameter :: group = 1, maxret = 100 
    integer, dimension(maxret) :: extno
    character(len=256), dimension(maxret) :: filename , online
    character(len=10) :: strtdate, strtime, stpdate, stptime

    integer :: nelo, nehi, ntheta
    integer :: i, j, unit, status, chatter, nret, nfound, nrow
    integer :: elocol, ehicol, thetacol, vigcol
    integer :: naxis
    integer, parameter :: maxdim = 3
    integer, dimension(maxdim) :: naxes
    real, dimension(:), allocatable :: elo, ehi
    real, dimension(:,:), allocatable :: vignet
    real :: e_mean

    real, parameter :: nullval = -1.0
    logical :: undef

    ! look for vignetting datasets in caldb
    call prep_gtcalf(date_obs, date_end, strtdate, strtime, &
                     stpdate, stptime, chatter)
    call wrchatty("get_vignetting", "Searching CALDB for component codename "//trim(codenam))
    call gtcalf (chatter, telescop, instrume, detnam, filter, &
                codenam, strtdate, strtime, stpdate, stptime, &
                expression, maxret, filename, extno, online, &
                nret, nfound, status) 

    ierr = process_caldb_status(status, nret, ic, "get_vignetting", codenam)!, ib)
    if(ierr/=0)&
  return

    ! open vignetting file
    status = 0
    call ftgiou(unit, status)
    call fitserr("ftgiou", status, ierr)
    call fttopn(unit, filename, ftread, status)
    call fitserr("fttopn", status, ierr)

    ! get table info
    call ftgnrw(unit, nrow, status)
    call fitserr("ftgnrw", status, ierr)
    call ftgcno(unit, casesens, "ENERG_LO", elocol, status)
    call fitserr("ftgcno", status, ierr, " (colname=ENERG_LO)")
    call ftgcno(unit, casesens, "ENERG_HI", ehicol, status)
    call fitserr("ftgcno", status, ierr, " (colname=ENERG_HI)")
    call ftgcno(unit, casesens, "THETA", thetacol, status)
    call fitserr("ftgcno", status, ierr, " (colname=THETA)")
    call ftgcno(unit, casesens, "VIGNET", vigcol, status)
    call fitserr("ftgcno", status, ierr, " (colname=VIGNET)")

    ! read vector columns
    call ftgtdm(unit, elocol, maxdim, naxis, naxes, status)
    call fitserr("ftgtdm", status, ierr)
    nelo = naxes(1)
    allocate (elo(nelo))
    call ftgcve(unit, elocol, 1, 1, naxes(1), nullval, elo, undef, status) 
    call fitserr("ftgcve", status, ierr)

    call ftgtdm(unit, ehicol, maxdim, naxis, naxes, status)
    call fitserr("ftgtdm", status, ierr)
    nehi = naxes(1)
    allocate (ehi(nehi))
    call ftgcve(unit, ehicol, 1, 1, naxes(1), nullval, ehi, undef, status) 
    call fitserr("ftgcve", status, ierr)

    call ftgtdm(unit, thetacol, maxdim, naxis, naxes, status)
    call fitserr("ftgtdm", status, ierr)
    ntheta = naxes(1) 
    allocate (theta(ntheta))
    call ftgcve(unit, thetacol, 1, 1, naxes(1), nullval, theta, undef, status)
    call fitserr("ftgcve", status, ierr)

    call ftgtdm(unit, vigcol, maxdim, naxis, naxes, status)
    call fitserr("ftgtdm", status, ierr)
    if ( nelo /= nehi .or. nelo /= naxes(1) .or. ntheta /= naxes(2) .or. 1 /= naxes(3) ) &
      call wrerr("get_vignetting", 3, "Mismatch of vignetting values")
    allocate (vignet(nelo,ntheta))
    call ftgcve(unit, vigcol, 1, 1, naxes(1)*naxes(2), nullval, vignet, undef, status) 
    call fitserr("ftgcve", status, ierr)

    call ftclos(unit, status)
    call fitserr("ftclos", status, ierr)

    if (status > 0) then
      call wrerr("get_vignetting", 3, "Could not read vignetting file")
      return
    end if 

    allocate (vig(nb,ntheta))

    ! select appropriate energy bins
    do i=1, nb
      e_mean = 0.5 * (e_min(i)+e_max(i))
      do j=1,nelo
        if (e_mean >= elo(j) .and. e_mean <= ehi(j))  exit
      end do
      if (j <= nelo) then
        vig(i,:) = vignet(j,:) 
      else
        ierr = 3
        call wrerr("get_vignetting", ierr, "No vignetting for energy band found")
      end if
    end do 

    deallocate (vignet, elo, ehi)
       
  end subroutine get_vignetting
  
!----------------------------------------------------------------------------
  subroutine get_evt_keywords(lun, date_obs, date_end, flagsel, filter, patsel)
!----------------------------------------------------------------------------
    ! Use in apetool.

    integer,           intent(in) :: lun
    character(len=19), intent(out) :: date_obs, date_end
    integer,           intent(out) :: flagsel
    character(len=30), intent(out) :: filter
    integer, optional, intent(out) :: patsel

    character(*), parameter :: functionName = "get_evt_keywords"
    integer :: ierr, status
    character(len=60) :: comment
    integer :: dummy_hdu_type

    ! Try to move to the PHDU.
    !
    status = 0
    call FTMAHD(lun, 1, dummy_hdu_type, status)
    call report_FITS_err(functionName, status, ierr, "- couldn't move to PHDU.")

    call ftgkys(lun, "DATE-OBS", date_obs, comment, status)
    if(status/=0) then
      status=0 ! IMS: this seems to be necessary.
      call FTMAHD(lun, 2, dummy_hdu_type, status) ! move to 2nd HDU.
      call report_FITS_err(functionName, status, ierr, "- couldn't move to HDU 2.")
      call ftgkys(lun, "DATE-OBS", date_obs, comment, status)
      call report_FITS_err(functionName, status, ierr, "- couldn't read keyword 'DATE-OBS'.")
    end if

    call fitserr("ftgkys", status, ierr)
    call ftgkys(lun, "DATE-END", date_end, comment, status)
    call fitserr("ftgkys", status, ierr)
    call ftgkyj(lun, "FLAGSEL", flagsel, comment, status)
    if (status /= 0) then
      call wrlog(functionName,"FLAGSEL keyword not found; set to 0.")
      flagsel=0
      status=0
    end if
    call ftgkys(lun, "FILTER", filter, comment, status)
    call fitserr("ftgkys", status, ierr)
    if (trim(adjustl(filter)) == "FILTER" .or. trim(adjustl(filter)) == "filter") then
      filter = "FILTER"
    else if(trim(adjustl(filter)) == "OPEN" .or. trim(adjustl(filter)) == "open") then
      filter = "OPEN"
    else
      call wrerr(functionName,2,"Unrecognized value '"//trim(filter)&
        //"' of FILTER keyword. The CALDB will return you files for all "&
        //"filters, which may not be what you want.")
      filter ="-"
    end if

    if(present(patsel)) then
      call ftgkyj(lun, "PAT_SEL", patsel, comment, status)
      if (status /= 0) then
        call wrlog(functionName,"PAT_SEL keyword not found; set to 0.")
        patsel=0
        status=0
      end if
    end if

  end subroutine get_evt_keywords 

end module expmap_legacy
