! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module expmap_mod

  !----------------------------------------------------------------------------
  ! Some notes about OpenMP in expmap (and eSASS)  [JSS]
  !
  !  * Functions called simultaneously in different threads need to
  !    be marked as *recursive* (due to the -fno-automatic used to compile
  !    eSASS), or their stacks will be clobbered, or
  !  * non-recursive functions must be called in a (named) critical section,
  !    ensuring that they are only called once in each thread at one time.
  !
  ! Threads wait if they hit a critical section and another thread got
  ! there first. fitsio, parameter access and calibration calls need to 
  ! be in critical sections as they are not thread-safe.
  !
  ! The wrschatty, wrslog and wrserr are special versions of the logging
  ! routines which wrap their calls in critical sections allowing them to
  ! be run in multiple threads simultaneously.
  !
  ! We also use critical sections to avoid variables being written to
  ! simultaneously in different threads (other options are available - 
  ! see OpenMP docs).
  !----------------------------------------------------------------------------

  use expmap_trans_mod
  use expmap_io
  use radec2pix_mod
  use fitsimage, only: shapehdr_type, stdhdr_type, wcs_imahdr_type, wcs_refhdr_type
  use srctool_generic, only: wrschatty, wrslog, wrserr

  use param
  use iso_fortran_env

  implicit none

logical, private, parameter :: doTest=.false.

  real(real64), parameter :: ftime = 0.05d0 ! frame time
!**** 20210428 IMS: switch to use the value in eroib_general.

  real(real64), parameter :: dt_chunk = ftime*ndead 
                                         ! interval for MIPS counter read out
  real, parameter :: diagdet = 1.46      ! field size for rolled detector maps
                                         !   in degrees
  real, parameter :: scansp = 0.005      ! max scan speed in deg/frame_time
                                         ! 4 hour scans: 0.00125
  real, parameter :: dxcen = 192.5, &    ! x/yraw center of detector map
                     dycen = 192.5       !

  !----------------------------------------------------------------------------
  ! Parameters of output images
  !----------------------------------------------------------------------------
  type imgpar_t
    type(shapehdr_type) :: shapehdr
    type(stdhdr_type) :: stdhdr
    type(wcs_refhdr_type) :: refhdr
    type(wcs_imahdr_type) :: wcshdr
    real(real64) :: proj_ra_deg, proj_dec_deg, proj_pix_x, proj_pix_y& ! WCS for sky images
                    , img_pixel_size_deg_x, img_pixel_size_deg_y       !
    integer :: size_xi, size_yi              ! size of sky images
    integer :: patch_sizes_i(2) ! size of detroll, detshift images
    real(real64) :: proj_ra, proj_dec, cos_dec_proj,sin_dec_proj
real :: rcen
    real, dimension(2) :: patch_ctr
  end type imgpar_t
  !----------------------------------------------------------------------------


  !----------------------------------------------------------------------------
  ! Input state for computing exposure map for a TM
  !----------------------------------------------------------------------------
  type tm_t
    character(len=30) :: inst                ! name of TM (TMx)
    integer :: tmi

    real, allocatable, dimension(:,:,:) :: detmap ! detector map for TM
    real, allocatable, dimension(:,:,:) :: camexp ! exposure map for TM
    type(badpixt), allocatable, dimension(:) :: badpix ! bad pixel map for TM

    integer :: flagsel                       ! FLAGSEL keyword
    integer :: patsel                        ! PAT_SEL keyword

    type(gti_t) :: gtis
    type(gti_t) :: ogtis ! start and stop times in output GTI

    type(tAtt), dimension(:), allocatable :: atts  ! attitudes for TM

    type(tDeadCorr), dimension(:), allocatable :: deadcs ! deadcorr data
    real, dimension(:), allocatable :: weights            ! area weight array

    real, dimension(2) :: img_pixel_sizes_ccd ! ratio of sky to detector pixel sizes.
    real(real64), dimension(2) :: img_pixel_size_tan ! Size of image pixels in tangent measure (essentially just radians)
    real(real64), dimension(2) :: ccd_pixel_size_tan ! Size of CCD pixels in tangent measure (essentially just radians)
    real(real64), dimension(2) :: optic_axis_ccd
  end type tm_t
  !----------------------------------------------------------------------------

  public :: imgpar_t, tm_t
  public :: get_imgpar, init_tm, dealloc_tm, expmap_tm

contains

  !----------------------------------------------------------------------------
  ! Get the parameters for the output images
  !----------------------------------------------------------------------------
  subroutine get_imgpar(templateimage, ip)
  !----------------------------------------------------------------------------
    character(*),   intent(in)  :: templateimage
    type(imgpar_t), intent(out) :: ip    ! output image parameters

    real :: frac_x, frac_y
    integer :: half_patch_size_xi, half_patch_size_yi

    call wrchatty ("get_imgpar", "Calling get_ima_keywords...")
    call get_ima_keywords (templateimage, ip%shapehdr, ip%stdhdr, ip%refhdr, ip%wcshdr)

    ip%proj_ra_deg          = ip%wcshdr%crval1
    ip%proj_dec_deg         = ip%wcshdr%crval2
    ip%proj_pix_x           = ip%wcshdr%crpix1
    ip%proj_pix_y           = ip%wcshdr%crpix2
    ip%img_pixel_size_deg_x = ip%wcshdr%cdelt1
    ip%img_pixel_size_deg_y = ip%wcshdr%cdelt2
    ip%size_xi              = ip%shapehdr%naxis1
    ip%size_yi              = ip%shapehdr%naxis2

    ! Here we calculate the size of the detroll and detshift 'patch'
    ! arrays. Note that detroll formally speaking doesn't need the extra
    ! room for 20 frames of scanning, but this is a small fractional
    ! addition, so it simplifies things to keep detroll and detshift the
    ! same dimensions.
    !
    ip%patch_sizes_i(1) = nint((diagdet+scansp*ndead)&
      /min(abs(ip%img_pixel_size_deg_x),abs(ip%img_pixel_size_deg_y)))+3
    ip%patch_sizes_i(2) = ip%patch_sizes_i(1)

    ip%proj_ra  = ip%proj_ra_deg *deg2rad
    ip%proj_dec = ip%proj_dec_deg*deg2rad
    ip%cos_dec_proj = cos(ip%proj_dec)
    ip%sin_dec_proj = sin(ip%proj_dec)

ip%rcen = real(ip%patch_sizes_i(1))/2.0+0.5

    ! Note that it is desirable to align detshift and camexp such that
    ! they match pixel to pixel (albeit they have different numbers of
    ! pixels) when the attitude shift is zero. Since camex is given the
    ! same dimensions and WCS as the input template image, we look to
    ! the latter. We obtain the fractional remainder of the WCS CRPIX1/2
    ! values and add this to a suitable integer to obtain the real-
    ! valued centre coordinates of detroll/detshift.
    !
    ! Note that we adhere to the WCS convention that the centre of the
    ! 1st pixel has coordinate value 1.0; thus the leading edge of that
    ! pixel (since image pixels are by definition of width 1 in pixel
    ! coordinates) has coordinate value 0.5, and the trailing edge, 1.5.
    !
    ! An example: suppose CRPIX1==10.2. Since 10.0 would be the middle
    ! of the 10th pixel (starting counting at 1), 10.2 is 70% across the
    ! 10th pixel. Then frac_x evaluates to 0.2. Suppose further that we
    ! choose a size of 4 pixels for the X axis of detroll etc. (Of
    ! course in reality the sizes are usually far greater.) Then
    ! half_patch_size_xi evaluates to ceiling(4*0.5) = 2. The final
    ! chosen patch_ctr_x value is 2.2, which is 70% across the 2nd pixel
    ! of detroll/detshift. The pixels will align at zero attitude offset.
    !
    frac_x = real(mod(ip%proj_pix_x, real(1.0, kind(ip%proj_pix_x))))
    frac_y = real(mod(ip%proj_pix_y, real(1.0, kind(ip%proj_pix_y))))
    half_patch_size_xi = ceiling(ip%patch_sizes_i(1)*0.5)
    half_patch_size_yi = ceiling(ip%patch_sizes_i(2)*0.5)
    ip%patch_ctr(1) = half_patch_size_xi + frac_x
    ip%patch_ctr(2) = half_patch_size_yi + frac_y

  end subroutine get_imgpar
  !----------------------------------------------------------------------------

  !----------------------------------------------------------------------------
  ! Initialize TM object, loading calibration data
  !----------------------------------------------------------------------------
  subroutine init_tm(tm, ilist, ci, task_info, ipar)
    type(tm_t),            intent(out)   :: tm
    integer,               intent(in)    :: ilist         ! index into TM lists
    integer,               intent(in)    :: ci            ! the TM number
    type(t_taskinfo),      intent(inout) :: task_info
    type(imgpar_t),        intent(in)    :: ipar          ! image parameters

    character(*), parameter :: functionName="init_tm"
    integer :: ierr=0
    character(len=19) :: date_obs, date_end
    character(len=30) :: filter
    real(real64) ::ccd_pixel_size_deg_x, ccd_pixel_size_deg_y      ! plate scale of detector pixels
                                                                   !   (from WCS)
    real(real64) :: xraw_ax, yraw_ax         ! position of optical axis on detector
                                             !   in RAWXR/RAWYR coordinates (WCS)
    integer :: natt
    type(gti_t) :: egtis


    tm%tmi = ci

    tm%inst  = task_info%tm_info(ci)%instrum !***** check lengths the same.
    date_obs = task_info%tm_info(ci)%date_obs !***** check lengths the same.
    date_end = task_info%tm_info(ci)%date_end !***** check lengths the same.
    filter   = task_info%tm_info(ci)%filter !***** check lengths the same.

    tm%flagsel = task_info%tm_info(ci)%flagsel
    tm%patsel  = task_info%tm_info(ci)%patsel

    ! allocation of arrays - ordering of dimensions to optimize cache access
    call wrschatty (functionName, "Allocating arrays...")
    allocate (tm%camexp(ipar%size_xi,task_info%n_bands,ipar%size_yi))
    allocate (tm%detmap(N_PIXELS_CCD,N_PIXELS_CCD,task_info%n_bands))
    allocate (tm%weights(task_info%n_bands))

    call wrschatty (functionName,  trim(tm%inst)//": Calling get_detmaps...")
    call get_detmaps(ci, tm%inst, task_info%n_bands&
      , task_info%params%e_min, task_info%params%e_max, date_obs, date_end, tm%flagsel&
      , tm%patsel, tm%detmap, xraw_ax, yraw_ax, ccd_pixel_size_deg_x, ccd_pixel_size_deg_y, ierr)
    ! Note that xraw_ax, yraw_ax give the position of the optic axis
    ! in CCD pixels, and ccd_pixel_size_deg_x, ccd_pixel_size_deg_y
    ! the size of CCD pixels in degrees.

    if (ierr == 3)&
call wrstop(functionName, trim(tm%inst)//": Obtaining detector maps failed")

    tm%optic_axis_ccd(:) = [xraw_ax,yraw_ax]

    if (.not. task_info%params%withdetmaps) tm%detmap=1.

    if (task_info%params%withfilebadpix) then
      tm%badpix = task_info%tm_info(ci)%badpix
      deallocate(task_info%tm_info(ci)%badpix)
    end if

    call wrschatty (functionName, trim(tm%inst)//": Calling get_area...")
    call get_area(ci, task_info%n_bands, tm%inst, date_obs, date_end&
      , task_info%params%e_min, task_info%params%e_max, filter, tm%weights, ierr)
    if (ierr == 3)&
call wrstop (functionName, trim(tm%inst)//": Obtaining area failed")
    if (.not. task_info%params%withweights) tm%weights=1./float(task_info%n_TMs)

!==== tabulated vignetting =========================
!     call get_vignetting(ic, inst, task_info%n_bands, task_info%params%e_min, task_info%params%e_max, date_obs, date_end, &
!                         theta, vig, ierr)
!===================================================

!==== parametric vignetting ========================
    if (task_info%params%withvig) then
      call wrschatty (functionName, trim(tm%inst)//": Calling get_vignetting2...")
      call get_vignetting2 (tm%inst, task_info%n_bands&
        , task_info%params%e_min, task_info%params%e_max, date_obs, date_end&
        , task_info%params%plindex, tm%detmap, ierr)
      if (ierr == 3)&
call wrstop (functionName, trim(tm%inst)//": Obtaining vignetting failed")
    end if
!===================================================

!==== tabulated vignetting ========================
!     if (task_info%params%withvig) then
!       call make_vigmaps(theta, vig, detmap,  &
!                         xraw_ax, yraw_ax, ccd_pixel_size_deg_x, ccd_pixel_size_deg_y, ierr)
!       if (ierr == 3) call wrstop(functionName, "Creation of vignetted detector maps failed")
!     end if
!===================================================

    tm%gtis = task_info%tm_info(ci)%gtis
    call task_info%tm_info(ci)%gtis%dealloc()

    tm%atts = task_info%tm_info(ci)%atts
    deallocate(task_info%tm_info(ci)%atts)

    natt = size(tm%atts)

    ! Convert attitude samples into positions in image pixel coordinates.
    !
    call wrschatty (functionName,  trim(tm%inst)//": Calling radec2xy...")
    call radec2pix(tm%atts%ra, tm%atts%dec, natt, ipar%proj_ra_deg, ipar%proj_dec_deg&
      , ipar%proj_pix_x, ipar%proj_pix_y, ipar%img_pixel_size_deg_x, ipar%img_pixel_size_deg_y&
      , tm%atts%x_pix, tm%atts%y_pix)

    tm%deadcs = task_info%tm_info(ci)%deadcs
    deallocate(task_info%tm_info(ci)%deadcs)

    tm%ogtis = tm%gtis
    if (task_info%params%withinmaps) then
      call wrschatty (functionName,  trim(tm%inst)//": Calling get_expmaps (WITHINMAPS)...")
      call get_expmaps(ilist, ci, task_info%n_bands&
        , task_info%params%exp_in_files, ipar%size_xi, ipar%size_yi&
        , tm%camexp, egtis, ierr)
      if (ierr == 3)&
call wrstop (functionName, trim(tm%inst)//": Reading of input exposure maps failed")

      ! merge GTI of old exposure map and new event file
      call wrschatty (functionName,  trim(tm%inst)//": Calling merge_gti...")
      call tm%ogtis%or_with(egtis)

      ! select new GTI, not yet included in exposure map
      call wrschatty (functionName,  trim(tm%inst)//": Calling cut_gti...")
      call tm%gtis%and_with_not(egtis)
    else
      tm%camexp = 0.
    end if
    call egtis%dealloc()

    tm%img_pixel_sizes_ccd(1) = real(abs(ipar%img_pixel_size_deg_x/ccd_pixel_size_deg_x)) ! Ratios of image pixel size and
    tm%img_pixel_sizes_ccd(2) = real(abs(ipar%img_pixel_size_deg_y/ccd_pixel_size_deg_y)) ! CCD pixel size (both in deg).
    tm%img_pixel_size_tan(1) = abs(ipar%img_pixel_size_deg_x)*PI/180.0
    tm%img_pixel_size_tan(2) = abs(ipar%img_pixel_size_deg_y)*PI/180.0
    tm%ccd_pixel_size_tan(1) = abs(ccd_pixel_size_deg_x)*PI/180.0
    tm%ccd_pixel_size_tan(2) = abs(ccd_pixel_size_deg_y)*PI/180.0

  end subroutine init_tm

  !----------------------------------------------------------------------------
  ! Dealloc allocated memory for TM object
  !----------------------------------------------------------------------------
  subroutine dealloc_tm(tm)
    class(tm_t), intent(inout) :: tm

    deallocate (tm%atts, tm%deadcs, tm%weights)
    if(allocated(tm%badpix)) deallocate(tm%badpix)
    deallocate (tm%camexp)
    call tm%gtis%dealloc()
    call tm%ogtis%dealloc()

  end subroutine dealloc_tm
  !----------------------------------------------------------------------------


  !----------------------------------------------------------------------------
  ! Compute the exposure map for one TM
  !----------------------------------------------------------------------------
  recursive subroutine expmap_tm(tm, gistart, giend, task_info, ipar, outcamexp)
  !----------------------------------------------------------------------------

    type(tm_t),             intent(in)    :: tm             ! camera
    integer,                intent(in)    :: gistart, giend ! gti range
    type(t_taskinfo),       intent(in)    :: task_info
    type(imgpar_t),         intent(in)    :: ipar           ! image parameters
    real, dimension(:,:,:), intent(inout) :: outcamexp      ! output exposure map for this tm

    character(*), parameter :: functionName="expmap_tm"

    ! Important variables:

    type(tAtt) :: att_mid ! interpolated xatt, yatt at the middle of the 20-frame time chunk
    type(tAtt), dimension(ndead) :: atts_frame ! interpolated xatt, yatt 
    logical, dimension(:), allocatable :: att_is_valid
    real :: fexp_new                     ! exposure in one frame time
    real :: deadc_mid                           ! dead time corr. value
    real :: tbin                             ! length of time bin
    real(real64) :: chunk_start_t, chunk_mid_t            ! start time, midpoint of time bin
    integer :: n_chunks                  ! nr of time bins in current gti 
                                         !   interval
    real, allocatable, dimension(:,:,:) :: baddetmap   ! detector maps with bad pixels removed (time dependent)
    real, allocatable, dimension(:,:,:) :: detroll     ! rotated detector maps in 
                                                       !   image pixels
    real, allocatable, dimension(:,:,:) :: detshift    ! shifted detector maps in 
                                                       !   image pixels
    real, allocatable, dimension(:,:,:) :: detshiftexp_old! detshift, multiplied with 
    real, allocatable, dimension(:,:,:) :: detshiftexp    !   exposure
    real, allocatable, dimension(:,:,:) :: camexp      ! accumulated exposure in this call

    integer :: n_frames_this_chunk       ! number of frame times in current
                                         !   time bin 
    integer :: iatime, idctime, istime   ! pointers to attitude & dreadc columns
                                         !   maps and shifted detector maps 
    integer, allocatable, dimension(:) :: offsets_xi, offsets_yi
                                         !   detector maps
    real :: xaxr, yaxr                       ! location of optical axis in rotated
                                             !   detector map image (detroll)          
    real(real64) :: tstart, tstop
    integer :: ti, fi, gi
    integer :: ierr_la2,ierr_la3,ierr_dc
    integer :: n_la2_err1,n_la2_err2,n_la3_err1,n_la3_err2,n_dc_err1,n_dc_err2
    character(len=81) :: line
    character(len=16) :: istr
    character(1024) :: errmsg,dstr0,dstr1
    logical :: no_overlaps
    real(real64) :: frac_time,time_lost_this_GTI
    logical :: have_done_at_least_one_full_pass,new_detroll_exists,new_detshift_exists,new_detshiftexp_exists
    integer :: n_no_overlap_frame,ti_start,ti_stop,total_n_no_overlap_chunk,total_n_no_overlap_frame

!===== tabulated vignetting ========================
! real, dimension(:), pointer :: theta
! real, dimension(:,:), pointer :: vig
!===================================================

    type(img_limits_t) :: roll_img_limits,shift_img_limits,shift_img_limits_old

    type(roll_decider_t)   :: roller
    type(shift_decider_t)  :: shifter
    type(exp_decider_t)    :: exposer

    logical :: shift_has_varied,exp_has_varied

    ! some of this code mixes real32 and real64 - is this necessary? [JSS] IMS: no.

    ! allocation of arrays - ordering of dimensions to optimize cache access
    call wrschatty (functionName, "Allocating arrays...")
    allocate (detroll(        ipar%patch_sizes_i(1),task_info%n_bands,ipar%patch_sizes_i(2))) 
    allocate (detshift(       ipar%patch_sizes_i(1),task_info%n_bands,ipar%patch_sizes_i(2)))
    allocate (detshiftexp_old(ipar%patch_sizes_i(1),task_info%n_bands,ipar%patch_sizes_i(2)))
    allocate (detshiftexp(    ipar%patch_sizes_i(1),task_info%n_bands,ipar%patch_sizes_i(2)))
    allocate (baddetmap(N_PIXELS_CCD,N_PIXELS_CCD,task_info%n_bands))
    allocate (camexp(ipar%size_xi,task_info%n_bands,ipar%size_yi))

    camexp = 0.0

    iatime = 1                         ! index for accessing attitude
    istime = 1                         ! another attitude index
    idctime = 1                        ! index for accessing dead time

    write (line, '(a, I5, a, I5, '' TSTART='', F14.2, '' TSTOP='', F14.2)') &
      trim(tm%inst)//": "//trim(task_info%params%gtitype)&
      //"=", gistart, " -", giend, tm%gtis%starts(gistart), tm%gtis%stops(giend)
    call wrslog(functionName, line)

    ! loop over gti ==================================================
    gti_loop: do gi = gistart, giend
      tstart = tm%gtis%starts(gi)
      tstop  = tm%gtis%stops(gi)

      if(tstart >= tstop) then
        call wrserr(functionName, warn, trim(tm%inst)//": GTI START>=STOP")
    cycle gti_loop
      end if

      n_chunks = ceiling((tstop-tstart)/dt_chunk)
      n_frames_this_chunk = ndead ! default
      chunk_start_t = tstart                       ! start of time step
      chunk_mid_t   = chunk_start_t + dt_chunk*0.5 ! midpoint of time step

      if(allocated(tm%badpix)) then
        ! Is only allocated if task_info%params%withfilebadpix.
        call wrschatty (functionName,  trim(tm%inst)//": Calling remove_badpix...")
        call remove_badpix (tm%badpix, tm%flagsel, tm%patsel&
          , tstart, tstop, task_info%n_bands, tm%detmap, baddetmap)
      else
        baddetmap = tm%detmap ! 20200917 IMS: we could move this outside the GTI loop, but in practice this gives only a negligible gain in processing time.
      end if

      allocate(att_is_valid(n_chunks))
      allocate(offsets_xi(n_chunks))
      allocate(offsets_yi(n_chunks))

      n_la2_err1 = 0
      n_la2_err2 = 0
      n_la3_err1 = 0
      n_la3_err2 = 0
      n_dc_err1 = 0
      n_dc_err2 = 0

      call roller%init()
      call shifter%init()
      call exposer%init()
      have_done_at_least_one_full_pass = .false.

      time_lost_this_GTI = 0.0d0
      total_n_no_overlap_chunk = 0
      total_n_no_overlap_frame = 0

      ! loop over time steps ========================================
      time_loop: do ti = 1, n_chunks ! each time step has duration=dt_chunk, which is the duration of ndead=20 frames.
        if (mod(ti,20000)==0) then
          write(line,'(F4.1,''%'')') 100.0*ti/n_chunks 
          call wrslog(functionName, line)
        end if

        att_is_valid(ti) = .false. ! default

        if (ti == n_chunks) then
          tbin = real(tstop-(tstart+(n_chunks-1)*dt_chunk)) ! last bin is shorter
          n_frames_this_chunk = nint(tbin/ftime)       ! number of frame times in last step

          if (n_frames_this_chunk == 0)&
      exit time_loop

          chunk_mid_t = tstop - 0.5*ftime*n_frames_this_chunk   ! mean time of last step
        end if

        ! Get interpolated attitude (in image pixel coords) at chunk_mid_t.
        !
        att_mid%time = chunk_mid_t
        call lininterp_att(tm%atts, iatime, att_mid, ierr_la3, do_angles=.true.)
        if (ierr_la3 == 3)&
call wrstop(functionName, trim(tm%inst)//": lininterp_att(): table interpolation error")
        if(ierr_la3==1) n_la3_err1 = n_la3_err1 + 1
        if(ierr_la3==2) n_la3_err2 = n_la3_err2 + 1

        if(ierr_la3 /= 0) then
          time_lost_this_GTI = time_lost_this_GTI + ftime*n_frames_this_chunk
      cycle time_loop ! because we don't have attitude
        end if

        ! Interpolate the attitude also for each frame in the time chunk.
        !
        atts_frame%time = chunk_start_t + ftime*([(fi,fi=1,ndead,1)] - 0.5)
        do fi = 1, n_frames_this_chunk
          call lininterp_att(tm%atts, istime, atts_frame(fi), ierr_la2)
          if (ierr_la2 == 3)&
call wrstop(functionName, trim(tm%inst)//": lininterp_att(): table interpolation error")

          if(ierr_la2==1) n_la2_err1 = n_la2_err1 + 1
          if(ierr_la2==2) n_la2_err2 = n_la2_err2 + 1
        end do

        if(any(.not.atts_frame(1:n_frames_this_chunk)%is_valid))&
          time_lost_this_GTI = time_lost_this_GTI&
            + ftime*count(.not.atts_frame(1:n_frames_this_chunk)%is_valid)

        if(all(.not.atts_frame%is_valid))&
      cycle time_loop

        att_is_valid(ti) = .true.

        ! Get interpolated deadtime at chunk_mid_t.
        !
        call linint_deadc(chunk_mid_t, tm%deadcs, idctime, deadc_mid, ierr_dc)
        if (ierr_dc == 3)&
call wrstop(functionName, trim(tm%inst)//": linint_deadc(): table interpolation error")
        if(ierr_dc==1) n_dc_err1 = n_dc_err1 + 1
        if(ierr_dc==2) n_dc_err2 = n_dc_err2 + 1

        new_detroll_exists = .false.
        new_detshift_exists = .false.
        new_detshiftexp_exists = .false.

        if(roller%need_new(att_mid)) then
          ! Returns True for the first valid att_mid of the present GTI,
          ! or if the (valid) att_mid%roll is sufficiently different
          ! from the last.
          !
          ! Note that the first valid att_mid will cause all of detroll,
          ! detshift and detshiftexp to be calculated in succession,
          ! and then for 'have_done_at_least_one_full_pass' to be set
          ! True. However, copy to camexp will not occur until the
          ! second good att_mid.

          call roll_img_limits%destroy()

          if(trim(task_info%params%accuracy)=="original") then
            call rotate_image_original(ipar%patch_sizes_i, task_info%n_bands&
              , baddetmap, att_mid%roll, tm%optic_axis_ccd, [ipar%rcen, ipar%rcen]&
              , tm%img_pixel_sizes_ccd, dxcen, dycen, detroll&
              , roll_img_limits, xaxr, yaxr)

          elseif(trim(task_info%params%accuracy)=="approximate") then
            call rotate_image_approx_slow(ipar%patch_sizes_i, task_info%n_bands&
              , baddetmap, tm%optic_axis_ccd, ipar%patch_ctr&
              , tm%img_pixel_size_tan, tm%ccd_pixel_size_tan, att_mid, ipar%proj_ra&
              , ipar%cos_dec_proj, ipar%sin_dec_proj, detroll, roll_img_limits)

          else ! "exact" should be the only choice left.
            call rotate_image_exact(ipar%patch_sizes_i, task_info%n_bands&
              , baddetmap, tm%optic_axis_ccd, ipar%patch_ctr&
              , tm%img_pixel_size_tan, tm%ccd_pixel_size_tan, att_mid, ipar%proj_ra&
              , ipar%cos_dec_proj, ipar%sin_dec_proj, detroll, roll_img_limits)
          end if

          call roller%update()

          new_detroll_exists = .true.
        end if ! end roll angle handling -------------------------------

        shift_has_varied = shifter%need_new(atts_frame, n_frames_this_chunk)
        ! I can't put this function call inside the 'if' test because
        ! although it needs to be called unilaterally, gcc does not
        ! guarantee to evaluate all tests in a logical expression.

        if(new_detroll_exists .or. shift_has_varied) then
          call shift_img_limits%destroy()

          if(trim(task_info%params%accuracy)=="original") then
            call shift_image_original(ipar%patch_sizes_i, task_info%n_bands, detroll&
              , roll_img_limits, n_frames_this_chunk, atts_frame&
              , att_mid%x_pix, att_mid%y_pix, detshift&
              , shift_img_limits, n_no_overlap_frame)

          else
            call shift_image_new(ipar%patch_sizes_i, task_info%n_bands, detroll&
              , roll_img_limits, n_frames_this_chunk, atts_frame&
              , att_mid%x_pix, att_mid%y_pix, detshift&
              , shift_img_limits, n_no_overlap_frame)
          end if

          total_n_no_overlap_frame = total_n_no_overlap_frame + n_no_overlap_frame

          call shifter%update()
          new_detshift_exists = .true.
        end if ! end of detector map shifting ------------------------

        fexp_new = real(ftime)*deadc_mid

        ! The exposure time only gets re-applied if it changed
        ! by more than 0.1% or if the shifted detector masks have changed.
        !
        exp_has_varied = exposer%need_new(fexp_new)
        ! I can't put this function call inside the 'if' test because
        ! although it needs to be called unilaterally, gcc does not
        ! guarantee to evaluate all tests in a logical expression.

        if(new_detshift_exists .or. exp_has_varied) then
          ! We do the following copy here to avoid copying detshiftexp
          ! unnecessarily immediately before the post-loop call to
          ! copy_to_camexp(). The test prevents us doing the copy
          ! before detshiftexp is set for the first time in the loop.
          !
          if(have_done_at_least_one_full_pass)&
            detshiftexp_old = detshiftexp

          ! multiply with time bin and deadc
          detshiftexp = detshift*fexp_new

          call exposer%update()
          new_detshiftexp_exists = .true.
        end if !-----------------------------------------------------

        if(trim(task_info%params%accuracy)=="original") then
          offsets_xi(ti) = nint(att_mid%x_pix - xaxr) ! Difference between attitude in the middle of the
          offsets_yi(ti) = nint(att_mid%y_pix - yaxr) ! present time step and the rotated optic axis.

        else
          offsets_xi(ti) = nint(att_mid%x_pix - ipar%patch_ctr(1))
          offsets_yi(ti) = nint(att_mid%y_pix - ipar%patch_ctr(2))
        end if

        if(new_detshiftexp_exists) then
          if(have_done_at_least_one_full_pass) then
            ! Finally, project rotated and shifted detector maps. Note
            ! that we accumulate the previous detshiftexp over all the
            ! time steps since its creation - but not for the present
            ! time step, since that is associated with the new detshiftexp.
            !
            call copy_to_camexp(ipar%patch_sizes_i, ipar%size_xi, ipar%size_yi&
              , task_info%n_bands, detshiftexp_old, shift_img_limits_old&
              , ti_start, ti-1, offsets_xi, offsets_yi, att_is_valid&
              , no_overlaps, camexp)

            if(no_overlaps)&
              total_n_no_overlap_chunk = total_n_no_overlap_chunk + 1
          end if

          ti_start = ti
          have_done_at_least_one_full_pass = .true.
          call shift_img_limits_old%destroy()
          shift_img_limits_old = shift_img_limits
!          detshiftexp_old = detshiftexp
        end if ! end of projection of detector maps ------------------

        chunk_start_t = chunk_start_t + dt_chunk
        chunk_mid_t   = chunk_mid_t   + dt_chunk
      end do time_loop !===================================================

      ! Handle last time bin.
      !
      if(have_done_at_least_one_full_pass) then
        if (n_frames_this_chunk<1) then
          ti_stop = n_chunks - 1
        else
          ti_stop = n_chunks
        end if

        call copy_to_camexp(ipar%patch_sizes_i, ipar%size_xi, ipar%size_yi&
          , task_info%n_bands, detshiftexp, shift_img_limits&
          , ti_start, ti_stop, offsets_xi, offsets_yi, att_is_valid&
          , no_overlaps, camexp)

        if(no_overlaps)&
          total_n_no_overlap_chunk = total_n_no_overlap_chunk + 1
      end if

      deallocate(offsets_xi, offsets_yi, att_is_valid)

      if(n_la2_err1>0) then
        write (istr, *) n_la2_err1
        errmsg = trim(tm%inst)&
          //": Status 'time out of range, extrapolating' occurred "&
          //trim(adjustl(istr))//" times."
        call wrserr("linint_att2", 1, errmsg)
      end if
      if(n_la2_err2>0) then
        write (istr, *) n_la2_err2
        errmsg = trim(tm%inst)&
          //": Status 'couldn't find requested time in attitude table' occurred "&
          //trim(adjustl(istr))//" times."
        call wrserr("linint_att2", 2, errmsg)
      end if

      if(n_la3_err1>0) then
        write (istr, *) n_la3_err1
        errmsg = trim(tm%inst)&
          //": Status 'time out of range, extrapolating' occurred "&
          //trim(adjustl(istr))//" times."
        call wrserr("linint_att3", 2, errmsg)
      end if
      if(n_la3_err2>0) then
        write (istr, *) n_la3_err2
        errmsg = trim(tm%inst)&
          //": Status 'couldn't find requested time in attitude table' occurred "&
          //trim(adjustl(istr))//" times."
        call wrserr("linint_att3", 2, errmsg)
      end if

      if(n_dc_err1>0) then
        write (istr, *) n_dc_err1
        errmsg = trim(tm%inst)&
          //": Status 'time out of range, extrapolating' occurred "&
          //trim(adjustl(istr))//" times."
        call wrserr("linint_deadc", 1, errmsg)
      end if
      if(n_dc_err2>0) then
        write (istr, *) n_dc_err2
        errmsg = trim(tm%inst)&
          //": Status 'couldn't find requested time in deadc table' occurred "&
          //trim(adjustl(istr))//" times."
        call wrserr("linint_deadc", 2, errmsg)
      end if

      if(total_n_no_overlap_chunk>0) then
        write (istr, *) total_n_no_overlap_chunk
        errmsg = trim(tm%inst)&
          //": No overlap of rolled and shifted CCD image occurred "&
          //trim(adjustl(istr))//" times."
        call wrserr(functionName, 1, errmsg)
      end if

      if(total_n_no_overlap_frame>0) then
        write (istr, *) total_n_no_overlap_frame
        errmsg = trim(tm%inst)&
          //": No overlap of rolled CCD image occurred "&
          //trim(adjustl(istr))//" times."
        call wrserr(functionName, 1, errmsg)
      end if

      if(time_lost_this_GTI > 0.0d0) then
        frac_time = time_lost_this_GTI/(tstop - tstart)
        write (dstr0, *) time_lost_this_GTI
        write (dstr1, *) frac_time*100.0d0
        errmsg = trim(tm%inst)//": "//trim(adjustl(dstr0))&
          //" seconds of exposure lost ("//trim(dstr1)//"% of the GTI)."
        call wrserr(functionName, 2, errmsg)
      end if
    end do gti_loop !======================================================

    ! Add computed exposure onto output. Ideally we would add a per-TM
    ! lock on this, but the overhead appears low

    !$OMP CRITICAL (addcamexp)
    outcamexp = outcamexp + camexp
    !$OMP END CRITICAL (addcamexp)

!    deallocate(roll_img_limits%rows)
!    deallocate(shift_img_limits%rows)
    call roll_img_limits%destroy()
    call shift_img_limits%destroy()
    call shift_img_limits_old%destroy()

    deallocate(detroll) 
    deallocate(detshift)
    deallocate(detshiftexp_old)
    deallocate(detshiftexp)
    deallocate(baddetmap)
    deallocate (camexp)

  end subroutine expmap_tm

end module expmap_mod
