! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module expmap_trans_mod

  use iso_fortran_env
  use erolib_general_mod
  use cal
  use expmap_io

  implicit none

logical, private, parameter :: doTest=.false.

  integer, parameter :: ndead = 20         ! MIPS counter readout every ndead 
                                         ! frames (=timestep for exposure maps)    
  type scan_t
    real(kind=real64) :: xshift=0.0d0,yshift=0.0d0
    logical :: is_valid=.false.
  end type scan_t

  type row_limits_t
    integer :: i_min,i_max
  end type row_limits_t

  type img_limits_t
    type(row_limits_t), dimension(:), allocatable :: rows
    integer :: yi_min,yi_max
  contains
    procedure :: init => init_img_limits
    procedure :: set_limits => set_lim_img_limits
    procedure :: destroy => destroy_img_limits
  end type img_limits_t

  type roll_decider_t
    logical, private :: old_roll_angle_is_valid = .false.
    real(real64), private :: roll_angle_deg_old = 0.0d0
    real(real64), private :: roll_angle_deg_new = 0.0d0
  contains
    procedure :: init => init_roll_decider_t
    procedure :: need_new => need_new_roll_decider_t
    procedure :: update => update_roll_decider_t
  end type roll_decider_t

  type shift_decider_t
    type(scan_t), private :: shift_old,shift_new
    logical, private :: old_atts_are_valid = .false.
    type(tAtt), dimension(ndead), private :: atts_old,atts_new
  contains
    procedure :: init => init_shift_decider_t
    procedure :: need_new => need_new_shift_decider_t
    procedure :: update => update_shift_decider_t
  end type shift_decider_t

  type exp_decider_t
    real, private :: fexp_old=0.0,fexp_new=0.0
    logical, private :: old_fexp_is_valid = .false.
  contains
    procedure :: init => init_exp_decider_t
    procedure :: need_new => need_new_exp_decider_t
    procedure :: update => update_exp_decider_t
  end type exp_decider_t

  real(real64), parameter :: pi=dacos(-1d0), deg2rad = pi/180d0
  integer, parameter :: n_blocks = 5         ! fraction of detmap and expmaps to fit into cpu cache

  real(real64), parameter :: droll = 0.1   ! roll angle increment in deg (should
                                           !   result in less than one pixel 
                                           !   error at the border of the fov)

  real, parameter :: dshift = 1.0          ! change of scan vector (in pixels)
                                           !  before shifted detector map gets
                                           !  recalculated 

  real, parameter :: dfexp = 1.001         ! required fractional accuracy of
                                           !   integration time step

  real(real64), parameter :: ccd_xoffset = 0.009d0 ! approximately the FOV radius in radians.

  enum, bind( C )
    enumerator :: CANNOT_DETERMINE=1, SCAN_HAS_CHANGED, SCAN_IS_UNCHANGED
  end enum

contains

  !.....................................................................
  recursive subroutine init_roll_decider_t(this)
    implicit none

    class(roll_decider_t) :: this

    this%old_roll_angle_is_valid = .false.
    this%roll_angle_deg_old = 0.0d0
    this%roll_angle_deg_new = 0.0d0

  end subroutine init_roll_decider_t

  !.....................................................................
  recursive function need_new_roll_decider_t(this, att_new) result(need_new_detroll)
    ! Returns True for the first valid att_new, or if the (valid) att_new%roll is sufficiently different from the last.
    implicit none

    class(roll_decider_t) :: this
    type(tAtt), intent(in) :: att_new
    logical :: need_new_detroll

    if(att_new%is_valid) then
      if(this%old_roll_angle_is_valid) then
        if(  abs(att_new%roll - this%roll_angle_deg_old         ) < droll& ! Have to
        .or. abs(att_new%roll - this%roll_angle_deg_old + 2.0*PI) < droll& ! cater for
        .or. abs(att_new%roll - this%roll_angle_deg_old - 2.0*PI) < droll) then
          need_new_detroll = .false.
        else
          need_new_detroll = .true.
        end if

      else ! no old valid roll angle, means this is the first call in the
           ! present loop over times, thus we need to calculate an initial detroll.
        need_new_detroll = .true.
      end if

      this%roll_angle_deg_new = att_new%roll

    else ! invalid attitude.
      need_new_detroll = .false.
    end if

  end function need_new_roll_decider_t

  !.....................................................................
  recursive subroutine update_roll_decider_t(this)
    implicit none

    class(roll_decider_t) :: this

    this%old_roll_angle_is_valid = .true.
    this%roll_angle_deg_old = this%roll_angle_deg_new

  end subroutine update_roll_decider_t

  !.....................................................................
  recursive subroutine init_shift_decider_t(this)
    implicit none

    class(shift_decider_t) :: this

    this%shift_old%is_valid = .false.
    this%old_atts_are_valid = .false.

  end subroutine init_shift_decider_t

  !.....................................................................
  recursive function need_new_shift_decider_t(this, atts_new, num_frames) result(need_new_detshift)
    implicit none

    class(shift_decider_t) :: this
    type(tAtt), dimension(ndead), intent(in) :: atts_new
    integer,                      intent(in) :: num_frames ! expected to be <=ndead
    logical :: need_new_detshift

    type(scan_t) :: shift_new
    integer :: scan_vector_status

    call compare_delta_att(this%shift_old, atts_new, num_frames, shift_new, scan_vector_status)

    need_new_detshift = .false. ! default

    ! shift detector map if n_frames_this_chunk/=ndead or
    !   roll angle or scan speed or direction changed
    !
    if(scan_vector_status/=SCAN_IS_UNCHANGED) then
      need_new_detshift = .true.
    else if(this%old_atts_are_valid .and. any(atts_new%is_valid .neqv. this%atts_old%is_valid)) then
      ! This test obviates the need to test whether num_frames/=ndead.
      need_new_detshift = .true.
    end if

    this%shift_new = shift_new
    this%atts_new(:) = atts_new(:)

  end function need_new_shift_decider_t

  !.....................................................................
  recursive subroutine update_shift_decider_t(this)
    implicit none

    class(shift_decider_t) :: this

    this%old_atts_are_valid = .true.
    this%shift_old = this%shift_new
    this%atts_old(:) = this%atts_new(:)

  end subroutine update_shift_decider_t

  !.....................................................................
  recursive subroutine init_exp_decider_t(this)
    implicit none

    class(exp_decider_t) :: this

    this%fexp_old = 0.0
    this%old_fexp_is_valid = .false.

  end subroutine init_exp_decider_t

  !.....................................................................
  recursive function need_new_exp_decider_t(this, fexp_new) result(need_new_detshiftexp)
    implicit none

    class(exp_decider_t) :: this
    real, intent(in) :: fexp_new
    logical :: need_new_detshiftexp

    if(this%old_fexp_is_valid) then
      if(fexp_new > this%fexp_old*dfexp .or. fexp_new < this%fexp_old/dfexp) then
        need_new_detshiftexp = .true.
      else
        need_new_detshiftexp = .false.
      end if
    else
      need_new_detshiftexp = .true.
    end if

    this%fexp_new = fexp_new

  end function need_new_exp_decider_t

  !.....................................................................
  recursive subroutine update_exp_decider_t(this)
    implicit none

    class(exp_decider_t) :: this

    this%old_fexp_is_valid = .true.
    this%fexp_old = this%fexp_new

  end subroutine update_exp_decider_t

  !.....................................................................
  recursive subroutine init_img_limits(this, xi_size, yi_size)
    implicit none

    class(img_limits_t) :: this
    integer, intent(in) :: xi_size, yi_size

    allocate(this%rows(yi_size))
    this%rows(:)%i_min = xi_size
    this%rows(:)%i_max = 1
    this%yi_min = yi_size
    this%yi_max = 1

  end subroutine init_img_limits

  !.....................................................................
  recursive subroutine set_lim_img_limits(this, xi_min, xi_max, yi)
    implicit none

    class(img_limits_t) :: this
    integer, intent(in) :: xi_min, xi_max, yi

    this%rows(yi)%i_min = min(this%rows(yi)%i_min,xi_min)
    this%rows(yi)%i_max = max(this%rows(yi)%i_max,xi_max)
    this%yi_min = min(this%yi_min,yi)
    this%yi_max = max(this%yi_max,yi)

  end subroutine set_lim_img_limits

  !.....................................................................
  recursive subroutine destroy_img_limits(this)
    implicit none

    class(img_limits_t) :: this

    if(allocated(this%rows)) deallocate(this%rows)

  end subroutine destroy_img_limits

  !.....................................................................
  recursive function vector_cross(v_a,v_b) result(v_c)
    implicit none
    real(real64), dimension(3), intent(in) :: v_a,v_b
    real(real64), dimension(3) :: v_c

    v_c(1) = v_a(2)*v_b(3) - v_a(3)*v_b(2)
    v_c(2) = v_a(3)*v_b(1) - v_a(1)*v_b(3)
    v_c(3) = v_a(1)*v_b(2) - v_a(2)*v_b(1)

  end function vector_cross

  !.....................................................................
  recursive subroutine calc_im_basis(ra, dec, basis, roll)
    ! This calculates basis vectors for an image-plane system which has axes defined as follows:
    !	Z^ from the sky toward the observer.
    !	X^ parallel to the equatorial axis, pointing in the direction of decreasing RA (i.e. to the right as one looks outward from the origin).
    !	Y^ pointing in the direction of increasing declination (i.e. pointing upward as one looks outward from the origin).
    !
    ! The coordinates of the vectors are expressed in the standard celestial basis which has:
    !	X^ in the direction of RA=0.
    !	Y^ in the direction of RA=90 deg.
    !	Z^ in the direction of the north celestial pole.
    !
    implicit none

    real(real64),           intent(in)  :: ra, dec ! both expected in radians.
    real(real64), dimension(3,3)&
      ,                     intent(out) :: basis
    real(real64), optional, intent(in)  :: roll

    real(real64) :: sinr,cosr,sind,cosd,sinroll,cosroll
    real(real64), dimension(2,2,3) :: temp
    integer :: ii

    sinr = sin(ra)
    cosr = cos(ra)
    sind = sin(dec)
    cosd = cos(dec)

    if(present(roll)) then
      sinroll = sin(roll)
      cosroll = cos(roll)
      temp(1,1,:) = [0.0d0, -cosroll, sinroll]
      temp(1,2,:) = [0.0d0,  sinroll, cosroll]
    else
      temp(1,1,:) = [0.0d0, -1.0d0, 0.0d0]
      temp(1,2,:) = [0.0d0,  0.0d0, 1.0d0]
    end if

    do ii=1,2
      temp(2,ii,:) = [cosd*temp(1,ii,1)-sind*temp(1,ii,3),temp(1,ii,2),cosd*temp(1,ii,3)+sind*temp(1,ii,1)]
    end do

    do ii=1,2
      basis(ii,:) = [cosr*temp(2,ii,1)-sinr*temp(2,ii,2),cosr*temp(2,ii,2)+sinr*temp(2,ii,1),temp(2,ii,3)]
    end do
    basis(3,:) = vector_cross(basis(1,:),basis(2,:)) 

  end subroutine calc_im_basis

  !.....................................................................
  recursive subroutine rotate_image_original(patch_sizes_i, n_bands, baddetmap&
    , roll, optic_axis_ccd, patch_ctr, img_pixel_sizes_ccd&
    , dxcen, dycen, detroll, img_limits, xaxr, yaxr)

    implicit none

    integer, dimension(2), intent(in)  :: patch_sizes_i
    integer,               intent(in)  :: n_bands
    real, dimension(N_PIXELS_CCD,N_PIXELS_CCD,n_bands)&
      ,                    intent(in)  :: baddetmap
    real(real64),          intent(in)  :: roll
    real(real64), dimension(2)&
      ,                    intent(in)  :: optic_axis_ccd
    real, dimension(2),    intent(in)  :: patch_ctr ! Nominal centre (in image pixels) of image.
    real, dimension(2),    intent(in)  :: img_pixel_sizes_ccd ! image pixel size in terms of CCD pixels.
    real,                  intent(in)  :: dxcen, dycen    ! nominal centre of CCD (in CCD pixel coords).
    real, dimension(patch_sizes_i(1),n_bands,patch_sizes_i(2))&
      ,                    intent(out) :: detroll
    type(img_limits_t),    intent(out) :: img_limits
    real,                  intent(out) :: xaxr, yaxr               ! rotated location of optical axis

    integer :: yi,xi,raw_xi,raw_yi
    real(real64) :: sinr,cosr
    real :: y,x,xax0,yax0,sinr_real,cosr_real

    sinr = dsin(roll*deg2rad)
    cosr = dcos(roll*deg2rad)
    sinr_real = real(sinr)
    cosr_real = real(cosr)

    call img_limits%init(patch_sizes_i(1), patch_sizes_i(2))

    detroll = 0.0
    do yi = 1, patch_sizes_i(2)
      y = real(yi) - patch_ctr(2)
      do xi = 1, patch_sizes_i(1)
        x = real(xi) - patch_ctr(1)
        ! 20200303 IMS: in evatt, the relation between sky axes (with X pointing west and Y pointing north) and the RAWX/Y axes projected on the celestial sphere is such that a positive roll angle corresponds to a clockwise rotation of the projected RAWX/Y on the sky, the zero point corresponding to RAWX pointing south and RAWY pointing west.
        raw_xi = nint((-sinr*x - cosr*y)*img_pixel_sizes_ccd(1) + dxcen)
        raw_yi = nint((-sinr*y + cosr*x)*img_pixel_sizes_ccd(2) + dycen)

        if (raw_xi >= 1 .and. raw_xi <= N_PIXELS_CCD .and. &
            raw_yi >= 1 .and. raw_yi <= N_PIXELS_CCD) then

          if (baddetmap(raw_xi,raw_yi,1) >= 0.0) then
            call img_limits%set_limits(xi, xi, yi)

            detroll(xi,:,yi) = baddetmap(raw_xi,raw_yi,:) 
          end if ! good raw pixel
        end if ! raw coord valid for CCD
      end do ! xi
    end do ! yi

    ! rotate optical axis
    xax0 = real(optic_axis_ccd(1)) - dxcen
    yax0 = real(optic_axis_ccd(2)) - dycen 
!*** 20200303 IMS: rotate through -roll.
    xaxr = (sinr_real*xax0 - cosr_real*yax0)/img_pixel_sizes_ccd(1) + patch_ctr(1)
    yaxr = (sinr_real*yax0 + cosr_real*xax0)/img_pixel_sizes_ccd(2) + patch_ctr(2)

  end subroutine rotate_image_original

  !.....................................................................
  recursive subroutine rotate_image_approx(patch_sizes_i, n_bands, baddetmap&
    , optic_axis_ccd, patch_ctr, img_pixel_size_tan, ccd_pixel_size_tan&
    , att_point, ra_proj, cos_dec_proj, sin_dec_proj, detroll, img_limits)

    implicit none

    integer, dimension(2), intent(in)  :: patch_sizes_i
    integer,               intent(in)  :: n_bands
    real,    dimension(N_PIXELS_CCD,N_PIXELS_CCD,n_bands)&
      ,                    intent(in)  :: baddetmap
    real(real64), dimension(2)&
      ,                    intent(in)  :: optic_axis_ccd
    real, dimension(2),    intent(in)  :: patch_ctr ! Nominal centre (in image pixels) of image.
    real(real64), dimension(2)&
      ,                    intent(in)  :: img_pixel_size_tan,ccd_pixel_size_tan
    type(tAtt),            intent(in)  :: att_point
    real(real64),          intent(in)  :: ra_proj,cos_dec_proj,sin_dec_proj
    real,    dimension(patch_sizes_i(1),n_bands,patch_sizes_i(2))&
      ,                    intent(out) :: detroll
    type(img_limits_t),    intent(out) :: img_limits

    real(real64) :: ra_point,dec_point,ccd_roll,det
    real(real64), dimension(3,3) :: p_basis,p_basis_im
    integer :: ii,xi,yi,raw_xi,raw_yi
    real(real64), dimension(patch_sizes_i(1)) :: xs_im,ys_im,xs_ccd,ys_ccd
    real(real64), dimension(2,2) :: t_mat,t_inv
    integer, dimension(patch_sizes_i(1)) :: rawxis,rawyis
    logical :: within_ccd_y_limits
!,x_min_is_increasing,x_max_is_decreasing

    call img_limits%init(patch_sizes_i(1), patch_sizes_i(2))

    ra_point  = att_point%ra  *deg2rad
    dec_point = att_point%dec *deg2rad
    ccd_roll  = (-att_point%roll - 90.0)*deg2rad

    call calc_im_basis(ra_point - ra_proj, dec_point, p_basis, ccd_roll)
    ! Note that p_hat = -p_basis(3,:).

    p_basis_im(:,1) = -p_basis(:,2)
    do ii=1,3
      p_basis_im(ii,3) = -cos_dec_proj*p_basis(ii,1) - sin_dec_proj*p_basis(ii,3)
      p_basis_im(ii,2) =  cos_dec_proj*p_basis(ii,3) - sin_dec_proj*p_basis(ii,1)
    end do

    ! The X and Y components of the X and Y vectors of the rotated p basis (i.e., we project these X and Y vectors onto the image plane) form columns of a linear transform matrix T. We have to invert this, after which we can transform image-plane positions to CCD positions via
    !
    !	v_ccd = T^{-1} v_im.
    !
    t_mat(:,1) = p_basis_im(1,1:2)
    t_mat(:,2) = p_basis_im(2,1:2)

    det = t_mat(1,1)*t_mat(2,2) - t_mat(1,2)*t_mat(2,1)
    t_inv(1,1) =  t_mat(2,2)/det
    t_inv(1,2) = -t_mat(1,2)/det
    t_inv(2,1) = -t_mat(2,1)/det
    t_inv(2,2) =  t_mat(1,1)/det

    xs_im(:) = img_pixel_size_tan(1)*(real([(xi,xi=1,patch_sizes_i(1),1)]) - patch_ctr(1))
    ys_im(:) = img_pixel_size_tan(2)*(real([(yi,yi=1,patch_sizes_i(2),1)]) - patch_ctr(2))

    within_ccd_y_limits = .false. ! default
    detroll = 0.0
    YLOOP: do yi = 1, patch_sizes_i(2)
      xs_ccd(:) = t_inv(1,1)*xs_im + t_inv(1,2)*ys_im(yi)
      ys_ccd(:) = t_inv(2,1)*xs_im + t_inv(2,2)*ys_im(yi)

      rawxis = nint(xs_ccd/ccd_pixel_size_tan(1) + optic_axis_ccd(1))
      rawyis = nint(ys_ccd/ccd_pixel_size_tan(2) + optic_axis_ccd(2))

!*** here would be the place to reverse the focal-plane corrections applied via cal_distcorr().

      XLOOP: do xi = 1, patch_sizes_i(1)
        raw_xi = rawxis(xi)
        raw_yi = rawyis(xi)
        if (raw_xi >= 1 .and. raw_xi <= N_PIXELS_CCD .and. &
            raw_yi >= 1 .and. raw_yi <= N_PIXELS_CCD) then

          call img_limits%set_limits(xi, xi, yi)
          detroll(xi,:,yi) = baddetmap(raw_xi,raw_yi,:)
          within_ccd_y_limits = .true.

        else if(within_ccd_y_limits) then
    exit YLOOP
        end if
      end do XLOOP
    end do YLOOP

  end subroutine rotate_image_approx

  !.....................................................................
  recursive subroutine rotate_image_approx_slow(patch_sizes_i, n_bands, baddetmap&
    , optic_axis_ccd, patch_ctr, img_pixel_size_tan, ccd_pixel_size_tan&
    , att_point, ra_proj, cos_dec_proj, sin_dec_proj, detroll, img_limits)

    implicit none

    integer, dimension(2), intent(in)  :: patch_sizes_i
    integer,               intent(in)  :: n_bands
    real,    dimension(N_PIXELS_CCD,N_PIXELS_CCD,n_bands)&
      ,                    intent(in)  :: baddetmap
    real(real64), dimension(2)&
      ,                    intent(in)  :: optic_axis_ccd
    real, dimension(2),    intent(in)  :: patch_ctr ! Nominal centre (in image pixels) of image.
    real(real64), dimension(2)&
      ,                    intent(in)  :: img_pixel_size_tan,ccd_pixel_size_tan
    type(tAtt),            intent(in)  :: att_point
    real(real64),          intent(in)  :: ra_proj,cos_dec_proj,sin_dec_proj
    real,    dimension(patch_sizes_i(1),n_bands,patch_sizes_i(2))&
      ,                    intent(out) :: detroll
    type(img_limits_t),    intent(out) :: img_limits

    real(real64) :: ra_point,dec_point,ccd_roll,det
    real(real64), dimension(3,3) :: p_basis,p_basis_im
    integer :: ii,xi,yi,raw_xi,raw_yi
    real(real64), dimension(patch_sizes_i(1)) :: xs_im,ys_im,xs_ccd,ys_ccd
    real(real64), dimension(2,2) :: t_mat,t_inv
    integer, dimension(patch_sizes_i(1)) :: rawxis,rawyis

    call img_limits%init(patch_sizes_i(1), patch_sizes_i(2))

    ra_point  = att_point%ra  *deg2rad
    dec_point = att_point%dec *deg2rad
    ccd_roll  = (-att_point%roll - 90.0)*deg2rad

    call calc_im_basis(ra_point - ra_proj, dec_point, p_basis, ccd_roll)
    ! Note that p_hat = -p_basis(3,:).

    p_basis_im(:,1) = -p_basis(:,2)
    do ii=1,3
      p_basis_im(ii,3) = -cos_dec_proj*p_basis(ii,1) - sin_dec_proj*p_basis(ii,3)
      p_basis_im(ii,2) =  cos_dec_proj*p_basis(ii,3) - sin_dec_proj*p_basis(ii,1)
    end do

    ! The X and Y components of the X and Y vectors of the rotated p basis (i.e., we project these X and Y vectors onto the image plane) form columns of a linear transform matrix T. We have to invert this, after which we can transform image-plane positions to CCD positions via
    !
    !	v_ccd = T^{-1} v_im.
    !
    t_mat(:,1) = p_basis_im(1,1:2)
    t_mat(:,2) = p_basis_im(2,1:2)

    det = t_mat(1,1)*t_mat(2,2) - t_mat(1,2)*t_mat(2,1)
    t_inv(1,1) =  t_mat(2,2)/det
    t_inv(1,2) = -t_mat(1,2)/det
    t_inv(2,1) = -t_mat(2,1)/det
    t_inv(2,2) =  t_mat(1,1)/det

    xs_im(:) = img_pixel_size_tan(1)*(real([(xi,xi=1,patch_sizes_i(1),1)]) - patch_ctr(1))
    ys_im(:) = img_pixel_size_tan(2)*(real([(yi,yi=1,patch_sizes_i(2),1)]) - patch_ctr(2))

    detroll = 0.0
    do yi = 1, patch_sizes_i(2)
      xs_ccd(:) = t_inv(1,1)*xs_im + t_inv(1,2)*ys_im(yi)
      ys_ccd(:) = t_inv(2,1)*xs_im + t_inv(2,2)*ys_im(yi)

      rawxis = nint(xs_ccd/ccd_pixel_size_tan(1) + optic_axis_ccd(1))
      rawyis = nint(ys_ccd/ccd_pixel_size_tan(2) + optic_axis_ccd(2))

!*** here would be the place to reverse the focal-plane corrections applied via cal_distcorr().

      do xi = 1, patch_sizes_i(1)
        raw_xi = rawxis(xi)
        raw_yi = rawyis(xi)
        if (raw_xi >= 1 .and. raw_xi <= N_PIXELS_CCD .and. &
            raw_yi >= 1 .and. raw_yi <= N_PIXELS_CCD) then

          call img_limits%set_limits(xi, xi, yi)

          detroll(xi,:,yi) = baddetmap(raw_xi,raw_yi,:)
        end if
      end do
    end do ! yi

  end subroutine rotate_image_approx_slow

  !.....................................................................
  recursive subroutine rotate_image_exact(patch_sizes_i, n_bands, baddetmap&
    , optic_axis_ccd, patch_ctr, img_pixel_size_tan, ccd_pixel_size_tan&
    , att_point, ra_proj, cos_dec_proj, sin_dec_proj, detroll, img_limits)

    implicit none

    integer, dimension(2), intent(in)  :: patch_sizes_i
    integer,               intent(in)  :: n_bands
    real,    dimension(N_PIXELS_CCD,N_PIXELS_CCD,n_bands)&
      ,                    intent(in)  :: baddetmap
    real(real64), dimension(2)&
      ,                    intent(in)  :: optic_axis_ccd
    real, dimension(2),    intent(in)  :: patch_ctr ! Nominal centre (in image pixels) of image.
    real(real64), dimension(2)&
      ,                    intent(in)  :: img_pixel_size_tan,ccd_pixel_size_tan
    type(tAtt),            intent(in)  :: att_point
    real(real64),          intent(in)  :: ra_proj,cos_dec_proj,sin_dec_proj
    real,    dimension(patch_sizes_i(1),n_bands,patch_sizes_i(2))&
      ,                    intent(out) :: detroll
    type(img_limits_t),    intent(out) :: img_limits

    real(real64) :: ra_point,dec_point,ccd_roll,cosroll,sinroll
    real(real64), dimension(3,3) :: p_basis,p_basis_im
    real(real64), dimension(3) :: p_hat_im
    integer :: ii,xi,yi,raw_xi,raw_yi
    real(real64), dimension(patch_sizes_i(1)) :: im_xs,im_ys,im_zs,scales,xs_p,ys_p,xs_ccd,ys_ccd
    real(real64), dimension(patch_sizes_i(1)) :: delta_xs_p,delta_ys_p,delta_zs_p
    integer, dimension(patch_sizes_i(1)) :: rawxis,rawyis

    ra_point  = att_point%ra  *deg2rad
    dec_point = att_point%dec *deg2rad
    ccd_roll  = att_point%roll*deg2rad

    call calc_im_basis(ra_point - ra_proj, dec_point, p_basis)
    ! Note that p_hat = -p_basis(3,:).

    cosroll = cos(-ccd_roll)
    sinroll = sin(-ccd_roll)

    ! After rotating by dec_proj, the p_basis is expressed in terms of the image basis.
    !
    p_basis_im(:,1) = -p_basis(:,2)
    do ii=1,3
      p_basis_im(ii,3) = -cos_dec_proj*p_basis(ii,1) - sin_dec_proj*p_basis(ii,3)
      p_basis_im(ii,2) =  cos_dec_proj*p_basis(ii,3) - sin_dec_proj*p_basis(ii,1)
    end do

    p_hat_im(:) = -p_basis_im(3,:)

    im_xs(:) = img_pixel_size_tan(1)*(real([(xi,xi=1,patch_sizes_i(1),1)]) - patch_ctr(1)) + p_hat_im(1)
    im_ys(:) = img_pixel_size_tan(2)*(real([(yi,yi=1,patch_sizes_i(2),1)]) - patch_ctr(2)) + p_hat_im(2)

    call img_limits%init(patch_sizes_i(1), patch_sizes_i(2))

    detroll = 0.0
    do yi = 1, patch_sizes_i(2)
      im_zs(:) = -sqrt(1.0d0 - im_xs(:)*im_xs(:) - im_ys(yi)*im_ys(yi))

      ! Take the cross product with the rotated p_hat to calculate scaling denominator to project onto the ccd tan plane:
      !
      scales(:) = im_xs*p_hat_im(1) + im_ys(yi)*p_hat_im(2) + im_zs*p_hat_im(3)

      delta_xs_p(:) = im_xs(:) /scales(:) - p_hat_im(1)
      delta_ys_p(:) = im_ys(yi)/scales(:) - p_hat_im(2)
      delta_zs_p(:) = im_zs(:) /scales(:) - p_hat_im(3)

      ! Project onto unrotated CCD plane:
      !
      xs_p(:) = (delta_xs_p(:)*p_basis_im(1,1) + delta_ys_p(:)*p_basis_im(1,2) + delta_zs_p(:)*p_basis_im(1,3))
      ys_p(:) = (delta_xs_p(:)*p_basis_im(2,1) + delta_ys_p(:)*p_basis_im(2,2) + delta_zs_p(:)*p_basis_im(2,3))

      ! Now finally, apply the position-angle rotation:
      !
      xs_ccd(:) = sinroll*xs_p - cosroll*ys_p
      ys_ccd(:) = sinroll*ys_p + cosroll*xs_p

      rawxis = nint(xs_ccd/ccd_pixel_size_tan(1) + optic_axis_ccd(1))
      rawyis = nint(ys_ccd/ccd_pixel_size_tan(2) + optic_axis_ccd(2))

!*** here would be the place to reverse the focal-plane corrections applied via cal_distcorr().

      do xi = 1, patch_sizes_i(1)
        raw_xi = rawxis(xi)
        raw_yi = rawyis(xi)
        if (raw_xi >= 1 .and. raw_xi <= N_PIXELS_CCD .and. &
            raw_yi >= 1 .and. raw_yi <= N_PIXELS_CCD) then

          call img_limits%set_limits(xi, xi, yi)

          detroll(xi,:,yi) = baddetmap(raw_xi,raw_yi,:)
        end if
      end do
    end do ! yi

  end subroutine rotate_image_exact

  !.....................................................................
  recursive subroutine calc_delta_att(atts_interp, nstep, datt)
    implicit none

    type(tAtt), dimension(:), intent(in)  :: atts_interp
    integer,                  intent(in)  :: nstep
    type(scan_t),             intent(out) :: datt

! 20201130 IMS: this routine could be cleverer, also I am wondering if we should not divide by nstep (or even better, by the time difference) to get the rate of shift. Need to look at HB original code to see what the comparison with dshift looked like there.

    if(atts_interp(1)%is_valid .and. atts_interp(nstep)%is_valid) then
      datt%xshift = atts_interp(nstep)%x_pix - atts_interp(1)%x_pix
      datt%yshift = atts_interp(nstep)%y_pix - atts_interp(1)%y_pix
      datt%is_valid = .true.
    else
      datt%xshift = 0.0 ! Just to give them
      datt%yshift = 0.0 ! some value.
      datt%is_valid = .false.
    end if

  end subroutine calc_delta_att

  !.....................................................................
  recursive subroutine compare_delta_att(datt_old, atts_interp, nstep, datt_new, datt_status)
    ! This function attempts to compare the scan direction and speed ('datt') of the present time step to the last good value. The return status can take one of three values:
    !	CANNOT_DETERMINE
    !	SCAN_HAS_CHANGED
    !	SCAN_IS_UNCHANGED
    !
    implicit none

    type(scan_t),             intent(in)  :: datt_old
    type(tAtt), dimension(:), intent(in)  :: atts_interp
    integer,                  intent(in)  :: nstep
    type(scan_t),             intent(out) :: datt_new
    integer,                  intent(out) :: datt_status

    call calc_delta_att(atts_interp,nstep,datt_new)

    datt_status = CANNOT_DETERMINE ! default

    if(datt_old%is_valid .and. datt_new%is_valid) then
      if(max(abs(datt_new%xshift-datt_old%xshift),abs(datt_new%yshift-datt_old%yshift)) > dshift) then
        datt_status = SCAN_HAS_CHANGED
      else
        datt_status = SCAN_IS_UNCHANGED
      end if
    end if

  end subroutine compare_delta_att

  !.....................................................................
  recursive subroutine shift_image_original(patch_sizes_i, n_bands, detroll&
    , roll_img_limits, nstep, atts_interp, xatt_mean, yatt_mean, detshift&
    , shift_img_limits, n_no_overlap)

    implicit none

    integer, dimension(2),               intent(in)  :: patch_sizes_i
    integer,                             intent(in)  :: n_bands
    real, dimension(patch_sizes_i(1),n_bands,patch_sizes_i(2))&
      ,                                  intent(in)  :: detroll
    type(img_limits_t),                  intent(in)  :: roll_img_limits
    integer,                             intent(in)  :: nstep
    type(tAtt), dimension(:)& ! dimensioned to ndead, which can be larger than nstep
       ,                                 intent(in)  :: atts_interp
    real(real64),                        intent(in)  :: xatt_mean, yatt_mean
    real, dimension(patch_sizes_i(1),n_bands,patch_sizes_i(2))&
      ,                                  intent(out) :: detshift
    type(img_limits_t),                  intent(out) :: shift_img_limits
    integer,                             intent(out) :: n_no_overlap

    integer :: nry,l,ix0,iy0,nsy1,nry1,nrx1_min,nrx1_max,nsx1_min,nsx1_max

    detshift = 0.0

    ! Defaults that exclude everything:
    !
    call shift_img_limits%init(patch_sizes_i(1),patch_sizes_i(2))

    n_no_overlap = 0
    do nry = roll_img_limits%yi_min, roll_img_limits%yi_max
      do l = 1, nstep ! cycle through the ~20 frames in this time step.
        if(.not. atts_interp(l)%is_valid)&
      cycle

        ! Calculate difference in pixels between the attitude in
        ! the middle of the whole time step (x/yatt1) and that
        ! for the present frame (y/yatt2).
        !
        ix0 = nint(atts_interp(l)%x_pix - xatt_mean)
        iy0 = nint(atts_interp(l)%y_pix - yatt_mean)

        nsy1 = nry + iy0

        if (nsy1<1) then
          n_no_overlap = n_no_overlap + 1
      cycle
        end if
        if (nsy1>patch_sizes_i(2)) then
          n_no_overlap = n_no_overlap + 1
      cycle
        end if

        nry1 = nry

        nrx1_min = roll_img_limits%rows(nry)%i_min
        nrx1_max = roll_img_limits%rows(nry)%i_max

        nsx1_min = nrx1_min + ix0
        if (nsx1_min>patch_sizes_i(1)) then
          n_no_overlap = n_no_overlap + 1
      cycle
        end if

        if (nsx1_min<1) then
          nsx1_min = 1
          nrx1_min = nsx1_min - ix0
          if(nrx1_min > patch_sizes_i(1)) then
            n_no_overlap = n_no_overlap + 1
      cycle
          end if
        end if   

        nsx1_max = nrx1_max + ix0
        if (nsx1_max<1) then
          n_no_overlap = n_no_overlap + 1
      cycle
        end if

        if (nsx1_max>patch_sizes_i(1)) then
          nsx1_max = patch_sizes_i(1)
          nrx1_max = nsx1_max - ix0
          if(nrx1_max < 1) then
            n_no_overlap = n_no_overlap + 1
      cycle
          end if
        end if   

        detshift(nsx1_min:nsx1_max,:,nsy1)&
          = detshift(nsx1_min:nsx1_max,:,nsy1)&
          + detroll(nrx1_min:nrx1_max,:,nry1)

        call shift_img_limits%set_limits(nsx1_min, nsx1_max, nsy1)
      end do ! next step
    end do ! next y value

  end subroutine shift_image_original

  !.....................................................................
  recursive subroutine shift_image_new(patch_sizes_i, n_bands, detroll&
    , roll_img_limits, nstep, atts_interp, xatt_mean, yatt_mean, detshift&
    , shift_img_limits, n_no_overlap)

    implicit none

    integer, dimension(2),               intent(in)  :: patch_sizes_i
    integer,                             intent(in)  :: n_bands
    real, dimension(patch_sizes_i(1),n_bands,patch_sizes_i(2))&
      ,                                  intent(in)  :: detroll
    type(img_limits_t),                  intent(in)  :: roll_img_limits
    integer,                             intent(in)  :: nstep
    type(tAtt), dimension(:)& ! dimensioned to ndead, which can be larger than nstep
       ,                                 intent(in)  :: atts_interp
    real(real64),                        intent(in)  :: xatt_mean, yatt_mean
    real, dimension(patch_sizes_i(1),n_bands,patch_sizes_i(2))&
      ,                                  intent(out) :: detshift
    type(img_limits_t),                  intent(out) :: shift_img_limits
    integer,                             intent(out) :: n_no_overlap

    integer :: l,offset_xi,offset_yi,roll_start_xi,roll_finis_xi,roll_start_yi,roll_finis_yi
    integer :: shift_start_xi,shift_finis_xi,roll_yi,shift_yi
    logical :: no_x_overlaps

    detshift = 0.0

    ! Defaults that exclude everything:
    !
    call shift_img_limits%init(patch_sizes_i(1),patch_sizes_i(2))

    n_no_overlap = 0
    do l = 1, nstep ! cycle through the ~20 frames in this time step.
      if(.not. atts_interp(l)%is_valid)&
    cycle

      offset_xi = nint(atts_interp(l)%x_pix - xatt_mean)
      offset_yi = nint(atts_interp(l)%y_pix - yatt_mean)

      roll_start_yi = max(roll_img_limits%yi_min,                1-offset_yi)
      roll_finis_yi = min(roll_img_limits%yi_max, patch_sizes_i(2)-offset_yi)

      if(roll_start_yi>patch_sizes_i(2) .or. roll_finis_yi<1) then
        n_no_overlap = n_no_overlap + 1
    cycle
      end if

      no_x_overlaps = .true. ! default
      do roll_yi=roll_start_yi,roll_finis_yi
        roll_start_xi = max(roll_img_limits%rows(roll_yi)%i_min,                1-offset_xi)
        roll_finis_xi = min(roll_img_limits%rows(roll_yi)%i_max, patch_sizes_i(1)-offset_xi)

        if(roll_start_xi>patch_sizes_i(1) .or. roll_finis_xi<1)&
      cycle

        no_x_overlaps = .false.

        shift_yi = roll_yi + offset_yi

        shift_start_xi = roll_start_xi + offset_xi
        shift_finis_xi = roll_finis_xi + offset_xi

        call shift_img_limits%set_limits(shift_start_xi, shift_finis_xi, shift_yi)

        detshift(    shift_start_xi:shift_finis_xi,:,shift_yi)&
          = detshift(shift_start_xi:shift_finis_xi,:,shift_yi)&
         + detroll(  roll_start_xi:roll_finis_xi, :, roll_yi)
      end do

      if(no_x_overlaps) then
        n_no_overlap = n_no_overlap + 1
      end if

    end do ! next step

  end subroutine shift_image_new

  !.....................................................................
  recursive subroutine copy_to_camexp(patch_sizes_i, size_xi, size_yi, n_bands&
    , patch, patch_img_limits, ti_start, ti_finis, offsets_xi, offsets_yi&
    , att_is_valid, no_overlaps, outimg)

    implicit none

    integer, dimension(2),  intent(in)    :: patch_sizes_i
    integer,                intent(in)    :: size_xi,size_yi,n_bands
    real, dimension(patch_sizes_i(1),n_bands,patch_sizes_i(2))&
      ,                     intent(in)    :: patch
    type(img_limits_t),     intent(in)    :: patch_img_limits
    integer,                intent(in)    :: ti_start,ti_finis
    integer, dimension(:),  intent(in)    :: offsets_xi,offsets_yi
    logical, dimension(:),  intent(in)    :: att_is_valid
    logical,                intent(out)   :: no_overlaps
    real, dimension(size_xi,n_bands,size_yi)&
      ,                     intent(inout) :: outimg

    integer :: ti,offset_xi,offset_yi,patch_start_xi,patch_start_yi,patch_finis_xi,patch_finis_yi
    integer :: start_xi,finis_xi,patch_yi,yi,iblock,local_n_blocks,all_start_yi,all_finis_yi,num_yi
    logical :: no_x_overlaps
    integer, dimension(n_blocks) :: patch_starts_yi,patch_stopps_yi

    no_overlaps = .false. ! default

    all_start_yi = patch_sizes_i(2)
    all_finis_yi = 1
    do ti = ti_start, ti_finis
      if(.not.att_is_valid(ti))&
    cycle

      all_start_yi = min(all_start_yi, max(patch_img_limits%yi_min,       1 - offsets_yi(ti)))
      all_finis_yi = max(all_finis_yi, min(patch_img_limits%yi_max, size_yi - offsets_yi(ti)))
    end do

    num_yi = 1 + all_finis_yi - all_start_yi
    if(all_finis_yi < all_start_yi) then
      no_overlaps = .true.
  return
    end if

    if(num_yi > n_blocks) then
      local_n_blocks = n_blocks
    else
      local_n_blocks = 1
    end if

    patch_starts_yi(1) = all_start_yi
    do iblock=1,local_n_blocks-1
      patch_stopps_yi(iblock) = patch_starts_yi(iblock)&
        + floor((1.0/real(local_n_blocks - (iblock-1)))*(all_finis_yi - patch_starts_yi(iblock)))
      patch_starts_yi(iblock+1) = patch_stopps_yi(iblock) + 1
    end do
    patch_stopps_yi(local_n_blocks) = all_finis_yi

    no_x_overlaps = .true. ! default
    do iblock = 1, local_n_blocks
      do ti = ti_start, ti_finis            
        if(.not.att_is_valid(ti))&
      cycle

        offset_xi = offsets_xi(ti) 
        offset_yi = offsets_yi(ti)                   

        patch_start_yi = max(patch_starts_yi(iblock),       1 - offset_yi)
        patch_finis_yi = min(patch_stopps_yi(iblock), size_yi - offset_yi)

!****is this necessary? We should have already sorted this when setting patch_img_limits!
        if(patch_start_yi>patch_sizes_i(2) .or. patch_finis_yi<1) then
      cycle
        end if

        do patch_yi = patch_start_yi, patch_finis_yi
          patch_start_xi = max(patch_img_limits%rows(patch_yi)%i_min,       1 - offset_xi)
          patch_finis_xi = min(patch_img_limits%rows(patch_yi)%i_max, size_xi - offset_xi)

!****is this necessary? We should have already sorted this when setting patch_img_limits!
          if(patch_start_xi>patch_sizes_i(1) .or. patch_finis_xi<1)&
        cycle

          no_x_overlaps = .false.

          yi = patch_yi + offset_yi

          start_xi = patch_start_xi + offset_xi
          finis_xi = patch_finis_xi + offset_xi

          outimg(    start_xi:finis_xi,:,yi)&
            = outimg(start_xi:finis_xi,:,yi)&
            + patch(patch_start_xi:patch_finis_xi,:,patch_yi)
        end do
      end do
    end do

    if(no_x_overlaps)&
      no_overlaps = .true.

  end subroutine copy_to_camexp

end module expmap_trans_mod

