! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module flaregti_mod

  use iso_fortran_env
  use erolib_gti_mod
  use expmap_legacy, only: get_attitude, tAtt, linint_att2
  use radec2pix_mod, only: radec2pix
  use param

  ! used wrapped fitsio commands
  use srctool_generic_fitsraw

  implicit none

  ! parameters read by the program
  type flaregti_pars_t
     ! event, mask and lightcurve filenames
     character(len=128) :: efile, mfile, lcfile, gtifile, threshimgfile
     ! pi min and max range included in flare creation
     integer :: pimin, pimax
     ! pi and max for mask creation
     integer :: mask_pimin, mask_pimax
     ! range of coordinates for image creation
     integer :: xmin, xmax, ymin, ymax
     ! number of spatial grid points
     integer :: ngrid
     ! image pixel size; unit: sky pixels
     integer :: xybin
     ! detection likelihood threshold for mask
     real :: detml
     ! source diameter in arcsec, for automatic threshold determination
     real :: sdiam
     ! source detection likelihood, for automatic threshold determination
     real :: sdetlike
     ! FoV in arcmin
     real :: fovrad_amin
     ! fixed threshold if >0, otherwise dynamical threshold
     real :: fthresh
     ! maximum threshold if >0, otherwise ignored
     real :: maxthresh
     ! time bin size in seconds
     real(kind=real64) :: tbin
     ! whether to write mask file or lightcurve file
     logical :: wrmask, wrlightcrv, wrthreshimg
     ! GTI HDU type
     character(len=7) :: gtitype
     ! whether to update gtis in event file
     logical :: updateevtgti
     ! number of iterations of doing masking, thresholding and filtering
     integer :: mask_iter
  end type flaregti_pars_t

  ! contains information about the opened fits event file
  type evtfile_t
     ! fits unit number and status
     integer :: unit, status
     ! number of rows in event table
     integer :: nrow
     ! column numbers for time, x, y, pi and telescope module
     integer :: tcol, xcol, ycol, picol, tmnrcol, flagcol
     ! WCS for x and y columns
     real(kind=real64) :: cdltx, cdlty, crpxx, crpxy, crvlx, crvly
     ! minimum and maximum x and y coordinates
     integer :: xmin, xmax, ymin, ymax
     ! selection of flag
     integer :: flagsel
  end type evtfile_t

  private
  public :: flaregti_pars_t, evtfile_t
  public :: &
       read_pars, &
       evtfile_open, &
       evtfile_goto_hdu, &
       evtfile_evt_col_info, &
       evtfile_read_gtis, &
       evtfile_mk_image, &
       evtfile_mk_lightcrv, &
       evtfile_prep_attitude, &
       evtfile_close, &
       copy_gti, validate_gti, validate_gtis, &
       update_flaregti, write_gtifile, &
       write_mask, write_lightcurve, &
       mk_mask, &
       mk_area, &
       calc_expos, remove_empty_tbins, &
       mk_thresh, write_thresh_img, &
       mk_common_flaregti, merge_flaregti, &
       poisl2, gammq, gammp, gammln

  private :: clean_gti, merge_gapfree_gti

contains

!----------------------------------------------------------------------------
  subroutine read_pars(lupar, pars, ierr)
!----------------------------------------------------------------------------
! Read parameters into flaregit_pars defined type
!----------------------------------------------------------------------------
    integer, intent(in) :: lupar
    type(flaregti_pars_t), intent(out) :: pars
    integer, intent(in out) :: ierr

    call rgetc(lupar, "eventfile", pars%efile, ierr)
    call rgetc(lupar, "gtifile", pars%gtifile, ierr)
    call rgeti(lupar, "pimin", pars%pimin, ierr)
    call rgeti(lupar, "pimax", pars%pimax, ierr)
    call rgeti(lupar, "mask_pimin", pars%mask_pimin, ierr)
    call rgeti(lupar, "mask_pimax", pars%mask_pimax, ierr)
    call rgeti(lupar, "xmin", pars%xmin, ierr)
    call rgeti(lupar, "xmax", pars%xmax, ierr)
    call rgeti(lupar, "ymin", pars%ymin, ierr)
    call rgeti(lupar, "ymax", pars%ymax, ierr)
    call rgeti(lupar, "gridsize",  pars%ngrid, ierr)
    call rgeti(lupar, "binsize", pars%xybin, ierr)
    call rgetf(lupar, "detml", pars%detml, ierr)
    call rgetd(lupar, "timebin", pars%tbin, ierr)
    call rgetf(lupar, "source_size", pars%sdiam, ierr)
    call rgetf(lupar, "source_like", pars%sdetlike, ierr)
    call rgetf(lupar, "fov_radius", pars%fovrad_amin, ierr)
    call rgetf(lupar, "threshold", pars%fthresh, ierr)
    call rgetf(lupar, "max_threshold", pars%maxthresh, ierr)
    call rgetl(lupar, "write_mask", pars%wrmask, ierr)
    call rgetc(lupar, "mask", pars%mfile, ierr)
    call rgetl(lupar, "write_lightcurve", pars%wrlightcrv, ierr)
    call rgetc(lupar, "lightcurve", pars%lcfile, ierr)
    call rgetc(lupar, "gtitype", pars%gtitype, ierr)
    call rgetl(lupar, "write_thresholdimg", pars%wrthreshimg, ierr)
    call rgetc(lupar, "thresholdimg", pars%threshimgfile, ierr)
    call rgeti(lupar, "mask_iter", pars%mask_iter, ierr)

    pars%updateevtgti = pars%gtifile == "none"
    pars%mask_iter = max(pars%mask_iter, 1)

  end subroutine read_pars

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Event file functions

!----------------------------------------------------------------------------
  subroutine evtfile_open(file, filename, lupar, update)
!----------------------------------------------------------------------------
! Open event file
!----------------------------------------------------------------------------

    type(evtfile_t), intent(out) :: file
    character(len=*), intent(in) :: filename
    integer, intent(in) :: lupar
    logical, intent(in) :: update

    integer :: filemode

    integer, parameter :: read=0, readwrite=1

    filemode = merge(readwrite, read, update)

    file%status = 0
    call ftgiou_w(file%unit, file%status)
    call fttopn_w(file%unit, filename, filemode, file%status)
    if (file%status > 0) call wrstop(lupar, "flaregti_mod", "Cannot open event file")
  end subroutine evtfile_open

!----------------------------------------------------------------------------
  subroutine evtfile_goto_hdu(file, hduname, ierr)
    type(evtfile_t) :: file
    character(len=*) :: hduname
    integer :: ierr

    integer, parameter :: binary_tbl=2

    ! goto events extension
    call ftmnhd_w(file%unit, binary_tbl, hduname, 0, file%status)
    call fitserr("ftmnhd", file%status, ierr)

  end subroutine evtfile_goto_hdu

!----------------------------------------------------------------------------
  subroutine evtfile_evt_col_info(file, ierr)
!----------------------------------------------------------------------------
! Read column numbers and ranges
!----------------------------------------------------------------------------
    type(evtfile_t) :: file
    integer :: ierr

    character(len=64) :: comment
    character(len=6) :: keyword

    integer, parameter :: casesens = 0

    call evtfile_goto_hdu(file, "EVENTS", ierr)

    ! get row and column info
    call ftgnrw_w(file%unit, file%nrow, file%status)
    call fitserr("ftgnrw", file%status, ierr)
    call ftgcno_w(file%unit, casesens, "TIME", file%tcol, file%status)
    call fitserr("ftgcno", file%status, ierr)
    call ftgcno_w(file%unit, casesens, "X", file%xcol, file%status)
    call fitserr("ftgcno", file%status, ierr)
    call ftgcno_w(file%unit, casesens, "Y", file%ycol, file%status)
    call fitserr("ftgcno", file%status, ierr)

    call ftgcno_w(file%unit, casesens, "PI", file%picol, file%status)
    if(file%status>0) then
       ! backward compatibility look for ENERGY column instead
       file%status = 0
       call ftgcno_w(file%unit, casesens, "ENERGY", file%picol, file%status)
       if(file%status <= 0) call wrlog("flaregti_mod", &
            "Warning: using obsolete ENERGY column instead of PI column")
       call fitserr("ftgcno", file%status, ierr)
    end if

    ! get telescope module number column
    call ftgcno_w(file%unit, casesens, "TM_NR", file%tmnrcol, file%status)
    if(file%status>0) then
       ! backward compatibility look for CCDNR column instead
       file%status = 0
       call ftgcno(file%unit, casesens, "CCDNR", file%tmnrcol, file%status)
       if(file%status <= 0) call wrlog("flaregti_mod", &
            "Warning: using obsolete CCDNR column instead of TM_NR column")
       call fitserr("ftgcno", file%status, ierr)
    end if

    call ftgcno_w(file%unit, casesens, "FLAG", file%flagcol, file%status)
    call fitserr("ftgcno", file%status, ierr)

    ! get ranges of X and Y columns
    write(keyword,"(A,i1)") "TLMIN", file%xcol
    call ftgkyj_w(file%unit, keyword, file%xmin, comment, file%status)
    write(keyword,"(A,i1)") "TLMAX", file%xcol
    call ftgkyj_w(file%unit, keyword, file%xmax, comment, file%status)
    if(file%status /= 0) then
       ! Y min and max keywords missing
       file%xmin = -108000
       file%xmax =  108000
       file%status = 0
    end if

    write(keyword,"(A,i1)") "TLMIN", file%ycol
    call ftgkyj_w(file%unit, keyword, file%ymin, comment, file%status)
    write(keyword,"(A,i1)") "TLMAX", file%ycol
    call ftgkyj_w(file%unit, keyword, file%ymax, comment, file%status)
    if(file%status /= 0) then
       ! Y min and max keywords missing
       file%ymin = -108000
       file%ymax =  108000
       file%status = 0
    end if

    ! Keywords to convert from X and Y to WCS
    write(keyword,"(A,i1)") "TCDLT", file%xcol
    call ftgkyd_w(file%unit, keyword, file%cdltx, comment, file%status)
    call fitserr("ftgkyd", file%status, ierr)
    write(keyword,"(A,i1)") "TCDLT", file%ycol
    call ftgkyd_w(file%unit, keyword, file%cdlty, comment, file%status)
    call fitserr("ftgkyd", file%status, ierr)

    write(keyword,"(A,i1)") "TCRPX", file%xcol
    call ftgkyd_w(file%unit, keyword, file%crpxx, comment, file%status)
    call fitserr("ftgkyd", file%status, ierr)
    write(keyword,"(A,i1)") "TCRPX", file%ycol
    call ftgkyd_w(file%unit, keyword, file%crpxy, comment, file%status)
    call fitserr("ftgkyd", file%status, ierr)

    write(keyword,"(A,i1)") "TCRVL", file%xcol
    call ftgkyd_w(file%unit, keyword, file%crvlx, comment, file%status)
    call fitserr("ftgkyd", file%status, ierr)
    write(keyword,"(A,i1)") "TCRVL", file%ycol
    call ftgkyd_w(file%unit, keyword, file%crvly, comment, file%status)
    call fitserr("ftgkyd", file%status, ierr)

    ! flag selection: default is 0
    call ftgkyj_w(file%unit, "FLAGSEL", file%flagsel, comment, file%status)
    if(file%status /= 0) then
       file%status=0
       file%flagsel=0
    end if

  end subroutine evtfile_evt_col_info

!----------------------------------------------------------------------------
  subroutine copy_gti(gtiin, gtiout)
!----------------------------------------------------------------------------
! Make a copy of the GTI array.
!----------------------------------------------------------------------------
    type(tGTIs), dimension(7), intent(in) :: gtiin
    type(tGTIs), dimension(7), intent(out) :: gtiout

    integer :: idet
    do idet = 1, 7
       call gtiout(idet)%realloc(gtiin(idet)%get_num())
       gtiout(idet)%starts(:) = gtiin(idet)%starts(:)
       gtiout(idet)%stops(:) = gtiin(idet)%stops(:)
    end do

  end subroutine copy_gti

!----------------------------------------------------------------------------
  subroutine validate_gti(gti)
!----------------------------------------------------------------------------
    type(tGTIs), intent(in) :: gti

    integer :: ierr

    if ( .not. gti%is_legal() ) then
      ierr = 200
      call wrerr("validate_gti", ierr, "Invalid GTI")
    end if
  end subroutine validate_gti

!----------------------------------------------------------------------------
  subroutine validate_gtis(gtis)
!----------------------------------------------------------------------------
    type(tGTIs), dimension(7), intent(in) :: gtis

    integer :: idet
    do idet = 1, 7
      call validate_gti(gtis(idet))
    end do
  end subroutine validate_gtis

!----------------------------------------------------------------------------
  subroutine merge_gapfree_gti(gti)

    type(tGTIs), intent(inout) :: gti

    integer :: igti, ngti
    real(kind=real64), dimension(:), allocatable :: tstart, tstop

    ! This is to work around a bug where evprep or evtool produces GTIs with no gap
    tstart = gti%starts
    tstop = gti%stops
    ngti = gti%get_num()

    igti = 1
    do while (igti < ngti-1)
      if( tstop(igti) == tstart(igti+1) ) then
        call wrlog("flaregti_mod", "Warning: merging gap-free adjacent GTIs")

        ! merge GTIs
        tstop(igti) = tstop(igti+1)
        if( igti+1 < ngti ) then
          tstart(igti+1:ngti-1) = tstart(igti+2:ngti)
          tstop(igti+1:ngti-1) = tstop(igti+2:ngti)
        end if
        ngti = ngti - 1
      end if
      igti = igti + 1
    end do

    call gti%realloc(ngti)
    gti%starts(:) = tstart(1:ngti)
    gti%stops(:) = tstop(1:ngti)

    deallocate(tstart, tstop)

  end subroutine merge_gapfree_gti



!----------------------------------------------------------------------------
  subroutine evtfile_read_gtis(file, gtitype, gti, tmin, tmax)
!----------------------------------------------------------------------------
! Read GTI values for each detector from fits file
!----------------------------------------------------------------------------
    type(evtfile_t) :: file
    character(*), intent(in) :: gtitype                 ! type of gti to get
    type(tGTIs), dimension(7), intent(out) :: gti       ! output gtis
    real(kind=real64), intent(out) :: tmin, tmax        ! min and max times

    integer idet, ierr
    integer status
    logical :: anygtis

    character(len=64) :: message

    tmin = 9d19
    tmax = 0d0
    anygtis = .false.

    do idet = 1, 7
      status = 0
      call read_gtis(gtitype, idet, file%unit, gti(idet), status)

      if( status == 0 .and. gti(idet)%get_num() > 0 ) then
        anygtis = .true.

        tmin = min(tmin, gti(idet)%starts(1))
        tmax = max(tmax, gti(idet)%stops(gti(idet)%get_num()))

        call merge_gapfree_gti(gti(idet))

      else
        ! no gti - should we abort instead?
        write(message, "(A,i1)") "Warning: No input GTIs found for TM", idet
        call wrlog("flaregti_mod", message)
        call gti(idet)%realloc(0)
      end if
    end do

    if( .not. anygtis ) then
       ierr = 202
       call wrerr("evtfile_read_gtis", ierr, "No GTIs found in event file")
    end if

    call validate_gtis(gti)

  end subroutine evtfile_read_gtis

! Construct image of events
!----------------------------------------------------------------------------
  subroutine evtfile_mk_image(file, xmin, ymin, xybin, pimin, pimax, &
       gti, image, ierr)
!----------------------------------------------------------------------------
    type(evtfile_t) :: file
    integer, intent(in) :: xmin, ymin, xybin
    integer, intent(in) :: pimin, pimax
    type(tGTIs), dimension(7), intent(in) :: gti
    integer, intent(out) :: ierr
    integer, dimension(:,:), intent(out) :: image

    integer, parameter :: nblock=2880
    real(kind=real64), parameter :: nullvald = 0d0
    integer, parameter :: nullvalj=-999999
    integer :: undef

    integer :: i, irow, ival, xsize, ysize, ix, iy
    integer, dimension(nblock) :: x, y, pi, tmnr, flag
    real(kind=real64), dimension(nblock) :: time

    call evtfile_goto_hdu(file, "EVENTS", ierr)

    xsize = size(image, 1)
    ysize = size(image, 2)
    image = 0

    ! loop through table
    do irow=1, file%nrow, nblock
       ival = min(nblock, file%nrow-irow+1)

       ! read columns:
       ! time
       call ftgcvd_w(file%unit, file%tcol, irow, 1, ival, nullvald, time, undef, file%status)
       call fitserr("ftgcvd", file%status, ierr)
       ! x
       call ftgcvj_w(file%unit, file%xcol, irow, 1, ival, nullvalj, x, undef, file%status)
       call fitserr("ftgcvj", file%status, ierr)
       ! y
       call ftgcvj_w(file%unit, file%ycol, irow, 1, ival, nullvalj, y, undef, file%status)
       call fitserr("ftgcvj", file%status, ierr)
       ! pi
       call ftgcvj_w(file%unit, file%picol, irow, 1, ival, nullvalj, pi, undef, file%status)
       call fitserr("ftgcvj", file%status, ierr)
       ! tmnr
       call ftgcvj_w(file%unit, file%tmnrcol, irow, 1, ival, nullvalj, &
            tmnr, undef, file%status)
       call fitserr("ftgcvj", file%status, ierr)
       ! flag
       call ftgcvj_w(file%unit, file%flagcol, irow, 1, ival, nullvalj, &
            flag, undef, file%status)
       call fitserr("ftgcvj", file%status, ierr)

       do i = 1, ival

          if (pi(i) >= pimin .and. pi(i) <= pimax .and. &
               x(i) >= xmin .and. y(i) >= ymin .and. &
               gti(tmnr(i))%contain(time(i)) .and. &
               iand(flag(i), file%flagsel)==0 ) then

             ix = (x(i)-xmin)/xybin + 1
             iy = (y(i)-ymin)/xybin + 1

             if (ix <= xsize .and. iy <= ysize) then
                image(ix, iy) = image(ix, iy) + 1
             end if

          end if

       end do
    end do

  end subroutine evtfile_mk_image


!----------------------------------------------------------------------------
  subroutine evtfile_mk_lightcrv(file, xmin, ymin, xybin, &
       pimin, pimax, mask, gti, tbin, tmin, &
       rfovpix, tsbin, xsarr, ysarr, &
       lcurve, ierr)
!----------------------------------------------------------------------------

    type(evtfile_t) :: file

    integer, intent(in) :: xmin, ymin, xybin
    integer, intent(in) :: pimin, pimax               ! range of PI values to include
    integer(kind=int8), dimension(:,:), intent(in) :: mask
    type(tGTIs), dimension(7), intent(in) :: gti
    real(kind=real64), intent(in) :: tbin, tmin             ! lightcurve start and bin size
    real, intent(in) :: rfovpix                             ! FoV in pixels
    real(kind=real64), intent(in) :: tsbin                  ! area calcn time bin size
    real(kind=real64), dimension(:), intent(in) :: xsarr, ysarr ! x and y for area calc
    real(kind=real64), dimension(:), intent(out) :: lcurve  ! lightcurve count array
    integer, intent(out) :: ierr

    integer, parameter :: nblock=2880
    integer, parameter :: nullvalj = -999999
    real(kind=real64), parameter :: nullvald = 0d0
    integer :: undef

    integer :: i, ibin, irow, ival, ix, iy, xsize, ysize, tsidx
    integer, dimension(nblock) :: x, y, pi, tmnr, flag
    real(kind=real64), dimension(nblock) :: time

    call evtfile_goto_hdu(file, "EVENTS", ierr)

    xsize = size(mask,1)
    ysize = size(mask,2)

    lcurve=0d0

    ! loop through table
    do irow=1, file%nrow, nblock
       ival=min(nblock, file%nrow-irow+1)

       ! time
       call ftgcvd_w(file%unit, file%tcol, irow, 1, ival, nullvald, time, undef, file%status)
       call fitserr("ftgcvd", file%status, ierr)
       ! x
       call ftgcvj_w(file%unit, file%xcol, irow, 1, ival, nullvalj, x, undef, file%status)
       call fitserr("ftgcvj", file%status, ierr)
       ! y
       call ftgcvj_w(file%unit, file%ycol, irow, 1, ival, nullvalj, y, undef, file%status)
       call fitserr("ftgcvj", file%status, ierr)
       ! pi
       call ftgcvj_w(file%unit, file%picol, irow, 1, ival, nullvalj, pi, undef, file%status)
       call fitserr("ftgcvj", file%status, ierr)
       ! tmnr
       call ftgcvj_w(file%unit, file%tmnrcol, irow, 1, ival, nullvalj, &
            tmnr, undef, file%status)
       call fitserr("ftgcvj", file%status, ierr)
       ! flag
       call ftgcvj_w(file%unit, file%flagcol, irow, 1, ival, nullvalj, &
            flag, undef, file%status)
       call fitserr("ftgcvj", file%status, ierr)

       do i = 1, ival

          if (pi(i) >= pimin .and. pi(i) <= pimax .and. &
               gti(tmnr(i))%contain(time(i)) .and. &
               iand(flag(i), file%flagsel)==0 ) then

             ix = (x(i)-xmin)/xybin+1
             iy = (y(i)-ymin)/xybin+1

             if (ix <= xsize .and. ix >= 1 .and. iy <= ysize .and. iy >= 1) then
                if (mask(ix,iy) == 1) then
                   ! check whether event is within the nominal FoV for
                   ! this time bin (using the same time binning as for
                   ! the area calculations)
                   tsidx = int((time(i)-tmin)/tsbin) + 1
                   if( (x(i)-xsarr(tsidx))**2 + (y(i)-ysarr(tsidx))**2 < rfovpix**2 ) then
                      ibin = int((time(i)-tmin)/tbin) + 1
                      lcurve(ibin)=lcurve(ibin)+1d0
                   end if

                end if ! mask
             end if ! in image

          end if ! pi, gti

       end do
    end do

  end subroutine evtfile_mk_lightcrv

!----------------------------------------------------------------------------
  subroutine evtfile_prep_attitude (file, gti, tmin, tmax, tbin, ntbin, &
       lc_xarr, lc_yarr, ierr)
!----------------------------------------------------------------------------
! Gets attitude from event file
! Calulates average pixel X and Y for ntbin bins (width tbin)
!----------------------------------------------------------------------------

    type(evtfile_t) :: file
    type(tGTIs), dimension(7), intent(in) :: gti          ! gtis for detectors
    integer, intent(in) :: ntbin                          ! number of lightcurve bins,
    real(kind=real64), intent(in) :: tmin, tmax           ! start and end of lightcurve,
    real(kind=real64), intent(in) :: tbin                 ! lightcurve bin size,
    real(kind=real64), dimension(:), intent(out) :: lc_xarr, lc_yarr  ! x/yarr: position
                                                          ! of optical axis of a telescope
    integer, intent(out) :: ierr

    type(tAtt), dimension(:), allocatable :: att
    integer :: natt
    real(kind=real64), dimension(:), allocatable :: att_time, att_ra, att_dec
    real(kind=real64), dimension(:), allocatable :: att_xarr, att_yarr

    integer :: ic, ibin, ia
    integer :: iatime

    real(kind=real64), dimension(ntbin) :: xexpsum, yexpsum, expsum, expos
    real(kind=real64) :: xarr1, yarr1, gtime
    character(len=128) :: message

    xexpsum = 0d0
    yexpsum = 0d0
    expsum = 0d0

    ! iterate over detectors and get x, y, expos for time bins
    do ic=1,7
       ! get attitude for detector
       call get_attitude(ic, file%unit, att, ierr)
       if(ierr>0) then
          write(message, '(a,i1)') "Attitude missing for detector", ic
          call wrlog("flaregti_mod", message)
          cycle
       end if

       natt = size(att)
       allocate(att_xarr(natt), att_yarr(natt), att_time(natt), att_ra(natt), att_dec(natt))
       do ia=1,natt
         att_time(ia) = att(ia)%time
         att_ra(ia) = att(ia)%ra
         att_dec(ia) = att(ia)%dec
       end do

       ! convert ra,dec -> pixels in attitude
       call radec2pix(att_ra, att_dec, natt, &
            file%crvlx, file%crvly, file%crpxx, file%crpxy, file%cdltx, file%cdlty, &
            att_xarr, att_yarr)

       ! get exposure in time bins for detector module
       call calc_expos(gti, tbin, tmin, ntbin, ic, ic, expos)

       ! now get attitude for each time bin
       iatime = 1
       do ibin=1,ntbin
          if(expos(ibin) > 0d0) then
             ! time in middle of bin
             gtime = tmin + (dble(ibin)-0.5d0)*tbin
             if(gtime > tmax) gtime = tmax

             call linint_att2(gtime, att, att_xarr, att_yarr, &
                 iatime, xarr1, yarr1, ierr)

             ! add weighted value to output
             xexpsum(ibin) = xexpsum(ibin) + xarr1*expos(ibin)
             yexpsum(ibin) = yexpsum(ibin) + yarr1*expos(ibin)
             expsum(ibin) = expsum(ibin) + expos(ibin)
          end if
       end do

       deallocate(att, att_xarr, att_yarr, att_time, att_ra, att_dec)
    end do

    where(expsum > 0d0)
       lc_xarr = xexpsum / expsum
       lc_yarr = yexpsum / expsum
    elsewhere
       lc_xarr = huge(0d0)
       lc_yarr = huge(0d0)
    end where

    if(sum(expsum) == 0d0) then
       ierr = 3
       call wrerr("evtfile_prep_attitude", ierr, "All time bins missing attitude information")
    end if

  end subroutine evtfile_prep_attitude

!----------------------------------------------------------------------------
  subroutine evtfile_close(file, ierr)
    type(evtfile_t), intent(out) :: file
    integer :: ierr

    call ftclos(file%unit, file%status)
    call fitserr("ftclos", file%status, ierr)
  end subroutine evtfile_close

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!----------------------------------------------------------------------------
  subroutine update_flaregti(unit, fgti, extname)
!----------------------------------------------------------------------------
! Update flare GTIs in fits file
!----------------------------------------------------------------------------
    integer, intent(in) :: unit
    type(tGTIs), dimension(7), intent(in) :: fgti
    character(len=*), intent(in) :: extname  ! prefix of extension name

    integer, parameter :: nblock = 2880
    character(len=5), dimension(2), parameter :: gttype=(/"START", "STOP "/)
    character(len=2), dimension(2), parameter :: gtform=(/"1D", "1D"/)
    character(len=1), dimension(2), parameter :: gtunit=(/"s", "s"/)

    character(len=9) :: fgtinam
    integer :: irow, inum, hdutype, ierr, idet
    integer :: status

    do idet=1,7
       write(fgtinam, "(A,i1)") trim(extname), idet

       ! remove old FLARGTIx extension if present
       status = 0
       call ftmnhd(unit, 2, fgtinam, 0, status)
       if (status == 0) then
          call ftdhdu(unit, hdutype, status)
          call fitserr("ftdhdu", status, ierr)
       else
          status = 0
       end if

       ! write new flare gti extension to event file
       call ftibin(unit, fgti(idet)%get_num(), 2, gttype, gtform, gtunit, &
            fgtinam, 0, status)
       call fitserr("ftibin", status, ierr)
       do irow=1, fgti(idet)%get_num(), nblock
          inum = min(nblock, fgti(idet)%get_num()-irow+1)
          call ftpcld(unit, 1, irow, 1, inum, fgti(idet)%starts(irow:irow+inum-1), status)
          call fitserr("ftpcld", status, ierr)
          call ftpcld(unit, 2, irow, 1, inum, fgti(idet)%stops(irow:irow+inum-1), status)
          call fitserr("ftpcld", status, ierr)
       end do
    end do

  end subroutine update_flaregti

!----------------------------------------------------------------------------
  subroutine write_gtifile(filename, fgti)
!----------------------------------------------------------------------------
! Write a separate gti file
!----------------------------------------------------------------------------
    character(len=*), intent(in) :: filename
    type(tGTIs), dimension(7), intent(in) :: fgti

    integer :: unit, status, ierr

    status = 0
    call ftgiou(unit, status)
    call fitserr("ftgiou", status, ierr)
    call ftinit(unit, "!"//filename, 0, status)
    call fitserr("ftinit", status, ierr)

    call update_flaregti(unit, fgti, "STDGTI")

    call ftclos(unit, status)
    call fitserr("ftclos", status, ierr)

  end subroutine write_gtifile

!----------------------------------------------------------------------------
  subroutine write_mask(filename, mask, image)
!----------------------------------------------------------------------------
! Write mask and image to filename
!----------------------------------------------------------------------------
    character(len=*), intent(in) :: filename
    integer(kind=int8), dimension(:,:), intent(in) :: mask
    integer, dimension(:,:), intent(in) :: image

    integer, dimension(2) :: naxes
    integer :: unit, status, ierr

    ! TODO: write WCS information

    status = 0
    naxes = shape(mask)
    call ftgiou(unit, status)
    call fitserr("ftgiou", status, ierr)
    call ftinit(unit, "!"//filename, 0, status)
    call fitserr("ftinit", status, ierr)

    ! write mask to primary HDU
    call ftphps(unit,8,2,naxes,status)
    call fitserr("ftphps", status, ierr)
    call ftp2db(unit, 0, naxes(1), naxes(1), naxes(2), mask, status)
    call fitserr("ftp2db", status, ierr)

    ! add image HDU with image
    call ftiimg(unit, 32, 2, naxes, status)
    call fitserr("ftiimg", status, ierr)
    call ftp2dj(unit, 0, naxes(1), naxes(1), naxes(2), image, status)
    call fitserr("ftp2dj", status, ierr)
    call ftpkys(unit, "EXTNAME", "IMAGE", "Image used to create mask", status)
    call fitserr("ftpkys", status, ierr)

    call ftclos(unit, status)
    call fitserr("ftclos", status, ierr)

  end subroutine write_mask

!----------------------------------------------------------------------------
  subroutine write_lightcurve(filename, ntbin, time, lcurve, &
       rcurve, filtrcurve, threscurve, expos, area, xarr, yarr)
!----------------------------------------------------------------------------
!Write lightcurve to filename other arguments are time, count, rate,
!filtered rate, exposure, area and x,y position arrays
!----------------------------------------------------------------------------
    character(len=*), intent(in) :: filename
    integer, intent(in) :: ntbin
    real(kind=real64), dimension(:), intent(in) :: time, lcurve, rcurve, &
         filtrcurve, expos, area, xarr, yarr
    real, dimension(:), intent(in) :: threscurve

    integer, parameter :: nblock = 2880
    character(len=*), parameter :: lcnam = "LIGHTCRV"
    character(len=9), dimension(9), parameter :: lctype = (/ &
         "TIME     ", "COUNTS   ", "RATE     ", "FILTRATE ", &
         "THRESHOLD", "EXPOSURE ", "AREA     ", &
         "X        ", "Y        " /)
    character(len=2), dimension(9), parameter :: lcform = (/ &
         "1D", "1E", "1E", "1E", "1E", "1E", "1E", "1E", "1E" /)
    character(len=11), dimension(9), parameter :: lcunit = (/ &
         "s          ", "cts        ", "cts/s/deg^2", "cts/s/deg^2", &
         "cts/s/deg^2", "s          ", "deg^2      ", &
         "           ", "           " /)

    integer :: unit, status, ierr
    integer :: irow, ival

    status = 0
    call ftgiou(unit, status)
    call fitserr("ftgiou", status, ierr)
    call ftinit(unit, "!"//filename, 0, status)
    call fitserr("ftinit", status, ierr)
    call ftibin(unit, ntbin, 9, lctype, lcform, lcunit, lcnam, 0, status)
    call fitserr("fibin", status, ierr)

    do irow=1, ntbin, nblock
       ival = min(nblock, ntbin-irow+1)
       call ftpcld(unit, 1, irow, 1, ival, time(irow:irow+ival-1), status)
       call fitserr("ftpcld", status, ierr)
       call ftpcle(unit, 2, irow, 1, ival, real(lcurve(irow:irow+ival-1)), status)
       call fitserr("ftpcle", status, ierr)
       call ftpcle(unit, 3, irow, 1, ival, real(rcurve(irow:irow+ival-1)), status)
       call fitserr("ftpcle", status, ierr)
       call ftpcle(unit, 4, irow, 1, ival, real(filtrcurve(irow:irow+ival-1)), status)
       call fitserr("ftpcle", status, ierr)
       call ftpcle(unit, 5, irow, 1, ival, threscurve(irow:irow+ival-1), status)
       call fitserr("ftpcle", status, ierr)
       call ftpcle(unit, 6, irow, 1, ival, real(expos(irow:irow+ival-1)), status)
       call fitserr("ftpcle", status, ierr)
       call ftpcle(unit, 7, irow, 1, ival, real(area(irow:irow+ival-1)), status)
       call fitserr("ftpcle", status, ierr)
       call ftpcle(unit, 8, irow, 1, ival, real(xarr(irow:irow+ival-1)), status)
       call fitserr("ftpcle", status, ierr)
       call ftpcle(unit, 9, irow, 1, ival, real(yarr(irow:irow+ival-1)), status)
       call fitserr("ftpcle", status, ierr)
    end do
    call ftclos(unit, status)
    call fitserr("ftclos", status, ierr)

  end subroutine write_lightcurve

!----------------------------------------------------------------------------
! Work out which pixels in mask have a significant detection above the local
! background (measured in 5x5 box centred on examined pixel)
!----------------------------------------------------------------------------
  subroutine mk_mask (image, detml, mask, ierr)
!----------------------------------------------------------------------------

    integer, dimension(:,:), intent(in) :: image
    real, intent(in) :: detml
    integer(kind=int8), dimension(:,:), intent(out) :: mask
    integer, intent(out) :: ierr

    integer :: xsize, ysize, ix, iy
    real :: bkgcnt, soucnt

    ! background calculated from surrounding bsize*bsize pixels around
    ! central pixel
    integer, parameter :: bsize = 5

    xsize = size(image, 1)
    ysize = size(image, 2)

    mask = 0
    do ix = bsize/2+1, xsize-bsize/2
       do iy = bsize/2+1, ysize-bsize/2
          soucnt = max(image(ix,iy), 0)
          bkgcnt = max(sum(image(ix-bsize/2:ix+bsize/2,iy-bsize/2:iy+bsize/2)) - &
               int(soucnt), 1)
          if (poisl2(soucnt, bkgcnt/real(bsize**2-1), ierr) < detml) then
             mask(ix, iy) = 1
          end if
       end do
    end do

  end subroutine mk_mask


!----------------------------------------------------------------------------
  subroutine mk_area(gti, mask, xmin, ymin, xybin, rfovpix, tmin, &
       ntbin, ntsbin, tbin, tsbin, xsarr, ysarr, area)
!----------------------------------------------------------------------------
! Calculates unmasked area in deg^2 per lightcurve time bin
!----------------------------------------------------------------------------

    type(tGTIs), dimension(7), intent(in) :: gti    ! gtis for detector
    integer(kind=int8), dimension(:,:), intent(in) :: mask
    integer, intent(in) :: xmin, ymin, xybin, ntbin, ntsbin ! nbin:
    real, intent(in) :: rfovpix
    real(kind=real64), intent(in) :: tmin        ! minimum time
    real(kind=real64), intent(in) :: tbin, tsbin ! tbin: light curve bin size,
                                                 ! tsbin: step width in seconds for
                                                 !  area calculation
    real(kind=real64), dimension(:) :: xsarr, ysarr ! averaged position of opt axis
                                                 ! in sky pixels for each lightcurve time bin
    real(kind=real64), dimension(:) :: area      ! unmasked area in deg^2 per time bin
    real(kind=real64), dimension(:), allocatable :: norm

    real(kind=real64), dimension(:), allocatable :: expos ! exposure in area calc bins

    real, parameter :: sky2deg = 1.929012e-10 ! skypixel to deg^2
    integer :: xsize, ysize, xs, ys, ix, iy, ixs2, rfovipix, rfovipix2, is, imask, ibin

    ! radius of fov in bins
    rfovipix = int(rfovpix/xybin)+1
    rfovipix2 = rfovipix**2

    xsize = size(mask, 1)
    ysize = size(mask, 2)

    allocate (norm(ntbin))
    area = 0d0
    norm = 0d0

    ! get exposures in each time bin for area calculation
    allocate(expos(ntsbin))
    call calc_expos(gti, tsbin, tmin, ntsbin, 1, 7, expos)

    ! iterate over area calculation time bins
    do is=1, ntsbin
       ! skip area if no exposure in area bin
       if ( expos(is) == 0d0 ) cycle

       imask = 0

       ! position of optical axis in time bin
       xs = int((xsarr(is)-xmin)/xybin + 1)
       ys = int((ysarr(is)-ymin)/xybin + 1)
       if ( xs <= xsize+rfovipix .and. xs >= 1-rfovipix .and. &
            ys <= ysize+rfovipix .and. ys >= 1-rfovipix ) then
          ! is a possible contribution to area
          do ix = max(xs-rfovipix, 1), min(xs+rfovipix, xsize)
             ixs2 = (ix-xs)**2
             do iy = max(ys-rfovipix, 1), min(ys+rfovipix, ysize)
                if ((iy-ys)**2+ixs2<rfovipix2) imask = imask + mask(ix, iy)
             end do
          end do
       end if
       ! convert area time bin to lightcurve time bin
       ibin = min(ntbin, int((dble(is)-0.5d0)*tsbin/tbin+1d0))
       ! add contribution to area
       area(ibin) = area(ibin) + imask
       norm(ibin) = norm(ibin) + 1d0
    end do

    ! normalize
    where (norm>0d0) area = area/norm
    ! convert to area in deg^2
    area = area*xybin**2*sky2deg

    deallocate(norm, expos)

  end subroutine mk_area

!----------------------------------------------------------------------------
  subroutine calc_expos(gti, tbin, tmin, ntbin, detmin, detmax, expos)
!----------------------------------------------------------------------------
! Compute exposure in each time bin (summing all cameras)

    type(tGTIs), dimension(7), intent(in) :: gti    ! gtis for detectors
    real(kind=real64), intent(in) :: tbin, tmin     ! light curve bin size, start time
    integer, intent(in) :: ntbin                    ! number of time bins
    integer, intent(in) :: detmin, detmax           ! detector min and max to compute
    real(kind=real64), dimension(:), intent(out) :: expos  ! output exposure time (sec)

    integer :: idet, iexp, igti, igti1s
    integer, dimension(7) :: igti1                   ! current gti start index
    real(kind=real64) :: binexp, t0, t1

    igti1 = 1

    ! iterate over time bins
    do iexp = 1, ntbin

       t1 = tmin + iexp*tbin
       t0 = t1 - tbin
       binexp = 0d0

       ! iterate over detectors
       do idet = detmin, detmax

          igti1s = igti1(idet)
          do igti = igti1(idet), gti(idet)%get_num()
             ! if time bin start was before gti start and time bin end
             ! is after gti start
             if (t0 < gti(idet)%starts(igti) .and. t1 > gti(idet)%starts(igti)) then
                binexp = binexp + &
                     min(t1-gti(idet)%starts(igti), gti(idet)%stops(igti)-gti(idet)%starts(igti))
             end if
             ! if time bin start is after gti start and before gti end
             if (t0 >= gti(idet)%starts(igti) .and. t0 < gti(idet)%stops(igti)) then
                binexp = binexp + min(t1-t0, gti(idet)%stops(igti)-t0)
             end if

             ! avoid repeating over gtis which are before this time bin
             if (t1 >= gti(idet)%stops(igti)) then
                igti1s = igti
             end if

             ! skip gtis where there's no further contribution
             if (t1 < gti(idet)%stops(igti)) then
                exit
             end if
          end do
          igti1(idet) = igti1s
       end do ! idet

       expos(iexp) = binexp
    end do ! iexp

  end subroutine calc_expos

  !----------------------------------------------------------------------------
  ! remove bins with zero exposure and outside of sky pixel range
  !----------------------------------------------------------------------------
  integer function remove_empty_tbins(ntbin, expos, area, lcurve, time, xarr, yarr)
    ! number of time bins (modified)
    integer, intent(in) :: ntbin
    ! input arrays (modified)
    real(kind=real64), dimension(:), intent(inout) :: expos, area, lcurve, &
         time, xarr, yarr

    integer :: iexp, iexp0

    iexp0 = 1
    do iexp = 1, ntbin
       if (lcurve(iexp) == 0d0 .and. expos(iexp)>5d0 .and. area(iexp)>0.25d0) then
          call wrlog("flaregti_mod", &
               "Warning: time bin with no counts found in input GTI")
       end if

       if (expos(iexp) > 0d0 .and. area(iexp) > 0d0) then
          expos(iexp0) = expos(iexp)
          area(iexp0) = area(iexp)
          lcurve(iexp0) = lcurve(iexp)
          time(iexp0) = time(iexp)
          xarr(iexp0) = xarr(iexp)
          yarr(iexp0) = yarr(iexp)
          iexp0=iexp0+1
       end if
    end do
    remove_empty_tbins = iexp0-1
  end function remove_empty_tbins

!----------------------------------------------------------------------------
  subroutine mk_thresh(ntbin, lcurve, rcurve, xarr, yarr, area, exposure, &
       sarea, fovrad, detlike, xmin, xmax, ymin, ymax, ngrid, thresarr, &
       xg, yg, ierr)
!----------------------------------------------------------------------------

    integer, intent(in) :: ntbin                   ! number of lightcurve bins
    real(kind=real64), dimension(:), intent(in) :: lcurve, rcurve ! lightcurve counts & rates
    real(kind=real64), dimension(:), intent(in) :: xarr, yarr     ! x and y positions in time bins
    real(kind=real64), dimension(:), intent(in) :: area           ! area for each time bin
    real(kind=real64), dimension(:), intent(in) :: exposure       ! exposure for each time bin
    real, intent(in) :: sarea                      ! source area in deg^2
    real, intent(in) :: fovrad                     ! fov radius in pixels
    real, intent(in) :: detlike                    ! detection likelihood for sources
    integer, intent(in) :: xmin, xmax, ymin, ymax  ! pixel range for flare screening
    integer, intent(in) :: ngrid                   ! number of grid points per dimension
    real(kind=real32), dimension(:), intent(out) :: thresarr ! threshold array, size is
                                                  ! number of spatial grid points
    integer, dimension(:), intent(out) :: xg, yg   ! spatial grid points
    integer, intent(out) :: ierr

    real, parameter :: pi=acos(-1.0)

    real(kind=real64) :: bgsb, logratemin, logratemax, logstep, tint
    real(kind=real64) :: soucnt, sourate, min_sourate, thresh, best_thresh, fovrad2, xdelta, ydelta
    integer :: i, ithresh, igrid, it, ix, iy

    integer, parameter :: nstep = 100
    real, parameter :: scstart=1.0 , scstep=0.5

    ! count how many time bins for each point and bins below threshold
    ! (for debugging)
    integer :: tbinnum, threshnum

    ! compute grid of points as xg, yg
    xdelta=(xmax-xmin)/ngrid
    ydelta=(ymax-ymin)/ngrid
    igrid=0
    do ix=1,ngrid
       do iy=1,ngrid
          igrid=igrid+1
          xg(igrid) = int(xmin+(real(ix)-0.5)*xdelta)
          yg(igrid) = int(ymin+(real(iy)-0.5)*ydelta)
       end do
    end do

    fovrad2 = fovrad**2

    ! get minimum and maximum log rates
    logratemax = log(maxval(pack(rcurve(1:ntbin), rcurve(1:ntbin)>0)))
    logratemin = log(minval(pack(rcurve(1:ntbin), rcurve(1:ntbin)>0)))

    ! if all the rates are the same, we should step over a wide range
    if( abs(logratemax-logratemin)<1.0d-5 ) then
       logratemin = logratemin-2
       logratemax = logratemax+2
    end if

    ! equal size of log steps between these
    logstep = (logratemax-logratemin)/nstep

    ! iterate over grid points
    do i=1,igrid

       ! keep track of threshold which minimizes detectable rate
       min_sourate = huge(0d0)
       best_thresh = -1d0

       ! iterate threshold rate over rate bins (order not important)
       do ithresh = 0, nstep
          ! threshold is in units of counts/deg^2/sec
          thresh = exp(logratemax-ithresh*logstep)

          ! integration time and background surface brightness
          ! (counts/deg^2) corresponding to threshold thresh
          tint = 0d0
          bgsb = 0d0
          tbinnum = 0
          threshnum = 0
          do it=1,ntbin
             ! is radius of grid point within fov?
             if ((xg(i)-xarr(it))**2+(yg(i)-yarr(it))**2 < fovrad2) then
                tbinnum = tbinnum + 1

                if (rcurve(it) <= thresh .and. rcurve(it) > 0d0) then
                   tint = tint + exposure(it) ! <== apply vignetting here
                   bgsb = bgsb + lcurve(it)/area(it)
                   threshnum = threshnum + 1
                end if
             end if
          end do ! it

          ! get number of counts needed for detection likelihood=detlike detection
          soucnt=scstart
          do while (poisl2(real(soucnt),real(bgsb)*sarea,ierr)<detlike)
             soucnt=soucnt+scstep
          end do

          ! source count rate (counts/sec)
          sourate=soucnt/tint

          ! if source rate is better than previous threshold
          if (sourate < min_sourate) then
             min_sourate = sourate
             best_thresh = thresh
          end if

       end do ! ithresh

       ! if we fail to find a best threshold (likely due to too few
       ! time bins), we do not filter
       if(best_thresh < 0) then
          best_thresh = maxval(rcurve)+1
       end if

       thresarr(i) = real(best_thresh)

    end do ! i

  end subroutine mk_thresh

!----------------------------------------------------------------------------
  subroutine write_thresh_img(xmin, xmax, ymin, ymax, xybin, thresarr, &
       xg, yg, filename)
!----------------------------------------------------------------------------
! Write an image of the threshold rate to the fits filename given
!----------------------------------------------------------------------------
    integer, intent(in) :: xmin, xmax, ymin, ymax, xybin ! pixel range
    real(kind=real32), dimension(:), intent(in) :: thresarr ! threshold array, size is
                                                  ! number of spatial grid points
    integer, dimension(:), intent(in) :: xg, yg   ! spatial grid points
    character(len=*), intent(in) :: filename

    ! image to write
    real(kind=real32), dimension(:,:), allocatable :: threshimg

    integer :: xsize, ysize
    integer :: ix, iy
    integer, dimension(1) :: igrid
    real :: cx, cy

    integer :: status, unit, ierr
    integer, dimension(2) :: naxes

    xsize = (xmax-xmin) / xybin
    ysize = (ymax-ymin) / xybin

    allocate(threshimg(xsize, ysize))

    do iy = 1, ysize
       do ix = 1, xsize
          ! find nearest grid point to pixel
          cx = (ix-0.5)*xybin + xmin
          cy = (iy-0.5)*xybin + ymin
          igrid = minloc((xg-cx)**2+(yg-cy)**2)
          ! update pixel
          threshimg(ix, iy) = thresarr(igrid(1))
       end do
    end do

    ! write image to filename
    status = 0
    naxes = shape(threshimg)
    call ftgiou(unit, status)
    call fitserr("ftgiou", status, ierr)
    call ftinit(unit, "!"//filename, 0, status)
    call fitserr("ftinit", status, ierr)

    ! write threshimg to primary HDU
    call ftphps(unit, -32, 2, naxes, status)
    call fitserr("ftphps", status, ierr)
    call ftp2de(unit, 0, naxes(1), naxes(1), naxes(2), threshimg, status)
    call fitserr("ftp2db", status, ierr)

    call ftclos(unit, status)
    call fitserr("ftclos", status, ierr)

    deallocate(threshimg)

  end subroutine write_thresh_img

!----------------------------------------------------------------------------
  subroutine mk_common_flaregti(rcurve, time, xarr, yarr, ntbin, tbin, &
       thresarr, xg, yg, flaregti, filtrcurve, threscurve)
!----------------------------------------------------------------------------
! Iterate over time bins, find threshold closest to pointing position
! Exclude from GTI if rate is above threshold
!----------------------------------------------------------------------------

    real(kind=real64), dimension(:), intent(in) :: rcurve ! lightcurve rate
    real(kind=real64), dimension(:), intent(in) :: time ! midpoint of time bins
    real(kind=real64), dimension(:), intent(in) :: xarr, yarr !
    integer, intent(in) :: ntbin                   ! number of lightcurve bins
    real(kind=real64), intent(in) :: tbin          ! lightcurve bin size
    real, dimension(:), intent(in) :: thresarr     ! threshold array, size is number of grid points

    integer, dimension(:), intent(in) :: xg, yg    ! grid points
    ! output GTI
    type(tGTIs), intent(inout) :: flaregti
    ! output filtered lightcurve, where filtered bins have invalid values
    real(kind=real64), dimension(:), intent(out) :: filtrcurve

    real(kind=real64), dimension(:), allocatable :: starts, stops
    ! output threshold in each time bin
    real, dimension(:), intent(out) :: threscurve

    logical :: ingti   ! whether time bin is part of current gti
    integer :: i
    integer :: ngti
    integer, dimension(1) :: igrid

    ! maximum number of gtis is number of time bins
    allocate(starts(ntbin), stops(ntbin))
    ngti = 0
    ingti = .false.

    do i=1, ntbin

       ! nearest grid point for attitude for time bin
       igrid = minloc((xg-xarr(i))**2+(yg-yarr(i))**2)
       ! this is the threshold for the time bin
       threscurve(i) = thresarr(igrid(1))

       if (rcurve(i) <= threscurve(i)) then
          ! rate in this bin is below the threshold
          if (ingti) then
             ! update end time of previous gti to midpoint of bin
             stops(ngti) = time(i)+(tbin*0.5d0)
          else
             ! create new gti
             ngti = ngti + 1
             starts(ngti) = time(i)-(tbin*0.5d0)
             stops(ngti) = time(i)+(tbin*0.5d0)
          end if
          ingti=.true.
          filtrcurve(i) = rcurve(i)
       else
          ingti=.false.
          ! set invalid rate in filtered rate
          filtrcurve(i) = -1
       end if

    end do

    call flaregti%realloc(ngti)
    flaregti%starts(:) = starts(1:ngti)
    flaregti%stops(:) = stops(1:ngti)

    deallocate(starts, stops)

  end subroutine mk_common_flaregti

!----------------------------------------------------------------------------
  subroutine clean_gti(gti)
!----------------------------------------------------------------------------
! Remove zero length GTIs and GTIs adjoining each other
!----------------------------------------------------------------------------

    type(tGTIs), intent(inout) :: gti

    integer :: i, j
    logical :: movenext
    real(kind=real64), dimension(:), allocatable :: starts, stops
    integer :: ngti

    ngti = gti%get_num()
    starts = gti%starts
    stops = gti%stops

    i = 1
    do while( i <= ngti )
      movenext = .true.

      if( starts(i) == stops(i) ) then
        ! replace this GTI by the next
        do j=i+1, ngti
          starts(j-1) = starts(j)
          stops(j-1) = stops(j)
        end do
        ngti = ngti - 1
        movenext = .false.
      else
        if( i+1<=ngti ) then
          ! is this a GTI joining the next?
          if( starts(i+1) == stops(i) ) then
            ! merge then replace next with later ones
            stops(i) = stops(i+1)
            do j=i+2, gti%get_num()
              starts(j-1) = starts(j)
              stops(j-1) = stops(j)
            end do
            ngti = ngti - 1
            movenext = .false.
          end if
        end if
      end if

      if(movenext) then
        i = i + 1
      end if

    end do

    call gti%realloc(ngti)
    gti%starts(:) = starts(1:ngti)
    gti%stops(:) = stops(1:ngti)

    deallocate(starts, stops)

  end subroutine clean_gti



!----------------------------------------------------------------------------
  subroutine merge_flaregti(gti, comflaregti, mgflaregti, ierr)
!----------------------------------------------------------------------------
! Merge common flaregti in comflaregti with existing gtis, producing
! mgflaregti
!----------------------------------------------------------------------------

    type(tGTIs), dimension(7), intent(in) :: gti
    type(tGTIs), intent(in) :: comflaregti
    type(tGTIs), dimension(7), intent(out) :: mgflaregti
    integer, intent(out) :: ierr
    integer :: idet

    ierr = 0
    do idet=1,7
      call mgflaregti(idet)%realloc(gti(idet)%get_num())
      mgflaregti(idet)%starts(:) = gti(idet)%starts(:)
      mgflaregti(idet)%stops(:) = gti(idet)%stops(:)

      call mgflaregti(idet)%and_with(comflaregti)

      ! clean up results
      call clean_gti(mgflaregti(idet))
    end do

  end subroutine merge_flaregti

!--------------------------------------------------------------------
! Below are math functions
!--------------------------------------------------------------------
! Jeremy Sanders: Change usage of gammln -> log_gamma which slightly
! improves accuracy in some cases and removes redundant code

!--------------------------------------------------------------------
  real function poisl2(measur, expect, ierr)
!--------------------------------------------------------------------

!  poiss. likelihood

!  author : s.doebereiner date: 11-sep-1996

!  get logarithm of the cumulative poisson probability (likelihood)
!  to have 'measur' or more counts if 'expect' counts are expected.

!  expect              r4  i/? expected value (background)
!  ierr                i4    o = 0 no error
!                              = error occured in ppois
!  measur              i4  i/? measured value
!  poisl2              r4      likelihood

    real, intent(in) :: expect, measur
    integer, intent(out) :: ierr

    ierr = 0

    if (measur > 0) then
       poisl2 = gammp(measur, expect, ierr)
    else
       poisl2 = 0.
    endif

  end function poisl2

!----------------------------------------------------------------
  real function gammp (a, x, ierr)
!----------------------------------------------------------------

!  returns the negative logarithm of the
!  incomplete gamma function p(a,x)
!  (regularized/normalized lower incomplete gamma function)
!  ( this is -numpy.log(scipy.special.gammainc(a, x)) )

!  author : s.doebereiner date: 11-sep-1996
!  returns the incomplete gamma function p(a,x)
!  in the maximum-likelihood representation -ln(p)
!  a and x must be >0 !
!  modified routines gcf and gser from press et al., numerical recipes,
!   6.2, p.162

    real, intent(in) :: a, x
    integer, intent(out) :: ierr

    real :: gln, gold
    real :: a0, a1, b0, b1, fac, an, ana, anf, g
    real :: ap, sum, del, gammcf
    integer :: n

    real, parameter :: eps = 3.e-7
    integer, parameter :: itmax = 1000
    character(*), parameter :: rtname = 'gammp'

    ierr = 1
    gammp = 0.

    if (x < a+1) then
       ! series representation
       gln = log_gamma(a)
       ap = a
       sum = 1./a
       del = sum
       do n = 1, itmax
          ap = ap+1.
          del = del*x/ap
          sum = sum+del
          if (abs(del) < abs(sum)*eps) then
             ierr = 0
             exit
          end if
       end do
       gammp = -alog(sum) + x - a*alog(x) + gln

    else

       gln = log_gamma(a)
       gold = 0.
       a0 = 1.
       a1 = x
       b0 = 0.
       b1 = 1.
       fac = 1.

       do n = 1, itmax
          an = float(n)
          ana = an-a
          a0 = (a1+a0*ana)*fac
          b0 = (b1+b0*ana)*fac
          anf = an*fac
          a1 = x*a0+anf*a1
          b1 = x*b0+anf*b1
          if (a1 /= 0) then
             fac = 1./a1
             g = b1*fac
             if (abs((g-gold)/g) < eps) then
                ierr = 0
                exit
             end if
             gold = g
          end if
       end do
       gammcf = exp(-x + a*alog(x) - gln) * g
       gammp = -alog(1.-gammcf)

    end if

  end function gammp

!----------------------------------------------------------------
  real function gammq (a, xv, ierr)
!----------------------------------------------------------------

!  returns the negative logarithm of the
!  complement of the incomplete gamma function p(a,x)

!  (does not appear as accurate as scipy scipy.special.gammaincc
!   according to mathematica (e.g. inputs 40, 20) - change?)

!  author : s.doebereiner date: 11-sep-1996
!  returns the incomplete gamma function p(a,x)
!  in the maximum-likelihood representation -ln(p)
!  a and x must be >0 !
!  modified routines gcf and gser from press et al., numerical recipes,
!   6.2, p.162

    real, intent(in) :: a, xv
    integer, intent(out) :: ierr

    real :: x, gln, gold
    real :: a0, a1, b0, b1, fac, an, ana, anf, g
    real :: ap, sum, del, gamser
    integer :: n

    real, parameter :: eps = 3.e-7
    integer, parameter :: itmax = 1000
    character(*), parameter :: rtname = 'gammq'

    ierr = 0
    gammq = 0.

    x = max(xv, eps)

    ierr = 1
    if (x < a+1) then
       gln = log_gamma(a)
       ap = a
       sum = 1./a
       del = sum
       do n = 1, itmax
          ap = ap+1.
          del = del*x/ap
          sum = sum+del
          if (abs(del) < abs(sum)*eps) then
             ierr = 0
             exit
          end if
       end do

       gamser = sum * exp( -x + a*log(x) - gln)
       gammq = -alog(1.0-gamser)

    else

       gln = log_gamma(a)
       gold = 0.
       a0 = 1.
       a1 = x
       b0 = 0.
       b1 = 1.
       fac = 1.

       do n = 1, itmax
          an = float(n)
          ana = an-a
          a0 = (a1+a0*ana)*fac
          b0 = (b1+b0*ana)*fac
          anf = an*fac
          a1 = x*a0+anf*a1
          b1 = x*b0+anf*b1
          if (a1 /= 0) then
             fac = 1./a1
             g = b1*fac
             if (abs((g-gold)/g) < eps) then
                ierr = 0
                exit
             end if
             gold = g
          end if
       end do

       gammq = x - a*alog(x) + gln - alog(g)

    end if

  end function gammq

!----------------------------------------------------------------
  real function gammln (a)
!----------------------------------------------------------------
! Log gamma function, wrapping library
!----------------------------------------------------------------
    real, intent(in) :: a
    gammln = log_gamma(a)
  end function gammln

end module flaregti_mod
