! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
!program test_radec2pix
!
!implicit none
!integer, parameter :: dp = selected_real_kind(11,30)
!
!real(dp), dimension(4) :: ra = (/ 0d0, 90d0, 180d0, 270d0 /)
!real(dp), dimension(4) :: dec = (/ -89d0, -89d0, -89d0, -89d0 /)
!real(dp), dimension(4) :: x, y
!
!call radec2xy (ra, dec, 4, 0d0, 90d0, 0.5d0, 0.5d0, -1d-3, 1d-3, x, y) 
!print *, x
!print *, y
!
!ra = (/ 0d0, 0d0, 1d0, 1d0 /)
!dec = (/ 0d0, 1d0, 0d0, 1d0 /)
!
!call radec2pix (ra, dec, 4, 0d0, 0d0, 0.5d0, 0.5d0, -1d-3, 1d-3, x, y) 
!print *, x
!print *, y
!
!end

module radec2pix_mod

use param

contains

  subroutine radec2pix (ra, de, n, ra0, de0, x0, y0, xd, yd, x, y)
  !-----------------------------------------------------------------------
  ! HB 20100326
  !-----------------------------------------------------------------------     
  ! This routine transforms a vector of ra[deg], dec[deg] sky locations 
  ! into x, y pixel coordinates (parallel projection) for a set of WCS 
  ! keywords. We use parallel projection because the transformation 
  ! requires less computation than other tangential projections 
  ! (formulae taken from Konrad Dennerl's "Survey II" document). 
  !-----------------------------------------------------------------------
  !
  ! input:
  !    ra      r*8 1d-array  deg 
  !    de      r*8 1d-array  deg
  !    n       i*4                  number of positions to transform
  !    ra0     r*8           deg    sky position of reference pixel (WCS)
  !    dec0    r*8           deg    sky position of reference pixel (WCS)
  !    x0      r*8                  reference pixel from WCS keywords       
  !    y0      r*8                  reference pixel from WCS keywords
  !    xd, yd  r*8           deg    pixel size (increment) from WCS
  ! output:
  !    x       r*8 1d-array         
  !    y       r*8 1d-array         

  implicit none
  integer, parameter :: dp = selected_real_kind(11,30)

  real(dp), dimension(*), intent(in) :: ra, de
  integer, intent(in) :: n
  real(dp), intent(in) :: ra0, de0, x0, y0, xd, yd

  real(dp), dimension(*), intent(out) :: x, y

  real(dp) :: rad2xpix, rad2ypix
  real(dp) :: ra0r, de0r
  real(dp) :: sr0, cr0, sd0, cd0

  real(dp), dimension(n) :: rar, der
  real(dp), dimension(n) :: sr, cr, sd, cd
  real(dp), dimension(n) :: x1, y1, y2, x3, y3 

  real(dp), parameter :: pi = 2d0*dacos(0d0), deg2rad = pi/180d0

  rad2xpix = 1d0/(xd*deg2rad)
  rad2ypix = 1d0/(yd*deg2rad)

  ra0r = ra0*deg2rad
  de0r = de0*deg2rad

  sr0 = dsin(ra0r)  
  cr0 = dcos(ra0r)

  sd0 = dsin(de0r)  
  cd0 = dcos(de0r)

! vector operations start here; could be changed into a 
! traditional do loop if this turns out to be faster

  rar = ra(1:n)*deg2rad
  der = de(1:n)*deg2rad

  sr = dsin(rar)    
  cr = dcos(rar)

  sd = dsin(der)
  cd = dcos(der)

  x1 = cd*cr
  y1 = cd*sr

  y2 = -cr0*x1-sr0*y1
  
  x3 = -sr0*x1+cr0*y1
  y3 =  sd0*y2+cd0*sd

  x(1:n) = x0 + x3*rad2xpix
  y(1:n) = y0 + y3*rad2ypix   

  end subroutine radec2pix

end module radec2pix_mod
