! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
program radec2xy

!-----------------------------------------------------------------------------
! HBrunner 20100430
!
! Program calculates X and Y sky pixel FITS table columns from input RA 
! and DEC columns, given an RA and DEC projection centre provided on the 
! command line. Input parameters:
! 
! file     FITS file name
! ra0      RA [deg] of projection centre
! dec0     DEC [deg] of projection centre 
!
! If not specified otherwise (FTOOLS/CFITIO convention), the RA and DEC columns
! are searched in the first FITS table extension. 
!-----------------------------------------------------------------------------
! Routines called:
!
!   RADEC2XY       performs local parallel projection of a vector of
!                    RA and DEC values, given a set of WCS keyword values    
!   KEYNAME        adds column number to WCS keyword name
!-----------------------------------------------------------------------------
! SASS environment calls:
! =======================   
!
!   Parameter i/o:
!   --------------
!
!   OPAST          opens either the PIL (interactive mode) or SASS (pipeline
!                    mode) parameter file depending on the environment variable 
!                    SASS_PARINT (value "pipeline": pipeline mode, not set or 
!                    any other value: interactive mode); in pipeline mode, the
!                    SASS logging environment is initialized and a status 
!                    message is written to the SASS parameter file
!
!                    Argument list:
!                    IERR      I4        O    error status
!                                                                            
!   CPAST          closes parameter file (PIL or SASS); in pipeline mode, the
!                    program exit status is written to the SASS parameter file
!
!                    Argument list:
!                    STATUS    I4       I     exit status of program; 
!                                               0: successful completion 
!                                              >1: failed
!                    IERR      I4        O    error status of routine
!                   
!   RGET<t>        reads a parameter of the specified type <t> from the PIL or
!                    the SASS parameter file; supported values for <t>: 
!                    J (i4), I (i2), B (byte), D (r8), C (character string), 
!                    L (logical), or the corresponding array data types JA, IA,
!                    BA, DA, FA, CA; if the last character of the routine name is 
!                    omitted, the data type is inferred from variable VALUE/S.   
!                       
!                    Argument list (single parameter):
!                    PARAM     C*(*)    I     name of parameter
!                    VALUE     <t>       O    parameter value, type as 
!                                               specified by <t>
!                    IERR      I4        O    error status of routine
!
!                    Argument list (array parameter):
!                    PARAM     C*(*)    I     name of parameter
!                    ISIZE     I4       I     size of array
!                    VALUES    <t>       O    array of returned values
!                    ICOUNT    I4        O    number of returned values
!                    IERR      I4        O    error status of routine    
!                       
!   Logging and error handling:   
!   ---------------------------
!
!   WRLOG          writes name of calling routine and arbitrary text string to 
!                    stdout and SASS log file (if pipeline mode is set);
!                    output is suppressed if SASS_VERBOSITY is set to "sparse";
!                    ideally, output using the write or print statement should
!                    be avoided as this will not be logged - use WRLOG or 
!                    WRCHATTY instead            
!
!                    Argument list:
!                    RTNAME    c*(*)   I      name of calling routine
!                    TEXT      c*(*)   I      output text                      
!
!   WRCHATTY       same as wrlog but output only gets written if SASS_VERBOSITY
!                    is set to "chatty"
!
!                    Argument list:
!                    RTNAME    c*(*)   I      name of calling routine
!                    TEXT      c*(*)   I      output text                      
!
!   WRERR          writes name of calling routine, error level, and abitrary
!                    text string to stdout and SASS log file (if pipeline 
!                    mode is set)
! 
!                    Error levels:  1  warning
!                                   2  error but attempt to continue program
!                                   3  severe error; no point in continuing
!                                        program execution, calling routine is
!                                        responsible to take appropriate action                                   
!                                                            
!                                   4  severe error; program is terminated
!                                        immediately (identical to wrstop)
!
!                    Argument list:
!                    RTNAME    c*(*)   I      name lf calling routine
!                    IERR      I4      I      error level          
!                    TEXT      C*(*)   I      output text                   
!                  
!   WRSTOP         sets error level to 3, calls WRERR and CPAST and then
!                    terminates program execution; note that WRSTOP typically
!                    should be called in the main program - in the case of
!                    severe errors in subroutines, WRERR with error level 3 
!                    should be called and the error status should be passed
!                    to the calling routine
!
!                    Argument list:
!                    RTNAME    C*(*)   I      name of task
!                    TEXT      C*(*)   I      output text          
!             
! Environment variables:
! ----------------------
!
! SASS_PARINT      values: "pipeline"                  pipeline (SASS) mode
!                          any other value or not set: interactive (PIL) mode
!
! SASS_VERBOSITY   values: "chatty"                    output from WRCHATTY, 
!                                                        WRLOG, and WRERR
!                          "sparse"                    only WRERR output
!                          any other value or not set: WRERR and WRLOG output
!-----------------------------------------------------------------------------

  use param
  use radec2pix_mod

  implicit none
  integer, parameter :: dp = selected_real_kind(11,30)

! CFITSIO constants
  integer, parameter :: readwrite = 1 
  logical, parameter :: casesens = .false.
  integer, parameter :: nblock=2880
  real(dp), parameter :: nullval = -1d3

! column buffers
  real(dp), dimension(nblock) :: ra, dec, x, y

! WCS keywords
  real(dp), parameter :: x0=0d0, y0=0d0, xd=-0.05/3600d0, yd=0.05/3600d0
  real(dp) :: ra0, dec0
  integer :: xtlmin=999999999, xtlmax=-999999999, ytlmin=999999999, ytlmax=-999999999
  
  integer :: unit, irow, nrow, ival
  integer :: racol, decol, xcol, ycol
  integer :: status = 0, ierr

  character(len=256) :: file
  character(len=80) :: line
  character(len=8) :: name
  logical :: undef

! open parameter interface, get command line parameters
  call opast(ierr)

  call rgetc("file", file, ierr)
  call rgetd("ra0", ra0, ierr)
  call rgetd("dec0", dec0, ierr)
  if (ierr > 0) call wrstop("radec2xy", "Couldn't read parameters")

! open FITS file
  call ftgiou(unit,status)
  call fitserr("ftgiou", status, ierr)
  call fttopn(unit, file, readwrite, status)
  call fitserr("fttopn", status, ierr)
  if (status > 0) call wrstop("radec2xy", "Can't read event table")

! get row and column info
  call ftgnrw(unit, nrow, status)
  call fitserr("ftgnrw", status, ierr)
  call ftgcno(unit, casesens, "RA", racol, status)
  call fitserr("ftgcno", status, ierr)
  call ftgcno(unit, casesens, "DEC", decol, status)
  call fitserr("ftgcno", status, ierr)
  if (status > 0) call wrstop("radec2xy", "Can't read event table")

! look for X, Y columns, if not present create them
  call ftgcno(unit, casesens, "X", xcol, status)
!  call fitserr("ftgcno", status, ierr)
  if (status > 0) then
    status=0
    xcol=max(racol,decol)+1
    call fticol(unit, xcol, "X", "J", status)
    call fitserr("fticol", status, ierr)
    if (status == 0) call wrlog("radec2xy", "Column 'X' not found; created")
  end if
  call ftgcno(unit, casesens, "Y", ycol, status)
!  call fitserr("ftgcno", status, ierr)
  if (status > 0) then
    status=0
    ycol=xcol+1
    call fticol(unit, ycol, "Y", "J", status)
    call fitserr("fticol", status, ierr)
    if (status == 0) call wrlog("radec2xy", "Column 'Y' not found; created")
  end if
  if (status > 0) call wrstop("radec2xy", "Can't create columns")

! write WCS keywords
  call keyname("TCTYP", xcol, name, ierr)
  call ftukys(unit, name, "RA---SIN", "WCS type: RA parallel projection", status)
  call fitserr("ftukys", status, ierr)
  call keyname("TCUNI", xcol, name, ierr)
  call ftukys(unit, name, "deg     ", "WCS physical unit of X axis", status)  
  call fitserr("ftukys", status, ierr)
  call keyname("TCRPX", xcol, name, ierr)
  call ftukyd(unit, name, x0, -14, "WCS X axis reference pixel", status)
  call fitserr("ftukyd", status, ierr)
  call keyname("TCRVL", xcol, name, ierr)
  call ftukyd(unit, name, ra0, -14, "RA at reference pixel", status)
  call fitserr("ftukyd", status, ierr)
  call keyname("TCDLT", xcol, name, ierr)
  call ftukyd(unit, name, xd, -14, "X increment at reference pixel", status)
  call fitserr("ftukyd", status, ierr)

  call keyname("TCTYP", ycol, name, ierr)
  call ftukys(unit, name, "DEC--SIN", "WCS type: DEC parallel projection", status)
  call fitserr("ftukys", status, ierr)
  call keyname("TCUNI", ycol, name, ierr)
  call ftukys(unit, name, "deg     ", "WCS physical unit of Y axis", status)
  call fitserr("ftukys", status, ierr)
  call keyname("TCRPX", ycol, name, ierr)
  call ftukyd(unit, name, y0, -14, "WCS YY axis reference pixel", status) 
  call fitserr("ftukyd", status, ierr)
  call keyname("TCRVL", ycol, name, ierr)
  call ftukyd(unit, name, dec0, -14, "DEC at reference pixel", status)
  call fitserr("ftukyd", status, ierr)
  call keyname("TCDLT", ycol, name, ierr)
  call ftukyd(unit, name, yd, -14, "Y increment at reference pixel", status)
  call fitserr("ftukyd", status, ierr)
  if (ierr == 3 .or. status > 0) call wrstop("radec2xy", "Couldn't write WCS keywords")    

  call ftukys(unit, "REFXCTYP", "RA---SIN", "WCS type: RA parallel projection", status) 
  call fitserr("ftukys", status, ierr)
  call ftukys(unit, "REFXCUNI", "deg     ", "WCS physical unit of X axis", status)  
  call fitserr("ftukys", status, ierr)
  call ftukyd(unit, "REFXCRPX", x0, -14, "WCS X axis reference pixel", status)
  call fitserr("ftukyd", status, ierr)
  call ftukyd(unit, "REFXCRVL", ra0, -14, "RA at reference pixel", status)
  call fitserr("ftukyd", status, ierr)
  call ftukyd(unit, "REFXCDLT", xd, -14, "X increment at reference pixel", status)
  call fitserr("ftukyd", status, ierr)

  call ftukys(unit, "REFYCTYP", "RA---SIN", "WCS type: DEC parallel projection", status) 
  call fitserr("ftukys", status, ierr)
  call ftukys (unit, "REFYCUNI", "deg     ", "WCS physical unit of Y axis", status)  
  call fitserr("ftukys", status, ierr)
  call ftukyd(unit, "REFYCRPX", y0, -14, "WCS Y axis reference pixel", status)
  call fitserr("ftukyd", status, ierr)
  call ftukyd(unit, "REFYCRVL", dec0, -14, "DEC at reference pixel", status)
  call fitserr("ftukyd", status, ierr)
  call ftukyd(unit, "REFYCDLT", yd, -14, "Y increment at reference pixel", status)
  call fitserr("ftukyd", status, ierr)
  if (ierr == 3 .or. status > 0) call wrstop("radec2xy", "Couldn't modify REF keywords")    

! loop through table   
  do irow=1, nrow, nblock
    ival=min(nblock,nrow-irow+1)

    call ftgcvd(unit, racol, irow, 1, ival, nullval, ra, undef, status)
    call fitserr("ftgcvd", status, ierr)
    call ftgcvd(unit, decol, irow, 1, ival, nullval, dec, undef, status)
    call fitserr("ftgcvd", status, ierr)

    call radec2pix(ra, dec, ival, ra0, dec0, x0, y0, xd, yd, x, y)

    xtlmin=min(xtlmin,nint(minval(x(1:ival))))
    xtlmax=max(xtlmax,nint(maxval(x(1:ival))))
    ytlmin=min(ytlmin,nint(minval(y(1:ival))))
    ytlmax=max(ytlmax,nint(maxval(y(1:ival))))

    call ftpclj(unit, xcol, irow, 1, ival, nint(x(1:ival)), status)
    call fitserr("ftpclj", status, ierr)
    call ftpclj(unit, ycol, irow, 1, ival, nint(y(1:ival)), status)
    call fitserr("ftpclj", status, ierr)

    if (status > 0) call wrstop("radec2xy", "Couldn't read or write FITS table data")

    write(line, *) irow+ival-1, " table rows processed"
    call wrchatty("radec2xy", trim(line))

  end do

! update tlmin/max keywords
  call keyname("TLMIN", xcol, name, ierr)
  call ftukyj(unit, name, xtlmin, "Minimum of X", status)
  call fitserr("ftukyj", status, ierr)
  call keyname("TLMAX", xcol, name, ierr)
  call ftukyj(unit, name, xtlmax, "Maximum of X", status)
  call fitserr("ftukyj", status, ierr)
  call keyname("TLMIN", ycol, name, ierr)
  call ftukyj(unit, name, ytlmin, "Minimum of Y", status)
  call fitserr("ftukyj", status, ierr)
  call keyname("TLMAX", ycol, name, ierr)
  call ftukyj(unit, name, ytlmax, "Maximum of Y", status)
  call fitserr("ftukyj", status, ierr)
  if (ierr == 3 .or. status > 0) call wrstop("radec2xy", "Couldn't update TLMIN/MAX keywords")    

  call ftukyj(unit, "REFXLMIN", xtlmin, "Minimum of X", status)
  call fitserr("ftukyj", status, ierr)
  call ftukyj(unit, "REFXLMAX", xtlmax, "Maximum of X", status)
  call fitserr("ftukyj", status, ierr)
  call ftukyj(unit, "REFYLMIN", ytlmin, "Minimum of Y", status)
  call fitserr("ftukyj", status, ierr)
  call ftukyj(unit, "REFYLMAX", ytlmax, "Maximum of Y", status)
  call fitserr("ftukyj", status, ierr)
  if (ierr == 3 .or. status > 0) call wrstop("radec2xy", "Couldn't modify REF keywords")    

  call ftclos(unit, status)
  call fitserr("ftclos", status, ierr)
  if (status > 0) call wrstop("radec2xy", "Couldn't close FITS table")
  write (line, *) nrow, " table rows processed"
  call wrlog("radec2xy", trim(line))

  call cpast(status, ierr)    

end

subroutine keyname(base, colnum, name, ierr)

  use param

  implicit none
  character(*), intent(in) :: base
  integer, intent(in) :: colnum
  character(len=8), intent(out) :: name
  integer, intent(out) :: ierr
  character(len=3) :: ccol

  if (colnum .lt. 10) then
    write(ccol, '(i1)') colnum
  else if (colnum .lt. 100) then
    write(ccol, '(i2)') colnum
  else if (colnum .lt. 1000) then
    write(ccol, '(i3)') colnum
  else
    ierr=3
    call wrerr("keyname", ierr, "Too many table columns")
  end if

  name = base // ccol

end
