! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module slib_constants


  use iso_fortran_env

   
  implicit none

  ! some of the constants below are in the tool erosita
  ! module with the same name, hence "my" at the front
  INTEGER, PARAMETER :: MAX_NUM_TELESCOPES=7
  DOUBLE PRECISION, PARAMETER :: myPI = 4.D0* ATAN(1.D0)
  DOUBLE PRECISION, PARAMETER :: myRAD2DEG = 180.0D0/myPI
  DOUBLE PRECISION, PARAMETER :: myDEG2RAD = myPI/180.0D0
  
  DOUBLE PRECISION, PARAMETER :: RAW_PIX_MICRON=0.075D0
  DOUBLE PRECISION, PARAMETER :: RAW_PIX_DEG=ATAN(RAW_PIX_MICRON/1600.0D0) * myRAD2DEG
  DOUBLE PRECISION, PARAMETER :: RAW_CCD_EXT=384.0D0

  !INTEGER, PARAMETER :: MAX_PSF_IMG_SIZE=500



  !atttiude file related stuff
  type :: ATT
     
     ! this is the attitude for a given telescope
     double precision, dimension (:), allocatable :: RA, DEC, TIME, ROLL
     
     ! size is the number of attitude extension rows
     ! ACTIVE_TEL is the instrument number that the extension corresponds to
     integer ::  SIZE, ACTIVE_TEL

  end type ATT


  !individual event related parameters
  type :: EVENT

     ! this is the attitude info ffor each telescope
     ! EV2SRC->lookback to src id for easy access
     ! ROLL will be the telescope ROLL ANGLE at the time of arrival
     ! ROLL ANGLE IS USED TO ROTATE PHOTON PSF. IT SHOULD BE
     ! UNITS OF DEG. IT IS CONVERTED TO RAD BY THE ROTATION ROUTINE
     integer :: nside=4096 ! corresponds to resolution ~51"
     logical :: HEALSORT=.FALSE. ! whether events have been sorted by position
     double precision, dimension (:), allocatable :: RA, DEC                          
     double precision, dimension (:), allocatable :: XD, YD, ZD
     integer, dimension (:), allocatable :: RAWX, RAWY                                                                            
     double precision, dimension (:), allocatable :: ENERGY, TIME, ROLL                                                           
     integer, dimension (:), allocatable :: CCDNR, EV2SRC                                                                         
     ! should I add shapelet coeff for each event? Or too MEM INTENSIVE?                                                          
     ! number of events in event file                                                                                             
     integer SIZE                                                                                                                 
     integer, dimension (:), allocatable :: INDEX1, INDEX2, IPIX

     
  end type EVENT                                                                                                                  


  type :: CCD                                                                                                                     

     ! this are ccd related constrats                                                                                             
     ! RAWX0, RAWY0 -> CCD CENTER                                                                                                 
     ! RAW_X_EXT, RAW_Y_EXT -> SIZE OF CCD IN PIXELS                                                                              
     ! RAW_XPIX-> X PIXEL SIZE IN MICRONS AND DEGREES. SAME FOR Y                                                                 
     integer, dimension (MAX_NUM_TELESCOPES) :: RAW_X0, RAW_Y0                                                                    

  end type CCD                                                                                                                    

  ! all together now...
  type :: EVENTFILE
     ! ATT is an array of all the attitude extensions in the EVTFILE
     ! NUM_ATT is the number of CORATT extenstion in the event file
     ! WHICH_ATT is an array with information on which INSTRUMENTS
     ! are present in the EVENT FILE, i.e. 1-7. In EVTTOOL this 
     ! corresponds to  this%ccdnr_infile(ii) (this of type tEventFile)
     ! which lists the ccds. One has to be careful because ccdnr_infile(ii) 
     ! has inverse order, i.e. ccdnr_infile(1)=7 ....  ccdnr_infile(7)=1 
     integer ::  NUM_ATT
     integer,  dimension (:), allocatable :: WHICH_CCDNR
     TYPE(ATT), dimension (:), allocatable :: ATT


     TYPE(EVENT) :: EVENT

     TYPE(CCD) :: CCD
  end type EVENTFILE
  
!  ! all together now...
!  type :: tEVENTFILE_EXT
!     type(tEventFile) EventFile
!     double precision, dimension (:), allocatable :: XD, YD, ZD
!     TYPE(CCD) :: CCD
!  end type tEVENTFILE_EXT



  !atttiude file related stuff
  ! C1 AND C2 APPEAR TO BE OBSOLETE. SHOULD REMOVE
  type :: SRC     
     ! this is RA, DEC and Extraction radius
     ! RAWX and RAWY for each telscope
     double precision, dimension (:), allocatable :: RA, DEC, RADIUS
     double precision, dimension (:), allocatable :: XD, YD, ZD
     double precision, dimension (:), allocatable :: XIMA, YIMA
     double precision, dimension (:,:), allocatable :: C1, C2 ! Real part Shapelet coefficients
     !complex, dimension (:,:), allocatable :: C1, C2 ! Shapelet Cartesian coefficients
     !complex, dimension (:,:), allocatable :: P1, P2 ! Shapelet Polar Coefficients
     integer  TOTAL ! total number of sources
  end type SRC




  !MODEL PSF related stuff
  type :: MODELPSF     
     ! this is model PSF information 

     !arrays of cartesian hermite coefficients as a function of position (2nd dim) and energy (3rd dim)
     !C1==core of PSF, C2==wings of PSF
     double precision, dimension (:,:,:), allocatable :: C0, C1, C2
     !double precision, dimension (:,:,:), allocatable :: P1, P2 place holder for polar

     ! 2-D array holding images 
     ! dimension 1, 2: X, Y
     ! dimension 2: source 
     ! depricated - dimension 3: energy - 09/25/2015 AGE: energy is not an input parameter any longer
     ! each evt file in the input list should correspond to a different energy interval.
     ! depricate - allocatable array for IMG: 15/10/2015 AGE: this array will 
     ! be simply hold the image of a single PSF at a time. 
     ! 10/06/16 AGE: IMG is the PSF image. It is an allocatebale array
     !               IMG_SHAPE1/2 are images of each shapelet order
     !               which are then simply multiplied by the corresponding 
     !               coefficients and stacked to generate IMG
     double precision, dimension (:,:,:), allocatable :: IMG_SHAPE0
     double precision, dimension (:,:,:), allocatable :: IMG_SHAPE1
     double precision, dimension (:,:,:), allocatable :: IMG_SHAPE2

     !arrays of cartesian hermite nx and ny order. This is to help 
     !understand which coefficient in C1 or C2 corresponds to which nx, ny order
     !the size of these arrays equals the size of first dimension of C1 or C2
     ! PN?, PM?: are the lookup tables of the polar coefficients 
     ! they hold information on N, M order of the polar shapelet coefficits
     integer,  dimension (:),  allocatable :: NX0, NY0, NX1, NY1, NX2, NY2
     integer,  dimension (:),  allocatable :: PN0, PM0, PN1, PM1, PN2, PM2

     ! this is the tranfomration matrices from cartesian<->polar
     ! they are defined ONCE, since it depends only on NX, NY, PN, PM ORDERS
     complex(kind=real64),  dimension (:,:),  allocatable :: MATRIX0, MATRIX_INV0
     complex(kind=real64),  dimension (:,:),  allocatable :: MATRIX1, MATRIX_INV1
     complex(kind=real64),  dimension (:,:),  allocatable :: MATRIX2, MATRIX_INV2
     
     !arrays of X, Y position on CCD. 
     !these have the sizw == 2nd dimension of C1 or C2
     double precision,  dimension (:),  allocatable :: XOFF, YOFF

     !arrays of the energies of the model PSFs
     !these have the size == 3rd dimension of C1 or C2
     double precision,  dimension (:),  allocatable :: ENERGY

     !Shapelet scale for core=BETA1 and wings=BETA2
     double precision BETA0, BETA1, BETA2     

     !Spatial MIN and MAX and STEO of PSF measuremnts
     double precision XMIN, XMAX, YMIN, YMAX, XBIN, YBIN

     ! max order of hermite coefficients for 1: core, 2: wings
     integer  NH0, NH1, NH2 

     ! 10/06/16 AGE: new structure parameter. 
     !   this is the side of the PSF image that will be generated. 
     !   this defines what is the spatial pixel scale in microns
     !   the shapelets are constructed in images of scale RAW_PIX_MICRON
     !   provided above. The reconstructed PSF image can be produced 
     !   in any scale defined by IMG_PIXEL_MM. IMG_PIXEL_MM  should be
     !   defined relative to  RAW_PIX_MICRON, i.e. IMG_PIXEL_MM = RAW_PIX_MICRON/2.0
     double precision IMG_PIXEL_MM   


     ! 10/06/16 AGE: new structure parameter. 
     !   this is the side of the PSF image that will be generated. 
     integer MAX_PSF_IMG_SIZE


  end type MODELPSF



  !OUTPUT STACKED PSF related stuff
  type :: STACKPSF     

     !arrays of cartesian hermite coefficients: 
     ! dimension 1: coefficients
     ! dimension 2: source 
     ! depricated - dimension 3: energy - 09/25/2015 AGE: energy is not an input parameter any longer
     ! each evt file in the input list should correspond to a different energy interval.
     !C1==core of PSF, C2==wings of PSF
     !double precision, dimension (:,:,:), allocatable :: C1, C2
     double precision, dimension (:,:), allocatable :: C0, C1, C2
     

     !arrays of cartesian hermite nx and ny order. This is to help 
     !understand which coefficient in C1 or C2 corresponds to which nx, ny order
     !the size of these arrays equals the size of first dimension of C1 or C2
     integer,  dimension (:),  allocatable ::  NX0, NY0, NX1, NY1, NX2, NY2
     ! added 19/09/2016 (GL)
     integer                               :: NUM_SRC
  end type STACKPSF


  ! CONSTANT PSF related stuff
  type :: CPSF     

     ! 2-D array holding images 
     ! dimension 1, 2: X, Y
     ! dimension 2: source 
     ! depricated - dimension 3: energy - 09/25/2015 AGE: energy is not an input parameter any longer
     ! each evt file in the input list should correspond to a different energy interval.
     ! depricate - allocatable array for IMG: 15/10/2015 AGE: this array will 
     ! be simply hold the image of a single PSF at a time. 
     ! 10/06/16 AGE: IMG is the PSF image. It is an allocatebale array
     !               IMG_SHAPE1/2 are images of each shapelet order
     !               which are then simply multiplied by the corresponding 
     !               coefficients and stacked to generate IMG
     double precision, dimension (:,:), allocatable   :: IMG
     double precision, dimension (:,:,:), allocatable :: IMG_SHAPE0
     double precision, dimension (:,:,:), allocatable :: IMG_SHAPE1
     double precision, dimension (:,:,:), allocatable :: IMG_SHAPE2

     !arrays of cartesian hermite nx and ny order. This is to help 
     !understand which coefficient in C1 or C2 corresponds to which nx, ny order
     !the size of these arrays equals the size of first dimension of C1 or C2
     ! PN?, PM?: are the lookup tables of the polar coefficients 
     ! they hold information on N, M order of the polar shapelet coefficits
     integer,  dimension (:),  allocatable ::  NX0, NY0, NX1, NY1, NX2, NY2
     integer,  dimension (:),  allocatable ::  PN0, PM0, PN1, PM1, PN2, PM2

     !Shapelet scale for core=BETA1 and wings=BETA2
     double precision BETA0, BETA1, BETA2     

     ! max order of hermite coefficients for 1: core, 2: wings
     integer  NH0, NH1, NH2 

     ! 10/06/16 AGE: new structure parameter. 
     !   this is the side of the PSF image that will be generated. 
     !   this defines what is the spatial pixel scale in microns
     !   the shapelets are constructed in images of scale RAW_PIX_MICRON
     !   provided above. The reconstructed PSF image can be produced 
     !   in any scale defined by IMG_PIXEL_MM. IMG_PIXEL_MM  should be
     !   defined relative to  RAW_PIX_MICRON, i.e. IMG_PIXEL_MM = RAW_PIX_MICRON/2.0
     double precision IMG_PIXEL_MM   


     ! 10/06/16 AGE: new structure parameter. 
     !   this is the side of the PSF image that will be generated. 
     integer MAX_PSF_IMG_SIZE
     


     ! this is the tranfomration matrices from cartesian<->polar
     ! they are defined ONCE, since it depends only on NX, NY, PN, PM ORDERS
     complex(kind=real64),  dimension (:,:),  allocatable :: MATRIX0, MATRIX_INV0
     complex(kind=real64),  dimension (:,:),  allocatable :: MATRIX1, MATRIX_INV1
     complex(kind=real64),  dimension (:,:),  allocatable :: MATRIX2, MATRIX_INV2

  end type CPSF

  ! CONSTANT PSF related stuff
  type :: FPSF     

     TYPE(CPSF) :: myCPSF

     TYPE(STACKPSF), dimension (:), allocatable :: mySPSF

     ! **** added  2nd STACKPSF and myModelPSF (GL 23/09/2016):
     TYPE(STACKPSF), dimension (:), allocatable :: evSPSF
     TYPE(MODELPSF), dimension (:) ,allocatable :: myModelPSF

  end type FPSF

end module slib_constants
