! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module slib_engine
    
  private 

  public :: slib_run, slib_run_heal, slib_run_wrap

contains
  
  
  ! this is the subroutine that does all the work
  ! 
  ! The assumoptions is that the input is a list of EVT files
  ! and a SINGLE source list. The different evt files may 
  ! corresponds to different energy bands of the same field
  ! A stackPSF structure is produced for each input evt file. 
  ! It containts shapelet coefficients (i.e. the PSFs) for
  ! each source in the source list. 
  !  
  ! the slib_run_array takes as input tEventFile structures
  ! which have been read via EVTTOOL outside the subroutine
  ! in the program that call the subroutine. 
  !
  !  slib_run takes as input the filenames of the evt files
  !  and open them within the subroutine. 
  !  slib_run  has become obsolte and shoudl be reomved
  !  at some point
  ! 
  ! slib_run does the following steps: 
  ! 
  ! 1. reads model PSF calibration files
  ! 2. reads input source list with src positions (RA, DEC) and
  !    radius within which to find photons and stack their PSFs
  ! 3. Use EROLIB routines input events
  ! 4. The routine slib_evttool_2_myevent converts from EROLIB 
  !    format to format understood by SLIB. It also reads
  !    the ATT extensions and assigne a ROLL ANGLE to each event.
  !    It assumes that the TIME column in the EVT and ATT extensions 
  !    is in ascending order. 
  ! 5. slib_assign assigns photons to sources within a search 
  !    radius provided in the input catalogue. It also 
  !    rotates PSFs and stacks them together.
  ! 6. slib_deallocate_event: deallocates all slib arrays related 
  !    to events
  ! 7. dealloc_eventfile: is the EROLIB deallocations
  ! 8. slib_deallocate_src finally it deallocates all slib
  !    arrays related to SRC


 subroutine slib_run_wrap(EventFile_in, mySRC, myFPSF, &
      MAX_PSF_IMG_SIZE, IMG_PIXEL_MM, HEAL)

   use slib_constants
   use erolib_eventfile_mod

   implicit none
   
   type(tEventFile), intent(in) :: EventFile_in(:)
   double precision, intent(in) :: IMG_PIXEL_MM   
   integer, intent(in) :: MAX_PSF_IMG_SIZE
   TYPE(FPSF), intent(out) :: myFPSF
   TYPE(SRC), intent(in) :: mySRC
   logical HEAL

   if(HEAL)then
      call  slib_run_heal(EventFile_in, mySRC, myFPSF, &
           MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
   else
      call  slib_run(EventFile_in, mySRC, myFPSF, &
           MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
   endif
   
 end subroutine slib_run_wrap


 
 subroutine slib_run(EventFile_in, mySRC, myFPSF, &
      MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
    
    

    use slib_constants
    use slib_data_psf
    use slib_data_evt
    use slib_data_src
    use slib_calc_space

    use erolib_eventfile_mod


    implicit none

    !character(len=*), intent(in) :: srcfile
    type(tEventFile), intent(in) :: EventFile_in(:)
    double precision, intent(in) :: IMG_PIXEL_MM   
    integer, intent(in) :: MAX_PSF_IMG_SIZE
    TYPE(FPSF), intent(out) :: myFPSF
    TYPE(SRC), intent(in) :: mySRC
    !TYPE(SRC) mySRC

    TYPE(CPSF) myCPSF
    TYPE(MODELPSF), dimension (MAX_NUM_TELESCOPES) :: myModelPSF
    TYPE(EVENTFILE) myEVENTFILE
    INTEGER NUM_EVENTS, IEVT
    INTEGER, dimension(:), allocatable :: NUMSPSF

           
    ! GET SIZES OF ARRAYS
    NUM_EVENTS=SIZE(EventFile_in)
    
    call slib_read_alltel_model_psf(myModelPSF, myCPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
    
   ! allocate new FullPSF parameter
   ! the NUMPSF array holds the number of PSFs
   ! to be estimated for each event file. It 
   ! defines the 2nd index of the arrays
   ! myStackPSF%C1(:, :). It can be the number
   ! of sources on an image OR the number of
   ! photons on an event file, in which case
   ! the NUMPSF may be different for each input
   ! event file, hence the definition as array
   allocate(NUMSPSF(NUM_EVENTS))
   do IEVT=1, NUM_EVENTS
      NUMSPSF(IEVT) =  mySRC%TOTAL
   enddo   
   call slib_init_FPSF(myFPSF, myCPSF, myModelPSF(1), NUM_EVENTS, NUMSPSF) 
   DEALLOCATE(NUMSPSF)
   ! allocate new FullPSF parameter



   
   ! loop through EVENT FILES AND PRODUCE PSFS
   do IEVT=1, NUM_EVENTS
      call slib_evttool_2_myevent(myEVENTFILE, EventFile_in(IEVT))
      call  slib_assign(myEVENTFILE, mySRC, myModelPSF, myFPSF%myCPSF, myFPSF%mySPSF(IEVT))      
      call slib_deallocate_event(myEVENTFILE)      
   enddo

   ! **** add myModelPSF to myFPS  (GL, 23/09/2016):
   myFPSF%myModelPSF = myModelPSF

   !call  slib_deallocate_src(mySRC)
   call  slib_dealloc_CPSF(myCPSF)
   call slib_dealloc_model_psf(myModelPSF)
   
   !call  slib_deallocate_src(mySRC)
   !call  slib_dealloc_CPSF(myCPSF)

   return

 end subroutine slib_run
 



subroutine slib_run_heal(EventFile_in, mySRC, myFPSF, &
      MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
    
    

    use slib_constants
    use slib_data_psf
    use slib_data_evt
    use slib_data_src
    use slib_calc_space

    use erolib_eventfile_mod


    implicit none

    !character(len=*), intent(in) :: srcfile
    type(tEventFile), intent(in) :: EventFile_in(:)
    double precision, intent(in) :: IMG_PIXEL_MM   
    integer, intent(in) :: MAX_PSF_IMG_SIZE
    TYPE(FPSF), intent(out) :: myFPSF
    TYPE(SRC), intent(in) :: mySRC
    !TYPE(SRC) mySRC

    TYPE(CPSF) myCPSF
    TYPE(MODELPSF), dimension (MAX_NUM_TELESCOPES) :: myModelPSF
    TYPE(EVENTFILE) myEVENTFILE
    INTEGER NUM_EVENTS, IEVT
    INTEGER, dimension(:), allocatable :: NUMSPSF

    ! GET SIZES OF ARRAYS
    NUM_EVENTS=SIZE(EventFile_in)
    
    call slib_read_alltel_model_psf(myModelPSF, myCPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)

   ! allocate new FullPSF parameter
   ! the NUMPSF array holds the number of PSFs
   ! to be estimated for each event file. It 
   ! defines the 2nd index of the arrays
   ! myStackPSF%C1(:, :). It can be the number
   ! of sources on an image OR the number of
   ! photons on an event file, in which case
   ! the NUMPSF may be different for each input
   ! event file, hence the definition as array
   allocate(NUMSPSF(NUM_EVENTS))
   do IEVT=1, NUM_EVENTS
      NUMSPSF(IEVT) =  mySRC%TOTAL
   enddo   
   call slib_init_FPSF(myFPSF, myCPSF, myModelPSF(1), NUM_EVENTS, NUMSPSF) 
   DEALLOCATE(NUMSPSF)
   ! allocate new FullPSF parameter

   ! loop through EVENT FILES AND PRODUCE PSFS
   do IEVT=1, NUM_EVENTS

      call slib_evttool_2_myevent(myEVENTFILE, EventFile_in(IEVT))
      call  slib_assign_heal(myEVENTFILE, mySRC, myModelPSF, myFPSF%myCPSF, myFPSF%mySPSF(IEVT))      

      call slib_deallocate_event(myEVENTFILE)
      
   enddo

   ! **** add myModelPSF to myFPS  (GL, 23/09/2016):
   myFPSF%myModelPSF = myModelPSF
   
   !call  slib_deallocate_src(mySRC)
   call  slib_dealloc_CPSF(myCPSF)
   call slib_dealloc_model_psf(myModelPSF)
 
   !call  slib_deallocate_src(mySRC)
   !call  slib_dealloc_CPSF(myCPSF)

   return

 end subroutine slib_run_heal
 









end module slib_engine



