#!/bin/tcsh -f

# -----------------------------------------------------------------------------
#  Author: T. Dwelly (dwelly@mpe.mpg.de)
# 
#  First version: 2018-03-23
#  Part of the eSASS software suite (eROSITA) 
# -----------------------------------------------------------------------------
# 
# -----------------------------------------------------------------------------
#   Summary: This is a script that uses SIXTE to build a set of simple
#   eROSITA eventfiles that are used to test SRCTOOL. Note that only a 
#   small subset of the grid of produced files are required in order to run the
#   srctool/test/test_srctool.csh script. 
# -----------------------------------------------------------------------------

set EROSIM = erosim
set ERO_CALEVENTS = ero_calevents

foreach HELPER ( $EROSIM $ERO_CALEVENTS ftlist sed gawk ftcreate ftverify ftcopy ftcalc gzip ftmerge stilts ftappend )
  if ( ! -X $EROSIM ) then 
    echo "Helper tool is not present or not executable: $HELPER" 
    exit 1
  endif
end

alias nrows "ftlist \!^ K include='NAXIS2' mode=q |& gawk 'BEGIN {result=-1} "\$"1~/^NAXIS2/ {result="\$"3;exit} NF > 1 {exit} END {print result}'"

set XML_ORIGINAL_STEM = "erosita_"
set XML_FINAL_STEM = "erosita_bigfov_"
set XML_STEM = ${XML_ORIGINAL_STEM}
#set XML_STEM = ${XML_FINAL_STEM}

#set TIDY_UP = "FALSE"
set TIDY_UP = "TRUE"


# Where to find the EROSITA XML files:
set EROSITA_XML_DIR = /home/tdwelly/software/linux/share/sixte/instruments/srg

# This script runs SIXTE to build a set of eventfiles that can be used
# to test the basic functionality of SRCTOOL. These event files are
# not intended to represent scientifically realistic sources or
# scenes, and are deliberately kept small in size. 



# This is where the final products will be publisted 
# use environment variable $SASS_TESTDATA if it is set
if ( $?SASS_TESTDATA ) then
  set PUBLISH_DIR = $SASS_TESTDATA
else
  set PUBLISH_DIR = /home/erosita/sw/eSASSdevel/sass/testdata
endif



# which sections to do?
set DO_ATTITUDE
set DO_SIMPUT   
set DO_PREP_SIXTE
set DO_SIXTE    
set DO_MAKECAT
set DO_PUBLISH

# uncomment the lines below to activate the section:
set DO_ATTITUDE = "TRUE"
set DO_SIMPUT   = "TRUE"
set DO_PREP_SIXTE = "TRUE"
set DO_SIXTE    = "TRUE"
set DO_MAKECAT   = "TRUE"
set DO_PUBLISH  = "TRUE"

# List of observation modes to test:
# 1) Pointed observation with rigidly fixed attitude
# 2) Pointed observation with slowly drifting attitude
# 3) Segment(s) of normal all-sky survey scans
# 5) Slow scan mapping with changes in scan direction

set OBS_MODE_LIST = ( POINTING DRIFT SURVEY )
#set OBS_MODE_LIST = ( POINTING DRIFT SURVEY SCAN )
#set OBS_MODE_LIST = ( POINTING ) 
#set OBS_MODE_LIST = ( SURVEY ) 
#set OBS_MODE_LIST = ( DRIFT ) 

# List of scenes to test:
# 1) A single compact source, pointlike or slightly extended (i.e. radius < few arcmin)
# 2) A single very extended source (i.e. radius >> few arcmin)
# 3) Multiple pointlike and compact extented sources described by a catalogue 

set SCENE_LIST = ( COMPACT VEXTENDED CATALOGUE )
#set SCENE_LIST = ( CATALOGUE )
#set SCENE_LIST = ( COMPACT )
#set SCENE_LIST = ( VEXTENDED CATALOGUE )
#set SCENE_LIST = ( VEXTENDED  )

# List of extended source models to test (for SCENE=COMPACT and SCENE=VEXTENDED)
# 1) TOPHAT
# 2) GAUSS
# 3) BETA
# 4) MAP

set COMPACT_EXTENT_MODEL_LIST   = ( POINTSRC TOPHAT GAUSS BETA MAP ) 
#set COMPACT_EXTENT_MODEL_LIST   = ( POINTSRC ) 
#set COMPACT_EXTENT_MODEL_LIST   = ( MAP ) 
set VEXTENDED_EXTENT_MODEL_LIST = ( TOPHAT GAUSS BETA MAP ) 
#set VEXTENDED_EXTENT_MODEL_LIST = ( TOPHAT BETA MAP ) 
#set VEXTENDED_EXTENT_MODEL_LIST = ( GAUSS BETA ) 
set CATALOGUE_EXTENT_MODEL_LIST = ( BETA )

set MJDREF = "51543.8750" # days - keep fixed to this value
set TIME0 = 500000.0        # s - not very important

# skycoor -d -n 7  -j 01:23:45.67 +01:23:45.67 J2000
# 20.9402917   1.3960194 J2000

set RA0  = 20.9402917     # deg, J2000
set DEC0 = +1.3960194     # deg, J2000
set ROLL0 = 0.0           # deg, E of N

# Params describing the instrument setup
set INST_LIST = ( 1 2 3 4 5 6 7 ) 
set BACKGROUND = "yes"
set TEXP  =  100.0        # s

set STEM = "test_"

# Params describing the spectrum of the source - exact values are not very important 
set NH22           = 0.03       # 10^22 cm^-2
set GAMMA          = 1.0        # deliberately hard to ensure that we get some high energy counts
set FLUX_EMIN      = 0.5         # keV  - energy range over which flux is defined
set FLUX_EMAX      = 2.0         # keV
#set FLUX0          = 1.0e-11     # erg cm^-2 s^-1

set CATALOGUE_FLUX0  = 1.0e-11
set COMPACT_FLUX0    = 5.0e-11
set VEXTENDED_FLUX0  = 5.0e-10

set CATALOGUE_SOURCE_RADIUS_DEG = 0.02 # deg = 1.2arcmin
set COMPACT_SOURCE_RADIUS_DEG   = 0.05 # deg = 3arcmin
set VEXTENDED_SOURCE_RADIUS_DEG = 0.25 # deg = 15arcmin


set CATALOGUE_NSRC = 10
set CATALOGUE_DRA  = 1.0 # deg
set CATALOGUE_DDEC = 1.0
set CATALOGUE_EXTENT_MODULUS = 3
set CATALOGUE_OFFAXIS_MAX_DEG = 2.0
set CATALOGUE_MIN_SRC_SEP_ARCSEC = `\wcalc -q "3*${CATALOGUE_SOURCE_RADIUS_DEG}*3600.0"`





# make a simput file with one or more sources
if ( "$DO_ATTITUDE" == "TRUE" ) then
  #######################################################################
  # Create the attitude files
  foreach OBS_MODE ( $OBS_MODE_LIST )
    set ATT_FILE = ${STEM}${OBS_MODE}_attitude.fits
    echo "Building Attitude file: $ATT_FILE"

    #######################################################################
    if ( "$OBS_MODE" == "POINTING" ) then
      set ATT_TSTEP = 10.0 
      # Make a static attitude file 
      gawk -v ra0=$RA0 -v dec0=$DEC0 -v roll0=$ROLL0 -v time0=$TIME0 -v dt=$ATT_TSTEP -v texp=$TEXP 'BEGIN{nsteps=int(texp/dt)+1;ra=ra0;dec=dec0;roll=roll0;for(i=0;i<nsteps;i++){t=time0+i*dt; printf("%12.3d %12.7f %12.7f %12.3f\n", t, ra, dec, roll)}}' > ${ATT_FILE:r}.txt
    #######################################################################
    else if ( "$OBS_MODE" == "DRIFT" ) then
  
      # random walk the pointing and slowly twist the roll angle
      set ATT_TSTEP = 1.0    # s
      set DATT_DT   = 0.0001   # deg/s 
      set DROLL_DT  = 0.001    # deg/s
  
      gawk -v ra0=$RA0 -v dec0=$DEC0 -v roll0=$ROLL0 -v time0=$TIME0 -v dt=$ATT_TSTEP -v texp=$TEXP -v datt_dt=$DATT_DT -v droll_dt=$DROLL_DT 'BEGIN{deg2rad=atan2(1.,1.)/45.0;nsteps=int(texp/dt)+1;ra=ra0;dec=dec0;roll=roll0;for(i=0;i<nsteps;i++){t=time0+i*dt;dr=dt*datt_dt*rand();a=360.*rand(); ra+=dr*sin(a*deg2rad)/cos(dec*deg2rad);dec+=dr*cos(a*deg2rad);roll+=(rand()-0.5)*dt*droll_dt;roll=roll%360.0; printf("%12.3d %12.7f %12.7f %12.3f\n", t, ra, dec, roll)}}' > ${ATT_FILE:r}.txt
  
    #######################################################################
    else if ( "$OBS_MODE" == "SURVEY" ) then
      #normal eROSITA All sky scan speed = 0.025 deg/s
      set ATT_TSTEP = 1.0    # s
      set DRA_DT    =  0.010   # deg/s
      set DDEC_DT   = -0.020   # deg/s  
      set DROLL_DT  = 0.000     # deg/s
  
      # Scan the pointing once over the scene (in RA and Dec) 
      gawk -v ra0=$RA0 -v dec0=$DEC0 -v roll0=$ROLL0 -v time0=$TIME0 -v dt=$ATT_TSTEP -v dra_dt=$DRA_DT -v ddec_dt=$DDEC_DT -v droll_dt=$DROLL_DT -v texp=$TEXP 'BEGIN{nsteps=int(texp/dt)+1;ra=ra0-(dra_dt*texp*0.5)/cos(dec0*deg2rad);dec=dec0-(ddec_dt*texp*0.5);roll=roll0;t=time0;for(i=0;i<nsteps;i++){printf("%12.3d %12.7f %12.7f %12.3f\n", t, ra, dec, roll);t+=dt; ra+=dra_dt*dt/cos(dec*deg2rad);dec+=ddec_dt*dt; roll+=dt*droll_dt;roll=roll%360.0; }}' > ${ATT_FILE:r}.txt
  
    #######################################################################
    else if ( "$OBS_MODE" == "SCAN" ) then
       echo "Warning - SCAN mode is not implemented here yet"
       continue
    #######################################################################
    else
       echo "Unknown OBS_MODE=$OBS_MODE"
       exit 1
    endif
    #######################################################################
  
    set TSTOP = `tail -n 1 ${ATT_FILE:r}.txt | awk '{print $1}'`
  
    echo "TIME 1D s\
RA 1D deg\
DEC 1D deg\
ROLL 1D deg" > cdfile

    echo "EXTNAME = 'ATTITUDE' \
MJDREF = $MJDREF\
TSTART = $TIME0\
TSTOP  = $TSTOP" > headfile
    ftcreate cdfile  ${ATT_FILE:r}.txt ${ATT_FILE}  headfile=headfile clobber=yes mode=q
    rm cdfile headfile  ${ATT_FILE:r}.txt
  end
  #######################################################################
endif
#######################################################################
#######################################################################


#######################################################################
#######################################################################
# make a simput file with one or more sources
if ( "$DO_SIMPUT" == "TRUE" ) then


  #######################################################################
  echo "Making spectrum extension: ${STEM}spectrum.xcm"
    echo "model phabs*powerlaw\
$NH22\
$GAMMA\
1.0" > ${STEM}spectrum.xcm
  #######################################################################

  #######################################################################
  # Make a map extension for each extended source type
  foreach SCENE ( $SCENE_LIST ) 

    set MAKE_MAP = "TRUE"
    if ("$SCENE" == "COMPACT" ) then
      set THIS_EXTENT_MODEL_LIST = ( $COMPACT_EXTENT_MODEL_LIST ) 
      set SOURCE_RADIUS_DEG = "$COMPACT_SOURCE_RADIUS_DEG"
      set EXTMAP_PIXELSCALE = 2.5e-4  # deg/pix
      set EXTMAP_NAXIS12 = 512
      set FLUX0 = "$COMPACT_FLUX0"
    else if ("$SCENE" == "VEXTENDED" ) then
      set THIS_EXTENT_MODEL_LIST = ( $VEXTENDED_EXTENT_MODEL_LIST ) 
      set SOURCE_RADIUS_DEG = "$VEXTENDED_SOURCE_RADIUS_DEG" 
      set EXTMAP_PIXELSCALE = 2.5e-3  # deg/pix
      set EXTMAP_NAXIS12 = 512
      set FLUX0 = "$VEXTENDED_FLUX0"
    else if ("$SCENE" == "CATALOGUE" ) then
      set SOURCE_RADIUS_DEG = "$CATALOGUE_SOURCE_RADIUS_DEG"
      set EXTMAP_PIXELSCALE = 2.5e-4  # deg/pix
      set EXTMAP_NAXIS12 = 256
      set THIS_EXTENT_MODEL_LIST = ( $CATALOGUE_EXTENT_MODEL_LIST ) 
      set FLUX0 = "$CATALOGUE_FLUX0"
    endif

    foreach EXTENT_MODEL ( $THIS_EXTENT_MODEL_LIST )
      set OUTPUT_STEM = ${STEM}${SCENE}_${EXTENT_MODEL}_
      set SOURCE_MAP = ${OUTPUT_STEM}extentmap.fits

      set SIMLOG = ${OUTPUT_STEM}simput.log
      if ( -e $SIMLOG ) rm $SIMLOG 
      echo "Working on ${STEM} ${SCENE} ${EXTENT_MODEL}" |& tee -a $SIMLOG

      set SIMPUT_IMPARS = ""
      if ( "${EXTENT_MODEL}" != "POINTSRC" ) then
        set SIMPUT_IMPARS = "ImageFile=${SOURCE_MAP}"

        set NAXIS1 = $EXTMAP_NAXIS12
        set NAXIS2 = $EXTMAP_NAXIS12
        set CRPIX1 = `\wcalc -q "0.5*($NAXIS1 + 1.0)"`
        set CRPIX2 = `\wcalc -q "0.5*($NAXIS2 + 1.0)"`
        set CDELT1 = "-${EXTMAP_PIXELSCALE}"
        set CDELT2 = "${EXTMAP_PIXELSCALE}"

        set A = 0.0
        set COSA = `\wcalc -q -P10 "cos($A)"`
        set SINA = `\wcalc -q -P10 "sin($A)"`
        set CRVAL1 = "$RA0"
        set CRVAL2 = "$DEC0"
    #    set CRVAL1 = `\wcalc -q -P10 "$RA0 + 0.02/cos($DEC0)"`
    #    set CRVAL2 = `\wcalc -q -P10 "$DEC0 + 0.03"`

        echo "CRPIX1 = $CRPIX1\
CRPIX2 = $CRPIX2 \
PC1_1 =  ${COSA} \
PC1_2 = -${SINA} \
PC2_1 =  ${SINA} \
PC2_2 =  ${COSA} \
CDELT1 = $CDELT1\
CDELT2 = $CDELT2\
CTYPE1 = 'RA---TAN'\
CTYPE2 = 'DEC--TAN'\
CRVAL1 = $CRVAL1\
CRVAL2 = $CRVAL2\
CUNIT1 = 'deg'\
CUNIT2 = 'deg'" > ${SOURCE_MAP:r}.head

        ftimgcreate bitpix=-32 naxes="$NAXIS1,$NAXIS2" datafile="none" outfile=${SOURCE_MAP} headfile=${SOURCE_MAP:r}.head mode=q clobber=yes history=no
        rm ${SOURCE_MAP:r}.head

        if ( "${EXTENT_MODEL}" == "TOPHAT" ) then
          # tophat source, with flat emission
          ftimgcalc outfile=${SOURCE_MAP} expr="(ANGSEP(a.RA,a.DEC,$RA0,$DEC0)<=${SOURCE_RADIUS_DEG}? 1.0 : a)" a=${SOURCE_MAP} clobber=yes mode=q history=no
        else if ( "${EXTENT_MODEL}" == "BETA" ) then
          set CUT_RADIUS_NSIGMAS = 1.0
          ftimgcalc outfile=${SOURCE_MAP} expr="(ANGSEP(a.RA,a.DEC,$RA0,$DEC0)<=${SOURCE_RADIUS_DEG}*${CUT_RADIUS_NSIGMAS} ? (1.0 + ANGSEP(a.RA,a.DEC,$RA0,$DEC0)/${SOURCE_RADIUS_DEG})**-1.5 : 0.0*a)" a=${SOURCE_MAP} clobber=yes mode=q history=no
        else if ( "${EXTENT_MODEL}" == "GAUSS" ) then
          set CUT_RADIUS_NSIGMAS = 1.0
          ftimgcalc outfile=${SOURCE_MAP} expr="(ANGSEP(a.RA,a.DEC,$RA0,$DEC0)<=${SOURCE_RADIUS_DEG}*${CUT_RADIUS_NSIGMAS} ? exp(-0.5*(ANGSEP(a.RA,a.DEC,$RA0,$DEC0)/${SOURCE_RADIUS_DEG})**2) : 0.0*a)" a=${SOURCE_MAP} clobber=yes mode=q history=no
        else if ( "${EXTENT_MODEL}" == "MAP" ) then
      #    # a circular source, with separate flat emission levels in each quadrant 
      #    ftimgcalc outfile=${SOURCE_MAP} expr="(ANGSEP(a.RA,a.DEC,$RA0,$DEC0)<=${SOURCE_RADIUS_DEG}?1.0:a)" a=${SOURCE_MAP} clobber=yes mode=q history=no
          # a rectagular source, with separate flat emission levels in each quadrant 
          ftimgcalc outfile=${SOURCE_MAP} expr="((abs(a.RA-${RA0})<${SOURCE_RADIUS_DEG}/cos(${DEC0}*#deg))&&abs(a.DEC-${DEC0})<${SOURCE_RADIUS_DEG}? 1.0 : a)" a=${SOURCE_MAP} clobber=yes mode=q history=no
      
          # this bit sets the quadrants
          ftimgcalc outfile=${SOURCE_MAP} expr="((a.RA<${RA0}&&a.DEC<${DEC0})?a*1.0:((a.RA>=${RA0}&&a.DEC<${DEC0})?a*2.0:((a.RA>=${RA0}&&a.DEC>=${DEC0})?a*3.0:a*4.0)))" a=${SOURCE_MAP} clobber=yes mode=q history=no
      
        else 
          echo "Unknown extent model: $EXTENT_MODEL"
          exit 1
        endif
      endif

      #######################################################################
      # now build the corresponding SIMPUT file 
      set SIMPUT_FILE = ${OUTPUT_STEM}simput.fits
      echo "Building SIMPUT_FILE: $SIMPUT_FILE"
      simputfile srcFlux=$FLUX0 Emin=$FLUX_EMIN Emax=$FLUX_EMAX Elow=0.05 Eup=12.5 Estep=0.005 MJDREF=$MJDREF RA=$RA0 Dec=$DEC0 XSPECFile="${STEM}spectrum.xcm" $SIMPUT_IMPARS Simput="$SIMPUT_FILE" Src_Name="TestSource" clobber=yes >&  $SIMLOG
      #######################################################################



      # For testing of CATALOGUE mode we need several sources (some pointlike, some extended) into the SIMPUT file and 
      # we need to make a fake CATPREP catalogue
      if ( "$SCENE" == "CATALOGUE" ) then
        echo "Modifying SIMPUT_FILE: $SIMPUT_FILE"
  
        # take a copy of the original 
        cp ${SIMPUT_FILE} temp0.fits
        ftcopy ${SIMPUT_FILE}+1 temp1.fits clobber=yes copyall=no history=no
        set TEMP_LIST = "${SIMPUT_FILE}+1"
        @ NSRC2 = $CATALOGUE_NSRC + $CATALOGUE_NSRC
        set I = 1
        while ( $I < $NSRC2 ) 
          set TEMP_LIST = "$TEMP_LIST temp1.fits+1"
           @ I ++
        end
        ftmerge "$TEMP_LIST" temp2.fits copyall=yes clobber=yes mode=q
        ftcalc temp2.fits+1 temp2.fits "SRC_ID"   "#row" clobber=yes mode=q
        ftcalc temp2.fits+1 temp2.fits "RA"       "$RA0+(random()-0.5)*$CATALOGUE_DRA/(cos(#deg*$DEC0))"   clobber=yes mode=q
        sleep 1 # ensures a new random seed for the dec column
        ftcalc temp2.fits+1 temp2.fits "DEC"      "$DEC0+(random()-0.5)*$CATALOGUE_DDEC" clobber=yes mode=q
  
        # always place source #1 at the centre
        ftcalc temp2.fits+1 temp2.fits "RA"  "(#row==1?${RA0}:RA)"    clobber=yes mode=q
        ftcalc temp2.fits+1 temp2.fits "DEC" "(#row==1?${DEC0}:DEC)"  clobber=yes mode=q
  
        # set one in every ${CATALOGUE_EXTENT_MODULUS} sources to be extended
        ftcalc temp2.fits+1 temp2.fits "IMAGE" "(#row % ${CATALOGUE_EXTENT_MODULUS} ==0?'[IMAGE,1]':'NULL')"  clobber=yes mode=q
  
        # filter out very close pairs
  #      echo "Number of sources before filtering: `nrows temp2.fits+1`" |& tee -a $SIMLOG
  #      ftlist "temp2.fits[1][col RA,DEC]" T rownum=no colheader=no mode=q |& tee -a $SIMLOG
        stilts tmatch1 in=temp2.fits matcher=sky params=${CATALOGUE_MIN_SRC_SEP_ARCSEC} values="RA DEC" action=keep1 ofmt=fits out=temp1.fits >>& $SIMLOG
        if ( `tail -n 10 $SIMLOG | grep -c 'Error: No matches were found'` >= 1 ) then
          cp temp2.fits temp1.fits
        else
          ftappend temp2.fits+2  temp1.fits mode=q history=no
        endif
        echo "Number of sources after proximity filtering: `nrows temp1.fits+1`" |& tee -a $SIMLOG
  #      ftlist "temp2.fits[1][col RA,DEC]" T rownum=no colheader=no mode=q |& tee -a $SIMLOG
        # curtail catalogue to desired length
        ftcopy "temp1.fits[1][#row<=$CATALOGUE_NSRC]" ${SIMPUT_FILE} copyall=yes mode=q clobber=yes
        ftappend "temp0.fits[3]" ${SIMPUT_FILE} mode=q history=no
        rm temp0.fits temp1.fits temp2.fits  
        echo "Number of sources in final catalogue: `nrows ${SIMPUT_FILE}+1`" |& tee -a $SIMLOG
  
      endif

    end

  #######################################################################



  end

  #######################################################################

  
endif
#######################################################################
#######################################################################

if ( "$DO_PREP_SIXTE" == "TRUE" ) then
  echo "Preparing SIXTE XML files: ${EROSITA_XML_DIR}/${XML_FINAL_STEM}?.xml"
  foreach INST ( 1 2 3 4 5 6 7 )
    set INFILE  = ${EROSITA_XML_DIR}/${XML_ORIGINAL_STEM}${INST}.xml  
    set OUTFILE = ${EROSITA_XML_DIR}/${XML_FINAL_STEM}${INST}.xml
    sed 's/<fov diameter="1.02"/<fov diameter="1.25"/g' $INFILE > ${OUTFILE}
  end
endif
 
## We do not need to tweak this now as SIXTE version 2.4.3/ calibration version instruments_srg-1.4.5.tar.gz
## includes most of the effects below
## #######################################################################
## #######################################################################
## if ( "$DO_PREP_SIXTE" == "TRUE" ) then
##   # wcalc "2.685739664E-3*384"   - from instmap files
##   set DIAM = "1.0313"
##   set ROTA = ( 270 210 90 90 330 330 210 ) 
## 
##   foreach INST ( $INST_LIST )
##     set XML = ${EROSITA_XML_DIR}/${XML_STEM}${INST}.xml
##     if ( -e $XML ) mv ${XML} ${XML}~
## 
##     #set ARF = "/home/erosita/caldb/data/erosita/tm${INST}/bcf/tm${INST}_arf_filter_000101v02.fits"
##     set ARF = "tm${INST}_arf_filter_000101v02.fits"
## 
##     if ( $INST == 2 || $INST == 5 ) then
##       set DESC1 = "only filter on filter wheel"
##     else
##       set DESC1 = "onchip filter, plus filter on filter wheel"
##     endif
## 
##     echo "Writing SIXTE XML file for TM${INST} -> $XML"
## 
##     echo '<?xml version="1.0"?>\
## <\!-- SRG eROSITA version 1.4 - modified by TD to match CALDB -->\
## <instrument telescop="eROSITA" instrume="TM'${INST}'">\
## \
## <telescope>\
##   <\!-- Telescope Module '${INST}': '${DESC1}' -->\
##   <arf filename="'${ARF}'"/>\
##   <fov diameter="'${DIAM}'"/>\
##   <focallength value="1.6"/>\
##   <vignetting filename="erosita_vignetting_v2.1.fits"/>\
##   <psf filename="erosita_psf_v3.1.fits"/>\
## </telescope>\
## \
## <detector>\
##   <dimensions xwidth="384" ywidth="384"/>\
##   <pixelborder x="0." y="0."/>\
##   <wcs xrpix="192.5" yrpix="192.5" xrval="0." yrval="0." \
##        xdelt="75.e-6" ydelt="75.e-6" rota="'${ROTA[$INST]}'"/>\
##   <rmf filename="sixte_erormf_normalized_singles_20170725.rmf"/>\
##   <cte value="1"/>\
##   <split type="exponential" par1="0.355"/>\
##   <phabackground filename="sixte_ero_particle_bkg_20180307.pha"/>\
##   <threshold_readout_lo_keV value="0."/>\
##   <threshold_event_lo_keV value="0.0"/>\
##   <threshold_split_lo_fraction value="0.01"/>\
##   <threshold_pattern_up_keV value="10."/>\
## \
##   <readout mode="time">\
##     <wait time="0.05"/>\
## \
##     <loop start="0" end="383" increment="1" variable="$i">\
##       <readoutline lineindex="0" readoutindex="$i"/>\
##       <lineshift/>\
##       <wait time="0.3e-6"/>\
##     </loop>\
## \
##     <newframe/>\
##   </readout>\
## </detector>\
## \
## </instrument>' > $XML
## 
##   end
## 
## endif
## #######################################################################
## #######################################################################


#######################################################################
#######################################################################
if ( "$DO_SIXTE" == "TRUE" ) then
  set EROSIM_CHATTER = 3
#  set EROSIM_CHATTER = 5

  set XML_PAR_LIST = ""
    foreach INST ( $INST_LIST )
    set XML_PAR_LIST = "${XML_PAR_LIST} XMLFile${INST}=${EROSITA_XML_DIR}/${XML_STEM}${INST}.xml"
  end 

  foreach OBS_MODE ( $OBS_MODE_LIST ) 
    set ATT_FILE = ${STEM}${OBS_MODE}_attitude.fits

    foreach SCENE ( $SCENE_LIST ) 
 
      if ("$SCENE" == "COMPACT" ) then
        set THIS_EXTENT_MODEL_LIST = ( $COMPACT_EXTENT_MODEL_LIST ) 
      else if ("$SCENE" == "VEXTENDED" ) then
        set THIS_EXTENT_MODEL_LIST = ( $VEXTENDED_EXTENT_MODEL_LIST ) 
      else if ("$SCENE" == "CATALOGUE" ) then
        set THIS_EXTENT_MODEL_LIST = ( $CATALOGUE_EXTENT_MODEL_LIST ) 
      else
        exit 1
      endif
  
      foreach EXTENT_MODEL ( $THIS_EXTENT_MODEL_LIST )
        set INPUT_STEM = ${STEM}${SCENE}_${EXTENT_MODEL}_
        set OUTPUT_STEM = ${STEM}${OBS_MODE}_${SCENE}_${EXTENT_MODEL}_
        set SIMPUT_FILE = ${INPUT_STEM}simput.fits

        rm ${OUTPUT_STEM}ccd*_evtlist1.fits ${OUTPUT_STEM}ccd*_evtlist2.fits >& /dev/null
        set SIMLOG = ${OUTPUT_STEM}sixte.log
        if ( -e $SIMLOG ) rm $SIMLOG 
        echo "\n###################################################################"
        echo "##########################start#####################################"
	echo "Working on OBS_MODE=${OBS_MODE} SCENE=${SCENE} EXTENT_MODEL=${EXTENT_MODEL}"
        echo "SIXTE log is here: $SIMLOG "

        ###############################################
        # run the SIXTE simulation
        echo "Running $EROSIM to make ${OUTPUT_STEM}*evtlist1.fits" | tee -a $SIMLOG
        set echo 
        $EROSIM Prefix="${OUTPUT_STEM}" EvtFile="evtlist1.fits" RA=$RA0 Dec=$DEC0 Simput="${SIMPUT_FILE}" MJDREF=$MJDREF TSTART=$TIME0 Exposure=$TEXP clobber=yes $XML_PAR_LIST Background=$BACKGROUND Attitude="${ATT_FILE}" RawData=none chatter=$EROSIM_CHATTER SkipInvalids=no >>& $SIMLOG
        unset echo
    

        # check the erosim outputs
        foreach INST ( $INST_LIST ) 
          if ( ! -es ${OUTPUT_STEM}ccd${INST}_evtlist1.fits ) then
            echo "Error, can't find: ${OUTPUT_STEM}ccd${INST}_evtlist1.fits " | tee -a $SIMLOG
            exit 1
          endif

          if ( `ftverify ${OUTPUT_STEM}ccd${INST}_evtlist1.fits errreport=e prstat=no mode=q | grep -c 'verification FAILED'` > 0 ) then
            echo "Error, ${OUTPUT_STEM}ccd${INST}_evtlist1.fits is corrupt" | tee -a $SIMLOG
            echo "Here is the ftverify report:" | tee -a $SIMLOG
            ftverify ${OUTPUT_STEM}ccd${INST}_evtlist1.fits  mode=q | tee -a $SIMLOG
            exit 1
          endif
        end

        # calibrate the events using ero_calevents
        echo "Running $ERO_CALEVENTS to make ${OUTPUT_STEM}*evtlist2.fits" |& tee -a $SIMLOG
        set echo
        foreach INST ( $INST_LIST ) 
          $ERO_CALEVENTS EvtFile=${OUTPUT_STEM}ccd${INST}_evtlist1.fits eroEvtFile=${OUTPUT_STEM}ccd${INST}_evtlist2.fits CCDNr=${INST} Attitude=${ATT_FILE} RA=$RA0 Dec=$DEC0 clobber=yes RefRA=$RA0 RefDec=$DEC0 >>& $SIMLOG
        end
        unset echo
        ###############################################


        # check the ero_calevents outputs
        foreach INST ( $INST_LIST ) 
          if ( ! -es ${OUTPUT_STEM}ccd${INST}_evtlist2.fits ) then
            echo "Error, can't find: ${OUTPUT_STEM}ccd${INST}_evtlist2.fits " | tee -a $SIMLOG
            exit 1
          endif

          if ( `ftverify ${OUTPUT_STEM}ccd${INST}_evtlist2.fits errreport=e prstat=no mode=q | grep -c 'verification FAILED'` > 0 ) then
            echo "Error, ${OUTPUT_STEM}ccd${INST}_evtlist2.fits is corrupt" | tee -a $SIMLOG
            echo "Here is the ftverify report:" | tee -a $SIMLOG
            ftverify ${OUTPUT_STEM}ccd${INST}_evtlist2.fits  mode=q | tee -a $SIMLOG
            exit 1
          endif
        end


 
#        set NVALID_LOG = ${OUTPUT_STEM}nvalid.txt
#        gethead -h NVALID NINVALID NPVALID NPINVALI ${OUTPUT_STEM}ccd?_evtlist1.fits | gawk 'BEGIN {t[2]="Valid fraction";t[3]="Invalid fraction";t[4]="Valid piled up fraction";t[5]="Invalid piled up fraction";} //{printf("%-50s %12s %12s %12s %12s\n", $1,$2,$3,$4,$5)} /evtlist1/{for(i=2;i<=5;i++){n[i]+=$i};} END {printf("%-50s %12d %12d %12d %12d\n\n", "Total", n[2],n[3],n[4],n[5]); ntotal=n[2]+n[3];for(i=2;i<=5;i++){printf("%-26s = %8.5f\n", t[i], n[i]/ntotal)}}' >& $NVALID_LOG 

        # set OUTPUT_STEM_LIST = ( `ls test_*_evtlist3.fits | gawk '//{split($1,a,"evtlist3");print a[1]}'` )
        # set SIMLOG = nvalid.log
        # foreach OUTPUT_STEM ( $OUTPUT_STEM_LIST ) 
        set NVALID_LOG = ${OUTPUT_STEM}nvalid.txt
        if ( -e ${NVALID_LOG} ) mv ${NVALID_LOG} ${NVALID_LOG}~

        echo "Valid events are counted here: $NVALID_LOG " | tee -a $SIMLOG
        
        gawk 'BEGIN {printf("#%-49s %12s %12s %12s %12s ", "FILENAME","NVALID", "NINVALID","NPVALID","NPINVALI");for(i=0;i<=12;i++){s=sprintf("NGRAD%d",i);printf("%8s ",s)};printf("\n");exit}' >> $NVALID_LOG 
        foreach FILE ( ${OUTPUT_STEM}ccd?_evtlist1.fits ) 
	   ftlist "${FILE}[1]" K include="NVALID,NINVALID,NPVALID,NPINVALI,NGRAD*" mode=q | gawk -v fname="$FILE" '//{split(substr($0,10),a,"/");n[substr($0,1,8)]=int(a[1]);} END {printf("%-50s %12d %12d %12d %12d ", fname, n["NVALID  "], n["NINVALID"], n["NPVALID"], n["NPINVALI"]);for(i=0;i<=12;i++){s=sprintf("NGRAD%-3d",i);printf("%8d ",n[s])}; printf("\n") }' >> $NVALID_LOG 
        end

        gawk '$1!~/^#/{n["NVALID  "]+=$2; n["NINVALID"]+=$3; n["NPVALID"]+=$4; n["NPINVALI"]+=$5;} END {n["TOT"]=n["NVALID  "]+n["NINVALID"]; printf("\n\n%-26s = %8.5f\n%-26s = %8.5f\n%-26s = %8.5f\n%-26s = %8.5f\n", "Valid fraction", n["NVALID  "]/n["TOT"],"Invalid fraction", n["NINVALID"]/n["TOT"], "Valid piled up fraction", n["NPVALID"]/n["TOT"], "Invalid piled up fraction", n["NPINVALI"]/n["TOT"] )} ' $NVALID_LOG  >> $NVALID_LOG 


        printf '\n#%-59s %8s %8s %8s %8s %8s\n' 'FILENAME' 'NTOT' 'Nsingle' 'Ndouble' 'Ntriple' 'Nquad'  >> $NVALID_LOG 
        foreach FILE ( ${OUTPUT_STEM}ccd?_evtlist2.fits ) 
          ftlist "${FILE}[1][col PAT_TYP]" T rownum=no colheader=no mode=q | gawk -v fname=$FILE '//{ntot++; n[$1]++;} END {printf("%-60s %8d %8d %8d %8d %8d\n", fname, ntot ,n[1],n[2],n[3],n[4])}'  >> $NVALID_LOG
        end

        # end 

        ###############################################
        # we have to run this in a separate shell or it will pollute the environment
        # Use evtool to merge events over all instruments
        set EVENTFILES1 = ( ${OUTPUT_STEM}ccd*_evtlist1.fits ) 
        set EVENTFILES2 = ( ${OUTPUT_STEM}ccd*_evtlist2.fits ) 
        set EVENTFILE3  = ${OUTPUT_STEM}evtlist3.fits 
        set EVENTFILE4  = ${OUTPUT_STEM}eventfile.fits 
        set SHELL_SCRIPT = throwaway_shell.csh
        if ( -e $SHELL_SCRIPT ) rm $SHELL_SCRIPT
        echo "#\!/bin/tcsh -f " > $SHELL_SCRIPT
        echo "source /home/erosita/sw/sass-setup.csh eSASSdevel " >> $SHELL_SCRIPT
        echo "evtool eventfiles="'"'"$EVENTFILES2"'"'" outfile=${EVENTFILE3} clobber=yes events=yes image=no pattern=15"  >> $SHELL_SCRIPT
        echo "Running evtool to make ${EVENTFILE3}" | tee -a $SIMLOG
#        cat $SHELL_SCRIPT
        chmod u+x $SHELL_SCRIPT
        ./$SHELL_SCRIPT >>& $SIMLOG
      
        # tweak the KEYWORDS
        fparkey 15 "${EVENTFILE3}[EVENTS]" PAT_SEL add=yes mode=q
        set NHDU = `ftlist ${EVENTFILE3} H mode=q | gawk '$1~/^HDU$/{nhdu=$2}END{print nhdu}'`
        set I = 1
        while ( $I < $NHDU ) 
          if ( "$OBS_MODE" == "POINTING" || "$OBS_MODE" == "DRIFT" ) then 
            fparkey "POINTING" ${EVENTFILE3}+$I OBS_MODE add=yes mode=q
          endif
          # no longer needed in SIXTE v2.4.5
          # fparkey "FILTER" ${EVENTFILE3}+$I FILTER add=yes mode=q
          @ I ++
        end
        ###############################################

        # filter out the events which are not within the circular DETMAP
        # 192^2 = 36864 =
        echo "Running ftcopy to make ${EVENTFILE4}" | tee -a $SIMLOG
        ftcopy "${EVENTFILE3}[1][(RAWX-192.5)**2+(RAWY-192.5)**2 <= 36864.0]" ${EVENTFILE4} clobber=yes mode=q copyall=yes history=yes	

        # tidy up
        if ( "$TIDY_UP" == "TRUE" ) then 
          rm ${EVENTFILES1} ${EVENTFILES2} ${EVENTFILE3} ${SHELL_SCRIPT}
        endif

        foreach FILE ( ${OUTPUT_STEM}evtlist3.fits ${OUTPUT_STEM}eventfile.fits ) 
          ftlist "${FILE}[1][col PAT_TYP]" T rownum=no colheader=no mode=q | gawk -v fname=$FILE '//{ntot++; n[$1]++;} END {printf("%-60s %8d %8d %8d %8d %8d\n", fname, ntot ,n[1],n[2],n[3],n[4])}'  >> $NVALID_LOG
        end

        echo "###########################done####################################"
        echo "###################################################################\n"

      end
    end
  end

endif
#######################################################################
#######################################################################



#######################################################################
#######################################################################
if ( "$DO_MAKECAT" == "TRUE" ) then

  foreach OBS_MODE ( $OBS_MODE_LIST ) 

    foreach SCENE ( $SCENE_LIST )
      if ( "$SCENE" == "CATALOGUE" ) then 
        set EXTENT_MODEL = "BETA"

        set SIMPUT_FILE = ${STEM}${SCENE}_${EXTENT_MODEL}_simput.fits
        set OUTPUT_STEM = ${STEM}${OBS_MODE}_${SCENE}_${EXTENT_MODEL}_
    
        # make a mock CATPREP catalogue (fake SCTS/BG_MAP columns)
        set CATPREP_FILE =  ${OUTPUT_STEM}catprep.fits
        echo "Building CATPREP_FILE: $CATPREP_FILE"
      
    #    ftcopy "${SIMPUT_FILE}[1][col ID_SRC=SRC_ID,RA=RA,DEC=DEC,SCTS(1E)=1e5*($CATALOGUE_FLUX0/3e-12),BG_MAP(1E)=7.5*($TEXP/1e4),EXT=(IMAGE=='IMAGE'?${CATALOGUE_SOURCE_RADIUS_DEG}*3600.0:0.0),X_IMA(1E)=(RA-$RA0)*cos(#deg*DEC)*-3600.0,Y_IMA(1E)=(DEC-$DEC0)*3600.0,AUTO_EXTRACT(1I)=1,AUTO_EXCLUDE(1I)=1,OFFAXIS_DEG(1E)=angsep($RA0,$DEC0,RA,DEC),LCPROD_PCN(3A)='',SPECPROD_PCN(3A)='']"  ${ERMLDET_FILE} copyall=no clobber=yes mode=q
    #    ftcalc "${ERMLDET_FILE}+1"  ${ERMLDET_FILE} AUTO_EXTRACT "(OFFAXIS_DEG<=${CATALOGUE_OFFAXIS_MAX_DEG}?1:0)" clobber=yes mode=q
    #    fparkey -1.0E-03 ${ERMLDET_FILE}+1 TCDLT7  comm="X increment at reference pixel" add=yes mode=q
    #    fparkey  1.0E-03 ${ERMLDET_FILE}+1 TCDLT8  comm="Y increment at reference pixel" add=yes mode=q
    
        # new format that matches http://erosita.mpe.mpg.de/eROdoc/products/SourceCatalogs_doc.html
        ftcopy "${SIMPUT_FILE}[1][col ID_SRC=SRC_ID,RA=RA,DEC=DEC,ML_CTS_0(1E)=${TEXP}*($CATALOGUE_FLUX0/3e-12),ML_BKG_0(1E)=1.0*($TEXP/1e3),EXT=(IMAGE=='IMAGE'?${CATALOGUE_SOURCE_RADIUS_DEG}*3600.0:0.0),AUTO_EXTRACT(1I)=1,AUTO_EXCLUDE(1I)=1,OFFAXIS_DEG(1E)=angsep($RA0,$DEC0,RA,DEC),LCPROD_PCN(3A)='',SPECPROD_PCN(3A)='']"  ${CATPREP_FILE} copyall=no clobber=yes mode=q
        ftcalc "${CATPREP_FILE}+1"  ${CATPREP_FILE} AUTO_EXTRACT "(OFFAXIS_DEG<=${CATALOGUE_OFFAXIS_MAX_DEG}?1:0)" clobber=yes mode=q
      endif
    end
  end
endif
#######################################################################
#######################################################################

  

if (  "$DO_PUBLISH" == "TRUE" ) then

  if ( ! -d "$PUBLISH_DIR" || ! -w  "$PUBLISH_DIR" ) then
    echo "Could not publish test files to $PUBLISH_DIR"
    exit 1
  endif 

  echo "Publishing test files to $PUBLISH_DIR"

  foreach INFILE ( test_SURVEY_COMPACT_POINTSRC_eventfile.fits \
test_POINTING_COMPACT_TOPHAT_eventfile.fits \
test_SURVEY_CATALOGUE_BETA_eventfile.fits \
test_SURVEY_CATALOGUE_BETA_catprep.fits \
test_POINTING_VEXTENDED_MAP_eventfile.fits \
test_VEXTENDED_MAP_extentmap.fits )

    if ( -e $INFILE ) then
      gzip -c $INFILE > ${PUBLISH_DIR}/${INFILE}.gz
      chown ${USER}:erosita ${PUBLISH_DIR}/${INFILE}.gz
      chmod ug+rw ${PUBLISH_DIR}/${INFILE}.gz
      echo "  Published ${PUBLISH_DIR}/${INFILE}.gz"
    else
      echo " Error, could not find $INFILE"
    endif 
  end

endif










exit










#
