! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_autoreg

  !-----------------------------------------------------------------------------
  ! Summary:
  ! Automatic region generation from source list
  !-----------------------------------------------------------------------------

  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic
  use srctool_generic_fitswrap
  use srctool_datatypes
  use srctool_misc
  use srctool_param
  use srctool_psf

  implicit none     ! insist on explicit typing

  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_AUTOREG_MOD'

  ! subroutine declarations and interfaces for module: srctool_autoreg
  public  :: srctool_autoreg_generate
  private :: srctool_autoreg_interp_psf
  private :: srctool_autoreg_setup_params
  private :: srctool_autoreg_setup_psfs
  private :: srctool_autoreg_src_inclusion
  private :: srctool_autoreg_build_excl_image
  private :: srctool_autoreg_get_incl_area
  private :: srctool_autoreg_calc_back
  private :: srctool_autoreg_remove_confusing_srcs
  private :: srctool_autoreg_back_remove_confusing_srcs

  private

  !--------------------------------------------------------------------
  ! the following are constants that control the behaviour where      !
  ! arbitrary/semi-arbitrary decisions must be made:                  !
  !-------------------------------------------------------------------!
  type param_type
     real(kind=real64) :: psf_energy_keV

     ! step sizes for iterations in radius:
     real(kind=real64) :: radial_stepsize_src
     real(kind=real64) :: radial_stepsize_back

     ! starting (minimum) radius for sources:
     real(kind=real64) :: minimum_source_radius

     ! No exclusion zones can overlap this close to a source:
     real(kind=real64) :: minimum_source_core

     ! no exclusion zones are allowed to be centred within this radius
     ! from the source :
     real(kind=real64) :: minimum_exclude_dist

     ! The minimum size of an exclusion region, must be smaller than
     ! minimum_exclude_dist:
     real(kind=real64) :: minimum_exclude_radius

     ! the minimum ratio of source radius to inner background annulus
     ! radius:
     real(kind=real64) :: initial_ratio_src_r_to_back_r1

     ! the maximum factor by which a source's surface brightness can
     ! contaminate its own background region:
     real(kind=real64) :: max_src_map_to_bg_map_ratio

     ! the size of the background annulus as a multiple of the area of
     ! the source region:
     real(kind=real64) :: back_to_src_area_ratio

     ! The maximum permitted ratio of surface brightness from a
     ! confusing source (within the source extraction region):
     real(kind=real64) :: max_conf_map_to_src_map_ratio

     ! The maximum permitted ratio of surface brightness from a
     ! confusing source (within the background extraction region):
     real(kind=real64) :: max_conf_map_to_back_map_ratio

     ! Upper limit to the inner background annulus wrt the 99%
     ! containment radius of the source:
     real(kind=real64) :: max_ratio_back_r1_to_radius_99pc

     ! Upper limit to the width of a background region
     real(kind=real64) :: max_back_annulus_width
  end type param_type
  !-------------------------------------------------------------------!

  ! For pixelization of images to calculate areas (degrees)
  real(kind=real64), parameter :: pixsize = 4.d0 / 3600.d0
  ! number of pixels per degree (inverse degrees)
  real(kind=real64), parameter :: invpixsize = 1.d0 / pixsize

contains

  !---------------------------------------------------------------------------
  ! Do PSF interpolation
  ! Sets ierr if there's an error
  subroutine srctool_autoreg_interp_psf(src, r, rpsf_out, area_out, ierr)
    character(len=*),    parameter   :: subroutinename='srctool_autoreg_interp_psf'
    type(src_type),      intent(in)  :: src
    real(kind=real64),   intent(in)  :: r
    real(kind=real64),   intent(out) :: rpsf_out, area_out
    integer(kind=int32), intent(inout) :: ierr

    integer(kind=int32) :: status

    status = 0
    call search_and_linear_interp(src%prpsf%pradius(:), src%prpsf%prpsf(:), &
         src%prpsf%parea(:), r, rpsf_out, area_out, status)

    if ( (status < 0) .or. is_zero(area_out) ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr(subroutinename, ierr, "Bad rpsf/area interpolation")
    end if
  end subroutine srctool_autoreg_interp_psf

  !---------------------------------------------------------------------------
  ! Read or set default parameters for automatic regions
  !---------------------------------------------------------------------------
  subroutine srctool_autoreg_setup_params(par)
    type(param_type), intent(out) :: par

    ! Set default parameters (optionally taken from environment)
    par%psf_energy_keV = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_PSF_ENERGY_KEV", 1.0d0)
    par%radial_stepsize_src = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_RADIAL_STEPSIZE_SRC", 1.0d0) * ARCSEC2DEG
    par%radial_stepsize_back = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_RADIAL_STEPSIZE_BACK", 2.0d0) * ARCSEC2DEG
    par%minimum_source_radius = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_MINIMUM_SOURCE_RADIUS", 10.0d0) * ARCSEC2DEG
    par%minimum_source_core = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_MINIMUM_SOURCE_CORE", 5.0d0) * ARCSEC2DEG
    par%minimum_exclude_dist = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_MINIMUM_EXCLUDE_DIST", 10.0d0) * ARCSEC2DEG
    par%minimum_exclude_radius = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_MINIMUM_EXCLUDE_RADIUS", 5.0d0) * ARCSEC2DEG
    par%initial_ratio_src_r_to_back_r1 = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_INITIAL_SRC_R_TO_BACK_R1", 2.0d0)
    par%max_src_map_to_bg_map_ratio = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_MAX_SRC_MAP_TO_BG_MAP_RATIO", 0.05d0)
    par%back_to_src_area_ratio = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_BACK_TO_SRC_AREA_RATIO", 150.0d0)
    par%max_conf_map_to_src_map_ratio = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_MAX_CONF_MAP_TO_SRC_MAP_RATIO", 0.2d0)
    par%max_conf_map_to_back_map_ratio = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_MAX_CONF_MAP_TO_BACK_MAP_RATIO", 0.1d0)
    par%max_ratio_back_r1_to_radius_99pc = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_MAX_RATIO_BACK_R1_TO_RADIUS_99PC", 3.0d0)
    par%max_back_annulus_width = &
         srctool_param_env_real64("SRCTOOL_AUTOREG_MAX_BACK_ANNULUS_WIDTH", 120.d0) * ARCSEC2DEG

  end subroutine srctool_autoreg_setup_params

  !---------------------------------------------------------------------------
  ! For each source, setup PSF
  !---------------------------------------------------------------------------
  subroutine srctool_autoreg_setup_psfs(instlist, srclist, par, ierr)
    character(len=*), parameter                :: subroutinename='srctool_autoreg_setup_psfs'
    type(instlist_type), intent(in), target    :: instlist
    type(srclist_type),  intent(in), target    :: srclist
    type(param_type),    intent(in)            :: par
    integer(kind=int32), intent(inout)         :: ierr

    character(len=STR_MAX)  :: str_buffer
    integer(kind=int32)     :: i_src, i_inst
    integer(kind=int32)     :: first_valid_inst
    integer(kind=int32)     :: status, counter
    integer(kind=int32)     :: psf_eband_index
    logical                 :: assume_survey_psf
    real(kind=real64)       ::  offaxis_mean

    type(psfmap_type), pointer  :: ppsfmap0
    type(rpsf_type), pointer    :: prpsf0
    type(src_type), pointer     :: psrc
    type(psfmap_type)           :: psfmap_temp

    !---------------------------------------------------------------------------
    ! Find the first valid survey psfmap/rpsf
    ! TODO - use the psf summed over all instruments
    ! TODO - use the psf merged over energy bands
    ! TODO - check that we are in survey mode, if not then use PSF based on position in focal plane

    ! if we find any survey/scan/slew obs then we will use the survey PSF

    if ( (instlist%num_survey_mode >= 1) .or. &
         (instlist%num_scan_mode   >= 1) .or. &
         (instlist%num_slew_mode   >= 1) ) then
       assume_survey_psf = .true.
    else if ( instlist%num_pointing_mode >= 1) then
       assume_survey_psf = .false.
    else
       ierr = ERROR_LEVEL_SEVERE
       call wrserr(subroutinename,ierr,"Unknown OBS_MODE configuration")
       return
    end if

    nullify(prpsf0, ppsfmap0)
    if ( assume_survey_psf ) then
       if ( instlist%inst(0)%survey_psf%is_valid ) then
          associate( psfmaplist => instlist%inst(0)%survey_psf%psfmaplist )
            ! search for the correct energy band index - aim to get close to psf_energy_keV
            call binary_search(psfmaplist%pe, par%psf_energy_keV, psf_eband_index, status)
            if ( (status == 1) .or. (status == 2) ) then
               ! do nothing, we have to take the extremal value
            else if ( status == 0 ) then
               ! choose which of the two possible PSF maps to use
               if ( abs(psfmaplist%pe(psf_eband_index)   - par%psf_energy_keV) > &
                    abs(psfmaplist%pe(psf_eband_index+1) - par%psf_energy_keV) ) then
                  psf_eband_index = psf_eband_index + 1
               end if
            else
               ierr = ERROR_LEVEL_SEVERE
               call wrserr(subroutinename, ierr, &
                    "Problem finding suitable PSF model at requested energy")
               return
            end if
          end associate

          ppsfmap0 => instlist%inst(0)%survey_psf%psfmaplist%psfmap(psf_eband_index)
          prpsf0 => ppsfmap0%rpsf
          call wrschatty(subroutinename, &
               "Assuming survey-averaged PSF when generating automatic regions")
       else
          ierr = ERROR_LEVEL_SEVERE
          call wrserr(subroutinename, ierr, &
               "Problem finding suitable RPSF description")
          return
       end if
    else
       ! choose best psf model for each source later
       call wrschatty(subroutinename, &
            "Assuming localised PSF model when generating automatic regions")
       ! we will do this in a moment, when processing each source

       ! TODO we should really use an RPSF model that is averaged over all valid instruments
       do first_valid_inst=1,MAX_INST
          if ( instlist%inst(first_valid_inst)%to_consider ) exit
       end do

    end if

    !---------------------------------------------------------------------------
    ! prepare RPSFs for all interesting sources
    do i_src = 1, srclist%num_srcs
       psrc => srclist%psrc(i_src)

       if ( .not. psrc%cat_extract .and. .not. psrc%cat_exclude ) then
          cycle
       end if

       ! choose a suitable PSF model based on location of source in focal plane
       if ( .not. assume_survey_psf ) then
          ! calculate the mean off-axis angle averaged over telescopes

          ! we need to know the average position of the source in the focal plane
          ! only the mean off-axis angle is interesting since the azimuth will be
          ! different for each instrument

          ! TODO - implement this properly - the following assumes that mean_ra,mean_dec
          ! of attitude is same as mean position of telescope axis
          offaxis_mean = 0.0d0
          counter = 0
          do i_inst=1,MAX_INST
             if ( instlist%inst(i_inst)%to_consider ) then
                offaxis_mean = offaxis_mean + calc_ang_sep(psrc%coord%long,psrc%coord%lat, &
                     instlist%inst(i_inst)%attlist%mean_ra, instlist%inst(i_inst)%attlist%mean_dec)
                counter = counter + 1
             end if
          end do
          offaxis_mean = offaxis_mean /  MAX(1.0d0,REAL(counter,kind=real64))

          ! Now choose the most appropriate psf model for this location
          call srctool_psf_get_psfmap_pointer( instlist%inst(first_valid_inst)%psf%psfmaplist, &
               offaxis_mean * DEG2ARCMIN, 0.0d0, par%psf_energy_keV, ppsfmap0)
          prpsf0 => ppsfmap0%rpsf

          write(str_buffer, '(F8.2)')  offaxis_mean*DEG2ARCMIN
          ! call wrschatty (subroutinename, "  DEBUG_CHOOSE_PSFMAP SRC " // trim(psrc%name) &
          !     // " OffAxisMean=" // trim(str_buffer) // &
          !     " arcmin => so I chose this PSF: " // trim(ppsfmap0%hduinfo%key_extname%value))
       end if

       ! TODO convolve the rPSF with the extent model
       if ( psrc%extent%code == EXTTYPE_CODE_POINT ) then
          psrc%prpsf => prpsf0  ! use a pointer to the default RPSF for a point source
       else
          call wrschatty(subroutinename, "Preparing RPSF for extended source: " //  trim(psrc%name))
          allocate(psrc%prpsf, STAT=status)
          if ( check_allocate_status(subroutinename, status, ierr) ) return

          call srctool_psf_empty_rpsf(psrc%prpsf)
          call srctool_psf_empty_psfmap(psfmap_temp)

          call srctool_psf_convolve_psfmap_with_extent(instlist,ppsfmap0,psfmap_temp,psrc%extent, ierr)
          if ( ierr /= ERROR_LEVEL_OK ) then
             call wrserr(subroutinename, ierr, "Problem convolving PSF with extent model")
             return
          end if

          call srctool_psf_calc_rpsf(psfmap_temp, psrc%prpsf, ierr)
          if ( ierr /= ERROR_LEVEL_OK ) then
             call wrserr(subroutinename, ierr, &
                  "Problem calculating RPSF from convolution of PSF and extent model")
             return
          end if

          call srctool_psf_empty_psfmap(psfmap_temp)
       end if

       ! write (str_buffer, '(A,A,A,F7.2,A,F7.2,A,F7.2,A)') "  DEBUG RPSF:", &
       !     trim(ppsfmap0%hduinfo%key_extname%value), &
       !     " HEW= ",           psrc%prpsf%radius_50pc_deg*2.0d0*DEG2ARCSEC, &
       !     " arcsec; 90%EW= ", psrc%prpsf%radius_90pc_deg*2.0d0*DEG2ARCSEC, &
       !     " arcsec; 99%EW= ", psrc%prpsf%radius_99pc_deg*2.0d0*DEG2ARCSEC, " arcsec"
       ! call wrschatty (subroutinename, str_buffer )

    end do

  end subroutine srctool_autoreg_setup_psfs


  !---------------------------------------------------------
  !  Compute size of source extraction radius
  !---------------------------------------------------------
  !
  !  * Inclusion policy for source regions:
  !  Include region around each source which maximises the nominal SNR
  !  given the local background surface brightness and the surface
  !  brightness profile of the PSF wings.
  !
  !  * Restrict ourselves to some minimum and maximum radii.
  !  * Use REEF (Radial EEF) PSF to determine optimal radius
  !    - for extended sources first convolve the source model and the REEF
  !
  !  * Algorithm:
  !  1) read columns from ermldetect/catprep catalogue: RA, DEC, SCTS, BG_MAP
  !  2) seach for radius which maximises the SNR:
  !     while (SNR_i+1 > SNR_i) r_i+i = r_i + dr
  !  where SNR_i = SCTS_eff / sqrt(BCTS_eff + SCTS_eff)
  !  where SCTS_eff = STCT*REEF(r_i) and BCTS_eff = BG_MAP*pi*r_i**2
  !  where BG_MAP is assumed to be in units of: cts s^-1 deg^-2
  !    - (this may not be true in ermldetect/catprep catalogues, so conversion needed)
  !  SCTS is background subtracted source counts, corrected for aperture
  !  losses, but including effects of ignetting.
  !  REEF(r_i) is fraction of energy enclosed in radius r_i
  !    - r_i is radius in deg
  !
  !---------------------------------------------------------

  !---------------------------------------------------------------------------
  subroutine srctool_autoreg_src_inclusion(src, par, src_r, ierr)
    character(len=*), parameter  :: subroutinename='srctool_autoreg_src_inclusion'

    type(src_type),      intent(inout) :: src
    type(param_type),    intent(in)    :: par
    real(kind=real64),   intent(out)   :: src_r   ! output source radius
    integer(kind=int32), intent(inout) :: ierr

    integer(kind=int32) :: status
    real(kind=real64)   :: dr, next_r
    real(kind=real64)   :: bcts_eff, scts_eff
    real(kind=real64)   :: snr, next_snr, reef_interp
    type(regcomp_type)  :: regcomp

    status = 0
    dr = par%radial_stepsize_src ! we will increment radius by 1arcsec per step

    ! start with some small minimum radius (note that this will be immediately incremented)
    src_r = par%minimum_source_radius - dr
    snr = 0.0d0
    do
       next_r = src_r + dr  ! increment radius

       call search_and_linear_interp(src%prpsf%pradius(:), src%prpsf%preef(:), next_r, &
            reef_interp, status)
       if ( status < 0 ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr(subroutinename, ierr, "Problem calculating source region radius")
          return
       end if

       scts_eff = src%cat_scts   * reef_interp
       bcts_eff = src%cat_bg_map * PI * next_r**2
       next_snr = scts_eff / sqrt(scts_eff + bcts_eff)

       ! !DEBUG
       ! write (str_buffer, '(A,A,A,F8.2,3(A,F10.5))'), &
       !      "DEBUG source: ", trim(src%name), &
       !      " next_r=", next_r*DEG2ARCSEC, &
       !      " scts_eff=", scts_eff, &
       !      " bcts_eff=", bcts_eff, &
       !      " next_snr=", next_snr
       ! call wrschatty (subroutinename, str_buffer )

       ! test if incrementing the radius has decreased the SNR
       if ( next_snr < snr ) then
          exit ! break out of the do loop
       else
          src_r = next_r
          snr = next_snr
       end if
       ! stop incrementing after we include 99pc of the energy
       if ( src_r >= src%prpsf%radius_99pc_deg ) then
          exit ! break out of the do loop
       end if

    end do

    ! !DEBUG
    ! write (str_buffer, '(A,A,A,F8.2,A,F12.3,A)'), &
    !      "DEBUG source: ", trim(src%name), " src_r=", src_r*DEG2ARCSEC, &
    !      "arcsec src_map_perim=", src_map_perim, " cts/deg2"
    ! call wrschatty (subroutinename, str_buffer )

    ! Make initial region
    ! A single circular region component
    call srctool_regions_init_regcomp(regcomp)
    regcomp%shape_code                       = REGCOMP_SHAPE_CODE_CIRCLE
    regcomp%coordsys_code                    = src%coord%coordsys_code
    regcomp%num_pars                         = REGCOMP_NUM_PARS_CIRCLE
    regcomp%pars(REGCOMP_PARS_CIRCLE_LONG)   = src%coord%long
    regcomp%pars(REGCOMP_PARS_CIRCLE_LAT)    = src%coord%lat
    regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS) = src_r
    regcomp%weight                           = 1
    regcomp%max_radius                       = src_r
    src%src_region%num_regcomps = 0
    call srctool_regions_add_regcomp(src%src_region, regcomp)

    call srctool_regions_region_calc_bounding_circle(src%src_region, src%coord, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
       call wrserr(subroutinename, ierr, "Problem computing auto regions (SRC)")
       return
    end if

  end subroutine srctool_autoreg_src_inclusion

  !---------------------------------------------------------------------------
  ! Build an "exclusion image", which is a bitmap with pixels set if
  ! in excluded region components
  !---------------------------------------------------------------------------
  ! image should be an allocatable 2D logical image, but deallocated
  ! Note: if there are no excluded regions, then image is not allocated
  !---------------------------------------------------------------------------
  subroutine srctool_autoreg_build_excl_image(src, region, image, ierr)
    character(len=*), parameter  :: subroutinename='srctool_autoreg_build_excl_image'

    type(src_type),       intent(in)  :: src         ! input source
    type(region_type),    intent(in)  :: region      ! input region
    logical, allocatable, intent(out) :: image(:,:)  ! output image (is allocated if used)
    integer(kind=int32),  intent(inout) :: ierr      ! set if error returned

    integer(kind=int32) :: i_comp
    integer(kind=int32) :: hsize
    real(kind=real64)   :: max_rad_deg, dist
    real(kind=real64)   :: rx, ry, rad_pix, rad_pix2
    integer(kind=int32) :: x1, x2, y1, y2, x, y

    ! iterate over regions to work out maximum radius
    max_rad_deg = -1d0
    do i_comp = 1, region%num_regcomps
       associate( comp => region%pregcomp(i_comp) )
         ! only consider excluded components
         if(comp%weight /= -1) cycle

         ! everything should be a circle
         if(comp%shape_code /= REGCOMP_SHAPE_CODE_CIRCLE) then
            ierr = ERROR_LEVEL_SEVERE
            call wrserr(subroutinename, ierr, "Only circular region components supported")
            return
         end if

         ! distance between source and centre of region
         dist = calc_ang_sep(src%coord%long, src%coord%lat, &
              comp%pars(REGCOMP_PARS_CIRCLE_LONG), comp%pars(REGCOMP_PARS_CIRCLE_LAT))

         max_rad_deg = max(max_rad_deg, dist+comp%pars(REGCOMP_PARS_CIRCLE_RADIUS))
       end associate
    end do

    ! no background region components
    if( max_rad_deg <= 0d0 ) then
       ! do not allocate image
       return
    end if

    ! allocate output with source at centre
    hsize = int(max_rad_deg*invpixsize) + 1
    allocate( image(hsize*2+1, hsize*2+1) )

    ! .false. are all the pixels without regions
    image = .false.

    ! now draw excluded components of region
    do i_comp = 1, region%num_regcomps
       associate( comp => region%pregcomp(i_comp) )
         ! only consider excluded regions
         if(comp%weight /= -1) cycle

         ! get coordinate of centre of circle
         call longlat2xy(dble(hsize+1), dble(hsize+1), &
              invpixsize, invpixsize, &
              src%coord%long, src%coord%lat, &
              0.d0, comp%pars(REGCOMP_PARS_CIRCLE_LONG), comp%pars(REGCOMP_PARS_CIRCLE_LAT), &
              rx, ry)
         rad_pix = comp%pars(REGCOMP_PARS_CIRCLE_RADIUS) * invpixsize

         ! these are box edges of circle
         x1 = max(1, int(rx - rad_pix)-1)
         x2 = min(hsize*2+1, int(rx + rad_pix)+1)
         y1 = max(1, int(ry - rad_pix)-1)
         y2 = min(hsize*2+1, int(ry + rad_pix)+1)

         ! draw circle in output as .true.
         rad_pix2 = rad_pix**2
         do y = y1, y2
            do x = x1, x2
               if( (dble(x)-rx)**2 + (dble(y)-ry)**2 < rad_pix2 ) then
                  image(x, y) = .true.
               end if
            end do
         end do

       end associate
    end do

  end subroutine srctool_autoreg_build_excl_image


  !---------------------------------------------------------------------------
  ! Return included region area for an annulus between radii of r1->r2
  ! Excludes pixels excluded in input image
  !---------------------------------------------------------------------------
  ! Note: if image is not allocated, the geometric area is returned
  !---------------------------------------------------------------------------
  real(kind=real64) function srctool_autoreg_get_incl_area(image, r1, r2)
    logical, allocatable, intent(in)  :: image(:,:)  ! image containing excluded pixels
    real(kind=real64),    intent(in)  :: r1, r2      ! radii in degrees for annulus (r1=0 for circle)

    real(kind=real64)   :: r1_2, r2_2, r_2
    integer(kind=int32) :: hsize, dim, x, y, it1, it2
    integer(kind=int32) :: totpix, inclpix

    if( .not. allocated(image) ) then
       ! just assume area geometric area
       srctool_autoreg_get_incl_area = PI * (r2**2-r1**2)
       return
    end if

    dim = size(image,1)
    hsize = (dim-1) / 2

    ! maximum x,y iteration ranges to encompass r2, rounding up
    it1 = hsize+1 - int(r2*invpixsize) - 1
    it2 = hsize+1 + int(r2*invpixsize) + 1

    ! precalculate radii squared for comparison
    r1_2 = (r1*invpixsize)**2
    r2_2 = (r2*invpixsize)**2
    totpix = 0
    inclpix = 0
    do y = it1, it2
       do x = it1, it2
          ! radius of pixel, squared
          r_2 = dble((x-hsize-1)**2 + (y-hsize-1)**2)
          if( (r_2>=r1_2) .and. (r_2<=r2_2) ) then
             ! area in annulus
             totpix = totpix + 1

             if( x>=1 .and. x<=dim .and. y>=1 .and. y<=dim ) then
                if( .not. image(x, y) ) then
                   ! area that is not masked out in image
                   inclpix = inclpix + 1
                end if
             else
                ! not in image, so not masked
                inclpix = inclpix + 1
             end if
          end if
       end do
    end do

    srctool_autoreg_get_incl_area = inclpix * pixsize**2
  end function srctool_autoreg_get_incl_area


  !---------------------------------------------------------
  !  Inclusion policy for background regions:
  !  ------------------------------------------------------------------
  !  Annulus with radii back_r1, back_r2, scaled such that area is
  !  at least NN times that within nominal source extraction
  !  radius.  Inner radius set such that the surface brightness of
  !  the target source's PSF wings is less than XX% of the local
  !  background surface brightness

  !  Now also consider confusing excluding sources from a previous
  !  iteration.  Use map of confusing sources to account for area
  !  removed to increase outer radius.

  !  --------------------------------------------------------- *
  !  compute size of background extraction annulus inner and outer radiii

  !---------------------------------------------------------------------------
  subroutine srctool_autoreg_calc_back(src, par, src_r, back_r2, ierr)
    character(len=*), parameter  :: subroutinename='srctool_autoreg_calc_back_radius'

    type(src_type),      intent(inout) :: src
    type(param_type),    intent(in)    :: par
    real(kind=real64),   intent(in)    :: src_r            ! source radius
    real(kind=real64),   intent(out)   :: back_r2          ! output background radii
    integer(kind=int32), intent(inout) :: ierr

    real(kind=real64)    :: back_r1, dr, next_r, src_map
    real(kind=real64)    :: rpsf_interp, area_interp

    logical, allocatable :: exclimage(:,:)
    real(kind=real64)    :: srcarea, targetbackarea, backarea

    type(regcomp_type)   :: regcomp

    dr = par%radial_stepsize_back ! we will increment radius by this amount per step

    ! start with some multiple of the source radius (note that this will be immediately incremented)
    back_r1 = (src_r*par%initial_ratio_src_r_to_back_r1) - dr

    do
       next_r = back_r1 + dr  ! increment radius

       call srctool_autoreg_interp_psf(src, next_r, rpsf_interp, area_interp, ierr)
       if ( ierr /= ERROR_LEVEL_OK ) return
       src_map  = src%cat_scts  * rpsf_interp / area_interp

       ! test if the src_map has dropped sufficiently below the bg map
       if ( src_map < src%cat_bg_map*par%max_src_map_to_bg_map_ratio ) then
          exit ! break out of the do loop
       else
          back_r1 = next_r ! continue
       end if
       ! stop incrementing after some very large radius
       if ( back_r1 >= src%prpsf%radius_99pc_deg * par%max_ratio_back_r1_to_radius_99pc ) then
          exit ! break out of the do loop
       end if
    end do

    ! now compute the outer radius such that the total area in the back region is
    ! a factor of back_to_src_area_ratio larger than src area
    ! area_back = pi * ( back_r2**2 - back_r1**2)
    ! area_back = par%back_to_src_area_ratio * area_src
    ! area_src  = pi * src_r**2
    ! back_r2 = sqrt(area_back/pi + back_r1**2)
    !         = sqrt(area_src*par%back_to_src_area_ratio/pi + back_r1**2)
    !         = sqrt(src_r**2*par%back_to_src_area_ratio    + back_r1**2)
    back_r2 =  sqrt(src_r**2*par%back_to_src_area_ratio   + back_r1**2)

    !print *, PI, PI*src_r**2, (back_r2**2-back_r1**2)/src_r**2

    srcarea = PI * src_r**2
    targetbackarea = par%back_to_src_area_ratio * srcarea

    ! make image with exclusion regions (if generated in previous iteration)
    call srctool_autoreg_build_excl_image(src, src%back_region, exclimage, ierr)
    if( ierr /= ERROR_LEVEL_OK ) return

    ! repeat making outer part of region larger until we reach the
    ! target area or we reach the maximum background width
    do
       backarea = srctool_autoreg_get_incl_area(exclimage, back_r1, back_r2)

       !print *, "radius=",back_r2, " area=", backarea, " target=", targetbackarea

       if( backarea >= targetbackarea*0.99d0 ) then
          ! if we're very close to target area, then exit
          exit
       end if

       ! if width is exceeded and the background area is at least the background
       if( backarea >= srcarea .and. back_r2 - back_r1 >  par%max_back_annulus_width ) then
          exit
       end if

       back_r2 = back_r2 + pixsize
    end do
    if( allocated(exclimage) ) deallocate(exclimage)

    ! A single annular region component
    call srctool_regions_init_regcomp(regcomp)
    regcomp%shape_code                         = REGCOMP_SHAPE_CODE_ANNULUS
    regcomp%coordsys_code                      = src%coord%coordsys_code
    regcomp%num_pars                           = REGCOMP_NUM_PARS_ANNULUS
    regcomp%pars(REGCOMP_PARS_ANNULUS_LONG)    = src%coord%long
    regcomp%pars(REGCOMP_PARS_ANNULUS_LAT)     = src%coord%lat
    regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS1) = back_r1
    regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS2) = back_r2
    regcomp%weight                             = 1
    regcomp%max_radius                         = back_r2
    src%back_region%num_regcomps = 0
    call srctool_regions_add_regcomp(src%back_region, regcomp)

    call srctool_regions_region_calc_bounding_circle(src%back_region, src%coord, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
       call wrserr(subroutinename, ierr, "Problem computing auto regions (BACK)")
       return
    end if

  end subroutine srctool_autoreg_calc_back

  !-------------------------------------------------------------------
  ! Now look for potentially confusing neighbouring sources
  !-------------------------------------------------------------------
  ! Exclusion policy for source regions:
  ! * Exclude region around each confusing source within which the surface
  ! brightness of the PSF wings is greater than XX% of the
  ! target source surface brightness.
  ! * Clip to some minimum and maximum radii.
  ! * Use RPSF of survey PSF to compute radius.
  ! * measure brightness along line joining the two sources.
  !-------------------------------------------------------------------

  !---------------------------------------------------------------------------
  subroutine srctool_autoreg_remove_confusing_srcs(srclist, par, i_src, src_r, ierr)
    character(len=*), parameter  :: subroutinename='srctool_autoreg_remove_confusing_srcs'
    type(srclist_type),  intent(inout), target :: srclist
    type(param_type),    intent(in)            :: par
    integer(kind=int32), intent(in)            :: i_src
    real(kind=real64),   intent(in)            :: src_r
    integer(kind=int32), intent(inout)         :: ierr

    integer(kind=int32)          :: i_src2
    type(src_type),     pointer  :: psrc, psrc2
    real(kind=real64)            :: rpsf_interp, area_interp
    real(kind=real64)            :: d, dr, r1, exclude_r
    real(kind=real64)            :: conf_map, src_map, src_map_perim

    type(regcomp_type)           :: regcomp

    psrc => srclist%psrc(i_src)

    ! calc source map at perimeter of source region
    call srctool_autoreg_interp_psf(psrc, src_r, rpsf_interp, area_interp, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) return
    src_map_perim  = psrc%cat_scts * rpsf_interp/area_interp

    dr = par%radial_stepsize_back
    do i_src2 = 1, srclist%num_srcs
       if ( i_src2 == i_src ) cycle    ! avoid self-exclusion
       psrc2 => srclist%psrc(i_src2)

       if ( .not. psrc2%cat_exclude ) cycle

       ! compute the radius which should be excluded
       ! TODO base this on rpsf, scts, bg_map etc
       !            exclude_r = 30.0d0 * ARCSEC2DEG
       ! calc separation between two sources:
       d = calc_ang_sep(psrc%coord%long, psrc%coord%lat, psrc2%coord%long, psrc2%coord%lat)

       ! test if confusing source is closer than some minimum distance
       if ( d > src_r + psrc2%prpsf%radius_99pc_deg ) then
          cycle ! we shouldn't consider this source further
       else if ( d > src_r ) then
          ! For confusing sources outside source region, test if they
          ! are close+bright enough to consider further.
          ! Compute the brightness of confusing source at perimeter of source region:
          exclude_r = d - src_r

          call srctool_autoreg_interp_psf(psrc2, exclude_r, rpsf_interp, area_interp, ierr)
          if ( ierr /= ERROR_LEVEL_OK ) return
          conf_map = psrc2%cat_scts * rpsf_interp / area_interp

          if ( conf_map < src_map_perim*par%max_conf_map_to_src_map_ratio ) then
             cycle  ! we can safely skip this confusing source
          end if
       else if ( d > par%minimum_exclude_dist ) then
          exclude_r = par%minimum_exclude_radius
       else
          cycle ! we shouldn't consider this source further
          ! since it is too close to the main source position to be excluded
       end if

       ! if we got to here then we should exclude this source from source region
       ! increment the exclusion region until we match the surface brightness
       ! constraints or hit the limits
       do
          r1 = d - exclude_r ! distance from source to perimeter of excluded region
          if ( r1 <= par%minimum_source_core ) then
             exclude_r = d - par%minimum_source_core
             exit ! stop incrementing
          end if

          call srctool_autoreg_interp_psf(psrc, r1, rpsf_interp, area_interp, ierr)
          if ( ierr /= ERROR_LEVEL_OK ) return
          src_map = psrc%cat_scts * rpsf_interp / area_interp

          call srctool_autoreg_interp_psf(psrc2, exclude_r, rpsf_interp, area_interp, ierr)
          if ( ierr /= ERROR_LEVEL_OK ) return
          conf_map = psrc2%cat_scts * rpsf_interp / area_interp

          if ( conf_map <= src_map*par%max_conf_map_to_src_map_ratio ) then
             exit ! we are done,  stop incrementing
          else
             exclude_r = exclude_r + dr  ! increment the radius
          end if
       end do

       ! Remove a circular region component
       call srctool_regions_init_regcomp(regcomp)
       regcomp%shape_code                       = REGCOMP_SHAPE_CODE_CIRCLE
       regcomp%coordsys_code                    = psrc2%coord%coordsys_code
       regcomp%num_pars                         = REGCOMP_NUM_PARS_CIRCLE
       regcomp%pars(REGCOMP_PARS_CIRCLE_LONG)   = psrc2%coord%long
       regcomp%pars(REGCOMP_PARS_CIRCLE_LAT)    = psrc2%coord%lat
       regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS) = exclude_r
       regcomp%weight                           = -1
       regcomp%max_radius                       = exclude_r
       call srctool_regions_add_regcomp(psrc%src_region, regcomp)

    end do

  end subroutine srctool_autoreg_remove_confusing_srcs

  ! now look for potentially confusing neighbouring sources
  ! ------------------------------------------------------------------
  ! Exclusion policy for background regions:
  ! ------------------------------------------------------------------
  ! Exclude region around each source within which the surface
  ! brightness of the PSF wings is greater than XX% of the local
  ! background surface brightness.  Clip to some minimum and maximum
  ! radii.  Use curve of growth of survey PSF to compute radius.
  ! ------------------------------------------------------------------

  !---------------------------------------------------------------------------
  subroutine srctool_autoreg_back_remove_confusing_srcs(srclist, par, i_src, back_r2, ierr)
    character(len=*), parameter :: subroutinename='srctool_autoreg_remove_confusing_backs'
    type(srclist_type),  intent(inout), target :: srclist
    type(param_type),    intent(in)            :: par
    integer(kind=int32), intent(in)            :: i_src
    real(kind=real64),   intent(in)            :: back_r2
    integer(kind=int32), intent(out)           :: ierr

    integer(kind=int32)      :: i_src2
    type(src_type), pointer  :: psrc, psrc2
    real(kind=real64)        :: dr, d, exclude_r, conf_map
    real(kind=real64)        :: area_interp, rpsf_interp
    type(regcomp_type)       :: regcomp

    psrc => srclist%psrc(i_src)

    do i_src2 = 1, srclist%num_srcs
       if ( i_src2 == i_src ) cycle    ! avoid self-exclusion
       psrc2 => srclist%psrc(i_src2)

       if ( .not. psrc2%cat_exclude ) cycle  ! only cat_exclude sources considered

       ! compute the radius which should be excluded
       dr = par%radial_stepsize_back ! radius increment per step
       ! compute distance of confusing source from target source:
       d = calc_ang_sep(psrc%coord%long, psrc%coord%lat, psrc2%coord%long, psrc2%coord%lat)

       if ( d < back_r2 + psrc2%prpsf%radius_99pc_deg ) then
          ! start with some small minimum radius
          exclude_r = par%minimum_exclude_radius
          do
             if ( exclude_r >= psrc2%prpsf%radius_99pc_deg ) then
                exit ! stop incrementing
             end if

             if ( exclude_r >= d ) then
                ! We do not want exclusion radius from source to be
                ! larger than radius between sources. This is
                ! important, for example, for AGN embedded in clusters.
                exit
             end if

             call srctool_autoreg_interp_psf(psrc2, exclude_r, rpsf_interp, area_interp, ierr)
             if ( ierr /= ERROR_LEVEL_OK ) return
             conf_map = psrc2%cat_scts * rpsf_interp / area_interp

             if ( conf_map <= psrc%cat_bg_map*par%max_conf_map_to_back_map_ratio ) then
                exit  ! stop incrementing
             else
                exclude_r = exclude_r + dr ! try a larger radius
             end if
          end do

          ! Remove a circular region component
          call srctool_regions_init_regcomp(regcomp)
          regcomp%shape_code                       = REGCOMP_SHAPE_CODE_CIRCLE
          regcomp%coordsys_code                    = psrc2%coord%coordsys_code
          regcomp%num_pars                         = REGCOMP_NUM_PARS_CIRCLE
          regcomp%pars(REGCOMP_PARS_CIRCLE_LONG)   = psrc2%coord%long
          regcomp%pars(REGCOMP_PARS_CIRCLE_LAT)    = psrc2%coord%lat
          regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS) = exclude_r
          regcomp%weight                           = -1
          regcomp%max_radius                       = exclude_r
          call srctool_regions_add_regcomp(psrc%back_region, regcomp)

       end if
    end do
  end subroutine srctool_autoreg_back_remove_confusing_srcs

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------------- Automatic region routines -------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  ! set up extraction regions automatically

  ! Algorithm: common to all AUTO modes types
  !
  ! foreach (source to be extracted)
  !   * compute size of source extraction radius
  !       - based on SRC_RATE vs BG_RATE, informed by EEF(r)
  !   * compute initial** size of inner and outer radii of background annulus
  !       - based on excluding source flux and including enough background counts
  !   * compute list of potentially confusing sources
  !       - anything which could touch the outer background radius
  !   * foreach (potentially confusing source)
  !       - compute source region exclusion radius
  !       - compute bkgnd region exclusion radius
  !       - test which exclusion regions overlap with source and background regions
  !     end
  !   * write source and bkgnd regions to source structure
  ! end
  !
  ! Note that sometimes we are only computing one of src or background regions
  ! in which case some of the above steps can be excluded
  ! ** In cases where too little background remains after excluding confusing sources

  !---------------------------------------------------------------------------
  subroutine srctool_autoreg_generate(instlist, srclist, ierr)
    character(len=*), parameter  :: subroutinename='srctool_autoreg_generate'
    type(instlist_type), intent(in), target   :: instlist
    type(srclist_type), intent(inout), target :: srclist
    integer(kind=int32), intent(out)          :: ierr

    character(len=STR_MAX)      :: str_buffer
    real(kind=real64)           :: src_r       ! source radius (deg)
    real(kind=real64)           :: back_r2     ! background outer radius (deg)
    integer(kind=int32)         :: i_src       ! source number
    integer(kind=int32)         :: backiter    ! background region iterations

    type(param_type)            :: par         ! parameters for region generation

    ierr = ERROR_LEVEL_OK

    ! read or setup default parameters
    call srctool_autoreg_setup_params(par)

    ! setup PSFs given source list
    call srctool_autoreg_setup_psfs(instlist, srclist, par, ierr)
    if( ierr /= ERROR_LEVEL_OK ) return

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Compute regions

    do i_src = 1, srclist%num_srcs
       associate ( src => srclist%psrc(i_src) )
         write(str_buffer, '(A,A,6A)') &
              "Computing auto regions for src: ", trim(src%name), &
              "  extract=", logical2string(src%cat_extract), &
              " exclude=", logical2string(src%cat_exclude), &
              " extent=", EXTTYPE_CODE_STR(src%extent%code)
         call wrschatty(subroutinename, str_buffer)

         if ( .not. src%cat_extract ) cycle

         ! if ( srclist%num_srcs > 1 ) then
         !   write(str_buffer, '(A,A,A,F10.2,A,A,F10.5,A,A,A,A,F8.1,A)'), &
         !       "  DEBUG Catalogue details for src: ", trim(src%name), &
         !       " SCTS=", src%cat_scts, " cts", &
         !       " BG_MAP=", src%cat_bg_map/SQDEG2SQARCMIN, " cts/arcmin2", &
         !       " EXTTYPE=", trim(EXTTYPE_CODE_STR(src%extent%code)), &
         !       " EXT=", src%cat_ext*DEG2ARCSEC, " arcsec"
         !   call wrschatty(subroutinename, str_buffer)
         ! end if

         if ( srclist%autoreg_code_src /= AUTOREG_CODE_NULL ) then
            ! generate source region
            call srctool_autoreg_src_inclusion(src, par, src_r, ierr)
            if( ierr /= ERROR_LEVEL_OK ) return

            ! remove confusing sources
            call srctool_autoreg_remove_confusing_srcs(srclist, par, i_src, src_r, ierr)
            if( ierr /= ERROR_LEVEL_OK ) return

            if ( srclist%autoreg_code_back /= AUTOREG_CODE_NULL ) then
               ! run this twice, to take account of area excluded by
               ! confusing sources in the background regions
               ! 1st it. -> generate confusing source list given background region
               ! 2nd it. -> expand background region given confusing sources
               do backiter=1,2
                  ! generate background region
                  call srctool_autoreg_calc_back(src, par, src_r, back_r2, ierr)
                  if( ierr /= ERROR_LEVEL_OK ) return

                  ! remove confusing sources
                  call srctool_autoreg_back_remove_confusing_srcs(srclist, par, i_src, back_r2, ierr)
                  if( ierr /= ERROR_LEVEL_OK ) return
               end do
            end if
         end if
       end associate
    end do

  end subroutine srctool_autoreg_generate

end module srctool_autoreg
