! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_badpix

  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling badpix data
  !
  !-----------------------------------------------------------------------------

  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         !
  use srctool_misc              !
  use srctool_generic_fitsraw

  implicit none     ! insist on explicit typing

  ! some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_BADPIX_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_badpix
  private :: srctool_badpix_process_badpix
  private :: srctool_badpix_setup_effmap
  private :: srctool_badpix_process_effmap_changes
  private :: srctool_badpix_get_effmap_value

  public ::  srctool_badpix_empty_badpix
  public ::  srctool_badpix_empty_effmap
  public ::  srctool_badpix_read_badpix_from_chdu
  public ::  srctool_badpix_process_all_badpix
  public ::  srctool_badpix_calculate_effmap
  public ::  srctool_badpix_get_effmap_values

  interface srctool_badpix_get_effmap
    module procedure srctool_badpix_get_effmap_value
    module procedure srctool_badpix_get_effmap_values
  end interface srctool_badpix_get_effmap

contains




  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- START of BADPIX handling --------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_badpix_empty_badpix(badpix)
    type(badpix_type), intent(inout)      :: badpix
    if ( allocated(badpix%pmap)) deallocate (badpix%pmap)
    badpix%class  = SRCTOOL_BADPIX_CLASS_NULL
    badpix%naxis1 = 0
    badpix%naxis2 = 0
    badpix%num_rows = 0
    if ( allocated(badpix%prawx   )) deallocate(badpix%prawx   )
    if ( allocated(badpix%prawy   )) deallocate(badpix%prawy   )
    if ( allocated(badpix%pyextent)) deallocate(badpix%pyextent)
    if ( allocated(badpix%ptype   )) deallocate(badpix%ptype   )
    if ( allocated(badpix%pbadflag)) deallocate(badpix%pbadflag)
    if ( allocated(badpix%ptimemin)) deallocate(badpix%ptimemin)
    if ( allocated(badpix%ptimemax)) deallocate(badpix%ptimemax)
    if ( allocated(badpix%pphamin )) deallocate(badpix%pphamin )
    if ( allocated(badpix%pphamax )) deallocate(badpix%pphamax )
    badpix%num_badpix = 0
    badpix%is_valid = .false.
    badpix%filter_for_patterns  = .true.
    call fitswrap_init_hduinfo(badpix%hduinfo)
  end subroutine srctool_badpix_empty_badpix
  !------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !read the badpix list/map from the current hdu of the fits file identified by unit
  subroutine srctool_badpix_read_badpix_from_chdu(unit, badpix, ierr)
    character(len=*), parameter :: subroutinename='srctool_badpix_read_badpix_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(badpix_type), intent(inout)  :: badpix
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status
    integer(kind=int32) :: group
    integer(kind=int8) :: nullval
    character(len=STR_MAX) :: str_buffer

!    !for writing to a badpix image
    integer(kind=int32) :: i,j

    integer(kind=int32) :: colnum_rawx
    integer(kind=int32) :: colnum_rawy
    integer(kind=int32) :: colnum_yextent
    integer(kind=int32) :: colnum_type
    integer(kind=int32) :: colnum_badflag
    integer(kind=int32) :: colnum_timemin
    integer(kind=int32) :: colnum_timemax
    integer(kind=int32) :: colnum_phamin
    integer(kind=int32) :: colnum_phamax

    integer(kind=int32) :: nrows_per_chunk
    integer(kind=int32) :: nrows_to_read
    integer(kind=int32) :: row
    integer(kind=int32) :: row_end
    integer(kind=int32) :: anyf
    integer(kind=int16) , parameter :: nullval_int16   = 0
    integer(kind=int32) , parameter :: nullval_int32   = 0
    real(kind=real64)   , parameter :: nullval_real64 = 0.0d0

    ierr = ERROR_LEVEL_OK
    status = 0
    anyf = 0
    group = 0
    nullval = 0
    call wrschatty (subroutinename, "Reading in Badpix info from CHDU")

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, badpix%hduinfo, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if

    write (str_buffer, '(A,A,A,I0,A,A)') "I am currently in extension:", &
        trim(badpix%hduinfo%key_extname%value), " index=", &
        badpix%hduinfo%hdu_index, &
        " of ", trim(badpix%hduinfo%filename)
    call wrschatty (subroutinename, str_buffer)

    ! Subsequent steps depend on the format of the BADPIX file
    ! either an image or a table of bad pixels
    ! JSS: Has this been tested?
    if ( badpix%hduinfo%hdu_type == FITSIO_IMAGE_HDU ) then
      call wrschatty (subroutinename, "This is a fits IMAGE extension")
      ! get the image dimension keywords
      badpix%class = SRCTOOL_BADPIX_CLASS_IMAGE
      badpix%naxis1 = 0
      badpix%naxis2 = 0
      call FTGKYJ_w(unit, "NAXIS1", badpix%naxis1, str_buffer, status)
      call FTGKYJ_w(unit, "NAXIS2", badpix%naxis2, str_buffer, status)
      if (status /= 0) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if

      if ( (badpix%naxis1 <= 0) .or. (badpix%naxis2 <= 0)  ) then
        write (str_buffer, '(A,I0,A,I0)') "Strange badpix image dimensions: (naxis1,naxis2)=", &
            badpix%naxis1,",", badpix%naxis2
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end if

      ! allocate some space
      allocate ( badpix%pmap(badpix%naxis1,badpix%naxis2), STAT = status)
      if (check_allocate_status(subroutinename, status, ierr)) return

      badpix%pmap(:,:) = 0 ! set to zero

      !read in the image data
      ! JSS: Change FTGPVB -> FTGPVJ as wrapper doesn't like passing int32 array to an int8 routine
      call FTGPVJ_w(unit,group,1,(badpix%naxis1*badpix%naxis2),nullval_int32, flatten(badpix%pmap),anyf,status)
      if (status /= 0) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if

      ! count the number of bad pixels
      badpix%num_badpix = 0
      do i=1,badpix%naxis1
        do j=1,badpix%naxis2
          if ( badpix%pmap(badpix%prawx(i),badpix%prawy(j)) /= 0 ) then
            badpix%num_badpix = badpix%num_badpix + 1
          end if
        end do
      end do

      write (str_buffer, '(A,I0,A,I0,A)') &
          "I read in: ", badpix%naxis1, "x", badpix%naxis2, " badpix image pixels"
      call wrschatty (subroutinename, str_buffer)

    else if (  badpix%hduinfo%hdu_type == FITSIO_BINARY_TBL ) then
      call wrschatty (subroutinename, "This is a fits BINARY_TBL extension")
      !--------------------------------------------------------------------
      ! Add in code to handle this this HDU type
      ! expect a list of badpixels in table format with coords in RAWX,RAWY
      ! read in the bad pixels as lists and then convert to mask images internally
!      badpix%naxis1 = inst%detmap%naxis1
!      badpix%naxis2 = inst%detmap%naxis2
      !get the number of rows in the extension
      call FTGNRW_w(unit, badpix%num_rows, status)
      if (status /= 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if
      badpix%class = SRCTOOL_BADPIX_CLASS_TABLE

      !check that have at least one bad pixel and then read in the column data
      if ( badpix%num_rows > 0 ) then
        write (str_buffer, '(A,I0,A,A,A,A,A)') "Reading in: ", badpix%num_rows, " rows of badpix info from: '", &
            trim(badpix%hduinfo%filename), "[", trim(badpix%hduinfo%key_extname%value), "]'"
        call wrschatty (subroutinename, str_buffer)

        call fitswrap_prep_col(unit, "RAWX",    badpix%num_rows, colnum_rawx,    badpix%prawx,    ierr)
        call fitswrap_prep_col(unit, "RAWY",    badpix%num_rows, colnum_rawy,    badpix%prawy,    ierr)
        call fitswrap_prep_col(unit, "YEXTENT", badpix%num_rows, colnum_yextent, badpix%pyextent, ierr)
        call fitswrap_prep_col(unit, "TYPE",    badpix%num_rows, colnum_type,    badpix%ptype,    ierr)
        call fitswrap_prep_col(unit, "BADFLAG", badpix%num_rows, colnum_badflag, badpix%pbadflag, ierr)
        call fitswrap_prep_col(unit, "TIMEMIN", badpix%num_rows, colnum_timemin, badpix%ptimemin, ierr)
        call fitswrap_prep_col(unit, "TIMEMAX", badpix%num_rows, colnum_timemax, badpix%ptimemax, ierr)
        call fitswrap_prep_col(unit, "PHAMIN",  badpix%num_rows, colnum_phamin,  badpix%pphamin,  ierr)
        call fitswrap_prep_col(unit, "PHAMAX",  badpix%num_rows, colnum_phamax,  badpix%pphamax,  ierr)

        if (ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Problem preparing column(s) for reading from HDU")
          return
        end if
        ! get the chunksize
        call FTGRSZ_w(unit, nrows_per_chunk, status)
        if (status /= 0) then
          call fitswrap_write_fits_error(subroutinename, status)
          ierr = ERROR_LEVEL_SEVERE
          return
        end if


        do row = 1, badpix%num_rows, nrows_per_chunk
          nrows_to_read = min(nrows_per_chunk,badpix%num_rows-row+1)
          row_end = row+nrows_to_read-1
          call FTGCVI_w(unit, colnum_rawx,   row,1,nrows_to_read,nullval_int16, badpix%prawx(row:row_end),    anyf,status)
          call FTGCVI_w(unit, colnum_rawy,   row,1,nrows_to_read,nullval_int16, badpix%prawy(row:row_end),    anyf,status)
          call FTGCVI_w(unit, colnum_yextent,row,1,nrows_to_read,nullval_int16, badpix%pyextent(row:row_end), anyf,status)
          call FTGCVI_w(unit, colnum_type,   row,1,nrows_to_read,nullval_int16, badpix%ptype(row:row_end),    anyf,status)
          call FTGCVI_w(unit, colnum_badflag,row,1,nrows_to_read,nullval_int16, badpix%pbadflag(row:row_end), anyf,status)
          call FTGCVD_w(unit, colnum_timemin,row,1,nrows_to_read,nullval_real64,badpix%ptimemin(row:row_end), anyf,status)
          call FTGCVD_w(unit, colnum_timemax,row,1,nrows_to_read,nullval_real64,badpix%ptimemax(row:row_end), anyf,status)
          call FTGCVI_w(unit, colnum_phamin, row,1,nrows_to_read,nullval_int16, badpix%pphamin(row:row_end),  anyf,status)
          call FTGCVI_w(unit, colnum_phamax, row,1,nrows_to_read,nullval_int16, badpix%pphamax(row:row_end),  anyf,status)

          if (status /= 0) then
            call fitswrap_write_fits_error(subroutinename, status)
            ierr = ERROR_LEVEL_SEVERE
            return
          end if
        end do



      end if

      ! now sort the bad pix in order of increasing rawx
      ! TODO

      ! now count the number of bad pixels

      !--------------------------------------------------------------------
    else
      ! error, unknown HDU type
      ierr = ERROR_LEVEL_SEVERE
      badpix%class = SRCTOOL_BADPIX_CLASS_NULL
      call wrserr (subroutinename, ierr, "Error, unknown HDU type")
      return
    end if
    badpix%is_valid = .true.

  end subroutine srctool_badpix_read_badpix_from_chdu
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! create a badpix map, count number of bad pixels etc
  subroutine srctool_badpix_process_badpix(inst, badpix, instpar, ierr)
    character(len=*), parameter :: subroutinename='srctool_badpix_process_badpix'
    type(inst_type), intent(in) :: inst
    type(badpix_type), intent(inout)  :: badpix
    type(instpar_type), intent(in) :: instpar
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status
    character(len=STR_MAX) :: str_buffer

    !for writing to a badpix image
    integer(kind=int32) :: i, j, ymax

    ierr = ERROR_LEVEL_OK
    status = 0

    if ( .not. badpix%is_valid ) then  ! skip silently
      return
    end if

    if ( badpix%class  == SRCTOOL_BADPIX_CLASS_NONE ) then
!      call wrschatty (subroutinename, "Skipping empty badpix Description")
    else if ( badpix%class == SRCTOOL_BADPIX_CLASS_IMAGE ) then ! badpix info was read from an image
      write (str_buffer, '(A,A)') "Processing badpix image info for inst: ", trim(inst%name)
      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,I0,A)') "Summary: There were ", badpix%num_badpix, " badpixels"
      call wrschatty (subroutinename, str_buffer)

    else if ( badpix%class == SRCTOOL_BADPIX_CLASS_TABLE ) then
      write (str_buffer, '(A,A)') "Processing badpix binary table info for inst: ", trim(inst%name)
      call wrschatty (subroutinename, str_buffer)

!      call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
!      if (ierr /= ERROR_LEVEL_OK) then
!        write (str_buffer, '(A,A)'), "Could not find suitable instpar for inst: ", trim(inst%name)
!        call wrserr (subroutinename, ierr, str_buffer)
!        return
!      end if

      badpix%naxis1 = instpar%naxis1
      badpix%naxis2 = instpar%naxis2

      ! ----------------------------------------------------
      ! insert the bad pixels into the image

      ! First check that the badpixmap is not already allocated
      if ( allocated(badpix%pmap)) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "badpixmap is already allocated")
        return
      end if

      ! allocate some space for the image
      allocate ( badpix%pmap(badpix%naxis1,badpix%naxis2), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
      badpix%pmap(:,:) = 0 ! set to zero

      ! the badpix map values contain the index of the first bad pixel in the column data
      badpix%num_badpix = 0
      do i=1,badpix%num_rows
        ymax = min(badpix%prawy(i)+badpix%pyextent(i)-1,badpix%naxis2)
        do j=badpix%prawy(i),ymax
          if ((badpix%prawx(i) > 0)              .and. &
              (badpix%prawx(i) <= badpix%naxis1) .and. &
              (j > 0)                            .and. &
              (j <= badpix%naxis2) ) then
            ! if pixel is not already recorded as being bad, then record the index
            if ( badpix%pmap(badpix%prawx(i),j) == 0 ) then
              badpix%pmap(badpix%prawx(i),j) = i
              badpix%num_badpix = badpix%num_badpix + 1
            end if
          end if
        end do
      end do
      ! ----------------------------------------------------

      write (str_buffer, '(A,I0,A)') "Summary: There were ", badpix%num_badpix, " badpixels"
      call wrschatty (subroutinename, str_buffer)
    else
      ! error, unknown HDU type
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Error, unknown BADPIX type")
      return
    end if

    badpix%filter_for_patterns = .true.

  end subroutine srctool_badpix_process_badpix


  subroutine srctool_badpix_process_all_badpix(instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_badpix_process_all_badpix'
    type(instlist_type), intent(inout), target :: instlist
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status

    integer(kind=int32) :: i, j

    type(inst_type), pointer :: pinst
    type(instpar_type), pointer :: pinstpar
    type(detmap_type), pointer :: pdetmap
    type(evtlist_type), pointer :: pevtlist
    integer(kind=int32)        :: i_first

    ierr = ERROR_LEVEL_OK
    status = 0
    i_first = 0

    do i=1,MAX_INST
      pinst => instlist%inst(i)
      if ( pinst%to_consider ) then

        if ( i_first == 0 ) i_first = i

        call srctool_misc_select_instpar_pointer(pinst,pinstpar,ierr)
        if (ierr /= ERROR_LEVEL_OK) then
          call wrserr (subroutinename, ierr, "Could not find suitable instpar")
          return
        end if
        call srctool_misc_select_detmap_pointer(pinst,pdetmap,ierr)
        if (ierr /= ERROR_LEVEL_OK) then
          call wrserr (subroutinename, ierr, "Could not find suitable detmap")
          return
        end if

        ! do the EVENTLIST specific BADPIX
        if ( pinst%evtlistlist%num_evtlists > 0 ) then
          do j=1,pinst%evtlistlist%num_evtlists
            pevtlist => pinst%evtlistlist%evtlist(j)
            call srctool_badpix_process_badpix(pinst, pevtlist%badpix, pinstpar, ierr)
            if ( ierr /= ERROR_LEVEL_OK ) then
              call wrserr (subroutinename, ierr, "Error processing evtlist badpix")
              return
            end if
          end do
        end if

        call srctool_badpix_setup_effmap(pinst, pinst%effmap, pdetmap, ierr)
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error setting up effmap")
          return
        end if

      end if
    end do

    ! now do the merged instrument
    pinst => instlist%inst(0)
    if ( i_first > 0 ) then
      call srctool_misc_select_instpar_pointer(instlist%inst(i_first),pinstpar,ierr)
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr (subroutinename, ierr, "Could not find suitable instpar")
        return
      end if
      call srctool_misc_select_detmap_pointer(instlist%inst(i_first),pdetmap,ierr)
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr (subroutinename, ierr, "Could not find suitable detmap")
        return
      end if
      ! reuse the temp_flagsel parameter from before - will be valid for the last processed instrument
      call srctool_badpix_setup_effmap(pinst, pinst%effmap, pdetmap, ierr)
      if ( ierr /= ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Error setting up effmap")
        return
      end if
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Could not find suitable inst")
      return
    end if

  end subroutine srctool_badpix_process_all_badpix


  !-------------------------------------------------------------------------------------
  !----------------------------- END of BADPIX handling --------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- START of EFFMAP handling --------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_badpix_empty_effmap(effmap)
    type(effmap_type), intent(inout) :: effmap
    integer(kind=int32) :: j
    if ( allocated(effmap%pmap)) deallocate (effmap%pmap)
    effmap%time_of_last_update = NULLVAL_REAL64
    effmap%time_of_next_update = -1.0d0 * huge(1d0)
    effmap%is_valid = .false.
!    effmap%npatt = 0
    effmap%naxis1 = 0
    effmap%naxis2 = 0
    if ( allocated(effmap%ptchange)) deallocate(effmap%ptchange)
    effmap%num_changes = 0
    do j=1,MAX_EVTFILES+1
      nullify(effmap%pbadpixlist(j)%p)
    end do
  end subroutine srctool_badpix_empty_effmap
  !------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_badpix_setup_effmap(inst, effmap, detmap, ierr)
    character(len=*), parameter :: subroutinename='srctool_badpix_setup_effmap'
    type(effmap_type), intent(inout) :: effmap
    type(detmap_type), intent(in)    :: detmap
    type(inst_type), intent(inout), target :: inst
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: status, j, k
!    character(len=STR_MAX) :: str_buffer

    ierr=ERROR_LEVEL_OK
    status = 0

!    effmap%npatt = SRCTOOL_PATT_NCLASS
    effmap%naxis1 = detmap%naxis1
    effmap%naxis2 = detmap%naxis2

    if ( allocated(effmap%pmap)) then
!      call wrserr(subroutinename,ierr,"")
    else
      ! allocate space, including an additional 1pixel buffer around the actual pixel array
      allocate(effmap%pmap(&
!          0:effmap%npatt+1,&
          0:effmap%naxis1+1,&
          0:effmap%naxis2+1),&
          STAT = status)
      if (check_allocate_status(subroutinename, status, ierr)) return
    end if
    effmap%pmap(:,:) = 0.0e0

    ierr = ERROR_LEVEL_OK
    status = 0


    !----------------------------------------------------------------------------------------
    ! prepare a list of all badpixel files that are potentially valid
    do j=1,MAX_EVTFILES+1
      nullify(effmap%pbadpixlist(j)%p)
    end do
    if ( inst%evtlistlist%num_evtlists > 0 ) then
      k = 1
      do j=1,inst%evtlistlist%num_evtlists
        if ( inst%evtlistlist%evtlist(j)%badpix%is_valid )  then
          effmap%pbadpixlist(k)%p => inst%evtlistlist%evtlist(j)%badpix
          k = k + 1
        end if
      end do
    end if
    !----------------------------------------------------------------------------------------

    ! now deal with the badpix changes array
    call srctool_badpix_process_effmap_changes(inst, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) return


  end subroutine srctool_badpix_setup_effmap
  !------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  recursive subroutine srctool_badpix_calculate_effmap(effmap, detmap, t, apply_fov_mask, ierr)
    character(len=*), parameter :: subroutinename='srctool_badpix_calculate_effmap'
    type(effmap_type), intent(inout) :: effmap
    type(detmap_type), intent(in) :: detmap
    ! this is a list of badpix extensions which should be considered
    real(kind=real64), intent(in) :: t   ! time at which to evaluate effmap
    logical, intent(in) :: apply_fov_mask
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i, j, k, ymin, ymax
    type(badpix_type), pointer :: pbadpix

    ierr=ERROR_LEVEL_OK

    ! first reset the effmap to the values in the detmap
    effmap%pmap(0,:)               = 0.0e0  ! there is a 1pixel buffer around the detmap
    effmap%pmap(effmap%naxis1+1,:) = 0.0e0  ! so set this to zero
    effmap%pmap(:,0)               = 0.0e0  !
    effmap%pmap(:,effmap%naxis2+1) = 0.0e0  !
    if( apply_fov_mask ) then
       ! FoV mask from CALDB
       effmap%pmap(1:effmap%naxis1,1:effmap%naxis2) = detmap%pmap(:,:)
    else
       ! no FoV mask
       effmap%pmap(1:effmap%naxis1,1:effmap%naxis2) = 1.0e0
    end if
!    do j=1,effmap%naxis2
!      do i=1,effmap%naxis1
!        effmap%pmap(i,j) = detmap%pmap(i + j * effmap%naxis1)
!      end do
!    end do

    ! see: http://erosita.mpe.mpg.de/eROdoc/tasks/kod/pattern_doc.html
    ! we only need to set pixels adjacent to bad pixels to zero, diagonals are ok
    !
    !       1   0   1                 0   0   0
    ! new:  0  bad  0       previous: 0  bad  0
    !       0  bad  0                 0  bad  0
    !       0  bad  0                 0  bad  0
    !        "    "                    "    "
    !       0  bad  0                 0  bad  0
    !       1   0   1                 0   0   0

    ! first make a map of all the badpixels at this time
    !
    ! for each badpixel list
    do k=1,size(effmap%pbadpixlist)
      if ( associated(effmap%pbadpixlist(k)%p)) then
        pbadpix => effmap%pbadpixlist(k)%p
        if ( pbadpix%is_valid ) then
          if ( k == 1 ) then
            ! trust that the first badpix contains the correct invalid_pattern_flag
            ! ok because this is checked in srctool_process_all_badpix()
            if ( pbadpix%filter_for_patterns ) then
              ! (this is always true currently)
              ! set the pixels next to the border to zero
              effmap%pmap(1,:)             = 0.0e0  ! Patterns next to the border cannot be valid
              effmap%pmap(effmap%naxis1,:) = 0.0e0  ! so set them to zero
              effmap%pmap(:,1)             = 0.0e0  !
              effmap%pmap(:,effmap%naxis2) = 0.0e0  !
            endif
          endif

          select case ( pbadpix%class )

          case ( SRCTOOL_BADPIX_CLASS_TABLE )
            do i=1,pbadpix%num_rows ! for each badpix entry
              if ((t >= pbadpix%ptimemin(i)) .and. &
                  (t <= pbadpix%ptimemax(i))) then  ! if covers this time
                if ((pbadpix%prawx(i) >= 1) .and. &
                    (pbadpix%prawx(i) <= pbadpix%naxis1 )) then ! if xcoord is sensible
                  ymin = max(pbadpix%prawy(i),1)
                  ymax = min(pbadpix%prawy(i)+pbadpix%pyextent(i)-1,pbadpix%naxis2)

                  if ( pbadpix%filter_for_patterns ) then
                    ! (this is always true currently)
                    ! set any neighbouring pixels to zero as well

                    ! set any adjacent pixels along RAWX to zero too
                    effmap%pmap(pbadpix%prawx(i)-1:pbadpix%prawx(i)+1,ymin:ymax) = 0.0e0
                    ! set the pixels immediately before and after along the rawy direction to zero too
                    effmap%pmap(pbadpix%prawx(i),max(ymin-1,1)) = 0.0e0
                    effmap%pmap(pbadpix%prawx(i),min(ymin+1,pbadpix%naxis2)) = 0.0e0

                  else
                    ! just set the bad pixels themselves to zero
                    effmap%pmap(pbadpix%prawx(i),ymin:ymax) = 0.0e0 ! set to zero
                  end if
                end if
              end if
            end do

          case ( SRCTOOL_BADPIX_CLASS_IMAGE )
            do i=1,effmap%naxis1
              do j=1,effmap%naxis2
                if ( pbadpix%pmap(i,j) > 0 ) then
                  if ( pbadpix%filter_for_patterns ) then
                    ! (this is always true currently)
                    ! set any neighbouring pixels to zero as well
                    effmap%pmap(i,j-1:j+1) = 0.0e0        ! three pixels in rawy
                    effmap%pmap(i-1:i+1,j) = 0.0e0        ! three pixels in rawx
                  else
                    ! just set the bad pixel itself to zero
                    effmap%pmap(i,j) = 0.0d0
                  endif
                end if
              end do
            end do

          case ( SRCTOOL_BADPIX_CLASS_NONE )
            ! do nothing

          case default
            ierr=ERROR_LEVEL_SEVERE
            call wrserr (subroutinename, ierr, "Unknown class of BADPIX description")
            return

          end select
        end if
      end if
    end do


    ! now work out when the next update should be
    i = 1
    do i=1,effmap%num_changes
      if ( t < effmap%ptchange(i)) then
        effmap%time_of_next_update = effmap%ptchange(i)
        exit ! leave the loop
      endif
    end do
    effmap%time_of_last_update = t


  end subroutine srctool_badpix_calculate_effmap
  !------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! calculate the list of times when any bad pixel status changes
  subroutine srctool_badpix_process_effmap_changes(inst, ierr)
    character(len=*), parameter :: subroutinename='srctool_badpix_process_effmap_changes'
    type(inst_type), intent(inout), target :: inst
    integer(kind=int32), intent(out)  :: ierr
    integer(kind=int32) :: i,j,k, status
    character(len=STR_MAX) :: str_buffer

    ierr = ERROR_LEVEL_OK
    status = 0

    ! first count up the total number of bad pixels
    inst%effmap%num_changes = 0
    do j=1,MAX_EVTFILES+1
      if ( associated(inst%effmap%pbadpixlist(j)%p) ) then
        if (inst%effmap%pbadpixlist(j)%p%is_valid .and. &
            (inst%effmap%pbadpixlist(j)%p%class == SRCTOOL_BADPIX_CLASS_TABLE)) then
          inst%effmap%num_changes = inst%effmap%num_changes + inst%effmap%pbadpixlist(j)%p%num_rows
        end if
      endif
    end do


    ! need space for min and max time for each badpix
    ! plus allow space for two extra bounding values
    inst%effmap%num_changes = (2*inst%effmap%num_changes) + 2

    allocate (inst%effmap%ptchange(1:inst%effmap%num_changes), STAT = status)

    if ( check_allocate_status(subroutinename, status, ierr) ) return

    ! now put the start/stop times into the array
    inst%effmap%ptchange(1)                       = -1.0d0 * huge(1.0d0)
    inst%effmap%ptchange(inst%effmap%num_changes) = huge(1.0d0)
    i = 2

    do j=1,MAX_EVTFILES+1
      if ( associated(inst%effmap%pbadpixlist(j)%p)) then
        if (inst%effmap%pbadpixlist(j)%p%is_valid .and. &
            (inst%effmap%pbadpixlist(j)%p%class == SRCTOOL_BADPIX_CLASS_TABLE)) then
          do k=1,inst%effmap%pbadpixlist(j)%p%num_rows
            inst%effmap%ptchange(i)   =  inst%effmap%pbadpixlist(j)%p%ptimemin(k)
            inst%effmap%ptchange(i+1) =  inst%effmap%pbadpixlist(j)%p%ptimemax(k)
            i = i + 2
          end do
        end if
      endif
    end do

    ! now sort them
    call sort_array(inst%effmap%ptchange)

    call filter_unique_real64(inst%effmap%ptchange(:),inst%effmap%num_changes)

    write (str_buffer, '(A,A,A,A,I0)') subroutinename, &
        " Inst= ", trim(inst%name), " effmap%num_changes=", inst%effmap%num_changes
    call wrschatty(subroutinename, str_buffer)

    ! TODO tidy up spare memory via a temporary array

  end subroutine srctool_badpix_process_effmap_changes
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_badpix_get_effmap_value(effmap, on_chip_flag, rawx_int, rawy_int, effmap_val)
    type(effmap_type), intent(in)        :: effmap
    logical          , intent(in)       :: on_chip_flag
    integer(kind=int32), intent(in)      :: rawx_int, rawy_int
    real(kind=real64), intent(out)       :: effmap_val

    if ( on_chip_flag ) then
      effmap_val = REAL(effmap%pmap(rawx_int,rawy_int), real64)
    else
      effmap_val = 0.0d0
    endif
  end subroutine srctool_badpix_get_effmap_value
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_badpix_get_effmap_values(effmap, samplelist)
    type(effmap_type), intent(in)        :: effmap
    type(samplelist_type), intent(inout) :: samplelist
    integer(kind=int32) :: i

    samplelist%peffmap(:) = 0.0d0  ! initialise to having all samples at zero sensitivity
    do i=1,samplelist%num_points
      if ( samplelist%pon_chip_flag(i) ) then
        samplelist%peffmap(i) = REAL(effmap%pmap(samplelist%prawx_int(i),samplelist%prawy_int(i)), real64)
      endif
    end do
  end subroutine srctool_badpix_get_effmap_values
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- END of EFFMAP handling ----------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

end module srctool_badpix
