! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_cal

  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling calibration data
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_generic_avign     ! generic analytic vignetting file handling routines
  use srctool_datatypes         !
  use srctool_misc              !
  use srctool_instpar           !
  use srctool_gti               !
  use srctool_ebounds           !
  use srctool_spec              !
  use srctool_att               !
!  use srctool_vign              !
  use srctool_lc                !
  use srctool_badpix            !
  use srctool_pattcal           !
  use srctool_psf               !
  use srctool_detmap            !
  use srctool_arf               !
  use srctool_rmf               !
  use srctool_deadcor           !
  use srctool_generic_fitsraw

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_CAL_MOD'


  ! subroutine declarations and interfaces for module: srctool_generic_cal
  public :: srctool_cal_get_filenames_from_caldb
  public :: srctool_cal_get_filename_from_caldb
  public :: srctool_cal_prep_gtcalf
  public :: srctool_cal_read_cal_from_files
  public :: srctool_cal_read_cal_from_file
  public :: srctool_cal_report_cal_status
  public :: srctool_cal_convert_keyval_to_index
  public :: srctool_cal_instrume_to_index

contains


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !----------------------- START of CAL/CALDB handling ---------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  subroutine srctool_cal_get_filenames_from_caldb(instlist, cal_type, filenames, ierr)
    character(len=*), parameter :: subroutinename='srctool_cal_get_filenames_from_caldb'
    type(instlist_type), intent(inout), target :: instlist
    type(srctool_cal_type),intent(in)   :: cal_type
    integer(kind=int32), intent(out)  :: ierr
    character(len=*),    intent(out), dimension(:) :: filenames
    integer(kind=int32) :: i
!    character(len=STR_MAX) :: str_buffer
    type(inst_type), pointer  :: pinst
    ierr = ERROR_LEVEL_OK

    do i=1,MAX_INST
      pinst => instlist%inst(i)
      if ( pinst%to_consider .and. (pinst%evtlistlist%num_evtlists .ge. 1) ) then
        call srctool_cal_get_filename_from_caldb(pinst, cal_type, &
            pinst%evtlistlist%hduinfo%key_date_obs%value, &
            pinst%evtlistlist%hduinfo%key_date_end%value, &
            filenames(i), ierr)
      else
        ! skip
      end if
      nullify(pinst)
    end do

  end subroutine srctool_cal_get_filenames_from_caldb

  !---------------------------------------------------------------------------------------------
  subroutine srctool_cal_get_filename_from_caldb(inst, cal_type, date_obs, date_end, filename, ierr)
    character(len=*), parameter :: subroutinename='srctool_cal_get_filename_from_caldb'
    type(inst_type), intent(inout), target :: inst
    type(srctool_cal_type),intent(in)   :: cal_type
    character(len=*), intent(in) :: date_obs, date_end
    integer(kind=int32), intent(out)  :: ierr
    character(len=*), intent(out) :: filename

    character(len=*), parameter :: str_detnam = "-"
    character(len=FITSIO_KEY_MAXLEN) :: str_instrume
    character(len=FITSIO_KEY_MAXLEN) :: str_filter

    integer(kind=int32), parameter :: group = 1, maxret = 256
    character(len=STR_MAX) :: str_expression
    character(len=30) :: str_codename
    character(len=10) :: start_date, start_time, stop_date, stop_time
    character(len=STR_MAX), dimension(maxret) :: pfilename , ponline
    integer(kind=int32), dimension(maxret) :: pextno
    integer(kind=int32) :: nret, nfound
    integer(kind=int32) :: chatter                          ! caldb verbosity

    integer(kind=int32) :: status
    integer(kind=int32) :: i
    character(len=STR_MAX) :: str_buffer
!    type(hduinfo_type), pointer :: phduinfo

    ierr = ERROR_LEVEL_OK
    status = 0
    filename = ""
    str_codename = ""
    str_filter = ""
    pextno(:) = 0
    pfilename(:) = ""
    ponline(:) = ""
    nret = 0
    nfound = 0


!    nullify(phduinfo)
!    str_expression = "ENERG.eq.1500"  ! DEBUG TODO
    str_expression = "-"  ! DEBUG TODO

    str_instrume = trim(inst%name)

!    str_filter   = trim(inst%evtlistlist%hduinfo%key_filter%value)
    str_filter   = "-"

    select case (cal_type%code)
    case (SRCTOOL_CALTYPE_ARF%code)
      str_codename   = "SPECRESP"
!      str_filter     = "FILTER"
      str_filter   = trim(inst%evtlistlist%hduinfo%key_filter%value)
      str_expression = "-"  ! DEBUG TODO

    case (SRCTOOL_CALTYPE_RMF%code)
      str_codename   = "RMF"
       str_expression = "-"  ! DEBUG TODO

    case (SRCTOOL_CALTYPE_EBOUNDS%code)
      str_codename   = "EBOUNDS"
      str_expression = "-"  ! DEBUG TODO

!    case (SRCTOOL_CALTYPE_VIGN%code)
!      str_codename   = "TVIGNET"
!      str_filter     = "-"
!      str_expression = "-"  ! DEBUG TODO

    case (SRCTOOL_CALTYPE_AVIGN%code)
      str_codename   = "AVIGNET2"
       str_expression = "-"  ! DEBUG TODO

    case (SRCTOOL_CALTYPE_PSF%code)
      str_codename   = "2D_PSF"
!      str_filter     = "ENERG.eq.500.0&&THETA.eq.0.0"
!      str_filter     = "ENERG(100.-2499.999)eV&&THETA(0.0-2.499)arcmin"
!      str_expression = "OBSMODE.eq.'POINTING'"  ! gtcalf() doesn't like this form
!      str_expression = 'OBSMODE.eq."POINTING"'
      str_expression = 'OBSMODE.eq."POINTING".and.ENERG.eq.500.0.and.THETA.eq.0.0'

    case (SRCTOOL_CALTYPE_SURVEYPSF%code)
      str_codename   = "2D_PSF"
!      str_filter     = "ENERG.eq.500.0"
!      str_filter     = "ENERG(100.-2499.999)eV&&THETA(0.0-2.499)arcmin"
!      str_expression = "OBSMODE.eq.'POINTING'"  ! gtcalf() doesn't like this form
      str_expression = 'OBSMODE.eq."SURVEY".and.ENERG.eq.1000.0'

    case (SRCTOOL_CALTYPE_DETMAP%code)
      str_codename   = "DETMAP"
      str_expression = "-"  ! DEBUG TODO

    case (SRCTOOL_CALTYPE_INSTPAR%code)
      str_codename   = "GEOM"
      str_expression = "-"  ! DEBUG TODO

    case (SRCTOOL_CALTYPE_PATTCAL%code)
      str_codename   = "PATFRAC"! TODO - fill in correct name convention here
      str_expression = "-"  ! DEBUG TODO

    case default
      ierr = ERROR_LEVEL_SEVERE
      return
    end select


    ! create test expression (TODO this is preliminary - needs to be improved

    ! prepare the cal
    call srctool_cal_prep_gtcalf(date_obs, date_end, &
        start_date, start_time, stop_date, stop_time, &
        chatter, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr(subroutinename, ierr,  "Problem preparing date/time strings for cal")
      return
    end if

    ! now get the cal filename
    nret = 0
    nfound = 0
    call gtcalf (chatter, &
        trim(inst%evtlistlist%hduinfo%key_telescop%value), &
        trim(str_instrume), &
        trim(str_detnam), &
        trim(str_filter), &
        trim(str_codename), &
        start_date, start_time, stop_date, stop_time, &
        trim(str_expression), &
        maxret, pfilename, pextno, ponline, &
        nret, nfound, &
        status)

    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, &
          "Problem reading filename+ext for cal_type='" // trim(str_codename) // &
          "' for inst='" // trim(inst%name) // "' filter='" // trim(str_filter) // &
          "' expression='" // trim(str_expression) // "'" )
      return
    end if

    if (nfound .gt. nret) then
      ierr = ERROR_LEVEL_WARNING
      call wrserr(subroutinename, ierr, &
          "*Warning* More files+extensions (" // trim(i2s(nfound)) // &
          ") were found than could be returned (" // trim(i2s(nret)) // ") for cal_type='" // &
          trim(str_codename) // "' for inst='" // trim(inst%name) // "' filter='" // trim(str_filter) // &
          "' expression='" // trim(str_expression) // "'")
      ierr = ERROR_LEVEL_OK
    endif

    ! This is a temporary kludge to fix issues with conflicts in the CALDB
    ! preference is to throw a warning and not an error when we read more than one CALDB filename
    ! TODO eventually replace with code below marked as !proper-way!
    if (nret .gt. 1) then
      write (str_buffer, '(A,I0,A,A,A,A,A)') &
          "POSSIBLE CALDB CONFLICT: More than one filename+ext (", nret ,") was returned for cal_type='", &
          trim(str_codename), "' for inst='", trim(inst%name), "'"
      if ((cal_type%code .eq. SRCTOOL_CALTYPE_PSF%code) .or. &
          (cal_type%code .eq. SRCTOOL_CALTYPE_SURVEYPSF%code) .or. &
          (cal_type%code .eq. SRCTOOL_CALTYPE_RMF%code) ) then
        ! we expect to get more than one valid cal file+ext for these cal types, so only throw a genetle warning
        call wrschatty(subroutinename, str_buffer)
      else
        ! give a more detailed warning for other cal types
        ierr = ERROR_LEVEL_WARNING
        call wrserr(subroutinename, ierr, str_buffer)
        do i=1,nret
          write (str_buffer, '(A,I0,A,A,A,I0,A)') &
              "POSSIBLE CALDB CONFLICT: ", i, " '", trim(pfilename(i)), "[",pextno(i), "]'"
          call wrserr(subroutinename, ierr, str_buffer)
        end do
        ierr = ERROR_LEVEL_OK
     end if
    end if

!proper-way!    if (nret .gt. 1) then
!proper-way!      write (str_buffer, '(A,I0,A,A,A,A,A)'), &
!proper-way!          "*Warning* More than one filename+ext (", nret ,") was returned for cal_type='", &
!proper-way!          trim(str_codename), "' for inst='", trim(inst%name), "'"
!proper-way!      if ((cal_type%code .eq. SRCTOOL_CALTYPE_PSF%code) .or. &
!proper-way!          (cal_type%code .eq. SRCTOOL_CALTYPE_CALBADPIX%code) .or. &
!proper-way!          (cal_type%code .eq. SRCTOOL_CALTYPE_RMF%code) ) then
!proper-way!        ! we are happy with getting more than one valid cal file+ext for these cal types
!proper-way!        call wrschatty(subroutinename, str_buffer)
!proper-way!      else
!proper-way!        ! but this is a failure for other cal types
!proper-way!        ierr = ERROR_LEVEL_SEVERE
!proper-way!        call wrserr(subroutinename, ierr, str_buffer)
!proper-way!        do i=1,nret
!proper-way!          write (str_buffer, '(A,I0,A,A,A,I0,A)'), &
!proper-way!              "DEBUG_INFO: ", i, " '", trim(pfilename(i)), "[",pextno(i), "]'"
!proper-way!          call wrschatty(subroutinename, str_buffer)
!proper-way!        end do
!proper-way!        return
!proper-way!      end if
!proper-way!    end if

    if (nret .lt. 1) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, &
          "Could not find a filename+ext for requested cal_type='" // &
          trim(str_codename) // "' for inst='" // trim(inst%name) // "'" // "' filter='" //trim(str_filter) // &
          "' expression='" // trim(str_expression) // "'" )
      return
    end if


!    ! also append the fits extension
!    write (str_buffer, '(I0)'), pextno(1)

!    phduinfo%filename  = pfilename(1)
!    phduinfo%hdu_index = pextno(1)
!    filename = pfilename(1)
    ! For PSF descriptions we will read the file structure later....TODO do this properly
    if ((cal_type%code .eq. SRCTOOL_CALTYPE_PSF%code) .or. &
        (cal_type%code .eq. SRCTOOL_CALTYPE_SURVEYPSF%code) ) then
      write (filename, '(A)')  trim(pfilename(1))
    else
      write (filename, '(A,A,I0)')  trim(pfilename(1)), "+",  pextno(1)
    end if

  end subroutine srctool_cal_get_filename_from_caldb

  ! adapted from /data12s/erosita/sw/fits_090304/erosita/task/expmap/expmap_io_mod.f90
  !----------------------------------------------------------------------------
  subroutine srctool_cal_prep_gtcalf(date_obs, date_end, start_date, start_time, stop_date, stop_time, chatter, ierr)
    character(len=*), parameter :: subroutinename='srctool_cal_prep_gtcalf'

    character(len=*), intent(in) :: date_obs, date_end
    character(len=*), intent(out) :: start_date, start_time, stop_date, stop_time
    integer(kind=int32), intent(out) :: chatter
    integer(kind=int32), intent(out)  :: ierr
    character(len=6) :: cenv
!    character(len=STR_MAX) :: str_buffer
    ierr = ERROR_LEVEL_OK

!    write(str_buffer, '(A,A,A,I0,A)'), "Extracting start_date,start_time from: '", &
!        trim(date_obs), "' len=(", len_trim(date_obs), ")"
!    call wrschatty(subroutinename, str_buffer)
!    write(str_buffer, '(A,A,A,I0,A)'), "Extracting end_date,end_time     from: '", &
!        trim(date_end), "' len=(", len_trim(date_end), ")"
!    call wrschatty(subroutinename, str_buffer)

    if ((len(date_obs) .lt. 19) .or. &
        (len(date_end) .lt. 19) ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr,  "Illegal date-obs/date-end string(s)")
    end if

    if ((len(start_date) .lt. 10) .or. &
        (len(start_time) .lt. 10) .or. &
        (len(stop_date)  .lt. 10) .or. &
        (len(stop_time)  .lt. 10) ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr,  "Return string(s) are too short")
      return
    end if

    ! stdheader times -> format expected by caldb
    start_date(:) = date_obs(1:10)
    stop_date(:)  = date_end(1:10)
    start_time(:) = date_obs(12:19)
    stop_time(:)  = date_end(12:19)

    ! SASS verbosity -> caldb verbosity
    call get_environment_variable("SASS_VERBOSITY", cenv)
    if (cenv .eq. "sparse") then
      chatter = 0
    else if (cenv .eq. "chatty") then
!      chatter = 5
      chatter = 3
    else
      chatter = 0
!      chatter = 2
    end if

  end subroutine srctool_cal_prep_gtcalf


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !----------------------- END of CAL/CALDB handling -----------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------START of GENERIC CAL file handling  ------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! read in a set of cal descriptions
  subroutine srctool_cal_read_cal_from_files(instlist, filenames, num_filenames, cal_type, ierr)
    character(len=*), parameter :: subroutinename='srctool_cal_read_cal_from_files'
    type(instlist_type), intent(inout), target  :: instlist
    character(len=*),    intent(in), dimension(:) :: filenames
    integer(kind=int32), intent(in)      :: num_filenames
    type(srctool_cal_type),intent(in)   :: cal_type
    integer(kind=int32), intent(out)     :: ierr
    type(inst_type), pointer  :: pinst
    integer(kind=int32)        :: i,j
    character(len=LSTR_MAX) :: str_filename
!    logical                  :: none_is_set

    ierr = ERROR_LEVEL_OK

    str_filename = ""

    if ( num_filenames .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, &
          "Problem reading calibration data from file (no filenames supplied)")
      return
    end if

    ! If there is exactly one "CALDB" entry then we need to read a CALDB entry for each eROSITA instrument
    if ( (num_filenames .eq. 1) .and. (trim(filenames(1)) .eq. "CALDB") ) then
      do i=1,MAX_INST
        pinst => instlist%inst(i)
        if ( instlist%inst(i)%to_consider .and. &
            (instlist%inst(i)%evtlistlist%num_evtlists .ge. 1) ) then

          if ((pinst%evtlistlist%hduinfo%key_date_obs%flag .eq. FITSWRAP_KEY_FLAG_BLANK) .or. &
              (pinst%evtlistlist%hduinfo%key_date_end%flag .eq. FITSWRAP_KEY_FLAG_BLANK) ) then
            ierr = ERROR_LEVEL_SEVERE
            call wrserr (subroutinename, ierr, "Problem with undefined DATE-OBS/DATE-END keywords")
            return
          endif


          call srctool_cal_get_filename_from_caldb(pinst, cal_type, &
              pinst%evtlistlist%hduinfo%key_date_obs%value, &
              pinst%evtlistlist%hduinfo%key_date_end%value, &
              str_filename, ierr)
          if ( ierr .ne. ERROR_LEVEL_OK ) then
!            call wrserr (subroutinename, ierr, "Problem getting filename from CALDB")
            call wrserr (subroutinename, ierr, "Problem getting (individual) filename from CALDB: " &
            // "inst="      // trim(pinst%name) &
            // " cal-type=" // trim(cal_type%name) &
            // " date-obs=" // trim(pinst%evtlistlist%hduinfo%key_date_obs%value) &
            // " date-end=" // trim(pinst%evtlistlist%hduinfo%key_date_end%value) &
            )
            return
          end if


          ! ignore blank filenames and 'NONE'
          if ((len_trim(str_filename) .gt. 0) .or. &
              (trim(str_filename) .eq. 'NONE') ) then
            call srctool_cal_read_cal_from_file(instlist, str_filename, cal_type, ierr)
            if ( ierr .ne. ERROR_LEVEL_OK ) then
              call wrserr (subroutinename, ierr, "Problem reading calibration data from file")
              return
            end if
          end if
        end if
      end do
    else

      ! now we need to work out a default instrument to choose when reading in generic cal data
      ! use the first to_consider instrument that has eventlists associated to it.
      j = -1
      do i=1,MAX_INST
        if ( instlist%inst(i)%to_consider .and. &
            (instlist%inst(i)%evtlistlist%num_evtlists .ge. 1) ) then
          j = i
          exit
        end if
      end do
      if ( j .eq. -1 ) then
        !      call wrserr (subroutinename, ierr, "Problem finding valid DATE-OBS/DATE-END keywords")
        call wrserr (subroutinename, ierr, "Couldn't find any valid event lists")
        ierr = ERROR_LEVEL_SEVERE
        return
      end if



      ! If there is exactly one filename or a single "CALDB" then this must apply to all eROSITA instruments
      ! if not then we expect one entry per instrument, in order, with blanks allowed
      do i=1,num_filenames
        str_filename = ""
        if ( trim(filenames(i)) .eq. "CALDB" ) then
!          if ( num_filenames .eq. 1 ) then
!            pinst => instlist%inst(j)
!          else
            pinst => instlist%inst(i)
!          end if

          if ( pinst%to_consider .and. (pinst%evtlistlist%num_evtlists .ge. 1 ) ) then

!            call fitswrap_print_hduinfo(pinst%evtlistlist%hduinfo)  ! DEBUG


            if ((pinst%evtlistlist%hduinfo%key_date_obs%flag .eq. FITSWRAP_KEY_FLAG_BLANK) .or. &
                (pinst%evtlistlist%hduinfo%key_date_end%flag .eq. FITSWRAP_KEY_FLAG_BLANK) ) then
              ierr = ERROR_LEVEL_SEVERE
              call wrserr (subroutinename, ierr, "Problem with undefined DATE-OBS/DATE-END keywords")
              return
            endif

            call srctool_cal_get_filename_from_caldb(pinst, cal_type, &
                pinst%evtlistlist%hduinfo%key_date_obs%value, &
                pinst%evtlistlist%hduinfo%key_date_end%value, &
                str_filename, ierr)
            if ( ierr .ne. ERROR_LEVEL_OK ) then
!              call wrserr (subroutinename, ierr, "Problem reading (shared) filename from CALDB")
              call wrserr (subroutinename, ierr, "Problem getting (shared) filename from CALDB: " &
                  // "inst="      // trim(pinst%name) &
                  // " cal-type=" // trim(cal_type%name) &
                  // " date-obs=" // trim(pinst%evtlistlist%hduinfo%key_date_obs%value) &
                  // " date-end=" // trim(pinst%evtlistlist%hduinfo%key_date_end%value) &
                  )
             return
            end if
          end if

        else
          str_filename = trim(filenames(i))
        end if


!        ! DEBUG : TODO - remove!!
!        if ( cal_type%code .eq. SRCTOOL_CALTYPE_DETMAP%code ) then
!          if ( i .eq. 1 ) then
!            str_filename = "./test_detmap.fits"
!          end if
!        end if

        ! ignore blank filenames and 'NONE'
        if ( (len_trim(str_filename) .gt. 0) .or. (trim(str_filename) .eq. 'NONE') ) then
        ! ignore blank filenames
!       if ( len_trim(str_filename) .gt. 0 ) then
          call srctool_cal_read_cal_from_file(instlist, str_filename, cal_type, ierr)
          if ( ierr .ne. ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Problem reading calibration data from file")
            return
          end if
        end if
      end do
    end if

    ! Some special extra treatment is required for PSF files:
    ! Make an index of the psf maps that have been read in
    if ( cal_type%code .eq. SRCTOOL_CALTYPE_PSF%code) then
      call srctool_psf_index_psfs(instlist, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem indexing PSFs")
        return
      end if
    end if

    ! for the merged instrument we need to take a copy of the EBOUNDS cal from the first valid instrument
    if ( cal_type%code .eq. SRCTOOL_CALTYPE_EBOUNDS%code) then
     do i=1,MAX_INST
       if ((instlist%inst(i)%is_valid          ) .and. &
           (instlist%inst(i)%ebounds%is_valid  )) then
         call srctool_ebounds_copy_ebounds(instlist%inst(i)%ebounds, instlist%inst(0)%ebounds, ierr)
         if ( ierr .ne. ERROR_LEVEL_OK ) then
           call wrserr (subroutinename, ierr, "Problem duplicating EBOUNDS for merged instrument")
           return
         end if
         exit ! from the do loop
       endif
     end do
   end if

   ! Merge the RMFs over the patterns
   if ( cal_type%code .eq. SRCTOOL_CALTYPE_RMF%code) then
     do i=1,MAX_INST
       pinst => instlist%inst(i)
       if ( pinst%to_consider .and. (pinst%evtlistlist%num_evtlists .ge. 1 ) ) then
         call srctool_rmf_merge_rmfs(pinst%refrmfset, pinst, &
             pinst%evtlistlist%evtlist(1)%key_pat_sel%value, ierr)
         if ( ierr .ne. ERROR_LEVEL_OK ) then
           call wrserr (subroutinename, ierr, "Problem merging RMFs over patterns")
           return
         end if
       end if
     end do
   end if


  end subroutine srctool_cal_read_cal_from_files
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !read all matching calibration extensions from the named file into the relevant structures
  subroutine srctool_cal_read_cal_from_file(instlist, filename, cal_type, ierr)
    character(len=*), parameter :: subroutinename='srctool_cal_read_cal_from_file'

    type(instlist_type), intent(inout), target :: instlist
    character(len=*),    intent(in)            :: filename
    type(srctool_cal_type),intent(in)          :: cal_type
    integer(kind=int32), intent(out)           :: ierr

    integer(kind=int32) :: hdu_type
    integer(kind=int32) :: num_hdus
    integer(kind=int32) :: status
    integer(kind=int32) :: unit

!    character(len=LSTR_MAX) :: str_buffer
    integer(kind=int32) :: hdu
    integer(kind=int32) :: nok
    integer(kind=int32) :: evtlist_index
    type(hduinfo_type)  :: hduinfo   ! local hdu info store

    logical                  :: have_inst_index
    logical                  :: hdu_is_valid
    integer(kind=int32)      :: inst_index
    type(inst_type), pointer :: pinst
    type(evtlist_type), pointer :: pevtlist
    status = 0
    ierr = ERROR_LEVEL_OK
    nok = 0


    call wrschatty (subroutinename, &
        "Reading cal/meta data (" // trim(cal_type%name) // ") from file: " // trim(filename))

    ! get an i/o unit
    call FTGIOU_w(unit, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if


    ! open the file and move to whichever extension is specified in filename
    call FTDOPN_w(unit, trim(filename), FITSIO_READONLY, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

!!    write (str_buffer, '(A,A)'), "Successfully opened calibration file: ", trim(filename)
!!    call wrschatty (subroutinename, str_buffer)
!    call wrschatty (subroutinename, "Successfully opened calibration file: " // trim(filename))


    !-------------------------------------------------------------------------
    ! We now search the file looking for HDUs with potentially valid EXTNAMEs
    !
    ! The following cal files will be given with fixed EXTNAME and no suffix:
    ! SPECRESP (ARF), EBOUNDS, VIGNET
    !   - INST index is read from the INSTRUME keyword
    ! The following cal types will be given with fixed EXTNAME format including INST suffix:
    ! GTIn, BADPIXn, CORRATTn, DEADCORn
    !   - INST index is derived from the 'n' but is checked against the INSTRUME keyword
    ! The following are given with free format EXTNAMES:
    ! 'XXXXeVYYarcmi' (PSFDESC) where XXX is energy in eV and YY is offaxis angle in arcmins
    !   - INST index is read from the INSTRUME keyword
    !-------------------------------------------------------------------------

    ! get the total number of HDUs in file
    call FTTHDU_w(unit, num_hdus, status)

    ! now go through all HDUs - starting with first image extension
    do hdu=1,num_hdus
      call fitswrap_init_hduinfo(hduinfo)

!      call fitswrap_print_hduinfo(hduinfo)

      ! move to the specified HDU
!      print *, "Moving to HDU index=", hdu

      call FTMAHD_w(unit, hdu, hdu_type, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      ! get the standard hdu info/keys
      call fitswrap_read_hduinfo_quick(unit, hduinfo, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem reading HDU info")
        return
      end if

      hdu_is_valid = .false.
      inst_index = -1

!      call fitswrap_print_hduinfo(hduinfo)


!      ! SUPERDEBUG stuff
!      write(str_buffer,'(A,A,A,A,1x,A,1x,A,A,I2,A,I2,1x,A,A,A,A)') &
!          "SUPERDEBUG Looking for ", &
!          trim(cal_type%extroot), "(", &
!          trim(logical2char(cal_type%allow_exact)), &
!          trim(logical2char(cal_type%allow_suffix)), &
!          trim(logical2char(cal_type%allow_free)), &
!          ") in HDU ", hdu, "/", &
!          num_hdus, trim(hduinfo%key_extname%name), &
!          "='", trim(hduinfo%key_extname%value),"'"
!      call wrschatty (subroutinename, str_buffer)

      have_inst_index = .false.

      ! check that the EXTNAME keyword has been supplied
      if ( hduinfo%key_extname%flag .ne. FITSWRAP_KEY_FLAG_SUPPLIED ) then
        ! special case for DETMAP caltype
        if ( cal_type%code .eq. SRCTOOL_CALTYPE_DETMAP%code ) then
          call srctool_cal_instrume_to_index(hduinfo, inst_index, ierr)
          if ( ierr .ne. ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Problem reading instrument index from INSTRUME keyword")
            return
          end if
          hdu_is_valid = .true.
          have_inst_index = .true.
          ! call wrschatty (subroutinename, &
          !         "DEBUG: Found valid HDU (INSTRUME keyword) hdu_index=" // trim(i2s(hdu)) // &
          !         " inst_index=" // trim(i2s(inst_index)) // &
          !         " extname='" // trim(hduinfo%key_extname%value) // "'" )

        else
          cycle   ! skip to end of do loop
!          ierr = ERROR_LEVEL_SEVERE
!          call wrserr (subroutinename, ierr, "Problem reading extname from CHDU")
!          return
        end if
      end if

!      if ( hduinfo%key_extname%flag .ne. FITSWRAP_KEY_FLAG_SUPPLIED .and. &
!          hdu .gt. 1 ) then
!      ! check to see if the HDU has an EXTNAME set - normally a problem except for get-around for detmaps
!!        print *, "Skipping HDU"
!        ! allow there to be no extname for DETMAPs - DEBUG TODO
!        if ( cal_type%code .eq. SRCTOOL_CALTYPE_DETMAP%code ) then
!          inst_index = 0
!          if ( ierr .ne. ERROR_LEVEL_OK ) then
!            call wrserr (subroutinename, ierr, "Problem reading instrument index from INSTRUME keyword")
!            return
!          end if
!          hdu_is_valid = .true.
!          write(str_buffer,'(A,I0,A,I0)') &
          ! DEBUG output, remove
!              "DEBUG: Found valid HDU (allow_unset) hdu_index=", hdu, " inst_index=", inst_index
!          call wrschatty (subroutinename, str_buffer)
!        else
!          cycle   ! skip to end of do loop
!        end if
!        ierr = ERROR_LEVEL_SEVERE
!        call wrserr (subroutinename, ierr, "Problem reading extname from CHDU")
!        return
!      end if


      ! see if the EXTNAME keyword exactly matches the required pattern
      ! in which case refer to INSTRUME keyword
      if ( .not. have_inst_index ) then
        if ( cal_type%allow_exact  ) then
          if ( trim(cal_type%extroot) .eq. trim(hduinfo%key_extname%value)) then
            call srctool_cal_instrume_to_index(hduinfo, inst_index, ierr)
            if ( ierr .ne. ERROR_LEVEL_OK ) then
              call wrserr (subroutinename, ierr, "Problem reading instrument index from INSTRUME keyword")
              return
            end if
            hdu_is_valid = .true.
            have_inst_index = .true.
            ! call wrschatty (subroutinename, &
            !       "DEBUG: Found valid HDU (allow_exact) hdu_index=" // trim(i2s(hdu)) // &
            !       " inst_index=" // trim(i2s(inst_index)) // &
            !       " extname='" // trim(hduinfo%key_extname%value) // "'" )

          end if
        end if
      end if

      !  see if the EXTNAME keyword matches extroot with an inst suffix
      if ( .not. have_inst_index ) then
        if ( .not. hdu_is_valid ) then
          if ( cal_type%allow_suffix  ) then
            call srctool_cal_convert_keyval_to_index(hduinfo%key_extname%value, cal_type%extroot, inst_index, ierr)
            if ( ierr .ne. ERROR_LEVEL_OK ) then
              call wrserr (subroutinename, ierr, "Problem reading instrument index from EXTNAME keyword")
              return
            end if
            if ((inst_index .ge. 0) .and. &
                (inst_index .le. MAX_INST)) then
              hdu_is_valid = .true.
              have_inst_index = .true.
              ! call wrschatty (subroutinename, &
              !     "DEBUG: Found valid HDU (allow_suffix) hdu_index=" // trim(i2s(hdu)) // &
              !     " inst_index=" // trim(i2s(inst_index)) // &
              !     " extname='" // trim(hduinfo%key_extname%value) // "'" )
            end if
          end if
        end if
      end if

      !  see if the EXTNAME keyword matches the free extroot format
      if ( .not. have_inst_index ) then
        if ( .not. hdu_is_valid ) then
          if ( cal_type%allow_free  ) then
            if ( index(trim(hduinfo%key_extname%value),trim(cal_type%extroot)) .gt. 0  ) then
              call srctool_cal_instrume_to_index(hduinfo, inst_index, ierr)
              if ( ierr .ne. ERROR_LEVEL_OK ) then
                call wrserr (subroutinename, ierr, "Problem reading Instrument index from INSTRUME keyword")
                return
              end if
              hdu_is_valid = .true.
              have_inst_index = .true.
              ! call wrschatty (subroutinename, &
              !     "DEBUG: Found valid HDU (allow_free) hdu_index=" // trim(i2s(hdu)) // &
              !     " inst_index=" // trim(i2s(inst_index)) // &
              !     " extname='" // trim(hduinfo%key_extname%value) // "'" )
            end if
          end if
        end if
      end if

      if ((hdu_is_valid ) .and. &
          (inst_index .ge. 0)         .and. &
          (inst_index .le. MAX_INST) ) then
        pinst => instlist%inst(inst_index)
        evtlist_index = pinst%evtlistlist%num_evtlists
        if ( pinst%to_consider .or. (inst_index .eq. 0) ) then
          ! we successfully found a matching extension!

          ! assume that these cal/meta data files apply to the most recently read in
          ! events list for each instrument - this assumption should be made more robust TODO
          ! the following is only needed for meta data within an eventslist file

!          print *, "Extname=", trim(hduinfo%key_extname%value), "  => inst_index=", inst_index

          if (cal_type%code .eq. SRCTOOL_CALTYPE_GTI%code      .or. &
              cal_type%code .eq. SRCTOOL_CALTYPE_QUALGTI%code  .or. &
              cal_type%code .eq. SRCTOOL_CALTYPE_FLAREGTI%code .or. &
              cal_type%code .eq. SRCTOOL_CALTYPE_BADPIX%code   .or. &
              cal_type%code .eq. SRCTOOL_CALTYPE_CORRATT%code  .or. &
              cal_type%code .eq. SRCTOOL_CALTYPE_DEADCOR%code ) then
            if ( evtlist_index .gt. 0 ) then
              pevtlist => pinst%evtlistlist%evtlist(evtlist_index)
            else
              ! there are no events for this instrument in this evtlist
              ! so we do not need to read any calibration data for this instrument
              call wrserr (subroutinename, ERROR_LEVEL_WARNING, &
                  "Reading CAL for illegal evtlist (CALTYPE: " // trim(i2s(cal_type%code)) // &
                  ") - no data for instrument TM" // trim(i2s(inst_index)) // "?")
              cycle
            end if
          end if

          ! so now choose the appropriate handling routine
          select case (cal_type%code)
          ! these are cal files which are an extension of the evtlist:
          case (SRCTOOL_CALTYPE_GTI%code,SRCTOOL_CALTYPE_QUALGTI%code,SRCTOOL_CALTYPE_FLAREGTI%code )
            call  srctool_gti_read_gtilist_from_chdu     (unit, pevtlist%gtilist, ierr)
          case (SRCTOOL_CALTYPE_BADPIX%code)
            call  srctool_badpix_read_badpix_from_chdu   (unit, pevtlist%badpix, ierr)
          case (SRCTOOL_CALTYPE_CORRATT%code)
            call  srctool_att_read_attlist_from_chdu     (unit, pevtlist%attlist, ierr)
          case (SRCTOOL_CALTYPE_DEADCOR%code)
            call  srctool_deadcor_read_deadcor_from_chdu (unit, pevtlist%deadcor, ierr)
            ! these are stand-alone cal files :
          case (SRCTOOL_CALTYPE_ARF%code)
            call  srctool_arf_read_arf_from_chdu         (unit, pinst%refarf,  ierr)
          case (SRCTOOL_CALTYPE_RMF%code)
            call  srctool_rmf_read_rmf_from_chdu         (unit, pinst%refrmfset,ierr)
          case (SRCTOOL_CALTYPE_EBOUNDS%code)
            call  srctool_ebounds_read_ebounds_from_chdu (unit, pinst%ebounds, ierr)
!          case (SRCTOOL_CALTYPE_VIGN%code)
!            call  srctool_vign_read_vign_from_chdu       (unit, pinst%vign,    ierr)
          case (SRCTOOL_CALTYPE_AVIGN%code)
            call  avign_read_avign_from_chdu             (unit, pinst%avign,   ierr)
          case (SRCTOOL_CALTYPE_PSF%code)
            call  srctool_psf_read_psf_from_chdu         (unit, pinst%psf,     ierr)
          case (SRCTOOL_CALTYPE_SURVEYPSF%code)
            call  srctool_psf_read_psf_from_chdu         (unit, pinst%survey_psf, ierr)
          case (SRCTOOL_CALTYPE_DETMAP%code)
            call  srctool_detmap_read_detmap_from_chdu   (unit, pinst%detmap,  ierr)
          case (SRCTOOL_CALTYPE_INSTPAR%code)
            call  srctool_instpar_read_instpar_from_chdu (unit, pinst%instpar, ierr)
          case (SRCTOOL_CALTYPE_PATTCAL%code)
            call  srctool_pattcal_read_pattcal_from_chdu (unit, pinst%pattcal, ierr)
          case default
            ierr = ERROR_LEVEL_SEVERE
            call wrserr (subroutinename, ierr, "Unknown CALTYPE: " // trim(i2s(cal_type%code)))
            return
          end select

          if (ierr .ne.  ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Error reading calibration data")
            return
          end if
          nok = nok + 1
        end if
      end if
    end do

    if ( nok .ge. 1 ) then
      call wrschatty (subroutinename,  &
          "Successfully read " // trim(i2s(nok)) // " (" // trim(cal_type%name) // ") extension(s)")
    else
      call wrschatty (subroutinename, "I found no valid (" // trim(cal_type%name) //  ") extensions" )
    end if
    call wrschatty (subroutinename, "")

    !tidy up
    !close the file
    call FTCLOS_w(unit, status)
    !release the unit number
    call FTFIOU_w(unit, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    if (((cal_type%code .eq. SRCTOOL_CALTYPE_GTI%code) .or. &
         (cal_type%code .eq. SRCTOOL_CALTYPE_QUALGTI%code) .or. &
         (cal_type%code .eq. SRCTOOL_CALTYPE_FLAREGTI%code)) .and. &
        (nok .le. 0) ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "I couldn't find any valid *GTIn extensions")
      return
    end if

  end subroutine srctool_cal_read_cal_from_file
  !---------------------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  ! write out a summary of all the calibration data that has been read in
  subroutine srctool_cal_report_cal_status(instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_cal_report_cal_status'
    type(instlist_type),intent(in)    :: instlist
    integer(kind=int32), intent(out) :: ierr
    character(len=LSTR_MAX) :: str_buffer
    character(len=STR_MAX) :: str_buffer1
    character(len=STR_MAX) :: str_buffer_i
    character(len=STR_MAX) :: str_buffer_e, str_buffer_g, str_buffer_b, str_buffer_a, str_buffer_d
    character(len=1) :: flags(0:MAX_TOT_INST)
    character(len=1) :: flags_e(0:MAX_TOT_INST), flags_g(0:MAX_TOT_INST), flags_b(0:MAX_TOT_INST)
    character(len=1) :: flags_a(0:MAX_TOT_INST), flags_d(0:MAX_TOT_INST)
    integer(kind=int32) :: i,e, num_valid

    ierr = ERROR_LEVEL_OK

    ! 1st deal with the 1-per-instrument cal data
    call wrschatty(subroutinename, "")
    call wrschatty(subroutinename, &
        "--------------------------------------------------------------------------------")
    write(str_buffer, '(A)') &
        "Here comes a summary of the calibration data sets (T=supplied, F=not supplied)"
    call wrschatty (subroutinename, str_buffer)

    write (str_buffer, '(A15,1x,A12)') "Cal/Meta type", "Instrument"
    call wrschatty (subroutinename, str_buffer)

    do i=0,MAX_TOT_INST
      if ( instlist%inst(i)%to_consider .or. (MERGED_TM_TO_IDX(i)>0) ) then
        write (flags(i),'(I0)') i
      else
        flags(i) = "_"
     end if
    end do
    call convert_array_to_string (flags(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer_i, "")

    write (str_buffer, '(A15,1x,A12)') "Instlist-->", trim(str_buffer_i)
    call wrschatty (subroutinename, str_buffer)

    !count up the REFARFs
    do i=0,MAX_TOT_INST
      if ( instlist%inst(i)%to_consider .or. (MERGED_TM_TO_IDX(i)>0) ) then
        flags(i) = logical2char(instlist%inst(i)%refarf%is_valid)
      else
        flags(i) = "_"
      end if
    end do
    call convert_array_to_string (flags(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer1, "")
    write (str_buffer, '(A15,1x,A12)') "REFARF", trim(str_buffer1)
    call wrschatty(subroutinename, str_buffer)

    !count up the REFRMFs
    do i=0,MAX_TOT_INST
      if ( instlist%inst(i)%to_consider .or. (MERGED_TM_TO_IDX(i)>0) ) then
        flags(i) = logical2char(instlist%inst(i)%refrmfset%rmf(SRCTOOL_PATT_CLASS_M)%is_valid)
      else
        flags(i) = "_"
      end if
    end do
    call convert_array_to_string (flags(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer1, "")
    write (str_buffer, '(A15,1x,A12)') "REFRMFSET", trim(str_buffer1)
    call wrschatty(subroutinename, str_buffer)

    !count up the EBOUNDs
    do i=0,MAX_TOT_INST
      if ( instlist%inst(i)%to_consider .or. (MERGED_TM_TO_IDX(i)>0) ) then
        flags(i) = logical2char(instlist%inst(i)%ebounds%is_valid)
      else
        flags(i) = "_"
      end if
    end do
    call convert_array_to_string (flags(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer1, "")
    write (str_buffer, '(A15,1x,A12)') "EBOUNDS", trim(str_buffer1)
    call wrschatty(subroutinename, str_buffer)

    !count up the AVIGNETs
    do i=0,MAX_TOT_INST
      if ( instlist%inst(i)%to_consider .or. (MERGED_TM_TO_IDX(i)>0) ) then
        flags(i) = logical2char(instlist%inst(i)%avign%is_valid)
      else
        flags(i) = "_"
      end if
    end do
    call convert_array_to_string (flags(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer1, "")
    write (str_buffer, '(A15,1x,A12)') "AVIGNET2", trim(str_buffer1)
    call wrschatty(subroutinename, str_buffer)

    !count up the PSFs
    do i=0,MAX_TOT_INST
      if ( instlist%inst(i)%to_consider .or. (MERGED_TM_TO_IDX(i)>0) ) then
        flags(i) = logical2char(instlist%inst(i)%psf%is_valid)
      else
        flags(i) = "_"
      end if
    end do
    call convert_array_to_string (flags(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer1, "")
    write (str_buffer, '(A15,1x,A12)') "PSF", trim(str_buffer1)
    call wrschatty(subroutinename, str_buffer)

    !count up the DETMAPs
    do i=0,MAX_TOT_INST
      if ( instlist%inst(i)%to_consider .or. (MERGED_TM_TO_IDX(i)>0) ) then
        flags(i) = logical2char(instlist%inst(i)%detmap%is_valid)
      else
        flags(i) = "_"
      end if
    end do
    call convert_array_to_string (flags(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer1, "")
    write (str_buffer, '(A15,1x,A12)') "DETMAP", trim(str_buffer1)
    call wrschatty(subroutinename, str_buffer)


    !count up the INSTPARs
    do i=0,MAX_TOT_INST
      if ( instlist%inst(i)%to_consider .or. (MERGED_TM_TO_IDX(i)>0) ) then
        flags(i) = logical2char(instlist%inst(i)%instpar%is_valid)
      else
        flags(i) = "_"
      end if
    end do
    call convert_array_to_string (flags(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer1, "")
    write (str_buffer, '(A15,1x,A12)') "INSTPAR", trim(str_buffer1)
    call wrschatty(subroutinename, str_buffer)



    ! now report the meta data associated with each eventlist
    ! do this horizontally
    call wrschatty(subroutinename, &
        "--------------------------------------------------------------------------------")
    write (str_buffer, '(A9,A6,5(1x,A12))') "Extension", "--->", &
        "EVENTS","GTIn","BADPIXn","ATTCORRn","DEADCORn"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A9,A6,5(1x,A12))') "Inst-list", "--->", &
        trim(str_buffer_i(2:MAX_INST)), trim(str_buffer_i), trim(str_buffer_i), trim(str_buffer_i), trim(str_buffer_i)
    call wrschatty(subroutinename, str_buffer)


    do e=1,MAX_EVTFILES
      num_valid = 0
      do i=0,MAX_TOT_INST
        ! first write flag if evt file contains events for this instrument
        if ( instlist%inst(i)%to_consider .or. (MERGED_TM_TO_IDX(i)>0) ) then
          if ( i .ge. 1) then
            flags_e(i) = logical2char(instlist%inst(i)%evtlistlist%evtlist(e)%is_valid)
            if ( instlist%inst(i)%evtlistlist%evtlist(e)%is_valid  ) then
              num_valid = num_valid + 1
            end if
          end if
          ! now the gti
          flags_g(i) = logical2char(instlist%inst(i)%evtlistlist%evtlist(e)%gtilist%is_valid)
          ! now the badpix
          flags_b(i) = logical2char(instlist%inst(i)%evtlistlist%evtlist(e)%badpix%is_valid)
          ! now the attlist
          flags_a(i) = logical2char(instlist%inst(i)%evtlistlist%evtlist(e)%attlist%is_valid)
          ! now the deadcorr
          flags_d(i) = logical2char(instlist%inst(i)%evtlistlist%evtlist(e)%deadcor%is_valid)
        else
          flags_e(i) = "_"
          flags_g(i) = "_"
          flags_b(i) = "_"
          flags_a(i) = "_"
          flags_d(i) = "_"
        end if
      end do

      if ( num_valid .ge. 1 ) then
        call convert_array_to_string (flags_e(1:MAX_INST), MAX_INST+0, str_buffer_e, "")
        call convert_array_to_string (flags_g(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer_g, "")
        call convert_array_to_string (flags_b(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer_b, "")
        call convert_array_to_string (flags_a(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer_a, "")
        call convert_array_to_string (flags_d(0:MAX_TOT_INST), MAX_TOT_INST+1, str_buffer_d, "")
        write (str_buffer, '(A8,1x,I6,5(1x,A12))') "Evtlist", e, &
            trim(str_buffer_e),trim(str_buffer_g),trim(str_buffer_b),trim(str_buffer_a),trim(str_buffer_d)
        call wrschatty(subroutinename, str_buffer)
      end if
    end do
    call wrschatty(subroutinename, &
        "--------------------------------------------------------------------------------")

    call srctool_psf_report_psf_info(instlist)
    call srctool_ebounds_report_ebounds_info(instlist)
    call srctool_rmf_report_rmf_info(instlist)
    call wrschatty(subroutinename, &
        "--------------------------------------------------------------------------------")
    call wrschatty(subroutinename, "")


  end subroutine srctool_cal_report_cal_status
  !-----------------------------------------------------------------------------------


  !-----------------------------------------------------------------------------------
  subroutine srctool_cal_convert_keyval_to_index(keyval, keyprefix, inst_index, ierr)
    character(len=*), parameter :: subroutinename='srctool_cal_convert_key_to_index'
    character(len=*),intent(in) :: keyval
    character(len=*),intent(in) :: keyprefix
    integer(kind=int32), intent(out) :: inst_index
    integer(kind=int32), intent(out) :: ierr

    character(len=STR_MAX) :: str_buffer
    character(len=STR_MAX) :: str_format
    integer(kind=int32) :: temp_index

    ierr = ERROR_LEVEL_OK
    temp_index = -999
    ! extract the instrument number from the keyword.
    ! we allow only the input format "PP...PPn", where PP...PP is the keyprefix string
    ! and n is the 1 digit instrument index, between 0 and MAX_INST
!    print *, "Converting keyval='", trim(keyval), "' using keyprefix='", trim(keyprefix), "'"

    ! check that the input string matches the format
    if ( index(trim(keyval),trim(keyprefix)) .eq. 1 ) then
      if ( len_trim(keyval) .eq. (len_trim(keyprefix) + 1) ) then
        write(str_format, '(A,I0,A)') "(T", len_trim(keyprefix) +1, ', I1)'
        read(keyval, str_format) temp_index  ! Extract the instrument number from the keyval
      end if
    end if

    ! do the following if string doesn't match the format
    if ( temp_index .eq. -999 ) then
      inst_index = -1
      return
    end if

    if ((temp_index .ge. 0) .and. &
        (temp_index .le. MAX_INST)) then
      inst_index = temp_index
    else
      inst_index = -1
      ierr = ERROR_LEVEL_SEVERE
      write(str_buffer, '(A,I0,A,A,A,A)') "Illegal instrument index (", temp_index, &
          ") read from keyval='", trim(keyval), "' (keyprefix='", trim(keyprefix), "')"
      call wrserr (subroutinename, ierr, str_buffer)
    end if


  end subroutine srctool_cal_convert_keyval_to_index
  !-----------------------------------------------------------------------------------



  subroutine srctool_cal_instrume_to_index(hduinfo, inst_index, ierr)
    character(len=*), parameter :: subroutinename='srctool_cal_instrume_to_index'
    type(hduinfo_type),intent(in)    :: hduinfo
    integer(kind=int32), intent(out) :: inst_index
    integer(kind=int32) :: temp_inst_index
    integer(kind=int32), intent(out) :: ierr
    character(len=STR_MAX) :: str_buffer
    ierr = ERROR_LEVEL_OK
    if ( len_trim(hduinfo%key_instrume%value) .eq. 0 ) then
      temp_inst_index = 0
    else

      call srctool_cal_convert_keyval_to_index(hduinfo%key_instrume%value, INST_STEM, temp_inst_index, ierr)

      ! handle instrument keywords descirbed in the old way
      if ( (temp_inst_index .lt. 0) .and. (ierr == ERROR_LEVEL_OK)) then
        call srctool_cal_convert_keyval_to_index(hduinfo%key_instrume%value, INST_STEM_OLD, temp_inst_index, ierr)

        ! if successful we should issue a warning
        if ( temp_inst_index .ge. 0 ) then
          call wrserr (subroutinename, ERROR_LEVEL_WARNING, &
              "Warning. Found an instrument labelled with '"// INST_STEM_OLD // &
              "*' (expected '" // INST_STEM // "*') in file: " // trim(hduinfo%filename))
        endif

      endif
    endif
!    print *, "Converted instrume='", trim(hduinfo%key_instrume%value), "' to inst=", temp_inst_index

    if ((temp_inst_index .lt. 0 ) .or. &
        (ierr .ne. ERROR_LEVEL_OK) ) then
      ierr = ERROR_LEVEL_SEVERE
      write(str_buffer,'(A,A,A)') &
          "Unable to extract sensible INST index from INSTRUME keyword (='", trim(hduinfo%key_instrume%value), "')"
      call wrserr(subroutinename, ierr, str_buffer)
    endif

    inst_index = temp_inst_index
  end subroutine srctool_cal_instrume_to_index
  !-----------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------END of GENERIC CAL file handling  ------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

end module srctool_cal
