! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_datatypes

  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of constants and defined types for srctool modules
  !
  !-----------------------------------------------------------------------------

  use iso_fortran_env ! a more standard replacement for numkind
  use iso_c_binding, only: C_PTR
  use ero_general    ! standard eROSITA module handling flag definitions etc
  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_regions   ! generic region handling routines
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_generic_avign     ! generic analytic vignetting file handling routines
  use srctool_extmap

  implicit none     ! insist on explicit typing

  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_DATATYPES_MOD'
  !-------------------------------------------------------------------------------
  public

  !-------------------------------------------------------------------------------
  ! subroutine declarations and interfaces for module: srctool_datatypes
  ! None!

  character(len=*), parameter :: PROCTIME_PREFIX='#PROCESSING_TIME_INFO:'

!  real(kind=real64),   parameter :: VERY_POSITIVE_DOUBLE =  1.0d300
!  real(kind=real64),   parameter :: VERY_NEGATIVE_DOUBLE = -1.0d300
  integer(kind=int32), parameter :: MAX_EVTFILES=12      ! max number of possible inst+filter combinations

  integer(kind=int32), parameter :: MAX_EXTPARS=8        ! max number of params used to describe source extent
  integer(kind=int32), parameter :: MAX_TODO_TASKS=12    ! max number of steps todo
  integer(kind=int32), parameter :: MAX_COORD_DIMS=3     ! max dimensions in which coordinates are expressed
  integer(kind=int32), parameter :: MAX_LCPARS=3         ! max number of params used to describe lc binning
  integer(kind=int32), parameter :: MAX_LCBANDS=10       ! max number of energy bands that can be extracted in one go
  integer(kind=int32), parameter :: MAX_XGRIDPARS=2      ! max number of params used to describe xgrid sampling
  integer(kind=int32), parameter :: MAX_PRODUCTS_PER_SRC=100 ! maximum number of products generated per source

  integer(kind=int32), parameter :: XGRIDPAR_SRC=1       !
  integer(kind=int32), parameter :: XGRIDPAR_BACK=2      !

  real(kind=real64),   parameter :: DEFAULT_E_MIN=0.5d0  ! keV
  real(kind=real64),   parameter :: DEFAULT_E_MAX=10.0d0 !
  real(kind=real64),   parameter :: EVAL_ENERGY_MIN_DELTAE=0.001d0 !keV
  real(kind=real64),   parameter :: MIN_EXPECTED_MEAN_GTI_DURATION = 5.0d0 ! seconds


!  real(kind=real64), dimension(1:12), parameter :: AVIGN_ENERGY_GRID = &
!      (/ 0.25d0, 0.75d0, 1.0d0, 1.5d0, 2.0d0, 2.5d0, 3.0d0, 4.0d0, 5.0d0, 6.0d0, 7.5d0, 8.4d0 /)  !


!  real(kind=real64),   parameter :: BACKGROUND_RESAMP_XGRID_FACTOR = 2.0d0   ! this is the linear ratio by which the
  !background region is sampled w.r.t. the source region sampling - this only affects the tests of whether a given
  !location in ccd coords is on a good pixel or not. Larger numbers mean sparser sampling.

  ! this sets the energy<-->channel scale for the eROSITA detectors
  real(kind=real32), parameter :: INV_ENERGY_PER_CHANNEL_IN_EV = 0.2

  integer(kind=int32), parameter :: OBS_MODE_NULL    = 0  ! not yet determined
  integer(kind=int32), parameter :: OBS_MODE_SURVEY  = 1
  integer(kind=int32), parameter :: OBS_MODE_POINTING = 2
  integer(kind=int32), parameter :: OBS_MODE_SLEW    = 3  ! TBD
  integer(kind=int32), parameter :: OBS_MODE_SCAN    = 4  ! TBD
  character(len=*), parameter :: OBS_MODE_STR(OBS_MODE_NULL:OBS_MODE_SCAN) &
      = (/ "NULL    " , "SURVEY  ", "POINTING", "SLEW    ", "SCAN    " /)


  integer(kind=int32), parameter :: SRCTOOL_PSF_CLASS_NULL      =0
  integer(kind=int32), parameter :: SRCTOOL_PSF_CLASS_NONE      =1
  integer(kind=int32), parameter :: SRCTOOL_PSF_CLASS_MAP       =2
  integer(kind=int32), parameter :: SRCTOOL_PSF_CLASS_ANALYTIC  =3
  integer(kind=int32), parameter :: SRCTOOL_PSF_CLASS_SHAPELETS =4

  character(len=*), parameter :: SRCTOOL_PSF_CLASS_STR(SRCTOOL_PSF_CLASS_NULL:SRCTOOL_PSF_CLASS_SHAPELETS) &
      = (/ "NULL     " , "NONE     ", "MAP      ", "ANALYTIC ", "SHAPELETS" /)

  integer(kind=int32), parameter :: SRCTOOL_BADPIX_CLASS_NULL  =0
  integer(kind=int32), parameter :: SRCTOOL_BADPIX_CLASS_NONE  =1
  integer(kind=int32), parameter :: SRCTOOL_BADPIX_CLASS_IMAGE =2
  integer(kind=int32), parameter :: SRCTOOL_BADPIX_CLASS_TABLE =3

  integer(kind=int32), parameter :: SRCTOOL_VIGN_CLASS_NULL      =0
  integer(kind=int32), parameter :: SRCTOOL_VIGN_CLASS_NONE      =1
  integer(kind=int32), parameter :: SRCTOOL_VIGN_CLASS_GRID_RPHI =2
  integer(kind=int32), parameter :: SRCTOOL_VIGN_CLASS_GRID_XY   =3

  integer(kind=int16), parameter :: SRCTOOL_PATT_CLASS_NULL =-1
  integer(kind=int16), parameter :: SRCTOOL_PATT_CLASS_M    =0 ! merged patterns
  integer(kind=int16), parameter :: SRCTOOL_PATT_CLASS_S    =1 ! singles
  integer(kind=int16), parameter :: SRCTOOL_PATT_CLASS_D    =2 ! doubles
  integer(kind=int16), parameter :: SRCTOOL_PATT_CLASS_T    =3 ! triples
  integer(kind=int16), parameter :: SRCTOOL_PATT_CLASS_Q    =4 ! quadruples
  integer(kind=int16), parameter :: SRCTOOL_PATT_NCLASS     =4

  integer(kind=int16), parameter :: SRCTOOL_PATT_MASK_NULL  =0 !
  integer(kind=int16), parameter :: SRCTOOL_PATT_MASK_S     =1 !
  integer(kind=int16), parameter :: SRCTOOL_PATT_MASK_D     =2 !
  integer(kind=int16), parameter :: SRCTOOL_PATT_MASK_T     =4 !
  integer(kind=int16), parameter :: SRCTOOL_PATT_MASK_Q     =8 !

  character(len=*), parameter :: SRCTOOL_PATT_CLASS_STR(SRCTOOL_PATT_CLASS_NULL:SRCTOOL_PATT_CLASS_Q) &
      = (/ "NULL   " , "MERGED ", "SINGLES", "DOUBLES", "TRIPLES", "QUADS  " /)



  integer(kind=int32), parameter :: TODO_CODE_NULL   =   0    ! these encode the steps that should be done for each target
  integer(kind=int32), parameter :: TODO_CODE_SPEC   =   1    ! they will be OR'ed together
  integer(kind=int32), parameter :: TODO_CODE_ARF    =   2
  integer(kind=int32), parameter :: TODO_CODE_RMF    =   4
  integer(kind=int32), parameter :: TODO_CODE_LC     =   8
  integer(kind=int32), parameter :: TODO_CODE_LCCORR =  16
  integer(kind=int32), parameter :: TODO_CODE_EVENTS =  32
  integer(kind=int32), parameter :: TODO_CODE_SRCGTI =  64
  integer(kind=int32), parameter :: TODO_CODE_ALL    =  &
      TODO_CODE_SPEC + TODO_CODE_ARF + TODO_CODE_RMF + TODO_CODE_LC + &
      TODO_CODE_LCCORR + TODO_CODE_EVENTS + TODO_CODE_SRCGTI
  integer(kind=int32), parameter :: TODO_CODE_SAMPLE = 128   ! debug outputs, not included in "ALL"
  integer(kind=int32), parameter :: TODO_CODE_STACK  = 256   !
  integer(kind=int32), parameter :: TODO_CODE_TEST   = 512   !



  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_SRC_SPEC  = 1
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_BACK_SPEC = 2
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_ARF      = 3
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_LC       = 4
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_SRC_EVT  = 5
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_BACK_EVT = 6
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_SRC_SAMPLELIST = 7
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_BACK_SAMPLELIST = 8
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_IMAGE    = 9
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_RMF      = 10
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_TARGZ    = 11
  integer(kind=int32), parameter :: SRCTOOL_FILETYPE_CODE_TARXZ    = 12

  ! these strings are written into the eSASS standard header keyword "DSTYPE" in the output products
  character(len=*),    parameter :: ESASS_DSTYPE_SRC_SPEC  = "SourceSpec"
  character(len=*),    parameter :: ESASS_DSTYPE_BACK_SPEC = "BackgrSpec"
  character(len=*),    parameter :: ESASS_DSTYPE_LC        = "LightCurve"
  character(len=*),    parameter :: ESASS_DSTYPE_SRC_EVT   = "EventList"
  character(len=*),    parameter :: ESASS_DSTYPE_BACK_EVT  = "BackgrEventList"
  character(len=*),    parameter :: ESASS_DSTYPE_ARF       = "ARF"
  character(len=*),    parameter :: ESASS_DSTYPE_RMF       = "RMF"
  character(len=*),    parameter :: ESASS_DSTYPE_TAR       = "SourceProducts"

  integer(kind=int32), parameter :: ESASS_FILTER_CODE_CLOSED = 0
  integer(kind=int32), parameter :: ESASS_FILTER_CODE_OPEN   = 1
  integer(kind=int32), parameter :: ESASS_FILTER_CODE_FILTER = 2
  integer(kind=int32), parameter :: ESASS_FILTER_CODE_FE55   = 3
  integer(kind=int32), parameter :: ESASS_FILTER_CODE_INDEF  = 4
  integer(kind=int32), parameter :: ESASS_FILTER_CODE_NULL   = 5
  character(len=*),    parameter :: ESASS_FILTER_STR(ESASS_FILTER_CODE_CLOSED:ESASS_FILTER_CODE_NULL) &
      = (/ 'CLOSED' , 'OPEN  ', 'FILTER', 'Fe55  ', 'INDEF ' , 'NULL  ' /)
  character(len=1),    parameter :: ESASS_FILTER_CHAR(ESASS_FILTER_CODE_CLOSED:ESASS_FILTER_CODE_NULL) &
      = (/ '0' , '1', '2', '3', '4', '?' /)

  integer(kind=int32), parameter :: EXTTYPE_CODE_POINT    =0
  integer(kind=int32), parameter :: EXTTYPE_CODE_GAUSSIAN =1
  integer(kind=int32), parameter :: EXTTYPE_CODE_BETA     =2
  integer(kind=int32), parameter :: EXTTYPE_CODE_TOPHAT   =3
  integer(kind=int32), parameter :: EXTTYPE_CODE_MAP      =4
!  integer(kind=int32), parameter :: EXTTYPE_CODE_PHOTON   =5
!  integer(kind=int32), parameter :: EXTTYPE_CODE_CATALOGUE=6
!  character(len=*),    parameter :: EXTTYPE_CODE_STR(EXTTYPE_CODE_POINT:EXTTYPE_CODE_CATALOGUE) &
!      = (/ "POINT    ", "GAUSSIAN " , "BETA     ", "TOPHAT   ", "MAP      ", "PHOTON   ", "CATALOGUE" /)
  character(len=*),    parameter :: EXTTYPE_CODE_STR(EXTTYPE_CODE_POINT:EXTTYPE_CODE_MAP) &
      = (/ "POINT   ", "GAUSSIAN" , "BETA    ", "TOPHAT  ", "MAP     "  /)

  real(kind=real64), parameter :: EXTENDED_SOURCE_DEFAULT_NSIGMAS = 5.0

  integer(kind=int32), parameter :: LCTYPE_CODE_NULL         =0
  integer(kind=int32), parameter :: LCTYPE_CODE_REGULAR      =1   ! strictly regular binning - with all empty bins given
  integer(kind=int32), parameter :: LCTYPE_CODE_REGULAR_PLUS =2   ! strictly regular binning - but with empty bins compacted
  integer(kind=int32), parameter :: LCTYPE_CODE_REGULAR_MINUS=3   ! strictly regular binning - without any empty bins
  integer(kind=int32), parameter :: LCTYPE_CODE_SRCGTI       =4   ! matches output source GTIs - with explicit empty bins
  integer(kind=int32), parameter :: LCTYPE_CODE_SRCGTI_MINUS =5   ! matches output source GTIs - without any empty bins
  integer(kind=int32), parameter :: LCTYPE_CODE_GTI          =6   ! matches input GTI (see also gtitype param)
  integer(kind=int32), parameter :: LCTYPE_CODE_FOLDED       =7

  integer(kind=int8), parameter :: EVT_CLASS_FLAG_NULL =0
  integer(kind=int8), parameter :: EVT_CLASS_FLAG_SRC  =1
  integer(kind=int8), parameter :: EVT_CLASS_FLAG_BACK =2
  character(len=*),   parameter :: EVT_CLASS_FLAG_STR(EVT_CLASS_FLAG_NULL:(EVT_CLASS_FLAG_SRC+EVT_CLASS_FLAG_BACK)) &
      = (/ "NULL", "SRC " , "BACK", "S+B " /)

  integer(kind=int32), parameter :: REG_CLASS_FLAG_SRC=1
  integer(kind=int32), parameter :: REG_CLASS_FLAG_BACK=2
  character(len=*),    parameter :: REG_CLASS_FLAG_STR(REG_CLASS_FLAG_SRC:REG_CLASS_FLAG_BACK) &
      = (/ "SRC " , "BACK" /)

  integer(kind=int32), parameter :: AUTOREG_CODE_NULL=0
  integer(kind=int32), parameter :: AUTOREG_CODE_AUTO1=1
!  integer(kind=int32), parameter :: AUTOREG_CODE_AUTO2=2

  integer(kind=int32), parameter :: SRCLIST_TYPE_CODE_NULL         =0
  integer(kind=int32), parameter :: SRCLIST_TYPE_CODE_INLINE       =1   !
  integer(kind=int32), parameter :: SRCLIST_TYPE_CODE_FITSCAT      =2   !
  integer(kind=int32), parameter :: SRCLIST_TYPE_CODE_ASCIICAT     =3   !
  character(len=*),    parameter :: SRCLIST_TYPE_CODE_STR(SRCLIST_TYPE_CODE_NULL:SRCLIST_TYPE_CODE_ASCIICAT) &
      = (/ "NULL  " , "SINGLE",  "FITS  ", "ASCII "/)


  integer(kind=int16), parameter :: PHA_QUALITY_GOOD=0
  integer(kind=int16), parameter :: PHA_QUALITY_BAD=1
  integer(kind=int16), parameter :: PHA_QUALITY_DUBIOUS=2
  integer(kind=int16), parameter :: PHA_QUALITY_BAD_MANUAL=5

  integer(kind=int16), parameter :: PHA_GROUPING_NEW=1
  integer(kind=int16), parameter :: PHA_GROUPING_CONT=-1
  integer(kind=int16), parameter :: PHA_GROUPING_INDEF=0


!  integer(kind=int32), parameter :: PSFTYPE_CODE_NONE       =0
!  integer(kind=int32), parameter :: PSFTYPE_CODE_2D_PSF     =1
!  integer(kind=int32), parameter :: PSFTYPE_CODE_SHAPELET   =2
!  character(len=*),    parameter :: PSFTYPE_CODE_STR(PSFTYPE_CODE_NONE:PSFTYPE_CODE_SHAPELET) &
!      = (/ "NONE    ", "2D      " , "SHAPELET" /)

  type srctool_cal_type
    integer(kind=int32)    :: code           ! a numerical code
    character(len=STR_MAX) :: name           ! user-friendly name
    character(len=STR_MAX) :: extroot        ! fits extname root string
    logical                :: allow_exact    ! if true then allow extensions exactly matching extroot
    logical                :: allow_suffix   ! if true then can allow extensions with digits 0-7 appended to extroot
    logical                :: allow_free     ! if true then allow any extensions with any part of name matching extroot
  end type srctool_cal_type

  !------------------------------------------------------------------------------------------!
  ! The following cal files have fixed EXTNAME and no suffix:                                !
  ! SPECRESP (ARF), EBOUNDS, VIGNET                                                          !
  !   - INST index is read from the INSTRUME keyword                                         !
  ! The following cal types wil be given with fixed EXTNAME format including INST suffix:    !
  ! GTIn, BADPIXn, CORRATTn, DEADCORn                                                        !
  !   - INST index is derived from the 'n' but is checked against the INSTRUME keyword       !
  ! The following are given with free format EXTNAMES:                                       !
  ! 'XXXXeVYYarcmi' (PSFDESC) where XXX is energy in eV and YY is offaxis angle in arcmins   !
  !   - INST index is read from the INSTRUME keyword                                         !
  ! 'SPECRESP_nPX' where n is the pattern ID (1, 2, 3 or 4)
  !--------------------------------------------------------------------------------------------------------------------------------------------!
  !------------------------------------------------------------------------- code name        extroot    exact?       suffix?      free?       !
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_NULL     =srctool_cal_type( 0,""         ,""        , .true.,.false.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_ARF      =srctool_cal_type( 1,"ARF"      ,"SPECRESP", .true.,.false.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_EBOUNDS  =srctool_cal_type( 2,"EBOUNDS"  ,"EBOUNDS" , .true.,.false.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_AVIGN    =srctool_cal_type( 4,"AVIGN2"   ,"AVIGNET2_COEFF", .true.,.false.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_PSF      =srctool_cal_type( 5,"PSF"      ,"eV"      ,.false.,.false., .true.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_SURVEYPSF=srctool_cal_type( 6,"SURVEYPSF","eV"      ,.false.,.false., .true.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_GTI      =srctool_cal_type( 7,"GTI"      ,"GTI"     ,.false., .true.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_QUALGTI  =srctool_cal_type( 8,"QUALGTI"  ,"QUALGTI" ,.false., .true.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_FLAREGTI =srctool_cal_type( 9,"FLAREGTI" ,"FLAREGTI",.false., .true.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_BADPIX   =srctool_cal_type(10,"BADPIX"   ,"BADPIX"  ,.false., .true.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_CORRATT  =srctool_cal_type(12,"CORRATT"  ,"CORRATT" ,.false., .true.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_DEADCOR  =srctool_cal_type(13,"DEADCOR"  ,"DEADCOR" ,.false., .true.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_DETMAP   =srctool_cal_type(14,"DETMAP"   ,"DETMAP"  ,.false.,.false.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_INSTPAR  =srctool_cal_type(15,"INSTPAR"  ,"INSTPAR" , .true.,.false.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_PATTCAL  =srctool_cal_type(16,"PATTCAL"  ,"PATTCAL" ,.false., .true.,.false.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_RMF      =srctool_cal_type(17,"RMF"      ,"SPECRESP",.false.,.false., .true.)
  type(srctool_cal_type),parameter::SRCTOOL_CALTYPE_SRCGTI   =srctool_cal_type(18,"SRCGTI"   ,"SRCGTI"  ,.false., .true.,.false.)
  !------------------------------------------------------------------------- code name        extroot    exact?       suffix?      free?       !
  !--------------------------------------------------------------------------------------------------------------------------------------------!



  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  ! Now a list of derived types - ie data structures


  !-------------------------------------------------------------------------------
  ! input parameter structure datatype
  ! we should expect lists of input files - one for each telescope/camera
  type :: srctool_paramlist_type

    !the input params handle
    integer(kind=int32)    :: unit
    ! variables to receive and control the input parameters
    character(len=STR_MAX) :: evtfiles(MAX_EVTFILES)
    integer(kind=int32)    :: num_evtfiles ! number of evtfile filenames that were supplied

    character(len=STR_MAX) :: prefix    ! stem for the output files
    character(len=STR_MAX) :: suffix    ! tail for the output files

    character(len=STR_MAX) :: srccoord    ! ... free format

    character(len=STR_MAX) :: todo(MAX_TODO_TASKS) ! Which operations to carry out (SPEC|LC|EVENT|ALL)
    integer(kind=int32)    :: num_todo             ! number of instruments that were supplied
    integer(kind=int32)    :: todo_code            ! a numerical version of the todo list

    integer(kind=int32)    :: insts(MAX_INST) ! the list of instruments to consider
    integer(kind=int32)    :: num_insts           ! number of instruments that were supplied

    integer(kind=int32)    :: writeinsts(NUM_TOT_INST) ! instruments to write
    integer(kind=int32)    :: num_writeinsts ! number to write

!    character(len=STR_MAX) :: refarfs(MAX_INST)
!    integer(kind=int32)    :: num_refarfs  ! number of refarf filenames that were supplied

!    character(len=STR_MAX) :: ebounds(MAX_INST)
!    integer(kind=int32)    :: num_ebounds  ! number of ebounds filenames that were supplied

!    character(len=STR_MAX) :: vigndesc(MAX_INST)
!    integer(kind=int32)    :: num_vigndesc ! number of vigndesc filenames that were supplied

!    character(len=STR_MAX) :: psfdesc(MAX_INST)
!    integer(kind=int32)    :: num_psfdesc  ! number of psfdesc filenames that were supplied


!    character(len=STR_MAX) :: badpix(MAX_INST)
!    integer(kind=int32)    :: num_badpix   ! number of additional badpix filenames that were supplied

    character(len=STR_MAX) :: srcreg
    character(len=STR_MAX) :: backreg

!    character(len=STR_MAX) :: extent       ! string describing the source shape
    character(len=STR_MAX) :: exttype      ! string describing the source shape
                                           ! one of POINT|GAUSSIAN|BETA|TOPHAT|PHOTON
!    integer(kind=int32)    :: extent_code  ! a numerical version of the above
    integer(kind=int32)    :: exttype_code  ! a numerical version of the above

    real(kind=real64)      :: extpars(MAX_EXTPARS)
    integer(kind=int32)    :: num_extpars  ! number of extent params that were supplied

    character(len=STR_MAX) :: extmap       ! string describing the extent map

    character(len=STR_MAX) :: lctype       ! string describing how the LC should be constructed
                                           ! one of REGULAR|GTI|ADAPTIVE ....
    integer(kind=int32)    :: lctype_code  ! a numerical version of the above

    real(kind=real64)      :: lcpars(MAX_LCPARS) ! parameters used to control binning of the LC
    integer(kind=int32)    :: num_lcpars    ! number of lcpars that were supplied

    real(kind=real64)      :: lcemin(MAX_LCBANDS) !  Light curve energy ranges (Emin keV)
    integer(kind=int32)    :: num_lcemin    ! number of lc emin params that were supplied

    real(kind=real64)      :: lcemax(MAX_LCBANDS) !  Light curve energy ranges (Emin keV)
    integer(kind=int32)    :: num_lcemax    ! number of lc emin params that were supplied

    real(kind=real64)      :: lcgamma       ! Photon index to use when weighting over lc energy ranges


    real(kind=real64)      :: tstep         ! the time steps (in s) over which sensitivity calculations are made
    real(kind=real64)      :: xgrid(2)      ! multipliers giving the spatial steps (in multiples of physical plixel size)
                                            ! over which sensitivity calculations are made
                                            ! First array element is for src region, second for background region
    integer(kind=int32)    :: num_xgrid     ! number of xgrid params that were supplied

    character(len=STR_MAX) :: gtitype       ! String describing which GTI type to use

    character(len=STR_MAX) :: psftype        ! String describing how to model PSF corrections (NONE,2D,SHAPELET)
    integer(kind=int32)    :: psftype_code   ! a numerical version of the psftype

!!    character(len=STR_MAX) :: str_pat_sel   ! string giving mask of which patterns to use

    character(len=STR_MAX) :: str_flagsel   ! string giving mask of which flags to reject
    character(len=STR_MAX) :: str_pat_sel   ! string giving mask of patterns to reject

    character(len=STR_MAX) :: out_srcreg    ! output regions for sources
    character(len=STR_MAX) :: out_backreg   ! output regions for backgrounds

    logical                :: continue_on_error ! keep processing if error occurs

    character(len=STR_MAX) :: tarball       ! parameter no, tar.gz or tar.xz
    logical                :: targz         ! interpreted from above: make tar.gz files
    logical                :: tarxz         ! interpreted from above: make tar.xz files
    logical                :: tarany        ! any tar file
    character(len=STR_MAX) :: tartmpdir     ! location of temporary files when making a tar output

    logical                :: clobber       ! Whether to clobber output files

  end type srctool_paramlist_type
  !-------------------------------------------------------------------------------


  ! only used in srctool_calc_fracrsp
  type :: current_gti_type
    logical              :: is_valid
    real   (kind=real64) :: tstart
    real   (kind=real64) :: tstop
  end type current_gti_type


  !-------------------------------------------------------------------------------
  ! structure to hold a list of gtis
  ! based on:
  ! http://www2011.mpe.mpg.de/erosita/internal/SASS-devel/eROSITA_TRoPIC_Eventfiles-4eng.html
  type :: gtilist_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    ! add in any extra header keywords here
    type(fitskey_real64_type)  :: key_telapse
    type(fitskey_real64_type)  :: key_ontime
    type(fitskey_real64_type)  :: key_livetime
    type(fitskey_real64_type)  :: key_exposure

!    Moved to std header
!    ! The following key combination is used for merged instrument output
!    ! for single instruments, do not print out key_ninst or key_instrum
!    type(fitskey_int32_type)  :: key_ninst
!    type(fitskey_string_type), dimension(1:MAX_INST)  :: pkey_instrum

    !arrays to hold the event properties - allocated at run time
    integer(kind=int32) :: num_rows
    real   (kind=real64), dimension(:), allocatable :: pstart
    real   (kind=real64), dimension(:), allocatable :: pstop

  end type gtilist_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  type :: badpix_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    integer(kind=int32) :: class

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    !this holds the (derived) map
    !values in the map give the index of the badpixel in the rows below
    !zeros indicate that this is a good pix for all energies and times
    integer(kind=int32) :: naxis1, naxis2
    integer(kind=int32), dimension(:,:), allocatable :: pmap

    !this holds the table based description
    integer(kind=int32) :: num_rows
    integer(kind=int16), dimension(:), allocatable :: prawx
    integer(kind=int16), dimension(:), allocatable :: prawy
    integer(kind=int16), dimension(:), allocatable :: pyextent
    integer(kind=int16), dimension(:), allocatable :: ptype
    integer(kind=int16), dimension(:), allocatable :: pbadflag
    real(kind=real64),   dimension(:), allocatable :: ptimemin
    real(kind=real64),   dimension(:), allocatable :: ptimemax
    integer(kind=int16), dimension(:), allocatable :: pphamin
    integer(kind=int16), dimension(:), allocatable :: pphamax

    integer(kind=int32) :: num_badpix

    ! whether to filter neighbouring pixels around bad pixels and edges
    logical :: filter_for_patterns

  end type badpix_type
  !-------------------------------------------------------------------------------

  type pbadpix_type
    type(badpix_type), pointer :: p
  end type pbadpix_type



  !-------------------------------------------------------------------------------
  ! structure to hold a list of telescope attitude information
  ! based on:
  ! http://www2011.mpe.mpg.de/erosita/internal/SASS-devel/eROSITA_TRoPIC_Eventfiles-4eng.html
  type :: attlist_type
    ! flag used to show if this structure is ready to be used
    logical :: is_valid

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    !arrays to hold the attitude entries - allocated at run time
    integer(kind=int32) :: num_rows
    real   (kind=real64), dimension(:), allocatable :: ptime
    real   (kind=real64), dimension(:), allocatable :: pra
    real   (kind=real64), dimension(:), allocatable :: pdec
    real   (kind=real64), dimension(:), allocatable :: proll

    ! derived parameters
    ! great circle distance travelled and bearing since previous entry (deg)
    real   (kind=real64), dimension(:), allocatable :: pdist
    real   (kind=real64), dimension(:), allocatable :: pcosdist
    real   (kind=real64), dimension(:), allocatable :: psindist
    real   (kind=real64), dimension(:), allocatable :: pbearing
    real   (kind=real64), dimension(:), allocatable :: pcosbearing
    real   (kind=real64), dimension(:), allocatable :: psinbearing
    real   (kind=real64), dimension(:), allocatable :: pcosdec
    real   (kind=real64), dimension(:), allocatable :: psindec

    ! statistics of the attitude - used mainly for pointing mode obs
    ! all in units of degrees,
    real   (kind=real64) :: mean_ra
    real   (kind=real64) :: mean_dec
    real   (kind=real64) :: mean_roll
!    real   (kind=real64) :: stdev_ra
!    real   (kind=real64) :: stdev_dec
!    real   (kind=real64) :: stdev_roll
    real   (kind=real64) :: stdev   ! vs the mean position above

  end type attlist_type
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! this type is used to hold an arf - either input or output
  ! see http://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002/cal_gen_92_002.html
  type :: arf_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid
    logical  :: to_consider

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    ! arf specific keywords
    type(fitskey_string_type) :: key_phafile

    type(fitskey_real64_type) :: key_telapse
    type(fitskey_real64_type) :: key_ontime
    type(fitskey_real64_type) :: key_livetime
    type(fitskey_real64_type) :: key_exposure

!    ! The following key combination is used for merged instrument output
!    ! for single instruments, do not print out key_ninst or key_instrum
!    type(fitskey_int32_type)  :: key_ninst
!    type(fitskey_string_type), dimension(1:MAX_INST)  :: pkey_instrum

    !arrays to hold the ARF entries - allocated at run time
    integer(kind=int32) :: num_rows
    real   (kind=real32), dimension(:), allocatable :: pelo        ! ENERG_LO column
    real   (kind=real32), dimension(:), allocatable :: pehi        ! ENERG_HI
    real   (kind=real64), dimension(:), allocatable :: presp       ! SPECRESP
    real   (kind=real64), dimension(:), allocatable :: pcorrcomb   ! CORRCOMB
    real   (kind=real64), dimension(:), allocatable :: pcorrpsf    ! CORRPSF
    real   (kind=real64), dimension(:), allocatable :: pcorrvign   ! CORRVIGN

!    real(kind=real64)  :: backratio  ! the average src/background area ratio
    real(kind=real64)  :: backscale_src   ! the average src area  (units deg^2)
    real(kind=real64)  :: backscale_back  ! the average background area   (units deg^2)

    real(kind=real64)  :: regarea_src, regarea_back ! region area of source+back (units deg^2)
    real(kind=real64)  :: regmodarea_src, regmodarea_back ! region+model areas
  end type arf_type
  !-------------------------------------------------------------------------------

  type :: arflist_type
    integer(kind=int32) :: num_arfs
    type(arf_type) :: arf(0:MAX_TOT_INST)
  end type arflist_type

  !-------------------------------------------------------------------------------
  ! this type is used to hold an rmf
  ! see http://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002/cal_gen_92_002.html
  type :: rmf_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo


    ! rmf specific keywords
    type(fitskey_string_type) :: key_phafile
    type(fitskey_string_type) :: key_chantype
    type(fitskey_int32_type)  :: key_detchans
!    type(fitskey_int32_type)  :: key_tlmin    !index is col number for the F_CHAN column
    type(fitskey_real64_type) :: key_lo_thres
    type(fitskey_int32_type)  :: key_numgrp
    type(fitskey_int32_type)  :: key_numelt

!    ! The following key combination is used for merged instrument output
!    ! for single instruments, do not print out key_ninst or key_instrum
!    type(fitskey_int32_type)  :: key_ninst
!    type(fitskey_string_type), dimension(1:MAX_INST)  :: pkey_instrum


    integer(kind=int32) :: patt_class         ! singles, doubles, triples or quadruples
    integer(kind=int16) :: patt_mask          ! singles, doubles, triples or quadruples

    ! this is the data area for the rmf values
    ! remember that each row entry is for a given range of input energies
    integer(kind=int32) :: num_rows
    real   (kind=real32), dimension(:),   allocatable :: pelo    ! ENERG_LO column
    real   (kind=real32), dimension(:),   allocatable :: pehi    ! ENERG_HI
    integer(kind=int16),  dimension(:),   allocatable :: pn_grp  ! N_GRP
!    integer(kind=int16),  dimension(:),   allocatable :: pf_chan ! F_CHAN
!    integer(kind=int16),  dimension(:),   allocatable :: pn_chan ! N_CHAN
    real   (kind=real32), dimension(:,:), allocatable :: pmatrix ! MATRIX (E_LO:E_HI,CHAN_LO:CHAN_HI)

    real   (kind=real32), dimension(:),   allocatable :: prowsum ! sum of pmatrix(e:e,1:detchans)

  end type rmf_type
  !-------------------------------------------------------------------------------

  type :: rmfset_type
    logical  :: is_valid
    type(rmf_type) :: rmf(SRCTOOL_PATT_CLASS_M:SRCTOOL_PATT_NCLASS)
  end type rmfset_type

!  type :: rmfsetlist_type
!    integer(kind=int32) :: num_rmfsets
!    type(rmfset_type) :: rmfset(0:MAX_INST)
!  end type rmfsetlist_type

  ! this type is used to contain the output RMFs (one per instrument and one for merged output)
  type :: rmflist_type
    integer(kind=int32) :: num_rmfs
    type(rmf_type) :: rmf(0:MAX_TOT_INST)
  end type rmflist_type



  !-------------------------------------------------------------------------------
  ! this type is used to hold an ebounds array
  ! see http://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002/cal_gen_92_002.html
  type :: ebounds_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    ! ebounds specific keywords
!    type(fitskey_string_type) :: key_phafile
!    type(fitskey_int32_type)  :: key_detchans
!    type(fitskey_int32_type)  :: key_tlmin

    integer(kind=int32) :: num_rows
    integer(kind=int32),  dimension(:), allocatable :: pchannel ! CHANNEL column
    real   (kind=real32), dimension(:), allocatable :: pemin    ! E_MIN
    real   (kind=real32), dimension(:), allocatable :: pemax    ! E_MAX

  end type ebounds_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! structure to hold info from a DEADCORn extension
  ! based on:
  ! http://www2011.mpe.mpg.de/erosita/internal/SASS-devel/eROSITA_TRoPIC_Eventfiles-4eng.html
  type :: deadcor_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    ! add in any extra header keywords here


    !arrays to hold the deadcor properties - allocated at run time
    integer(kind=int32) :: num_rows
    real   (kind=real64), dimension(:), allocatable :: ptime
    real   (kind=real32), dimension(:), allocatable :: pdeadc  !
    ! the following is derived from ptime() array
    real   (kind=real64), dimension(:), allocatable :: pstart ! start of the validity of this bin
!    real   (kind=real64), dimension(:), allocatable :: pstop

  end type deadcor_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! Date: Wed, 18 Jun 2014 17:18:09 +0200
  ! From: Hermann Brunner <hbrunner@mpe.mpg.de>
  !
  !      this morning we discussed the need for a pattern calibration
  !      file. Konrad proposed the following FITS table format:
  !
  !  ENERG_LO  E    keV
  !  ENERG_HI  E    keV
  !  S         E
  !  D         E
  !  T         E
  !  Q         E
  !
  !  The first two columns are the same as in the arf and the energy
  !  binning will also be the same. There are only four valid pattern types
  !  (s, d, t, q). The pattern columns will contain dimensionless values
  !  between 0.0 and 1.0. These need to be added up for the patter types
  !  included in the event file and multiplied with the effective areas.
  !
  !  The pattern selection will be coded in the PAT_SEL keyword of the
  !  input event file (as a bit mask, e.g. s+d: 3, s+d+t: 7, etc.).
  !-------------------------------------------------------------------------------
  ! structure to hold info from a Pattern Calibration file
  type :: pattcal_type
    logical  :: is_valid

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    ! add in any extra header keywords here
    integer(kind=int32) :: num_rows

    real   (kind=real32), dimension(:), allocatable :: pelo        ! ENERG_LO column
    real   (kind=real32), dimension(:), allocatable :: pehi        ! ENERG_HI
    real   (kind=real32), dimension(:), allocatable :: ps          ! Singles
    real   (kind=real32), dimension(:), allocatable :: pd          ! Doubles
    real   (kind=real32), dimension(:), allocatable :: pt          ! Triples
    real   (kind=real32), dimension(:), allocatable :: pq          ! Quadruples

  end type pattcal_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! this contains a live map of the current efficiency of the detector to events of
  ! each pattern type (singles, doubles, triples, quads)
  ! this gets updated whenever a new badpixel turns on or off
  type :: effmap_type
    logical  :: is_valid

!    integer(kind=int32) :: npatt
    integer(kind=int32) :: naxis1, naxis2
!    real (kind=real32), dimension(:,:,:), allocatable :: pmap  ! [PAT_CLASS][RAWX][RAWY]
    real (kind=real32), dimension(:,:), allocatable :: pmap  ! [RAWX][RAWY]

    ! the following is used to record the times when the status of one or more bad pixels changes
    ! and hence when the effmap should be updated
    integer(kind=int32) :: num_changes
    real(kind=real64),   dimension(:), allocatable :: ptchange

    real(kind=real64) :: time_of_last_update  ! s
    real(kind=real64) :: time_of_next_update  ! s

    type(pbadpix_type) :: pbadpixlist(MAX_EVTFILES+1)


  end type effmap_type
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! structure to hold a radially averaged PSF description
  ! will be computed internally from a 2D map psf description
  type :: rpsf_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    real(kind=real64) :: dradius_deg
    real(kind=real64) :: radius_50pc_deg   ! HEW/2
    real(kind=real64) :: radius_90pc_deg   ! 90%EW/2
    real(kind=real64) :: radius_99pc_deg   ! 99%EW/2
    ! 1d array to hold the radius for each step - this is the outer radius of each annulus - measured in deg
    real(kind=real64), dimension(:), allocatable :: pradius
    ! 1d array to hold the area of the annulus at each step - measured in deg2
    real(kind=real64), dimension(:), allocatable :: parea
    ! 1d array to hold the radially summed differential PSF within an annulus
    real(kind=real64), dimension(:), allocatable :: prpsf
    ! 1D array to hold the cumulative EEF within radius
    real(kind=real64), dimension(:), allocatable :: preef

  end type rpsf_type


  !-------------------------------------------------------------------------------
  ! structure to hold info from PSF small map extension
  type :: psfmap_type

    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    ! generic hdu info and keywords
    type(hduinfo_type)  :: hduinfo

    ! extra information keys
!    type(fitskey_real64_type) :: key_energ_lo
!    type(fitskey_real64_type) :: key_energ_hi
!    type(fitskey_real64_type) :: key_theta
!    type(fitskey_real64_type) :: key_phi
    type(fitskey_real64_type) :: key_energy
    type(fitskey_real64_type) :: key_backgrnd
    type(fitskey_real64_type) :: key_crpix1
    type(fitskey_real64_type) :: key_crpix2
    type(fitskey_real64_type) :: key_crval1
    type(fitskey_real64_type) :: key_crval2
    type(fitskey_real64_type) :: key_cdelt1
    type(fitskey_real64_type) :: key_cdelt2
    type(fitskey_string_type) :: key_cunit1
    type(fitskey_string_type) :: key_cunit2
    type(fitskey_int32_type)  :: key_naxis1
    type(fitskey_int32_type)  :: key_naxis2

    type(fitskey_string_type) :: key_caldb_energ ! key_CBD10001
    type(fitskey_string_type) :: key_caldb_theta ! key_CBD20001
    type(fitskey_string_type) :: key_caldb_obsmode ! key_CBD30001


    ! the naxis1,2 values are copied here from key_naxis1,2 keys for ease of access:
    integer(kind=int32) :: naxis1, naxis2

    ! 2d array to hold the map:
    ! use single precision because was taking up a huge amount of memory
    real(kind=real32), dimension(:,:), allocatable :: ppix

    ! index of the pixels that bracket the non-zero entries in the PSF map
    integer(kind=int32) :: pix_x_min, pix_x_max
    integer(kind=int32) :: pix_y_min, pix_y_max

    real(kind=real64) :: focal_length ! in mm, to be updated from instpar cal file
    real(kind=real64) :: pixscale_x, pixscale_y ! in deg/pix
    real(kind=real64) :: inv_pixscale_x, inv_pixscale_y ! in pix/deg
    real(kind=real64) :: area_per_pix  ! in deg

!    real(kind=real64) :: eff_energy  ! effective energy of this PSF in eV
    real(kind=real64) :: energy_lo   ! minimum energy of this PSF in keV
    real(kind=real64) :: energy_hi   ! maximum energy of this PSF in keV
    real(kind=real64) :: energy_eff  ! effective energy of this PSF in keV

    real(kind=real64) :: theta_lo   ! minimum offaxis angle of this PSF in arcmin
    real(kind=real64) :: theta_hi   ! maximum offaxis angle of this PSF in arcmin
    real(kind=real64) :: theta_eff  ! effective offaxis angle of this PSF in arcmin

    real(kind=real64) :: phi_lo   ! minimum azimuthal angle of this PSF in deg
    real(kind=real64) :: phi_hi   ! maximum azimuthal angle of this PSF in deg
    real(kind=real64) :: phi_eff  ! effective azimuthal angle of this PSF in deg

    ! this will contain a description of the radially averaged PSF/EEF
    type(rpsf_type)     :: rpsf


  end type psfmap_type
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  ! the following allows arrays of pointers
  type psfmap_pointer
    type(psfmap_type), pointer :: p
  end type psfmap_pointer
  !-------------------------------------------------------------------------------




  type :: psfmaplist_type
   ! flag used to show if this structure is ready to be used
    logical  :: is_valid
    integer(kind=int32) :: num_psfmaps
    type(psfmap_type), allocatable, dimension(:) :: psfmap

    ! used to record the discrete values of energy,theta, and phi are represented by the psfmaplist
    integer(kind=int32) :: num_e
    integer(kind=int32) :: num_theta
    integer(kind=int32) :: num_phi

    real(kind=real64), dimension(:), allocatable :: pe            ! energy in keV
    real(kind=real64), dimension(:), allocatable :: ptheta_arcmin ! offaxis in arcmin
    real(kind=real64), dimension(:), allocatable :: pphi          ! azimuthal angle in degrees

    integer(kind=int32), dimension(:,:,:), allocatable    :: pindex  !

  end type psfmaplist_type

  !-------------------------------------------------------------------------------
  ! structure to hold a set of info from a PSF description file
  type :: psf_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    ! generic hdu info and keywords - copy from representative values
    type(hduinfo_type)  :: hduinfo

    integer(kind=int32) :: class   ! a code which tells us the type of PSF description

    type(psfmaplist_type)   :: psfmaplist
    ! TODO - add in other types of psf description

  end type psf_type
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! structure to hold info from a vignetting description file
  ! Vignetting description format based on:
  !   http://www2011.mpe.mpg.de/erosita/simul/synthetic/Release_jan2013/Calib/sim1_tvignet_100302v01.fits
  type :: vign_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    integer(kind=int32) :: class

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    ! add in any extra header keywords here

    !arrays to hold the vignetting properties - allocated at run time
    integer(kind=int32) :: num_rows
    integer(kind=int32) :: num_e
    real   (kind=real32), dimension(:), allocatable :: pelo ! keV
    real   (kind=real32), dimension(:), allocatable :: pehi ! keV
    integer(kind=int32) :: num_theta
    real   (kind=real32), dimension(:), allocatable :: ptheta
    integer(kind=int32) :: num_phi
    real   (kind=real32), dimension(:), allocatable :: pphi
    integer(kind=int32) :: num_vign
    real   (kind=real32), dimension(:), allocatable :: pvign

    ! derived parameters:
    ! this is the monochromatic energy at which we should assume the vignetting is evaluated
    real   (kind=real32), dimension(:), allocatable :: pe ! keV

  end type vign_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! structure to hold info about the detector layout
  ! based on detector map example at
  ! /data12s/erosita/caldb/data/erosita/sim1/bcf/sim1_detmap_100602v01.fits
  type :: detmap_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    !specialised keywords
    type(fitskey_int32_type)  :: key_naxis1
    type(fitskey_int32_type)  :: key_naxis2

    ! copies of keyword values
    integer(kind=int32) :: naxis1,naxis2                    ! ccd format

    real (kind=real32), dimension(:,:), allocatable :: pmap  ! the detmap data itself
!    real (kind=real32), dimension(:), allocatable :: pmap  ! the detmap data itself

  end type detmap_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! structure to hold generic info about the instrument layout
  ! based on example at
  ! /data12s/erosita/caldb/data/erosita/sim1/bcf/sim1_instpar_100705v01.fits
  !
  !  No. Type     EXTNAME      BITPIX Dimensions(columns)      PCOUNT  GCOUNT
  !
  !   0  PRIMARY                  8     0                           0    1
  !   1  BINTABLE INSTPAR         8     68(10) 1                    0    1
  !     Column Name                Format     Dims       Units     TLMIN  TLMAX
  !      1 FOCAL_LENGTH               1D                  mm
  !      2 MM_XPIX                    1D                  mm/pix
  !      3 MM_YPIX                    1D                  mm/pix
  !      4 X_PLATESCALE               1D                  arcsec/pix
  !      5 Y_PLATESCALE               1D                  arcsec/pix
  !      6 X_CCDPIX                   1J
  !      7 Y_CCDPIX                   1J
  !      8 X_OPTAX                    1D                  pixel
  !      9 Y_OPTAX                    1D                  pixel
  !     10 FOV_RAD                    1E                  deg

  type :: instpar_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    ! add in any extra header keywords here

    ! Now come the instpar values - each is stored as a column entry
    ! we expect to get only a single row, but keep it generic for now
    integer(kind=int32) :: num_rows       ! the number rows
    real   (kind=real64), allocatable, dimension(:) :: pfocal_length
    real   (kind=real64), allocatable, dimension(:) :: pmm_xpix
    real   (kind=real64), allocatable, dimension(:) :: pmm_ypix
    real   (kind=real64), allocatable, dimension(:) :: px_platescale
    real   (kind=real64), allocatable, dimension(:) :: py_platescale
    integer(kind=int32),  allocatable, dimension(:) :: px_ccdpix
    integer(kind=int32),  allocatable, dimension(:) :: py_ccdpix
    real   (kind=real64), allocatable, dimension(:) :: px_optax
    real   (kind=real64), allocatable, dimension(:) :: py_optax
    real   (kind=real32), allocatable, dimension(:) :: pfov_rad

    ! copies and parameters derived from above values
    integer(kind=int32) :: naxis1,naxis2                    ! ccd format
    real(kind=real64)   :: pixscale_x, pixscale_y           ! pixel scale in deg/pix
    real(kind=real64)   :: inv_pixscale_x, inv_pixscale_y   ! inverse pixel scale in pix/deg
    real(kind=real64)   :: oa_rawx, oa_rawy                 ! pixel coords of optical axis
    real(kind=real64)   :: fov_radius_arcmin                !

  end type instpar_type
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! Structure to hold WCS information for event list
  type :: evtwcs_type
     real(kind=real64) :: rval_x, rval_y
     real(kind=real64) :: rpix_x, rpix_y
     real(kind=real64) :: delta_x, delta_y
     character(len=STR_MAX) :: type_x, type_y
     character(len=STR_MAX) :: unit_x, unit_y
  end type evtwcs_type

  !-------------------------------------------------------------------------------
  ! structure to hold an eventlist
  ! columns based on:
  ! http://www2011.mpe.mpg.de/erosita/internal/SASS-devel/eROSITA_TRoPIC_Eventfiles-4eng.html
  ! Calibrated event file columns (as archived in the products archive):

  type :: evtlist_type
    ! flag used to show if this structure is ready to be used
    logical  :: is_valid

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    ! add in any extra header keywords here
    type(fitskey_int32_type) :: key_flagsel
    type(fitskey_int16_type) :: key_pat_sel

    ! arrays to hold the event properties - allocated at run time
    integer(kind=int32) :: num_rows       ! the number of events recorded in arrays
    real   (kind=real64), allocatable, dimension(:) :: ptime
    real   (kind=real64), allocatable, dimension(:) :: pra
    real   (kind=real64), allocatable, dimension(:) :: pdec
    integer(kind=int32),  allocatable, dimension(:) :: px
    integer(kind=int32),  allocatable, dimension(:) :: py
    real   (kind=real32), allocatable, dimension(:) :: penergy
    real   (kind=real32), allocatable, dimension(:) :: pev_weight
    integer(kind=int16),  allocatable, dimension(:) :: prawx     ! TODO
    integer(kind=int16),  allocatable, dimension(:) :: prawy
    real   (kind=real64), allocatable, dimension(:) :: psubx
    real   (kind=real64), allocatable, dimension(:) :: psuby
    integer(kind=int16),  allocatable, dimension(:) :: ppha
    integer(kind=int16),  allocatable, dimension(:) :: ppat_typ
    integer(kind=int8),   allocatable, dimension(:) :: ppat_inf
!    integer(kind=int8),   allocatable, dimension(:) :: pccdnr
    integer(kind=int8),   allocatable, dimension(:) :: ptm_nr
    integer(kind=int32),  allocatable, dimension(:) :: pflag

!    integer(kind=int32),  allocatable, dimension(:) :: pfile_index   ! which input evtlist did this come from?
!    integer(kind=int32),  allocatable, dimension(:) :: pfile_row     ! which row of the input evtlist did this come from?

    ! mask used to store location of event (logical OR of EVT_CLASS_FLAG_SRC and/or EVT_CLASS_FLAG_BACK)
    integer(kind=int8),   allocatable, dimension(:) :: pclass_mask
    ! now the associated metadata extensions

    ! WCS information from event file
    type (evtwcs_type) :: wcs

    ! A structure containing the gti info for these events
    type (gtilist_type) :: gtilist

    ! A structure containing the baxpix info for these events
    type (badpix_type)  :: badpix

    ! A structure holding the attitude information
    type (attlist_type) :: attlist

    ! A structure holding the livetime information
    type (deadcor_type) :: deadcor

    ! an integer that encodes the information in the hduinfo%key_obs_mode keyword
    ! i.e. either OBS_MODE_SURVEY or OBS_MODE_POINTING
    integer(kind=int32) :: obs_mode

  end type evtlist_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! structure to hold a list of eventlists
  type :: evtlistlist_type
    ! generic hdu info and keywords that apply to products made from all input evtlists
    type(hduinfo_type) :: hduinfo

    integer(kind=int32) :: num_evtlists
    type(evtlist_type)  :: evtlist(1:MAX_EVTFILES)

  end type evtlistlist_type
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! this type is used to tidy things up inside srctool_calc_fracrsp
  type :: lcbin_type
    integer(kind=int32) :: i      ! index of current lc bin
    real(kind=real64)   :: tstart
    real(kind=real64)   :: tstop
    real(kind=real64)   :: t_tot
    real(kind=real64)   :: t_tot_live
!    logical  :: flag
    real(kind=real64)   :: backratio
    real(kind=real64)   :: backscale_src, backscale_back
    real(kind=real64)   :: fracareaflat
    real(kind=real64)   :: offaxis
!    real(kind=real64)   :: backscale_src! backscale of the source in this bin (units=deg^2)
!    real(kind=real64)   :: backscale_back! backscale of the background in this bin (units=deg^2)
!    logical  :: to_consider

  end type lcbin_type
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! this type is used to tidy things up inside srctool_calc_fracrsp
  type :: tstep_type
    real(kind=real64)  :: tmin     ! time at the start of the time step
    real(kind=real64)  :: tmid     ! time at the middle of the time step
    real(kind=real64)  :: tmax     ! time at the end of the time step
    real(kind=real64)  :: dt       ! length of the time step (tmax-tmin)
    real(kind=real64)  :: dt_live  ! length of the time step, multiplied by livetime

    real(kind=real64)  :: backratio ! backscale ratio in this time step
    real(kind=real64)  :: backscale_src! backscale of the source in this tstep (units=deg^2)
    real(kind=real64)  :: backscale_back! backscale of the background in this tspec (units=deg^2)
    real(kind=real64)  :: fracareaflat ! ratio of backscale to total in tstep
  end type tstep_type
  !-------------------------------------------------------------------------------

!  !-------------------------------------------------------------------------------
!  ! this type is used to tidy things up inside srctool_calc__fracrsp
!  type :: eval_type
!
!!    integer(kind=int32) :: vign_nenergies
!!    real(kind=real64), dimension(:), allocatable :: vign_energies
!!    real(kind=real64), dimension(:), allocatable :: vign_values
!
!    integer(kind=int32) :: num_e   ! number of unique energies at which we evaluate psf or vign
!    real(kind=real64),   dimension(:), allocatable :: pe         ! energy values - keV
!    real(kind=real64),   dimension(:), allocatable :: values     ! total over step
!    real(kind=real64),   dimension(:), allocatable :: values_arf ! total over whole observation
!    real(kind=real64),   dimension(:), allocatable :: values_lc  ! just for this lightcurve bin
!  end type eval_type
!  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! this type is used to hold a light curve
  ! see: https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/summary/ogip_93_003_summary.html
  !  "The Proposed Timing FITS File Format for High Energy Astrophysics Data"
  !     Desc. Gives the standard keywords and formats to be used for various types of 'light curve', published in Legacy, 3, 32.
  !     Code: OGIP/93-003
  !     Authors: Angelini Pence and Tennant


  type :: lc_type
    logical  :: to_consider

    character(len=STR_MAX) :: lctype       ! string describing how the LC should be constructed
    integer(kind=int32) :: lctype_code   ! what type of lc this is
    integer(kind=int32) :: num_lcpars
    real(kind=real64),  dimension(MAX_LCPARS) :: lcpars

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    ! lc-specific keywords
    type(fitskey_real64_type)  :: key_timedel
    type(fitskey_real64_type)  :: key_timezero

    type(fitskey_real64_type)  :: key_telapse
    type(fitskey_real64_type)  :: key_ontime
    type(fitskey_real64_type)  :: key_livetime
    type(fitskey_real64_type)  :: key_exposure

    type(fitskey_real64_type)  :: key_tierrela
    type(fitskey_real64_type)  :: key_tierabso
    type(fitskey_logical_type) :: key_clockcor
    type(fitskey_logical_type) :: key_backapp
    type(fitskey_logical_type) :: key_deadapp
    type(fitskey_logical_type) :: key_vignapp

    type(fitskey_int32_type) :: key_flagsel
    type(fitskey_int16_type) :: key_pat_sel

    type(fitskey_vlstring_type)  :: key_src_region
    type(fitskey_vlstring_type)  :: key_back_region

    ! The following key combination is used for merged instrument output
    ! for single instruments, do not print out key_ninst or key_instrum
!    type(fitskey_int32_type)  :: key_ninst
!    type(fitskey_string_type), dimension(1:MAX_INST)  :: pkey_instrum

    ! the following describes the ebands over which the lcurves are compiled
    integer(kind=int32) :: num_ebands
    type(fitskey_real64_type) :: key_emin(MAX_LCBANDS)
    type(fitskey_real64_type) :: key_emax(MAX_LCBANDS)

    type(fitskey_int32_type)   :: key_ncounts_src(MAX_LCBANDS)
    type(fitskey_int32_type)   :: key_ncounts_back(MAX_LCBANDS)

    integer(kind=int32) :: int_nenergies                         ! number of discrete energies which we integrate over
    real(kind=real64),  dimension(:), allocatable :: pint_e      ! energies used when integrating over ebands
    real(kind=real64),  dimension(:,:), allocatable :: pint_weights ! weights used when integrating over ebands - TODO separate weights for each lc band
    real(kind=real64)  :: lcgamma  ! copied from param list


    integer(kind=int32) :: num_rows
    real(kind=real64),  dimension(:),   allocatable :: ptime          ! TIME                                 column
    real(kind=real64),  dimension(:),   allocatable :: ptimedel       ! TIMEDEL                              column
    integer(kind=int32),dimension(:,:), allocatable :: pcounts        ! COUNTS      (num_ebands wide vector) column
    real(kind=real32),  dimension(:,:), allocatable :: prate          ! RATE        (num_ebands wide vector) column
    real(kind=real32),  dimension(:,:), allocatable :: prate_err      ! RATE_ERR    (num_ebands wide vector) column
    real(kind=real32),  dimension(:,:), allocatable :: prate_estposerr ! RATE_ESTPOSERR (num_ebands wide vector) column
    real(kind=real32),  dimension(:,:), allocatable :: prate_estnegerr ! RATE_ESTNEGERR (num_ebands wide vector) column

    real(kind=real32),  dimension(:,:), allocatable :: pback_counts   ! BACK_COUNTS (num_ebands wide vector) column
    real(kind=real32),  dimension(:,:), allocatable :: pfracarea      ! FRACAREA    (num_ebands wide vector) column
    real(kind=real32),  dimension(:),   allocatable :: pfracflatarea  ! FRACFLATAREA                         column
    real(kind=real32),  dimension(:),   allocatable :: pfractime      ! FRACTIME                             column
    real(kind=real32),  dimension(:,:), allocatable :: pfracexp       ! FRACEXP     (num_ebands wide vector) column
    real(kind=real32),  dimension(:),   allocatable :: pfracflatexp   ! FRACFLATEXP                          column
    real(kind=real32),  dimension(:),   allocatable :: poffaxis       ! OFFAXIS                              column
    real(kind=real32),  dimension(:),   allocatable :: pbackscal_src  ! BACKSCAL_SRC                         column
    real(kind=real32),  dimension(:),   allocatable :: pbackscal_back ! BACKSCAL_BACK                        column
    real(kind=real64),  dimension(:),   allocatable :: pbackratio     ! BACKRATIO                            column

    ! the following columns are not written out
    real(kind=real64),  dimension(:),   allocatable :: ptmin          ! TIME at start of bin
    real(kind=real64),  dimension(:),   allocatable :: ptmax          ! TIME at end of bin

  end type lc_type
  !-------------------------------------------------------------------------------

  type :: lclist_type
    integer(kind=int32) :: num_lcs
    type(lc_type) :: lc(0:MAX_TOT_INST)
  end type lclist_type
  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  ! this type is used to hold a spectrum
  ! see:
  ! http://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007/ogip_92_007.html
  ! "The OGIP Spectral File Format"
  ! Desc. Gives the standard keywords and formats to be used for XSPEC-compatible PHA files. published in Legacy, 2, 65.
  ! Code: OGIP/92-007
  ! Authors: Arnaud, George and Tennant.
  ! Notes: Published in Legacy, 2, 65. There is also an addendum
  ! available (OGIP/92-007a) which lists all the changes which have
  ! been made to the formats since the Legacy article.

  type :: spec_type
    integer(kind=int8)  :: evt_class  ! set to EVT_CLASS_FLAG_SRC _OR_ EVT_CLASS_FLAG_BACK

    ! generic hdu info and keywords
    type(hduinfo_type) :: hduinfo

    ! spec-specific keywords
    type(fitskey_real64_type)  :: key_telapse
    type(fitskey_real64_type)  :: key_ontime
    type(fitskey_real64_type)  :: key_livetime
    type(fitskey_real64_type)  :: key_exposure

    type(fitskey_string_type)  :: key_backfile
    type(fitskey_string_type)  :: key_corrfile
    type(fitskey_real64_type)  :: key_corrscal
    type(fitskey_string_type)  :: key_respfile
    type(fitskey_string_type)  :: key_ancrfile
    type(fitskey_logical_type) :: key_poisserr
    type(fitskey_string_type)  :: key_chantype
    type(fitskey_int32_type)   :: key_detchans
    type(fitskey_real64_type)  :: key_areascal
    type(fitskey_real64_type)  :: key_backscal
    type(fitskey_int32_type)   :: key_grouping
    type(fitskey_int32_type)   :: key_quality
    type(fitskey_real64_type)  :: key_geomregarea
    type(fitskey_real64_type)  :: key_geomregmodarea

!    Moved to std header
!    ! The following key combination is used for merged instrument output
!    ! for single instruments, do not print out key_ninst or key_instrum
!    type(fitskey_int32_type)  :: key_ninst
!    type(fitskey_string_type), dimension(1:MAX_INST)  :: pkey_instrum

    type(fitskey_int32_type)   :: key_ncounts
    type(fitskey_int32_type)   :: key_ncounts_s
    type(fitskey_int32_type)   :: key_ncounts_d
    type(fitskey_int32_type)   :: key_ncounts_t
    type(fitskey_int32_type)   :: key_ncounts_q

    type(fitskey_int32_type) :: key_flagsel
    type(fitskey_int16_type) :: key_pat_sel

    type(fitskey_vlstring_type)  :: key_src_region
    type(fitskey_vlstring_type)  :: key_back_region

    integer(kind=int32) :: num_rows
    integer(kind=int32), dimension(:), allocatable :: pchannel  ! CHANNEL column
    integer(kind=int32), dimension(:), allocatable :: pcounts   ! COUNTS
    integer(kind=int32), dimension(:), allocatable :: pcounts_s  ! COUNTS_S
    integer(kind=int32), dimension(:), allocatable :: pcounts_d  ! COUNTS_D
    integer(kind=int32), dimension(:), allocatable :: pcounts_t  ! COUNTS_T
    integer(kind=int32), dimension(:), allocatable :: pcounts_q  ! COUNTS_Q
  end type spec_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  type :: speclist_type
    integer(kind=int32) :: num_spec

    ! in the following array
    !  index=0 pha data is summed over all instruments
    !  index=x: spectra for TMx
    !  index=8: sum of TMs 1,2,3,4,6
    !  index=9: sum of TMs 5,7
    type(spec_type) :: src_spec(0:MAX_TOT_INST)
    type(spec_type) :: back_spec(0:MAX_TOT_INST)
    logical :: do_back
  end type speclist_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  type :: samplelist_type
    ! the following are used for the sample points
    integer(kind=int8)  :: evt_class       ! set to EVT_CLASS_FLAG_SRC _OR_ EVT_CLASS_FLAG_BACK
    integer(kind=int32) :: psf_class       ! SRCTOOL_PSF_CLASS_NONE, etc

    integer(kind=int32) :: num_points      ! initially nx*ny, but then filtered
    integer(kind=int32) :: num_points_region ! sample points within region
    integer(kind=int32) :: num_points_region_model ! within region + model
    integer(kind=int32) :: nx
    integer(kind=int32) :: ny
    real(kind=real64)   :: area_per_point          ! in deg^2
    real(kind=real64)   :: area_per_point_sqarcsec ! in arcsec^2
    real(kind=real64)   :: dx, dy              ! step between points, in deg
    real(kind=real64)   :: weight_per_point    !

    real(kind=real64)  :: src_ra       ! coords of the source in various systems
    real(kind=real64)  :: src_dec
    real(kind=real64)  :: src_rawx
    real(kind=real64)  :: src_rawy
    logical :: src_on_chip_flag
    real(kind=real64)  :: src_theta_arcmin    ! offaxis in arcmin
    real(kind=real64)  :: src_phi      ! polar angle in deg

    integer(kind=int32) :: num_eval_e   ! =size(peval_e) number of energy bands to evaluate samples
    integer(kind=int32) :: num_psf_e    ! =size(ppsf_e)  number of energy bands to evaluate summed psfs

    ! these are arrays in energy:
    real(kind=real64),  dimension(:),   allocatable :: peval_e   ! energy levels, keV
    real(kind=real64),  dimension(:),   allocatable :: ppsf_e    ! energy levels, keV
    real(kind=real64),  dimension(:),   allocatable :: ppsfsum   ! summed psf values (at native psf energy levels)
                                                                 ! size(ppsfsum(:)) = num_psf_e
    real(kind=real64),  dimension(:),   allocatable :: pfracarea ! fractional effective area at at eval energy levels)
                                                                 ! size(pfracarea(:)) = num_eval_e

    ! these are arrays over the sample points:
    real(kind=real64),  dimension(:),   allocatable :: pra           ! 1d arrays,
    real(kind=real64),  dimension(:),   allocatable :: pdec          !
    real(kind=real64),  dimension(:),   allocatable :: pcosdec       ! computed once for speed
    real(kind=real64),  dimension(:),   allocatable :: psindec       !    ""     ""
    real(kind=real64),  dimension(:),   allocatable :: pdist         ! instantaneous dist from source coord, arcmin!!!!!!!!
    real(kind=real64),  dimension(:),   allocatable :: prawx         ! instantaneous position in rawxy coords
    real(kind=real64),  dimension(:),   allocatable :: prawy         !
    integer(kind=int32),dimension(:),   allocatable :: prawx_int     ! rounded version of the above
    integer(kind=int32),dimension(:),   allocatable :: prawy_int     !
    logical, dimension(:),   allocatable :: pon_chip_flag !
!    logical, dimension(:),   allocatable :: pin_region_flag !
    integer(kind=int32),dimension(:),   allocatable :: pi            !
    integer(kind=int32),dimension(:),   allocatable :: pj            !
    real(kind=real64),  dimension(:),   allocatable :: pmodel        ! Relative weight of sample point,
                                                                     ! derived from source extent model
!    real(kind=real64),  dimension(:),   allocatable :: parea         ! sky area per sample point (arcsec**2)
    integer(kind=int32),dimension(:),   allocatable :: pimage        ! used to sum counts into an image,optional

    real(kind=real64),  dimension(:), allocatable   :: peffmap       ! effective instrument sensitivity at sample

    ! these are arrays over the sample points and with second dimension spanning the eval energy levels:
    real(kind=real64),  dimension(:,:), allocatable :: pexpmap       ! 2nd dimension is eval energy bands
    real(kind=real64),  dimension(:,:), allocatable :: pvign         ! 2nd dimension is eval energy bands

    ! these used to be in the eval_type structure - now moved here for simplicity
    ! they are arrays in energy, with size()=num_eval_e
    real(kind=real64),   dimension(:), allocatable :: eval_values     ! fracarea*dt_live in a time bin
    real(kind=real64),   dimension(:), allocatable :: eval_values_arf ! sum of eval_values
    real(kind=real64),   dimension(:), allocatable :: eval_sum_psf    ! sum interpolated psfsum*dt_live
    real(kind=real64),   dimension(:), allocatable :: eval_sum_vign   ! sum interpolated pvign*dt_live
    real(kind=real64),   dimension(:), allocatable :: eval_values_lc  ! just for this lightcurve bin

    real(kind=real64)  :: t_tot          ! sum over the 'tstep%dt's
    real(kind=real64)  :: t_tot_live     ! sum over the 'tstep%dt_live's

    real(kind=real64)  :: ra0,dec0,roll ! current orientation of instrument
    real(kind=real64)  :: cosdec0,sindec0,cosroll,sinroll ! derived from current orientation of instrument

    real(kind=real64)  :: smallest_sample_radius_deg   ! offaxis in deg
    real(kind=real64)  :: largest_sample_radius_deg    ! offaxis in deg

    real(kind=real64)  :: backscale


  end type samplelist_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! this type is used to define the time intervals over which any integrations/summations should be performed
  ! the idea is that this is derived from a set of input *GTIn extensions
  type :: interval_type
    real(kind=real64)  :: t1     ! beginning of time interval
    real(kind=real64)  :: t2     ! end of time interval
    real(kind=real64)  :: tmid   ! middle of time interval
    real(kind=real64)  :: tstep  ! time step for calculations within this interval
    logical :: inst_flag(0:MAX_TOT_INST) ! flag showing if should consider each instrument for this time bin
    integer(kind=int32) :: ninst  ! total number of insts to consider in this time interval

  end type interval_type
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  type :: intervallist_type
    integer(kind=int32) :: num_rows
    type(interval_type), dimension(:), allocatable :: pinterval
    integer(kind=int32)  :: num_tsteps  ! total number of time steps within all intervals where at least one inst is valid
    real(kind=real64)    :: total_time  ! sum of the durations of the intervals where at least one inst is valid
  end type intervallist_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  type :: filelist_type
    ! TODO: make allocatable
    character(len=LSTR_MAX), dimension(MAX_PRODUCTS_PER_SRC) :: names
    integer(kind=int32) :: num
  end type filelist_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! pointers are used here to work around a gfortan bug - see srctool_allocate_inst
  type :: inst_type
    ! these are the index number and name used to identify this instrument
    integer(kind=int32) :: index
    character(len=FITSIO_KEY_MAXLEN) :: name

    ! flag used to show if this structure is ready to be used
    logical :: is_valid

    ! this is set to true when we want to consider data from this instrument
    logical :: to_consider

    ! this is whether to write output files for this instrument
    logical :: to_write

    ! whether to use FoV mask
    logical :: use_fov_mask

    ! this is a pointer to the corresponding generic instrument data structure
    ! (note this isn't a pointer to workaround the gfortran bug - this is necesary!)
    type(inst_type), pointer :: pinst0

    ! this is the structure containing all the eventlists for this instrument
    type(evtlistlist_type), pointer :: evtlistlist

    ! A data structure holding the on-axis reference ARF information
    type(arf_type), pointer :: refarf

    ! A data structure holding the CALDB RMF information
    ! Need one RMF per pattern type, plus keep space for one merged pattern RMF
    type(rmfset_type), pointer :: refrmfset

    ! A data structure holding the energy bounds info
    type(ebounds_type), pointer :: ebounds

   ! A data structure holding an analytic vignetting description
    type(avign_type), pointer :: avign

    ! A data structure holding the instantaneous PSF description
    type(psf_type), pointer :: psf

    ! A data structure holding the survey PSF description
    type(psf_type), pointer :: survey_psf

    ! A data structure holding the detector map info
    type(detmap_type), pointer :: detmap

    ! A data structure holding the instrument parameters
    type(instpar_type), pointer :: instpar

    ! A structure holding the merged (across input eventlist) attitude information
    type(attlist_type), pointer :: attlist

    ! A structure holding the merged (across input eventlist) deadcor information
    type(deadcor_type), pointer :: deadcor

    ! this is a merged and filtered (across input eventlist) input GTI information
    type(gtilist_type), pointer :: gtilist

    ! A structure holding the pattcal info
    type(pattcal_type), pointer :: pattcal

    ! a structure holding the current efficiency of each pixel to events of each class (s,d,q,t)
    type(effmap_type), pointer :: effmap

    ! this is updated to contain the time intervals when each source
    ! lies within the FOV
    type(gtilist_type), pointer :: src_gtilist

    ! copies of input params - for ease of processing later
    real(kind=real64) :: tstep
    real(kind=real64) :: xgrid(MAX_XGRIDPARS)

    ! this is used when calculating fractional response
    type(samplelist_type), pointer :: samplelist_src
    type(samplelist_type), pointer :: samplelist_back

    integer(kind=int32) :: filter_code

  end type inst_type
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  type :: instlist_type
    integer(kind=int32) :: num_inst_to_consider
    ! in the following array, the inst_index=0 cal files area for all instruments
    type(inst_type), dimension(0:MAX_TOT_INST) :: inst

    real(kind=real64)        :: min_tstart  ! derived from the extreme tstart/tstop values in the evtlists
    real(kind=real64)        :: max_tstop   ! TODO check that these are calculated correctly

    type(intervallist_type) :: intervallist

    ! copies of input params - for ease of processing later
    type(srctool_cal_type) :: gtitype_to_read

    integer(kind=int16) :: cl_pat_sel
    integer(kind=int32) :: cl_flagsel

    ! these are used for extended source handling
    ! they give the common dimensions and pixel scale of the images that
    ! should be FFT'ed when convolving extemnt model and PSF models
    ! fft_npix = sum of the linear dimensions of the largest PSF map and the largest extent model
    ! fft_pixscale = the pixel scale on which the PSF and extent models are sampled (arcsec/pix)
    ! and is typicaly the same as the native resolution of the PSF maps
    ! it is an error if there is more than one pixel scale within the set of PSF maps
    integer(kind=int32) :: fft_npix
    real(kind=real64)   :: fft_pixscale
    
    ! these are plans for doing FFT transforms of the PSF
    type(C_PTR)         :: fftw_plan_r2c, fftw_plan_c2r

    ! number of eventlists of each type
    integer(kind=int32) :: num_survey_mode
    integer(kind=int32) :: num_pointing_mode
    integer(kind=int32) :: num_scan_mode
    integer(kind=int32) :: num_slew_mode

  end type instlist_type
  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  !----- SOURCE DESCRIPTION TYPES ------------------------------------------------
  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  type :: extent_type
    integer(kind=int32) :: code
    integer(kind=int32) :: num_pars
    real(kind=real64)   :: pars(1:MAX_EXTPARS) ! meaning is extent-code specific

    ! image of extent with same pixel size as psf
    real(kind=real32), allocatable, dimension(:,:) :: xy

    integer(kind=int32) :: num_pix ! size of extent images in pixels

    real(kind=real64)   :: norm_const ! multiplier that makes integral over radial range = 1,  units [arcsec**-2]
  end type extent_type
  !-------------------------------------------------------------------------------

  ! ownership of sources
  !-------------------------------------------------------------------------------
  ! ownership values
  integer(kind=int32), parameter :: SRCTOOL_OWNER_UNSET=0
  integer(kind=int32), parameter :: SRCTOOL_OWNER_YES=1
  integer(kind=int32), parameter :: SRCTOOL_OWNER_NO=2
  integer(kind=int32), parameter :: SRCTOOL_OWNER_MIXED=3

  !-------------------------------------------------------------------------------
  type :: owner_type
     ! has above ownership values
     integer(kind=int32) :: mpe        ! has MPE ownership
     integer(kind=int32) :: iki        ! has IKI ownership
  end type owner_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! a src type describes all params for an input source (for which we will extract spec/lc)
  type :: src_type
    character(len=STR_MAX)  :: name
    integer(kind=int32)     :: index
    type(coord_type)        :: coord
    type(extent_type)       :: extent
    type(extmap_type)       :: extmap
    type(region_type)       :: src_region
    type(region_type)       :: back_region
    logical                 :: do_back   ! set to true if we are extracting background
    type(owner_type)        :: owner

    character(len=:), allocatable :: str_src_region
    character(len=:), allocatable :: str_back_region
    integer(kind=int32)     :: nevents_src_i(1:MAX_INST)
    integer(kind=int32)     :: nevents_back_i(1:MAX_INST)

    logical                 :: file_clobber  ! whether to clobber output files
    character(len=LSTR_MAX) :: file_dirname_tar ! directory of output (tar) products
    character(len=LSTR_MAX) :: file_dirname_product ! directory of fits products
    character(len=LSTR_MAX) :: file_prefix   ! filename prefix (excluding directory)
    character(len=LSTR_MAX) :: file_suffix   ! filename suffix

    integer(kind=int32)     :: todo_code     ! a bitmask version of the todo list

    real(kind=real64)       :: cat_scts      ! Units [cts/exposure]
                                             ! The catalogued value of the max likelihood source counts
                                             ! This is expected to be:
                                             !    - background subtracted
                                             !    - corrected for vignetting effects
                                             !    - corrected to infinite aperture
                                             !    - for the "total" energy band
    real(kind=real64)       :: cat_bg_map    ! Units [cts/deg2/exposure]
                                             ! The catalogued value of the local background rate
                                             !    - for the "total" energy band
                                             ! converted immediately after reading from input units of [cts/pix/exposure] using
                                             ! pixel scale is read from sourcelist header keywords

    real(kind=real64)       :: cat_ext       ! Units [deg]
                                             ! The catalogued value of the object extent
                                             ! converted immediately after reading from input units of [arcsec]


    logical                 :: cat_extract   ! Flag column (whether to extract source products)
    logical                 :: cat_exclude   ! Flag column (whether to exclude from auto regions)

    type(rpsf_type), pointer :: prpsf      ! Will point to the relevant RPSF structure for this source
                                            ! For extended sources, we will create a new RPSF

    ! keep track of products generated for each source
    type(filelist_type) :: filelist

  end type src_type
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  type :: srclist_type
    integer(kind=int32) :: type_code         ! How the sourcelist was specified (single source, FITS catalogue, ascii list etc)

    integer(kind=int32) :: num_srcs          ! the number of objects recorded in array

    integer(kind=int32) :: autoreg_code_src  ! These tell us if we need to automatically generate
    integer(kind=int32) :: autoreg_code_back ! regions for sources and background regions

    logical             :: auto_columns_supplied


    integer(kind=int32) :: n_to_extract
    integer(kind=int32) :: n_to_exclude

    integer(kind=int32) :: n_extended        ! the number of sources flagged as extended
    integer(kind=int32) :: n_extended_to_extract
    integer(kind=int32) :: n_extended_to_exclude

    type(src_type),  allocatable, dimension(:) :: psrc
  end type srclist_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

contains

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !------------------------ SUBROUTINES ------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------



!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
end module  srctool_datatypes
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
