! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_detmap

  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling DETMAP data/files
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         !
  use srctool_misc              !
  use srctool_generic_fitsraw

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_DETMAP_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_detmap
  private :: srctool_detmap_report_detmap_info
  public ::  srctool_detmap_empty_detmap
  public ::  srctool_detmap_read_detmap_from_chdu

contains




  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------Start of DETMAP file handling-------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_detmap_empty_detmap(detmap)
    type(detmap_type), intent(inout)      :: detmap
    if ( allocated(detmap%pmap))   deallocate (detmap%pmap)
    detmap%naxis1 = 0
    detmap%naxis2 = 0
    detmap%is_valid = .false.

    call fitswrap_init_hduinfo(detmap%hduinfo)
    call fitswrap_init_key(detmap%key_naxis1, "NAXIS1")
    call fitswrap_init_key(detmap%key_naxis2, "NAXIS2")


  end subroutine srctool_detmap_empty_detmap
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !read the detmap description from the current hdu of the fits file identified by unit
  subroutine srctool_detmap_read_detmap_from_chdu(unit, detmap, ierr)
    character(len=*), parameter :: subroutinename='srctool_deadcor_read_detmap_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(detmap_type), intent(inout)  :: detmap
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status
!    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: anyf
    integer(kind=int32), parameter :: group=0
    real(kind=real32), parameter   :: nullval = 0.0e0
!    integer(kind=int32) :: i,j

    ierr = ERROR_LEVEL_OK
    status = 0
!    call wrschatty (subroutinename, "Reading in DETMAP extension from CHDU")

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, detmap%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if

    call fitswrap_read_key(unit, detmap%key_naxis1, ierr)
    call fitswrap_read_key(unit, detmap%key_naxis2, ierr)

    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading NAXIS1,2 keywords")
      return
    end if

    !- check that vital keader keywords are present - TODO
    ! HDUCLASS= 'OGIP    '
    ! HDUCLAS1= 'IMAGE   '
    ! HDUCLAS2= 'DETMAP  '

!    ! read the other keywords
!    call  fitswrap_read_key(unit, detmap%key_cdelt1,         ierr)
!    if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, detmap%key_cdelt2,         ierr)
!    if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, detmap%key_xraw_ax,         ierr)
!    if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, detmap%key_yraw_ax,         ierr)
!    if ( ierr .ne. ERROR_LEVEL_OK ) return

!    ! massive fudge step to correct for the different optical axis values in the SIXTE simulations
!    call wrschatty (subroutinename, "Fudging the optical-axis location to match SIXTE datasets!")
!    detmap%key_xraw_ax%value = 192.5d0
!    detmap%key_yraw_ax%value = 192.5d0

    ! copy the naxis keyword values to an easier location
!    detmap%naxis1 = detmap%hduinfo%key_naxis1%value
!    detmap%naxis2 = detmap%hduinfo%key_naxis2%value
    detmap%naxis1 = detmap%key_naxis1%value
    detmap%naxis2 = detmap%key_naxis2%value

!    detmap%oa_rawx  = detmap%key_xraw_ax%value
!    detmap%oa_rawy  = detmap%key_yraw_ax%value
!
!    detmap%pixscale_x = detmap%key_cdelt1%value
!    detmap%pixscale_y = detmap%key_cdelt2%value

!    if ( abs(detmap%pixscale_x) .gt. 0.0d0 ) then
!      detmap%inv_pixscale_x = 1.0d0 / detmap%pixscale_x
!    else
!      ierr = ERROR_LEVEL_SEVERE
!      call wrserr (subroutinename, ierr, "Problem interpreting detmap keyword: CDELT1")
!      return
!    end if
!
!    if ( abs(detmap%pixscale_y) .gt. 0.0d0 ) then
!      detmap%inv_pixscale_y = 1.0d0 / detmap%pixscale_y
!    else
!      ierr = ERROR_LEVEL_SEVERE
!      call wrserr (subroutinename, ierr, "Problem interpreting detmap keyword: CDELT2")
!      return
!    end if


    ! now allocate some space
    if ( (detmap%naxis1 .ge. 1) .and.  (detmap%naxis2 .ge. 1) ) then
      allocate(detmap%pmap(detmap%naxis1,detmap%naxis2), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return


!      write (str_buffer, '(A,I0,A,I0,A,A,A,A,A,A,F6.2,A,F6.2,A,A,F8.3,A,F8.3,A)') &
!          "Reading in: ", detmap%naxis1, "x", detmap%naxis2, " pixels of DETMAP info from extname='", &
!          trim(detmap%hduinfo%key_extname%value), "' instrument='", &
!          trim(detmap%hduinfo%key_instrume%value), "' ", &
!          "pixscale=[", detmap%pixscale_x*deg2arcsec, ",", detmap%pixscale_y*deg2arcsec, "] arcsec/pix ", &
!          "raw_ax=[", detmap%oa_rawx, ",", detmap%oa_rawy, "]"
!      call wrschatty (subroutinename, str_buffer)
      call wrschatty (subroutinename, &
          "Reading in: " // trim(i2s(detmap%naxis1)) // "x" // trim(i2s(detmap%naxis2)) // &
          " pixels of DETMAP info from extname='" // trim(detmap%hduinfo%key_extname%value) // &
          "' instrument='" //  trim(detmap%hduinfo%key_instrume%value) // "' ")



      ! now read the pixel data
!      call FTGPVJ(unit,group,1,detmap%naxis1*detmap%naxis2,nullval, detmap%pmap(:,:),anyf,status)
!      call FTGPVE(unit,group,1,detmap%naxis1*detmap%naxis2,nullval, detmap%pmap(:,:),anyf,status)
      call FTG2DE_w(unit, group, nullval, detmap%naxis1, detmap%naxis1, detmap%naxis2, &
          flatten(detmap%pmap(:,:)), anyf, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Error reading DETMAP data array")
        return
      end if

!      ! write out a few values from the detmap to check we have everything right
!      i=1;  j=1;   write (*,'(A,1x,I8,1x,I8,1x,F10.4)'), "DEBUG_DETMAP", i, j, detmap%pmap(i,j)
!      i=10; j=10;  write (*,'(A,1x,I8,1x,I8,1x,F10.4)'), "DEBUG_DETMAP", i, j, detmap%pmap(i,j)
!      i=10; j=100; write (*,'(A,1x,I8,1x,I8,1x,F10.4)'), "DEBUG_DETMAP", i, j, detmap%pmap(i,j)
!      i=100;j=10;  write (*,'(A,1x,I8,1x,I8,1x,F10.4)'), "DEBUG_DETMAP", i, j, detmap%pmap(i,j)
!      i=100;j=100; write (*,'(A,1x,I8,1x,I8,1x,F10.4)'), "DEBUG_DETMAP", i, j, detmap%pmap(i,j)
!      i=200;j=200; write (*,'(A,1x,I8,1x,I8,1x,F10.4)'), "DEBUG_DETMAP", i, j, detmap%pmap(i,j)


    else
      call wrschatty (subroutinename, "strange detmap image dimensions: (" // &
          trim(i2s(detmap%naxis1)) // "x" // trim(i2s(detmap%naxis2)) // ")")
    end if



    !----------------------------------------------
    ! now report the detmap info
    call srctool_detmap_report_detmap_info(detmap)
    !----------------------------------------------

    detmap%is_valid = .true.

  end subroutine srctool_detmap_read_detmap_from_chdu
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine srctool_detmap_report_detmap_info(detmap)
    character(len=*), parameter :: subroutinename='srctool_deadcor_report_detmap_info'
    type(detmap_type), intent(in)    :: detmap

    character(len=STR_MAX) :: str_buffer

!    write (str_buffer, '(A12,1x,A6,6(1x,A8))'), &
!        "#DetmapInfo", "Inst", "NAXIS1", "NAXIS2", "CDELT1", "CDELT2", "XRAW_AX", "YRAW_AX"
!    call wrschatty(subroutinename, str_buffer)
!    write (str_buffer, '(A12,1x,A6,2(1x,I8),2(1x,F8.6),2(1x,F8.2))') &
!        "DetmapInfo", trim(detmap%hduinfo%key_instrume%value), &
!        detmap%naxis1, detmap%naxis2, detmap%key_cdelt1%value, detmap%key_cdelt2%value, &
!        detmap%key_xraw_ax%value, detmap%key_yraw_ax%value
!    call wrschatty(subroutinename, str_buffer)

    write (str_buffer, '(A12,1x,A6,2(1x,A8))') &
        "#DetmapInfo", "Inst", "NAXIS1", "NAXIS2"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A12,1x,A6,2(1x,I8))') &
        "DetmapInfo", trim(detmap%hduinfo%key_instrume%value), &
        detmap%naxis1, detmap%naxis2
    call wrschatty(subroutinename, str_buffer)
  end subroutine srctool_detmap_report_detmap_info
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------END of DETMAP file handling  -------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------





end module srctool_detmap
