! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_generic_avign

  !-----------------------------------------------------------------------------
  ! Summary:

  ! Originally:
  !
  !   A stand-alone module of routines for handling computation of
  !   the analytic vignetting function for eROSITA
  !   See here for details: https://wiki.mpe.mpg.de/eRosita/EroCat/AnalyticVignetting
  !
  ! Now updated to use Konrad Dennerl's interpolation function by Jeremy Sanders

  !-----------------------------------------------------------------------------

  !-----------------------------------------------------------------------------
  !
  !  AVIGN CALDB FILE STRUCTURE
  !  ==========================
  !
  ! FITS Binary table HDU with the following columns:
  !
  ! ColName Format   Units     Description
  ! COEFFS      nD*  []        Array of coefficients defining the vignetting function
  ! OA_X        1D   [pix]     Pixel coordinates of energy dependent optical axis - x-axis
  ! OA_Y        1D   [pix]     Pixel coordinates of energy dependent optical axis - y-axis
  ! PIXSCALE_X  1D   [deg/pix] Pixel scale of detector in x-axis
  ! PIXSCALE_Y  1D   [deg/pix] Pixel scale of detector in y-axis
  !
  ! * n is an integer >=1 depending on the analytic form of the vignetting function
  !
  ! Each row of the table gives the set of params+coefficients describing the vignetting
  ! function at a given monochromatic energy
  !
  ! Note that column order is not important and column names are case-insensitive
  !-------------------------------------------------------------------------------


  !--------------------------------------------------------------------------------
  !
  !  HOW TO USE THESE ROUTINES
  !  =========================
  !
  ! 1) Include this module in the calling module:
  !
  !  use srctool_generic_avign     ! use generic analytic vignetting model routines
  !
  !  And make sure that this file is listed in the Makefile modules variable for your task
  !
  ! 2) At the beginning of the calling routine define an avign data type (one for each instrument):
  !
  !    type(avign_type)   :: avign             ! an avign for a single instrument
  !
  !    alternatively:
  !
  !    type(avign_type), dimension(7) :: avign_list  ! array of avigns for all instruments
  !
  ! 3) Always initialise the avign data structure before use:
  !
  !    call avign_empty_avign(avign)
  !
  ! 4) Read the avign coefficients from the CALDB for each instrument (giving the instrument name,
  !    and the observation start+end date strings) and check the value of the resulting ierr status flag
  !
  !    call avign_read_avign_from_caldb(avign, instname, date_obs, date_end, ierr)
  !
  !    if ( ierr /= 0 ) then
  !      call wrserr (subroutinename, ierr, "Problem reading AVIGN from CALDB")
  !      return
  !    end if
  !
  !    ierr has the standard defined eSASS values:
  !    0 -> ERROR_LEVEL_OK
  !    1 -> ERROR_LEVEL_WARNING
  !    2 -> ERROR_LEVEL_ERROR
  !    3 -> ERROR_LEVEL_SEVERE
  !
  ! 5) [Optional] Report the contents of the AVIGN calibration file to stdout
  !
  !    call avign_report_avign_info(avign)
  !
  ! 6) Call the relevant routine to compute the value of the vignetting function for a given
  !    pixel position (raw coordinate system) and photon energy (keV):
  !
  !    call avign_calc_vign_at_xye(avign, x, y, energy, vign, ierr)
  !
  ! This is actually an interface to four versions of the subroutine, allowing the
  ! input parameters x,y and energies to be scalars or 1D arrays.
  !    vign receives the computed values, and so must have the following dimensions:
  !    a) x,y and e are scalars -> vign is a scalar
  !    b) x,y are scalars, energy is a 1D array -> vign must be a 1D array of length equal size(energy)
  !    c) x,y are 1D arrays, energy is a scalar -> vign must be a 1D array of length equal size(x) (=size(y))
  !    d) x,y are 1D arrays, energy is a 1D array -> vign must be a 2D array with shape (size(x),size(energy))

  ! 7) Repeat step 6) as required
  !
  ! 8) when finished, free up the allocated memory
  !
  !    call avign_empty_avign(avign)
  !
  !-------------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param           ! standard eROSITA module handling parameter/logging/error interface
  use ero_general     ! standard eROSITA module handling flag definitions etc

  ! for safe error reporting
  use srctool_generic, only: wrserr, wrschatty, wrslog, flatten
  use srctool_generic_fitsraw

  implicit none     ! insist on explicit typing

  !-------------------------------------------------------------------------------
  !some private constants
  character(len=*),    private, parameter :: modulename='SRCTOOL_GENERIC_AVIGN_MOD'
  integer(kind=int32), private, parameter :: STR_MAX=256     ! normal string buffer length
!  integer(kind=int32), private, parameter :: STR_MAX=512     ! normal string buffer length
  real(kind=real64), private, parameter :: EMIN_HARDLIMIT = 0.0d0   ! keV
  real(kind=real64), private, parameter :: EMAX_HARDLIMIT = 15.0d0 ! keV

  !standard wr error levels
  integer(kind=int32), private, parameter :: ERROR_LEVEL_OK      = 0
  integer(kind=int32), private, parameter :: ERROR_LEVEL_WARNING = 1
  integer(kind=int32), private, parameter :: ERROR_LEVEL_ERROR   = 2
  integer(kind=int32), private, parameter :: ERROR_LEVEL_SEVERE  = 3
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! CFITSIO constants - annoying that these are not available by default
  integer(kind=int32), private, parameter :: FITSIO_READWRITE = 1
  integer(kind=int32), private, parameter :: FITSIO_READONLY  = 0

  logical,             private, parameter :: FITSIO_CASESENS   = .true.
  logical,             private, parameter :: FITSIO_CASEINSENS = .false.

  ! maximum length of a string header keyword
  integer(kind=int32), private, parameter :: FITSIO_KEY_MAXLEN = 72
  !-------------------------------------------------------------------------------

  ! number of coefficients in interpolation scheme
  integer(kind=int32), private, parameter :: num_coeffs = 21

  ! these parameters are converted into this number of parameters for each energy
  integer(kind=int32), private, parameter :: num_epars = 9

  public

  ! subroutine declarations and interfaces for module: srctool_generic_avign
  private :: i2s
  private :: clip
  private :: check_fitsio_iserror
  public  :: avign_empty_avign
  public  :: avign_read_avign_from_chdu
  public  :: avign_read_avign_from_file
  public  :: avign_read_avign_from_caldb
  public  :: avign_report_avign_info
  private :: avign_make_cache
  private :: avign_calc_weights

  private  :: avign_calc_vign_at_xye_arrays
  private  :: avign_calc_vign_at_xye_earray
  private  :: avign_calc_vign_at_xye_xyarray
  private  :: avign_calc_vign_at_xye_scalar

  private  :: intpar
  private  :: df4
!  public  :: binary_search

  interface avign_calc_vign_at_xye
    module procedure avign_calc_vign_at_xye_arrays
    module procedure avign_calc_vign_at_xye_earray
    module procedure avign_calc_vign_at_xye_xyarray
    module procedure avign_calc_vign_at_xye_scalar
  end interface avign_calc_vign_at_xye

  !-------------------------------------------------------------------------------
  ! A structure to hold the parameters of the analytic vignetting
  ! description for a single eROSITA telescope module
  !
  ! We assume that the radial+azimuthal vignetting function
  ! is well described at a number of discrete energies, and that
  ! linear interpolation is used between these values
  !
  type :: avign_type
     ! flag used to show if this structure is ready to be used
     logical           :: is_valid

     ! A 2d array of coefficients which describes the vignetting function at each energy
     ! as a function of the off-axis angle and the azimuthal angles
     ! first axis is energy, second is index of the coefficient
     real(kind=real64) :: coeffs(num_coeffs)

     ! valid ranges of calculation (keV)
     real(kind=real64) :: e_min, e_max

     ! energies that the vignetting is sampled at (keV)
     real(kind=real64), dimension(:), allocatable :: e_sample

     ! number of pixels on detector
     integer(kind=int32) :: n_x, n_y

     ! The location of the optical axis (for the purposes of vignetting)
     ! in the raw_x,raw_y detector pixel coordinate system
     real(kind=real64) :: oa_x, oa_y  ! physical CCD pixels

     ! used to convert between raw_x,raw_y coords (pixels) and off-axis angle (deg)
     real(kind=real64) :: pixscale_x, pixscale_y  ! deg/pixel
     real(kind=real64) :: pixscale_x_arcmin, pixscale_y_arcmin  ! arcmin/pixel

     ! cache results at sample energies over CCD
     real(kind=real32), dimension(:,:,:), allocatable :: cache
  end type avign_type


contains

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------Start of AVIGN algorithmic routines-------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  ! This is Konrad Dennerl's code to compute the vignetting, translated to F90
  !-------------------------------------------------------------------------------------


  !--------------------------------------------------------------------------------------
  ! compute parameters for a particular energy
  !--------------------------------------------------------------------------------------
  recursive subroutine intpar(dp, dkev, dpar)

    ! input parameters for TM
    real(kind=real64), intent(in) :: dp(num_coeffs)
    ! energy (keV) to evaluate for
    real(kind=real64), intent(in) :: dkev
    ! output energy parameters
    real(kind=real64), intent(out) :: dpar(num_epars)

    real(kind=real64) :: dkevl

    dkevl = log(dkev)

    dpar(1) = dp(1) + ( dp(2) + dp(3) * dkevl ) * dkevl + &
         dp(4) * dkevl**3
    dpar(2) = dp(5) + ( dp(6) + dp(7) * dkev ) * dkev

    if (dkev < 3.d0) then
       dpar(3) = dp(8)  + dp(9)  * dkevl
    else
       dpar(3) = dp(10) + dp(11) * dkevl
    end if

    dpar(4) = dp(12) + ( dp(13) + dp(14) * dkev ) * dkev
    dpar(5) = dp(15) + ( dp(16) + dp(17) * dkev ) * dkev

    dpar(6) = dp(18)
    dpar(7) = dp(19)
    dpar(8) = dp(20)
    dpar(9) = dp(21)

  end subroutine intpar

  !--------------------------------------------------------------------------------------
  ! Compute the vignetting given position and energy
  !--------------------------------------------------------------------------------------
  real(kind=real64) recursive function df4(dpar, dx, dy, dkev, dr)

    ! input parameters from intpar
    real(kind=real64), intent(in) :: dpar(num_epars)
    ! offset dx,dy (arcmin)
    real(kind=real64), intent(in) :: dx, dy
    ! input energy (keV)
    real(kind=real64), intent(in) :: dkev
    ! output offset (arcmin)
    real(kind=real64), intent(out) :: dr

    real(kind=real64) :: drl, drr
    real(kind=real64) :: dyl, dyr, dyll, dyrr
    real(kind=real64) :: da, db, dc, dd, dh1, dh2
    real(kind=real64) :: dx0, dy0

    real(kind=real64), parameter :: deps = 1.d-3

    dx0 = dpar(4)
    dy0 = dpar(5)

    drl = dpar(8) - dkev * dpar(9)
    drr = dpar(7) - dkev * dpar(6)
    dr = sqrt( (dx-dx0)**2 + (dy-dy0)**2 )

    if (dr < drl) then

       df4 = exp(-(dr*dpar(1))**2)

    else if (dr > drr) then

       df4 = dpar(2) * dr**dpar(3)

    else

       dyl = exp(-(drl*dpar(1))**2)
       dyr = dpar(2) * drr**dpar(3)

       dyll = (exp(-((drl+deps)*dpar(1))**2)-dyl)/deps
       dyrr = (dpar(2)*(drr+deps)**dpar(3)-dyr)/deps

       dh1 = (dyll-dyrr)*(drl**2-drr**2) - &
            2.d0*(dyll*drl-dyrr*drr-dyl+dyr)*(drl-drr)
       dh2 = 3.d0*(drl**2-drr**2)**2 - 4.d0*(drl**3-drr**3)*(drl-drr)

       dd = dh1 / dh2
       dc = ( dyll-dyrr-3.d0*dd*(drl**2-drr**2) ) / (2.d0*(drl-drr) )
       db =  dyrr - 2.d0*dc*drr - 3.d0 * dd * drr**2
       da = dyl - db*drl - dc*drl**2 - dd*drl**3

       df4 = da + db*dr + dc*dr**2 + dd*dr**3

    end if

  end function df4


  !--------------------------------------------------------------------------------------
  ! Interface routines are below
  !--------------------------------------------------------------------------------------


  !--------------------------------------------------------------------------------------
  ! calculates the vignetting at raw pixel location x,y, for photon energies e(:) (keV)
  recursive subroutine avign_calc_vign_at_xye_arrays(avign, x, y, e, vign, ierr)
    character(len=*), parameter :: subroutinename='avign_calc_vign_at_xye_arrays'
    type(avign_type), intent(in) :: avign
    real(kind=real64), dimension(:), intent(in)  :: x, y
    real(kind=real64), dimension(:), intent(in)  :: e
    real(kind=real64), dimension(:,:), intent(out)  :: vign
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32) :: ie, ipos

    integer(kind=int32) :: idx
    real(kind=real64) :: weight
    integer(kind=int32) :: ix, iy

    ierr = ERROR_LEVEL_OK

    if ( .not. avign%is_valid ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, &
            "Analytic vignetting datatype was not initialised before use")
       return
    end if

    if ( size(e) /= size(vign(1,:))  .or. &
         size(x) /= size(y)          .or. &
         size(x) /= size(vign(:,1)) ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "input/output array sizes do not match")
       return
    end if

    do ie=1, size(e)
       call avign_calc_weights(avign, e(ie), idx, weight)

       do ipos=1, size(x)
          ix = nint(x(ipos))
          iy = nint(y(ipos))

          if( ix>=1 .and. iy>=1 .and. ix<=avign%n_x .and. iy<=avign%n_y ) then
             vign(ipos, ie) = weight*avign%cache(ix,iy,idx) + &
                  (1-weight)*avign%cache(ix,iy,idx+1)
          else
             vign(ipos, ie) = 0
          end if
       end do
    end do

  end subroutine avign_calc_vign_at_xye_arrays
  !-------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  ! calculates the vignetting at raw pixel location x,y, for photon energies e(:) (keV)
  recursive subroutine avign_calc_vign_at_xye_earray(avign, x, y, e, vign, ierr)
    character(len=*), parameter :: subroutinename='avign_calc_vign_at_xye_earray'
    type(avign_type), intent(in) :: avign
    real(kind=real64), intent(in)  :: x, y
    real(kind=real64), dimension(:), intent(in)  :: e
    real(kind=real64), dimension(:), intent(out)  :: vign
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32) :: ie

    integer(kind=int32) :: idx
    real(kind=real64) :: weight
    integer(kind=int32) :: ix, iy

    ierr = ERROR_LEVEL_OK

    if ( .not. avign%is_valid ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, &
           "Analytic vignetting datatype was not initialised before use")
      return
    end if

    if ( size(e) /= size(vign) ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, &
           "input e(:) and output vign(:) array sizes do not match")
      return
    end if

    ix = nint(x)
    iy = nint(y)
    if(ix<1 .or. iy<1 .or. ix>avign%n_x .or. iy>avign%n_y) then
       vign = 0
    else
       do ie=1, size(e)
          call avign_calc_weights(avign, e(ie), idx, weight)
          vign(ie) = weight*avign%cache(ix, iy, idx) + &
               (1-weight)*avign%cache(ix, iy, idx+1)
       end do
    end if

  end subroutine avign_calc_vign_at_xye_earray
  !-------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  ! calculates the vignetting at list of raw pixel locations x,y, for single photon energy e (keV)
  recursive subroutine avign_calc_vign_at_xye_xyarray(avign, x, y, e, vign, ierr)
    character(len=*), parameter :: subroutinename='avign_calc_vign_at_xye_xyarray'
    type(avign_type), intent(in) :: avign
    real(kind=real64), dimension(:), intent(in)  :: x, y
    real(kind=real64), intent(in)  :: e
    real(kind=real64), dimension(:), intent(out)  :: vign
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32) :: ipos

    integer(kind=int32) :: idx
    real(kind=real64) :: weight
    integer(kind=int32) :: ix, iy

    ierr = ERROR_LEVEL_OK

    if ( .not. avign%is_valid ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, &
           "Analytic vignetting datatype was not initialised before use")
      return
    end if

    if (size(x) /= size(y)  .or. &
        size(x) /= size(vign)) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "input/output array sizes do not match")
      return
    end if

    call avign_calc_weights(avign, e, idx, weight)

    do ipos=1, size(x)
       ix = nint(x(ipos))
       iy = nint(y(ipos))
       if( ix>=1 .and. iy>=1 .and. ix<=avign%n_x .and. iy<=avign%n_y ) then
          vign(ipos) = weight*avign%cache(ix,iy,idx) + &
               (1-weight)*avign%cache(ix,iy,idx+1)
       else
          vign(ipos) = 0
       end if
    end do

  end subroutine avign_calc_vign_at_xye_xyarray
  !-------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  ! calculates the vignetting at single raw pixel location x,y, for single photon energy e (keV)
  recursive subroutine avign_calc_vign_at_xye_scalar(avign, x, y, e, vign, ierr)
    character(len=*), parameter :: subroutinename='avign_calc_vign_at_xye_scalar'
    type(avign_type), intent(in) :: avign
    real(kind=real64), intent(in)  :: x, y
    real(kind=real64), intent(in)  :: e
    real(kind=real64), intent(out)  :: vign
    integer(kind=int32), intent(out)   :: ierr
    real(kind=real64) :: vign_arr(1)

    call avign_calc_vign_at_xye_earray(avign, x, y, [e], vign_arr(:), ierr)
    vign = vign_arr(1)
  end subroutine avign_calc_vign_at_xye_scalar
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !---------------------------End of AVIGN algorithmic routines-------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------Start of AVIGN file/type handling---------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine avign_empty_avign(avign)
    type(avign_type), intent(inout)      :: avign

    avign%coeffs = 0
    avign%e_min = 0
    avign%e_max = 0
    avign%n_x = 0
    avign%n_y = 0
    avign%oa_x = 0
    avign%oa_y = 0
    avign%pixscale_x = 0
    avign%pixscale_y = 0
    avign%pixscale_x_arcmin = 0
    avign%pixscale_y_arcmin = 0

    if(allocated(avign%e_sample)) deallocate(avign%e_sample)
    if(allocated(avign%cache)) deallocate(avign%cache)

    avign%is_valid = .false.
  end subroutine avign_empty_avign
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! cache vignetting in pixels for each sample energy
  subroutine avign_make_cache(avign)
    type(avign_type), intent(inout)   :: avign

    real(kind=real64) :: dpar(num_epars)
    integer(kind=int32) :: ie, ix, iy
    real(kind=real64) :: dx, dy, dr

    allocate(avign%cache(avign%n_x, avign%n_y, size(avign%e_sample)))

    ! for each energy and pixel calculate vignetting
    do ie=1, size(avign%e_sample)

       call intpar(avign%coeffs, avign%e_sample(ie), dpar)

       do iy=1, avign%n_y
          do ix=1, avign%n_x
             dx = (ix-avign%oa_x) * avign%pixscale_x_arcmin
             dy = (iy-avign%oa_y) * avign%pixscale_y_arcmin
             avign%cache(ix, iy, ie) = &
                  real(df4(dpar, dx, dy, avign%e_sample(ie), dr), real32)
          end do
       end do
    end do

  end subroutine avign_make_cache

  !-----------------------------------------------------------------------------
  ! calculate index into arrays and weighting factors for interpolation
  ! given energy
  !
  ! idx: lower index to weight
  ! weight: weighting factor for that index
  ! use idx+1, 1-weight for other factor
  !-----------------------------------------------------------------------------
  subroutine avign_calc_weights(avign, energy, idx, weight)

    type(avign_type), intent(in)   :: avign
    real(kind=real64), intent(in) :: energy

    ! output array index
    integer(kind=int32), intent(out) :: idx
    ! output array weight
    real(kind=real64), intent(out) :: weight

    integer(kind=int32) :: i
    real(kind=real64) :: clip_e

    clip_e = clip(energy, avign%e_min, avign%e_max)

    associate( evals => avign%e_sample )
      if(clip_e <= evals(1)) then
         idx = 1
         weight = 1
      else
         ! linear scan is efficient for short arrays
         ! TODO: see whether binary search is better
         i=1
         do while (evals(i) < clip_e .and. i < size(evals))
            i = i+1
         end do
         if(evals(i) < clip_e) then
            ! clip at end
            idx = size(evals)-1
            weight = 0
         else
            idx = i-1
            weight = (evals(i)-clip_e)/(evals(i)-evals(i-1))
         end if
      end if
  end associate

  end subroutine avign_calc_weights

  !-------------------------------------------------------------------------------
  ! read the avign description from the current hdu of the fits file identified by unit
  subroutine avign_read_avign_from_chdu(unit, avign, ierr)
    character(len=*), parameter :: subroutinename='avign_read_avign_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(avign_type), intent(inout)   :: avign
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: num_rows
    integer(kind=int32) :: status
    integer(kind=int32) :: file_num_coeffs, dummy_int
    integer(kind=int32) :: num_samples
    integer(kind=int32) :: colnum_coeffs, colnum_e_sample
    integer(kind=int32) :: colnum_n_x, colnum_n_y
    integer(kind=int32) :: colnum_e_min, colnum_e_max
    integer(kind=int32) :: colnum_oa_x, colnum_oa_y, colnum_pixscale_x, colnum_pixscale_y
    integer(kind=int32) :: colfmt_coeffs, colfmt_e_sample
    integer(kind=int32) :: anyf
    real(kind=real64), parameter :: huge_real64 = huge(huge_real64)

    ierr = ERROR_LEVEL_OK
    status = 0
!    call wrschatty (subroutinename, "Reading in AVIGNET extension from CHDU")

    ! get the hdu info/keys
!    call fitswrap_read_hduinfo(unit, avign%hduinfo, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if
    !- check the following keader keywords are present - TODO
    !- HDUCLASS       'OGIP'
    !- HDUCLAS1       'RESPONSE'
    !- HDUCLAS2       'AVIGNET2'

    ! get the number of rows in the extension - this is the number of energies
    call FTGNRW_w(unit, num_rows, status)
    if (check_fitsio_iserror(subroutinename, status, ierr)) return

    ! check that have at least some entries and then read in the column data
    if ( num_rows /= 1 ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, &
            "Analytic vignetting file should only contain one row")
       return
    end if

    ! get column numbers
    call FTGCNO_w(unit, 0, "COEFFS",     colnum_coeffs,     status)
    call FTGCNO_w(unit, 0, "E_MIN",      colnum_e_min,      status)
    call FTGCNO_w(unit, 0, "E_MAX",      colnum_e_max,      status)
    call FTGCNO_w(unit, 0, "E_SAMPLE",   colnum_e_sample,   status)
    call FTGCNO_w(unit, 0, "N_X",        colnum_n_x,        status)
    call FTGCNO_w(unit, 0, "N_Y",        colnum_n_y,        status)
    call FTGCNO_w(unit, 0, "OA_X",       colnum_oa_x,       status)
    call FTGCNO_w(unit, 0, "OA_Y",       colnum_oa_y,       status)
    call FTGCNO_w(unit, 0, "PIXSCALE_X", colnum_pixscale_x, status)
    call FTGCNO_w(unit, 0, "PIXSCALE_Y", colnum_pixscale_y, status)
    if (check_fitsio_iserror(subroutinename, status, ierr)) return

    ! get dimensions of the COEFFS column
    call FTGTCL_w(unit, colnum_coeffs, colfmt_coeffs, file_num_coeffs, dummy_int, status)
    if (check_fitsio_iserror(subroutinename, status, ierr)) return

    ! enforce correct number of coefficients
    if (file_num_coeffs /= num_coeffs) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Incorrect number of columns")
       return
    end if

    ! get dimensions of the E_SAMPLE column
    call FTGTCL_w(unit, colnum_e_sample, colfmt_e_sample, num_samples, dummy_int, status)
    if (check_fitsio_iserror(subroutinename, status, ierr)) return

    ! enforce correct number of coefficients
    if (num_samples < 2) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Too few sample energies")
       return
    end if
    allocate(avign%e_sample(num_samples))

    !------------------------------------------------------------------------------
    ! read the data into variables or arrays
    call FTGCVD_w(unit,colnum_e_min,     1,1,1,huge_real64,flatten(avign%e_min),     anyf,status)
    call FTGCVD_w(unit,colnum_e_max,     1,1,1,huge_real64,flatten(avign%e_max),     anyf,status)
    call FTGCVJ_w(unit,colnum_n_x ,      1,1,1,-1,         flatten(avign%n_x),       anyf,status)
    call FTGCVJ_w(unit,colnum_n_y ,      1,1,1,-1,         flatten(avign%n_y),       anyf,status)
    call FTGCVD_w(unit,colnum_oa_x,      1,1,1,huge_real64,flatten(avign%oa_x),      anyf,status)
    call FTGCVD_w(unit,colnum_oa_y,      1,1,1,huge_real64,flatten(avign%oa_y),      anyf,status)
    call FTGCVD_w(unit,colnum_pixscale_x,1,1,1,huge_real64,flatten(avign%pixscale_x),anyf,status)
    call FTGCVD_w(unit,colnum_pixscale_y,1,1,1,huge_real64,flatten(avign%pixscale_y),anyf,status)

    ! read multi-value columns
    call FTGCVD_w(unit,colnum_coeffs,  1,1,num_coeffs, huge_real64, &
         avign%coeffs(:),  anyf,status)
    call FTGCVD_w(unit,colnum_e_sample,1,1,num_samples,huge_real64, &
         avign%e_sample(:),anyf,status)

    !------------------------------------------------------------------------------
    if (check_fitsio_iserror(subroutinename, status, ierr)) return

    ! convert to arcmin
    avign%pixscale_x_arcmin = avign%pixscale_x * 60
    avign%pixscale_y_arcmin = avign%pixscale_y * 60

    ! construct cache
    call avign_make_cache(avign)

    !----------------------------------------------
    ! ! now report the scope of the avignetting data
    !    call avign_report_avign_info(avign)
    !----------------------------------------------

    avign%is_valid = .true.

  end subroutine avign_read_avign_from_chdu
  !---------------------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  ! read the avign description from the current hdu of the fits file identified by unit
  subroutine avign_read_avign_from_file(filename, avign, ierr)
    character(len=*), parameter :: subroutinename='avign_read_avign_from_file'
    character(len=*),    intent(in)   :: filename
    type(avign_type), intent(inout)   :: avign
    integer(kind=int32), intent(out)  :: ierr
    integer(kind=int32) :: status
    integer(kind=int32) :: unit

    status = 0
    ierr = ERROR_LEVEL_OK

    call wrschatty (subroutinename, "Reading cal/meta data (AVIGN2) from file: " &
         // trim(filename))

    ! get an i/o unit
    call FTGIOU_w(unit, status)
    if (check_fitsio_iserror(subroutinename, status, ierr)) return


    ! open the file and move to whichever extension is specified in filename
    call FTDOPN_w(unit, trim(filename), FITSIO_READONLY, status)
    if (check_fitsio_iserror(subroutinename, status, ierr)) return


    ! now read in the data
    call avign_read_avign_from_chdu(unit, avign, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading AVIGN2 from file: " &
           // trim(filename))
      return
    end if

    ! tidy up
    ! close the file
    call FTCLOS_w(unit, status)
    if (check_fitsio_iserror(subroutinename, status, ierr)) return

    ! release the unit number
    call FTFIOU_w(unit, status)
    if (check_fitsio_iserror(subroutinename, status, ierr)) return

    call wrschatty (subroutinename, &
         "Finished reading cal/meta data (AVIGN2) from file: " // trim(filename))


  end subroutine avign_read_avign_from_file
  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  ! read the avign description from the current hdu of the fits file identified by unit
  subroutine avign_read_avign_from_caldb(avign, instname, date_obs, date_end, ierr)
    character(len=*), parameter :: subroutinename='avign_read_avign_from_caldb'
    type(avign_type), intent(inout)   :: avign
    character(len=*),    intent(in)   :: instname
    character(len=*),    intent(in)   :: date_obs
    character(len=*),    intent(in)   :: date_end
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status, i
    character(len=STR_MAX) :: filename

    character(len=*), parameter :: str_telescop = "eROSITA"
    character(len=*), parameter :: str_detnam = "-"
    character(len=*), parameter :: str_filter = "NONE"
    character(len=*), parameter :: str_codename = "AVIGNET2"
    character(len=*), parameter :: str_expression = "-"

    character(len=6) :: cenv
    integer(kind=int32) :: chatter
    character(len=10) :: start_date, start_time, stop_date, stop_time

    integer(kind=int32), parameter :: group = 1, maxret = 3
    character(len=STR_MAX), dimension(maxret) :: pfilename, ponline
    integer(kind=int32), dimension(maxret) :: pextno
    integer(kind=int32) :: nret, nfound

    ierr = ERROR_LEVEL_OK

    ! SASS verbosity -> caldb verbosity
    call get_environment_variable("SASS_VERBOSITY", cenv)
    if (cenv == "sparse") then
       chatter = 1
    else if (cenv == "chatty") then
       !      chatter = 5  ! this doesn't seem to add any useful information
       chatter = 3
    else
       chatter = 2
    end if

    if ( len(date_obs) < 19 .or. len(date_end) < 19 ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr(subroutinename, ierr,  "Illegal date-obs/date-end string(s)")
    end if

    ! stdheader times -> format expected by caldb
    start_date(:) = date_obs(1:10)
    stop_date(:)  = date_end(1:10)
    start_time(:) = date_obs(12:19)
    stop_time(:)  = date_end(12:19)

    ! now get the caldb filenames+extensions
    nret = 0
    nfound = 0
    call gtcalf( chatter, &
         str_telescop, &
         trim(instname), &
         str_detnam, &
         str_filter, &
         str_codename, &
         start_date, start_time, stop_date, stop_time, &
         str_expression, &
         maxret, &
         pfilename, &
         pextno, &
         ponline, &
         nret, &
         nfound, &
         status)

    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr(subroutinename, ierr, &
            "Problem reading filename+ext for cal_type='" // trim(str_codename) // &
            "' for inst='" // trim(instname) // "'")
       return
    end if

    !------------------------------------------------------------------------------
    ! report an error is no CALDB files are found
    if (nret < 1 ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr(subroutinename, ierr, &
            "CALDB does not contain files for cal_type='" // trim(str_codename) // &
          "' for inst='" // trim(instname) // "'")
       return
    end if
    !------------------------------------------------------------------------------
    !------------------------------------------------------------------------------
    !------------------------------------------------------------------------------
    !------------------------------------------------------------------------------
    !------------------------------------------------------------------------------
    !------------------------------------------------------------------------------
    !------------------------------------------------------------------------------

    if (nfound > nret) then
       call wrserr(subroutinename, ERROR_LEVEL_WARNING, &
            "*Warning* More files+extensions (" // trim(i2s(nfound)) // &
            ") were found than could be returned (" // trim(i2s(nret)) // ") for cal_type='" // &
            trim(str_codename) // "' for inst='" // trim(instname) // "'")
    end if

    ! Throw a warning when we read more than one CALDB filename
    if (nret > 1) then
       ! give a more detailed warning for other cal types
       call wrserr(subroutinename, ERROR_LEVEL_WARNING, &
            "POSSIBLE CALDB CONFLICT: More than one filename+ext (" // trim(i2s(nret)) // &
            ") was returned for cal_type='" // &
            trim(str_codename) // "' for inst='" // trim(instname) // "'")
       do i=1,nret
          call wrserr(subroutinename, ERROR_LEVEL_WARNING, &
               "POSSIBLE CALDB CONFLICT: " // trim(i2s(i)) // " '" // &
               trim(pfilename(i)) // "[" // trim(i2s(pextno(i))) // "]'" )
       end do
    end if

    ! we will always use the first entry in the list
    ! append the FITS extension to the chosen filename
    write (filename, '(A,A,I0)')  trim(pfilename(1)), "+",  pextno(1)

    ! now read the file from disk
    call avign_read_avign_from_file(filename, avign, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
       call wrserr (subroutinename, ierr, "Problem reading " // trim(str_codename) // " from CALDB")
       return
    end if

  end subroutine avign_read_avign_from_caldb
  !-------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  ! report the contents of the avign parameter structure
  subroutine avign_report_avign_info(avign)
    character(len=*), parameter :: subroutinename='avign_report_avign_info'
    type(avign_type), intent(in)    :: avign

    integer(kind=int32) :: c
    character(len=STR_MAX) :: str_buffer
!
    call wrschatty(subroutinename, "Report of AVIGN structure.")
    call wrschatty(subroutinename, "Coefficients:")
    do c=1, num_coeffs
       write(str_buffer,'(ES13.6E2)') avign%coeffs(c)
       call wrschatty(subroutinename, str_buffer)
    end do
    write(str_buffer, '(A,F8.3,A,F8.3)') " E_min(keV)=", avign%e_min, &
         " E_max(keV)=", avign%e_max
    call wrschatty(subroutinename, str_buffer)
    write(str_buffer, '(A,9999F6.2)') "E_sample(keV)=", avign%e_sample
    call wrschatty(subroutinename, str_buffer)
    write(str_buffer, '(A,F8.3,A,F8.3)') " OptAx_x=", avign%oa_x, &
         " OptAx_y=", avign%oa_y
    call wrschatty(subroutinename, str_buffer)
    write(str_buffer, '(A,F8.5,A,F8.5)') " deg/pix_x=", avign%pixscale_x, &
         " deg/pix_y=", avign%pixscale_y
    call wrschatty(subroutinename, str_buffer)
!
  end subroutine avign_report_avign_info
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------END of AVIGN file/type handling  ---------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !---------------------------- START OF TEST ROUTINES  -----------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  subroutine avign_test_harness(ierr)
    character(len=*), parameter :: subroutinename='avign_test_harness'
    type(avign_type)   :: avign
    integer(kind=int32), intent(out)  :: ierr
    character(len=STR_MAX) :: str_buffer
    real(kind=real64) :: x,y,v,e
    integer(kind=int32) :: i,j
    character(len=STR_MAX) :: instname
    logical, parameter :: chatty = .false.

    ierr = ERROR_LEVEL_OK

    ! DEBUG test of AVIGN module
    !    call random_seed()
    do i=1,7
       call avign_empty_avign(avign)

       write (instname,'(A,I0)') 'tm', i

       call avign_read_avign_from_caldb(avign, trim(instname), &
            "2010-01-01T00:00:00", "2010-01-02T00:00:00", ierr)
       if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Problem reading AVIGN from CALDB")
          return
       end if

       if ( chatty ) then
          call wrschatty(subroutinename,  "AVIGN read in successfully for inst: " // trim(instname))

          !----------------------------------------------
          ! now report the scope of the avignetting data
          call avign_report_avign_info(avign)
          !----------------------------------------------

          ! Now test a few random positions/energies
          write(str_buffer,'(A,x,A4,x,A4,3(x,A10),x,A12,x,A15)') &
               "#AVIGNET_TEST", "INST", "j", "x", "y", "E(keV)", "Offaxis(deg)", "Vign"
          call wrschatty(subroutinename,str_buffer)
       end if

       do j=1,100000
          call random_number(x)
          call random_number(y)
          call random_number(e)
          x = x * 384.0d0 + 0.5d0
          y = y * 384.0d0 + 0.5d0
          e = 0.01d0 + e*9.99d0
          call avign_calc_vign_at_xye(avign, x, y, e, v, ierr)
          if ( ierr /= ERROR_LEVEL_OK ) then
             call wrserr (subroutinename, ierr, "Problem computing AVIGN")
             return
          end if

          if ( v < 0.0d0 .or. v > 1.0d0 ) then
             ierr = ERROR_LEVEL_SEVERE
             write(str_buffer,'(A,x,A,4(x,F12.4),x,ES15.8E2)') &
                  "Computed AVIGN value is out of valid range [0.0:1.0]: inst,x,y,e,oa,v=", &
                  trim(instname),x,y,e,v
             call wrserr (subroutinename, ierr, str_buffer)
             return
          end if

          if ( chatty ) then
             write(str_buffer,'(A,x,A4,x,I4,3(x,F10.4),x,F12.8,x,ES15.8E2)') &
                  " AVIGNET_TEST", trim(instname), j, x, y, e, v
             call wrschatty(subroutinename,str_buffer)
          end if
       end do
       call avign_empty_avign(avign)

    end do

  end subroutine avign_test_harness

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !---------------------------- END OF TEST ROUTINES  ----------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !---------------------------- START OF UTILITY ROUTINES  -----------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  ! make sure value is between ranges given
  pure recursive real(kind=real64) function clip(val, minval, maxval)
    real(kind=real64), intent(in) :: val, minval, maxval

    clip = max(minval, min(maxval, val))
  end function clip

  !-------------------------------------------------------------------------------------
  ! Convert a 32bit integer to a string inline
  ! - for best results, wrap the resulting string in trim()
  recursive pure character(len=32) function i2s(k)
    integer, intent(in) :: k

    write (i2s, '(I0)') k
    i2s = adjustl(i2s)
  end function i2s


  !-----------------------------------------------------------------------
  ! report the result of an FITSIO operation
  function check_fitsio_iserror (subroutinename, status, ierr)
    character(len=*), intent(in) ::  subroutinename
    integer(kind=int32), intent(in) :: status
    integer(kind=int32), intent(out)  :: ierr
    logical :: check_fitsio_iserror

    character(len=30) :: str_errtext

    check_fitsio_iserror = status /= 0
    if (check_fitsio_iserror) then
      ierr = ERROR_LEVEL_SEVERE
      call FTGERR_w(status, str_errtext)
      call wrserr(trim(subroutinename), ierr, &
          "FITSIO error (status=" // trim(i2s(status)) // "): " // trim(str_errtext))
    end if
  end function check_fitsio_iserror
  !-----------------------------------------------------------------------

  ! !-----------------------------------------------------------------------
  ! !
  ! ! branch-free binary search
  ! !
  ! ! returns insertion position of key to maintain order (past end of
  ! ! array if it is larger than the last element)
  ! !
  ! !-----------------------------------------------------------------------
  ! recursive function binary_search(vector, key)
  !   integer(kind=int32) :: binary_search
  !   real(kind=real32), dimension(:), contiguous, intent(in) :: vector
  !   real(kind=real32), intent(in) :: key

  !   integer(kind=int32) :: vsize, midoff, offset, pow2size, jump

  !   vsize = size(vector)

  !   ! if the array size isn't a power of two, we should do an
  !   ! additional comparison at the start and move the starting
  !   ! position if necessary
  !   pow2size = 2**(bit_size(vsize)-leadz(vsize-1));
  !   if(vsize == pow2size) then
  !      offset = 1
  !   else
  !      midoff = vsize - pow2size/2 + 1
  !      offset = merge(midoff, 1, vector(midoff) <= key)
  !   end if

  !   jump = vsize/2
  !   do while (jump>0)
  !      offset = offset + merge(jump, 0, vector(offset+jump) <= key)
  !      jump = jump / 2
  !   end do

  !   binary_search = merge(1, offset+1, vector(offset) > key)
  ! end function binary_search


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !------------------------------ END OF UTILITY ROUTINES  -----------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------




end module srctool_generic_avign
