! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_generic_fitswrap

!-----------------------------------------------------------------------------
! Summary:
! module containing wrappers to fitsio functions
!
!-----------------------------------------------------------------------------
!  use numkind   ! standard eRosita module giving defined "kind"s
  use iso_fortran_env ! replacement for numkind
  use param           ! standard eRosita module handling parameter/logging/error interface

  use srctool_generic  ! generic module with some universal constants
  use srctool_generic_fitsraw

  implicit none     ! insist on explicit typing

  !-------------------------------------------------------------------------------
  !some constants
  character(len=*), private, parameter :: str_modname='SRCTOOL_GENERIC_FITSWRAP_MOD'

  public
  ! CFITSIO constants - annoying that these are not available by default
  integer(kind=int32), parameter :: FITSIO_READWRITE = 1
  integer(kind=int32), parameter :: FITSIO_READONLY  = 0

  integer(kind=int32), parameter :: FITSIO_IMAGE_HDU  =  0
  integer(kind=int32), parameter :: FITSIO_ASCII_TBL  =  1
  integer(kind=int32), parameter :: FITSIO_BINARY_TBL =  2
  integer(kind=int32), parameter :: FITSIO_ANY_HDU    = -1

  integer(kind=int32), parameter :: FITSIO_BLOCKSIZE  = 1

  integer(kind=int32), parameter :: FITSIO_CASESENS   = 1
  integer(kind=int32), parameter :: FITSIO_CASEINSENS = 0

  integer(kind=int32), parameter :: FITSIO_EXTVER_IGNORE = 0

  integer(kind=int32), parameter :: FITSIO_ERROR_BAD_HDU_NUM = 301
  integer(kind=int32), parameter :: FITSIO_ERROR_KEYWORD_NOEXIST = 202  ! specified keyword name was not found in the header
  integer(kind=int32), parameter :: FITSIO_ERROR_VALUE_UNDEFINED = 204  ! specified keyword value is undefined

  integer(kind=int32), parameter :: FITSIO_KEY_MAXLEN = 72  ! maximum length of a string header keyword

  integer(kind=int32), parameter :: FITSWRAP_KEY_FLAG_BLANK        = 0
  integer(kind=int32), parameter :: FITSWRAP_KEY_FLAG_SUPPLIED     = 1
  integer(kind=int32), parameter :: FITSWRAP_KEY_FLAG_NOT_SUPPLIED = 2

  ! fits table extension column datatypes
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TBIT         =   1
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TBYTE        =  11
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TSBYTE       =  12
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TLOGICAL     =  14
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TSTRING      =  16
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TUSHORT      =  20
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TSHORT       =  21
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TINT         =  31
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TUINT        =  30
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TULONG       =  40
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TLONG        =  41
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TFLOAT       =  42
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TLONGLONG    =  81
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TDOUBLE      =  82
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TCOMPLEX     =  83
  integer(kind=int32), parameter :: FITSIO_COL_DATATYPE_TDBLCOMPLEX  = 163

  integer(kind=int32), parameter :: FITSIO_DECIMALS = -10

  ! This is an obscure string value that is very unlikely to be used in practice
  character(len=*), parameter :: FITSWRAP_SET_KEY_BLANK = "__-_+#@[{$}]_1__"
  ! this is just used for debug printouts
  character(len=*), parameter :: FITSWRAP_UNKNOWN_KEY_VALUE = "__________________________"


  ! subroutine declarations and interfaces for module: srctool_generic_fitswrap
  public  :: fitswrap_check_status
  public  :: fitswrap_write_fits_error
  public  :: fitswrap_test_for_col
  public  :: fitswrap_get_colnum

  private :: fitswrap_prep_col_int8
  private :: fitswrap_prep_col_int16
  private :: fitswrap_prep_col_int32
  private :: fitswrap_prep_col_int64
  private :: fitswrap_prep_col_real32
  private :: fitswrap_prep_col_real64

  private :: fitswrap_proc_col_int8
  private :: fitswrap_proc_col_int16
  private :: fitswrap_proc_col_int32
  private :: fitswrap_proc_col_int64
  private :: fitswrap_proc_col_real32
  private :: fitswrap_proc_col_real64
  private :: fitswrap_proc_vec_col_real32

  public  :: fitswrap_read_hduinfo_quick
  public  :: fitswrap_write_hduinfo
  public  :: fitswrap_write_hduinfo_quiet
  public  :: fitswrap_read_hduinfo

  private :: fitswrap_init_key_logical
  private :: fitswrap_init_key_string
  private :: fitswrap_init_key_vlstring
  private :: fitswrap_init_key_int16
  private :: fitswrap_init_key_int32
  private :: fitswrap_init_key_int64
  private :: fitswrap_init_key_real32
  private :: fitswrap_init_key_real64
  private :: fitswrap_init_key_wcomment_logical
  private :: fitswrap_init_key_wcomment_string
  private :: fitswrap_init_key_wcomment_vlstring
  private :: fitswrap_init_key_wcomment_int16
  private :: fitswrap_init_key_wcomment_int32
  private :: fitswrap_init_key_wcomment_int64
  private :: fitswrap_init_key_wcomment_real32
  private :: fitswrap_init_key_wcomment_real64

  private :: fitswrap_set_key_logical
  private :: fitswrap_set_key_string
  private :: fitswrap_set_key_vlstring
  private :: fitswrap_set_key_int16
  private :: fitswrap_set_key_int32
  private :: fitswrap_set_key_int64
  private :: fitswrap_set_key_real32
  private :: fitswrap_set_key_real64

  private :: fitswrap_set_key_value_logical
  private :: fitswrap_set_key_value_string
  private :: fitswrap_set_key_value_vlstring
  private :: fitswrap_set_key_value_int16
  private :: fitswrap_set_key_value_int32
  private :: fitswrap_set_key_value_int64
  private :: fitswrap_set_key_value_real32
  private :: fitswrap_set_key_value_real64

  private :: fitswrap_set_key_null_logical
  private :: fitswrap_set_key_null_string
  private :: fitswrap_set_key_null_vlstring
  private :: fitswrap_set_key_null_int16
  private :: fitswrap_set_key_null_int32
  private :: fitswrap_set_key_null_int64
  private :: fitswrap_set_key_null_real32
  private :: fitswrap_set_key_null_real64

  private :: fitswrap_read_key_logical
  private :: fitswrap_read_key_string
  private :: fitswrap_read_key_vlstring
  private :: fitswrap_read_key_int16
  private :: fitswrap_read_key_int32
  private :: fitswrap_read_key_int64
  private :: fitswrap_read_key_real32
  private :: fitswrap_read_key_real64

  private :: fitswrap_read_key_quick_string
  private :: fitswrap_read_key_quick_real64

  private :: fitswrap_write_key_logical
  private :: fitswrap_write_key_string
  private :: fitswrap_write_key_vlstring
  private :: fitswrap_write_key_int16
  private :: fitswrap_write_key_int32
  private :: fitswrap_write_key_int64
  private :: fitswrap_write_key_real32
  private :: fitswrap_write_key_real64

  public  :: fitswrap_sprint_key_header
  private :: fitswrap_sprint_key_logical
  private :: fitswrap_sprint_key_string
  private :: fitswrap_sprint_key_vlstring
  private :: fitswrap_sprint_key_int16
  private :: fitswrap_sprint_key_int32
  private :: fitswrap_sprint_key_int64
  private :: fitswrap_sprint_key_real32
  private :: fitswrap_sprint_key_real64

  public  :: fitswrap_print_hduinfo
!  public  :: fitswrap_write_hduinfo_keys_quiet
  public  :: fitswrap_test_if_is_fits

  private :: fitswrap_read_column_real64
  private :: fitswrap_read_column_real32
  private :: fitswrap_read_column_int64
  private :: fitswrap_read_column_int32
  private :: fitswrap_read_column_int16
  private :: fitswrap_read_column_int8

  interface fitswrap_prep_col
!     module procedure fitswrap_prep_col_logical
     module procedure fitswrap_prep_col_int8
     module procedure fitswrap_prep_col_int16
     module procedure fitswrap_prep_col_int32
     module procedure fitswrap_prep_col_int64
     module procedure fitswrap_prep_col_real32
     module procedure fitswrap_prep_col_real64
  end interface

  interface fitswrap_proc_col
!     module procedure fitswrap_proc_col_logical
     module procedure fitswrap_proc_col_int8
     module procedure fitswrap_proc_col_int16
     module procedure fitswrap_proc_col_int32
     module procedure fitswrap_proc_col_int64
     module procedure fitswrap_proc_col_real32
     module procedure fitswrap_proc_col_real64
  end interface

  interface fitswrap_proc_vec_col
     module procedure fitswrap_proc_vec_col_real32
  end interface

  interface fitswrap_read_key
     module procedure fitswrap_read_key_logical
     module procedure fitswrap_read_key_string
     module procedure fitswrap_read_key_vlstring
     module procedure fitswrap_read_key_int16
     module procedure fitswrap_read_key_int32
     module procedure fitswrap_read_key_int64
     module procedure fitswrap_read_key_real32
     module procedure fitswrap_read_key_real64
  end interface

  interface fitswrap_read_key_quick
     module procedure fitswrap_read_key_quick_string
     module procedure fitswrap_read_key_quick_real64
  end interface fitswrap_read_key_quick

  interface fitswrap_write_key
     module procedure fitswrap_write_key_logical
     module procedure fitswrap_write_key_string
     module procedure fitswrap_write_key_vlstring
     module procedure fitswrap_write_key_int16
     module procedure fitswrap_write_key_int32
     module procedure fitswrap_write_key_int64
     module procedure fitswrap_write_key_real32
     module procedure fitswrap_write_key_real64
  end interface

  interface fitswrap_write_key_if_supplied
     module procedure fitswrap_write_key_if_supplied_logical
     module procedure fitswrap_write_key_if_supplied_string
     module procedure fitswrap_write_key_if_supplied_vlstring
     module procedure fitswrap_write_key_if_supplied_int16
     module procedure fitswrap_write_key_if_supplied_int32
     module procedure fitswrap_write_key_if_supplied_int64
     module procedure fitswrap_write_key_if_supplied_real32
     module procedure fitswrap_write_key_if_supplied_real64
  end interface

  interface fitswrap_init_key
     module procedure fitswrap_init_key_logical
     module procedure fitswrap_init_key_string
     module procedure fitswrap_init_key_vlstring
     module procedure fitswrap_init_key_int16
     module procedure fitswrap_init_key_int32
     module procedure fitswrap_init_key_int64
     module procedure fitswrap_init_key_real32
     module procedure fitswrap_init_key_real64
     module procedure fitswrap_init_key_wcomment_logical
     module procedure fitswrap_init_key_wcomment_string
     module procedure fitswrap_init_key_wcomment_vlstring
     module procedure fitswrap_init_key_wcomment_int16
     module procedure fitswrap_init_key_wcomment_int32
     module procedure fitswrap_init_key_wcomment_int64
     module procedure fitswrap_init_key_wcomment_real32
     module procedure fitswrap_init_key_wcomment_real64
  end interface

  interface fitswrap_set_key
     module procedure fitswrap_set_key_logical        ! set all parameters of key
     module procedure fitswrap_set_key_string
     module procedure fitswrap_set_key_vlstring
     module procedure fitswrap_set_key_int16
     module procedure fitswrap_set_key_int32
     module procedure fitswrap_set_key_int64
     module procedure fitswrap_set_key_real32
     module procedure fitswrap_set_key_real64
     module procedure fitswrap_set_key_value_logical   ! only set the value
     module procedure fitswrap_set_key_value_string
     module procedure fitswrap_set_key_value_vlstring
     module procedure fitswrap_set_key_value_int16
     module procedure fitswrap_set_key_value_int32
     module procedure fitswrap_set_key_value_int64
     module procedure fitswrap_set_key_value_real32
     module procedure fitswrap_set_key_value_real64
  end interface

  interface fitswrap_set_key_null
     module procedure fitswrap_set_key_null_logical
     module procedure fitswrap_set_key_null_string
     module procedure fitswrap_set_key_null_vlstring
     module procedure fitswrap_set_key_null_int16
     module procedure fitswrap_set_key_null_int32
     module procedure fitswrap_set_key_null_int64
     module procedure fitswrap_set_key_null_real32
     module procedure fitswrap_set_key_null_real64
  end interface

  interface fitswrap_sprint_key
     module procedure fitswrap_sprint_key_logical
     module procedure fitswrap_sprint_key_string
     module procedure fitswrap_sprint_key_vlstring
     module procedure fitswrap_sprint_key_int16
     module procedure fitswrap_sprint_key_int32
     module procedure fitswrap_sprint_key_int64
     module procedure fitswrap_sprint_key_real32
     module procedure fitswrap_sprint_key_real64
  end interface

  interface fitswrap_read_column
     module procedure fitswrap_read_column_real64
     module procedure fitswrap_read_column_real32
     module procedure fitswrap_read_column_int64
     module procedure fitswrap_read_column_int32
     module procedure fitswrap_read_column_int16
     module procedure fitswrap_read_column_int8
  end interface fitswrap_read_column


!  !----------------------------------------------------------------
!  ! this type holds one column of fits data and associated meta data
!  ! convention is to store all column data (vector or otherwise) in a 1D array
!  ! with length =  num_rows *
!  type fcol_type
!    character(len=FITSIO_KEY_MAXLEN) :: name           ! fits column name matching pattern
!    character(len=FITSIO_KEY_MAXLEN) :: matched_name   ! actual fits column name
!    integer(kind=int32)              :: num     ! fits position of column in HDU
!    integer(kind=int32)              :: anyf
!
!    integer(kind=int32) :: datacode     ! original fits datacode
!    integer(kind=int32) :: kindcode     ! internal data kind code
!
!    integer(kind=int8),  dimension(:), allocatable :: pdata_b
!    integer(kind=int16), dimension(:), allocatable :: pdata_i
!    integer(kind=int32), dimension(:), allocatable :: pdata_j
!    integer(kind=int64), dimension(:), allocatable :: pdata_k
!    real   (kind=real32),dimension(:), allocatable :: pdata_f
!    real   (kind=real64),dimension(:), allocatable :: pdata_d
!    logical           ,  dimension(:), allocatable :: pdata_l
!    character,           dimension(:), allocatable :: pdata_s
!
!    integer(kind=int64) :: num_rows  ! number of rows of data
!    integer(kind=int32) :: width     ! dimension of vector column
!    integer(kind=int32) :: group     ! length of individual groups (for rAw character cols)
!
!
!  end type fcol_type
!  !----------------------------------------------------------------


  !-----------------------------------------------------
  type :: fitskey_string_type
    character(len=FITSIO_KEY_MAXLEN) :: name
    character(len=STR_MAX)           :: value  ! note that we need to write these out with  FTPKLS()
    character(len=FITSIO_KEY_MAXLEN) :: comment
    integer(kind=int32)              :: flag
    logical                          :: compulsory
    logical                          :: is_null
  end type fitskey_string_type

  ! long strings needed for comments/history etc
  type :: fitskey_vlstring_type
    character(len=FITSIO_KEY_MAXLEN) :: name
    character(len=VLSTR_MAX)         :: value  ! note that we need to write these out with  FTPKLS()
    character(len=FITSIO_KEY_MAXLEN) :: comment
    integer(kind=int32)              :: flag
    logical                          :: compulsory
    logical                          :: is_null
  end type fitskey_vlstring_type

  type :: fitskey_int16_type
    character(len=FITSIO_KEY_MAXLEN) :: name
    integer(kind=int16)              :: value
    character(len=FITSIO_KEY_MAXLEN) :: comment
    integer(kind=int32)              :: flag
    logical                          :: compulsory
    logical                          :: is_null
  end type fitskey_int16_type

  type :: fitskey_int32_type
    character(len=FITSIO_KEY_MAXLEN) :: name
    integer(kind=int32)              :: value
    character(len=FITSIO_KEY_MAXLEN) :: comment
    integer(kind=int32)              :: flag
    logical                          :: compulsory
    logical                          :: is_null
  end type fitskey_int32_type

  type :: fitskey_int64_type
    character(len=FITSIO_KEY_MAXLEN) :: name
    integer(kind=int64)              :: value
    character(len=FITSIO_KEY_MAXLEN) :: comment
    integer(kind=int32)              :: flag
    logical                          :: compulsory
    logical                          :: is_null
  end type fitskey_int64_type

  type :: fitskey_real32_type
    character(len=FITSIO_KEY_MAXLEN) :: name
    real(kind=real32)                :: value
    character(len=FITSIO_KEY_MAXLEN) :: comment
    integer(kind=int32)              :: flag
    logical                          :: compulsory
    logical                          :: is_null
  end type fitskey_real32_type

  type :: fitskey_real64_type
    character(len=FITSIO_KEY_MAXLEN) :: name
    real(kind=real64)                :: value
    character(len=FITSIO_KEY_MAXLEN) :: comment
    integer(kind=int32)              :: flag
    logical                          :: compulsory
    logical                          :: is_null
  end type fitskey_real64_type

  type :: fitskey_logical_type
    character(len=FITSIO_KEY_MAXLEN) :: name
    logical                          :: value
    character(len=FITSIO_KEY_MAXLEN) :: comment
    integer(kind=int32)              :: flag
    logical                          :: compulsory
    logical                          :: is_null
  end type fitskey_logical_type
  !-----------------------------------------------------


  !-----------------------------------------------------------------------------
  ! erosita standard header keywords are described here:
  ! http://www2011.mpe.mpg.de/erosita/internal/SASS-devel/eROSITA_TRoPIC_Eventfiles-4eng.html
  !-----------------------------------------------------------------------------
  ! Should appear in events extension and be copied to the primary & extension header of all
  ! data/products files.
  !-----------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! this is used to hold meta info about an input/output fits extension
  type :: hduinfo_type

    ! perhaps surplus but handy to have here
    character(len=LSTR_MAX)    :: filename

    ! now some file meta info that can be easily extracted
    integer(kind=int32)        :: hdu_index
    integer(kind=int32)        :: hdu_type

    ! special case keywords:

    ! now some commonly used standard header keywords - not in standard eROSITA header
    type(fitskey_string_type) :: key_extname
    type(fitskey_string_type) :: key_hduclass
    type(fitskey_string_type) :: key_hduclas1
    type(fitskey_string_type) :: key_hduclas2
    type(fitskey_string_type) :: key_hduclas3
    type(fitskey_string_type) :: key_hduclas4
    type(fitskey_string_type) :: key_hduvers

    ! Keywords from eROSITA standard header are indicated below:
    type(fitskey_string_type) :: key_origin      ! eROSITA STD HDR
    type(fitskey_string_type) :: key_creator     ! eROSITA STD HDR
    type(fitskey_string_type) :: key_owner       ! eROSITA STD HDR
    type(fitskey_string_type) :: key_mission     ! eROSITA STD HDR
    type(fitskey_string_type) :: key_telescop    ! eROSITA STD HDR
    type(fitskey_string_type) :: key_instrume    ! eROSITA STD HDR
    type(fitskey_string_type) :: key_obs_mode    ! eROSITA STD HDR
    type(fitskey_string_type) :: key_datamode    ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_frametim    ! eROSITA STD HDR
    type(fitskey_string_type) :: key_filter      ! eROSITA STD HDR
    type(fitskey_string_type) :: key_skyfield    ! eROSITA STD HDR
    type(fitskey_string_type) :: key_obs_id      ! eROSITA STD HDR
    type(fitskey_string_type) :: key_exp_id      ! eROSITA STD HDR
    type(fitskey_string_type) :: key_observer    ! eROSITA STD HDR
    type(fitskey_string_type) :: key_object      ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_ra_obj      ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_dec_obj     ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_ra_cen      ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_dec_cen     ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_fieldlon    ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_fieldlat    ! eROSITA STD HDR
    type(fitskey_string_type) :: key_date_obs    ! eROSITA STD HDR
    type(fitskey_string_type) :: key_date_end    ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_tstart      ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_tstop       ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_mjdref      ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_timezero    ! eROSITA STD HDR
    type(fitskey_string_type) :: key_timeunit    ! eROSITA STD HDR
    type(fitskey_string_type) :: key_timesys     ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_ra_pnt      ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_dec_pnt     ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_pa_pnt      ! eROSITA STD HDR
    type(fitskey_string_type) :: key_radecsys    ! eROSITA STD HDR
    type(fitskey_real64_type) :: key_equinox     ! eROSITA STD HDR
    type(fitskey_string_type) :: key_longstr     ! eROSITA STD HDR

    !------------------------------------------------------------------------------------
    ! The following key combination is used for merged instrument output
    ! i.e,. when key_instrume = "MERGED"
    ! for HDUs which contain data from a single instrument, should not write key_ninst nor key_instrum
    type(fitskey_int32_type)  :: key_ninst
    type(fitskey_string_type), dimension(1:MAX_INST)  :: pkey_instrum
    !------------------------------------------------------------------------------------

    !------------------------------------------------------------------------------------
    ! These keywords are used to describe the components of the standard filenaming scheme.
    ! They were added to the eROSITA standard header around 18/06/2018
    type(fitskey_string_type) :: key_prefix      ! eROSITA STD HDR
    type(fitskey_string_type) :: key_dscode      ! eROSITA STD HDR
    type(fitskey_string_type) :: key_dstype      ! eROSITA STD HDR
    type(fitskey_string_type) :: key_version     ! eROSITA STD HDR
    type(fitskey_string_type) :: key_config      ! eROSITA STD HDR
    !
    ! from: http://erosita.mpe.mpg.de/eROdoc/pipeline/SASS-products_file_naming_04.html
    ! -----------------------------------------------------------------------------------
    !           1         2         3         4
    !  1234567890123456789012345678901234567890123456
    !  sm01_015056_721_MaxLikSourceList_012_r001.fits
    !  |    |      ||| |                |   ||
    !  |    |      ||| |                |   ||----  processing config
    !  |    |      ||| |                |   |----- "r": release, "i": internal, "t": testing, etc.
    !  |    |--|   ||| |-------|        |
    !  |       |   |||         |        |----- dataset version
    !  ObsMode |   |||===|||   |
    !          |         |||   |-- product type descriptor
    !          SkyMap    |||
    !          /obsID    |||-- energy band
    !                    ||
    !                    ||--- filter wheel position
    !                    |
    !                    |---- telescope ID
    !-----------------------------------------------------------------------------------


    type(fitskey_vlstring_type) :: key_sasshist    ! eROSITA STD HDR


  end type hduinfo_type
  !-------------------------------------------------------------------------------


contains

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !------------------------ SUBROUTINES ------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine fitswrap_write_fits_error(subroutinename, status)

    !---------------------------------------------------------
    ! Writes out a standard error message that contains the
    ! human-readable FITSIO error text
    !---------------------------------------------------------
    character(len=*), intent(in)  :: subroutinename
    integer(kind=int32) , intent(in)  :: status

!    character(len=12) :: str_outbuff
    character(len=30) :: str_errtext

    call FTGERR_w(status, str_errtext)

!    write (str_status, '(A,I0,A2)'), "FITSIO error (status=", status, "): "
!    write (str_outbuff, '(A,I0,A2)'), "FITSIO error (status=", status, "): "
!    write (str_outbuff, '(A,I0,A2,1X,A)'), "FITSIO error (status=", status, "): ",  trim(str_errtext)
    call wrserr(trim(subroutinename), ERROR_LEVEL_SEVERE, &
        "FITSIO error (status=" // trim(i2s(status)) // "): " // trim(str_errtext))

  end subroutine fitswrap_write_fits_error
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------- fitswrap_prep_col ------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! quietly test for existence of column
  ! set exists to true if column exists
  subroutine fitswrap_test_for_col(unit, colname, exists)
    character(len=*), parameter :: subroutinename='fitswrap_test_for_col'
    integer(kind=int32), intent(in)    :: unit
    character(len=*),    intent(in)    :: colname
    logical,             intent(out)   :: exists
    integer(kind=int32) :: status
    integer(kind=int32) :: colnum
!    character(len=STR_MAX) :: str_buffer
    status = 0
    exists = .false.

    ! try to get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
    if ( (status .eq. 0) .and. (colnum .gt. 0) ) then
      exists = .true.
    else
 !     if ( status .ne. 0 ) then
 !       call fitswrap_write_fits_error(subroutinename, status)
 !     end if
      exists = .false.
    end if
!    write(str_buffer, '(A,x,I0,x,I0,x,A)'),  trim(colname), colnum, status, logical2string(exists)
!    call wrschatty(subroutinename, str_buffer)
    return
  end subroutine fitswrap_test_for_col
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! get column number (with error message wrapping
  subroutine fitswrap_get_colnum(unit, colname, colnum, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_get_colnum'
    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32), intent(inout)  :: colnum
    integer(kind=int32), intent(inout)  :: ierr
    integer(kind=int32) :: status

    status = 0
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

  end subroutine fitswrap_get_colnum
  !---------------------------------------------------------------------------------------------




  !-------------------------------------------------------------------------------
  ! get column number, then allocate some space
  ! datatype: integer(kind=int8)
  subroutine fitswrap_prep_col_int8(unit, colname, num_rows, colnum, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_prep_col_int8'
    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    integer(kind=int8), dimension(:), allocatable, intent(inout) :: pdata
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout) :: colnum
    integer(kind=int32), intent(inout)  :: ierr
    integer(kind=int32) :: status

    status = 0
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, trim(colname), str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
  end subroutine fitswrap_prep_col_int8
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! get column number, then allocate some space
  ! datatype: integer(kind=int16)
  subroutine fitswrap_prep_col_int16(unit, colname, num_rows, colnum, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_prep_col_int16'
    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    integer(kind=int16), dimension(:), allocatable, intent(inout) :: pdata
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout) :: colnum
    integer(kind=int32), intent(inout)  :: ierr
    integer(kind=int32) :: status

    status = 0
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, trim(colname), str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
  end subroutine fitswrap_prep_col_int16
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! get column number, then allocate some space
  ! datatype: integer(kind=int32)
  subroutine fitswrap_prep_col_int32(unit, colname, num_rows, colnum, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_prep_col_int32'
    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    integer(kind=int32), dimension(:), allocatable, intent(inout) :: pdata
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout) :: colnum
    integer(kind=int32), intent(inout)  :: ierr
    integer(kind=int32) :: status

    status = 0
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    ! get column number
!    call FTGCNN(unit, FITSIO_CASEINSENS, trim(colname), str_buffer, colnum, status)
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
  end subroutine fitswrap_prep_col_int32
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! get column number, then allocate some space
  ! datatype: integer(kind=int64)
  subroutine fitswrap_prep_col_int64(unit, colname, num_rows, colnum, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_prep_col_int64'
    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    integer(kind=int64), dimension(:), allocatable, intent(inout) :: pdata
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout) :: colnum
    integer(kind=int32), intent(inout)  :: ierr
    integer(kind=int32) :: status
!    character(len=STR_MAX) :: str_buffer

    status = 0
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, trim(colname), str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
  end subroutine fitswrap_prep_col_int64
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! get column number, then allocate some space
  ! datatype: integer(kind=real32)
  subroutine fitswrap_prep_col_real32(unit, colname, num_rows, colnum, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_prep_col_int8'
    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    real(kind=real32), dimension(:), allocatable, intent(inout) :: pdata
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout) :: colnum
    integer(kind=int32), intent(inout)  :: ierr
    integer(kind=int32) :: status
!    character(len=STR_MAX) :: str_buffer

    status = 0
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, trim(colname), str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
  end subroutine fitswrap_prep_col_real32
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! get column number, then allocate some space
  ! datatype: integer(kind=real64)
  subroutine fitswrap_prep_col_real64(unit, colname, num_rows, colnum, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_prep_col_int8'
    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    real(kind=real64), dimension(:), allocatable, intent(inout) :: pdata
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout) :: colnum
    integer(kind=int32), intent(inout)  :: ierr
    integer(kind=int32) :: status
!    character(len=STR_MAX) :: str_buffer

    status = 0
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, trim(colname), str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
  end subroutine fitswrap_prep_col_real64
  !---------------------------------------------------------------------------------------------















  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------- fitswrap_proc_col ------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------






! this fails because unlimited polymorphic types are not supported in the current gfortran version (v4.6.4)
!  !-------------------------------------------------------------------------------
!  ! get column number, allocate some space, and read in the column data
!  subroutine fitswrap_proc_col_generic(unit, colname, num_rows, pdata, ierr)
!    character(len=*), parameter :: subroutinename='fitswrap_proc_col_generic'
!
!    integer(kind=int32),  intent(in)    :: unit
!    character(len=*), intent(in)        :: colname
!    integer(kind=int32),  intent(in)    :: num_rows
!    !------------------------------------------------------------------------------
!!    integer(kind=int8), dimension(:), allocatable, intent(inout) :: pdata
!    class(*), intent(inout), pointer :: pdata
!    integer(kind=int8)                 :: nullval
!    !------------------------------------------------------------------------------
!    integer(kind=int32), intent(inout)  :: ierr
!
!    integer(kind=int32) :: colnum
!    integer(kind=int32) :: status
!    integer(kind=int32) :: anyf
!    character(len=STR_MAX) :: str_buffer
!
!    status = 0
!    nullval = 0
!
!    ! get column number
!    call FTGCNN(unit, FITSIO_CASEINSENS, colname, str_buffer, colnum, status)
!    if ( colnum <= 0 ) then
!       write (str_buffer, '(A,A)'), "Could not find fits column : ", trim(colname)
!       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, str_buffer)
!    end if
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!
!    if ( .not. allocated(pdata) ) then
!      ! allocate some space for the array
!      allocate ( pdata(num_rows), STAT = status)
!      if ( check_allocate_status(subroutinename, status, ierr) ) return
!    end if
!    !read the column data into the arrays
!    select type (pdata)
!    type is (integer)
!
!      !------------------------------------------------------------------------------
!      ! - This is one of the two lines of this subroutine which differ depending on datatype
!      call FTGCVB(unit,colnum,1,1,num_rows,nullval,pdata,anyf,status)
!      !------------------------------------------------------------------------------
!
!    end select
!
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!  end subroutine fitswrap_proc_col_generic
!  !---------------------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  ! get column number, allocate some space, and read in the column data
  ! datatype: integer(kind=int8)
  subroutine fitswrap_proc_col_int8(unit, colname, num_rows, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_proc_col_int8'

    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    integer(kind=int8), dimension(:), allocatable, intent(inout) :: pdata
    integer(kind=int8)                 :: nullval
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout)  :: ierr

    integer(kind=int32) :: colnum
    integer(kind=int32) :: status
    integer(kind=int32) :: anyf

    status = 0
    nullval = 0

    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, colname, str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
    !read the column data into the arrays
    !------------------------------------------------------------------------------
    ! - This is one of the two lines of this subroutine which differ depending on datatype
    call FTGCVB_w(unit,colnum,1,1,num_rows,nullval,pdata,anyf,status)
    !------------------------------------------------------------------------------

    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
  end subroutine fitswrap_proc_col_int8
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! get column number, allocate some space, and read in the column data
  ! datatype: integer(kind=int16)
  subroutine fitswrap_proc_col_int16(unit, colname, num_rows, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_proc_col_int16'

    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    integer(kind=int16), dimension(:), allocatable, intent(inout) :: pdata
    integer(kind=int16)                 :: nullval
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout)  :: ierr

    integer(kind=int32) :: colnum
    integer(kind=int32) :: status
    integer(kind=int32) :: anyf

    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    status = 0
    nullval = 0
    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, colname, str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
    !read the column data into the arrays
    !------------------------------------------------------------------------------
    ! - This is one of the two lines of this subroutine which differ depending on datatype
    call FTGCVI_w(unit,colnum,1,1,num_rows,nullval,pdata,anyf,status)
    !------------------------------------------------------------------------------

    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
  end subroutine fitswrap_proc_col_int16
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! get column number, allocate some space, and read in the column data
  ! datatype: integer(kind=int32)
  subroutine fitswrap_proc_col_int32(unit, colname, num_rows, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_proc_col_int32'

    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    integer(kind=int32), dimension(:), allocatable, intent(inout) :: pdata
    integer(kind=int32)                 :: nullval
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout)  :: ierr

    integer(kind=int32) :: colnum
    integer(kind=int32) :: status
    integer(kind=int32) :: anyf

    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    status = 0
    nullval = 0

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, colname, str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
    !read the column data into the arrays
    !------------------------------------------------------------------------------
    ! - This is one of the two lines of this subroutine which differ depending on datatype
    call FTGCVJ_w(unit,colnum,1,1,num_rows,nullval,pdata,anyf,status)
    !------------------------------------------------------------------------------

    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
  end subroutine fitswrap_proc_col_int32
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! get column number, allocate some space, and read in the column data
  ! datatype: integer(kind=int64)
  subroutine fitswrap_proc_col_int64(unit, colname, num_rows, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_proc_col_int64'

    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    integer(kind=int64), dimension(:), allocatable, intent(inout) :: pdata
    integer(kind=int64)                 :: nullval
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout)  :: ierr

    integer(kind=int32) :: colnum
    integer(kind=int32) :: status
    integer(kind=int32) :: anyf

    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    status = 0
    nullval = 0

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, colname, str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
    !read the column data into the arrays
    !------------------------------------------------------------------------------
    ! - This is one of the two lines of this subroutine which differ depending on datatype
    call FTGCVK_w(unit,colnum,1,1,num_rows,nullval,pdata,anyf,status)
    !------------------------------------------------------------------------------

    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
  end subroutine fitswrap_proc_col_int64
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! get column number, allocate some space, and read in the column data
  ! datatype: real(kind=real32)
  subroutine fitswrap_proc_col_real32(unit, colname, num_rows, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_proc_col_real32'

    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    real(kind=real32), dimension(:), allocatable, intent(inout) :: pdata
    real(kind=real32)                 :: nullval
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout)  :: ierr

    integer(kind=int32) :: colnum
    integer(kind=int32) :: status
    integer(kind=int32) :: anyf

    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    status = 0
    nullval = 0.0

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, colname, str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname))
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
    !read the column data into the arrays
    !------------------------------------------------------------------------------
    ! - This is one of the two lines of this subroutine which differ depending on datatype
    call FTGCVE_w(unit,colnum,1,1,num_rows,nullval,pdata,anyf,status)
    !------------------------------------------------------------------------------

    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
  end subroutine fitswrap_proc_col_real32
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! get column number, allocate some space, and read in the column data
  ! datatype: real(kind=real64)
  subroutine fitswrap_proc_col_real64(unit, colname, num_rows, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_proc_col_real64'

    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    !------------------------------------------------------------------------------
    real(kind=real64), dimension(:), allocatable, intent(inout) :: pdata
    real(kind=real64)                 :: nullval
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout)  :: ierr

    integer(kind=int32) :: colnum
    integer(kind=int32) :: status
    integer(kind=int32) :: anyf

    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    status = 0
    nullval = 0.0d0

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, colname, str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE,  "Could not find fits column : " // trim(colname) )
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
    !read the column data into the arrays
    !------------------------------------------------------------------------------
    ! - This is one of the two lines of this subroutine which differ depending on datatype
    call FTGCVD_w(unit,colnum,1,1,num_rows,nullval,pdata,anyf,status)
    !------------------------------------------------------------------------------

    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
  end subroutine fitswrap_proc_col_real64
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! get column number, allocate some space, and read in the column data
  subroutine fitswrap_proc_vec_col_real32(unit, colname, num_rows, width, pdata, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_vec_col_real32'

    integer(kind=int32),  intent(in)    :: unit
    character(len=*), intent(in)        :: colname
    integer(kind=int32),  intent(in)    :: num_rows
    integer(kind=int32),  intent(out)    :: width  ! dimension of vector column
    !------------------------------------------------------------------------------
    real(kind=real32), dimension(:), allocatable, intent(inout) :: pdata
    real(kind=real32)                 :: nullval
    !------------------------------------------------------------------------------
    integer(kind=int32), intent(inout)  :: ierr

    integer(kind=int32) :: colnum
    integer(kind=int32) :: datacode
    integer(kind=int32) :: group
    integer(kind=int32) :: status
    integer(kind=int32) :: anyf

    if ( ierr .ne. ERROR_LEVEL_OK ) then
      return
    end if

    status = 0
    nullval = 0.0

    ! get column number
    call FTGCNO_w(unit, FITSIO_CASEINSENS, trim(colname), colnum, status)
!    call FTGCNN(unit, FITSIO_CASEINSENS, colname, str_buffer, colnum, status)
    if ( colnum <= 0 ) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Could not find fits column : " // trim(colname) )
    end if
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    ! get column dimensions
    call FTGTCL_w(unit,colnum, datacode, width,group,status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if


    if ( .not. allocated(pdata) ) then
      ! allocate some space for the array
      allocate ( pdata(num_rows*width), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if
    !read the column data into the arrays
    !------------------------------------------------------------------------------
    ! - This is one of the two lines of this subroutine which differ depending on datatype
    call FTGCVE_w(unit,colnum,1,1,num_rows*width,nullval,pdata,anyf,status)
    !------------------------------------------------------------------------------

    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
  end subroutine fitswrap_proc_vec_col_real32
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_hduinfo(hduinfo)
    character(len=*), parameter :: subroutinename='fitswrap_init_hduinfo'
    type(hduinfo_type), intent(inout) :: hduinfo
    integer(kind=int32) :: i

    hduinfo%filename = "NONE"
    hduinfo%hdu_index = 0
    hduinfo%hdu_type  = FITSIO_ANY_HDU
    call fitswrap_init_key(hduinfo%key_extname , "EXTNAME" , "Name of FITS extension")
    call fitswrap_init_key(hduinfo%key_hduclass, "HDUCLASS", "Standard OGIP keyword. Class of format")
    call fitswrap_init_key(hduinfo%key_hduclas1, "HDUCLAS1", "Standard OGIP keyword. Sub-class of format")
    call fitswrap_init_key(hduinfo%key_hduclas2, "HDUCLAS2", "Standard OGIP keyword. Sub-class of format")
    call fitswrap_init_key(hduinfo%key_hduclas3, "HDUCLAS3", "Standard OGIP keyword. Sub-class of format")
    call fitswrap_init_key(hduinfo%key_hduclas4, "HDUCLAS4", "Standard OGIP keyword. Sub-class of format" )
    call fitswrap_init_key(hduinfo%key_hduvers , "HDUVERS" , "Standard OGIP keyword. Version of format" )

    call fitswrap_init_key(hduinfo%key_origin  , "ORIGIN"  , "Where is the file originate?" )
    call fitswrap_init_key(hduinfo%key_creator , "CREATOR" , "Software which created this file")
    call fitswrap_init_key(hduinfo%key_owner,    "OWNER",    "Data rights (MPE|IKI|IKI+MPE|MIXED|INVALID)")
    call fitswrap_init_key(hduinfo%key_mission , "MISSION" , "Location of the camera")
    call fitswrap_init_key(hduinfo%key_telescop, "TELESCOP", "Name of the telescope")
    call fitswrap_init_key(hduinfo%key_instrume, "INSTRUME", "eROSITA instrument identifier")
    call fitswrap_init_key(hduinfo%key_obs_mode, "OBS_MODE", "Observation mode (SURVEY|POINTING|SCAN)")
    call fitswrap_init_key(hduinfo%key_datamode, "DATAMODE", "Camera data acquisition mode")
    call fitswrap_init_key(hduinfo%key_frametim, "FRAMETIM", "[ms] Nominal frame time")
    call fitswrap_init_key(hduinfo%key_filter  , "FILTER"  , "Filter wheel position (OPEN|FILTER|CLOSED|Fe55)" )
    call fitswrap_init_key(hduinfo%key_skyfield, "SKYFIELD", "Six-digit all-sky survey field number")
    call fitswrap_init_key(hduinfo%key_obs_id  , "OBS_ID"  , "Observation identifier")
    call fitswrap_init_key(hduinfo%key_exp_id  , "EXP_ID"  , "Exposure identifier")
    call fitswrap_init_key(hduinfo%key_observer, "OBSERVER", "Name of observer")
    call fitswrap_init_key(hduinfo%key_object  , "OBJECT"  , "Target name or name from mission planning")
    call fitswrap_init_key(hduinfo%key_ra_obj  , "RA_OBJ"  , "[deg] Coord. of target in RADECSYS system")
    call fitswrap_init_key(hduinfo%key_dec_obj , "DEC_OBJ" , "[deg] Coord. of target in RADECSYS system")
    call fitswrap_init_key(hduinfo%key_ra_cen  , "RA_CEN"  , "[deg] Coord. of all-sky survey field in RADECSYS system")
    call fitswrap_init_key(hduinfo%key_dec_cen , "DEC_CEN" , "[deg] Coord. of all-sky survey field in RADECSYS system")
    call fitswrap_init_key(hduinfo%key_fieldlon, "FIELDLON", "[deg] Extent of field along Ecliptic longitude")
    call fitswrap_init_key(hduinfo%key_fieldlat, "FIELDLAT", "[deg] Extent of field along Ecliptic latitude")
    call fitswrap_init_key(hduinfo%key_date_obs, "DATE-OBS", "[YYYY-MM-DDThh:mm:ss.s] UTC Start time of exposure")
    call fitswrap_init_key(hduinfo%key_date_end, "DATE-END", "[YYYY-MM-DDThh:mm:ss.s] UTC End time of exposure")
    call fitswrap_init_key(hduinfo%key_tstart  , "TSTART"  , "[TIMEUNIT] Start time of exposure in TIMESYS system")
    call fitswrap_init_key(hduinfo%key_tstop   , "TSTOP"   , "[TIMEUNIT] End time of exposure in TIMESYS system")
    call fitswrap_init_key(hduinfo%key_mjdref  , "MJDREF"  , "[day] MJD at TIMEZERO")
    call fitswrap_init_key(hduinfo%key_timezero, "TIMEZERO", "[TIMEUNIT] Clock correction in TIMESYS system")
    call fitswrap_init_key(hduinfo%key_timeunit, "TIMEUNIT", "Units of TIMESYS system")
    call fitswrap_init_key(hduinfo%key_timesys , "TIMESYS" , "Standard name of time system")
    call fitswrap_init_key(hduinfo%key_ra_pnt  , "RA_PNT"  , "[deg] Actual average RA of pointing")
    call fitswrap_init_key(hduinfo%key_dec_pnt , "DEC_PNT" , "[deg] Actual average Dec of pointing")
    call fitswrap_init_key(hduinfo%key_pa_pnt  , "PA_PNT"  , "[deg] Actual average Position Angle of pointing")
    call fitswrap_init_key(hduinfo%key_radecsys, "RADECSYS", "Coordinate system identifier" )
    call fitswrap_init_key(hduinfo%key_equinox , "EQUINOX" , "[year] Equinox of coordinate system")
    call fitswrap_init_key(hduinfo%key_longstr , "LONGSTR" , "How long strings should be treated")

    call fitswrap_init_key(hduinfo%key_ninst   , "NINST"   , "Number of instruments that were merged")
    do i=1,MAX_INST
      call fitswrap_init_key(hduinfo%pkey_instrum(i), "INSTRUM" // trim(i2s(i)), "Contributing instrument name")
    end do

    call fitswrap_init_key(hduinfo%key_prefix  , "PREFIX"  , "Prefix part of pipeline filename")
    call fitswrap_init_key(hduinfo%key_dscode  , "DSCODE"  , "Dataset code, codes telescope/filter/band")
    call fitswrap_init_key(hduinfo%key_dstype  , "DSTYPE"  , "Dataset type: EventFile, Image, etc.")
    call fitswrap_init_key(hduinfo%key_version , "VERSION" , "File version - part of pipeline fname")
    call fitswrap_init_key(hduinfo%key_config  , "CONFIG"  , "Pipeline configuration - part of pipeline fname")

    call fitswrap_init_key(hduinfo%key_sasshist, "SASSHIST", "History of SASS processing" )

  end subroutine fitswrap_init_hduinfo


  !---------------------------------------------------------------------------------------------
  ! read in very short subset of standard header keywords from a previously opened fits header
  subroutine fitswrap_read_hduinfo_quick(unit, hduinfo, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_hduinfo_quick'

    integer(kind=int32),  intent(in)    :: unit
    type(hduinfo_type), intent(inout) :: hduinfo
    integer(kind=int32), intent(inout)  :: ierr

    integer(kind=int32) :: status
!    character(len=VLSTR_MAX) :: str_buffer

    ierr = ERROR_LEVEL_OK
    status = 0

    !record the filename associated with unit
    call FTFLNM_w(unit, hduinfo%filename, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    ! read which extension we are in
    call FTGHDN_w(unit, hduinfo%hdu_index)

    ! and the HDU type
    call FTGHDT_w(unit, hduinfo%hdu_type, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if
!    ! now read the vital keywords
!    call  fitswrap_read_key(unit, hduinfo%key_extname  ,"EXTNAME",  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, hduinfo%key_hduclass ,"HDUCLASS", ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, hduinfo%key_hduclas1 ,"HDUCLAS1", ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, hduinfo%key_hduclas2 ,"HDUCLAS2", ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, hduinfo%key_hduclas3 ,"HDUCLAS3", ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, hduinfo%key_hduclas4 ,"HDUCLAS4", ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, hduinfo%key_hduvers  ,"HDUVERS",  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, hduinfo%key_instrume ,"INSTRUME", ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return

    call fitswrap_read_key(unit, hduinfo%key_extname  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_hduclass , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_hduclas1 , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_hduclas2 , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_hduclas3 , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_hduclas4 , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_hduvers  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_instrume , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return

    call fitswrap_read_key(unit, hduinfo%key_sasshist , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return

!    write(*,'(A,A,1x,A,A)') subroutinename, " Reading hduinfo quick: filename= '", trim(hduinfo%filename),"'"
!    write(*,'(A,A,1x,I0)')  subroutinename, " Reading hduinfo quick: ext_index=", hduinfo%hdu_index
!    write(*,'(A,A,1x,I0)')  subroutinename, " Reading hduinfo quick: ext_type= ", hduinfo%hdu_type
!
!    write(*,'(A,A,A8,A,A,A,A,A,I0)') &
!         subroutinename, " Reading hduinfo quick: key name=", trim(hduinfo%key_extname%name), " value='", &
!        trim(hduinfo%key_extname%value),  "' comment='", trim(hduinfo%key_extname%comment), &
!        "' status=", status
!    write(*,'(A,A,A8,A,A,A,A,A,I0)') &
!         subroutinename, " Reading hduinfo quick: key name=", trim(hduinfo%key_instrume%name), " value='", &
!        trim(hduinfo%key_instrume%value), "' comment='", trim(hduinfo%key_instrume%comment), &
!        "' status=", status



  end subroutine fitswrap_read_hduinfo_quick


  !---------------------------------------------------------------------------------------------
  ! write all the standard eRosita header keywords into a previously opened fits header
  subroutine fitswrap_write_hduinfo(unit, hduinfo, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_write_hduinfo'

    integer(kind=int32),  intent(in)    :: unit
    type(hduinfo_type),   intent(in) :: hduinfo
    integer(kind=int32), intent(inout)  :: ierr
    integer(kind=int32)    :: i

    ! local versions of standard keywords that are always updated when we write files
    type(fitskey_vlstring_type) :: key_sasshist
    type(fitskey_string_type) :: key_creator


    ! update the SASSHIST keyword with information from this run of SRCTOOL
    ! the sass_history parameter is a global variable from module "param"
    call fitswrap_init_key(key_sasshist, "SASSHIST", "History of SASS processing")
    call fitswrap_set_key(key_sasshist, trim(hduinfo%key_sasshist%value) // trim(sass_history))

    ! update the CREATOR keyword to give the standard SRCTOOL name and version
    call fitswrap_init_key(key_creator, "CREATOR", "Software which created this file")
    call fitswrap_set_key(key_creator, CREATOR_STRING)

    ierr = ERROR_LEVEL_OK

    !--------------
    ! now write the keys one by one
    call fitswrap_write_key(unit, hduinfo%key_origin   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_creator  , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_owner    , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_mission  , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_telescop , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_instrume , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_obs_mode , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_datamode , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_frametim , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_filter   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_skyfield , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_obs_id   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_exp_id   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_observer , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_object   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_ra_obj   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_dec_obj  , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_ra_cen   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_dec_cen  , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_fieldlon , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_fieldlat , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_date_obs , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_date_end , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_tstart   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_tstop    , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_mjdref   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_timezero , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_timeunit , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_timesys  , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_ra_pnt   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_dec_pnt  , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_pa_pnt   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_radecsys , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_equinox  , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_longstr  , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return

!    write (*,'(A,A,A,A,A,I0)') "blah3 EXTNAME=", &
!        trim(hduinfo%key_extname%value), "-> INSTRUME= ", trim(hduinfo%key_instrume%value), &
!        " NINST=", hduinfo%key_ninst%value

    ! only write out NINST and INSTRUMx keywords when INSTRUME == "MERGED"
    if ( trim(hduinfo%key_instrume%value) .eq. INSTRUME_IS_MERGED ) then
      call fitswrap_write_key (unit, hduinfo%key_ninst,    ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) return
      if ((hduinfo%key_ninst%value .ge. 1) .and. (hduinfo%key_ninst%value .le. MAX_INST) ) then
        do i=1,hduinfo%key_ninst%value
          call fitswrap_write_key (unit, hduinfo%pkey_instrum(i), ierr)
          if ( ierr .ne. ERROR_LEVEL_OK ) return
        end do
      else
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Illegal number in NINST keyword")
        return
      endif
    end if

    call fitswrap_write_key(unit, hduinfo%key_prefix   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_dscode   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_dstype   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_version  , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_write_key(unit, hduinfo%key_config   , ierr) ; if ( ierr .ne. ERROR_LEVEL_OK ) return

  end subroutine fitswrap_write_hduinfo
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_write_hduinfo_quiet(unit, hduinfo, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_write_hduinfo_quiet'
    integer(kind=int32), intent(in)    :: unit
    type(hduinfo_type), intent(in) :: hduinfo
    integer(kind=int32), intent(out)   :: ierr

    ! local versions of standard keywords that are always updated when we write files
    type(fitskey_vlstring_type) :: key_sasshist
    type(fitskey_string_type) :: key_creator
    type(fitskey_string_type) :: key_longstrn
    integer(kind=int32) :: i

    ! update the SASSHIST keyword with information from this run of SRCTOOL
    ! the sass_history parameter is a global variable from module "param"
    !associated(hduinfo%key_sasshist)
    call fitswrap_init_key(key_sasshist, "SASSHIST", "History of SASS processing")
    call fitswrap_set_key(key_sasshist, trim(hduinfo%key_sasshist%value) // trim(sass_history))

    !write(*,'(A,A)'), "Setting SASSHIST key for file: ", trim(hduinfo%filename)

    ! update the CREATOR keyword to give the standard SRCTOOL name and version
    call fitswrap_init_key(key_creator, "CREATOR", "Software which created this file")
    call fitswrap_set_key(key_creator, CREATOR_STRING)

    call fitswrap_init_key(key_longstrn, "LONGSTRN", "Uses OGIP long string convention")
    call fitswrap_set_key(key_longstrn, "OGIP 1.0")

    ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key_if_supplied(unit, hduinfo%key_extname , ierr)
    call fitswrap_write_key_if_supplied(unit, key_longstrn,         ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_hduclass, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_hduclas1, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_hduclas2, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_hduclas3, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_hduclas4, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_hduvers , ierr)

    call fitswrap_write_key_if_supplied(unit, hduinfo%key_origin  , ierr)
    call fitswrap_write_key_if_supplied(unit, key_creator, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_owner,    ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_mission , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_telescop, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_instrume, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_obs_mode, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_datamode, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_frametim, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_filter  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_skyfield, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_obs_id  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_exp_id  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_observer, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_object  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_ra_obj  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_dec_obj , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_ra_cen  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_dec_cen , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_fieldlon, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_fieldlat, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_date_obs, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_date_end, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_tstart  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_tstop   , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_mjdref  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_timezero, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_timeunit, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_timesys , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_ra_pnt  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_dec_pnt , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_pa_pnt  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_radecsys, ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_equinox , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_longstr , ierr)

!    write (*,'(A,A,A,A,A,I0)') "blah4 EXTNAME=", &
!       trim(hduinfo%key_extname%value), "-> INSTRUME= ", trim(hduinfo%key_instrume%value), &
!        " NINST=", hduinfo%key_ninst%value

    ! only write out NINST and INSTRUMx keywords when INSTRUME == "MERGED"
    if ( trim(hduinfo%key_instrume%value) .eq. INSTRUME_IS_MERGED ) then
      call fitswrap_write_key_if_supplied (unit, hduinfo%key_ninst,    ierr)
      if ((hduinfo%key_ninst%value .ge. 1) .and. (hduinfo%key_ninst%value .le. MAX_INST) ) then
        do i=1,hduinfo%key_ninst%value
          call fitswrap_write_key_if_supplied (unit, hduinfo%pkey_instrum(i), ierr)
        end do
      endif
    end if

    call fitswrap_write_key_if_supplied(unit, hduinfo%key_prefix  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_dscode  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_dstype  , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_version , ierr)
    call fitswrap_write_key_if_supplied(unit, hduinfo%key_config  , ierr)

    call fitswrap_write_key(unit, key_sasshist, ierr)


    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing standard header keywords to CHDU")
      return
    end if

  end subroutine fitswrap_write_hduinfo_quiet

  !---------------------------------------------------------------------------------------------

!  !---------------------------------------------------------------------------------------------
!  ! write all the standard eRosita header keywords into a previously opened fits header
!  ! do not stop in case of errors
!  subroutine fitswrap_write_hduinfo_quiet(unit, hduinfo, ierr)
!    character(len=*), parameter :: subroutinename='fitswrap_write_hduinfo_quiet'
!
!    integer(kind=int32),  intent(in)    :: unit
!    type(hduinfo_type),   intent(in) :: hduinfo
!    integer(kind=int32), intent(inout)  :: ierr
!    integer(kind=int32) :: i
!    ierr = ERROR_LEVEL_OK
!
!    !--------------
!    ! now write the keys one by one, reset the error flag after each
!    call fitswrap_write_key(unit, hduinfo%key_origin   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_creator  , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_owner    , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_mission  , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_telescop , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_instrume , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_obs_mode , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_datamode , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_frametim , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_filter   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_skyfield , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_obs_id   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_exp_id   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_observer , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_object   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_ra_obj   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_dec_obj  , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_ra_cen   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_dec_cen  , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_fieldlon , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_fieldlat , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_date_obs , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_date_end , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_tstart   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_tstop    , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_mjdref   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_timezero , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_timeunit , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_timesys  , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_ra_pnt   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_dec_pnt  , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_pa_pnt   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_radecsys , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_equinox  , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_longstr  , ierr); ierr = ERROR_LEVEL_OK
!
!    write (*,'(A,A,A,A,A,I0)') "blah4 EXTNAME=", &
!        trim(hduinfo%key_extname%value), "-> INSTRUME= ", trim(hduinfo%key_instrume%value), &
!        " NINST=", hduinfo%key_ninst%value
!
!    ! only write out NINST and INSTRUMx keywords when INSTRUME == "MERGED"
!    if ( trim(hduinfo%key_instrume%value) .eq. "MERGED" ) then
!
!      call fitswrap_write_key (unit, hduinfo%key_ninst,    ierr)
!      ierr = ERROR_LEVEL_OK
!      if ((hduinfo%key_ninst%value .ge. 1) .and. (hduinfo%key_ninst%value .le. MAX_INST) ) then
!        do i=1,hduinfo%key_ninst%value
!          call fitswrap_write_key (unit, hduinfo%pkey_instrum(i), ierr)
!          ierr = ERROR_LEVEL_OK
!        end do
!      endif
!    end if
!
!    call fitswrap_write_key(unit, hduinfo%key_prefix   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_dscode   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_dstype   , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_version  , ierr); ierr = ERROR_LEVEL_OK
!    call fitswrap_write_key(unit, hduinfo%key_config   , ierr); ierr = ERROR_LEVEL_OK
!
!!    call fitswrap_write_key(unit, hduinfo%key_naxis1   , ierr); ierr = ERROR_LEVEL_OK
!!    call fitswrap_write_key(unit, hduinfo%key_naxis2   , ierr); ierr = ERROR_LEVEL_OK
!
!
!  end subroutine fitswrap_write_hduinfo_quiet
!  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  ! read in all the standard eRosita header keywords from a previously opened fits header
  subroutine fitswrap_read_hduinfo(unit, hduinfo, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_hduinfo'

    integer(kind=int32),  intent(in)    :: unit
    type(hduinfo_type), intent(inout) :: hduinfo
    integer(kind=int32), intent(inout)  :: ierr

    integer(kind=int32) :: status, i
!    character(len=VLSTR_MAX) :: str_buffer

    ierr = ERROR_LEVEL_OK
    status = 0

    !-------------------------------------------------------------------------------
    ! set if any keys to be compulsory
    ! hduinfo%key_extname%compulsory = .true.

!    call wrschatty (subroutinename, "Reading in hduinfo from CHDU")
    call fitswrap_read_hduinfo_quick(unit, hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) return

    !--------------
    ! now read in the standard keys one by one
    call fitswrap_read_key(unit, hduinfo%key_origin   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_creator  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_owner    , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_mission  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_telescop , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_instrume , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_obs_mode , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_datamode , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_frametim , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_filter   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_skyfield , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_obs_id   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_exp_id   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_observer , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_object   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_ra_obj   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_dec_obj  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_ra_cen   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_dec_cen  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_fieldlon , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_fieldlat , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_date_obs , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_date_end , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_tstart   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_tstop    , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_mjdref   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_timezero , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_timeunit , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_timesys  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_ra_pnt   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_dec_pnt  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_pa_pnt   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_radecsys , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_equinox  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_longstr  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return

    call fitswrap_read_key(unit, hduinfo%key_prefix   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_dscode   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_dstype   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_version  , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call fitswrap_read_key(unit, hduinfo%key_config   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return


    if ((hduinfo%key_instrume%flag .ne. FITSWRAP_KEY_FLAG_BLANK) .and. &
        (trim(hduinfo%key_instrume%value) .eq. INSTRUME_IS_MERGED) ) then
      call fitswrap_read_key (unit, hduinfo%key_ninst, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) return
      if ( (hduinfo%key_ninst%value .ge. 1) .and. (hduinfo%key_ninst%value .le. MAX_INST) ) then
        do i=1,hduinfo%key_ninst%value
          call fitswrap_read_key (unit, hduinfo%pkey_instrum(i), ierr);
        end do
      else
        ierr = ERROR_LEVEL_SEVERE
        return
      endif
    end if


!    call fitswrap_read_key(unit, hduinfo%key_naxis1   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call fitswrap_read_key(unit, hduinfo%key_naxis2   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return


!    write (str_buffer, '(A,A,A,A,A,I0,A)'), &
!        "I read HDU info from:", &
!        trim(hduinfo%filename), &
!        "[", &
!        trim(hduinfo%key_extname%value), &
!        "] (hduindex=", &
!        hduinfo%hdu_index, &
!        ")"
!    call wrschatty (subroutinename, str_buffer)

  end subroutine fitswrap_read_hduinfo
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_logical (fitskey, keyname, compulsory)
    type(fitskey_logical_type), intent(inout) :: fitskey
    character(len=*), intent(in)              :: keyname
    logical         , intent(in), optional    :: compulsory
    fitskey%name    = trim(keyname(1:MIN(LEN(keyname),FITSIO_KEY_MAXLEN)))
    fitskey%value   = .false.
    fitskey%comment = ""
    fitskey%flag    = FITSWRAP_KEY_FLAG_BLANK
    fitskey%compulsory = .false.
    if (present(compulsory)) fitskey%compulsory = compulsory
    fitskey%is_null = .false.

  end subroutine fitswrap_init_key_logical
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_string (fitskey, keyname, compulsory)
    type(fitskey_string_type), intent(inout) :: fitskey
    character(len=*), intent(in)             :: keyname
    logical           , intent(in), optional :: compulsory
    fitskey%name    = trim(keyname(1:MIN(LEN(keyname),FITSIO_KEY_MAXLEN)))
    fitskey%value   = ""
    fitskey%comment = ""
    fitskey%flag    = FITSWRAP_KEY_FLAG_BLANK
    fitskey%compulsory = .false.
    if (present(compulsory)) fitskey%compulsory = compulsory
    fitskey%is_null = .false.
  end subroutine fitswrap_init_key_string
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_vlstring (fitskey, keyname, compulsory)
    type(fitskey_vlstring_type), intent(inout) :: fitskey
    character(len=*), intent(in)               :: keyname
    logical           , intent(in), optional   :: compulsory
    fitskey%name    = trim(keyname(1:MIN(LEN(keyname),FITSIO_KEY_MAXLEN)))
    fitskey%value   = ""
    fitskey%comment = ""
    fitskey%flag    = FITSWRAP_KEY_FLAG_BLANK
    fitskey%compulsory = .false.
    if (present(compulsory)) fitskey%compulsory = compulsory
    fitskey%is_null = .false.
  end subroutine fitswrap_init_key_vlstring
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_int16 (fitskey, keyname, compulsory)
    type(fitskey_int16_type), intent(inout)  :: fitskey
    character(len=*), intent(in)             :: keyname
    logical           , intent(in), optional :: compulsory
    fitskey%name    = trim(keyname(1:MIN(LEN(keyname),FITSIO_KEY_MAXLEN)))
    fitskey%value   = 0
    fitskey%comment = ""
    fitskey%flag    = FITSWRAP_KEY_FLAG_BLANK
    fitskey%compulsory = .false.
    if (present(compulsory)) fitskey%compulsory = compulsory
    fitskey%is_null = .false.
  end subroutine fitswrap_init_key_int16
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_int32 (fitskey, keyname, compulsory)
    type(fitskey_int32_type), intent(inout)  :: fitskey
    character(len=*), intent(in)             :: keyname
    logical           , intent(in), optional :: compulsory
    fitskey%name    = trim(keyname(1:MIN(LEN(keyname),FITSIO_KEY_MAXLEN)))
    fitskey%value   = 0
    fitskey%comment = ""
    fitskey%flag    = FITSWRAP_KEY_FLAG_BLANK
    fitskey%compulsory = .false.
    if (present(compulsory)) fitskey%compulsory = compulsory
    fitskey%is_null = .false.
  end subroutine fitswrap_init_key_int32
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_int64 (fitskey, keyname, compulsory)
    type(fitskey_int64_type), intent(inout)  :: fitskey
    character(len=*), intent(in)             :: keyname
    logical           , intent(in), optional :: compulsory
    fitskey%name    = trim(keyname(1:MIN(LEN(keyname),FITSIO_KEY_MAXLEN)))
    fitskey%value   = 0
    fitskey%comment = ""
    fitskey%flag    = FITSWRAP_KEY_FLAG_BLANK
    fitskey%compulsory = .false.
    if (present(compulsory)) fitskey%compulsory = compulsory
    fitskey%is_null = .false.
  end subroutine fitswrap_init_key_int64
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_real32 (fitskey, keyname, compulsory)
    type(fitskey_real32_type), intent(inout) :: fitskey
    character(len=*), intent(in)             :: keyname
    logical           , intent(in), optional :: compulsory
    fitskey%name    = trim(keyname(1:MIN(LEN(keyname),FITSIO_KEY_MAXLEN)))
    fitskey%value   = 0.0
    fitskey%comment = ""
    fitskey%flag    = FITSWRAP_KEY_FLAG_BLANK
    fitskey%compulsory = .false.
    if (present(compulsory)) fitskey%compulsory = compulsory
    fitskey%is_null = .false.
  end subroutine fitswrap_init_key_real32
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_real64 (fitskey, keyname, compulsory)
    type(fitskey_real64_type), intent(inout) :: fitskey
    character(len=*), intent(in)             :: keyname
    logical           , intent(in), optional :: compulsory
    fitskey%name    = trim(keyname(1:MIN(LEN(keyname),FITSIO_KEY_MAXLEN)))
    fitskey%value   = 0.0d0
    fitskey%comment = ""
    fitskey%flag    = FITSWRAP_KEY_FLAG_BLANK
    fitskey%compulsory = .false.
    if (present(compulsory)) fitskey%compulsory = compulsory
    fitskey%is_null = .false.
  end subroutine fitswrap_init_key_real64
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_wcomment_logical (fitskey, keyname, comment, compulsory)
    type(fitskey_logical_type), intent(inout) :: fitskey
    character(len=*), intent(in)              :: keyname
    character(len=*), intent(in)              :: comment
    logical           , intent(in), optional :: compulsory
    call fitswrap_init_key (fitskey, keyname, compulsory)
    fitskey%comment = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
  end subroutine fitswrap_init_key_wcomment_logical
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_wcomment_string (fitskey, keyname, comment, compulsory)
    type(fitskey_string_type), intent(inout) :: fitskey
    character(len=*), intent(in)              :: keyname
    character(len=*), intent(in)              :: comment
    logical           , intent(in), optional :: compulsory
    call fitswrap_init_key (fitskey, keyname, compulsory)
    fitskey%comment = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
  end subroutine fitswrap_init_key_wcomment_string
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_wcomment_vlstring (fitskey, keyname, comment, compulsory)
    type(fitskey_vlstring_type), intent(inout) :: fitskey
    character(len=*), intent(in)              :: keyname
    character(len=*), intent(in)              :: comment
    logical           , intent(in), optional :: compulsory
    call fitswrap_init_key (fitskey, keyname, compulsory)
    fitskey%comment = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
  end subroutine fitswrap_init_key_wcomment_vlstring
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_wcomment_int16 (fitskey, keyname, comment, compulsory)
    type(fitskey_int16_type), intent(inout) :: fitskey
    character(len=*), intent(in)              :: keyname
    character(len=*), intent(in)              :: comment
    logical           , intent(in), optional :: compulsory
    call fitswrap_init_key (fitskey, keyname, compulsory)
    fitskey%comment = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
  end subroutine fitswrap_init_key_wcomment_int16
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_wcomment_int32 (fitskey, keyname, comment, compulsory)
    type(fitskey_int32_type), intent(inout) :: fitskey
    character(len=*), intent(in)              :: keyname
    character(len=*), intent(in)              :: comment
    logical           , intent(in), optional :: compulsory
    call fitswrap_init_key (fitskey, keyname, compulsory)
    fitskey%comment = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
  end subroutine fitswrap_init_key_wcomment_int32
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_wcomment_int64 (fitskey, keyname, comment, compulsory)
    type(fitskey_int64_type), intent(inout) :: fitskey
    character(len=*), intent(in)              :: keyname
    character(len=*), intent(in)              :: comment
    logical           , intent(in), optional :: compulsory
    call fitswrap_init_key (fitskey, keyname, compulsory)
    fitskey%comment = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
  end subroutine fitswrap_init_key_wcomment_int64
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_wcomment_real32 (fitskey, keyname, comment, compulsory)
    type(fitskey_real32_type), intent(inout) :: fitskey
    character(len=*), intent(in)              :: keyname
    character(len=*), intent(in)              :: comment
    logical           , intent(in), optional :: compulsory
    call fitswrap_init_key (fitskey, keyname, compulsory)
    fitskey%comment = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
  end subroutine fitswrap_init_key_wcomment_real32
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_init_key_wcomment_real64 (fitskey, keyname, comment, compulsory)
    type(fitskey_real64_type), intent(inout) :: fitskey
    character(len=*), intent(in)              :: keyname
    character(len=*), intent(in)              :: comment
    logical           , intent(in), optional :: compulsory
    call fitswrap_init_key (fitskey, keyname, compulsory)
    fitskey%comment = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
  end subroutine fitswrap_init_key_wcomment_real64
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------



  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_null_logical (fitskey)
    type(fitskey_logical_type), intent(inout) :: fitskey
    fitskey%is_null = .true.
  end subroutine fitswrap_set_key_null_logical

  recursive subroutine fitswrap_set_key_null_string (fitskey)
    type(fitskey_string_type), intent(inout) :: fitskey
    fitskey%is_null = .true.
  end subroutine fitswrap_set_key_null_string

  recursive subroutine fitswrap_set_key_null_vlstring (fitskey)
    type(fitskey_vlstring_type), intent(inout) :: fitskey
    fitskey%is_null = .true.
  end subroutine fitswrap_set_key_null_vlstring

  recursive subroutine fitswrap_set_key_null_int16 (fitskey)
    type(fitskey_int16_type), intent(inout) :: fitskey
    fitskey%is_null = .true.
  end subroutine fitswrap_set_key_null_int16

  recursive subroutine fitswrap_set_key_null_int32 (fitskey)
    type(fitskey_int32_type), intent(inout) :: fitskey
    fitskey%is_null = .true.
  end subroutine fitswrap_set_key_null_int32

  recursive subroutine fitswrap_set_key_null_int64 (fitskey)
    type(fitskey_int64_type), intent(inout) :: fitskey
    fitskey%is_null = .true.
  end subroutine fitswrap_set_key_null_int64

  recursive subroutine fitswrap_set_key_null_real32 (fitskey)
    type(fitskey_real32_type), intent(inout) :: fitskey
    fitskey%is_null = .true.
  end subroutine fitswrap_set_key_null_real32

  recursive subroutine fitswrap_set_key_null_real64 (fitskey)
    type(fitskey_real64_type), intent(inout) :: fitskey
    fitskey%is_null = .true.
  end subroutine fitswrap_set_key_null_real64
  !---------------------------------------------------------------------------------------------



  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_write_key_if_supplied_logical (unit, fitskey, ierr)
    type(fitskey_logical_type), intent(in) :: fitskey
    integer(kind=int32), intent(in)           :: unit
    integer(kind=int32), intent(inout)        :: ierr
    if ( fitskey%flag  .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call fitswrap_write_key (unit, fitskey, ierr)
    endif
  end subroutine fitswrap_write_key_if_supplied_logical

  recursive subroutine fitswrap_write_key_if_supplied_string (unit, fitskey, ierr)
    type(fitskey_string_type), intent(in) :: fitskey
    integer(kind=int32), intent(in)           :: unit
    integer(kind=int32), intent(inout)        :: ierr
    if ( fitskey%flag  .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call fitswrap_write_key (unit, fitskey, ierr)
    endif
  end subroutine fitswrap_write_key_if_supplied_string

  recursive subroutine fitswrap_write_key_if_supplied_vlstring (unit, fitskey, ierr)
    type(fitskey_vlstring_type), intent(in) :: fitskey
    integer(kind=int32), intent(in)           :: unit
    integer(kind=int32), intent(inout)        :: ierr
    if ( fitskey%flag  .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call fitswrap_write_key (unit, fitskey, ierr)
    endif
  end subroutine fitswrap_write_key_if_supplied_vlstring

  recursive subroutine fitswrap_write_key_if_supplied_int16 (unit, fitskey, ierr)
    type(fitskey_int16_type), intent(in) :: fitskey
    integer(kind=int32), intent(in)           :: unit
    integer(kind=int32), intent(inout)        :: ierr
    if ( fitskey%flag  .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call fitswrap_write_key (unit, fitskey, ierr)
    endif
  end subroutine fitswrap_write_key_if_supplied_int16

  recursive subroutine fitswrap_write_key_if_supplied_int32 (unit, fitskey, ierr)
    type(fitskey_int32_type), intent(in) :: fitskey
    integer(kind=int32), intent(in)           :: unit
    integer(kind=int32), intent(inout)        :: ierr
    if ( fitskey%flag  .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call fitswrap_write_key (unit, fitskey, ierr)
    endif
  end subroutine fitswrap_write_key_if_supplied_int32

  recursive subroutine fitswrap_write_key_if_supplied_int64 (unit, fitskey, ierr)
    type(fitskey_int64_type), intent(in) :: fitskey
    integer(kind=int32), intent(in)           :: unit
    integer(kind=int32), intent(inout)        :: ierr
    if ( fitskey%flag  .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call fitswrap_write_key (unit, fitskey, ierr)
    endif
  end subroutine fitswrap_write_key_if_supplied_int64

  recursive subroutine fitswrap_write_key_if_supplied_real32 (unit, fitskey, ierr)
    type(fitskey_real32_type), intent(in) :: fitskey
    integer(kind=int32), intent(in)           :: unit
    integer(kind=int32), intent(inout)        :: ierr
    if ( fitskey%flag  .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call fitswrap_write_key (unit, fitskey, ierr)
    endif
  end subroutine fitswrap_write_key_if_supplied_real32

  recursive subroutine fitswrap_write_key_if_supplied_real64 (unit, fitskey, ierr)
    type(fitskey_real64_type), intent(in) :: fitskey
    integer(kind=int32), intent(in)           :: unit
    integer(kind=int32), intent(inout)        :: ierr
    if ( fitskey%flag  .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call fitswrap_write_key (unit, fitskey, ierr)
    endif
  end subroutine fitswrap_write_key_if_supplied_real64
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_logical (fitskey, name, key_value, comment)
    type(fitskey_logical_type), intent(inout) :: fitskey
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: comment
    logical           , intent(in) :: key_value
    if ( trim(name) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%name       = trim(name(1:MIN(LEN(name),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%value      = key_value
    if ( trim(comment) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%comment    = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%flag       = FITSWRAP_KEY_FLAG_SUPPLIED
    fitskey%is_null = .false.
  end subroutine fitswrap_set_key_logical
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_string (fitskey, name, key_value, comment)
    type(fitskey_string_type), intent(inout) :: fitskey
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: comment
    character(len=*), intent(in) :: key_value
    if ( trim(name) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%name       = trim(name(1:MIN(LEN(name),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%value      = trim(key_value(1:MIN(LEN(key_value),STR_MAX))) ! make sure not to overrun the buffer
    if ( trim(comment) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%comment    = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%flag       = FITSWRAP_KEY_FLAG_SUPPLIED
    fitskey%is_null = .false.
  end subroutine fitswrap_set_key_string
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_vlstring (fitskey, name, key_value, comment)
    type(fitskey_vlstring_type), intent(inout) :: fitskey
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: comment
    character(len=*), intent(in) :: key_value
    if ( trim(name) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%name       = trim(name(1:MIN(LEN(name),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%value      = trim(key_value(1:MIN(LEN(key_value),VLSTR_MAX))) ! make sure not to overrun the buffer
    if ( trim(comment) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%comment    = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%flag       = FITSWRAP_KEY_FLAG_SUPPLIED
    fitskey%is_null = .false.
  end subroutine fitswrap_set_key_vlstring
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_int16 (fitskey, name, key_value, comment)
    type(fitskey_int16_type), intent(inout) :: fitskey
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: comment
    integer(kind=int16), intent(in) :: key_value
    if ( trim(name) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%name       = trim(name(1:MIN(LEN(name),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%value      = key_value
    if ( trim(comment) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%comment    = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%flag       = FITSWRAP_KEY_FLAG_SUPPLIED
    fitskey%is_null = .false.
  end subroutine fitswrap_set_key_int16
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_int32 (fitskey, name, key_value, comment)
    type(fitskey_int32_type), intent(inout) :: fitskey
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: comment
    integer(kind=int32), intent(in) :: key_value
    if ( trim(name) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%name       = trim(name(1:MIN(LEN(name),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%value      = key_value
    if ( trim(comment) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%comment    = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%flag       = FITSWRAP_KEY_FLAG_SUPPLIED
    fitskey%is_null = .false.
  end subroutine fitswrap_set_key_int32
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_int64 (fitskey, name, key_value, comment)
    type(fitskey_int64_type), intent(inout) :: fitskey
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: comment
    integer(kind=int64), intent(in) :: key_value
    if ( trim(name) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%name       = trim(name(1:MIN(LEN(name),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%value      = key_value
    if ( trim(comment) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%comment    = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%flag       = FITSWRAP_KEY_FLAG_SUPPLIED
    fitskey%is_null = .false.
  end subroutine fitswrap_set_key_int64
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_real32 (fitskey, name, key_value, comment)
    type(fitskey_real32_type), intent(inout) :: fitskey
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: comment
    real(kind=real32), intent(in) :: key_value
    if ( trim(name) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%name       = trim(name(1:MIN(LEN(name),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%value      = key_value
    if ( trim(comment) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%comment    = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%flag       = FITSWRAP_KEY_FLAG_SUPPLIED
    fitskey%is_null = .false.
  end subroutine fitswrap_set_key_real32
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_real64 (fitskey, name, key_value, comment)
    type(fitskey_real64_type), intent(inout) :: fitskey
    character(len=*),  intent(in) :: name
    character(len=*),  intent(in) :: comment
    real(kind=real64), intent(in) :: key_value
    if ( trim(name) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%name       = trim(name(1:MIN(LEN(name),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%value      = key_value
    if ( trim(comment) .ne. FITSWRAP_SET_KEY_BLANK ) then
      fitskey%comment    = trim(comment(1:MIN(LEN(comment),FITSIO_KEY_MAXLEN)))
    end if
    fitskey%flag       = FITSWRAP_KEY_FLAG_SUPPLIED
    fitskey%is_null = .false.
  end subroutine fitswrap_set_key_real64
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_value_logical (fitskey, key_value)
    type(fitskey_logical_type), intent(inout) :: fitskey
    logical           , intent(in) :: key_value
    call fitswrap_set_key(fitskey, FITSWRAP_SET_KEY_BLANK, key_value, FITSWRAP_SET_KEY_BLANK)
  end subroutine fitswrap_set_key_value_logical
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_value_string (fitskey, key_value)
    type(fitskey_string_type), intent(inout) :: fitskey
    character(len=*),  intent(in) :: key_value
    call fitswrap_set_key(fitskey, FITSWRAP_SET_KEY_BLANK, key_value, FITSWRAP_SET_KEY_BLANK)
  end subroutine fitswrap_set_key_value_string
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_value_vlstring (fitskey, key_value)
    type(fitskey_vlstring_type), intent(inout) :: fitskey
    character(len=*),  intent(in) :: key_value
    call fitswrap_set_key(fitskey, FITSWRAP_SET_KEY_BLANK, key_value, FITSWRAP_SET_KEY_BLANK)
  end subroutine fitswrap_set_key_value_vlstring
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_value_int16 (fitskey, key_value)
    type(fitskey_int16_type), intent(inout) :: fitskey
    integer(kind=int16), intent(in) :: key_value
    call fitswrap_set_key(fitskey, FITSWRAP_SET_KEY_BLANK, key_value, FITSWRAP_SET_KEY_BLANK)
  end subroutine fitswrap_set_key_value_int16
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_value_int32 (fitskey, key_value)
    type(fitskey_int32_type), intent(inout) :: fitskey
    integer(kind=int32), intent(in) :: key_value
    call fitswrap_set_key(fitskey, FITSWRAP_SET_KEY_BLANK, key_value, FITSWRAP_SET_KEY_BLANK)
  end subroutine fitswrap_set_key_value_int32
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_value_int64 (fitskey, key_value)
    type(fitskey_int64_type), intent(inout) :: fitskey
    integer(kind=int64), intent(in) :: key_value
    call fitswrap_set_key(fitskey, FITSWRAP_SET_KEY_BLANK, key_value, FITSWRAP_SET_KEY_BLANK)
  end subroutine fitswrap_set_key_value_int64
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_value_real32 (fitskey, key_value)
    type(fitskey_real32_type), intent(inout) :: fitskey
    real(kind=real32), intent(in) :: key_value
    call fitswrap_set_key(fitskey, FITSWRAP_SET_KEY_BLANK, key_value, FITSWRAP_SET_KEY_BLANK)
  end subroutine fitswrap_set_key_value_real32
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  recursive subroutine fitswrap_set_key_value_real64 (fitskey, key_value)
    type(fitskey_real64_type), intent(inout) :: fitskey
    real(kind=real64), intent(in) :: key_value
    call fitswrap_set_key(fitskey, FITSWRAP_SET_KEY_BLANK, key_value, FITSWRAP_SET_KEY_BLANK)
  end subroutine fitswrap_set_key_value_real64
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------



  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  !-----------------------Start of fitswrap_read_key_* subroutines------------------------------
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_read_key_logical (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_key_logical'

    integer(kind=int32), intent(in)             :: unit
    type(fitskey_logical_type), intent(inout)  :: fitskey
    integer(kind=int32), intent(inout)          :: ierr

    integer(kind=int32) :: temp
    integer(kind=int32) :: status
    character(len=FITSIO_KEY_MAXLEN) :: comment_copy

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    comment_copy = trim(fitskey%comment)

!    ierr = ERROR_LEVEL_OK
    status = 0

    if ( len_trim(fitskey%name) > 0 ) then
      ! do nothing
    else
      fitskey%flag = FITSWRAP_KEY_FLAG_BLANK
      return ! we therefore do not use this key
    end if

    ! now actually try to read the key from the file
    call FTGKYL_w(unit, trim(fitskey%name), temp, fitskey%comment, status)
    if( temp == 0 ) then
      fitskey%value = .false.
    else
      fitskey%value = .true.
    end if

    ! revert to default keyword comment when no keyword is supplied
    if ( len_trim(fitskey%comment) .eq. 0 ) fitskey%comment = trim(comment_copy)

    if ( status == 0 ) then ! all is well
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
    else if ( status == FITSIO_ERROR_VALUE_UNDEFINED ) then
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
      call fitswrap_set_key_null(fitskey)
    else if ( status == FITSIO_ERROR_KEYWORD_NOEXIST ) then
      if ( fitskey%compulsory ) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        call wrserr(subroutinename,ierr,"Could not find compulsory key: " // trim(fitskey%name))
        return
      else ! just quietly ignore this missing key
        fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      end if
    else ! there must be some other kind of fitsio error
      ierr = ERROR_LEVEL_SEVERE
      fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      call fitswrap_write_fits_error(subroutinename, status)
      call wrserr(subroutinename,ierr,"Problem reading key: " // trim(fitskey%name))
      return
    end if

  end subroutine fitswrap_read_key_logical
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_read_key_string (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_key_string'

    integer(kind=int32), intent(in)          :: unit
    type(fitskey_string_type), intent(inout) :: fitskey
    integer(kind=int32), intent(inout)       :: ierr

    integer(kind=int32) :: status
    character(len=FITSIO_KEY_MAXLEN) :: comment_copy

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    comment_copy = trim(fitskey%comment)

!    ierr = ERROR_LEVEL_OK
    status = 0

    if ( len_trim(fitskey%name) > 0 ) then
      ! do nothing
    else
      fitskey%flag = FITSWRAP_KEY_FLAG_BLANK
      return ! we therefore do not use this key
    end if

    ! now actually try to read the key from the file
    call FTGKYS_w(unit, trim(fitskey%name), fitskey%value, fitskey%comment, status)

    ! revert to default keyword comment when no keyword is supplied
    if ( len_trim(fitskey%comment) .eq. 0 ) fitskey%comment = trim(comment_copy)

    if ( status == 0 ) then ! all is well
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
    else if ( status == FITSIO_ERROR_VALUE_UNDEFINED ) then
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
      call fitswrap_set_key_null(fitskey)
    else if ( status == FITSIO_ERROR_KEYWORD_NOEXIST ) then
      if ( fitskey%compulsory ) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        call wrserr(subroutinename,ierr,"Could not find compulsory key: " // trim(fitskey%name))
        return
      else ! just quietly ignore this missing key
        fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      end if
      fitskey%value = ''
    else ! there must be some other kind of fitsio error
      ierr = ERROR_LEVEL_SEVERE
      fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      call fitswrap_write_fits_error(subroutinename, status)
      call wrserr(subroutinename,ierr,"Problem reading key: " // trim(fitskey%name))
      return
    end if

!    write(*,'(A,A,A8,A,I0,A,A,A,A,A,I0)') &
!        subroutinename, &
!        " Reading fitskey%name='", trim(fitskey%name), "' len_trim(fitskey%name)=", len_trim(fitskey%name), " value='", &
!        trim(fitskey%value), "' comment='", trim(fitskey%comment), "' status=", status

  end subroutine fitswrap_read_key_string
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_read_key_vlstring (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_key_vlstring'

    integer(kind=int32), intent(in)          :: unit
    type(fitskey_vlstring_type), intent(inout) :: fitskey
    integer(kind=int32), intent(inout)       :: ierr

    integer(kind=int32) :: status
    character(len=FITSIO_KEY_MAXLEN) :: comment_copy

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    comment_copy = trim(fitskey%comment)

!    ierr = ERROR_LEVEL_OK
    status = 0

    if ( len_trim(fitskey%name) > 0 ) then
      ! do nothing
    else
      fitskey%flag = FITSWRAP_KEY_FLAG_BLANK
      return ! we therefore do not use this key
    end if

    ! now actually try to read the key from the file
    call FTGKYS_w(unit, trim(fitskey%name), fitskey%value, fitskey%comment, status)

    ! revert to default keyword comment when no keyword is supplied
    if ( len_trim(fitskey%comment) .eq. 0 ) fitskey%comment = trim(comment_copy)

    if ( status == 0 ) then ! all is well
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
    else if ( status == FITSIO_ERROR_VALUE_UNDEFINED ) then
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
      call fitswrap_set_key_null(fitskey)
    else if ( status == FITSIO_ERROR_KEYWORD_NOEXIST ) then
      if ( fitskey%compulsory ) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        call wrserr(subroutinename,ierr,"Could not find compulsory key: " // trim(fitskey%name))
        return
      else ! just quietly ignore this missing key
        fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      end if
      fitskey%value = ''
    else ! there must be some other kind of fitsio error
      ierr = ERROR_LEVEL_SEVERE
      fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      call fitswrap_write_fits_error(subroutinename, status)
      call wrserr(subroutinename,ierr,"Problem reading key: " // trim(fitskey%name))
      return
    end if

  end subroutine fitswrap_read_key_vlstring
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_read_key_int16 (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_key_int16'

    integer(kind=int32), intent(in)             :: unit
    type(fitskey_int16_type), intent(inout)  :: fitskey
    integer(kind=int32), intent(inout)          :: ierr

    integer(kind=int32) :: status
    integer(kind=int32) :: temp_int32  ! we read value into this 32bit integer as there is no FTGKYI command
    character(len=FITSIO_KEY_MAXLEN) :: comment_copy

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    comment_copy = trim(fitskey%comment)

!    ierr = ERROR_LEVEL_OK
    status = 0

    if ( len_trim(fitskey%name) > 0 ) then
      ! do nothing
    else
      fitskey%flag = FITSWRAP_KEY_FLAG_BLANK
      return ! we therefore do not use this key
    end if

    ! now actually try to read the key from the file
    call FTGKYJ_w(unit, trim(fitskey%name), temp_int32, fitskey%comment, status)
    fitskey%value = INT(temp_int32, int16)

    ! revert to default keyword comment when no keyword is supplied
    if ( len_trim(fitskey%comment) .eq. 0 ) fitskey%comment = trim(comment_copy)

    if ( status == 0 ) then ! all is well
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
    else if ( status == FITSIO_ERROR_VALUE_UNDEFINED ) then
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
      call fitswrap_set_key_null(fitskey)
    else if ( status == FITSIO_ERROR_KEYWORD_NOEXIST ) then
      if ( fitskey%compulsory ) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        call wrserr(subroutinename,ierr,"Could not find compulsory key: " // trim(fitskey%name))
        return
      else ! just quietly ignore this missing key
        fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      end if
    else ! there must be some other kind of fitsio error
      ierr = ERROR_LEVEL_SEVERE
      fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      call fitswrap_write_fits_error(subroutinename, status)
      call wrserr(subroutinename,ierr,"Problem reading key: " // trim(fitskey%name))
      return
    end if

!    write(*,'(A,1x,A10,1x,A,I0,A,A,A,I0)') &
!        "Reading key name=", trim(fitskey%name), "value=", &
!        fitskey%value, " comment='", trim(fitskey%comment), "' status=", status

  end subroutine fitswrap_read_key_int16
  !---------------------------------------------------------------------------------------------



  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_read_key_int32 (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_key_int32'

    integer(kind=int32), intent(in)             :: unit
    type(fitskey_int32_type), intent(inout)  :: fitskey
    integer(kind=int32), intent(inout)          :: ierr
!    character(len=*), intent(in)            :: keyname

    integer(kind=int32) :: status
    character(len=FITSIO_KEY_MAXLEN) :: comment_copy

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    comment_copy = trim(fitskey%comment)

!    ierr = ERROR_LEVEL_OK
    status = 0

    if ( len_trim(fitskey%name) > 0 ) then
      ! do nothing
    else
      fitskey%flag = FITSWRAP_KEY_FLAG_BLANK
      return ! we therefore do not use this key
    end if

    ! now actually try to read the key from the file
    call FTGKYJ_w(unit, trim(fitskey%name), fitskey%value, fitskey%comment, status)

    ! revert to default keyword comment when no keyword is supplied
    if ( len_trim(fitskey%comment) .eq. 0 ) fitskey%comment = trim(comment_copy)

    if ( status == 0 ) then ! all is well
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
    else if ( status == FITSIO_ERROR_VALUE_UNDEFINED ) then
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
      call fitswrap_set_key_null(fitskey)
    else if ( status == FITSIO_ERROR_KEYWORD_NOEXIST ) then
      if ( fitskey%compulsory ) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        call wrserr(subroutinename,ierr,"Could not find compulsory key: " // trim(fitskey%name))
        return
      else ! just quietly ignore this missing key
        fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      end if
    else ! there must be some other kind of fitsio error
      ierr = ERROR_LEVEL_SEVERE
      fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      call fitswrap_write_fits_error(subroutinename, status)
      call wrserr(subroutinename,ierr,"Problem reading key: " // trim(fitskey%name))
      return
    end if

!    write(*,'(A,1x,A10,1x,A,I0,A,A,A,I0)') &
!        "Reading key name=", trim(fitskey%name), "value=", &
!        fitskey%value, " comment='", trim(fitskey%comment), "' status=", status

  end subroutine fitswrap_read_key_int32
  !---------------------------------------------------------------------------------------------



  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_read_key_int64 (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_key_int64'

    integer(kind=int32), intent(in)             :: unit
    type(fitskey_int64_type), intent(inout)  :: fitskey
    integer(kind=int32), intent(inout)          :: ierr

    integer(kind=int32) :: status
    character(len=FITSIO_KEY_MAXLEN) :: comment_copy

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    comment_copy = trim(fitskey%comment)

!    ierr = ERROR_LEVEL_OK
    status = 0

    if ( len_trim(fitskey%name) > 0 ) then
      ! do nothing
    else
      fitskey%flag = FITSWRAP_KEY_FLAG_BLANK
      return ! we therefore do not use this key
    end if

    ! now actually try to read the key from the file
    call FTGKYK_w(unit, trim(fitskey%name), fitskey%value, fitskey%comment, status)
!    call FTGKYJ(unit, trim(fitskey%name), fitskey%value, fitskey%comment, status)

    ! revert to default keyword comment when no keyword is supplied
    if ( len_trim(fitskey%comment) .eq. 0 ) fitskey%comment = trim(comment_copy)

    if ( status == 0 ) then ! all is well
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
    else if ( status == FITSIO_ERROR_VALUE_UNDEFINED ) then
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
      call fitswrap_set_key_null(fitskey)
    else if ( status == FITSIO_ERROR_KEYWORD_NOEXIST ) then
      if ( fitskey%compulsory ) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        call wrserr(subroutinename,ierr,"Could not find compulsory key: " // trim(fitskey%name))
        return
      else ! just quietly ignore this missing key
        fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      end if
    else ! there must be some other kind of fitsio error
      ierr = ERROR_LEVEL_SEVERE
      fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      call fitswrap_write_fits_error(subroutinename, status)
      call wrserr(subroutinename,ierr,"Problem reading key: " // trim(fitskey%name))
      return
    end if

!    write(*,'(A,1x,A10,1x,A,I0,A,A,A,I0)') &
!        "Reading key name=", trim(fitskey%name), "value=", &
!        fitskey%value, " comment='", trim(fitskey%comment), "' status=", status

  end subroutine fitswrap_read_key_int64
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_read_key_real32 (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_key_real32'

    integer(kind=int32), intent(in)             :: unit
    type(fitskey_real32_type), intent(inout) :: fitskey
    integer(kind=int32), intent(inout)          :: ierr

    integer(kind=int32) :: status
    character(len=FITSIO_KEY_MAXLEN) :: comment_copy

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    comment_copy = trim(fitskey%comment)

 !   ierr = ERROR_LEVEL_OK
    status = 0
    if ( len_trim(fitskey%name) > 0 ) then
      ! do nothing
    else
      fitskey%flag = FITSWRAP_KEY_FLAG_BLANK
      return ! we therefore do not use this key
    end if

    ! now actually try to read the key from the file
    call FTGKYE_w(unit, trim(fitskey%name), fitskey%value, fitskey%comment, status)

    ! revert to default keyword comment when no keyword is supplied
    if ( len_trim(fitskey%comment) .eq. 0 ) fitskey%comment = trim(comment_copy)

    if ( status == 0 ) then ! all is well
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
    else if ( status == FITSIO_ERROR_VALUE_UNDEFINED ) then
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
      call fitswrap_set_key_null(fitskey)
    else if ( status == FITSIO_ERROR_KEYWORD_NOEXIST ) then
      if ( fitskey%compulsory ) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        call wrserr(subroutinename,ierr,"Could not find compulsory key: " // trim(fitskey%name))
        return
      else ! just quietly ignore this missing key
        fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      end if
    else ! there must be some other kind of fitsio error
      ierr = ERROR_LEVEL_SEVERE
      fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      call fitswrap_write_fits_error(subroutinename, status)
      call wrserr(subroutinename,ierr,"Problem reading key: " // trim(fitskey%name))
      return
    end if

!    write(*,'(A,1x,A10,1x,A,E12.7,A,A,A,I0)') &
!        "Reading key name=", trim(fitskey%name), "value=", &
!        fitskey%value, " comment='", trim(fitskey%comment), "' status=", status

  end subroutine fitswrap_read_key_real32
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_read_key_real64 (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_key_real64'

    integer(kind=int32), intent(in)             :: unit
    type(fitskey_real64_type), intent(inout) :: fitskey
    integer(kind=int32), intent(inout)          :: ierr

    integer(kind=int32) :: status
    character(len=FITSIO_KEY_MAXLEN) :: comment_copy

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    comment_copy = trim(fitskey%comment)

!    ierr = ERROR_LEVEL_OK
    status = 0
    if ( len_trim(fitskey%name) > 0 ) then
      ! do nothing
    else
      fitskey%flag = FITSWRAP_KEY_FLAG_BLANK
      return ! we therefore do not use this key
    end if

    ! now actually try to read the key from the file
    call FTGKYD_w(unit, trim(fitskey%name), fitskey%value, fitskey%comment, status)

    ! revert to default keyword comment when no keyword is supplied
    if ( len_trim(fitskey%comment) .eq. 0 ) fitskey%comment = trim(comment_copy)

    if ( status == 0 ) then ! all is well
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
    else if ( status == FITSIO_ERROR_VALUE_UNDEFINED ) then
      fitskey%flag = FITSWRAP_KEY_FLAG_SUPPLIED
      call fitswrap_set_key_null(fitskey)
    else if ( status == FITSIO_ERROR_KEYWORD_NOEXIST ) then
      if ( fitskey%compulsory ) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        call wrserr(subroutinename,ierr,"Could not find compulsory key: " // trim(fitskey%name))
        return
      else ! just quietly ignore this missing key
        fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      end if
    else ! there must be some other kind of fitsio error
      ierr = ERROR_LEVEL_SEVERE
      fitskey%flag = FITSWRAP_KEY_FLAG_NOT_SUPPLIED
      call fitswrap_write_fits_error(subroutinename, status)
      call wrserr(subroutinename,ierr,"Problem reading key: " // trim(fitskey%name))
      return
    end if

!    write(*,'(A,1x,A10,1x,A,E12.7,A,A,A,I0)') &
!        "Reading key name=", trim(fitskey%name), "value=", &
!        fitskey%value, " comment='", trim(fitskey%comment), "' status=", status

  end subroutine fitswrap_read_key_real64

  !---------------------------------------------------------------------------------------------


  ! routines to easily get a fits header value
  subroutine fitswrap_read_key_quick_string(unit, key, outval, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_key_quick_string'

    integer(kind=int32), intent(in)        :: unit
    character(len=*), intent(in)           :: key
    character(len=*), intent(out)          :: outval
    integer(kind=int32), intent(inout)     :: ierr

    character(len=FITSIO_KEY_MAXLEN)       :: comment, val
    integer(kind=int32) :: status

    if(ierr .ne. ERROR_LEVEL_OK) return
    status = 0
    call FTGKYS_w(unit, trim(key), val, comment, status)
    if ( status .ne. 0 ) then
       call wrserr(subroutinename, ierr, "Problem reading key: " // trim(key))
       ierr = ERROR_LEVEL_SEVERE
    else
       outval = trim(val)
    end if
  end subroutine fitswrap_read_key_quick_string

  subroutine fitswrap_read_key_quick_real64(unit, key, outval, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_read_key_quick_real64'

    integer(kind=int32), intent(in)        :: unit
    character(len=*), intent(in)           :: key
    real(kind=real64), intent(out)         :: outval
    integer(kind=int32), intent(inout)     :: ierr

    character(len=FITSIO_KEY_MAXLEN)       :: comment
    integer(kind=int32) :: status

    if(ierr .ne. ERROR_LEVEL_OK) return
    status = 0
    call FTGKYD_w(unit, trim(key), outval, comment, status)
    if ( status .ne. 0 ) then
       call wrserr(subroutinename, ierr, "Problem reading key: " // trim(key))
       ierr = ERROR_LEVEL_SEVERE
    end if
  end subroutine fitswrap_read_key_quick_real64

  ! extend here...

  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  !-------------------------End of fitswrap_read_key_* subroutines------------------------------
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------


!  !---------------------------------------------------------------------------------------------
!  ! allocate and read the given general column - must have had column name set previously
!  ! read data into the most suitable internal data format
!  subroutine fitswrap_alloc_and_read_col (unit, col, ierr)
!    character(len=*), parameter :: subroutinename='fitswrap_alloc_and_read_col'
!
!    integer(kind=int32), intent(in)        :: unit
!    type(fcol_type), intent(inout), target :: col
!    integer(kind=int32), intent(inout)     :: ierr
!
!    character(len=STR_MAX) :: str_buffer
!    integer(kind=int32) :: status
!    integer(kind=int32) :: anyf
!
!    ierr = ERROR_LEVEL_OK
!    status = 0
!
!    if ( len_trim(col%name) .le. 0 ) then
!       ierr = ERROR_LEVEL_SEVERE
!       write (str_buffer, '(A,A,A)') &
!           "Column name is not sensible: '", trim(col%name), "'"
!       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, str_buffer)
!       return
!    end if
!
!    ! get the fits column number
!    call FTGCNN(unit, FITSIO_CASEINSENS, col%name, col%matched_name, col%num, status)
!    if ( col%num <= 0 ) then
!       write (str_buffer, '(A,A,A)') &
!           "Could not find fits column: '", trim(col%name), "'"
!       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, str_buffer)
!    end if
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!
!
!  end subroutine fitswrap_alloc_and_read_fcol
!  !---------------------------------------------------------------------------------------------







  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_write_key_real32 (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_write_key_real32'
    integer(kind=int32), intent(in)             :: unit
    type(fitskey_real32_type), intent(in)       :: fitskey
    integer(kind=int32), intent(inout)          :: ierr
    integer(kind=int32) :: status

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    ierr = ERROR_LEVEL_OK
    status = 0

!    write(*,'(A,A,A)') subroutinename, ": Attempting to write out key: ", trim(fitskey%name)

    if ( fitskey%is_null ) then
      call FTPKYU_w(unit,trim(fitskey%name),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    ! write the key to the file
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call FTPKYE_w(unit,trim(fitskey%name),fitskey%value,FITSIO_DECIMALS,trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        return
      end if
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr, "Attempted to write out uninitialised key: " // trim(fitskey%name) )
      return
    end if
  end subroutine fitswrap_write_key_real32
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_write_key_real64 (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_write_key_real64'
    integer(kind=int32), intent(in)             :: unit
    type(fitskey_real64_type), intent(in)       :: fitskey
    integer(kind=int32), intent(inout)          :: ierr
    integer(kind=int32) :: status

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    ierr = ERROR_LEVEL_OK
    status = 0

!    write(*,'(A,A,A)') subroutinename, ": Attempting to write out key: ", trim(fitskey%name)

    if ( fitskey%is_null ) then
      call FTPKYU_w(unit,trim(fitskey%name),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    ! write the key to the file
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call FTPKYD_w(unit,trim(fitskey%name),fitskey%value,FITSIO_DECIMALS,trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr,"Problem writing key: " //  trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr, "Attempted to write out uninitialised key: " // trim(fitskey%name) )
      return
    end if
  end subroutine fitswrap_write_key_real64
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_write_key_int16 (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_write_key_int16'
    integer(kind=int32), intent(in)            :: unit
    type(fitskey_int16_type), intent(in)       :: fitskey
    integer(kind=int32), intent(inout)         :: ierr
    integer(kind=int32) :: status

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    ierr = ERROR_LEVEL_OK
    status = 0

!    write(*,'(A,A,A)') subroutinename, ": Attempting to write out key: ", trim(fitskey%name)

    if ( fitskey%is_null ) then
      call FTPKYU_w(unit,trim(fitskey%name),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    ! write the key to the file
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call FTPKYJ_w(unit,trim(fitskey%name),INT(fitskey%value,int32),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr, "Attempted to write out uninitialised key: " // trim(fitskey%name) )
      return
    end if
  end subroutine fitswrap_write_key_int16
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_write_key_int32 (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_write_key_int32'
    integer(kind=int32), intent(in)            :: unit
    type(fitskey_int32_type), intent(in)       :: fitskey
    integer(kind=int32), intent(inout)         :: ierr
    integer(kind=int32) :: status

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    ierr = ERROR_LEVEL_OK
    status = 0

!    write(*,'(A,A,A)') subroutinename, ": Attempting to write out key: ", trim(fitskey%name)

    ! write null key to the file
    if ( fitskey%is_null ) then
      call FTPKYU_w(unit,trim(fitskey%name),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    ! write the key to the file
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call FTPKYJ_w(unit,trim(fitskey%name),INT(fitskey%value,int32),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr, "Attempted to write out uninitialised key: " // trim(fitskey%name) )
      return
    end if
  end subroutine fitswrap_write_key_int32
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_write_key_int64 (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_write_key_int64'
    integer(kind=int32), intent(in)            :: unit
    type(fitskey_int64_type), intent(in)       :: fitskey
    integer(kind=int32), intent(inout)         :: ierr
    integer(kind=int32) :: status

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    ierr = ERROR_LEVEL_OK
    status = 0

!    write(*,'(A,A,A)') subroutinename, ": Attempting to write out key: ", trim(fitskey%name)

    ! write nul key to the file
    if ( fitskey%is_null ) then
      call FTPKYU_w(unit,trim(fitskey%name),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        return
      end if
    ! write the key to the file
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      call FTPKYK_w(unit,trim(fitskey%name),INT(fitskey%value,int64),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr, "Attempted to write out uninitialised key: " // trim(fitskey%name) )
      return
    end if
  end subroutine fitswrap_write_key_int64
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_write_key_string (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_write_key_string'
    integer(kind=int32), intent(in)            :: unit
    type(fitskey_string_type), intent(in)      :: fitskey
    integer(kind=int32), intent(inout)         :: ierr
    integer(kind=int32) :: status

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    ierr = ERROR_LEVEL_OK
    status = 0

!    write(*,'(A,A,A)') subroutinename, ": Attempting to write out key: ", trim(fitskey%name)

    ! write null key to the file
    if ( fitskey%is_null ) then
      call FTPKYU_w(unit,trim(fitskey%name),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    ! write the key to the file
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      ! write as a long string
      call FTPKLS_w(unit,trim(fitskey%name),trim(fitskey%value),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr, "Attempted to write out uninitialised key: " // trim(fitskey%name) )
      return
    end if
  end subroutine fitswrap_write_key_string
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_write_key_vlstring (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_write_key_vlstring'
    integer(kind=int32), intent(in)            :: unit
    type(fitskey_vlstring_type), intent(in)      :: fitskey
    integer(kind=int32), intent(inout)         :: ierr
    integer(kind=int32) :: status

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

    ierr = ERROR_LEVEL_OK
    status = 0

!    write(*,'(A,A,A)') subroutinename, ": Attempting to write out key: ", trim(fitskey%name)

    ! write null key to the file
    if ( fitskey%is_null ) then
      call FTPKYU_w(unit,trim(fitskey%name),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    ! write the key to the file
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      ! write as a long string
      call FTPKLS_w(unit,trim(fitskey%name),trim(fitskey%value),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr, "Attempted to write out uninitialised key: " // trim(fitskey%name) )
      return
    end if
  end subroutine fitswrap_write_key_vlstring
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_write_key_logical (unit, fitskey, ierr)
    character(len=*), parameter :: subroutinename='fitswrap_write_key_logical'
    integer(kind=int32), intent(in)            :: unit
    type(fitskey_logical_type), intent(in)     :: fitskey
    integer(kind=int32), intent(inout)         :: ierr
    integer(kind=int32) :: status
    integer(kind=int32) :: cvalue

    if ( ierr .ne.  ERROR_LEVEL_OK ) return

!    integer(kind=int32) :: value
    ierr = ERROR_LEVEL_OK
    status = 0
    ! write the key to the file

!    write(*,'(A,A,A)') subroutinename, ": Attempting to write out key: ", trim(fitskey%name)


!    if ( fitskey%value ) then
!      value = 1
!    else
!      value = 0
!    end if
!    call FTPKYS(unit,trim(fitskey%name),trim(logical2char(fitskey%value)),trim(fitskey%comment), status)
    ! write null key to the file
    if ( fitskey%is_null ) then
      call FTPKYU_w(unit,trim(fitskey%name),trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    ! write the key to the file
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then

      ! fitsio code seems to assume C standard logical values
      if( fitskey%value ) then
         cvalue = 1
      else
         cvalue = 0
      end if

      call FTPKYL_w(unit, trim(fitskey%name), cvalue, trim(fitskey%comment), status)
      if  ( status .ne. 0 ) then ! there must be some other kind of fitsio error
        ierr = ERROR_LEVEL_SEVERE
        call wrserr(subroutinename,ierr, "Problem writing key: " // trim(fitskey%name))
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr, "Attempted to write out uninitialised key: " // trim(fitskey%name) )
      return
    end if
  end subroutine fitswrap_write_key_logical
  !---------------------------------------------------------------------------------------------






  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  !-----------------------Start of fitswrap_sprint_key_*() routines-----------------------------
  !---------------------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------------------
  ! print key to string
  subroutine fitswrap_sprint_key_header (outstr)
    character(len=*), intent(out) :: outstr
    write(outstr,'(A12,1x,A1,1x,A24,1x,A,1x,A)') "#KEYNAME","=", "VALUE", "//","COMMENT"
  end subroutine fitswrap_sprint_key_header

  subroutine fitswrap_sprint_key_real64 (fitskey, outstr)
    type(fitskey_real64_type), intent(in) :: fitskey
    character(len=*), intent(out) :: outstr
    outstr = ""
    if ( fitskey%is_null ) then
      write(outstr,'(A12,1x,A,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", "NULL", "//",trim(fitskey%comment)
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      write(outstr,'(A12,1x,A1,1x,ES24.7,1x,A,1x,A)') &
          trim(fitskey%name),"=", fitskey%value, "//",trim(fitskey%comment)
    else if ( len_trim(fitskey%name) > 0 ) then
      write(outstr,'(A12,1x,A,1x,A,1x,A)') &
          trim(fitskey%name), FITSWRAP_UNKNOWN_KEY_VALUE, "//", "!!value unset/unread!!"
    else
      write(*, '(A,A)') "Error: Attempted to print uninitialised key: ", trim(fitskey%name)
    end if
  end subroutine fitswrap_sprint_key_real64

  subroutine fitswrap_sprint_key_real32 (fitskey, outstr)
    type(fitskey_real32_type), intent(in) :: fitskey
    character(len=*), intent(out) :: outstr
    outstr = ""
    if ( fitskey%is_null ) then
      write(outstr,'(A12,1x,A,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", "NULL", "//",trim(fitskey%comment)
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      write(outstr,'(A12,1x,A1,1x,ES24.7,1x,A,1x,A)') &
          trim(fitskey%name),"=", fitskey%value, "//",trim(fitskey%comment)
    else if ( len_trim(fitskey%name) > 0 ) then
      write(outstr,'(A12,1x,A,1x,A,1x,A)') &
          trim(fitskey%name), FITSWRAP_UNKNOWN_KEY_VALUE, "//", "!!value unset/unread!!"
    else
      write(*, '(A,A)') "Error: Attempted to print uninitialised key: ", trim(fitskey%name)
    end if
  end subroutine fitswrap_sprint_key_real32

  subroutine fitswrap_sprint_key_int16 (fitskey, outstr)
    type(fitskey_int16_type), intent(in) :: fitskey
    character(len=*), intent(out) :: outstr
    outstr = ""
    if ( fitskey%is_null ) then
      write(outstr,'(A12,1x,A,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", "NULL", "//",trim(fitskey%comment)
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      write(outstr,'(A12,1x,A,1x,I24,1x,A,1x,A)') &
          trim(fitskey%name),"=", fitskey%value, "//",trim(fitskey%comment)
    else if ( len_trim(fitskey%name) > 0 ) then
      write(outstr,'(A12,1x,A,1x,A,1x,A)') &
          trim(fitskey%name), FITSWRAP_UNKNOWN_KEY_VALUE, "//", "!!value unset/unread!!"
    else
      write(*, '(A,A)') "Error: Attempted to print uninitialised key: ", trim(fitskey%name)
    end if
  end subroutine fitswrap_sprint_key_int16

  subroutine fitswrap_sprint_key_int32 (fitskey, outstr)
    type(fitskey_int32_type), intent(in) :: fitskey
    character(len=*), intent(out) :: outstr
    outstr = ""
    if ( fitskey%is_null ) then
      write(outstr,'(A12,1x,A,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", "NULL", "//",trim(fitskey%comment)
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      write(outstr,'(A12,1x,A,1x,I24,1x,A,1x,A)') &
          trim(fitskey%name),"=", fitskey%value, "//",trim(fitskey%comment)
    else if ( len_trim(fitskey%name) > 0 ) then
      write(outstr,'(A12,1x,A,1x,A,1x,A)') &
          trim(fitskey%name),FITSWRAP_UNKNOWN_KEY_VALUE, "//", "!!value unset/unread!!"
    else
      write(*, '(A,A)') "Error: Attempted to print uninitialised key: ", trim(fitskey%name)
    end if
  end subroutine fitswrap_sprint_key_int32

  subroutine fitswrap_sprint_key_int64 (fitskey, outstr)
    type(fitskey_int64_type), intent(in) :: fitskey
    character(len=*), intent(out) :: outstr
    outstr = ""
    if ( fitskey%is_null ) then
      write(outstr,'(A12,1x,A,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", "NULL", "//",trim(fitskey%comment)
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      write(outstr,'(A12,1x,A1,1x,I24,1x,A,1x,A)') &
          trim(fitskey%name),"=", fitskey%value, "//",trim(fitskey%comment)
    else if ( len_trim(fitskey%name) > 0 ) then
      write(outstr,'(A12,1x,A,1x,A,1x,A)') &
          trim(fitskey%name),FITSWRAP_UNKNOWN_KEY_VALUE, "//", "!!value unset/unread!!"
    else
      write(*, '(A,A)') "Error: Attempted to print uninitialised key: ", trim(fitskey%name)
    end if
  end subroutine fitswrap_sprint_key_int64
  !---------------------------------------------------------------------------------------------

  subroutine fitswrap_sprint_key_string (fitskey, outstr)
    type(fitskey_string_type), intent(in) :: fitskey
    character(len=*), intent(out) :: outstr
    outstr = ""
    if ( fitskey%is_null ) then
      write(outstr,'(A12,1x,A,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", "NULL", "//",trim(fitskey%comment)
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      write(outstr,'(A12,1x,A1,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", trim(fitskey%value), "//",trim(fitskey%comment)
    else if ( len_trim(fitskey%name) > 0 ) then
      write(outstr,'(A12,1x,A,1x,A,1x,A)') &
          trim(fitskey%name), FITSWRAP_UNKNOWN_KEY_VALUE, "//", "!!value unset/unread!!"
    else
      write(*, '(A,A)') "Error: Attempted to print uninitialised key: ", trim(fitskey%name)
    end if
  end subroutine fitswrap_sprint_key_string
  !---------------------------------------------------------------------------------------------

  subroutine fitswrap_sprint_key_vlstring (fitskey, outstr)
    type(fitskey_vlstring_type), intent(in) :: fitskey
    character(len=*), intent(out) :: outstr
    outstr = ""
    if ( fitskey%is_null ) then
      write(outstr,'(A12,1x,A,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", "NULL", "//",trim(fitskey%comment)
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      write(outstr,'(A12,1x,A1,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", trim(fitskey%value), "//",trim(fitskey%comment)
    else if ( len_trim(fitskey%name) > 0 ) then
      write(outstr,'(A12,1x,A,1x,A,1x,A)') &
          trim(fitskey%name), FITSWRAP_UNKNOWN_KEY_VALUE, "//", "!!value unset/unread!!"
    else
      write(*, '(A,A)') "Error: Attempted to print uninitialised key: ", trim(fitskey%name)
    end if
  end subroutine fitswrap_sprint_key_vlstring
  !---------------------------------------------------------------------------------------------

  subroutine fitswrap_sprint_key_logical (fitskey, outstr)
    type(fitskey_logical_type), intent(in) :: fitskey
    character(len=*), intent(out) :: outstr
    outstr = ""
    if ( fitskey%is_null ) then
      write(outstr,'(A12,1x,A,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", "NULL", "//",trim(fitskey%comment)
    else if ( fitskey%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      write(outstr,'(A12,1x,A1,1x,A24,1x,A,1x,A)') &
          trim(fitskey%name),"=", logical2char(fitskey%value), "//", trim(fitskey%comment)
    else if ( len_trim(fitskey%name) > 0 ) then
      write(outstr,'(A12,1x,A,1x,A,1x,A)') &
          trim(fitskey%name), FITSWRAP_UNKNOWN_KEY_VALUE, "//", "!!value unset/unread!!"
    else
      write(*, '(A,A)') "Error: Attempted to print uninitialised key: ", trim(fitskey%name)
    end if
  end subroutine fitswrap_sprint_key_logical

  !---------------------------------------------------------------------------------------------
  !-------------------------End of fitswrap_sprint_key_*() routines-----------------------------
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  subroutine fitswrap_print_hduinfo(hduinfo)
    character(len=*), parameter :: subroutinename='fitswrap_print_hduinfo'
    type(hduinfo_type), intent(in) :: hduinfo
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32)    :: i

    write(*,'(A,1x,A)') subroutinename, "Printing hduinfo for filename:"
    write(*,'(A,1x,A)') subroutinename, trim(hduinfo%filename)
    write(*,'(A,1x,A,I0,A,I0)') subroutinename, "HDUINDEX=", hduinfo%hdu_index, " HDUTYPE=", hduinfo%hdu_type
    write(*,'(A,1x,A)') subroutinename, "START"
    call fitswrap_sprint_key(hduinfo%key_extname , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_hduclass, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_hduclas1, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_hduclas2, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_hduclas3, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_hduclas4, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_hduvers , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)

    call fitswrap_sprint_key(hduinfo%key_origin  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_creator , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_owner,    str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_mission , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_telescop, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_instrume, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_obs_mode, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_datamode, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_frametim, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_skyfield, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_filter  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_obs_id  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_exp_id  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_observer, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_object  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_ra_obj  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_dec_obj , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_ra_cen  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_dec_cen , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_fieldlon, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_fieldlat, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_date_obs, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_date_end, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_tstart  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_tstop   , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_mjdref  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_timezero, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_timeunit, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_timesys , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_ra_pnt  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_dec_pnt , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_pa_pnt  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_radecsys, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_equinox , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_longstr , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)

    ! only write out NINST and INSTRUMx keywords when INSTRUME == "MERGED"
    if ( trim(hduinfo%key_instrume%value) .eq. INSTRUME_IS_MERGED ) then
      call fitswrap_sprint_key (hduinfo%key_ninst, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)

      if ((hduinfo%key_ninst%value .ge. 1) .and. (hduinfo%key_ninst%value .le. MAX_INST) ) then
        do i=1,hduinfo%key_ninst%value
          call fitswrap_sprint_key(hduinfo%pkey_instrum(i), str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
        end do
      endif
    end if

    call fitswrap_sprint_key(hduinfo%key_prefix  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_dscode  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_dstype  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_version , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
    call fitswrap_sprint_key(hduinfo%key_config  , str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)

!    if ( associated(hduinfo%key_sasshist) ) then
      call fitswrap_sprint_key(hduinfo%key_sasshist, str_buffer); write(*,'(A,1x,A)') subroutinename, trim(str_buffer)
!    endif

    write(*,'(A,1x,A)') subroutinename, "END"

  end subroutine fitswrap_print_hduinfo

  !---------------------------------------------------------------------------------------------



  !------------------------------------------------------
  ! set is_fits to +1 if file is openable and file is fits
  ! set is_fits to  0 if file is openable and file is not fits
  ! set is_fits to -1 otherwise
  subroutine fitswrap_test_if_is_fits ( filename, is_fits )
    character(len=*), parameter :: subroutinename='fitswrap_test_if_is_fits'
    character(len=*)  , intent(in)  :: filename
    integer(kind=int32), intent(out) :: is_fits
    integer(kind=int32)     :: unit
    integer                 :: status
    logical                 :: exist_flag
    character(len=STR_MAX)  :: read_state
!    character(len=:), allocatable :: filename_copy
    character(len=LSTR_MAX) :: filename_copy
    character(len=VLSTR_MAX) :: str_buffer
    status = 0
    is_fits = -1

    if ( len_trim(adjustl(filename)) .lt. 1 ) then
      is_fits = -1
      return
    end if

!    ! make a copy of the filename
!    ALLOCATE(character(len=LEN_TRIM(ADJUSTL(filename))) :: filename_copy, &
!        STAT = status)
!    if (status .ne. 0) then
!      is_fits = -1
!      return
!    end if
     if ( LEN_TRIM(ADJUSTL(filename)) .gt. len(filename_copy)) then
       call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "filename string is too long to process")
       is_fits = -1
       return
     endif



    ! try to trim off any bits of the filename which are fits HDU/selection expressions
    call FTRTNM_w(TRIM(ADJUSTL(filename)), filename_copy, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      is_fits = -1
      return
    end if

    ! check to see if the a file exists with this filename
    inquire(FILE=TRIM(filename_copy), IOSTAT=status, EXIST=exist_flag, READ=read_state)

    ! if we were successful
    if ((status .eq. 0) .and. exist_flag ) then    ! ok great, I found the name of a file which exists

      ! check that we can actually read this file:
      if ( TRIM(read_state) .ne. "YES" ) then
        write (str_buffer, '(A,A,A,A)') &
            "I do not have read access to file: '",TRIM(filename_copy),"'"
        call wrserr (subroutinename, ERROR_LEVEL_SEVERE, str_buffer)
        is_fits = -1
        return
      end if

      ! get an i/o unit - use the fitsio routine for this
      call FTGIOU_w(unit, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        is_fits = -1
        return
      end if

      ! now try to open as a fits file
      ! open file
      call FTTOPN_w(unit, filename_copy, FITSIO_READONLY, status)
      if (status .eq. 0) then
        is_fits = 1
      else
        is_fits = 0
      end if
      status = 0

      !close the file
      call FTCLOS_w(unit, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        is_fits = -1
      end if

      !and release the unit number
      call FTFIOU_w(unit, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        is_fits = -1
      end if

    end if
    ! free up the allocated space
!    if ( allocated(filename_copy) ) deallocate(filename_copy)

  end subroutine fitswrap_test_if_is_fits
  !------------------------------------------------------


  function fitswrap_check_status (subroutinename, status, ierr) result(flag)
    character(len=*), intent(in) ::  subroutinename
    integer(kind=int32), intent(in) :: status
    integer(kind=int32), intent(inout)  :: ierr
    logical            :: flag
    character(len=30) :: str_errtext
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call FTGERR_w(status, str_errtext)
      call wrserr(trim(subroutinename), ierr, &
          "FITSIO error (status=" // trim(i2s(status)) // "): " // trim(str_errtext))
      flag = .true.
    else
      flag = .false.
    end if
  end function fitswrap_check_status




!---------------------------------------------------------------------------------------------

! Wrappers to column reading routines

subroutine fitswrap_read_column_real64(unit, colnum, frow, felem, nelements, &
     nullval, values, anyf, status)

  integer(kind=int32), intent(in) :: unit
  integer(kind=int32), intent(in) :: colnum
  integer(kind=int32), intent(in) :: frow
  integer(kind=int32), intent(in) :: felem
  integer(kind=int32), intent(in) :: nelements

  real(kind=real64), intent(in) :: nullval
  real(kind=real64), intent(out), dimension(:) :: values
  integer(kind=int32), intent(out) :: anyf
  integer(kind=int32), intent(out) :: status

  call FTGCVD_w(unit, colnum, frow, felem, nelements, nullval, values, anyf, status)
end subroutine fitswrap_read_column_real64

subroutine fitswrap_read_column_real32(unit, colnum, frow, felem, nelements, &
     nullval, values, anyf, status)

  integer(kind=int32), intent(in) :: unit
  integer(kind=int32), intent(in) :: colnum
  integer(kind=int32), intent(in) :: frow
  integer(kind=int32), intent(in) :: felem
  integer(kind=int32), intent(in) :: nelements

  real(kind=real32), intent(in) :: nullval
  real(kind=real32), intent(out), dimension(:) :: values
  integer(kind=int32), intent(out) :: anyf
  integer(kind=int32), intent(out) :: status

  call FTGCVE_w(unit, colnum, frow, felem, nelements, nullval, values, anyf, status)
end subroutine fitswrap_read_column_real32

subroutine fitswrap_read_column_int64(unit, colnum, frow, felem, nelements, &
     nullval, values, anyf, status)

  integer(kind=int32), intent(in) :: unit
  integer(kind=int32), intent(in) :: colnum
  integer(kind=int32), intent(in) :: frow
  integer(kind=int32), intent(in) :: felem
  integer(kind=int32), intent(in) :: nelements

  integer(kind=int64), intent(in) :: nullval
  integer(kind=int64), intent(out), dimension(:) :: values
  integer(kind=int32), intent(out) :: anyf
  integer(kind=int32), intent(out) :: status

  call FTGCVK_w(unit, colnum, frow, felem, nelements, nullval, values, anyf, status)
end subroutine fitswrap_read_column_int64

subroutine fitswrap_read_column_int32(unit, colnum, frow, felem, nelements, &
     nullval, values, anyf, status)

  integer(kind=int32), intent(in) :: unit
  integer(kind=int32), intent(in) :: colnum
  integer(kind=int32), intent(in) :: frow
  integer(kind=int32), intent(in) :: felem
  integer(kind=int32), intent(in) :: nelements

  integer(kind=int32), intent(in) :: nullval
  integer(kind=int32), intent(out), dimension(:) :: values
  integer(kind=int32), intent(out) :: anyf
  integer(kind=int32), intent(out) :: status

  call FTGCVJ_w(unit, colnum, frow, felem, nelements, nullval, values, anyf, status)
end subroutine fitswrap_read_column_int32

subroutine fitswrap_read_column_int16(unit, colnum, frow, felem, nelements, &
     nullval, values, anyf, status)

  integer(kind=int32), intent(in) :: unit
  integer(kind=int32), intent(in) :: colnum
  integer(kind=int32), intent(in) :: frow
  integer(kind=int32), intent(in) :: felem
  integer(kind=int32), intent(in) :: nelements

  integer(kind=int16), intent(in) :: nullval
  integer(kind=int16), intent(out), dimension(:) :: values
  integer(kind=int32), intent(out) :: anyf
  integer(kind=int32), intent(out) :: status

  call FTGCVI_w(unit, colnum, frow, felem, nelements, nullval, values, anyf, status)
end subroutine fitswrap_read_column_int16

subroutine fitswrap_read_column_int8(unit, colnum, frow, felem, nelements, &
     nullval, values, anyf, status)

  integer(kind=int32), intent(in) :: unit
  integer(kind=int32), intent(in) :: colnum
  integer(kind=int32), intent(in) :: frow
  integer(kind=int32), intent(in) :: felem
  integer(kind=int32), intent(in) :: nelements

  integer(kind=int8), intent(in) :: nullval
  integer(kind=int8), intent(out), dimension(:) :: values
  integer(kind=int32), intent(out) :: anyf
  integer(kind=int32), intent(out) :: status

  call FTGCVB_w(unit, colnum, frow, felem, nelements, nullval, values, anyf, status)
end subroutine fitswrap_read_column_int8

end module  srctool_generic_fitswrap
