! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_gti

  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling GTIs
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         !
  use srctool_generic_fitsraw

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_GTI_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_gti
  public ::  srctool_gti_merge_gtilists_for_inst
  public ::  srctool_gti_setup_src_gtilist
  public ::  srctool_gti_update_src_gtilist
  public ::  srctool_gti_append_gtilist_hdu
  public ::  srctool_gti_read_gtilist_from_chdu
  public ::  srctool_gti_empty_gtilist
  public ::  srctool_gti_merge_gtilists_for_instlist
  public ::  srctool_gti_duplicate_gtilist
  public ::  srctool_gti_test_overlap_with_gtilist
  public ::  srctool_gti_test_if_in_gtilist
  public ::  srctool_gti_get_gti_index
  public ::  srctool_gti_get_gti_bounds
  public ::  srctool_gti_get_next_gti_index
contains

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- START of GTI handling -----------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_gti_empty_gtilist(gtilist)
    type(gtilist_type), intent(inout)      :: gtilist
!    integer(kind=int32) :: i
    if ( allocated(gtilist%pstart)) deallocate (gtilist%pstart)
    if ( allocated(gtilist%pstop))  deallocate (gtilist%pstop)
    gtilist%num_rows = 0
    gtilist%is_valid = .false.
    call fitswrap_init_hduinfo(gtilist%hduinfo)
    call fitswrap_init_key(gtilist%key_telapse,  "TELAPSE" , "[s] Total observation duration")
    call fitswrap_init_key(gtilist%key_ontime,   "ONTIME"  , "[s] On-target exposure time (pre deadtime corr)")
    call fitswrap_init_key(gtilist%key_livetime, "LIVETIME", "[s] On-target active-detector exposure time")
    call fitswrap_init_key(gtilist%key_exposure, "EXPOSURE", "[s] Fully corrected effective exposure time")


!    call fitswrap_init_key(gtilist%key_ninst, "NINST", "Number of instruments that were merged")
!    do i=1,MAX_INST
!      call fitswrap_init_key(gtilist%pkey_instrum(i), "INSTRUM" // trim(i2s(i)), "Contributing instrument name")
!    end do


  end subroutine srctool_gti_empty_gtilist
  !---------------------------------------------------------------------------------------------
!  !-------------------------------------------------------------------------------
!  subroutine srctool_gti_empty_gtilists(gtilists)
!    type(gtilist_type), dimension(:), intent(inout)      :: gtilists
!    integer(kind=int32) :: i
!    do i=1, size(gtilists)
!      call srctool_gti_empty_gtilist(gtilists(i))
!    end do
!  end subroutine srctool_gti_empty_gtilists
!  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_gti_duplicate_gtilist(gtilist_in, gtilist_out, ierr)
    character(len=*), parameter :: subroutinename='srctool_gti_duplicate_gtilist'
    type(gtilist_type), intent(in)     :: gtilist_in
    type(gtilist_type), intent(inout)  :: gtilist_out
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: status

    ierr=ERROR_LEVEL_OK
    status = 0

    call srctool_gti_empty_gtilist(gtilist_out)
    gtilist_out%is_valid = gtilist_in%is_valid
    gtilist_out%hduinfo  = gtilist_in%hduinfo

    gtilist_out%key_telapse   = gtilist_out%key_telapse
    gtilist_out%key_ontime    = gtilist_out%key_ontime
    gtilist_out%key_livetime  = gtilist_out%key_livetime
    gtilist_out%key_exposure  = gtilist_out%key_exposure

    gtilist_out%num_rows  = gtilist_in%num_rows
    if ( gtilist_out%num_rows .ge. 1 ) then
      allocate(gtilist_out%pstart(gtilist_out%num_rows),&
          gtilist_out%pstop(gtilist_out%num_rows), &
          STAT = status)
      if (check_allocate_status(subroutinename, status, ierr)) return
      gtilist_out%pstart(1:gtilist_out%num_rows) = gtilist_in%pstart(1:gtilist_out%num_rows)
      gtilist_out%pstop(1:gtilist_out%num_rows)  = gtilist_in%pstop(1:gtilist_out%num_rows)

    end if
  end subroutine srctool_gti_duplicate_gtilist
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_gti_read_gtilist_from_chdu(unit, gtilist, ierr)
    character(len=*), parameter :: subroutinename='srctool_gti_read_gtilist_from_chdu'

    integer(kind=int32), intent(in)    :: unit
    type(gtilist_type),  intent(inout) :: gtilist
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32) :: status
    character(len=STR_MAX) :: str_buffer

    integer(kind=int32) :: colnum_start
    integer(kind=int32) :: colnum_stop
    integer(kind=int32) :: nrows_per_chunk
    integer(kind=int32) :: nrows_to_read
    integer(kind=int32) :: row
    integer(kind=int32) :: row_end
    integer(kind=int32) :: anyf
    real(kind=real64)  , parameter :: nullval_real64  = 0.0d0

    ierr = ERROR_LEVEL_OK
    status = 0

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, gtilist%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if

    !get the number of rows in the table extension
    call FTGNRW_w(unit, row, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if
    gtilist%num_rows = int(row, int32)

    ! check that have at least some rows, and if so read in the column data
    if ( gtilist%num_rows .gt. 0 ) then
      write (str_buffer, '(A,I0,A,A,A,A,A)') &
          "Reading in: ", gtilist%num_rows, " rows of GTI info from extname='", &
          trim(gtilist%hduinfo%key_extname%value), "' instrument='", &
          trim(gtilist%hduinfo%key_instrume%value), "'"
      call wrschatty (subroutinename, str_buffer)

!      call fitswrap_proc_col(unit, "START", gtilist%num_rows, gtilist%pstart, ierr)
!      call fitswrap_proc_col(unit, "STOP" , gtilist%num_rows, gtilist%pstop, ierr)
!      if (ierr .ne. 0) then
!        ierr = ERROR_LEVEL_SEVERE
!        call wrserr (subroutinename, ierr, "Problem reading column(s) from HDU")
!        return
!      end if
      call fitswrap_prep_col(unit, "START", gtilist%num_rows, colnum_start, gtilist%pstart, ierr)
      call fitswrap_prep_col(unit, "STOP" , gtilist%num_rows, colnum_stop,  gtilist%pstop,  ierr)
      if (ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem preparing column(s) for reading from HDU")
        return
      end if
      ! get the chunksize
      call FTGRSZ_w(unit, nrows_per_chunk, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      do row = 1, gtilist%num_rows, nrows_per_chunk
        nrows_to_read = min(nrows_per_chunk,gtilist%num_rows-row+1)
        row_end = row+nrows_to_read-1
        call FTGCVD_w(unit, colnum_start,row,1,nrows_to_read,nullval_real64,gtilist%pstart(row:row_end), anyf,status)
        call FTGCVD_w(unit, colnum_stop, row,1,nrows_to_read,nullval_real64,gtilist%pstop(row:row_end),  anyf,status)
        if (status .ne. 0) then
          call fitswrap_write_fits_error(subroutinename, status)
          ierr = ERROR_LEVEL_SEVERE
          return
        end if
      end do

      if ( gtilist%num_rows .gt. 0 ) then
        write (str_buffer, '(A,F12.2,A,I0,A,F12.2)') &
            "GTI details: START[1]= ", gtilist%pstart(1),  &
            " STOP[", gtilist%num_rows, "]= ", gtilist%pstop(gtilist%num_rows)
       call wrschatty (subroutinename, str_buffer)
     end if

      if ( status .ne. 0 ) then
        ierr  = ERROR_LEVEL_SEVERE
        return
      end if
    else
      call wrslog (subroutinename, "Warning: no GTI entries found in extension")
    end if
    gtilist%is_valid = .true.
  end subroutine srctool_gti_read_gtilist_from_chdu
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! merge gtis across all eventlists for an instrument
! place result in inst%gtilist
  subroutine srctool_gti_merge_gtilists_for_inst(inst, ierr)
    character(len=*), parameter :: subroutinename='srctool_gti_merge_gtilists_for_inst'
    type(inst_type),  intent(inout) :: inst
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i, j1,j2,n, k
    integer(kind=int32) :: status
    real(kind=real64), allocatable, dimension(:) :: tstart,tstop
    integer(kind=int32)  :: num_rows
    character(len=STR_MAX) :: str_buffer
    ierr = ERROR_LEVEL_OK
    status = 0

    call srctool_gti_empty_gtilist(inst%gtilist)

    ! copy the header from the first evtlist
    if ( inst%evtlistlist%num_evtlists > 0 ) then
      inst%gtilist%hduinfo = inst%evtlistlist%evtlist(1)%gtilist%hduinfo
      ! TODO update the correct values of the TSTART, TSTOP, DATE-OBS, DATE-END etc keywords
    else
      ! exit quietly
      return
    end if

    ! make sure the keywords are set to null

    call fitswrap_set_key_null(inst%gtilist%key_telapse)
    call fitswrap_set_key_null(inst%gtilist%key_ontime)
    call fitswrap_set_key_null(inst%gtilist%key_livetime)
    call fitswrap_set_key_null(inst%gtilist%key_exposure)

    num_rows = 0
    do i=1,inst%evtlistlist%num_evtlists
      if ( inst%evtlistlist%evtlist(i)%gtilist%is_valid ) then
        num_rows = num_rows + inst%evtlistlist%evtlist(i)%gtilist%num_rows
      end if
    end do

    ! exit if there are no valid GTIs
    if( num_rows == 0 ) then
       write (str_buffer, *) "No GTIs found for inst: ", trim(inst%name)
       call wrslog (subroutinename, str_buffer)
       return
    end if

    allocate(tstart(num_rows), &
        tstop(num_rows), &
        STAT = status)
    if (check_allocate_status(subroutinename, status, ierr)) return

    j1 = 1
    do i=1,inst%evtlistlist%num_evtlists
      if ( inst%evtlistlist%evtlist(i)%gtilist%is_valid ) then
        n = inst%evtlistlist%evtlist(i)%gtilist%num_rows
        j2 = j1 + n - 1

        tstart(j1:j2) = inst%evtlistlist%evtlist(i)%gtilist%pstart(1:n)
        tstop(j1:j2)  = inst%evtlistlist%evtlist(i)%gtilist%pstop(1:n)
        j1 = j2+1
      end if
    end do

    call cosort_arrays(tstart, tstop)

    ! first go through and make sure that any intervals with matching tstart are filtered out
    ! also check for overlapping/consecutive gtis and merge them
!    k = 1
!    do i=1,num_rows
!      if ( eq_real(tstart(i), tstart(k))) then
!        if ( tstop(i) .gt. tstop(k) ) tstop(k) = tstop(i)
!      else
!        if (tstart(i) .le. tstop(k)) then
!          if ( tstop(i) .gt. tstop(k) ) tstop(k) = tstop(i)
!        else
!          k = k + 1
!          tstart(k) = tstart(i)
!          tstop(k) = tstop(i)
!        end if
!      end if
!    end do
!    num_rows = k

    k = 1
    do i = 1, num_rows
       if ( tstart(i) <= tstart(k))then
          tstop(k) = MAX(tstop(i),tstop(k))
       else
          if (tstart(i) <= tstop(k)) then
             tstop(k) = MAX(tstop(i),tstop(k))
          else
             k = k + 1
             tstart(k) = tstart(i)
             tstop(k) = tstop(i)
          end if
       end if
    end do
    num_rows = k

    ! now copy over the merged gtis to the instlist%inst(0)%src_gtilist
    allocate(inst%gtilist%pstart(num_rows),&
        inst%gtilist%pstop(num_rows),&
        STAT = status)
    if (check_allocate_status(subroutinename, status, ierr)) return

    inst%gtilist%num_rows = num_rows
    inst%gtilist%pstart(:) = tstart(1:num_rows)
    inst%gtilist%pstop(:)  = tstop(1:num_rows)

    if ( allocated(tstart)) deallocate(tstart)
    if ( allocated(tstop))  deallocate(tstop)

    inst%gtilist%is_valid = .true.

  end subroutine srctool_gti_merge_gtilists_for_inst


  !-------------------------------------------------------------------------------
  ! merge input gtis across all instruments
  ! place result in instlist%inst(0)%gtilist
  subroutine srctool_gti_merge_gtilists_for_instlist(instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_gti_merge_gtilists_for_instlist'
    type(instlist_type),  intent(inout) :: instlist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i, k, j
    integer(kind=int32) :: status
    real(kind=real64), allocatable, dimension(:) :: tstart,tstop
    real(kind=real64) :: ttotal
    integer(kind=int32)  :: num_rows
    character(len=STR_MAX) :: str_buffer
    ierr = ERROR_LEVEL_OK
    status = 0

    ! FIXME: is this necessary for TMs 8,9?

    call srctool_gti_empty_gtilist(instlist%inst(0)%gtilist)

    ! make sure the keywords are set to null

    call fitswrap_set_key_null(instlist%inst(0)%gtilist%key_telapse)
    call fitswrap_set_key_null(instlist%inst(0)%gtilist%key_ontime)
    call fitswrap_set_key_null(instlist%inst(0)%gtilist%key_livetime)
    call fitswrap_set_key_null(instlist%inst(0)%gtilist%key_exposure)

    ! now merge the src GTIs from the separate instruments
    ! to give a merged GTI which covers intervals when the source region is covered by >=1 eROSITA instrument
    num_rows = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        num_rows = num_rows +  instlist%inst(i)%gtilist%num_rows
      end if
    end do

    if ( num_rows .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Could not find any valid GTIs")
      return
    end if


    allocate(tstart(num_rows),tstop(num_rows), &
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    ! copy the data to the new array
    k = 1
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        do j=1,instlist%inst(i)%gtilist%num_rows
          tstart(k) = instlist%inst(i)%gtilist%pstart(j)
          tstop(k)  = instlist%inst(i)%gtilist%pstop(j)
          k = k + 1
        end do
      end if
    end do

    call cosort_arrays(tstart, tstop)

    ! first go through and make sure that any intervals with matching tstart are filtered out
    ! also check for overlapping/consecutive gtis and merge them
    k = 1
    do i=1,num_rows
      if ( tstart(i) .le. tstart(k))then
        tstop(k) = MAX(tstop(i),tstop(k))
      else
        if (tstart(i) .le. tstop(k)) then
          tstop(k) = MAX(tstop(i),tstop(k))
        else
          k = k + 1
          tstart(k) = tstart(i)
          tstop(k) = tstop(i)
        end if
      end if
    end do
    num_rows = k

    if ( num_rows .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Could not find any valid GTIs")
      return
    end if

    ! now copy over the merged gtis to the instlist%inst(0)%gtilist
    allocate(instlist%inst(0)%gtilist%pstart(num_rows),&
        instlist%inst(0)%gtilist%pstop(num_rows),&
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    instlist%inst(0)%gtilist%num_rows = num_rows
    instlist%inst(0)%gtilist%pstart(:) = tstart(1:num_rows)
    instlist%inst(0)%gtilist%pstop(:)  = tstop(1:num_rows)
    instlist%inst(0)%gtilist%is_valid = .true.

    if ( allocated(tstart)) deallocate(tstart)
    if ( allocated(tstop))  deallocate(tstop)

    do i=0,MAX_TOT_INST
      if ( instlist%inst(i)%to_consider .or. (MERGED_TM_TO_IDX(i)>0) ) then

        ! add up the GTIs
        ttotal = 0.0d0
        do j=1,instlist%inst(i)%gtilist%num_rows
          if ( instlist%inst(i)%gtilist%is_valid ) then
            ttotal = ttotal + instlist%inst(i)%gtilist%pstop(j) - instlist%inst(i)%gtilist%pstart(j)
          endif
        end do

        write (str_buffer, '(A,I8,A,I0,A,A,A,F9.2,A)') &
            "There were ", instlist%inst(i)%gtilist%num_rows, &
            " input GTIs for inst index=",instlist%inst(i)%index, &
            " (name=", trim(instlist%inst(i)%name), ") Total time=", &
            ttotal, "s"
        call wrschatty (subroutinename, str_buffer)
      end if
    end do

  end subroutine srctool_gti_merge_gtilists_for_instlist



  !-------------------------------------------------------------------------------
  ! return the amount of time spanned by t1--t2 that falls within gtilist
  ! assumes that gtis are sorted in ascending order of tstart
  ! assumes that t1-t2 with overlap with at most 1 gti
  recursive subroutine srctool_gti_test_overlap_with_gtilist(gtilist, t1, t2, overlap)
    character(len=*), parameter :: subroutinename='srctool_gti_test_overlap_with_gtilist'
    type(gtilist_type), intent(in)      :: gtilist
    real(kind=real64), intent(in)       :: t1
    real(kind=real64), intent(in)       :: t2
    real(kind=real64), intent(out)      :: overlap
    integer(kind=int32) :: i1,i2,i
    integer(kind=int32) :: status
    overlap = 0.0d0

    if ( gtilist%num_rows .eq. 1 ) then
      overlap = MAX(0.0d0,MIN(t2,gtilist%pstop(1)) - MAX(t1, gtilist%pstart(1)))
    else if ( gtilist%num_rows .gt. 1 ) then
      ! find the first gti where t1 >= pstop(i)
      call binary_search(gtilist%pstop(1:gtilist%num_rows), t1, i1, status)
      if ((status .eq. 0) .or. (status .eq. 2) ) then
        i1 = MAX(1,i1 -1) ! so this must be the last gti where t1 < ptstop(i)
      else
        return  ! there can be no overlap between [t1:t2] and the gtilist
      endif

       ! find the first gti where t2 >= pstart(i)
      call binary_search(gtilist%pstart(1:gtilist%num_rows), t2, i2, status)
      if ((status .ne. 0) .and. (status .ne. 2) ) then
        return  ! there can be no overlap between [t1:t2] and the gtilist
      endif
      do i=i1,i2
        overlap = overlap + MAX(0.0d0,MIN(t2,gtilist%pstop(i)) - MAX(t1, gtilist%pstart(i)))
      end do
    endif

  end subroutine srctool_gti_test_overlap_with_gtilist


  !-------------------------------------------------------------------------------
  ! return the index of the gti which contains time t
  ! return i=-1 if not in a gti
  ! assumes that gtis are sorted in ascending order of tstart and do not overlap
  recursive subroutine srctool_gti_get_gti_index(gtilist, t, i)
    character(len=*), parameter :: subroutinename='srctool_gti_get_gti_index'
    type(gtilist_type), intent(in)      :: gtilist
    real(kind=real64), intent(in)       :: t
    integer(kind=int32), intent(out)     :: i
    integer(kind=int32) :: status
    i = -1
    if ( gtilist%num_rows .eq. 1 ) then
      i = 1
      return
    else  if ( gtilist%num_rows .gt. 1 ) then
      call binary_search(gtilist%pstart(1:gtilist%num_rows), t, i, status)
      if ( (status .eq. 0 ) .or. (status .eq. 2 )) then
        if (t .ge. gtilist%pstop(i))  then
          i = -1
        endif
      else
        i = -1
      endif
    end if
!    if ( i < 0 ) then
!      write (*,'(A,x,F15.5,x,I8)') subroutinename, t, i
!    else
!      write (*,'(A,x,F15.5,x,I8,2(x,F15.5))') subroutinename, t, i, gtilist%pstart(i),  gtilist%pstop(i)
!    endif
  end subroutine srctool_gti_get_gti_index

  !-------------------------------------------------------------------------------
  ! return the index of the gti after the one which contains time t
  ! return i=-1 if t lands after all gtis
  ! assumes that gtis are sorted in ascending order of tstart and do not overlap
  recursive subroutine srctool_gti_get_next_gti_index(gtilist, t, i)
    character(len=*), parameter :: subroutinename='srctool_gti_get_next_gti_index'
    type(gtilist_type), intent(in)      :: gtilist
    real(kind=real64), intent(in)       :: t
    integer(kind=int32), intent(out)     :: i
    integer(kind=int32) :: status
    i = -1
    if ( gtilist%num_rows .gt. 1 ) then
      call binary_search(gtilist%pstop(1:gtilist%num_rows), t, i, status)
      if ( (status .eq. 0 ) .or. (status .eq. 1 )) then
        if (t .gt. gtilist%pstart(i+1))  then
          i = -1
        endif
      else
        i = -1
      endif
    end if

!    if ( i < 0 ) then
!      write (*,'(A,x,F12.5,x,I8)') subroutinename, t, i
!    else
!      write (*,'(A,x,F12.5,x,I8,2(x,F15.5))') subroutinename, t, i, gtilist%pstart(i),  gtilist%pstop(i)
!    endif
  end subroutine srctool_gti_get_next_gti_index

  recursive subroutine srctool_gti_get_gti_bounds(gtilist, t, in_flag, tstart, tstop)
    character(len=*), parameter :: subroutinename='srctool_gti_get_gti_bounds'
    type(gtilist_type), intent(in)      :: gtilist
    real(kind=real64), intent(in)       :: t
    logical           , intent(out)     :: in_flag
    real(kind=real64), intent(out)       :: tstart
    real(kind=real64), intent(out)       :: tstop
    integer(kind=int32) :: i
    tstart = huge(tstart)
    tstop  = -huge(tstop)
    in_flag = .false.

    call srctool_gti_get_gti_index(gtilist, t, i)
    if ( i .gt. 0 ) then
      tstart = gtilist%pstart(i)
      tstop  = gtilist%pstop(i)
      in_flag = .true.
    endif

  end subroutine srctool_gti_get_gti_bounds




  !-------------------------------------------------------------------------------
  ! return .true. if inside at least one gti or .false. if not
  ! assumes that gtis are sorted in ascending order of tstart
  recursive function srctool_gti_test_if_in_gtilist(gtilist, t)
    character(len=*), parameter :: subroutinename='srctool_gti_test_if_in_gtilist'
    type(gtilist_type), intent(in)  :: gtilist
    real(kind=real64), intent(in)   :: t

    integer(kind=int32) :: i
    logical :: srctool_gti_test_if_in_gtilist
    integer(kind=int32) :: status

    status = 0
    srctool_gti_test_if_in_gtilist = .false.

    if ( gtilist%num_rows .ge. 1 ) then
      if ( gtilist%num_rows .eq. 1 ) then
        if ((t .ge. gtilist%pstart(1)) .and. &
            (t .lt. gtilist%pstop(1)) ) then
          srctool_gti_test_if_in_gtilist = .true.
        end if
      else
        call binary_search(gtilist%pstart(1:gtilist%num_rows), t, i, status)
        if ( (status .eq. 0) .or. (status .eq. 2) ) then
!          if ((i .ge. 1) .and. &
!              (i .lt. gtilist%num_rows)) then
            if ((t .ge. gtilist%pstart(i)) .and. &
                (t .lt. gtilist%pstop(i)) ) then
              srctool_gti_test_if_in_gtilist = .true.
            end if
!          end if
        end if
      end if
    end if

  end function srctool_gti_test_if_in_gtilist
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! return .true. if inside at least one gti or .false. if not
  ! makes no assumptions about the ordering of gtis - so could be sped up for a guaranteed ordered set of GTIs
  recursive function srctool_test_if_in_gtilist_slow(gtilist, t)
    character(len=*), parameter :: subroutinename='srctool_gti_test_if_in_gtilist_slow'
    type(gtilist_type), intent(in)      :: gtilist
    real(kind=real64), intent(in)       :: t

    integer(kind=int32) :: i
    logical :: srctool_test_if_in_gtilist_slow
    srctool_test_if_in_gtilist_slow = .false.

    do i=1,gtilist%num_rows
      if ((t .ge. gtilist%pstart(i)) .and. &
          (t .lt. gtilist%pstop(i)) ) then
        srctool_test_if_in_gtilist_slow = .true.
        return
      end if
    end do

  end function srctool_test_if_in_gtilist_slow
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! prepare a src_gtilist based on the supplied intervallist
  subroutine srctool_gti_setup_src_gtilist(gtilist, gtilist_template, intervallist, ierr)
    character(len=*), parameter :: subroutinename='srctool_gti_setup_src_gtilist'
    type(gtilist_type), intent(inout)       :: gtilist
    type(gtilist_type), intent(in)          :: gtilist_template
    type(intervallist_type), intent(in)     :: intervallist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: status
    integer(kind=int32) :: array_length
    status = 0
    ierr=ERROR_LEVEL_OK

    call srctool_gti_empty_gtilist(gtilist)
    !----------------------------------------------------------------------------------------
    ! allocate space for the source gtis - conservative number of intervals is total exposure/MIN_EXPECTED_MEAN_GTI_DURATION
    if (  intervallist%total_time .le. 0.0d0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Intervallist is of duration <=0.0sec")
      return
    end if

    array_length = 1 + nint(intervallist%total_time / MIN_EXPECTED_MEAN_GTI_DURATION)

    allocate(gtilist%pstart(1:array_length), &
        gtilist%pstop(1:array_length), &
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    gtilist%pstart(:) = -huge(gtilist%pstart(1)) ! VERY_NEGATIVE_DOUBLE
    gtilist%pstop(:)  = -huge(gtilist%pstop(1)) ! VERY_NEGATIVE_DOUBLE
    gtilist%num_rows = 1
    gtilist%is_valid  = .true.


    ! now set up the HDU keywords
    gtilist%hduinfo = gtilist_template%hduinfo
!    call fitswrap_set_key(gtilist%hduinfo%key_hduclass, "HDUCLASS", "OGIP" , FITSWRAP_SET_KEY_BLANK)
!    call fitswrap_set_key(gtilist%hduinfo%key_hduclas1, "HDUCLAS1", "GTI"  , FITSWRAP_SET_KEY_BLANK)
    call fitswrap_set_key(gtilist%hduinfo%key_hduclas2, "SRCGTI")
!    call fitswrap_set_key(gtilist%hduinfo%key_telescop, "INSTRUME" ,trim(inst%name), "Instrument name")

  end subroutine srctool_gti_setup_src_gtilist
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  subroutine srctool_gti_append_gtilist_hdu(unit, gtilist, ierr)
    character(len=*), parameter :: subroutinename='srctool_gti_append_gtilist_hdu'
    integer(kind=int32), intent(in)    :: unit
    type(gtilist_type), intent(in)  :: gtilist
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32), parameter :: ncols=2
    character(len=*),    parameter :: ttype(ncols) = (/ "START","STOP " /)
    character(len=*),    parameter :: tunit(ncols) = (/ "s"    ,"s"     /)
    character(len=*),    parameter :: tform(ncols) = (/ "1D"   , "1D"   /)
    integer(kind=int32), parameter :: COLNUM_START  =  1
    integer(kind=int32), parameter :: COLNUM_STOP   =  2
    integer(kind=int32), parameter :: varidat=0
    integer(kind=int32), parameter :: frow=1, felem=1

    integer(kind=int32) :: status

    ierr = ERROR_LEVEL_OK
    status = 0

    ! first append a new BINTABLE HDU to open file at unit
    call FTIBIN_w(unit,gtilist%num_rows,ncols,ttype,tform,tunit, &
        trim(gtilist%hduinfo%key_extname%value),varidat,status)

    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    !-----------------------------------
    ! now write the relevant keywords
!    call fitswrap_write_key (unit, gtilist%hduinfo%key_telescop, ierr)
!    call fitswrap_write_key (unit, gtilist%hduinfo%key_instrume, ierr)
!    call fitswrap_write_key (unit, gtilist%hduinfo%key_filter,   ierr)
!    call fitswrap_write_key (unit, gtilist%hduinfo%key_hduclass, ierr)
!    call fitswrap_write_key (unit, gtilist%hduinfo%key_hduclas1, ierr)
!    call fitswrap_write_key (unit, gtilist%hduinfo%key_hduclas2, ierr)
!!    call fitswrap_write_key (unit, gtilist%hduinfo%key_hduvers,  ierr)



    call fitswrap_write_hduinfo_quiet(unit, gtilist%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing fits keywords")
      return
    end if

    call fitswrap_write_key_if_supplied (unit, gtilist%key_telapse,  ierr)
    call fitswrap_write_key_if_supplied (unit, gtilist%key_ontime,   ierr)
    call fitswrap_write_key_if_supplied (unit, gtilist%key_livetime, ierr)
    call fitswrap_write_key_if_supplied (unit, gtilist%key_exposure, ierr)

!    if ( trim(gtilist%hduinfo%key_instrume%value) .eq. "MERGED" ) then
!      call fitswrap_write_key (unit, gtilist%key_ninst,    ierr)
!      do i=1,gtilist%key_ninst%value
!        call fitswrap_write_key (unit, gtilist%pkey_instrum(i), ierr)
!      end do
!    end if

    !-----------------------------------

    ! TODO - write out data in chunks
     ! write out columns
    call  FTPCLD_w(unit,COLNUM_START,frow,felem,gtilist%num_rows,gtilist%pstart,status)
    ! write COUNTS column
    call  FTPCLD_w(unit,COLNUM_STOP ,frow,felem,gtilist%num_rows,gtilist%pstop,status)


  end subroutine srctool_gti_append_gtilist_hdu
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------------
!! check if we need to update the gtilist
  recursive subroutine srctool_gti_update_src_gtilist (gtilist, in_fov_now, t1, t2, ierr)
    character(len=*), parameter :: subroutinename='srctool_gti_update_src_gtilist'
    type(gtilist_type), intent(inout)   :: gtilist
    logical           , intent(in)      :: in_fov_now
    real(kind=real64), intent(in)       :: t1
    real(kind=real64), intent(in)       :: t2
    integer(kind=int32), intent(out)    :: ierr

    character(len=STR_MAX) :: str_buffer

    ierr=ERROR_LEVEL_OK

    if ( gtilist%num_rows .gt. size(gtilist%pstart) ) then
      ierr = ERROR_LEVEL_SEVERE
      write (str_buffer, '(A,I0,A,I0,A)') &
          "Too many source GTIs; num_rows=", gtilist%num_rows, "(max_rows=", size(gtilist%pstart), ")"
      call wrserr (subroutinename, ierr, str_buffer)
      return
    end if

    !determine the previous state from the value in gtilist%pstart(gti_i)
    !(should be set to a very negative number at init stage)
    if ( gtilist%pstart(gtilist%num_rows) .le. -1d0*huge(gtilist%pstart(1)) ) then !  VERY_NEGATIVE_DOUBLE ) then
      ! we were previously outside a GTI
      if ( in_fov_now ) then ! and src has just entered the FOV
        gtilist%pstart(gtilist%num_rows) = t1
      end if
    else
      ! we were previously in a GTI
      if ( .not. in_fov_now ) then ! and src has just left the FOV
        gtilist%pstop(gtilist%num_rows) = t2
        gtilist%num_rows = gtilist%num_rows + 1

        ! if we go beyond the size of the arrays, we expand
        if( gtilist%num_rows > size(gtilist%pstart) ) then
           call expand_array(gtilist%pstart)
           call expand_array(gtilist%pstop)
           gtilist%pstart(gtilist%num_rows:) = -huge(gtilist%pstart(1)) ! VERY_NEGATIVE_DOUBLE
           gtilist%pstop(gtilist%num_rows:)  = -huge(gtilist%pstop(1)) ! VERY_NEGATIVE_DOUBLE
        end if

      end if
    end if

  end subroutine srctool_gti_update_src_gtilist
  !-------------------------------------------------------------------------------------





  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- END of GTI handling -----------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
end module srctool_gti
