! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_instpar

  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling instpar cal data
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         !
  use srctool_misc              !

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_INSTPAR_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_instpar
  private :: srctool_instpar_report_instpar_info
  public  :: srctool_instpar_empty_instpar
  public  :: srctool_instpar_read_instpar_from_chdu

contains


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------Start of INSTPAR file handling-------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_instpar_empty_instpar(instpar)
    type(instpar_type), intent(inout)      :: instpar
    if ( allocated(instpar%pfocal_length))   deallocate (instpar%pfocal_length)
    if ( allocated(instpar%pmm_xpix     ))   deallocate (instpar%pmm_xpix     )
    if ( allocated(instpar%pmm_ypix     ))   deallocate (instpar%pmm_ypix     )
    if ( allocated(instpar%px_platescale))   deallocate (instpar%px_platescale)
    if ( allocated(instpar%py_platescale))   deallocate (instpar%py_platescale)
    if ( allocated(instpar%px_ccdpix    ))   deallocate (instpar%px_ccdpix    )
    if ( allocated(instpar%py_ccdpix    ))   deallocate (instpar%py_ccdpix    )
    if ( allocated(instpar%px_optax     ))   deallocate (instpar%px_optax     )
    if ( allocated(instpar%py_optax     ))   deallocate (instpar%py_optax     )
    if ( allocated(instpar%pfov_rad     ))   deallocate (instpar%pfov_rad     )

    instpar%naxis1 = 0
    instpar%naxis2 = 0
    instpar%pixscale_x = 0.0d0
    instpar%pixscale_y = 0.0d0
    instpar%inv_pixscale_x = 0.0d0
    instpar%inv_pixscale_y = 0.0d0
    instpar%oa_rawx = 0.0d0
    instpar%oa_rawy = 0.0d0

    instpar%is_valid = .false.

    call fitswrap_init_hduinfo(instpar%hduinfo)

  end subroutine srctool_instpar_empty_instpar
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !read the instpar description from the current hdu of the fits file identified by unit
  subroutine srctool_instpar_read_instpar_from_chdu(unit, instpar, ierr)
    character(len=*), parameter :: subroutinename='srctool_instpar_read_instpar_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(instpar_type), intent(inout)  :: instpar
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status
    integer(kind=int32), parameter :: group=0
!    character(len=STR_MAX) :: str_buffer

    ierr = ERROR_LEVEL_OK
    status = 0
    call wrschatty (subroutinename, "Reading in INSTPAR extension from CHDU")

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, instpar%hduinfo, ierr)
    if ( fitswrap_check_status(subroutinename, status, ierr) ) return

    !- check that vital keader keywords are present - TODO
    ! HDUCLASS= 'OGIP    '
    ! HDUCLAS1= 'DETCHAR '           / Telescope geometry data
    ! HDUCLAS2= 'GEOMETRY'           / Telescope geometry data

    ! read any other keywords

    ! now do the steps to read the column values

    !get the number of rows in the table extension
    call FTGNRW_w(unit, instpar%num_rows, status)
    if ( fitswrap_check_status(subroutinename, status, ierr) ) return

    ! check that have at least some rows, and if so read in the column data
    if ( instpar%num_rows .gt. 0 ) then
      call wrschatty (subroutinename, "Reading in: " // trim(i2s(instpar%num_rows)) // " rows")
      call fitswrap_proc_col(unit, "FOCAL_LENGTH", instpar%num_rows, instpar%pfocal_length, ierr)
      call fitswrap_proc_col(unit, "MM_XPIX",      instpar%num_rows, instpar%pmm_xpix     , ierr)
      call fitswrap_proc_col(unit, "MM_YPIX",      instpar%num_rows, instpar%pmm_ypix     , ierr)
      call fitswrap_proc_col(unit, "X_PLATESCALE", instpar%num_rows, instpar%px_platescale, ierr)
      call fitswrap_proc_col(unit, "Y_PLATESCALE", instpar%num_rows, instpar%py_platescale, ierr)
      call fitswrap_proc_col(unit, "X_CCDPIX",     instpar%num_rows, instpar%px_ccdpix    , ierr)
      call fitswrap_proc_col(unit, "Y_CCDPIX",     instpar%num_rows, instpar%py_ccdpix    , ierr)
      call fitswrap_proc_col(unit, "X_OPTAX",      instpar%num_rows, instpar%px_optax     , ierr)
      call fitswrap_proc_col(unit, "Y_OPTAX",      instpar%num_rows, instpar%py_optax     , ierr)
      call fitswrap_proc_col(unit, "FOV_RAD",      instpar%num_rows, instpar%pfov_rad     , ierr)

      if (ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem reading column(s) from HDU")
        return
      end if

    else
      call wrslog (subroutinename, "Warning: no entries found in INSTPAR extension")
    end if

    ! make some copies to make things shorter later
    instpar%naxis1  = instpar%px_ccdpix(1)
    instpar%naxis2  = instpar%py_ccdpix(1)

    instpar%oa_rawx  = instpar%px_optax(1)
    instpar%oa_rawy  = instpar%py_optax(1)

    instpar%fov_radius_arcmin = instpar%pfov_rad(1)

    instpar%pixscale_x = instpar%px_platescale(1) / 3600.0d0
    instpar%pixscale_y = instpar%py_platescale(1) / 3600.0d0

    if ( abs(instpar%pixscale_x) .gt. 0.0d0 ) then
      instpar%inv_pixscale_x = 1.0d0 / instpar%pixscale_x
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Problem interpreting value of instpar entry: X_PLATESCALE")
      return
    end if

    if ( abs(instpar%pixscale_y) .gt. 0.0d0 ) then
      instpar%inv_pixscale_y = 1.0d0 / instpar%pixscale_y
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Problem interpreting value of instpar entry: Y_PLATESCALE")
      return
    end if

    !----------------------------------------------
    ! now report the instpar info
    call srctool_instpar_report_instpar_info(instpar)
    !----------------------------------------------

    instpar%is_valid = .true.

  end subroutine srctool_instpar_read_instpar_from_chdu
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine srctool_instpar_report_instpar_info(instpar)
    character(len=*), parameter :: subroutinename='srctool_instpar_report_instpar_info'
    type(instpar_type), intent(in) :: instpar

    character(len=STR_MAX) :: str_buffer

    write (str_buffer, '(A12,1x,A6,6(1x,A8))') &
        "#InstparInfo", "Inst", "NAXIS1", "NAXIS2", "PSCALEX", "PSCALEY", "X_OPTAX", "Y_OPTAX"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A12,1x,A6,2(1x,I8),2(1x,F8.6),2(1x,F8.2))') &
        "InstparInfo", trim(instpar%hduinfo%key_instrume%value), &
        instpar%naxis1, instpar%naxis2, instpar%pixscale_x, instpar%pixscale_y, &
        instpar%oa_rawx, instpar%oa_rawy
    call wrschatty(subroutinename, str_buffer)

  end subroutine srctool_instpar_report_instpar_info
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------END of INSTPAR file handling  -------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------



end module srctool_instpar
