! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_lc

  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling LCs
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         ! module with some srctool-specific constants and datatypes
  use srctool_misc              ! module with some srctool-specific constants and datatypes
  use srctool_gti               ! module with some srctool-specific constants and datatypes
  use srctool_owner             ! handle event ownership
  use srctool_generic_fitsraw

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_LC_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_lc
  public  :: srctool_lc_empty_lc
  public  :: srctool_lc_empty_lclist
  private :: srctool_lc_write_lc_keys
  private :: srctool_lc_calc_lc_rates
  public  :: srctool_lc_calc_lclist_rates
  private :: srctool_lc_write_lc
  public  :: srctool_lc_write_lclist
  private :: srctool_lc_prepare_lc_template
  private :: srctool_lc_setup_lc
  public  :: srctool_lc_setup_lclist
  public  :: srctool_lc_accumulate_lc
  public  :: srctool_lc_accumulate_lclist
  public  :: srctool_lc_coadd_lclist
  public  :: srctool_lc_get_lc_index
  public  :: srctool_lc_get_next_lc_index

  private :: srctool_lc_like_src_bg
  private :: srctool_lc_maxlike_given_bgsub
  private :: srctool_lc_uncert_binary_search
  private :: srctool_lc_bgsub_calc_errors

contains


  !-----------------------------------------------------------------------------------
  !-----------------------------------------------------------------------------------
  !--------------------------- START of Lightcurve handling --------------------------
  !-----------------------------------------------------------------------------------
  !-----------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_lc_empty_lc(lc)
    type(lc_type), intent(inout)   :: lc
    integer(kind=int32) :: e!, i
!    character(len=FITSIO_KEY_MAXLEN)             :: str_buffer1
!    character(len=FITSIO_KEY_MAXLEN)             :: str_buffer2
    call fitswrap_init_hduinfo(lc%hduinfo)

    lc%lctype      =  "NULL"
    lc%lctype_code =  LCTYPE_CODE_NULL
    lc%lcpars(:)   =  0.0d0
    lc%num_lcpars  =  0

    call fitswrap_init_key(lc%key_timedel , "TIMEDEL" , "" )
    call fitswrap_init_key(lc%key_timezero, "TIMEZERO", "[s]")

    call fitswrap_init_key(lc%key_telapse,  "TELAPSE" , "[s] Total observation duration"                 )
    call fitswrap_init_key(lc%key_ontime,   "ONTIME"  , "[s] On-target exposure time (pre deadtime corr)")
    call fitswrap_init_key(lc%key_livetime, "LIVETIME", "[s] On-target active-detector exposure time"    )
    call fitswrap_init_key(lc%key_exposure, "EXPOSURE", "[s] Fully corrected effective exposure time")

    call fitswrap_init_key(lc%key_tierrela, "TIERRELA", "")
    call fitswrap_init_key(lc%key_tierabso, "TIERABSO", "[s]")
    call fitswrap_init_key(lc%key_clockcor, "CLOCKCOR", "True if clock correction was applied"      )
    call fitswrap_init_key(lc%key_backapp , "BACKAPP" , "True if background was subtracted"         )
    call fitswrap_init_key(lc%key_deadapp , "DEADAPP" , "True if deadtime correction was applied"   )
    call fitswrap_init_key(lc%key_vignapp , "VIGNAPP" , "True if vignetting correction was applied" )

    call fitswrap_init_key(lc%key_flagsel , "FLAGSEL" , "Flag selection mask which has been applied" )
    call fitswrap_init_key(lc%key_pat_sel , "PAT_SEL" , "Pattern mask which has been applied")

    call fitswrap_init_key(lc%key_src_region,  "SRC_REG",  "Source extraction (ds9) region used")
    call fitswrap_init_key(lc%key_back_region, "BACK_REG", "Background extraction (ds9) region used")

    do e=1,MAX_LCBANDS
!      write(str_buffer1, '(A,I0)') "EMIN", e
!      write(str_buffer2, '(A,I0)') "[keV] Min. energy bound,band ", e
      call fitswrap_init_key(lc%key_emin(e), &
          "EMIN" // trim(i2s(e)), "[keV] Min. energy bound,band " // trim(i2s(e)))

!      write(str_buffer1, '(A,I0)') "EMAX", e
!      write(str_buffer2, '(A,I0)') "[keV] Max. energy bound,band ", e
!      call fitswrap_init_key(lc%key_emax(e), trim(str_buffer1), trim(str_buffer2))
      call fitswrap_init_key(lc%key_emax(e), &
          "EMAX" // trim(i2s(e)), "[keV] Max. energy bound,band " // trim(i2s(e)))

!      write(str_buffer1, '(A,I0)') "CTS_SC", e
!      write(str_buffer2, '(A,I0)') "Cts in source region,band ", e
!      call fitswrap_init_key(lc%key_ncounts_src(e), trim(str_buffer1), trim(str_buffer2))
      call fitswrap_init_key(lc%key_ncounts_src(e), &
          "CTS_SC" // trim(i2s(e)), "Cts in source region,band " // trim(i2s(e)))

!      write(str_buffer1, '(A,I0)') "CTS_BK", e
!      write(str_buffer2, '(A,I0)') "Cts in bkgnd region,band ", e
!      call fitswrap_init_key(lc%key_ncounts_back(e), trim(str_buffer1), trim(str_buffer2))
      call fitswrap_init_key(lc%key_ncounts_back(e), &
          "CTS_BK" // trim(i2s(e)), "Cts in bkgnd region,band " // trim(i2s(e)))
    end do

!    call fitswrap_init_key(lc%key_ninst, "NINST", "Number of instruments that were merged")
!    do i=1,MAX_INST
!!      write(str_buffer1, '(A,I0)') "INSTRUM", i
!     call fitswrap_init_key(lc%pkey_instrum(i), "INSTRUM" // trim(i2s(i)), "Contributing instrument name")
!    end do

    if ( allocated(lc%ptime         ))  deallocate (lc%ptime         )
    if ( allocated(lc%ptimedel      ))  deallocate (lc%ptimedel      )
    if ( allocated(lc%pcounts       ))  deallocate (lc%pcounts       )
    if ( allocated(lc%prate         ))  deallocate (lc%prate         )
    if ( allocated(lc%prate_err     ))  deallocate (lc%prate_err     )
    if ( allocated(lc%prate_estposerr))  deallocate (lc%prate_estposerr)
    if ( allocated(lc%prate_estnegerr))  deallocate (lc%prate_estnegerr)
    if ( allocated(lc%pback_counts  ))  deallocate (lc%pback_counts  )
    if ( allocated(lc%pfracarea     ))  deallocate (lc%pfracarea     )
    if ( allocated(lc%pfracflatarea ))  deallocate (lc%pfracflatarea )
    if ( allocated(lc%pfractime     ))  deallocate (lc%pfractime     )
    if ( allocated(lc%pfracexp      ))  deallocate (lc%pfracexp      )
    if ( allocated(lc%pfracflatexp  ))  deallocate (lc%pfracflatexp  )
    if ( allocated(lc%poffaxis      ))  deallocate (lc%poffaxis      )
    if ( allocated(lc%pbackscal_src ))  deallocate (lc%pbackscal_src )
    if ( allocated(lc%pbackscal_back))  deallocate (lc%pbackscal_back)
    if ( allocated(lc%pbackratio    ))  deallocate (lc%pbackratio    )
    if ( allocated(lc%ptmin         ))  deallocate (lc%ptmin         )
    if ( allocated(lc%ptmax         ))  deallocate (lc%ptmax         )
    lc%num_rows   = 0
    lc%num_ebands = 0
    lc%to_consider = .false.

    lc%lcgamma = 2.0d0
    lc%int_nenergies = 1
    if ( allocated(lc%pint_e))       deallocate (lc%pint_e)
    if ( allocated(lc%pint_weights))  deallocate (lc%pint_weights)


  end subroutine srctool_lc_empty_lc
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_lc_empty_lclist(lclist)
    type(lclist_type), intent(inout)   :: lclist
    integer(kind=int32) :: i
    do i=0,MAX_TOT_INST
      call srctool_lc_empty_lc(lclist%lc(i))
    end do
    lclist%num_lcs = 0
  end subroutine srctool_lc_empty_lclist
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! write lc fits keys info to the chdu
  ! TODO - rest of keywords - check existing ones too
  subroutine srctool_lc_write_lc_keys(unit, lc, ierr)
    character(len=*), parameter :: subroutinename='srctool_lc_write_lc_keys'
    integer(kind=int32), intent(in)  :: unit
    type(lc_type),       intent(in)  :: lc
    integer(kind=int32), intent(out) :: ierr
    integer(kind=int32)    :: status
    integer(kind=int32)    :: e!, i
    ierr = ERROR_LEVEL_OK
    status = 0


    !-----------------------------------
    ! write the standard keywords to the binary table HDU:
    call fitswrap_write_hduinfo_quiet(unit, lc%hduinfo, ierr)
    if (ierr /= ERROR_LEVEL_OK) then
      return
    end if


    call fitswrap_write_key (unit, lc%key_timedel,  ierr)

    call fitswrap_write_key (unit, lc%key_telapse,  ierr)
    call fitswrap_write_key (unit, lc%key_ontime,   ierr)
    call fitswrap_write_key (unit, lc%key_livetime, ierr)
    call fitswrap_write_key (unit, lc%key_exposure, ierr)

    call fitswrap_write_key (unit, lc%key_tierrela, ierr)
    call fitswrap_write_key (unit, lc%key_tierabso, ierr)
    call fitswrap_write_key (unit, lc%key_clockcor, ierr)
    call fitswrap_write_key (unit, lc%key_backapp,  ierr)
    call fitswrap_write_key (unit, lc%key_deadapp,  ierr)
    call fitswrap_write_key (unit, lc%key_vignapp,  ierr)

    call fitswrap_write_key (unit, lc%key_flagsel, ierr)
    call fitswrap_write_key (unit, lc%key_pat_sel, ierr)

    ! write the energy band keywords
    do e=1,lc%num_ebands
      call fitswrap_write_key (unit, lc%key_emin(e), ierr)
      call fitswrap_write_key (unit, lc%key_emax(e), ierr)
      call fitswrap_write_key (unit, lc%key_ncounts_src(e), ierr)
      call fitswrap_write_key (unit, lc%key_ncounts_back(e), ierr)
    end do

!    if ( trim(lc%hduinfo%key_instrume%value) == "MERGED" ) then
!      call fitswrap_write_key (unit, lc%key_ninst,    ierr)
!      do i=1,lc%key_ninst%value
!        call fitswrap_write_key (unit, lc%pkey_instrum(i), ierr)
!      end do
!   end if

    call fitswrap_write_key (unit, lc%key_src_region,    ierr)
    call fitswrap_write_key (unit, lc%key_back_region,   ierr)


    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    ! TODO - rest of keywords
    !-----------------------------------

  end subroutine srctool_lc_write_lc_keys



  !-------------------------------------------------------------------------------
  ! Calculate the Poisson likelihood of src_cts and bg_cts, given src_mod and bg_mod
  ! counts
  !
  ! If backratio<=0 then just use source likelihood
  !-------------------------------------------------------------------------------
  function srctool_lc_like_src_bg(src_cts, bg_cts, src_mod, bg_mod, backratio)
    real(kind=real64), intent(in) :: src_cts, bg_cts, src_mod, bg_mod, backratio
    real(kind=real64) :: srctool_lc_like_src_bg

    if( backratio > 0d0 ) then
      ! combined poisson likelihood of src and bg
      srctool_lc_like_src_bg = src_cts*log(src_mod)-src_mod + &
          bg_cts*log(bg_mod)-bg_mod
    else
      ! no background, so use src likelihood
      srctool_lc_like_src_bg = src_cts*log(src_mod)-src_mod
    end if

  end function srctool_lc_like_src_bg

  !-------------------------------------------------------------------------------
  ! Calculate the maximum likelihood given a background subtracted number of counts
  ! given the observed number of background and source counts
  !
  ! In detail
  ! bgsubrate = the background subtracted number of counts we want to examine
  ! backratio = the ratio to use when subtracting background from foreground
  ! src_cts = the real number of counts in the source region
  ! bg_cts = the real number of counts in the background region
  ! like = output likelihood
  !-------------------------------------------------------------------------------
  ! Derivation using Mathematica:
  !  L[F_] := f*Log[F] - F + b*Log[(F - u)/R] - (F - u)/R
  !  FortranForm[ Solve[D[L[F], F] == 0, F] ]
  ! (where F=src_mod, u=bgsubcts, f=src_cts, b=bg_cts)
  ! See paper Barlow R., PHYSTAT2003, SLAC, Stanford, California, September 8-11, 2003
  ! pg 250
  ! https://www.slac.stanford.edu/econf/C030908/papers/SLAC-R-703.pdf
  !
  ! Idea is to maximize likelihood for a given value of the bgsubcts, in order to
  ! look for the delta log like -0.5 range about the most likely value
  !
  ! If there is no background, we just use the Poisson likelihood
  !-------------------------------------------------------------------------------
  function srctool_lc_maxlike_given_bgsub(bgsubcts, backratio, src_cts, bg_cts)
    real(kind=real64), intent(in) :: bgsubcts, backratio, src_cts, bg_cts
    real(kind=real64) :: srctool_lc_maxlike_given_bgsub

    real(kind=real64) :: src_mod, bg_mod

    if( backratio > 0d0 ) then
      src_mod = ( (src_cts+bg_cts)*backratio + bgsubcts*(1+backratio) + &
          sqrt( -4*src_cts*(1+backratio)*backratio*bgsubcts + &
          ((src_cts+bg_cts)*backratio + bgsubcts*(1+backratio))**2)) / &
          (2 * (1+backratio))
      bg_mod = (src_mod - bgsubcts)/backratio
    else
      ! no background, so use normal poisson likelihood below
      src_mod = bgsubcts
      bg_mod = 0
    end if

    srctool_lc_maxlike_given_bgsub = srctool_lc_like_src_bg(src_cts, bg_cts, src_mod, bg_mod, backratio)
  end function srctool_lc_maxlike_given_bgsub
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! Given a starting range between rates of bgsubcts_a and bgsubcts_b,
  ! find the rate which gives a decrease in log likelihood of -0.5 from like0
  ! Assumes a monotonic shape of the functional form
  ! ierr=1 returned in case of error
  !-------------------------------------------------------------------------------
  function srctool_lc_uncert_binary_search(bgsubcts_a, bgsubcts_b, backratio, src_cts, bg_cts, like0, ierr)
    real(kind=real64), value :: bgsubcts_a, bgsubcts_b
    real(kind=real64), intent(in) :: backratio, src_cts, bg_cts, like0
    integer(kind=int32), intent(inout) :: ierr
    real(kind=real64) :: srctool_lc_uncert_binary_search

    integer, parameter :: maxiter = 200
    real(kind=real64), parameter :: tol = 1d-6

    real(kind=real64) :: bgsubcts_mid, like_target, like_a, like_b, like_mid
    integer :: numiter

    like_a = srctool_lc_maxlike_given_bgsub(bgsubcts_a, backratio, src_cts, bg_cts)
    like_b = srctool_lc_maxlike_given_bgsub(bgsubcts_b, backratio, src_cts, bg_cts)
    like_target = like0 - 0.5d0

    ! If bounds don't span target then something has gone wrong
    if( (like_a<like_target .and. like_b<like_target) .or. &
        (like_a>like_target .and. like_b>like_target) ) then
      ierr = 1
      srctool_lc_uncert_binary_search = 0
      return
    end if

    ! usual binary search: we calculate value at a midpoint and update one of the
    ! ranges
    do numiter=1,maxiter

      bgsubcts_mid = 0.5d0 * (bgsubcts_a + bgsubcts_b)
      like_mid = srctool_lc_maxlike_given_bgsub(bgsubcts_mid, backratio, src_cts, bg_cts)

      ! reached numerical target
      if( abs(like_mid-like_target) < tol ) then
        exit
      end if

      if( (like_mid>=like_target .and. like_a>like_target .and. like_b< like_target) .or. &
          (like_mid< like_target .and. like_a<like_target .and. like_b>=like_target )) then
        ! mid point better than bound a
        like_a = like_mid
        bgsubcts_a = bgsubcts_mid
      else if ( (like_mid>=like_target .and. like_b>like_target .and. like_a< like_target) .or. &
                (like_mid< like_target .and. like_b<like_target .and. like_a>=like_target ) ) then
        ! mid point better than bound b
        like_b = like_mid
        bgsubcts_b = bgsubcts_mid
      else
        ! Error!
        bgsubcts_mid = 0
        ierr = 2
        exit
      end if
    end do

    srctool_lc_uncert_binary_search = bgsubcts_mid
  end function srctool_lc_uncert_binary_search
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! Given the backratio, number of counts, calculate the error bars for the rate
  ! Note, if number of counts is 0, then we make it 1
  !-------------------------------------------------------------------------------
  subroutine srctool_lc_bgsub_calc_errors(backratio, src_cts, bg_cts, poserr, negerr, ierr)
    real(kind=real64), intent(in) :: backratio   ! input back ratio
    real(kind=real64), value :: src_cts, bg_cts  ! input number of counts
    real(kind=real64), intent(out) :: poserr, negerr  ! error bars (with sign)
    integer(kind=int32), intent(inout) :: ierr  ! (set to >0 on error)

    real(kind=real64) :: like_0, bgsubcts_0, bound
    real(kind=real64) :: bgsubcts_lo, bgsubcts_hi

    ! If we have zero counts, then we assume 1 - is this ok?
    src_cts = max(1d0, src_cts)
    bg_cts = max(1d0, bg_cts)

    ! this is the maximum likelihood
    like_0 = srctool_lc_like_src_bg(src_cts, bg_cts, src_cts, bg_cts, backratio)
    ! rate at the maximim likelihood
    bgsubcts_0 = src_cts - backratio*bg_cts

    ! if no background subtraction, then rate must by positive
    if(backratio > 0d0) then
      bound = bgsubcts_0-src_cts*1000
    else
      bound = 1d-10
    end if

    bgsubcts_lo = srctool_lc_uncert_binary_search(bgsubcts_0, bound, &
        backratio, src_cts, bg_cts, like_0, ierr)

    ! hit bound and we can't have a negative rate: force bound to be zero
    if(backratio <= 0d0 .and. ierr == 1) then
      bgsubcts_lo = 0
      ierr = 0
    end if

    bgsubcts_hi = srctool_lc_uncert_binary_search(bgsubcts_0, bgsubcts_0+src_cts*1000, &
        backratio, src_cts, bg_cts, like_0, ierr)

    poserr = bgsubcts_hi - bgsubcts_0
    negerr = bgsubcts_lo - bgsubcts_0
  end subroutine srctool_lc_bgsub_calc_errors
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_lc_calc_lc_rates(lc, ierr)
    character(len=*), parameter :: subroutinename='srctool_lc_calc_lc_rates'
    type(lc_type),  intent(inout) :: lc
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)    :: i, e, j
    real(kind=real64)      :: temp_fracrsp    ! working variable
    real(kind=real64)      :: err_s, err_b    ! working variables
    real(kind=real64)      :: poserr, negerr
    logical             :: last_bin_empty
    logical             :: this_bin_empty
    integer :: error_error
!    character(len=STR_MAX) :: str_buffer
    ierr=ERROR_LEVEL_OK

    do i=1,lc%num_rows
      do e=1,lc%num_ebands
        ! net rate = (cts_src - cts_back*backratio)/fracarea*fractime*timedel
        temp_fracrsp = real(lc%ptimedel(i) * lc%pfracexp(e,i), real64)
        if ( temp_fracrsp > 0.0e0 ) then
          lc%prate(e,i) = real((real(lc%pcounts(e,i)) - &
              real(lc%pback_counts(e,i))*lc%pbackratio(i))/ &
              temp_fracrsp, real32)

          ! Combine foreground and background errors providing we have enough counts
          ! (Gehrels doesn't work with error combination formula)
          if( lc%pcounts(e,i) + lc%pback_counts(e,i) >= 25 ) then
            err_s = sqrt(real(lc%pcounts(e,i)))
            err_b = sqrt(real(lc%pback_counts(e,i)))*lc%pbackratio(i)
            lc%prate_err(e,i) = real(sqrt(err_s**2 + err_b**2) / &
                temp_fracrsp, real32)
          else
            lc%prate_err(e,i) = NULLVAL_real32
          end if

          ! Estimate asymmetric rate errors
          error_error = 0
          call srctool_lc_bgsub_calc_errors(lc%pbackratio(i), &
              real(lc%pcounts(e,i), real64), real(lc%pback_counts(e,i), real64), &
              poserr, negerr, error_error)
          if( error_error == 0 .and. lc%pcounts(e,i)>1 .and. &
            ( lc%pback_counts(e,i)>1 .or. lc%pbackratio(i)<=0d0 ) ) then
            lc%prate_estposerr(e,i) = real(poserr / temp_fracrsp, real32)
            lc%prate_estnegerr(e,i) = real(negerr / temp_fracrsp, real32)
          else
            lc%prate_estposerr(e,i) = NULLVAL_REAL32
            lc%prate_estnegerr(e,i) = NULLVAL_REAL32
          end if

        else
          lc%prate(e,i)     = NULLVAL_REAL32
          lc%prate_err(e,i) = NULLVAL_REAL32
          lc%prate_estposerr(e,i) = NULLVAL_REAL32
          lc%prate_estnegerr(e,i) = NULLVAL_REAL32
          lc%pbackratio(i) = NULLVAL_REAL64
        end if
      end do
    end do

    ! If the lc typecode is LCTYPE_CODE_*MINUS then filter the lc to remove all the zero bins
    if ((lc%lctype_code == LCTYPE_CODE_REGULAR_MINUS) .or. &
        (lc%lctype_code == LCTYPE_CODE_SRCGTI_MINUS) ) then
      j = 0
      do i=1,lc%num_rows
        this_bin_empty = ( &
            (MAXVAL(lc%pcounts(:,i))       <= 0    ) .and. &
            (MAXVAL(lc%pback_counts(:,i))  <= 0    ) .and. &
            (MAXVAL(lc%pfracexp(:,i))      <= 0.0e0) )

        if ( this_bin_empty ) then
          ! do nothing
        else
          j = j + 1
          lc%ptime(j)              = lc%ptime(i)
          lc%ptimedel(j)           = lc%ptimedel(i)
          lc%pcounts(:,j)          = lc%pcounts(:,i)
          lc%prate(:,j)            = lc%prate(:,i)
          lc%prate_err(:,j)        = lc%prate_err(:,i)
          lc%prate_estposerr(:,j)  = lc%prate_estposerr(:,i)
          lc%prate_estnegerr(:,j)  = lc%prate_estnegerr(:,i)
          lc%pback_counts(:,j)     = lc%pback_counts(:,i)
          lc%pfracarea(:,j)        = lc%pfracarea(:,i)
          lc%pfracflatarea(j)      = lc%pfracflatarea(i)
          lc%pfractime(j)          = lc%pfractime(i)
          lc%pfracexp(:,j)         = lc%pfracexp(:,i)
          lc%pfracflatexp(j)       = lc%pfracflatexp(i)
          lc%poffaxis(j)           = lc%poffaxis(i)
          lc%pbackscal_src(j)      = lc%pbackscal_src(i)
          lc%pbackscal_back(j)     = lc%pbackscal_back(i)
          lc%pbackratio(j)         = lc%pbackratio(i)
        end if
      end do

      ! call wrschatty (subroutinename, &
      !     " DEBUG: Removed empty lc bins: new num_rows=" // &
      !     trim(i2s(j)) // " (was " // trim(i2s(lc%num_rows)) // ")" )
      lc%num_rows = j
    end if


    ! If the lc typecode is LCTYPE_CODE_REGULAR_PLUS then filter the lc to compact all the consecutive zero bins
    if ( lc%lctype_code == LCTYPE_CODE_REGULAR_PLUS ) then
      last_bin_empty = .false.
      j = 0
      do i=1,lc%num_rows

        this_bin_empty = ( &
            (MAXVAL(lc%pcounts(:,i))       <= 0    ) .and. &
            (MAXVAL(lc%pback_counts(:,i))  <= 0    ) .and. &
            (MAXVAL(lc%pfracexp(:,i))      <= 0.0e0) )

        if ( this_bin_empty .and.  last_bin_empty ) then
          lc%ptime(j)    = lc%ptime(j)    + 0.5d0*lc%ptimedel(i)
          lc%ptimedel(j) = lc%ptimedel(j) + lc%ptimedel(i)
          lc%ptmax(j)    = lc%ptmax(i)
        else
          j = j + 1
          lc%ptime(j)              = lc%ptime(i)
          lc%ptimedel(j)           = lc%ptimedel(i)
          lc%pcounts(:,j)          = lc%pcounts(:,i)
          lc%prate(:,j)            = lc%prate(:,i)
          lc%prate_err(:,j)        = lc%prate_err(:,i)
          lc%prate_estposerr(:,j)  = lc%prate_estposerr(:,i)
          lc%prate_estnegerr(:,j)  = lc%prate_estnegerr(:,i)
          lc%pback_counts(:,j)     = lc%pback_counts(:,i)
          lc%pfracarea(:,j)        = lc%pfracarea(:,i)
          lc%pfracflatarea(j)      = lc%pfracflatarea(i)
          lc%pfractime(j)          = lc%pfractime(i)
          lc%pfracexp(:,j)         = lc%pfracexp(:,i)
          lc%pfracflatexp(j)       = lc%pfracflatexp(i)
          lc%poffaxis(j)           = lc%poffaxis(i)
          lc%pbackscal_src(j)      = lc%pbackscal_src(i)
          lc%pbackscal_back(j)     = lc%pbackscal_back(i)
          lc%pbackratio(j)         = lc%pbackratio(i)

        end if
        last_bin_empty = this_bin_empty
      end do

      ! call wrschatty (subroutinename, &
      !     " DEBUG: Compacted number of lc bins: new num_rows=" // &
      !     trim(i2s(j)) // " (was " // trim(i2s(lc%num_rows)) // ")" )

      lc%num_rows = j
    end if

  end subroutine srctool_lc_calc_lc_rates
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  subroutine srctool_lc_calc_lclist_rates(lclist, ierr)
    character(len=*), parameter :: subroutinename='srctool_lc_calc_lclist_rates'
    type(lclist_type),  intent(inout) :: lclist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i
    ierr = ERROR_LEVEL_OK

    do i=0,MAX_TOT_INST
      if ( (MERGED_TM_TO_IDX(i)>0) .or. lclist%lc(i)%to_consider ) then
        call srctool_lc_calc_lc_rates(lclist%lc(i), ierr)
      end if
      if ( ierr /= ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem calculating rate column for lightcurves")
        return
      end if
    end do

  end subroutine srctool_lc_calc_lclist_rates
  !-------------------------------------------------------------------------------






  !-------------------------------------------------------------------------------
  ! write a lc file to disk
  ! we expect that all relevant parameters are already set up
  subroutine srctool_lc_write_lc(src, lc, gtilist, ierr)
    character(len=*), parameter :: subroutinename='srctool_lc_write_lc'

    type(src_type), intent(inout) :: src
    type(lc_type),      intent(inout) :: lc
    type(gtilist_type),  intent(in) :: gtilist
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32), parameter :: blank_bitpix = 16
    integer(kind=int32), parameter :: blank_naxis = 0
    integer(kind=int32), parameter :: blank_naxes(1) = (/ 0 /)
!    logical(kind=log32), parameter :: simple = .true.
!    logical,             parameter :: simple = .true.

    integer(kind=int32), parameter :: COLNUM_TIME          =  1
    integer(kind=int32), parameter :: COLNUM_TIMEDEL       =  2
    integer(kind=int32), parameter :: COLNUM_COUNTS        =  3
    integer(kind=int32), parameter :: COLNUM_RATE          =  4
    integer(kind=int32), parameter :: COLNUM_RATE_ERR      =  5
    integer(kind=int32), parameter :: COLNUM_RATE_ESTPOSERR=  6
    integer(kind=int32), parameter :: COLNUM_RATE_ESTNEGERR=  7
    integer(kind=int32), parameter :: COLNUM_BACK_COUNTS   =  8
    integer(kind=int32), parameter :: COLNUM_FRACAREA      =  9
    integer(kind=int32), parameter :: COLNUM_FRACFLATAREA  = 10
    integer(kind=int32), parameter :: COLNUM_FRACTIME      = 11
    integer(kind=int32), parameter :: COLNUM_FRACEXP       = 12
    integer(kind=int32), parameter :: COLNUM_FRACFLATEXP   = 13
    integer(kind=int32), parameter :: COLNUM_OFFAXIS       = 14
    integer(kind=int32), parameter :: COLNUM_BACKSCAL_SRC  = 15
    integer(kind=int32), parameter :: COLNUM_BACKSCAL_BACK = 16
    integer(kind=int32), parameter :: COLNUM_BACKRATIO     = 17
    integer(kind=int32), parameter :: ncols=17

    character(len=*),    parameter :: ttype(ncols) = &
        (/"TIME          ","TIMEDEL       ","COUNTS        ","RATE          ",&
          "RATE_ERR      ","RATE_ESTPOSERR","RATE_ESTNEGERR",&
          "BACK_COUNTS   ",&
          "FRACAREA      ","FRACFLATAREA  ",&
          "FRACTIME      ",&
          "FRACEXP       ","FRACFLATEXP   ",&
          "OFFAXIS       ",&
          "BACKSCAL_SRC  ","BACKSCAL_BACK ",&
          "BACKRATIO     "/)
    character(len=*),    parameter :: tunit(ncols) = &
        (/"s     "       ,"s     "         ,"count "        ,"ct/s  "        ,&
          "ct/s  "       ,"ct/s  "         ,"ct/s  ",&
          "count "       ,&
          "ratio "       ,"ratio ",&
          "ratio "       ,&
          "ratio "       ,"ratio ",&
          "deg   "       ,&
          "deg**2"       ,"deg**2",&
          "ratio "       /)
    character(len=4)  :: tform(ncols) ! to be filled in depending on number of ebands

!    character(len=*),    parameter :: extname = "RATE"
    integer(kind=int32), parameter :: varidat=0
    integer(kind=int32), parameter :: felem=1

!    character(len=STR_MAX) :: str_buffer
    integer(kind=int32)    :: unit
    integer(kind=int32)    :: status
    integer(kind=int32)    :: irow, nrow, nblock

    ierr = ERROR_LEVEL_OK
    status = 0
    unit = 0
!    str_buffer = ""

    if ( len_trim(lc%hduinfo%filename) <= 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Illegal filename for lc file")
      return
    end if

!    write (str_buffer,'(A,A)')  "Writing lc file, filename= ", trim(lc%hduinfo%filename)
    call wrschatty (subroutinename,  "Writing lc file, filename= " // trim(lc%hduinfo%filename))

    ! set up the tform params
    write  (tform(COLNUM_TIME)         ,'(A)')    "D"
    write  (tform(COLNUM_TIMEDEL)      ,'(A)')    "D"
    write  (tform(COLNUM_COUNTS)       ,'(I0,A)') lc%num_ebands, "J"
    write  (tform(COLNUM_RATE)         ,'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_RATE_ERR)     ,'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_RATE_ESTPOSERR),'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_RATE_ESTNEGERR),'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_BACK_COUNTS)  ,'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_FRACAREA)     ,'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_FRACFLATAREA) ,'(A)')    "E"
    write  (tform(COLNUM_FRACTIME)     ,'(A)')    "E"
    write  (tform(COLNUM_FRACEXP)      ,'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_FRACFLATEXP)  ,'(A)')    "E"
    write  (tform(COLNUM_OFFAXIS)      ,'(A)')    "E"
    write  (tform(COLNUM_BACKSCAL_SRC) ,'(A)')    "E"
    write  (tform(COLNUM_BACKSCAL_BACK),'(A)')    "E"
    write  (tform(COLNUM_BACKRATIO)    ,'(A)')    "D"

    !-----------------------------------
    ! get a fortran unit
    call FTGIOU_w( unit, status)
    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! try to open the fits file for writing
    call FTINIT_w(unit,lc%hduinfo%filename,FITSIO_BLOCKSIZE, status)
    if (status /= 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
    !-----------------------------------

    call srctool_misc_log_product(src, lc%hduinfo%filename)

    ! append a new empty primary image hdu and write date keyword to the HDU
    call FTIIMG_w(unit,blank_bitpix,blank_naxis,blank_naxes, status)
    call FTPDAT_w(unit, status)
    if (status /= 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
!
!    !-----------------------------------
!    ! write the standard FITS header:
!    call FTPHPS(unit,bitpix,naxis,naxes,status)
!    if (status /= 0) then
!      ierr = ERROR_LEVEL_SEVERE
!      call fitswrap_write_fits_error(subroutinename, status)
!      return
!    end if
!    !-----------------------------------
!
!    !-----------------------------------
!    ! write the necessary keywords to the primary HDU
!    ! TODO
!    ! write date keyword
!    call  FTPDAT(unit, status)
!    if (status /= 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------

    !-----------------------------------
    ! append a new binary table HDU
    !FTIBIN(unit,nrows,tfields,ttype,tform,tunit,extname,varidat > status)
    call FTIBIN_w(unit,lc%num_rows,ncols,ttype,tform,tunit,trim(lc%hduinfo%key_extname%value),varidat,status)
    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------
!
!    !-----------------------------------
!    ! append a new table HDU
!    call FTCRHD(unit, status)
!    if (status /= 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------
!
!    !-----------------------------------
!    ! write header info for the binary table:
!    call FTPHBN(unit,lc%num_rows,ncols,ttype,tform,tunit,extname,varidat,status)
!    if (status /= 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------

    ! update TSTART/TSTOP using GTI
    if(gtilist%is_valid .and. gtilist%num_rows >= 1) then
       call fitswrap_set_key(lc%hduinfo%key_tstart, gtilist%pstart(1))
       call fitswrap_set_key(lc%hduinfo%key_tstop, gtilist%pstop(gtilist%num_rows))
    end if

    !-----------------------------------
    call srctool_lc_write_lc_keys(unit, lc, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing lc fits keys")
      return
    end if
    !-----------------------------------

    !-----------------------------------
    ! now write the column data to the file
    ! format: FTPCL[SLBIJKEDCM](unit,colnum,frow,felem,nelements,values, > status)

    ! white data in chuncks of size nblock (found from fitsio)
    call FTGRSZ_w(unit, nblock, status)
    do irow=1, lc%num_rows, nblock
      nrow = min(nblock, lc%num_rows-irow+1)

      call FTPCLD_w(unit, COLNUM_TIME, irow, felem, nrow, &
          lc%ptime(irow:irow+nrow-1), status)
      call FTPCLD_w(unit, COLNUM_TIMEDEL, irow, felem, nrow, &
          lc%ptimedel(irow:irow+nrow-1), status)
      call FTPCLJ_w(unit, COLNUM_COUNTS, irow, felem, nrow*lc%num_ebands, &
          flatten(lc%pcounts(:,irow:irow+nrow-1)), status)
      call FTPCNE_w(unit, COLNUM_RATE, irow, felem, nrow*lc%num_ebands, &
          flatten(lc%prate(:,irow:irow+nrow-1)), NULLVAL_REAL32, status)
      call FTPCNE_w(unit, COLNUM_RATE_ERR, irow, felem, nrow*lc%num_ebands, &
          flatten(lc%prate_err(:,irow:irow+nrow-1)), NULLVAL_REAL32, status)
      call FTPCNE_w(unit, COLNUM_RATE_ESTPOSERR, irow, felem, nrow*lc%num_ebands, &
          flatten(lc%prate_estposerr(:,irow:irow+nrow-1)), NULLVAL_REAL32, status)
      call FTPCNE_w(unit, COLNUM_RATE_ESTNEGERR, irow, felem, nrow*lc%num_ebands, &
          flatten(lc%prate_estnegerr(:,irow:irow+nrow-1)), NULLVAL_REAL32, status)
      call FTPCLE_w(unit, COLNUM_BACK_COUNTS, irow, felem, nrow*lc%num_ebands, &
          flatten(lc%pback_counts(:,irow:irow+nrow-1)), status)
      call FTPCLE_w(unit, COLNUM_FRACAREA, irow, felem, nrow*lc%num_ebands, &
          flatten(lc%pfracarea(:,irow:irow+nrow-1)), status)
      call FTPCLE_w(unit, COLNUM_FRACFLATAREA , irow, felem, nrow, &
          lc%pfracflatarea(irow:irow+nrow-1), status)
      call FTPCLE_w(unit, COLNUM_FRACTIME, irow, felem, nrow, &
          lc%pfractime(irow:irow+nrow-1), status)
      call FTPCLE_w(unit, COLNUM_FRACEXP, irow, felem, nrow*lc%num_ebands, &
          flatten(lc%pfracexp(:,irow:irow+nrow-1)), status)
      call FTPCLE_w(unit, COLNUM_FRACFLATEXP, irow, felem, nrow, &
          lc%pfracflatexp(irow:irow+nrow-1), status)
      call FTPCNE_w(unit, COLNUM_OFFAXIS, irow, felem, nrow, &
          lc%poffaxis(irow:irow+nrow-1), NULLVAL_REAL32, status)
      call FTPCNE_w(unit, COLNUM_BACKSCAL_SRC , irow, felem, nrow, &
          lc%pbackscal_src(irow:irow+nrow-1), NULLVAL_REAL32, status)
      call FTPCNE_w(unit, COLNUM_BACKSCAL_BACK, irow, felem, nrow, &
          lc%pbackscal_back(irow:irow+nrow-1), NULLVAL_REAL32, status)
      call FTPCND_w(unit, COLNUM_BACKRATIO, irow, felem, nrow, &
          lc%pbackratio(irow:irow+nrow-1), NULLVAL_REAL64, status)
    end do

    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    !-----------------------------------


    !-----------------------------------
    ! append the gtilist as a new hdu
    if ( gtilist%is_valid ) then
      call srctool_gti_append_gtilist_hdu(unit, gtilist, ierr)
      if ( ierr /= ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem appending GTI extension")
        return
      end if
    end if
    !-----------------------------------


    !-----------------------------------
    ! now close the fits file
    call FTCLOS_w(unit, status)
    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! deallocate the fortran unit
    call FTFIOU_w( unit, status)
    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------


  end subroutine srctool_lc_write_lc
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! write a set of lc files to disk
  ! we expect that all relevant parameters are already set up
  subroutine srctool_lc_write_lclist(src, lclist, instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_lc_write_lclist'

    type(src_type), intent(inout) :: src
    type(lclist_type),  intent(inout) :: lclist
    type(instlist_type),  intent(in) :: instlist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i
    type(gtilist_type) :: dummy_gtilist
    ierr = ERROR_LEVEL_OK
    dummy_gtilist%is_valid = .false.

    do i=0,MAX_TOT_INST
      if ( MERGED_TM_TO_IDX(i)>0 .and. instlist%inst(i)%to_write ) then
        call srctool_lc_write_lc(src, lclist%lc(i), dummy_gtilist, ierr)
      else if ( instlist%inst(i)%to_consider .and. instlist%inst(i)%to_write ) then
        call srctool_lc_write_lc(src, lclist%lc(i), instlist%inst(i)%src_gtilist, ierr)
      end if
      if ( ierr /= ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem writing lightcurve file")
        return
      end if
    end do

  end subroutine srctool_lc_write_lclist
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! set up binning for an lc
  ! assume that input params have already been copied into the lc structure
  subroutine srctool_lc_prepare_lc_template(lc, instlist, ierr )
    character(len=*), parameter :: subroutinename='srctool_lc_prepare_lc_template'
    type(lc_type),  intent(inout) :: lc
    type(instlist_type), intent(in),target :: instlist
    integer(kind=int32), intent(inout) :: ierr
    integer(kind=int32)                :: status
    real(kind=real64)        :: timedel, half_timedel
    real(kind=real64)        :: tstart, tstop
    real(kind=real64)        :: t1, t2
    integer(kind=int32)      :: nbins
    integer(kind=int32)      :: i,ii
    character(len=STR_MAX) :: str_buffer
    real(kind=real64), parameter :: nullval_real64  = -999.9d0
    type(gtilist_type), pointer :: pgtilist

    if ( allocated(lc%ptime))    deallocate(lc%ptime)
    if ( allocated(lc%ptimedel)) deallocate(lc%ptimedel)
    if ( allocated(lc%ptmin))    deallocate(lc%ptmin)
    if ( allocated(lc%ptmax))    deallocate(lc%ptmax)

    ! how to proceed depends on the lctype parameter
    select case ( lc%lctype_code )
    case ( LCTYPE_CODE_REGULAR, LCTYPE_CODE_REGULAR_PLUS, LCTYPE_CODE_REGULAR_MINUS ) ! evenly spaced bins
      ! interpretation depends on the number of parameters supplied
      select case ( lc%num_lcpars )
      case ( 1 ) ! param 1 is the bin time, tstart and tstop derived from eventlist TSTART and TSTOP
        timedel = lc%lcpars(1)
        tstart  = instlist%min_tstart
        tstop   = instlist%max_tstop
      case ( 2 ) !  param 1 is the bin time, param 2 is tstart, tstop derived from eventlist TSTOP
        timedel = lc%lcpars(1)
        tstart  = lc%lcpars(2)
        tstop   = instlist%max_tstop
      case ( 3 ) !  param 1 is the bin time, param 2 is tstart, param 3 is tstop
        timedel = lc%lcpars(1)
        tstart  = lc%lcpars(2)
        tstop   = lc%lcpars(3)

      case default
        write (str_buffer, '(A,I0,A)') &
            "Invalid number of lcpars (", lc%num_lcpars, ") were supplied: "
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end select


      ! make some sanity checks on the bin params
      if (  timedel <= 0.0d0 ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,F18.4,A,I0,A)') &
            "Invalid value of LC binning param: timedel=", timedel, &
            " (npars_supplied=",  lc%num_lcpars, ")"
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end if
      if (  tstart >= tstop ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(2(A,F18.4) ,I0,A)') &
            "Invalid value of LC binning params: tstart=", tstart, " tstop=", tstop, &
            " (npars_supplied=",  lc%num_lcpars, ")"
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end if

      call fitswrap_set_key(lc%key_timedel, timedel)

      nbins = int((tstop - tstart ) / timedel, int32) + 1
!      nbins = int(((tstop - tstart ) / timedel)+0.5d0, int32)
      if ( tstart + timedel*real(nbins-1,kind=real64) >= tstop ) then
        nbins = nbins - 1
      endif
      half_timedel = 0.5d0 * timedel

      if (  nbins < 0 ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,I0,A,I0,A)') &
            "Invalid value of LC binning params: nbins=", nbins, &
            " (npars_supplied=",  lc%num_lcpars, ")"
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end if

      write (str_buffer, '(A,I0,A,I0,A)') &
          "Working with following ", lc%num_lcpars, " LC parameters (npars_supplied=",  lc%num_lcpars, "):"
      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,F16.4)') "timedel= ", timedel
      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,F16.4)') "tstart=  ", tstart
      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,F16.4)') "tstop=   ", tstop
      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,I16)')   "nbins=   ", nbins
      call wrschatty (subroutinename, str_buffer)


      lc%num_rows = nbins
      allocate(lc%ptime(nbins), &
          lc%ptimedel(nbins), &
          lc%ptmin(nbins), &
          lc%ptmax(nbins), &
          STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
!      t1 = tstart - half_timedel
!      t2 = t1     + half_timedel
      t1 = tstart
      t2 = t1     + timedel
      do i=1,lc%num_rows
        t2 = MIN(t2,tstop)
        lc%ptimedel(i) = t2 - t1
        lc%ptime(i)    = 0.5d0*(t1 + t2)
        lc%ptmin(i)    = t1 ! lc%ptime(i) - 0.5d0*lc%ptimedel(i)
        lc%ptmax(i)    = t2 ! lc%ptime(i) + 0.5d0*lc%ptimedel(i)
        t1 = t2
        t2 = t2 + timedel
      end do
!      if ( lc%ptimedel(lc%num_rows) <= 0.0d0 ) then
!        lc%num_rows = lc%num_rows -1
!      endif

      ! TODO - deal more efficiently with gaps between GTIs for REGULAR+ binning??

      if ( ierr /= ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Error when interpreting lc descriptions")
        return
      end if


    case ( LCTYPE_CODE_SRCGTI, LCTYPE_CODE_GTI , LCTYPE_CODE_SRCGTI_MINUS )
      nullify(pgtilist)
      if ( (lc%lctype_code == LCTYPE_CODE_SRCGTI) .or. &
           (lc%lctype_code == LCTYPE_CODE_SRCGTI_MINUS ) ) then
        pgtilist => instlist%inst(0)%src_gtilist
      else if ( lc%lctype_code == LCTYPE_CODE_GTI ) then
        pgtilist => instlist%inst(0)%gtilist
      endif
      ! Use bins which match the src gtis, with a gap between each pair of gtis
      lc%num_rows = (2 * pgtilist%num_rows) -1
      allocate(lc%ptime(lc%num_rows), &
          lc%ptimedel(lc%num_rows), &
          lc%ptmin(lc%num_rows), &
          lc%ptmax(lc%num_rows), &
          STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return

      ii = 1
      do i=1,lc%num_rows
        if ( MODULO(i,2) == 1 ) then
          lc%ptmin(i) = pgtilist%pstart(ii)
          lc%ptmax(i) = pgtilist%pstop(ii)
          ii = ii + 1
        else
          lc%ptmin(i) = lc%ptmax(i-1)       ! end of the last bin
          lc%ptmax(i) = pgtilist%pstart(ii) ! start of the next bin
        endif
        lc%ptime(i) = 0.5d0 * (lc%ptmin(i) + lc%ptmax(i))
        lc%ptimedel(i) = (lc%ptmax(i) - lc%ptmin(i))
      end do
      call fitswrap_set_key_null(lc%key_timedel)

    case ( LCTYPE_CODE_FOLDED )
      write (str_buffer, '(A,A,A,I0,A)') "Unimplemented LC type: ", trim(lc%lctype), " (status=", status, ")"
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return

    case default
      write (str_buffer, '(A,A,A,I0,A)') "Unknown LC type: ", trim(lc%lctype), " (status=", status, ")"
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return
    end select      ! end select on lctype_code
    !--------------------------------

  end subroutine srctool_lc_prepare_lc_template

  !-------------------------------------------------------------------------------
  ! set up keywords,arrays etc for an lc file
  subroutine srctool_lc_setup_lc(lc, lc_template, inst, src, ierr )
    character(len=*), parameter :: subroutinename='srctool_lc_setup_lc'
    type(lc_type),      intent(inout) :: lc
    type(lc_type),      intent(in)    :: lc_template
    type(inst_type),     intent(in),target :: inst
    type(src_type),      intent(in)    :: src
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)                :: status
!    character(len=STR_MAX)             :: str_buffer
!    character(len=STR_MAX)             :: str_buffer2
    integer(kind=int32)      :: i,j,e,flag
    real(kind=real64)        :: emid, weight_sum
    type(arf_type),    pointer :: prefarf
    ierr = ERROR_LEVEL_OK
    status = 0
    lc%to_consider = .true.

    ! base the lc structure on the lc template
    lc%lctype_code = lc_template%lctype_code
    lc%num_rows   = lc_template%num_rows
    lc%num_ebands = lc_template%num_ebands
!    lc%int_nenergies = lc_template%int_nenergies

    if ( .not. allocated(lc%ptime) ) then

      allocate(lc%ptime(lc%num_rows),&
          lc%ptimedel      (lc%num_rows),&
          lc%ptmin         (lc%num_rows),&
          lc%ptmax         (lc%num_rows),&
          lc%pcounts       (lc%num_ebands,lc%num_rows),&
          lc%prate         (lc%num_ebands,lc%num_rows),&
          lc%prate_err     (lc%num_ebands,lc%num_rows),&
          lc%prate_estposerr(lc%num_ebands,lc%num_rows),&
          lc%prate_estnegerr(lc%num_ebands,lc%num_rows),&
          lc%pback_counts  (lc%num_ebands,lc%num_rows),&
          lc%pfracarea     (lc%num_ebands,lc%num_rows),&
          lc%pfracflatarea (lc%num_rows), &
          lc%pfractime     (lc%num_rows), &
          lc%pfracexp      (lc%num_ebands,lc%num_rows),&
          lc%pfracflatexp  (lc%num_rows), &
          lc%poffaxis      (lc%num_rows), &
          lc%pbackscal_src (lc%num_rows), &
          lc%pbackscal_back(lc%num_rows), &
          lc%pbackratio    (lc%num_rows), &
          STAT = status)

      if ( check_allocate_status(subroutinename, status, ierr) ) return

    end if

    ! initialise the values in the arrays
    lc%ptime          = 0.0d0
    lc%ptimedel       = 0.0d0
    lc%ptmin          = 0.0d0
    lc%ptmax          = 0.0d0
    lc%pcounts        = 0
    lc%prate          = 0.0e0
    lc%prate_err      = 0.0e0
    lc%prate_estposerr= 0.0e0
    lc%prate_estnegerr= 0.0e0
    lc%pback_counts   = 0
    lc%pfracarea      = 0.0e0
    lc%pfracflatarea  = 0.0e0
    lc%pfractime      = 0.0e0
    lc%pfracexp       = 0.0e0
    lc%pfracflatexp   = 0.0e0
    lc%poffaxis       = 0.0e0
    lc%pbackscal_src  = 0.0e0
    lc%pbackscal_back = 0.0e0
    lc%pbackratio     = 0.0d0
    ! copy the time bins over from the lc template
    lc%ptime(:)       = lc_template%ptime(:)
    lc%ptimedel(:)    = lc_template%ptimedel(:)
    lc%lcgamma        = lc_template%lcgamma
    lc%ptmin(:)       = lc_template%ptmin(:)
    lc%ptmax(:)       = lc_template%ptmax(:)
!    ! and the integration binning stuff
!    lc%pint_e      = lc_template%pint_e
!    lc%pint_weights = lc_template%pint_weights
!    lc%pint_index1 = lc_template%pint_index1
!    lc%pint_index2 = lc_template%pint_index2

    ! and copy over the relevant keywords from the lc template
    lc%key_timedel    = lc_template%key_timedel
    lc%key_timezero   = lc_template%key_timezero

    lc%key_telapse    = lc_template%key_telapse
    lc%key_ontime     = lc_template%key_ontime
    lc%key_livetime   = lc_template%key_livetime
    lc%key_exposure   = lc_template%key_exposure

    lc%key_tierrela   = lc_template%key_tierrela
    lc%key_tierabso   = lc_template%key_tierabso
    lc%key_clockcor   = lc_template%key_clockcor
    lc%key_backapp    = lc_template%key_backapp
    lc%key_deadapp    = lc_template%key_deadapp
    lc%key_vignapp    = lc_template%key_vignapp
    lc%key_emin       = lc_template%key_emin
    lc%key_emax       = lc_template%key_emax

    ! NOT SURE THIS IS RIGHT - CHECK
    !if ( inst%index == 0 ) then
!
    if ( MERGED_TM_TO_IDX(inst%index)>0 ) then

!      !TODO
!      ! we put some placeholder values here for now, but should eventually derive these from TM1-7 values
!      call fitswrap_set_key(lc%hduinfo%key_telescop, "TELESCOP" ,  "eROSITA", "Telescope name")
!      call fitswrap_set_key(lc%hduinfo%key_instrume, "INSTRUME" ,      "TM0", "Instrument name")
!      call fitswrap_set_key(lc%hduinfo%key_filter,   "FILTER"   ,  "UNKNOWN", "CCD filter used")
!      call fitswrap_set_key(lc%hduinfo%key_tstart,   "TSTART"   ,      0.0d0, "[s]")
!      call fitswrap_set_key(lc%hduinfo%key_tstop,    "TSTOP"    ,      1.0d0, "[s]")
!      call fitswrap_set_key(lc%hduinfo%key_date_obs, "DATE-OBS" , "YYYY-MM-DDThh:mm:ss.s", "[YYYY-MM-DDThh:mm:ss.s]")
!      call fitswrap_set_key(lc%hduinfo%key_date_end, "DATE-END" , "YYYY-MM-DDThh:mm:ss.s", "[YYYY-MM-DDThh:mm:ss.s]")
    else
      ! copy the hduinfo from the first evtlist in the inst structure
      ! but first check that there is at least one valid evtlist
      if ( inst%evtlistlist%num_evtlists <= 0  ) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Instrument structure contains no evtlists")
        return
      end if
      lc%hduinfo = inst%evtlistlist%hduinfo
    end if

    ! set the output filename and DSTYPE
    call srctool_misc_set_filename(lc%hduinfo%filename, inst, src, SRCTOOL_FILETYPE_CODE_LC )
    call fitswrap_set_key(lc%hduinfo%key_dstype, ESASS_DSTYPE_LC )

    call fitswrap_set_key(lc%hduinfo%key_dscode, &
        srctool_misc_dscode(inst%index,ESASS_FILTER_STR(inst%filter_code),0) )



    ! now set some more appropriate values for an OGIP compliant lc file
    call fitswrap_set_key(lc%hduinfo%key_extname,   "RATE"      )
    call fitswrap_set_key(lc%hduinfo%key_hduclass,  "OGIP"      )
    call fitswrap_set_key(lc%hduinfo%key_hduclas1,  "LIGHTCURVE")
    call fitswrap_set_key(lc%hduinfo%key_hduclas2,  "TOTAL"     )
    call fitswrap_set_key(lc%hduinfo%key_hduclas3,  "COUNT"     )

    call fitswrap_set_key(lc%hduinfo%key_hduvers,  "1994Nov30" )
!    call fitswrap_set_key(lc%hduinfo%key_creator,  "")

!   TODO: Check if these keywords should just be populated from the input evtlist
    call fitswrap_set_key(lc%hduinfo%key_object  , trim(src%name))
    call fitswrap_set_key(lc%hduinfo%key_ra_obj  , src%coord%long)
    call fitswrap_set_key(lc%hduinfo%key_dec_obj , src%coord%lat)
    call fitswrap_set_key(lc%hduinfo%key_radecsys, COORDSYS_STR(src%coord%coordsys_code))
    call fitswrap_set_key(lc%hduinfo%key_equinox , 2000.0d0)

    ! ownership
    call fitswrap_set_key(lc%hduinfo%key_owner, srctool_owner_getkeyword_owner(src%owner))

!    ! TODO - put the correct values in the following
!    call fitswrap_set_key(lc%hduinfo%key_date_end, "DATE-END" , "YYYY-MM-DDThh:mm:ss.s", "[YYYY-MM-DDThh:mm:ss.s]")

    call fitswrap_set_key(lc%key_clockcor, .false.)
    call fitswrap_set_key(lc%key_backapp,  .false.)
    call fitswrap_set_key(lc%key_deadapp,  .false.)
    call fitswrap_set_key(lc%key_vignapp,  .false.)
    call fitswrap_set_key(lc%key_timezero,  0.0d0)

    call fitswrap_set_key(lc%key_telapse,  0.0d0) ! these get set later
    call fitswrap_set_key(lc%key_ontime,   0.0d0)
    call fitswrap_set_key(lc%key_livetime, 0.0d0)
    call fitswrap_set_key(lc%key_exposure, 0.0d0)

    call fitswrap_set_key(lc%key_tierrela, 0.0d0)
    call fitswrap_set_key(lc%key_tierabso, 0.0d0)

    call fitswrap_set_key(lc%key_src_region,  trim(adjustl(src%str_src_region)))
    call fitswrap_set_key(lc%key_back_region, trim(adjustl(src%str_back_region)))

    ! set the FILTER keyword
    call fitswrap_set_key(lc%hduinfo%key_filter, ESASS_FILTER_STR(inst%filter_code) )

    !--------------------------------
    ! now set up the lc weighting and energies arrays
    ! Base this on the refarf and the lcgamma param
    ! weight(E) = E^-gamma * ARF(E) * dE / (Emax - Emin)
    ! where dE is the bin width
    call srctool_misc_select_refarf_pointer(inst,prefarf,ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Could not find suitable reference ARF")
      return
    end if

    ! TODO TODO
    ! count the number of arf energy bins that fall within at least one of the lc ebands
    lc%int_nenergies = 0
    do i=1,prefarf%num_rows
      flag = 0
      do e=1,lc%num_ebands
        emid = 0.5d0*(prefarf%pelo(i) + prefarf%pehi(i))
        if (emid >= lc%key_emin(e)%value .and. &
            emid < lc%key_emax(e)%value   ) then
          flag = 1
          exit
        end if
      end do
      if ( flag == 1 ) lc%int_nenergies = lc%int_nenergies + 1
    end do

    if ( .not. allocated(lc%pint_e) ) then
      allocate(lc%pint_e(lc%int_nenergies), &
          lc%pint_weights(lc%num_ebands,lc%int_nenergies), &
          STAT = status)

      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if

    ! now set the energies and weights
    j = 1
    lc%pint_weights(:,:) = 0.0d0
    do i=1,prefarf%num_rows
      flag = 0
      do e=1,lc%num_ebands
        emid = 0.5d0*(prefarf%pelo(i) + prefarf%pehi(i))
        if (emid >= lc%key_emin(e)%value .and. &
            emid < lc%key_emax(e)%value   ) then
          flag = 1
          lc%pint_e(j) = emid
          lc%pint_weights(e,j) = prefarf%presp(i)* &
              (emid**(-1.0d0 * lc%lcgamma)) * &
              (prefarf%pehi(i) - prefarf%pelo(i))
        end if
      end do
      if ( flag == 1 ) j = j + 1
    end do

    ! now normalise the weights to unity in each lc band
    do e=1,lc%num_ebands
      weight_sum = sum(lc%pint_weights(e,:))
      if ( weight_sum > 0.0d0 ) then
        lc%pint_weights(e,:) = lc%pint_weights(e,:) / weight_sum
      end if
    end do

    do e=1,lc%num_ebands
!      write (str_buffer,'(A,I0)'), "CTS_SC", e
!      write (str_buffer2,'(A,I0)'), "Cts in source reg,band ", e
!      call fitswrap_set_key(lc%key_ncounts_src(e), trim(str_buffer),   0, trim(str_buffer2))
!      write (str_buffer,'(A,I0)'), "CTS_BK", e
!      write (str_buffer2,'(A,I0)'), "Cts in bkgnd reg,band ", e
!      call fitswrap_set_key(lc%key_ncounts_back(e), trim(str_buffer),  0, trim(str_buffer2))
      call fitswrap_set_key(lc%key_ncounts_src(e),  0)
      call fitswrap_set_key(lc%key_ncounts_back(e), 0)
    end do


!    if ( inst%index == 1 ) then
!      write (*, '(A,1x,A6,2(1x,A6,1x,A8,1x,A12))'), "#DEBUG_LCINT", &
!          "INST", "lc_e", "lc_emin", "lc_emax", &
!          "j", "e(j)", "weight(e,j)"
!      do e=1,lc%num_ebands
!        do j=1,lc%int_nenergies
!          write (*, '(A,1x,A6,2(1x,I6,1x,F8.3,1x,F12.8))'), "DEBUG_LCINT ", &
!              trim(inst%name), e, lc%key_emin(e)%value, lc%key_emax(e)%value, &
!              j, lc%pint_e(j), lc%pint_weights(e,j)
!        end do
!      end do
!    end if

  end subroutine srctool_lc_setup_lc
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! set up keywords,arrays etc for a lclist
  subroutine srctool_lc_setup_lclist(lclist, lc_template, instlist, src, ierr )
    character(len=*), parameter :: subroutinename='srctool_lc_setup_lclist'
    type(lclist_type),    intent(inout) :: lclist
    type(lc_type),        intent(inout) :: lc_template
    type(instlist_type),  intent(in)    :: instlist
    type(src_type),       intent(in)    :: src
    integer(kind=int32),  intent(out)   :: ierr
    integer(kind=int32)                 :: i, mi, mtm
    integer(kind=int32)                 :: nok
!    character(len=STR_MAX)             :: str_buffer


    ierr = ERROR_LEVEL_OK
    call srctool_lc_prepare_lc_template(lc_template, instlist, ierr )
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Error preparing lc template structure")
      return
    end if

    do i=0,MAX_TOT_INST
      if ( (MERGED_TM_TO_IDX(i)>0) .or. instlist%inst(i)%to_consider ) then
        call srctool_lc_setup_lc(lclist%lc(i), lc_template, instlist%inst(i), src, ierr )
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error setting up src lc data structure")
          return
        end if
      end if
    end do

!    ! now set up the hduinfo values in the merged instrument lc
!    ! derive them from the first valid lc - with INSTRUME and filename adapted accordingly
!    do i=1,MAX_INST
!      if ( (instlist%inst(i)%to_consider ) ) then
!        lclist%lc(0)%hduinfo = lclist%lc(i)%hduinfo
!        call fitswrap_set_key(lclist%lc(0)%hduinfo%key_instrume, "INSTRUME",trim(instlist%inst(0)%name),"Instrument name")
!        ! sort out the output filename
!        call srctool_misc_set_filename(lclist%lc(0)%hduinfo%filename, &
!            instlist%inst(0), src, SRCTOOL_FILETYPE_CODE_LC )
!        exit
!      end if
!    end do

    do mi=1, NUM_MERGED_INST
      mtm = MERGED_INST(mi)

      ! copy the basic keywords from the first valid instrument for the merged instrument files
      do i=1,MAX_INST
        if ( instlist%inst(i)%to_consider .and. MERGED_INST_CMPT(mtm, i) ) then
          lclist%lc(mtm)%hduinfo  = lclist%lc(i)%hduinfo
          exit
        end if
      end do

      ! now set up the hduinfo values in the "MERGED" lc files
      ! derive them from the first valid lc file - with INSTRUME and filename adapted accordingly
      call fitswrap_set_key(lclist%lc(mtm)%hduinfo%key_instrume, INSTRUME_IS_MERGED)
      call srctool_misc_set_filename(lclist%lc(mtm)%hduinfo%filename, &
          instlist%inst(mtm), src, SRCTOOL_FILETYPE_CODE_LC )

      ! now deal with the NINST and INSTRUMn keywords
      nok = 0
      do i=1,MAX_INST
        if ( instlist%inst(i)%to_consider ) then
          nok = nok + 1
          !write (str_buffer, '(A,I0)') "INSTRUM", nok
          call fitswrap_set_key(lclist%lc(mtm)%hduinfo%pkey_instrum(nok), &
              trim(instlist%inst(i)%name))
        end if
      end do
      call fitswrap_set_key(lclist%lc(mtm)%hduinfo%key_ninst, nok)
    end do

  end subroutine srctool_lc_setup_lclist
  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  ! gather lc info from an evtlist and add to a lc file
  ! assume that the lc file has been setup
  subroutine srctool_lc_accumulate_lc(lc, evtlist, ierr )
    character(len=*), parameter :: subroutinename='srctool_lc_accumulate_lc'
    type(lc_type),      intent(inout) :: lc
    type(evtlist_type),  intent(in)    :: evtlist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i
    integer(kind=int32) :: tbin, ebin, ei
    integer(kind=int32) :: status
    real(kind=real64)   :: t
    real(kind=real32)   :: e_keV
    real(kind=real64)   :: t1,t2
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32), dimension(1:MAX_LCBANDS)  :: ncounts_src, ncounts_back  !counters

    ierr = ERROR_LEVEL_OK
    status = 0

    ncounts_src(:) = 0
    ncounts_back(:) = 0

    if ( evtlist%num_rows <= 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Error accumulating lc (no events)")
      return
    endif

    if ( lc%num_rows > 0 ) then

  !    t1 = lc%ptime(1) - 0.5d0*lc%ptimedel(1)
  !    t2 = lc%ptime(lc%num_rows) + 0.5d0*lc%ptimedel(lc%num_rows)
      t1 = lc%ptmin(1)
      t2 = lc%ptmax(lc%num_rows)
      do i=1,evtlist%num_rows
        if ( evtlist%pclass_mask(i) /= EVT_CLASS_FLAG_NULL ) then
          t = evtlist%ptime(i)
          if ( (t >= t1) .and.  (t < t2) ) then
          ! determine the lc bin (tbin) in which this event falls
            if ( lc%num_rows > 1 ) then
              call binary_search(lc%ptmin(:), t, tbin, status)
              if ((status /= 0) .and. (status /= 2) ) then
                tbin = 0  ! make sure that this event is not recorded
  !              ierr = ERROR_LEVEL_SEVERE
  !              call wrserr (subroutinename, ierr, "Could not find lc bin")
  !              return
              end if
            else
              tbin = 1
            endif
            if ( tbin > 0 ) then
              ! now refine the bin:
              tbin=MAX(1,tbin-1)
              do tbin=tbin,lc%num_rows
                !            if ((t >= (lc%ptime(tbin) - 0.5*lc%ptimedel(tbin))) .and. &
                !                (t < (lc%ptime(tbin) + 0.5*lc%ptimedel(tbin))) ) then
                if ((t >= lc%ptmin(tbin)) .and. &
                    (t < lc%ptmax(tbin))) then
                  exit ! get out of do loop
                end if
              end do

              ! work out which energy bands the event falls within
              e_keV = real(EV2KEV * evtlist%penergy(i), real32)
              do ebin=1,lc%num_ebands
                if ((e_keV >= lc%key_emin(ebin)%value) .and. &
                    (e_keV < lc%key_emax(ebin)%value)) then
                  ei = (tbin-1)*lc%num_ebands + ebin
                  if ( IAND(evtlist%pclass_mask(i),EVT_CLASS_FLAG_SRC) /= 0 ) then
                    lc%pcounts(ebin,tbin) = lc%pcounts(ebin,tbin) + 1
                    ncounts_src(ebin) = ncounts_src(ebin) + 1
                  end if
                  if ( IAND(evtlist%pclass_mask(i),EVT_CLASS_FLAG_BACK) /= 0 ) then
                    lc%pback_counts(ebin,tbin) = lc%pback_counts(ebin,tbin) + 1
                    ncounts_back(ebin) = ncounts_back(ebin) + 1
                  end if
                end if
              end do
           endif
          else
            ! silently ignore any out-of-bounds events
          end if

       end if
      end do
    endif

!     ! debug
! !    write (str_buffer, '(A,A6,3(1x,I8))'), &
! !       "DEBUG LC_ACCUM: ",  trim(evtlist%hduinfo%key_instrume%value), &
! !        evtlist%num_rows, ncounts_src, ncounts_back
!     write (str_buffer, '(A,2(1x,A8),2(1x,I8))'), &
!         "DEBUG LC_ACCUM: ",  "SRC", trim(evtlist%hduinfo%key_instrume%value), &
!         evtlist%num_rows, ncounts_src
!     call wrschatty (subroutinename, str_buffer)
!     write (str_buffer, '(A,2(1x,A8),2(1x,I8))'), &
!         "DEBUG LC_ACCUM: ",  "BACK", trim(evtlist%hduinfo%key_instrume%value), &
!         evtlist%num_rows, ncounts_back
!     call wrschatty (subroutinename, str_buffer)

    ! copy the flagsel and pat_sel keywords from the evtlist
    lc%key_flagsel = evtlist%key_flagsel
    lc%key_pat_sel = evtlist%key_pat_sel

    do ebin=1,lc%num_ebands
!      write (str_buffer, '(A,2(1x,A10),2(1x,F10.3),1(1x,I10))'), &
!          "DEBUG LC_ACCUM: ",  "SRC", trim(evtlist%hduinfo%key_instrume%value), &
!          lc%key_emin(ebin)%value, lc%key_emax(ebin)%value, &
!          ncounts_src(ebin)
!      call wrschatty (subroutinename, str_buffer)
!      write (str_buffer, '(A,2(1x,A10),2(1x,F10.3),1(1x,I10))'), &
!          "DEBUG LC_ACCUM: ",  "BACK", trim(evtlist%hduinfo%key_instrume%value), &
!          lc%key_emin(ebin)%value, lc%key_emax(ebin)%value, &
!          ncounts_back(ebin)
!      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,1(1x,A10),2(1x,F10.3),2(1x,I10))') &
          "DEBUG LC_ACCUM: ", trim(evtlist%hduinfo%key_instrume%value), &
          lc%key_emin(ebin)%value, lc%key_emax(ebin)%value, &
          ncounts_src(ebin), ncounts_back(ebin)
      call wrschatty (subroutinename, str_buffer)

      call fitswrap_set_key(lc%key_ncounts_src(ebin),  ncounts_src(ebin))
      call fitswrap_set_key(lc%key_ncounts_back(ebin), ncounts_back(ebin))

    end do

  end subroutine srctool_lc_accumulate_lc
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! gather lc info from an evtlist and add to a lc file
  ! assume that the lc file has been setup
  subroutine srctool_lc_accumulate_lclist(lclist, instlist, ierr )
    character(len=*), parameter :: subroutinename='srctool_lc_accumulate_lc'
    type(lclist_type),    intent(inout) :: lclist
    type(instlist_type),  intent(in)    :: instlist
    integer(kind=int32),  intent(out)   :: ierr
    integer(kind=int32) :: i,j
    character(len=STR_MAX) :: str_buffer

    ierr = ERROR_LEVEL_OK
    write (str_buffer, '(A,5(1x,A10))') &
        "DEBUG LC_ACCUM: ", &
        "INSTR", "Emin(keV)", "Emax(keV)", "Src_cnts", "Back_cnts"
    call wrschatty (subroutinename, str_buffer)

    !-----------------------------------------------
    do i=1,MAX_INST
      do j=1,instlist%inst(i)%evtlistlist%num_evtlists
        call srctool_lc_accumulate_lc(lclist%lc(i), instlist%inst(i)%evtlistlist%evtlist(j), ierr )
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error accumulating lc")
          return
        end if
      end do
    end do
    !-----------------------------------------------

!    ! now add up the counts to make the the merged instrument lc
!    ! sum the lcs over the seven telescopes
!    call srctool_coadd_lclist(lclist, instlist )

  end subroutine srctool_lc_accumulate_lclist
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! coadd the lc data across all 7 telescope modules
  subroutine srctool_lc_coadd_lclist(lclist, instlist, comb_tm)
    character(len=*), parameter :: subroutinename='srctool_lc_coadd_lclist'
    type(lclist_type), target, intent(inout) :: lclist
    type(instlist_type), intent(in) :: instlist
    integer(kind=int32), intent(in) :: comb_tm

    integer(kind=int32) :: itm, it, ib
    real(kind=real64) :: tstart, tstop
    integer(kind=int32) :: first_valid_inst
    real(kind=real64) :: weight, tot_weight, tot_backscal_src, tot_backscal_back, tot_offaxis, &
        tot_fracflatarea, tot_fracflatexp

    integer(kind=int32) :: n_ok, n_tot

    type(lc_type), pointer :: out_lc

    call wrschatty(subroutinename, "Coadding lightcurve for merged instrument TM" // &
        trim(i2s(comb_tm)))

    out_lc => lclist%lc(comb_tm)

    first_valid_inst = -1

    out_lc%pcounts       = 0
    out_lc%pback_counts  = 0
    out_lc%pfracarea     = 0
    out_lc%pfracflatarea = 0
    out_lc%pfractime     = 0
    out_lc%pfracexp      = 0
    out_lc%pfracflatexp  = 0
    out_lc%poffaxis      = 0
    out_lc%pbackscal_src = 0
    out_lc%pbackscal_back= 0
    out_lc%pbackratio    = 0

    tstart = huge(0.0)
    tstop = -huge(0.0)

    ! output time bins
    do it=1,out_lc%num_rows

      ! compute backscals, backratio, offaxis and fractime
      ! (weight non-fractime quantities by fractime - assume timedel is same for all LCs)
      tot_weight = 0
      tot_backscal_src = 0
      tot_backscal_back = 0
      tot_offaxis = 0
      tot_fracflatarea = 0
      tot_fracflatexp = 0

      ! number of TMs in bin with GTIs
      n_ok = 0
      ! total number of TMs to consider
      n_tot = 0

      do itm=1,MAX_INST
        if (instlist%inst(itm)%to_consider .and. MERGED_INST_CMPT(comb_tm,itm)) then
          if (lclist%lc(itm)%pbackscal_src(it) > 0 .and. lclist%lc(itm)%pfractime(it) > 0) then
            if ( first_valid_inst < 0 ) first_valid_inst = itm

            weight = lclist%lc(itm)%pfractime(it)
            tot_weight = tot_weight + weight
            tot_backscal_src = tot_backscal_src + lclist%lc(itm)%pbackscal_src(it) * weight
            tot_backscal_back = tot_backscal_back + lclist%lc(itm)%pbackscal_back(it) * weight
            tot_offaxis = tot_offaxis + lclist%lc(itm)%poffaxis(it) * weight
            tot_fracflatarea = tot_fracflatarea + lclist%lc(itm)%pfracflatarea(it) * weight
            tot_fracflatexp = tot_fracflatexp + lclist%lc(itm)%pfracflatexp(it)

            n_ok = n_ok + 1
          end if
          n_tot = n_tot + 1
        end if
      end do ! itm

      if ( tot_weight > 0 ) then
        ! we have some signal in this bin, so compute the quantities which do not depend on energy
        out_lc%pbackscal_src(it) = real(tot_backscal_src / tot_weight, real32)
        out_lc%pbackscal_back(it) = real(tot_backscal_back / tot_weight, real32)
        if ( tot_backscal_back > 0d0 ) then
          out_lc%pbackratio(it) = tot_backscal_src / tot_backscal_back
        else
          out_lc%pbackratio(it) = 0e0
        end if
        out_lc%poffaxis(it) = real(tot_offaxis / tot_weight, real32)
        out_lc%pfracflatarea(it) = real(tot_fracflatarea / tot_weight, real32)
        out_lc%pfractime(it) = real(tot_weight / n_tot, real32)
        out_lc%pfracflatexp(it) = real(tot_fracflatexp / n_tot, real32)

        ! now compute the energy-band specific quantities
        do ib=1,out_lc%num_ebands
          do itm=1,MAX_INST
            if (instlist%inst(itm)%to_consider .and. MERGED_INST_CMPT(comb_tm,itm)) then
              if (lclist%lc(itm)%pbackscal_src(it) > 0 .and. lclist%lc(itm)%pfractime(it) > 0) then
                out_lc%pcounts(ib,it) = out_lc%pcounts(ib,it) + lclist%lc(itm)%pcounts(ib,it)
                out_lc%pback_counts(ib,it) = out_lc%pback_counts(ib,it) + lclist%lc(itm)%pback_counts(ib,it)
                out_lc%pfracarea(ib,it) = out_lc%pfracarea(ib,it) + lclist%lc(itm)%pfracarea(ib,it) / n_ok
                out_lc%pfracexp(ib,it) = out_lc%pfracexp(ib,it) + lclist%lc(itm)%pfracexp(ib,it) / n_tot
              end if

            end if
          end do ! itm
        end do ! ib

      else
        ! write some nulls where appropriate
        out_lc%pbackratio(it) = NULLVAL_REAL64
        out_lc%poffaxis(it) = NULLVAL_REAL32
      end if

    end do ! it

    ! Compute the total counts for the header keywords
    do ib=1,out_lc%num_ebands
      out_lc%key_ncounts_src(ib)%value = sum(out_lc%pcounts(ib,:))
      out_lc%key_ncounts_back(ib)%value = int(sum(out_lc%pback_counts(ib,:)), int32)
    end do

    ! combine tstart/tstop with individual gtis
    do itm=1,MAX_INST
      if (instlist%inst(itm)%to_consider .and. MERGED_INST_CMPT(comb_tm,itm)) then
        if ( lclist%lc(itm)%to_consider ) then
          associate(gtilist => instlist%inst(itm)%src_gtilist)
            if(gtilist%is_valid .and. gtilist%num_rows >= 1) then
              tstart = min(gtilist%pstart(1), tstart)
              tstop = max(gtilist%pstop(gtilist%num_rows), tstop)
            end if
          end associate
        end if
      end if
    end do

    if ( first_valid_inst > 0 ) then
      ! copy the flagsel and pat_sel keywords from the first valid evtlist
      out_lc%key_flagsel  = lclist%lc(first_valid_inst)%key_flagsel
      out_lc%key_pat_sel  = lclist%lc(first_valid_inst)%key_pat_sel
    else
      ! nothing valid found, so use some invalid values
      call wrserr(subroutinename, ERROR_LEVEL_WARNING, "No valid inputs found for TM" // i2s(comb_tm))
      call fitswrap_set_key(out_lc%key_flagsel, 0_int32)
      call fitswrap_set_key(out_lc%key_pat_sel, 0_int16)
    end if

    ! set the filter keyword
    call fitswrap_set_key(out_lc%hduinfo%key_filter, ESASS_FILTER_STR(instlist%inst(0)%filter_code))

    ! update tstart/tstop
    if(tstart /= huge(0.0)) call fitswrap_set_key(out_lc%hduinfo%key_tstart, tstart)
    if(tstop /= -huge(0.0)) call fitswrap_set_key(out_lc%hduinfo%key_tstop, tstop)

    !---------------------------------------------------------------

    nullify(out_lc)

  end subroutine srctool_lc_coadd_lclist
  !-------------------------------------------------------------------------------


  recursive subroutine srctool_lc_get_lc_index (lc, t, i)
    character(len=*), parameter :: subroutinename='srctool_lc_get_lc_index'
    type(lc_type),    intent(in)      :: lc
    real(kind=real64), intent(in)  :: t
    integer(kind=int32), intent(out) :: i
    integer(kind=int32) :: status
    status = 0
    i = 0

    if ( lc%to_consider ) then

      ! check if t is outside bounds of lc
      if ((t < lc%ptmin(1)) .or. &
          (t >= lc%ptmax(lc%num_rows))) then
        i = 0
      else if ( lc%num_rows == 1 ) then
        i = 1
      else
        ! do binary search to find index of lc bin spanning time=t
        call binary_search(lc%ptmin(:),t,i,status)
        if ((status /= BINARY_SEARCH_STATUS_OK) .and. &
            (status /= BINARY_SEARCH_STATUS_GTMAX) ) then
          ! this should never happen
          i = -1
          call wrserr (subroutinename, ERROR_LEVEL_WARNING, &
              "Problem moving to next lcbin (code:" // trim(i2s(status)) // ")" )
          return
        endif
      endif
    endif

  end subroutine srctool_lc_get_lc_index


  recursive subroutine srctool_lc_get_next_lc_index (lc, t, i)
    character(len=*), parameter :: subroutinename='srctool_lc_get_next_lc_index'
    type(lc_type),    intent(in)      :: lc
    real(kind=real64), intent(in)  :: t
    integer(kind=int32), intent(out) :: i
    integer(kind=int32) :: status
    status = 0
    i = -1
    if ( lc%to_consider ) then

      if ( lc%num_rows > 1 ) then
        call binary_search(lc%ptmax(1:lc%num_rows), t, i, status)
        if ((status == BINARY_SEARCH_STATUS_OK ) .or. &
            (status == BINARY_SEARCH_STATUS_LTMIN )) then
          if (t > lc%ptmin(i+1))  then
            i = -1
          endif
        else
          i = -1
        endif
      else if ( t < lc%ptmax(1) ) then  ! t is before fist and only time bin
        i = 1
      end if
    endif

  end subroutine srctool_lc_get_next_lc_index

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------------- END of Lightcurve handling ------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------



end module srctool_lc
