! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

program srctool

!-----------------------------------------------------------------------------
! Summary:
! Task which extracts source and background spectra and creates an
! ARF for an eRosita source.
!
!-----------------------------------------------------------------------------
!
!-----------------------------------------------------------------------------
! Routines called:
!
!
!-----------------------------------------------------------------------------
!
!-----------------------------------------------------------------------------
! SASS environment calls:
! =======================
!
!   Parameter i/o:
!   --------------
!
!   Logging and error handling:
!   ---------------------------
!
! Environment variables:
! ----------------------
!
! SASS_PARINT      values: "pipeline"                  pipeline (SASS) mode
!                          any other value or not set: interactive (PIL) mode
!
! SASS_VERBOSITY   values: "chatty"                    output from WRCHATTY,
!                                                      WRLOG, and WRERR
!                          "sparse"                    only WRERR output
!                          any other value or not set: WRERR and WRLOG output
!-----------------------------------------------------------------------------


  use iso_fortran_env           ! replacement for numkind
  use param                     ! standard eRosita module handling parameter/logging/error interface

#ifdef _OPENMP
  use omp_lib
#endif

  use srctool_generic           ! generic module with some universal constants
  use srctool_datatypes
  use srctool_param
  use srctool_lc
  use srctool_badpix
  use srctool_cal
  use srctool_engine
  use srctool_evtlist
  use srctool_autoreg
  use srctool_test

  implicit none     ! insist on explicit typing

  character(len=*), parameter :: str_taskname='srctool'


!  character(len=*), dimension(0:MAX_INST), parameter :: str_caldb = &
!      (/ 'CALDB', 'CALDB', 'CALDB', 'CALDB', 'CALDB', 'CALDB', 'CALDB', 'CALDB' /)

  character(len=STR_MAX), dimension(0:MAX_TOT_INST) :: str_caldb
!  character(len=STR_MAX), dimension(0:MAX_TOT_INST) :: str_debugrmf

  !some useful admin vars
  integer(kind=int32) :: status, ierr, src_i, i, c, im, mtm
  integer(kind=int32) :: clock_count0, clock_rate, clock_count
  logical :: found_any

  character(len=LSTR_MAX) :: str_buffer
  integer(kind=int32) :: nfiles

  !input params structure
  type(srctool_paramlist_type)  :: paramlist

  ! list of structures which hold the info for each instrument
  type(instlist_type),target    :: instlist

  ! lc_template is used to describe the lc format
  type(lc_type)              :: lc_template

  !  this will hold the input source list (if required)
  type(srclist_type), target  :: srclist

#ifdef _OPENMP
  character(len=STR_MAX) :: str_env_num_threads
!  integer(kind=int32) :: int_env_num_threads
#endif



  type(srctool_cal_type), parameter :: caltype_list(1:7) &
      = (/ SRCTOOL_CALTYPE_DETMAP, &
      SRCTOOL_CALTYPE_INSTPAR, &
!      SRCTOOL_CALTYPE_PATTCAL, &   ! uncomment this when caldb contains PATTCAL files
      SRCTOOL_CALTYPE_ARF, &
      SRCTOOL_CALTYPE_RMF, &
      SRCTOOL_CALTYPE_EBOUNDS, &
!      SRCTOOL_CALTYPE_VIGN, &
      SRCTOOL_CALTYPE_PSF, &
!      SRCTOOL_CALTYPE_CALBADPIX, &
      SRCTOOL_CALTYPE_AVIGN /)

  character(len=STR_MAX) :: str_argval

  !-----------------------------------------------------------------------------------!
  ! init the clock
  call system_clock(clock_count0, clock_rate)
  !-----------------------------------------------------------------------------------!
  status = 0
  ierr = ERROR_LEVEL_OK


  !-----------------------------------------------------------------------------------!
  ! set the creator string for later use:
  call set_creator_string()
  !-----------------------------------------------------------------------------------!

  !-----------------------------------------------------------------------------------!
  ! check if there is a "--version" or "-v" as the first and only argument to the task
  if ( IARGC() .eq. 1 ) then ! get number of command line arguments
    call GETARG(1,str_argval)  ! get value of 1st command line argument
    if ( (trim(str_argval) .eq. "--version") .or. &
         (trim(str_argval) .eq. "-v") )  then
      write (*,'(A)') trim(CREATOR_STRING)
      stop
    endif
  endif
  !-----------------------------------------------------------------------------------!

  !-----------------------------------------------------------------------------------!
  ! initialise the random number generator
  call random_seed()
  !-----------------------------------------------------------------------------------!

#ifdef _OPENMP
  ! manage multi-threading
  ! manage the cases where the OMP_NUM_THREADS environment variable is not set
  ! want to default to OMP_NUM_THREADS=1 in these cases

  ! https://gcc.gnu.org/onlinedocs/gfortran/GET_005fENVIRONMENT_005fVARIABLE.html
  ! STATUS is -1 if VALUE is present but too short for the environment
  ! variable; it is 1 if the environment variable does not exist and 2
  ! if the processor does not support environment variables; in all
  ! other cases STATUS is zero.

  call get_environment_variable(NAME="OMP_NUM_THREADS", VALUE=str_env_num_threads, STATUS=status)
  if ( (status .eq. 1) .or. (status .eq. 2) ) then
    call wrschatty(str_taskname, "OMP_NUM_THREADS undefined, setting to 1")
    call OMP_set_num_threads(1)
  endif

!  int_env_num_threads = omp_get_max_threads()
!  call wrschatty(str_taskname, "OMP: Running with max threads: " // trim(i2s(int_env_num_threads))  )
  call wrschatty(str_taskname, &
      "OMP: Running with maximum threads= " // trim(i2s(omp_get_max_threads())))

#endif

!  stop 0

  call wrslog(str_taskname, "Initialising data structures")

  !----------------------------------------------------------------------------
  ! write some initialisation info to the main data structures
  call srctool_allocate_instlist(instlist, ierr)
  if (ierr >= ERROR_LEVEL_SEVERE) call wrstop(paramlist%unit, str_taskname, &
      "Error allocating memory")
  call srctool_empty_instlist(instlist)
  call srctool_param_empty_srclist(srclist)
  call srctool_lc_empty_lc(lc_template)
  !----------------------------------------------------------------------------

  !----------------------------------------------------------------------------
  ! open parameter interface
  call srctool_param_init_pars(paramlist,ierr)
  if (ierr >= ERROR_LEVEL_SEVERE) call wrstop(paramlist%unit, str_taskname, &
      "Error initialising input parameter handler")
  !----------------------------------------------------------------------------

  !----------------------------------------------------------------------------
  ! get input parameters from command line
  call srctool_param_read_pars(paramlist,ierr)
  if (ierr >= ERROR_LEVEL_SEVERE) call wrstop(paramlist%unit, str_taskname, &
      "Error reading input parameters")
  !----------------------------------------------------------------------------

!  write (*,'(A)') trim(sass_history)

  !----------------------------------------------------------------------------
  ! interpret the values of relevant params
  call wrslog(str_taskname, "Interpreting input parameters")
  call srctool_param_interpret_pars(paramlist, instlist, srclist, ierr)
  if (ierr >= ERROR_LEVEL_SEVERE) call wrstop(paramlist%unit, str_taskname, &
      "Error interpreting input parameters")

  !----------------------------------------------------------------------------
  ! This is where we will call the internal test harness
  if (IAND(paramlist%todo_code,TODO_CODE_TEST)   .ne. 0 ) then
    call wrslog(str_taskname, "Starting up test harness")
    call srctool_test_harness (instlist, ierr) ! TODO write some test routines
    if (ierr >= ERROR_LEVEL_SEVERE) then
      call wrstop(paramlist%unit, str_taskname, &
          "Error running internal test routines")
    endif
    ! exit gracefully
    stop 0
  endif
  !----------------------------------------------------------------------------

  !----------------------------------------------------------------------------
  ! print parameter values
  call srctool_param_print_pars(paramlist)
  !----------------------------------------------------------------------------


  !----------------------------------------------------------------------------
  ! verify params
  call srctool_param_verify_pars(paramlist,ierr)
  if (ierr >= ERROR_LEVEL_SEVERE) call wrstop(paramlist%unit, str_taskname, &
      "Error when verifying input parameters")
  !----------------------------------------------------------------------------

!  ! DEBUG
!  call srctool_regions_extract_central_coords (paramlist%srccoord, str_buffer)
!  return




  !----------------------------------------------------------------------------
  ! read in the evtlist(s) and their contained GTIn, BADPIXn, CORRATTn, DEADCORn extensions
  call wrslog(str_taskname, "Reading in eventfile(s)")
  call srctool_evtlist_read_evtlists_from_file(instlist,paramlist%evtfiles,paramlist%num_evtfiles, ierr)
  if (ierr >= ERROR_LEVEL_SEVERE) then
    call convert_string_array_to_string (paramlist%evtfiles, paramlist%num_evtfiles, str_buffer, ",")
    call wrstop(paramlist%unit, str_taskname, "Error reading input events tables: " //  trim(str_buffer))
  endif
  !----------------------------------------------------------------------------
  ! check that at least one eventlist was read for each instrument to be considered
  ! if not, emit a warning, and flag instrument to not be considered further
  do i=1,MAX_INST
    if ( instlist%inst(i)%to_consider .and. &
        (instlist%inst(i)%evtlistlist%num_evtlists .eq. 0)) then
      call wrserr(str_taskname, ERROR_LEVEL_WARNING, &
          "You included " // trim(instlist%inst(i)%name) // &
          " in the 'insts' parameter but there are zero event lists corresponding to this instrument.")
      instlist%inst(i)%to_consider = .false.
    endif

    if ( instlist%inst(i)%to_write .and. &
        (instlist%inst(i)%evtlistlist%num_evtlists == 0)) then
      call wrserr(str_taskname, ERROR_LEVEL_WARNING, &
          "You included " // trim(instlist%inst(i)%name) // &
          " in the 'writeinsts' parameter but there are zero event lists corresponding to this instrument or it is not included in 'insts'.")
      instlist%inst(i)%to_write = .false.
    end if
  end do


  ! point the pinst0 pointer at the first instrument to consider
  do im=1, NUM_MERGED_INST
    mtm = MERGED_INST(im)
    nullify(instlist%inst(mtm)%pinst0)
    found_any = .false.
    do i=1,MAX_INST
      if ( MERGED_INST_CMPT(mtm, i) .and. instlist%inst(i)%to_consider .and. &
          instlist%inst(i)%is_valid ) then
        instlist%inst(mtm)%pinst0 => instlist%inst(i)
        found_any = .true.
        exit
      end if
    end do

    ! Nothing to do!
    if ( .not. associated(instlist%inst(0)%pinst0) ) then
      call wrstop(paramlist%unit, str_taskname, "Error finding any valid instruments")
    endif

    ! Set FoV mask flag for instruments
    call srctool_set_instlist_fov_mask(instlist, ierr)
    if( ierr /= ERROR_LEVEL_OK ) then
      call wrstop(paramlist%unit, str_taskname, "Error setting FoV mask flag")
    end if

    ! fixup the pointer for merged TMs which don't have any instruments
    ! this is a bit of a hack, but a lot of downstream code relies on this
    if ( .not. found_any ) then
      instlist%inst(mtm)%pinst0 => instlist%inst(0)%pinst0
    end if

    ! don't write any output files for merged instrument if there are no valid TMs for it
    if ( instlist%inst(mtm)%to_write .and. .not. found_any ) then
      call wrserr(str_taskname, ERROR_LEVEL_WARNING, &
          "You included " // trim(instlist%inst(mtm)%name) // &
          " in the 'writeinsts' parameter but there are zero event lists corresponding to this instrument or its constituents not included in 'insts'.")
      instlist%inst(mtm)%to_consider = .false.
      instlist%inst(mtm)%to_write = .false.
    end if

  end do

  ! prepare the str_caldb - no need to read cal for instruments that we will never use
  str_caldb(:) = ""
  do i=1,MAX_INST
    if ( instlist%inst(i)%to_consider ) then
      str_caldb(i) = "CALDB"
    endif
  end do


!  str_debugrmf(:) = ""
!  do i=1,MAX_INST
!    if ( instlist%inst(i)%to_consider ) then
!      str_caldb(i) = "CALDB"
!      ! DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG
!      ! DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG
!      write(str_debugrmf(i),'(A,I0,A,I0,A)') &
!          "/home/erosita/caldb/data/erosita/fm",i,"/bcf/fm",i,"_rmf_141103v01.fits"
!      ! DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG
!      ! DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG
!    endif
!  end do


  !----------------------------------------------------------------------------
  ! now read each of the various cal types from the CALDB
  call wrslog(str_taskname, "Reading calibration files from the CALDB")
  do c=1,size(caltype_list)
    ! TODO add other exceptions depending on contents of todo parameter

    ! check if we really need PSFs
    if ( (caltype_list(c)%code   .eq. SRCTOOL_CALTYPE_PSF%code) .and. &
         (paramlist%psftype_code .eq. SRCTOOL_PSF_CLASS_NONE) ) then
      call wrschatty(str_taskname, &
          "Will skip reading cal (" // trim(caltype_list(c)%name) // ") descriptions from: CALDB")

      do i=0, MAX_TOT_INST
         instlist%inst(i)%psf%class = SRCTOOL_PSF_CLASS_NONE
         instlist%inst(i)%psf%is_valid = .true.
      end do
      cycle
    endif

    ! if we got to here then we should read in this particular caltype
    nfiles = MAX_INST
    call srctool_cal_read_cal_from_files(instlist,str_caldb(1:MAX_INST), nfiles, caltype_list(c), ierr)
    if (ierr .ge. ERROR_LEVEL_SEVERE) then
      call wrstop(paramlist%unit, str_taskname, &
          "Error reading cal (" // trim(caltype_list(c)%name) // ") descriptions from: CALDB")
    endif
  end do
  !----------------------------------------------------------------------------

  !----------------------------------------------------------------------------
  ! check if we need to compute survey PSFs for auto region handling
  if ((srclist%autoreg_code_src  .ne. AUTOREG_CODE_NULL) .or. &
      (srclist%autoreg_code_back .ne. AUTOREG_CODE_NULL) ) then
    ! TODO test if we are using survey or pointing mode data
    ! If we are working on survey mode data then we need to
    ! read in the survey PSFs, if not then we can skip this step
    call wrslog(str_taskname, "Building survey-averaged PSF model")
    call srctool_build_survey_psfs_for_instlist(instlist, ierr )
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrstop (paramlist%unit, str_taskname, "Problem building survey PSFs")
    endif
  endif
  !----------------------------------------------------------------------------



  !----------------------------------------------------------------------------
  call srctool_badpix_process_all_badpix(instlist, ierr)
  if (ierr .ne. ERROR_LEVEL_OK) then
    call wrstop(paramlist%unit, str_taskname, "Problem processing badpix info")
  endif
  !----------------------------------------------------------------------------

  !----------------------------------------------------------------------------
  call srctool_cal_report_cal_status(instlist, ierr)
  if (ierr .ne. ERROR_LEVEL_OK) then
    call wrstop(paramlist%unit, str_taskname, "Problem reporting calibration status")
  endif
  !----------------------------------------------------------------------------

  !----------------------------------------------------------------------------
  ! wait until now to interpret lc params in case we need to derive some values from input evtlists
  if (IAND(paramlist%todo_code,TODO_CODE_LC)     .ne. 0 .or. &
      IAND(paramlist%todo_code,TODO_CODE_LCCORR) .ne. 0 ) then
    call wrslog(str_taskname, "Setting up lightcurve data structures according to input parameters")
    call srctool_param_interpret_lc_pars(paramlist, lc_template, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrstop(paramlist%unit, str_taskname, "Error interpreting lc parameters")
    endif
  endif
  !----------------------------------------------------------------------------


  !----------------------------------------------------------------------------
  call srctool_setup_intervallist(instlist%intervallist, instlist, paramlist%tstep, ierr)
  if ( ierr .ne. ERROR_LEVEL_OK ) then
    call wrstop(paramlist%unit, str_taskname, "Error setting up intervallist")
  end if
  !----------------------------------------------------------------------------

  !----------------------------------------------------------------------------
  if ( (srclist%n_extended_to_extract .gt. 0) .or. &
       (srclist%n_extended_to_exclude .gt. 0) ) then
    call wrslog(str_taskname, "Preparing for the handling of extended sources")
    call srctool_setup_extended_source_handling(instlist, srclist, paramlist%psftype_code, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrstop(paramlist%unit, str_taskname, "Error setting up extended source handling")
    end if
  endif
  !----------------------------------------------------------------------------


  !----------------------------------------------------------------------------
  ! check if we need to generate the source or background regions automatically
  ! Must be done after preparing DDTs of PSF maps and extent models etc
  if ((srclist%autoreg_code_src  .ne. AUTOREG_CODE_NULL) .or. &
      (srclist%autoreg_code_back .ne. AUTOREG_CODE_NULL) ) then
    call wrslog(str_taskname, "Generating extraction regions automatically from input source catalogue")
    call srctool_autoreg_generate(instlist, srclist,  ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrstop(paramlist%unit, str_taskname, "Error when automatically generating extraction regions")
    endif
  end if
  !----------------------------------------------------------------------------

  if (srclist%num_srcs <= 0) then
     call wrslog(str_taskname, "WARNING: No sources in source list")
  end if

  !----------------------------------------------------------------------------
  call system_clock(clock_count)
  write (str_buffer, '(A,A,F12.5,A)') PROCTIME_PREFIX, "Setup completed: Total setup time was     ", &
      real((clock_count - clock_count0), real32)/real(clock_rate, real32), "s"
  call wrslog(str_taskname,str_buffer )
  call wrslog(str_taskname,"------------------------------------------------------------------------------")
  call wrslog(str_taskname, "" )
  call wrslog(str_taskname,"------------------------------------------------------------------------------")

  !----------------------------------------------------------------------------

  !----------------------------------------------------------------------------
  !----------------------------------------------------------------------------
  !----------------------------------------------------------------------------
  !------------------------- Source loop --------------------------------------
  !----------------------------------------------------------------------------
  !----------------------------------------------------------------------------
  ! now step through the various requested tasks for each input source
  call wrslog(str_taskname, &
      "Source products will be extracted for " // trim(i2s(srclist%n_to_extract)) // " source" // N2S(srclist%n_to_extract) )
  if ( srclist%num_srcs .gt. 1 ) then
    call wrslog(str_taskname, "Number of sources in sourcelist:       " // trim(i2s(srclist%num_srcs)) )
    call wrslog(str_taskname, "Number of sources to extract:          " // trim(i2s(srclist%n_to_extract)) )
    call wrslog(str_taskname, "Number of sources to exclude:          " // trim(i2s(srclist%n_to_exclude))  )
    call wrslog(str_taskname, "Number of extended sources:            " // trim(i2s(srclist%n_extended)) )
    call wrslog(str_taskname, "Number of extended sources to extract: " // trim(i2s(srclist%n_extended_to_extract)) )
    call wrslog(str_taskname, "Number of extended sources to exclude: " // trim(i2s(srclist%n_extended_to_exclude)) )
  endif

  do src_i=1,srclist%num_srcs
    if ( srclist%psrc(src_i)%cat_extract ) then
      call wrslog(str_taskname, "Working on source " // trim(i2s(src_i)) // "/" // trim(i2s(srclist%num_srcs)))
      call srctool_do_all_processing_for_source (instlist, srclist%psrc(src_i), lc_template, paramlist, ierr)

      if (ierr /= ERROR_LEVEL_OK) then
         if( .not. paramlist%continue_on_error ) then
            call wrstop(paramlist%unit, str_taskname, "Problem when working on source " // trim(i2s(src_i)) )
         else
            call wrserr(str_taskname, ERROR_LEVEL_WARNING, &
                 "Problem when working on source " // trim(i2s(src_i)) // ". IGNORING ERROR AND CONTINUING.")
            ierr = ERROR_LEVEL_OK
         end if
      end if
    else
      call wrslog(str_taskname, "Skipping source   " // trim(i2s(src_i)) // "/" // trim(i2s(srclist%num_srcs)))
    end if
  end do
  !----------------------------------------------------------------------------
  !----------------------------------------------------------------------------
  !------------------------- End Source loop ----------------------------------
  !----------------------------------------------------------------------------
  !----------------------------------------------------------------------------
  !----------------------------------------------------------------------------



  !----------------------------------------------------------------------------
  ! write output source regions
  if (paramlist%out_srcreg /= "none" .and. paramlist%out_srcreg /= "") then
    call wrslog(str_taskname, "Writing source regions to " // trim(paramlist%out_srcreg) )

    call srctool_write_regions(paramlist%out_srcreg, srclist, paramlist%clobber, .true., ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrstop(paramlist%unit, str_taskname, "Error when writing source regions to a file")
    end if
  end if
  ! write output background regions
  if (paramlist%out_backreg /= "none" .and. paramlist%out_backreg /= "") then
    call wrslog(str_taskname, "Writing background regions to " // trim(paramlist%out_backreg) )

    call srctool_write_regions(paramlist%out_backreg, srclist, paramlist%clobber, .false., ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrstop(paramlist%unit, str_taskname, "Error when writing background regions to a file")
    end if
  end if
  !----------------------------------------------------------------------------


  !free the memory in the main data arrays
  call srctool_empty_instlist(instlist)
  call srctool_deallocate_instlist(instlist)
  call srctool_param_empty_srclist(srclist)
  call srctool_lc_empty_lc(lc_template)

  call wrslog(str_taskname,"------------------------------------------------------------------------------")
  call system_clock(clock_count)
  write (str_buffer, '(A,A,F12.5,A)') PROCTIME_PREFIX, "Task completed: Total processing time was ", &
      real((clock_count - clock_count0), real32)/real(clock_rate, real32), "s"
  call wrslog(str_taskname, str_buffer)
  call wrslog(str_taskname,"------------------------------------------------------------------------------")

  ! close parameter interface
  if (ierr < ERROR_LEVEL_SEVERE) then
    call cpast(str_taskname, paramlist%unit, status, ierr)
  endif

end program srctool


! useful command for checking subroutines in a file:
! set STEM = "srctool" ; set F90FILE = ${STEM}_mod.f90 ; gawk '$0~/^module/ {print "  \! subroutine declarations and interfaces for module:", $2} $1~/^end$/ && $2~/^subroutine$/ {print "  private ::",$3}' $F90FILE ; egrep "call.*${STEM}" *.f90 | grep -v $F90FILE | gawk -v thisone="$F90FILE" '$1!~thisone {split($0,a,"call");split(a[2],b,"("); n[b[1]]++; } END { for(i in n){print "  public ::", i}}'



! egrep -c 'subroutine declarations and interfaces' *.f90

! These generic calls should be gradually inserted to replace current code:
!
!  if ( fitswrap_check_status (subroutinename, status, ierr)) return
!
!  if ( check_allocate_status (subroutinename, status, ierr)) return

! this gets rid of spurious commas after write statements that cause
! lots of warnings when compiling on mac clang complier.
! some minor editing required afterwards to catch a few oddities
! foreach FILE ( *.f90 )
!   cp -pv ${FILE} ${FILE}.old
!   gawk '$1~/^write/{i=match($0,"),");j=match($0,"TRIM|trim");if(i>0&&(i<j||j==0)){print substr($0,1,i),substr($0,i+3)}else{print $0}} $1!~/^write/{print $0}' ${FILE}.old > $FILE
! end
