! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_owner

  !-------------------------------------------------------------------------------
  ! Summary: Handle ownership of events
  !-------------------------------------------------------------------------------

  use iso_fortran_env
  use ero_general
  use srctool_datatypes

  implicit none
  character(len=*), private, parameter :: modulename='SRCTOOL_OWNER_MOD'

  public

  public :: srctool_owner_setup_owner
  public :: srctool_owner_applyflag_owner
  public :: srctool_owner_getkeyword_owner
  public :: srctool_owner_getfilenamecode_owner
  public :: srctool_owner_warnon_owner
  public :: srctool_owner_modify_src_prefix

contains

  !-------------------------------------------------------------------------------
  ! initialise ownership data type
  !-------------------------------------------------------------------------------
  recursive subroutine srctool_owner_setup_owner(owner)
    type(owner_type), intent(inout) :: owner

    owner%mpe = SRCTOOL_OWNER_UNSET
    owner%iki = SRCTOOL_OWNER_UNSET
  end subroutine srctool_owner_setup_owner

  !-------------------------------------------------------------------------------
  ! take event flag and update ownership structure
  !-------------------------------------------------------------------------------
  recursive subroutine srctool_owner_applyflag_owner(owner, flag)
    type(owner_type), intent(inout) :: owner
    integer(kind=int32), intent(in) :: flag

    logical :: evt_owner_mpe, evt_owner_iki

    ! is this event an MPE one?
    evt_owner_mpe = IAND(flag, FLAG_MPE_OWNER)==0
    if(owner%mpe==SRCTOOL_OWNER_UNSET) then
       ! we don't know yet what the ownership is
       if(evt_owner_mpe) then
          owner%mpe=SRCTOOL_OWNER_YES  ! MPE
       else
          owner%mpe=SRCTOOL_OWNER_NO   ! not MPE
       end if
    else
       ! if is an MPE event but the previous event was not MPE, set as mixed
       if(evt_owner_mpe .and. owner%mpe==SRCTOOL_OWNER_NO) &
            owner%mpe=SRCTOOL_OWNER_MIXED
       ! if is not an MPE event, but the previous event was MPE, set as mixed
       if(.not. evt_owner_mpe .and. owner%mpe==SRCTOOL_OWNER_YES) &
            owner%mpe=SRCTOOL_OWNER_MIXED
    end if

    ! is this event an IKI one?
    evt_owner_iki = IAND(flag, FLAG_IKI_OWNER)==0
    if(owner%iki==SRCTOOL_OWNER_UNSET) then
       ! we don't know yet what the ownership is
       if(evt_owner_iki) then
          owner%iki=SRCTOOL_OWNER_YES  ! IKI
       else
          owner%iki=SRCTOOL_OWNER_NO   ! not IKI
       end if
    else
       ! if is an IKI event but the previous event was not IKI, set as mixed
       if(evt_owner_iki .and. owner%iki==SRCTOOL_OWNER_NO) &
            owner%iki=SRCTOOL_OWNER_MIXED
       ! if is not an IKI event, but the previous event was IKI, set as mixed
       if(.not. evt_owner_iki .and. owner%iki==SRCTOOL_OWNER_YES) &
            owner%iki=SRCTOOL_OWNER_MIXED
    end if

  end subroutine srctool_owner_applyflag_owner

  !-------------------------------------------------------------------------------
  ! get the text to place in a keyword head according to the ownership
  !-------------------------------------------------------------------------------
  recursive function srctool_owner_getkeyword_owner(owner) result(res)
    type(owner_type), intent(in) :: owner
    character(len=8) :: res

    if      (owner%mpe==SRCTOOL_OWNER_YES   .and. owner%iki==SRCTOOL_OWNER_NO   ) then
       res="MPE"
    else if (owner%mpe==SRCTOOL_OWNER_NO    .and. owner%iki==SRCTOOL_OWNER_YES  ) then
       res="IKI"
    else if (owner%mpe==SRCTOOL_OWNER_YES   .and. owner%iki==SRCTOOL_OWNER_YES  ) then
       res="IKI+MPE"
    else if (owner%mpe==SRCTOOL_OWNER_NO    .and. owner%iki==SRCTOOL_OWNER_NO   ) then
       ! no ownership by anybody
       res="INVALID"
    else if (owner%mpe==SRCTOOL_OWNER_UNSET .and. owner%iki==SRCTOOL_OWNER_UNSET) then
       ! no events
       res="INVALID"
    else if (owner%mpe==SRCTOOL_OWNER_MIXED .or.  owner%iki==SRCTOOL_OWNER_MIXED) then
       ! mixed ownership
       res="MIXED"
    else
       ! bug!
       res="INVALID"
    end if

  end function srctool_owner_getkeyword_owner


  !-------------------------------------------------------------------------------
  ! get the text to place in a keyword head according to the ownership
  !-------------------------------------------------------------------------------
  recursive function srctool_owner_getfilenamecode_owner(owner) result(res)
    type(owner_type), intent(in) :: owner
    character(len=1) :: res

    if      (owner%mpe==SRCTOOL_OWNER_YES   .and. owner%iki==SRCTOOL_OWNER_NO   ) then
       res="m"
    else if (owner%mpe==SRCTOOL_OWNER_NO    .and. owner%iki==SRCTOOL_OWNER_YES  ) then
       res="i"
    else if (owner%mpe==SRCTOOL_OWNER_YES   .and. owner%iki==SRCTOOL_OWNER_YES  ) then
       res="b"
    else if (owner%mpe==SRCTOOL_OWNER_NO    .and. owner%iki==SRCTOOL_OWNER_NO   ) then
       ! no ownership by anybody
       res="o"
    else if (owner%mpe==SRCTOOL_OWNER_UNSET .and. owner%iki==SRCTOOL_OWNER_UNSET) then
       ! no events
       res="o"
    else if (owner%mpe==SRCTOOL_OWNER_MIXED .or.  owner%iki==SRCTOOL_OWNER_MIXED) then
       ! mixed ownership
       res="o"
    else
       ! bug!
       res="o"
    end if

  end function srctool_owner_getfilenamecode_owner


  !-------------------------------------------------------------------------------
  ! print out warning if the ownership information has a problem
  !-------------------------------------------------------------------------------
  recursive subroutine srctool_owner_warnon_owner(owner)
    character(len=*), parameter :: subroutinename='srctool_owner_warnon_owner'
    type(owner_type), intent(in) :: owner
    integer(kind=int32) :: ierr

    ierr=0
    if      (owner%mpe==SRCTOOL_OWNER_YES   .and. owner%iki==SRCTOOL_OWNER_NO   ) then
       ! ok: MPE
    else if (owner%mpe==SRCTOOL_OWNER_NO    .and. owner%iki==SRCTOOL_OWNER_YES  ) then
       ! ok: IKI
    else if (owner%mpe==SRCTOOL_OWNER_YES   .and. owner%iki==SRCTOOL_OWNER_YES  ) then
       ! ok IKI+MPE
    else if (owner%mpe==SRCTOOL_OWNER_NO    .and. owner%iki==SRCTOOL_OWNER_NO   ) then
       ! no ownership by anybody
       ierr = ERROR_LEVEL_WARNING
       call wrserr(subroutinename, ierr, "Events were neither owned by MPE nor IKI")
    else if (owner%mpe==SRCTOOL_OWNER_UNSET .and. owner%iki==SRCTOOL_OWNER_UNSET) then
       ! no events
       ierr = ERROR_LEVEL_WARNING
       call wrserr(subroutinename, ierr, "No events, so ownership could not be determined")
    else if (owner%mpe==SRCTOOL_OWNER_MIXED .or.  owner%iki==SRCTOOL_OWNER_MIXED) then
       ! mixed ownership
       ierr = ERROR_LEVEL_WARNING
       call wrserr(subroutinename, ierr, "Inconsistent or mixed ownership of events by MPE and IKI")
    else
       ! bug!
       ierr = ERROR_LEVEL_SEVERE
       call wrserr(subroutinename, ierr, "Internal error finding event ownership")
    end if

  end subroutine srctool_owner_warnon_owner


  !-------------------------------------------------------------------------------
  ! for a source type, update the prefix according to the source owner
  ! if in pipeline mode (or testpipe for testing)
  !-------------------------------------------------------------------------------
  recursive subroutine srctool_owner_modify_src_prefix(src)
    type(src_type), intent(inout) :: src

    character(len=8) :: environ
    integer :: length, status

    call get_environment_variable("SASS_PARINT", environ, length, status)
    if (status==0 .and. (environ=="pipeline" .or. environ=="PIPELINE" .or. &
        environ=="testpipe")) then
      ! update prefix if in pipeline mode with owner

      if( src%file_prefix(2:2) == "b" ) then
        src%file_prefix(2:2) = srctool_owner_getfilenamecode_owner(src%owner)
      end if

    end if

  end subroutine srctool_owner_modify_src_prefix


end module srctool_owner
