! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

module srctool_pattcal

  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling pattern calibration data
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         !
  use srctool_misc              !
  use srctool_generic_fitsraw

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_PATTCAL_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_pattcal
  public ::  srctool_pattcal_read_pattcal_from_chdu
  public ::  srctool_pattcal_empty_pattcal


contains

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- START of PATTCAL handling ---------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_pattcal_empty_pattcal(pattcal)
    type(pattcal_type), intent(inout)      :: pattcal
    if ( allocated(pattcal%pelo))   deallocate (pattcal%pelo)
    if ( allocated(pattcal%pehi))   deallocate (pattcal%pehi)
    if ( allocated(pattcal%ps))     deallocate (pattcal%ps)
    if ( allocated(pattcal%pd))     deallocate (pattcal%pd)
    if ( allocated(pattcal%pt))     deallocate (pattcal%pt)
    if ( allocated(pattcal%pq))     deallocate (pattcal%pq)

    pattcal%num_rows = 0
    pattcal%is_valid = .false.
    call fitswrap_init_hduinfo(pattcal%hduinfo)

  end subroutine srctool_pattcal_empty_pattcal
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !read the pattcal from the current hdu of the fits file identified by unit
  subroutine srctool_pattcal_read_pattcal_from_chdu(unit, pattcal, ierr)
    character(len=*), parameter :: subroutinename='srctool_pattcal_read_pattcal_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(pattcal_type), intent(inout) :: pattcal
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status
!    character(len=STR_MAX) :: str_buffer

    integer(kind=int32) :: colnum_elo
    integer(kind=int32) :: colnum_ehi
    integer(kind=int32) :: colnum_s
    integer(kind=int32) :: colnum_d
    integer(kind=int32) :: colnum_t
    integer(kind=int32) :: colnum_q
    integer(kind=int32) :: nrows_per_chunk
    integer(kind=int32) :: nrows_to_read
    integer(kind=int32) :: row
    integer(kind=int32) :: row_end
    integer(kind=int32) :: anyf
    real(kind=real32)  , parameter :: nullval_real32  = 0.0

    ierr = ERROR_LEVEL_OK
    status = 0
!    call wrschatty (subroutinename, "Reading in PATTCAL extension from CHDU")

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, pattcal%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if
    !- check the following keader keywords are present
    !- HDUCLASS       'OGIP'
    !- HDUCLAS1
    !- HDUCLAS2

    !get the number of rows in the extension
    call FTGNRW_w(unit, pattcal%num_rows, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    !check that have at least some entries and then read in the column data
    if ( pattcal%num_rows .gt. 0 ) then
      call wrschatty (subroutinename, &
          "Reading in: " // trim(i2s(pattcal%num_rows)) // " rows of PATTCAL info from extname='" // &
          trim(pattcal%hduinfo%key_extname%value) //  "' instrument='" // &
          trim(pattcal%hduinfo%key_instrume%value) // "'")

      call fitswrap_prep_col(unit, "ENERG_LO", pattcal%num_rows, colnum_elo,  pattcal%pelo,  ierr)
      call fitswrap_prep_col(unit, "ENERG_HI", pattcal%num_rows, colnum_ehi,  pattcal%pehi,  ierr)
      call fitswrap_prep_col(unit, "S",        pattcal%num_rows, colnum_s,    pattcal%ps,    ierr)
      call fitswrap_prep_col(unit, "D",        pattcal%num_rows, colnum_d,    pattcal%pd,    ierr)
      call fitswrap_prep_col(unit, "T",        pattcal%num_rows, colnum_t,    pattcal%pt,    ierr)
      call fitswrap_prep_col(unit, "Q",        pattcal%num_rows, colnum_q,    pattcal%pq,    ierr)
      if (ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem preparing column(s) for reading from HDU")
        return
      end if

      ! get the chunksize
      call ftgrsz_w(unit, nrows_per_chunk, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      do row = 1, pattcal%num_rows, nrows_per_chunk
        nrows_to_read = min(nrows_per_chunk,pattcal%num_rows-row+1)
        row_end = row+nrows_to_read-1
        call FTGCVE_w(unit, colnum_elo,   row,1,nrows_to_read,nullval_real32,pattcal%pelo(row:row_end), anyf,status)
        call FTGCVE_w(unit, colnum_ehi,   row,1,nrows_to_read,nullval_real32,pattcal%pehi(row:row_end), anyf,status)
        call FTGCVE_w(unit, colnum_s,     row,1,nrows_to_read,nullval_real32,pattcal%ps(row:row_end),   anyf,status)
        call FTGCVE_w(unit, colnum_d,     row,1,nrows_to_read,nullval_real32,pattcal%pd(row:row_end),   anyf,status)
        call FTGCVE_w(unit, colnum_t,     row,1,nrows_to_read,nullval_real32,pattcal%pt(row:row_end),   anyf,status)
        call FTGCVE_w(unit, colnum_q,     row,1,nrows_to_read,nullval_real32,pattcal%pq(row:row_end),   anyf,status)

        if (status .ne. 0) then
          call fitswrap_write_fits_error(subroutinename, status)
          ierr = ERROR_LEVEL_SEVERE
          return
        end if
      end do

    end if

    pattcal%is_valid = .true.
  end subroutine srctool_pattcal_read_pattcal_from_chdu
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- END of PATTCAL handling ---------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

end module srctool_pattcal
