  !---------------------------------------------------------------------------------
  ! Combined merge and insertion sort
  ! This is a very simplified form of the "timsort" algorithm
  !
  ! This "argsort" variant modifies an array of indexes into the original,
  ! but leaves the original untouched.

    DATATYPE, dimension(:), intent(in) :: data
    integer(kind=int32), dimension(:), intent(out) :: idxs

    integer(kind=int32) :: num, chunksize, startp, endp, ssize, leftp, midp, rightp, i
    integer(kind=int32), dimension(:), allocatable :: tmp_left, tmp_right

    ! For testing
    ! DATATYPE :: temp

    num = size(data)
    do i=1, num
      idxs(i) = i
    end do

    if ( num < 2 ) then
       return
    end if

    allocate( tmp_left(num), tmp_right(num) )

    ! initial chunk size
    chunksize = find_chunksize(num)

    ! run insertion sort on smaller chunks
    do startp=1, num, chunksize
      endp = min(startp+chunksize, num)
      call do_insertion_sort(startp, endp)
    end do

    ! repeatedly run merge sort on the chunks, doubling the size each time
    ssize = chunksize
    do while( ssize < num )
      do leftp=1, num, 2*ssize
        midp = min(num, leftp+ssize-1)
        rightp = min(leftp+2*ssize-1, num)

        call do_merge(leftp, midp, rightp)
      end do

      ssize = 2 * ssize
    end do

    ! For testing
    ! temp = data(idxs(1))
    ! do i=2, num
    !   if(data(idxs(i)) < temp ) stop
    !   temp = data(idxs(i))
    ! end do

    deallocate( tmp_left, tmp_right )

  contains

    !----------------------------------------------------------------------
    ! do an insertion sort on the array between leftp and rightp
    recursive subroutine do_insertion_sort(leftp, rightp)
      integer(kind=int32), intent(in) :: leftp, rightp

      integer(kind=int32) :: i, j
      integer(kind=int32) :: tmp

      do i=leftp+1, rightp
        tmp = idxs(i)
        j = i-1
        do
          if( j<leftp ) exit
          if( data(tmp)>=data(idxs(j)) ) exit
          idxs(j+1) = idxs(j)
          j = j-1
        end do
        idxs(j+1) = tmp
      end do

    end subroutine do_insertion_sort

    !----------------------------------------------------------------------
    ! do a merge sort of leftp->midp and midp->rightp
    recursive subroutine do_merge(leftp, midp, rightp)

      integer(kind=int32), intent(in) :: leftp, midp, rightp
      integer(kind=int32) :: len1, len2, i, j, k

      len1 = midp - leftp + 1
      len2 = rightp - midp

      tmp_left(:len1) = idxs(leftp:leftp+len1-1)
      tmp_right(:len2) = idxs(midp+1:midp+1+len2-1)

      i = 1
      j = 1
      k = leftp

      do while ( j <= len2 .and. i <= len1 )
        if( data(tmp_left(i)) <= data(tmp_right(j)) ) then
          idxs(k) = tmp_left(i)
          i = i + 1
        else
          idxs(k) = tmp_right(j)
          j = j + 1
        end if
        k = k + 1
      end do

      do while ( i <= len1 )
        idxs(k) = tmp_left(i)
        k = k + 1
        i = i + 1
      end do

      do while ( j <= len2 )
        idxs(k) = tmp_right(j)
        k = k + 1
        j = j + 1
      end do
      
    end subroutine do_merge

    !----------------------------------------------------------------------
    ! find optimal chunck size so number of chunks is close to power of two
    recursive function find_chunksize(n_in) result(x)
      integer(kind=int32), intent(in) :: n_in
      integer(kind=int32) :: n, x, q
      integer(kind=int32), parameter :: MINIMUM_RUN=64

      n = n_in
      q = 0
      do while ( n >= MINIMUM_RUN )
        q = ior(q, iand(n, 1))
        n = n / 2
      end do

      x = n + q
    end function find_chunksize
