#!/bin/tcsh -f 

# Test routines for srctool

# Setup the output prefix:
alias echoe 'echo "`date`: ${0}: ##ERROR##  " \!* '
alias echoc 'echo "`date`: ${0}: ##COMMENT##" \!* '

set DIVIDER = "###----------------------------------------------------------------###"

set DEBUG_MODE = "FALSE"
set VALGRIND_MODE = "FALSE"
set VALGRIND_PREFIX = ""
if ( $#argv == 1 ) then
  if ( "$argv[1]" == "DEBUG" || "$argv[1]" == "debug" ) then
    set DEBUG_MODE = "TRUE"
    echoc "Running in debug mode"
  else if ( "$argv[1]" == "MEMCHECK" || "$argv[1]" == "memcheck" ) then
    set VALGRIND_PREFIX = "valgrind --tool=memcheck -v --leak-check=full --track-origins=yes --max-stackframe=512000000 "
    set VALGRIND_MODE = "TRUE"
    set DEBUG_MODE = "TRUE"
    echoc "Running in debug mode and with valgrind memcheck"
    if ( ! -X valgrind ) then
      echoe "You need to have valgrind in your PATH to run in this mode"
      exit 1
    endif
  else
    echoe "Unknown argument: $argv[1] "
    exit 1
  endif
endif
 

set TASKNAME = "srctool"

set RETVAL = 0

####################################################
######  Basic installation checks ##################
####################################################

####################################################
# check if SASS environment is set up:
if ( $?SASS_ROOT == 0 || $?SASS_DIR == 0 || $?SASS_HOME == 0 || $?SASS_TESTDATA == 0 ) then
  echoe "You must first run the eSASS setup script"
  @ RETVAL ++
endif
####################################################


####################################################
# check if the executable exists and is in the path:
if ( ! -X ${TASKNAME} ) then
  echoe "Couldn't find ${TASKNAME} executable in path"
  @ RETVAL ++
endif
####################################################


####################################################
# check if necessary helper apps exist:
foreach HELPER_TASK ( awk plist ) 
  if ( ! -X  ${HELPER_TASK}  ) then
    echoe "Couldn't find ${HELPER_TASK} in your path"
    @ RETVAL ++
  endif
end
####################################################


####################################################
# check that the parameter file can be found
if ( "`plist ${TASKNAME} |& cat`" == 'plist: unable to initialize tool/parameter file "${TASKNAME}".' ) then
  echoe "Couldn't initialise input parameters for ${TASKNAME}"
  @ RETVAL ++
endif
####################################################



####################################################
# Check if we need to give up already
if ( $RETVAL != 0 ) then
  echoe "Failed basic installation checks"
  exit  $RETVAL
endif
####################################################
####################################################
####################################################



####################################################
######  Internal functionality checks ##############
####################################################

####################################################
# run the internal test harness
${TASKNAME} eventfiles="" srccoord="" todo=TEST >& /dev/null
if ( $? > 0 ) then
  echoe "${TASKNAME}'s internal test harness returned with an error"
  @ RETVAL ++
endif


####################################################
# Check if we need to give up already
if ( $RETVAL != 0 ) then
  echoe "Failed internal functionality checks"
  exit  $RETVAL
endif
####################################################
####################################################
####################################################


####################################################
######  Basic functionality checks #################
####################################################
# Now test the basic functionality of ${TASKNAME}
# Look at outputs produced from a 

# Add some checks here.

###################################################################
###################################################################
###################################################################
# Run all of the usage examples from the documentation web page
set DOC_HTML = "${SASS_ROOT}/erosita/task/${TASKNAME}/${TASKNAME}_doc.html" 
#set DOC_HTML = ~/working/SASS_devel/task/${TASKNAME}/${TASKNAME}_doc.html
if ( ! -e "${DOC_HTML}" ) then
  echoe "Could not find task documentation file: $DOC_HTML"
  exit 1
endif

if ( ! $?SASS_TESTDATA ) then
  setenv SASS_TESTDATA /home/erosita/sw/eSASSdevel/sass/testdata
  echoe "Environment variable SASS_TESTDATA giving eSASS test data directory was not set, assuming: $SASS_TESTDATA"
endif

if ( ! -d "${SASS_TESTDATA}" ) then
  echoe "Could not find eSASS test data directory: $SASS_TESTDATA"
  exit 1
endif

set TEST_OUTDIR = "/tmp/${USER}/test_eSASS/${TASKNAME}_${HOST:r:r:r}_$$"
mkdir -p $TEST_OUTDIR
if ( ! -d $TEST_OUTDIR ) then
  echoe "Could not create a new test output directory: $TEST_OUTDIR"
  exit 1
endif


set EXAMPLE_LIST = (`awk '$1=="<pre" && $0~/name="example/ {split($0,a,"name=\"");split(a[2],aa,"\""); print aa[1]}' $DOC_HTML` ) 
foreach EXAMPLE ( $EXAMPLE_LIST )
  rm -r ${TEST_OUTDIR}/srctool_${EXAMPLE} >& /dev/null

  set SCRIPT_FILE =  ${TEST_OUTDIR}/run_srctool_${EXAMPLE}.csh
  echo "\n\n${DIVIDER}"
  echo "Running usage example: $EXAMPLE"
  echo "#\!/bin/tcsh -f\
cd ${TEST_OUTDIR}" > ${SCRIPT_FILE}

  if ( "${VALGRIND_MODE}" == "TRUE" ) then
    echo 'alias srctool '"'"${VALGRIND_PREFIX} srctool"'" >> ${SCRIPT_FILE}
  endif

  awk -v example="${EXAMPLE}" '$1~/<\/pre>/{flag=0}\
flag==1 {\
  gsub("&quot;","\"",$0);\
  gsub("eventfiles=\"","eventfiles=\"${SASS_TESTDATA}/",$0);\
  gsub("extmap=\"","extmap=\"${SASS_TESTDATA}/",$0);\
  gsub("srccoord=\"test_","srccoord=\"${SASS_TESTDATA}/test_",$0); \
  gsub("tstep=0.05","tstep=1.0",$0); \
  print $0;\
} \
$1=="<pre" && $0~/name="example/ && $0~example {flag=1}' $DOC_HTML >>  ${SCRIPT_FILE}

  chmod u+x  ${SCRIPT_FILE}
  ${SCRIPT_FILE} >& ${TEST_OUTDIR}/run_srctool_${EXAMPLE}.log

  if ( $? > 0 ) then
    echoe "Usage example ${EXAMPLE} returned with an error"
    echoe "Command was: " 
    echo "\n${DIVIDER}"
    cat  ${SCRIPT_FILE}
    echo "${DIVIDER}\n"
    echoe "Output directory: $TEST_OUTDIR"
    echoe "Detailed log here (and below): ${TEST_OUTDIR}/run_srctool_${EXAMPLE}.log"
    echo "\n${DIVIDER}"
    cat ${TEST_OUTDIR}/run_srctool_${EXAMPLE}.log
    echo "${DIVIDER}\n"
    exit  1
    @ RETVAL ++
  else
    # Remove the output directory quietly
    if ( "$DEBUG_MODE" == "TRUE" ) then
      echo "${DIVIDER}\n"
      echoc "Output directory: $TEST_OUTDIR"
      echoc "Script file: ${SCRIPT_FILE}"
      echoc "Detailed log here: ${TEST_OUTDIR}/run_srctool_${EXAMPLE}.log"
      echo "\n${DIVIDER}"
    else
      rm -r ${TEST_OUTDIR}/srctool_${EXAMPLE} >& /dev/null
    endif
  endif
  echo "${DIVIDER}\n"
end
echo "All usage tests ran successfully"

if ( "$DEBUG_MODE" != "TRUE" ) then
  rm -r ${TEST_OUTDIR}/
endif

###################################################################
###################################################################
###################################################################



####################################################
# Check if we need to give up already
if ( $RETVAL != 0 ) then
  echoe "Failed basic functionality checks"
  exit  $RETVAL
endif
####################################################
####################################################
####################################################



####################################################
exit $RETVAL
####################################################
