#!/usr/bin/perl -w
#
# vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
# Copyright (C) 2019-2021:
#   Max Planck Institut fuer extraterrestrische Physik (MPE)
#   Leibniz Institute for Astrophysics Potsdam (AIP)
#
# This file is part of eROSITA's Science Analysis Software System
# (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
# USA, or see https://www.gnu.org/licenses/.
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Description: this is designed to be run from a Makefile which will
# recurse through a list of subdirectories, running a subsiduary
# Makefile in each. We assume the subdirectories are packages which have
# an interdependency relation encoded in one or more *.libs files in
# each subdirectory, where the *.libs file for task A lists the packages
# that task A depends on. The purpose of the present script is to output
# a list of packages in correct build order, such that all the libraries
# needed by any task A have been built earlier in the sequence.

  require 5.10.0;

  use English;
  use strict;

  my (%temp_hash, @fields, @depends);

  my @package_list = ();

  if (!defined($ARGV[0])) {
my_exit(\@package_list, 0);
  }

  my $base_dir = $ARGV[0]; # something like '$SASS_ROOT/erosita/task'.

  my $do_test = 0;
  my %omit_package;
  if (defined($ARGV[1])) { # if present, it is expected to be a string
                           # containing a list of subdirectory names to omit.

    $do_test = 1; # regardless of what the 2nd argument is.

    my $omit_str = $ARGV[1]; # e.g. 'expmerge telgti sim_att'
    @fields = split(/\s+/, $omit_str);
    foreach my $field (@fields){
      $omit_package{"$field"} = 1; # value is just a dummy. We use a
      # hash to avoid looping many times through a list.
    }
  }

  # Make a hash, the keys of which are the names of all immediate
  # subdirs that contain a Makefile. The values will each be a hash
  # reference, with the keys 'done' (value 0 or 1), 'listed' (value
  # also 0 or 1), and 'depends on', the latter value being a reference
  # to the list of dependency names compiled from all the *.libs files
  # in the subdirectory.
  #
  my @subdir_list = `ls -1 $base_dir`;
  my %package_hash;

  foreach my $subdir (@subdir_list){
    chomp($subdir);
  next if(!-d "$base_dir/$subdir" || !-e "$base_dir/$subdir/Makefile" || defined($omit_package{"$subdir"}));
    my @libs_files = `ls -1 $base_dir/$subdir/*.libs 2>/dev/null`;
    # the redirect of stderr is to prevent things like 'ls: cannot
    # access blah/*.libs: No such file or directory' appearing in @libs_files.

    $package_hash{"$subdir"} = {
      "done" => 0, "listed" => 0, "depends on" => get_depends(\@libs_files, $do_test, \%omit_package),
    };
  }

  # Check that all 'depends on' names exist as keys of package_hash,
  # and that no package depends on itself.
  #
  foreach my $package (keys(%package_hash)){
    %temp_hash = %{$package_hash{"$package"}};
    @depends = @{$temp_hash{"depends on"}};
    foreach my $depend (@depends){
      if(!defined($package_hash{"$depend"})){
        if($do_test){
die "Depend entry '$depend' is not found in the list of packages.\n";
        }else{
exit 0;
        }
      }

#*** better to do this inside get_depends().
      if($depend eq $package){
        if($do_test){
die "Package $package depends on itself!\n";
        }else{
exit 0;
        }
      }
    }
  }

  # Compile a list of the subdir names in dependency order.
  #
  my @build_list = ();
  my %build_hash;
  my $all_packages_are_done;
  do{
    $all_packages_are_done = 1;
    foreach my $package (keys(%package_hash)){
      %temp_hash = %{$package_hash{"$package"}};
      if(!$package_hash{"$package"}{"done"}){
        $all_packages_are_done = 0;

        # To decide whether we can now process the present package, we
        # check to see if all the packages mentioned in its 'depends on'
        # are already in the build_list.
        #
        @depends = @{$temp_hash{"depends on"}};
        my $all_depends_already_built = 1; # default
        foreach my $depend (@depends){
          if(!defined($build_hash{"$depend"})){
            $all_depends_already_built = 0;
        last;
          }
        }

        if($all_depends_already_built){ # add this package to the build list.
          push @build_list, $package;
          $build_hash{"$package"} = 1;
          $package_hash{"$package"}{"done"} = 1;
        }
      }
    }
  }until($all_packages_are_done);

my_exit(\@build_list, $do_test);

  #.....................................................................
  sub get_depends{
    # This gets a list of all the *.libs files in a package. Its job is to collate them into a single list.
    my ($libs_files_ref, $do_test, $omit_package_ref) = @_;
    my @libs_files = @{$libs_files_ref};
    my %omit_package = %{$omit_package_ref};

    my %temp_hash;
    foreach my $path (@libs_files){
      chomp($path);

      my @lines = `cat $path`;
      foreach my $line (@lines){
        chomp($line);
        if($line =~ /^\$E_LIB\/(\w+)\.a\s*/){ # should match for example '$E_LIB/erolib.a \'.
          my $depend = $1;
          if(!defined($omit_package{"$depend"})){
            $temp_hash{"$1"} = 1; # value is just a dummy. We use a hash to avoid looping many times through a list.
          }
        }elsif($line !~ /^\s*$/){
          if($do_test){
die "Unrecognized line '$line' in libs file $path";
          }else{
exit 0;
          }
        }
      }
    }

    my @keys = keys(%temp_hash);
  return \@keys;
  }

  #.....................................................................
  sub my_exit{
    # print the list of packages before exiting.
    my ($package_list_ref, $do_test) = @_;

    my @package_list = @{$package_list_ref};
    my $i = 0;
    if(@package_list > 0){
      my $out_str = "$package_list[$i]";
      foreach $i (1 .. $#package_list){
        $out_str .= " $package_list[$i]";
      }
      print "$out_str\n";
    }

exit 0;
  }


