'''
	Copyright (C) 2019-2023:
	Max Planck Institut fuer extraterrestrische Physik (MPE)
	Leibniz Institute for Astrophysics Potsdam (AIP)

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of 
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

 	You should have received a copy of the GNU General Public License along
 	with this program; if not, write to the Free Software Foundation, Inc.,
 	51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''	

import numpy 
from astropy.io import fits
from astropy.coordinates import FK5
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from scipy.stats import poisson
import sys, time
from astropy.table import Table
from astropy.wcs import WCS
from mocpy import MOC
from  mocpy import mocpy
import os

def create_moc_from_expmap(fitsFILE, max_order=14):

    from mocpy import MOC

    '''
    creating a MOC are description file from the exposure map
    of an eROSITA tile

    INPUT:    
    fitsFILE: filename of the exposure map

    max_order : maximum healpix order of the MOC. Controls the angular size
    of the smallest healpix cells that will be used to describe the 
    tile area.

    OUTPUT:

    myStruct: MOC object

    '''

    hdu = fits.open(fitsFILE)
    mask = hdu[0].data>0
    m = MOC.from_fits_image(hdu[0], max_order, mask=mask)
    hdu.close()
    return m
    

def read_area_curve(fitsFILE, ECF, IEXTENT = 3):

    '''
    reading the apetool sensitvity curve (count-rate vs area in deg2)

    INPUT:    
    fitsFILE: filename of the fits apetool SensMap product

    ECF : Energy to Count Conversion Factor for the turn
    count rate to energy flux

    IEXTENT = 3 : hdu extension of the area curve. For apetool
    generated sensitivity maps the area curves are in extension 3. 

    OUTPUT:

    myStruct: Structure that stores the area curve and 
    corresponding ECF

    HARD-CODED PARAMETERS
    IEXTENT: hdu extension of the area curve
    '''

    # the area curve is the HDU EXTENSION 3
    #IEXTENT = 3

    # initalise
    myStruct={}
    
    hdu = fits.open(fitsFILE)

    myStruct['area_curve'] =  hdu[IEXTENT].data['AREA'] 
    myStruct['crt_area_curve'] = hdu[IEXTENT].data['LGCR']
    myStruct['ECF'] = ECF


    hdu.close()

    return myStruct


def read_srclist(fitsFILE, BAND=1):
    '''
    reading the source list that includes apetool aperture photometry

    INPUT:    
    fitsFILE: filename of the fits source list

    BAND : band number within the source list. 

    OUTPUT:

    data: the source list to be used to reconstruct the point-source logN-logS
    Extended sources are excluded. Only point sources that have poisson false
    detection propability < POISSON_FDP = 4e-6 are kept. 

    HARD-CODED PARAMETERS:
    POISSON_FDP = 4e-6: Poisson False Detection Propability. 
    '''

    # POISSON FALSE DETECTION PTOBABILITY
    POISSON_FDP = 4e-6
    
    hdu = fits.open(fitsFILE)
    data = hdu[1].data

    # exclude extended sources
    m = data['EXT_LIKE']==0  

    data = data[m]
        
    # this is the part where sources are selected in a specific energy band
    m = numpy.logical_and(data['APE_POIS_{}'.format(BAND)]>=0, data['APE_POIS_{}'.format(BAND)]<=4e-6)
    data= data[m]

    return data

def get_logn_logs(lgflux):

    # this term accounts for Eddington bias, i.e. the
    # fact there are more faint than bright soures and
    # therefore the chance of faint source fluctuating into
    # a detection is larger. It also acts as a prior. 
    
    dnds=numpy.zeros_like(lgflux)

    b1 = -1.58
    b2 = -2.50
    lgknorm = 0.0
    lgfb14  = -0.04
    lgf14=lgflux-numpy.log10(1.0e-14)

    lgknorm_prime=lgknorm + (b1-b2) * lgfb14

    for i in range(dnds.size):
        if(lgf14[i]>lgfb14):
            dnds[i]=lgknorm_prime + b2 * lgf14[i]
        else:
            dnds[i]=lgknorm + b1 * lgf14[i]

    return dnds;


def produce_dnds(srcdata=[], myStruct={}, BAND=1):
    '''
    construnction of logN-logS using the apetool sensitivity map
    and the apetool aperture photometry of individual sources

    INPUT:

    srcdata: eROSITA source catalogue with aperture photometry columns
    APE_CTS_{$BAND}: total number of counts within an aperture 
    for the spectral band $BAND (integer)
    APE_BKG_{$BAND}: expected background level within the aperture 
    for the spectral band $BAND (real) 
    APE_EXP_{$BAND}: mean exposure time within the aperture 
    for the spectral band $BAND (real)
    APE_EEF_{$BAND}: Encircled Energy Fraction of the aperture 
    for the spectral band $BAND
    
    myStruct: dictionaty that contains the area curve and 
    the Energy to Flux Conversion Factor (ECF)

    BAND: integer that represents the spectral band $BAND in 
    the source list
     
    OUTPUT:
    dictionary that containts the differential and cumulative number counts
    "LGFLUX": logarithmic flux bin 
    "LGDNDS": logarithm of the differential number counts dN/dS 
    "CUMULATIVE": cumulative number counts 

    '''

    from scipy import stats

    # setup arrays
    total = srcdata['APE_CTS_{}'.format(BAND)]
    expt = srcdata['APE_EXP_{}'.format(BAND)]
    bkg = srcdata['APE_BKG_{}'.format(BAND)]
    eef = srcdata['APE_EEF_{}'.format(BAND)]


    # CR is the countrate
    # the logN-logS is estimated at each point of the vector
    lgCR=numpy.linspace(-3.0,1.0, 100, endpoint=True, retstep=False)   

    sum_prob=numpy.zeros_like(lgCR)
    sum_prob1=numpy.zeros_like(lgCR)
    cumulative_sum=numpy.zeros_like(lgCR)
    CR=10**lgCR

    # determines dlogCR
    dlgCR = (lgCR[-1] - lgCR[0])/lgCR.size

    # reads the area curve - should be part of myStruct disctionaty
    # see function read_area_curve
    # below is reads the value of the area curve at a given CR
    crt_area_curve = myStruct['crt_area_curve']
    area_curve = myStruct['area_curve'] 

    # determines area curve at each point of the CR vector
    area = numpy.interp(lgCR, crt_area_curve, area_curve)


    # the assumption is that the are curve is in CR
    # in which case an ECF is needed to conver to flux.
    lgflux = lgCR - numpy.log10(myStruct['ECF']) 

    # estimate dS (linear flux) from dlogCR (log CR)
    dS = dlgCR * numpy.log(10) * 10**lgflux


    # loop through sources
    for isrc in range(bkg.size):

        # for each source estimate the expected total number of counts
        # within the aperture, i.e. CR * exposure_time * PSF_EEF +background
        # where PSF_EEF is the encicrcled energy fraction of the aperture,
        # i.e. what fraction of the photons are enclosed within the given aperture
        expected = CR * expt[isrc] * eef[isrc] + bkg[isrc]

        #  define a Poisson PDF instance p
        p = stats.poisson(expected)

        # this is the "prior", i.e. the logn-logs
        # it represents the probability of a source
        # having a given flux
        dnds_factor=10**(get_logn_logs(lgflux)) 

        # this estimates the probability of a sources with a given
        # number of observed counts (source + background) to have a flux
        # of CR. The multiplication with dnds_factor is kind of a prior and provides
        # a measure of the number of sources with a given CR or flux from which the
        # source in question could be drawn. The additional factor CR has to do with the
        # fact that the probabiloty is per dlogS not dS and so a transformation is
        # needed. See notes in  function get_logn_logs(lgflux):
        prob=p.pmf(total[isrc]) * dnds_factor * CR 


        if(prob.max()>0):

            # normalise to unity
            prob=prob/prob.sum()            

            # devide with the area curve
            prob = prob /area            

            # sum up
            sum_prob = sum_prob + prob
            #print prob
            cumulative_sum = cumulative_sum+numpy.cumsum(numpy.flip(prob))
                        
    # differential number counts dN/dS can compare with observed logN-logS: e.g. https://members.noa.gr/age/data.html
    dnds = sum_prob / dS

    #for i in range(lgCR.size):
    #    if(dnds[i]>0):
    #        print( lgflux[i], sum_prob[i], numpy.log10(dnds[i]))

    # cumulative can be compared with the observed logN-logS: e.g. https://members.noa.gr/age/data.html
    #for i in range(lgCR.size):
    #    print (lgflux[lgCR.size-1-i], cumulative_sum[i])


    return {"LGFLUX": lgflux, "LGDNDS": numpy.log10(dnds), "CUMULATIVE": numpy.flip(cumulative_sum)}




def get_tile_boundaries(tile, skymapfile="./DATA/SKYMAPS_052022.fits"):

    data = Table.read(skymapfile)
    if (isinstance(tile, str)):
        tile=int(tile)
    return data[data['SRVMAP'] == tile];

def is_in_tile(coords, tile_info):
    
    m1 = numpy.logical_and(coords.ra.value>tile_info['RA_MIN'].data, coords.ra.value<=tile_info['RA_MAX'].data)
    m2 = numpy.logical_and(coords.dec.value>tile_info['DE_MIN'].data, coords.dec.value<=tile_info['DE_MAX'].data)
    return numpy.logical_and(m1,m2)

def read_tiles_from_file(fname):

    f=open(fname,"r")
    tiles=[]
    for line in f.readlines():
        line=line.rstrip("\n")
        tiles.append(line)
    f.close()
    return tiles


def create_moc(hplist, max_order, mocfilename):
    # WATCH OUT IT USES PYMOC 
    from pymoc import MOC as MOC1

    l = numpy.unique(hplist)
    m = MOC1(max_order, l)
    m.normalize(max_order=max_order)
    m.write(mocfilename, filetype='fits', overwrite=True)
        
def create_moc_from_expmap(expmapfile, max_order):
    hdu = fits.open(expmapfile)
    mask = hdu[0].data>0
    m = MOC.from_fits_image(hdu[0], max_order, mask=mask)
    hdu.close()
    return m
  
def flatten(moc):
    max_depth = moc.max_order
    #print(max_depth)
    #pix_id= core.flatten_pixels(moc._interval_set._intervals, max_depth)
    pix_id= mocpy.flatten_pixels(moc._interval_set._intervals, max_depth)
    return pix_id


def merge_mocs(tiles):

    i=-1
    for tile in tiles:

        search1 = os.path.join("MOCOUT", "{}_uni_indie.moc".format(tile))
                
        if (os.path.isfile(search1)):
            i=i+1
            moc = MOC.from_fits(search1)

            if(i==0):
                union = moc
            else:
                union=union.union(moc)
            del moc;

    #union.write(outfile, format='fits', overwrite=True)
    return union


def get_wcs(expmapfile, extension=0):
    hdu = fits.open(expmapfile)    
    wcs=WCS(hdu[extension].header)    
    data = hdu[extension].data
    header = hdu[extension].header
    hdu.close()
    return wcs, data, header

def coadd_acurves(tiles=[], suffix="HEALAREA", band="024"):

    for i, tile in enumerate(tiles):

        filename = "{}_{}_{}.fits".format(suffix, tile, band)
        print(filename)
        hdu = fits.open(filename)    
        data = hdu[1].data
        if(i==0):
            header = hdu[1].header
            all = data['AREA']
        else:
            all=all+data['AREA']        
        hdu.close()
        
    filename="{}_{}_{}.fits".format(suffix, "ALL", band)
        
    c=[]
    c.append(fits.Column(name='LGCR', array=data['LGCRT'], format='E'))
    c.append(fits.Column(name='AREA', array=all, format='E'))
    cols = fits.ColDefs(c)
    hdu = fits.BinTableHDU.from_columns(cols, header=header)
    hdu.writeto(filename, overwrite=True)
    print("SAVED AREA CURVE TO {}".format(filename))
         



def do_calc(obs, eef=0.75, poisson_thresh=4e-6, lgcrt_min=-4, lgcrt_max=1,  bin_lgflux=0.01):

    '''
    functions that claculates the sensitivity / area curve based on aperture photometry products. 
    
    INPUTS: 

    obs: dictionary containing a list of positions that cover the field-of-view over which the 
    sensitvity curve needs to be estimated and provide the background level integreated 
    within an aperture of a given EEF (dictionary name 'bkgeef') and the mean exposure time 
    within the same aperture (dictionary name 'expeef'). These quantities can be taken from 
    the APETOOL ApeSensMap sensitvity map product. 

    eef = 0.75 : the Encircled Energy Fraction of the aperture used to extract counts. This is fixed for 
    all apertures. In ERASS1 a value of 0.75 is adopted, i.e. the aperture has a fixed EEF of
    75 per cent. It is recommended not to change this value.

    poisson_thresh = 4e-6 : Poisson False Detection Probability.

    lgcrt_min = -4 : minimum logarithmic count rate at which the sensitivity curve is estimated

    lgcrt_max = 1:  maximum  logarithmic count rate at which the sensitivity curve is estimated
    
    bin_lgflux=0.01: logarithic count rate step size. 

    '''
    
    expt = obs['expeef'] * eef
    bg = obs['bkgeef']

    mask=bg>0
    bg=bg[mask]
    expt=expt[mask]

    # this is the sensitivity map. The minimum number of
    # counts that an aperture should contain so that the
    # probability of the background (bg) fluctuating above
    # min_counts is less than the Poisson False Detection
    # Threshold. Here the Inverse Survival Function of the
    # Poisson distribution is used to determine min_counts
    min_counts = poisson.isf(poisson_thresh, bg)

    #print("SHAPES:",min_counts.shape, bg.shape, expt.shape)
    
    lgcrt = numpy.arange(lgcrt_min, lgcrt_max,  bin_lgflux)    

    #print("LGCRT:",lgcrt)
    
    crt1=10**lgcrt 

    area_pix=numpy.zeros_like(lgcrt)

    # loop through all values of expt / bg to determine for
    # each on the probability of a source having count rate CR 
    do_loop(expt, crt1, bg, min_counts, area_pix, 0,len(crt1))

    # multiply by the pixel size to get area in deg2
    S1=area_pix * obs['hpix_area'].value

    return {"LGCRT": lgcrt, "AREA": S1}


def do_basic_calc(expt, crt, bg, min_counts):

    # expect = expt * crt * 0.7 + bg
    # i multiple with 0.7 expt before the loop

    expect = expt * crt + bg
    
    #mask = expect<=0

    prob = poisson.sf(min_counts, expect )
    
    A = prob.sum()
    
    return A;


def do_loop(expt, crt, bg, min_counts, area_pix, START, END):


    for i in range(START,END):    

        A1 = do_basic_calc(expt, crt[i], bg, min_counts)

        area_pix[i]=A1

        #print (i, crt[i], area_pix[i], bg.size, min_counts.size)


    return area_pix


def  save_area(sense, tile, band, out="AREA"):

    filename = "{}_{}_{}.fits".format(out, tile, band)

    c=[]
    c.append(fits.Column(name='LGCRT', array=sense['LGCRT'], format='E'))
    c.append(fits.Column(name='AREA', array=sense['AREA'], format='E'))
    cols = fits.ColDefs(c)
    hdu = fits.BinTableHDU.from_columns(cols)
    hdu.writeto(filename, overwrite=True)
    print("SAVED AREA CURVE TO {}".format(filename))
 
