c Copyright (C) 2019-2023:
c   Max Planck Institut fuer extraterrestrische Physik (MPE)
c   Leibniz Institute for Astrophysics Potsdam (AIP)
c
c This file is part of eROSITA's Science Analysis Software System
c (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
c redistribute it and/or modify it under the terms of the GNU General
c Public License as published by the Free Software Foundation; either
c version 2 of the License, or (at your option) any later version.
c
c This program is distributed in the hope that it will be useful,
c but WITHOUT ANY WARRANTY; without even the implied warranty of
c MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
c General Public License for more details.
c
c You should have received a copy of the GNU General Public License
c along with this library; if not, write to the Free Software
c Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
c USA, or see https://www.gnu.org/licenses/.
c-----------------------------------------------------------------------
c
      module astrolib

      use param
   
      implicit none
      real*8, parameter :: dpi    =  dacos(-1d0)
      real*8, parameter :: d2pi   =  2d0*dpi 
      real*8, parameter :: dpid2  =  dpi/2d0
      real*8, parameter :: dpi180 =  dpi/180d0
      real*8, parameter :: d180pi =  180d0/dpi
      real*8, parameter :: pi=dpi
      character(len=256) :: outpt

      contains

      subroutine ANGADD(DALPH0,DDELT0,DANG,DPOS,DALPHA,DDELTA,IERR)
C
CC  Calculates a new position out of the separation and position angle
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:33
C
CA  author : KOD        date: 07-FEB-1991 19:22
CU  update : KOD        date: 18-AUG-1994 12:37
CU  update : AGU        date: 13 May 2020 10:51 changed D line into C lines 
C
CT  status: tested
C
C   general description 
CG
CG  This subroutine calculates the coordinates of a position which
CG  is supplied by a reference position and an offset. The offset is
CG  specified by the angular separation and the position angle. The
CG  position angle is defined as the angle between the great circle
CG  passing through both positions and the local north meridian as
CG  seen from the reference position (DALPH0, DDELT0). It is counted
CG  counterclockwise from North over East, South, West to North.
C
C   call_var.          type I/O description 
CP  DALPH0              R8  I   Right Ascension of reference  (radians)
CP  DDELT0              R8  I   Declination     of reference  (radians)
CP  DANG                R8  I   angular separation            (radians)
CP  DPOS                R8  I   position angle                (radians)
CP  DALPHA              R8  I/O resulting Right Ascension     (radians)
CP  DDELTA              R8    O resulting Declination         (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
CR  DRADRS              R8      
CR  ERRSTC              SR      
CR  EULER1              SR      
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DALPH0, DDELT0, DALPHA, DDELTA, DANG, DPOS,
!HB  &              D1, D2, D3, DPH1, DTH1, DPH2, DTH2, DRADRS
     &              D1, D2, D3, DPH1, DTH1, DPH2, DTH2
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT ! , IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'ANGADD'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (max(abs(DALPH0),abs(DALPHA)).gt.D2PI) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Unusual value for the R.A.')
      endif
C
      if (DPOS.lt.-DPI .or. DPOS.gt.D2PI) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Unusual value for DPOS')
      endif
C
      if (DANG.lt.0.D0) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'ANG < 0')
      endif
C
C
      if (IERR.lt.2) then
C
        D1 =  0.D0
        D2 =  DRADRS ( DDELT0 - DPID2 )
        D3 =  DRADRS ( -DPID2 - DALPH0 )
C
        DPH1 = DRADRS ( DPID2 - DPOS )
        DTH1 = DANG
C
        call EULER1(D1,D2,D3,DPH1,DTH1,DPH2,DTH2,ISTAT)
C
        if (ISTAT.gt.0) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
C
        else
C
          DALPHA = DPH2
          DDELTA = DPID2 - DTH2
C
        endif
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: !<output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine ANGCQ1(LEQU,DL,DB,DEPS,DANGCQ,IERR)
C
CC    Calculates the angle between eclipt. and equat. north directions
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:33
C
CA  author : KOD        date: 22-JUN-1990 13:19
CU  update : KOD        date: 26-JUN-1990 17:04
CU  update : KOD        date: 17-AUG-1992 10:00 (`1.D-6')
C
CT  status: tested
C
C   general description 
CG
CG  Subroutine to calculate the angle between the ecliptical and
CG  equatorial north meridians for a specified direction on the sky.
CG  This angle is defined counterclockwise from the equatorial to
CG  the ecliptical north meridian (as viewed from the inside of
CG  a sphere, as usual) and will be in the interval (0,2*pi).
CG  The direction on the sky can be supplied in equatorial or
CG  ecliptical coordinates. Use the subroutine JDEPS1 to calculate
CG  the obliquity of the ecliptic.
CG  This subroutine has been optimized for high speed.
C
C   call_var.          type I/O description 
CP  LEQU                L4  I   equatorial or ecliptical ccordinates:
CP                                LEQU = .true.  : equatorial coord.
CP                                LEQU = .false. : ecliptical coord.
CP  DL                  R8  I   R.A. or ecliptical longitude (radians)
CP  DB                  R8  I   Dec. or ecliptical latitude  (radians)
CP  DEPS                R8  I   obliquity of the ecliptic    (radians)
CP  DANGCQ              R8    O angle between the meridians  (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  DRADRS              R8      `resets' radians
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
!HB   REAL*8        DL, DB, DEPS, DANGCQ, DX, DY, DH, DRADRS,
      REAL*8        DL, DB, DEPS, DANGCQ, DX, DY, DH,
     &              DSE, DCE, DSA, DCA, DSD, DCD, DSL, DCL, DSB, DCB,
     &              DLM, DAL, DANG ! , DBT, DDE
c      REAL*4        HFLAG
      INTEGER*4     IERR !, IOS
      LOGICAL*4     LEQU
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'ANGCQ1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     check input values:
C
      if (DL.lt.-1.D-6 .or. DL.GT.D2PI+1.D-6) then
        IERR = 1
        call WRERR(RTNAME,IERR,'Warning: unusual value for DL')
      endif
C
      if (dabs(DB).gt.DPID2+1.D-6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameter DB')
      endif
C
C
      if (IERR.lt.2) then
C
        DSE = dsin(DEPS)
        DCE = dcos(DEPS)
C
        if (LEQU) then
C
          DSA = dsin(DL)
          DCA = dcos(DL)
          DSD = dsin(DB)
          DCD = dcos(DB)
C
          DX = DCD * DCA
          DY = DCE * DCD * DSA + DSE * DSD
C
          DLM = datan2(DY,DX)
C
          DSL = dsin(DLM)
          DCL = dcos(DLM)
C
        else
C
          DSL = dsin(DL)
          DCL = dcos(DL)
          DSB = dsin(DB)
          DCB = dcos(DB)
C
          DX = DCB * DCL
          DY = DCE * DCB * DSL - DSE * DSB
C
          DAL = datan2(DY,DX)
C
          DSA = dsin(DAL)
          DCA = dcos(DAL)
C
        endif
C
        DANG = dacos ( DSL * DSA + DCL * DCA * DCE )
C
        DH = DRADRS(DL+DPID2) - DPI
        if (DH.lt.0.D0) DANG = -DANG
C
        DANGCQ = DRADRS(DANG)
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine ANGPOS(DALPH0,DDELT0,DALPHA,DDELTA,DPOS,IERR)
C
CC    Calculates the position angle
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:33
C
CA  author : KOD        date: 06-FEB-1991 19:24
CU  update : KOD        date: 19-FEB-1991 18:21
C
CT  status: tested
C
C   general description 
CG
CG  Subroutine to calculate the position angle between two positions.
CG  This value is defined as the angle between the great circle
CG  passing through both positions and the local north meridian as
CG  seen from the position (DALPH0, DDELT0). It is counted
CG  counterclockwise from North over East, South, West to North.
C
C   call_var.          type I/O description 
CP  DALPH0              R8  I   Right Ascension of reference  (radians)
CP  DDELT0              R8  I   Declination     of reference  (radians)
CP  DALPHA              R8  I   Right Ascension               (radians)
CP  DDELTA              R8  I   Declination                   (radians)
CP  DPOS                R8    O position angle                (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  EULER1              SR      general coordinate transformation
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
CR  DRADRS              R8      'resets' radians
CR  ERRSTC              SR      continue error stack for one level
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DALPH0, DDELT0, DALPHA, DDELTA, DPOS,
!HB  &              D1, D2, D3, DPH1, DTH1, DPH2, DTH2, DRADRS
     &              D1, D2, D3, DPH1, DTH1, DPH2, DTH2
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT ! , IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'ANGPOS'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (max(abs(DALPH0),abs(DALPHA)).gt.D2PI) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Unusual value for the R.A.')
      endif
C
      if (max(abs(DDELT0), abs(DDELTA)).gt.DPID2) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameters')
      endif
C
C
      if (IERR.lt.2) then
C
        D1 =  DRADRS ( DPID2 + DALPH0 )
        D2 =  DPID2 - DDELT0
        D3 =  0.D0
C
        DPH1 = DALPHA
        DTH1 = DPID2 - DDELTA
C
        call EULER1(D1,D2,D3,DPH1,DTH1,DPH2,DTH2,ISTAT)
C
        if (ISTAT.gt.0) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
C
        else
C
          DPOS = DRADRS ( DPID2 - DPH2 )
C
        endif
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: !<output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine ANGSEP(DLONG1,DLAT1,DLONG2,DLAT2,DANG,IERR)
C
CC    Calculates the angular separation between two celestial bodies
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:33
C
CA  author : KOD        date: 06-SEP-1989 16:40
CU  update : KOD        date: 20-MAR-1990 15:45
CU  update : KOD        date: 24-OCT-1991 13:57  (sin --> dsin, ...)
CU  update : KOD        date: 17-AUG-1992 09:58  (`1.D-6')
C
CT  status: tested
C
C   general description 
CG
CG  Subroutine to calculate the angular separation between two
CG  celestial bodies. The 'latitude' is defined as to have
CG  0 degrees at the equator and +90 degrees at the north pole.
CG  Special care is taken for small angles, according to
CG    R.W.Sinnott, 'Sky and Telescope', August 1984, p.159.
C
C   call_var.          type I/O description 
CP  DLONG1              R8  I   'longitude 1' (e.g. R.A.; radians)
CP  DLAT1               R8  I   'latitude  1' (e.g. Dec.; radians)
CP  DLONG2              R8  I   'longitude 2' (e.g. R.A.; radians)
CP  DLAT2               R8  I   'latitude  2' (e.g. Dec.; radians)
CP  DANG                R8    O angular separation       (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DLONG1, DLAT1, DLONG2, DLAT2, DANG, D1, D2, DR
c      REAL*4        HFLAG
      INTEGER*4     IERR ! ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'ANGSEP'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (max(abs(DLONG1),abs(DLONG2)).gt.D2PI+1.D-6) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Unusual value for the longitude')
      endif
 
      if (max(abs(DLAT1),abs(DLAT2)).gt.DPID2+1.D-6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameters')
      endif
C
      if (IERR.lt.2) then
C
        D1 = dsin ( 0.5D0 * ( DLAT1 -  DLAT2 ) )
        D2 = dsin ( 0.5D0 * (DLONG1 - DLONG2 ) )
        DR = D1 * D1 + D2 * D2 * dcos(DLAT1) * dcos(DLAT2)
        DANG = 2.D0 * datan2 ( dsqrt(DR), dsqrt(1.D0-DR) )
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: !<output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CALJD1(IY,IM,ID,IHR,IMT,SEC,LCHECK,DJD,IERR)
C
CC    Converts the calendar date to the Julian Date
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:33
C
CA  author : KOD        date: 18-JUL-1989 15:03
CU  update : KOD        date: 28-MAR-1990 15:54
CU  update : KOD        date: 28-MAR-1990 19:23
CU  update : KOD        date: 4-JUN-1992 08:40  ('ihr=24 imt=0 sec=0.0')
CU  update : CRI        date: 18-DEC-1992 13:29 FORMATS
C
CT  status: tested (08-AUG-1989 17:45 / KOD)
C
C   general description 
CG  Subroutine to calculate the Julian Date out of the calendar date.
CG  The calendar date is assumed to be given in the Gregorian system
CG  for dates starting with October 15, 1582, and in the Julian
CG  system until October 4, 1582 (in between it is undefined).
CG
CG  This subroutine is valid for the years  -4712  to  +3500 ,
CG  and has been written according to
CG     van Flandern and Pulkkinen:
CG     'Low-Precision Formulae for Planetary Positions',
CG     Astrophysical Journal Supplement Series 41, 1979, p.392.
CG  and
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.23-25 .
CG
CG  For negative years all the other parameters must be positive.
CG  An extended input check is possible for valid dates.
C
C   call_var.          type I/O description 
CP  IY                  I4  I   year         (-4712 .. +3500)
CP  IM                  I4  I   month        (1 .. 12)
CP  ID                  I4  I   day of month (1 .. 28,29,30,31)
CP  IHR                 I4  I   hours        (0 .. 23)
CP  IMT                 I4  I   minutes      (0 .. 59)
CP  SEC                 R4  I   seconds      (0.0 .. 59.999...)
CP  LCHECK              L4  I   extended input check if LCHECK=.TRUE.
CP  DJD                 R8    O corresponding Julian Date
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  CNVFRC              SR      conversion to fractional number
CR  JDCAL1              SR      converts Jul. Date into Calendar Date
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      INTEGER*4     IY, IM, ID, IERR, IYR, IMTH, IDAY, IHR, IMT, ISTAT,
     &              JD, IYY, IMM, IHR1, IMT1, IDAT !, IOS
      REAL*4        SEC, SEC1 !HFLAG, 
      REAL*8        DJD, DFRC, DH
      LOGICAL*4     LCHECK, LPLUS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CALJD1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
CR!   if wrong input
C
      if ( IY .lt.-4712 .or. IY .gt.3500 .or.
     &     IM .lt.1     .or. IM .gt.12   .or.
     &     ID .lt.1     .or. ID .gt.31   .or.
     &     IHR.lt.0     .or. IHR.gt.24   .or.
     &    (IHR.eq.24 .and. (IMT.gt.0 .or. SEC.gt.0.0)) .or.
     &     IMT.lt.0     .or. IMT.gt.59   .or.
     &     SEC.lt.0.0   .or. SEC.gt.60.0 )   then
C
        IERR = 2
        WRITE(OUTPT,9)  IY,IM,ID
9       FORMAT('Error in input parameters IY,IM,ID ',3(I6,1X))
        call WRERR(RTNAME,IERR,OUTPT)
C
      endif
C
C
      if  (IERR.lt.2)  then
C
C       calculate the Julian Day for 0:00 UT
C
        IDAT = isign((iabs(IY)*100+IM)*100+ID,IY)
C
        if (IDAT.ge.15821015) then
C
C         assume Gregorian calendar date as input:
C
          JD = 367*IY - 7*(IY+(IM+9)/12)/4 - 3*((IY+(IM-9)/7)/100+1)/4
     &       + 275*IM/9 + ID + 1721029
C
        else if (IDAT.le.15821004) then
C
C         assume Julian calendar date as input:
C
          if (IM.gt.2) then
            IYY = IY
            IMM = IM
          else
            IYY = IY - 1
            IMM = IM + 12
          endif
C
          if (IYY.gt.0) then
            DH=0.D0
          else
            DH=0.75D0
          endif
C
          JD = idint(365.25d0*dble(IYY)-DH) +
     &         idint(30.6001d0*dble(IMM+1)) +
     &         ID + 1720995
C
          else
C
            IERR = max(IERR,2)
            WRITE(OUTPT,10)  IDAT
10          FORMAT('Error in date: calendar reform ',I10)
            call WRERR(RTNAME,IERR,OUTPT)
C
        endif
C
C       calculate the fractional part:
C
        LPLUS = .true.
        IHR1 = IHR
        IMT1 = IMT
        SEC1 = SEC
        call CNVFRC(LPLUS,IHR1,IMT1,SEC1,DFRC,ISTAT)
C
CN!     ERROR handling
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
C       add both parts
C
        DJD = dble(JD) + DFRC/24.D0 - 0.5D0
C
        IF (LCHECK) THEN
C
CJ!       extended input check
C
          DH = dble(JD) - 0.5D0*0.0
          call JDCAL1(DH,IYR,IMTH,IDAY,IHR1,IMT1,SEC1,ISTAT)
C
CN!       ERROR handling
C
          if  (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,20)
          endif
C
CR!       if inconsistent results
C
          if (IYR .ne.IY  .or. IMTH.ne.IM .or. IDAY.ne.ID) then
            IERR = max(IERR,2)
            WRITE(OUTPT,11) IYR,IY,IMTH,IM,IDAY,ID
11          FORMAT('Error: Date does not exist ',6(I8,1X))
            call WRERR(RTNAME,IERR,OUTPT)
          endif
        endif
      endif
C
CH!   CJD
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      SUBROUTINE CALJD2(ITD,ITS,DJD,IERR)
C
CC    Calculates the Julian Date out of the Calendar Date.
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:33
C
CA  author : KOD        date: 12-MAR-1992 14:25
CU  update : KOD        date: 16-MAR-1992 16:04
C
CT  status: not tested
C
C   general description 
CG
CG  This subroutine calculates the Julian Date out of the Canedar Date
CG  for the years 1950 to 2049. The Calendar Date is contained in the
CG  integers values ITD (YYMMDD) and ITS (msec of the day in UT).
C
C   call_var.          type I/O description 
CP  ITD                 I4  I   Calendar Date  (YYMMDD)
CP  ITS                 I4  I   Universal Time (msec of the day)
CP  DJD                 R8  ?/? Julian Date
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
CR  CALJD1              SR      calendar to JD conversion
CR  ERRSTC              SR      continue error stack for one level
CR  WRERR               SR      formats error code in standard form
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD
      REAL*4        SEC !HFLAG, 
      INTEGER*4     IERR, ISTAT, ITD, ITS, IY, IM, ID, IHR, IMT, IH !, IOS
      LOGICAL       LCHECK
      CHARACTER*6   CH6
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CALJD2'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      LCHECK = .true.
C
C     simple check of input values:
C
      if (ITD.lt.0 .or. ITS.lt.0 .or. ITS.gt.86400000) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in ITD or ITS')
      endif
C
C
      if (IERR.lt.2) then
C
C       determine the year, month, day:
C
        write(CH6,'(I6)') ITD
        read(CH6,'(3I2)') IY,IM,ID
C
        if (IY.lt.50) then
          IY = 2000 + IY
        else
          IY = 1900 + IY
        endif
C
C       determine the hour, minute, second:
C
        IHR = ITS / 3600000
        IH  = IHR * 3600000
        IMT = (ITS - IH) / 60000
        SEC = 0.001 * real(ITS - IH - IMT * 60000)
C
C       calculate the Julian Date:
C
        call CALJD1(IY,IM,ID,IHR,IMT,SEC,LCHECK,DJD,ISTAT)
C
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
      RETURN
      END
      subroutine CHDATE(IY,IM,ID,IHR,IMT,SEC,LCHECK,CHDAT,IERR)
C
CC    Produces a string of the form 'YYYY.MM.DD / HH:MM:SS.SS'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:33
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 28-MAR-1990 19:33
CU  update : KOD        date: 30-MAR-1990 16:50
C
CT  status: tested (08-AUG-1989 19:29 / KOD)
C
C   general description 
CG
CG  Subroutine to produce a string of the form
CG  'YYYY.MM.DD / HH:MM:SS.SS'  (Maximum resolution: 10 msec).
CG  There will be a rough check on the validity of the given date;
CG  a detailed check is also possible (see LCHECK).
C
C   call_var.          type I/O description 
CP  IY                  I4  I   year         (-999 .. +9999)
CP  IM                  I4  I   month        (1 .. 12)
CP  ID                  I4  I   day of month (1 .. 28,29,30,31)
CP  IHR                 I4  I   hours        (0 .. 23)
CP  IMT                 I4  I   minutes      (0 .. 59)
CP  SEC                 R4  I   seconds      (0.0 .. 59.999...)
CP  LCHECK              L4  I   extended input check if LCHECK=.true.
CP  CHDAT               C   I/O string containing the calendar date
CP                              'YYYY' to 'YYYY.MM.DD / HH:MM:SS.SS'
CP                               possible lengths 'L' of CHDAT :
CP                               L =  4 : 'YYYY'
CP                               L =  7 : 'YYYY.MM'
CP                               L = 10 : 'YYYY.MM.DD'
CP                               L = 15 : 'YYYY.MM.DD / HH'
CP                               L = 18 : 'YYYY.MM.DD / HH:MM'
CP                               L = 21 : 'YYYY.MM.DD / HH:MM:SS'
CP                               L = 23 : 'YYYY.MM.DD / HH:MM:SS.S'
CP                               L = 24 : 'YYYY.MM.DD / HH:MM:SS.SS'
CP                               The values for the year, month
CP                               and day will not be rounded,
CP                               the values for the hour, minute
CP                               and second will be rounded.
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  JDCAL1              SR      Julian Date to Calendar Date
CR  CALJD1              SR      Calendar Date to Julian Date
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD
      REAL*4        SEC, F, FSEC, SEC1 !, H, HFLAG, 
      INTEGER*4     IERR, IY, IM, ID, IHR, IMT, N, NDIG, ISEC,
     &              ISEC1, ISEC2, IY1, IM1, ID1, IHR1, IMT1,
     &              ISTAT !I1, I2, IOS
      LOGICAL*4     LCHECK, LCHK
      CHARACTER*3   CHDIG
      CHARACTER*(*) CHDAT
      CHARACTER*70  CHFMT
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CHDATE'/
C
CD     if (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) then
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     determine length of string:
C
      N = len(CHDAT)
C
C     check length of string:
C
      ISTAT = 1
      if (N.eq. 4 .or. N.eq. 7 .or. N.eq.10 .or. N.eq.15 .or.
     &    N.eq.18 .or. N.eq.21 .or. N.eq.23 .or. N.eq.24)  ISTAT = 0
C
      IERR = max(IERR,ISTAT)
C
      if (IERR.gt.0)
     &   call WRERR(RTNAME,IERR,'Error in length of parameter CHDAT')
C
C     (rough) check for meaningful date:
c
      if (IY .lt.-999 .or.  IY.gt.9999  .or.
     &    IM .lt.1    .or.  IM.gt.12    .or.
     &    ID .lt.1    .or.  ID.gt.31    .or.
     &    IHR.lt.0    .or.  IHR.gt.23   .or.
     &    IMT.lt.0    .or.  IMT.gt.59   .or.
     &    SEC.lt.0.0  .or.  SEC.gt.60.0) then
C
        ISTAT = 2
        IERR = max(IERR,ISTAT)
        call WRERR(RTNAME,IERR,'Error in input parameters IY .. SEC')
C
      endif
C
C
C     precise check for valid date:
C
      if (IERR.lt.2) then
C
C       copy input parameters:
C
        IY1  = IY
        IM1  = IM
        ID1  = ID
        IHR1 = IHR
        IMT1 = IMT
        SEC1 = SEC
C
        if (LCHECK) then
C
          LCHK = LCHECK
          call CALJD1(IY1,IM1,ID1,IHR1,IMT1,SEC1,LCHK,DJD,ISTAT)
C
          if (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,10)
          endif
C
        endif
C
        SEC1 = SEC-1.E-6
C
C       round values:
C
        NDIG  = min( max(N-22,0), 3 )
        F     = real( 10**NDIG )
        ISEC  = nint( SEC1*F )
        FSEC  = real( ISEC ) / F
        ISEC1 = int( FSEC )
        ISEC2 = ISEC - ISEC1 * 10**NDIG
C
C
C       increase minutes:
C
        if ( (ISEC1.ge.60) .or.
     &       (N.le.18 .and. SEC1.gt.30.0) ) then
          SEC1  = 0.0
          ISEC1 = 0
          ISEC2 = 0
          IMT1 = IMT1 + 1
        endif
C
C
C       increase hours:
C
        if ( (IMT1.ge.60) .or.
     &       (N.le.15 .and.
     &        (IMT1.gt.30 .or. (IMT1.eq.30 .and. SEC1.gt.0.0))) ) then
          SEC1  = 0.0
          ISEC1 = 0
          ISEC2 = 0
          IMT1 = 0
          IHR1 = IHR1 + 1
        endif
C
C
C       increase the day:
C
        if (IHR1.eq.24) then
C
          IHR1 = 0
C
          LCHK=.true.
          call CALJD1(IY1,IM1,ID1,IHR1,IMT1,SEC1,LCHK,DJD,ISTAT)
C
          if (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,20)
          endif
C
          if (IERR.lt.2) then
            DJD = DJD + 1.D0
            call JDCAL1(DJD,IY1,IM1,ID1,IHR1,IMT1,SEC1,ISTAT)
C
            if (ISTAT.gt.1) then
              IERR = max(IERR,ISTAT)
              call ERRSTC(RTNAME,30)
            endif
C
          endif
        endif
C
C
C       assign result to CHDAT:
C
        if (N.eq. 4) write(CHDAT, 4) IY1
        if (N.eq. 7) write(CHDAT, 7) IY1,IM1
        if (N.eq.10) write(CHDAT,10) IY1,IM1,ID1
        if (N.eq.15) write(CHDAT,15) IY1,IM1,ID1,IHR1
        if (N.eq.18) write(CHDAT,18) IY1,IM1,ID1,IHR1,IMT1
        if (N.eq.21) write(CHDAT,21) IY1,IM1,ID1,IHR1,IMT1,ISEC1
        if (N.GE.23) then
           write(CHDIG,'(I1,''.'',I1)') NDIG,NDIG
           CHFMT='(I4,''.'',I2.2,''.'',I2.2,'' / '',' //
     $           'I2.2,'':'',I2.2,'':'',I2.2,''.'',I' // CHDIG // ')'
           write(CHDAT,CHFMT) IY1,IM1,ID1,IHR1,IMT1,ISEC1,ISEC2
        endif
      endif
C
    4 FORMAT(I4)
    7 FORMAT(I4,'.',I2.2)
   10 FORMAT(I4,'.',I2.2,'.',I2.2)
   15 FORMAT(I4,'.',I2.2,'.',I2.2,' / ',I2.2)
   18 FORMAT(I4,'.',I2.2,'.',I2.2,' / ',I2.2,':',I2.2)
   21 FORMAT(I4,'.',I2.2,'.',I2.2,' / ',I2.2,':',I2.2,':',I2.2)
C
C
CD     if (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) then
CD       write (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CHDEG1(DRAD,CHDG,IERR)
C
CC    Converts radians into degrees and writes a string 'DDD:MM:SS.SSSS'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:33
C
CA  author : KOD        date: 22-AUG-1989 16:41
CU  update : KOD        date: 28-MAR-1990 19:08
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert radians into a degrees and write a string
CG  of the form 'DDD:MM:SS.SSSS'
C
C   call_var.          type I/O description 
CP  DRAD                R8  I   angle to be converted (rad,-pi/2..2*pi)
CP  CHDG                C   I/O string containing the degrees
CP                              'DDD:MM:SS.SSSS'
CP                               possible lengths 'L' of CHDG :
CP                               L =  3 : 'DDD'
CP                               L =  6 : 'DDD:MM'
CP                               L =  9 : 'DDD:MM:SS'
CP                               L = 11 : 'DDD:MM:SS.S'
CP                               L = 12 : 'DDD:MM:SS.SS'
CP                               L = 13 : 'DDD:MM:SS.SSS'
CP                               L = 14 : 'DDD:MM:SS.SSSS'
CP                               In each case the values
CP                               will be rounded appropriately;
CP                               the value of DDD must be greater
CP                               than -99 and less than 999.
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  CNVDMS              SR      splits decimal degrees
CR  HFLAG               R4      output flag handling routine
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DRAD, DDG, DDGABS
      REAL*4        RSEC, FSEC, F !HFLAG, 
      INTEGER*4     IERR, IDG, IMIN, N, NDIG, ISEC, ISEC1, ISEC2,
     &              ISTAT ! ,IOS
      LOGICAL*4     LUNIT, LPLUS, LH
      CHARACTER*3   CHDIG
      CHARACTER*35  CHFMT
      CHARACTER*(*) CHDG
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CHDEG1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     initialize values:
C
      N = len(CHDG)
      DDG = DRAD * D180PI
      LUNIT = (DDG.ge.-90.D0 .and. DDG.lt.360.D0)
C
C     check input values:
C
      ISTAT = 1
      if (N.eq.3 .or. N.eq.6 .or. N.eq.9 .or. (N.ge.11 .and. N.le.14))
     &  ISTAT = 0
      if (ISTAT.ne.0) then
        IERR = max(IERR,ISTAT)
        call WRERR(RTNAME,IERR,'Error in length of CHDG')
        N = min(N,14)
      endif
C
      ISTAT = 0
      if (DDG.lt. -90.D0 .or. DDG.ge. 360.D0) ISTAT = 1
      if (DDG.le.-100.D0 .or. DDG.ge.1000.D0) ISTAT = 2
C
      if (ISTAT.ne.0) then
        IERR = max(IERR,ISTAT)
        call WRERR(RTNAME,IERR,'Error in the input parameter DDG')
      endif
C
C
      if (IERR.lt.2) then
C
C       store sign:
C
        LPLUS = (DDG.ge.0.D0)
        DDGABS = dabs(DDG)
C
C       extract hours, minutes and seconds:
C
        call CNVDMS(DDGABS,LH,IDG,IMIN,RSEC,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       round values:
C
        NDIG = max(N-10,0)
        F = real(10**NDIG)
        ISEC = nint(RSEC*F)
        FSEC = real(ISEC)/F
        ISEC1 = int(FSEC)
        ISEC2 = ISEC - ISEC1*10**NDIG
        RSEC  = nint(RSEC*10000)*1.E-4
C
        if ( (ISEC1.ge.60) .or.
     &      (N.le.6 .and. (ISEC1.gt.30 .or. RSEC.gt.30.0)) ) then
          RSEC  = 0.0
          ISEC1 = 0
          ISEC2 = 0
          IMIN  = IMIN + 1
        endif
C
        if ( (IMIN.ge.60) .or.
     &       (N.eq.3 .and.
     &        (IMIN.gt.30 .or. (IMIN.eq.30 .and. RSEC.gt.0.0))) ) then
          RSEC  = 0.0
          ISEC1 = 0
          ISEC2 = 0
          IMIN  = 0
          IDG   = IDG + 1
        endif
C
        if (LUNIT .and. IDG.ge.360) IDG = IDG - 360
C
        if (.not.LPLUS .and. IDG.eq.0) then
          if (N.eq. 3) LPLUS=.true.
          if (N.eq. 6 .and. IMIN.eq.0) LPLUS=.true.
          if (N.eq. 9 .and. IMIN.eq.0 .and. ISEC1.eq.0) LPLUS=.true.
          if (N.ge.11 .and. IMIN.eq.0 .and. ISEC1.eq.0 .and. ISEC2.eq.0)
     &       LPLUS=.true.
        endif
C
        if (.not.LPLUS) IDG = -IDG
C
C       assign result to CHDG:
C
        if (N.eq.3) write(CHDG,'(I3.1)') IDG
        if (N.eq.6) write(CHDG,'(I3.1,'':'',I2.2)') IDG,IMIN
        if (N.eq.9) write(CHDG,'(I3.1,'':'',I2.2,'':'',I2.2)') IDG,IMIN,
     &                                                         ISEC1
        if (N.ge.11) then
           write(CHDIG,'(I1,''.'',I1)') NDIG,NDIG
           CHFMT = '(I3.1,'':'',I2.2,'':'',I2.2,''.'',I' // CHDIG // ')'
           write(CHDG,CHFMT) IDG,IMIN,ISEC1,ISEC2
        endif
C
        if (.not.LPLUS .and. IDG.eq.0) CHDG(2:2)='-'
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CHDEG2(DRAD,CHDG,IERR)
C
CC    Converts radians into degrees and writes a string 'DDD:MM.MMMMM'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:33
C
CA  author : KOD        date: 25-AUG-1989 14:41
CU  update : KOD        date: 28-MAR-1990 19:10
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert radians into a degrees and write a string
CG  of the form 'DDD:MM.MMMMM'
C
C   call_var.          type I/O description 
CP  DRAD                R8  I   angle to be converted (rad,-pi/2..2*pi)
CP  CHDG                C   I/O string containing the degrees
CP                              'DDD:MM.MMMMM'
CP                               possible lengths 'L' of CHDG :
CP                               L =  3 : 'DDD'
CP                               L =  6 : 'DDD:MM'
CP                               L =  8 : 'DDD:MM.M'
CP                               L =  9 : 'DDD:MM.MM'
CP                               L = 10 : 'DDD:MM.MMM'
CP                               L = 11 : 'DDD:MM.MMMM'
CP                               L = 12 : 'DDD:MM.MMMMM'
CP                               In each case the values
CP                               will be rounded appropriately;
CP                               the value of DDD must be greater
CP                               than -99 and less than 999.
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  CNVDMS              SR      splits decimal degrees
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DRAD, DDG, DDGABS
      REAL*4        F, RMIN, RSEC !HFLAG, 
      INTEGER*4     IERR, IDG, IMIN, N, NDIG, IMIN1, ISTAT ! ,IOS
      LOGICAL*4     LUNIT, LPLUS, LH
      CHARACTER*3   CHDIG
      CHARACTER*35  CHFMT
      CHARACTER*(*) CHDG
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CHDEG2'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     initialize values:
C
      N = len(CHDG)
      DDG = DRAD * D180PI
      LUNIT = (DDG.ge.-90.D0 .and. DDG.lt.360.D0)
C
C     check input values:
C
      ISTAT = 1
      if (N.eq.3 .or. N.eq.6 .or. (N.ge.8 .and. N.le.12))
     &  ISTAT = 0
      if (ISTAT.ne.0) then
        IERR = max(IERR,ISTAT)
        call WRERR(RTNAME,IERR,'Error in length of CHDG')
        N = min(N,12)
      endif
C
      ISTAT = 0
      if (DDG.lt. -90.D0 .or. DDG.ge. 360.D0) ISTAT = 1
      if (DDG.le.-100.D0 .or. DDG.ge.1000.D0) ISTAT = 2
C
      if (ISTAT.ne.0) then
        IERR = max(IERR,ISTAT)
        call WRERR(RTNAME,IERR,'Error in the input parameter DDG')
      endif
C
C
      if (IERR.lt.2) then
C
C       store sign:
C
        LPLUS = (DDG.ge.0.D0)
        DDGABS = dabs(DDG)
C
C       extract hours, minutes and seconds:
C
        call CNVDMS(DDGABS,LH,IDG,IMIN,RSEC,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        else
          RMIN = RSEC / 60.0
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       round values:
C
        NDIG = max(N-7,0)
        F = real(10**NDIG)
        IMIN1 = nint(RMIN*F)
C
        if ( (IMIN1.eq.10**NDIG) .or.
     &       (N.le.6 .and. (IMIN1.gt.0.5*F .or. RMIN.gt.0.5)) ) then
          RMIN = 0.0
          IMIN1 = 0
          IMIN = IMIN + 1
        endif
C
        if ( (IMIN.ge.60) .or.
     &       (N.eq.3 .and.
     &        (IMIN.gt.30 .or. (IMIN.eq.30 .and. RMIN.gt.0.0))) ) then
          RMIN = 0.0
          IMIN = 0
          IDG = IDG + 1
        endif
C
        if (LUNIT .and. IDG.ge.360) IDG = IDG - 360
C
        if (.not.LPLUS) IDG = -IDG
C
C       assign result to CHDG:
C
        if (N.eq.3) write(CHDG,'(I3.1)') IDG
        if (N.eq.6) write(CHDG,'(I3.1,'':'',I2.2)') IDG,IMIN
        if (N.ge.8) then
           write(CHDIG,'(I1,''.'',I1)') NDIG,NDIG
           CHFMT = '(I3.1,'':'',I2.2,''.'',I' // CHDIG // ')'
           write(CHDG,CHFMT) IDG,IMIN,IMIN1
        endif
C
        if (.not.LPLUS .and. IDG.eq.0) CHDG(2:2)='-'
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CHHR1(DRAD,CHHR,IERR)
C
CC    Converts radians into hours and write a string 'HH:MM:SS.SSSS'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 22-AUG-1989 17:10
CU  update : KOD        date: 28-MAR-1990 19:11
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert radians into a hours and write a string
CG  of the form 'HH:MM:SS.SSSS'
C
C   call_var.          type I/O description 
CP  DRAD                R8  I   angle to be converted (rad,0..2*pi)
CP  CHHR                C   I/O string containing the hours
CP                              'HH:MM:SS.SSSS'
CP                               possible lengths 'L' of CHHR :
CP                               L =  2 : 'HH'
CP                               L =  5 : 'HH:MM'
CP                               L =  8 : 'HH:MM:SS'
CP                               L = 10 : 'HH:MM:SS.S'
CP                               L = 11 : 'HH:MM:SS.SS'
CP                               L = 12 : 'HH:MM:SS.SSS'
CP                               L = 13 : 'HH:MM:SS.SSSS'
CP                               In each case the values
CP                               will be rounded appropriately;
CP                               the value of HH must be positive
CP                               (or zero) and less than 100.
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  CNVDMS              SR      splits decimal degrees
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DRAD, DHR
      REAL*4        RSEC, FSEC, F !HFLAG, 
      INTEGER*4     IERR, IHR, IMIN, N, NDIG, ISEC, ISEC1, ISEC2,
     &              ISTAT ! ,IOS
      LOGICAL*4     LUNIT, LH
      CHARACTER*3   CHDIG
      CHARACTER*33  CHFMT
      CHARACTER*(*) CHHR
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CHHR1 '/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     initialize values:
C
      N = len(CHHR)
      DHR = DRAD * 24.D0 / D2PI
      LUNIT = (DHR.ge.0.D0 .and. DHR.lt.24.D0)
C
C     check input values:
C
      ISTAT = 1
      if (N.eq.2 .or. N.eq.5 .or. N.eq.8 .or. (N.ge.10 .and. N.le.13))
     &  ISTAT = 0
      if (ISTAT.ne.0) then
        IERR = max(IERR,ISTAT)
        call WRERR(RTNAME,IERR,'Error in length of CHHR')
        N = min(N,13)
      endif
C
      ISTAT = 0
      if (DHR.ge.24.D0) ISTAT = 1
      if (DHR.lt.0.D0 .or. DHR.ge.100.D0) ISTAT = 2
C
      if (ISTAT.ne.0) then
        IERR = max(IERR,ISTAT)
        call WRERR(RTNAME,IERR,'Error in the input parameter DHR')
      endif
C
C
      if (IERR.lt.2) then
C
C       extract hours, minutes and seconds:
C
        call CNVDMS(DHR,LH,IHR,IMIN,RSEC,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
      endif
C
C
      if (IERR.lt.2) then
C
C       round values:
C
        NDIG = max(N-9,0)
        F = real(10**NDIG)
        ISEC = nint(RSEC*F)
        FSEC = real(ISEC)/F
        ISEC1 = int(FSEC)
        ISEC2 = ISEC - ISEC1*10**NDIG
        RSEC  = nint(RSEC*10000)*1.E-4
C
        if ( (ISEC1.ge.60) .or.
     &      (N.le.5 .and. (ISEC1.gt.30 .or. RSEC.gt.30.0)) ) then
          RSEC  = 0.0
          ISEC1 = 0
          ISEC2 = 0
          IMIN  = IMIN + 1
        endif
C
        if ( (IMIN.ge.60) .or.
     &       (N.eq.2 .and.
     &        (IMIN.gt.30 .or. (IMIN.eq.30 .and. RSEC.gt.0.0))) ) then
          RSEC  = 0.0
          ISEC1 = 0
          ISEC2 = 0
          IMIN  = 0
          IHR = IHR + 1
        endif
C
        if (N.eq.2.and.(IMIN.gt.30 .or. (IMIN.eq.30 .and. ISEC1.gt.0)))
     &     IHR = IHR + 1
        if (LUNIT .and. IHR.ge.24) IHR = IHR - 24
C
C       assign result to chhr:
C
        if (N.eq.2) write(CHHR,'(I2.2)') IHR
        if (N.eq.5) write(CHHR,'(I2.2,'':'',I2.2)') IHR,IMIN
        if (N.eq.8) write(CHHR,'(I2.2,'':'',I2.2,'':'',I2.2)') IHR,IMIN,
     &                                                         ISEC1
        if (N.ge.10) then
           write(CHDIG,'(I1,''.'',I1)') NDIG,NDIG
           CHFMT = '(I2.2,'':'',I2.2,'':'',I2.2,''.'',I' // CHDIG // ')'
           write(CHHR,CHFMT) IHR,IMIN,ISEC1,ISEC2
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CHHR2(DRAD,CHHR,IERR)
C
CC    Converts radians into hours and writes a string 'HH:MM.MMMMM'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 28-AUG-1989 15:04
CU  update : KOD        date: 28-MAR-1990 19:13
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert radians into a hours and write a string
CG  of the form 'HH:MM.MMMMM'
C
C   call_var.          type I/O description 
CP  DRAD                R8  I   angle to be converted (rad,0..2*pi)
CP  CHHR                C   I/O  string containing the hours
CP                              'HH:MM.MMMMM'
CP                               possible lengths 'L' of CHHR :
CP                               L =  2 : 'HH'
CP                               L =  5 : 'HH:MM'
CP                               L =  7 : 'HH:MM.M'
CP                               L =  8 : 'HH:MM.MM'
CP                               L =  9 : 'HH:MM.MMM'
CP                               L = 10 : 'HH:MM.MMMM'
CP                               L = 11 : 'HH:MM.MMMMM'
CP                               In each case the values
CP                               will be rounded appropriately;
CP                               the value of HH must be positive
CP                               (or zero) and less than 100.
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  CNVDMS              SR      splits decimal degrees
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DRAD, DHR
      REAL*4        RMIN, RSEC, F ! HFLAG, 
      INTEGER*4     IERR, IHR, IMIN, N, NDIG, IMIN1, ISTAT !  ,IOS
      LOGICAL*4     LUNIT, LH
      CHARACTER*3   CHDIG
      CHARACTER*33  CHFMT
      CHARACTER*(*) CHHR
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CHHR2 '/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     initialize values:
C
      N = len(CHHR)
      DHR = DRAD * 24.D0 / D2PI
      LUNIT = (DHR.ge.0.D0 .and. DHR.lt.24.D0)
C
C     check input values:
C
      ISTAT = 1
      if (N.eq.2 .or. N.eq.5 .or. (N.ge.7 .and. N.le.11))
     &  ISTAT = 0
      if (ISTAT.ne.0) then
        IERR = max(IERR,ISTAT)
        call WRERR(RTNAME,IERR,'Error in length of CHHR')
        N = min(N,11)
      endif
C
      ISTAT = 0
      if (DHR.ge.24.D0) ISTAT = 1
      if (DHR.lt.0.D0 .or. DHR.ge.100.D0) ISTAT = 2
C
      if (ISTAT.ne.0) then
        IERR = max(IERR,ISTAT)
        call WRERR(RTNAME,IERR,'Error in the input parameter DHR')
      endif
C
C
      if (IERR.lt.2) then
C
C       extract hours, minutes and seconds:
C
        call CNVDMS(DHR,LH,IHR,IMIN,RSEC,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        else
          RMIN = RSEC / 60.0
        endif
      endif
C
C
      if (IERR.lt.2) then
C
C       round values:
C
        NDIG = max(N-6,0)
        F = real(10**NDIG)
        IMIN1 = nint(RMIN*F)
C
        if ( (IMIN1.eq.10**NDIG) .or.
     &       (N.le.5 .and. (IMIN1.gt.0.5*F .or. RMIN.gt.0.5)) ) then
          RMIN = 0.0
          IMIN1 = 0
          IMIN = IMIN + 1
        endif
C
        if ( (IMIN.ge.60) .or.
     &       (N.eq.2 .and.
     &        (IMIN.gt.30 .or. (IMIN.eq.30 .and. RMIN.gt.0.0))) ) then
          RMIN = 0.0
          IMIN = 0
          IHR = IHR + 1
        endif
C
        if (LUNIT .and. IHR.ge.24) IHR = IHR - 24
C
C       assign result to chhr:
C
        if (N.eq.2) write(CHHR,'(I2.2)') IHR
        if (N.eq.5) write(CHHR,'(I2.2,'':'',I2.2)') IHR,IMIN
        if (N.ge.7) then
           write(CHDIG,'(I1,''.'',I1)') NDIG,NDIG
           CHFMT = '(I2.2,'':'',I2.2,''.'',I' // CHDIG // ')'
           write(CHHR,CHFMT) IHR,IMIN,IMIN1
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CHWDAY(DJD,CHDAY,IERR)
C
CC    Determines the Day of Week out of the Julian Date
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 23-AUG-1989 19:23
CU  update : KOD        date: 28-MAR-1990 19:14
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the Day of Week out of the Julian Date,
CG  according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.27 .
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Date
CP  CHDAY               C   I/O string containing the day of week
CP                              useful lengths of CHDAY: 2,3,9 charact.
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD
c      REAL*4        HFLAG
      INTEGER*4     IERR, IDAY, N, ISTAT !, IOS
      CHARACTER*(*) CHDAY
      CHARACTER*9   CHDAYS(7)
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CHWDAY'/
C
      DATA          CHDAYS / 'Monday   ' , 'Tuesday  ' , 'Wednesday' ,
     &                       'Thursday ' , 'Friday   ' , 'Saturday ' ,
     &                       'Sunday   ' /
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DJD.lt.0.D0 .or. DJD.gt.2999773.5D0) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
      if (IERR.lt.2) then
C
        ISTAT = 1
        N = len(CHDAY)
        if (N.eq.2 .or. N.eq.3 .or. N.eq.9) ISTAT = 0
C
        if (ISTAT.ne.0) then
          IERR = max(IERR,ISTAT)
          call WRERR(RTNAME,IERR,'Warning: unusual length of CHDAY')
        endif
C
        IDAY = int(DJD + 0.5D0)
        IDAY = MOD(IDAY,7) + 1
C
        CHDAY=CHDAYS(IDAY)(1:N)
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CNVDMS(DDEG,LPLUS,IDEG,IMIN,RSEC,IERR)
C
CC    Calculates minutes and seconds out of a decimal number
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 28-MAY-1991 10:46  ( IMIN,RSEC < 60 )
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the minutes and seconds out of a decimal
CG  number (degrees or hours). Negative values are allowed.
C
C   call_var.          type I/O description 
CP  DDEG                R8  I   number to be converted (degr. or hours)
CP  LPLUS               L4    O sign of DDEG ( > 0  if  LPLUS=.true.)
CP  IDEG                I4    O integer part (positive, degr. or hours)
CP  IMIN                I4    O minutes (positive, < 60 )
CP  RSEC                R4    O seconds (positive, < 60 )
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DDEG, DH
      REAL*4        RSEC ! HFLAG, 
      INTEGER*4     IERR, IDEG, IMIN !  ,IOS
      LOGICAL*4     LPLUS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CNVDMS'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (abs(DDEG).gt.1.D3) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Possible error in input parameters')
      endif
C
      LPLUS = (DDEG.ge.0.D0)
C
      DH   = dabs(DDEG)
      IDEG = idint(DH)
      IMIN = idint((DH-dble(IDEG))*60.D0)
      RSEC = sngl(((DH-dble(IDEG))*60.D0-dble(IMIN))*60.D0)
C
      if (RSEC.ge.60.0) then
        RSEC = 0.0
        IMIN = IMIN + 1
      endif
C
      if (IMIN.ge.60) then
        IMIN = 0
        IDEG = IDEG + 1
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CNVFRC(LPLUS,IINT,IM,SEC,DDMS,IERR)
C
CC    Converts minutes and seconds into a decimal fraction
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 20-MAR-1990 14:41
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert minutes and seconds into a decimal fraction
CG  of degrees (or hours) and to append it to the degree (or hour)
CG  value.
CG  The sign has to be given explicitely; the values for IINT,IM,SEC
CG  should be positive, possible signs there will be ignored.
C
C   call_var.          type I/O description 
CP  LPLUS               L4  I   sign (LPLUS=.true. for positive values)
CP  IINT                I4  I   integer part (degrees or hours)
CP  IM                  I4  I   minutes
CP  SEC                 R4  I   seconds
CP  DDMS                R8    O decimal value (degrees or hours)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DDMS, DM, DS, DH
      REAL*4        SEC ! HFLAG, 
      INTEGER*4     IERR, IINT, IM !  ,IOS
      LOGICAL*4     LPLUS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CNVFRC'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      DM = dabs(dble(IM))
      DS = dabs(dble(SEC))
C
      if (dmax1(DM,DS).ge.60.D0) then
C
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in input parameters DM,DS')
C
      else
C
         DH = dabs(dble(IINT)) + ( DM + DS/60.D0 ) / 60.D0
C
         if (LPLUS) then
            DDMS =  DH
         else
            DDMS = -DH
         endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CPREC1(DJD1,DJD2,DLOLD,DBOLD,DLNEW,DBNEW,IERR)
C
CC    Converts equatorial coordinates from one equinox to another
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 26-SEP-1989 15:55
CU  update : KOD        date: 28-MAR-1990 20:03
CU  update : KOD        date: 17-AUG-1992 10:04 (`1.D-6')
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert the equatorial coordinates from one
CG  equinox to another, according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.65,66.
CG
CGT Notes: the effect of the proper motion must be calculated BEFORE
CGT        converting the coordinates to another epoch.
CGT
CGT        The equinox of 1900.0 corresponds to JD 2415020.313,
CGT        the equinox of 1950.0 corresponds to JD 2433282.423,
CGT        according to J.Meeus, p.65 .
CGT        The equinox of 2000.0 corresponds to JD 2451545.000.
C
C   call_var.          type I/O description 
CP  DJD1                R8  I   Julian Date for the initial epoch
CP  DJD2                R8  I   Julian Date for the  final  epoch
CP  DLOLD               R8  I   old longitude  (radians)
CP  DBOLD               R8  I   old latitude   (radians)
CP  DLNEW               R8    O new longitude  (radians)
CP  DBNEW               R8    O new latitude   (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  CVEQUI              SR      calculates quantities for epoch conv.
CR  REPOL1              SR      rectangular to polar coord. conversion
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DJD1, DJD2, DLOLD, DBOLD, DLNEW, DBNEW, DZETA, DZ,
     &              DTHETA, DL, DCL, DSL, DCB, DSB, DCT, DST, DA, DB,
!HB  &              DC, DH1, DR, DPHI, DH2, DRADRS, DJ1, DJ2
     &              DC, DH1, DR, DPHI, DH2, DJ1, DJ2
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CPREC1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DLOLD.lt.-1.D-6 .or. DLOLD.gt.D2PI+1.D-6) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Warning: unusual input value for DLOLD')
      endif
C
      if (dabs(DBOLD).gt.DPID2+1.D-6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameter DBOLD')
      endif
C
      if (IERR.lt.2) then
        DJ1 = DJD1
        DJ2 = DJD2
        call CVEQUI(DJ1,DJ2,DZETA,DZ,DTHETA,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
      if (IERR.lt.2) then
C
        DL = DLOLD + DZETA
        DCL = cos(DL)
        DSL = sin(DL)
C
        DCB = cos(DBOLD)
        DSB = sin(DBOLD)
C
        DCT = cos(DTHETA)
        DST = sin(DTHETA)
C
        DA = DCB * DSL
        DB = DCT * DCB * DCL - DST * DSB
        DC = DST * DCB * DCL + DCT * DSB
C
        DH1 = 0.D0
        call REPOL1(DB,DA,DH1,DR,DPHI,DH2)
        DLNEW = DRADRS(DPHI + DZ)
C
        DBNEW = asin(DC)
        if (abs(DBNEW).gt.0.5D0*DPID2) DBNEW = sign(acos(DR),DBNEW)
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CPREC2(LDIR,DLOLD,DBOLD,DLNEW,DBNEW,IERR)
C
CC    Converts equatorial coordinates between 1950.0 and 2000.0
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 28-SEP-1989 17:10
CU  update : KOD        date: 28-MAR-1990 20:05
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert the equatorial coordinates between
CG  the equinox 1950.0 and 2000.0, according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.65,66.
CG
CG  Notes: the effect of the proper motion must be calculated BEFORE
CG         converting the coordinates to another epoch.
CG
C
C   call_var.          type I/O description 
CP  LDIR                L4  I   direction of the conversion:
CP                                LDIR = .true.  : 1950.0 --> 2000.0
CP                                LDIR = .false. : 1950.0 <-- 2000.0
CP  DLOLD               R8  I   old longitude  (radians)
CP  DBOLD               R8  I   old latitude   (radians)
CP  DLNEW               R8    O new longitude  (radians)
CP  DBNEW               R8    O new latitude   (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  CPREC1              SR      converts equ.coord. for precession
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DLOLD, DBOLD, DLNEW, DBNEW, DJD1, DJD2, DL1, DB1,
     &              DL2, DB2
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      LOGICAL       LDIR
      CHARACTER*8   RTNAME
C
C     Julian Day for equinox 1950.0:
      PARAMETER ( DJD1 = 2433282.423D0 )
C
C     Julian Day for equinox 2000.0:
      PARAMETER ( DJD2 = 2451545.000D0 )
C
      DATA          RTNAME /'CPREC2'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      DL1 = DLOLD
      DB1 = DBOLD
C
      if (LDIR) then
         call CPREC1(DJD1,DJD2,DL1,DB1,DL2,DB2,ISTAT)
      else
         call CPREC1(DJD2,DJD1,DL1,DB1,DL2,DB2,ISTAT)
      endif
C
      if  (ISTAT.gt.1) then
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
      else
        DLNEW = DL2
        DBNEW = DB2
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CPREC3(DZETA,DZ,DTHETA,DLOLD,DBOLD,DLNEW,DBNEW,IERR)
C
CC    Converts equatorial coordinates from one equinox to another
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 08-FEB-1990 14:08
CU  update : KOD        date: 20-MAR-1990 14:42
CU  update : KOD        date: 17-AUG-1992 10:02 (`1.D-6')
C
CT  status: tested
C
C   general description 
CG
CG  Subroutine to convert the equatorial coordinates from one
CG  equinox to another, according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.65,66.
CG
CGT Notes: The effect of the proper motion must be calculated BEFORE
CGT        converting the coordinates to another epoch.
CGT
CGT        This subroutine should be used for converting many
CGT        coordinates at the same time to avoid unneccessary
CGT        re-calculations of the quantities required for the
CGT        equinox conversion.
C
C   call_var.          type I/O description 
CP  DZETA               R8  I   'zeta'   (radians)
CP  DZ                  R8  I   'z'      (radians)
CP  DTHETA              R8  I   'theta'  (radians)
CP  DLOLD               R8  I   old longitude  (radians)
CP  DBOLD               R8  I   old latitude   (radians)
CP  DLNEW               R8    O new longitude  (radians)
CP  DBNEW               R8    O new latitude   (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  REPOL1              SR      rectangular to polar coord. conversion
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DZETA, DZ, DTHETA, DLOLD, DBOLD, DLNEW, DBNEW,
     &              DL, DCL, DSL, DCB, DSB, DCT, DST, DA, DB, DC,
!HB  &              DH1, DR, DPHI, DH2, DRADRS
     &              DH1, DR, DPHI, DH2
c      REAL*4        HFLAG
      INTEGER*4     IERR !,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CPREC3'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (dmax1(dabs(DZETA),dabs(DZ),dabs(DTHETA)).gt.1.D-2 .or.
     &    DLOLD.lt.-1.D-6 .or. DLOLD.gt.D2PI+1.D-6) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Warning: unusual input values')
      endif
C
      if (dabs(DBOLD).gt.DPID2+1.D-6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameter DBOLD')
      endif
C
      if (IERR.lt.2) then
C
        DL = DLOLD + DZETA
        DCL = cos(DL)
        DSL = sin(DL)
C
        DCB = cos(DBOLD)
        DSB = sin(DBOLD)
C
        DCT = cos(DTHETA)
        DST = sin(DTHETA)
C
        DA = DCB * DSL
        DB = DCT * DCB * DCL - DST * DSB
        DC = DST * DCB * DCL + DCT * DSB
C
        DH1 = 0.D0
        call REPOL1(DB,DA,DH1,DR,DPHI,DH2)
        DLNEW = DRADRS(DPHI + DZ)
C
        DBNEW = asin(DC)
        if (abs(DBNEW).gt.0.5D0*DPID2) DBNEW = sign(acos(DR),DBNEW)
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CPROP1(DJD0,DJD,DLOLD,DBOLD,DLP,DBP,DLNEW,DBNEW,IERR)
C
CC    Calculates the change of the equ. coordinates due to proper motion
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 26-SEP-1989 16:50
CU  update : KOD        date: 20-MAR-1990 14:42
CU  update : KOD        date: 17-AUG-1992 10:08 (`1.D-6')
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the effect of proper motion on the
CG  equatorial coordinates. This effect has to be taken into
CG  account BEFORE converting the coordinates to another epoch,
CG  according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.66 .
C
C   call_var.          type I/O description 
CP  DJD0                R8  I   Julian Date for the initial epoch
CP  DJD                 R8  I   Julian Date for the current epoch
CP  DLOLD               R8  I   old longitude  (radians)
CP  DBOLD               R8  I   old latitude   (radians)
CP  DLP                 R8  I   annual proper motion in R.A. (radians)
CP  DBP                 R8  I   annual proper motion in Dec. (radians)
CP  DLNEW               R8    O new longitude  (radians)
CP  DBNEW               R8    O new latitude   (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DJD0, DJD, DLOLD, DBOLD, DLP, DBP, DLNEW, DBNEW, DT
c      REAL*4        HFLAG
      INTEGER*4     IERR ! ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CPROP1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DLOLD.lt.-1.D-6 .or. DLOLD.gt.D2PI+1.D-6) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Warning: unusual input value for DLOLD')
      endif
C
      if (dabs(DBOLD).gt.DPID2+1.D-6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in the input parameter DBOLD')
      endif
C
      if ( max( abs(DLP*cos(DBOLD)), abs(DBP) ).gt.1.D-4) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameters DLP, DBP')
      endif
C
      if (min(DJD0,DJD).lt.0.D0 .or. max(DJD0,DJD).gt.3.D6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
      if (IERR.lt.2) then
        DT = (DJD - DJD0) / 365.242199D0
        DLNEW = DLOLD + DT * DLP
        DBNEW = DBOLD + DT * DBP
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine CVEQUI(DJD1,DJD2,DZETA,DZ,DTHETA,IERR)
C
CC    Calculates quantities needed for equinox conversions
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 26-SEP-1989 15:09
CU  update : KOD        date: 20-MAR-1990 14:44
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the quantities needed for the accurate
CG  reduction of positions from one equinox to another,
CG  according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.65 .
C
C   call_var.          type I/O description 
CP  DJD1                R8  I   Julian Date for the initial epoch
CP  DJD2                R8  I   Julian Date for the  final  epoch
CP  DZETA               R8    O 'zeta'   (radians)
CP  DZ                  R8    O 'z'      (radians)
CP  DTHETA              R8    O 'theta'  (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD1, DJD2, DZETA, DZ, DTHETA, DJD0, DYEAR,
     &              DT0, DT, DH, DF
c      REAL*4        HFLAG
      INTEGER*4     IERR ! ,IOS
      CHARACTER*8   RTNAME
C
      PARAMETER ( DJD0  = 2415020.313D0  )
      PARAMETER ( DYEAR =  36524.2199D0  )
      PARAMETER ( DF    = 4.848136811D-6 )
C
      DATA          RTNAME /'CVEQUI'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (min(DJD1,DJD2).lt.0.D0 .or. max(DJD1,DJD2).gt.3.D6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
      if (IERR.lt.2) then
C
        DT0 = ( DJD1 - DJD0 ) / DYEAR
        DT  = ( DJD2 - DJD1 ) / DYEAR
C
        DH = 2304.250D0 + 1.396D0 * DT0
        DZETA = ( DH + ( 0.302D0 + 0.018D0 * DT ) * DT ) * DT
C
        DZ = DZETA + ( 0.791D0 + 0.001D0 * DT ) * DT * DT
C
        DH = 2004.682D0 - 0.853D0 * DT0
        DTHETA = ( DH - ( 0.426D0 + 0.042D0 * DT ) * DT ) * DT
C
        DZETA  = DZETA  * DF
        DZ     = DZ     * DF
        DTHETA = DTHETA * DF
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine DATEJD(DT,LCHECK,DJD,IERR)
C
CC    Calculates the Julian Date out of the Calendar Date
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 22-AUG-1989 18:31
CU  update : KOD        date: 28-MAR-1990 20:11
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the Julian Date out of the calendar date.
CG  The calendar date is assumed to be given in the Gregorian system
CG  for dates starting with October 15, 1582, and in the Julian
CG  system until October 4, 1582 (in between it is undefined).
CG
CG  This subroutine is valid for the years  -4712  to  +3500 ,
CG  and has been written according to
CG     van Flandern and Pulkkinen:
CG     'Low-Precision Formulae for Planetary Positions',
CG     Astrophysical Journal Supplement Series 41, 1979, p.392.
CG  and
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.23-25 .
CG
CG  An extended input check is possible for valid dates.
C
C   call_var.          type I/O description 
CP  DT                  R8  I   Calendar Date  (YYYYMMDD.HHMMSSS)
CP  LCHECK              L4  I   extended input check if LCHECK=.TRUE.
CP  DJD                 R8    O Julian Date
CP  IERR                I4    O = 0 no error
CP                              = 1 warning (unusual input values)
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  YMDHMS              SR      splits the variable DT into its parts
CR  CALJD1              SR      converts the Calendar Date to J.D.
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DT, DH, DJD
      REAL*4        SEC ! HFLAG, 
      INTEGER*4     IERR, IY, IM, ID, IHR, IMT, ISTAT ! , IOS
      LOGICAL*4     LCHECK, LCHK
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'DATEJD'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR=0
C
      DH = DT
      CALL YMDHMS(DH,IY,IM,ID,IHR,IMT,SEC,ISTAT)
CN!   ERROR handling
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      LCHK=LCHECK
      CALL CALJD1(IY,IM,ID,IHR,IMT,SEC,LCHK,DH,IERR)
CN!       ERROR handling
          if  (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,20)
          endif
C
      DJD=DH
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      real*8 function DDEGRS(DDEG)
C
CC    'Resets' degrees to the interval  0 .. 360
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:34
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 28-MAR-1990 20:16
C
CT  status: tested
C
C   general description 
CG  'Resets' degrees to the interval  0 .. 360.
C
C   call_var.          type I/O description 
CP  DDEG                R8  I   degrees (may be negative)
CP  DDEGRS              R8    O degrees in the interval  0 .. 360
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  RESVAL              SR      'resets' periodic variables
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DDEG, DXMIN, DXMAX, DXOLD, DXNEW
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'DDEGRS'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      DXMIN =   0.D0
      DXMAX = 360.D0
      DXOLD = DDEG
C
      CALL RESVAL(DXMIN,DXMAX,DXOLD,DXNEW,ISTAT)
C
      if (ISTAT.gt.1) then
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
      endif
C
      if (IERR.lt.2) DDEGRS = DXNEW
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      real*8 function DHRRES(DHR)
C
CC    'Resets' hours to the interval  0 .. 24
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 28-MAR-1990 20:18
C
CT  status: tested
C
C   general description 
CG  'Resets' hours to the interval  0 .. 24.
C
C   call_var.          type I/O description 
CP  DHR                 R8  I   hours (may be negative)
CP  DHRRES              R8    O hours inside the interval  0 .. 24
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  RESVAL              SR      'resets' periodic variables
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DHR, DXMIN, DXMAX, DXOLD, DXNEW
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'DHRRES'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      DXMIN =  0.D0
      DXMAX = 24.D0
      DXOLD = DHR
C
      CALL RESVAL(DXMIN,DXMAX,DXOLD,DXNEW,ISTAT)
C
      if (ISTAT.gt.1) then
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
      endif
C
      if (IERR.lt.2) DHRRES = DXNEW
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine DMMDEC(DDMMM,DDEG,IERR)
C
CC    Calc. the value of a number given as 'DDD.MMMM' or 'HH.MMMM'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 25-AUG-1989 14:48
CU  update : KOD        date: 20-MAR-1990 14:48
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the value of a number given in the form
CG  'DDD.MMM' or 'HH.MMM'. Negative values are allowed.
C
C   call_var.          type I/O description 
CP  DDMMM               R8  I   number to be converted (-360 .. 360)
CP                              (DDD.MMM or HH.MMM)
CP  DDEG                R8    O corresponding decimal value
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DDMMM, DDEG, DMM
c      REAL*4        HFLAG
      INTEGER*4     IERR, IDEG, IMT, IMT1 !  ,IOS
      LOGICAL*4     LPLUS
      CHARACTER*8   RTNAME
      CHARACTER*17  CHDEG17
C
      DATA          RTNAME /'DMMDEC'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      LPLUS = (DDMMM.ge.0.D0)
C
      DMM = dabs(DDMMM)
      if (DMM.gt.360.D0) then
        IERR = 1
        if (DMM.ge.1.D6) IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameter DDMMM')
      endif
C
      if (IERR.lt.2) then
C
C       split input value:
C
        write(CHDEG17,'(F17.10)') DMM
C
C       read individual parts:
C
        read(CHDEG17,'(I6,1X,I2,I8)') IDEG, IMT, IMT1
C
        if (IMT.ge.60) then
          IERR = max(IERR,2)
          call WRERR(RTNAME,IERR,'Error in input parameter DDMMM')
        else
          DDEG = dble(IDEG) + dble(IMT)/60.D0 + dble(IMT1)/60.D8
          if (.not.LPLUS) DDEG = -DDEG
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine DMMRAD(DDMMM,DRAD,IERR)
C
CC    Calculates radians of a number given in the form 'DDD.MMMM'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 05-SEP-1989 19:17
CU  update : KOD        date: 28-MAR-1990 20:23
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert a number given in the form 'DDD.MMMM'
CG  (degrees) into radians. Negative values are allowed.
C
C   call_var.          type I/O description 
CP  DDMMM               R8  I   number to be converted (-360 .. 360)
CP                              (DDD.MMMM)
CP  DRAD                R8    O corresponding radians
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  DMMDEC              SR      converts 'DDD.MMMM' into dec. degrees
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DDMMM, DRAD, DDEG, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'DMMRAD'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      DH = DDMMM
      call DMMDEC(DH,DDEG,ISTAT)
C
      if  (ISTAT.gt.1) then
C
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
C
      else
C
        DRAD = DDEG * DPI180
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine DMSDEC(DDMMSS,DDEG,IERR)
C
CC    Calc. the value of a number given as 'DDD.MMSSS' or 'HH.MMSSS'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 22-AUG-1989 15:48
CU  update : KOD        date: 20-MAR-1990 14:48
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the value of a number given in the form
CG  'DDD.MMSSS' or 'HH.MMSSS'. Negative values are allowed.
C
C   call_var.          type I/O description 
CP  DDMMSS              R8  I   number to be converted (-360 .. 360)
CP                              (DDD.MMSSS or HH.MMSSS)
CP  DDEG                R8    O corresponding decimal value
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DDMMSS, DDEG, DMS
c      REAL*4        HFLAG
      INTEGER*4     IERR, IDEG, IMT, ISEC1, ISEC2 !  ,IOS
      LOGICAL*4     LPLUS
      CHARACTER*8   RTNAME
      CHARACTER*19  CHDEG19
C
      DATA          RTNAME /'DMSDEC'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      LPLUS = (DDMMSS.ge.0.D0)
C
      DMS = dabs(DDMMSS)
      if (DMS.gt.360.D0) then
        IERR = 1
        if (DMS.ge.1.D6) IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameter DDMMSS')
      endif
C
      if (IERR.lt.2) then
C
C       split input value:
C
        write(CHDEG19,'(F19.12)') DMS
C
C       read individual parts:
C
        read(CHDEG19,'(I6,1X,I2,I2,I8)') IDEG, IMT, ISEC1, ISEC2
C
        if (max(IMT,ISEC1).ge.60) then
          IERR = max(IERR,2)
          call WRERR(RTNAME,IERR,'Error in input parameter DDMMSS')
        else
          DDEG = dble(IDEG) + dble(IMT)/60.D0 + dble(ISEC1)/3600.D0
     &                                        + dble(ISEC2)/3600.D8
          if (.not.LPLUS) DDEG = -DDEG
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine DMSRAD(DDMMSS,DRAD,IERR)
C
CC    Calculates radians of a number given in the form 'DDD.MMSS'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 05-SEP-1989 18:32
CU  update : KOD        date: 20-MAR-1990 14:49
CU  update : KOD        date: 28-MAR-1990 20:24
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert a number given in the form 'DDD.MMSSS'
CG  (degrees) into radians. Negative values are allowed.
C
C   call_var.          type I/O description 
CP  DDMMSS              R8  I   number to be converted (-360 .. 360)
CP                              (DDD.MMSSS)
CP  DRAD                R8    O corresponding radians
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  DMSDEC              SR      converts 'DDD.MMSS' into dec. degrees
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DDMMSS, DRAD, DDEG, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'DMSRAD'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      DH = DDMMSS
      call DMSDEC(DH,DDEG,ISTAT)
C
      if  (ISTAT.gt.1) then
C
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
C
      else
C
        DRAD = DDEG * DPI180
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine DOYYMD(IY,IDOY,IM,ID,IERR)
C
CC    Calculates the Month and Day from the Year and Day of the Year
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 23-AUG-1989 20:04
CU  update : KOD        date: 28-MAR-1990 15:49
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the Month and Day from the Year and
CG  Day of the Year. The Day of the Year is assumed to be given
CG  in the Gregorian system for dates starting with October 15,
CG  1582, and in the Julian system until October 4, 1582 (in
CG  between it is undefined). For the year 1582 this subroutine
CG  will count the ACTUAL days of the year (which has 355 days)
CG  and will give a warning message for days after October 4.
CG
CG  This subroutine is valid for the years -4712  to  +3500 ,
CG  and has been written according to
CG     van Flandern and Pulkkinen:
CG     'Low-Precision Formulae for Planetary Positions',
CG     Astrophysical Journal Supplement Series 41, 1979, p.392.
CG  and
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.23-25 .
CG
CG  The input date will be checked for validity.
C
C   call_var.          type I/O description 
CP  IY                  I4  I   year         (-4712 .. +3500)
CP  IDOY                I4  I   day of year  (1 .. 365,366)
CP  IM                  I4    O month        (1 .. 12)
CP  ID                  I4    O day of month (1 .. 28,29,30,31)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  CALJD1              SR      converts Calendar Date into Julian Date
CR  JDCAL1              SR      converts Julian Date into Calendar Date
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD
      REAL*4        sec ! HFLAG, 
      INTEGER*4     IERR, IY, IDOY, IM, ID, IYEAR, IDAY, IMONTH,
     &              IHR, IMT, ISTAT ! , IOS
      LOGICAL*4     LCHECK
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'DOYYMD'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     rough input check:
C
      if (IDOY.gt.366) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Day of Year > 366')
      endif
C
      if (IY.eq.1582 .and. IDOY.gt.277) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,
     &      'Warning: Calendar Reform --> Day of the Year not unique')
      endif
C
      if (IERR.lt.2) then
C
C       calculate the Julian Date for the first day of the year:
C
        IYEAR  = IY
        IMONTH = 1
        IDAY   = 1
        IHR    = 12
        IMT    = 0
        SEC    = 0.0
        LCHECK = .true.
C
        call CALJD1(IYEAR,IMONTH,IDAY,IHR,IMT,SEC,LCHECK,DJD,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
      if (IERR.lt.2) then
C
C       add Day of Year:
C
        DJD = DJD + dble(IDOY-1)
C
C       calculate the Calendar Date:
C
        call JDCAL1(DJD,IYEAR,IMONTH,IDAY,IHR,IMT,SEC,IERR)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        else
          if (IYEAR.eq.IY) then
            IM = IMONTH
            ID = IDAY
          else
            IERR = max(IERR,2)
            call WRERR(RTNAME,IERR,'No leap year')
          endif
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine EANOM1(DMANOM,DECC,DEANOM,IERR)
C
CC    Calculates the eccentric anomaly from the mean anomaly
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 31-AUG-1989 13:57
CU  update : KOD        date: 20-MAR-1990 14:50
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the eccentric anomaly from the mean
CG  anomaly according to the second method described by
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.94-95 .
CG
CG  This subroutine has been tested with a random number generator
CG  used to produce random mean anomalies between 0.0 and 2*pi,
CG  and random eccentricities:
CG
CGT   DECC = 0.90 .. 0.97   n = 100000  It = 5.8  7.6 CPU sec (IBM)
CGT   DECC = 0.00 .. 0.97   n = 100000  It = 4.7  6.8 CPU sec (IBM)
CG
CG  In every case the algorithm converged to an accurracy of 1.E-12
CG  in the eccentric anomaly; above DECC = 0.97 errors occurred.
C
C   call_var.          type I/O description 
CP  DMANOM              R8  I   mean anomaly  (radians, 0 .. 2*pi)
CP  DECC                R8  I   numeric eccentricity (0.0 .. 0.97)
CP  DEANOM              R8    O eccentric anomaly (radians, 0 .. 2*pi)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
!HB   REAL*8        DMANOM, DECC, DEANOM, DACC, DE, DH, DRADRS
      REAL*8        DMANOM, DECC, DEANOM, DACC, DE, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, MAXIT, IT !  ,IOS
      LOGICAL*4     LCONV
      CHARACTER*8   RTNAME
C
      PARAMETER ( DACC = 1.D-12 )
C
      PARAMETER ( MAXIT = 100 )
C
      DATA          RTNAME /'EANOM1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DMANOM.lt.0.d0 .or. DMANOM.gt.D2PI) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Input parameter DMANOM out of range')
      endif
C
      if (DECC.lt.0.D0 .or. DECC.ge.0.97D0) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Input parameter DECC out of range')
      endif
C
      if (IERR.lt.2) then
C
        IT = 0
        LCONV = .false.
        DE = DMANOM
C
        do while (.not.LCONV .and. IT.lt.MAXIT)
          IT = IT + 1
          DH = (DMANOM + DECC*sin(DE) - DE) / (1.D0 - DECC*cos(DE))
          DE = DE + DH
          LCONV = (abs(DH).lt.DACC)
        end do
C
        if (.not.LCONV) then
          IERR = max(IERR,2)
          call WRERR(RTNAME,IERR,'No sufficient convergence')
        else
          DEANOM = DRADRS(DE)
        endif
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine EQUECL(DJD,LMEAN,LDIR,DLOLD,DBOLD,DLNEW,DBNEW,IERR)
C
CC    Converts coordinates between equatorial and ecliptical system
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 29-MAR-1990 15:13
CU  update : KOD        date: 17-AUG-1992 10:09 (`1.D-6')
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert coordinates between the equatorial and the
CG  ecliptical system (LII,BII).
CG
CG  To convert the TRUE positions, use the MEAN obliquity of the
CG  ecliptic (normal case);
CG  for the APPARENT positions (affected by aberration and nutation)
CG  use the TRUE obliquity.
CG  If the equatorial coordinates refer to a standard equinox
CG  (1950, 2000, ...), the obliquity of the ecliptic for this
CG  epoch should be used.
CG
CG  (Jean Meeus, 'Astronomical Formulae for Calculators',
CG                third edition, 1985, p.43-44)
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Day for the conversion
CP  LMEAN               L4  I   mean or true obliquity of the ecliptic:
CP                                LMEAN = .true.  : mean obliquity
CP                                LMEAN = .false. : true obliquity
CP  LDIR                L4  I   transformation direction:
CP                                LDIR = .true.  : equ --> ecl
CP                                LDIR = .false. : equ <-- ecl
CP  DLOLD               R8  I   old longitude  (radians)
CP  DBOLD               R8  I   old latitude   (radians)
CP  DLNEW               R8    O new longitude  (radians)
CP  DBNEW               R8    O new latitude   (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  EULER1              SR      general coordinate transformation
CR  JDEPS1              SR      calculates mean/true obl. of the ecl.
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DJD, DLOLD, DBOLD, DLNEW, DBNEW, D1, D2, D3,
     &              DEPS, DPHI1, DTHETA1, DPHI2, DTHETA2, DJ
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      LOGICAL*4     LMEAN, LDIR, LMN
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'EQUECL'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C
      if (DLOLD.lt.-1.D-6 .or. DLOLD.GT.D2PI+1.D-6) then
        IERR = 1
        call WRERR(RTNAME,IERR,'Warning: unusual value for DLOLD')
      endif
C
      if (dabs(DBOLD).gt.DPID2+1.D-6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameter DBOLD')
      endif
C
      if (DJD.lt.0.D0 .or. DJD.gt.3.D6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
C
      if  (IERR.lt.2)  then
C
        DJ = DJD
        LMN = LMEAN
        call JDEPS1(DJ,LMN,DEPS,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
C
      if  (IERR.lt.2)  then
C
        D1 = 0.D0
        D2 = DEPS
        D3 = 0.D0
C
        DPHI1   = DLOLD
        DTHETA1 = DPID2 - DBOLD
C
        if (LDIR) then
          call EULER1(D1,D2,D3,DPHI1,DTHETA1,DPHI2,DTHETA2,ISTAT)
        else
          call EULER1(-D3,-D2,-D1,DPHI1,DTHETA1,DPHI2,DTHETA2,ISTAT)
        endif
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        else
          DLNEW = DPHI2
          DBNEW = DPID2 - DTHETA2
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine EQUGAL(DJD,LDIR,DLOLD,DBOLD,DLNEW,DBNEW,IERR)
C
CC    Converts coordinates between equatorial and new galactic system
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 19-JUL-1989 19:45
CU  update : KOD        date: 20-MAR-1990 14:56
CU  update : KOD        date: 17-AUG-1992 10:10 (`1.D-6')
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert coordinates between the equatorial and the
CG  new galactic system (LII,BII).
CG  Parameters extracted from:
CG     F.C.Bertiau, 'Programmes for pocket calculators HP-67
CG                   and HP-97 in the field of theoretical and
CG                   observational astronomy',
CG                   1977 Leuven University Press, 'ASTR. COOR. 2'
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Day for the conversion
CP  LDIR                L4  I   transformation direction:
CP                                LDIR = .true.  : equ --> gal II
CP                                LDIR = .false. : equ <-- gal II
CP  DLOLD               R8  I   old longitude  (radians)
CP  DBOLD               R8  I   old latitude   (radians)
CP  DLNEW               R8    O new longitude  (radians)
CP  DBNEW               R8    O new latitude   (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  EULER1              SR      general coordinate transformation
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C

      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DJD, DLOLD, DBOLD, DLNEW, DBNEW, D11, D21, D31,
     &              D12, D22, D32, D1, D2, D3, DJD50, DPHI1, DTHETA1,
     &              DPHI2, DTHETA2, DT
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      LOGICAL*4     LDIR
      CHARACTER*8   RTNAME
C
C     julian day for 1950.01.01 / 00:00:00
      PARAMETER ( DJD50 = 2433282.5D0 )
C
      PARAMETER ( D11 = 4.92619181375D0 )
      PARAMETER ( D21 = 1.09257611174D0 )
      PARAMETER ( D31 = 5.70722665400D0 )
C
      PARAMETER ( D12 = 5.82362268000D-7 )
      PARAMETER ( D22 = 2.59857718339D-7 )
      PARAMETER ( D32 = 0.64223753852D-7 )
C
      DATA          RTNAME /'EQUGAL'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DLOLD.lt.-1.D-6 .or. DLOLD.GT.D2PI+1.D-6) then
        IERR = 1
        call WRERR(RTNAME,IERR,'Warning: unusual value for DLOLD')
      endif
C
      if (dabs(DBOLD).gt.DPID2+1.D-6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameter DBOLD')
      endif
C
      if (IERR.lt.2) then
C
        DT = DJD - DJD50
C
        ISTAT = 0
        if (dabs(DT).gt.5.D4) ISTAT = 1
        if (dabs(DT).gt.5.D6) ISTAT = 2
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call WRERR(RTNAME,IERR,'Error in input parameter DJD')
        endif
C
      endif
C
      if  (IERR.lt.2)  then
C
        D1 = D11 + D12 * DT
        D2 = D21 + D22 * DT
        D3 = D31 + D32 * DT
C
        DPHI1   = DLOLD
        DTHETA1 = DPID2 - DBOLD
C
        if (LDIR) then
          call EULER1(D1,D2,D3,DPHI1,DTHETA1,DPHI2,DTHETA2,ISTAT)
        else
          call EULER1(-D3,-D2,-D1,DPHI1,DTHETA1,DPHI2,DTHETA2,ISTAT)
        endif
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        else
          DLNEW = DPHI2
          DBNEW = DPID2 - DTHETA2
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine ERRBX1(DL0,DB0,DBW,DBH,DBT,DDL,DDB,IERR)
C
CC    Calculates the coordinates of the corners of an error box
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 28-JUN-1990 20:34
CU  update : KOD        date: 26-OCT-1990 18:23 (no warning in EULER1)
CU  update : KOD        date: 17-AUG-1992 10:11 (`1.D-6')
C
CT  status: tested
C
C   general description 
CG
CG  Subroutine to calculate the coordinates of the corners of an
CG  error box which is specified by its width, height, tilt, and
CG  the coordinates of its center. The height parameter determines
CG  the extent of the `vertical' boundaries, which are part of a
CG  great circle on the sphere, while the `horizontal' boundaries
CG  specified by the width parameter are determined as small circles
CG  of equal length on the sphere. The tilt of the error box is
CG  the angle between the symmetry plane of the width boundaries
CG  and the declination line (DB0) running through the center of
CG  the error box, measured counterclockwise. Note that the dimensions
CG  of the source box are required in `true' degrees (i.e., the values
CG  are independent of the declination) to be consistent with any
CG  tilts. The calculation is done in full spherical geometry.
C
C   call_var.          type I/O description 
CP  DL0                 R8  I   Right Ascension of the center  (radians)
CP  DB0                 R8  I   Declination of the center      (radians)
CP  DBW                 R8  I   full width  of the error box   (radians)
CP  DBH                 R8  I   full height of the error box   (radians)
CP  DBT                 R8  I   tilt of the error box          (radians)
CP  DDL                 R8    O Right Ascension of the corners (radians)
CP  DDB                 R8    O Declination     of the corners (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  EULER1              SR      transformation of spherical coordinates
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to flag output stream
CR  DRADRS              R8      'resets' radians
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DL0, DB0, DBW, DBH, DBT, DDL(4), DDB(4), D1, D2, D3,
!HB  &              DPH1, DTH1, DPH2, DTH2, DW, DH, DRADRS
     &              DPH1, DTH1, DPH2, DTH2, DW, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT, I ! , IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'ERRBX1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     check input values:
C
      if (max(abs(DL0),DBW,abs(DBT)).gt.D2PI+1.D-6) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Unusual input values')
      endif
C
      if ( abs(DB0).gt.DPID2+1.D-6 .or. DBH.gt.DPI .or.
     &     DBW.lt.0.D0 .or. DBH.lt.0.D0 ) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input values')
      endif
C
C
      if (IERR.lt.2) then
C
        do I = 1, 4
C
          if (I.le.2) then
            DW = D2PI - 0.5D0 * DBW
          else
            DW = 0.5D0 * DBW
          endif
C
          if (I.eq.1 .or. I.eq.4) then
            DH =  0.5D0 * DBH
          else
            DH = -0.5D0 * DBH
          endif
C
          D1 = 0.D0
          D2 =  DBT
          D3 = 0.D0
C
          DPH1 = DRADRS(DW)
          DTH1 = DPID2 - DH
C
          CALL EULER1(D1,D2,D3,DPH1,DTH1,DPH2,DTH2,ISTAT)
C
          if (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,10)
          endif
 
          if (IERR.lt.2) then
C
            D1 =  DPID2
            D2 =  DB0
            D3 = DRADRS(-DPID2-DL0)
C
            DPH1 = DPH2
            DTH1 = DTH2
C
            CALL EULER1(D1,D2,D3,DPH1,DTH1,DPH2,DTH2,ISTAT)
C
            if (ISTAT.gt.1) then
              IERR = max(IERR,ISTAT)
              call ERRSTC(RTNAME,20)
            else
              DDL(I) = DPH2
              DDB(I) = DPID2 - DTH2
            endif
C
          endif
C
        enddo
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine ETANOM(DECC,LDIR,DANOM1,DANOM2,IERR)
C
CC    Converts between the eccentric and true anomaly
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : KOD        date: 25-SEP-1989 13:29
CU  update : KOD        date: 20-MAR-1990 15:40
C
CT  status: tested
C
C   general description 
CG    Subroutine to convert between the eccentric and true anomaly
CG    for an elliptic orbit, according to
CG       Jean Meeus, 'Astronomical Formulae for Calculators',
CG                   third edition, 1985, p.122 .
C
C   call_var.          type I/O description 
CP  DECC                R8  I   eccentricity  ( 0.0 <= DECC < 1.0 )
CP  LDIR                L4  I   conversion direction:
CP                                LDIR = .true.  : ecc --> true anom.
CP                                LDIR = .false. : ecc <-- true anom.
CP  DANOM1              R8  I   'old' anomaly  (radians)
CP  DANOM2              R8    O 'new' anomaly  (radians, 0 .. 2 * pi )
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DECC, DANOM1, DANOM2, DH1, DH2, DHS, DHC, DH
!HB  &              DRADRS
c      REAL*4        HFLAG
      INTEGER*4     IERR ! ,IOS
      LOGICAL*4     LDIR
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'ETANOM'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (dabs(DANOM1).gt.12.56D0) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Error in anomaly')
      endif
C
      if (DECC.lt.0.D0 .or. DECC.ge.1.D0) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in eccentricity')
      endif
C
      if (IERR.lt.2) then
C
        if (LDIR) then
          DH1 = sqrt ( (1.D0 + DECC) / (1.D0 - DECC) )
        else
          DH1 = sqrt ( (1.D0 - DECC) / (1.D0 + DECC) )
        endif
C
        DH2 = 0.5D0 * DANOM1
C
        DHS = sin(DH2)
        DHC = cos(DH2)
C
        DH = 2.D0 * atan2(DH1*DHS,DHC)
C
        DANOM2 = DRADRS(DH)
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine EULER1(D1,D2,D3,DPH1,DTH1,DPH2,DTH2,IERR)
C
CC    General transformation of spherical coordinates
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : KOD        date: 15-JUL-1989 18:31
CU  update : KOD        date: 20-MAR-1990 14:57
CU  update : KOD        date: 17-AUG-1992 09:38 (`1.D-6')
C
CT  status: tested
C
C   general description 
CG
CG  Subroutine to convert spherical coordinates between coordinate
CG  systems which are rotated with respect to each other.
CG
CG  This subroutine converts the azimuthal and polar angles
CG  (DPH1,DTH1), given in mathematical notation and in radians,
CG  into the corresponding angles with respect to a coordinate
CG  system which orientation is given by the EULER angles (D1,D2,D3)
CG  in the following way:
CG
CG  The original cartesian coordinate system (X0,Y0,Z0) is rotated
CG  in three steps:
CG
CGT   1) rotation around the Z0-axis by the angle D1  --->  (X1,Y1,Z1)
CGT   2) rotation around the X1-axis by the angle D2  --->  (X2,Y2,Z2)
CGT   3) rotation around the Z2-axis by the angle D3  --->  (X3,Y3,Z3)
CG
CG  The calculation is accelerated by using differences of angles.
C
C   call_var.          type I/O description 
CP  D1                  R8  I   EULER angle 1   (radians, -2 pi .. 2 pi)
CP  D2                  R8  I   EULER angle 2   (radians, -2 pi .. 2 pi)
CP  D3                  R8  I   EULER angle 3   (radians, -2 pi .. 2 pi)
CP  DPH1                R8  I   original  azimuthal angle (rad, 0..2 pi)
CP  DTH1                R8  I   original    polar   angle (rad, 0..pi)
CP  DPH2                R8    O resulting azimuthal angle (rad, 0..2 pi)
CP  DTH2                R8    O resulting   polar   angle (rad, 0..pi)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  REPOL1              SR      cartesian to polar coordinate conversion
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        D1, D2, D3, DPH1, DPH2, DTH1, DTH2,
     &              DST, DCT, DH, DS1, DC1, DS2, DC2, DX, DY, DZ,
!HB  &              DR, DPHI, DTHETA, DRADRS
     &              DR, DPHI, DTHETA
c      REAL*4        HFLAG
      INTEGER*4     IERR ! ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'EULER1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (max(dabs(D1),dabs(D2),dabs(D3)).gt.D2PI .or.
     &    DPH1.lt.0.D0) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Warning: Unusual input parameters')
      endif
C
      if (DTH1.lt.-1.D-6 .or. dabs(DTH1).gt.DPI+1.D-6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameters')
      else
C
        DST = sin(DTH1)
        DCT = cos(DTH1)
C
        DH  = DPH1 - D1
        DS1 = sin(DH)
        DC1 = cos(DH)
C
        DS2 = sin(D2)
        DC2 = cos(D2)
C
        DX = DST*DC1
        DY = DST*DS1*DC2 + DCT*DS2
        DZ =-DST*DS1*DS2 + DCT*DC2
C
        call REPOL1(DX,DY,DZ,DR,DPHI,DTHETA)
C
        if (DABS(DR-1.D0).gt.1.E-6) then
          IERR = max(IERR,2)
        else
          DPH2 = DRADRS(DPHI - D3)
          DTH2 = DTHETA
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      SUBROUTINE FLDCEN(NZONE,IFIELD,NOMRA,NOMDE,IERR)
C
CC  calculate center of FIELD
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : GRU        date: 25-OCT-1989 16:53
CU  update : GRU        date: 10-AUG-1994 17:57
CU  update : CRI        date: 11-AUG-1994 15:14 reset IERR
C
CT  status: not tested
C
C   general description 
CG
C
C   call_var.          type I/O description 
CP  NZONE              II4  ?/? # of zones used for field definition
CP  IFIELD             II4  ?/? field#
CP  NOMDE              II4  ?/? nominal RA    [0.5 arcsec]
CP  NOMRA              II4  ?/? nominal DEC   [0.5 arcsec]
CP  IERR                I4    O 
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
CR  ERRSTC              SR      
CR  FLDCNV              SR      convert field# to latitude/longitude
CR                              zones and vice versa
CR  ZONCEN              SR      determine center of field
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
c      REAL*4        HFLAG
      INTEGER*4     IERR,IFIELD,NOMRA,NOMDE,ISTAT,LAT,LON,NZONE ! ,IOS
      CHARACTER     RTNAME*8,FLAG*6
C
      DATA          RTNAME /'FLDCEN'/
C
CD     IF (HFLAG(RTNAME,'in: SEQ# ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) IFIELD
CD 101   FORMAT ( I)
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C**********************************************************************
C generate latitude and longitude zones from field#
      FLAG = 'TOZONE'
      CALL FLDCNV(LAT,LON,IFIELD,FLAG,ISTAT)
      IF (ISTAT.GT.1) THEN
         IERR=3
         CALL ERRSTC(RTNAME,2)
      ELSE
         CALL ZONCEN(NZONE,LAT,LON,NOMRA,NOMDE,ISTAT)
         IF (ISTAT.GT.1) THEN
            IERR=3
            CALL ERRSTC(RTNAME,2)
         ENDIF
      ENDIF
C
CD     IF (HFLAG(RTNAME,'out: IERR').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) IERR
CD9901   FORMAT (I)
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
      RETURN
      END
      SUBROUTINE FLDCNV(LAT,LON,NFLD,FLAG,IERR)
C
CC  de/construct field# from/into latitude zone and and longitude zone
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : GRU        date: 9-JUN-1994 15:43
CU  update : GRU        date: 10-AUG-1994 17:42
CU  update : GRU        date: 10-AUG-1994 17:43
CU  update : GRU        date: 10-AUG-1994 17:59
C
CT  status: not tested
C
C   general description 
CG  dependent on FLAG the field# is deconstructed into the latitude and
CG  longitude zone # or the field# is constructed from both zone numbers
C
C   call_var.          type I/O description 
CP  IERR                I4    O = 0 no error
CP                              = 3
CP  FLAG                C   I   = 'TOFLD': generate field# from latidude
CP                                         and longitude zone
CP                              = 'TOZONE':generate latitude & longitude
CP                                         zone from field#
CP  LAT                 I4  I/O latitude zone
CP  LON                 I4  I/O longitude zone
CP  NFLD                I4    O field#
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameters and buffers
C
C   routines_called    type     description 
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
CR  WRERR               SR      
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
c      REAL*4        HFLAG
      INTEGER*4     IERR, LAT,LON,NFLD ! ,IOS
      CHARACTER     FLAG*(*)
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'FLDCNV'/
C
CD     IF (HFLAG(RTNAME,'in: FLAG').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) FLAG
CD 101   FORMAT (A)
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR=0
      IF (FLAG.EQ.'TOFLD') THEN
         NFLD = LAT*1000 + LON
C LAT has two digits, LON has 2 or three digits dependend on the number
C of latitude zones. NFLD has 5 digits
      ELSE IF (FLAG.EQ.'TOZONE') THEN
C decompose NFLD into LAT and LON
           LON = MOD(NFLD,1000)
           LAT = (NFLD - LON)/1000
      ELSE
C flag not known
         IERR=3
         WRITE (OUTPT,110) FLAG
110      FORMAT('unknown flag ',A)
         CALL WRERR(RTNAME,IERR,OUTPT)
      ENDIF
C
CD     IF (HFLAG(RTNAME,'out: NFLD,LAT,LON').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) NFLD,LAT,LON
CD9901   FORMAT (I5,X,I2,X,I3.3)
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
      RETURN
      END
      SUBROUTINE FLDLIM(NZONE,IFIELD,IRAMIN,IRAMAX,IDEMIN,IDEMAX,IERR)
C
CC  calculate border lines of FIELD
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : GRU        date: 25-OCT-1989 16:53
CU  update : GRU        date: 10-AUG-1994 18:00
CU  update : CRI        date: 11-AUG-1994 15:15 reset IERR
C
CT  status: not tested
C
C   general description 
CG
C
C   call_var.          type I/O description 
CP  NZONE               I4  ?/? # of zones used for field definition
CP  IFIELD              I4  ?/? field#
CP  IRAMAX              I4  ?/? upper RA border of field [0.5 arcsec]
CP  IRAMIN              I4  ?/? lower RA border of field [0.5 arcsec]
CP  IDEMAX              I4  ?/? upper DEC border of field [0.5 arcsec]
CP  IDEMIN              I4  ?/? lower DEC border of field [0.5 arcsec]
CP  IERR                I4    O error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
CR  ERRSTC              SR      
CR  FLDCNV              SR      convert field# to longitudinal and
CR                              latitudinal zones
CR  ZONLIM              SR      determine field boundaries
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
c      REAL*4        HFLAG
      INTEGER*4     IERR,IFIELD,IRAMIN,IRAMAX,IDEMIN,IDEMAX,
     *              ISTAT,LAT,LON,NZONE ! ,IOS
      CHARACTER     RTNAME*8,FLAG*6
C
      DATA          RTNAME /'FLDLIM'/
C
CD     IF (HFLAG(RTNAME,'in: SEQ# ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) IFIELD
CD 101   FORMAT ( I)
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C**********************************************************************
C generate latitude and longitude zones from field#
      FLAG = 'TOZONE'
      CALL FLDCNV(LAT,LON,IFIELD,FLAG,ISTAT)
      IF (ISTAT.GT.1) THEN
         IERR=3
         CALL ERRSTC(RTNAME,2)
      ELSE
         CALL ZONLIM(NZONE,LAT,LON,IRAMIN,IRAMAX,IDEMIN,IDEMAX,ISTAT)
         IF (ISTAT.GT.1) THEN
            IERR=3
            CALL ERRSTC(RTNAME,2)
         ENDIF
      ENDIF
C
CD     IF (HFLAG(RTNAME,'out: IERR').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) IERR
CD9901   FORMAT (I)
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
      RETURN
      END
      subroutine GAL12(LDIR,DLOLD,DBOLD,DLNEW,DBNEW,IERR)
C
CC    Converts coordinates between the old and new galactic system
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 20-MAR-1990 14:58
CU  update : KOD        date: 17-AUG-1992 11:14 (`1.D-6')
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert coordinates between the galactic systems
CG  I (old) and II (new), according to
CG     F.C.Bertiau, 'Programmes for pocket calculators HP-67
CG                   and HP-97 in the field of theoretical and
CG                   observational astronomy',
CG                   1977 Leuven University Press, 'ASTR. COOR. 2'
C
C   call_var.          type I/O description 
CP  LDIR                L4  I   transformation direction:
CP                                LDIR = .true.  : gal I --> gal II
CP                                LDIR = .false. : gal I <-- gal II
CP  DLOLD               R8  I   old longitude  (radians)
CP  DBOLD               R8  I   old latitude   (radians)
CP  DLNEW               R8    O new longitude  (radians)
CP  DBNEW               R8    O new latitude   (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  EULER1              SR      general coordinate transformation
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DLOLD, DBOLD, DLNEW, DBNEW, D1, D2, D3,
     &              DPHI1, DTHETA1, DPHI2, DTHETA2 !, DT,  DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      LOGICAL       LDIR
      CHARACTER*8   RTNAME
C
      PARAMETER ( D1 = 1.35524816417D0 )
      PARAMETER ( D2 = 0.02594606466D0 )
      PARAMETER ( D3 = 4.36420103061D0 )
C
      DATA          RTNAME /'GAL12 '/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DLOLD.lt.-1.D-6 .or. DLOLD.gt.D2PI+1.D-6) then
        IERR = 1
        call WRERR(RTNAME,IERR,'Warning: unusual value for DLOLD')
      endif
C
      if (dabs(DBOLD).gt.DPID2+1.D-6) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in input parameter DBOLD')
      endif
C
      if  (IERR.lt.2)  then
C
        DPHI1   = DLOLD
        DTHETA1 = DPID2 - DBOLD
C
        if (LDIR) then
          call EULER1(D1,D2,D3,DPHI1,DTHETA1,DPHI2,DTHETA2,ISTAT)
        else
          call EULER1(-D3,-D2,-D1,DPHI1,DTHETA1,DPHI2,DTHETA2,ISTAT)
        endif
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        else
          DLNEW = DPHI2
          DBNEW = DPID2 - DTHETA2
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine GEORCO(DPHI,DHT,DRSPHI,DRCPHI,IERR)
C
CC    Calculates the GEOcentric Rectangular COordinates of an observer
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : KOD        date: 29-AUG-1989 19:16
CU  update : KOD        date: 30-MAR-1990 16:10
CU  update : KOD        date: 17-AUG-1992 11:15 (`1.D-6')
CU  update : PAUL       date: 16-NOV-1992 16:04 null charactes deleted
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the geocentric rectangular coordinates
CG  of an observer, according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.37-38 .
C
C   call_var.          type I/O description 
CP  DPHI                R8  I   geographical latitude  (radians)
CP  DHT                 R8  I   height above sea level  (meters)
CP  DRSPHI              R8    O rho * sin(geocentric latitude)
CP  DRCPHI              R8    O rho * cos(geocentric latitude)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
CR  WRERR               SR      formats error code in standard form
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DPHI, DHT, DRSPHI, DRCPHI, DBA, DHT0, DU, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR ! ,IOS
      CHARACTER*8   RTNAME
C
      PARAMETER ( DBA   = 0.99664719D0 )
      PARAMETER ( DHT0  = 6378140.D0 )
C
      DATA          RTNAME /'GEORCO'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
      if (DHT.lt.-1.D2 .or. DHT.gt.1.D4) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Error in the input parameter DHT')
      endif
      if (dabs(DPHI).gt.DPID2+1.D-6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in the input parameter DPID2')
      endif
C
      if (IERR.lt.2) then
C
        if (abs(abs(DPHI)-DPID2).lt.1.D-6) then
          DU = sign(DPID2,DPHI)
        else
          DU = atan( DBA * tan(DPHI) )
        endif
C
        DH = DHT / DHT0
C
        DRSPHI = DBA * sin(DU) + DH * sin(DPHI)
        DRCPHI = cos(DU)       + DH * cos(DPHI)
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine GREJUL(LDIR,IYOLD,IMOLD,IDOLD,IYNEW,IMNEW,IDNEW,IERR)
C
CC    Converts between the Gregorian and Julian Calendar system
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : KOD        date: 23-AUG-1989 13:56
CU  update : KOD        date: 20-MAR-1990 14:59
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert between the Gregorian and Julian Calendar
CG  system. The input values will be checked for validity.
CG
CG  This subroutine is valid for the years  -4712  to  +3500 ,
CG  (cf. 'Sky and Telescope, May 1984, p.455) and has been written
CG  according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.24,26.
C
C   call_var.          type I/O description 
CP  LDIR                L4  I   direction of conversion
CP                                LDIR = .true.  :  Greg --> Jul
CP                                LDIR = .false. :  Greg <-- Jul
CP  IYOLD               I4  I   'old' year          (-4712 .. 3500)
CP  IMOLD               I4  I   'old' month         (1..12)
CP  IDOLD               I4  I   'old' day of month  (1..28,29,30,31)
CP  IYNEW               I4    O 'new' year          (-4712 .. 3500)
CP  IMNEW               I4    O 'new' month         (1..12)
CP  IDNEW               I4    O 'new' day of month  (1..28,29,30,31)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, IYOLD, IMOLD, IDOLD, IYNEW, IMNEW,
     &              IDNEW, IYY, IMM, JD, IGREG, I, IA, IB, IC, IG, IE,
     &              IY, IM, ID ! , IOS
      LOGICAL*4     LDIR, LCHECK
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'GREJUL'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR=0
C
C     rough input check:
C
      if ( (.not.LDIR .and. IYOLD .lt.-4712) .or.
     &     IYOLD .gt.3500 .or.
     &     IMOLD .lt.1     .or. IMOLD .gt.12   .or.
     &     IDOLD .lt.1     .or. IDOLD .gt.31 )  then
C
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameters')
C
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate the Julian Date:
C
        if (IMOLD.gt.2) then
          IYY = IYOLD
          IMM = IMOLD
        else
          IYY = IYOLD - 1
          IMM = IMOLD + 12
        endif
C
        if (IYY.gt.0) then
          DH=0.D0
        else
          DH=0.75D0
        endif
C
        JD = idint(365.25d0*dble(IYY)-DH) +
     &       idint(30.6001d0*dble(IMM+1)) +
     &       IDOLD + 1720995
C
        if (LDIR) then
          IGREG = IYY/100
          IGREG = 2 - IGREG + IGREG/4
          JD = JD + IGREG
        endif
C
        if (JD.lt.0) then
          IERR = max(IERR,2)
          call WRERR(RTNAME,IERR,'Error in input parameters')
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate the Calendar Date:
C
        do I = 1, 2
C
          LCHECK = (I.eq.1)
           if (LCHECK.eqv.LDIR) then
             IA = idint( (dble(JD)-1867216.25D0) / 36524.25D0 )
             IA = JD + 1 + IA - IA/4
           else
             IA = JD
           endif
C
           IB = IA + 1524
           IC = idint( (dble(IB)-122.1D0) / 365.25D0 )
           IG = idint( 365.25D0 * dble(IC) )
           IE = idint( dble(IB-IG) / 30.6001D0 )
C
           ID = IB - IG - int( 30.6004 * real(IE) )
C
           if (IE.le.13) then
             IM = IE - 1
           else
             IM = IE - 13
           endif
C
           if (IM.ge.3) then
             IY = IC - 4716
           else
             IY = IC - 4715
           endif
C
          if (LCHECK) then
            if ( IY.ne.IYOLD .or. IM.ne.IMOLD .or. ID.ne.IDOLD )  then
               IERR = max(IERR,2)
               call WRERR(RTNAME,IERR,'input date invalid')
            endif
          else
            IYNEW = IY
            IMNEW = IM
            IDNEW = ID
          endif
C
        end do
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine HMMRAD(DHMMM,DRAD,IERR)
C
CC    Calculates radians of a number given in the form 'HH.MMMM'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : KOD        date: 05-SEP-1989 18:41
CU  update : KOD        date: 29-MAR-1990 15:32
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert a number given in the form 'HH.MMMM'
CG  (hours) into radians. Negative values are allowed.
C
C   call_var.          type I/O description 
CP  DHMMM               R8  I   number to be converted (-24 .. 24)
CP                              (HH.MMMM)
CP  DRAD                R8    O corresponding radians
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  DMMDEC              SR      converts 'HH.MMMM' into dec. hours
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DHMMM, DRAD, DHR, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'HMMRAD'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (dabs(DHMMM).gt.24.D0) then
        IERR = 1
        call WRERR(RTNAME,IERR,'Error in input parameter DHMMM')
      endif
C
      DH = DHMMM
      call DMMDEC(DH,DHR,ISTAT)
C
      if  (ISTAT.gt.1) then
C
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
C
      else
C
        DRAD = DHR * 15.D0 * DPI180
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine HMSRAD(DHMMSS,DRAD,IERR)
C
CC    Calculates radians of a number given in the form 'HH.MMSS'
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : KOD        date: 05-SEP-1989 18:39
CU  update : KOD        date: 29-MAR-1990 15:33
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert a number given in the form 'HH.MMSSS'
CG  (hours) into radians. Negative values are allowed.
C
C   call_var.          type I/O description 
CP  DHMMSS              R8  I   number to be converted (-24 .. 24)
CP                              (HH.MMSSS)
CP  DRAD                R8    O corresponding radians
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  DMSDEC              SR      converts 'HH.MMSS' into dec. hours
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DHMMSS, DRAD, DHR, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'HMSRAD'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (dabs(DHMMSS).gt.24.D0) then
        IERR = 1
        call WRERR(RTNAME,IERR,'Error in input parameter DHMMSS')
      endif
C
      DH = DHMMSS
      call DMSDEC(DH,DHR,ISTAT)
C
      if  (ISTAT.gt.1) then
C
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
C
      else
C
        DRAD = DHR * 15.D0 * DPI180
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine JDCAL1(DJD,IY,IM,ID,IHR,IMT,SEC,IERR)
C
CC    Calculates the Calendar Date out of the Julian Date
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:36
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 29-MAR-1990 15:36
C
CT  status: tested (08-AUG-1989 17:45 / KOD)
C
C   general description 
CG  Subroutine to calculate the Calendar Date out of the Julian Date.
CG
CG  The calendar date will be given in the Gregorian system for
CG  dates starting with October 15, 1582, and in the Julian system
CG  until October 4, 1582  (which precedes the Oct.15 immediately).
CG
CG  This subroutine is valid for the years  -4712  to  +3500 ,
CG  (cf. 'Sky and Telescope, May 1984, p.455) and has been written
CG  according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.26 .
CG  For negative years only the variable IY will be negative.
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Date
CP  IY                  I4    O year         (-4712 .. +3500)
CP  IM                  I4    O month        (1 .. 12)
CP  ID                  I4    O day of month (1 .. 28,29,30,31)
CP  IHR                 I4    O hours        (0 .. 23)
CP  IMT                 I4    O minutes      (0 .. 59)
CP  SEC                 R4    O seconds      (0.0 .. 59.999...)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  CNVDMS              SR      splits fractional part
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD, DH, DF
      REAL*4        SEC, SEC1 ! HFLAG, 
      INTEGER*4     IERR, IY, IM, ID, IHR, IMT, JD, IA, IB, IC,
     &              IG, IE, IHR1, IMT1, ISTAT !  ,IOS
      LOGICAL       LPLUS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'JDCAL1 '/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DJD.lt.0.D0 .or. DJD.gt.2999773.5D0) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
      if (IERR.lt.2) then
        DH = DJD + 0.5D0
        JD = idint(DH)
        DF = DH - dble(JD)
C
        if (JD.lt.2299161) then
          IA = JD
        else
          IA = idint( (dble(JD)-1867216.25D0) / 36524.25D0 )
          IA = JD + 1 + IA - IA/4
        endif
C
        IB = IA + 1524
        IC = idint( (dble(IB)-122.1D0) / 365.25D0 )
        IG = idint( 365.25D0 * dble(IC) )
        IE = idint( dble(IB-IG) / 30.6001D0 )
C
        ID = IB - IG - int( 30.6004 * real(IE) )
C
        if (IE.le.13) then
          IM = IE - 1
        else
          IM = IE - 13
        endif
C
        if (IM.ge.3) then
          IY = IC - 4716
        else
          IY = IC - 4715
        endif
C
        DF = DF * 24.D0
        LPLUS = .TRUE.
        call CNVDMS(DF,LPLUS,IHR1,IMT1,SEC1,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        else
          IHR = IHR1
          IMT = IMT1
          SEC = SEC1
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine JDCAL2(DJD,ITD,ITS,IERR)
C
CC    Calculates the Calendar Date out of the Julian Date.
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 9-MAR-1992 16:50
CU  update : KOD        date: 11-MAR-1992 16:18
CU  update : KOD        date: 11-MAR-1992 16:58
C
CT  status: tested
C
C   general description 
CG
CG  This subroutine calculates the Calendar Date out of the Julian Date
CG  for the years 1950 to 2049. The Calendar Date is contained in the
CG  integers values ITD (YYMMDD) and ITS (msec of the day in UT).
C
C   call_var.          type I/O description 
CP  DJD                 R8  I/? Julian Date
CP  ITD                 I4    O Calendar Date  (YYMMDD)
CP  ITS                 I4    O Universal Time (msec of the day)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  JDCAL1              SR      calculates the calendar date
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DJD
      REAL*4        SEC ! HFLAG, 
      INTEGER*4     IERR, ISTAT, ITD, ITS, IY, IM, ID, IHR, IMT ! , IOS
      CHARACTER*4   CH4
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'JDCAL2'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     check input values:
C
      if (DJD.lt.2433282.5D0 .or. DJD.gt.2469807.5D0) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in Julian Date; < 1950 or > 2049')
      endif
C
C
      if (IERR.lt.2) then
C
        call JDCAL1(DJD,IY,IM,ID,IHR,IMT,SEC,ISTAT)
C
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        else
          write(CH4,'(I4)') IY
          read(CH4,'(2X,I2)') IY
          ITD = IY * 10000 + IM * 100 + ID
          ITS = (IHR * 3600 + IMT * 60) * 1000 + nint(SEC*1.E3)
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine JDDAT1(DJD,DT,IERR)
C
CC    Calculates the Calendar Date out of the Julian Date
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 22-AUG-1989 17:28
CU  update : KOD        date: 29-MAR-1990 15:44
CU  update : PAUL       date: 08-FEB-1994 14:00 ...,IERR) -> ...,ISTAT)
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the Calendar Date out of the Julian Date.
CG
CG  The calendar date will be given in the Gregorian system for
CG  dates starting with October 15, 1582, and in the Julian system
CG  until October 4, 1582  (which precedes the Oct.15 immediately).
CG
CG  This subroutine is valid for the years  -4712  to  +3500 ,
CG  (cf. 'Sky and Telescope, May 1984, p.455) and has been written
CG  according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.26 .
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Date
CP  DT                  R8    O Calendar Date (YYYYMMDD.HHMMSSS)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  JDCAL1              SR      Julian Date to Calendar Date
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD, DT, DJ
      REAL*4        SEC ! HFLAG, 
      INTEGER*4     IERR, IYR, IMTH, IDAY, IHR, IMIN, MSEC, ISTAT !  ,IOS
      CHARACTER*19  CH19
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'JDDAT1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR=0
C
      DJ = DJD
      call JDCAL1(DJ,IYR,IMTH,IDAY,IHR,IMIN,SEC,ISTAT)
CN!     ERROR handling
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      MSEC = INT(SEC*1.E3)
C
      WRITE(CH19,10) IYR, IMTH, IDAY, IHR, IMIN, MSEC
   10 FORMAT(I5,2I2.2,'.',2I2.2,I5.5)
      READ(CH19,'(F17.7)') DT
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine JDDAT2(DJD,CHDAT,IERR)
C
CC    Calculates the Calendar Date out of the Julian Date
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 22-AUG-1989 18:31
CU  update : KOD        date: 29-MAR-1990 15:49
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the Calendar Date out of the Julian Date.
CG
CG  The calendar date will be given in the Gregorian system for
CG  dates starting with October 15, 1582, and in the Julian system
CG  until October 4, 1582  (which precedes the Oct.15 immediately).
CG
CG  This subroutine is valid for the years  -4712  to  +3500 ,
CG  (cf. 'Sky and Telescope, May 1984, p.455) and has been written
CG  according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.26 .
CG
CG  This subroutine produces a string of the form
CG  'YYYY.MM.DD / HH:MM:SS.SS'  (Maximum resolution: 10 msec)
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Date
CP  CHDAT               C   ?/? string containing the calendar date
CP                              'YYYY' to 'YYYY.MM.DD / HH:MM:SS.SS'
CP                               possible lengths 'L' of CHDAT :
CP                               L =  4 : 'YYYY'
CP                               L =  7 : 'YYYY.MM'
CP                               L = 10 : 'YYYY.MM.DD'
CP                               L = 15 : 'YYYY.MM.DD / HH'
CP                               L = 18 : 'YYYY.MM.DD / HH:MM'
CP                               L = 21 : 'YYYY.MM.DD / HH:MM:SS'
CP                               L = 23 : 'YYYY.MM.DD / HH:MM:SS.S'
CP                               L = 24 : 'YYYY.MM.DD / HH:MM:SS.SS'
CP                               The values for the year, month
CP                               and day will not be rounded,
CP                               the values for the hour, minute
CP                               and second will be rounded.
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  JDCAL1              SR      Julian Date to Calendar Date
CR  CHDATE              SR      converts Calendar Date into a string
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD, DJ
      REAL*4        SEC ! HFLAG, 
      INTEGER*4     IERR, IYR, IMTH, IDAY, IHR, IMIN, ISTAT !  ,IOS, MSEC
      LOGICAL*4     LCHECK
      CHARACTER*(*) CHDAT
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'JDDAT2'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      DJ = DJD
      call JDCAL1(DJ,IYR,IMTH,IDAY,IHR,IMIN,SEC,ISTAT)
CN!     ERROR handling
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      LCHECK=.true.
      call CHDATE(IYR,IMTH,IDAY,IHR,IMIN,SEC,LCHECK,CHDAT,ISTAT)
C
CN!   ERROR handling
      if  (ISTAT.gt.1) then
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,20)
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine JDEPS1(DJD,LMEAN,DEPS,IERR)
C
CC    Calculates the MEAN/TRUE obliquity of the ecl. for the Jul. Date
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 29-AUG-1989 17:22
CU  update : KOD        date: 29-MAR-1990 15:53
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the MEAN/TRUE obliquity of the ecliptic
CG  for a given Julian Date, according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.44,69,70,81.
CG
CG  The parameters for calculating the mean obliquity have been
CG  taken from 'THE ASTRONOMICAL ALMANACH', 1984, S21.
CG
CG  For APPARENT positions (affected by aberration and nutation)
CG  use the TRUE obliquity, otherwise the MEAN obliquity.
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Date
CP  LMEAN               L4  I   mean or true obliquity of the ecliptic:
CP                                LMEAN = .true.  : mean obliquity
CP                                LMEAN = .false. : true obliquity
CP  DEPS                R8    O obliquity of the ecliptic  (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  JDNUT1              SR      calculates the nutation parameters
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD, DEPS, DJD0, DYEAR, D1, D2, D3, D4, DT, DP,
     &              DE, DJ
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN
      CHARACTER*8   RTNAME
C
      PARAMETER ( DJD0 = 2451545.D0 )
      PARAMETER ( DYEAR =  36525.D0 )
C
      PARAMETER ( D1 =  0.40909280422D0 )
      PARAMETER ( D2 = -2.26965525D-4 )
      PARAMETER ( D3 = -2.86040D-9 )
      PARAMETER ( D4 =  8.78967D-9 )
C
      DATA          RTNAME /'JDEPS1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DJD.lt.0.D0 .or. DJD.gt.3.D6) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
      if (IERR.lt.2) then
C
        DT = (DJD - DJD0) / DYEAR
C
        DEPS = D1 + ( D2 + ( D3 + D4 * DT ) * DT ) * DT
C
        if (.not.LMEAN) then
C
          DJ = DJD
          call JDNUT1(DJ,DP,DE,ISTAT)
C
          if  (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,10)
          else
            DEPS = DEPS + DE
          endif
C
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine JDLST1(DJD,LMEAN,DLONG,DLSID,IERR)
C
CC    Calculates the local mean/apparent sidereal time for a given JD
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 05-SEP-1989 16:40
CU  update : KOD        date: 29-MAR-1990 15:56
CU  update : KOD        date: 17-AUG-1992 09:57 (DPI+1.D-6)
C
CT  status: tested
C
C   general description 
CG
CG  Subroutine to calculate the local mean/apparent sidereal time
CG  from the Julian Date, according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.39-41,44.
CG
CG  For APPARENT positions (affected by aberration and nutation)
CG  use the APPARENT sid. time, otherwise the MEAN sidereal time.
CG
CG  For consistency with the other subroutines radians are used for
CG  the geographical longitude and sidereal time.
CG
CG  According to a resolution of the IAU east longitudes are counted
CG  positive, west long. negative (Sky and Tel, June, 1984, p.559).
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Date
CP  LMEAN               L4  I   mean or app. sidereal time:
CP                                LMEAN = .true.  : mean sid. time
CP                                LMEAN = .false. : app. sid. time
CP  DLONG               R8  I   geographical longitude ( radians ! )
CP  DLSID               R8    O local sidereal time    ( radians ! )
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  JDSID1              SR      calc. the sid. time at Greenwich
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
!HB   REAL*8        DJD, DLONG, DLSID, DSID, DRADRS, DJ
      REAL*8        DJD, DLONG, DLSID, DSID, DJ
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN, LMN
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'JDLST1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (dabs(DLONG).gt.DPI+1.D-6) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DLONG')
      endif
C
      if (IERR.lt.2) then
        DJ = DJD
        LMN = LMEAN
        call JDSID1(DJ,LMN,DSID,ISTAT)
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        else
          DLSID = DRADRS ( DSID + DLONG )
        endif
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine JDNUT1(DJD,DP,DE,IERR)
C
CC    Calculates the nutation in longitude and in obliquity from the JD
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 29.08L-1989 14:45
CU  update : KOD        date: 20-MAR-1990 15:04
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the nutation in longitude and the
CG  nutation in obliquity which are needed for the calculation
CG  of the APPARENT place of a star and for that of the APPARENT
CG  sidereal time, according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.69-70 .
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Date
CP  DP                  R8    O nutation in longitude (radians)
CP  DE                  R8    O nutation in obliquity (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD, DP, DE, DJD0, DYEAR, DT, DLS, DLM,
     &              DMS, DMM, DOM, DH, DLS2, DLM2, DOM2, DF
c      REAL*4        HFLAG
      INTEGER*4     IERR !,IOS
      CHARACTER*8   RTNAME
C
      PARAMETER ( DJD0 = 2415020.D0 )
      PARAMETER ( DYEAR =  36525.D0 )
C
      PARAMETER ( DF = 4.848136811D-6 )
C
      DATA          RTNAME /'JDNUT1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DJD.lt.0.D0 .or. DJD.gt.3.D6) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
      if (IERR.lt.2) then
C
        DT = (DJD - DJD0) / DYEAR
C
        DLS = 4.8816283D0 + (628.3319506D0 + 5.2883D-6 * DT) * DT
        DLM = 4.7199672D0 + (8399.709144D0 - 1.9775D-5 * DT) * DT
        DMS = 6.2565830D0 + (628.3019466D0 - 2.6180D-6 * DT) * DT
        DMM = 5.1680002D0 + (8328.691104D0 + 1.6043D-4 * DT) * DT
        DOM = 4.5236020D0 - (33.75714610D0 + 3.6268D-5 * DT) * DT
C
        DLS2 = 2.D0 * DLS
        DLM2 = 2.D0 * DLM
        DOM2 = 2.D0 * DOM
C
        DH = - (17.2327D0 + 0.01737D0 * DT ) * sin(DOM)
     &       - ( 1.2729D0 + 0.00013D0 * DT ) * sin(DLS2)
     &       + 0.2088D0 * sin(DOM2) - 0.2037D0 * sin(DLM2)
     &       + ( 0.1261D0 - 0.00031D0 * DT) * sin(DMS)
     &       + 0.0675 * sin(DMM)
     &       - (0.0497D0 - 0.00012D0 * DT) * sin(DLS2+DMS)
     &       - 0.0342D0 * sin(DLM2-DOM)
     &       - 0.0261D0 * sin(DLM2+DMM)
     &       + 0.0214D0 * sin(DLS2-DMS)
     &       - 0.0149D0 * sin(DLS2-DLM2+DMM)
     &       + 0.0124D0 * sin(DLS2-DOM)
     &       + 0.0114D0 * sin(DLM2-DMM)
C
        DP = DH * DF
C
        DH =   ( 9.2100D0 + 0.00091D0 * DT ) * cos(DOM)
     &       + ( 0.5522D0 - 0.00029D0 * DT ) * cos(DLS2)
     &       - 0.0904D0 * cos(DOM2)
     &       + 0.0884D0 * cos(DLM2)
     &       + 0.0216D0 * cos(DLS2+DMS)
     &       + 0.0183D0 * cos(DLM2-DOM)
     &       + 0.0113D0 * cos(DLM2+DMM)
     &       - 0.0093D0 * cos(DLS2-DMS)
     &       - 0.0066D0 * cos(DLS2-DOM)
C
        DE = DH * DF
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine JDSID1(DJD,LMEAN,DSID,IERR)
C
CC    Calculates the mean/apparent sid. time at Greenwich from the JD
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 05-SEP-1989 16:20
CU  update : KOD        date: 30-MAR-1990 16:11
CU  update : PAUL       date: 16-NOV-1992 16:05 null characters deleted
C
CT  status: tested
C
C   general description 
CG
CG  Subroutine to calculate the mean/apparent sidereal time
CG  at Greenwich from the Julian Date, according to
CG     R.W.Sinnott, 'Astronomical Computing',
CG                   Sky and Telescope, 1984, p.558-559 .
CG
CG  The formula used has been adopted by the International
CG  Astronomical Union (Sky and Telescope, June, 1982, p.562).
CG
CG  Values to check the results (Sky and Telescope, June 1984, p.559):
CG
CGT        JD:        0.0   GMST = 16:13:22
CGT        JD: 170 5426.1   GMST = 01:45:59
CGT        JD: 244 5700.5   GMST = 06:39:23
CGT        JD: 245 1545.0   GMST = 18:41:51
CGT        JD: 299 9624.2   GMST = 13:41:28
CG
CG  For APPARENT positions (affected by aberration and nutation)
CG  use the APPARENT sid. time, otherwise the MEAN sidereal time.
CG
CG  For consistency with the other subroutines the sidereal time
CG  will be given in radians.
C
C   call_var.          type I/O description 
CP  DJD                 R8  I   Julian Date
CP  LMEAN               L4  I   mean or true siderial time:
CP                                LMEAN = .true.  : mean sid. time
CP                                LMEAN = .false. : app. sid. time
CP  DSID                R8    O sidereal time ( radians ! )
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  DRADRS              R8      'resets' radians
CR  JDEPS1              SR      calculates obliquity of the ecliptic
CR  JDNUT1              SR      calculates the nutation parameters
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DJD, DSID, DJD0, DYEAR, DA, DB, DC, DD, DF, DJD0UT,
!HB  &              DT, DFRC, DRADRS, DSID0, DEPS, DP, DE, DH, DJ
     &              DT, DFRC, DSID0, DEPS, DP, DE, DH, DJ
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN, LMN
      CHARACTER*8   RTNAME
C
      PARAMETER ( DJD0 = 2451545.D0 )
      PARAMETER ( DYEAR =  36525.D0 )
C
      PARAMETER ( DA =   1.7533685592D0  )
      PARAMETER ( DB = 628.33197068884D0 )
      PARAMETER ( DC =   6.770714D-6     )
      PARAMETER ( DD =  -4.509D-10       )
      PARAMETER ( DF =   1.002737908D0   )
C
      DATA          RTNAME /'JDSID1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DJD.lt.0.D0 .or. DJD.gt.3.D6) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
      if (IERR.lt.2) then
C
C       calculate the nearest Julian Date at 00:00 UT:
C
        DJD0UT = DNINT(DJD-0.5D0)+0.5D0
C
C       calculate the corresponding sidereal time:
C
        DT = (DJD0UT-DJD0) / DYEAR
        DSID0 = DA + (DB + (DC + DD * DT ) * DT ) * DT
C
C       find sidereal time for the given hours:
C
        DFRC = DJD - DJD0UT
        DT = DSID0 + DFRC * D2PI * DF
        DSID = DRADRS(DT)
C
        if (.not.LMEAN) then
C
          DJ = DJD
          LMN = LMEAN
          call JDEPS1(DJ,LMN,DEPS,ISTAT)
          if  (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,10)
          endif
C
          if (IERR.lt.2) then
            call JDNUT1(DJ,DP,DE,IERR)
            if  (ISTAT.gt.1) then
              IERR = max(IERR,ISTAT)
              call ERRSTC(RTNAME,20)
            endif
          endif
C
          if (IERR.lt.2) then
            DH = DP * cos(DEPS)
            DSID = DRADRS ( DSID + DH )
          endif
C
        endif
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine LSTJD1(DJD0,DLONG,DLSID,LMEAN,DJD,IERR)
C
CC    Calculates the time from the mean/apparent local sid.time
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 05-SEP-1989 16:10
CU  update : KOD        date: 29-MAR-1990 16:09
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the Universal Time corresponding
CG  to a given mean/apparent local sidereal time. This time
CG  will be selected to be nearest to the given date.
CG
CG  For consistency with the other subroutines the sidereal time
CG  and geographical longitude are required in radians.
CG
CG  According to a resolution of the IAU east longitudes are counted
CG  positive, west long. negative (Sky and Tel, June, 1984, p.559).
C
C   call_var.          type I/O description 
CP  DJD0                R8  I   Julian Date (nearest time)
CP  DLONG               R8  I   geographical longitude ( radians ! )
CP  DLSID               R8  I   local sidereal time ( radians ! )
CP  LMEAN               L4  I   mean or app. sidereal time:
CP                                LMEAN = .true.  : mean sid. time
CP                                LMEAN = .false. : app. sid. time
CP  DJD                 R8    O Julian Date corresponding to DSID
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  JDLST1              SR      calc. the local mean/app. sid. time
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
!HB   REAL*8        DJD0, DLONG, DLSID, DSID, DRADRS, DLSID0, DJD, DT,
      REAL*8        DJD0, DLONG, DLSID, DSID, DLSID0, DJD, DT,
     &              DJ0, DLG
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN, LMN
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'LSTJD1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DSID.lt.0.D0 .or. DSID.gt.D2PI) then
        IERR = 1
        call WRERR(RTNAME,IERR,'Error in the input parameter DSID')
        DSID = DRADRS(DSID)
      endif
C
      if (IERR.lt.2) then
C
C       calculate the local sidereal time at DJD0:
C
        DJ0 = DJD0
        LMN = LMEAN
        DLG = DLONG
        call JDLST1(DJ0,LMN,DLG,DLSID0,ISTAT)
C
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
      if (IERR.lt.2) then
C
C       calculate the difference in local sidereal time:
C
        DT = DRADRS ( DLSID - DLSID0 ) / D2PI
C
        if (DT.gt.0.5D0) DT = DT - 1.D0
C
C       convert the difference into solar time:
C
        DT = DT / 1.002737908D0
C
        DJD = DJD0 + DT
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine POLRE1(DR,DPHI,DTHETA,DX,DY,DZ,IERR)
C
CC    Converts polar coordinates into rectangular coordinates
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 19-JUN-1989 18:31
CU  update : KOD        date: 20-MAR-1990 15:08
CU  update : KOD        date: 17-AUG-1992 11:16 (`1.D-6')
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert the polar coordinates (DR,DPHI,DTHETA)
CG  (mathematical notation) into the rectangular coordinates
CG  (DX,DY,DZ).
C
C   call_var.          type I/O description 
CP  DR                  R8  I   radius  (same unit as DX,DY,DZ)
CP  DPHI                R8  I   azimuthal angle  (radians)
CP  DTHETA              R8  I   polar     angle  (radians)
CP  DX                  R8    O x-value (same unit as DX,DY,DZ)
CP  DY                  R8    O y-value (same unit as DX,DY,DZ)
CP  DZ                  R8    O z-value (same unit as DX,DY,DZ)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DR, DPHI, DTHETA, DX, DY, DZ, DST
c      REAL*4        HFLAG
      INTEGER*4     IERR !,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'POLRE1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR=0
C
      if (DR.lt.-1.D-6 .or. DPHI.lt.-1.D-6 .or. DPHI.gt.D2PI+1.D-6 .or.
     $    DTHETA.lt.-1.D-6 .or. DTHETA.gt.DPI+1.D-6) then
        IERR = 1
        call WRERR(RTNAME,IERR,'Error in input parameters')
      endif
C
      DST = dsin(DTHETA)
C
      DX = DR * DST * dcos(DPHI)
      DY = DR * DST * dsin(DPHI)
      DZ = DR * dcos(DTHETA)
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine REPOL1(DX,DY,DZ,DR,DPHI,DTHETA)
C
CC    Converts rectangular coordinates into polar coordinates
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 15-JUN-1989 18:31
CU  update : KOD        date: 20-MAR-1990 15:09
C
CT  status: tested
C
C   general description 
CG  Subroutine to convert the rectangular coordinates (DX,DY,DZ)
CG  into the polar coordinates (DR,DPHI,DTHETA) - mathematical
CG  notation.
C
C   call_var.          type I/O description 
CP  DX                  R8  I   x-value
CP  DY                  R8  I   y-value
CP  DZ                  R8  I   z-value
CP  DR                  R8    O radius   (same unit as DX,DY,DZ)
CP  DPHI                R8    O azimuthal angle  (radians)
CP  DTHETA              R8    O  polar    angle  (radians)
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DX, DY, DZ, DR, DPHI, DTHETA, DH
c      REAL*4        HFLAG
c      INTEGER*4     IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'REPOL1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      DR = dsqrt(DX*DX+DY*DY+DZ*DZ)
C
      if (DR.eq.0.0D0) then
        DPHI   = 0.0D0
        DTHETA = 0.0D0
      else
        DTHETA = dacos(DZ/DR)
        DH = dsqrt(DX*DX+DY*DY)
        if (DH.eq.0.0D0) then
          DPHI = 0.0D0
        else
          DPHI = dasin(DY/DH)
          if (DX  .lt. 0.0D0) DPHI = DPI  - DPHI
          if (DPHI.lt. 0.0D0) DPHI = DPHI + D2PI
        endif
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine RESVAL(DXMIN,DXMAX,DXOLD,DXNEW,IERR)
C
CC    'Resets' periodic variables to their basic interval
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 20-MAR-1990 15:10
C
CT  status: tested
C
C   general description 
CG  Subroutine to 'reset' the values of periodically defined
CG  variables to their basic interval.
C
C   call_var.          type I/O description 
CP  DXMIN               R8  I   left  boundary of basic interval
CP  DXMAX               R8  I   right boundary of basic interval
CP  DXOLD               R8  I   value before resetting
CP  DXNEW               R8    O value after  resetting
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DXMIN, DXMAX, DXOLD, DXNEW, DDIFF, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR !,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'RESVAL'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DXMIN.ge.DXMAX) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in input parameters DXMIN,DXMAX')
      endif
C
      if (IERR.lt.2) then
        DDIFF = DXMAX - DXMIN
        DH = (DXOLD-DXMIN) / DDIFF
        DH = DH - DINT(DH)
        if (DH.lt.0.D0) DH = DH + 1.D0
        DXNEW = DXMIN + DH * DDIFF
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine SIDJD1(DJD0,DSID,LMEAN,DJD,IERR)
C
CC    Calculates the time from the mean/apparent sid.time at Greenwich
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 02-OCT-1989 14:20
CU  update : KOD        date: 29-MAR-1990 16:43
CU  update : KOD        date: 29-MAR-1990 16:45
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the Universal Time corresponding
CG  to a given mean/apparent sidereal time at Greenwich.
CG  This time will be selected to be nearest to the given date.
CG
CG  For consistency with the other subroutines the sidereal time
CG  is required in radians.
C
C   call_var.          type I/O description 
CP  DJD0                R8  I   Julian Date (nearest time)
CP  DSID                R8  I   sidereal time ( radians ! )
CP  LMEAN               L4  I   mean or app. sidereal time:
CP                                LMEAN = .true.  : mean sid. time
CP                                LMEAN = .false. : app. sid. time
CP  DJD                 R8    O Julian Date corresponding to DSID
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  JDSID1              SR      calculates the mean/app. sidereal time
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DJD0, DSID, DJD, DSID0, DT, DJ0, DSD
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN, LMN
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'SIDJD1'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      if (DSID.lt.0.D0 .or. DSID.gt.D2PI) then
        IERR = 1
        call WRERR(RTNAME,IERR,'Error in the input parameter DSID')
        DSD = DSID
        DSD = DRADRS(DSD)
      endif
C
      if (IERR.lt.2) then
C
C       calculate the sidereal time at DJD0:
C
        DJ0 = DJD0
        LMN = LMEAN
        call JDSID1(DJ0,LMN,DSID0,ISTAT)
C
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
      if (IERR.lt.2) then
C
C       calculate the difference in sidereal time:
C
        DT = DRADRS ( DSD - DSID0 ) / D2PI
C
        if (DT.gt.0.5D0) DT = DT - 1.D0
C
C       convert the difference into solar time:
C
        DT = DT / 1.002737908D0
C
        DJD = DJD0 + DT
C
      endif
C
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine smbnd3(nband,iband,iseq,lcen,dalphm,dalphp,
     &                  ddeltm,ddeltp,ierr)
C
CC    Calculates the boundaries of a given small map
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 21-APR-1993
CU  update : GRU        date: 4-AUG-1994 15:05
CU  update : GRU        date: 4-AUG-1994 15:11
CU  update : GRU        date: 4-AUG-1994 15:12
C
CT  status: not tested
C
C   general description 
CG  Subroutine to calculate the boundaries of a given small map.
CG  Two polar caps, one band centered on the equator, with RA=0
CG  at the center (lcen=true) or border of small maps with iseq=1.
C
C   call_var.          type I/O description 
CP  NBAND               I4  I/? number of bands on the the sphere
CP                              including the equatorial band and
CP                              the polar caps
CP                              ( NBAND > 0  and  not even )
CP  IBAND               I4  I   actual number of the band     (1..nband)
CP  ISEQ                I4  I   sequence number along the band
CP  LCEN                L4  I   lcen=.true.  --> RA=0 at center
CP                              lcen=.false. --> RA=0 at border
CP  DALPHM              R8    O lower  Right Ascension     (deg, 0..360)
CP  DALPHP              R8    O higher Right Ascension     (deg, 0..360)
CP  DDELTM              R8    O lower Declination          (deg,-90..90)
CP  DDELTP              R8    O higher Declination         (deg,-90..90)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  SMCRT2              SR      
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      implicit none
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
c
      integer*4     maxbnd
c
      parameter     ( maxbnd = 100 )
c
      real*8        dalphm, dalphp, ddeltm, ddeltp, dh
      real*4        sbsm(maxbnd), rms
      integer*4     ierr, istat, nband, nband0, nbsm(maxbnd),
     &              nsm, iband, iseq!, ios, i, ism
      logical       lcen, lfirst, lstart
      character*8   rtname
      save          nband0, nbsm
c
      data          rtname /'smbnd3'/
      data          lfirst / .true. /
C
      ierr = 0
c
c     check the input values:
c
      if (mod(nband,2).eq.0 .or. nband.gt.maxbnd .or.
     &    min(nband,iband,iseq).lt.1 .or. iband.gt.nband) then
        ierr = max(ierr,2)
        call wrerr(rtname,ierr,'Error in input values')
      endif
c
      if (ierr.lt.2) then
c
        if (lfirst) then
          lstart = .true.
        else
          lstart = (nband.ne.nband0)
        endif
c
        if (lstart) then
c
c         split the sphere into segments:
c
          call smcrt2(nband,nbsm,sbsm,nsm,rms,istat)
c
          if (istat.gt.1) then
            ierr = max(ierr,istat)
            call errstc(rtname,10)
          else
            nband0 = nband
            lfirst = .false.
          endif
c
        endif
      endif
c
      if (iseq.gt.nbsm(iband)) then
        ierr = max(ierr,2)
        write (outpt,105) iseq,iband,nbsm(iband)
105     format('iseq ',I5,' too large for band ',I5,' (max: ',I5,')')
        call wrerr(rtname,ierr,outpt)
      endif
c
      if (ierr.lt.2) then
c
c       calculate the declination boundaries:
c
        if (nband.eq.1) then
          ddeltm = -90.d0
          ddeltp =  90.d0
        else
          dh = 180.d0 / dble(nband-1)
          ddeltm = (0.5d0 * dble(nband) - dble(iband)) * dh
          ddeltp = dmin1( ddeltm + dh, 90.d0 )
          ddeltm = dmax1( ddeltm, -90.d0 )
        endif
c
c       calculate the right ascension boundaries:
c
        if (nband.eq.1 .or. ddeltm.eq.-90.d0 .or. ddeltp.eq.90.d0) then
          dalphm =   0.d0
          dalphp = 360.d0
        else
          dh = 360.d0 / dble(nbsm(iband))
          if (lcen) then
            dalphm = (dble(iseq)-1.5d0) * dh
          else
            dalphm = (dble(iseq)-1.0d0) * dh
          endif
          dalphp = dalphm + dh
          if (dalphm.lt.0.d0) dalphm = dalphm + 360.d0
        endif
c
      endif
c
      return
      end
      subroutine smcen3(nband,iband,iseq,lcen,dalph0,ddelt0,ierr)
C
CC    Calculates the center of a given small map
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:37
C
CA  author : KOD        date: 13-JUN-1994
C
CT  status: not tested
C
C   general description 
CG  Subroutine to calculate the center of a given small map.
CG  Two polar caps, one band centered on the equator, with RA=0
CG  at the center (lcen=true) or border of small maps with iseq=1.
C
C   call_var.          type I/O description 
CP  NBAND               I4  I/? number of bands on the the sphere
CP                              including the equatorial band and
CP                              the polar caps
CP                              ( NBAND > 0  and  not even )
CP  IBAND               I4  I   actual number of the band     (1..nband)
CP  ISEQ                I4  I   sequence number along the band
CP  LCEN                L4  I   lcen=.true.  --> RA=0 at center
CP                              lcen=.false. --> RA=0 at border
CP  DALPH0              R8    O central Right Ascension      (deg, 0..360)
CP  DDELT0              R8    O central Declination          (deg,-90..90)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI 
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  SMCRT2              SR      
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      implicit none
c
      integer*4     maxbnd
c
      parameter     ( maxbnd = 100 )
c
      real*8        dalph0, ddelt0, dh
      real*4        sbsm(maxbnd), rms
      integer*4     ierr, istat, nband, nband0, nbsm(maxbnd),
     &              nsm, iband, iseq!, ios, i, ism
      logical       lcen, lfirst, lstart
      character*8   rtname
      save          nband0, nbsm
c
      data          rtname /'smcen3'/
      data          lfirst / .true. /
C
      ierr = 0
c
c     check the input values:
c
      if (mod(nband,2).eq.0 .or. nband.gt.maxbnd .or.
     &    min(nband,iband,iseq).lt.1 .or. iband.gt.nband) then
        ierr = max(ierr,2)
        call wrerr(rtname,ierr,'Error in input values')
      endif
c
      if (ierr.lt.2) then
c
        if (lfirst) then
          lstart = .true.
        else
          lstart = (nband.ne.nband0)
        endif
c
        if (lstart) then
c
c         split the sphere into segments:
c
          call smcrt2(nband,nbsm,sbsm,nsm,rms,istat)
c
          if (istat.gt.1) then
            ierr = max(ierr,istat)
            call errstc(rtname,10)
          else
            nband0 = nband
            lfirst = .false.
          endif
c
        endif
      endif
c
      if (iseq.gt.nbsm(iband)) then
        ierr = max(ierr,2)
        call wrerr(rtname,ierr,'iseq too large')
      endif
c
      if (ierr.lt.2) then
c
c       calculate the central declination:
c
        if (nband.eq.1) then
          ddelt0 =  90.d0
        else
          dh = 180.d0 / dble(nband-1)
          ddelt0 = (0.5d0 * dble(nband+1) - dble(iband)) * dh
        endif
c
c       calculate the central right ascension:
c
        if (nband.eq.1 .or. ddelt0.eq.-90.d0 .or. ddelt0.eq.90.d0) then
          dalph0 =   0.d0
        else
          dh = 360.d0 / dble(nbsm(iband))
          if (lcen) then
            dalph0 = (dble(iseq)-1.0d0) * dh
          else
            dalph0 = (dble(iseq)-0.5d0) * dh
          endif
          if (dalph0.lt.0.d0) dalph0 = dalph0 + 360.d0
        endif
c
      endif
c
      return
      end
      subroutine SMCRT2(NBAND,NBSM,SBSM,NSM,RMS,IERR)
C
CC    Creates small maps of sufficient overlap covering the whole sphere
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:38
C
CA  author : KOD        date: 26-AUG-1992 10:42
CU  update : KOD        date: 26-AUG-1992 10:51
CU  update : KOD        date: 26-AUG-1992 11:20
C
CT  status: not tested
C
C   general description 
CG
CG  Subroutine to split the surface of a sphere into small maps.
CG  These small maps all have the same extent with respect to
CG  the polar angle; there will be two circular polar caps with
CG  identical size and one band with the equator in its middle.
CG  The small maps will be determined so that they fit into the
CG  same `square'.
C
C   call_var.          type I/O description 
CP  NBAND               I4  I   number of bands on the the sphere
CP                              including the equatorial band and
CP                              the polar caps
CP                              ( NBAND  > 0  and  not even )
CP  NBSM                I4    O number of small maps per band
CP  SBSM                R4    O solid angle covered by one small map
CP  NSM                 I4    O total number of small maps
CP  RMS                 R4    O relative mean quadratic deviation
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      INTEGER*4     IERR, NBAND, NBSM(NBAND), NSM, N, I, NUMB! ,IOS
      REAL*4        PDIFF, PID2, PSIZE, PI2, ANG1, ANG2,
     &              TSIZE, SBSM(NBAND), RMS, H, SIZEM, TDIFF !HFLAG, 
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'SMCRT2'/
C
      PI2  = 2.0 * real(PI)
      PID2 = 0.5 * real(PI)
C
      IERR = 0
C
C     check input value:
C
      if (NBAND.le.0 .or. mod(NBAND,2).eq.0) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameter')
      endif
C
C
      if (IERR.lt.2) then
C
        if (NBAND.eq.1) then
C
          NBSM(1) = 1
          SBSM(1) = 2.0 * PI2
          NSM = 1
C
        else
C
          N = NBAND / 2
C
C         determine extent in polar angle:
C
          PDIFF = real(PI) / real(NBAND-1)
C
          NBSM(NBAND) = 1
C
C         calculate the size of the polar cap:
C
          PSIZE = PI2 * ( 1.0 - cos (0.5 * PDIFF) )
          SBSM(NBAND) = PSIZE
C
C
C         find the number of sm per band with approx.the same max extent
C
          do I = 1, N
C
C           calculate the length of the larger small circle
C           defining the boundary of the band:
C
            ANG1 = (real(I) - 1.5) * PDIFF
            ANG2 = ANG1 + PDIFF
C
            TSIZE = PI2 * ( sin(ANG2) - sin(ANG1) )
            TDIFF = PI2 * cos(ANG1)
C
C           find the number of small maps:
C
            NUMB = int(TDIFF / PDIFF)
            if (real(NUMB)*PDIFF.lt.TDIFF) NUMB = NUMB + 1
C
            NBSM(N+I) = NUMB
            SBSM(N+I) = TSIZE / real(NUMB)
C
          end do
C
C         copy upper part of the sphere to lower part:
C
          do I = 1, N
            NBSM(I) = NBSM(NBAND+1-I)
            SBSM(I) = SBSM(NBAND+1-I)
          end do
C
C         calculate the total number of small maps:
C
          NSM = 0
          do I = 1, NBAND
            NSM = NSM + NBSM(I)
            H = H + real(NBSM(I)) * SBSM(I)
          end do
C
C         calculate the mean quadratic deviation:
C
          SIZEM = 2.0 * PI2 / real (NSM)
C
          RMS = 0.0
          do I = 1, NBAND
            H = (SBSM(I)-SIZEM)
            RMS = RMS + real(NBSM(I)) * H * H
          end do
          RMS = sqrt(RMS) / (real(NSM) * SIZEM)
C
        endif
C
      endif
C
      RETURN
      END
      subroutine YMDDOY(IY,IM,ID,IDOY,IERR)
C
CC    Calculates the Day of the Year for a given date
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:38
C
CA  author : KOD        date: 23-AUG-1989 19:39
CU  update : KOD        date: 28-MAR-1990 15:50
C
CT  status: tested
C
C   general description 
CG  Subroutine to calculate the Day of the Year for a given Cal.Date.
CG
CG  The calendar date is assumed to be given in the Gregorian system
CG  for dates starting with October 15, 1582, and in the Julian
CG  system until October 4, 1582 (in between it is undefined).
CG  For the year 1582 this subroutine will count the ACTUAL days
CG  of the year (which has 355 days) and will give a warning message
CG  for days after October 4.
CG
CG  This subroutine is valid for the years  -4712  to  +3500 ,
CG  and has been written according to
CG     van Flandern and Pulkkinen:
CG     'Low-Precision Formulae for Planetary Positions',
CG     Astrophysical Journal Supplement Series 41, 1979, p.392.
CG  and
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.23-25 .
CG
CG  The input date will be checked for validity.
C
C   call_var.          type I/O description 
CP  IY                  I4  I   year         (-4712 .. +3500)
CP  IM                  I4  I   month        (1 .. 12)
CP  ID                  I4  I   day of month (1 .. 28,29,30,31)
CP  IDOY                I4    O day of year  (1 .. 365,366)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  CALJD1              SR      converts Calendar Date into Julian Date
CR  ERRSTC              SR      continue error stack for one level
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DJD, DJD0
      REAL*4        SEC ! HFLAG, 
      INTEGER*4     IERR, IY, IM, ID, IDOY, IYEAR, IMONTH, IDAY,
     &              IHR, IMT, ISTAT!, IOS
      LOGICAL*4     LCHECK
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'YMDDOY'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     calculate the Julian Date:
C
      IYEAR  = IY
      IMONTH = IM
      IDAY   = ID
      IHR    = 0
      IMT    = 0
      SEC    = 0.0
      LCHECK = .true.
C
      call CALJD1(IYEAR,IMONTH,IDAY,IHR,IMT,SEC,LCHECK,DJD,ISTAT)
C
      if (ISTAT.gt.1) then
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
      endif
C
      if (IERR.lt.2) then
C
C       calculate the Julian Date for the first day of the year:
C
        IMONTH = 1
        IDAY   = 1
C
        call CALJD1(IYEAR,IMONTH,IDAY,IHR,IMT,SEC,LCHECK,DJD0,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        endif
C
C       determine the difference:
C
        IDOY = nint(DJD - DJD0) + 1
C
C       Calendar reform:
C
        if (IY.eq.1582 .and. IDOY.gt.277) then
          IERR = max(IERR,1)
          call WRERR(RTNAME,IERR,
     &        'Warning: Calendar Reform --> Day of the Year not unique')
        endif
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      subroutine YMDHMS(DT,IY,IM,ID,IHR,IMT,SEC,IERR)
C
CC    Splits the calendar date into its components
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:38
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 28-MAR-1990 15:52
C
CT  status: tested (02.08.1989 / K.Dennerl)
C
C   general description 
CG  Subroutine to split the calendar date, supplied as a single
CG  number in the form YYYYMMDD.HHMMSSS, into its components.
CG  The years may cover the range -9999 .. +9999.
CG  In case of negative years only the year IY will be negative.
CG  Resolution (given by REAL*8): 10~msec.
C
C   call_var.          type I/O description 
CP  DT                  R8  I   calendar date (YYYYMMDD.HHMMSSS)
CP  IY                  I4    O year
CP  IM                  I4    O month
CP  ID                  I4    O day
CP  IHR                 I4    O hour
CP  IMT                 I4    O minute
CP  SEC                 R4    O second
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description 
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
C
      REAL*8        DT, DH
      REAL*4        SEC ! HFLAG, 
      INTEGER*4     IERR, IY, IM, ID, IHR, IMT, I1, I2, ISEC1,
     &              ISEC2!, IOS
      LOGICAL*4     LPLUS
      CHARACTER*8   RTNAME, CHDAT1, CHDAT2
C
      DATA          RTNAME /'YMDHMS'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
C     extract sign:
C
      LPLUS = (DT.ge.0.D0)
      DH = dabs(DT)
C
C     check length of calendar date:
C
      if (DH.gt.99991231.2359599D0) then
C
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in IY')
C
      else
C
C       split according to decimal point:
C
        I1 = int(DH)
        I2 = nint((DH-int(DH))*1.D8)
        write(CHDAT1,'(I8.8)') I1
        write(CHDAT2,'(I8.8)') I2
C
C       read individual parts:
C
        read(CHDAT1,'(I4,I2,I2)') IY, IM, ID
        read(CHDAT2,'(I2,I2,I2,I2)') IHR, IMT, ISEC1, ISEC2
C
C       reconstruct seconds:  (rounded to 10 msec)
C
        SEC = real(ISEC1) + real(ISEC2)*1.E-2
C
C       (rough) check for valid calendar date:
C
        if (IM.lt.1 .or. IM.gt.12 .or. ID.lt.1 .or. ID.gt.31 .or.
     &      IHR.gt.23 .or. IMT.gt.59 .or. SEC.gt.60.0) then
          IERR = max(IERR,2)
          call WRERR(RTNAME,IERR,'Error in IM, ID, IHR, IMT, SEC')
        endif
C
C       add sign:
C
        if (.not. LPLUS)  IY = -IY
C
      endif
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END
      SUBROUTINE ZONCEN(NZONE,NLAT,NLON,NOMRA,NOMDE,IERR)
C
CC    Calculates the field center from the lat/long zone number/counter
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:38
C
CA  author : KOD        date: 14-JUN-1994
CU  update : GRU        date: 4-AUG-1994 10:28
C
CT  status: not tested
C
C   general description 
CG  Subroutine to calculate the field center from the latitude zone
CG  number and longitude zone counter.
C
C   call_var.          type I/O description 
CP  NZONE               I4  ?/? number of zones on the the sphere
CP                              including the equatorial zone and
CP                              the polar caps
CP                              ( NZONE > 0  and  not even )
CP  NLAT                I4  ?/? latitude zone number
CP  NLON                I4  ?/? longitude zone counter
CP  NOMRA               I4    O RA  of center of field [0.5 arcsec]
CP  NOMDE               I4    O DEC of center of field [0.5 arcsec]
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:IGENL.INC         
C
C   routines_called    type     description 
CR  ERRSTC              SR      continue error stack for one level
CR  SMCEN3              SR      
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:IGENL.INC'
c
      REAL*8        DALPH0, DDELT0
      INTEGER*4     NZONE, NLAT, NLON, NOMRA, NOMDE, IERR, ISTAT
      LOGICAL       LCEN
      CHARACTER*8   RTNAME
c
      DATA          RTNAME /'ZONCEN'/
c
c     R.A. = 0.0 is at the field boundary:
c
      PARAMETER ( LCEN = .FALSE. )
c
      IERR = 0
c
      CALL SMCEN3(NZONE,NLAT,NLON,LCEN,DALPH0,DDELT0,ISTAT)
C
      IF (ISTAT.NE.0) THEN
        IERR = MAX(IERR,ISTAT)
        CALL ERRSTC(RTNAME,10)
      ELSE
        NOMRA = NINT(DALPH0*7200.D0)
        NOMDE = NINT(DDELT0*7200.D0)
      ENDIF
C
      RETURN
      END

      SUBROUTINE ZONLIM(NZONE,NLAT,NLON,NRAMIN,NRAMAX,NDEMIN,NDEMAX,
     &                  IERR)
C
CC    Calc. the field boundary from the lat/long zone number/counter
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:38
C
CA  author : KOD        date: 14-JUN-1994
CU  update : GRU        date: 4-AUG-1994 10:30
C
CT  status: not tested
C
C   general description 
CG  Subroutine to calculate the field boundaries from the latitude zone
CG  number and longitude zone counter.
C
C   call_var.          type I/O description 
CP  NZONE               I4  ?/? number of zones on the the sphere
CP                              including the equatorial zone and
CP                              the polar caps
CP                              ( NZONE > 0  and  not even )
CP  NLAT                I4  ?/? latitude zone number
CP  NLON                I4  ?/? longitude zone counter
CP  NRAMIN              I4    O western RA field boundary  [0.5 arcsec]
CP  NRAMAX              I4    O eastern RA field boundary  [0.5 arcsec]
CP  NDEMIN              I4    O southern Dec field bound.  [0.5 arcsec]
CP  NDEMAX              I4    O northern Dec field bound.  [0.5 arcsec]
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description 
CI  R_INCLUDE:IGENL.INC         
C
C   routines_called    type     description 
CR  ERRSTC              SR      continue error stack for one level
CR  SMBND3              SR      
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:IGENL.INC'
c
      REAL*8        DALPHM, DALPHP, DDELTM, DDELTP
      INTEGER*4     NZONE, NLAT, NLON, NRAMIN, NRAMAX, NDEMIN, NDEMAX,
     &              IERR, ISTAT
      LOGICAL       LCEN
      CHARACTER*8   RTNAME
c
      DATA          RTNAME /'ZONLIM'/
c
c     R.A. = 0.0 is at the field boundary:
c
      PARAMETER ( LCEN = .FALSE. )
c
      IERR = 0
c
      CALL SMBND3(NZONE,NLAT,NLON,LCEN,DALPHM,DALPHP,DDELTM,DDELTP,
     &            ISTAT)
C
      IF (ISTAT.NE.0) THEN
        IERR = MAX(IERR,ISTAT)
        CALL ERRSTC(RTNAME,10)
      ELSE
        NRAMIN = NINT(DALPHM*7200.D0)
        NRAMAX = NINT(DALPHP*7200.D0)
        NDEMIN = NINT(DDELTM*7200.D0)
        NDEMAX = NINT(DDELTP*7200.D0)
      ENDIF
C
      RETURN
      END

      real*8 function DRADRS(DRAD)
C
CC    'Resets' radians to the interval  0 .. 2 * pi
C
C************************ FFORM VERSION 1.2 ************  4-DEC-98 16:35
C
CA  author : KOD        date: 19-JUL-1989 18:31
CU  update : KOD        date: 29-MAR-1990 15:08
C
CT  status: tested
C
C   general description 
CG  'Resets' radians to the interval  0  .. 2 * pi.
C
C   call_var.          type I/O description 
CP  DRAD                R8  I   radians (may be negative)
CP  DRADRS              R8    O radians inside the interval  0 .. 2 * pi
C
C   include_block name          description 
CI  R_INCLUDE:ICGENL.INC        general parameter common block
CI  R_INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description 
CR  RESVAL              SR      'resets' periodic variables
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
!      INCLUDE 'R_INCLUDE:ICGENL.INC'
!      INCLUDE 'R_INCLUDE:IGENL.INC'
C
      REAL*8        DRAD, DXMIN, DXMAX, DXOLD, DXNEW
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'DRADRS'/
C
CD     IF (HFLAG(RTNAME,'in: <input_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,101,IOSTAT=IOS) !<input_parameters>
CD 101   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      IERR = 0
C
      DXMIN = 0.D0
      DXMAX = D2PI
      DXOLD = DRAD
C
      CALL RESVAL(DXMIN,DXMAX,DXOLD,DXNEW,ISTAT)
C
      if (ISTAT.gt.1) then
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
      endif
C
      if (IERR.lt.2) DRADRS = DXNEW
C
CD     IF (HFLAG(RTNAME,'out: <output_parameters> ').NE.0) THEN
CD       WRITE (OUTPT,9901,IOSTAT=IOS) !<output_parameters>
CD9901   FORMAT (  )
CD       CALL WRFLAG(OUTPT)
CD     ENDIF
C
      RETURN
      END function dradrs

      subroutine BACOR1(DJD,DJD0,DALPHA,DDELTA,DTSEC,IERR)
C
CC    Calculates the barycentric correction for a given object and date
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:43
C
CA  author : KOD        date: 12-OCT-1989 17:24
CU  update : KOD        date: 28-MAR-1990 19:20
CU  update : KOD        date: 22-OCT-1991 11:33
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the time correction which has to be
CG  ADDED to reduce the photon arrival times measured at the
CG  centre of the earth to the center of the sun - not to the solar
CG  system barycenter ! @@@@@@
CG
CG  This subroutine can be used for converting the pulse arrival
CG  times at any point on earth or in a low-earth orbit (h < 1000 km)
CG  if an accuracy of 25 msec is sufficient.
CG
CG  Without this correction the difference between topocentric and
CG  barycentric photon arrival times can became as large as 500 sec.
C
C   call_var.          type I/O description
CP  DJD                 R8  I   Julian Date for the baryc. correction
CP  DJD0                R8  I   Julian Date defining the equinox
CP
CP  DALPHA              R8  I   right ascension at DJD0   (radians)
CP  DDELTA              R8  I   declination at DJD0       (radians)
CP
CP  DTSEC               R8    O time correction           (seconds)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  CSUN02              SR      calculates rectangular sun coordinates
CR  POLRE1              SR      converts polar into rectang. coord.
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DJD, DJD0, DALPHA, DDELTA, DTSEC, DX, DY, DZ,
     &              DR, DTH, DX0, DY0, DZ0, DF, DJ, DJ0, DA ! ,DL, DB
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      CHARACTER*8   RTNAME
C
      PARAMETER ( DF = 499.004D0 )
C
      DATA          RTNAME /'BACOR1'/
C
      IERR = 0
C
      if (DALPHA.lt.0.d0 .or. DALPHA.gt.D2PI .or.
     &    abs(DDELTA).gt.DPID2) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input coordinates')
      endif
C
      if (IERR.lt.2) then
C
C       calculate the rectangular coordinates of the sun:
C
        DJ  = DJD
        DJ0 = DJD0
        call CSUN02(DJ,DJ0,DX,DY,DZ,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
      if (IERR.lt.2) then
C
C       calculate the rectangular coordinates of the star:
C
        DR = 1.D0
        DTH = DPID2 - DDELTA
        DA = DALPHA
        call POLRE1(DR,DA,DTH,DX0,DY0,DZ0,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        endif
C
      endif
C
      if (IERR.lt.2) then
C
        DR = DX * DX0 + DY * DY0 + DZ * DZ0
        DTSEC = -DR * DF
C
      endif
C
      RETURN
      end subroutine bacor1
      


      subroutine CALJD3(chdat,LCHECK,DJD,IERR)
C
CC    Converts the calendar date to the Julian Date
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:43
C
CA  author : KOD        date: 15-NOV-2016
C
C   general description
CG  Subroutine to calculate the Julian Date out of the calendar date.
CG  The calendar date is assumed to be given in the Gregorian system
CG  for dates starting with October 15, 1582, and in the Julian
CG  system until October 4, 1582 (in between it is undefined).
CG
CG  This subroutine is valid for the years  -4712  to  +3500 ,
CG  and has been written according to
CG     van Flandern and Pulkkinen:
CG     'Low-Precision Formulae for Planetary Positions',
CG     Astrophysical Journal Supplement Series 41, 1979, p.392.
CG  and
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.23-25 .
CG
CG  For negative years all the other parameters must be positive.
CG  An extended input check is possible for valid dates.
C
C   call_var.          type I/O description
cp  chdat               c19 i  yyyy-mm-ddThh:mm:ss
CP
CP  LCHECK              L4  I   extended input check if LCHECK=.TRUE.
CP
CP  DJD                 R8    O corresponding Julian Date
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description
CR  CNVFRC              SR      conversion to fractional number
CR  JDCAL1              SR      converts Jul. Date into Calendar Date
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
C
      INTEGER*4     IY, IM, ID, IERR, IYR, IMTH, IDAY, IHR, IMT, ISTAT,
     &              JD, IYY, IMM, IHR1, IMT1, IDAT, isc!, IOS
      REAL*4        SEC, SEC1 ! HFLAG, 
      REAL*8        DJD, DFRC, DH
      LOGICAL*4     LCHECK, LPLUS
      character*(*) chdat
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CALJD1'/
C
      IERR = 0
c
      read(chdat,'(i4,5(1x,i2))') iy,im,id,ihr,imt,isc
      sec = real(isc)
C
CR!   if wrong input
C
      if ( IY .lt.-4712 .or. IY .gt.3500 .or.
     &     IM .lt.1     .or. IM .gt.12   .or.
     &     ID .lt.1     .or. ID .gt.31   .or.
     &     IHR.lt.0     .or. IHR.gt.24   .or.
     &    (IHR.eq.24 .and. (IMT.gt.0 .or. SEC.gt.0.0)) .or.
     &     IMT.lt.0     .or. IMT.gt.59   .or.
     &     SEC.lt.0.0   .or. SEC.gt.60.0 )   then
C
        IERR = 2
        WRITE(OUTPT,9)  IY,IM,ID
9       FORMAT('Error in input parameters IY,IM,ID ',3(I6,1X))
        call WRERR(RTNAME,IERR,OUTPT)
C
      endif
C
C
      if  (IERR.lt.2)  then
C
C       calculate the Julian Day for 0:00 UT
C
        IDAT = isign((iabs(IY)*100+IM)*100+ID,IY)
C
        if (IDAT.ge.15821015) then
C
C         assume Gregorian calendar date as input:
C
          JD = 367*IY - 7*(IY+(IM+9)/12)/4 - 3*((IY+(IM-9)/7)/100+1)/4
     &       + 275*IM/9 + ID + 1721029
C
        else if (IDAT.le.15821004) then
C
C         assume Julian calendar date as input:
C
          if (IM.gt.2) then
            IYY = IY
            IMM = IM
          else
            IYY = IY - 1
            IMM = IM + 12
          endif
C
          if (IYY.gt.0) then
            DH=0.D0
          else
            DH=0.75D0
          endif
C
          JD = idint(365.25d0*dble(IYY)-DH) +
     &         idint(30.6001d0*dble(IMM+1)) +
     &         ID + 1720995
C
          else
C
            IERR = max(IERR,2)
            WRITE(OUTPT,10)  IDAT
10          FORMAT('Error in date: calendar reform ',I10)
            call WRERR(RTNAME,IERR,OUTPT)
C
        endif
C
C       calculate the fractional part:
C
        LPLUS = .true.
        IHR1 = IHR
        IMT1 = IMT
        SEC1 = SEC
        call CNVFRC(LPLUS,IHR1,IMT1,SEC1,DFRC,ISTAT)
C
CN!     ERROR handling
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
C       add both parts
C
        DJD = dble(JD) + DFRC/24.D0 - 0.5D0
C
        IF (LCHECK) THEN
C
CJ!       extended input check
C
          DH = dble(JD) - 0.5D0*0.0
          call JDCAL1(DH,IYR,IMTH,IDAY,IHR1,IMT1,SEC1,ISTAT)
C
CN!       ERROR handling
C
          if  (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,20)
          endif
C
CR!       if inconsistent results
C
          if (IYR .ne.IY  .or. IMTH.ne.IM .or. IDAY.ne.ID) then
            IERR = max(IERR,2)
            WRITE(OUTPT,11) IYR,IY,IMTH,IM,IDAY,ID
11          FORMAT('Error: Date does not exist ',6(I8,1X))
            call WRERR(RTNAME,IERR,OUTPT)
          endif
        endif
      endif
C
CH!   CJD
C
      RETURN
      END

      subroutine CHDAT3(IY,IM,ID,IHR,IMT,SEC,LCHECK,CHDAT,IERR)
C
CC    Produces a string of the form 'YYYY.MM.DD / HH:MM:SS.SS'
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:52
C
CA  author : KOD        date: 15-NOV-2016
C
CT  status: tested
C
C   general description
CG
CG  Subroutine to produce a string of the form
CG  'YYYY.MM.DD / HH:MM:SS.SS'  (Maximum resolution: 10 msec).
CG  There will be a rough check on the validity of the given date;
CG  a detailed check is also possible (see LCHECK).
C
C   call_var.          type I/O description
CP  IY                  I4  I   year         (-999 .. +9999)
CP  IM                  I4  I   month        (1 .. 12)
CP  ID                  I4  I   day of month (1 .. 28,29,30,31)
CP  IHR                 I4  I   hours        (0 .. 23)
CP  IMT                 I4  I   minutes      (0 .. 59)
CP  SEC                 R4  I   seconds      (0.0 .. 59.999...)
CP
CP  LCHECK              L4  I   extended input check if LCHECK=.true.
CP
CP  CHDAT               C   I/O string containing the calendar date
CP                              'YYYY' to 'YYYY.MM.DD / HH:MM:SS.SS'
CP
CP                               possible lengths 'L' of CHDAT :
CP                               L = 19 : 'yyyy-mm-ddThh:mm:ss'
CP
CP                               The values for the year, month
CP                               and day will not be rounded,
CP                               the values for the hour, minute
CP                               and second will be rounded.
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$INCLUDE:ICGENL.INC        general parameter common block
C
C   routines_called    type     description
CR  JDCAL1              SR      Julian Date to Calendar Date
CR  CALJD1              SR      Calendar Date to Julian Date
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
C
      REAL*8        DJD
      REAL*4        SEC, F, FSEC, SEC1 ! , H, HFLAG 
      INTEGER*4     IERR, IY, IM, ID, IHR, IMT, N, NDIG, ISEC,
     &              ISEC1, ISEC2, IY1, IM1, ID1, IHR1, IMT1,
     &              ISTAT! ,IOS, I1, I2
      LOGICAL*4     LCHECK, LCHK
c      CHARACTER*3   CHDIG
      CHARACTER*(*) CHDAT
c      CHARACTER*70  CHFMT
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CHDATE'/
C
      IERR = 0
C
      N = 19
C
C     (rough) check for meaningful date:
c
      if (IY .lt.-999 .or.  IY.gt.9999  .or.
     &    IM .lt.1    .or.  IM.gt.12    .or.
     &    ID .lt.1    .or.  ID.gt.31    .or.
     &    IHR.lt.0    .or.  IHR.gt.23   .or.
     &    IMT.lt.0    .or.  IMT.gt.59   .or.
     &    SEC.lt.0.0  .or.  SEC.gt.60.0) then
C
        ISTAT = 2
        IERR = max(IERR,ISTAT)
        call WRERR(RTNAME,IERR,'Error in input parameters IY .. SEC')
C
      endif
C
C
C     precise check for valid date:
C
      if (IERR.lt.2) then
C
C       copy input parameters:
C
        IY1  = IY
        IM1  = IM
        ID1  = ID
        IHR1 = IHR
        IMT1 = IMT
        SEC1 = SEC
C
        if (LCHECK) then
C
          LCHK = LCHECK
          call CALJD1(IY1,IM1,ID1,IHR1,IMT1,SEC1,LCHK,DJD,ISTAT)
C
          if (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,10)
          endif
C
        endif
C
        SEC1 = SEC-1.E-6
C
C       round values:
C
        NDIG  = min( max(N-22,0), 3 )
        F     = real( 10**NDIG )
        ISEC  = nint( SEC1*F )
        FSEC  = real( ISEC ) / F
        ISEC1 = int( FSEC )
        ISEC2 = ISEC - ISEC1 * 10**NDIG
C
C
C       increase minutes:
C
        if ( (ISEC1.ge.60) .or.
     &       (N.le.18 .and. SEC1.gt.30.0) ) then
          SEC1  = 0.0
          ISEC1 = 0
          ISEC2 = 0
          IMT1 = IMT1 + 1
        endif
C
C
C       increase hours:
C
        if ( (IMT1.ge.60) .or.
     &       (N.le.15 .and.
     &        (IMT1.gt.30 .or. (IMT1.eq.30 .and. SEC1.gt.0.0))) ) then
          SEC1  = 0.0
          ISEC1 = 0
          ISEC2 = 0
          IMT1 = 0
          IHR1 = IHR1 + 1
        endif
C
C
C       increase the day:
C
        if (IHR1.eq.24) then
C
          IHR1 = 0
C
          LCHK=.true.
          call CALJD1(IY1,IM1,ID1,IHR1,IMT1,SEC1,LCHK,DJD,ISTAT)
C
          if (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,20)
          endif
C
          if (IERR.lt.2) then
            DJD = DJD + 1.D0
            call JDCAL1(DJD,IY1,IM1,ID1,IHR1,IMT1,SEC1,ISTAT)
C
            if (ISTAT.gt.1) then
              IERR = max(IERR,ISTAT)
              call ERRSTC(RTNAME,30)
            endif
C
          endif
        endif
C
C
C       assign result to CHDAT:
C
        write(chdat,'(i4,5(a,i2))')
     &        iy1,'-',im1,'-',id1,'T',ihr1,':',imt1,':',isec1
C
      endif
C
      RETURN
      END

      subroutine CMOON1(DJD,DLECL,DBECL,DPRLX,IERR)
C
CC    Calculates the ecl. coordinates and the parallax of the moon
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:54
C
CA  author : KOD        date: 31-AUG-1989 16:14
CU  update : KOD        date: 20-MAR-1990 14:39
C
CT  status: tested (01-SEP-1989 17:18 KOD)
C
C   general description
CG  Subroutine to calculate the ecliptical coordinates and the
CG  parallax of the moon, according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.147-153 .
CG
CGT    Accuracy:  10 arcsec in the ecliptical longitude,
CGT                3 arcsec in the ecliptical latitude,
CGT              0.2 arcsec in the parallax.
C
C   call_var.          type I/O description
CP  DJD                 R8  I   Julian Date ( Ephemeris time !)
CP
CP  DLECL               R8    O ecliptical longitude (radians, 0..2*pi)
CP  DBECL               R8    O ecliptical  latitude (radians,-0.5..0.5)
CP  DPRLX               R8    O parallax             (radians, < 0.02 )
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  DDEGRS              R8      'resets' degrees
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
!HB   REAL*8        DJD, DJD0, DT, DYEAR, DDEGRS, DLECL, DBECL,
      REAL*8        DJD, DJD0, DT, DYEAR, DLECL, DBECL,
     &              DPRLX, DLL, DM, DMM, DD, DF, DOM, DH, DF2, DF3,
     &              DD2, DD3, DD4, DM2, DMM2, DMM3, DMM4, DE1, DE2, DL,
     &              DB, DP, DDLL(4), DDM(4), DDMM(4), DDD(4), DDF(4),
     &              DDOM(4)
c      REAL*4        HFLAG
      INTEGER*4     IERR !,IOS
      CHARACTER*8   RTNAME
C
      PARAMETER ( DJD0 = 2415020.D0 )
      PARAMETER ( DYEAR =  36525.D0 )
C
      DATA          RTNAME /'CMOON1'/
C
      DATA DDLL / 270.434164D0, 481267.8831D0,-0.001133D0, 0.0000019D0 /
      DATA DDM  / 358.475833D0,  35999.0498D0,-0.000150D0,-0.0000033D0 /
      DATA DDMM / 296.104608D0, 477198.8491D0, 0.009192D0, 0.0000144D0 /
      DATA DDD  / 350.737486D0, 445267.1142D0,-0.001436D0, 0.0000019D0 /
      DATA DDF  /  11.250889D0, 483202.0251D0,-0.003211D0,-0.0000003D0 /
      DATA DDOM / 259.183275D0,  -1934.1420D0, 0.002078D0, 0.0000022D0 /
C
      IERR = 0
C
      if (DJD.lt.0.D0 .or. DJD.gt.3.D6) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
      if (IERR.lt.2) then
C
        DT = (DJD - DJD0) / DYEAR
C
        DLL = DDLL(1) + (DDLL(2) + (DDLL(3) + DDLL(4) * DT) * DT ) * DT
        DM  = DDM (1) + (DDM (2) + (DDM (3) + DDM (4) * DT) * DT ) * DT
        DMM = DDMM(1) + (DDMM(2) + (DDMM(3) + DDMM(4) * DT) * DT ) * DT
        DD  = DDD (1) + (DDD (2) + (DDD (3) + DDD (4) * DT) * DT ) * DT
        DF  = DDF (1) + (DDF (2) + (DDF (3) + DDF (4) * DT) * DT ) * DT
        DOM = DDOM(1) + (DDOM(2) + (DDOM(3) + DDOM(4) * DT) * DT ) * DT
C
        DLL = DDEGRS(DLL)
        DM  = DDEGRS(DM )
        DMM = DDEGRS(DMM)
        DD  = DDEGRS(DD )
        DF  = DDEGRS(DF )
        DOM = DDEGRS(DOM)
C
        DH = sin ( 0.8936086D0 + 0.3525565D0 * DT )
C
        DLL = DLL + 0.000233 * DH
        DM  = DM  - 0.001778 * DH
        DMM = DMM + 0.000817 * DH
        DD  = DD  + 0.002011 * DH
C
        DH = sin (6.048613D0 + ( 2.319019D0 - 1.601008D-4 * DT ) * DT)
C
        DH = 0.003964D0 * DH
C
        DLL = DLL + DH
        DMM = DMM + DH
        DD  = DD  + DH
        DF  = DF  + DH
C
        DH = sin ( DOM * DPI180 )
C
        DLL = DLL + 0.001964D0 * DH
        DMM = DMM + 0.002541D0 * DH
        DD  = DD  + 0.001964D0 * DH
        DF  = DF  - 0.024691D0 * DH
C
        DH = sin ( (DOM + 275.05D0 - 2.30D0) * DPI180 )
C
        DF = DF - 0.004328D0 * DH
C
        DM  = DM  * DPI180
        DMM = DMM * DPI180
        DD  = DD  * DPI180
        DF  = DF  * DPI180
        DOM = DOM * DPI180
C
        DF2  = DF   + DF
        DF3  = DF2  + DF
        DD2  = DD   + DD
        DD3  = DD2  + DD
        DD4  = DD3  + DD
        DM2  = DM   + DM
        DMM2 = DMM  + DMM
        DMM3 = DMM2 + DMM
        DMM4 = DMM3 + DMM
C
        DE1 = 1.D0 - ( 0.002495D0 + 0.00000752D0 * DT ) * DT
        DE2 = DE1 * DE1
C
        DL = DLL
        DL = DL + 6.288750D0 * sin(DMM)
        DL = DL + 1.274018D0 * sin(DD2-DMM)
        DL = DL + 0.658309D0 * sin(DD2)
        DL = DL + 0.213616D0 * sin(DMM2)
        DL = DL - 0.185596D0 * sin(DM)           * DE1
        DL = DL - 0.114336D0 * sin(DF2)
        DL = DL + 0.058793D0 * sin(DD2-DMM2)
        DL = DL + 0.057212D0 * sin(DD2-DM-DMM)   * DE1
        DL = DL + 0.053320D0 * sin(DD2+DMM)
        DL = DL + 0.045874D0 * sin(DD2-DM)       * DE1
        DL = DL + 0.041024D0 * sin(DMM-DM)       * DE1
        DL = DL - 0.034718D0 * sin(DD)
        DL = DL - 0.030465D0 * sin(DM+DMM)       * DE1
        DL = DL + 0.015326D0 * sin(DD2-DF2)
        DL = DL - 0.012528D0 * sin(DF2+DMM)
        DL = DL - 0.010980D0 * sin(DF2-DMM)
        DL = DL + 0.010674D0 * sin(DD4-DMM)
        DL = DL + 0.010034D0 * sin(DMM3)
        DL = DL + 0.008548D0 * sin(DD4-DMM2)
        DL = DL - 0.007910D0 * sin(DM-DMM+DD2)    * DE1
        DL = DL - 0.006783D0 * sin(DD2+DM)        * DE1
        DL = DL + 0.005162D0 * sin(DMM-DD)
        DL = DL + 0.005000D0 * sin(DM+DD)         * DE1
        DL = DL + 0.004049D0 * sin(DMM-DM+DD2)    * DE1
        DL = DL + 0.003996D0 * sin(DMM2+DD2)
        DL = DL + 0.003862D0 * sin(DD4)
        DL = DL + 0.003665D0 * sin(DD2-DMM3)
        DL = DL + 0.002695D0 * sin(DMM2-DM)       * DE1
        DL = DL + 0.002602D0 * sin(DMM-DF2-DD2)
        DL = DL + 0.002396D0 * sin(DD2-DM-DMM2)   * DE1
        DL = DL - 0.002349D0 * sin(DMM+DD)
        DL = DL + 0.002249D0 * sin(DD2-DM2)       * DE2
        DL = DL - 0.002125D0 * sin(DMM2+DM)       * DE1
        DL = DL - 0.002079D0 * sin(DM2)           * DE2
        DL = DL + 0.002059D0 * sin(DD2-DMM-DM2)   * DE2
        DL = DL - 0.001773D0 * sin(DMM+DD2-DF2)
        DL = DL - 0.001595D0 * sin(DF2+DD2)
        DL = DL + 0.001220D0 * sin(DD4-DM-DMM)    * DE1
        DL = DL - 0.001110D0 * sin(DMM2+DF2)
        DL = DL + 0.000892D0 * sin(DMM-DD3)
        DL = DL - 0.000811D0 * sin(DM+DMM+DD2)    * DE1
        DL = DL + 0.000761D0 * sin(DD4-DM-DMM2)   * DE1
        DL = DL + 0.000717D0 * sin(DMM-DM2)       * DE2
        DL = DL + 0.000704D0 * sin(DMM-DM2-DD2)   * DE2
        DL = DL + 0.000693D0 * sin(DM-DMM2+DD2)   * DE1
        DL = DL + 0.000598D0 * sin(DD2-DM-DF2)    * DE1
        DL = DL + 0.000550D0 * sin(DMM+DD4)
        DL = DL + 0.000538D0 * sin(DMM4)
        DL = DL + 0.000521D0 * sin(DD4-DM)        * DE1
        DL = DL + 0.000486D0 * sin(DMM2-DD)
C
        DB =      5.128189D0 * sin(DF)
        DB = DB + 0.280606D0 * sin(DMM+DF)
        DB = DB + 0.277693D0 * sin(DMM-DF)
        DB = DB + 0.173238D0 * sin(DD2-DF)
        DB = DB + 0.055413D0 * sin(DD2+DF-DMM)
        DB = DB + 0.046272D0 * sin(DD2-DF-DMM)
        DB = DB + 0.032573D0 * sin(DD2+DF)
        DB = DB + 0.017198D0 * sin(DMM2+DF)
        DB = DB + 0.009267D0 * sin(DD2+DMM-DF)
        DB = DB + 0.008823D0 * sin(DMM2-DF)
        DB = DB + 0.008247D0 * sin(DD2-DM-DF)     * DE1
        DB = DB + 0.004323D0 * sin(DD2-DF-DMM2)
        DB = DB + 0.004200D0 * sin(DD2+DF+DMM)
        DB = DB + 0.003372D0 * sin(DF-DM-DD2)     * DE1
        DB = DB + 0.002472D0 * sin(DD2+DF-DM-DMM) * DE1
        DB = DB + 0.002222D0 * sin(DD2+DF-DM)     * DE1
        DB = DB + 0.002072D0 * sin(DD2-DF-DM-DMM) * DE1
        DB = DB + 0.001877D0 * sin(DF-DM+DMM)     * DE1
        DB = DB + 0.001828D0 * sin(DD4-DF-DMM)
        DB = DB - 0.001803D0 * sin(DF+DM)         * DE1
        DB = DB - 0.001750D0 * sin(DF3)
        DB = DB + 0.001570D0 * sin(DMM-DM-DF)     * DE1
        DB = DB - 0.001487D0 * sin(DF+DD)
        DB = DB - 0.001481D0 * sin(DF+DM+DMM)     * DE1
        DB = DB + 0.001417D0 * sin(DF-DM-DMM)     * DE1
        DB = DB + 0.001350D0 * sin(DF-DM)         * DE1
        DB = DB + 0.001330D0 * sin(DF-DD)
        DB = DB + 0.001106D0 * sin(DF+DMM3)
        DB = DB + 0.001020D0 * sin(DD4-DF)
        DB = DB + 0.000833D0 * sin(DF+DD4-DMM)
        DB = DB + 0.000781D0 * sin(DMM-DF3)
        DB = DB + 0.000670D0 * sin(DF+DD4-DMM2)
        DB = DB + 0.000606D0 * sin(DD2-DF3)
        DB = DB + 0.000597D0 * sin(DD2+DMM2-DF)
        DB = DB + 0.000492D0 * sin(DD2+DMM-DM-DF) * DE1
        DB = DB + 0.000450D0 * sin(DMM2-DF-DD2)
        DB = DB + 0.000439D0 * sin(DMM3-DF)
        DB = DB + 0.000423D0 * sin(DF+DD2+DMM2)
        DB = DB + 0.000422D0 * sin(DD2-DF-DMM3)
        DB = DB - 0.000367D0 * sin(DM+DF+DD2-DMM) * DE1
        DB = DB - 0.000353D0 * sin(DM+DF+DD2)     * DE1
        DB = DB + 0.000331D0 * sin(DF+DD4)
        DB = DB + 0.000313D0 * sin(DD2+DF-DM+DMM) * DE1
        DB = DB + 0.000306D0 * sin(DD2-DM2-DF)    * DE2
        DB = DB - 0.000283D0 * sin(DMM+DF3)
C
        DH =      0.0004664 * cos(DOM)
        DH = DH + 0.0000754 * cos(DOM + 4.8005281D0 - 0.04014257D0*DT)
C
        DB = DB * (1.D0 - DH)
C
        DP =      0.950724
        DP = DP + 0.051818 * cos(DMM)
        DP = DP + 0.009531 * cos(DD2-DMM)
        DP = DP + 0.007843 * cos(DD2)
        DP = DP + 0.002824 * cos(DMM2)
        DP = DP + 0.000857 * cos(DD2+DMM)
        DP = DP + 0.000533 * cos(DD2-DM)          * DE1
        DP = DP + 0.000401 * cos(DD2-DM-DMM)      * DE1
        DP = DP + 0.000320 * cos(DMM-DM)          * DE1
        DP = DP - 0.000271 * cos(DD)
        DP = DP - 0.000264 * cos(DM+DMM)          * DE1
        DP = DP - 0.000198 * cos(DF2-DMM)
        DP = DP + 0.000173 * cos(DMM3)
        DP = DP + 0.000167 * cos(DD4-DMM)
        DP = DP - 0.000111 * cos(DM)              * DE1
        DP = DP + 0.000103 * cos(DD4-DMM2)
        DP = DP - 0.000084 * cos(DMM2-DD2)
        DP = DP - 0.000083 * cos(DD2+DM)          * DE1
        DP = DP + 0.000079 * cos(DD2+DMM2)
        DP = DP + 0.000072 * cos(DD4)
        DP = DP + 0.000064 * cos(DD2-DM+DMM)      * DE1
        DP = DP - 0.000063 * cos(DD2+DM-DMM)      * DE1
        DP = DP + 0.000041 * cos(DM+DD)           * DE1
        DP = DP + 0.000035 * cos(DMM2-DM)         * DE1
        DP = DP - 0.000033 * cos(DMM3-DD2)
        DP = DP - 0.000030 * cos(DMM+DD)
        DP = DP - 0.000029 * cos(DF2-DD2)
        DP = DP - 0.000029 * cos(DMM2+DM)         * DE1
        DP = DP + 0.000026 * cos(DD2-DM2)         * DE2
        DP = DP - 0.000023 * cos(DF2-DD2+DMM)
        DP = DP + 0.000019 * cos(DD4-DM-DMM)      * DE1
C
        DL    = DL * DPI180
        DLECL = DDEGRS(DL)
C
        DBECL = DB * DPI180
        DPRLX = DP * DPI180
C
      endif
C
      RETURN
      END

      subroutine CMOON2(DJD,LMEAN,DALPHA,DDELTA,DDIST,IERR)
C
CC    Calculates the equ. coordinates and the distance of the moon
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:54
C
CA  author : KOD        date: 05-SEP-1989 16:15
CU  update : KOD        date: 28-MAR-1990 19:50
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the equatorial coordinates and the
CG  distance of the moon, according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.147-153 .
CG
CG  To calculate the TRUE position, use the MEAN obliquity of the
CG  ecliptic (normal case);
CG  for the APPARENT position (affected by aberration and nutation)
CG  use the TRUE obliquity.
CG  In the equatorial coordinates shall refer to a standard equinox
CG  (1950, 2000, ...), the obliquity of the ecliptic for this epoch
CG  should be used.
CG  (Jean Meeus, 'Astronomical Formulae for Calculators',
CG                third edition, 1985, p.43-44)
CG
CG  Accurracy: 10 arcsec in position, 25 km in distance
C
C   call_var.          type I/O description
CP  DJD                 R8  I   Julian Date ( Ephemeris time !)
CP
CP  LMEAN               L4  I   mean or true obliquity of the ecliptic:
CP                                LMEAN = .true.  : mean obliquity
CP                                LMEAN = .false. : true obliquity
CP
CP  DALPHA              R8    O right ascension      (radians)
CP  DDELTA              R8    O declination          (radians)
CP  DDIST               R8    O distance             (km)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
C
C   routines_called    type     description
CR  CMOON1              SR      calculates ecl. position of the moon
CR  EQUECL              SR      converts ecl. to equatorial coorindates
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
C
      LOGICAL*4     LMEAN, LDIR, LMN
      REAL*8        DJD, DALPHA, DDELTA, DDIST, DLECL, DBECL, DPRLX, DJ,
     &              DA, DD
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CMOON2'/
C
      IERR = 0
C
      if (DJD.lt.0.D0 .or. DJD.gt.3.D6) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
      if (IERR.lt.2) then
C
        DJ = DJD
        call CMOON1(DJ,DLECL,DBECL,DPRLX,ISTAT)
C
        if (ISTAT.gt.1) then
C
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
C
        else
C
          LDIR = .false.
          LMN = LMEAN
          call EQUECL(DJ,LMN,LDIR,DLECL,DBECL,DA,DD,ISTAT)
C
          if (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,20)
          else
            DALPHA = DA
            DDELTA = DD
            DDIST = 6378.14D0 / sin (DPRLX)
          endif
C
        endif
C
      endif
C
      RETURN
      END

      subroutine CPSAB1(DJD,DLOLD,DBOLD,LDIR,DLNEW,DBNEW,IERR)
C
CC  corrects ecliptical positions for the effect of aberration
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:55
C
CA  author : KOD        date: 06-MAR-1991 14:45
CU  update : KOD        date: 22-OCT-1991 10:47
CU  update : KOD        date: 25-OCT-1991 14:40
C
CT  status: not tested
C
C   general description
CG
CG  Subroutine to correct ecliptical positions for the effect of
CG  (annual) aberration. It can be applied in both directions (see
CG  LDIR). Here the term 'uncorrected' specifies the true position
CG  while 'corrected' refers to the apparent position.
CG  The method applied has been described by F.Schmeidler, in:
CG    G.D.Roth: `Handbuch fuer Sternfreunde' (I), p.305-306,
CG    (Springer Verlag, vierte Auflage)
CG  It provides only an approximate correction since the Earth is
CG  assumed to orbit the sun in a circle with constant velocity.
CG  ||| This method is not applicable to the ecliptical poles |||
CG
C
C   call_var.          type I/O description
CP
CP  DJD                 R8  I   Julian Date for the correction
CP
CP  DLOLD               R8  I   old ecl. longitude  (radians)
CP  DBOLD               R8  I   old ecl. latitude   (radians)
CP
CP  LDIR                L   I   direction of the correction:
CP                               = .true. :  uncorrected ---> corrected
CP                               = .false.:  uncorrected <--- corrected
CP
CP  DLNEW               R8    O new longitude  (radians)
CP  DBNEW               R8    O new latitude   (radians)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  CSUN01              SR      calculates the ecl. coord. of the sun
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DJD, DLOLD, DBOLD, DLNEW, DBNEW, DLONG, DDIST,
!HB     &           DC, DLDIFF, DL, DB, DH, DRADRS
     &              DC, DLDIFF, DL, DB, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL       LDIR
      CHARACTER*8   RTNAME
C
C     constant for aberration (radians):
C
      PARAMETER   ( DC = 9.93674D-5 )
C
      DATA          RTNAME /'CPSAB1'/
C
      IERR = 0
C
      if (DLOLD.lt.0.D0 .or. DLOLD.gt.D2PI) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Warning: unusual input value for DLOLD')
      endif
C
      if (dabs(DBOLD).gt.DPID2) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in the input parameter DBOLD')
      endif
C
      if (DJD.lt.0.D0 .or. DJD.gt.3.D6) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate the ecliptical longitude of the sun:
C
        call CSUN01(DJD,.true.,DLONG,DDIST,ISTAT)
C
        if (ISTAT.ne.0) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,IERR)
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
        DLDIFF = DLOLD - DLONG
C
        DH = dcos(DBOLD)
        if (dabs(DH).lt.1.D-6) DH = dsign(1.D-6,DH)
C
        DL = -DC * dcos(DLDIFF) / DH
        DB =  DC * dsin(DLDIFF) * dsin(DBOLD)
C
        if (LDIR) then
          DLNEW = DRADRS ( DLOLD + DL )
          DBNEW = DBOLD + DB
        else
          DLNEW = DRADRS ( DLOLD - DL )
          DBNEW = DBOLD - DB
        endif
C
        if (dabs(DBNEW).gt.DPID2) then
          IERR = max(IERR,2)
          call WRERR(RTNAME,IERR,
     &        'The method used is not applicable for the ecl. poles')
          DLNEW = DRADRS ( DLNEW + DPI )
          DBNEW = dsign(DPI-dabs(DBNEW),DBNEW)
        endif
C
      endif
C
C
      RETURN
      END

      subroutine CPSAB2(DVX,DVY,DVZ,LDIR,DXOLD,DYOLD,DZOLD,
     &                  DXNEW,DYNEW,DZNEW,ISTAT)
C
CC  corrects positions for the effect of (non-relativistic) aberration
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:55
C
CA  author : KOD        date: 22-OCT-1991 10:47
CU  update : KOD        date: 25-OCT-1991 15:19
C
CT  status: not tested
C
C   general description
CG
CG  Subroutine to correct positions for the effect of (non-relativistic)
CG  aberration. It can be applied in both directions (see LDIR).
CG  All positions and velocities refer to a rectangular coordinate
CG  system which can be chosen arbitrarily but must be used in a
CG  consistent way. The velocities have to be specified in units of
CG  the speed of light. For the positions unit vectors should/will
CG  be used.
C
C   call_var.          type I/O description
CP
CP  DVX                 R8  I   velocity of the observer in x-direction
CP  DVY                 R8  I   velocity of the observer in y-direction
CP  DVZ                 R8  I   velocity of the observer in z-direction
CP
CP  LDIR                L   I   direction of the correction:
CP                               = .true. :  true pos. --> apparent pos.
CP                               = .false.:  true pos. <-- apparent pos.
CP
CP  DXOLD               R8  I   x-component of the original position
CP  DYOLD               R8  I   y-component of the original position
CP  DZOLD               R8  I   z-component of the original position
CP
CP  DXNEW               R8    O x-component of the resulting position
CP  DYNEW               R8    O y-component of the resulting position
CP  DZNEW               R8    O z-component of the resulting position
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DVX, DVY, DVZ, DXOLD, DYOLD, DZOLD, DXNEW, DYNEW,
     &              DZNEW, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT!, IOS
      LOGICAL       LDIR
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CPSAB2'/
C
C
      IERR = 0
      ISTAT = 0 ! 20210903 IMS
      if(LDIR) then
        ! 20210903 IMS: do nothing - this just avoids warnings from -Wall about the unused argument.
      end if
C
      DH = dmax1(dabs(DVX),dabs(DVY),dabs(DVZ))
C
      if (DH.gt.1.D0) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in velocity')
      endif
C
      if (DH.gt.0.1D0) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,
     &      'Warning: correction for aberration is non-relativistic')
      endif
C
      DH = DXOLD * DXOLD + DYOLD * DYOLD + DZOLD * DZOLD
C
      if (DH.eq.0.D0) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in position')
      endif
C
      if (dabs(DH-1.D0).gt.1.D-3) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,
     &      'Warning: Position is not specified by a unit vector')
      endif
C
C
      if (IERR.lt.2) then
C
        DH = 1.D0 / DH
C
        DXNEW = DXOLD * DH + DVX
        DYNEW = DYOLD * DH + DVY
        DZNEW = DZOLD * DH + DVZ
C
        DH = DXNEW * DXNEW + DYNEW * DYNEW + DZNEW * DZNEW
C
        DH = 1.D0 / DH
C
        DXNEW = DXNEW * DH
        DYNEW = DYNEW * DH
        DZNEW = DZNEW * DH
C
      endif
C
C
      RETURN
      END

      subroutine CSUN01(DJD,LMEAN,DLONG,DDIST,IERR)
C
CC    Calculates the longitude and distance of the sun for a given JD
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:56
C
CA  author : KOD        date: 07-SEP-1989 10:39
CU  update : KOD        date: 20-MAR-1990 14:43
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the true/apparent longitude of the sun
CG  for a given Julian Date, according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.79-80 .
CG
CG  The true longitude is referred to the mean equinox of the date,
CG  required e.g. for calculating the geocentric planetary positions.
CG  The apparent longitude is referred to the true equinox of the
CG  date, corrected for nutation and aberration.
CG
CG  The longitude with respect to the standard equinox of 1950.0 can
CG  be calculated with a sufficient accuracy for the 20th century as
CG
CG    DLONG(1950.0) = DLONG - 2.4365E-4 * (year - 1950)  (radians)
CG
CG  Accuracy: 2 arc seconds
C
C   call_var.          type I/O description
CP  DJD                 R8  I   Julian Date
CP
CP  LMEAN               L4  I   true or app. longitude:
CP                                LMEAN = .true.  : true longitude
CP                                LMEAN = .false. : app. longitude
CP
CP  DLONG               R8    O longitude of the sun    (radians)
CP  DDIST               R8    O distance  of the sun  (astr. units)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  EANOM1              SR      calculates eccentric anomaly
CR  ETANOM              SR      converts eccentric into true anomaly
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DJD, DLONG, DJD0, DYEAR, DT, DLL(3), DMM(4), DEE(3),
!HB     &           DL, DM, DE, DTM, DRADRS, DEM, DH, DOM, DDIST,
     &              DL, DM, DE, DTM, DEM, DH, DOM, DDIST,
     &              DHA, DHB, DHC, DHD, DHE, DHH, DCOR
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN, LDIR
      CHARACTER*8   RTNAME
C
      PARAMETER ( DJD0 = 2415020.D0 )
      PARAMETER ( DYEAR =  36525.D0 )
C
      DATA          RTNAME /'CSUN01'/
C
      DATA DLL / 279.69668D0, 36000.76892D0,  0.0003025D0 /
      DATA DMM / 358.47583D0, 35999.04975D0, -0.000150D0, -0.0000033D0 /
      DATA DEE / 0.01675104D0, -0.0000418D0, -0.000000126D0 /
C
      IERR = 0
C
      if (DJD.lt.0.D0 .or. DJD.gt.3.D6) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
      if (IERR.lt.2) then
C
        DT = (DJD - DJD0) / DYEAR
C
C       calculate the geocentric mean longitude of the sun:
C
        DL = DLL(1) + ( DLL(2) + DLL(3) * DT ) * DT
        DL = DL * DPI180
        DL = DRADRS(DL)
C
C       calculate the mean anomaly:
C
        DM = DMM(1) + ( DMM(2) + ( DMM(3) + DMM(4) * DT ) * DT ) * DT
        DM = DM * DPI180
        DM = DRADRS(DM)
C
C       calculate the eccentricity of the earth`s orbit:
C
        DE = DEE(1) + ( DEE(2) + DEE(3) * DT ) * DT
C
C       calculate the eccentric anomaly:
C
        call EANOM1(DM,DE,DEM,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        else
C
C         calculate the true anomaly:
C
          LDIR=.true.
          call ETANOM(DE,LDIR,DEM,DTM,ISTAT)
C
          if (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,20)
          else
            DH = DL + DTM - DM
            if (.not.LMEAN) then
              DOM = 4.52354D0 - 33.75715D0 * DT
              DH = DH - 9.9309D-5 - 8.360D-5 * sin(DOM)
            endif
C
            DHA = 153.23D0 + 22518.7541D0 * DT
            DHB = 216.57D0 + 45037.5082D0 * DT
            DHC = 312.69D0 + 32964.3577D0 * DT
            DHD = 350.74D0 + (445267.1142D0 - 0.00144D0 * DT) * DT
            DHE = 231.19D0 + 20.20D0 * DT
            DHH = 353.40D0 + 65928.7155D0 * DT
C
            DHA = DHA * DPI180
            DHB = DHB * DPI180
            DHC = DHC * DPI180
            DHD = DHD * DPI180
            DHE = DHE * DPI180
            DHH = DHH * DPI180
C
            DCOR = 0.00134D0 * cos(DHA)
     &           + 0.00154D0 * cos(DHB)
     &           + 0.00200D0 * cos(DHC)
     &           + 0.00179D0 * sin(DHD)
     &           + 0.00178D0 * sin(DHE)
C
            DH = DH + DCOR * DPI180
            DLONG = DRADRS(DH)
C
            DH = 1.0000002D0 * (1.D0 - DE * cos(DEM))
C
            DCOR = 0.00000543D0 * sin(DHA)
     &           + 0.00001575D0 * sin(DHB)
     &           + 0.00001627D0 * sin(DHC)
     &           + 0.00003076D0 * cos(DHD)
     &           + 0.00000927D0 * sin(DHH)
C
            DDIST = DH + DCOR
C
          endif
C
        endif
      endif
C
      RETURN
      END

      subroutine CSUN02(DJD,DJD0,DX,DY,DZ,IERR)
C
CC    Calculates the rect. coordinates of the sun for a given equinox
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:56
C
CA  author : KOD        date: 26-SEP-1989 18:25
CU  update : KOD        date: 30-MAR-1990 14:46
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the rectangular coordinates of the sun
CG  for a given Julian Date with respect to an arbitrary equinox,
CG  according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.85-87 .
CG
CGT Note: The equinox of 1950.0 corresponds to JD 2433282.423,
CGT       the equinox of 1900.0 corresponds to JD 2415020.313,
CGT       according to J.Meeus, p.65 .
C
C   call_var.          type I/O description
CP  DJD                 R8  I   Julian Date for the sun position
CP  DJD0                R8  I   Julian Date defining the equinox
CP
CP  DX                  R8    O x-coordinate (to vernal equinox)
CP  DY                  R8    O y-coordinate (to ecl.longitude 90 deg)
CP  DZ                  R8    O z-coordinate (to north pole)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
C
C   routines_called    type     description
CR  CVEQUI              SR      calculates quantities for epoch conv.
CR  CSUN01              SR      calculates long. and dist. of the sun
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
CR  JDEPS1              SR      calc. the obliquity of the ecliptic
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
C
      REAL*8        DJD, DJD0, DX, DY, DZ, DZETA, DZT, DTHETA, DLONG,
     &              DDIST, DEPS, DCU, DSU, DCZ, DSZ, DCT, DST, DXX,
     &              DXY, DXZ, DYX, DYY, DYZ, DZX, DZY, DZZ, DX0, DY0,
     &              DZ0, DJ, DJ0
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'CSUN02'/
C
      IERR = 0
C
C     check input values:
C
      if (min(DJD0,DJD).lt.0.D0 .or. max(DJD0,DJD).gt.3.D6) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
C     calculate parameters for converting the equinox:
C
      if (IERR.lt.2) then
        DJ = DJD
        DJ0 = DJD0
        call CVEQUI(DJ,DJ0,DZETA,DZT,DTHETA,ISTAT)
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
      endif
C
C     calculate position of the sun for the initial equinox:
C
      if (IERR.lt.2) then
        LMEAN = .true.
        call CSUN01(DJ,LMEAN,DLONG,DDIST,ISTAT)
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        endif
      endif
C
C     calculate obliquity of the ecliptic for the initial equinox:
C
      if (IERR.lt.2) then
        LMEAN = .true.
        call JDEPS1(DJ,LMEAN,DEPS,ISTAT)
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,30)
        endif
      endif
C
      if (IERR.lt.2) then
C
C       calculate rectangular positions for the initial equinox:
C
        DX0 = DDIST * cos(DLONG)
        DY0 = DDIST * sin(DLONG) * cos(DEPS)
        DZ0 = DDIST * sin(DLONG) * sin(DEPS)
C
C       calculate parameters for converting the equinox:
C
        DCU = cos(DZETA)
        DSU = sin(DZETA)
C
        DCZ = cos(DZT)
        DSZ = sin(DZT)
C
        DCT = cos(DTHETA)
        DST = sin(DTHETA)
C
        DXX =  DCU * DCZ * DCT - DSU * DSZ
        DXY =  DSU * DCZ + DCU * DSZ * DCT
        DXZ =  DCU * DST
C
        DYX = -DCU * DSZ - DSU * DCZ * DCT
        DYY =  DCU * DCZ - DSU * DSZ * DCT
        DYZ = -DSU * DST
C
        DZX = -DCZ * DST
        DZY = -DSZ * DST
        DZZ =  DCT
C
C       convert rectangular coordinates to the final equinox:
C
        DX = DXX * DX0 + DYX * DY0 + DZX * DZ0
        DY = DXY * DX0 + DYY * DY0 + DZY * DZ0
        DZ = DXZ * DX0 + DYZ * DY0 + DZZ * DZ0
C
      endif
C
      RETURN
      END

      subroutine CSUN03(DJD,LMEAN,DVX,DVY,IERR)
C
CC    Calculates the velocity of the Earth in rect.ecl.coordinates
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:56
C
CA  author : KOD        date: 22-OCT-1991 11:00:10
CU  update : KOD        date: 25-OCT-1991 15:29
C
CT  status: not tested
C
C   general description :
CG
CG  Subroutine to calculate the velocity of the Earth around the Sun
CG  in rectangular ecliptical coordinates for a given Julian Date.
CG  For an easy application to aberration corrections the velocity is
CG  expressed in fractions of the speed of light.
CG
CG  The velocity of the Earth is calculated with respect to the CENTER
CG  of the Sun. This value may differ from the velocity with respect
CG  to the solar system barycenter; the maximum separation between the
CG  center of the sun and the barycenter is about 2.2 solar radii.
CG  The velocity is calculated by the following (approximate) procedure:
CG  first the rectangular ecliptical coordinates of the sun are
CG  determined for two time instants separated by an appropriate
CG  time delay (e.g. one hour) and centered onto the Julian Date
CG  supplied. Then the (negative) difference vector is calculated,
CG  divided by the time delay and normalized with respect to the speed
CG  of light.
CG
CG  Note that the rectangular velocity components refer to the equinox
CG  of date. If they are needed for a standard equinox to a high
CG  precision, they should be precessed appropriately.
CG
CGT Note: The equinox of 2000.0 corresponds to JD 2451545.000,
CGT       the equinox of 1950.0 corresponds to JD 2433282.423,
CGT       the equinox of 1900.0 corresponds to JD 2415020.313.
C
C
C   call_var.          type I/O description
CP
CP  DJD                 R8  I   Julian Date for the Earth velocity
CP
CP  LMEAN               L4  I   true or app. longitude:
CP                                LMEAN = .true.  : true longitude
CP                                LMEAN = .false. : app. longitude
CP
CP  DVX                 R8    O x-velocity (to vernal equinox)
CP  DVY                 R8    O y-velocity (to ecliptical long. 90 deg)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
C
C   routines_called    type     description
CR  CSUN01              SR      calculates long. and dist. of the sun
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
C
      REAL*8        DJD, DJD1, DJD2, DT, DLONG1, DLONG2, DDIST1, DDIST2,
     &              DX1, DY1, DX2, DY2, DVX, DVY, DC
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN
      CHARACTER*8   RTNAME
C
      PARAMETER   ( DT = 0.04D0 )
      PARAMETER   ( DC = 5.775518304D-3 )
C
      DATA          RTNAME /'CSUN03'/
C
      IERR = 0
C
C     check input values:
C
      if (DJD.lt.0.D0 .or. DJD.gt.3.D6) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DJD')
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate 'first' position of the Sun:
C
        DJD1 = DJD - 0.5D0 * DT
C
        call CSUN01(DJD1,LMEAN,DLONG1,DDIST1,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate 'second' position of the Sun:
C
        DJD2 = DJD + 0.5D0 * DT
C
        call CSUN01(DJD2,LMEAN,DLONG2,DDIST2,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate rectangular positions of the Sun:
C
        DX1 = DDIST1 * cos(DLONG1)
        DY1 = DDIST1 * sin(DLONG1)
C
        DX2 = DDIST2 * cos(DLONG2)
        DY2 = DDIST2 * sin(DLONG2)
C
C       determine the velocity of the Earth:
C
        DVX = DC * ( DX1 - DX2 ) / DT
        DVY = DC * ( DY1 - DY2 ) / DT
C
      endif
C
C
      RETURN
      END
 
      subroutine DISTP1(DDIST,DPRLX,IERR)
C
CC    Calculates the parallax for a given distance
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:57
C
CA  author : KOD        date: 04-OCT-1989 19:49
CU  update : KOD        date: 20-MAR-1990 14:47
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the parallax for a given distance,
CG  according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.143 .
C
C   call_var.          type I/O description
CP  DDIST               R8  I   distance of the object  (Astr.Units)
CP
CP  DPRLX               R8    O corresponding parallax  (radians)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
C
C   routines_called    type     description
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
C
      REAL*8        DDIST, DPRLX, DF
c      REAL*4        HFLAG
      INTEGER*4     IERR !,IOS
      CHARACTER*8   RTNAME
C
      parameter ( DF = 4.2634515D-5 )
C
      DATA          RTNAME /'DISTP1'/
C
      IERR = 0
C
      if (DDIST.le.0.D0) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in the input parameter DDIST')
      else
        DPRLX = asin ( DF / DDIST )
      endif
C
      RETURN
      END

      subroutine ECEQAB(DVX,DVY,DVZ,LDIR,DL,DB,DEPS,DA,DD,IERR)
C
CC  converts ecl. coord. into aberration-corrected equ. coord.
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:55
C
CA  author : KOD        date: 22-OCT-1991 10:47
CU  update : KOD        date: 30-OCT-1991 08:11
C
CT  status: not tested
C
C   general description
CG
CG  Subroutine to convert ecliptical coordinates into equatorial
CG  coordinates which are corrected for the effect of aberration.
CG  This correction can be applied in both directions (see LDIR).
CG  The velocities have to be supplied in units of the speed of light
CG  in a right-handed rectangular coordinates system with the x-axis
CG  pointing towards the vernal equinox and the y-axis towards an
CG  ecliptical longitude of 90 degrees. The correction for aberration
CG  performed is non-relativistic.
CG
CG  This routine has been optimized for a frequent conversion/correction
CG  of different positions within a small interval of time. Therefore,
CG  the velocity (DVX,DVY,DVZ) and the obliquity of the ecliptic have
CG  to be provided as arguments. Since these quantities do not change
CG  very quickly (the obliquity being a constant for most applications),
CG  they can be calculated before or may have to be updated only after
CG  a number of calls to this subroutine.
CG
C
C   call_var.          type I/O description
CP
CP  DVX                 R8  I   velocity of the observer in x-direction
CP  DVY                 R8  I   velocity of the observer in y-direction
CP  DVZ                 R8  I   velocity of the observer in z-direction
CP
CP  LDIR                L   I   direction of the correction:
CP                               = .true. :  true pos. --> apparent pos.
CP                               = .false.:  true pos. <-- apparent pos.
CP
CP  DL                  R8  I   ecliptical longitude          (radians)
CP  DB                  R8  I   ecliptical latitude           (radians)
CP
CP  DEPS                R8  I   obliquity of the ecliptic     (radians)
CP
CP  DA                  R8    O Right Ascension               (radians)
CP  DD                  R8    O Declination                   (radians)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DVX, DVY, DVZ, DL, DB, DEPS, DA, DD, DH,
     &              DX1, DY1, DZ1, DX2, DY2, DZ2, DX3, DY3, DZ3,
     &              DSE, DCE, DR, DPH, DTH
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT!, IOS
      LOGICAL       LDIR
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'ECEQAB'/
C
C
      IERR = 0
C
      DH = dmax1(dabs(DVX),dabs(DVY),dabs(DVZ))
C
      if (DH.gt.1.D0) then
         IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in velocity')
      endif
C
      if (DH.gt.0.1D0) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,
     &      'Warning: correction for aberration is non-relativistic')
      endif
C
      if (DL.lt.0.D0 .or. DL.gt.D2PI) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Warning: unusual value for DL')
      endif
C
      if (dabs(DB).gt.DPID2) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameter DB')
      endif
C
      if (dabs(DEPS-0.409D0).gt.1.D-2) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input parameter DEPS')
      endif
C
C
      if (IERR.lt.2) then
C
C       convert the ecliptical coordinates into rectangular coordinates:
C
        DR  = 1.D0
        DPH = DL
        DTH = DPID2 - DB
C
        call POLRE1(DR,DPH,DTH,DX1,DY1,DZ1,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       correct coordinates for aberration:
C
        call CPSAB2(DVX,DVY,DVZ,LDIR,DX1,DY1,DZ1,
     &              DX2,DY2,DZ2,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       convert (rect.) ecl.coord. into (rect.) equatorial coord.:
C
        DSE = dsin(DEPS)
        DCE = dcos(DEPS)
C
        DX3 = DX2
        DY3 = DY2 * DCE - DZ2 * DSE
        DZ3 = DY2 * DSE + DZ2 * DCE
C
C       convert rect. coordinates into spherical coord.:
C
        call REPOL1(DX3,DY3,DZ3,DR,DPH,DTH)
C
        DA = DPH
        DD = DPID2 - DTH
C
      endif
C
C
      RETURN
      END

      subroutine ELEVA1(DLAT,DDELTA,DHGHT,LREF,DHR,DAZ,IERR)
C
CC    Calculates the hour angle corresponding to a given elevation
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:59
C
CA  author : KOD        date: 29-SEP-1989 16:36
CU  update : KOD        date: 30-MAR-1990 15:00
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the hour angle (and azimuth) for
CG  a celestial object when its apparent/true elevation DECREASES
CG  below a certain value (equivalent to an hour angle between
CG  0 and +12 hours). The apparent elevation takes the effect
CG  of atmospheric refraction into account.
CG
CG  If the elevation cannot be reached, the hour angle is set
CG  according to the best approximation (0 or pi) and a warning
CG  message will be given ( IERR = 1 ) .
CG
CG  The azimuth is defined to start from south and increase
CG  towards west.
CG
CG  To determine the hour angle for apparent rising (or setting),
CG  take the following values:
CG
CGT    apparent height    true height
CGT    ---------------   -------------
CGT       0' =  0.000    -34' = -0.010    for stars and planets,
CGT     -13' = -0.004    -50' = -0.015    for the sun, and
CGT     +36' =  0.010    + 8' = +0.002    for the moon.
CG
CG  According to Jean Meeus, 'Astronomical Formulae for Calculators',
CG                           third edition, 1985, p.44, 207, 208 .
C
C   call_var.          type I/O description
CP  DLAT                R8  I   geographical latitude     (radians)
CP
CP  DDELTA              R8  I   declination               (radians)
CP
CP  DHGHT               R8  I   height above the horizon  (radians)
CP
CP  LREF                L4  I   correction for atmospheric refraction:
CP                                LREF = .true.  : with correction
CP                                LREF = .false. : without correction
CP
CP  DHR                 R8    O hour angle     ( 0 .. pi , radians)
CP  DAZ                 R8    O azimuth        ( 0 .. pi , radians)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  REFRA1              SR      corrects for atmospheric refraction
CR  REPOL1              SR      converts rectangular into polar coord.
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DHGHT, DDELTA, DLAT, DHR, DAZ, DH1, DH2, DH,
     &              DR, DTHETA, DHT
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LDIR, LREF
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'ELEVA1'/
C
      IERR = 0
C
      if (max(abs(DLAT),abs(DDELTA),abs(DHGHT)).gt.DPID2) then
C
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in input values')
C
      else
C
C       correct for atmospheric refraction
C
        if (.not.LREF) then
          DHT = DHGHT
        else
          LDIR=.true.
          DH = DHGHT
!HB          call REFRA1(DH,LDIR,DHT,ISTAT)
          istat = 0 ! 20210903 IMS
          if  (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,10)
          endif
        endif
C
      endif
C
      if (IERR.lt.2) then
C
        DH1 = sin(DHT) - sin(DLAT) * sin(DDELTA)
        DH2 = cos(DLAT) * cos(DDELTA)
C
        if (DH2.eq.0.D0) then
C
          DHR = 0.D0
          DAZ = 0.D0
C
          if ( .not. ((cos(DLAT)  .eq.0.D0 .and. DDELTA.eq.DHT) .or.
     &                (cos(DDELTA).eq.0.D0 .and. DLAT  .eq.DHT) )) then
            IERR = max(IERR,1)
            call WRERR(RTNAME,IERR,'Elevation cannot be reached')
          endif
C
        else
C
          DH = DH1 / DH2
C
          if (dabs(DH).gt.1.D0) then
            IERR = max(IERR,1)
            call WRERR(RTNAME,IERR,'Elevation cannot be reached')
            DH = sign(1.D0,DH)
          endif
C
          DHR = acos(DH)
C
C         calculate the azimuth:
          DH1 = DH * sin(DLAT) - tan(DDELTA) * cos(DLAT)
          DH2 = sin(DHR)
          DH = 0.D0
          call REPOL1(DH1,DH2,DH,DR,DAZ,DTHETA)
C
        endif
C
      endif
C
      RETURN
      END

      subroutine ELEVA2(DJD,DLONG,DLAT,DALPHA,DDELTA,DHGHT,LREF,
     &                  DJD1,DAZ1,DJD2,DAZ2,IERR)
C
CC    Calculates the times corresponding to a given elevation
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:59
C
CA  author : KOD        date: 29-SEP-1989 16:56
CU  update : KOD        date: 30-MAR-1990 15:05
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the times (and corresponding azimuths)
CG  during which the apparent/true elevation of a celestial object
CG  EXCEEDS a given value. The apparent elevation takes the effect
CG  of atmospheric refraction into account.
CG
CG  If the elevation cannot be reached the times will be calculated
CG  for the corresponding transit and a warning message will be
CG  given ( IERR = 1 ) .  The azimuth is defined to start from
CG  south and increase towards west.
CG
CG  According to a resolution of the IAU east longitudes are counted
CG  positive, west longitudes negative
CG                                  (Sky and Tel, June, 1984, p.559).
CG
CG  To determine the hour angle for apparent rising (or setting),
CG  take the following values:
CG
CGT   apparent height    true height
CGT   ---------------   -------------
CGT      0' =  0.000    -34' = -0.010   for stars and planets,
CGT    -13' = -0.004    -50' = -0.015   for the sun, and
CGT    -13' = -0.004    -50' = -0.015   for the moon (with par.cor.)
CGT    +36' =  0.010    + 8' = +0.002   for the moon ( no  par.cor.)
CG
CG  The times are given in Julian Ephemeris Times (approx. UT) and
CG  are selected to be as close to the given date as possible.
cg  +++  problem: djd < djd1 < djd2 possible  +++
CG
CG  According to Jean Meeus, 'Astronomical Formulae for Calculators',
CG                           third edition, 1985, p.44, 207, 208 .
C
C   call_var.          type I/O description
CP  DJD                 R8  I   Julian Day to be analysed
CP
CP  DLONG               R8  I   geographical longitude    (radians)
CP  DLAT                R8  I   geographical latitude     (radians)
CP
CP  DALPHA              R8  I   right ascension           (radians)
CP  DDELTA              R8  I   declination               (radians)
CP
CP  DHGHT               R8  I   height above the horizon  (radians)
CP
CP  LREF                L4  I   correction for atmospheric refraction:
CP                                LREF = .true.  : with correction
CP                                LREF = .false. : without correction
CP
CP  DJD1                R8    O J.D. for ascending        ( < DJD2 )
CP  DAZ1                R8    O azimuth for DJD1          (radians)
CP
CP  DJD2                R8    O J.D. for descending       ( > DJD1 )
CP  DAZ2                R8    O azimuth for DJD2          (radians)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  ELEVA1              SR      calculates the hour angle for elevation
CR  LSTJD1              SR      calculates the time from sid. time
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DJD, DLONG, DLAT, DALPHA, DDELTA, DHGHT, DJD1,
     &              DAZ1, DJD2, DAZ2, DHR, DAZ, DLSID, DJD0, DT,
     &              DJ, DLG, DLT, DD, DHG
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN, LREF, LRF
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'ELEVA2'/
C
      IERR = 0
C
      if (DLONG.lt.-DPI .or. DLONG.gt.D2PI) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DLONG')
      endif
C
      if (DALPHA.lt.0.D0 .or. DALPHA.gt.D2PI) then
        IERR = 2
        call WRERR(RTNAME,IERR,'Error in the input parameter DALPHA')
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate the hour angle:
C
        DLT = DLAT
        DD = DDELTA
        DHG = DHGHT
        LRF = LREF
        call ELEVA1(DLT,DD,DHG,LRF,DHR,DAZ,ISTAT)
C
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate time of nearest transit:
C
        if (dabs(DHR-DPI).lt.1.D-6) then
          DHR = 0.D0
          DLSID = DALPHA + DPI
        else
          DLSID = DALPHA
        endif
C
        LMEAN = .false.
        DJ = DJD
        DLG = DLONG
        call LSTJD1(DJ,DLG,DLSID,LMEAN,DJD0,ISTAT)
C
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        else
C
          DT = DHR / ( 1.002737908D0 * D2PI )
C
          DJD1 = DJD0 - DT
          DAZ1 = DAZ
C
          DJD2 = DJD0 + DT
          DAZ2 = D2PI - DAZ
C
        endif
      endif
C
C
      RETURN
      END

      subroutine ELMON1(DJD,DLONG,DLAT,DHGHT,LREF,
     &                  DJD1,DAZ1,DJD2,DAZ2,IERR)
C
CC    Calculates the times corresp. to a given elevation of the moon
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 16:59
C
CA  author : KOD        date: 04-OCT-1989 18:15
CU  update : KOD        date: 30-MAR-1990 15:55
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the times (and corresponding azimuths)
CG  during which the true or apparent elevation (corrected for
CG  atmospheric refraction) of the moon EXCEEDS a given value. The
CG  apparent movement of the moon and the topocentric parallax will
CG  be taken into account.
CG
CG  If the elevation cannot be reached the times will be calculated
CG  for the corresponding transit and a warning message will be
CG  given ( IERR = 1 ) .  The azimuth is defined to start from
CG  south and increase towards west.
CG
CG  According to a resolution of the IAU east longitudes are counted
CG  positive, west long. negative (Sky and Tel, June, 1984, p.559).
CG
CGT   To determine the times for apparent rising or setting of the moon,
CGT   take an apparent elevation of  -13' = -0.004
CGT          or a true elevation of  -50' = -0.015 .
CG
CG  The times are given in Julian Ephemeris Times (approx. UT) and
CG  are selected to be as close to the given date as possible.
CG
C
C   call_var.          type I/O description
CP  DJD                 R8  I   Julian Day to be analysed
CP
CP  DLONG               R8  I   geographical longitude    (radians)
CP  DLAT                R8  I   geographical latitude     (radians)
CP
CP  DHGHT               R8  I   height above the horizon  (radians)
CP
CP  LREF                L4  I   correction for atmospheric refraction:
CP                                LREF = .true.  : with correction
CP                                LREF = .false. : without correction
CP
CP  DJD1                R8    O J.D. for ascending        ( < DJD2 )
CP  DAZ1                R8    O azimuth for DJD1          (radians)
CP
CP  DJD2                R8    O J.D. for descending       ( > DJD1 )
CP  DAZ2                R8    O azimuth for DJD2          (radians)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
C
C   routines_called    type     description
CR  CMOON1              SR      calculates the position of the moon
CR  EQUECL              SR      converts equat. and ecl. coordinates
CR  ELEVA1              SR      determines the hour angle to an elev.
CR  ELEVA2              SR      determines the times to an elevation
CR  PARLX1              SR      corrects for topocentric parallax
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
C
      REAL*8        DJD, DLONG, DLAT, DHGHT, DJD1, DAZ1, DJD2, DAZ2,
     &              DALPHA, DDELTA, DJD0, DA1, DA2, DEPS,
     &              DLMON, DBMON, DJ1, DJ2, DPRLX, DHR, DAZ, DALPH0,
     &              DDELT0, DJ, DLG, DLT, DHG
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT, IT, ITMAX! ,IOS
      LOGICAL*4     LMEAN, LDIR, LCONV, LSTOP, LREF, LPAR, LRF
      CHARACTER*8   RTNAME
C
      PARAMETER ( ITMAX = 10 )
      PARAMETER ( DEPS = 1.D-12 )
C
      DATA          RTNAME /'ELMON1'/
C
      IERR = 0
C
C     calculate the apparent position of the moon for DJD:
C
      LMEAN = .false.
      DJ = DJD
      call CMOON1(DJ,DLMON,DBMON,DPRLX,ISTAT)
C
      if  (ISTAT.gt.1) then
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
      endif
C
      if (IERR.lt.2) then
C
        LDIR = .false.
        call EQUECL(DJ,LMEAN,LDIR,DLMON,DBMON,DALPHA,DDELTA,ISTAT)
C
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate the hour angle:
C
        DLT = DLAT
        DHG = DHGHT
        LRF = LREF
        call ELEVA1(DLT,DDELTA,DHG,LRF,DHR,DAZ,ISTAT)
C
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,30)
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate the approx. transition times for the elevation:
C
        DLG = DLONG
        call ELEVA2(DJ,DLG,DLT,DALPHA,DDELTA,DHG,LRF,
     &              DJ1,DA1,DJ2,DA2,ISTAT)
        DJD1 = DJ1
        DJD2 = DJ2
C
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,40)
        endif
C
C
C       improve the ascending position of the moon:
C
        LCONV = .false.
        LSTOP = .false.
        LPAR = .true.
        IT = 0
        do while (.not.LCONV .and. IT.lt.ITMAX .and. .not.LSTOP)
          IT = IT + 1
          DJD0 = DJD1
          call CMOON1(DJD0,DLMON,DBMON,DPRLX,ISTAT)
          LSTOP = (ISTAT.gt.1)
          call EQUECL(DJD0,LMEAN,LDIR,DLMON,DBMON,DALPH0,DDELT0,ISTAT)
          LSTOP = (ISTAT.gt.1)
          call PARLX1(DLT,DHG,DPRLX,-DHR,DALPH0,DDELT0,LPAR,
     &                  DALPHA,DDELTA,ISTAT)
          LSTOP = (ISTAT.gt.1)
          call ELEVA2(DJD0,DLG,DLT,DALPHA,DDELTA,DHG,LRF,
     &                DJ1,DA1,DJ2,DA2,ISTAT)
          DJD1 = DJ1
          DAZ1 = DA1
          LSTOP = (ISTAT.gt.1)
          LCONV = (abs(DJD1-DJD0).lt.DEPS)
        end do
C
        if (LSTOP) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,50)
        endif
C
C
C       improve the descending position of the moon:
C
        LCONV = .false.
        LSTOP = .false.
        LPAR = .true.
        IT = 0
        do while (.not.LCONV .and. IT.lt.ITMAX .and. .not.LSTOP)
          IT = IT + 1
          DJD0 = DJD2
          call CMOON1(DJD0,DLMON,DBMON,DPRLX,ISTAT)
          LSTOP = (ISTAT.gt.1)
          call EQUECL(DJD0,LMEAN,LDIR,DLMON,DBMON,DALPH0,DDELT0,ISTAT)
          LSTOP = (ISTAT.gt.1)
          call PARLX1(DLT,DHG,DPRLX,DHR,DALPH0,DDELT0,LPAR,
     &                  DALPHA,DDELTA,ISTAT)
          LSTOP = (ISTAT.gt.1)
          call ELEVA2(DJD0,DLG,DLT,DALPHA,DDELTA,DHG,LRF,
     &                DJ1,DA1,DJ2,DA2,ISTAT)
          DJD2 = DJ2
          DAZ2 = DA2
          LSTOP = (ISTAT.gt.1)
          LCONV = (abs(DJD2-DJD0).lt.DEPS)
        end do
C
        if (LSTOP) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,60)
        endif
C
      endif
C
C
      RETURN
      END

      subroutine EQUHOR(DJD,LMEAN,DLONG,DLAT,LDIR,DLOLD,DBOLD,
     &                  DLNEW,DBNEW,IERR)
C
CC    Converts coordinates between the equatorial and horizontal system
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 17:02
C
CA  author : KOD        date: 05-SEP-1989 16:45
CU  update : KOD        date: 29-MAR-1990 15:21
CU  update : KOD        date: 01-FEB-1995           (dradrs)
C
CT  status: not tested
C
C   general description
CG  Subroutine to convert coordinates between the equatorial and the
CG  horizontal system (the azimuthal angle starting from south and
CG  running to west).
CG
CG  For the equatorial coordinates the current equinox should be used.
CG  (APPARENT positions are affected by aberration and nutation.)
CG  The height above the horizon { is assumed to be / will be }
CG  corrected for atmospheric refraction.
CG
CG  According to a resolution of the IAU east longitudes are counted
CG  positive, west long. negative (Sky and Tel, June, 1984, p.559).
C
C   call_var.          type I/O description
CP  DJD                 R8  I   Julian Date
CP
CP  LMEAN               L4  I   mean or apparent positions:
CP                                LMEAN = .true.  :  mean    positions
CP                                LMEAN = .false. : apparent positions
CP
CP  DLONG               R8  I   geographical longitude ( radians ! )
CP  DLAT                R8  I   geographical latitude  ( radians ! )
CP
CP  LDIR                L4  I   transformation direction:
CP                                LDIR = .true.  : equ --> hor
CP                                LDIR = .false. : equ <-- hor
CP
CP  DLOLD               R8  I   old longitude / azimuth    (radians)
CP  DBOLD               R8  I   old latitude / elevation   (radians)
CP
CP  DLNEW               R8    O new longitude / azimuth    (radians)
CP  DBNEW               R8    O new latitude / elevation   (radians)
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  REFRA1              SR      corrects for atmospheric refraction
CR  JDLST1              SR      calculates local siderial time
CR  EULER1              SR      general coordinate transformation
CR  DRADRS              R8      'resets' radians
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DJD, DLONG, DLAT, DLOLD, DBOLD, DLNEW, DBNEW,
!HB  &              DRADRS, DLSID, D1, D2, D3, DPHI1, DTHETA1,
     &              DLSID, D1, D2, D3, DPHI1, DTHETA1,
     &              DPHI2, DTHETA2, DBLD, DBNW, DJ, DLNG, DB
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN, LDIR, LCONV, LMN
      CHARACTER*8   RTNAME
      logical       lfirst
C
      DATA          RTNAME /'EQUHOR'/
      data lfirst / .true. /
C
      IERR = 0
C
      if (DLOLD.lt.0.D0 .or. DLOLD.gt.D2PI .or. dabs(DBOLD).gt.DPID2)
     &  then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'Error in input parameters DLOLD, DBOLD')
      endif
C
C     calculate the local sidereal time:
C
      DJ = DJD
      LMN = LMEAN
      DLNG = DLONG
      call JDLST1(DJ,LMN,DLNG,DLSID,ISTAT)
C
      if (ISTAT.gt.1) then
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
      endif
C
      if (IERR.lt.2) then
C
C       correct for atmospheric refraction:
C
        if (lfirst) then
          write(*,*) '+++ equhor: atmospheric refraction changed +++'
c          write(*,*) '+++ equhor: atm. refraction not changed +++'
          lfirst = .false.
        endif
        if (LDIR .or. DBOLD.lt.0.d0) then
          DBLD = DBOLD
        else
          LCONV = .true.
          DB = DBOLD
****          call REFRA1(DB,LCONV,DBLD,ISTAT)
          dbld = db  !@@@
C
          if (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,20)
          endif
C
        endif
      endif
C
      if (IERR.lt.2) then
C
        D1 =  dradrs(DLSID + DPID2)
        D2 =  DPID2 - DLAT
        D3 = -DPID2
C
        if (ldir) then
          DPHI1 = DLOLD
        else
          dphi1 = d2pi - dlold
        endif
C
        DTHETA1 = DPID2 - DBLD
C
        if (LDIR) then
           call EULER1(D1,D2,D3,DPHI1,DTHETA1,DPHI2,DTHETA2,ISTAT)
        else
           call EULER1(-D3,-D2,-D1,DPHI1,DTHETA1,DPHI2,DTHETA2,ISTAT)
        endif
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,30)
        endif
      endif
C
      if (IERR.lt.2) then
C
        if (ldir) then
          DLNEW = D2PI - DPHI2
        else
          dlnew = dphi2
        endif
C
        DBNW  = DPID2 - DTHETA2
C
C       correct for atmospheric refraction:
C
        if (.not.LDIR .or.dbnw.lt.0.d0) then
          DBNEW = DBNW
        else
          LCONV = .false.
****          call REFRA1(DBNW,LCONV,DBNEW,ISTAT)
          dbnew = dbnw   !@@@@
C
          if (ISTAT.gt.1) then
            IERR = max(IERR,ISTAT)
            call ERRSTC(RTNAME,40)
          endif
C
        endif
      endif
C
      RETURN
      END

      subroutine KMOON1(DJD,DPHANG,DFRC,IERR)
C
CC    Calc. the phase angle and the illum. fract. of the moon's disk
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 17:07
C
CA  author : KOD        date: 05-OCT-1989 14:18
CU  update : KOD        date: 29-MAR-1990 16:02
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the (geocentric) phase angle and the
CG  illuminated fraction of the moon for a given date, according to
CG
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                  third edition, 1985, p.155-157 .
CG
CG  The phase angle is defined as the angular distance sun-earth
CG  as seen from the moon.
C
C   call_var.          type I/O description
CP  DJD                 R8  I   Julian Date  (Ephemeris time)
CP
CP  DPHANG              R8    O phase angle  (radians)
CP
CP  DFRC                R8    O illuminated fraction of the moon's disk
CP
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  CSUN01              SR      calculates the position of the sun
CR  CMOON1              SR      calculates the position of the moon
CR  DRADRS              R8      'resets' radians
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DJD, DPHANG, DFRC, DLSUN, DDIST, DLMON, DBMON,
!HB  &              DPRLX, DJD0, DYEAR, DT, DRADRS, DMS, DMM,
     &              DPRLX, DJD0, DYEAR, DT, DMS, DMM,
     &              DDMS(4), DDMM(4), DH, DH1, DH2, DJ
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT! ,IOS
      LOGICAL*4     LMEAN
      CHARACTER*8   RTNAME
C
      PARAMETER ( DJD0 = 2415020.D0 )
      PARAMETER ( DYEAR =  36525.D0 )
C
      DATA          RTNAME /'KMOON1'/
C
      DATA DDMS / 358.475833D0,  35999.0498D0,-0.000150D0,-0.0000033D0 /
      DATA DDMM / 296.104608D0, 477198.8491D0, 0.009192D0, 0.0000144D0 /
C
      IERR = 0
C
C     calculate the ecliptical position of the sun:
C
      LMEAN = .true.
      DJ = DJD
      call CSUN01(DJ,LMEAN,DLSUN,DDIST,ISTAT)
C
      if  (ISTAT.gt.1) then
        IERR = max(IERR,ISTAT)
        call ERRSTC(RTNAME,10)
      endif
C
      if (IERR.lt.2) then
C
C       calculate the ecliptical position of the moon:
C
        call CMOON1(DJ,DLMON,DBMON,DPRLX,ISTAT)
C
        if  (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        endif
C
      endif
C
      if (IERR.lt.2) then
C
C       calculate the mean anomalies of the sun and moon:
C
        DT = (DJD - DJD0) / DYEAR
C
        DMS = DDMS(1) + (DDMS(2) + (DDMS(3) + DDMS(4) * DT) * DT) * DT
        DMM = DDMM(1) + (DDMM(2) + (DDMM(3) + DDMM(4) * DT) * DT) * DT
C
        DMS = DMS * DPI180
        DMM = DMM * DPI180
C
        DMS = DRADRS(DMS)
        DMM = DRADRS(DMM)
C
C       calculate the phase angle:
C
        DH = acos( cos(DLMON-DLSUN) * cos(DBMON) )
C
        DH1 = 1.D0 - 0.0549D0 * sin(DMM)
        DH2 = 1.D0 - 0.0167D0 * sin(DMS)
C
        DH = 180.D0 - DH * D180PI - 0.1468D0 * sin(DH) * DH1 / DH2
C
        DH = DH * DPI180
        DPHANG = DRADRS(DH)
C
C       calculate the illuminated fraction:
C
        DFRC = 0.5D0 * ( 1.D0 + cos(DPHANG) )
C
      endif
C
      RETURN
      END

      subroutine PARLX1(DLAT,DHT,DPRLX,DHR,DAOLD,DDOLD,LDIR,
     &                  DANEW,DDNEW,IERR)
C
CC    Corrects the equatorial coordinates for parallax
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 17:08
C
CA  author : KOD        date: 04-OCT-1989 18:35
CU  update : KOD        date: 29-MAR-1990 16:24
C
CT  status: tested
C
C   general description
CG  Subroutine to convert between GEOcentric and TOPOcentric
CG  equatorial coordinates (of the Moon, Sun, a planet or comet),
CG  according to
CG     Jean Meeus, 'Astronomical Formulae for Calculators',
CG                 third edition, 1985, p.143-146 .
CG
CG  The conversion from topocentric to geocentric coordinates is
CG  less accurrate; this difference might however only be noticed
CG  for the moon where the error is usually around 20 arc seconds.
C
C   call_var.          type I/O description
CP  DLAT                R8  I   geographical latitude  (radians)
CP  DHT                 R8  I   height above sea level  (meters)
CP  DPRLX               R8  I   parallax               (radians)
CP  DHR                 R8  I   hour angle             (radians)
CP  DAOLD               R8  I   'old' right ascension  (radians)
CP  DDOLD               R8  I   'old' declination      (radians)
CP
CP  LDIR                L4  I   transformation direction:
CP                                LDIR = .true.  : geo. --> topo.
CP                                LDIR = .false. : geo. <-- topo.
CP
CP  DANEW               R8    O 'new' right ascension  (radians)
CP  DDNEW               R8    O 'new' declination      (radians)
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  GEORCO              SR      calculates geocentric rect. coordinates
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DLAT, DHT, DPRLX, DHR, DAOLD, DDOLD, DANEW, DDNEW,
     &              DRSPHI, DRCPHI, DH1, DH2, DACOR, DDCOR, DSPRLX,
     &              DH, DLT
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT !  ,IOS
      LOGICAL*4     LDIR
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'PARLX1'/
C
      IERR = 0
C     check input values
C
      if (DPRLX.lt.0.D0 .or. DPRLX.gt.0.02D0 .or. dabs(DHR).gt.DPI .or.
     &    DAOLD.lt.0.D0 .or. DAOLD.gt.D2PI .or. dabs(DDOLD).gt.DPI) then
C
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error in the input values')
      endif
C
      if (IERR.lt.2) then
C
C       calculate the geocentric rectangular coordinates
C
        DLT = DLAT
        DH = DHT
        call GEORCO(DLT,DH,DRSPHI,DRCPHI,ISTAT)
C
        if  (ISTAT.gt.1) then
C
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
C
        else
C
          DSPRLX = sin(DPRLX)
C
          DH1 = -DRCPHI * DSPRLX * sin(DHR)
          DH2 = cos(DDOLD) - DRCPHI * DSPRLX * cos(DHR)
          DACOR = atan2 ( DH1, DH2 )
C
          DH1 = ( sin(DDOLD) - DRSPHI * DSPRLX ) * cos(DACOR)
          DH2 = cos(DDOLD) - DRCPHI * DSPRLX * cos(DHR)
          DH = atan2 ( DH1, DH2 )
          DDCOR = DH - DDOLD
C
          if (LDIR) then
            DANEW = DAOLD + DACOR
            DDNEW = DDOLD + DDCOR
          else
            DANEW = DAOLD - DACOR
            DDNEW = DDOLD - DDCOR
          endif
C
        endif
      endif
C
      RETURN
      END

      subroutine PHASE1(DT0,DPER,DT,DPH,NPER,IERR)
C
CC    Calculates the phase and number of extrapolated periods
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 17:08
C
CA  author : KOD        date: 13-OCT-1989 14:48
CU  update : KOD        date: 20-MAR-1990 15:07
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the phase and number of extrapolated
CG  (full) periods.
C
C   call_var.          type I/O description
CP  DT0                 R8  I   reference epoch
CP  DPER                R8  I   period             (same unit as DT0)
CP
CP  DT                  R8  I   time of interest   (same unit as DT0)
CP
CP  DPH                 R8    O corresponding phase  ( 0 .. 1 )
CP  NPER                I4    O number of extrapolated full periods
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
C
C   routines_called    type     description
CR  WRERR               SR      formats error code in standard form
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
C
      REAL*8        DT0, DPER, DT, DPH, DH
c      REAL*4        HFLAG
      INTEGER*4     IERR, NPER! ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'PHASE1'/
C
      IERR = 0
C
      if (DPER.le.0.D0) then
        IERR = max(IERR,2)
        call WRERR(RTNAME,IERR,'Error: negative period')
      endif
C
      if (IERR.lt.2) then
C
        DH = (DT - DT0) / DPER
        NPER = int(DH)
        DH = DH - dble(NPER)
        if (DH.lt.0.D0) DH = DH + 1.D0
        DPH = DH
C
      endif
C
      RETURN
      END

      subroutine PHASE2(DT0,DPER,DT1,DPH1,DT,NPER,IERR)
C
CC    Calculates the time for a certain phase nearest to a given time
C
C************************ FFORM VERSION 1.2 ************ 30-MAR-90 17:08
C
CA  author : KOD        date: 13-OCT-1989 14:48
CU  update : KOD        date: 29-MAR-1990 16:33
C
CT  status: tested
C
C   general description
CG  Subroutine to calculate the time nearest to a given time
CG  for which a certain phase occurs.
C
C   call_var.          type I/O description
CP  DT0                 R8  I   reference epoch
CP  DPER                R8  I   period             (same unit as DT0)
CP
CP  DT1                 R8  I   time of interest   (same unit as DT0)
CP  DPH1                R8  I   phase of interest    ( 0 .. 1)
CP
CP  DT                  R8    O corresponding time (same unit as DT0)
CP  NPER                I4    O number of extrapolated full periods
CP  IERR                I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
C
C   routines_called    type     description
CR  PHASE1              SR      calculates the phase and number of per.
CR  RESVAL              SR      'resets' value
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
CR  HFLAG               R4      output flag handling routine
CR  WRFLAG              SR      writes to FLAG output stream
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
C
      REAL*8        DT0, DPER, DT1, DPH1, DT, DPH, DPHMIN, DPHMAX, DH,
     &              D0, D1, DP
c      REAL*4        HFLAG
      INTEGER*4     IERR, NPER, N, ISTAT ! ,IOS
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'PHASE2'/
C
      IERR = 0
C
      if (DPH1.lt.0.D0 .or. DPH1.gt.1.D0) then
        IERR = max(IERR,1)
        call WRERR(RTNAME,IERR,'DPH1 not between 0 and 1')
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate phase for DT1:
C
        D0 = DT0
        DP = DPER
        D1 = DT1
        call PHASE1(D0,DP,D1,DPH,N,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,10)
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate minimum difference in phase:
C
        DH = DPH1 - DPH
        DPHMIN = 0.D0
        DPHMAX = 1.D0
        call RESVAL(DPHMIN,DPHMAX,DH,DPH,ISTAT)
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,20)
        else
          if (DPH.ge.0.5D0) DPH = DPH - 1.D0
          DT = DT1 + DPH * DPER
        endif
C
      endif
C
C
      if (IERR.lt.2) then
C
C       calculate number of extrapolated full periods:
C
        call PHASE1(D0,DP,DT,DPH,N,ISTAT)
        NPER = N
C
        if (ISTAT.gt.1) then
          IERR = max(IERR,ISTAT)
          call ERRSTC(RTNAME,30)
        endif
C
      endif
C
C
      RETURN
      END

      subroutine TREQU1(DL0,DB0,DL,DB,DLEQU,DBEQU)
C
CC    Transforms spherical coordinates to the equator
C
C************************ FFORM VERSION 1.1 ************ DD-MMM-YY HH:MM
C
CA  author : KOD        date:  8-OCT-1990 15:41
CU  update : KOD        date: 8-OCT-1990 15:48
CU  update : KOD        date: 8-OCT-1990 15:50
CU  update : KOD        date: 8-OCT-1990 16:13
CU  update : KOD        date: 10-OCT-1990 14:08
C
CT  status: not tested
C
C   general description:
CG
C
C   call_var.          type I/O description
CP
CP  DL0                 R8  I   longitude of reference       (radians)
CP  DB0                 R8  I   latitude  of reference       (radians)
CP
CP  DL                  R8  I   longitude to be transformed  (radians)
CP  DB                  R8  I   latitude  to be transformed  (radians)
CP
CP  DLEQU               R8    O transformed longitude        (radians)
CP  DBEQU               R8    O transformed latitude         (radians)
CP
CP  (ierr)              I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block_name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DL0, DB0, DL, DB, DLEQU, DBEQU,
     &              D1, D2, D3, DPH1, DTH1, DPH2, DTH2
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'TREQU1'/
C
      IERR = 0
C
      D1 =  DPID2 + DL0
      D2 =        - DB0
      D3 = -DPID2
C
      DPH1 = DL
      DTH1 = DPID2 - DB
C
      CALL EULER1(D1,D2,D3,DPH1,DTH1,DPH2,DTH2,ISTAT)
C
      DLEQU = DPH2
      DBEQU = DPID2 - DTH2
C
C
      RETURN
      END

      subroutine TRPOL1(DALPH0,DDELT0,DALPHA,DDELTA,DAPOL,DDPOL)
C
CC    Transforms equatorial coordinates to the polar cap
C
C************************ FFORM VERSION 1.1 ************ DD-MMM-YY HH:MM
C
CA  author : KOD        date:  9-AUG-1990 14:57
C
CT  status: not tested
C
C   general description:
CG
C
C   call_var.          type I/O description
CP
CP  DALPH0              R8  I   Right Ascension of reference  (radians)
CP  DDELT0              R8  I   Declination     of reference  (radians)
CP
CP  DALPHA              R8  I   Right Ascension to be transformed (rad)
CP  DDELTA              R8  I   Declination     to be transformed (rad)
CP
CP  DAPOL               R8    O transformed Right Ascension   (radians)
CP  DDPOL               R8    O transformed Declination       (radians)
CP
CP  (ierr)              I4    O = 0 no error
CP                              = 1 warning
CP                              = 2 error
CP                              = 3 severe error
C
C   include_block_name          description
CI  R$COMMON:CGENL.CMN          general parameter common block
CI  R$INCLUDE:IGENL.INC         parameter block with trig. constants
C
C   routines_called    type     description
CR  WRERR               SR      formats error code in standard form
CR  ERRSTC              SR      continue error stack for one level
C
C   extensions/system calls     description
CX
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c*
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c      CHARACTER*265 OUTPT                                ! R$INCLUDE:ICGENL.INC
      REAL*4        PI, RAD, DEG                         ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( PI  = 3.141592654 )                  ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( RAD = 0.0174532925 )                 ! R$INCLUDE:ICGENL.INC
      PARAMETER   ( DEG = 57.29577951 )                  ! R$INCLUDE:ICGENL.INC
c     ---------------------------------------------------! R$INCLUDE:ICGENL.INC
c*
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
      REAL*8        DPI, D2PI, DPID2, DPI180, D180PI     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI    =  3.141592653589793238D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( D2PI   =  6.283185307179586477D0 )     ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPID2  =  1.5707963267948966192D0 )    ! R$INCLUDE:IGENL.INC
      PARAMETER ( DPI180 =  0.017453292519943295769D0 )  ! R$INCLUDE:IGENL.INC
      PARAMETER ( D180PI =  57.29577951308232088D0 )     ! R$INCLUDE:IGENL.INC
c     ---------------------------------------------------! R$INCLUDE:IGENL.INC
C
      REAL*8        DALPH0, DDELT0, DALPHA, DDELTA, DAPOL, DDPOL,
     &              D1, D2, D3, DPH1, DTH1, DPH2, DTH2
c      REAL*4        HFLAG
      INTEGER*4     IERR, ISTAT
      CHARACTER*8   RTNAME
C
      DATA          RTNAME /'TRPOL1'/
C
      IERR = 0
C
      D1 =  DPID2 + DALPH0
      D2 =  DPID2 - DDELT0
      D3 = -DPID2 - DALPH0
C
      DPH1 = DALPHA
      DTH1 = DPID2 - DDELTA
C
      CALL EULER1(D1,D2,D3,DPH1,DTH1,DPH2,DTH2,ISTAT)
C
      DAPOL= DPH2
      DDPOL= DPID2 - DTH2
C
C
      RETURN
      END

      end module astrolib
