! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
program expmap_main

  ! HB 20100323
  !
  !---------------------------------------------------------------------
  ! Summary:
  !
  ! Creates vignetted and unvignetted exposure maps per instrument
  ! and energy band (only vignetted) in sky coordinates; WCS keywords
  ! and image dimensions are taken from an input template image.
  !
  ! Algorithm:
  !
  ! Optimized for speed for survey but should also work for pointing;
  ! pointing exposure maps could be done much faster by using an attitude
  ! histogram.
  !
  ! The program exploits two characteristic times, a) the frame duration (the
  ! attitude moves by about one sky image pixel within one frame duration and this
  ! also is the accuracy with which arrival times are known), and b) the time
  ! interval in which the MIPS counter is read out (assumed to be 20 (=ndead)
  ! frame times; this is the time interval for which the dead time correction
  ! is known). In addition, it is assumed that the roll angle and scan speed
  ! don't change within ndead frame times.
  !
  ! 1) vignetting will be multiplied onto detector map once at program start
  !    for each energy band (if withvignetting=yes)
  ! 2) loop over GTI intervals in time steps of length ndead x frame time
  ! 3) whenever the roll angle changed by more than 0.1 deg, bin detector maps
  !    in image pixel and rotate by roll angle
  ! 4) smear and add detector maps over ndead frame times; repeat whenever
  !    the slew speed or roll angle changed, or the end of the GTI interval
  !    was reached
  ! 5) multiply with frame time and dead time correction; repeat whenever
  !    (4) was performed or exposure changed by more then 0.1%
  ! 6) for each time bin of ndead frame times, add rotated and
  !    smeared detector maps onto sky image according to attitude
  !---------------------------------------------------------------------

  use expmap_io
  use expmap_mod
  use param
  use iso_fortran_env
#ifdef _OPENMP
  use omp_lib
#endif

  implicit none

  type(t_taskinfo) :: task_info
  integer :: ierr,status
  integer :: ilist,ib,ngti,ci
  real(real64) :: expmap_start, expmap_stop    ! Time runtime
  character(len=80) :: line
  type(imgpar_t) :: imgpar
  type(gti_t), dimension(N_TELESCOPES) :: ogtis
  real, allocatable, dimension(:,:,:) :: texp
  type(tm_t), allocatable, dimension(:) :: tms
  character(1024) :: istr
  real(real64) :: ttot, tchunk

  ! List of chunks to process (TM index, GTI index 0, GTI index 1)
  type chunk_t
     integer :: ilist, gistart, giend
  end type chunk_t
  type(chunk_t), allocatable, dimension(:) :: chunks
  integer :: maxchunks, ichunk, nchunk, gi, gistart

  ! get command line parameters =======================================
  call opast(ierr)

  call cpu_time(expmap_start)

  call wrchatty ("main", "Calling get_param...")
  call task_info%obtain(ierr)
!*** test ierr
  call get_imgpar(task_info%params%templateimage, imgpar)
!*** eventually move imgpar to be an attribute of task_info.

  write(istr, *) task_info%n_TMs
  call wrlog("main", "Number of TMs: "//adjustl(trim(istr)))

#ifdef _OPENMP
  ! following srctool, we default to a single thread if OMP_NUM_THREADS
  ! is not set

  call get_environment_variable(name="OMP_NUM_THREADS", status=status)
  if ( (status==1) .or. (status==2) ) then
    call wrchatty("main", "OMP_NUM_THREADS undefined, setting to 1")
    call OMP_set_num_threads(1)
  end if
#endif

  ! Setup and load calibration data for each TM
  ! and add up total time in GTIs
  allocate (tms(task_info%n_TMs))
  ttot = 0
  ilist = 1
  do ci = 1, N_TELESCOPES
    if(.not.task_info%tm_info(ci)%is_present)&
  cycle

    call init_tm(tms(ilist), ilist, ci, task_info, imgpar)
    ngti = tms(ilist)%gtis%get_num()
    do gi=1, ngti
       ttot = ttot + (tms(ilist)%gtis%stops(gi)-tms(ilist)%gtis%starts(gi))
    end do

    ilist = ilist + 1
  end do

  ! Split total time in all TMs into roughly maxchunks chunks
  maxchunks = 128
  ichunk = 1
  tchunk = 0
  allocate(chunks(maxchunks+task_info%n_TMs+1))
  do ilist = 1, task_info%n_TMs
    gistart = 1
    ngti = tms(ilist)%gtis%get_num()
    do gi = 1, ngti
      tchunk = tchunk + (tms(ilist)%gtis%stops(gi)-tms(ilist)%gtis%starts(gi))
      if(tchunk>=(ttot/maxchunks) .or. gi==ngti) then
        chunks(ichunk)%ilist = ilist
        chunks(ichunk)%gistart = gistart
        chunks(ichunk)%giend = gi
        tchunk = 0
        gistart = gi + 1
        ichunk = ichunk + 1
      end if
    end do
  end do
  nchunk = ichunk - 1

  !$OMP PARALLEL DO
  do ichunk = 1, nchunk
     ilist = chunks(ichunk)%ilist
     call expmap_tm(tms(ilist), chunks(ichunk)%gistart, chunks(ichunk)%giend &
          , task_info, imgpar, tms(ilist)%camexp)
  end do
  !$OMP END PARALLEL DO

  ! Write out single maps
  if (task_info%params%withsingle) then
    do ilist = 1, task_info%n_TMs !**** size(tms) is more robust - also for the other similar loops.
      call wrschatty ("main", trim(tms(ilist)%inst)//": Calling write_cam_expmaps...")
      call write_cam_expmaps(tms(ilist)%tmi, ilist &
           , tms(ilist)%inst, task_info%n_bands&
           , task_info%params%e_min, task_info%params%e_max &
           , imgpar%stdhdr, imgpar%refhdr, imgpar%wcshdr &
           , imgpar%size_xi, imgpar%size_yi, tms(ilist)%ogtis &
           , task_info%params%exp_out_files, tms(ilist)%camexp)
    end do
  end if

  ! Compute and write total exposure maps
  if (task_info%params%withmerged) then
    allocate (texp(imgpar%size_xi,task_info%n_bands,imgpar%size_yi))
    texp = 0.

    ! add up TM exposure maps with weights, and collect output GTIs
    do ilist = 1, task_info%n_TMs
      do ib=1,task_info%n_bands
        texp(:,ib,:) = texp(:,ib,:) + tms(ilist)%weights(ib) * tms(ilist)%camexp(:,ib,:)
      end do
      ogtis(ilist) = tms(ilist)%ogtis
    end do

    ! write out total exposure maps
    call wrchatty ("main",  "Calling write_merged_expmaps...")
    call write_merged_expmaps(tms%inst, tms%tmi, task_info%n_bands&
        , task_info%params%e_min, task_info%params%e_max &
        , imgpar%stdhdr, imgpar%refhdr, imgpar%wcshdr &
        , imgpar%size_xi, imgpar%size_yi, ogtis&
        , task_info%params%merged_out_files, texp)

    do ilist = 1, task_info%n_TMs
      call ogtis(ilist)%dealloc()
    end do

    deallocate(texp)
  end if

  ! clean up TMs before exiting
  do ilist = 1, task_info%n_TMs
    call dealloc_tm(tms(ilist))
  end do
  deallocate(chunks)

  call cpu_time(expmap_stop)
  write (line, '(f10.3)') expmap_stop-expmap_start
  line = "Processing duration = "//trim(adjustl(line))//" seconds."
  call wrlog("main", line)

  call cpast(0, ierr)

end program expmap_main
